/* Generated by Cython 3.2.4 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "breezy/python-compat.h"
        ],
        "include_dirs": [
            "breezy"
        ],
        "name": "breezy._known_graph_pyx",
        "sources": [
            "breezy/_known_graph_pyx.pyx"
        ]
    },
    "module_name": "breezy._known_graph_pyx"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API)
  #if !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef _MSC_VER
  #pragma message ("Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.")
  #else
  #warning Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_2_4"
#define CYTHON_HEX_VERSION 0x030204F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #ifndef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 1
  #endif
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
  #if defined(CYTHON_IMMORTAL_CONSTANTS) && PY_VERSION_HEX < 0x030C0000
    #undef CYTHON_IMMORTAL_CONSTANTS
    #define CYTHON_IMMORTAL_CONSTANTS 0  // definitely won't work
  #elif !defined(CYTHON_IMMORTAL_CONSTANTS)
    #define CYTHON_IMMORTAL_CONSTANTS (PY_VERSION_HEX >= 0x030C0000 && !CYTHON_USE_MODULE_STATE && CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_COMPRESS_STRINGS
  #define CYTHON_COMPRESS_STRINGS 1
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL)
#endif
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#include <stdint.h>
typedef uintptr_t  __pyx_uintptr_t;
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CInitCode */
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef Py_TPFLAGS_IMMUTABLETYPE
  #define Py_TPFLAGS_IMMUTABLETYPE (1UL << 8)
#endif
#ifndef Py_TPFLAGS_DISALLOW_INSTANTIATION
  #define Py_TPFLAGS_DISALLOW_INSTANTIATION (1UL << 7)
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if PY_VERSION_HEX < 0x03090000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) GraalPyFrame_SetLineNumber((frame), (lineno))
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && PyUnstable_Object_IsUniquelyReferenced(obj)) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#elif CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
enum __Pyx_ReferenceSharing {
  __Pyx_ReferenceSharing_DefinitelyUnique, // We created it so we know it's unshared - no need to check
  __Pyx_ReferenceSharing_OwnStrongReference,
  __Pyx_ReferenceSharing_FunctionArgument,
  __Pyx_ReferenceSharing_SharedReference, // Never trust it to be unshared because it's a global or similar
};
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && PY_VERSION_HEX >= 0x030E0000
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing)\
    (sharing == __Pyx_ReferenceSharing_DefinitelyUnique ? 1 :\
      (sharing == __Pyx_ReferenceSharing_FunctionArgument ? PyUnstable_Object_IsUniqueReferencedTemporary(o) :\
      (sharing == __Pyx_ReferenceSharing_OwnStrongReference ? PyUnstable_Object_IsUniquelyReferenced(o) : 0)))
#elif (CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING) || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)sharing), Py_REFCNT(o) == 1)
#else
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)o), ((void)sharing), 0)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !CYTHON_COMPILING_IN_LIMITED_API && CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) (__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared) ?\
    __Pyx_NewRef(PyList_GET_ITEM(o, i)) : __Pyx_PyList_GetItemRef(o, i))
#else
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) __Pyx_PyList_GetItemRef(o, i)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__breezy___known_graph_pyx
#define __PYX_HAVE_API__breezy___known_graph_pyx
/* Early includes */
#include "python-compat.h"
#include <string.h>
#include <stdio.h>

    #if PY_MAJOR_VERSION <= 2
    #define PyDict_GetItemWithError _PyDict_GetItemWithError
    #endif

    #if __PYX_LIMITED_VERSION_HEX < 0x030d0000
    static CYTHON_INLINE int
    __Pyx_CAPI_PyDict_GetItemStringRef(PyObject *mp, const char *key, PyObject **result)
    {
        int res;
        PyObject *key_obj = PyUnicode_FromString(key);
        if (key_obj == NULL) {
            *result = NULL;
            return -1;
        }
        res = __Pyx_PyDict_GetItemRef(mp, key_obj, result);
        Py_DECREF(key_obj);
        return res;
    }
    #else
    #define __Pyx_CAPI_PyDict_GetItemStringRef PyDict_GetItemStringRef
    #endif
    #if PY_VERSION_HEX < 0x030d0000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030F0000)
    static CYTHON_INLINE int
    __Pyx_CAPI_PyDict_SetDefaultRef(PyObject *d, PyObject *key, PyObject *default_value,
                        PyObject **result)
    {
        PyObject *value;
        if (__Pyx_PyDict_GetItemRef(d, key, &value) < 0) {
            // get error
            if (result) {
                *result = NULL;
            }
            return -1;
        }
        if (value != NULL) {
            // present
            if (result) {
                *result = value;
            }
            else {
                Py_DECREF(value);
            }
            return 1;
        }

        // missing: set the item
        if (PyDict_SetItem(d, key, default_value) < 0) {
            // set error
            if (result) {
                *result = NULL;
            }
            return -1;
        }
        if (result) {
            Py_INCREF(default_value);
            *result = default_value;
        }
        return 0;
    }
    #else
    #define __Pyx_CAPI_PyDict_SetDefaultRef PyDict_SetDefaultRef
    #endif
    

    #if __PYX_LIMITED_VERSION_HEX < 0x030d0000
    static CYTHON_INLINE PyObject *
    __Pyx_CAPI_PyList_GetItemRef(PyObject *list, Py_ssize_t index)
    {
        PyObject *item = PyList_GetItem(list, index);
        Py_XINCREF(item);
        return item;
    }
    #else
    #define __Pyx_CAPI_PyList_GetItemRef PyList_GetItemRef
    #endif

    #if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030d0000
    static CYTHON_INLINE int
    __Pyx_CAPI_PyList_Extend(PyObject *list, PyObject *iterable)
    {
        return PyList_SetSlice(list, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, iterable);
    }

    static CYTHON_INLINE int
    __Pyx_CAPI_PyList_Clear(PyObject *list)
    {
        return PyList_SetSlice(list, 0, PY_SSIZE_T_MAX, NULL);
    }
    #else
    #define __Pyx_CAPI_PyList_Extend PyList_Extend
    #define __Pyx_CAPI_PyList_Clear PyList_Clear
    #endif
    
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifdef CYTHON_FREETHREADING_COMPATIBLE
#if CYTHON_FREETHREADING_COMPATIBLE
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "breezy/_known_graph_pyx.pyx",
  "(tree fragment)",
  "cpython/type.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* Atomics.proto (used by UnpackUnboundCMethod) */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_nonatomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_nonatomic_ptr_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #define __pyx_nonatomic_ptr_type void*
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_atomic_ptr_type old = _InterlockedCompareExchangePointer(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif

/* CriticalSectionsDefinition.proto (used by CriticalSections) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection void*
#define __Pyx_PyCriticalSection2 void*
#define __Pyx_PyCriticalSection_End(cs)
#define __Pyx_PyCriticalSection2_End(cs)
#else
#define __Pyx_PyCriticalSection PyCriticalSection
#define __Pyx_PyCriticalSection2 PyCriticalSection2
#define __Pyx_PyCriticalSection_End PyCriticalSection_End
#define __Pyx_PyCriticalSection2_End PyCriticalSection2_End
#endif

/* CriticalSections.proto (used by ParseKeywordsImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection_Begin(cs, arg) (void)(cs)
#define __Pyx_PyCriticalSection2_Begin(cs, arg1, arg2) (void)(cs)
#else
#define __Pyx_PyCriticalSection_Begin PyCriticalSection_Begin
#define __Pyx_PyCriticalSection2_Begin PyCriticalSection2_Begin
#endif
#if PY_VERSION_HEX < 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_BEGIN_CRITICAL_SECTION(o) {
#define __Pyx_END_CRITICAL_SECTION() }
#else
#define __Pyx_BEGIN_CRITICAL_SECTION Py_BEGIN_CRITICAL_SECTION
#define __Pyx_END_CRITICAL_SECTION Py_END_CRITICAL_SECTION
#endif

/* IncludeStructmemberH.proto (used by FixUpExtensionType) */
#include <structmember.h>

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode;
struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph;
struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode;
struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter;

/* "breezy/_known_graph_pyx.pyx":45
 * 
 * 
 * cdef class _KnownGraphNode:             # <<<<<<<<<<<<<<
 *     """Represents a single object in the known graph."""
 * 
*/
struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_vtab;
  PyObject *key;
  PyObject *parents;
  PyObject *children;
  long gdfo;
  int seen;
  PyObject *extra;
};


/* "breezy/_known_graph_pyx.pyx":174
 * cdef class _MergeSorter
 * 
 * cdef class KnownGraph:             # <<<<<<<<<<<<<<
 *     """This is a class which assumes we already know the full graph."""
 * 
*/
struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph {
  PyObject_HEAD
  struct __pyx_vtabstruct_6breezy_16_known_graph_pyx_KnownGraph *__pyx_vtab;
  PyObject *_nodes;
  PyObject *_known_heads;
  int do_cache;
};


/* "breezy/_known_graph_pyx.pyx":650
 * 
 * 
 * cdef class _MergeSortNode:             # <<<<<<<<<<<<<<
 *     """Tracks information about a node during the merge_sort operation."""
 * 
*/
struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_vtab;
  PyObject *key;
  long merge_depth;
  PyObject *end_of_merge;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *left_parent;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *left_pending_parent;
  PyObject *pending_parents;
  long _revno_first;
  long _revno_second;
  long _revno_last;
  int is_first_child;
  int seen_by_child;
  int completed;
};


/* "breezy/_known_graph_pyx.pyx":172
 * 
 * 
 * cdef class _MergeSorter             # <<<<<<<<<<<<<<
 * 
 * cdef class KnownGraph:
*/
struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter {
  PyObject_HEAD
  struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSorter *__pyx_vtab;
  struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *graph;
  PyObject *_depth_first_stack;
  Py_ssize_t _last_stack_item;
  PyObject *_revno_to_branch_count;
  PyObject *_scheduled_nodes;
};



/* "breezy/_known_graph_pyx.pyx":45
 * 
 * 
 * cdef class _KnownGraphNode:             # <<<<<<<<<<<<<<
 *     """Represents a single object in the known graph."""
 * 
*/

struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__KnownGraphNode {
  PyObject *(*clear_references)(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *);
};
static struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_vtabptr_6breezy_16_known_graph_pyx__KnownGraphNode;


/* "breezy/_known_graph_pyx.pyx":708
 * 
 * 
 * cdef class _MergeSorter:             # <<<<<<<<<<<<<<
 *     """This class does the work of computing the merge_sort ordering.
 * 
*/

struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSorter {
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *(*_get_ms_node)(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *, struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *);
  PyObject *(*_push_node)(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *, struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *, long);
  PyObject *(*_pop_node)(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *);
  PyObject *(*_schedule_stack)(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *);
  PyObject *(*topo_order)(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *);
};
static struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSorter *__pyx_vtabptr_6breezy_16_known_graph_pyx__MergeSorter;


/* "breezy/_known_graph_pyx.pyx":174
 * cdef class _MergeSorter
 * 
 * cdef class KnownGraph:             # <<<<<<<<<<<<<<
 *     """This is a class which assumes we already know the full graph."""
 * 
*/

struct __pyx_vtabstruct_6breezy_16_known_graph_pyx_KnownGraph {
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *(*_get_or_create_node)(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *, PyObject *);
  PyObject *(*_populate_parents)(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *, struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *, PyObject *);
};
static struct __pyx_vtabstruct_6breezy_16_known_graph_pyx_KnownGraph *__pyx_vtabptr_6breezy_16_known_graph_pyx_KnownGraph;


/* "breezy/_known_graph_pyx.pyx":650
 * 
 * 
 * cdef class _MergeSortNode:             # <<<<<<<<<<<<<<
 *     """Tracks information about a node during the merge_sort operation."""
 * 
*/

struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSortNode {
  int (*has_pending_parents)(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *);
  PyObject *(*_revno)(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *);
};
static struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_vtabptr_6breezy_16_known_graph_pyx__MergeSortNode;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* TupleAndListFromArray.proto (used by fastcall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto (used by BytesEquals) */
#include <string.h>

/* BytesEquals.proto (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_PySequence_ITEM(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_NewRef(__Pyx_PyTuple_GET_ITEM(args, i))
#else
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_XNewRef(PyTuple_GetItem(args, i))
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_ArgRef_FASTCALL(args, i) __Pyx_NewRef(args[i])
    #define __Pyx_NumKwargs_FASTCALL(kwds) __Pyx_PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
#else
    #define __Pyx_ArgRef_FASTCALL __Pyx_ArgRef_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#if CYTHON_METH_FASTCALL || (CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(args + start, stop - start)
#else
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* py_dict_items.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* CallCFunction.proto (used by CallUnboundCMethod0) */
#define __Pyx_CallCFunction(cfunc, self, args)\
    ((PyCFunction)(void(*)(void))(cfunc)->func)(self, args)
#define __Pyx_CallCFunctionWithKeywords(cfunc, self, args, kwargs)\
    ((PyCFunctionWithKeywords)(void(*)(void))(cfunc)->func)(self, args, kwargs)
#define __Pyx_CallCFunctionFast(cfunc, self, args, nargs)\
    ((__Pyx_PyCFunctionFast)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs)
#define __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, nargs, kwnames)\
    ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs, kwnames)

/* PyObjectCall.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto (used by PyObjectCallOneArg) */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* PyObjectCallOneArg.proto (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectGetAttrStr.proto (used by UnpackUnboundCMethod) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* UnpackUnboundCMethod.proto (used by CallUnboundCMethod0) */
typedef struct {
    PyObject *type;
    PyObject **method_name;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && CYTHON_ATOMICS
    __pyx_atomic_int_type initialized;
#endif
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
static CYTHON_INLINE int __Pyx_CachedCFunction_GetAndSetInitializing(__Pyx_CachedCFunction *cfunc) {
#if !CYTHON_ATOMICS
    return 1;
#else
    __pyx_nonatomic_int_type expected = 0;
    if (__pyx_atomic_int_cmp_exchange(&cfunc->initialized, &expected, 1)) {
        return 0;
    }
    return expected;
#endif
}
static CYTHON_INLINE void __Pyx_CachedCFunction_SetFinishedInitializing(__Pyx_CachedCFunction *cfunc) {
#if CYTHON_ATOMICS
    __pyx_atomic_store(&cfunc->initialized, 2);
#endif
}
#else
#define __Pyx_CachedCFunction_GetAndSetInitializing(cfunc) 2
#define __Pyx_CachedCFunction_SetFinishedInitializing(cfunc)
#endif

/* CallUnboundCMethod0.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* py_dict_values.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d);

/* OwnedDictNext.proto (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue);
#else
CYTHON_INLINE
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue);
#endif

/* RaiseDoubleKeywords.proto (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywordsImpl.export */
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name
);
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* CallUnboundCMethod2.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* ParseKeywords.proto */
static CYTHON_INLINE int __Pyx_ParseKeywords(
    PyObject *kwds, PyObject *const *kwvalues, PyObject ** const argnames[],
    PyObject *kwds2, PyObject *values[],
    Py_ssize_t num_pos_args, Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* BuildPyUnicode.proto (used by COrdinalToPyUnicode) */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, const char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* COrdinalToPyUnicode.proto (used by CIntToPyUnicode) */
static CYTHON_INLINE int __Pyx_CheckUnicodeValue(int value);
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromOrdinal_Padded(int value, Py_ssize_t width, char padding_char);

/* GCCDiagnostics.proto (used by CIntToPyUnicode) */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* IncludeStdlibH.proto (used by CIntToPyUnicode) */
#include <stdlib.h>

/* CIntToPyUnicode.proto */
#define __Pyx_PyUnicode_From_long(value, width, padding_char, format_char) (\
    ((format_char) == ('c')) ?\
        __Pyx_uchar___Pyx_PyUnicode_From_long(value, width, padding_char) :\
        __Pyx____Pyx_PyUnicode_From_long(value, width, padding_char, format_char)\
    )
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_long(long value, Py_ssize_t width, char padding_char);
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_long(long value, Py_ssize_t width, char padding_char, char format_char);

/* JoinPyUnicode.export */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* RejectKeywords.export */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds);

/* PyErrExceptionMatches.proto (used by GetAttr3) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto (used by PyErrFetchRestore) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto (used by GetAttr3) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyObjectGetAttrStrNoError.proto (used by GetBuiltinName) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto (used by GetModuleGlobalName) */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyDictVersioning.proto (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __Pyx_XNewRef(__pyx_dict_cached_value);\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* PyTypeError_Check.proto */
#define __Pyx_PyExc_TypeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_TypeError)

/* RaiseException.export */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyObjectVectorCallKwBuilder.proto (used by PyObjectVectorCallMethodKwBuilder) */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* PyObjectVectorCallMethodKwBuilder.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_VectorcallMethod_CallFromBuilder PyObject_VectorcallMethod
#else
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#endif

/* PyObjectFastCallMethod.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyObject_FastCallMethod(name, args, nargsf) PyObject_VectorcallMethod(name, args, nargsf, NULL)
#else
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf);
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck, unsafe_shared) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck, int unsafe_shared);

/* PyValueError_Check.proto */
#define __Pyx_PyExc_ValueError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_ValueError)

/* PyObjectCallNoArg.proto (used by pyfrozenset_new) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* PyKeyError_Check.proto */
#define __Pyx_PyExc_KeyError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_KeyError)

/* py_dict_pop.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value);

/* PyObjectCall2Args.proto (used by CallUnboundCMethod1) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* CallUnboundCMethod1.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00a2
    return PyList_Extend(L, v);
#elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* AssertionsEnabled.proto */
#if CYTHON_COMPILING_IN_LIMITED_API  ||  PY_VERSION_HEX >= 0x030C0000
  static int __pyx_assertions_enabled_flag;
  #define __pyx_assertions_enabled() (__pyx_assertions_enabled_flag)
  #if __clang__ || __GNUC__
  __attribute__((no_sanitize("thread")))
  #endif
  static int __Pyx_init_assertions_enabled(void) {
    PyObject *builtins, *debug, *debug_str;
    int flag;
    builtins = PyEval_GetBuiltins();
    if (!builtins) goto bad;
    debug_str = PyUnicode_FromStringAndSize("__debug__", 9);
    if (!debug_str) goto bad;
    debug = PyObject_GetItem(builtins, debug_str);
    Py_DECREF(debug_str);
    if (!debug) goto bad;
    flag = PyObject_IsTrue(debug);
    Py_DECREF(debug);
    if (flag == -1) goto bad;
    __pyx_assertions_enabled_flag = flag;
    return 0;
  bad:
    __pyx_assertions_enabled_flag = 1;
    return -1;
  }
#else
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (!Py_OptimizeFlag)
#endif

/* PyAssertionError_Check.proto */
#define __Pyx_PyExc_AssertionError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_AssertionError)

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* CIntToPyUnicode.proto */
#define __Pyx_PyUnicode_From_int(value, width, padding_char, format_char) (\
    ((format_char) == ('c')) ?\
        __Pyx_uchar___Pyx_PyUnicode_From_int(value, width, padding_char) :\
        __Pyx____Pyx_PyUnicode_From_int(value, width, padding_char, format_char)\
    )
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char);
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char);

/* PyRuntimeError_Check.proto */
#define __Pyx_PyExc_RuntimeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_RuntimeError)

/* PyLongBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static CYTHON_INLINE PyObject* __Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyLong_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyObjectGetMethod.proto (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);
#endif

/* PyObjectCallMethod0.proto (used by pop) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* pop.proto */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L);
#define __Pyx_PyObject_Pop(L) (likely(PyList_CheckExact(L)) ?\
    __Pyx_PyList_Pop(L) : __Pyx__PyObject_Pop(L))
#else
#define __Pyx_PyList_Pop(L)  __Pyx__PyObject_Pop(L)
#define __Pyx_PyObject_Pop(L)  __Pyx__PyObject_Pop(L)
#endif

/* ArgTypeTestFunc.export */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))

/* AllocateExtensionType.proto */
static PyObject *__Pyx_AllocateExtensionType(PyTypeObject *t, int is_final);

/* CallTypeTraverse.proto */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#define __Pyx_call_type_traverse(o, always_call, visit, arg) 0
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg);
#endif

/* LimitedApiGetTypeDict.proto (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp);
#endif

/* SetItemOnTypeDict.proto (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v);
#define __Pyx_SetItemOnTypeDict(tp, k, v) __Pyx__SetItemOnTypeDict((PyTypeObject*)tp, k, v)

/* FixUpExtensionType.proto */
static CYTHON_INLINE int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);

/* ValidateBasesTuple.proto (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto (used by MergeVTables) */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
static int __Pyx_MergeVtables(PyTypeObject *type);

/* DelItemOnTypeDict.proto (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k);
#define __Pyx_DelItemOnTypeDict(tp, k) __Pyx__DelItemOnTypeDict((PyTypeObject*)tp, k)

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_2_4
#define __PYX_HAVE_RT_ImportType_proto_3_2_4
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_2_4 {
   __Pyx_ImportType_CheckSize_Error_3_2_4 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_2_4 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_2_4 = 2
};
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size);
#endif

/* HasAttr.proto (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* ImportImpl.export */
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level);

/* Import.proto */
static CYTHON_INLINE PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* dict_setdefault.proto (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value);

/* AddModuleRef.proto (used by FetchSharedCythonModule) */
#if ((CYTHON_COMPILING_IN_CPYTHON_FREETHREADING ) ||\
     __PYX_LIMITED_VERSION_HEX < 0x030d0000)
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name);
#else
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#endif

/* FetchSharedCythonModule.proto (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto (used by CommonTypesMetaclass) */
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases);

/* CommonTypesMetaclass.proto (used by CythonFunctionShared) */
static int __pyx_CommonTypesMetaclass_init(PyObject *module);
#define __Pyx_CommonTypesMetaclass_USED

/* PyMethodNew.proto (used by CythonFunctionShared) */
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ);

/* PyVectorcallFastCallDict.proto (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto (used by CythonFunction) */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && CYTHON_METH_FASTCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_dict;
#endif
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    PyObject *defaults;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_mstate_global->__pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_mstate_global->__pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void));
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func,
                                                         PyTypeObject *defaults_type);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CLineInTraceback.proto (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto (used by AddTraceback) */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CheckUnpickleChecksum.proto */
static CYTHON_INLINE int __Pyx_CheckUnpickleChecksum(long checksum, long checksum1, long checksum2, long checksum3, const char *members);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* UpdateUnpickledDict.proto */
static int __Pyx_UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
static unsigned long __Pyx_cached_runtime_version = 0;
static void __Pyx_init_runtime_version(void);
#else
#define __Pyx_init_runtime_version()
#endif
static unsigned long __Pyx_get_runtime_version(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* DecompressString.proto */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."

static PyObject *__pyx_f_6breezy_16_known_graph_pyx_15_KnownGraphNode_clear_references(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_self); /* proto*/
static struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_f_6breezy_16_known_graph_pyx_10KnownGraph__get_or_create_node(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_key); /* proto*/
static PyObject *__pyx_f_6breezy_16_known_graph_pyx_10KnownGraph__populate_parents(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node, PyObject *__pyx_v_parent_keys); /* proto*/
static int __pyx_f_6breezy_16_known_graph_pyx_14_MergeSortNode_has_pending_parents(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6breezy_16_known_graph_pyx_14_MergeSortNode__revno(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self); /* proto*/
static struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_f_6breezy_16_known_graph_pyx_12_MergeSorter__get_ms_node(CYTHON_UNUSED struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *__pyx_v_self, struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node); /* proto*/
static PyObject *__pyx_f_6breezy_16_known_graph_pyx_12_MergeSorter__push_node(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *__pyx_v_self, struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node, long __pyx_v_merge_depth); /* proto*/
static PyObject *__pyx_f_6breezy_16_known_graph_pyx_12_MergeSorter__pop_node(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6breezy_16_known_graph_pyx_12_MergeSorter__schedule_stack(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6breezy_16_known_graph_pyx_12_MergeSorter_topo_order(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *__pyx_v_self); /* proto*/

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "cpython.pyport" */

/* Module declarations from "cpython.dict" */

/* Module declarations from "cpython.list" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "cpython.tuple" */

/* Module declarations from "breezy._known_graph_pyx" */
static PyObject *__pyx_v_6breezy_16_known_graph_pyx_NULL_REVISION = 0;
static struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_f_6breezy_16_known_graph_pyx__get_list_node(PyObject *, Py_ssize_t); /*proto*/
static struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_f_6breezy_16_known_graph_pyx__get_tuple_node(PyObject *, Py_ssize_t); /*proto*/
static PyObject *__pyx_f_6breezy_16_known_graph_pyx__sort_list_nodes(PyObject *, int); /*proto*/
static PyObject *__pyx_f_6breezy_16_known_graph_pyx___pyx_unpickle__KnownGraphNode__set_state(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6breezy_16_known_graph_pyx___pyx_unpickle_KnownGraph__set_state(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *, PyObject *); /*proto*/
static PyObject *__pyx_f_6breezy_16_known_graph_pyx___pyx_unpickle__MergeSortNode__set_state(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6breezy_16_known_graph_pyx___pyx_unpickle__MergeSorter__set_state(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *, PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "breezy._known_graph_pyx"
extern int __pyx_module_is_main_breezy___known_graph_pyx;
int __pyx_module_is_main_breezy___known_graph_pyx = 0;

/* Implementation of "breezy._known_graph_pyx" */
/* #### Code section: global_var ### */
/* #### Code section: string_decls ### */
static const char __pyx_k_known_heads__nodes_do_cache[] = "_known_heads, _nodes, do_cache";
static const char __pyx_k_revno_first__revno_last__revno[] = "_revno_first, _revno_last, _revno_second, completed, end_of_merge, is_first_child, key, left_parent, left_pending_parent, merge_depth, pending_parents, seen_by_child";
static const char __pyx_k_children_extra_gdfo_key_parents[] = "children, extra, gdfo, key, parents, seen";
static const char __pyx_k_depth_first_stack__last_stack_i[] = "_depth_first_stack, _last_stack_item, _revno_to_branch_count, _scheduled_nodes, graph";
static const char __pyx_k_Implementation_of_Graph_algorith[] = "Implementation of Graph algorithms when we have already loaded everything.\n";
/* #### Code section: decls ### */
static int __pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode___init__(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_10child_keys___get__(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_11parent_keys___get__(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_2__repr__(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_4gdfo___get__(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_self); /* proto */
static int __pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_4gdfo_2__set__(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_4__reduce_cython__(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_6__setstate_cython__(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_get_key(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_node); /* proto */
static int __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph___init__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_parent_map, PyObject *__pyx_v_do_cache); /* proto */
static void __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_2__dealloc__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_4_initialize_nodes(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_parent_map); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_6_find_tails(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_8_find_tips(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_10_find_gdfo(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_12add_node(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_parent_keys); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_14heads(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_keys); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_16topo_sort(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_18gc_sort(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_20merge_sort(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_tip_key); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_22get_parent_keys(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_24get_child_keys(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_6_nodes___get__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self); /* proto */
static int __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_6_nodes_2__set__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_6_nodes_4__del__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_12_known_heads___get__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self); /* proto */
static int __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_12_known_heads_2__set__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_12_known_heads_4__del__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_8do_cache___get__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self); /* proto */
static int __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_8do_cache_2__set__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_26__reduce_cython__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_28__setstate_cython__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode___init__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_2__repr__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_5revno___get__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_3key___get__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self); /* proto */
static int __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_3key_2__set__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_3key_4__del__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_11merge_depth___get__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self); /* proto */
static int __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_11merge_depth_2__set__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_12end_of_merge___get__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self); /* proto */
static int __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_12end_of_merge_2__set__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_12end_of_merge_4__del__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_4__reduce_cython__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_6__setstate_cython__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6breezy_16_known_graph_pyx_12_MergeSorter___init__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *__pyx_v_self, PyObject *__pyx_v_known_graph, PyObject *__pyx_v_tip_key); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_12_MergeSorter_2__reduce_cython__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_12_MergeSorter_4__setstate_cython__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_2__pyx_unpickle__KnownGraphNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_4__pyx_unpickle_KnownGraph(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_6__pyx_unpickle__MergeSortNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_8__pyx_unpickle__MergeSorter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_6breezy_16_known_graph_pyx__KnownGraphNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6breezy_16_known_graph_pyx__MergeSorter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6breezy_16_known_graph_pyx_KnownGraph(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6breezy_16_known_graph_pyx__MergeSortNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  PyObject *__pyx_type_6breezy_16_known_graph_pyx__KnownGraphNode;
  PyObject *__pyx_type_6breezy_16_known_graph_pyx__MergeSorter;
  PyObject *__pyx_type_6breezy_16_known_graph_pyx_KnownGraph;
  PyObject *__pyx_type_6breezy_16_known_graph_pyx__MergeSortNode;
  PyTypeObject *__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode;
  PyTypeObject *__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter;
  PyTypeObject *__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph;
  PyTypeObject *__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_values;
  __Pyx_CachedCFunction __pyx_umethod_PyList_Type_pop;
  PyObject *__pyx_codeobj_tab[24];
  PyObject *__pyx_string_tab[185];
  PyObject *__pyx_number_tab[6];
/* #### Code section: module_state_contents ### */
/* CommonTypesMetaclass.module_state_decls */
PyTypeObject *__pyx_CommonTypesMetaclassType;

/* CachedMethodType.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API
PyObject *__Pyx_CachedMethodType;
#endif

/* CythonFunctionShared.module_state_decls */
PyTypeObject *__pyx_CyFunctionType;

/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_ __pyx_string_tab[0]
#define __pyx_kp_u_Note_that_Cython_is_deliberately __pyx_string_tab[1]
#define __pyx_kp_u_Parent_key_mismatch_existing_nod __pyx_string_tab[2]
#define __pyx_kp_u_Something_wrong_with __pyx_string_tab[3]
#define __pyx_kp_u__2 __pyx_string_tab[4]
#define __pyx_kp_u__3 __pyx_string_tab[5]
#define __pyx_kp_u__4 __pyx_string_tab[6]
#define __pyx_kp_u__5 __pyx_string_tab[7]
#define __pyx_kp_u__6 __pyx_string_tab[8]
#define __pyx_kp_u_add_note __pyx_string_tab[9]
#define __pyx_kp_u_breezy __pyx_string_tab[10]
#define __pyx_kp_u_breezy__known_graph_pyx_pyx __pyx_string_tab[11]
#define __pyx_kp_u_child __pyx_string_tab[12]
#define __pyx_kp_u_depth __pyx_string_tab[13]
#define __pyx_kp_u_disable __pyx_string_tab[14]
#define __pyx_kp_u_enable __pyx_string_tab[15]
#define __pyx_kp_u_first __pyx_string_tab[16]
#define __pyx_kp_u_gc __pyx_string_tab[17]
#define __pyx_kp_u_gdfo __pyx_string_tab[18]
#define __pyx_kp_u_ghost_nodes_should_not_be_pushed __pyx_string_tab[19]
#define __pyx_kp_u_has_parents_of __pyx_string_tab[20]
#define __pyx_kp_u_isenabled __pyx_string_tab[21]
#define __pyx_kp_u_key_2 __pyx_string_tab[22]
#define __pyx_kp_u_lst_or_tpl_must_be_a_list_or_tup __pyx_string_tab[23]
#define __pyx_kp_u_not __pyx_string_tab[24]
#define __pyx_kp_u_not_in_nodes __pyx_string_tab[25]
#define __pyx_kp_u_par __pyx_string_tab[26]
#define __pyx_kp_u_parent_map_should_be_a_dict_of_k __pyx_string_tab[27]
#define __pyx_kp_u_rev __pyx_string_tab[28]
#define __pyx_kp_u_seen __pyx_string_tab[29]
#define __pyx_kp_u_tree_fragment __pyx_string_tab[30]
#define __pyx_n_u_GraphCycleError __pyx_string_tab[31]
#define __pyx_n_u_KnownGraph __pyx_string_tab[32]
#define __pyx_n_u_KnownGraphNode __pyx_string_tab[33]
#define __pyx_n_u_KnownGraphNode___reduce_cython __pyx_string_tab[34]
#define __pyx_n_u_KnownGraphNode___setstate_cytho __pyx_string_tab[35]
#define __pyx_n_u_KnownGraph___reduce_cython __pyx_string_tab[36]
#define __pyx_n_u_KnownGraph___setstate_cython __pyx_string_tab[37]
#define __pyx_n_u_KnownGraph__find_gdfo __pyx_string_tab[38]
#define __pyx_n_u_KnownGraph__find_tails __pyx_string_tab[39]
#define __pyx_n_u_KnownGraph__find_tips __pyx_string_tab[40]
#define __pyx_n_u_KnownGraph__initialize_nodes __pyx_string_tab[41]
#define __pyx_n_u_KnownGraph_add_node __pyx_string_tab[42]
#define __pyx_n_u_KnownGraph_gc_sort __pyx_string_tab[43]
#define __pyx_n_u_KnownGraph_get_child_keys __pyx_string_tab[44]
#define __pyx_n_u_KnownGraph_get_parent_keys __pyx_string_tab[45]
#define __pyx_n_u_KnownGraph_heads __pyx_string_tab[46]
#define __pyx_n_u_KnownGraph_merge_sort __pyx_string_tab[47]
#define __pyx_n_u_KnownGraph_topo_sort __pyx_string_tab[48]
#define __pyx_n_u_MergeSortNode __pyx_string_tab[49]
#define __pyx_n_u_MergeSortNode___reduce_cython __pyx_string_tab[50]
#define __pyx_n_u_MergeSortNode___setstate_cython __pyx_string_tab[51]
#define __pyx_n_u_MergeSorter __pyx_string_tab[52]
#define __pyx_n_u_MergeSorter___reduce_cython __pyx_string_tab[53]
#define __pyx_n_u_MergeSorter___setstate_cython __pyx_string_tab[54]
#define __pyx_n_u_NULL_REVISION __pyx_string_tab[55]
#define __pyx_n_u_Pyx_PyDict_NextRef __pyx_string_tab[56]
#define __pyx_n_u__3 __pyx_string_tab[57]
#define __pyx_n_u_add_node __pyx_string_tab[58]
#define __pyx_n_u_append __pyx_string_tab[59]
#define __pyx_n_u_asyncio_coroutines __pyx_string_tab[60]
#define __pyx_n_u_breezy__known_graph_pyx __pyx_string_tab[61]
#define __pyx_n_u_candidate_nodes __pyx_string_tab[62]
#define __pyx_n_u_child_2 __pyx_string_tab[63]
#define __pyx_n_u_child_keys __pyx_string_tab[64]
#define __pyx_n_u_child_node __pyx_string_tab[65]
#define __pyx_n_u_class __pyx_string_tab[66]
#define __pyx_n_u_class_getitem __pyx_string_tab[67]
#define __pyx_n_u_cleanup __pyx_string_tab[68]
#define __pyx_n_u_clear __pyx_string_tab[69]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[70]
#define __pyx_n_u_collections __pyx_string_tab[71]
#define __pyx_n_u_deque __pyx_string_tab[72]
#define __pyx_n_u_dict __pyx_string_tab[73]
#define __pyx_n_u_dict_2 __pyx_string_tab[74]
#define __pyx_n_u_do_cache __pyx_string_tab[75]
#define __pyx_n_u_errors __pyx_string_tab[76]
#define __pyx_n_u_existing_parent_keys __pyx_string_tab[77]
#define __pyx_n_u_find_gdfo __pyx_string_tab[78]
#define __pyx_n_u_find_tails __pyx_string_tab[79]
#define __pyx_n_u_find_tips __pyx_string_tab[80]
#define __pyx_n_u_func __pyx_string_tab[81]
#define __pyx_n_u_gc __pyx_string_tab[82]
#define __pyx_n_u_gc_sort __pyx_string_tab[83]
#define __pyx_n_u_get_child_keys __pyx_string_tab[84]
#define __pyx_n_u_get_key __pyx_string_tab[85]
#define __pyx_n_u_get_parent_keys __pyx_string_tab[86]
#define __pyx_n_u_getstate __pyx_string_tab[87]
#define __pyx_n_u_heads __pyx_string_tab[88]
#define __pyx_n_u_heads_key __pyx_string_tab[89]
#define __pyx_n_u_initialize_nodes __pyx_string_tab[90]
#define __pyx_n_u_is_coroutine __pyx_string_tab[91]
#define __pyx_n_u_items __pyx_string_tab[92]
#define __pyx_n_u_key __pyx_string_tab[93]
#define __pyx_n_u_keys __pyx_string_tab[94]
#define __pyx_n_u_known_graph __pyx_string_tab[95]
#define __pyx_n_u_last_item __pyx_string_tab[96]
#define __pyx_n_u_main __pyx_string_tab[97]
#define __pyx_n_u_maybe_heads __pyx_string_tab[98]
#define __pyx_n_u_maybe_node __pyx_string_tab[99]
#define __pyx_n_u_merge_sort __pyx_string_tab[100]
#define __pyx_n_u_min_gdfo __pyx_string_tab[101]
#define __pyx_n_u_module __pyx_string_tab[102]
#define __pyx_n_u_name __pyx_string_tab[103]
#define __pyx_n_u_new __pyx_string_tab[104]
#define __pyx_n_u_next_gdfo __pyx_string_tab[105]
#define __pyx_n_u_node __pyx_string_tab[106]
#define __pyx_n_u_node2 __pyx_string_tab[107]
#define __pyx_n_u_parent_gdfo __pyx_string_tab[108]
#define __pyx_n_u_parent_keys __pyx_string_tab[109]
#define __pyx_n_u_parent_map __pyx_string_tab[110]
#define __pyx_n_u_parent_node __pyx_string_tab[111]
#define __pyx_n_u_parents __pyx_string_tab[112]
#define __pyx_n_u_pending __pyx_string_tab[113]
#define __pyx_n_u_pending_append __pyx_string_tab[114]
#define __pyx_n_u_pending_popleft __pyx_string_tab[115]
#define __pyx_n_u_pop __pyx_string_tab[116]
#define __pyx_n_u_popleft __pyx_string_tab[117]
#define __pyx_n_u_pos __pyx_string_tab[118]
#define __pyx_n_u_prefix __pyx_string_tab[119]
#define __pyx_n_u_prefix_tips __pyx_string_tab[120]
#define __pyx_n_u_pyx_checksum __pyx_string_tab[121]
#define __pyx_n_u_pyx_result __pyx_string_tab[122]
#define __pyx_n_u_pyx_state __pyx_string_tab[123]
#define __pyx_n_u_pyx_type __pyx_string_tab[124]
#define __pyx_n_u_pyx_unpickle_KnownGraph __pyx_string_tab[125]
#define __pyx_n_u_pyx_unpickle__KnownGraphNode __pyx_string_tab[126]
#define __pyx_n_u_pyx_unpickle__MergeSortNode __pyx_string_tab[127]
#define __pyx_n_u_pyx_unpickle__MergeSorter __pyx_string_tab[128]
#define __pyx_n_u_pyx_vtable __pyx_string_tab[129]
#define __pyx_n_u_qualname __pyx_string_tab[130]
#define __pyx_n_u_real_node __pyx_string_tab[131]
#define __pyx_n_u_reduce __pyx_string_tab[132]
#define __pyx_n_u_reduce_cython __pyx_string_tab[133]
#define __pyx_n_u_reduce_ex __pyx_string_tab[134]
#define __pyx_n_u_replace __pyx_string_tab[135]
#define __pyx_n_u_result __pyx_string_tab[136]
#define __pyx_n_u_reverse __pyx_string_tab[137]
#define __pyx_n_u_revision __pyx_string_tab[138]
#define __pyx_n_u_self __pyx_string_tab[139]
#define __pyx_n_u_set_name __pyx_string_tab[140]
#define __pyx_n_u_setdefault __pyx_string_tab[141]
#define __pyx_n_u_setstate __pyx_string_tab[142]
#define __pyx_n_u_setstate_cython __pyx_string_tab[143]
#define __pyx_n_u_sort __pyx_string_tab[144]
#define __pyx_n_u_sorter __pyx_string_tab[145]
#define __pyx_n_u_state __pyx_string_tab[146]
#define __pyx_n_u_tails __pyx_string_tab[147]
#define __pyx_n_u_temp __pyx_string_tab[148]
#define __pyx_n_u_temp_key __pyx_string_tab[149]
#define __pyx_n_u_temp_node __pyx_string_tab[150]
#define __pyx_n_u_temp_parent_keys __pyx_string_tab[151]
#define __pyx_n_u_test __pyx_string_tab[152]
#define __pyx_n_u_tip_key __pyx_string_tab[153]
#define __pyx_n_u_tip_nodes __pyx_string_tab[154]
#define __pyx_n_u_tips __pyx_string_tab[155]
#define __pyx_n_u_topo_order __pyx_string_tab[156]
#define __pyx_n_u_topo_sort __pyx_string_tab[157]
#define __pyx_n_u_update __pyx_string_tab[158]
#define __pyx_n_u_use_setstate __pyx_string_tab[159]
#define __pyx_n_u_values __pyx_string_tab[160]
#define __pyx_kp_b_iso88591_6 __pyx_string_tab[161]
#define __pyx_kp_b_iso88591_9A __pyx_string_tab[162]
#define __pyx_kp_b_iso88591_AV1 __pyx_string_tab[163]
#define __pyx_kp_b_iso88591_A_1D_c_A_t9Cq_avQ_M_q_q_O4q_q_1 __pyx_string_tab[164]
#define __pyx_kp_b_iso88591_A_4_1A_a_k_auAZq_4_1A_6 __pyx_string_tab[165]
#define __pyx_kp_b_iso88591_A_IQa_nAT_s_81_G1_q_IQ_Q_ha_8_E __pyx_string_tab[166]
#define __pyx_kp_b_iso88591_A_QfA_v __pyx_string_tab[167]
#define __pyx_kp_b_iso88591_A_a_k_iq_V1A_A_t9CuC_7q_Jc_HA_Qg __pyx_string_tab[168]
#define __pyx_kp_b_iso88591_A_l_9Cr_S_Yb_Q_Q_O1IRq_j_1_1_A_t __pyx_string_tab[169]
#define __pyx_kp_b_iso88591_A_l_O1IRq_j_1_1_A_F_A_O1D_at_a_R __pyx_string_tab[170]
#define __pyx_kp_b_iso88591_A_q_a_k_iq_V1A_A_at_c_QfA_q __pyx_string_tab[171]
#define __pyx_kp_b_iso88591_A_t7_4q __pyx_string_tab[172]
#define __pyx_kp_b_iso88591_A_t7_4q_2 __pyx_string_tab[173]
#define __pyx_kp_b_iso88591_A_t_a_a_I_q_fCs_4vS_T_q_uCq_1Jaq __pyx_string_tab[174]
#define __pyx_kp_b_iso88591_QfA __pyx_string_tab[175]
#define __pyx_kp_b_iso88591_T_D_G4vT_4q_G1F_a_vWE_Q_q_t_WE __pyx_string_tab[176]
#define __pyx_kp_b_iso88591_T_IT_G1F_a_vWE_Q_q_t_S_HG1_q_4q __pyx_string_tab[177]
#define __pyx_kp_b_iso88591_T_N_6Fd_VZZiimm_C_C_I_I_M_M __pyx_string_tab[178]
#define __pyx_kp_b_iso88591_T_d_TAZZ_qquuv_G1F_a_vWE_Q_q_t __pyx_string_tab[179]
#define __pyx_kp_b_iso88591_q __pyx_string_tab[180]
#define __pyx_kp_b_iso88591_q_0_kQR_1_7_1_2DNRS_1 __pyx_string_tab[181]
#define __pyx_kp_b_iso88591_q_0_kQR_7_0_1B_PQ_1 __pyx_string_tab[182]
#define __pyx_kp_b_iso88591_q_0_kQR_XQa_7_A_1 __pyx_string_tab[183]
#define __pyx_kp_b_iso88591_q_0_kQR_xq_7_a_nA_1 __pyx_string_tab[184]
#define __pyx_int_0 __pyx_number_tab[0]
#define __pyx_int_1 __pyx_number_tab[1]
#define __pyx_int_25525219 __pyx_number_tab[2]
#define __pyx_int_50889907 __pyx_number_tab[3]
#define __pyx_int_238330808 __pyx_number_tab[4]
#define __pyx_int_249680612 __pyx_number_tab[5]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode);
  Py_CLEAR(clear_module_state->__pyx_type_6breezy_16_known_graph_pyx__KnownGraphNode);
  Py_CLEAR(clear_module_state->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter);
  Py_CLEAR(clear_module_state->__pyx_type_6breezy_16_known_graph_pyx__MergeSorter);
  Py_CLEAR(clear_module_state->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph);
  Py_CLEAR(clear_module_state->__pyx_type_6breezy_16_known_graph_pyx_KnownGraph);
  Py_CLEAR(clear_module_state->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode);
  Py_CLEAR(clear_module_state->__pyx_type_6breezy_16_known_graph_pyx__MergeSortNode);
  for (int i=0; i<24; ++i) { Py_CLEAR(clear_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<185; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<6; ++i) { Py_CLEAR(clear_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_clear_contents ### */
/* CommonTypesMetaclass.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_clear_end ### */
return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode);
  Py_VISIT(traverse_module_state->__pyx_type_6breezy_16_known_graph_pyx__KnownGraphNode);
  Py_VISIT(traverse_module_state->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter);
  Py_VISIT(traverse_module_state->__pyx_type_6breezy_16_known_graph_pyx__MergeSorter);
  Py_VISIT(traverse_module_state->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph);
  Py_VISIT(traverse_module_state->__pyx_type_6breezy_16_known_graph_pyx_KnownGraph);
  Py_VISIT(traverse_module_state->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode);
  Py_VISIT(traverse_module_state->__pyx_type_6breezy_16_known_graph_pyx__MergeSortNode);
  for (int i=0; i<24; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<185; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<6; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_traverse_contents ### */
/* CommonTypesMetaclass.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_traverse_end ### */
return 0;
}
#endif
/* #### Code section: module_code ### */

/* "breezy/_known_graph_pyx.pyx":55
 *     cdef object extra
 * 
 *     def __init__(self, key):             # <<<<<<<<<<<<<<
 *         self.key = key
 *         self.parents = None
*/

/* Python wrapper */
static int __pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_key = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 55, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 55, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 55, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, i); __PYX_ERR(0, 55, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 55, __pyx_L3_error)
    }
    __pyx_v_key = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 55, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx._KnownGraphNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode___init__(((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_v_self), __pyx_v_key);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode___init__(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "breezy/_known_graph_pyx.pyx":56
 * 
 *     def __init__(self, key):
 *         self.key = key             # <<<<<<<<<<<<<<
 *         self.parents = None
 * 
*/
  __Pyx_INCREF(__pyx_v_key);
  __Pyx_GIVEREF(__pyx_v_key);
  __Pyx_GOTREF(__pyx_v_self->key);
  __Pyx_DECREF(__pyx_v_self->key);
  __pyx_v_self->key = __pyx_v_key;

  /* "breezy/_known_graph_pyx.pyx":57
 *     def __init__(self, key):
 *         self.key = key
 *         self.parents = None             # <<<<<<<<<<<<<<
 * 
 *         self.children = []
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->parents);
  __Pyx_DECREF(__pyx_v_self->parents);
  __pyx_v_self->parents = Py_None;

  /* "breezy/_known_graph_pyx.pyx":59
 *         self.parents = None
 * 
 *         self.children = []             # <<<<<<<<<<<<<<
 *         # Greatest distance from origin
 *         self.gdfo = -1
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->children);
  __Pyx_DECREF(__pyx_v_self->children);
  __pyx_v_self->children = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":61
 *         self.children = []
 *         # Greatest distance from origin
 *         self.gdfo = -1             # <<<<<<<<<<<<<<
 *         self.seen = 0
 *         self.extra = None
*/
  __pyx_v_self->gdfo = -1L;

  /* "breezy/_known_graph_pyx.pyx":62
 *         # Greatest distance from origin
 *         self.gdfo = -1
 *         self.seen = 0             # <<<<<<<<<<<<<<
 *         self.extra = None
 * 
*/
  __pyx_v_self->seen = 0;

  /* "breezy/_known_graph_pyx.pyx":63
 *         self.gdfo = -1
 *         self.seen = 0
 *         self.extra = None             # <<<<<<<<<<<<<<
 * 
 *     property child_keys:
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->extra);
  __Pyx_DECREF(__pyx_v_self->extra);
  __pyx_v_self->extra = Py_None;

  /* "breezy/_known_graph_pyx.pyx":55
 *     cdef object extra
 * 
 *     def __init__(self, key):             # <<<<<<<<<<<<<<
 *         self.key = key
 *         self.parents = None
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy._known_graph_pyx._KnownGraphNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":66
 * 
 *     property child_keys:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef _KnownGraphNode child
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_10child_keys_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_10child_keys_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_10child_keys___get__(((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_10child_keys___get__(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_self) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_child = 0;
  PyObject *__pyx_v_keys = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "breezy/_known_graph_pyx.pyx":69
 *             cdef _KnownGraphNode child
 * 
 *             keys = []             # <<<<<<<<<<<<<<
 *             for child in self.children:
 *                 PyList_Append(keys, child.key)
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_keys = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":70
 * 
 *             keys = []
 *             for child in self.children:             # <<<<<<<<<<<<<<
 *                 PyList_Append(keys, child.key)
 *             return keys
*/
  if (likely(PyList_CheckExact(__pyx_v_self->children)) || PyTuple_CheckExact(__pyx_v_self->children)) {
    __pyx_t_1 = __pyx_v_self->children; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_self->children); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 70, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 70, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        __pyx_t_4 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_2, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_2;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 70, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2));
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2);
        #endif
        ++__pyx_t_2;
      }
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 70, __pyx_L1_error)
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 70, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_4);
    if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode))))) __PYX_ERR(0, 70, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_child, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "breezy/_known_graph_pyx.pyx":71
 *             keys = []
 *             for child in self.children:
 *                 PyList_Append(keys, child.key)             # <<<<<<<<<<<<<<
 *             return keys
 * 
*/
    __pyx_t_4 = __pyx_v_child->key;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_5 = PyList_Append(__pyx_v_keys, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 71, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "breezy/_known_graph_pyx.pyx":70
 * 
 *             keys = []
 *             for child in self.children:             # <<<<<<<<<<<<<<
 *                 PyList_Append(keys, child.key)
 *             return keys
*/
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":72
 *             for child in self.children:
 *                 PyList_Append(keys, child.key)
 *             return keys             # <<<<<<<<<<<<<<
 * 
 *     property parent_keys:
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_keys);
  __pyx_r = __pyx_v_keys;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":66
 * 
 *     property child_keys:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef _KnownGraphNode child
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("breezy._known_graph_pyx._KnownGraphNode.child_keys.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_keys);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":75
 * 
 *     property parent_keys:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.parents is None:
 *                 return None
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_11parent_keys_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_11parent_keys_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_11parent_keys___get__(((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_11parent_keys___get__(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_self) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_parent = 0;
  PyObject *__pyx_v_keys = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "breezy/_known_graph_pyx.pyx":76
 *     property parent_keys:
 *         def __get__(self):
 *             if self.parents is None:             # <<<<<<<<<<<<<<
 *                 return None
 * 
*/
  __pyx_t_1 = (__pyx_v_self->parents == Py_None);
  if (__pyx_t_1) {

    /* "breezy/_known_graph_pyx.pyx":77
 *         def __get__(self):
 *             if self.parents is None:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *             cdef _KnownGraphNode parent
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "breezy/_known_graph_pyx.pyx":76
 *     property parent_keys:
 *         def __get__(self):
 *             if self.parents is None:             # <<<<<<<<<<<<<<
 *                 return None
 * 
*/
  }

  /* "breezy/_known_graph_pyx.pyx":81
 *             cdef _KnownGraphNode parent
 * 
 *             keys = []             # <<<<<<<<<<<<<<
 *             for parent in self.parents:
 *                 PyList_Append(keys, parent.key)
*/
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_keys = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "breezy/_known_graph_pyx.pyx":82
 * 
 *             keys = []
 *             for parent in self.parents:             # <<<<<<<<<<<<<<
 *                 PyList_Append(keys, parent.key)
 *             return keys
*/
  if (likely(PyList_CheckExact(__pyx_v_self->parents)) || PyTuple_CheckExact(__pyx_v_self->parents)) {
    __pyx_t_2 = __pyx_v_self->parents; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_self->parents); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 82, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 82, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        __pyx_t_5 = __Pyx_PyList_GetItemRefFast(__pyx_t_2, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_3;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 82, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3));
        #else
        __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3);
        #endif
        ++__pyx_t_3;
      }
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 82, __pyx_L1_error)
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 82, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_5);
    if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode))))) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_parent, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "breezy/_known_graph_pyx.pyx":83
 *             keys = []
 *             for parent in self.parents:
 *                 PyList_Append(keys, parent.key)             # <<<<<<<<<<<<<<
 *             return keys
 * 
*/
    __pyx_t_5 = __pyx_v_parent->key;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_6 = PyList_Append(__pyx_v_keys, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "breezy/_known_graph_pyx.pyx":82
 * 
 *             keys = []
 *             for parent in self.parents:             # <<<<<<<<<<<<<<
 *                 PyList_Append(keys, parent.key)
 *             return keys
*/
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/_known_graph_pyx.pyx":84
 *             for parent in self.parents:
 *                 PyList_Append(keys, parent.key)
 *             return keys             # <<<<<<<<<<<<<<
 * 
 *     cdef clear_references(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_keys);
  __pyx_r = __pyx_v_keys;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":75
 * 
 *     property parent_keys:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.parents is None:
 *                 return None
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("breezy._known_graph_pyx._KnownGraphNode.parent_keys.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_parent);
  __Pyx_XDECREF(__pyx_v_keys);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":86
 *             return keys
 * 
 *     cdef clear_references(self):             # <<<<<<<<<<<<<<
 *         self.parents = None
 *         self.children = None
*/

static PyObject *__pyx_f_6breezy_16_known_graph_pyx_15_KnownGraphNode_clear_references(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear_references", 0);

  /* "breezy/_known_graph_pyx.pyx":87
 * 
 *     cdef clear_references(self):
 *         self.parents = None             # <<<<<<<<<<<<<<
 *         self.children = None
 * 
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->parents);
  __Pyx_DECREF(__pyx_v_self->parents);
  __pyx_v_self->parents = Py_None;

  /* "breezy/_known_graph_pyx.pyx":88
 *     cdef clear_references(self):
 *         self.parents = None
 *         self.children = None             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->children);
  __Pyx_DECREF(__pyx_v_self->children);
  __pyx_v_self->children = Py_None;

  /* "breezy/_known_graph_pyx.pyx":86
 *             return keys
 * 
 *     cdef clear_references(self):             # <<<<<<<<<<<<<<
 *         self.parents = None
 *         self.children = None
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":90
 *         self.children = None
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         cdef _KnownGraphNode node
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_2__repr__(((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_2__repr__(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_self) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node = 0;
  PyObject *__pyx_v_parent_keys = NULL;
  PyObject *__pyx_v_child_keys = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10[10];
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "breezy/_known_graph_pyx.pyx":93
 *         cdef _KnownGraphNode node
 * 
 *         parent_keys = []             # <<<<<<<<<<<<<<
 *         if self.parents is not None:
 *             for node in self.parents:
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_parent_keys = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":94
 * 
 *         parent_keys = []
 *         if self.parents is not None:             # <<<<<<<<<<<<<<
 *             for node in self.parents:
 *                 parent_keys.append(node.key)
*/
  __pyx_t_2 = (__pyx_v_self->parents != Py_None);
  if (__pyx_t_2) {

    /* "breezy/_known_graph_pyx.pyx":95
 *         parent_keys = []
 *         if self.parents is not None:
 *             for node in self.parents:             # <<<<<<<<<<<<<<
 *                 parent_keys.append(node.key)
 *         child_keys = []
*/
    if (likely(PyList_CheckExact(__pyx_v_self->parents)) || PyTuple_CheckExact(__pyx_v_self->parents)) {
      __pyx_t_1 = __pyx_v_self->parents; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_self->parents); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 95, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 95, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 95, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          __pyx_t_5 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_3;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 95, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3));
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3);
          #endif
          ++__pyx_t_3;
        }
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 95, __pyx_L1_error)
      } else {
        __pyx_t_5 = __pyx_t_4(__pyx_t_1);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 95, __pyx_L1_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_5);
      if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode))))) __PYX_ERR(0, 95, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_5));
      __pyx_t_5 = 0;

      /* "breezy/_known_graph_pyx.pyx":96
 *         if self.parents is not None:
 *             for node in self.parents:
 *                 parent_keys.append(node.key)             # <<<<<<<<<<<<<<
 *         child_keys = []
 *         if self.children is not None:
*/
      __pyx_t_5 = __pyx_v_node->key;
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_parent_keys, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 96, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "breezy/_known_graph_pyx.pyx":95
 *         parent_keys = []
 *         if self.parents is not None:
 *             for node in self.parents:             # <<<<<<<<<<<<<<
 *                 parent_keys.append(node.key)
 *         child_keys = []
*/
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":94
 * 
 *         parent_keys = []
 *         if self.parents is not None:             # <<<<<<<<<<<<<<
 *             for node in self.parents:
 *                 parent_keys.append(node.key)
*/
  }

  /* "breezy/_known_graph_pyx.pyx":97
 *             for node in self.parents:
 *                 parent_keys.append(node.key)
 *         child_keys = []             # <<<<<<<<<<<<<<
 *         if self.children is not None:
 *             for node in self.children:
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_child_keys = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":98
 *                 parent_keys.append(node.key)
 *         child_keys = []
 *         if self.children is not None:             # <<<<<<<<<<<<<<
 *             for node in self.children:
 *                 child_keys.append(node.key)
*/
  __pyx_t_2 = (__pyx_v_self->children != Py_None);
  if (__pyx_t_2) {

    /* "breezy/_known_graph_pyx.pyx":99
 *         child_keys = []
 *         if self.children is not None:
 *             for node in self.children:             # <<<<<<<<<<<<<<
 *                 child_keys.append(node.key)
 *         return '%s(%s  gdfo:%s par:%s child:%s)' % (
*/
    if (likely(PyList_CheckExact(__pyx_v_self->children)) || PyTuple_CheckExact(__pyx_v_self->children)) {
      __pyx_t_1 = __pyx_v_self->children; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_self->children); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 99, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 99, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          __pyx_t_5 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_3;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 99, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3));
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3);
          #endif
          ++__pyx_t_3;
        }
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 99, __pyx_L1_error)
      } else {
        __pyx_t_5 = __pyx_t_4(__pyx_t_1);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 99, __pyx_L1_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_5);
      if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode))))) __PYX_ERR(0, 99, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_5));
      __pyx_t_5 = 0;

      /* "breezy/_known_graph_pyx.pyx":100
 *         if self.children is not None:
 *             for node in self.children:
 *                 child_keys.append(node.key)             # <<<<<<<<<<<<<<
 *         return '%s(%s  gdfo:%s par:%s child:%s)' % (
 *             self.__class__.__name__, self.key, self.gdfo,
*/
      __pyx_t_5 = __pyx_v_node->key;
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_child_keys, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 100, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "breezy/_known_graph_pyx.pyx":99
 *         child_keys = []
 *         if self.children is not None:
 *             for node in self.children:             # <<<<<<<<<<<<<<
 *                 child_keys.append(node.key)
 *         return '%s(%s  gdfo:%s par:%s child:%s)' % (
*/
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":98
 *                 parent_keys.append(node.key)
 *         child_keys = []
 *         if self.children is not None:             # <<<<<<<<<<<<<<
 *             for node in self.children:
 *                 child_keys.append(node.key)
*/
  }

  /* "breezy/_known_graph_pyx.pyx":101
 *             for node in self.children:
 *                 child_keys.append(node.key)
 *         return '%s(%s  gdfo:%s par:%s child:%s)' % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__, self.key, self.gdfo,
 *             parent_keys, child_keys)
*/
  __Pyx_XDECREF(__pyx_r);

  /* "breezy/_known_graph_pyx.pyx":102
 *                 child_keys.append(node.key)
 *         return '%s(%s  gdfo:%s par:%s child:%s)' % (
 *             self.__class__.__name__, self.key, self.gdfo,             # <<<<<<<<<<<<<<
 *             parent_keys, child_keys)
 * 
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_t_5), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_v_self->key), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyUnicode_From_long(__pyx_v_self->gdfo, 0, ' ', 'd'); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "breezy/_known_graph_pyx.pyx":103
 *         return '%s(%s  gdfo:%s par:%s child:%s)' % (
 *             self.__class__.__name__, self.key, self.gdfo,
 *             parent_keys, child_keys)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_8 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_v_parent_keys), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_v_child_keys), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10[0] = __pyx_t_1;
  __pyx_t_10[1] = __pyx_mstate_global->__pyx_kp_u_;
  __pyx_t_10[2] = __pyx_t_5;
  __pyx_t_10[3] = __pyx_mstate_global->__pyx_kp_u_gdfo;
  __pyx_t_10[4] = __pyx_t_7;
  __pyx_t_10[5] = __pyx_mstate_global->__pyx_kp_u_par;
  __pyx_t_10[6] = __pyx_t_8;
  __pyx_t_10[7] = __pyx_mstate_global->__pyx_kp_u_child;
  __pyx_t_10[8] = __pyx_t_9;
  __pyx_t_10[9] = __pyx_mstate_global->__pyx_kp_u__2;

  /* "breezy/_known_graph_pyx.pyx":101
 *             for node in self.children:
 *                 child_keys.append(node.key)
 *         return '%s(%s  gdfo:%s par:%s child:%s)' % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__, self.key, self.gdfo,
 *             parent_keys, child_keys)
*/
  __pyx_t_11 = __Pyx_PyUnicode_Join(__pyx_t_10, 10, __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1) + 1 * 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5) + 7 * 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7) + 5 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9));
  if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_11;
  __pyx_t_11 = 0;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":90
 *         self.children = None
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         cdef _KnownGraphNode node
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("breezy._known_graph_pyx._KnownGraphNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF(__pyx_v_parent_keys);
  __Pyx_XDECREF(__pyx_v_child_keys);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":51
 *     cdef object parents
 *     cdef object children
 *     cdef public long gdfo             # <<<<<<<<<<<<<<
 *     cdef int seen
 *     cdef object extra
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_4gdfo_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_4gdfo_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_4gdfo___get__(((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_4gdfo___get__(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_long(__pyx_v_self->gdfo); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy._known_graph_pyx._KnownGraphNode.gdfo.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_4gdfo_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_4gdfo_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_4gdfo_2__set__(((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_4gdfo_2__set__(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyLong_As_long(__pyx_v_value); if (unlikely((__pyx_t_1 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 51, __pyx_L1_error)
  __pyx_v_self->gdfo = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("breezy._known_graph_pyx._KnownGraphNode.gdfo.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_15_KnownGraphNode_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_4__reduce_cython__(((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_4__reduce_cython__(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.children, self.extra, self.gdfo, self.key, self.parents, self.seen)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __pyx_t_1 = __Pyx_PyLong_From_long(__pyx_v_self->gdfo); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_v_self->seen); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(6); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->children);
  __Pyx_GIVEREF(__pyx_v_self->children);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->children) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->extra);
  __Pyx_GIVEREF(__pyx_v_self->extra);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self->extra) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_1) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->key);
  __Pyx_GIVEREF(__pyx_v_self->key);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_self->key) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->parents);
  __Pyx_GIVEREF(__pyx_v_self->parents);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_v_self->parents) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.children, self.extra, self.gdfo, self.key, self.parents, self.seen)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.children, self.extra, self.gdfo, self.key, self.parents, self.seen)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_5 = (__pyx_v__dict != Py_None);
  if (__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(1, 7, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict) != (0)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.children is not None or self.extra is not None or self.key is not None or self.parents is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.children, self.extra, self.gdfo, self.key, self.parents, self.seen)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.children is not None or self.extra is not None or self.key is not None or self.parents is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__KnownGraphNode, (type(self), 0x1857be3, None), state
*/
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->children != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->extra != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->key != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->parents != Py_None);
    __pyx_t_4 = __pyx_t_5;
    __pyx_L6_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.children is not None or self.extra is not None or self.key is not None or self.parents is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__KnownGraphNode, (type(self), 0x1857be3, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.children is not None or self.extra is not None or self.key is not None or self.parents is not None
 *     if use_setstate:
 *         return __pyx_unpickle__KnownGraphNode, (type(self), 0x1857be3, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__KnownGraphNode, (type(self), 0x1857be3, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__KnownGraphNode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_25525219);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_25525219);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_mstate_global->__pyx_int_25525219) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.children is not None or self.extra is not None or self.key is not None or self.parents is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__KnownGraphNode, (type(self), 0x1857be3, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__KnownGraphNode, (type(self), 0x1857be3, None), state
 *     else:
 *         return __pyx_unpickle__KnownGraphNode, (type(self), 0x1857be3, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__KnownGraphNode__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__KnownGraphNode); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_25525219);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_25525219);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_mstate_global->__pyx_int_25525219) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy._known_graph_pyx._KnownGraphNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__KnownGraphNode, (type(self), 0x1857be3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__KnownGraphNode__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_15_KnownGraphNode_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx._KnownGraphNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_6__setstate_cython__(((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_15_KnownGraphNode_6__setstate_cython__(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__KnownGraphNode, (type(self), 0x1857be3, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__KnownGraphNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(1, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_6breezy_16_known_graph_pyx___pyx_unpickle__KnownGraphNode__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__KnownGraphNode, (type(self), 0x1857be3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__KnownGraphNode__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy._known_graph_pyx._KnownGraphNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":106
 * 
 * 
 * cdef _KnownGraphNode _get_list_node(lst, Py_ssize_t pos):             # <<<<<<<<<<<<<<
 *     cdef PyObject *temp_node
 * 
*/

static struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_f_6breezy_16_known_graph_pyx__get_list_node(PyObject *__pyx_v_lst, Py_ssize_t __pyx_v_pos) {
  PyObject *__pyx_v_temp_node;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_list_node", 0);

  /* "breezy/_known_graph_pyx.pyx":109
 *     cdef PyObject *temp_node
 * 
 *     temp_node = PyList_GET_ITEM(lst, pos)             # <<<<<<<<<<<<<<
 *     return <_KnownGraphNode>temp_node
 * 
*/
  __pyx_v_temp_node = PyList_GET_ITEM(__pyx_v_lst, __pyx_v_pos);

  /* "breezy/_known_graph_pyx.pyx":110
 * 
 *     temp_node = PyList_GET_ITEM(lst, pos)
 *     return <_KnownGraphNode>temp_node             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_v_temp_node));
  __pyx_r = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_v_temp_node);
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":106
 * 
 * 
 * cdef _KnownGraphNode _get_list_node(lst, Py_ssize_t pos):             # <<<<<<<<<<<<<<
 *     cdef PyObject *temp_node
 * 
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":113
 * 
 * 
 * cdef _KnownGraphNode _get_tuple_node(tpl, Py_ssize_t pos):             # <<<<<<<<<<<<<<
 *     cdef PyObject *temp_node
 * 
*/

static struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_f_6breezy_16_known_graph_pyx__get_tuple_node(PyObject *__pyx_v_tpl, Py_ssize_t __pyx_v_pos) {
  PyObject *__pyx_v_temp_node;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_tuple_node", 0);

  /* "breezy/_known_graph_pyx.pyx":116
 *     cdef PyObject *temp_node
 * 
 *     temp_node = PyTuple_GET_ITEM(tpl, pos)             # <<<<<<<<<<<<<<
 *     return <_KnownGraphNode>temp_node
 * 
*/
  __pyx_v_temp_node = PyTuple_GET_ITEM(__pyx_v_tpl, __pyx_v_pos);

  /* "breezy/_known_graph_pyx.pyx":117
 * 
 *     temp_node = PyTuple_GET_ITEM(tpl, pos)
 *     return <_KnownGraphNode>temp_node             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_v_temp_node));
  __pyx_r = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_v_temp_node);
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":113
 * 
 * 
 * cdef _KnownGraphNode _get_tuple_node(tpl, Py_ssize_t pos):             # <<<<<<<<<<<<<<
 *     cdef PyObject *temp_node
 * 
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":120
 * 
 * 
 * def get_key(node):             # <<<<<<<<<<<<<<
 *     cdef _KnownGraphNode real_node
 *     real_node = node
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_1get_key(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_1get_key = {"get_key", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_1get_key, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_1get_key(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_node = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_key (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_node,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 120, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 120, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_key", 0) < (0)) __PYX_ERR(0, 120, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_key", 1, 1, 1, i); __PYX_ERR(0, 120, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 120, __pyx_L3_error)
    }
    __pyx_v_node = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_key", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 120, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx.get_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_get_key(__pyx_self, __pyx_v_node);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_get_key(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_node) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_real_node = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_key", 0);

  /* "breezy/_known_graph_pyx.pyx":122
 * def get_key(node):
 *     cdef _KnownGraphNode real_node
 *     real_node = node             # <<<<<<<<<<<<<<
 *     return real_node.key
 * 
*/
  __pyx_t_1 = __pyx_v_node;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode))))) __PYX_ERR(0, 122, __pyx_L1_error)
  __pyx_v_real_node = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":123
 *     cdef _KnownGraphNode real_node
 *     real_node = node
 *     return real_node.key             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_real_node->key);
  __pyx_r = __pyx_v_real_node->key;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":120
 * 
 * 
 * def get_key(node):             # <<<<<<<<<<<<<<
 *     cdef _KnownGraphNode real_node
 *     real_node = node
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy._known_graph_pyx.get_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_real_node);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":126
 * 
 * 
 * cdef object _sort_list_nodes(object lst_or_tpl, int reverse):             # <<<<<<<<<<<<<<
 *     """Sort a list of _KnownGraphNode objects.
 * 
*/

static PyObject *__pyx_f_6breezy_16_known_graph_pyx__sort_list_nodes(PyObject *__pyx_v_lst_or_tpl, int __pyx_v_reverse) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node1 = 0;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node2 = 0;
  int __pyx_v_do_swap;
  int __pyx_v_is_tuple;
  Py_ssize_t __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_sort_list_nodes", 0);
  __Pyx_INCREF(__pyx_v_lst_or_tpl);

  /* "breezy/_known_graph_pyx.pyx":136
 *     cdef Py_ssize_t length
 * 
 *     is_tuple = PyTuple_CheckExact(lst_or_tpl)             # <<<<<<<<<<<<<<
 *     if not (is_tuple or PyList_CheckExact(lst_or_tpl)):
 *         raise TypeError('lst_or_tpl must be a list or tuple.')
*/
  __pyx_v_is_tuple = PyTuple_CheckExact(__pyx_v_lst_or_tpl);

  /* "breezy/_known_graph_pyx.pyx":137
 * 
 *     is_tuple = PyTuple_CheckExact(lst_or_tpl)
 *     if not (is_tuple or PyList_CheckExact(lst_or_tpl)):             # <<<<<<<<<<<<<<
 *         raise TypeError('lst_or_tpl must be a list or tuple.')
 *     length = len(lst_or_tpl)
*/
  __pyx_t_2 = (__pyx_v_is_tuple != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = PyList_CheckExact(__pyx_v_lst_or_tpl);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  __pyx_t_2 = (!__pyx_t_1);
  if (unlikely(__pyx_t_2)) {

    /* "breezy/_known_graph_pyx.pyx":138
 *     is_tuple = PyTuple_CheckExact(lst_or_tpl)
 *     if not (is_tuple or PyList_CheckExact(lst_or_tpl)):
 *         raise TypeError('lst_or_tpl must be a list or tuple.')             # <<<<<<<<<<<<<<
 *     length = len(lst_or_tpl)
 *     if length == 0 or length == 1:
*/
    __pyx_t_4 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_lst_or_tpl_must_be_a_list_or_tup};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 138, __pyx_L1_error)

    /* "breezy/_known_graph_pyx.pyx":137
 * 
 *     is_tuple = PyTuple_CheckExact(lst_or_tpl)
 *     if not (is_tuple or PyList_CheckExact(lst_or_tpl)):             # <<<<<<<<<<<<<<
 *         raise TypeError('lst_or_tpl must be a list or tuple.')
 *     length = len(lst_or_tpl)
*/
  }

  /* "breezy/_known_graph_pyx.pyx":139
 *     if not (is_tuple or PyList_CheckExact(lst_or_tpl)):
 *         raise TypeError('lst_or_tpl must be a list or tuple.')
 *     length = len(lst_or_tpl)             # <<<<<<<<<<<<<<
 *     if length == 0 or length == 1:
 *         return lst_or_tpl
*/
  __pyx_t_6 = PyObject_Length(__pyx_v_lst_or_tpl); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 139, __pyx_L1_error)
  __pyx_v_length = __pyx_t_6;

  /* "breezy/_known_graph_pyx.pyx":140
 *         raise TypeError('lst_or_tpl must be a list or tuple.')
 *     length = len(lst_or_tpl)
 *     if length == 0 or length == 1:             # <<<<<<<<<<<<<<
 *         return lst_or_tpl
 *     if length == 2:
*/
  switch (__pyx_v_length) {
    case 0:
    case 1:

    /* "breezy/_known_graph_pyx.pyx":141
 *     length = len(lst_or_tpl)
 *     if length == 0 or length == 1:
 *         return lst_or_tpl             # <<<<<<<<<<<<<<
 *     if length == 2:
 *         if is_tuple:
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_lst_or_tpl);
    __pyx_r = __pyx_v_lst_or_tpl;
    goto __pyx_L0;

    /* "breezy/_known_graph_pyx.pyx":140
 *         raise TypeError('lst_or_tpl must be a list or tuple.')
 *     length = len(lst_or_tpl)
 *     if length == 0 or length == 1:             # <<<<<<<<<<<<<<
 *         return lst_or_tpl
 *     if length == 2:
*/
    break;
    default: break;
  }

  /* "breezy/_known_graph_pyx.pyx":142
 *     if length == 0 or length == 1:
 *         return lst_or_tpl
 *     if length == 2:             # <<<<<<<<<<<<<<
 *         if is_tuple:
 *             node1 = _get_tuple_node(lst_or_tpl, 0)
*/
  __pyx_t_2 = (__pyx_v_length == 2);
  if (__pyx_t_2) {

    /* "breezy/_known_graph_pyx.pyx":143
 *         return lst_or_tpl
 *     if length == 2:
 *         if is_tuple:             # <<<<<<<<<<<<<<
 *             node1 = _get_tuple_node(lst_or_tpl, 0)
 *             node2 = _get_tuple_node(lst_or_tpl, 1)
*/
    __pyx_t_2 = (__pyx_v_is_tuple != 0);
    if (__pyx_t_2) {

      /* "breezy/_known_graph_pyx.pyx":144
 *     if length == 2:
 *         if is_tuple:
 *             node1 = _get_tuple_node(lst_or_tpl, 0)             # <<<<<<<<<<<<<<
 *             node2 = _get_tuple_node(lst_or_tpl, 1)
 *         else:
*/
      __pyx_t_3 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_tuple_node(__pyx_v_lst_or_tpl, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 144, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_node1 = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "breezy/_known_graph_pyx.pyx":145
 *         if is_tuple:
 *             node1 = _get_tuple_node(lst_or_tpl, 0)
 *             node2 = _get_tuple_node(lst_or_tpl, 1)             # <<<<<<<<<<<<<<
 *         else:
 *             node1 = _get_list_node(lst_or_tpl, 0)
*/
      __pyx_t_3 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_tuple_node(__pyx_v_lst_or_tpl, 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 145, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_node2 = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "breezy/_known_graph_pyx.pyx":143
 *         return lst_or_tpl
 *     if length == 2:
 *         if is_tuple:             # <<<<<<<<<<<<<<
 *             node1 = _get_tuple_node(lst_or_tpl, 0)
 *             node2 = _get_tuple_node(lst_or_tpl, 1)
*/
      goto __pyx_L7;
    }

    /* "breezy/_known_graph_pyx.pyx":147
 *             node2 = _get_tuple_node(lst_or_tpl, 1)
 *         else:
 *             node1 = _get_list_node(lst_or_tpl, 0)             # <<<<<<<<<<<<<<
 *             node2 = _get_list_node(lst_or_tpl, 1)
 *         if reverse:
*/
    /*else*/ {
      __pyx_t_3 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_list_node(__pyx_v_lst_or_tpl, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 147, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_node1 = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "breezy/_known_graph_pyx.pyx":148
 *         else:
 *             node1 = _get_list_node(lst_or_tpl, 0)
 *             node2 = _get_list_node(lst_or_tpl, 1)             # <<<<<<<<<<<<<<
 *         if reverse:
 *             do_swap = PyObject_RichCompareBool(node1.key, node2.key, Py_LT)
*/
      __pyx_t_3 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_list_node(__pyx_v_lst_or_tpl, 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_node2 = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_3);
      __pyx_t_3 = 0;
    }
    __pyx_L7:;

    /* "breezy/_known_graph_pyx.pyx":149
 *             node1 = _get_list_node(lst_or_tpl, 0)
 *             node2 = _get_list_node(lst_or_tpl, 1)
 *         if reverse:             # <<<<<<<<<<<<<<
 *             do_swap = PyObject_RichCompareBool(node1.key, node2.key, Py_LT)
 *         else:
*/
    __pyx_t_2 = (__pyx_v_reverse != 0);
    if (__pyx_t_2) {

      /* "breezy/_known_graph_pyx.pyx":150
 *             node2 = _get_list_node(lst_or_tpl, 1)
 *         if reverse:
 *             do_swap = PyObject_RichCompareBool(node1.key, node2.key, Py_LT)             # <<<<<<<<<<<<<<
 *         else:
 *             do_swap = PyObject_RichCompareBool(node2.key, node1.key, Py_LT)
*/
      __pyx_t_3 = __pyx_v_node1->key;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = __pyx_v_node2->key;
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_2 = PyObject_RichCompareBool(__pyx_t_3, __pyx_t_4, Py_LT); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 150, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_do_swap = __pyx_t_2;

      /* "breezy/_known_graph_pyx.pyx":149
 *             node1 = _get_list_node(lst_or_tpl, 0)
 *             node2 = _get_list_node(lst_or_tpl, 1)
 *         if reverse:             # <<<<<<<<<<<<<<
 *             do_swap = PyObject_RichCompareBool(node1.key, node2.key, Py_LT)
 *         else:
*/
      goto __pyx_L8;
    }

    /* "breezy/_known_graph_pyx.pyx":152
 *             do_swap = PyObject_RichCompareBool(node1.key, node2.key, Py_LT)
 *         else:
 *             do_swap = PyObject_RichCompareBool(node2.key, node1.key, Py_LT)             # <<<<<<<<<<<<<<
 *         if not do_swap:
 *             return lst_or_tpl
*/
    /*else*/ {
      __pyx_t_4 = __pyx_v_node2->key;
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_3 = __pyx_v_node1->key;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_2 = PyObject_RichCompareBool(__pyx_t_4, __pyx_t_3, Py_LT); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 152, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_do_swap = __pyx_t_2;
    }
    __pyx_L8:;

    /* "breezy/_known_graph_pyx.pyx":153
 *         else:
 *             do_swap = PyObject_RichCompareBool(node2.key, node1.key, Py_LT)
 *         if not do_swap:             # <<<<<<<<<<<<<<
 *             return lst_or_tpl
 *         if is_tuple:
*/
    __pyx_t_2 = (!(__pyx_v_do_swap != 0));
    if (__pyx_t_2) {

      /* "breezy/_known_graph_pyx.pyx":154
 *             do_swap = PyObject_RichCompareBool(node2.key, node1.key, Py_LT)
 *         if not do_swap:
 *             return lst_or_tpl             # <<<<<<<<<<<<<<
 *         if is_tuple:
 *             return (node2, node1)
*/
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_lst_or_tpl);
      __pyx_r = __pyx_v_lst_or_tpl;
      goto __pyx_L0;

      /* "breezy/_known_graph_pyx.pyx":153
 *         else:
 *             do_swap = PyObject_RichCompareBool(node2.key, node1.key, Py_LT)
 *         if not do_swap:             # <<<<<<<<<<<<<<
 *             return lst_or_tpl
 *         if is_tuple:
*/
    }

    /* "breezy/_known_graph_pyx.pyx":155
 *         if not do_swap:
 *             return lst_or_tpl
 *         if is_tuple:             # <<<<<<<<<<<<<<
 *             return (node2, node1)
 *         else:
*/
    __pyx_t_2 = (__pyx_v_is_tuple != 0);
    if (__pyx_t_2) {

      /* "breezy/_known_graph_pyx.pyx":156
 *             return lst_or_tpl
 *         if is_tuple:
 *             return (node2, node1)             # <<<<<<<<<<<<<<
 *         else:
 *             # Swap 'in-place', since lists are mutable
*/
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF((PyObject *)__pyx_v_node2);
      __Pyx_GIVEREF((PyObject *)__pyx_v_node2);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_node2)) != (0)) __PYX_ERR(0, 156, __pyx_L1_error);
      __Pyx_INCREF((PyObject *)__pyx_v_node1);
      __Pyx_GIVEREF((PyObject *)__pyx_v_node1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_node1)) != (0)) __PYX_ERR(0, 156, __pyx_L1_error);
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "breezy/_known_graph_pyx.pyx":155
 *         if not do_swap:
 *             return lst_or_tpl
 *         if is_tuple:             # <<<<<<<<<<<<<<
 *             return (node2, node1)
 *         else:
*/
    }

    /* "breezy/_known_graph_pyx.pyx":159
 *         else:
 *             # Swap 'in-place', since lists are mutable
 *             Py_INCREF(node1)             # <<<<<<<<<<<<<<
 *             PyList_SetItem(lst_or_tpl, 1, node1)
 *             Py_INCREF(node2)
*/
    /*else*/ {
      Py_INCREF(((PyObject *)__pyx_v_node1));

      /* "breezy/_known_graph_pyx.pyx":160
 *             # Swap 'in-place', since lists are mutable
 *             Py_INCREF(node1)
 *             PyList_SetItem(lst_or_tpl, 1, node1)             # <<<<<<<<<<<<<<
 *             Py_INCREF(node2)
 *             PyList_SetItem(lst_or_tpl, 0, node2)
*/
      __pyx_t_7 = PyList_SetItem(__pyx_v_lst_or_tpl, 1, ((PyObject *)__pyx_v_node1)); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 160, __pyx_L1_error)

      /* "breezy/_known_graph_pyx.pyx":161
 *             Py_INCREF(node1)
 *             PyList_SetItem(lst_or_tpl, 1, node1)
 *             Py_INCREF(node2)             # <<<<<<<<<<<<<<
 *             PyList_SetItem(lst_or_tpl, 0, node2)
 *             return lst_or_tpl
*/
      Py_INCREF(((PyObject *)__pyx_v_node2));

      /* "breezy/_known_graph_pyx.pyx":162
 *             PyList_SetItem(lst_or_tpl, 1, node1)
 *             Py_INCREF(node2)
 *             PyList_SetItem(lst_or_tpl, 0, node2)             # <<<<<<<<<<<<<<
 *             return lst_or_tpl
 *     # For all other sizes, we just use 'sorted()'
*/
      __pyx_t_7 = PyList_SetItem(__pyx_v_lst_or_tpl, 0, ((PyObject *)__pyx_v_node2)); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 162, __pyx_L1_error)

      /* "breezy/_known_graph_pyx.pyx":163
 *             Py_INCREF(node2)
 *             PyList_SetItem(lst_or_tpl, 0, node2)
 *             return lst_or_tpl             # <<<<<<<<<<<<<<
 *     # For all other sizes, we just use 'sorted()'
 *     if is_tuple:
*/
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_lst_or_tpl);
      __pyx_r = __pyx_v_lst_or_tpl;
      goto __pyx_L0;
    }

    /* "breezy/_known_graph_pyx.pyx":142
 *     if length == 0 or length == 1:
 *         return lst_or_tpl
 *     if length == 2:             # <<<<<<<<<<<<<<
 *         if is_tuple:
 *             node1 = _get_tuple_node(lst_or_tpl, 0)
*/
  }

  /* "breezy/_known_graph_pyx.pyx":165
 *             return lst_or_tpl
 *     # For all other sizes, we just use 'sorted()'
 *     if is_tuple:             # <<<<<<<<<<<<<<
 *         # Note that sorted() is just list(iterable).sort()
 *         lst_or_tpl = list(lst_or_tpl)
*/
  __pyx_t_2 = (__pyx_v_is_tuple != 0);
  if (__pyx_t_2) {

    /* "breezy/_known_graph_pyx.pyx":167
 *     if is_tuple:
 *         # Note that sorted() is just list(iterable).sort()
 *         lst_or_tpl = list(lst_or_tpl)             # <<<<<<<<<<<<<<
 *     lst_or_tpl.sort(key=get_key, reverse=reverse)
 *     return lst_or_tpl
*/
    __pyx_t_3 = PySequence_List(__pyx_v_lst_or_tpl); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_lst_or_tpl, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "breezy/_known_graph_pyx.pyx":165
 *             return lst_or_tpl
 *     # For all other sizes, we just use 'sorted()'
 *     if is_tuple:             # <<<<<<<<<<<<<<
 *         # Note that sorted() is just list(iterable).sort()
 *         lst_or_tpl = list(lst_or_tpl)
*/
  }

  /* "breezy/_known_graph_pyx.pyx":168
 *         # Note that sorted() is just list(iterable).sort()
 *         lst_or_tpl = list(lst_or_tpl)
 *     lst_or_tpl.sort(key=get_key, reverse=reverse)             # <<<<<<<<<<<<<<
 *     return lst_or_tpl
 * 
*/
  __pyx_t_4 = __pyx_v_lst_or_tpl;
  __Pyx_INCREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_get_key); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyLong_From_int(__pyx_v_reverse); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 2 : 0)] = {__pyx_t_4, NULL};
    __pyx_t_10 = __Pyx_MakeVectorcallBuilderKwds(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_key, __pyx_t_8, __pyx_t_10, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 168, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_reverse, __pyx_t_9, __pyx_t_10, __pyx_callargs+1, 1) < (0)) __PYX_ERR(0, 168, __pyx_L1_error)
    __pyx_t_3 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_sort, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "breezy/_known_graph_pyx.pyx":169
 *         lst_or_tpl = list(lst_or_tpl)
 *     lst_or_tpl.sort(key=get_key, reverse=reverse)
 *     return lst_or_tpl             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_lst_or_tpl);
  __pyx_r = __pyx_v_lst_or_tpl;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":126
 * 
 * 
 * cdef object _sort_list_nodes(object lst_or_tpl, int reverse):             # <<<<<<<<<<<<<<
 *     """Sort a list of _KnownGraphNode objects.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("breezy._known_graph_pyx._sort_list_nodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node1);
  __Pyx_XDECREF((PyObject *)__pyx_v_node2);
  __Pyx_XDECREF(__pyx_v_lst_or_tpl);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":181
 *     cdef public int do_cache
 * 
 *     def __init__(self, parent_map, do_cache=True):             # <<<<<<<<<<<<<<
 *         """Create a new KnownGraph instance.
 * 
*/

/* Python wrapper */
static int __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph___init__, "Create a new KnownGraph instance.\n\n        :param parent_map: A dictionary mapping key => parent_keys\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_6breezy_16_known_graph_pyx_10KnownGraph___init__;
#endif
static int __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_parent_map = 0;
  PyObject *__pyx_v_do_cache = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_parent_map,&__pyx_mstate_global->__pyx_n_u_do_cache,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 181, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 181, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 181, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 181, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_True));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, i); __PYX_ERR(0, 181, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 181, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 181, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_True));
    }
    __pyx_v_parent_map = values[0];
    __pyx_v_do_cache = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 181, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph___init__(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self), __pyx_v_parent_map, __pyx_v_do_cache);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph___init__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_parent_map, PyObject *__pyx_v_do_cache) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "breezy/_known_graph_pyx.pyx":187
 *         """
 *         # tests at pre-allocating the node dict actually slowed things down
 *         self._nodes = {}             # <<<<<<<<<<<<<<
 *         # Maps {sorted(revision_id, revision_id): heads}
 *         self._known_heads = {}
*/
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_nodes);
  __Pyx_DECREF(__pyx_v_self->_nodes);
  __pyx_v_self->_nodes = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":189
 *         self._nodes = {}
 *         # Maps {sorted(revision_id, revision_id): heads}
 *         self._known_heads = {}             # <<<<<<<<<<<<<<
 *         self.do_cache = int(do_cache)
 *         # TODO: consider disabling gc since we are allocating a lot of nodes
*/
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_known_heads);
  __Pyx_DECREF(__pyx_v_self->_known_heads);
  __pyx_v_self->_known_heads = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":190
 *         # Maps {sorted(revision_id, revision_id): heads}
 *         self._known_heads = {}
 *         self.do_cache = int(do_cache)             # <<<<<<<<<<<<<<
 *         # TODO: consider disabling gc since we are allocating a lot of nodes
 *         #       that won't be collectable anyway. real world testing has not
*/
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_v_do_cache); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->do_cache = __pyx_t_2;

  /* "breezy/_known_graph_pyx.pyx":194
 *         #       that won't be collectable anyway. real world testing has not
 *         #       shown a specific impact, yet.
 *         self._initialize_nodes(parent_map)             # <<<<<<<<<<<<<<
 *         self._find_gdfo()
 * 
*/
  __pyx_t_3 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_parent_map};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_initialize_nodes, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":195
 *         #       shown a specific impact, yet.
 *         self._initialize_nodes(parent_map)
 *         self._find_gdfo()             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
*/
  __pyx_t_3 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_find_gdfo, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":181
 *     cdef public int do_cache
 * 
 *     def __init__(self, parent_map, do_cache=True):             # <<<<<<<<<<<<<<
 *         """Create a new KnownGraph instance.
 * 
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":197
 *         self._find_gdfo()
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         cdef _KnownGraphNode child
 *         cdef Py_ssize_t pos
*/

/* Python wrapper */
static void __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_2__dealloc__(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_2__dealloc__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_child = 0;
  Py_ssize_t __pyx_v_pos;
  PyObject *__pyx_v_temp_node;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "breezy/_known_graph_pyx.pyx":202
 *         cdef PyObject *temp_node
 * 
 *         while PyDict_Next(self._nodes, &pos, NULL, &temp_node):             # <<<<<<<<<<<<<<
 *             child = <_KnownGraphNode>temp_node
 *             child.clear_references()
*/
  while (1) {
    __pyx_t_1 = __pyx_v_self->_nodes;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = (PyDict_Next(__pyx_t_1, (&__pyx_v_pos), NULL, (&__pyx_v_temp_node)) != 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_2) break;

    /* "breezy/_known_graph_pyx.pyx":203
 * 
 *         while PyDict_Next(self._nodes, &pos, NULL, &temp_node):
 *             child = <_KnownGraphNode>temp_node             # <<<<<<<<<<<<<<
 *             child.clear_references()
 * 
*/
    __pyx_t_1 = ((PyObject *)__pyx_v_temp_node);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_child, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":204
 *         while PyDict_Next(self._nodes, &pos, NULL, &temp_node):
 *             child = <_KnownGraphNode>temp_node
 *             child.clear_references()             # <<<<<<<<<<<<<<
 * 
 *     cdef _KnownGraphNode _get_or_create_node(self, key):
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_v_child->__pyx_vtab)->clear_references(__pyx_v_child); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "breezy/_known_graph_pyx.pyx":197
 *         self._find_gdfo()
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         cdef _KnownGraphNode child
 *         cdef Py_ssize_t pos
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("breezy._known_graph_pyx.KnownGraph.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_RefNannyFinishContext();
}

/* "breezy/_known_graph_pyx.pyx":206
 *             child.clear_references()
 * 
 *     cdef _KnownGraphNode _get_or_create_node(self, key):             # <<<<<<<<<<<<<<
 *         cdef PyObject *temp_node
 *         cdef _KnownGraphNode node
*/

static struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_f_6breezy_16_known_graph_pyx_10KnownGraph__get_or_create_node(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_v_temp_node;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node = 0;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_or_create_node", 0);

  /* "breezy/_known_graph_pyx.pyx":210
 *         cdef _KnownGraphNode node
 * 
 *         temp_node = PyDict_GetItem(self._nodes, key)             # <<<<<<<<<<<<<<
 *         if temp_node == NULL:
 *             node = _KnownGraphNode(key)
*/
  __pyx_t_1 = __pyx_v_self->_nodes;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_temp_node = PyDict_GetItem(__pyx_t_1, __pyx_v_key);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":211
 * 
 *         temp_node = PyDict_GetItem(self._nodes, key)
 *         if temp_node == NULL:             # <<<<<<<<<<<<<<
 *             node = _KnownGraphNode(key)
 *             PyDict_SetItem(self._nodes, key, node)
*/
  __pyx_t_2 = (__pyx_v_temp_node == NULL);
  if (__pyx_t_2) {

    /* "breezy/_known_graph_pyx.pyx":212
 *         temp_node = PyDict_GetItem(self._nodes, key)
 *         if temp_node == NULL:
 *             node = _KnownGraphNode(key)             # <<<<<<<<<<<<<<
 *             PyDict_SetItem(self._nodes, key, node)
 *         else:
*/
    __pyx_t_3 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_key};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_1);
    }
    __pyx_v_node = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":213
 *         if temp_node == NULL:
 *             node = _KnownGraphNode(key)
 *             PyDict_SetItem(self._nodes, key, node)             # <<<<<<<<<<<<<<
 *         else:
 *             node = <_KnownGraphNode>temp_node
*/
    __pyx_t_1 = __pyx_v_self->_nodes;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = PyDict_SetItem(__pyx_t_1, __pyx_v_key, ((PyObject *)__pyx_v_node)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":211
 * 
 *         temp_node = PyDict_GetItem(self._nodes, key)
 *         if temp_node == NULL:             # <<<<<<<<<<<<<<
 *             node = _KnownGraphNode(key)
 *             PyDict_SetItem(self._nodes, key, node)
*/
    goto __pyx_L3;
  }

  /* "breezy/_known_graph_pyx.pyx":215
 *             PyDict_SetItem(self._nodes, key, node)
 *         else:
 *             node = <_KnownGraphNode>temp_node             # <<<<<<<<<<<<<<
 *         return node
 * 
*/
  /*else*/ {
    __pyx_t_1 = ((PyObject *)__pyx_v_temp_node);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_node = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "breezy/_known_graph_pyx.pyx":216
 *         else:
 *             node = <_KnownGraphNode>temp_node
 *         return node             # <<<<<<<<<<<<<<
 * 
 *     cdef _populate_parents(self, _KnownGraphNode node, parent_keys):
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_node);
  __pyx_r = __pyx_v_node;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":206
 *             child.clear_references()
 * 
 *     cdef _KnownGraphNode _get_or_create_node(self, key):             # <<<<<<<<<<<<<<
 *         cdef PyObject *temp_node
 *         cdef _KnownGraphNode node
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph._get_or_create_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":218
 *         return node
 * 
 *     cdef _populate_parents(self, _KnownGraphNode node, parent_keys):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t num_parent_keys, pos
 *         cdef _KnownGraphNode parent_node
*/

static PyObject *__pyx_f_6breezy_16_known_graph_pyx_10KnownGraph__populate_parents(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node, PyObject *__pyx_v_parent_keys) {
  Py_ssize_t __pyx_v_num_parent_keys;
  Py_ssize_t __pyx_v_pos;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_parent_node = 0;
  PyObject *__pyx_v_parent_nodes = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_populate_parents", 0);

  /* "breezy/_known_graph_pyx.pyx":222
 *         cdef _KnownGraphNode parent_node
 * 
 *         num_parent_keys = len(parent_keys)             # <<<<<<<<<<<<<<
 *         # We know how many parents, so we pre allocate the tuple
 *         parent_nodes = PyTuple_New(num_parent_keys)
*/
  __pyx_t_1 = PyObject_Length(__pyx_v_parent_keys); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 222, __pyx_L1_error)
  __pyx_v_num_parent_keys = __pyx_t_1;

  /* "breezy/_known_graph_pyx.pyx":224
 *         num_parent_keys = len(parent_keys)
 *         # We know how many parents, so we pre allocate the tuple
 *         parent_nodes = PyTuple_New(num_parent_keys)             # <<<<<<<<<<<<<<
 *         for pos from 0 <= pos < num_parent_keys:
 *             # Note: it costs us 10ms out of 40ms to lookup all of these
*/
  __pyx_t_2 = PyTuple_New(__pyx_v_num_parent_keys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_parent_nodes = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "breezy/_known_graph_pyx.pyx":225
 *         # We know how many parents, so we pre allocate the tuple
 *         parent_nodes = PyTuple_New(num_parent_keys)
 *         for pos from 0 <= pos < num_parent_keys:             # <<<<<<<<<<<<<<
 *             # Note: it costs us 10ms out of 40ms to lookup all of these
 *             #       parents, it doesn't seem to be an allocation overhead,
*/
  __pyx_t_1 = __pyx_v_num_parent_keys;
  for (__pyx_v_pos = 0; __pyx_v_pos < __pyx_t_1; __pyx_v_pos++) {

    /* "breezy/_known_graph_pyx.pyx":233
 *             #       node.
 *             # We use [] because parent_keys may be a tuple or list
 *             parent_node = self._get_or_create_node(parent_keys[pos])             # <<<<<<<<<<<<<<
 *             # PyTuple_SET_ITEM will steal a reference, so INCREF first
 *             Py_INCREF(parent_node)
*/
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_parent_keys, __pyx_v_pos, Py_ssize_t, 1, PyLong_FromSsize_t, 0, 1, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 233, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self->__pyx_vtab)->_get_or_create_node(__pyx_v_self, __pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 233, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_parent_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "breezy/_known_graph_pyx.pyx":235
 *             parent_node = self._get_or_create_node(parent_keys[pos])
 *             # PyTuple_SET_ITEM will steal a reference, so INCREF first
 *             Py_INCREF(parent_node)             # <<<<<<<<<<<<<<
 *             PyTuple_SET_ITEM(parent_nodes, pos, parent_node)
 *             PyList_Append(parent_node.children, node)
*/
    Py_INCREF(((PyObject *)__pyx_v_parent_node));

    /* "breezy/_known_graph_pyx.pyx":236
 *             # PyTuple_SET_ITEM will steal a reference, so INCREF first
 *             Py_INCREF(parent_node)
 *             PyTuple_SET_ITEM(parent_nodes, pos, parent_node)             # <<<<<<<<<<<<<<
 *             PyList_Append(parent_node.children, node)
 *         node.parents = parent_nodes
*/
    PyTuple_SET_ITEM(__pyx_v_parent_nodes, __pyx_v_pos, ((PyObject *)__pyx_v_parent_node));

    /* "breezy/_known_graph_pyx.pyx":237
 *             Py_INCREF(parent_node)
 *             PyTuple_SET_ITEM(parent_nodes, pos, parent_node)
 *             PyList_Append(parent_node.children, node)             # <<<<<<<<<<<<<<
 *         node.parents = parent_nodes
 * 
*/
    __pyx_t_3 = __pyx_v_parent_node->children;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = PyList_Append(__pyx_t_3, ((PyObject *)__pyx_v_node)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 237, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "breezy/_known_graph_pyx.pyx":238
 *             PyTuple_SET_ITEM(parent_nodes, pos, parent_node)
 *             PyList_Append(parent_node.children, node)
 *         node.parents = parent_nodes             # <<<<<<<<<<<<<<
 * 
 *     def _initialize_nodes(self, parent_map):
*/
  __Pyx_INCREF(__pyx_v_parent_nodes);
  __Pyx_GIVEREF(__pyx_v_parent_nodes);
  __Pyx_GOTREF(__pyx_v_node->parents);
  __Pyx_DECREF(__pyx_v_node->parents);
  __pyx_v_node->parents = __pyx_v_parent_nodes;

  /* "breezy/_known_graph_pyx.pyx":218
 *         return node
 * 
 *     cdef _populate_parents(self, _KnownGraphNode node, parent_keys):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t num_parent_keys, pos
 *         cdef _KnownGraphNode parent_node
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph._populate_parents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_parent_node);
  __Pyx_XDECREF(__pyx_v_parent_nodes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":240
 *         node.parents = parent_nodes
 * 
 *     def _initialize_nodes(self, parent_map):             # <<<<<<<<<<<<<<
 *         """Populate self._nodes.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_5_initialize_nodes(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_4_initialize_nodes, "Populate self._nodes.\n\n        After this has finished:\n        - self._nodes will have an entry for every entry in parent_map.\n        - ghosts will have a parent_keys = None,\n        - all nodes found will also have child_keys populated with all known\n          child keys,\n        ");
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_5_initialize_nodes = {"_initialize_nodes", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_5_initialize_nodes, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_4_initialize_nodes};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_5_initialize_nodes(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_parent_map = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_initialize_nodes (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_parent_map,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 240, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 240, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_initialize_nodes", 0) < (0)) __PYX_ERR(0, 240, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_initialize_nodes", 1, 1, 1, i); __PYX_ERR(0, 240, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 240, __pyx_L3_error)
    }
    __pyx_v_parent_map = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_initialize_nodes", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 240, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph._initialize_nodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_4_initialize_nodes(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self), __pyx_v_parent_map);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_4_initialize_nodes(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_parent_map) {
  PyObject *__pyx_v_temp_key;
  PyObject *__pyx_v_temp_parent_keys;
  Py_ssize_t __pyx_v_pos;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node = 0;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_parent_keys = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_initialize_nodes", 0);

  /* "breezy/_known_graph_pyx.pyx":256
 *         cdef _KnownGraphNode parent_node
 * 
 *         if not PyDict_CheckExact(parent_map):             # <<<<<<<<<<<<<<
 *             raise TypeError('parent_map should be a dict of {key:parent_keys}')
 *         # for key, parent_keys in parent_map.iteritems():
*/
  __pyx_t_1 = (!PyDict_CheckExact(__pyx_v_parent_map));
  if (unlikely(__pyx_t_1)) {

    /* "breezy/_known_graph_pyx.pyx":257
 * 
 *         if not PyDict_CheckExact(parent_map):
 *             raise TypeError('parent_map should be a dict of {key:parent_keys}')             # <<<<<<<<<<<<<<
 *         # for key, parent_keys in parent_map.iteritems():
 *         pos = 0
*/
    __pyx_t_3 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_parent_map_should_be_a_dict_of_k};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 257, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 257, __pyx_L1_error)

    /* "breezy/_known_graph_pyx.pyx":256
 *         cdef _KnownGraphNode parent_node
 * 
 *         if not PyDict_CheckExact(parent_map):             # <<<<<<<<<<<<<<
 *             raise TypeError('parent_map should be a dict of {key:parent_keys}')
 *         # for key, parent_keys in parent_map.iteritems():
*/
  }

  /* "breezy/_known_graph_pyx.pyx":259
 *             raise TypeError('parent_map should be a dict of {key:parent_keys}')
 *         # for key, parent_keys in parent_map.iteritems():
 *         pos = 0             # <<<<<<<<<<<<<<
 *         while PyDict_Next(parent_map, &pos, &temp_key, &temp_parent_keys):
 *             key = <object>temp_key
*/
  __pyx_v_pos = 0;

  /* "breezy/_known_graph_pyx.pyx":260
 *         # for key, parent_keys in parent_map.iteritems():
 *         pos = 0
 *         while PyDict_Next(parent_map, &pos, &temp_key, &temp_parent_keys):             # <<<<<<<<<<<<<<
 *             key = <object>temp_key
 *             parent_keys = <object>temp_parent_keys
*/
  while (1) {
    __pyx_t_1 = (PyDict_Next(__pyx_v_parent_map, (&__pyx_v_pos), (&__pyx_v_temp_key), (&__pyx_v_temp_parent_keys)) != 0);
    if (!__pyx_t_1) break;

    /* "breezy/_known_graph_pyx.pyx":261
 *         pos = 0
 *         while PyDict_Next(parent_map, &pos, &temp_key, &temp_parent_keys):
 *             key = <object>temp_key             # <<<<<<<<<<<<<<
 *             parent_keys = <object>temp_parent_keys
 *             node = self._get_or_create_node(key)
*/
    __pyx_t_2 = ((PyObject *)__pyx_v_temp_key);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "breezy/_known_graph_pyx.pyx":262
 *         while PyDict_Next(parent_map, &pos, &temp_key, &temp_parent_keys):
 *             key = <object>temp_key
 *             parent_keys = <object>temp_parent_keys             # <<<<<<<<<<<<<<
 *             node = self._get_or_create_node(key)
 *             self._populate_parents(node, parent_keys)
*/
    __pyx_t_2 = ((PyObject *)__pyx_v_temp_parent_keys);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_parent_keys, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "breezy/_known_graph_pyx.pyx":263
 *             key = <object>temp_key
 *             parent_keys = <object>temp_parent_keys
 *             node = self._get_or_create_node(key)             # <<<<<<<<<<<<<<
 *             self._populate_parents(node, parent_keys)
 * 
*/
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self->__pyx_vtab)->_get_or_create_node(__pyx_v_self, __pyx_v_key)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "breezy/_known_graph_pyx.pyx":264
 *             parent_keys = <object>temp_parent_keys
 *             node = self._get_or_create_node(key)
 *             self._populate_parents(node, parent_keys)             # <<<<<<<<<<<<<<
 * 
 *     def _find_tails(self):
*/
    __pyx_t_2 = ((struct __pyx_vtabstruct_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self->__pyx_vtab)->_populate_parents(__pyx_v_self, __pyx_v_node, __pyx_v_parent_keys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "breezy/_known_graph_pyx.pyx":240
 *         node.parents = parent_nodes
 * 
 *     def _initialize_nodes(self, parent_map):             # <<<<<<<<<<<<<<
 *         """Populate self._nodes.
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph._initialize_nodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_parent_keys);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":266
 *             self._populate_parents(node, parent_keys)
 * 
 *     def _find_tails(self):             # <<<<<<<<<<<<<<
 *         cdef PyObject *temp_node
 *         cdef _KnownGraphNode node
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_7_find_tails(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_7_find_tails = {"_find_tails", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_7_find_tails, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_7_find_tails(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_find_tails (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_find_tails", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_find_tails", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_6_find_tails(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_6_find_tails(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self) {
  PyObject *__pyx_v_temp_node;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node = 0;
  Py_ssize_t __pyx_v_pos;
  PyObject *__pyx_v_tails = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_find_tails", 0);

  /* "breezy/_known_graph_pyx.pyx":271
 *         cdef Py_ssize_t pos
 * 
 *         tails = []             # <<<<<<<<<<<<<<
 *         pos = 0
 *         while PyDict_Next(self._nodes, &pos, NULL, &temp_node):
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_tails = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":272
 * 
 *         tails = []
 *         pos = 0             # <<<<<<<<<<<<<<
 *         while PyDict_Next(self._nodes, &pos, NULL, &temp_node):
 *             node = <_KnownGraphNode>temp_node
*/
  __pyx_v_pos = 0;

  /* "breezy/_known_graph_pyx.pyx":273
 *         tails = []
 *         pos = 0
 *         while PyDict_Next(self._nodes, &pos, NULL, &temp_node):             # <<<<<<<<<<<<<<
 *             node = <_KnownGraphNode>temp_node
 *             if node.parents is None or PyTuple_GET_SIZE(node.parents) == 0:
*/
  while (1) {
    __pyx_t_1 = __pyx_v_self->_nodes;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = (PyDict_Next(__pyx_t_1, (&__pyx_v_pos), NULL, (&__pyx_v_temp_node)) != 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_2) break;

    /* "breezy/_known_graph_pyx.pyx":274
 *         pos = 0
 *         while PyDict_Next(self._nodes, &pos, NULL, &temp_node):
 *             node = <_KnownGraphNode>temp_node             # <<<<<<<<<<<<<<
 *             if node.parents is None or PyTuple_GET_SIZE(node.parents) == 0:
 *                 node.gdfo = 1
*/
    __pyx_t_1 = ((PyObject *)__pyx_v_temp_node);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":275
 *         while PyDict_Next(self._nodes, &pos, NULL, &temp_node):
 *             node = <_KnownGraphNode>temp_node
 *             if node.parents is None or PyTuple_GET_SIZE(node.parents) == 0:             # <<<<<<<<<<<<<<
 *                 node.gdfo = 1
 *                 PyList_Append(tails, node)
*/
    __pyx_t_3 = (__pyx_v_node->parents == Py_None);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = __pyx_v_node->parents;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = (PyTuple_GET_SIZE(__pyx_t_1) == 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __pyx_t_3;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_2) {

      /* "breezy/_known_graph_pyx.pyx":276
 *             node = <_KnownGraphNode>temp_node
 *             if node.parents is None or PyTuple_GET_SIZE(node.parents) == 0:
 *                 node.gdfo = 1             # <<<<<<<<<<<<<<
 *                 PyList_Append(tails, node)
 *         return tails
*/
      __pyx_v_node->gdfo = 1;

      /* "breezy/_known_graph_pyx.pyx":277
 *             if node.parents is None or PyTuple_GET_SIZE(node.parents) == 0:
 *                 node.gdfo = 1
 *                 PyList_Append(tails, node)             # <<<<<<<<<<<<<<
 *         return tails
 * 
*/
      __pyx_t_4 = PyList_Append(__pyx_v_tails, ((PyObject *)__pyx_v_node)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 277, __pyx_L1_error)

      /* "breezy/_known_graph_pyx.pyx":275
 *         while PyDict_Next(self._nodes, &pos, NULL, &temp_node):
 *             node = <_KnownGraphNode>temp_node
 *             if node.parents is None or PyTuple_GET_SIZE(node.parents) == 0:             # <<<<<<<<<<<<<<
 *                 node.gdfo = 1
 *                 PyList_Append(tails, node)
*/
    }
  }

  /* "breezy/_known_graph_pyx.pyx":278
 *                 node.gdfo = 1
 *                 PyList_Append(tails, node)
 *         return tails             # <<<<<<<<<<<<<<
 * 
 *     def _find_tips(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_tails);
  __pyx_r = __pyx_v_tails;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":266
 *             self._populate_parents(node, parent_keys)
 * 
 *     def _find_tails(self):             # <<<<<<<<<<<<<<
 *         cdef PyObject *temp_node
 *         cdef _KnownGraphNode node
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph._find_tails", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF(__pyx_v_tails);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":280
 *         return tails
 * 
 *     def _find_tips(self):             # <<<<<<<<<<<<<<
 *         cdef PyObject *temp_node
 *         cdef _KnownGraphNode node
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_9_find_tips(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_9_find_tips = {"_find_tips", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_9_find_tips, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_9_find_tips(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_find_tips (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_find_tips", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_find_tips", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_8_find_tips(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_8_find_tips(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self) {
  PyObject *__pyx_v_temp_node;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node = 0;
  Py_ssize_t __pyx_v_pos;
  PyObject *__pyx_v_tips = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_find_tips", 0);

  /* "breezy/_known_graph_pyx.pyx":285
 *         cdef Py_ssize_t pos
 * 
 *         tips = []             # <<<<<<<<<<<<<<
 *         pos = 0
 *         while PyDict_Next(self._nodes, &pos, NULL, &temp_node):
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_tips = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":286
 * 
 *         tips = []
 *         pos = 0             # <<<<<<<<<<<<<<
 *         while PyDict_Next(self._nodes, &pos, NULL, &temp_node):
 *             node = <_KnownGraphNode>temp_node
*/
  __pyx_v_pos = 0;

  /* "breezy/_known_graph_pyx.pyx":287
 *         tips = []
 *         pos = 0
 *         while PyDict_Next(self._nodes, &pos, NULL, &temp_node):             # <<<<<<<<<<<<<<
 *             node = <_KnownGraphNode>temp_node
 *             if PyList_GET_SIZE(node.children) == 0:
*/
  while (1) {
    __pyx_t_1 = __pyx_v_self->_nodes;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = (PyDict_Next(__pyx_t_1, (&__pyx_v_pos), NULL, (&__pyx_v_temp_node)) != 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_2) break;

    /* "breezy/_known_graph_pyx.pyx":288
 *         pos = 0
 *         while PyDict_Next(self._nodes, &pos, NULL, &temp_node):
 *             node = <_KnownGraphNode>temp_node             # <<<<<<<<<<<<<<
 *             if PyList_GET_SIZE(node.children) == 0:
 *                 PyList_Append(tips, node)
*/
    __pyx_t_1 = ((PyObject *)__pyx_v_temp_node);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":289
 *         while PyDict_Next(self._nodes, &pos, NULL, &temp_node):
 *             node = <_KnownGraphNode>temp_node
 *             if PyList_GET_SIZE(node.children) == 0:             # <<<<<<<<<<<<<<
 *                 PyList_Append(tips, node)
 *         return tips
*/
    __pyx_t_1 = __pyx_v_node->children;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = (PyList_GET_SIZE(__pyx_t_1) == 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_2) {

      /* "breezy/_known_graph_pyx.pyx":290
 *             node = <_KnownGraphNode>temp_node
 *             if PyList_GET_SIZE(node.children) == 0:
 *                 PyList_Append(tips, node)             # <<<<<<<<<<<<<<
 *         return tips
 * 
*/
      __pyx_t_3 = PyList_Append(__pyx_v_tips, ((PyObject *)__pyx_v_node)); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 290, __pyx_L1_error)

      /* "breezy/_known_graph_pyx.pyx":289
 *         while PyDict_Next(self._nodes, &pos, NULL, &temp_node):
 *             node = <_KnownGraphNode>temp_node
 *             if PyList_GET_SIZE(node.children) == 0:             # <<<<<<<<<<<<<<
 *                 PyList_Append(tips, node)
 *         return tips
*/
    }
  }

  /* "breezy/_known_graph_pyx.pyx":291
 *             if PyList_GET_SIZE(node.children) == 0:
 *                 PyList_Append(tips, node)
 *         return tips             # <<<<<<<<<<<<<<
 * 
 *     def _find_gdfo(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_tips);
  __pyx_r = __pyx_v_tips;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":280
 *         return tails
 * 
 *     def _find_tips(self):             # <<<<<<<<<<<<<<
 *         cdef PyObject *temp_node
 *         cdef _KnownGraphNode node
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph._find_tips", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF(__pyx_v_tips);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":293
 *         return tips
 * 
 *     def _find_gdfo(self):             # <<<<<<<<<<<<<<
 *         cdef _KnownGraphNode node
 *         cdef _KnownGraphNode child
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_11_find_gdfo(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_11_find_gdfo = {"_find_gdfo", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_11_find_gdfo, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_11_find_gdfo(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_find_gdfo (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_find_gdfo", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_find_gdfo", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_10_find_gdfo(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_10_find_gdfo(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node = 0;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_child = 0;
  Py_ssize_t __pyx_v_pos;
  Py_ssize_t __pyx_v_last_item;
  long __pyx_v_next_gdfo;
  PyObject *__pyx_v_pending = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_find_gdfo", 0);

  /* "breezy/_known_graph_pyx.pyx":302
 *         cdef long next_gdfo
 * 
 *         pending = self._find_tails()             # <<<<<<<<<<<<<<
 * 
 *         last_item = PyList_GET_SIZE(pending) - 1
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_find_tails, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_pending = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":304
 *         pending = self._find_tails()
 * 
 *         last_item = PyList_GET_SIZE(pending) - 1             # <<<<<<<<<<<<<<
 *         while last_item >= 0:
 *             # Avoid pop followed by push, instead, peek, and replace
*/
  __pyx_v_last_item = (PyList_GET_SIZE(__pyx_v_pending) - 1);

  /* "breezy/_known_graph_pyx.pyx":305
 * 
 *         last_item = PyList_GET_SIZE(pending) - 1
 *         while last_item >= 0:             # <<<<<<<<<<<<<<
 *             # Avoid pop followed by push, instead, peek, and replace
 *             # timing shows this is 930ms => 770ms for OOo
*/
  while (1) {
    __pyx_t_4 = (__pyx_v_last_item >= 0);
    if (!__pyx_t_4) break;

    /* "breezy/_known_graph_pyx.pyx":308
 *             # Avoid pop followed by push, instead, peek, and replace
 *             # timing shows this is 930ms => 770ms for OOo
 *             node = _get_list_node(pending, last_item)             # <<<<<<<<<<<<<<
 *             last_item = last_item - 1
 *             next_gdfo = node.gdfo + 1
*/
    __pyx_t_1 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_list_node(__pyx_v_pending, __pyx_v_last_item)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":309
 *             # timing shows this is 930ms => 770ms for OOo
 *             node = _get_list_node(pending, last_item)
 *             last_item = last_item - 1             # <<<<<<<<<<<<<<
 *             next_gdfo = node.gdfo + 1
 *             for pos from 0 <= pos < PyList_GET_SIZE(node.children):
*/
    __pyx_v_last_item = (__pyx_v_last_item - 1);

    /* "breezy/_known_graph_pyx.pyx":310
 *             node = _get_list_node(pending, last_item)
 *             last_item = last_item - 1
 *             next_gdfo = node.gdfo + 1             # <<<<<<<<<<<<<<
 *             for pos from 0 <= pos < PyList_GET_SIZE(node.children):
 *                 child = _get_list_node(node.children, pos)
*/
    __pyx_v_next_gdfo = (__pyx_v_node->gdfo + 1);

    /* "breezy/_known_graph_pyx.pyx":311
 *             last_item = last_item - 1
 *             next_gdfo = node.gdfo + 1
 *             for pos from 0 <= pos < PyList_GET_SIZE(node.children):             # <<<<<<<<<<<<<<
 *                 child = _get_list_node(node.children, pos)
 *                 if next_gdfo > child.gdfo:
*/
    __pyx_t_1 = __pyx_v_node->children;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = PyList_GET_SIZE(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (__pyx_v_pos = 0; __pyx_v_pos < __pyx_t_5; __pyx_v_pos++) {

      /* "breezy/_known_graph_pyx.pyx":312
 *             next_gdfo = node.gdfo + 1
 *             for pos from 0 <= pos < PyList_GET_SIZE(node.children):
 *                 child = _get_list_node(node.children, pos)             # <<<<<<<<<<<<<<
 *                 if next_gdfo > child.gdfo:
 *                     child.gdfo = next_gdfo
*/
      __pyx_t_1 = __pyx_v_node->children;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_2 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_list_node(__pyx_t_1, __pyx_v_pos)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_child, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "breezy/_known_graph_pyx.pyx":313
 *             for pos from 0 <= pos < PyList_GET_SIZE(node.children):
 *                 child = _get_list_node(node.children, pos)
 *                 if next_gdfo > child.gdfo:             # <<<<<<<<<<<<<<
 *                     child.gdfo = next_gdfo
 *                 child.seen = child.seen + 1
*/
      __pyx_t_4 = (__pyx_v_next_gdfo > __pyx_v_child->gdfo);
      if (__pyx_t_4) {

        /* "breezy/_known_graph_pyx.pyx":314
 *                 child = _get_list_node(node.children, pos)
 *                 if next_gdfo > child.gdfo:
 *                     child.gdfo = next_gdfo             # <<<<<<<<<<<<<<
 *                 child.seen = child.seen + 1
 *                 if child.seen == PyTuple_GET_SIZE(child.parents):
*/
        __pyx_v_child->gdfo = __pyx_v_next_gdfo;

        /* "breezy/_known_graph_pyx.pyx":313
 *             for pos from 0 <= pos < PyList_GET_SIZE(node.children):
 *                 child = _get_list_node(node.children, pos)
 *                 if next_gdfo > child.gdfo:             # <<<<<<<<<<<<<<
 *                     child.gdfo = next_gdfo
 *                 child.seen = child.seen + 1
*/
      }

      /* "breezy/_known_graph_pyx.pyx":315
 *                 if next_gdfo > child.gdfo:
 *                     child.gdfo = next_gdfo
 *                 child.seen = child.seen + 1             # <<<<<<<<<<<<<<
 *                 if child.seen == PyTuple_GET_SIZE(child.parents):
 *                     # This child is populated, queue it to be walked
*/
      __pyx_v_child->seen = (__pyx_v_child->seen + 1);

      /* "breezy/_known_graph_pyx.pyx":316
 *                     child.gdfo = next_gdfo
 *                 child.seen = child.seen + 1
 *                 if child.seen == PyTuple_GET_SIZE(child.parents):             # <<<<<<<<<<<<<<
 *                     # This child is populated, queue it to be walked
 *                     last_item = last_item + 1
*/
      __pyx_t_2 = __pyx_v_child->parents;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_4 = (__pyx_v_child->seen == PyTuple_GET_SIZE(__pyx_t_2));
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_4) {

        /* "breezy/_known_graph_pyx.pyx":318
 *                 if child.seen == PyTuple_GET_SIZE(child.parents):
 *                     # This child is populated, queue it to be walked
 *                     last_item = last_item + 1             # <<<<<<<<<<<<<<
 *                     if last_item < PyList_GET_SIZE(pending):
 *                         Py_INCREF(child) # SetItem steals a ref
*/
        __pyx_v_last_item = (__pyx_v_last_item + 1);

        /* "breezy/_known_graph_pyx.pyx":319
 *                     # This child is populated, queue it to be walked
 *                     last_item = last_item + 1
 *                     if last_item < PyList_GET_SIZE(pending):             # <<<<<<<<<<<<<<
 *                         Py_INCREF(child) # SetItem steals a ref
 *                         PyList_SetItem(pending, last_item, child)
*/
        __pyx_t_4 = (__pyx_v_last_item < PyList_GET_SIZE(__pyx_v_pending));
        if (__pyx_t_4) {

          /* "breezy/_known_graph_pyx.pyx":320
 *                     last_item = last_item + 1
 *                     if last_item < PyList_GET_SIZE(pending):
 *                         Py_INCREF(child) # SetItem steals a ref             # <<<<<<<<<<<<<<
 *                         PyList_SetItem(pending, last_item, child)
 *                     else:
*/
          Py_INCREF(((PyObject *)__pyx_v_child));

          /* "breezy/_known_graph_pyx.pyx":321
 *                     if last_item < PyList_GET_SIZE(pending):
 *                         Py_INCREF(child) # SetItem steals a ref
 *                         PyList_SetItem(pending, last_item, child)             # <<<<<<<<<<<<<<
 *                     else:
 *                         PyList_Append(pending, child)
*/
          __pyx_t_6 = PyList_SetItem(__pyx_v_pending, __pyx_v_last_item, ((PyObject *)__pyx_v_child)); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 321, __pyx_L1_error)

          /* "breezy/_known_graph_pyx.pyx":319
 *                     # This child is populated, queue it to be walked
 *                     last_item = last_item + 1
 *                     if last_item < PyList_GET_SIZE(pending):             # <<<<<<<<<<<<<<
 *                         Py_INCREF(child) # SetItem steals a ref
 *                         PyList_SetItem(pending, last_item, child)
*/
          goto __pyx_L9;
        }

        /* "breezy/_known_graph_pyx.pyx":323
 *                         PyList_SetItem(pending, last_item, child)
 *                     else:
 *                         PyList_Append(pending, child)             # <<<<<<<<<<<<<<
 *                     # We have queued this node, we don't need to track it
 *                     # anymore
*/
        /*else*/ {
          __pyx_t_6 = PyList_Append(__pyx_v_pending, ((PyObject *)__pyx_v_child)); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 323, __pyx_L1_error)
        }
        __pyx_L9:;

        /* "breezy/_known_graph_pyx.pyx":326
 *                     # We have queued this node, we don't need to track it
 *                     # anymore
 *                     child.seen = 0             # <<<<<<<<<<<<<<
 * 
 *     def add_node(self, key, parent_keys):
*/
        __pyx_v_child->seen = 0;

        /* "breezy/_known_graph_pyx.pyx":316
 *                     child.gdfo = next_gdfo
 *                 child.seen = child.seen + 1
 *                 if child.seen == PyTuple_GET_SIZE(child.parents):             # <<<<<<<<<<<<<<
 *                     # This child is populated, queue it to be walked
 *                     last_item = last_item + 1
*/
      }
    }
  }

  /* "breezy/_known_graph_pyx.pyx":293
 *         return tips
 * 
 *     def _find_gdfo(self):             # <<<<<<<<<<<<<<
 *         cdef _KnownGraphNode node
 *         cdef _KnownGraphNode child
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph._find_gdfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_pending);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":328
 *                     child.seen = 0
 * 
 *     def add_node(self, key, parent_keys):             # <<<<<<<<<<<<<<
 *         """Add a new node to the graph.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_13add_node(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_12add_node, "Add a new node to the graph.\n\n        If this fills in a ghost, then the gdfos of all children will be\n        updated accordingly.\n\n        :param key: The node being added. If this is a duplicate, this is a\n            no-op.\n        :param parent_keys: The parents of the given node.\n        :return: None (should we return if this was a ghost, etc?)\n        ");
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_13add_node = {"add_node", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_13add_node, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_12add_node};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_13add_node(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_parent_keys = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_node (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,&__pyx_mstate_global->__pyx_n_u_parent_keys,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 328, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 328, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 328, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "add_node", 0) < (0)) __PYX_ERR(0, 328, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("add_node", 1, 2, 2, i); __PYX_ERR(0, 328, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 328, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 328, __pyx_L3_error)
    }
    __pyx_v_key = values[0];
    __pyx_v_parent_keys = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_node", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 328, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.add_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_12add_node(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self), __pyx_v_key, __pyx_v_parent_keys);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_12add_node(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_parent_keys) {
  PyObject *__pyx_v_maybe_node;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node = 0;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_parent_node = 0;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_child_node = 0;
  long __pyx_v_parent_gdfo;
  long __pyx_v_next_gdfo;
  PyObject *__pyx_v_existing_parent_keys = NULL;
  PyObject *__pyx_v_pending = NULL;
  PyObject *__pyx_v_pending_popleft = NULL;
  PyObject *__pyx_v_pending_append = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11[6];
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  long __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_node", 0);
  __Pyx_INCREF(__pyx_v_parent_keys);

  /* "breezy/_known_graph_pyx.pyx":343
 *         cdef long parent_gdfo, next_gdfo
 * 
 *         maybe_node = PyDict_GetItem(self._nodes, key)             # <<<<<<<<<<<<<<
 *         if maybe_node != NULL:
 *             node = <_KnownGraphNode>maybe_node
*/
  __pyx_t_1 = __pyx_v_self->_nodes;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_maybe_node = PyDict_GetItem(__pyx_t_1, __pyx_v_key);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":344
 * 
 *         maybe_node = PyDict_GetItem(self._nodes, key)
 *         if maybe_node != NULL:             # <<<<<<<<<<<<<<
 *             node = <_KnownGraphNode>maybe_node
 *             if node.parents is None:
*/
  __pyx_t_2 = (__pyx_v_maybe_node != NULL);
  if (__pyx_t_2) {

    /* "breezy/_known_graph_pyx.pyx":345
 *         maybe_node = PyDict_GetItem(self._nodes, key)
 *         if maybe_node != NULL:
 *             node = <_KnownGraphNode>maybe_node             # <<<<<<<<<<<<<<
 *             if node.parents is None:
 *                 # We are filling in a ghost
*/
    __pyx_t_1 = ((PyObject *)__pyx_v_maybe_node);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_node = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":346
 *         if maybe_node != NULL:
 *             node = <_KnownGraphNode>maybe_node
 *             if node.parents is None:             # <<<<<<<<<<<<<<
 *                 # We are filling in a ghost
 *                 self._populate_parents(node, parent_keys)
*/
    __pyx_t_2 = (__pyx_v_node->parents == Py_None);
    if (__pyx_t_2) {

      /* "breezy/_known_graph_pyx.pyx":348
 *             if node.parents is None:
 *                 # We are filling in a ghost
 *                 self._populate_parents(node, parent_keys)             # <<<<<<<<<<<<<<
 *                 # We can't trust cached heads anymore
 *                 self._known_heads.clear()
*/
      __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self->__pyx_vtab)->_populate_parents(__pyx_v_self, __pyx_v_node, __pyx_v_parent_keys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "breezy/_known_graph_pyx.pyx":350
 *                 self._populate_parents(node, parent_keys)
 *                 # We can't trust cached heads anymore
 *                 self._known_heads.clear()             # <<<<<<<<<<<<<<
 *             else: # Ensure that the parent_key list matches
 *                 existing_parent_keys = []
*/
      __pyx_t_3 = __pyx_v_self->_known_heads;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
        __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_clear, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 350, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "breezy/_known_graph_pyx.pyx":346
 *         if maybe_node != NULL:
 *             node = <_KnownGraphNode>maybe_node
 *             if node.parents is None:             # <<<<<<<<<<<<<<
 *                 # We are filling in a ghost
 *                 self._populate_parents(node, parent_keys)
*/
      goto __pyx_L4;
    }

    /* "breezy/_known_graph_pyx.pyx":352
 *                 self._known_heads.clear()
 *             else: # Ensure that the parent_key list matches
 *                 existing_parent_keys = []             # <<<<<<<<<<<<<<
 *                 for parent_node in node.parents:
 *                     existing_parent_keys.append(parent_node.key)
*/
    /*else*/ {
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_existing_parent_keys = ((PyObject*)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "breezy/_known_graph_pyx.pyx":353
 *             else: # Ensure that the parent_key list matches
 *                 existing_parent_keys = []
 *                 for parent_node in node.parents:             # <<<<<<<<<<<<<<
 *                     existing_parent_keys.append(parent_node.key)
 *                 # Make sure we use a list for the comparison, in case it was a
*/
      if (likely(PyList_CheckExact(__pyx_v_node->parents)) || PyTuple_CheckExact(__pyx_v_node->parents)) {
        __pyx_t_1 = __pyx_v_node->parents; __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = 0;
        __pyx_t_6 = NULL;
      } else {
        __pyx_t_5 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_node->parents); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 353, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_6 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 353, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_6)) {
          if (likely(PyList_CheckExact(__pyx_t_1))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
              #if !CYTHON_ASSUME_SAFE_SIZE
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 353, __pyx_L1_error)
              #endif
              if (__pyx_t_5 >= __pyx_temp) break;
            }
            __pyx_t_3 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_5, __Pyx_ReferenceSharing_OwnStrongReference);
            ++__pyx_t_5;
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
              #if !CYTHON_ASSUME_SAFE_SIZE
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 353, __pyx_L1_error)
              #endif
              if (__pyx_t_5 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5));
            #else
            __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_5);
            #endif
            ++__pyx_t_5;
          }
          if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 353, __pyx_L1_error)
        } else {
          __pyx_t_3 = __pyx_t_6(__pyx_t_1);
          if (unlikely(!__pyx_t_3)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 353, __pyx_L1_error)
              PyErr_Clear();
            }
            break;
          }
        }
        __Pyx_GOTREF(__pyx_t_3);
        if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode))))) __PYX_ERR(0, 353, __pyx_L1_error)
        __Pyx_XDECREF_SET(__pyx_v_parent_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_3));
        __pyx_t_3 = 0;

        /* "breezy/_known_graph_pyx.pyx":354
 *                 existing_parent_keys = []
 *                 for parent_node in node.parents:
 *                     existing_parent_keys.append(parent_node.key)             # <<<<<<<<<<<<<<
 *                 # Make sure we use a list for the comparison, in case it was a
 *                 # tuple, etc
*/
        __pyx_t_3 = __pyx_v_parent_node->key;
        __Pyx_INCREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_existing_parent_keys, __pyx_t_3); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 354, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "breezy/_known_graph_pyx.pyx":353
 *             else: # Ensure that the parent_key list matches
 *                 existing_parent_keys = []
 *                 for parent_node in node.parents:             # <<<<<<<<<<<<<<
 *                     existing_parent_keys.append(parent_node.key)
 *                 # Make sure we use a list for the comparison, in case it was a
*/
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "breezy/_known_graph_pyx.pyx":357
 *                 # Make sure we use a list for the comparison, in case it was a
 *                 # tuple, etc
 *                 parent_keys = list(parent_keys)             # <<<<<<<<<<<<<<
 *                 if existing_parent_keys == parent_keys:
 *                     # Exact match, nothing more to do
*/
      __pyx_t_1 = PySequence_List(__pyx_v_parent_keys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 357, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_parent_keys, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "breezy/_known_graph_pyx.pyx":358
 *                 # tuple, etc
 *                 parent_keys = list(parent_keys)
 *                 if existing_parent_keys == parent_keys:             # <<<<<<<<<<<<<<
 *                     # Exact match, nothing more to do
 *                     return
*/
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_existing_parent_keys, __pyx_v_parent_keys, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 358, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (likely(__pyx_t_2)) {

        /* "breezy/_known_graph_pyx.pyx":360
 *                 if existing_parent_keys == parent_keys:
 *                     # Exact match, nothing more to do
 *                     return             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError('Parent key mismatch, existing node %s'
*/
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        goto __pyx_L0;

        /* "breezy/_known_graph_pyx.pyx":358
 *                 # tuple, etc
 *                 parent_keys = list(parent_keys)
 *                 if existing_parent_keys == parent_keys:             # <<<<<<<<<<<<<<
 *                     # Exact match, nothing more to do
 *                     return
*/
      }

      /* "breezy/_known_graph_pyx.pyx":362
 *                     return
 *                 else:
 *                     raise ValueError('Parent key mismatch, existing node %s'             # <<<<<<<<<<<<<<
 *                         ' has parents of %s not %s'
 *                         % (key, existing_parent_keys, parent_keys))
*/
      /*else*/ {
        __pyx_t_3 = NULL;

        /* "breezy/_known_graph_pyx.pyx":364
 *                     raise ValueError('Parent key mismatch, existing node %s'
 *                         ' has parents of %s not %s'
 *                         % (key, existing_parent_keys, parent_keys))             # <<<<<<<<<<<<<<
 *         else:
 *             node = _KnownGraphNode(key)
*/
        __pyx_t_8 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_v_key), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 364, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_v_existing_parent_keys), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 364, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_v_parent_keys), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 364, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11[0] = __pyx_mstate_global->__pyx_kp_u_Parent_key_mismatch_existing_nod;
        __pyx_t_11[1] = __pyx_t_8;
        __pyx_t_11[2] = __pyx_mstate_global->__pyx_kp_u_has_parents_of;
        __pyx_t_11[3] = __pyx_t_9;
        __pyx_t_11[4] = __pyx_mstate_global->__pyx_kp_u_not;
        __pyx_t_11[5] = __pyx_t_10;

        /* "breezy/_known_graph_pyx.pyx":362
 *                     return
 *                 else:
 *                     raise ValueError('Parent key mismatch, existing node %s'             # <<<<<<<<<<<<<<
 *                         ' has parents of %s not %s'
 *                         % (key, existing_parent_keys, parent_keys))
*/
        __pyx_t_12 = __Pyx_PyUnicode_Join(__pyx_t_11, 6, 35 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8) + 16 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9) + 5 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10));
        if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 362, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_4 = 1;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_12};
          __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 362, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
        }
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(0, 362, __pyx_L1_error)
      }
    }
    __pyx_L4:;

    /* "breezy/_known_graph_pyx.pyx":344
 * 
 *         maybe_node = PyDict_GetItem(self._nodes, key)
 *         if maybe_node != NULL:             # <<<<<<<<<<<<<<
 *             node = <_KnownGraphNode>maybe_node
 *             if node.parents is None:
*/
    goto __pyx_L3;
  }

  /* "breezy/_known_graph_pyx.pyx":366
 *                         % (key, existing_parent_keys, parent_keys))
 *         else:
 *             node = _KnownGraphNode(key)             # <<<<<<<<<<<<<<
 *             PyDict_SetItem(self._nodes, key, node)
 *             self._populate_parents(node, parent_keys)
*/
  /*else*/ {
    __pyx_t_12 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_v_key};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 366, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_1);
    }
    __pyx_v_node = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":367
 *         else:
 *             node = _KnownGraphNode(key)
 *             PyDict_SetItem(self._nodes, key, node)             # <<<<<<<<<<<<<<
 *             self._populate_parents(node, parent_keys)
 *         parent_gdfo = 0
*/
    __pyx_t_1 = __pyx_v_self->_nodes;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_13 = PyDict_SetItem(__pyx_t_1, __pyx_v_key, ((PyObject *)__pyx_v_node)); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 367, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":368
 *             node = _KnownGraphNode(key)
 *             PyDict_SetItem(self._nodes, key, node)
 *             self._populate_parents(node, parent_keys)             # <<<<<<<<<<<<<<
 *         parent_gdfo = 0
 *         for parent_node in node.parents:
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self->__pyx_vtab)->_populate_parents(__pyx_v_self, __pyx_v_node, __pyx_v_parent_keys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "breezy/_known_graph_pyx.pyx":369
 *             PyDict_SetItem(self._nodes, key, node)
 *             self._populate_parents(node, parent_keys)
 *         parent_gdfo = 0             # <<<<<<<<<<<<<<
 *         for parent_node in node.parents:
 *             if parent_node.gdfo == -1:
*/
  __pyx_v_parent_gdfo = 0;

  /* "breezy/_known_graph_pyx.pyx":370
 *             self._populate_parents(node, parent_keys)
 *         parent_gdfo = 0
 *         for parent_node in node.parents:             # <<<<<<<<<<<<<<
 *             if parent_node.gdfo == -1:
 *                 # This is a newly introduced ghost, so it gets gdfo of 1
*/
  if (likely(PyList_CheckExact(__pyx_v_node->parents)) || PyTuple_CheckExact(__pyx_v_node->parents)) {
    __pyx_t_1 = __pyx_v_node->parents; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_node->parents); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 370, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 370, __pyx_L1_error)
          #endif
          if (__pyx_t_5 >= __pyx_temp) break;
        }
        __pyx_t_12 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_5, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_5;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 370, __pyx_L1_error)
          #endif
          if (__pyx_t_5 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_12 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5));
        #else
        __pyx_t_12 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_5);
        #endif
        ++__pyx_t_5;
      }
      if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 370, __pyx_L1_error)
    } else {
      __pyx_t_12 = __pyx_t_6(__pyx_t_1);
      if (unlikely(!__pyx_t_12)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 370, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_12);
    if (!(likely(((__pyx_t_12) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_12, __pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode))))) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_parent_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_12));
    __pyx_t_12 = 0;

    /* "breezy/_known_graph_pyx.pyx":371
 *         parent_gdfo = 0
 *         for parent_node in node.parents:
 *             if parent_node.gdfo == -1:             # <<<<<<<<<<<<<<
 *                 # This is a newly introduced ghost, so it gets gdfo of 1
 *                 parent_node.gdfo = 1
*/
    __pyx_t_2 = (__pyx_v_parent_node->gdfo == -1L);
    if (__pyx_t_2) {

      /* "breezy/_known_graph_pyx.pyx":373
 *             if parent_node.gdfo == -1:
 *                 # This is a newly introduced ghost, so it gets gdfo of 1
 *                 parent_node.gdfo = 1             # <<<<<<<<<<<<<<
 *             if parent_gdfo < parent_node.gdfo:
 *                 parent_gdfo = parent_node.gdfo
*/
      __pyx_v_parent_node->gdfo = 1;

      /* "breezy/_known_graph_pyx.pyx":371
 *         parent_gdfo = 0
 *         for parent_node in node.parents:
 *             if parent_node.gdfo == -1:             # <<<<<<<<<<<<<<
 *                 # This is a newly introduced ghost, so it gets gdfo of 1
 *                 parent_node.gdfo = 1
*/
    }

    /* "breezy/_known_graph_pyx.pyx":374
 *                 # This is a newly introduced ghost, so it gets gdfo of 1
 *                 parent_node.gdfo = 1
 *             if parent_gdfo < parent_node.gdfo:             # <<<<<<<<<<<<<<
 *                 parent_gdfo = parent_node.gdfo
 *         node.gdfo = parent_gdfo + 1
*/
    __pyx_t_2 = (__pyx_v_parent_gdfo < __pyx_v_parent_node->gdfo);
    if (__pyx_t_2) {

      /* "breezy/_known_graph_pyx.pyx":375
 *                 parent_node.gdfo = 1
 *             if parent_gdfo < parent_node.gdfo:
 *                 parent_gdfo = parent_node.gdfo             # <<<<<<<<<<<<<<
 *         node.gdfo = parent_gdfo + 1
 *         # Now fill the gdfo to all children
*/
      __pyx_t_14 = __pyx_v_parent_node->gdfo;
      __pyx_v_parent_gdfo = __pyx_t_14;

      /* "breezy/_known_graph_pyx.pyx":374
 *                 # This is a newly introduced ghost, so it gets gdfo of 1
 *                 parent_node.gdfo = 1
 *             if parent_gdfo < parent_node.gdfo:             # <<<<<<<<<<<<<<
 *                 parent_gdfo = parent_node.gdfo
 *         node.gdfo = parent_gdfo + 1
*/
    }

    /* "breezy/_known_graph_pyx.pyx":370
 *             self._populate_parents(node, parent_keys)
 *         parent_gdfo = 0
 *         for parent_node in node.parents:             # <<<<<<<<<<<<<<
 *             if parent_node.gdfo == -1:
 *                 # This is a newly introduced ghost, so it gets gdfo of 1
*/
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":376
 *             if parent_gdfo < parent_node.gdfo:
 *                 parent_gdfo = parent_node.gdfo
 *         node.gdfo = parent_gdfo + 1             # <<<<<<<<<<<<<<
 *         # Now fill the gdfo to all children
 *         # Note that this loop is slightly inefficient, in that we may visit the
*/
  __pyx_v_node->gdfo = (__pyx_v_parent_gdfo + 1);

  /* "breezy/_known_graph_pyx.pyx":385
 *         # than DFD. So that if a longer path is possible, we walk it before we
 *         # get to the final child
 *         pending = collections.deque([node])             # <<<<<<<<<<<<<<
 *         pending_popleft = pending.popleft
 *         pending_append = pending.append
*/
  __pyx_t_12 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_collections); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_deque); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)__pyx_v_node);
  __Pyx_GIVEREF((PyObject *)__pyx_v_node);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_node)) != (0)) __PYX_ERR(0, 385, __pyx_L1_error);
  __pyx_t_4 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
    assert(__pyx_t_12);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_10);
    __Pyx_INCREF(__pyx_t_12);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_10, __pyx__function);
    __pyx_t_4 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_10, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_pending = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":386
 *         # get to the final child
 *         pending = collections.deque([node])
 *         pending_popleft = pending.popleft             # <<<<<<<<<<<<<<
 *         pending_append = pending.append
 *         while pending:
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_pending, __pyx_mstate_global->__pyx_n_u_popleft); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pending_popleft = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":387
 *         pending = collections.deque([node])
 *         pending_popleft = pending.popleft
 *         pending_append = pending.append             # <<<<<<<<<<<<<<
 *         while pending:
 *             node = pending_popleft()
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_pending, __pyx_mstate_global->__pyx_n_u_append); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pending_append = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":388
 *         pending_popleft = pending.popleft
 *         pending_append = pending.append
 *         while pending:             # <<<<<<<<<<<<<<
 *             node = pending_popleft()
 *             next_gdfo = node.gdfo + 1
*/
  while (1) {
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_pending); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 388, __pyx_L1_error)
    if (!__pyx_t_2) break;

    /* "breezy/_known_graph_pyx.pyx":389
 *         pending_append = pending.append
 *         while pending:
 *             node = pending_popleft()             # <<<<<<<<<<<<<<
 *             next_gdfo = node.gdfo + 1
 *             for child_node in node.children:
*/
    __pyx_t_10 = NULL;
    __Pyx_INCREF(__pyx_v_pending_popleft);
    __pyx_t_3 = __pyx_v_pending_popleft; 
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      assert(__pyx_t_10);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 389, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode))))) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":390
 *         while pending:
 *             node = pending_popleft()
 *             next_gdfo = node.gdfo + 1             # <<<<<<<<<<<<<<
 *             for child_node in node.children:
 *                 if child_node.gdfo < next_gdfo:
*/
    __pyx_v_next_gdfo = (__pyx_v_node->gdfo + 1);

    /* "breezy/_known_graph_pyx.pyx":391
 *             node = pending_popleft()
 *             next_gdfo = node.gdfo + 1
 *             for child_node in node.children:             # <<<<<<<<<<<<<<
 *                 if child_node.gdfo < next_gdfo:
 *                     # This child is being updated, we need to check its
*/
    if (likely(PyList_CheckExact(__pyx_v_node->children)) || PyTuple_CheckExact(__pyx_v_node->children)) {
      __pyx_t_1 = __pyx_v_node->children; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_node->children); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 391, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 391, __pyx_L1_error)
            #endif
            if (__pyx_t_5 >= __pyx_temp) break;
          }
          __pyx_t_3 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_5, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_5;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 391, __pyx_L1_error)
            #endif
            if (__pyx_t_5 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5));
          #else
          __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_5);
          #endif
          ++__pyx_t_5;
        }
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 391, __pyx_L1_error)
      } else {
        __pyx_t_3 = __pyx_t_6(__pyx_t_1);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 391, __pyx_L1_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_3);
      if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode))))) __PYX_ERR(0, 391, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_child_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "breezy/_known_graph_pyx.pyx":392
 *             next_gdfo = node.gdfo + 1
 *             for child_node in node.children:
 *                 if child_node.gdfo < next_gdfo:             # <<<<<<<<<<<<<<
 *                     # This child is being updated, we need to check its
 *                     # children
*/
      __pyx_t_2 = (__pyx_v_child_node->gdfo < __pyx_v_next_gdfo);
      if (__pyx_t_2) {

        /* "breezy/_known_graph_pyx.pyx":395
 *                     # This child is being updated, we need to check its
 *                     # children
 *                     child_node.gdfo = next_gdfo             # <<<<<<<<<<<<<<
 *                     pending_append(child_node)
 * 
*/
        __pyx_v_child_node->gdfo = __pyx_v_next_gdfo;

        /* "breezy/_known_graph_pyx.pyx":396
 *                     # children
 *                     child_node.gdfo = next_gdfo
 *                     pending_append(child_node)             # <<<<<<<<<<<<<<
 * 
 *     def heads(self, keys):
*/
        __pyx_t_10 = NULL;
        __Pyx_INCREF(__pyx_v_pending_append);
        __pyx_t_12 = __pyx_v_pending_append; 
        __pyx_t_4 = 1;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_12);
          assert(__pyx_t_10);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_12, __pyx__function);
          __pyx_t_4 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_10, ((PyObject *)__pyx_v_child_node)};
          __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_12, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 396, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "breezy/_known_graph_pyx.pyx":392
 *             next_gdfo = node.gdfo + 1
 *             for child_node in node.children:
 *                 if child_node.gdfo < next_gdfo:             # <<<<<<<<<<<<<<
 *                     # This child is being updated, we need to check its
 *                     # children
*/
      }

      /* "breezy/_known_graph_pyx.pyx":391
 *             node = pending_popleft()
 *             next_gdfo = node.gdfo + 1
 *             for child_node in node.children:             # <<<<<<<<<<<<<<
 *                 if child_node.gdfo < next_gdfo:
 *                     # This child is being updated, we need to check its
*/
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "breezy/_known_graph_pyx.pyx":328
 *                     child.seen = 0
 * 
 *     def add_node(self, key, parent_keys):             # <<<<<<<<<<<<<<
 *         """Add a new node to the graph.
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.add_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_parent_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_child_node);
  __Pyx_XDECREF(__pyx_v_existing_parent_keys);
  __Pyx_XDECREF(__pyx_v_pending);
  __Pyx_XDECREF(__pyx_v_pending_popleft);
  __Pyx_XDECREF(__pyx_v_pending_append);
  __Pyx_XDECREF(__pyx_v_parent_keys);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":398
 *                     pending_append(child_node)
 * 
 *     def heads(self, keys):             # <<<<<<<<<<<<<<
 *         """Return the heads from amongst keys.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_15heads(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_14heads, "Return the heads from amongst keys.\n\n        This is done by searching the ancestries of each key.  Any key that is\n        reachable from another key is not returned; all the others are.\n\n        This operation scales with the relative depth between any two keys. It\n        uses gdfo to avoid walking all ancestry.\n\n        :param keys: An iterable of keys.\n        :return: A set of the heads. Note that as a set there is no ordering\n            information. Callers will need to filter their input to create\n            order if they need it.\n        ");
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_15heads = {"heads", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_15heads, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_14heads};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_15heads(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_keys = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("heads (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_keys,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 398, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 398, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "heads", 0) < (0)) __PYX_ERR(0, 398, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("heads", 1, 1, 1, i); __PYX_ERR(0, 398, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 398, __pyx_L3_error)
    }
    __pyx_v_keys = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("heads", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 398, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.heads", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_14heads(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self), __pyx_v_keys);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_14heads(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_keys) {
  PyObject *__pyx_v_maybe_node;
  PyObject *__pyx_v_maybe_heads;
  PyObject *__pyx_v_temp_node;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node = 0;
  Py_ssize_t __pyx_v_pos;
  Py_ssize_t __pyx_v_last_item;
  long __pyx_v_min_gdfo;
  PyObject *__pyx_v_heads_key = NULL;
  PyObject *__pyx_v_candidate_nodes = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_cleanup = NULL;
  PyObject *__pyx_v_pending = NULL;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_parent_node = NULL;
  PyObject *__pyx_v_heads = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8[3];
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  long __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("heads", 0);

  /* "breezy/_known_graph_pyx.pyx":419
 *         cdef long min_gdfo
 * 
 *         heads_key = frozenset(keys)             # <<<<<<<<<<<<<<
 *         maybe_heads = PyDict_GetItem(self._known_heads, heads_key)
 *         if maybe_heads != NULL:
*/
  __pyx_t_1 = __Pyx_PyFrozenSet_New(__pyx_v_keys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_heads_key = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":420
 * 
 *         heads_key = frozenset(keys)
 *         maybe_heads = PyDict_GetItem(self._known_heads, heads_key)             # <<<<<<<<<<<<<<
 *         if maybe_heads != NULL:
 *             return <object>maybe_heads
*/
  __pyx_t_1 = __pyx_v_self->_known_heads;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_maybe_heads = PyDict_GetItem(__pyx_t_1, __pyx_v_heads_key);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":421
 *         heads_key = frozenset(keys)
 *         maybe_heads = PyDict_GetItem(self._known_heads, heads_key)
 *         if maybe_heads != NULL:             # <<<<<<<<<<<<<<
 *             return <object>maybe_heads
 *         # Not cached, compute it ourselves
*/
  __pyx_t_2 = (__pyx_v_maybe_heads != NULL);
  if (__pyx_t_2) {

    /* "breezy/_known_graph_pyx.pyx":422
 *         maybe_heads = PyDict_GetItem(self._known_heads, heads_key)
 *         if maybe_heads != NULL:
 *             return <object>maybe_heads             # <<<<<<<<<<<<<<
 *         # Not cached, compute it ourselves
 *         candidate_nodes = {}
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_maybe_heads));
    __pyx_r = ((PyObject *)__pyx_v_maybe_heads);
    goto __pyx_L0;

    /* "breezy/_known_graph_pyx.pyx":421
 *         heads_key = frozenset(keys)
 *         maybe_heads = PyDict_GetItem(self._known_heads, heads_key)
 *         if maybe_heads != NULL:             # <<<<<<<<<<<<<<
 *             return <object>maybe_heads
 *         # Not cached, compute it ourselves
*/
  }

  /* "breezy/_known_graph_pyx.pyx":424
 *             return <object>maybe_heads
 *         # Not cached, compute it ourselves
 *         candidate_nodes = {}             # <<<<<<<<<<<<<<
 *         for key in keys:
 *             maybe_node = PyDict_GetItem(self._nodes, key)
*/
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_candidate_nodes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":425
 *         # Not cached, compute it ourselves
 *         candidate_nodes = {}
 *         for key in keys:             # <<<<<<<<<<<<<<
 *             maybe_node = PyDict_GetItem(self._nodes, key)
 *             if maybe_node == NULL:
*/
  if (likely(PyList_CheckExact(__pyx_v_keys)) || PyTuple_CheckExact(__pyx_v_keys)) {
    __pyx_t_1 = __pyx_v_keys; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_keys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 425, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 425, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        __pyx_t_5 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_3;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 425, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3));
        #else
        __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3);
        #endif
        ++__pyx_t_3;
      }
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 425, __pyx_L1_error)
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_1);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 425, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "breezy/_known_graph_pyx.pyx":426
 *         candidate_nodes = {}
 *         for key in keys:
 *             maybe_node = PyDict_GetItem(self._nodes, key)             # <<<<<<<<<<<<<<
 *             if maybe_node == NULL:
 *                 raise KeyError('key %s not in nodes' % (key,))
*/
    __pyx_t_5 = __pyx_v_self->_nodes;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_v_maybe_node = PyDict_GetItem(__pyx_t_5, __pyx_v_key);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "breezy/_known_graph_pyx.pyx":427
 *         for key in keys:
 *             maybe_node = PyDict_GetItem(self._nodes, key)
 *             if maybe_node == NULL:             # <<<<<<<<<<<<<<
 *                 raise KeyError('key %s not in nodes' % (key,))
 *             PyDict_SetItem(candidate_nodes, key, <object>maybe_node)
*/
    __pyx_t_2 = (__pyx_v_maybe_node == NULL);
    if (unlikely(__pyx_t_2)) {

      /* "breezy/_known_graph_pyx.pyx":428
 *             maybe_node = PyDict_GetItem(self._nodes, key)
 *             if maybe_node == NULL:
 *                 raise KeyError('key %s not in nodes' % (key,))             # <<<<<<<<<<<<<<
 *             PyDict_SetItem(candidate_nodes, key, <object>maybe_node)
 *         maybe_node = PyDict_GetItem(candidate_nodes, NULL_REVISION)
*/
      __pyx_t_6 = NULL;
      __pyx_t_7 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_v_key), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 428, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8[0] = __pyx_mstate_global->__pyx_kp_u_key_2;
      __pyx_t_8[1] = __pyx_t_7;
      __pyx_t_8[2] = __pyx_mstate_global->__pyx_kp_u_not_in_nodes;
      __pyx_t_9 = __Pyx_PyUnicode_Join(__pyx_t_8, 3, 4 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7) + 13, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7));
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 428, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_10 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_9};
        __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_KeyError)), __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 428, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 428, __pyx_L1_error)

      /* "breezy/_known_graph_pyx.pyx":427
 *         for key in keys:
 *             maybe_node = PyDict_GetItem(self._nodes, key)
 *             if maybe_node == NULL:             # <<<<<<<<<<<<<<
 *                 raise KeyError('key %s not in nodes' % (key,))
 *             PyDict_SetItem(candidate_nodes, key, <object>maybe_node)
*/
    }

    /* "breezy/_known_graph_pyx.pyx":429
 *             if maybe_node == NULL:
 *                 raise KeyError('key %s not in nodes' % (key,))
 *             PyDict_SetItem(candidate_nodes, key, <object>maybe_node)             # <<<<<<<<<<<<<<
 *         maybe_node = PyDict_GetItem(candidate_nodes, NULL_REVISION)
 *         if maybe_node != NULL:
*/
    __pyx_t_11 = PyDict_SetItem(__pyx_v_candidate_nodes, __pyx_v_key, ((PyObject *)__pyx_v_maybe_node)); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 429, __pyx_L1_error)

    /* "breezy/_known_graph_pyx.pyx":425
 *         # Not cached, compute it ourselves
 *         candidate_nodes = {}
 *         for key in keys:             # <<<<<<<<<<<<<<
 *             maybe_node = PyDict_GetItem(self._nodes, key)
 *             if maybe_node == NULL:
*/
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":430
 *                 raise KeyError('key %s not in nodes' % (key,))
 *             PyDict_SetItem(candidate_nodes, key, <object>maybe_node)
 *         maybe_node = PyDict_GetItem(candidate_nodes, NULL_REVISION)             # <<<<<<<<<<<<<<
 *         if maybe_node != NULL:
 *             # NULL_REVISION is only a head if it is the only entry
*/
  __pyx_t_1 = __pyx_v_6breezy_16_known_graph_pyx_NULL_REVISION;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_maybe_node = PyDict_GetItem(__pyx_v_candidate_nodes, __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":431
 *             PyDict_SetItem(candidate_nodes, key, <object>maybe_node)
 *         maybe_node = PyDict_GetItem(candidate_nodes, NULL_REVISION)
 *         if maybe_node != NULL:             # <<<<<<<<<<<<<<
 *             # NULL_REVISION is only a head if it is the only entry
 *             candidate_nodes.pop(NULL_REVISION)
*/
  __pyx_t_2 = (__pyx_v_maybe_node != NULL);
  if (__pyx_t_2) {

    /* "breezy/_known_graph_pyx.pyx":433
 *         if maybe_node != NULL:
 *             # NULL_REVISION is only a head if it is the only entry
 *             candidate_nodes.pop(NULL_REVISION)             # <<<<<<<<<<<<<<
 *             if not candidate_nodes:
 *                 return frozenset([NULL_REVISION])
*/
    __pyx_t_1 = __Pyx_PyDict_Pop(__pyx_v_candidate_nodes, __pyx_v_6breezy_16_known_graph_pyx_NULL_REVISION, ((PyObject *)NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":434
 *             # NULL_REVISION is only a head if it is the only entry
 *             candidate_nodes.pop(NULL_REVISION)
 *             if not candidate_nodes:             # <<<<<<<<<<<<<<
 *                 return frozenset([NULL_REVISION])
 *             # The keys changed, so recalculate heads_key
*/
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_candidate_nodes); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 434, __pyx_L1_error)
    __pyx_t_12 = (!__pyx_t_2);
    if (__pyx_t_12) {

      /* "breezy/_known_graph_pyx.pyx":435
 *             candidate_nodes.pop(NULL_REVISION)
 *             if not candidate_nodes:
 *                 return frozenset([NULL_REVISION])             # <<<<<<<<<<<<<<
 *             # The keys changed, so recalculate heads_key
 *             heads_key = frozenset(candidate_nodes)
*/
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 435, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_6breezy_16_known_graph_pyx_NULL_REVISION);
      __Pyx_GIVEREF(__pyx_v_6breezy_16_known_graph_pyx_NULL_REVISION);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_6breezy_16_known_graph_pyx_NULL_REVISION) != (0)) __PYX_ERR(0, 435, __pyx_L1_error);
      __pyx_t_5 = __Pyx_PyFrozenSet_New(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 435, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_r = __pyx_t_5;
      __pyx_t_5 = 0;
      goto __pyx_L0;

      /* "breezy/_known_graph_pyx.pyx":434
 *             # NULL_REVISION is only a head if it is the only entry
 *             candidate_nodes.pop(NULL_REVISION)
 *             if not candidate_nodes:             # <<<<<<<<<<<<<<
 *                 return frozenset([NULL_REVISION])
 *             # The keys changed, so recalculate heads_key
*/
    }

    /* "breezy/_known_graph_pyx.pyx":437
 *                 return frozenset([NULL_REVISION])
 *             # The keys changed, so recalculate heads_key
 *             heads_key = frozenset(candidate_nodes)             # <<<<<<<<<<<<<<
 *         if PyDict_Size(candidate_nodes) < 2:
 *             return heads_key
*/
    __pyx_t_5 = __Pyx_PyFrozenSet_New(__pyx_v_candidate_nodes); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF_SET(__pyx_v_heads_key, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "breezy/_known_graph_pyx.pyx":431
 *             PyDict_SetItem(candidate_nodes, key, <object>maybe_node)
 *         maybe_node = PyDict_GetItem(candidate_nodes, NULL_REVISION)
 *         if maybe_node != NULL:             # <<<<<<<<<<<<<<
 *             # NULL_REVISION is only a head if it is the only entry
 *             candidate_nodes.pop(NULL_REVISION)
*/
  }

  /* "breezy/_known_graph_pyx.pyx":438
 *             # The keys changed, so recalculate heads_key
 *             heads_key = frozenset(candidate_nodes)
 *         if PyDict_Size(candidate_nodes) < 2:             # <<<<<<<<<<<<<<
 *             return heads_key
 * 
*/
  __pyx_t_3 = PyDict_Size(__pyx_v_candidate_nodes); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 438, __pyx_L1_error)
  __pyx_t_12 = (__pyx_t_3 < 2);
  if (__pyx_t_12) {

    /* "breezy/_known_graph_pyx.pyx":439
 *             heads_key = frozenset(candidate_nodes)
 *         if PyDict_Size(candidate_nodes) < 2:
 *             return heads_key             # <<<<<<<<<<<<<<
 * 
 *         cleanup = []
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_heads_key);
    __pyx_r = __pyx_v_heads_key;
    goto __pyx_L0;

    /* "breezy/_known_graph_pyx.pyx":438
 *             # The keys changed, so recalculate heads_key
 *             heads_key = frozenset(candidate_nodes)
 *         if PyDict_Size(candidate_nodes) < 2:             # <<<<<<<<<<<<<<
 *             return heads_key
 * 
*/
  }

  /* "breezy/_known_graph_pyx.pyx":441
 *             return heads_key
 * 
 *         cleanup = []             # <<<<<<<<<<<<<<
 *         pending = []
 *         # we know a gdfo cannot be longer than a linear chain of all nodes
*/
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_cleanup = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "breezy/_known_graph_pyx.pyx":442
 * 
 *         cleanup = []
 *         pending = []             # <<<<<<<<<<<<<<
 *         # we know a gdfo cannot be longer than a linear chain of all nodes
 *         min_gdfo = PyDict_Size(self._nodes) + 1
*/
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_pending = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "breezy/_known_graph_pyx.pyx":444
 *         pending = []
 *         # we know a gdfo cannot be longer than a linear chain of all nodes
 *         min_gdfo = PyDict_Size(self._nodes) + 1             # <<<<<<<<<<<<<<
 *         # Build up nodes that need to be walked, note that starting nodes are
 *         # not added to seen()
*/
  __pyx_t_5 = __pyx_v_self->_nodes;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_3 = PyDict_Size(__pyx_t_5); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_min_gdfo = (__pyx_t_3 + 1);

  /* "breezy/_known_graph_pyx.pyx":447
 *         # Build up nodes that need to be walked, note that starting nodes are
 *         # not added to seen()
 *         pos = 0             # <<<<<<<<<<<<<<
 *         while PyDict_Next(candidate_nodes, &pos, NULL, &temp_node):
 *             node = <_KnownGraphNode>temp_node
*/
  __pyx_v_pos = 0;

  /* "breezy/_known_graph_pyx.pyx":448
 *         # not added to seen()
 *         pos = 0
 *         while PyDict_Next(candidate_nodes, &pos, NULL, &temp_node):             # <<<<<<<<<<<<<<
 *             node = <_KnownGraphNode>temp_node
 *             if node.parents is not None:
*/
  while (1) {
    __pyx_t_12 = (PyDict_Next(__pyx_v_candidate_nodes, (&__pyx_v_pos), NULL, (&__pyx_v_temp_node)) != 0);
    if (!__pyx_t_12) break;

    /* "breezy/_known_graph_pyx.pyx":449
 *         pos = 0
 *         while PyDict_Next(candidate_nodes, &pos, NULL, &temp_node):
 *             node = <_KnownGraphNode>temp_node             # <<<<<<<<<<<<<<
 *             if node.parents is not None:
 *                 pending.extend(node.parents)
*/
    __pyx_t_5 = ((PyObject *)__pyx_v_temp_node);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "breezy/_known_graph_pyx.pyx":450
 *         while PyDict_Next(candidate_nodes, &pos, NULL, &temp_node):
 *             node = <_KnownGraphNode>temp_node
 *             if node.parents is not None:             # <<<<<<<<<<<<<<
 *                 pending.extend(node.parents)
 *             if node.gdfo < min_gdfo:
*/
    __pyx_t_12 = (__pyx_v_node->parents != Py_None);
    if (__pyx_t_12) {

      /* "breezy/_known_graph_pyx.pyx":451
 *             node = <_KnownGraphNode>temp_node
 *             if node.parents is not None:
 *                 pending.extend(node.parents)             # <<<<<<<<<<<<<<
 *             if node.gdfo < min_gdfo:
 *                 min_gdfo = node.gdfo
*/
      __pyx_t_5 = __pyx_v_node->parents;
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_13 = __Pyx_PyList_Extend(__pyx_v_pending, __pyx_t_5); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 451, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "breezy/_known_graph_pyx.pyx":450
 *         while PyDict_Next(candidate_nodes, &pos, NULL, &temp_node):
 *             node = <_KnownGraphNode>temp_node
 *             if node.parents is not None:             # <<<<<<<<<<<<<<
 *                 pending.extend(node.parents)
 *             if node.gdfo < min_gdfo:
*/
    }

    /* "breezy/_known_graph_pyx.pyx":452
 *             if node.parents is not None:
 *                 pending.extend(node.parents)
 *             if node.gdfo < min_gdfo:             # <<<<<<<<<<<<<<
 *                 min_gdfo = node.gdfo
 * 
*/
    __pyx_t_12 = (__pyx_v_node->gdfo < __pyx_v_min_gdfo);
    if (__pyx_t_12) {

      /* "breezy/_known_graph_pyx.pyx":453
 *                 pending.extend(node.parents)
 *             if node.gdfo < min_gdfo:
 *                 min_gdfo = node.gdfo             # <<<<<<<<<<<<<<
 * 
 *         # Now do all the real work
*/
      __pyx_t_14 = __pyx_v_node->gdfo;
      __pyx_v_min_gdfo = __pyx_t_14;

      /* "breezy/_known_graph_pyx.pyx":452
 *             if node.parents is not None:
 *                 pending.extend(node.parents)
 *             if node.gdfo < min_gdfo:             # <<<<<<<<<<<<<<
 *                 min_gdfo = node.gdfo
 * 
*/
    }
  }

  /* "breezy/_known_graph_pyx.pyx":456
 * 
 *         # Now do all the real work
 *         last_item = PyList_GET_SIZE(pending) - 1             # <<<<<<<<<<<<<<
 *         while last_item >= 0:
 *             node = _get_list_node(pending, last_item)
*/
  __pyx_v_last_item = (PyList_GET_SIZE(__pyx_v_pending) - 1);

  /* "breezy/_known_graph_pyx.pyx":457
 *         # Now do all the real work
 *         last_item = PyList_GET_SIZE(pending) - 1
 *         while last_item >= 0:             # <<<<<<<<<<<<<<
 *             node = _get_list_node(pending, last_item)
 *             last_item = last_item - 1
*/
  while (1) {
    __pyx_t_12 = (__pyx_v_last_item >= 0);
    if (!__pyx_t_12) break;

    /* "breezy/_known_graph_pyx.pyx":458
 *         last_item = PyList_GET_SIZE(pending) - 1
 *         while last_item >= 0:
 *             node = _get_list_node(pending, last_item)             # <<<<<<<<<<<<<<
 *             last_item = last_item - 1
 *             if node.seen:
*/
    __pyx_t_5 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_list_node(__pyx_v_pending, __pyx_v_last_item)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "breezy/_known_graph_pyx.pyx":459
 *         while last_item >= 0:
 *             node = _get_list_node(pending, last_item)
 *             last_item = last_item - 1             # <<<<<<<<<<<<<<
 *             if node.seen:
 *                 # node already appears in some ancestry
*/
    __pyx_v_last_item = (__pyx_v_last_item - 1);

    /* "breezy/_known_graph_pyx.pyx":460
 *             node = _get_list_node(pending, last_item)
 *             last_item = last_item - 1
 *             if node.seen:             # <<<<<<<<<<<<<<
 *                 # node already appears in some ancestry
 *                 continue
*/
    __pyx_t_12 = (__pyx_v_node->seen != 0);
    if (__pyx_t_12) {

      /* "breezy/_known_graph_pyx.pyx":462
 *             if node.seen:
 *                 # node already appears in some ancestry
 *                 continue             # <<<<<<<<<<<<<<
 *             PyList_Append(cleanup, node)
 *             node.seen = 1
*/
      goto __pyx_L15_continue;

      /* "breezy/_known_graph_pyx.pyx":460
 *             node = _get_list_node(pending, last_item)
 *             last_item = last_item - 1
 *             if node.seen:             # <<<<<<<<<<<<<<
 *                 # node already appears in some ancestry
 *                 continue
*/
    }

    /* "breezy/_known_graph_pyx.pyx":463
 *                 # node already appears in some ancestry
 *                 continue
 *             PyList_Append(cleanup, node)             # <<<<<<<<<<<<<<
 *             node.seen = 1
 *             if node.gdfo <= min_gdfo:
*/
    __pyx_t_11 = PyList_Append(__pyx_v_cleanup, ((PyObject *)__pyx_v_node)); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 463, __pyx_L1_error)

    /* "breezy/_known_graph_pyx.pyx":464
 *                 continue
 *             PyList_Append(cleanup, node)
 *             node.seen = 1             # <<<<<<<<<<<<<<
 *             if node.gdfo <= min_gdfo:
 *                 continue
*/
    __pyx_v_node->seen = 1;

    /* "breezy/_known_graph_pyx.pyx":465
 *             PyList_Append(cleanup, node)
 *             node.seen = 1
 *             if node.gdfo <= min_gdfo:             # <<<<<<<<<<<<<<
 *                 continue
 *             if node.parents is not None and PyTuple_GET_SIZE(node.parents) > 0:
*/
    __pyx_t_12 = (__pyx_v_node->gdfo <= __pyx_v_min_gdfo);
    if (__pyx_t_12) {

      /* "breezy/_known_graph_pyx.pyx":466
 *             node.seen = 1
 *             if node.gdfo <= min_gdfo:
 *                 continue             # <<<<<<<<<<<<<<
 *             if node.parents is not None and PyTuple_GET_SIZE(node.parents) > 0:
 *                 for pos from 0 <= pos < PyTuple_GET_SIZE(node.parents):
*/
      goto __pyx_L15_continue;

      /* "breezy/_known_graph_pyx.pyx":465
 *             PyList_Append(cleanup, node)
 *             node.seen = 1
 *             if node.gdfo <= min_gdfo:             # <<<<<<<<<<<<<<
 *                 continue
 *             if node.parents is not None and PyTuple_GET_SIZE(node.parents) > 0:
*/
    }

    /* "breezy/_known_graph_pyx.pyx":467
 *             if node.gdfo <= min_gdfo:
 *                 continue
 *             if node.parents is not None and PyTuple_GET_SIZE(node.parents) > 0:             # <<<<<<<<<<<<<<
 *                 for pos from 0 <= pos < PyTuple_GET_SIZE(node.parents):
 *                     parent_node = _get_tuple_node(node.parents, pos)
*/
    __pyx_t_2 = (__pyx_v_node->parents != Py_None);
    if (__pyx_t_2) {
    } else {
      __pyx_t_12 = __pyx_t_2;
      goto __pyx_L20_bool_binop_done;
    }
    __pyx_t_5 = __pyx_v_node->parents;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_2 = (PyTuple_GET_SIZE(__pyx_t_5) > 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_12 = __pyx_t_2;
    __pyx_L20_bool_binop_done:;
    if (__pyx_t_12) {

      /* "breezy/_known_graph_pyx.pyx":468
 *                 continue
 *             if node.parents is not None and PyTuple_GET_SIZE(node.parents) > 0:
 *                 for pos from 0 <= pos < PyTuple_GET_SIZE(node.parents):             # <<<<<<<<<<<<<<
 *                     parent_node = _get_tuple_node(node.parents, pos)
 *                     last_item = last_item + 1
*/
      __pyx_t_5 = __pyx_v_node->parents;
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_3 = PyTuple_GET_SIZE(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      for (__pyx_v_pos = 0; __pyx_v_pos < __pyx_t_3; __pyx_v_pos++) {

        /* "breezy/_known_graph_pyx.pyx":469
 *             if node.parents is not None and PyTuple_GET_SIZE(node.parents) > 0:
 *                 for pos from 0 <= pos < PyTuple_GET_SIZE(node.parents):
 *                     parent_node = _get_tuple_node(node.parents, pos)             # <<<<<<<<<<<<<<
 *                     last_item = last_item + 1
 *                     if last_item < PyList_GET_SIZE(pending):
*/
        __pyx_t_5 = __pyx_v_node->parents;
        __Pyx_INCREF(__pyx_t_5);
        __pyx_t_1 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_tuple_node(__pyx_t_5, __pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF_SET(__pyx_v_parent_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "breezy/_known_graph_pyx.pyx":470
 *                 for pos from 0 <= pos < PyTuple_GET_SIZE(node.parents):
 *                     parent_node = _get_tuple_node(node.parents, pos)
 *                     last_item = last_item + 1             # <<<<<<<<<<<<<<
 *                     if last_item < PyList_GET_SIZE(pending):
 *                         Py_INCREF(parent_node) # SetItem steals a ref
*/
        __pyx_v_last_item = (__pyx_v_last_item + 1);

        /* "breezy/_known_graph_pyx.pyx":471
 *                     parent_node = _get_tuple_node(node.parents, pos)
 *                     last_item = last_item + 1
 *                     if last_item < PyList_GET_SIZE(pending):             # <<<<<<<<<<<<<<
 *                         Py_INCREF(parent_node) # SetItem steals a ref
 *                         PyList_SetItem(pending, last_item, parent_node)
*/
        __pyx_t_12 = (__pyx_v_last_item < PyList_GET_SIZE(__pyx_v_pending));
        if (__pyx_t_12) {

          /* "breezy/_known_graph_pyx.pyx":472
 *                     last_item = last_item + 1
 *                     if last_item < PyList_GET_SIZE(pending):
 *                         Py_INCREF(parent_node) # SetItem steals a ref             # <<<<<<<<<<<<<<
 *                         PyList_SetItem(pending, last_item, parent_node)
 *                     else:
*/
          Py_INCREF(((PyObject *)__pyx_v_parent_node));

          /* "breezy/_known_graph_pyx.pyx":473
 *                     if last_item < PyList_GET_SIZE(pending):
 *                         Py_INCREF(parent_node) # SetItem steals a ref
 *                         PyList_SetItem(pending, last_item, parent_node)             # <<<<<<<<<<<<<<
 *                     else:
 *                         PyList_Append(pending, parent_node)
*/
          __pyx_t_11 = PyList_SetItem(__pyx_v_pending, __pyx_v_last_item, ((PyObject *)__pyx_v_parent_node)); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 473, __pyx_L1_error)

          /* "breezy/_known_graph_pyx.pyx":471
 *                     parent_node = _get_tuple_node(node.parents, pos)
 *                     last_item = last_item + 1
 *                     if last_item < PyList_GET_SIZE(pending):             # <<<<<<<<<<<<<<
 *                         Py_INCREF(parent_node) # SetItem steals a ref
 *                         PyList_SetItem(pending, last_item, parent_node)
*/
          goto __pyx_L24;
        }

        /* "breezy/_known_graph_pyx.pyx":475
 *                         PyList_SetItem(pending, last_item, parent_node)
 *                     else:
 *                         PyList_Append(pending, parent_node)             # <<<<<<<<<<<<<<
 *         heads = []
 *         pos = 0
*/
        /*else*/ {
          __pyx_t_11 = PyList_Append(__pyx_v_pending, ((PyObject *)__pyx_v_parent_node)); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 475, __pyx_L1_error)
        }
        __pyx_L24:;
      }

      /* "breezy/_known_graph_pyx.pyx":467
 *             if node.gdfo <= min_gdfo:
 *                 continue
 *             if node.parents is not None and PyTuple_GET_SIZE(node.parents) > 0:             # <<<<<<<<<<<<<<
 *                 for pos from 0 <= pos < PyTuple_GET_SIZE(node.parents):
 *                     parent_node = _get_tuple_node(node.parents, pos)
*/
    }
    __pyx_L15_continue:;
  }

  /* "breezy/_known_graph_pyx.pyx":476
 *                     else:
 *                         PyList_Append(pending, parent_node)
 *         heads = []             # <<<<<<<<<<<<<<
 *         pos = 0
 *         while PyDict_Next(candidate_nodes, &pos, NULL, &temp_node):
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_heads = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":477
 *                         PyList_Append(pending, parent_node)
 *         heads = []
 *         pos = 0             # <<<<<<<<<<<<<<
 *         while PyDict_Next(candidate_nodes, &pos, NULL, &temp_node):
 *             node = <_KnownGraphNode>temp_node
*/
  __pyx_v_pos = 0;

  /* "breezy/_known_graph_pyx.pyx":478
 *         heads = []
 *         pos = 0
 *         while PyDict_Next(candidate_nodes, &pos, NULL, &temp_node):             # <<<<<<<<<<<<<<
 *             node = <_KnownGraphNode>temp_node
 *             if not node.seen:
*/
  while (1) {
    __pyx_t_12 = (PyDict_Next(__pyx_v_candidate_nodes, (&__pyx_v_pos), NULL, (&__pyx_v_temp_node)) != 0);
    if (!__pyx_t_12) break;

    /* "breezy/_known_graph_pyx.pyx":479
 *         pos = 0
 *         while PyDict_Next(candidate_nodes, &pos, NULL, &temp_node):
 *             node = <_KnownGraphNode>temp_node             # <<<<<<<<<<<<<<
 *             if not node.seen:
 *                 PyList_Append(heads, node.key)
*/
    __pyx_t_1 = ((PyObject *)__pyx_v_temp_node);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":480
 *         while PyDict_Next(candidate_nodes, &pos, NULL, &temp_node):
 *             node = <_KnownGraphNode>temp_node
 *             if not node.seen:             # <<<<<<<<<<<<<<
 *                 PyList_Append(heads, node.key)
 *         heads = frozenset(heads)
*/
    __pyx_t_12 = (!(__pyx_v_node->seen != 0));
    if (__pyx_t_12) {

      /* "breezy/_known_graph_pyx.pyx":481
 *             node = <_KnownGraphNode>temp_node
 *             if not node.seen:
 *                 PyList_Append(heads, node.key)             # <<<<<<<<<<<<<<
 *         heads = frozenset(heads)
 *         for pos from 0 <= pos < PyList_GET_SIZE(cleanup):
*/
      __pyx_t_1 = __pyx_v_node->key;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_11 = PyList_Append(__pyx_v_heads, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 481, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "breezy/_known_graph_pyx.pyx":480
 *         while PyDict_Next(candidate_nodes, &pos, NULL, &temp_node):
 *             node = <_KnownGraphNode>temp_node
 *             if not node.seen:             # <<<<<<<<<<<<<<
 *                 PyList_Append(heads, node.key)
 *         heads = frozenset(heads)
*/
    }
  }

  /* "breezy/_known_graph_pyx.pyx":482
 *             if not node.seen:
 *                 PyList_Append(heads, node.key)
 *         heads = frozenset(heads)             # <<<<<<<<<<<<<<
 *         for pos from 0 <= pos < PyList_GET_SIZE(cleanup):
 *             node = _get_list_node(cleanup, pos)
*/
  __pyx_t_1 = __Pyx_PyFrozenSet_New(__pyx_v_heads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_heads, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":483
 *                 PyList_Append(heads, node.key)
 *         heads = frozenset(heads)
 *         for pos from 0 <= pos < PyList_GET_SIZE(cleanup):             # <<<<<<<<<<<<<<
 *             node = _get_list_node(cleanup, pos)
 *             node.seen = 0
*/
  __pyx_t_3 = PyList_GET_SIZE(__pyx_v_cleanup);
  for (__pyx_v_pos = 0; __pyx_v_pos < __pyx_t_3; __pyx_v_pos++) {

    /* "breezy/_known_graph_pyx.pyx":484
 *         heads = frozenset(heads)
 *         for pos from 0 <= pos < PyList_GET_SIZE(cleanup):
 *             node = _get_list_node(cleanup, pos)             # <<<<<<<<<<<<<<
 *             node.seen = 0
 *         if self.do_cache:
*/
    __pyx_t_1 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_list_node(__pyx_v_cleanup, __pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":485
 *         for pos from 0 <= pos < PyList_GET_SIZE(cleanup):
 *             node = _get_list_node(cleanup, pos)
 *             node.seen = 0             # <<<<<<<<<<<<<<
 *         if self.do_cache:
 *             PyDict_SetItem(self._known_heads, heads_key, heads)
*/
    __pyx_v_node->seen = 0;
  }

  /* "breezy/_known_graph_pyx.pyx":486
 *             node = _get_list_node(cleanup, pos)
 *             node.seen = 0
 *         if self.do_cache:             # <<<<<<<<<<<<<<
 *             PyDict_SetItem(self._known_heads, heads_key, heads)
 *         return heads
*/
  __pyx_t_12 = (__pyx_v_self->do_cache != 0);
  if (__pyx_t_12) {

    /* "breezy/_known_graph_pyx.pyx":487
 *             node.seen = 0
 *         if self.do_cache:
 *             PyDict_SetItem(self._known_heads, heads_key, heads)             # <<<<<<<<<<<<<<
 *         return heads
 * 
*/
    __pyx_t_1 = __pyx_v_self->_known_heads;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_11 = PyDict_SetItem(__pyx_t_1, __pyx_v_heads_key, __pyx_v_heads); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":486
 *             node = _get_list_node(cleanup, pos)
 *             node.seen = 0
 *         if self.do_cache:             # <<<<<<<<<<<<<<
 *             PyDict_SetItem(self._known_heads, heads_key, heads)
 *         return heads
*/
  }

  /* "breezy/_known_graph_pyx.pyx":488
 *         if self.do_cache:
 *             PyDict_SetItem(self._known_heads, heads_key, heads)
 *         return heads             # <<<<<<<<<<<<<<
 * 
 *     def topo_sort(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_heads);
  __pyx_r = __pyx_v_heads;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":398
 *                     pending_append(child_node)
 * 
 *     def heads(self, keys):             # <<<<<<<<<<<<<<
 *         """Return the heads from amongst keys.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.heads", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF(__pyx_v_heads_key);
  __Pyx_XDECREF(__pyx_v_candidate_nodes);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_cleanup);
  __Pyx_XDECREF(__pyx_v_pending);
  __Pyx_XDECREF((PyObject *)__pyx_v_parent_node);
  __Pyx_XDECREF(__pyx_v_heads);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":490
 *         return heads
 * 
 *     def topo_sort(self):             # <<<<<<<<<<<<<<
 *         """Return the nodes in topological order.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_17topo_sort(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_16topo_sort, "Return the nodes in topological order.\n\n        All parents must occur before all children.\n        ");
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_17topo_sort = {"topo_sort", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_17topo_sort, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_16topo_sort};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_17topo_sort(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("topo_sort (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("topo_sort", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("topo_sort", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_16topo_sort(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_16topo_sort(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node = 0;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_child = 0;
  Py_ssize_t __pyx_v_pos;
  Py_ssize_t __pyx_v_last_item;
  PyObject *__pyx_v_pending = NULL;
  PyObject *__pyx_v_topo_order = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("topo_sort", 0);

  /* "breezy/_known_graph_pyx.pyx":507
 *         cdef Py_ssize_t last_item
 * 
 *         pending = self._find_tails()             # <<<<<<<<<<<<<<
 *         if PyList_GET_SIZE(pending) == 0 and len(self._nodes) > 0:
 *             raise errors.GraphCycleError(self._nodes)
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_find_tails, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 507, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_pending = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":508
 * 
 *         pending = self._find_tails()
 *         if PyList_GET_SIZE(pending) == 0 and len(self._nodes) > 0:             # <<<<<<<<<<<<<<
 *             raise errors.GraphCycleError(self._nodes)
 * 
*/
  __pyx_t_5 = (PyList_GET_SIZE(__pyx_v_pending) == 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = __pyx_v_self->_nodes;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_6 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 508, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = (__pyx_t_6 > 0);
  __pyx_t_4 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_4)) {

    /* "breezy/_known_graph_pyx.pyx":509
 *         pending = self._find_tails()
 *         if PyList_GET_SIZE(pending) == 0 and len(self._nodes) > 0:
 *             raise errors.GraphCycleError(self._nodes)             # <<<<<<<<<<<<<<
 * 
 *         topo_order = []
*/
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_errors); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_GraphCycleError); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_8);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_8, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_self->_nodes};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_8, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 509, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 509, __pyx_L1_error)

    /* "breezy/_known_graph_pyx.pyx":508
 * 
 *         pending = self._find_tails()
 *         if PyList_GET_SIZE(pending) == 0 and len(self._nodes) > 0:             # <<<<<<<<<<<<<<
 *             raise errors.GraphCycleError(self._nodes)
 * 
*/
  }

  /* "breezy/_known_graph_pyx.pyx":511
 *             raise errors.GraphCycleError(self._nodes)
 * 
 *         topo_order = []             # <<<<<<<<<<<<<<
 * 
 *         last_item = PyList_GET_SIZE(pending) - 1
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_topo_order = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":513
 *         topo_order = []
 * 
 *         last_item = PyList_GET_SIZE(pending) - 1             # <<<<<<<<<<<<<<
 *         while last_item >= 0:
 *             # Avoid pop followed by push, instead, peek, and replace
*/
  __pyx_v_last_item = (PyList_GET_SIZE(__pyx_v_pending) - 1);

  /* "breezy/_known_graph_pyx.pyx":514
 * 
 *         last_item = PyList_GET_SIZE(pending) - 1
 *         while last_item >= 0:             # <<<<<<<<<<<<<<
 *             # Avoid pop followed by push, instead, peek, and replace
 *             # timing shows this is 930ms => 770ms for OOo
*/
  while (1) {
    __pyx_t_4 = (__pyx_v_last_item >= 0);
    if (!__pyx_t_4) break;

    /* "breezy/_known_graph_pyx.pyx":517
 *             # Avoid pop followed by push, instead, peek, and replace
 *             # timing shows this is 930ms => 770ms for OOo
 *             node = _get_list_node(pending, last_item)             # <<<<<<<<<<<<<<
 *             last_item = last_item - 1
 *             if node.parents is not None:
*/
    __pyx_t_1 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_list_node(__pyx_v_pending, __pyx_v_last_item)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":518
 *             # timing shows this is 930ms => 770ms for OOo
 *             node = _get_list_node(pending, last_item)
 *             last_item = last_item - 1             # <<<<<<<<<<<<<<
 *             if node.parents is not None:
 *                 # We don't include ghost parents
*/
    __pyx_v_last_item = (__pyx_v_last_item - 1);

    /* "breezy/_known_graph_pyx.pyx":519
 *             node = _get_list_node(pending, last_item)
 *             last_item = last_item - 1
 *             if node.parents is not None:             # <<<<<<<<<<<<<<
 *                 # We don't include ghost parents
 *                 PyList_Append(topo_order, node.key)
*/
    __pyx_t_4 = (__pyx_v_node->parents != Py_None);
    if (__pyx_t_4) {

      /* "breezy/_known_graph_pyx.pyx":521
 *             if node.parents is not None:
 *                 # We don't include ghost parents
 *                 PyList_Append(topo_order, node.key)             # <<<<<<<<<<<<<<
 *             for pos from 0 <= pos < PyList_GET_SIZE(node.children):
 *                 child = _get_list_node(node.children, pos)
*/
      __pyx_t_1 = __pyx_v_node->key;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_9 = PyList_Append(__pyx_v_topo_order, __pyx_t_1); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 521, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "breezy/_known_graph_pyx.pyx":519
 *             node = _get_list_node(pending, last_item)
 *             last_item = last_item - 1
 *             if node.parents is not None:             # <<<<<<<<<<<<<<
 *                 # We don't include ghost parents
 *                 PyList_Append(topo_order, node.key)
*/
    }

    /* "breezy/_known_graph_pyx.pyx":522
 *                 # We don't include ghost parents
 *                 PyList_Append(topo_order, node.key)
 *             for pos from 0 <= pos < PyList_GET_SIZE(node.children):             # <<<<<<<<<<<<<<
 *                 child = _get_list_node(node.children, pos)
 *                 if child.gdfo == -1:
*/
    __pyx_t_1 = __pyx_v_node->children;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_6 = PyList_GET_SIZE(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (__pyx_v_pos = 0; __pyx_v_pos < __pyx_t_6; __pyx_v_pos++) {

      /* "breezy/_known_graph_pyx.pyx":523
 *                 PyList_Append(topo_order, node.key)
 *             for pos from 0 <= pos < PyList_GET_SIZE(node.children):
 *                 child = _get_list_node(node.children, pos)             # <<<<<<<<<<<<<<
 *                 if child.gdfo == -1:
 *                     # We know we have a graph cycle because a node has a parent
*/
      __pyx_t_1 = __pyx_v_node->children;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_8 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_list_node(__pyx_t_1, __pyx_v_pos)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 523, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_child, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_8));
      __pyx_t_8 = 0;

      /* "breezy/_known_graph_pyx.pyx":524
 *             for pos from 0 <= pos < PyList_GET_SIZE(node.children):
 *                 child = _get_list_node(node.children, pos)
 *                 if child.gdfo == -1:             # <<<<<<<<<<<<<<
 *                     # We know we have a graph cycle because a node has a parent
 *                     # which we couldn't find
*/
      __pyx_t_4 = (__pyx_v_child->gdfo == -1L);
      if (unlikely(__pyx_t_4)) {

        /* "breezy/_known_graph_pyx.pyx":527
 *                     # We know we have a graph cycle because a node has a parent
 *                     # which we couldn't find
 *                     raise errors.GraphCycleError(self._nodes)             # <<<<<<<<<<<<<<
 *                 child.seen = child.seen + 1
 *                 if child.seen == PyTuple_GET_SIZE(child.parents):
*/
        __pyx_t_1 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_errors); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 527, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_GraphCycleError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 527, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_3 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          assert(__pyx_t_1);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
          __pyx_t_3 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_self->_nodes};
          __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 527, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
        }
        __Pyx_Raise(__pyx_t_8, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __PYX_ERR(0, 527, __pyx_L1_error)

        /* "breezy/_known_graph_pyx.pyx":524
 *             for pos from 0 <= pos < PyList_GET_SIZE(node.children):
 *                 child = _get_list_node(node.children, pos)
 *                 if child.gdfo == -1:             # <<<<<<<<<<<<<<
 *                     # We know we have a graph cycle because a node has a parent
 *                     # which we couldn't find
*/
      }

      /* "breezy/_known_graph_pyx.pyx":528
 *                     # which we couldn't find
 *                     raise errors.GraphCycleError(self._nodes)
 *                 child.seen = child.seen + 1             # <<<<<<<<<<<<<<
 *                 if child.seen == PyTuple_GET_SIZE(child.parents):
 *                     # All parents of this child have been yielded, queue this
*/
      __pyx_v_child->seen = (__pyx_v_child->seen + 1);

      /* "breezy/_known_graph_pyx.pyx":529
 *                     raise errors.GraphCycleError(self._nodes)
 *                 child.seen = child.seen + 1
 *                 if child.seen == PyTuple_GET_SIZE(child.parents):             # <<<<<<<<<<<<<<
 *                     # All parents of this child have been yielded, queue this
 *                     # one to be yielded as well
*/
      __pyx_t_8 = __pyx_v_child->parents;
      __Pyx_INCREF(__pyx_t_8);
      __pyx_t_4 = (__pyx_v_child->seen == PyTuple_GET_SIZE(__pyx_t_8));
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (__pyx_t_4) {

        /* "breezy/_known_graph_pyx.pyx":532
 *                     # All parents of this child have been yielded, queue this
 *                     # one to be yielded as well
 *                     last_item = last_item + 1             # <<<<<<<<<<<<<<
 *                     if last_item < PyList_GET_SIZE(pending):
 *                         Py_INCREF(child) # SetItem steals a ref
*/
        __pyx_v_last_item = (__pyx_v_last_item + 1);

        /* "breezy/_known_graph_pyx.pyx":533
 *                     # one to be yielded as well
 *                     last_item = last_item + 1
 *                     if last_item < PyList_GET_SIZE(pending):             # <<<<<<<<<<<<<<
 *                         Py_INCREF(child) # SetItem steals a ref
 *                         PyList_SetItem(pending, last_item, child)
*/
        __pyx_t_4 = (__pyx_v_last_item < PyList_GET_SIZE(__pyx_v_pending));
        if (__pyx_t_4) {

          /* "breezy/_known_graph_pyx.pyx":534
 *                     last_item = last_item + 1
 *                     if last_item < PyList_GET_SIZE(pending):
 *                         Py_INCREF(child) # SetItem steals a ref             # <<<<<<<<<<<<<<
 *                         PyList_SetItem(pending, last_item, child)
 *                     else:
*/
          Py_INCREF(((PyObject *)__pyx_v_child));

          /* "breezy/_known_graph_pyx.pyx":535
 *                     if last_item < PyList_GET_SIZE(pending):
 *                         Py_INCREF(child) # SetItem steals a ref
 *                         PyList_SetItem(pending, last_item, child)             # <<<<<<<<<<<<<<
 *                     else:
 *                         PyList_Append(pending, child)
*/
          __pyx_t_9 = PyList_SetItem(__pyx_v_pending, __pyx_v_last_item, ((PyObject *)__pyx_v_child)); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 535, __pyx_L1_error)

          /* "breezy/_known_graph_pyx.pyx":533
 *                     # one to be yielded as well
 *                     last_item = last_item + 1
 *                     if last_item < PyList_GET_SIZE(pending):             # <<<<<<<<<<<<<<
 *                         Py_INCREF(child) # SetItem steals a ref
 *                         PyList_SetItem(pending, last_item, child)
*/
          goto __pyx_L13;
        }

        /* "breezy/_known_graph_pyx.pyx":537
 *                         PyList_SetItem(pending, last_item, child)
 *                     else:
 *                         PyList_Append(pending, child)             # <<<<<<<<<<<<<<
 *                     # We have queued this node, we don't need to track it
 *                     # anymore
*/
        /*else*/ {
          __pyx_t_9 = PyList_Append(__pyx_v_pending, ((PyObject *)__pyx_v_child)); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 537, __pyx_L1_error)
        }
        __pyx_L13:;

        /* "breezy/_known_graph_pyx.pyx":540
 *                     # We have queued this node, we don't need to track it
 *                     # anymore
 *                     child.seen = 0             # <<<<<<<<<<<<<<
 *         # We started from the parents, so we don't need to do anymore work
 *         return topo_order
*/
        __pyx_v_child->seen = 0;

        /* "breezy/_known_graph_pyx.pyx":529
 *                     raise errors.GraphCycleError(self._nodes)
 *                 child.seen = child.seen + 1
 *                 if child.seen == PyTuple_GET_SIZE(child.parents):             # <<<<<<<<<<<<<<
 *                     # All parents of this child have been yielded, queue this
 *                     # one to be yielded as well
*/
      }
    }
  }

  /* "breezy/_known_graph_pyx.pyx":542
 *                     child.seen = 0
 *         # We started from the parents, so we don't need to do anymore work
 *         return topo_order             # <<<<<<<<<<<<<<
 * 
 *     def gc_sort(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_topo_order);
  __pyx_r = __pyx_v_topo_order;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":490
 *         return heads
 * 
 *     def topo_sort(self):             # <<<<<<<<<<<<<<
 *         """Return the nodes in topological order.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.topo_sort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_pending);
  __Pyx_XDECREF(__pyx_v_topo_order);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":544
 *         return topo_order
 * 
 *     def gc_sort(self):             # <<<<<<<<<<<<<<
 *         """Return a reverse topological ordering which is 'stable'.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_19gc_sort(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_18gc_sort, "Return a reverse topological ordering which is 'stable'.\n\n        There are a few constraints:\n          1) Reverse topological (all children before all parents)\n          2) Grouped by prefix\n          3) 'stable' sorting, so that we get the same result, independent of\n             machine, or extra data.\n        To do this, we use the same basic algorithm as topo_sort, but when we\n        aren't sure what node to access next, we sort them lexicographically.\n        ");
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_19gc_sort = {"gc_sort", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_19gc_sort, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_18gc_sort};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_19gc_sort(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("gc_sort (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("gc_sort", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("gc_sort", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_18gc_sort(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_18gc_sort(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self) {
  PyObject *__pyx_v_temp;
  Py_ssize_t __pyx_v_pos;
  Py_ssize_t __pyx_v_last_item;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node = 0;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_parent_node = 0;
  PyObject *__pyx_v_tips = NULL;
  PyObject *__pyx_v_prefix_tips = NULL;
  PyObject *__pyx_v_prefix = NULL;
  PyObject *__pyx_v_tip_nodes = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_pending = NULL;
  PyObject *__pyx_v_parents = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("gc_sort", 0);

  /* "breezy/_known_graph_pyx.pyx":559
 *         cdef _KnownGraphNode node, node2, parent_node
 * 
 *         tips = self._find_tips()             # <<<<<<<<<<<<<<
 *         # Split the tips based on prefix
 *         prefix_tips = {}
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_find_tips, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_tips = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":561
 *         tips = self._find_tips()
 *         # Split the tips based on prefix
 *         prefix_tips = {}             # <<<<<<<<<<<<<<
 *         for pos from 0 <= pos < PyList_GET_SIZE(tips):
 *             node = _get_list_node(tips, pos)
*/
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_prefix_tips = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":562
 *         # Split the tips based on prefix
 *         prefix_tips = {}
 *         for pos from 0 <= pos < PyList_GET_SIZE(tips):             # <<<<<<<<<<<<<<
 *             node = _get_list_node(tips, pos)
 *             if PyBytes_CheckExact(node.key) or len(node.key) == 1:
*/
  __pyx_t_4 = PyList_GET_SIZE(__pyx_v_tips);
  for (__pyx_v_pos = 0; __pyx_v_pos < __pyx_t_4; __pyx_v_pos++) {

    /* "breezy/_known_graph_pyx.pyx":563
 *         prefix_tips = {}
 *         for pos from 0 <= pos < PyList_GET_SIZE(tips):
 *             node = _get_list_node(tips, pos)             # <<<<<<<<<<<<<<
 *             if PyBytes_CheckExact(node.key) or len(node.key) == 1:
 *                 prefix = ''
*/
    __pyx_t_1 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_list_node(__pyx_v_tips, __pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":564
 *         for pos from 0 <= pos < PyList_GET_SIZE(tips):
 *             node = _get_list_node(tips, pos)
 *             if PyBytes_CheckExact(node.key) or len(node.key) == 1:             # <<<<<<<<<<<<<<
 *                 prefix = ''
 *             else:
*/
    __pyx_t_1 = __pyx_v_node->key;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_6 = PyBytes_CheckExact(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = __pyx_v_node->key;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_7 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = (__pyx_t_7 == 1);
    __pyx_t_5 = __pyx_t_6;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_5) {

      /* "breezy/_known_graph_pyx.pyx":565
 *             node = _get_list_node(tips, pos)
 *             if PyBytes_CheckExact(node.key) or len(node.key) == 1:
 *                 prefix = ''             # <<<<<<<<<<<<<<
 *             else:
 *                 prefix = node.key[0]
*/
      __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u__3);
      __Pyx_XDECREF_SET(__pyx_v_prefix, __pyx_mstate_global->__pyx_kp_u__3);

      /* "breezy/_known_graph_pyx.pyx":564
 *         for pos from 0 <= pos < PyList_GET_SIZE(tips):
 *             node = _get_list_node(tips, pos)
 *             if PyBytes_CheckExact(node.key) or len(node.key) == 1:             # <<<<<<<<<<<<<<
 *                 prefix = ''
 *             else:
*/
      goto __pyx_L5;
    }

    /* "breezy/_known_graph_pyx.pyx":567
 *                 prefix = ''
 *             else:
 *                 prefix = node.key[0]             # <<<<<<<<<<<<<<
 *             temp = PyDict_GetItem(prefix_tips, prefix)
 *             if temp == NULL:
*/
    /*else*/ {
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_node->key, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_prefix, __pyx_t_1);
      __pyx_t_1 = 0;
    }
    __pyx_L5:;

    /* "breezy/_known_graph_pyx.pyx":568
 *             else:
 *                 prefix = node.key[0]
 *             temp = PyDict_GetItem(prefix_tips, prefix)             # <<<<<<<<<<<<<<
 *             if temp == NULL:
 *                 prefix_tips[prefix] = [node]
*/
    __pyx_v_temp = PyDict_GetItem(__pyx_v_prefix_tips, __pyx_v_prefix);

    /* "breezy/_known_graph_pyx.pyx":569
 *                 prefix = node.key[0]
 *             temp = PyDict_GetItem(prefix_tips, prefix)
 *             if temp == NULL:             # <<<<<<<<<<<<<<
 *                 prefix_tips[prefix] = [node]
 *             else:
*/
    __pyx_t_5 = (__pyx_v_temp == NULL);
    if (__pyx_t_5) {

      /* "breezy/_known_graph_pyx.pyx":570
 *             temp = PyDict_GetItem(prefix_tips, prefix)
 *             if temp == NULL:
 *                 prefix_tips[prefix] = [node]             # <<<<<<<<<<<<<<
 *             else:
 *                 tip_nodes = <object>temp
*/
      __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF((PyObject *)__pyx_v_node);
      __Pyx_GIVEREF((PyObject *)__pyx_v_node);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_node)) != (0)) __PYX_ERR(0, 570, __pyx_L1_error);
      if (unlikely((PyDict_SetItem(__pyx_v_prefix_tips, __pyx_v_prefix, __pyx_t_1) < 0))) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "breezy/_known_graph_pyx.pyx":569
 *                 prefix = node.key[0]
 *             temp = PyDict_GetItem(prefix_tips, prefix)
 *             if temp == NULL:             # <<<<<<<<<<<<<<
 *                 prefix_tips[prefix] = [node]
 *             else:
*/
      goto __pyx_L8;
    }

    /* "breezy/_known_graph_pyx.pyx":572
 *                 prefix_tips[prefix] = [node]
 *             else:
 *                 tip_nodes = <object>temp             # <<<<<<<<<<<<<<
 *                 PyList_Append(tip_nodes, node)
 * 
*/
    /*else*/ {
      __pyx_t_1 = ((PyObject *)__pyx_v_temp);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_tip_nodes, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "breezy/_known_graph_pyx.pyx":573
 *             else:
 *                 tip_nodes = <object>temp
 *                 PyList_Append(tip_nodes, node)             # <<<<<<<<<<<<<<
 * 
 *         result = []
*/
      __pyx_t_8 = PyList_Append(__pyx_v_tip_nodes, ((PyObject *)__pyx_v_node)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 573, __pyx_L1_error)
    }
    __pyx_L8:;
  }

  /* "breezy/_known_graph_pyx.pyx":575
 *                 PyList_Append(tip_nodes, node)
 * 
 *         result = []             # <<<<<<<<<<<<<<
 *         for prefix in sorted(prefix_tips):
 *             temp = PyDict_GetItem(prefix_tips, prefix)
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":576
 * 
 *         result = []
 *         for prefix in sorted(prefix_tips):             # <<<<<<<<<<<<<<
 *             temp = PyDict_GetItem(prefix_tips, prefix)
 *             assert temp != NULL
*/
  __pyx_t_1 = PySequence_List(__pyx_v_prefix_tips); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely((PyList_Sort(__pyx_t_1) < 0))) __PYX_ERR(0, 576, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
      #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 576, __pyx_L1_error)
      #endif
      if (__pyx_t_4 >= __pyx_temp) break;
    }
    __pyx_t_1 = __Pyx_PyList_GetItemRefFast(__pyx_t_2, __pyx_t_4, __Pyx_ReferenceSharing_OwnStrongReference);
    ++__pyx_t_4;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_prefix, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":577
 *         result = []
 *         for prefix in sorted(prefix_tips):
 *             temp = PyDict_GetItem(prefix_tips, prefix)             # <<<<<<<<<<<<<<
 *             assert temp != NULL
 *             tip_nodes = <object>temp
*/
    __pyx_v_temp = PyDict_GetItem(__pyx_v_prefix_tips, __pyx_v_prefix);

    /* "breezy/_known_graph_pyx.pyx":578
 *         for prefix in sorted(prefix_tips):
 *             temp = PyDict_GetItem(prefix_tips, prefix)
 *             assert temp != NULL             # <<<<<<<<<<<<<<
 *             tip_nodes = <object>temp
 *             pending = _sort_list_nodes(tip_nodes, 1)
*/
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_5 = (__pyx_v_temp != NULL);
      if (unlikely(!__pyx_t_5)) {
        __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), 0, 0, 0);
        __PYX_ERR(0, 578, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 578, __pyx_L1_error)
    #endif

    /* "breezy/_known_graph_pyx.pyx":579
 *             temp = PyDict_GetItem(prefix_tips, prefix)
 *             assert temp != NULL
 *             tip_nodes = <object>temp             # <<<<<<<<<<<<<<
 *             pending = _sort_list_nodes(tip_nodes, 1)
 *             last_item = PyList_GET_SIZE(pending) - 1
*/
    __pyx_t_1 = ((PyObject *)__pyx_v_temp);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_tip_nodes, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":580
 *             assert temp != NULL
 *             tip_nodes = <object>temp
 *             pending = _sort_list_nodes(tip_nodes, 1)             # <<<<<<<<<<<<<<
 *             last_item = PyList_GET_SIZE(pending) - 1
 *             while last_item >= 0:
*/
    __pyx_t_1 = __pyx_f_6breezy_16_known_graph_pyx__sort_list_nodes(__pyx_v_tip_nodes, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_pending, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/_known_graph_pyx.pyx":581
 *             tip_nodes = <object>temp
 *             pending = _sort_list_nodes(tip_nodes, 1)
 *             last_item = PyList_GET_SIZE(pending) - 1             # <<<<<<<<<<<<<<
 *             while last_item >= 0:
 *                 node = _get_list_node(pending, last_item)
*/
    __pyx_v_last_item = (PyList_GET_SIZE(__pyx_v_pending) - 1);

    /* "breezy/_known_graph_pyx.pyx":582
 *             pending = _sort_list_nodes(tip_nodes, 1)
 *             last_item = PyList_GET_SIZE(pending) - 1
 *             while last_item >= 0:             # <<<<<<<<<<<<<<
 *                 node = _get_list_node(pending, last_item)
 *                 last_item = last_item - 1
*/
    while (1) {
      __pyx_t_5 = (__pyx_v_last_item >= 0);
      if (!__pyx_t_5) break;

      /* "breezy/_known_graph_pyx.pyx":583
 *             last_item = PyList_GET_SIZE(pending) - 1
 *             while last_item >= 0:
 *                 node = _get_list_node(pending, last_item)             # <<<<<<<<<<<<<<
 *                 last_item = last_item - 1
 *                 if node.parents is None:
*/
      __pyx_t_1 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_list_node(__pyx_v_pending, __pyx_v_last_item)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 583, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "breezy/_known_graph_pyx.pyx":584
 *             while last_item >= 0:
 *                 node = _get_list_node(pending, last_item)
 *                 last_item = last_item - 1             # <<<<<<<<<<<<<<
 *                 if node.parents is None:
 *                     # Ghost
*/
      __pyx_v_last_item = (__pyx_v_last_item - 1);

      /* "breezy/_known_graph_pyx.pyx":585
 *                 node = _get_list_node(pending, last_item)
 *                 last_item = last_item - 1
 *                 if node.parents is None:             # <<<<<<<<<<<<<<
 *                     # Ghost
 *                     continue
*/
      __pyx_t_5 = (__pyx_v_node->parents == Py_None);
      if (__pyx_t_5) {

        /* "breezy/_known_graph_pyx.pyx":587
 *                 if node.parents is None:
 *                     # Ghost
 *                     continue             # <<<<<<<<<<<<<<
 *                 PyList_Append(result, node.key)
 *                 # Sorting the parent keys isn't strictly necessary for stable
*/
        goto __pyx_L11_continue;

        /* "breezy/_known_graph_pyx.pyx":585
 *                 node = _get_list_node(pending, last_item)
 *                 last_item = last_item - 1
 *                 if node.parents is None:             # <<<<<<<<<<<<<<
 *                     # Ghost
 *                     continue
*/
      }

      /* "breezy/_known_graph_pyx.pyx":588
 *                     # Ghost
 *                     continue
 *                 PyList_Append(result, node.key)             # <<<<<<<<<<<<<<
 *                 # Sorting the parent keys isn't strictly necessary for stable
 *                 # sorting of a given graph. But it does help minimize the
*/
      __pyx_t_1 = __pyx_v_node->key;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_8 = PyList_Append(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 588, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "breezy/_known_graph_pyx.pyx":595
 *                 #   4.73ms  no sort
 *                 #   7.73ms  RichCompareBool sort
 *                 parents = _sort_list_nodes(node.parents, 1)             # <<<<<<<<<<<<<<
 *                 for pos from 0 <= pos < len(parents):
 *                     if PyTuple_CheckExact(parents):
*/
      __pyx_t_1 = __pyx_v_node->parents;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_9 = __pyx_f_6breezy_16_known_graph_pyx__sort_list_nodes(__pyx_t_1, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 595, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_parents, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "breezy/_known_graph_pyx.pyx":596
 *                 #   7.73ms  RichCompareBool sort
 *                 parents = _sort_list_nodes(node.parents, 1)
 *                 for pos from 0 <= pos < len(parents):             # <<<<<<<<<<<<<<
 *                     if PyTuple_CheckExact(parents):
 *                         parent_node = _get_tuple_node(parents, pos)
*/
      __pyx_t_7 = PyObject_Length(__pyx_v_parents); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 596, __pyx_L1_error)
      for (__pyx_v_pos = 0; __pyx_v_pos < __pyx_t_7; __pyx_v_pos++) {

        /* "breezy/_known_graph_pyx.pyx":597
 *                 parents = _sort_list_nodes(node.parents, 1)
 *                 for pos from 0 <= pos < len(parents):
 *                     if PyTuple_CheckExact(parents):             # <<<<<<<<<<<<<<
 *                         parent_node = _get_tuple_node(parents, pos)
 *                     else:
*/
        __pyx_t_5 = PyTuple_CheckExact(__pyx_v_parents);
        if (__pyx_t_5) {

          /* "breezy/_known_graph_pyx.pyx":598
 *                 for pos from 0 <= pos < len(parents):
 *                     if PyTuple_CheckExact(parents):
 *                         parent_node = _get_tuple_node(parents, pos)             # <<<<<<<<<<<<<<
 *                     else:
 *                         parent_node = _get_list_node(parents, pos)
*/
          __pyx_t_9 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_tuple_node(__pyx_v_parents, __pyx_v_pos)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 598, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_XDECREF_SET(__pyx_v_parent_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_9));
          __pyx_t_9 = 0;

          /* "breezy/_known_graph_pyx.pyx":597
 *                 parents = _sort_list_nodes(node.parents, 1)
 *                 for pos from 0 <= pos < len(parents):
 *                     if PyTuple_CheckExact(parents):             # <<<<<<<<<<<<<<
 *                         parent_node = _get_tuple_node(parents, pos)
 *                     else:
*/
          goto __pyx_L16;
        }

        /* "breezy/_known_graph_pyx.pyx":600
 *                         parent_node = _get_tuple_node(parents, pos)
 *                     else:
 *                         parent_node = _get_list_node(parents, pos)             # <<<<<<<<<<<<<<
 *                     # TODO: GraphCycle detection
 *                     parent_node.seen = parent_node.seen + 1
*/
        /*else*/ {
          __pyx_t_9 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_list_node(__pyx_v_parents, __pyx_v_pos)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 600, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_XDECREF_SET(__pyx_v_parent_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_9));
          __pyx_t_9 = 0;
        }
        __pyx_L16:;

        /* "breezy/_known_graph_pyx.pyx":602
 *                         parent_node = _get_list_node(parents, pos)
 *                     # TODO: GraphCycle detection
 *                     parent_node.seen = parent_node.seen + 1             # <<<<<<<<<<<<<<
 *                     if (parent_node.seen
 *                         == PyList_GET_SIZE(parent_node.children)):
*/
        __pyx_v_parent_node->seen = (__pyx_v_parent_node->seen + 1);

        /* "breezy/_known_graph_pyx.pyx":604
 *                     parent_node.seen = parent_node.seen + 1
 *                     if (parent_node.seen
 *                         == PyList_GET_SIZE(parent_node.children)):             # <<<<<<<<<<<<<<
 *                         # All children have been processed, queue up this
 *                         # parent
*/
        __pyx_t_9 = __pyx_v_parent_node->children;
        __Pyx_INCREF(__pyx_t_9);
        __pyx_t_5 = (__pyx_v_parent_node->seen == PyList_GET_SIZE(__pyx_t_9));
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "breezy/_known_graph_pyx.pyx":603
 *                     # TODO: GraphCycle detection
 *                     parent_node.seen = parent_node.seen + 1
 *                     if (parent_node.seen             # <<<<<<<<<<<<<<
 *                         == PyList_GET_SIZE(parent_node.children)):
 *                         # All children have been processed, queue up this
*/
        if (__pyx_t_5) {

          /* "breezy/_known_graph_pyx.pyx":607
 *                         # All children have been processed, queue up this
 *                         # parent
 *                         last_item = last_item + 1             # <<<<<<<<<<<<<<
 *                         if last_item < PyList_GET_SIZE(pending):
 *                             Py_INCREF(parent_node) # SetItem steals a ref
*/
          __pyx_v_last_item = (__pyx_v_last_item + 1);

          /* "breezy/_known_graph_pyx.pyx":608
 *                         # parent
 *                         last_item = last_item + 1
 *                         if last_item < PyList_GET_SIZE(pending):             # <<<<<<<<<<<<<<
 *                             Py_INCREF(parent_node) # SetItem steals a ref
 *                             PyList_SetItem(pending, last_item, parent_node)
*/
          __pyx_t_5 = (__pyx_v_last_item < PyList_GET_SIZE(__pyx_v_pending));
          if (__pyx_t_5) {

            /* "breezy/_known_graph_pyx.pyx":609
 *                         last_item = last_item + 1
 *                         if last_item < PyList_GET_SIZE(pending):
 *                             Py_INCREF(parent_node) # SetItem steals a ref             # <<<<<<<<<<<<<<
 *                             PyList_SetItem(pending, last_item, parent_node)
 *                         else:
*/
            Py_INCREF(((PyObject *)__pyx_v_parent_node));

            /* "breezy/_known_graph_pyx.pyx":610
 *                         if last_item < PyList_GET_SIZE(pending):
 *                             Py_INCREF(parent_node) # SetItem steals a ref
 *                             PyList_SetItem(pending, last_item, parent_node)             # <<<<<<<<<<<<<<
 *                         else:
 *                             PyList_Append(pending, parent_node)
*/
            __pyx_t_8 = PyList_SetItem(__pyx_v_pending, __pyx_v_last_item, ((PyObject *)__pyx_v_parent_node)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 610, __pyx_L1_error)

            /* "breezy/_known_graph_pyx.pyx":608
 *                         # parent
 *                         last_item = last_item + 1
 *                         if last_item < PyList_GET_SIZE(pending):             # <<<<<<<<<<<<<<
 *                             Py_INCREF(parent_node) # SetItem steals a ref
 *                             PyList_SetItem(pending, last_item, parent_node)
*/
            goto __pyx_L18;
          }

          /* "breezy/_known_graph_pyx.pyx":612
 *                             PyList_SetItem(pending, last_item, parent_node)
 *                         else:
 *                             PyList_Append(pending, parent_node)             # <<<<<<<<<<<<<<
 *                         parent_node.seen = 0
 *         return result
*/
          /*else*/ {
            __pyx_t_8 = PyList_Append(__pyx_v_pending, ((PyObject *)__pyx_v_parent_node)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 612, __pyx_L1_error)
          }
          __pyx_L18:;

          /* "breezy/_known_graph_pyx.pyx":613
 *                         else:
 *                             PyList_Append(pending, parent_node)
 *                         parent_node.seen = 0             # <<<<<<<<<<<<<<
 *         return result
 * 
*/
          __pyx_v_parent_node->seen = 0;

          /* "breezy/_known_graph_pyx.pyx":603
 *                     # TODO: GraphCycle detection
 *                     parent_node.seen = parent_node.seen + 1
 *                     if (parent_node.seen             # <<<<<<<<<<<<<<
 *                         == PyList_GET_SIZE(parent_node.children)):
 *                         # All children have been processed, queue up this
*/
        }
      }
      __pyx_L11_continue:;
    }

    /* "breezy/_known_graph_pyx.pyx":576
 * 
 *         result = []
 *         for prefix in sorted(prefix_tips):             # <<<<<<<<<<<<<<
 *             temp = PyDict_GetItem(prefix_tips, prefix)
 *             assert temp != NULL
*/
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/_known_graph_pyx.pyx":614
 *                             PyList_Append(pending, parent_node)
 *                         parent_node.seen = 0
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def merge_sort(self, tip_key):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":544
 *         return topo_order
 * 
 *     def gc_sort(self):             # <<<<<<<<<<<<<<
 *         """Return a reverse topological ordering which is 'stable'.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.gc_sort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_parent_node);
  __Pyx_XDECREF(__pyx_v_tips);
  __Pyx_XDECREF(__pyx_v_prefix_tips);
  __Pyx_XDECREF(__pyx_v_prefix);
  __Pyx_XDECREF(__pyx_v_tip_nodes);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_pending);
  __Pyx_XDECREF(__pyx_v_parents);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":616
 *         return result
 * 
 *     def merge_sort(self, tip_key):             # <<<<<<<<<<<<<<
 *         """Compute the merge sorted graph output."""
 *         cdef _MergeSorter sorter
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_21merge_sort(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_20merge_sort, "Compute the merge sorted graph output.");
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_21merge_sort = {"merge_sort", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_21merge_sort, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_20merge_sort};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_21merge_sort(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_tip_key = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("merge_sort (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_tip_key,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 616, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 616, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "merge_sort", 0) < (0)) __PYX_ERR(0, 616, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("merge_sort", 1, 1, 1, i); __PYX_ERR(0, 616, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 616, __pyx_L3_error)
    }
    __pyx_v_tip_key = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("merge_sort", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 616, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.merge_sort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_20merge_sort(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self), __pyx_v_tip_key);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_20merge_sort(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_tip_key) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *__pyx_v_sorter = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("merge_sort", 0);

  /* "breezy/_known_graph_pyx.pyx":623
 *         #       that won't be collectable anyway. real world testing has not
 *         #       shown a specific impact, yet.
 *         sorter = _MergeSorter(self, tip_key)             # <<<<<<<<<<<<<<
 *         return sorter.topo_order()
 * 
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, ((PyObject *)__pyx_v_self), __pyx_v_tip_key};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter, __pyx_callargs+__pyx_t_3, (3-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 623, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_sorter = ((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":624
 *         #       shown a specific impact, yet.
 *         sorter = _MergeSorter(self, tip_key)
 *         return sorter.topo_order()             # <<<<<<<<<<<<<<
 * 
 *     def get_parent_keys(self, key):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSorter *)__pyx_v_sorter->__pyx_vtab)->topo_order(__pyx_v_sorter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":616
 *         return result
 * 
 *     def merge_sort(self, tip_key):             # <<<<<<<<<<<<<<
 *         """Compute the merge sorted graph output."""
 *         cdef _MergeSorter sorter
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.merge_sort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_sorter);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":626
 *         return sorter.topo_order()
 * 
 *     def get_parent_keys(self, key):             # <<<<<<<<<<<<<<
 *         """Get the parents for a key
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_23get_parent_keys(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_22get_parent_keys, "Get the parents for a key\n\n        Returns a list containing the parents keys. If the key is a ghost,\n        None is returned. A KeyError will be raised if the key is not in\n        the graph.\n\n        :param keys: Key to check (eg revision_id)\n        :return: A list of parents\n        ");
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_23get_parent_keys = {"get_parent_keys", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_23get_parent_keys, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_22get_parent_keys};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_23get_parent_keys(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_parent_keys (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 626, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 626, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_parent_keys", 0) < (0)) __PYX_ERR(0, 626, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_parent_keys", 1, 1, 1, i); __PYX_ERR(0, 626, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 626, __pyx_L3_error)
    }
    __pyx_v_key = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_parent_keys", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 626, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.get_parent_keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_22get_parent_keys(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self), __pyx_v_key);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_22get_parent_keys(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_parent_keys", 0);

  /* "breezy/_known_graph_pyx.pyx":636
 *         :return: A list of parents
 *         """
 *         return self._nodes[key].parent_keys             # <<<<<<<<<<<<<<
 * 
 *     def get_child_keys(self, key):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_self->_nodes, __pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_parent_keys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":626
 *         return sorter.topo_order()
 * 
 *     def get_parent_keys(self, key):             # <<<<<<<<<<<<<<
 *         """Get the parents for a key
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.get_parent_keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":638
 *         return self._nodes[key].parent_keys
 * 
 *     def get_child_keys(self, key):             # <<<<<<<<<<<<<<
 *         """Get the children for a key
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_25get_child_keys(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_24get_child_keys, "Get the children for a key\n\n        Returns a list containing the children keys. A KeyError will be raised\n        if the key is not in the graph.\n\n        :param keys: Key to check (eg revision_id)\n        :return: A list of children\n        ");
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_25get_child_keys = {"get_child_keys", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_25get_child_keys, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_24get_child_keys};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_25get_child_keys(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_child_keys (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 638, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 638, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_child_keys", 0) < (0)) __PYX_ERR(0, 638, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_child_keys", 1, 1, 1, i); __PYX_ERR(0, 638, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 638, __pyx_L3_error)
    }
    __pyx_v_key = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_child_keys", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 638, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.get_child_keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_24get_child_keys(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self), __pyx_v_key);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_24get_child_keys(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_child_keys", 0);

  /* "breezy/_known_graph_pyx.pyx":647
 *         :return: A list of children
 *         """
 *         return self._nodes[key].child_keys             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_self->_nodes, __pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_child_keys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":638
 *         return self._nodes[key].parent_keys
 * 
 *     def get_child_keys(self, key):             # <<<<<<<<<<<<<<
 *         """Get the children for a key
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.get_child_keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":177
 *     """This is a class which assumes we already know the full graph."""
 * 
 *     cdef public object _nodes             # <<<<<<<<<<<<<<
 *     cdef public object _known_heads
 *     cdef public int do_cache
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_6_nodes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_6_nodes_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_6_nodes___get__(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_6_nodes___get__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_nodes);
  __pyx_r = __pyx_v_self->_nodes;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_6_nodes_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_6_nodes_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_6_nodes_2__set__(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_6_nodes_2__set__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_nodes);
  __Pyx_DECREF(__pyx_v_self->_nodes);
  __pyx_v_self->_nodes = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_6_nodes_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_6_nodes_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_6_nodes_4__del__(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_6_nodes_4__del__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_nodes);
  __Pyx_DECREF(__pyx_v_self->_nodes);
  __pyx_v_self->_nodes = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":178
 * 
 *     cdef public object _nodes
 *     cdef public object _known_heads             # <<<<<<<<<<<<<<
 *     cdef public int do_cache
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_12_known_heads_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_12_known_heads_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_12_known_heads___get__(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_12_known_heads___get__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_known_heads);
  __pyx_r = __pyx_v_self->_known_heads;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_12_known_heads_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_12_known_heads_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_12_known_heads_2__set__(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_12_known_heads_2__set__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_known_heads);
  __Pyx_DECREF(__pyx_v_self->_known_heads);
  __pyx_v_self->_known_heads = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_12_known_heads_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_12_known_heads_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_12_known_heads_4__del__(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_12_known_heads_4__del__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_known_heads);
  __Pyx_DECREF(__pyx_v_self->_known_heads);
  __pyx_v_self->_known_heads = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":179
 *     cdef public object _nodes
 *     cdef public object _known_heads
 *     cdef public int do_cache             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, parent_map, do_cache=True):
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_8do_cache_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_8do_cache_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_8do_cache___get__(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_8do_cache___get__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->do_cache); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.do_cache.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_8do_cache_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_8do_cache_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_8do_cache_2__set__(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_8do_cache_2__set__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 179, __pyx_L1_error)
  __pyx_v_self->do_cache = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.do_cache.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_27__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_27__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_27__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_27__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_26__reduce_cython__(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_26__reduce_cython__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._known_heads, self._nodes, self.do_cache)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->do_cache); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->_known_heads);
  __Pyx_GIVEREF(__pyx_v_self->_known_heads);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->_known_heads) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_nodes);
  __Pyx_GIVEREF(__pyx_v_self->_nodes);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self->_nodes) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._known_heads, self._nodes, self.do_cache)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self._known_heads, self._nodes, self.do_cache)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(1, 7, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict) != (0)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._known_heads is not None or self._nodes is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._known_heads, self._nodes, self.do_cache)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._known_heads is not None or self._nodes is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_KnownGraph, (type(self), 0xe34a3b8, None), state
*/
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->_known_heads != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->_nodes != Py_None);
    __pyx_t_3 = __pyx_t_4;
    __pyx_L6_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._known_heads is not None or self._nodes is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_KnownGraph, (type(self), 0xe34a3b8, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._known_heads is not None or self._nodes is not None
 *     if use_setstate:
 *         return __pyx_unpickle_KnownGraph, (type(self), 0xe34a3b8, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_KnownGraph, (type(self), 0xe34a3b8, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_KnownGraph); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_238330808);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_238330808);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_mstate_global->__pyx_int_238330808) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._known_heads is not None or self._nodes is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_KnownGraph, (type(self), 0xe34a3b8, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_KnownGraph, (type(self), 0xe34a3b8, None), state
 *     else:
 *         return __pyx_unpickle_KnownGraph, (type(self), 0xe34a3b8, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_KnownGraph__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_KnownGraph); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_238330808);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_238330808);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_mstate_global->__pyx_int_238330808) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_KnownGraph, (type(self), 0xe34a3b8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_KnownGraph__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_29__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_29__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_29__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_29__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_28__setstate_cython__(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_10KnownGraph_28__setstate_cython__(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_KnownGraph, (type(self), 0xe34a3b8, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_KnownGraph__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(1, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_6breezy_16_known_graph_pyx___pyx_unpickle_KnownGraph__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_KnownGraph, (type(self), 0xe34a3b8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_KnownGraph__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy._known_graph_pyx.KnownGraph.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":670
 *     cdef int completed # Fully Processed
 * 
 *     def __init__(self, key):             # <<<<<<<<<<<<<<
 *         self.key = key
 *         self.merge_depth = -1
*/

/* Python wrapper */
static int __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_key = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 670, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 670, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 670, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, i); __PYX_ERR(0, 670, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 670, __pyx_L3_error)
    }
    __pyx_v_key = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 670, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSortNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode___init__(((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_v_self), __pyx_v_key);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode___init__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "breezy/_known_graph_pyx.pyx":671
 * 
 *     def __init__(self, key):
 *         self.key = key             # <<<<<<<<<<<<<<
 *         self.merge_depth = -1
 *         self.left_parent = None
*/
  __Pyx_INCREF(__pyx_v_key);
  __Pyx_GIVEREF(__pyx_v_key);
  __Pyx_GOTREF(__pyx_v_self->key);
  __Pyx_DECREF(__pyx_v_self->key);
  __pyx_v_self->key = __pyx_v_key;

  /* "breezy/_known_graph_pyx.pyx":672
 *     def __init__(self, key):
 *         self.key = key
 *         self.merge_depth = -1             # <<<<<<<<<<<<<<
 *         self.left_parent = None
 *         self.left_pending_parent = None
*/
  __pyx_v_self->merge_depth = -1L;

  /* "breezy/_known_graph_pyx.pyx":673
 *         self.key = key
 *         self.merge_depth = -1
 *         self.left_parent = None             # <<<<<<<<<<<<<<
 *         self.left_pending_parent = None
 *         self.pending_parents = None
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->left_parent);
  __Pyx_DECREF((PyObject *)__pyx_v_self->left_parent);
  __pyx_v_self->left_parent = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)Py_None);

  /* "breezy/_known_graph_pyx.pyx":674
 *         self.merge_depth = -1
 *         self.left_parent = None
 *         self.left_pending_parent = None             # <<<<<<<<<<<<<<
 *         self.pending_parents = None
 *         self._revno_first = -1
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->left_pending_parent);
  __Pyx_DECREF((PyObject *)__pyx_v_self->left_pending_parent);
  __pyx_v_self->left_pending_parent = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)Py_None);

  /* "breezy/_known_graph_pyx.pyx":675
 *         self.left_parent = None
 *         self.left_pending_parent = None
 *         self.pending_parents = None             # <<<<<<<<<<<<<<
 *         self._revno_first = -1
 *         self._revno_second = -1
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->pending_parents);
  __Pyx_DECREF(__pyx_v_self->pending_parents);
  __pyx_v_self->pending_parents = Py_None;

  /* "breezy/_known_graph_pyx.pyx":676
 *         self.left_pending_parent = None
 *         self.pending_parents = None
 *         self._revno_first = -1             # <<<<<<<<<<<<<<
 *         self._revno_second = -1
 *         self._revno_last = -1
*/
  __pyx_v_self->_revno_first = -1L;

  /* "breezy/_known_graph_pyx.pyx":677
 *         self.pending_parents = None
 *         self._revno_first = -1
 *         self._revno_second = -1             # <<<<<<<<<<<<<<
 *         self._revno_last = -1
 *         self.is_first_child = 0
*/
  __pyx_v_self->_revno_second = -1L;

  /* "breezy/_known_graph_pyx.pyx":678
 *         self._revno_first = -1
 *         self._revno_second = -1
 *         self._revno_last = -1             # <<<<<<<<<<<<<<
 *         self.is_first_child = 0
 *         self.seen_by_child = 0
*/
  __pyx_v_self->_revno_last = -1L;

  /* "breezy/_known_graph_pyx.pyx":679
 *         self._revno_second = -1
 *         self._revno_last = -1
 *         self.is_first_child = 0             # <<<<<<<<<<<<<<
 *         self.seen_by_child = 0
 *         self.completed = 0
*/
  __pyx_v_self->is_first_child = 0;

  /* "breezy/_known_graph_pyx.pyx":680
 *         self._revno_last = -1
 *         self.is_first_child = 0
 *         self.seen_by_child = 0             # <<<<<<<<<<<<<<
 *         self.completed = 0
 * 
*/
  __pyx_v_self->seen_by_child = 0;

  /* "breezy/_known_graph_pyx.pyx":681
 *         self.is_first_child = 0
 *         self.seen_by_child = 0
 *         self.completed = 0             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
*/
  __pyx_v_self->completed = 0;

  /* "breezy/_known_graph_pyx.pyx":670
 *     cdef int completed # Fully Processed
 * 
 *     def __init__(self, key):             # <<<<<<<<<<<<<<
 *         self.key = key
 *         self.merge_depth = -1
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":683
 *         self.completed = 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '%s(%s depth:%s rev:%s,%s,%s first:%s seen:%s)' % (
 *             self.__class__.__name__, self.key,
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_2__repr__(((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_2__repr__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9[16];
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "breezy/_known_graph_pyx.pyx":684
 * 
 *     def __repr__(self):
 *         return '%s(%s depth:%s rev:%s,%s,%s first:%s seen:%s)' % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__, self.key,
 *             self.merge_depth,
*/
  __Pyx_XDECREF(__pyx_r);

  /* "breezy/_known_graph_pyx.pyx":685
 *     def __repr__(self):
 *         return '%s(%s depth:%s rev:%s,%s,%s first:%s seen:%s)' % (
 *             self.__class__.__name__, self.key,             # <<<<<<<<<<<<<<
 *             self.merge_depth,
 *             self._revno_first, self._revno_second, self._revno_last,
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_t_2), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_v_self->key), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "breezy/_known_graph_pyx.pyx":686
 *         return '%s(%s depth:%s rev:%s,%s,%s first:%s seen:%s)' % (
 *             self.__class__.__name__, self.key,
 *             self.merge_depth,             # <<<<<<<<<<<<<<
 *             self._revno_first, self._revno_second, self._revno_last,
 *             self.is_first_child, self.seen_by_child)
*/
  __pyx_t_3 = __Pyx_PyUnicode_From_long(__pyx_v_self->merge_depth, 0, ' ', 'd'); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 686, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "breezy/_known_graph_pyx.pyx":687
 *             self.__class__.__name__, self.key,
 *             self.merge_depth,
 *             self._revno_first, self._revno_second, self._revno_last,             # <<<<<<<<<<<<<<
 *             self.is_first_child, self.seen_by_child)
 * 
*/
  __pyx_t_4 = __Pyx_PyUnicode_From_long(__pyx_v_self->_revno_first, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyUnicode_From_long(__pyx_v_self->_revno_second, 0, ' ', 'd'); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyUnicode_From_long(__pyx_v_self->_revno_last, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "breezy/_known_graph_pyx.pyx":688
 *             self.merge_depth,
 *             self._revno_first, self._revno_second, self._revno_last,
 *             self.is_first_child, self.seen_by_child)             # <<<<<<<<<<<<<<
 * 
 *     cdef int has_pending_parents(self): # cannot_raise
*/
  __pyx_t_7 = __Pyx_PyUnicode_From_int(__pyx_v_self->is_first_child, 0, ' ', 'd'); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyUnicode_From_int(__pyx_v_self->seen_by_child, 0, ' ', 'd'); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9[0] = __pyx_t_1;
  __pyx_t_9[1] = __pyx_mstate_global->__pyx_kp_u_;
  __pyx_t_9[2] = __pyx_t_2;
  __pyx_t_9[3] = __pyx_mstate_global->__pyx_kp_u_depth;
  __pyx_t_9[4] = __pyx_t_3;
  __pyx_t_9[5] = __pyx_mstate_global->__pyx_kp_u_rev;
  __pyx_t_9[6] = __pyx_t_4;
  __pyx_t_9[7] = __pyx_mstate_global->__pyx_kp_u__4;
  __pyx_t_9[8] = __pyx_t_5;
  __pyx_t_9[9] = __pyx_mstate_global->__pyx_kp_u__4;
  __pyx_t_9[10] = __pyx_t_6;
  __pyx_t_9[11] = __pyx_mstate_global->__pyx_kp_u_first;
  __pyx_t_9[12] = __pyx_t_7;
  __pyx_t_9[13] = __pyx_mstate_global->__pyx_kp_u_seen;
  __pyx_t_9[14] = __pyx_t_8;
  __pyx_t_9[15] = __pyx_mstate_global->__pyx_kp_u__2;

  /* "breezy/_known_graph_pyx.pyx":684
 * 
 *     def __repr__(self):
 *         return '%s(%s depth:%s rev:%s,%s,%s first:%s seen:%s)' % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__, self.key,
 *             self.merge_depth,
*/
  __pyx_t_10 = __Pyx_PyUnicode_Join(__pyx_t_9, 16, __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1) + 1 * 4 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2) + 7 * 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3) + 5 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7) + 6 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2));
  if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":683
 *         self.completed = 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '%s(%s depth:%s rev:%s,%s,%s first:%s seen:%s)' % (
 *             self.__class__.__name__, self.key,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSortNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":690
 *             self.is_first_child, self.seen_by_child)
 * 
 *     cdef int has_pending_parents(self): # cannot_raise             # <<<<<<<<<<<<<<
 *         if self.left_pending_parent is not None or self.pending_parents:
 *             return 1
*/

static int __pyx_f_6breezy_16_known_graph_pyx_14_MergeSortNode_has_pending_parents(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "breezy/_known_graph_pyx.pyx":691
 * 
 *     cdef int has_pending_parents(self): # cannot_raise
 *         if self.left_pending_parent is not None or self.pending_parents:             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
*/
  __pyx_t_2 = (((PyObject *)__pyx_v_self->left_pending_parent) != Py_None);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_self->pending_parents); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 691, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "breezy/_known_graph_pyx.pyx":692
 *     cdef int has_pending_parents(self): # cannot_raise
 *         if self.left_pending_parent is not None or self.pending_parents:
 *             return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
*/
    __pyx_r = 1;
    goto __pyx_L0;

    /* "breezy/_known_graph_pyx.pyx":691
 * 
 *     cdef int has_pending_parents(self): # cannot_raise
 *         if self.left_pending_parent is not None or self.pending_parents:             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
*/
  }

  /* "breezy/_known_graph_pyx.pyx":693
 *         if self.left_pending_parent is not None or self.pending_parents:
 *             return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef object _revno(self):
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":690
 *             self.is_first_child, self.seen_by_child)
 * 
 *     cdef int has_pending_parents(self): # cannot_raise             # <<<<<<<<<<<<<<
 *         if self.left_pending_parent is not None or self.pending_parents:
 *             return 1
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSortNode.has_pending_parents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":695
 *         return 0
 * 
 *     cdef object _revno(self):             # <<<<<<<<<<<<<<
 *         if self._revno_first == -1:
 *             if self._revno_second != -1:
*/

static PyObject *__pyx_f_6breezy_16_known_graph_pyx_14_MergeSortNode__revno(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_revno", 0);

  /* "breezy/_known_graph_pyx.pyx":696
 * 
 *     cdef object _revno(self):
 *         if self._revno_first == -1:             # <<<<<<<<<<<<<<
 *             if self._revno_second != -1:
 *                 raise RuntimeError('Something wrong with: %s' % (self,))
*/
  __pyx_t_1 = (__pyx_v_self->_revno_first == -1L);
  if (__pyx_t_1) {

    /* "breezy/_known_graph_pyx.pyx":697
 *     cdef object _revno(self):
 *         if self._revno_first == -1:
 *             if self._revno_second != -1:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError('Something wrong with: %s' % (self,))
 *             return (self._revno_last,)
*/
    __pyx_t_1 = (__pyx_v_self->_revno_second != -1L);
    if (unlikely(__pyx_t_1)) {

      /* "breezy/_known_graph_pyx.pyx":698
 *         if self._revno_first == -1:
 *             if self._revno_second != -1:
 *                 raise RuntimeError('Something wrong with: %s' % (self,))             # <<<<<<<<<<<<<<
 *             return (self._revno_last,)
 *         else:
*/
      __pyx_t_3 = NULL;
      __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(((PyObject *)__pyx_v_self)), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 698, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_Something_wrong_with, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 698, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
        __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_RuntimeError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 698, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 698, __pyx_L1_error)

      /* "breezy/_known_graph_pyx.pyx":697
 *     cdef object _revno(self):
 *         if self._revno_first == -1:
 *             if self._revno_second != -1:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError('Something wrong with: %s' % (self,))
 *             return (self._revno_last,)
*/
    }

    /* "breezy/_known_graph_pyx.pyx":699
 *             if self._revno_second != -1:
 *                 raise RuntimeError('Something wrong with: %s' % (self,))
 *             return (self._revno_last,)             # <<<<<<<<<<<<<<
 *         else:
 *             return (self._revno_first, self._revno_second, self._revno_last)
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyLong_From_long(__pyx_v_self->_revno_last); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2) != (0)) __PYX_ERR(0, 699, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "breezy/_known_graph_pyx.pyx":696
 * 
 *     cdef object _revno(self):
 *         if self._revno_first == -1:             # <<<<<<<<<<<<<<
 *             if self._revno_second != -1:
 *                 raise RuntimeError('Something wrong with: %s' % (self,))
*/
  }

  /* "breezy/_known_graph_pyx.pyx":701
 *             return (self._revno_last,)
 *         else:
 *             return (self._revno_first, self._revno_second, self._revno_last)             # <<<<<<<<<<<<<<
 * 
 *     property revno:
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __Pyx_PyLong_From_long(__pyx_v_self->_revno_first); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyLong_From_long(__pyx_v_self->_revno_second); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyLong_From_long(__pyx_v_self->_revno_last); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5) != (0)) __PYX_ERR(0, 701, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2) != (0)) __PYX_ERR(0, 701, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3) != (0)) __PYX_ERR(0, 701, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "breezy/_known_graph_pyx.pyx":695
 *         return 0
 * 
 *     cdef object _revno(self):             # <<<<<<<<<<<<<<
 *         if self._revno_first == -1:
 *             if self._revno_second != -1:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSortNode._revno", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":704
 * 
 *     property revno:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._revno()
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_5revno_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_5revno_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_5revno___get__(((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_5revno___get__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "breezy/_known_graph_pyx.pyx":705
 *     property revno:
 *         def __get__(self):
 *             return self._revno()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_v_self->__pyx_vtab)->_revno(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":704
 * 
 *     property revno:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._revno()
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSortNode.revno.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":654
 * 
 *     # Public api
 *     cdef public object key             # <<<<<<<<<<<<<<
 *     cdef public long merge_depth
 *     cdef public object end_of_merge # True/False Is this the end of the current merge
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_3key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_3key_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_3key___get__(((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_3key___get__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->key);
  __pyx_r = __pyx_v_self->key;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_3key_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_3key_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_3key_2__set__(((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_3key_2__set__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->key);
  __Pyx_DECREF(__pyx_v_self->key);
  __pyx_v_self->key = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_3key_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_3key_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_3key_4__del__(((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_3key_4__del__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->key);
  __Pyx_DECREF(__pyx_v_self->key);
  __pyx_v_self->key = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":655
 *     # Public api
 *     cdef public object key
 *     cdef public long merge_depth             # <<<<<<<<<<<<<<
 *     cdef public object end_of_merge # True/False Is this the end of the current merge
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_11merge_depth_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_11merge_depth_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_11merge_depth___get__(((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_11merge_depth___get__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_long(__pyx_v_self->merge_depth); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSortNode.merge_depth.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_11merge_depth_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_11merge_depth_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_11merge_depth_2__set__(((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_11merge_depth_2__set__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyLong_As_long(__pyx_v_value); if (unlikely((__pyx_t_1 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 655, __pyx_L1_error)
  __pyx_v_self->merge_depth = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSortNode.merge_depth.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":656
 *     cdef public object key
 *     cdef public long merge_depth
 *     cdef public object end_of_merge # True/False Is this the end of the current merge             # <<<<<<<<<<<<<<
 * 
 *     # Private api, used while computing the information
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_12end_of_merge_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_12end_of_merge_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_12end_of_merge___get__(((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_12end_of_merge___get__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->end_of_merge);
  __pyx_r = __pyx_v_self->end_of_merge;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_12end_of_merge_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_12end_of_merge_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_12end_of_merge_2__set__(((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_12end_of_merge_2__set__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->end_of_merge);
  __Pyx_DECREF(__pyx_v_self->end_of_merge);
  __pyx_v_self->end_of_merge = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_12end_of_merge_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_12end_of_merge_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_12end_of_merge_4__del__(((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_12end_of_merge_4__del__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->end_of_merge);
  __Pyx_DECREF(__pyx_v_self->end_of_merge);
  __pyx_v_self->end_of_merge = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_14_MergeSortNode_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_4__reduce_cython__(((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_4__reduce_cython__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._revno_first, self._revno_last, self._revno_second, self.completed, self.end_of_merge, self.is_first_child, self.key, self.left_parent, self.left_pending_parent, self.merge_depth, self.pending_parents, self.seen_by_child)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __pyx_t_1 = __Pyx_PyLong_From_long(__pyx_v_self->_revno_first); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_From_long(__pyx_v_self->_revno_last); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyLong_From_long(__pyx_v_self->_revno_second); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyLong_From_int(__pyx_v_self->completed); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyLong_From_int(__pyx_v_self->is_first_child); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyLong_From_long(__pyx_v_self->merge_depth); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyLong_From_int(__pyx_v_self->seen_by_child); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyTuple_New(12); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_3) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_t_4) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->end_of_merge);
  __Pyx_GIVEREF(__pyx_v_self->end_of_merge);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 4, __pyx_v_self->end_of_merge) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 5, __pyx_t_5) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->key);
  __Pyx_GIVEREF(__pyx_v_self->key);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 6, __pyx_v_self->key) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->left_parent);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->left_parent);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 7, ((PyObject *)__pyx_v_self->left_parent)) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->left_pending_parent);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->left_pending_parent);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 8, ((PyObject *)__pyx_v_self->left_pending_parent)) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 9, __pyx_t_6) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->pending_parents);
  __Pyx_GIVEREF(__pyx_v_self->pending_parents);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 10, __pyx_v_self->pending_parents) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 11, __pyx_t_7) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._revno_first, self._revno_last, self._revno_second, self.completed, self.end_of_merge, self.is_first_child, self.key, self.left_parent, self.left_pending_parent, self.merge_depth, self.pending_parents, self.seen_by_child)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_8 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_v__dict = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "(tree fragment)":7
 *     state = (self._revno_first, self._revno_last, self._revno_second, self.completed, self.end_of_merge, self.is_first_child, self.key, self.left_parent, self.left_pending_parent, self.merge_depth, self.pending_parents, self.seen_by_child)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_10 = (__pyx_v__dict != Py_None);
  if (__pyx_t_10) {
  } else {
    __pyx_t_9 = __pyx_t_10;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(1, 7, __pyx_L1_error)
  __pyx_t_9 = __pyx_t_10;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_9) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v__dict) != (0)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_7 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.end_of_merge is not None or self.key is not None or self.left_parent is not None or self.left_pending_parent is not None or self.pending_parents is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._revno_first, self._revno_last, self._revno_second, self.completed, self.end_of_merge, self.is_first_child, self.key, self.left_parent, self.left_pending_parent, self.merge_depth, self.pending_parents, self.seen_by_child)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.end_of_merge is not None or self.key is not None or self.left_parent is not None or self.left_pending_parent is not None or self.pending_parents is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__MergeSortNode, (type(self), 0x30884b3, None), state
*/
  /*else*/ {
    __pyx_t_10 = (__pyx_v_self->end_of_merge != Py_None);
    if (!__pyx_t_10) {
    } else {
      __pyx_t_9 = __pyx_t_10;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_10 = (__pyx_v_self->key != Py_None);
    if (!__pyx_t_10) {
    } else {
      __pyx_t_9 = __pyx_t_10;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_10 = (((PyObject *)__pyx_v_self->left_parent) != Py_None);
    if (!__pyx_t_10) {
    } else {
      __pyx_t_9 = __pyx_t_10;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_10 = (((PyObject *)__pyx_v_self->left_pending_parent) != Py_None);
    if (!__pyx_t_10) {
    } else {
      __pyx_t_9 = __pyx_t_10;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_10 = (__pyx_v_self->pending_parents != Py_None);
    __pyx_t_9 = __pyx_t_10;
    __pyx_L6_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_9;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.end_of_merge is not None or self.key is not None or self.left_parent is not None or self.left_pending_parent is not None or self.pending_parents is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__MergeSortNode, (type(self), 0x30884b3, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.end_of_merge is not None or self.key is not None or self.left_parent is not None or self.left_pending_parent is not None or self.pending_parents is not None
 *     if use_setstate:
 *         return __pyx_unpickle__MergeSortNode, (type(self), 0x30884b3, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__MergeSortNode, (type(self), 0x30884b3, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__MergeSortNode); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_50889907);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_50889907);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_mstate_global->__pyx_int_50889907) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 2, Py_None) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_8) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_7 = 0;
    __pyx_t_8 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.end_of_merge is not None or self.key is not None or self.left_parent is not None or self.left_pending_parent is not None or self.pending_parents is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__MergeSortNode, (type(self), 0x30884b3, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__MergeSortNode, (type(self), 0x30884b3, None), state
 *     else:
 *         return __pyx_unpickle__MergeSortNode, (type(self), 0x30884b3, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__MergeSortNode__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__MergeSortNode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_50889907);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_50889907);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_mstate_global->__pyx_int_50889907) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_8) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_6 = 0;
    __pyx_t_8 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSortNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__MergeSortNode, (type(self), 0x30884b3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__MergeSortNode__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_14_MergeSortNode_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSortNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_6__setstate_cython__(((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_14_MergeSortNode_6__setstate_cython__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__MergeSortNode, (type(self), 0x30884b3, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__MergeSortNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(1, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_6breezy_16_known_graph_pyx___pyx_unpickle__MergeSortNode__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__MergeSortNode, (type(self), 0x30884b3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__MergeSortNode__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSortNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":727
 *     cdef object _scheduled_nodes # List of nodes ready to be yielded
 * 
 *     def __init__(self, known_graph, tip_key):             # <<<<<<<<<<<<<<
 *         cdef _KnownGraphNode node
 * 
*/

/* Python wrapper */
static int __pyx_pw_6breezy_16_known_graph_pyx_12_MergeSorter_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6breezy_16_known_graph_pyx_12_MergeSorter_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_known_graph = 0;
  PyObject *__pyx_v_tip_key = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_known_graph,&__pyx_mstate_global->__pyx_n_u_tip_key,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 727, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 727, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 727, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 727, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, i); __PYX_ERR(0, 727, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 727, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 727, __pyx_L3_error)
    }
    __pyx_v_known_graph = values[0];
    __pyx_v_tip_key = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 727, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSorter.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_12_MergeSorter___init__(((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *)__pyx_v_self), __pyx_v_known_graph, __pyx_v_tip_key);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_16_known_graph_pyx_12_MergeSorter___init__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *__pyx_v_self, PyObject *__pyx_v_known_graph, PyObject *__pyx_v_tip_key) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "breezy/_known_graph_pyx.pyx":730
 *         cdef _KnownGraphNode node
 * 
 *         self.graph = known_graph             # <<<<<<<<<<<<<<
 *         # self._ms_nodes = {}
 *         self._revno_to_branch_count = {}
*/
  __pyx_t_1 = __pyx_v_known_graph;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph))))) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->graph);
  __Pyx_DECREF((PyObject *)__pyx_v_self->graph);
  __pyx_v_self->graph = ((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":732
 *         self.graph = known_graph
 *         # self._ms_nodes = {}
 *         self._revno_to_branch_count = {}             # <<<<<<<<<<<<<<
 *         self._depth_first_stack = []
 *         self._last_stack_item = -1
*/
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_revno_to_branch_count);
  __Pyx_DECREF(__pyx_v_self->_revno_to_branch_count);
  __pyx_v_self->_revno_to_branch_count = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":733
 *         # self._ms_nodes = {}
 *         self._revno_to_branch_count = {}
 *         self._depth_first_stack = []             # <<<<<<<<<<<<<<
 *         self._last_stack_item = -1
 *         self._scheduled_nodes = []
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 733, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_depth_first_stack);
  __Pyx_DECREF(__pyx_v_self->_depth_first_stack);
  __pyx_v_self->_depth_first_stack = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":734
 *         self._revno_to_branch_count = {}
 *         self._depth_first_stack = []
 *         self._last_stack_item = -1             # <<<<<<<<<<<<<<
 *         self._scheduled_nodes = []
 *         if (tip_key is not None and tip_key != NULL_REVISION
*/
  __pyx_v_self->_last_stack_item = -1L;

  /* "breezy/_known_graph_pyx.pyx":735
 *         self._depth_first_stack = []
 *         self._last_stack_item = -1
 *         self._scheduled_nodes = []             # <<<<<<<<<<<<<<
 *         if (tip_key is not None and tip_key != NULL_REVISION
 *             and tip_key != (NULL_REVISION,)):
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_scheduled_nodes);
  __Pyx_DECREF(__pyx_v_self->_scheduled_nodes);
  __pyx_v_self->_scheduled_nodes = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":736
 *         self._last_stack_item = -1
 *         self._scheduled_nodes = []
 *         if (tip_key is not None and tip_key != NULL_REVISION             # <<<<<<<<<<<<<<
 *             and tip_key != (NULL_REVISION,)):
 *             node = self.graph._nodes[tip_key]
*/
  __pyx_t_3 = (__pyx_v_tip_key != Py_None);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }

  /* "breezy/_known_graph_pyx.pyx":737
 *         self._scheduled_nodes = []
 *         if (tip_key is not None and tip_key != NULL_REVISION
 *             and tip_key != (NULL_REVISION,)):             # <<<<<<<<<<<<<<
 *             node = self.graph._nodes[tip_key]
 *             self._push_node(node, 0)
*/
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_tip_key, __pyx_v_6breezy_16_known_graph_pyx_NULL_REVISION, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 736, __pyx_L1_error)

  /* "breezy/_known_graph_pyx.pyx":736
 *         self._last_stack_item = -1
 *         self._scheduled_nodes = []
 *         if (tip_key is not None and tip_key != NULL_REVISION             # <<<<<<<<<<<<<<
 *             and tip_key != (NULL_REVISION,)):
 *             node = self.graph._nodes[tip_key]
*/
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }

  /* "breezy/_known_graph_pyx.pyx":737
 *         self._scheduled_nodes = []
 *         if (tip_key is not None and tip_key != NULL_REVISION
 *             and tip_key != (NULL_REVISION,)):             # <<<<<<<<<<<<<<
 *             node = self.graph._nodes[tip_key]
 *             self._push_node(node, 0)
*/
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 737, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_6breezy_16_known_graph_pyx_NULL_REVISION);
  __Pyx_GIVEREF(__pyx_v_6breezy_16_known_graph_pyx_NULL_REVISION);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_6breezy_16_known_graph_pyx_NULL_REVISION) != (0)) __PYX_ERR(0, 737, __pyx_L1_error);
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_tip_key, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 737, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 737, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;

  /* "breezy/_known_graph_pyx.pyx":736
 *         self._last_stack_item = -1
 *         self._scheduled_nodes = []
 *         if (tip_key is not None and tip_key != NULL_REVISION             # <<<<<<<<<<<<<<
 *             and tip_key != (NULL_REVISION,)):
 *             node = self.graph._nodes[tip_key]
*/
  if (__pyx_t_2) {

    /* "breezy/_known_graph_pyx.pyx":738
 *         if (tip_key is not None and tip_key != NULL_REVISION
 *             and tip_key != (NULL_REVISION,)):
 *             node = self.graph._nodes[tip_key]             # <<<<<<<<<<<<<<
 *             self._push_node(node, 0)
 * 
*/
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_self->graph->_nodes, __pyx_v_tip_key); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode))))) __PYX_ERR(0, 738, __pyx_L1_error)
    __pyx_v_node = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "breezy/_known_graph_pyx.pyx":739
 *             and tip_key != (NULL_REVISION,)):
 *             node = self.graph._nodes[tip_key]
 *             self._push_node(node, 0)             # <<<<<<<<<<<<<<
 * 
 *     cdef _MergeSortNode _get_ms_node(self, _KnownGraphNode node):
*/
    __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSorter *)__pyx_v_self->__pyx_vtab)->_push_node(__pyx_v_self, __pyx_v_node, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 739, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "breezy/_known_graph_pyx.pyx":736
 *         self._last_stack_item = -1
 *         self._scheduled_nodes = []
 *         if (tip_key is not None and tip_key != NULL_REVISION             # <<<<<<<<<<<<<<
 *             and tip_key != (NULL_REVISION,)):
 *             node = self.graph._nodes[tip_key]
*/
  }

  /* "breezy/_known_graph_pyx.pyx":727
 *     cdef object _scheduled_nodes # List of nodes ready to be yielded
 * 
 *     def __init__(self, known_graph, tip_key):             # <<<<<<<<<<<<<<
 *         cdef _KnownGraphNode node
 * 
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSorter.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":741
 *             self._push_node(node, 0)
 * 
 *     cdef _MergeSortNode _get_ms_node(self, _KnownGraphNode node):             # <<<<<<<<<<<<<<
 *         cdef PyObject *temp_node
 *         cdef _MergeSortNode ms_node
*/

static struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_f_6breezy_16_known_graph_pyx_12_MergeSorter__get_ms_node(CYTHON_UNUSED struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *__pyx_v_self, struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_ms_node = 0;
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_ms_node", 0);

  /* "breezy/_known_graph_pyx.pyx":745
 *         cdef _MergeSortNode ms_node
 * 
 *         if node.extra is None:             # <<<<<<<<<<<<<<
 *             ms_node = _MergeSortNode(node.key)
 *             node.extra = ms_node
*/
  __pyx_t_1 = (__pyx_v_node->extra == Py_None);
  if (__pyx_t_1) {

    /* "breezy/_known_graph_pyx.pyx":746
 * 
 *         if node.extra is None:
 *             ms_node = _MergeSortNode(node.key)             # <<<<<<<<<<<<<<
 *             node.extra = ms_node
 *         else:
*/
    __pyx_t_3 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_node->key};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 746, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_2);
    }
    __pyx_v_ms_node = ((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "breezy/_known_graph_pyx.pyx":747
 *         if node.extra is None:
 *             ms_node = _MergeSortNode(node.key)
 *             node.extra = ms_node             # <<<<<<<<<<<<<<
 *         else:
 *             ms_node = <_MergeSortNode>node.extra
*/
    __Pyx_INCREF((PyObject *)__pyx_v_ms_node);
    __Pyx_GIVEREF((PyObject *)__pyx_v_ms_node);
    __Pyx_GOTREF(__pyx_v_node->extra);
    __Pyx_DECREF(__pyx_v_node->extra);
    __pyx_v_node->extra = ((PyObject *)__pyx_v_ms_node);

    /* "breezy/_known_graph_pyx.pyx":745
 *         cdef _MergeSortNode ms_node
 * 
 *         if node.extra is None:             # <<<<<<<<<<<<<<
 *             ms_node = _MergeSortNode(node.key)
 *             node.extra = ms_node
*/
    goto __pyx_L3;
  }

  /* "breezy/_known_graph_pyx.pyx":749
 *             node.extra = ms_node
 *         else:
 *             ms_node = <_MergeSortNode>node.extra             # <<<<<<<<<<<<<<
 *         return ms_node
 * 
*/
  /*else*/ {
    __pyx_t_2 = __pyx_v_node->extra;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_ms_node = ((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_t_2);
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "breezy/_known_graph_pyx.pyx":750
 *         else:
 *             ms_node = <_MergeSortNode>node.extra
 *         return ms_node             # <<<<<<<<<<<<<<
 * 
 *     cdef _push_node(self, _KnownGraphNode node, long merge_depth):
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_ms_node);
  __pyx_r = __pyx_v_ms_node;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":741
 *             self._push_node(node, 0)
 * 
 *     cdef _MergeSortNode _get_ms_node(self, _KnownGraphNode node):             # <<<<<<<<<<<<<<
 *         cdef PyObject *temp_node
 *         cdef _MergeSortNode ms_node
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSorter._get_ms_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ms_node);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":752
 *         return ms_node
 * 
 *     cdef _push_node(self, _KnownGraphNode node, long merge_depth):             # <<<<<<<<<<<<<<
 *         cdef _KnownGraphNode parent_node
 *         cdef _MergeSortNode ms_node, ms_parent_node
*/

static PyObject *__pyx_f_6breezy_16_known_graph_pyx_12_MergeSorter__push_node(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *__pyx_v_self, struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node, long __pyx_v_merge_depth) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_parent_node = 0;
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_ms_node = 0;
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_ms_parent_node = 0;
  Py_ssize_t __pyx_v_pos;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_push_node", 0);

  /* "breezy/_known_graph_pyx.pyx":757
 *         cdef Py_ssize_t pos
 * 
 *         ms_node = self._get_ms_node(node)             # <<<<<<<<<<<<<<
 *         ms_node.merge_depth = merge_depth
 *         if node.parents is None:
*/
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSorter *)__pyx_v_self->__pyx_vtab)->_get_ms_node(__pyx_v_self, __pyx_v_node)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ms_node = ((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":758
 * 
 *         ms_node = self._get_ms_node(node)
 *         ms_node.merge_depth = merge_depth             # <<<<<<<<<<<<<<
 *         if node.parents is None:
 *             raise RuntimeError('ghost nodes should not be pushed'
*/
  __pyx_v_ms_node->merge_depth = __pyx_v_merge_depth;

  /* "breezy/_known_graph_pyx.pyx":759
 *         ms_node = self._get_ms_node(node)
 *         ms_node.merge_depth = merge_depth
 *         if node.parents is None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('ghost nodes should not be pushed'
 *                                ' onto the stack: %s' % (node,))
*/
  __pyx_t_2 = (__pyx_v_node->parents == Py_None);
  if (unlikely(__pyx_t_2)) {

    /* "breezy/_known_graph_pyx.pyx":760
 *         ms_node.merge_depth = merge_depth
 *         if node.parents is None:
 *             raise RuntimeError('ghost nodes should not be pushed'             # <<<<<<<<<<<<<<
 *                                ' onto the stack: %s' % (node,))
 *         if PyTuple_GET_SIZE(node.parents) > 0:
*/
    __pyx_t_3 = NULL;

    /* "breezy/_known_graph_pyx.pyx":761
 *         if node.parents is None:
 *             raise RuntimeError('ghost nodes should not be pushed'
 *                                ' onto the stack: %s' % (node,))             # <<<<<<<<<<<<<<
 *         if PyTuple_GET_SIZE(node.parents) > 0:
 *             parent_node = _get_tuple_node(node.parents, 0)
*/
    __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(((PyObject *)__pyx_v_node)), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 761, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "breezy/_known_graph_pyx.pyx":760
 *         ms_node.merge_depth = merge_depth
 *         if node.parents is None:
 *             raise RuntimeError('ghost nodes should not be pushed'             # <<<<<<<<<<<<<<
 *                                ' onto the stack: %s' % (node,))
 *         if PyTuple_GET_SIZE(node.parents) > 0:
*/
    __pyx_t_5 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_ghost_nodes_should_not_be_pushed, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 760, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_RuntimeError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 760, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 760, __pyx_L1_error)

    /* "breezy/_known_graph_pyx.pyx":759
 *         ms_node = self._get_ms_node(node)
 *         ms_node.merge_depth = merge_depth
 *         if node.parents is None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('ghost nodes should not be pushed'
 *                                ' onto the stack: %s' % (node,))
*/
  }

  /* "breezy/_known_graph_pyx.pyx":762
 *             raise RuntimeError('ghost nodes should not be pushed'
 *                                ' onto the stack: %s' % (node,))
 *         if PyTuple_GET_SIZE(node.parents) > 0:             # <<<<<<<<<<<<<<
 *             parent_node = _get_tuple_node(node.parents, 0)
 *             ms_node.left_parent = parent_node
*/
  __pyx_t_1 = __pyx_v_node->parents;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = (PyTuple_GET_SIZE(__pyx_t_1) > 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "breezy/_known_graph_pyx.pyx":763
 *                                ' onto the stack: %s' % (node,))
 *         if PyTuple_GET_SIZE(node.parents) > 0:
 *             parent_node = _get_tuple_node(node.parents, 0)             # <<<<<<<<<<<<<<
 *             ms_node.left_parent = parent_node
 *             if parent_node.parents is None: # left-hand ghost
*/
    __pyx_t_1 = __pyx_v_node->parents;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_tuple_node(__pyx_t_1, 0)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 763, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_parent_node = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "breezy/_known_graph_pyx.pyx":764
 *         if PyTuple_GET_SIZE(node.parents) > 0:
 *             parent_node = _get_tuple_node(node.parents, 0)
 *             ms_node.left_parent = parent_node             # <<<<<<<<<<<<<<
 *             if parent_node.parents is None: # left-hand ghost
 *                 ms_node.left_pending_parent = None
*/
    __Pyx_INCREF((PyObject *)__pyx_v_parent_node);
    __Pyx_GIVEREF((PyObject *)__pyx_v_parent_node);
    __Pyx_GOTREF((PyObject *)__pyx_v_ms_node->left_parent);
    __Pyx_DECREF((PyObject *)__pyx_v_ms_node->left_parent);
    __pyx_v_ms_node->left_parent = __pyx_v_parent_node;

    /* "breezy/_known_graph_pyx.pyx":765
 *             parent_node = _get_tuple_node(node.parents, 0)
 *             ms_node.left_parent = parent_node
 *             if parent_node.parents is None: # left-hand ghost             # <<<<<<<<<<<<<<
 *                 ms_node.left_pending_parent = None
 *                 ms_node.left_parent = None
*/
    __pyx_t_2 = (__pyx_v_parent_node->parents == Py_None);
    if (__pyx_t_2) {

      /* "breezy/_known_graph_pyx.pyx":766
 *             ms_node.left_parent = parent_node
 *             if parent_node.parents is None: # left-hand ghost
 *                 ms_node.left_pending_parent = None             # <<<<<<<<<<<<<<
 *                 ms_node.left_parent = None
 *             else:
*/
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_GOTREF((PyObject *)__pyx_v_ms_node->left_pending_parent);
      __Pyx_DECREF((PyObject *)__pyx_v_ms_node->left_pending_parent);
      __pyx_v_ms_node->left_pending_parent = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)Py_None);

      /* "breezy/_known_graph_pyx.pyx":767
 *             if parent_node.parents is None: # left-hand ghost
 *                 ms_node.left_pending_parent = None
 *                 ms_node.left_parent = None             # <<<<<<<<<<<<<<
 *             else:
 *                 ms_node.left_pending_parent = parent_node
*/
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_GOTREF((PyObject *)__pyx_v_ms_node->left_parent);
      __Pyx_DECREF((PyObject *)__pyx_v_ms_node->left_parent);
      __pyx_v_ms_node->left_parent = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)Py_None);

      /* "breezy/_known_graph_pyx.pyx":765
 *             parent_node = _get_tuple_node(node.parents, 0)
 *             ms_node.left_parent = parent_node
 *             if parent_node.parents is None: # left-hand ghost             # <<<<<<<<<<<<<<
 *                 ms_node.left_pending_parent = None
 *                 ms_node.left_parent = None
*/
      goto __pyx_L5;
    }

    /* "breezy/_known_graph_pyx.pyx":769
 *                 ms_node.left_parent = None
 *             else:
 *                 ms_node.left_pending_parent = parent_node             # <<<<<<<<<<<<<<
 *         if PyTuple_GET_SIZE(node.parents) > 1:
 *             ms_node.pending_parents = []
*/
    /*else*/ {
      __Pyx_INCREF((PyObject *)__pyx_v_parent_node);
      __Pyx_GIVEREF((PyObject *)__pyx_v_parent_node);
      __Pyx_GOTREF((PyObject *)__pyx_v_ms_node->left_pending_parent);
      __Pyx_DECREF((PyObject *)__pyx_v_ms_node->left_pending_parent);
      __pyx_v_ms_node->left_pending_parent = __pyx_v_parent_node;
    }
    __pyx_L5:;

    /* "breezy/_known_graph_pyx.pyx":762
 *             raise RuntimeError('ghost nodes should not be pushed'
 *                                ' onto the stack: %s' % (node,))
 *         if PyTuple_GET_SIZE(node.parents) > 0:             # <<<<<<<<<<<<<<
 *             parent_node = _get_tuple_node(node.parents, 0)
 *             ms_node.left_parent = parent_node
*/
  }

  /* "breezy/_known_graph_pyx.pyx":770
 *             else:
 *                 ms_node.left_pending_parent = parent_node
 *         if PyTuple_GET_SIZE(node.parents) > 1:             # <<<<<<<<<<<<<<
 *             ms_node.pending_parents = []
 *             for pos from 1 <= pos < PyTuple_GET_SIZE(node.parents):
*/
  __pyx_t_5 = __pyx_v_node->parents;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_2 = (PyTuple_GET_SIZE(__pyx_t_5) > 1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_2) {

    /* "breezy/_known_graph_pyx.pyx":771
 *                 ms_node.left_pending_parent = parent_node
 *         if PyTuple_GET_SIZE(node.parents) > 1:
 *             ms_node.pending_parents = []             # <<<<<<<<<<<<<<
 *             for pos from 1 <= pos < PyTuple_GET_SIZE(node.parents):
 *                 parent_node = _get_tuple_node(node.parents, pos)
*/
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 771, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_v_ms_node->pending_parents);
    __Pyx_DECREF(__pyx_v_ms_node->pending_parents);
    __pyx_v_ms_node->pending_parents = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "breezy/_known_graph_pyx.pyx":772
 *         if PyTuple_GET_SIZE(node.parents) > 1:
 *             ms_node.pending_parents = []
 *             for pos from 1 <= pos < PyTuple_GET_SIZE(node.parents):             # <<<<<<<<<<<<<<
 *                 parent_node = _get_tuple_node(node.parents, pos)
 *                 if parent_node.parents is None: # ghost
*/
    __pyx_t_5 = __pyx_v_node->parents;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_7 = PyTuple_GET_SIZE(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (__pyx_v_pos = 1; __pyx_v_pos < __pyx_t_7; __pyx_v_pos++) {

      /* "breezy/_known_graph_pyx.pyx":773
 *             ms_node.pending_parents = []
 *             for pos from 1 <= pos < PyTuple_GET_SIZE(node.parents):
 *                 parent_node = _get_tuple_node(node.parents, pos)             # <<<<<<<<<<<<<<
 *                 if parent_node.parents is None: # ghost
 *                     continue
*/
      __pyx_t_5 = __pyx_v_node->parents;
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_1 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_tuple_node(__pyx_t_5, __pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_parent_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "breezy/_known_graph_pyx.pyx":774
 *             for pos from 1 <= pos < PyTuple_GET_SIZE(node.parents):
 *                 parent_node = _get_tuple_node(node.parents, pos)
 *                 if parent_node.parents is None: # ghost             # <<<<<<<<<<<<<<
 *                     continue
 *                 PyList_Append(ms_node.pending_parents, parent_node)
*/
      __pyx_t_2 = (__pyx_v_parent_node->parents == Py_None);
      if (__pyx_t_2) {

        /* "breezy/_known_graph_pyx.pyx":775
 *                 parent_node = _get_tuple_node(node.parents, pos)
 *                 if parent_node.parents is None: # ghost
 *                     continue             # <<<<<<<<<<<<<<
 *                 PyList_Append(ms_node.pending_parents, parent_node)
 * 
*/
        goto __pyx_L7_continue;

        /* "breezy/_known_graph_pyx.pyx":774
 *             for pos from 1 <= pos < PyTuple_GET_SIZE(node.parents):
 *                 parent_node = _get_tuple_node(node.parents, pos)
 *                 if parent_node.parents is None: # ghost             # <<<<<<<<<<<<<<
 *                     continue
 *                 PyList_Append(ms_node.pending_parents, parent_node)
*/
      }

      /* "breezy/_known_graph_pyx.pyx":776
 *                 if parent_node.parents is None: # ghost
 *                     continue
 *                 PyList_Append(ms_node.pending_parents, parent_node)             # <<<<<<<<<<<<<<
 * 
 *         ms_node.is_first_child = 1
*/
      __pyx_t_1 = __pyx_v_ms_node->pending_parents;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_8 = PyList_Append(__pyx_t_1, ((PyObject *)__pyx_v_parent_node)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 776, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_L7_continue:;
    }

    /* "breezy/_known_graph_pyx.pyx":770
 *             else:
 *                 ms_node.left_pending_parent = parent_node
 *         if PyTuple_GET_SIZE(node.parents) > 1:             # <<<<<<<<<<<<<<
 *             ms_node.pending_parents = []
 *             for pos from 1 <= pos < PyTuple_GET_SIZE(node.parents):
*/
  }

  /* "breezy/_known_graph_pyx.pyx":778
 *                 PyList_Append(ms_node.pending_parents, parent_node)
 * 
 *         ms_node.is_first_child = 1             # <<<<<<<<<<<<<<
 *         if ms_node.left_parent is not None:
 *             ms_parent_node = self._get_ms_node(ms_node.left_parent)
*/
  __pyx_v_ms_node->is_first_child = 1;

  /* "breezy/_known_graph_pyx.pyx":779
 * 
 *         ms_node.is_first_child = 1
 *         if ms_node.left_parent is not None:             # <<<<<<<<<<<<<<
 *             ms_parent_node = self._get_ms_node(ms_node.left_parent)
 *             if ms_parent_node.seen_by_child:
*/
  __pyx_t_2 = (((PyObject *)__pyx_v_ms_node->left_parent) != Py_None);
  if (__pyx_t_2) {

    /* "breezy/_known_graph_pyx.pyx":780
 *         ms_node.is_first_child = 1
 *         if ms_node.left_parent is not None:
 *             ms_parent_node = self._get_ms_node(ms_node.left_parent)             # <<<<<<<<<<<<<<
 *             if ms_parent_node.seen_by_child:
 *                 ms_node.is_first_child = 0
*/
    __pyx_t_1 = ((PyObject *)__pyx_v_ms_node->left_parent);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSorter *)__pyx_v_self->__pyx_vtab)->_get_ms_node(__pyx_v_self, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 780, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_ms_parent_node = ((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "breezy/_known_graph_pyx.pyx":781
 *         if ms_node.left_parent is not None:
 *             ms_parent_node = self._get_ms_node(ms_node.left_parent)
 *             if ms_parent_node.seen_by_child:             # <<<<<<<<<<<<<<
 *                 ms_node.is_first_child = 0
 *             ms_parent_node.seen_by_child = 1
*/
    __pyx_t_2 = (__pyx_v_ms_parent_node->seen_by_child != 0);
    if (__pyx_t_2) {

      /* "breezy/_known_graph_pyx.pyx":782
 *             ms_parent_node = self._get_ms_node(ms_node.left_parent)
 *             if ms_parent_node.seen_by_child:
 *                 ms_node.is_first_child = 0             # <<<<<<<<<<<<<<
 *             ms_parent_node.seen_by_child = 1
 *         self._last_stack_item = self._last_stack_item + 1
*/
      __pyx_v_ms_node->is_first_child = 0;

      /* "breezy/_known_graph_pyx.pyx":781
 *         if ms_node.left_parent is not None:
 *             ms_parent_node = self._get_ms_node(ms_node.left_parent)
 *             if ms_parent_node.seen_by_child:             # <<<<<<<<<<<<<<
 *                 ms_node.is_first_child = 0
 *             ms_parent_node.seen_by_child = 1
*/
    }

    /* "breezy/_known_graph_pyx.pyx":783
 *             if ms_parent_node.seen_by_child:
 *                 ms_node.is_first_child = 0
 *             ms_parent_node.seen_by_child = 1             # <<<<<<<<<<<<<<
 *         self._last_stack_item = self._last_stack_item + 1
 *         if self._last_stack_item < PyList_GET_SIZE(self._depth_first_stack):
*/
    __pyx_v_ms_parent_node->seen_by_child = 1;

    /* "breezy/_known_graph_pyx.pyx":779
 * 
 *         ms_node.is_first_child = 1
 *         if ms_node.left_parent is not None:             # <<<<<<<<<<<<<<
 *             ms_parent_node = self._get_ms_node(ms_node.left_parent)
 *             if ms_parent_node.seen_by_child:
*/
  }

  /* "breezy/_known_graph_pyx.pyx":784
 *                 ms_node.is_first_child = 0
 *             ms_parent_node.seen_by_child = 1
 *         self._last_stack_item = self._last_stack_item + 1             # <<<<<<<<<<<<<<
 *         if self._last_stack_item < PyList_GET_SIZE(self._depth_first_stack):
 *             Py_INCREF(node) # SetItem steals a ref
*/
  __pyx_v_self->_last_stack_item = (__pyx_v_self->_last_stack_item + 1);

  /* "breezy/_known_graph_pyx.pyx":785
 *             ms_parent_node.seen_by_child = 1
 *         self._last_stack_item = self._last_stack_item + 1
 *         if self._last_stack_item < PyList_GET_SIZE(self._depth_first_stack):             # <<<<<<<<<<<<<<
 *             Py_INCREF(node) # SetItem steals a ref
 *             PyList_SetItem(self._depth_first_stack, self._last_stack_item,
*/
  __pyx_t_5 = __pyx_v_self->_depth_first_stack;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_2 = (__pyx_v_self->_last_stack_item < PyList_GET_SIZE(__pyx_t_5));
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_2) {

    /* "breezy/_known_graph_pyx.pyx":786
 *         self._last_stack_item = self._last_stack_item + 1
 *         if self._last_stack_item < PyList_GET_SIZE(self._depth_first_stack):
 *             Py_INCREF(node) # SetItem steals a ref             # <<<<<<<<<<<<<<
 *             PyList_SetItem(self._depth_first_stack, self._last_stack_item,
 *                            node)
*/
    Py_INCREF(((PyObject *)__pyx_v_node));

    /* "breezy/_known_graph_pyx.pyx":787
 *         if self._last_stack_item < PyList_GET_SIZE(self._depth_first_stack):
 *             Py_INCREF(node) # SetItem steals a ref
 *             PyList_SetItem(self._depth_first_stack, self._last_stack_item,             # <<<<<<<<<<<<<<
 *                            node)
 *         else:
*/
    __pyx_t_5 = __pyx_v_self->_depth_first_stack;
    __Pyx_INCREF(__pyx_t_5);

    /* "breezy/_known_graph_pyx.pyx":788
 *             Py_INCREF(node) # SetItem steals a ref
 *             PyList_SetItem(self._depth_first_stack, self._last_stack_item,
 *                            node)             # <<<<<<<<<<<<<<
 *         else:
 *             PyList_Append(self._depth_first_stack, node)
*/
    __pyx_t_8 = PyList_SetItem(__pyx_t_5, __pyx_v_self->_last_stack_item, ((PyObject *)__pyx_v_node)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 787, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "breezy/_known_graph_pyx.pyx":785
 *             ms_parent_node.seen_by_child = 1
 *         self._last_stack_item = self._last_stack_item + 1
 *         if self._last_stack_item < PyList_GET_SIZE(self._depth_first_stack):             # <<<<<<<<<<<<<<
 *             Py_INCREF(node) # SetItem steals a ref
 *             PyList_SetItem(self._depth_first_stack, self._last_stack_item,
*/
    goto __pyx_L12;
  }

  /* "breezy/_known_graph_pyx.pyx":790
 *                            node)
 *         else:
 *             PyList_Append(self._depth_first_stack, node)             # <<<<<<<<<<<<<<
 * 
 *     cdef _pop_node(self):
*/
  /*else*/ {
    __pyx_t_5 = __pyx_v_self->_depth_first_stack;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_8 = PyList_Append(__pyx_t_5, ((PyObject *)__pyx_v_node)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_L12:;

  /* "breezy/_known_graph_pyx.pyx":752
 *         return ms_node
 * 
 *     cdef _push_node(self, _KnownGraphNode node, long merge_depth):             # <<<<<<<<<<<<<<
 *         cdef _KnownGraphNode parent_node
 *         cdef _MergeSortNode ms_node, ms_parent_node
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSorter._push_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_parent_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_ms_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_ms_parent_node);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":792
 *             PyList_Append(self._depth_first_stack, node)
 * 
 *     cdef _pop_node(self):             # <<<<<<<<<<<<<<
 *         cdef PyObject *temp
 *         cdef _MergeSortNode ms_node, ms_parent_node, ms_prev_node
*/

static PyObject *__pyx_f_6breezy_16_known_graph_pyx_12_MergeSorter__pop_node(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *__pyx_v_self) {
  PyObject *__pyx_v_temp;
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_ms_node = 0;
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_ms_parent_node = 0;
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_ms_prev_node = 0;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node = 0;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_prev_node = 0;
  long __pyx_v_base_revno;
  PyObject *__pyx_v_branch_count = NULL;
  PyObject *__pyx_v_root_count = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  long __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_pop_node", 0);

  /* "breezy/_known_graph_pyx.pyx":797
 *         cdef _KnownGraphNode node, parent_node, prev_node
 * 
 *         node = _get_list_node(self._depth_first_stack, self._last_stack_item)             # <<<<<<<<<<<<<<
 *         ms_node = <_MergeSortNode>node.extra
 *         self._last_stack_item = self._last_stack_item - 1
*/
  __pyx_t_1 = __pyx_v_self->_depth_first_stack;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_list_node(__pyx_t_1, __pyx_v_self->_last_stack_item)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_node = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "breezy/_known_graph_pyx.pyx":798
 * 
 *         node = _get_list_node(self._depth_first_stack, self._last_stack_item)
 *         ms_node = <_MergeSortNode>node.extra             # <<<<<<<<<<<<<<
 *         self._last_stack_item = self._last_stack_item - 1
 *         if ms_node.left_parent is not None:
*/
  __pyx_t_2 = __pyx_v_node->extra;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_ms_node = ((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "breezy/_known_graph_pyx.pyx":799
 *         node = _get_list_node(self._depth_first_stack, self._last_stack_item)
 *         ms_node = <_MergeSortNode>node.extra
 *         self._last_stack_item = self._last_stack_item - 1             # <<<<<<<<<<<<<<
 *         if ms_node.left_parent is not None:
 *             # Assign the revision number from the left-hand parent
*/
  __pyx_v_self->_last_stack_item = (__pyx_v_self->_last_stack_item - 1);

  /* "breezy/_known_graph_pyx.pyx":800
 *         ms_node = <_MergeSortNode>node.extra
 *         self._last_stack_item = self._last_stack_item - 1
 *         if ms_node.left_parent is not None:             # <<<<<<<<<<<<<<
 *             # Assign the revision number from the left-hand parent
 *             ms_parent_node = <_MergeSortNode>ms_node.left_parent.extra
*/
  __pyx_t_3 = (((PyObject *)__pyx_v_ms_node->left_parent) != Py_None);
  if (__pyx_t_3) {

    /* "breezy/_known_graph_pyx.pyx":802
 *         if ms_node.left_parent is not None:
 *             # Assign the revision number from the left-hand parent
 *             ms_parent_node = <_MergeSortNode>ms_node.left_parent.extra             # <<<<<<<<<<<<<<
 *             if ms_node.is_first_child:
 *                 # First child just increments the final digit
*/
    __pyx_t_2 = __pyx_v_ms_node->left_parent->extra;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_ms_parent_node = ((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "breezy/_known_graph_pyx.pyx":803
 *             # Assign the revision number from the left-hand parent
 *             ms_parent_node = <_MergeSortNode>ms_node.left_parent.extra
 *             if ms_node.is_first_child:             # <<<<<<<<<<<<<<
 *                 # First child just increments the final digit
 *                 ms_node._revno_first = ms_parent_node._revno_first
*/
    __pyx_t_3 = (__pyx_v_ms_node->is_first_child != 0);
    if (__pyx_t_3) {

      /* "breezy/_known_graph_pyx.pyx":805
 *             if ms_node.is_first_child:
 *                 # First child just increments the final digit
 *                 ms_node._revno_first = ms_parent_node._revno_first             # <<<<<<<<<<<<<<
 *                 ms_node._revno_second = ms_parent_node._revno_second
 *                 ms_node._revno_last = ms_parent_node._revno_last + 1
*/
      __pyx_t_4 = __pyx_v_ms_parent_node->_revno_first;
      __pyx_v_ms_node->_revno_first = __pyx_t_4;

      /* "breezy/_known_graph_pyx.pyx":806
 *                 # First child just increments the final digit
 *                 ms_node._revno_first = ms_parent_node._revno_first
 *                 ms_node._revno_second = ms_parent_node._revno_second             # <<<<<<<<<<<<<<
 *                 ms_node._revno_last = ms_parent_node._revno_last + 1
 *             else:
*/
      __pyx_t_4 = __pyx_v_ms_parent_node->_revno_second;
      __pyx_v_ms_node->_revno_second = __pyx_t_4;

      /* "breezy/_known_graph_pyx.pyx":807
 *                 ms_node._revno_first = ms_parent_node._revno_first
 *                 ms_node._revno_second = ms_parent_node._revno_second
 *                 ms_node._revno_last = ms_parent_node._revno_last + 1             # <<<<<<<<<<<<<<
 *             else:
 *                 # Not the first child, make a new branch
*/
      __pyx_v_ms_node->_revno_last = (__pyx_v_ms_parent_node->_revno_last + 1);

      /* "breezy/_known_graph_pyx.pyx":803
 *             # Assign the revision number from the left-hand parent
 *             ms_parent_node = <_MergeSortNode>ms_node.left_parent.extra
 *             if ms_node.is_first_child:             # <<<<<<<<<<<<<<
 *                 # First child just increments the final digit
 *                 ms_node._revno_first = ms_parent_node._revno_first
*/
      goto __pyx_L4;
    }

    /* "breezy/_known_graph_pyx.pyx":811
 *                 # Not the first child, make a new branch
 *                 #  (mainline_revno, branch_count, 1)
 *                 if ms_parent_node._revno_first == -1:             # <<<<<<<<<<<<<<
 *                     # Mainline ancestor, the increment is on the last digit
 *                     base_revno = ms_parent_node._revno_last
*/
    /*else*/ {
      __pyx_t_3 = (__pyx_v_ms_parent_node->_revno_first == -1L);
      if (__pyx_t_3) {

        /* "breezy/_known_graph_pyx.pyx":813
 *                 if ms_parent_node._revno_first == -1:
 *                     # Mainline ancestor, the increment is on the last digit
 *                     base_revno = ms_parent_node._revno_last             # <<<<<<<<<<<<<<
 *                 else:
 *                     base_revno = ms_parent_node._revno_first
*/
        __pyx_t_4 = __pyx_v_ms_parent_node->_revno_last;
        __pyx_v_base_revno = __pyx_t_4;

        /* "breezy/_known_graph_pyx.pyx":811
 *                 # Not the first child, make a new branch
 *                 #  (mainline_revno, branch_count, 1)
 *                 if ms_parent_node._revno_first == -1:             # <<<<<<<<<<<<<<
 *                     # Mainline ancestor, the increment is on the last digit
 *                     base_revno = ms_parent_node._revno_last
*/
        goto __pyx_L5;
      }

      /* "breezy/_known_graph_pyx.pyx":815
 *                     base_revno = ms_parent_node._revno_last
 *                 else:
 *                     base_revno = ms_parent_node._revno_first             # <<<<<<<<<<<<<<
 *                 temp = PyDict_GetItem(self._revno_to_branch_count,
 *                                       base_revno)
*/
      /*else*/ {
        __pyx_t_4 = __pyx_v_ms_parent_node->_revno_first;
        __pyx_v_base_revno = __pyx_t_4;
      }
      __pyx_L5:;

      /* "breezy/_known_graph_pyx.pyx":816
 *                 else:
 *                     base_revno = ms_parent_node._revno_first
 *                 temp = PyDict_GetItem(self._revno_to_branch_count,             # <<<<<<<<<<<<<<
 *                                       base_revno)
 *                 if temp == NULL:
*/
      __pyx_t_2 = __pyx_v_self->_revno_to_branch_count;
      __Pyx_INCREF(__pyx_t_2);

      /* "breezy/_known_graph_pyx.pyx":817
 *                     base_revno = ms_parent_node._revno_first
 *                 temp = PyDict_GetItem(self._revno_to_branch_count,
 *                                       base_revno)             # <<<<<<<<<<<<<<
 *                 if temp == NULL:
 *                     branch_count = 1
*/
      __pyx_t_1 = __Pyx_PyLong_From_long(__pyx_v_base_revno); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 817, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "breezy/_known_graph_pyx.pyx":816
 *                 else:
 *                     base_revno = ms_parent_node._revno_first
 *                 temp = PyDict_GetItem(self._revno_to_branch_count,             # <<<<<<<<<<<<<<
 *                                       base_revno)
 *                 if temp == NULL:
*/
      __pyx_v_temp = PyDict_GetItem(__pyx_t_2, __pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "breezy/_known_graph_pyx.pyx":818
 *                 temp = PyDict_GetItem(self._revno_to_branch_count,
 *                                       base_revno)
 *                 if temp == NULL:             # <<<<<<<<<<<<<<
 *                     branch_count = 1
 *                 else:
*/
      __pyx_t_3 = (__pyx_v_temp == NULL);
      if (__pyx_t_3) {

        /* "breezy/_known_graph_pyx.pyx":819
 *                                       base_revno)
 *                 if temp == NULL:
 *                     branch_count = 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     branch_count = (<object>temp) + 1
*/
        __Pyx_INCREF(__pyx_mstate_global->__pyx_int_1);
        __pyx_v_branch_count = __pyx_mstate_global->__pyx_int_1;

        /* "breezy/_known_graph_pyx.pyx":818
 *                 temp = PyDict_GetItem(self._revno_to_branch_count,
 *                                       base_revno)
 *                 if temp == NULL:             # <<<<<<<<<<<<<<
 *                     branch_count = 1
 *                 else:
*/
        goto __pyx_L6;
      }

      /* "breezy/_known_graph_pyx.pyx":821
 *                     branch_count = 1
 *                 else:
 *                     branch_count = (<object>temp) + 1             # <<<<<<<<<<<<<<
 *                 PyDict_SetItem(self._revno_to_branch_count, base_revno,
 *                                branch_count)
*/
      /*else*/ {
        __pyx_t_1 = __Pyx_PyLong_AddObjC(((PyObject *)__pyx_v_temp), __pyx_mstate_global->__pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 821, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_v_branch_count = __pyx_t_1;
        __pyx_t_1 = 0;
      }
      __pyx_L6:;

      /* "breezy/_known_graph_pyx.pyx":822
 *                 else:
 *                     branch_count = (<object>temp) + 1
 *                 PyDict_SetItem(self._revno_to_branch_count, base_revno,             # <<<<<<<<<<<<<<
 *                                branch_count)
 *                 ms_node._revno_first = base_revno
*/
      __pyx_t_1 = __pyx_v_self->_revno_to_branch_count;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyLong_From_long(__pyx_v_base_revno); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 822, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);

      /* "breezy/_known_graph_pyx.pyx":823
 *                     branch_count = (<object>temp) + 1
 *                 PyDict_SetItem(self._revno_to_branch_count, base_revno,
 *                                branch_count)             # <<<<<<<<<<<<<<
 *                 ms_node._revno_first = base_revno
 *                 ms_node._revno_second = branch_count
*/
      __pyx_t_5 = PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_v_branch_count); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 822, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "breezy/_known_graph_pyx.pyx":824
 *                 PyDict_SetItem(self._revno_to_branch_count, base_revno,
 *                                branch_count)
 *                 ms_node._revno_first = base_revno             # <<<<<<<<<<<<<<
 *                 ms_node._revno_second = branch_count
 *                 ms_node._revno_last = 1
*/
      __pyx_v_ms_node->_revno_first = __pyx_v_base_revno;

      /* "breezy/_known_graph_pyx.pyx":825
 *                                branch_count)
 *                 ms_node._revno_first = base_revno
 *                 ms_node._revno_second = branch_count             # <<<<<<<<<<<<<<
 *                 ms_node._revno_last = 1
 *         else:
*/
      __pyx_t_4 = __Pyx_PyLong_As_long(__pyx_v_branch_count); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 825, __pyx_L1_error)
      __pyx_v_ms_node->_revno_second = __pyx_t_4;

      /* "breezy/_known_graph_pyx.pyx":826
 *                 ms_node._revno_first = base_revno
 *                 ms_node._revno_second = branch_count
 *                 ms_node._revno_last = 1             # <<<<<<<<<<<<<<
 *         else:
 *             temp = PyDict_GetItem(self._revno_to_branch_count, 0)
*/
      __pyx_v_ms_node->_revno_last = 1;
    }
    __pyx_L4:;

    /* "breezy/_known_graph_pyx.pyx":800
 *         ms_node = <_MergeSortNode>node.extra
 *         self._last_stack_item = self._last_stack_item - 1
 *         if ms_node.left_parent is not None:             # <<<<<<<<<<<<<<
 *             # Assign the revision number from the left-hand parent
 *             ms_parent_node = <_MergeSortNode>ms_node.left_parent.extra
*/
    goto __pyx_L3;
  }

  /* "breezy/_known_graph_pyx.pyx":828
 *                 ms_node._revno_last = 1
 *         else:
 *             temp = PyDict_GetItem(self._revno_to_branch_count, 0)             # <<<<<<<<<<<<<<
 *             if temp == NULL:
 *                 # The first root node doesn't have a 3-digit revno
*/
  /*else*/ {
    __pyx_t_2 = __pyx_v_self->_revno_to_branch_count;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_temp = PyDict_GetItem(__pyx_t_2, __pyx_mstate_global->__pyx_int_0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "breezy/_known_graph_pyx.pyx":829
 *         else:
 *             temp = PyDict_GetItem(self._revno_to_branch_count, 0)
 *             if temp == NULL:             # <<<<<<<<<<<<<<
 *                 # The first root node doesn't have a 3-digit revno
 *                 root_count = 0
*/
    __pyx_t_3 = (__pyx_v_temp == NULL);
    if (__pyx_t_3) {

      /* "breezy/_known_graph_pyx.pyx":831
 *             if temp == NULL:
 *                 # The first root node doesn't have a 3-digit revno
 *                 root_count = 0             # <<<<<<<<<<<<<<
 *                 ms_node._revno_first = -1
 *                 ms_node._revno_second = -1
*/
      __Pyx_INCREF(__pyx_mstate_global->__pyx_int_0);
      __pyx_v_root_count = __pyx_mstate_global->__pyx_int_0;

      /* "breezy/_known_graph_pyx.pyx":832
 *                 # The first root node doesn't have a 3-digit revno
 *                 root_count = 0
 *                 ms_node._revno_first = -1             # <<<<<<<<<<<<<<
 *                 ms_node._revno_second = -1
 *                 ms_node._revno_last = 1
*/
      __pyx_v_ms_node->_revno_first = -1L;

      /* "breezy/_known_graph_pyx.pyx":833
 *                 root_count = 0
 *                 ms_node._revno_first = -1
 *                 ms_node._revno_second = -1             # <<<<<<<<<<<<<<
 *                 ms_node._revno_last = 1
 *             else:
*/
      __pyx_v_ms_node->_revno_second = -1L;

      /* "breezy/_known_graph_pyx.pyx":834
 *                 ms_node._revno_first = -1
 *                 ms_node._revno_second = -1
 *                 ms_node._revno_last = 1             # <<<<<<<<<<<<<<
 *             else:
 *                 root_count = (<object>temp) + 1
*/
      __pyx_v_ms_node->_revno_last = 1;

      /* "breezy/_known_graph_pyx.pyx":829
 *         else:
 *             temp = PyDict_GetItem(self._revno_to_branch_count, 0)
 *             if temp == NULL:             # <<<<<<<<<<<<<<
 *                 # The first root node doesn't have a 3-digit revno
 *                 root_count = 0
*/
      goto __pyx_L7;
    }

    /* "breezy/_known_graph_pyx.pyx":836
 *                 ms_node._revno_last = 1
 *             else:
 *                 root_count = (<object>temp) + 1             # <<<<<<<<<<<<<<
 *                 ms_node._revno_first = 0
 *                 ms_node._revno_second = root_count
*/
    /*else*/ {
      __pyx_t_2 = __Pyx_PyLong_AddObjC(((PyObject *)__pyx_v_temp), __pyx_mstate_global->__pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_root_count = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "breezy/_known_graph_pyx.pyx":837
 *             else:
 *                 root_count = (<object>temp) + 1
 *                 ms_node._revno_first = 0             # <<<<<<<<<<<<<<
 *                 ms_node._revno_second = root_count
 *                 ms_node._revno_last = 1
*/
      __pyx_v_ms_node->_revno_first = 0;

      /* "breezy/_known_graph_pyx.pyx":838
 *                 root_count = (<object>temp) + 1
 *                 ms_node._revno_first = 0
 *                 ms_node._revno_second = root_count             # <<<<<<<<<<<<<<
 *                 ms_node._revno_last = 1
 *             PyDict_SetItem(self._revno_to_branch_count, 0, root_count)
*/
      __pyx_t_4 = __Pyx_PyLong_As_long(__pyx_v_root_count); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 838, __pyx_L1_error)
      __pyx_v_ms_node->_revno_second = __pyx_t_4;

      /* "breezy/_known_graph_pyx.pyx":839
 *                 ms_node._revno_first = 0
 *                 ms_node._revno_second = root_count
 *                 ms_node._revno_last = 1             # <<<<<<<<<<<<<<
 *             PyDict_SetItem(self._revno_to_branch_count, 0, root_count)
 *         ms_node.completed = 1
*/
      __pyx_v_ms_node->_revno_last = 1;
    }
    __pyx_L7:;

    /* "breezy/_known_graph_pyx.pyx":840
 *                 ms_node._revno_second = root_count
 *                 ms_node._revno_last = 1
 *             PyDict_SetItem(self._revno_to_branch_count, 0, root_count)             # <<<<<<<<<<<<<<
 *         ms_node.completed = 1
 *         if PyList_GET_SIZE(self._scheduled_nodes) == 0:
*/
    __pyx_t_2 = __pyx_v_self->_revno_to_branch_count;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_5 = PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_int_0, __pyx_v_root_count); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 840, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "breezy/_known_graph_pyx.pyx":841
 *                 ms_node._revno_last = 1
 *             PyDict_SetItem(self._revno_to_branch_count, 0, root_count)
 *         ms_node.completed = 1             # <<<<<<<<<<<<<<
 *         if PyList_GET_SIZE(self._scheduled_nodes) == 0:
 *             # The first scheduled node is always the end of merge
*/
  __pyx_v_ms_node->completed = 1;

  /* "breezy/_known_graph_pyx.pyx":842
 *             PyDict_SetItem(self._revno_to_branch_count, 0, root_count)
 *         ms_node.completed = 1
 *         if PyList_GET_SIZE(self._scheduled_nodes) == 0:             # <<<<<<<<<<<<<<
 *             # The first scheduled node is always the end of merge
 *             ms_node.end_of_merge = True
*/
  __pyx_t_2 = __pyx_v_self->_scheduled_nodes;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_t_2) == 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "breezy/_known_graph_pyx.pyx":844
 *         if PyList_GET_SIZE(self._scheduled_nodes) == 0:
 *             # The first scheduled node is always the end of merge
 *             ms_node.end_of_merge = True             # <<<<<<<<<<<<<<
 *         else:
 *             prev_node = _get_list_node(self._scheduled_nodes,
*/
    __Pyx_INCREF(Py_True);
    __Pyx_GIVEREF(Py_True);
    __Pyx_GOTREF(__pyx_v_ms_node->end_of_merge);
    __Pyx_DECREF(__pyx_v_ms_node->end_of_merge);
    __pyx_v_ms_node->end_of_merge = Py_True;

    /* "breezy/_known_graph_pyx.pyx":842
 *             PyDict_SetItem(self._revno_to_branch_count, 0, root_count)
 *         ms_node.completed = 1
 *         if PyList_GET_SIZE(self._scheduled_nodes) == 0:             # <<<<<<<<<<<<<<
 *             # The first scheduled node is always the end of merge
 *             ms_node.end_of_merge = True
*/
    goto __pyx_L8;
  }

  /* "breezy/_known_graph_pyx.pyx":846
 *             ms_node.end_of_merge = True
 *         else:
 *             prev_node = _get_list_node(self._scheduled_nodes,             # <<<<<<<<<<<<<<
 *                                     PyList_GET_SIZE(self._scheduled_nodes) - 1)
 *             ms_prev_node = <_MergeSortNode>prev_node.extra
*/
  /*else*/ {
    __pyx_t_2 = __pyx_v_self->_scheduled_nodes;
    __Pyx_INCREF(__pyx_t_2);

    /* "breezy/_known_graph_pyx.pyx":847
 *         else:
 *             prev_node = _get_list_node(self._scheduled_nodes,
 *                                     PyList_GET_SIZE(self._scheduled_nodes) - 1)             # <<<<<<<<<<<<<<
 *             ms_prev_node = <_MergeSortNode>prev_node.extra
 *             if ms_prev_node.merge_depth < ms_node.merge_depth:
*/
    __pyx_t_1 = __pyx_v_self->_scheduled_nodes;
    __Pyx_INCREF(__pyx_t_1);

    /* "breezy/_known_graph_pyx.pyx":846
 *             ms_node.end_of_merge = True
 *         else:
 *             prev_node = _get_list_node(self._scheduled_nodes,             # <<<<<<<<<<<<<<
 *                                     PyList_GET_SIZE(self._scheduled_nodes) - 1)
 *             ms_prev_node = <_MergeSortNode>prev_node.extra
*/
    __pyx_t_6 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_list_node(__pyx_t_2, (PyList_GET_SIZE(__pyx_t_1) - 1))); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 846, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_prev_node = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_6);
    __pyx_t_6 = 0;

    /* "breezy/_known_graph_pyx.pyx":848
 *             prev_node = _get_list_node(self._scheduled_nodes,
 *                                     PyList_GET_SIZE(self._scheduled_nodes) - 1)
 *             ms_prev_node = <_MergeSortNode>prev_node.extra             # <<<<<<<<<<<<<<
 *             if ms_prev_node.merge_depth < ms_node.merge_depth:
 *                 # The previously pushed node is to our left, so this is the end
*/
    __pyx_t_6 = __pyx_v_prev_node->extra;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_v_ms_prev_node = ((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_t_6);
    __pyx_t_6 = 0;

    /* "breezy/_known_graph_pyx.pyx":849
 *                                     PyList_GET_SIZE(self._scheduled_nodes) - 1)
 *             ms_prev_node = <_MergeSortNode>prev_node.extra
 *             if ms_prev_node.merge_depth < ms_node.merge_depth:             # <<<<<<<<<<<<<<
 *                 # The previously pushed node is to our left, so this is the end
 *                 # of this right-hand chain
*/
    __pyx_t_3 = (__pyx_v_ms_prev_node->merge_depth < __pyx_v_ms_node->merge_depth);
    if (__pyx_t_3) {

      /* "breezy/_known_graph_pyx.pyx":852
 *                 # The previously pushed node is to our left, so this is the end
 *                 # of this right-hand chain
 *                 ms_node.end_of_merge = True             # <<<<<<<<<<<<<<
 *             elif (ms_prev_node.merge_depth == ms_node.merge_depth
 *                   and prev_node not in node.parents):
*/
      __Pyx_INCREF(Py_True);
      __Pyx_GIVEREF(Py_True);
      __Pyx_GOTREF(__pyx_v_ms_node->end_of_merge);
      __Pyx_DECREF(__pyx_v_ms_node->end_of_merge);
      __pyx_v_ms_node->end_of_merge = Py_True;

      /* "breezy/_known_graph_pyx.pyx":849
 *                                     PyList_GET_SIZE(self._scheduled_nodes) - 1)
 *             ms_prev_node = <_MergeSortNode>prev_node.extra
 *             if ms_prev_node.merge_depth < ms_node.merge_depth:             # <<<<<<<<<<<<<<
 *                 # The previously pushed node is to our left, so this is the end
 *                 # of this right-hand chain
*/
      goto __pyx_L9;
    }

    /* "breezy/_known_graph_pyx.pyx":853
 *                 # of this right-hand chain
 *                 ms_node.end_of_merge = True
 *             elif (ms_prev_node.merge_depth == ms_node.merge_depth             # <<<<<<<<<<<<<<
 *                   and prev_node not in node.parents):
 *                 # The next node is not a direct parent of this node
*/
    __pyx_t_7 = (__pyx_v_ms_prev_node->merge_depth == __pyx_v_ms_node->merge_depth);
    if (__pyx_t_7) {
    } else {
      __pyx_t_3 = __pyx_t_7;
      goto __pyx_L10_bool_binop_done;
    }

    /* "breezy/_known_graph_pyx.pyx":854
 *                 ms_node.end_of_merge = True
 *             elif (ms_prev_node.merge_depth == ms_node.merge_depth
 *                   and prev_node not in node.parents):             # <<<<<<<<<<<<<<
 *                 # The next node is not a direct parent of this node
 *                 ms_node.end_of_merge = True
*/
    __pyx_t_7 = (__Pyx_PySequence_ContainsTF(((PyObject *)__pyx_v_prev_node), __pyx_v_node->parents, Py_NE)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 854, __pyx_L1_error)
    __pyx_t_3 = __pyx_t_7;
    __pyx_L10_bool_binop_done:;

    /* "breezy/_known_graph_pyx.pyx":853
 *                 # of this right-hand chain
 *                 ms_node.end_of_merge = True
 *             elif (ms_prev_node.merge_depth == ms_node.merge_depth             # <<<<<<<<<<<<<<
 *                   and prev_node not in node.parents):
 *                 # The next node is not a direct parent of this node
*/
    if (__pyx_t_3) {

      /* "breezy/_known_graph_pyx.pyx":856
 *                   and prev_node not in node.parents):
 *                 # The next node is not a direct parent of this node
 *                 ms_node.end_of_merge = True             # <<<<<<<<<<<<<<
 *             else:
 *                 ms_node.end_of_merge = False
*/
      __Pyx_INCREF(Py_True);
      __Pyx_GIVEREF(Py_True);
      __Pyx_GOTREF(__pyx_v_ms_node->end_of_merge);
      __Pyx_DECREF(__pyx_v_ms_node->end_of_merge);
      __pyx_v_ms_node->end_of_merge = Py_True;

      /* "breezy/_known_graph_pyx.pyx":853
 *                 # of this right-hand chain
 *                 ms_node.end_of_merge = True
 *             elif (ms_prev_node.merge_depth == ms_node.merge_depth             # <<<<<<<<<<<<<<
 *                   and prev_node not in node.parents):
 *                 # The next node is not a direct parent of this node
*/
      goto __pyx_L9;
    }

    /* "breezy/_known_graph_pyx.pyx":858
 *                 ms_node.end_of_merge = True
 *             else:
 *                 ms_node.end_of_merge = False             # <<<<<<<<<<<<<<
 *         PyList_Append(self._scheduled_nodes, node)
 * 
*/
    /*else*/ {
      __Pyx_INCREF(Py_False);
      __Pyx_GIVEREF(Py_False);
      __Pyx_GOTREF(__pyx_v_ms_node->end_of_merge);
      __Pyx_DECREF(__pyx_v_ms_node->end_of_merge);
      __pyx_v_ms_node->end_of_merge = Py_False;
    }
    __pyx_L9:;
  }
  __pyx_L8:;

  /* "breezy/_known_graph_pyx.pyx":859
 *             else:
 *                 ms_node.end_of_merge = False
 *         PyList_Append(self._scheduled_nodes, node)             # <<<<<<<<<<<<<<
 * 
 *     cdef _schedule_stack(self):
*/
  __pyx_t_6 = __pyx_v_self->_scheduled_nodes;
  __Pyx_INCREF(__pyx_t_6);
  __pyx_t_5 = PyList_Append(__pyx_t_6, ((PyObject *)__pyx_v_node)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "breezy/_known_graph_pyx.pyx":792
 *             PyList_Append(self._depth_first_stack, node)
 * 
 *     cdef _pop_node(self):             # <<<<<<<<<<<<<<
 *         cdef PyObject *temp
 *         cdef _MergeSortNode ms_node, ms_parent_node, ms_prev_node
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSorter._pop_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ms_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_ms_parent_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_ms_prev_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_prev_node);
  __Pyx_XDECREF(__pyx_v_branch_count);
  __Pyx_XDECREF(__pyx_v_root_count);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":861
 *         PyList_Append(self._scheduled_nodes, node)
 * 
 *     cdef _schedule_stack(self):             # <<<<<<<<<<<<<<
 *         cdef _KnownGraphNode last_node, next_node
 *         cdef _MergeSortNode ms_node, ms_last_node, ms_next_node
*/

static PyObject *__pyx_f_6breezy_16_known_graph_pyx_12_MergeSorter__schedule_stack(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *__pyx_v_self) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_last_node = 0;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_next_node = 0;
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_ms_last_node = 0;
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_ms_next_node = 0;
  long __pyx_v_next_merge_depth;
  CYTHON_UNUSED PyObject *__pyx_v_ordered = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_t_7;
  long __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_schedule_stack", 0);

  /* "breezy/_known_graph_pyx.pyx":865
 *         cdef _MergeSortNode ms_node, ms_last_node, ms_next_node
 *         cdef long next_merge_depth
 *         ordered = []             # <<<<<<<<<<<<<<
 *         while self._last_stack_item >= 0:
 *             # Peek at the last item on the stack
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ordered = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":866
 *         cdef long next_merge_depth
 *         ordered = []
 *         while self._last_stack_item >= 0:             # <<<<<<<<<<<<<<
 *             # Peek at the last item on the stack
 *             last_node = _get_list_node(self._depth_first_stack,
*/
  while (1) {
    __pyx_t_2 = (__pyx_v_self->_last_stack_item >= 0);
    if (!__pyx_t_2) break;

    /* "breezy/_known_graph_pyx.pyx":868
 *         while self._last_stack_item >= 0:
 *             # Peek at the last item on the stack
 *             last_node = _get_list_node(self._depth_first_stack,             # <<<<<<<<<<<<<<
 *                                        self._last_stack_item)
 *             if last_node.gdfo == -1:
*/
    __pyx_t_1 = __pyx_v_self->_depth_first_stack;
    __Pyx_INCREF(__pyx_t_1);

    /* "breezy/_known_graph_pyx.pyx":869
 *             # Peek at the last item on the stack
 *             last_node = _get_list_node(self._depth_first_stack,
 *                                        self._last_stack_item)             # <<<<<<<<<<<<<<
 *             if last_node.gdfo == -1:
 *                 # if _find_gdfo skipped a node, that means there is a graph
*/
    __pyx_t_3 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_list_node(__pyx_t_1, __pyx_v_self->_last_stack_item)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 868, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_last_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "breezy/_known_graph_pyx.pyx":870
 *             last_node = _get_list_node(self._depth_first_stack,
 *                                        self._last_stack_item)
 *             if last_node.gdfo == -1:             # <<<<<<<<<<<<<<
 *                 # if _find_gdfo skipped a node, that means there is a graph
 *                 # cycle, error out now
*/
    __pyx_t_2 = (__pyx_v_last_node->gdfo == -1L);
    if (unlikely(__pyx_t_2)) {

      /* "breezy/_known_graph_pyx.pyx":873
 *                 # if _find_gdfo skipped a node, that means there is a graph
 *                 # cycle, error out now
 *                 raise errors.GraphCycleError(self.graph._nodes)             # <<<<<<<<<<<<<<
 *             ms_last_node = <_MergeSortNode>last_node.extra
 *             if not ms_last_node.has_pending_parents():
*/
      __pyx_t_1 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_errors); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 873, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_GraphCycleError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 873, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
        assert(__pyx_t_1);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
        __pyx_t_6 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_self->graph->_nodes};
        __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 873, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 873, __pyx_L1_error)

      /* "breezy/_known_graph_pyx.pyx":870
 *             last_node = _get_list_node(self._depth_first_stack,
 *                                        self._last_stack_item)
 *             if last_node.gdfo == -1:             # <<<<<<<<<<<<<<
 *                 # if _find_gdfo skipped a node, that means there is a graph
 *                 # cycle, error out now
*/
    }

    /* "breezy/_known_graph_pyx.pyx":874
 *                 # cycle, error out now
 *                 raise errors.GraphCycleError(self.graph._nodes)
 *             ms_last_node = <_MergeSortNode>last_node.extra             # <<<<<<<<<<<<<<
 *             if not ms_last_node.has_pending_parents():
 *                 # Processed all parents, pop this node
*/
    __pyx_t_3 = __pyx_v_last_node->extra;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_ms_last_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "breezy/_known_graph_pyx.pyx":875
 *                 raise errors.GraphCycleError(self.graph._nodes)
 *             ms_last_node = <_MergeSortNode>last_node.extra
 *             if not ms_last_node.has_pending_parents():             # <<<<<<<<<<<<<<
 *                 # Processed all parents, pop this node
 *                 self._pop_node()
*/
    __pyx_t_7 = ((struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_v_ms_last_node->__pyx_vtab)->has_pending_parents(__pyx_v_ms_last_node); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 875, __pyx_L1_error)
    __pyx_t_2 = (!(__pyx_t_7 != 0));
    if (__pyx_t_2) {

      /* "breezy/_known_graph_pyx.pyx":877
 *             if not ms_last_node.has_pending_parents():
 *                 # Processed all parents, pop this node
 *                 self._pop_node()             # <<<<<<<<<<<<<<
 *                 continue
 *             while ms_last_node.has_pending_parents():
*/
      __pyx_t_3 = ((struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSorter *)__pyx_v_self->__pyx_vtab)->_pop_node(__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "breezy/_known_graph_pyx.pyx":878
 *                 # Processed all parents, pop this node
 *                 self._pop_node()
 *                 continue             # <<<<<<<<<<<<<<
 *             while ms_last_node.has_pending_parents():
 *                 if ms_last_node.left_pending_parent is not None:
*/
      goto __pyx_L3_continue;

      /* "breezy/_known_graph_pyx.pyx":875
 *                 raise errors.GraphCycleError(self.graph._nodes)
 *             ms_last_node = <_MergeSortNode>last_node.extra
 *             if not ms_last_node.has_pending_parents():             # <<<<<<<<<<<<<<
 *                 # Processed all parents, pop this node
 *                 self._pop_node()
*/
    }

    /* "breezy/_known_graph_pyx.pyx":879
 *                 self._pop_node()
 *                 continue
 *             while ms_last_node.has_pending_parents():             # <<<<<<<<<<<<<<
 *                 if ms_last_node.left_pending_parent is not None:
 *                     # recurse depth first into the primary parent
*/
    while (1) {
      __pyx_t_7 = ((struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_v_ms_last_node->__pyx_vtab)->has_pending_parents(__pyx_v_ms_last_node); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 879, __pyx_L1_error)
      __pyx_t_2 = (__pyx_t_7 != 0);
      if (!__pyx_t_2) break;

      /* "breezy/_known_graph_pyx.pyx":880
 *                 continue
 *             while ms_last_node.has_pending_parents():
 *                 if ms_last_node.left_pending_parent is not None:             # <<<<<<<<<<<<<<
 *                     # recurse depth first into the primary parent
 *                     next_node = ms_last_node.left_pending_parent
*/
      __pyx_t_2 = (((PyObject *)__pyx_v_ms_last_node->left_pending_parent) != Py_None);
      if (__pyx_t_2) {

        /* "breezy/_known_graph_pyx.pyx":882
 *                 if ms_last_node.left_pending_parent is not None:
 *                     # recurse depth first into the primary parent
 *                     next_node = ms_last_node.left_pending_parent             # <<<<<<<<<<<<<<
 *                     ms_last_node.left_pending_parent = None
 *                 else:
*/
        __pyx_t_3 = ((PyObject *)__pyx_v_ms_last_node->left_pending_parent);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_XDECREF_SET(__pyx_v_next_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_3));
        __pyx_t_3 = 0;

        /* "breezy/_known_graph_pyx.pyx":883
 *                     # recurse depth first into the primary parent
 *                     next_node = ms_last_node.left_pending_parent
 *                     ms_last_node.left_pending_parent = None             # <<<<<<<<<<<<<<
 *                 else:
 *                     # place any merges in right-to-left order for scheduling
*/
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        __Pyx_GOTREF((PyObject *)__pyx_v_ms_last_node->left_pending_parent);
        __Pyx_DECREF((PyObject *)__pyx_v_ms_last_node->left_pending_parent);
        __pyx_v_ms_last_node->left_pending_parent = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)Py_None);

        /* "breezy/_known_graph_pyx.pyx":880
 *                 continue
 *             while ms_last_node.has_pending_parents():
 *                 if ms_last_node.left_pending_parent is not None:             # <<<<<<<<<<<<<<
 *                     # recurse depth first into the primary parent
 *                     next_node = ms_last_node.left_pending_parent
*/
        goto __pyx_L9;
      }

      /* "breezy/_known_graph_pyx.pyx":892
 *                     #       left most, which will display nicely (you get
 *                     #       smaller trees at the top of the combined merge).
 *                     next_node = ms_last_node.pending_parents.pop()             # <<<<<<<<<<<<<<
 *                 ms_next_node = self._get_ms_node(next_node)
 *                 if ms_next_node.completed:
*/
      /*else*/ {
        __pyx_t_3 = __Pyx_PyObject_Pop(__pyx_v_ms_last_node->pending_parents); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 892, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode))))) __PYX_ERR(0, 892, __pyx_L1_error)
        __Pyx_XDECREF_SET(__pyx_v_next_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_3));
        __pyx_t_3 = 0;
      }
      __pyx_L9:;

      /* "breezy/_known_graph_pyx.pyx":893
 *                     #       smaller trees at the top of the combined merge).
 *                     next_node = ms_last_node.pending_parents.pop()
 *                 ms_next_node = self._get_ms_node(next_node)             # <<<<<<<<<<<<<<
 *                 if ms_next_node.completed:
 *                     # this parent was completed by a child on the
*/
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSorter *)__pyx_v_self->__pyx_vtab)->_get_ms_node(__pyx_v_self, __pyx_v_next_node)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_XDECREF_SET(__pyx_v_ms_next_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "breezy/_known_graph_pyx.pyx":894
 *                     next_node = ms_last_node.pending_parents.pop()
 *                 ms_next_node = self._get_ms_node(next_node)
 *                 if ms_next_node.completed:             # <<<<<<<<<<<<<<
 *                     # this parent was completed by a child on the
 *                     # call stack. skip it.
*/
      __pyx_t_2 = (__pyx_v_ms_next_node->completed != 0);
      if (__pyx_t_2) {

        /* "breezy/_known_graph_pyx.pyx":897
 *                     # this parent was completed by a child on the
 *                     # call stack. skip it.
 *                     continue             # <<<<<<<<<<<<<<
 *                 # otherwise transfer it from the source graph into the
 *                 # top of the current depth first search stack.
*/
        goto __pyx_L7_continue;

        /* "breezy/_known_graph_pyx.pyx":894
 *                     next_node = ms_last_node.pending_parents.pop()
 *                 ms_next_node = self._get_ms_node(next_node)
 *                 if ms_next_node.completed:             # <<<<<<<<<<<<<<
 *                     # this parent was completed by a child on the
 *                     # call stack. skip it.
*/
      }

      /* "breezy/_known_graph_pyx.pyx":901
 *                 # top of the current depth first search stack.
 * 
 *                 if next_node is ms_last_node.left_parent:             # <<<<<<<<<<<<<<
 *                     next_merge_depth = ms_last_node.merge_depth
 *                 else:
*/
      __pyx_t_2 = (__pyx_v_next_node == __pyx_v_ms_last_node->left_parent);
      if (__pyx_t_2) {

        /* "breezy/_known_graph_pyx.pyx":902
 * 
 *                 if next_node is ms_last_node.left_parent:
 *                     next_merge_depth = ms_last_node.merge_depth             # <<<<<<<<<<<<<<
 *                 else:
 *                     next_merge_depth = ms_last_node.merge_depth + 1
*/
        __pyx_t_8 = __pyx_v_ms_last_node->merge_depth;
        __pyx_v_next_merge_depth = __pyx_t_8;

        /* "breezy/_known_graph_pyx.pyx":901
 *                 # top of the current depth first search stack.
 * 
 *                 if next_node is ms_last_node.left_parent:             # <<<<<<<<<<<<<<
 *                     next_merge_depth = ms_last_node.merge_depth
 *                 else:
*/
        goto __pyx_L11;
      }

      /* "breezy/_known_graph_pyx.pyx":904
 *                     next_merge_depth = ms_last_node.merge_depth
 *                 else:
 *                     next_merge_depth = ms_last_node.merge_depth + 1             # <<<<<<<<<<<<<<
 *                 self._push_node(next_node, next_merge_depth)
 *                 # and do not continue processing parents until this 'call'
*/
      /*else*/ {
        __pyx_v_next_merge_depth = (__pyx_v_ms_last_node->merge_depth + 1);
      }
      __pyx_L11:;

      /* "breezy/_known_graph_pyx.pyx":905
 *                 else:
 *                     next_merge_depth = ms_last_node.merge_depth + 1
 *                 self._push_node(next_node, next_merge_depth)             # <<<<<<<<<<<<<<
 *                 # and do not continue processing parents until this 'call'
 *                 # has recursed.
*/
      __pyx_t_3 = ((struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSorter *)__pyx_v_self->__pyx_vtab)->_push_node(__pyx_v_self, __pyx_v_next_node, __pyx_v_next_merge_depth); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 905, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "breezy/_known_graph_pyx.pyx":908
 *                 # and do not continue processing parents until this 'call'
 *                 # has recursed.
 *                 break             # <<<<<<<<<<<<<<
 * 
 *     cdef topo_order(self):
*/
      goto __pyx_L8_break;
      __pyx_L7_continue:;
    }
    __pyx_L8_break:;
    __pyx_L3_continue:;
  }

  /* "breezy/_known_graph_pyx.pyx":861
 *         PyList_Append(self._scheduled_nodes, node)
 * 
 *     cdef _schedule_stack(self):             # <<<<<<<<<<<<<<
 *         cdef _KnownGraphNode last_node, next_node
 *         cdef _MergeSortNode ms_node, ms_last_node, ms_next_node
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSorter._schedule_stack", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_last_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_next_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_ms_last_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_ms_next_node);
  __Pyx_XDECREF(__pyx_v_ordered);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/_known_graph_pyx.pyx":910
 *                 break
 * 
 *     cdef topo_order(self):             # <<<<<<<<<<<<<<
 *         cdef _MergeSortNode ms_node
 *         cdef _KnownGraphNode node
*/

static PyObject *__pyx_f_6breezy_16_known_graph_pyx_12_MergeSorter_topo_order(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *__pyx_v_self) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v_ms_node = 0;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v_node = 0;
  Py_ssize_t __pyx_v_pos;
  PyObject *__pyx_v_ordered = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("topo_order", 0);

  /* "breezy/_known_graph_pyx.pyx":921
 *         #       We might consider moving the attributes into the base
 *         #       KnownGraph object.
 *         self._schedule_stack()             # <<<<<<<<<<<<<<
 * 
 *         # We've set up the basic schedule, now we can continue processing the
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSorter *)__pyx_v_self->__pyx_vtab)->_schedule_stack(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":931
 *         #       7ms is computing the return tuple
 *         #       4ms is PyList_Append()
 *         ordered = []             # <<<<<<<<<<<<<<
 *         # output the result in reverse order, and separate the generated info
 *         for pos from PyList_GET_SIZE(self._scheduled_nodes) > pos >= 0:
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ordered = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":933
 *         ordered = []
 *         # output the result in reverse order, and separate the generated info
 *         for pos from PyList_GET_SIZE(self._scheduled_nodes) > pos >= 0:             # <<<<<<<<<<<<<<
 *             node = _get_list_node(self._scheduled_nodes, pos)
 *             ms_node = <_MergeSortNode>node.extra
*/
  __pyx_t_1 = __pyx_v_self->_scheduled_nodes;
  __Pyx_INCREF(__pyx_t_1);
  for (__pyx_v_pos = PyList_GET_SIZE(__pyx_t_1)-1; __pyx_v_pos >= 0; __pyx_v_pos--) {

    /* "breezy/_known_graph_pyx.pyx":934
 *         # output the result in reverse order, and separate the generated info
 *         for pos from PyList_GET_SIZE(self._scheduled_nodes) > pos >= 0:
 *             node = _get_list_node(self._scheduled_nodes, pos)             # <<<<<<<<<<<<<<
 *             ms_node = <_MergeSortNode>node.extra
 *             PyList_Append(ordered, ms_node)
*/
    __pyx_t_2 = __pyx_v_self->_scheduled_nodes;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = ((PyObject *)__pyx_f_6breezy_16_known_graph_pyx__get_list_node(__pyx_t_2, __pyx_v_pos)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 934, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "breezy/_known_graph_pyx.pyx":935
 *         for pos from PyList_GET_SIZE(self._scheduled_nodes) > pos >= 0:
 *             node = _get_list_node(self._scheduled_nodes, pos)
 *             ms_node = <_MergeSortNode>node.extra             # <<<<<<<<<<<<<<
 *             PyList_Append(ordered, ms_node)
 *             node.extra = None
*/
    __pyx_t_3 = __pyx_v_node->extra;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_ms_node, ((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "breezy/_known_graph_pyx.pyx":936
 *             node = _get_list_node(self._scheduled_nodes, pos)
 *             ms_node = <_MergeSortNode>node.extra
 *             PyList_Append(ordered, ms_node)             # <<<<<<<<<<<<<<
 *             node.extra = None
 *         # Clear out the scheduled nodes now that we're done
*/
    __pyx_t_4 = PyList_Append(__pyx_v_ordered, ((PyObject *)__pyx_v_ms_node)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 936, __pyx_L1_error)

    /* "breezy/_known_graph_pyx.pyx":937
 *             ms_node = <_MergeSortNode>node.extra
 *             PyList_Append(ordered, ms_node)
 *             node.extra = None             # <<<<<<<<<<<<<<
 *         # Clear out the scheduled nodes now that we're done
 *         self._scheduled_nodes = []
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_node->extra);
    __Pyx_DECREF(__pyx_v_node->extra);
    __pyx_v_node->extra = Py_None;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":939
 *             node.extra = None
 *         # Clear out the scheduled nodes now that we're done
 *         self._scheduled_nodes = []             # <<<<<<<<<<<<<<
 *         return ordered
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_scheduled_nodes);
  __Pyx_DECREF(__pyx_v_self->_scheduled_nodes);
  __pyx_v_self->_scheduled_nodes = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/_known_graph_pyx.pyx":940
 *         # Clear out the scheduled nodes now that we're done
 *         self._scheduled_nodes = []
 *         return ordered             # <<<<<<<<<<<<<<
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ordered);
  __pyx_r = __pyx_v_ordered;
  goto __pyx_L0;

  /* "breezy/_known_graph_pyx.pyx":910
 *                 break
 * 
 *     cdef topo_order(self):             # <<<<<<<<<<<<<<
 *         cdef _MergeSortNode ms_node
 *         cdef _KnownGraphNode node
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSorter.topo_order", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ms_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF(__pyx_v_ordered);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_12_MergeSorter_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_12_MergeSorter_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_12_MergeSorter_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_12_MergeSorter_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_12_MergeSorter_2__reduce_cython__(((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_12_MergeSorter_2__reduce_cython__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._depth_first_stack, self._last_stack_item, self._revno_to_branch_count, self._scheduled_nodes, self.graph)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __pyx_t_1 = PyLong_FromSsize_t(__pyx_v_self->_last_stack_item); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->_depth_first_stack);
  __Pyx_GIVEREF(__pyx_v_self->_depth_first_stack);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->_depth_first_stack) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_revno_to_branch_count);
  __Pyx_GIVEREF(__pyx_v_self->_revno_to_branch_count);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_self->_revno_to_branch_count) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_scheduled_nodes);
  __Pyx_GIVEREF(__pyx_v_self->_scheduled_nodes);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_v_self->_scheduled_nodes) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->graph);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->graph);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, ((PyObject *)__pyx_v_self->graph)) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._depth_first_stack, self._last_stack_item, self._revno_to_branch_count, self._scheduled_nodes, self.graph)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self._depth_first_stack, self._last_stack_item, self._revno_to_branch_count, self._scheduled_nodes, self.graph)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(1, 7, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict) != (0)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._depth_first_stack is not None or self._revno_to_branch_count is not None or self._scheduled_nodes is not None or self.graph is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._depth_first_stack, self._last_stack_item, self._revno_to_branch_count, self._scheduled_nodes, self.graph)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._depth_first_stack is not None or self._revno_to_branch_count is not None or self._scheduled_nodes is not None or self.graph is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__MergeSorter, (type(self), 0xee1d2e4, None), state
*/
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->_depth_first_stack != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->_revno_to_branch_count != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->_scheduled_nodes != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->graph) != Py_None);
    __pyx_t_3 = __pyx_t_4;
    __pyx_L6_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._depth_first_stack is not None or self._revno_to_branch_count is not None or self._scheduled_nodes is not None or self.graph is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__MergeSorter, (type(self), 0xee1d2e4, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._depth_first_stack is not None or self._revno_to_branch_count is not None or self._scheduled_nodes is not None or self.graph is not None
 *     if use_setstate:
 *         return __pyx_unpickle__MergeSorter, (type(self), 0xee1d2e4, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__MergeSorter, (type(self), 0xee1d2e4, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__MergeSorter); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_249680612);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_249680612);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_mstate_global->__pyx_int_249680612) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._depth_first_stack is not None or self._revno_to_branch_count is not None or self._scheduled_nodes is not None or self.graph is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__MergeSorter, (type(self), 0xee1d2e4, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__MergeSorter, (type(self), 0xee1d2e4, None), state
 *     else:
 *         return __pyx_unpickle__MergeSorter, (type(self), 0xee1d2e4, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__MergeSorter__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__MergeSorter); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_249680612);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_249680612);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_mstate_global->__pyx_int_249680612) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSorter.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__MergeSorter, (type(self), 0xee1d2e4, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__MergeSorter__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_12_MergeSorter_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_12_MergeSorter_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_12_MergeSorter_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_12_MergeSorter_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSorter.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_12_MergeSorter_4__setstate_cython__(((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_12_MergeSorter_4__setstate_cython__(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__MergeSorter, (type(self), 0xee1d2e4, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__MergeSorter__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(1, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_6breezy_16_known_graph_pyx___pyx_unpickle__MergeSorter__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__MergeSorter, (type(self), 0xee1d2e4, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__MergeSorter__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy._known_graph_pyx._MergeSorter.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle__KnownGraphNode(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x1857be3, 0x4591477, 0x1dbac4a, b'children, extra, gdfo, key, parents, seen')
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_3__pyx_unpickle__KnownGraphNode(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_3__pyx_unpickle__KnownGraphNode = {"__pyx_unpickle__KnownGraphNode", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_3__pyx_unpickle__KnownGraphNode, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_3__pyx_unpickle__KnownGraphNode(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__KnownGraphNode (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle__KnownGraphNode", 0) < (0)) __PYX_ERR(1, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__KnownGraphNode", 1, 3, 3, i); __PYX_ERR(1, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__KnownGraphNode", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx.__pyx_unpickle__KnownGraphNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(1, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_2__pyx_unpickle__KnownGraphNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_2__pyx_unpickle__KnownGraphNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__KnownGraphNode", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle__KnownGraphNode(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x1857be3, 0x4591477, 0x1dbac4a, b'children, extra, gdfo, key, parents, seen')             # <<<<<<<<<<<<<<
 *     __pyx_result = _KnownGraphNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0x1857be3, 0x4591477, 0x1dbac4a, __pyx_k_children_extra_gdfo_key_parents); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x1857be3, 0x4591477, 0x1dbac4a, b'children, extra, gdfo, key, parents, seen')
 *     __pyx_result = _KnownGraphNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__KnownGraphNode__set_state(<_KnownGraphNode> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x1857be3, 0x4591477, 0x1dbac4a, b'children, extra, gdfo, key, parents, seen')
 *     __pyx_result = _KnownGraphNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__KnownGraphNode__set_state(<_KnownGraphNode> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = _KnownGraphNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__KnownGraphNode__set_state(<_KnownGraphNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__KnownGraphNode__set_state(_KnownGraphNode __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(1, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_6breezy_16_known_graph_pyx___pyx_unpickle__KnownGraphNode__set_state(((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x1857be3, 0x4591477, 0x1dbac4a, b'children, extra, gdfo, key, parents, seen')
 *     __pyx_result = _KnownGraphNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__KnownGraphNode__set_state(<_KnownGraphNode> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__KnownGraphNode__set_state(<_KnownGraphNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__KnownGraphNode__set_state(_KnownGraphNode __pyx_result, __pyx_state: tuple):
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.extra = __pyx_state[1]; __pyx_result.gdfo = __pyx_state[2]; __pyx_result.key = __pyx_state[3]; __pyx_result.parents = __pyx_state[4]; __pyx_result.seen = __pyx_state[5]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle__KnownGraphNode(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x1857be3, 0x4591477, 0x1dbac4a, b'children, extra, gdfo, key, parents, seen')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy._known_graph_pyx.__pyx_unpickle__KnownGraphNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__KnownGraphNode__set_state(<_KnownGraphNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__KnownGraphNode__set_state(_KnownGraphNode __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.extra = __pyx_state[1]; __pyx_result.gdfo = __pyx_state[2]; __pyx_result.key = __pyx_state[3]; __pyx_result.parents = __pyx_state[4]; __pyx_result.seen = __pyx_state[5]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 6)
*/

static PyObject *__pyx_f_6breezy_16_known_graph_pyx___pyx_unpickle__KnownGraphNode__set_state(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  long __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__KnownGraphNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__KnownGraphNode__set_state(_KnownGraphNode __pyx_result, __pyx_state: tuple):
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.extra = __pyx_state[1]; __pyx_result.gdfo = __pyx_state[2]; __pyx_result.key = __pyx_state[3]; __pyx_result.parents = __pyx_state[4]; __pyx_result.seen = __pyx_state[5]             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 6)
*/
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->children);
  __Pyx_DECREF(__pyx_v___pyx_result->children);
  __pyx_v___pyx_result->children = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->extra);
  __Pyx_DECREF(__pyx_v___pyx_result->extra);
  __pyx_v___pyx_result->extra = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_long(__pyx_t_1); if (unlikely((__pyx_t_2 == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->gdfo = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->key);
  __Pyx_DECREF(__pyx_v___pyx_result->key);
  __pyx_v___pyx_result->key = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->parents);
  __Pyx_DECREF(__pyx_v___pyx_result->parents);
  __pyx_v___pyx_result->parents = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->seen = __pyx_t_3;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle__KnownGraphNode__set_state(_KnownGraphNode __pyx_result, __pyx_state: tuple):
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.extra = __pyx_state[1]; __pyx_result.gdfo = __pyx_state[2]; __pyx_result.key = __pyx_state[3]; __pyx_result.parents = __pyx_state[4]; __pyx_result.seen = __pyx_state[5]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 6)             # <<<<<<<<<<<<<<
*/
  __pyx_t_3 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 6); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle__KnownGraphNode__set_state(<_KnownGraphNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__KnownGraphNode__set_state(_KnownGraphNode __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.extra = __pyx_state[1]; __pyx_result.gdfo = __pyx_state[2]; __pyx_result.key = __pyx_state[3]; __pyx_result.parents = __pyx_state[4]; __pyx_result.seen = __pyx_state[5]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 6)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy._known_graph_pyx.__pyx_unpickle__KnownGraphNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_KnownGraph(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe34a3b8, 0xbb3a387, 0x6ffbab1, b'_known_heads, _nodes, do_cache')
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_5__pyx_unpickle_KnownGraph(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_5__pyx_unpickle_KnownGraph = {"__pyx_unpickle_KnownGraph", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_5__pyx_unpickle_KnownGraph, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_5__pyx_unpickle_KnownGraph(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_KnownGraph (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_KnownGraph", 0) < (0)) __PYX_ERR(1, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_KnownGraph", 1, 3, 3, i); __PYX_ERR(1, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_KnownGraph", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx.__pyx_unpickle_KnownGraph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(1, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_4__pyx_unpickle_KnownGraph(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_4__pyx_unpickle_KnownGraph(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_KnownGraph", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle_KnownGraph(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe34a3b8, 0xbb3a387, 0x6ffbab1, b'_known_heads, _nodes, do_cache')             # <<<<<<<<<<<<<<
 *     __pyx_result = KnownGraph.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0xe34a3b8, 0xbb3a387, 0x6ffbab1, __pyx_k_known_heads__nodes_do_cache); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe34a3b8, 0xbb3a387, 0x6ffbab1, b'_known_heads, _nodes, do_cache')
 *     __pyx_result = KnownGraph.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_KnownGraph__set_state(<KnownGraph> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe34a3b8, 0xbb3a387, 0x6ffbab1, b'_known_heads, _nodes, do_cache')
 *     __pyx_result = KnownGraph.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_KnownGraph__set_state(<KnownGraph> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = KnownGraph.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_KnownGraph__set_state(<KnownGraph> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_KnownGraph__set_state(KnownGraph __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(1, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_6breezy_16_known_graph_pyx___pyx_unpickle_KnownGraph__set_state(((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe34a3b8, 0xbb3a387, 0x6ffbab1, b'_known_heads, _nodes, do_cache')
 *     __pyx_result = KnownGraph.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_KnownGraph__set_state(<KnownGraph> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_KnownGraph__set_state(<KnownGraph> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_KnownGraph__set_state(KnownGraph __pyx_result, __pyx_state: tuple):
 *     __pyx_result._known_heads = __pyx_state[0]; __pyx_result._nodes = __pyx_state[1]; __pyx_result.do_cache = __pyx_state[2]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_KnownGraph(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe34a3b8, 0xbb3a387, 0x6ffbab1, b'_known_heads, _nodes, do_cache')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy._known_graph_pyx.__pyx_unpickle_KnownGraph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_KnownGraph__set_state(<KnownGraph> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_KnownGraph__set_state(KnownGraph __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._known_heads = __pyx_state[0]; __pyx_result._nodes = __pyx_state[1]; __pyx_result.do_cache = __pyx_state[2]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 3)
*/

static PyObject *__pyx_f_6breezy_16_known_graph_pyx___pyx_unpickle_KnownGraph__set_state(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_KnownGraph__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_KnownGraph__set_state(KnownGraph __pyx_result, __pyx_state: tuple):
 *     __pyx_result._known_heads = __pyx_state[0]; __pyx_result._nodes = __pyx_state[1]; __pyx_result.do_cache = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 3)
*/
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_known_heads);
  __Pyx_DECREF(__pyx_v___pyx_result->_known_heads);
  __pyx_v___pyx_result->_known_heads = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_nodes);
  __Pyx_DECREF(__pyx_v___pyx_result->_nodes);
  __pyx_v___pyx_result->_nodes = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->do_cache = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_KnownGraph__set_state(KnownGraph __pyx_result, __pyx_state: tuple):
 *     __pyx_result._known_heads = __pyx_state[0]; __pyx_result._nodes = __pyx_state[1]; __pyx_result.do_cache = __pyx_state[2]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 3)             # <<<<<<<<<<<<<<
*/
  __pyx_t_2 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle_KnownGraph__set_state(<KnownGraph> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_KnownGraph__set_state(KnownGraph __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._known_heads = __pyx_state[0]; __pyx_result._nodes = __pyx_state[1]; __pyx_result.do_cache = __pyx_state[2]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 3)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy._known_graph_pyx.__pyx_unpickle_KnownGraph__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle__MergeSortNode(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x30884b3, 0xd890a73, 0x7a2ed2f, b'_revno_first, _revno_last, _revno_second, completed, end_of_merge, is_first_child, key, left_parent, left_pending_parent, merge_depth, pending_parents, seen_by_child')
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_7__pyx_unpickle__MergeSortNode(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_7__pyx_unpickle__MergeSortNode = {"__pyx_unpickle__MergeSortNode", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_7__pyx_unpickle__MergeSortNode, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_7__pyx_unpickle__MergeSortNode(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__MergeSortNode (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle__MergeSortNode", 0) < (0)) __PYX_ERR(1, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__MergeSortNode", 1, 3, 3, i); __PYX_ERR(1, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__MergeSortNode", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx.__pyx_unpickle__MergeSortNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(1, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_6__pyx_unpickle__MergeSortNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_6__pyx_unpickle__MergeSortNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__MergeSortNode", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle__MergeSortNode(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x30884b3, 0xd890a73, 0x7a2ed2f, b'_revno_first, _revno_last, _revno_second, completed, end_of_merge, is_first_child, key, left_parent, left_pending_parent, merge_depth, pending_parents, seen_by_child')             # <<<<<<<<<<<<<<
 *     __pyx_result = _MergeSortNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0x30884b3, 0xd890a73, 0x7a2ed2f, __pyx_k_revno_first__revno_last__revno); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x30884b3, 0xd890a73, 0x7a2ed2f, b'_revno_first, _revno_last, _revno_second, completed, end_of_merge, is_first_child, key, left_parent, left_pending_parent, merge_depth, pending_parents, seen_by_child')
 *     __pyx_result = _MergeSortNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__MergeSortNode__set_state(<_MergeSortNode> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x30884b3, 0xd890a73, 0x7a2ed2f, b'_revno_first, _revno_last, _revno_second, completed, end_of_merge, is_first_child, key, left_parent, left_pending_parent, merge_depth, pending_parents, seen_by_child')
 *     __pyx_result = _MergeSortNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__MergeSortNode__set_state(<_MergeSortNode> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = _MergeSortNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__MergeSortNode__set_state(<_MergeSortNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__MergeSortNode__set_state(_MergeSortNode __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(1, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_6breezy_16_known_graph_pyx___pyx_unpickle__MergeSortNode__set_state(((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x30884b3, 0xd890a73, 0x7a2ed2f, b'_revno_first, _revno_last, _revno_second, completed, end_of_merge, is_first_child, key, left_parent, left_pending_parent, merge_depth, pending_parents, seen_by_child')
 *     __pyx_result = _MergeSortNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__MergeSortNode__set_state(<_MergeSortNode> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__MergeSortNode__set_state(<_MergeSortNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__MergeSortNode__set_state(_MergeSortNode __pyx_result, __pyx_state: tuple):
 *     __pyx_result._revno_first = __pyx_state[0]; __pyx_result._revno_last = __pyx_state[1]; __pyx_result._revno_second = __pyx_state[2]; __pyx_result.completed = __pyx_state[3]; __pyx_result.end_of_merge = __pyx_state[4]; __pyx_result.is_first_child = __pyx_state[5]; __pyx_result.key = __pyx_state[6]; __pyx_result.left_parent = __pyx_state[7]; __pyx_result.left_pending_parent = __pyx_state[8]; __pyx_result.merge_depth = __pyx_state[9]; __pyx_result.pending_parents = __pyx_state[10]; __pyx_result.seen_by_child = __pyx_state[11]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle__MergeSortNode(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x30884b3, 0xd890a73, 0x7a2ed2f, b'_revno_first, _revno_last, _revno_second, completed, end_of_merge, is_first_child, key, left_parent, left_pending_parent, merge_depth, pending_parents, seen_by_child')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy._known_graph_pyx.__pyx_unpickle__MergeSortNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__MergeSortNode__set_state(<_MergeSortNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__MergeSortNode__set_state(_MergeSortNode __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._revno_first = __pyx_state[0]; __pyx_result._revno_last = __pyx_state[1]; __pyx_result._revno_second = __pyx_state[2]; __pyx_result.completed = __pyx_state[3]; __pyx_result.end_of_merge = __pyx_state[4]; __pyx_result.is_first_child = __pyx_state[5]; __pyx_result.key = __pyx_state[6]; __pyx_result.left_parent = __pyx_state[7]; __pyx_result.left_pending_parent = __pyx_state[8]; __pyx_result.merge_depth = __pyx_state[9]; __pyx_result.pending_parents = __pyx_state[10]; __pyx_result.seen_by_child = __pyx_state[11]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 12)
*/

static PyObject *__pyx_f_6breezy_16_known_graph_pyx___pyx_unpickle__MergeSortNode__set_state(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  long __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__MergeSortNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__MergeSortNode__set_state(_MergeSortNode __pyx_result, __pyx_state: tuple):
 *     __pyx_result._revno_first = __pyx_state[0]; __pyx_result._revno_last = __pyx_state[1]; __pyx_result._revno_second = __pyx_state[2]; __pyx_result.completed = __pyx_state[3]; __pyx_result.end_of_merge = __pyx_state[4]; __pyx_result.is_first_child = __pyx_state[5]; __pyx_result.key = __pyx_state[6]; __pyx_result.left_parent = __pyx_state[7]; __pyx_result.left_pending_parent = __pyx_state[8]; __pyx_result.merge_depth = __pyx_state[9]; __pyx_result.pending_parents = __pyx_state[10]; __pyx_result.seen_by_child = __pyx_state[11]             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 12)
*/
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_long(__pyx_t_1); if (unlikely((__pyx_t_2 == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_revno_first = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_long(__pyx_t_1); if (unlikely((__pyx_t_2 == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_revno_last = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_long(__pyx_t_1); if (unlikely((__pyx_t_2 == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_revno_second = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->completed = __pyx_t_3;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->end_of_merge);
  __Pyx_DECREF(__pyx_v___pyx_result->end_of_merge);
  __pyx_v___pyx_result->end_of_merge = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_first_child = __pyx_t_3;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->key);
  __Pyx_DECREF(__pyx_v___pyx_result->key);
  __pyx_v___pyx_result->key = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->left_parent);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->left_parent);
  __pyx_v___pyx_result->left_parent = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->left_pending_parent);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->left_pending_parent);
  __pyx_v___pyx_result->left_pending_parent = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_long(__pyx_t_1); if (unlikely((__pyx_t_2 == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->merge_depth = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->pending_parents);
  __Pyx_DECREF(__pyx_v___pyx_result->pending_parents);
  __pyx_v___pyx_result->pending_parents = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->seen_by_child = __pyx_t_3;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle__MergeSortNode__set_state(_MergeSortNode __pyx_result, __pyx_state: tuple):
 *     __pyx_result._revno_first = __pyx_state[0]; __pyx_result._revno_last = __pyx_state[1]; __pyx_result._revno_second = __pyx_state[2]; __pyx_result.completed = __pyx_state[3]; __pyx_result.end_of_merge = __pyx_state[4]; __pyx_result.is_first_child = __pyx_state[5]; __pyx_result.key = __pyx_state[6]; __pyx_result.left_parent = __pyx_state[7]; __pyx_result.left_pending_parent = __pyx_state[8]; __pyx_result.merge_depth = __pyx_state[9]; __pyx_result.pending_parents = __pyx_state[10]; __pyx_result.seen_by_child = __pyx_state[11]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 12)             # <<<<<<<<<<<<<<
*/
  __pyx_t_3 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 12); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle__MergeSortNode__set_state(<_MergeSortNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__MergeSortNode__set_state(_MergeSortNode __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._revno_first = __pyx_state[0]; __pyx_result._revno_last = __pyx_state[1]; __pyx_result._revno_second = __pyx_state[2]; __pyx_result.completed = __pyx_state[3]; __pyx_result.end_of_merge = __pyx_state[4]; __pyx_result.is_first_child = __pyx_state[5]; __pyx_result.key = __pyx_state[6]; __pyx_result.left_parent = __pyx_state[7]; __pyx_result.left_pending_parent = __pyx_state[8]; __pyx_result.merge_depth = __pyx_state[9]; __pyx_result.pending_parents = __pyx_state[10]; __pyx_result.seen_by_child = __pyx_state[11]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 12)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy._known_graph_pyx.__pyx_unpickle__MergeSortNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle__MergeSorter(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xee1d2e4, 0xd40361d, 0x6f5320f, b'_depth_first_stack, _last_stack_item, _revno_to_branch_count, _scheduled_nodes, graph')
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_9__pyx_unpickle__MergeSorter(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_16_known_graph_pyx_9__pyx_unpickle__MergeSorter = {"__pyx_unpickle__MergeSorter", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_9__pyx_unpickle__MergeSorter, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_16_known_graph_pyx_9__pyx_unpickle__MergeSorter(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__MergeSorter (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle__MergeSorter", 0) < (0)) __PYX_ERR(1, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__MergeSorter", 1, 3, 3, i); __PYX_ERR(1, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__MergeSorter", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy._known_graph_pyx.__pyx_unpickle__MergeSorter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(1, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_6breezy_16_known_graph_pyx_8__pyx_unpickle__MergeSorter(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_16_known_graph_pyx_8__pyx_unpickle__MergeSorter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__MergeSorter", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle__MergeSorter(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xee1d2e4, 0xd40361d, 0x6f5320f, b'_depth_first_stack, _last_stack_item, _revno_to_branch_count, _scheduled_nodes, graph')             # <<<<<<<<<<<<<<
 *     __pyx_result = _MergeSorter.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0xee1d2e4, 0xd40361d, 0x6f5320f, __pyx_k_depth_first_stack__last_stack_i); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xee1d2e4, 0xd40361d, 0x6f5320f, b'_depth_first_stack, _last_stack_item, _revno_to_branch_count, _scheduled_nodes, graph')
 *     __pyx_result = _MergeSorter.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__MergeSorter__set_state(<_MergeSorter> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xee1d2e4, 0xd40361d, 0x6f5320f, b'_depth_first_stack, _last_stack_item, _revno_to_branch_count, _scheduled_nodes, graph')
 *     __pyx_result = _MergeSorter.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__MergeSorter__set_state(<_MergeSorter> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = _MergeSorter.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__MergeSorter__set_state(<_MergeSorter> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__MergeSorter__set_state(_MergeSorter __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(1, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_6breezy_16_known_graph_pyx___pyx_unpickle__MergeSorter__set_state(((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xee1d2e4, 0xd40361d, 0x6f5320f, b'_depth_first_stack, _last_stack_item, _revno_to_branch_count, _scheduled_nodes, graph')
 *     __pyx_result = _MergeSorter.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__MergeSorter__set_state(<_MergeSorter> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__MergeSorter__set_state(<_MergeSorter> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__MergeSorter__set_state(_MergeSorter __pyx_result, __pyx_state: tuple):
 *     __pyx_result._depth_first_stack = __pyx_state[0]; __pyx_result._last_stack_item = __pyx_state[1]; __pyx_result._revno_to_branch_count = __pyx_state[2]; __pyx_result._scheduled_nodes = __pyx_state[3]; __pyx_result.graph = __pyx_state[4]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle__MergeSorter(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xee1d2e4, 0xd40361d, 0x6f5320f, b'_depth_first_stack, _last_stack_item, _revno_to_branch_count, _scheduled_nodes, graph')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy._known_graph_pyx.__pyx_unpickle__MergeSorter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__MergeSorter__set_state(<_MergeSorter> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__MergeSorter__set_state(_MergeSorter __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._depth_first_stack = __pyx_state[0]; __pyx_result._last_stack_item = __pyx_state[1]; __pyx_result._revno_to_branch_count = __pyx_state[2]; __pyx_result._scheduled_nodes = __pyx_state[3]; __pyx_result.graph = __pyx_state[4]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 5)
*/

static PyObject *__pyx_f_6breezy_16_known_graph_pyx___pyx_unpickle__MergeSorter__set_state(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__MergeSorter__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__MergeSorter__set_state(_MergeSorter __pyx_result, __pyx_state: tuple):
 *     __pyx_result._depth_first_stack = __pyx_state[0]; __pyx_result._last_stack_item = __pyx_state[1]; __pyx_result._revno_to_branch_count = __pyx_state[2]; __pyx_result._scheduled_nodes = __pyx_state[3]; __pyx_result.graph = __pyx_state[4]             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 5)
*/
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_depth_first_stack);
  __Pyx_DECREF(__pyx_v___pyx_result->_depth_first_stack);
  __pyx_v___pyx_result->_depth_first_stack = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_last_stack_item = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_revno_to_branch_count);
  __Pyx_DECREF(__pyx_v___pyx_result->_revno_to_branch_count);
  __pyx_v___pyx_result->_revno_to_branch_count = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_scheduled_nodes);
  __Pyx_DECREF(__pyx_v___pyx_result->_scheduled_nodes);
  __pyx_v___pyx_result->_scheduled_nodes = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->graph);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->graph);
  __pyx_v___pyx_result->graph = ((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle__MergeSorter__set_state(_MergeSorter __pyx_result, __pyx_state: tuple):
 *     __pyx_result._depth_first_stack = __pyx_state[0]; __pyx_result._last_stack_item = __pyx_state[1]; __pyx_result._revno_to_branch_count = __pyx_state[2]; __pyx_result._scheduled_nodes = __pyx_state[3]; __pyx_result.graph = __pyx_state[4]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 5)             # <<<<<<<<<<<<<<
*/
  __pyx_t_3 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 5); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle__MergeSorter__set_state(<_MergeSorter> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__MergeSorter__set_state(_MergeSorter __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._depth_first_stack = __pyx_state[0]; __pyx_result._last_stack_item = __pyx_state[1]; __pyx_result._revno_to_branch_count = __pyx_state[2]; __pyx_result._scheduled_nodes = __pyx_state[3]; __pyx_result.graph = __pyx_state[4]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 5)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy._known_graph_pyx.__pyx_unpickle__MergeSorter__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
/* #### Code section: module_exttypes ### */
static struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__KnownGraphNode __pyx_vtable_6breezy_16_known_graph_pyx__KnownGraphNode;

static PyObject *__pyx_tp_new_6breezy_16_known_graph_pyx__KnownGraphNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6breezy_16_known_graph_pyx__KnownGraphNode;
  p->key = Py_None; Py_INCREF(Py_None);
  p->parents = Py_None; Py_INCREF(Py_None);
  p->children = Py_None; Py_INCREF(Py_None);
  p->extra = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6breezy_16_known_graph_pyx__KnownGraphNode(PyObject *o) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *p = (struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6breezy_16_known_graph_pyx__KnownGraphNode) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->key);
  Py_CLEAR(p->parents);
  Py_CLEAR(p->children);
  Py_CLEAR(p->extra);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_6breezy_16_known_graph_pyx__KnownGraphNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *p = (struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->key) {
    e = (*v)(p->key, a); if (e) return e;
  }
  if (p->parents) {
    e = (*v)(p->parents, a); if (e) return e;
  }
  if (p->children) {
    e = (*v)(p->children, a); if (e) return e;
  }
  if (p->extra) {
    e = (*v)(p->extra, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6breezy_16_known_graph_pyx__KnownGraphNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *p = (struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)o;
  tmp = ((PyObject*)p->key);
  p->key = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->parents);
  p->parents = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->children);
  p->children = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->extra);
  p->extra = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6breezy_16_known_graph_pyx_15_KnownGraphNode_child_keys(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_10child_keys_1__get__(o);
}

static PyObject *__pyx_getprop_6breezy_16_known_graph_pyx_15_KnownGraphNode_parent_keys(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_11parent_keys_1__get__(o);
}

static PyObject *__pyx_getprop_6breezy_16_known_graph_pyx_15_KnownGraphNode_gdfo(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_4gdfo_1__get__(o);
}

static int __pyx_setprop_6breezy_16_known_graph_pyx_15_KnownGraphNode_gdfo(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_4gdfo_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_6breezy_16_known_graph_pyx__KnownGraphNode[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6breezy_16_known_graph_pyx__KnownGraphNode[] = {
  {"child_keys", __pyx_getprop_6breezy_16_known_graph_pyx_15_KnownGraphNode_child_keys, 0, 0, 0},
  {"parent_keys", __pyx_getprop_6breezy_16_known_graph_pyx_15_KnownGraphNode_parent_keys, 0, 0, 0},
  {"gdfo", __pyx_getprop_6breezy_16_known_graph_pyx_15_KnownGraphNode_gdfo, __pyx_setprop_6breezy_16_known_graph_pyx_15_KnownGraphNode_gdfo, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6breezy_16_known_graph_pyx__KnownGraphNode_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6breezy_16_known_graph_pyx__KnownGraphNode},
  {Py_tp_repr, (void *)__pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Represents a single object in the known graph.")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6breezy_16_known_graph_pyx__KnownGraphNode},
  {Py_tp_clear, (void *)__pyx_tp_clear_6breezy_16_known_graph_pyx__KnownGraphNode},
  {Py_tp_methods, (void *)__pyx_methods_6breezy_16_known_graph_pyx__KnownGraphNode},
  {Py_tp_getset, (void *)__pyx_getsets_6breezy_16_known_graph_pyx__KnownGraphNode},
  {Py_tp_init, (void *)__pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6breezy_16_known_graph_pyx__KnownGraphNode},
  {0, 0},
};
static PyType_Spec __pyx_type_6breezy_16_known_graph_pyx__KnownGraphNode_spec = {
  "breezy._known_graph_pyx._KnownGraphNode",
  sizeof(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6breezy_16_known_graph_pyx__KnownGraphNode_slots,
};
#else

static PyTypeObject __pyx_type_6breezy_16_known_graph_pyx__KnownGraphNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "breezy._known_graph_pyx.""_KnownGraphNode", /*tp_name*/
  sizeof(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6breezy_16_known_graph_pyx__KnownGraphNode, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  __pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Represents a single object in the known graph."), /*tp_doc*/
  __pyx_tp_traverse_6breezy_16_known_graph_pyx__KnownGraphNode, /*tp_traverse*/
  __pyx_tp_clear_6breezy_16_known_graph_pyx__KnownGraphNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6breezy_16_known_graph_pyx__KnownGraphNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6breezy_16_known_graph_pyx__KnownGraphNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6breezy_16_known_graph_pyx_15_KnownGraphNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6breezy_16_known_graph_pyx__KnownGraphNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSorter __pyx_vtable_6breezy_16_known_graph_pyx__MergeSorter;

static PyObject *__pyx_tp_new_6breezy_16_known_graph_pyx__MergeSorter(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *)o);
  p->__pyx_vtab = __pyx_vtabptr_6breezy_16_known_graph_pyx__MergeSorter;
  p->graph = ((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)Py_None); Py_INCREF(Py_None);
  p->_depth_first_stack = Py_None; Py_INCREF(Py_None);
  p->_revno_to_branch_count = Py_None; Py_INCREF(Py_None);
  p->_scheduled_nodes = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6breezy_16_known_graph_pyx__MergeSorter(PyObject *o) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *p = (struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6breezy_16_known_graph_pyx__MergeSorter) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->graph);
  Py_CLEAR(p->_depth_first_stack);
  Py_CLEAR(p->_revno_to_branch_count);
  Py_CLEAR(p->_scheduled_nodes);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_6breezy_16_known_graph_pyx__MergeSorter(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *p = (struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->graph) {
    e = (*v)(((PyObject *)p->graph), a); if (e) return e;
  }
  if (p->_depth_first_stack) {
    e = (*v)(p->_depth_first_stack, a); if (e) return e;
  }
  if (p->_revno_to_branch_count) {
    e = (*v)(p->_revno_to_branch_count, a); if (e) return e;
  }
  if (p->_scheduled_nodes) {
    e = (*v)(p->_scheduled_nodes, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6breezy_16_known_graph_pyx__MergeSorter(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *p = (struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *)o;
  tmp = ((PyObject*)p->graph);
  p->graph = ((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_depth_first_stack);
  p->_depth_first_stack = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_revno_to_branch_count);
  p->_revno_to_branch_count = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_scheduled_nodes);
  p->_scheduled_nodes = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6breezy_16_known_graph_pyx__MergeSorter[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_12_MergeSorter_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_12_MergeSorter_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6breezy_16_known_graph_pyx__MergeSorter_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6breezy_16_known_graph_pyx__MergeSorter},
  {Py_tp_doc, (void *)PyDoc_STR("This class does the work of computing the merge_sort ordering.\n\n    We have some small advantages, in that we get all the extra information\n    that KnownGraph knows, like knowing the child lists, etc.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6breezy_16_known_graph_pyx__MergeSorter},
  {Py_tp_clear, (void *)__pyx_tp_clear_6breezy_16_known_graph_pyx__MergeSorter},
  {Py_tp_methods, (void *)__pyx_methods_6breezy_16_known_graph_pyx__MergeSorter},
  {Py_tp_init, (void *)__pyx_pw_6breezy_16_known_graph_pyx_12_MergeSorter_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6breezy_16_known_graph_pyx__MergeSorter},
  {0, 0},
};
static PyType_Spec __pyx_type_6breezy_16_known_graph_pyx__MergeSorter_spec = {
  "breezy._known_graph_pyx._MergeSorter",
  sizeof(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6breezy_16_known_graph_pyx__MergeSorter_slots,
};
#else

static PyTypeObject __pyx_type_6breezy_16_known_graph_pyx__MergeSorter = {
  PyVarObject_HEAD_INIT(0, 0)
  "breezy._known_graph_pyx.""_MergeSorter", /*tp_name*/
  sizeof(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6breezy_16_known_graph_pyx__MergeSorter, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("This class does the work of computing the merge_sort ordering.\n\n    We have some small advantages, in that we get all the extra information\n    that KnownGraph knows, like knowing the child lists, etc.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6breezy_16_known_graph_pyx__MergeSorter, /*tp_traverse*/
  __pyx_tp_clear_6breezy_16_known_graph_pyx__MergeSorter, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6breezy_16_known_graph_pyx__MergeSorter, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6breezy_16_known_graph_pyx_12_MergeSorter_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6breezy_16_known_graph_pyx__MergeSorter, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6breezy_16_known_graph_pyx_KnownGraph __pyx_vtable_6breezy_16_known_graph_pyx_KnownGraph;

static PyObject *__pyx_tp_new_6breezy_16_known_graph_pyx_KnownGraph(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)o);
  p->__pyx_vtab = __pyx_vtabptr_6breezy_16_known_graph_pyx_KnownGraph;
  p->_nodes = Py_None; Py_INCREF(Py_None);
  p->_known_heads = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6breezy_16_known_graph_pyx_KnownGraph(PyObject *o) {
  struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *p = (struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6breezy_16_known_graph_pyx_KnownGraph) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_nodes);
  Py_CLEAR(p->_known_heads);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_6breezy_16_known_graph_pyx_KnownGraph(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *p = (struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->_nodes) {
    e = (*v)(p->_nodes, a); if (e) return e;
  }
  if (p->_known_heads) {
    e = (*v)(p->_known_heads, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6breezy_16_known_graph_pyx_KnownGraph(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *p = (struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *)o;
  tmp = ((PyObject*)p->_nodes);
  p->_nodes = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_known_heads);
  p->_known_heads = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6breezy_16_known_graph_pyx_10KnownGraph__nodes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_6_nodes_1__get__(o);
}

static int __pyx_setprop_6breezy_16_known_graph_pyx_10KnownGraph__nodes(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_6_nodes_3__set__(o, v);
  }
  else {
    return __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_6_nodes_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6breezy_16_known_graph_pyx_10KnownGraph__known_heads(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_12_known_heads_1__get__(o);
}

static int __pyx_setprop_6breezy_16_known_graph_pyx_10KnownGraph__known_heads(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_12_known_heads_3__set__(o, v);
  }
  else {
    return __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_12_known_heads_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6breezy_16_known_graph_pyx_10KnownGraph_do_cache(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_8do_cache_1__get__(o);
}

static int __pyx_setprop_6breezy_16_known_graph_pyx_10KnownGraph_do_cache(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_8do_cache_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_6breezy_16_known_graph_pyx_KnownGraph[] = {
  {"_initialize_nodes", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_5_initialize_nodes, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_4_initialize_nodes},
  {"_find_tails", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_7_find_tails, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_find_tips", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_9_find_tips, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_find_gdfo", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_11_find_gdfo, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"add_node", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_13add_node, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_12add_node},
  {"heads", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_15heads, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_14heads},
  {"topo_sort", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_17topo_sort, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_16topo_sort},
  {"gc_sort", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_19gc_sort, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_18gc_sort},
  {"merge_sort", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_21merge_sort, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_20merge_sort},
  {"get_parent_keys", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_23get_parent_keys, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_22get_parent_keys},
  {"get_child_keys", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_25get_child_keys, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph_24get_child_keys},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_27__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_29__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6breezy_16_known_graph_pyx_KnownGraph[] = {
  {"_nodes", __pyx_getprop_6breezy_16_known_graph_pyx_10KnownGraph__nodes, __pyx_setprop_6breezy_16_known_graph_pyx_10KnownGraph__nodes, 0, 0},
  {"_known_heads", __pyx_getprop_6breezy_16_known_graph_pyx_10KnownGraph__known_heads, __pyx_setprop_6breezy_16_known_graph_pyx_10KnownGraph__known_heads, 0, 0},
  {"do_cache", __pyx_getprop_6breezy_16_known_graph_pyx_10KnownGraph_do_cache, __pyx_setprop_6breezy_16_known_graph_pyx_10KnownGraph_do_cache, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6breezy_16_known_graph_pyx_KnownGraph_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6breezy_16_known_graph_pyx_KnownGraph},
  {Py_tp_doc, (void *)PyDoc_STR("This is a class which assumes we already know the full graph.")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6breezy_16_known_graph_pyx_KnownGraph},
  {Py_tp_clear, (void *)__pyx_tp_clear_6breezy_16_known_graph_pyx_KnownGraph},
  {Py_tp_methods, (void *)__pyx_methods_6breezy_16_known_graph_pyx_KnownGraph},
  {Py_tp_getset, (void *)__pyx_getsets_6breezy_16_known_graph_pyx_KnownGraph},
  {Py_tp_init, (void *)__pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6breezy_16_known_graph_pyx_KnownGraph},
  {0, 0},
};
static PyType_Spec __pyx_type_6breezy_16_known_graph_pyx_KnownGraph_spec = {
  "breezy._known_graph_pyx.KnownGraph",
  sizeof(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6breezy_16_known_graph_pyx_KnownGraph_slots,
};
#else

static PyTypeObject __pyx_type_6breezy_16_known_graph_pyx_KnownGraph = {
  PyVarObject_HEAD_INIT(0, 0)
  "breezy._known_graph_pyx.""KnownGraph", /*tp_name*/
  sizeof(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6breezy_16_known_graph_pyx_KnownGraph, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("This is a class which assumes we already know the full graph."), /*tp_doc*/
  __pyx_tp_traverse_6breezy_16_known_graph_pyx_KnownGraph, /*tp_traverse*/
  __pyx_tp_clear_6breezy_16_known_graph_pyx_KnownGraph, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6breezy_16_known_graph_pyx_KnownGraph, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6breezy_16_known_graph_pyx_KnownGraph, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6breezy_16_known_graph_pyx_10KnownGraph_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6breezy_16_known_graph_pyx_KnownGraph, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6breezy_16_known_graph_pyx__MergeSortNode __pyx_vtable_6breezy_16_known_graph_pyx__MergeSortNode;

static PyObject *__pyx_tp_new_6breezy_16_known_graph_pyx__MergeSortNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6breezy_16_known_graph_pyx__MergeSortNode;
  p->key = Py_None; Py_INCREF(Py_None);
  p->end_of_merge = Py_None; Py_INCREF(Py_None);
  p->left_parent = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)Py_None); Py_INCREF(Py_None);
  p->left_pending_parent = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)Py_None); Py_INCREF(Py_None);
  p->pending_parents = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6breezy_16_known_graph_pyx__MergeSortNode(PyObject *o) {
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *p = (struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6breezy_16_known_graph_pyx__MergeSortNode) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->key);
  Py_CLEAR(p->end_of_merge);
  Py_CLEAR(p->left_parent);
  Py_CLEAR(p->left_pending_parent);
  Py_CLEAR(p->pending_parents);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_6breezy_16_known_graph_pyx__MergeSortNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *p = (struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->key) {
    e = (*v)(p->key, a); if (e) return e;
  }
  if (p->end_of_merge) {
    e = (*v)(p->end_of_merge, a); if (e) return e;
  }
  if (p->left_parent) {
    e = (*v)(((PyObject *)p->left_parent), a); if (e) return e;
  }
  if (p->left_pending_parent) {
    e = (*v)(((PyObject *)p->left_pending_parent), a); if (e) return e;
  }
  if (p->pending_parents) {
    e = (*v)(p->pending_parents, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6breezy_16_known_graph_pyx__MergeSortNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *p = (struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *)o;
  tmp = ((PyObject*)p->key);
  p->key = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->end_of_merge);
  p->end_of_merge = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->left_parent);
  p->left_parent = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->left_pending_parent);
  p->left_pending_parent = ((struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->pending_parents);
  p->pending_parents = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6breezy_16_known_graph_pyx_14_MergeSortNode_revno(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_5revno_1__get__(o);
}

static PyObject *__pyx_getprop_6breezy_16_known_graph_pyx_14_MergeSortNode_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_3key_1__get__(o);
}

static int __pyx_setprop_6breezy_16_known_graph_pyx_14_MergeSortNode_key(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_3key_3__set__(o, v);
  }
  else {
    return __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_3key_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6breezy_16_known_graph_pyx_14_MergeSortNode_merge_depth(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_11merge_depth_1__get__(o);
}

static int __pyx_setprop_6breezy_16_known_graph_pyx_14_MergeSortNode_merge_depth(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_11merge_depth_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6breezy_16_known_graph_pyx_14_MergeSortNode_end_of_merge(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_12end_of_merge_1__get__(o);
}

static int __pyx_setprop_6breezy_16_known_graph_pyx_14_MergeSortNode_end_of_merge(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_12end_of_merge_3__set__(o, v);
  }
  else {
    return __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_12end_of_merge_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_6breezy_16_known_graph_pyx__MergeSortNode[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6breezy_16_known_graph_pyx__MergeSortNode[] = {
  {"revno", __pyx_getprop_6breezy_16_known_graph_pyx_14_MergeSortNode_revno, 0, 0, 0},
  {"key", __pyx_getprop_6breezy_16_known_graph_pyx_14_MergeSortNode_key, __pyx_setprop_6breezy_16_known_graph_pyx_14_MergeSortNode_key, 0, 0},
  {"merge_depth", __pyx_getprop_6breezy_16_known_graph_pyx_14_MergeSortNode_merge_depth, __pyx_setprop_6breezy_16_known_graph_pyx_14_MergeSortNode_merge_depth, 0, 0},
  {"end_of_merge", __pyx_getprop_6breezy_16_known_graph_pyx_14_MergeSortNode_end_of_merge, __pyx_setprop_6breezy_16_known_graph_pyx_14_MergeSortNode_end_of_merge, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6breezy_16_known_graph_pyx__MergeSortNode_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6breezy_16_known_graph_pyx__MergeSortNode},
  {Py_tp_repr, (void *)__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Tracks information about a node during the merge_sort operation.")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6breezy_16_known_graph_pyx__MergeSortNode},
  {Py_tp_clear, (void *)__pyx_tp_clear_6breezy_16_known_graph_pyx__MergeSortNode},
  {Py_tp_methods, (void *)__pyx_methods_6breezy_16_known_graph_pyx__MergeSortNode},
  {Py_tp_getset, (void *)__pyx_getsets_6breezy_16_known_graph_pyx__MergeSortNode},
  {Py_tp_init, (void *)__pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6breezy_16_known_graph_pyx__MergeSortNode},
  {0, 0},
};
static PyType_Spec __pyx_type_6breezy_16_known_graph_pyx__MergeSortNode_spec = {
  "breezy._known_graph_pyx._MergeSortNode",
  sizeof(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6breezy_16_known_graph_pyx__MergeSortNode_slots,
};
#else

static PyTypeObject __pyx_type_6breezy_16_known_graph_pyx__MergeSortNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "breezy._known_graph_pyx.""_MergeSortNode", /*tp_name*/
  sizeof(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6breezy_16_known_graph_pyx__MergeSortNode, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Tracks information about a node during the merge_sort operation."), /*tp_doc*/
  __pyx_tp_traverse_6breezy_16_known_graph_pyx__MergeSortNode, /*tp_traverse*/
  __pyx_tp_clear_6breezy_16_known_graph_pyx__MergeSortNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6breezy_16_known_graph_pyx__MergeSortNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6breezy_16_known_graph_pyx__MergeSortNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6breezy_16_known_graph_pyx_14_MergeSortNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6breezy_16_known_graph_pyx__MergeSortNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_6breezy_16_known_graph_pyx_NULL_REVISION = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_6breezy_16_known_graph_pyx__KnownGraphNode = &__pyx_vtable_6breezy_16_known_graph_pyx__KnownGraphNode;
  __pyx_vtable_6breezy_16_known_graph_pyx__KnownGraphNode.clear_references = (PyObject *(*)(struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *))__pyx_f_6breezy_16_known_graph_pyx_15_KnownGraphNode_clear_references;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6breezy_16_known_graph_pyx__KnownGraphNode_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode)) __PYX_ERR(0, 45, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6breezy_16_known_graph_pyx__KnownGraphNode_spec, __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode) < (0)) __PYX_ERR(0, 45, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode = &__pyx_type_6breezy_16_known_graph_pyx__KnownGraphNode;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode) < (0)) __PYX_ERR(0, 45, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode->tp_dictoffset && __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode, __pyx_vtabptr_6breezy_16_known_graph_pyx__KnownGraphNode) < (0)) __PYX_ERR(0, 45, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode) < (0)) __PYX_ERR(0, 45, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_KnownGraphNode, (PyObject *) __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode) < (0)) __PYX_ERR(0, 45, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode) < (0)) __PYX_ERR(0, 45, __pyx_L1_error)
  __pyx_vtabptr_6breezy_16_known_graph_pyx__MergeSorter = &__pyx_vtable_6breezy_16_known_graph_pyx__MergeSorter;
  __pyx_vtable_6breezy_16_known_graph_pyx__MergeSorter._get_ms_node = (struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *(*)(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *, struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *))__pyx_f_6breezy_16_known_graph_pyx_12_MergeSorter__get_ms_node;
  __pyx_vtable_6breezy_16_known_graph_pyx__MergeSorter._push_node = (PyObject *(*)(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *, struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *, long))__pyx_f_6breezy_16_known_graph_pyx_12_MergeSorter__push_node;
  __pyx_vtable_6breezy_16_known_graph_pyx__MergeSorter._pop_node = (PyObject *(*)(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *))__pyx_f_6breezy_16_known_graph_pyx_12_MergeSorter__pop_node;
  __pyx_vtable_6breezy_16_known_graph_pyx__MergeSorter._schedule_stack = (PyObject *(*)(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *))__pyx_f_6breezy_16_known_graph_pyx_12_MergeSorter__schedule_stack;
  __pyx_vtable_6breezy_16_known_graph_pyx__MergeSorter.topo_order = (PyObject *(*)(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSorter *))__pyx_f_6breezy_16_known_graph_pyx_12_MergeSorter_topo_order;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6breezy_16_known_graph_pyx__MergeSorter_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter)) __PYX_ERR(0, 708, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6breezy_16_known_graph_pyx__MergeSorter_spec, __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter) < (0)) __PYX_ERR(0, 708, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter = &__pyx_type_6breezy_16_known_graph_pyx__MergeSorter;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter) < (0)) __PYX_ERR(0, 708, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter->tp_dictoffset && __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter, __pyx_vtabptr_6breezy_16_known_graph_pyx__MergeSorter) < (0)) __PYX_ERR(0, 708, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter) < (0)) __PYX_ERR(0, 708, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_MergeSorter, (PyObject *) __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter) < (0)) __PYX_ERR(0, 708, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter) < (0)) __PYX_ERR(0, 708, __pyx_L1_error)
  __pyx_vtabptr_6breezy_16_known_graph_pyx_KnownGraph = &__pyx_vtable_6breezy_16_known_graph_pyx_KnownGraph;
  __pyx_vtable_6breezy_16_known_graph_pyx_KnownGraph._get_or_create_node = (struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *(*)(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *, PyObject *))__pyx_f_6breezy_16_known_graph_pyx_10KnownGraph__get_or_create_node;
  __pyx_vtable_6breezy_16_known_graph_pyx_KnownGraph._populate_parents = (PyObject *(*)(struct __pyx_obj_6breezy_16_known_graph_pyx_KnownGraph *, struct __pyx_obj_6breezy_16_known_graph_pyx__KnownGraphNode *, PyObject *))__pyx_f_6breezy_16_known_graph_pyx_10KnownGraph__populate_parents;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6breezy_16_known_graph_pyx_KnownGraph_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph)) __PYX_ERR(0, 174, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6breezy_16_known_graph_pyx_KnownGraph_spec, __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph) < (0)) __PYX_ERR(0, 174, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph = &__pyx_type_6breezy_16_known_graph_pyx_KnownGraph;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph) < (0)) __PYX_ERR(0, 174, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph->tp_dictoffset && __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 174, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_6breezy_16_known_graph_pyx_10KnownGraph___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_6breezy_16_known_graph_pyx_10KnownGraph___init__.doc = __pyx_doc_6breezy_16_known_graph_pyx_10KnownGraph___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_6breezy_16_known_graph_pyx_10KnownGraph___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph, __pyx_vtabptr_6breezy_16_known_graph_pyx_KnownGraph) < (0)) __PYX_ERR(0, 174, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph) < (0)) __PYX_ERR(0, 174, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_KnownGraph, (PyObject *) __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph) < (0)) __PYX_ERR(0, 174, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph) < (0)) __PYX_ERR(0, 174, __pyx_L1_error)
  __pyx_vtabptr_6breezy_16_known_graph_pyx__MergeSortNode = &__pyx_vtable_6breezy_16_known_graph_pyx__MergeSortNode;
  __pyx_vtable_6breezy_16_known_graph_pyx__MergeSortNode.has_pending_parents = (int (*)(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *))__pyx_f_6breezy_16_known_graph_pyx_14_MergeSortNode_has_pending_parents;
  __pyx_vtable_6breezy_16_known_graph_pyx__MergeSortNode._revno = (PyObject *(*)(struct __pyx_obj_6breezy_16_known_graph_pyx__MergeSortNode *))__pyx_f_6breezy_16_known_graph_pyx_14_MergeSortNode__revno;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6breezy_16_known_graph_pyx__MergeSortNode_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode)) __PYX_ERR(0, 650, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6breezy_16_known_graph_pyx__MergeSortNode_spec, __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode) < (0)) __PYX_ERR(0, 650, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode = &__pyx_type_6breezy_16_known_graph_pyx__MergeSortNode;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode) < (0)) __PYX_ERR(0, 650, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode->tp_dictoffset && __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode, __pyx_vtabptr_6breezy_16_known_graph_pyx__MergeSortNode) < (0)) __PYX_ERR(0, 650, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode) < (0)) __PYX_ERR(0, 650, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_MergeSortNode, (PyObject *) __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode) < (0)) __PYX_ERR(0, 650, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode) < (0)) __PYX_ERR(0, 650, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_2_4(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  0, 0,
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__known_graph_pyx(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__known_graph_pyx},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, __Pyx_FREETHREADING_COMPATIBLE},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "_known_graph_pyx",
      __pyx_k_Implementation_of_Graph_algorith, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit__known_graph_pyx(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__known_graph_pyx(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
    PY_INT64_T current_id = GraalPyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__known_graph_pyx(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_known_graph_pyx' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "_known_graph_pyx" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
    PyErr_Clear();
    __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
    if (!__Pyx_RefNanny)
        Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  
__Pyx_RefNannySetupContext("PyInit__known_graph_pyx", 0);
  __Pyx_init_runtime_version();
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Library function declarations ---*/
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__pyx_module_is_main_breezy___known_graph_pyx) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name, __pyx_mstate_global->__pyx_n_u_main) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "breezy._known_graph_pyx")) {
      if (unlikely((PyDict_SetItemString(modules, "breezy._known_graph_pyx", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  (void)__Pyx_modinit_function_export_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_type_init_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  (void)__Pyx_modinit_function_import_code(__pyx_mstate);
  /*--- Execution code ---*/

  /* "breezy/_known_graph_pyx.pyx":35
 *                             PyTuple_GET_SIZE, PyTuple_New, PyTuple_SET_ITEM)
 * 
 * import collections             # <<<<<<<<<<<<<<
 * import gc
 * 
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_collections, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_collections, __pyx_t_2) < (0)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/_known_graph_pyx.pyx":36
 * 
 * import collections
 * import gc             # <<<<<<<<<<<<<<
 * 
 * from . import errors, revision
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_gc, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_gc, __pyx_t_2) < (0)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/_known_graph_pyx.pyx":38
 * import gc
 * 
 * from . import errors, revision             # <<<<<<<<<<<<<<
 * 
 * 
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_errors,__pyx_mstate_global->__pyx_n_u_revision};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u__3, __pyx_imported_names, 2, __pyx_mstate_global->__pyx_kp_u_breezy, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 38, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_errors,__pyx_mstate_global->__pyx_n_u_revision};
    for (__pyx_t_3=0; __pyx_t_3 < 2; __pyx_t_3++) {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 38, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 38, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/_known_graph_pyx.pyx":42
 * 
 * cdef object NULL_REVISION
 * NULL_REVISION = revision.NULL_REVISION             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_revision); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_NULL_REVISION); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6breezy_16_known_graph_pyx_NULL_REVISION);
  __Pyx_DECREF_SET(__pyx_v_6breezy_16_known_graph_pyx_NULL_REVISION, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_15_KnownGraphNode_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_KnownGraphNode___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_4) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__KnownGraphNode, (type(self), 0x1857be3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__KnownGraphNode__set_state(self, __pyx_state)
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_15_KnownGraphNode_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_KnownGraphNode___setstate_cytho, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__KnownGraphNode, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_4) < (0)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "breezy/_known_graph_pyx.pyx":120
 * 
 * 
 * def get_key(node):             # <<<<<<<<<<<<<<
 *     cdef _KnownGraphNode real_node
 *     real_node = node
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_1get_key, 0, __pyx_mstate_global->__pyx_n_u_get_key, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[2])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_get_key, __pyx_t_4) < (0)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "breezy/_known_graph_pyx.pyx":240
 *         node.parents = parent_nodes
 * 
 *     def _initialize_nodes(self, parent_map):             # <<<<<<<<<<<<<<
 *         """Populate self._nodes.
 * 
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_5_initialize_nodes, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_KnownGraph__initialize_nodes, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[3])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph, __pyx_mstate_global->__pyx_n_u_initialize_nodes, __pyx_t_4) < (0)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "breezy/_known_graph_pyx.pyx":266
 *             self._populate_parents(node, parent_keys)
 * 
 *     def _find_tails(self):             # <<<<<<<<<<<<<<
 *         cdef PyObject *temp_node
 *         cdef _KnownGraphNode node
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_7_find_tails, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_KnownGraph__find_tails, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[4])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph, __pyx_mstate_global->__pyx_n_u_find_tails, __pyx_t_4) < (0)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "breezy/_known_graph_pyx.pyx":280
 *         return tails
 * 
 *     def _find_tips(self):             # <<<<<<<<<<<<<<
 *         cdef PyObject *temp_node
 *         cdef _KnownGraphNode node
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_9_find_tips, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_KnownGraph__find_tips, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[5])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph, __pyx_mstate_global->__pyx_n_u_find_tips, __pyx_t_4) < (0)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "breezy/_known_graph_pyx.pyx":293
 *         return tips
 * 
 *     def _find_gdfo(self):             # <<<<<<<<<<<<<<
 *         cdef _KnownGraphNode node
 *         cdef _KnownGraphNode child
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_11_find_gdfo, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_KnownGraph__find_gdfo, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[6])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph, __pyx_mstate_global->__pyx_n_u_find_gdfo, __pyx_t_4) < (0)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "breezy/_known_graph_pyx.pyx":328
 *                     child.seen = 0
 * 
 *     def add_node(self, key, parent_keys):             # <<<<<<<<<<<<<<
 *         """Add a new node to the graph.
 * 
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_13add_node, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_KnownGraph_add_node, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[7])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph, __pyx_mstate_global->__pyx_n_u_add_node, __pyx_t_4) < (0)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "breezy/_known_graph_pyx.pyx":398
 *                     pending_append(child_node)
 * 
 *     def heads(self, keys):             # <<<<<<<<<<<<<<
 *         """Return the heads from amongst keys.
 * 
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_15heads, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_KnownGraph_heads, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[8])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph, __pyx_mstate_global->__pyx_n_u_heads, __pyx_t_4) < (0)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "breezy/_known_graph_pyx.pyx":490
 *         return heads
 * 
 *     def topo_sort(self):             # <<<<<<<<<<<<<<
 *         """Return the nodes in topological order.
 * 
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_17topo_sort, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_KnownGraph_topo_sort, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[9])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph, __pyx_mstate_global->__pyx_n_u_topo_sort, __pyx_t_4) < (0)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "breezy/_known_graph_pyx.pyx":544
 *         return topo_order
 * 
 *     def gc_sort(self):             # <<<<<<<<<<<<<<
 *         """Return a reverse topological ordering which is 'stable'.
 * 
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_19gc_sort, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_KnownGraph_gc_sort, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[10])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph, __pyx_mstate_global->__pyx_n_u_gc_sort, __pyx_t_4) < (0)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "breezy/_known_graph_pyx.pyx":616
 *         return result
 * 
 *     def merge_sort(self, tip_key):             # <<<<<<<<<<<<<<
 *         """Compute the merge sorted graph output."""
 *         cdef _MergeSorter sorter
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_21merge_sort, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_KnownGraph_merge_sort, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[11])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph, __pyx_mstate_global->__pyx_n_u_merge_sort, __pyx_t_4) < (0)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "breezy/_known_graph_pyx.pyx":626
 *         return sorter.topo_order()
 * 
 *     def get_parent_keys(self, key):             # <<<<<<<<<<<<<<
 *         """Get the parents for a key
 * 
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_23get_parent_keys, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_KnownGraph_get_parent_keys, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[12])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph, __pyx_mstate_global->__pyx_n_u_get_parent_keys, __pyx_t_4) < (0)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "breezy/_known_graph_pyx.pyx":638
 *         return self._nodes[key].parent_keys
 * 
 *     def get_child_keys(self, key):             # <<<<<<<<<<<<<<
 *         """Get the children for a key
 * 
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_25get_child_keys, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_KnownGraph_get_child_keys, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[13])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph, __pyx_mstate_global->__pyx_n_u_get_child_keys, __pyx_t_4) < (0)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_27__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_KnownGraph___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[14])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_4) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_KnownGraph, (type(self), 0xe34a3b8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_KnownGraph__set_state(self, __pyx_state)
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_10KnownGraph_29__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_KnownGraph___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[15])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx_KnownGraph, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_4) < (0)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_14_MergeSortNode_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_MergeSortNode___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[16])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_4) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__MergeSortNode, (type(self), 0x30884b3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__MergeSortNode__set_state(self, __pyx_state)
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_14_MergeSortNode_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_MergeSortNode___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[17])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSortNode, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_4) < (0)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_12_MergeSorter_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_MergeSorter___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[18])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_4) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__MergeSorter, (type(self), 0xee1d2e4, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__MergeSorter__set_state(self, __pyx_state)
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_12_MergeSorter_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_MergeSorter___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[19])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_16_known_graph_pyx__MergeSorter, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_4) < (0)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle__KnownGraphNode(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x1857be3, 0x4591477, 0x1dbac4a, b'children, extra, gdfo, key, parents, seen')
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_3__pyx_unpickle__KnownGraphNode, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__KnownGraphNode, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[20])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__KnownGraphNode, __pyx_t_4) < (0)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * cdef extern from *:             # <<<<<<<<<<<<<<
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_5__pyx_unpickle_KnownGraph, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_KnownGraph, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[21])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_KnownGraph, __pyx_t_4) < (0)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle__MergeSortNode(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x30884b3, 0xd890a73, 0x7a2ed2f, b'_revno_first, _revno_last, _revno_second, completed, end_of_merge, is_first_child, key, left_parent, left_pending_parent, merge_depth, pending_parents, seen_by_child')
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_7__pyx_unpickle__MergeSortNode, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__MergeSortNode, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[22])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__MergeSortNode, __pyx_t_4) < (0)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * cdef extern from *:             # <<<<<<<<<<<<<<
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_16_known_graph_pyx_9__pyx_unpickle__MergeSorter, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__MergeSorter, NULL, __pyx_mstate_global->__pyx_n_u_breezy__known_graph_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[23])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__MergeSorter, __pyx_t_4) < (0)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "breezy/_known_graph_pyx.pyx":1
 * # Copyright (C) 2009, 2010 Canonical Ltd             # <<<<<<<<<<<<<<
 * #
 * # This program is free software; you can redistribute it and/or modify
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_4) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init breezy._known_graph_pyx", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init breezy._known_graph_pyx");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */
/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);

  /* Cached unbound methods */
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.method_name = &__pyx_mstate->__pyx_n_u_items;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.method_name = &__pyx_mstate->__pyx_n_u_values;
  __pyx_mstate->__pyx_umethod_PyList_Type_pop.type = (PyObject*)&PyList_Type;
  __pyx_mstate->__pyx_umethod_PyList_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  return 0;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);
  __Pyx_RefNannyFinishContext();
  return 0;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  {
    const struct { const unsigned int length: 10; } index[] = {{1},{179},{35},{22},{1},{0},{1},{1},{1},{8},{7},{27},{7},{7},{7},{6},{7},{2},{7},{49},{16},{9},{4},{35},{5},{13},{5},{48},{5},{6},{15},{15},{10},{15},{33},{35},{28},{30},{21},{22},{21},{28},{19},{18},{25},{26},{16},{21},{20},{14},{32},{34},{12},{30},{32},{13},{20},{0},{8},{6},{18},{23},{15},{5},{10},{10},{9},{17},{7},{5},{18},{11},{5},{8},{5},{8},{6},{20},{10},{11},{10},{8},{2},{7},{14},{7},{15},{12},{5},{9},{17},{13},{5},{3},{4},{11},{9},{8},{11},{10},{10},{8},{10},{8},{7},{9},{4},{5},{11},{11},{10},{11},{7},{7},{14},{15},{3},{7},{3},{6},{11},{14},{12},{11},{10},{25},{30},{29},{27},{14},{12},{9},{10},{17},{13},{7},{6},{7},{8},{4},{12},{10},{12},{19},{4},{6},{5},{5},{4},{8},{9},{16},{8},{7},{9},{4},{10},{9},{6},{12},{6},{11},{14},{11},{315},{89},{589},{24},{90},{276},{203},{72},{17},{17},{408},{11},{155},{115},{292},{192},{11},{57},{57},{55},{55}};
    #if (CYTHON_COMPRESS_STRINGS) == 3 && __PYX_LIMITED_VERSION_HEX >= 0x030e0000 /* compression: zstd (2725 bytes) */
const char* const cstring = "(\265/\375`\264\023\335T\000\212t\374\030:\360\0327\314\300\010)$e\220c\204F\301t\246X\370 \23707\205\265\202\010\261\271B$\342\370=\226+%\005\204\240\n\000\000\010\000$HI\002\000\256Z\250\367\214(HH|n\001v\001\210\001g\231\n\331l\246\032\010\314=]\264\255\264\020\211c%$\"I\241\241\036\243\236\244\234\"\022\010\010Ih\325FA\244\nT\267\216Z\365\317\314\332Gn\275W\225%\313(R_\235h>\367\231\236\300\223E\317\274I3={\233$KR\227\255\251W\222,\021\"\177Vu]s4Z\nov@~Qy\325\353\376xa\307\027*-\026\313\004\346\037\371\222\236%\032\005\266\352\211\216\231m\177#\n\334\344\252\222\233EM\366\004\216l\366\215\343\313T\2317\253\356T\365\261-p\324\274q\233h\253\3726\253\372\242^\324\236\0315\216e\324\313f\327\2637\246Z\243\274\314\236&\360\305DO_9\226X\t\201\244\017\230\327(2\255R\323)\322\244Dx\245\365\312\323\244if\232\222\021\220\361\305\267Ou\255\362m\375S\215Fm\252\351\270\331\323J\217#\276r\347\037{\376F\241\236\032w\252\327LT\242Y4\353h\252s=\315\321#*=\311T=w\336j\232f\217qkvL5\233\201\243\272GS\353tlU\341\330\353\031KX\r\024\3561=7z\266\274\261\247\216\254f\331\214\226\274\226\035\331]j)\275\216f\232\307\\K\216\204(\207hLW\275\313s\335\354(D\365\316\253\346\235\2506G\366\034S\325\267\2057\343\316\276=\352\255\n}\0049\257\004\003\200\261\006\347\356\343\030\017\273\216\362\342.\270\263\214\256\303e\\\"\001\206\263\327B\001\014\353\021|\242{p\212\002\030\327\223\270\210\353`\000\320\360u\353\261\016\344\362\320\374\024\277\230\3010\357\304Qx\026\256\267p\333,\326uY\357\340X\301!\257\204\0028\273\020~%c\366g\t\360\316\013x!\317\361\033\247H\240q\035\022\342\374\256\214\177\264\305\273\256\353!\234\242\026?\324b\234\017\360\0373\030\327o\335\344\0304\200\201/\334\020\341\rz\002\377Q\313\365!8\330=:D\017rw\236\252\347\336Q\373\265Y\034\3447\334\305\335^\014\t\260\256o8|\207\004\230\306K\027\306\013\241\000\246\275\006\036\247\343:\224\013a\000p\361\245\353\302\275I^\347Gt\256\213\027\202;\223\305\260.\332\273\342o>D\006\303\272X\335\242\231J\263-\260]\265\271\252\347f\365\006\262""\244\331C5{jv\325\272\314(\320J\223h\326+\232\333 /o\252U\243S\345\tL\333\215\266\254^\177-;Uj\253\263\247W \273{\032\005\252@\222E\251\322\277\234j\2645\325\364\374i\2360=\363\217\334,\311\236$YF\221d\231\266=\222d\265\366\225\271$U\323l\203\332dK\215\242j\311\323e^7\211\226+\260M\263\376\351}\222$I\356l\236f\232KV\267\352\252\226%j\020\037\242`\307u\274\347\372\2704?\325yxE\356>\034\307m7n\353I\216x-&\032\r\251\336\034\353o|\257o0\374\370\202\305\3355\332[z\030\256\240\327l\254\305\023E\374\245!\246u\tNq\013\276\034\270\333/;:\022_r\037N\306\337\301m\003\\%\2770~/p\211\031\316P\267\371\r\217\3065\205Z\274\203\371\017\211\337A\344\r\231\336\332\013{\215|\315\257\375\246\273\320\240\356F\306N\307yh\305\222\3503\224\261x\374\233\336\352^\201\223op\232\001\300\001\214J\327{\025n\203\001\300\302\227\256+\350\3050\345\271\202^9\030Cwr{h\334\200\026e\341\346\316\276\353\034\022\261P\342&\001\343\336\333K\323\033\372\225\033Q\313\360\007}\254\3010\242\324\273\361\322^\223\335\2606\374\370N\265\"J\221\202x4\324\rcF\276\314\253\300:\356\000\343\005\270\016\302i\264Y|\200\363 3\001\\*\350\213|O\307=\360\006\276\202\253\352;\020\024\237\301 {\033\357\3227\317\034\306u(\210\335\006\037\363\037\021?\033\304\267.t#`\335\226\020\253l\374\207\241\304\233\343Gl\277\005\343g\356s!f\360\216k\354\336\036/\363\301~\301'\022`\234\273S\020\343x\013g\335\030\014\177\224f\276\343=\356\203\001@\210\316\343\274H\304\317\335\372\307\215\177\300\206\301\267\251\035\3414>\025\3538\216\203\270\r\035\326G\206\0378/\004\031\254+\246C.\237}\t\314\212\307\300\212%\303\247%\376\262#\372%D\033\334\227\35038\243\242/D\304A\371Y\0372\270h\367\247\020\202\034#\350!b\333\320\343%\341\305r\21655\305\214\214bCC1 \240\330\316N\254\206\313\346\233\303\224\241~\253\274\211\355S\265\277\r\376\240\356\037\reB\276\331;h?\307\273b\024\021f\270\263\275I\214*McOO1&\246XRRl\267\213\371\370\3006\360\003}\337\227\177>\372GD\337\250(P\214 \346\004\360\002r\016xz\n\370\371\t\330\t\330\211\t\205\261\266\215\221d\314\311)""\026\024\024\353\351\211\335\304n`\\\000\000_[\277\377;\377\331\364\233\222\376H\321$\352\254\374\215W\350\026\323\212\036D\254\227 \347Ep\305\035\"\217\321+\360F\370\031\332{\344{\272\254c\004\343\\\254\"\376\014\225d\357\t\303\017b]\375+\366\263R_\252\367{\345/\241>\224\322W*\372f1\245\250\211UC\002\007\245\035\007\335\210\032\317q\225\274\250\272\010d\260\216+\203\203\305\334NA\307\3416\367\310kA\000\240\340\353\355\371\235'\301\371\226\256\001R\361\363\272T\300\271\002\016\276\371\345\352E\375\005\366\362\274O\347\230\270y\032n\353\010\\\242\006\337\206\261a\207\025\307\250\305Y\367\246qpx\276\273W\3654\017\022\361\257|\216\342\333h7ov_\017\310\301<\257b+\004\3676\274\006\014\267\027+\202\025\250\262\242\022\t!RJdD$M\ni\014\202\030\004\306\2424\2227{\022*EI\230A\246dDd\244\244\214\222\224\244\320\030\200\247\274\356j\211\341S\351\264\335\236s)\014\265V\016\326\3059n\024\n\327z\327\243-pf\340\206Bh\226kR\325\033\023\311\303`\255\333\326\203\375\254W\n\331\377\352\205\222\016\031=3\311+1\360?\353\265\006\213]\251\227\306n\\\224\263\363Td\006\2622T\242\022\330\320\211\231V\234\337\215\260\227\021\240\003\350x\023b\021e+\365\336\010\373\222@\274\261\304LJ$K\267\302PoU\376\234\3616\310\353\300\275(\341\251\307\377Z2\346\262\214\211\312\302\347\026M\"\327\203r\\X\243\014L`\024>\324S\361\244\362}w\025\320\003\336\204\366\205\220\243w\275C*\rt\271\347\033\301\177\233\312\231I\305h\3067\262\031&\260\325s%\256|\\\372\321\274(y\201\016\323\360\254rO\332\373{\3547+\330\251\2076\270#\352\"\316:\241\r\327\025O\240\330v\307[\323l\032\361fW)\022\301^O\212\001\303j\250\211\000-\230\\x)\311\013\026{w\001b\000\003)\350?\025\372\246\354\270\274\265\010\335\032\310W\361,t\246\367D\234\255dK\336\013D\033*\274M\277\264\207\346a\0279uk\025\216\036$\251H/\352\325\346\022-\226N\255\033\235\213\204\225\200XY\tb?\205\305\213\265F~l\223p\212v\364\317\r$\261^\374\304?\263\322\271\245A\326\016\225qN\373\241\240\034\0160$\215s\351\037D\230eb\030\326\272\266V\347t&*\345{7ER\200p\007\205\032!""\n\315\312\242R\"\357\346U\227\362\313\212:\230\314[\361\336\267\201E\233n\3275\367\236\377\327\250\236\325\300\250i\200\307\203\251\0136\227\354\027\221\025\312=u_h\314\301I\354\020\374q\n\231\2751U\2633j\004I\302$1\313\346BH!\252\265\326hr\235\364\213l\017\262*\300\242x2\3060\017\371\300\213w\021@\024\360&\021\t\221\013\0004\262\203\005\302C\334\301%o\315\371\2032\030\370Bd\237\037\221\240;\262R \3410\321n$4\374\364%h\177\310\337y\247\277\204\303\023q&\275q\375\267\311\277A\312\0100\340\313\2564\315GnMez\273'\365\302\177j\205\037\366\004n\223/\311\375\0354\357\372\257\200$\360\300\3049\245\357T\226\3661\302\201\361\010\024\031\340Dsd<\372bP\370{\311\321\022Y\025\322\262\330\010Us\027\224\311\000T\224\225\224\327$\340\016\236\233\300\322\365\260;[?v\336\311V\202i&\303\271\315Sci\021\310\377\267\014\026\0046\267|\205\372X4\360p\347\232uNC\035H\"\022]\240S\000\020Y\225\275\265@y\205P\\\035\333\262\276cE\205\331GT\016\336!\303~\212!\266\274hS\302S\253\266pH\207A\342%\235\360\016\\\343\330JK\3656\222\337e\341'\013|\355\353[\242\314\324\332C\275\371\255Hl\007\025){\"\235u\313\3345\315\264\016{\004\244$\365\021.\205r\363c\244LR).\006o\021}\302~r\242C\214G\211\370\227\210\352\003c\024`$Y\003\316\016Y\315\361\210\241\367\003\343\322\212#bq\206\034G.\246\352\331\334\247\262\247x\273\200\310\000\244@\322A\225`\026\t\023n*G\2258#\206<\200\267\235\210\210\037\217\304\327\373\335\034\203!>\206\352\363*\010\3654>Z\330\033q\020\246/\215K\030wo\331-C\317\255\035\001\017\177Y\350\025\274\306\377U[L\360\2127N\177\361TQ\333\324\325R\222\021W\344q\367\306\035\377\302\302-\272a\022V<\312H\026 \312D\025\342\235\036[r\273\210\2305\240&nl+d\261\214\317\014\362i\016\207\323\233\370\304\357H\205\256\225?\n\336n\212&\210\320D\206\245\326\224;\226K\\\002\207\241,\0176gW\377\353\300\202\331\365\276wd\024\260\032\321\314\3636R\270\242*\223X\204L\242\372\003\347\321\360*\030g3\0069RP\333\022:\202\265\001";
    PyObject *data = __Pyx_DecompressString(cstring, 2725, 3);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) == 2 /* compression: bz2 (2796 bytes) */
const char* const cstring = "BZh91AY&SY\377\307\204\022\000\002,\377\377\377\377\377\377\377\377\377\377\377\377\377\373\277\377\377\377\300@@@@@@@@@@@@\000@\000`\n\177[\0362\366\323\266\024\204\364b\t\330\325(c\200\000\340\022\210(\232SM=M= \361L\324\364#\304m56L\230\215OS\3653SQ\262\246\2306\251\342\237\252z=\032\002G\246S\304\206\207\264Q\355!\250$\220#D\301\014\223FFD\3012\243\312i\350\2324\003 \000i\350\3224\003\324\003C@\000i\240\365\017P*\200\320\320\214\230\t\2020\232d\300\t\200&\206L\000\23054\323\000\004h\304`\215\000\321\243@\224\320\202\005\000\220\364F\215\2051\032\003A\240\000\000\000\000\000\000\000\003@\003@A\200\002`\000#\000\000\214\000\021\243\023L\0010\000\004\3020\214\004\311\210\304\014\t\024\215\023\324\324 z\236\243\324i\246i\0324\032h\000\000\000\000\000\r\003@\0004\000\006\200\365.$)\006\004\256\254\346o\346\314`\316ii\241\221iT\204c6&\217\236G\341d\237\223\364\254\367\\\346\016x\215\"\215\211\356}\320\243\325\3129\212\350\250\025\271|~z\372Z655\254\352lmmz\275\217\341\307\260\327\300\265\376oo,\262\307,\371e\216Yc\226:r\313#V\037\202\211m\rz\316\n(\242\217\372\204\\`\301\204\211)\027\305\352\206\01051o\021\"\225\252\005\030\224\304$%]\202m6\333m\24640m\211\203m\032\002\265U\tL\246\221I\005\025\\5\r\242j\200\220 \256\3046!\353h\n\005Q#X\020\303g\211\377\014^X\3453\t\004\350<\001\236ha$\230\022\020\206J\303\3303\245D\230\314\256\230j^\3311#4\265\030\2013\347\252\351S\260U}zM\224V\246\273c\210\332\004\331\257\230\325jd\204\311\032\227P\302CSC\321\031\334^&\031\230\251\275\361\020\250i\242\302\242\350`\213J\0367\200\256\017 \327\316e\324(\201\"\224\325\320v\302\321\\\2313S4\241H\352Uu\331\010\231\332.\335\230\355\"\234\314\352V\325Glp45\316\317[\025b\250m\241[+e\300\001\260D\006\tw\324\371\3100\230\005\r\365\0218B\251e\010\031\276\264\223L%\020\300\374\004\212\0060\202\260{-\320@\"\345\203\270\203\331\306{!\345\302\177\342\211Z\343\310\372\252\203\352\2156\030\037Ga\352\353:\271Q\277\245Q\332\215\354r\002\257\003\233\254 \210tK\301\034t\317Gg[\332qE\014K[\035.<\364\321)Vz""\236)\247\224T\365\352v8\232\225tcf\367\335#L~\366\371\321_o\213N\027[\276s\243D\203\034\256\242\217Z\312a8\032\317o\356Nh\237\203\217\317\3760\231y\321\321\336k\030\350g9\316\0146\360\340\215K\274\302\277\242\350\312\335\347i\234\200\\\260\325\272-\005\222\010\275\014b\275\002\237\234\240\342QR\032\322\312\036B\223*Jj\264\213\035\230\245\250\276?sLK3[\025YL\244\210\370\363<yE\317\027\235\357\035\\\276\377\216\270\2421\237\371\327\340\360\371\257\353\344\333\2732\314X\234Yb\326(C\177\256\340(=\354^\237J\370\234\034\321\331),\311\3536\310$B\003\2423\301\234\336$\211\303\314\257\354\343\307eJ\252[K\335\231{\035\237\323\236a0\317[\302\243\331\273\306\315\265\013\274\302\340D\225W\034\004*\216j\252\360\370m\244\245h\014\324\345\032\310\327\037\273\351\255\376b\307u\331Hn\255\254\0146\177\077\077bf\300B\363\227?\222I%\321\353\355\341\207\243\300\310*r\366\367\216\233J\252\200k\002\221qc\312\317\2003\240\376\227\273S\355\211\"\271\232)\204W\256\026Y\236\376 %qu\335\313t\253m)T\020l\376\324\213\256\306\250\023N\230Ua\233IEUPw\273a\367\362\367k\\\033\006\210[\227\300\251\372z|\026k]\333\232\\4\342\341-\304\017\331>\342f\3432\252HO\267\014\316\033\362\016\254\277\225Jy*&\274ne&\3654F\210\337\rO\026+!\226\021\340\347x5\n\\a\305xUO'[\202\2613,\254W\251\245R\262\004#\306>\232\336D%C\312XJk\021\031\"45\256\300\223+Y*t5.s\337\t\217\030v\027\271\231\032\2656\336\351\033\233E8[\037\265\243/k8rM\250Z\341\344=o\355\360\257Z\363\337\350\316\253\022C;^\307w\234\nm\2128\237F\000>nr\227\256%a\244t\014LlWb\177\261\020q!\007\305dQ\344\227\266LFv\273\077\077X\277LS\007L:%\317+\263Y\335\345cw]\251f:#\262P\255\333Z\342\250\227\205\361\353\3534?\233\177\242\374\246\211\247\210\332e\273a\344\347(\301\234\341\226\206a\310\252\330g]\242Yd\344\331hzZ\205\014\345`\357V\331\302\240zT\336\266\347m\273\372t\361\032*t,\354\216\232\020\346\320\002t\203\212\272;O\021\316r\n\245\024Bd\222\313i\177d\345\221\240m\351\222\306\324\212s#Zd\014\361\212Kr\261rJ\275g\006I\213\\\235V)]T\014\351\026cWK{\340X\303b""\327\252\026.f\245dlW]\222(\364\306a!\246\251\024l\331\037\0049\204\026\256\357\026\323\320mQ+\025n:\214\004\367\220\347\334n\347\345\342\301.n\207\316\266FB!\253S\272bSUOV\030\346\330\205X\310\032\306U\346\244\223\232\014&\226\207$p\304\317YdLi\014|\214hm\342\241q\330aT8\360\357\212\310\251\331[\330\221\235\346\014(S\\\236\355\325\327\330SF\317;Y\n\312\014M\241\273\335Fx\367\324a\240\317\237\313y\245\222\334\301 a\356\232-7\356\016\016\026\220p\361\036\340]Q\217@F\0178\034\241\315\023\261\211\276\261\024>\002Q]\303U\311\330m\204\252\333\273\325\313\236\316\362\034%\016l\352\0357Q\2363t\270\214\301\014\020\223\262x+\300\313[\203\205\306\263\240}\021{BL\n#DQ\0270\320\326\004;\351\005\245PPR\t\262\265\233x\266\025\336\002\336\333\025\335\200\303\014\303\002\350\316\270\375.\024\212\272\352q'TfL\370\344Lw!P3\003\"Uf\013\033[\312\007\211\030h\276\361\213~E\301\227iu\322\324\334\005\360\337\252\207\302\324\221x;*\247\231\026S\022\225\016\355X\006\025o\277\2272Dy%.\236\236\227t(Im\261,q\302\003\223\311\255\033\276\004QE\243~T]\302`\215\362(?6,d\347\r\253\237\361\2029\024e,x\031\030\223\355[\016\032P\244M\362rCvo\031\035\275\007\315\334fH\332\210\262\305?\r\316F\217\270oRa+\252\315kG\242\336\227r\367\312\351\220\301-\r\n\253\234\212yb\350\320\255`\203\260\275\023u\025]\033\263RT\310\223B\202\213\350N;'\315d8\366y\267Q\245\306.\355j^\354\005\322\3210!\"2{\334\002N\347P\373D8\306\030\306\021\014\tl\233:\325\264\333F0\273<\337e\213\340\310\201n%d\353\256\206P\246\025\324[\036\250Q\270oJ\311Q:\035P\016\252\226nt,\227qq\234\003\364\361\256\371\253\313y\310\223\263\363D\204x\030\312\360\323\270\306\335\365\033\261\304\307\014\3636\261gPSGI\311.\033\004\206\327\303\325\331\275P\312\326I\303\230\300\333\203\315\276zw\240\333Oe\307\035\373\023\322!\312!\333JXF\202\335J\362KdV6Y\254\244\326a\302\023\321[=f{\311U\035\355g}\330\342\250L\371\326\243\304\235\024\016oh/r<\014LF@b\003\016\233@\321\217\320\202 46'Y\244\370\036\3061c\214\036\315\030Pc\023\017\360B\010`\330w\006,\234^XX\010""\370Y\375pi\254\275cq\267\030)\354f\250\271\037F\330s\250\214P\331)\334Stb*\007\007\200\336T\2363`\252\0147Ub\014\347\032zT\354\264$V\322p\273\261\006\252\331\r\267Y1wXb\313\037\266\254\025W\270\213\343\306\304\305\365G\"z\306&\363\327\271\310\305C\03439}\r\370\027\032j\306=\022\255\332\337o!\305\003\034\rY\247+\t\210\340Y\242\370\023\326\331\357yo\363\257B\257{\201\367#\033\367Gn\354\243\227\341r~\013\360D\341lCM8_\315\006\343\n\026\240c\336\304r\263^\341\000\256=\244V\t\032\363\371\341\242\213G6\234B\276\232\303\0323^N\214g'{7w8\274\222\n\252\220\350\303\013#\003y\326\241I\247B\234U\np\265\254\321\200\377%\240\241\337\246\257c\257\\\014=\224\277S\323.\010lB\300.\263\245\224\372\\4\\\321\214\205\366!`\352l\323\203$\261M\314\247\304\264\207\335\367R\205\010\021s`JM\362QS2\016\2409\303\2230^\334\305\2319\026l\373-\037<\343\346\320\326\256\214p\311\317\024\376&[\235^w\343.i\224mKI\367\202\244\353i\022\177\277\215rL\266\032CI\250\255\t\221\272\344\221V\276\312\354\345\n\231)\257\257'\215F)\240\232\006A\346\222\222\327C\"A\347\275\013Jd\377y\322eJ%##\3100W[\273+\222I\267\215\240c\222\212:\261\321\225O\024\241\315\253Z\316?\220\317\347\324\017n\030\300\330\303J\206\016v\022\364{\316\216\325.l\nW\333\364;\346Y\237\255ipc\326\332\220\324>C\305\237S\261_\3547\273\013B\325\251\335>\266A\345@\340/\240\023\356\233\225{u\223\201\341)\324\212\374\240\375/\310|\266~\271S~8:\257\302\306\314\017\275\317\2576\222n\035S\251T\327\312%\361%\273\3575\210p!'\364\320\277\302~mW_\257KS\257N\300=\013\271\274A\031;\274]\t\235\214\247+*1\020\026r\256\017\016\026\251gp\245\252\2514r\320\324\222\264\261e$=-5\263\244\251PR\345\2514w**T\265\235\310\"\254\235M\237\353z \262{\336\261v*\021Ib\267+jDZR\233\004\274\322)H\253\312E\355\020\312\253\272\3254BC\t\tq\323|7\255\331\031\340\363\240f\324BC\177\342\356H\247\n\022\037\370\360\202@";
    PyObject *data = __Pyx_DecompressString(cstring, 2796, 2);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) != 0 /* compression: zlib (2712 bytes) */
const char* const cstring = "x\332\235V\313_\333V\026\306\2064\3200\301\006\363\010d\202\234R\036)!uJ\323\204t\232q\t!\220\204\304@\2236\371\245\372\t\351\032+\310\222\255\007\340\244\323d\311R\313\273\324\362.\265\364\322K/\265\364R\177B\376\2049\347\312`\033\332L;\013\351\\\335\307y|\347;\347jv\303\260\211`\027$[X\256\330\005C\027TKP\210\246\356\020S\262\211V\021,\333Te\233\230\270I\027\236\255<\273\276x{Q\220tE0\311\033\"\333\226`9;\262&Y\026\261\004#/\3548\252f\253\272`WJ\304Z\020\326\362B\305p\004\235\020E\260\r\241\004\373\332\017\330\005\242\013\026\261q \314H\272n\330\222\255\032\272\010\307U}wFPT\023\214\250\373\004O?\2204\213,<\223L\242\333\302\036\251\010E\325*J\266\\\230\027\310\241j\201\331]A7\024\"l\031Eb\027\360\363\3004\360\255\332\205%an~\341\236\244(\"\030!;&!o+\013\221\270!\356\351\306\201.\356\232R\251 \226*\207\013\360\010rA\325\224%@\243\004\207\025\325\222v4Bt|\013y\325\264\354\245]Y\020v\225\274\261\264[0,\233[\206\340\n\206\243)\360a\013;D(9V\001\"7t\360\036C\264lI\336[\022\204\202d\001\026\030\007\007M\265\"\305\n\006\245Y\266h\230\242]\322\204\242cq5\222\240Ax\202\001YpJ\032Y\340\352\371\013\220\216\354\202\266\245H\243X\224J\307n\360\263\n$\020\255\274\003\355\307{`h\375\007R\270\277\004\360\023}i\326\006$\204\274)\355\026ayn\025\221X\256\310\032Y1M\303|\204\360\3609\2615\334\000\273\247>\027D\321$\212#\023Q\346l\022\305\263\033 \333\200\202\335\332\322\332q\366x\307\332'O\346U]\0211\033g&mI\325\254\263\263j\251cR\325U[\2254\365-\0219\240mk\021g\024\3226\265+\213\226a\332\3553\304\0269c8\264\247\026\332@o[)\020Ii\377.\022s\227\234\326k\033%\203\317\211Opy\013F\034\370\216\257?\300\375\364\372\031\360Z;\210\331>\376\224.\276zF\323\306O\217\037\213\233+\317\327\266\326\236n\210\342\263\312!<\367\201t\342\0069\2647I\376\030@\251T\"\272\"Y\025]V\215\005\3310\r\007j\226X\315b<]\2052t\031UAK<#\034\335\026\304\321\010WD\221\267\023p\2649\000\310U\233\024\361\223H\272SBa\312\032\230\2024\213\266)\311d\007\352P64\r\233\213\241[\n);\240\007+E\214\204b\210\262$\027\010A\376[""\307\355\245=\223-\312\265\361\254E.Q\314;\272,\212\273r\223,\235\014\301/\220\247\330!\242\357\021\272\"\247\007\177\341\322\031\202\212\252%\236@\210\341Z\260\013u\264\241\010h\330b\004EQ\202\320\341]\331!\221\346h\210\252Z\274+\302\036\036\220X4\024G#\010\251.\025#I\016\360uh\363\035x\016\237\233M\347q\256-\216V'j\216po\263\341!\007\000\312\246\020#R\034\177\225\014\350oy\033\304\311\310*\231$\257\036F\357&\264\300\016\300\222\310{\226S\214\276Lb9\232\035\215\233\000\342\020\257\241h\344\350%U\336\323\332[\326\251\205\323\275\355\324jg\375\375\331\"\024\023_\332\267\261\241#peG\322\"\020M\"iM\3066KL<[l\307\023\344\020\207%\r\310\032\305\006\315\232\000\023A\250\026r\226hy^\214\315\014\301H\201R\343 \234\224\250\370\007\345\212\211\266\270\247|\236\323\026(R\302\007\223\307%z\311\007\035\344\264\t\360\tS\300\367\251\3216\013S\302\333\224a*\304<iXN\tk\327\261\310\211\013\373\222\346\020\353C\254\321s\215\245\331-?\371!\026\366$\\\370\276pt\307\315\342\302<\313\262\347~\346C\366\343\225\256\276\021\367W/\343\335g}~\262\321{\341\350\256+\323d\243\177(\030\232\362\262\215\376\201#\333\275C\227i9L\244\202\324\264'y\373,\007c\367\2117\342\301\344\214Wn\300\327So\021\006\251Yv\236\225\375\013\325\314\307\317\272\222\223T\361\322^\246\221\000]\263\254\233e\302\324p\230\032\247\327\274\254\227\013'\204@\270\345Ka\377\220{\217o\353\037\247i\272H+\354\034\3135\372\023A\342*L\337b\340\324\210+5z\373\217\236\302j\031=zG\247\275\036/\033&&\350m<8p\364\033\275I_\261X#q\205\356\341\211\376\243\207\356c/\356\245?\016v\365\r\271_\321\347\020\"Xm\364\216\323\031\0176\\\242\243^\254\321{\361HjD\0164\372\307h\017}\340]\305\240\023\356E\364\035<w\227\340\350M/\007\361\014O\322\202'5R\350\026B't\365\365\037-\006\003\202\027\003\215\375\203\356\034\315\320l\310u\302k\217&\351<\300\345\000\326/\375r5\206[fi\032\343\234E{C\356b0\372\025\313\260l{\264\240\370ZW\337\260\273FsT\302\320u\360{\233M\302yH\316\367\256E\371\321\3334\023B,i\014u\325\005\014.\321+^\231\365\2605?\027A\364\005\3155""\022#n\201J\301\344m\264\016\273\203\211y\266\342\217U\001\200a\314z0\365-l?Nz\330?\201\010\003.<\353\220\270Q\267\202\301y\351\020\320\351\303A\264Y\nFgY\234qO\300\217A\027\346\341\025\366\016\271w\251Dm\357\016\373\221I\200Y\337\300\t\030\3018\306\272\342\217DXt\362k\325\343\326\016\370\341L4{\213\016R\230\345>\241\252a Y\306[c\233\254\214*\337\320nX\006\007~\000\277\346X\006\023\370y\224>\356}\230H\"*I:\207\n\023n/\215\035\353\035\242\331\006\256\036\333\376\226\365\004\363\337W\263\325\355\332x\375f=\207\204^\243\257\202Y\300\255\232\252\3460\3477\374\244?\345\277\251\305\032)\301\373\234]\205\254\245F\335\267\336\240w\317O\373\231\306X\232\3033\206\025R\301\n\010\307\276\364r\236\212\316b\361\265\321\342\023H,\272PM\377\204\324\357B\316\223x\262\017\"\340Y^s_y\003P\\\261\216\230\243\270 )\213G\345\343\nz\317\3660\303\003Ge \323\305\256\276A\3672h\314\203\r\230\333\00751\230\377\274\253/\331\351\026\235\242*P\350\034v\006?{\266\0038\336r0\363\035\370\334S]\257\311\365$\242\364\020\221\214\374\315\035[\274\212\0257\345j\021Y\356\201Kw\274e\317d)\266\345\307\374\224\377Ku\247\306\213a:\030\233e1\230\317\205@Gx\245\376 \303\220\371\177\244\376$\307\2347!\267\257\261)&!\034}t\302{\n\001\334\007h\023c\364\"\244\303fw}\tk\371[\372\031\2265\326\362\004\235\016\204\257\271?\220\356a\367g\372\2237\315)\335\334'\007\351\014\244\374K\037\2717\234\376{)\307#\343t\314K\206ML\006O0\371{Qv\364\244O\205\266D7!A\340\340%4\373?\"\n\377.\205O\342\211\210\2238*\3775\342\274\347\365|\227\311p\265\36045I\210\220\244P\217\355~\307)\333\304\250s\342J4\001\035%<\276\002\316\024\3014\303\252\016\006\320\365)/\317\226\231\005\261,\372\373\325- Y\002\352?\204W\212n\363\320\232\207\2563~\2158\3562\226\333\004\264\355u\010\031.\263\313\364\241\327$4\\#ao\322\345w\311\272\233\247Yh\247\355\307\021S\336\200\306h/v\314\221`\004q{\347\363\233d\000\360\350\363\007\001T\340\3715\274>\261\275\375\336l\010`g\335\333\301\353\n\356\000h\261\335\036^\215\221\335\202\247\2604Tmr\"\030\307\373{\221\275""\255&\233\r\t\357P\236\264`t\216%\031O\225\341\347|\265Z\016a\250\373Y(.)LM\322Cpd\272\032\257\246\033\2511z\001\032\340\330\004\275\007\332\200-\220\313K_z\353l\007@\204\331%o\323{\357K~\271q\031\223-5.\243\325;\376#\320y\031\255\314!3\276\360n\003/y\346\340\367\341:\313\261\274\017\177\022\037{\273\316]<\332v'\350}\257\027.\237U\216<v\321\305z\271\321\303\257&\367\001^\203\215\236\363\037\366\217^\270+\220\213\\#\2721F\335r\010/\033|x\001\355p\010\210}\276zP#\365o\352J\260\371\"x\361k\360\253\034\310\371 \377&xc\005\326\333\340\355;\324\003\344\203\214\177\355\367@\274/\252\217j\007u H\307L\271\345\332$\374)\254\261m \340_v\347\0076\352\237\363\267\240\317=\254\255\3263-\223_A6\312\376y\377]m\246\236\014;g\322\235&7\374\251\340\326\203\232R\237\017\236\277\014^\252\201Z\014\212\277\177\354\352z\037[\216\203X\216\257\241X\213?A\361$\376\n\305\253\270\210B\214;(\234x\005E%\276\332\rb\265\373\021\212G\335\257Q\274\356\226PH\335;\335\377O\\+\265\321\372\271\372V\360l;\330\226\002\251\020\024\300\273RP\262\321ll\035\315\256\307s(r\361\347(\236\307\177A\361K\3745\212\327q\035\205~\354\347~\274\205\321\rfC\027X\255\336\255\355\326sa\347\214t\202Q\220\234\366\224\340\332\277\252\333A\026\340\201D\227\203\262\0238\373\1779\232`\342\006;\360I\365\233\252\022\374\010* \014 J\204\363{\360\352\337\361Utn5\376\030\305\343\370S\024O\343[(\266\342/Q\274\214K(\244\370N\233\377\327\221+\376Lu\002S\037v\316\360\177g\274\262FX\031\177\252'i\031/\221\013\020\332^\220\333l\364\340\237\344,\224\017\347\314o\330\312\032\275\320s\203\233\367k\033\301\346\026\376\201g\376\344\340\017\336\030/\351\326A\270\235\202\314\217\265\205\340Y\356S\007\227\350\317\274n[\007\361\357\376uu\241\036\373\324\261\357\241G\224\241\005\265\216-0\211\275\257\352\365,?\366_8\260\023X";
    PyObject *data = __Pyx_DecompressString(cstring, 2712, 1);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #else /* compression: none (5300 bytes) */
const char* const bytes = "(Note that Cython is deliberately stricter than PEP-484 and rejects subclasses of builtin types. If you need to pass subclasses then set the 'annotation_typing' directive to False.Parent key mismatch, existing node Something wrong with: ),.?add_notebreezy.breezy/_known_graph_pyx.pyx child: depth:disableenable first:gc  gdfo:ghost nodes should not be pushed onto the stack:  has parents of isenabledkey lst_or_tpl must be a list or tuple. not  not in nodes par:parent_map should be a dict of {key:parent_keys} rev: seen:(tree fragment)GraphCycleErrorKnownGraph_KnownGraphNode_KnownGraphNode.__reduce_cython___KnownGraphNode.__setstate_cython__KnownGraph.__reduce_cython__KnownGraph.__setstate_cython__KnownGraph._find_gdfoKnownGraph._find_tailsKnownGraph._find_tipsKnownGraph._initialize_nodesKnownGraph.add_nodeKnownGraph.gc_sortKnownGraph.get_child_keysKnownGraph.get_parent_keysKnownGraph.headsKnownGraph.merge_sortKnownGraph.topo_sort_MergeSortNode_MergeSortNode.__reduce_cython___MergeSortNode.__setstate_cython___MergeSorter_MergeSorter.__reduce_cython___MergeSorter.__setstate_cython__NULL_REVISION__Pyx_PyDict_NextRefadd_nodeappendasyncio.coroutinesbreezy._known_graph_pyxcandidate_nodeschildchild_keyschild_node__class____class_getitem__cleanupclearcline_in_tracebackcollectionsdeque__dict___dictdo_cacheerrorsexisting_parent_keys_find_gdfo_find_tails_find_tips__func__gcgc_sortget_child_keysget_keyget_parent_keys__getstate__headsheads_key_initialize_nodes_is_coroutineitemskeykeysknown_graphlast_item__main__maybe_headsmaybe_nodemerge_sortmin_gdfo__module____name____new__next_gdfonodenode2parent_gdfoparent_keysparent_mapparent_nodeparentspendingpending_appendpending_popleftpoppopleftposprefixprefix_tips__pyx_checksum__pyx_result__pyx_state__pyx_type__pyx_unpickle_KnownGraph__pyx_unpickle__KnownGraphNode__pyx_unpickle__MergeSortNode__pyx_unpickle__MergeSorter__pyx_vtable____qualname__real_node__reduce____reduce_cython____reduce_ex__replaceresultreverserevisionself__set_name__setde""fault__setstate____setstate_cython__sortsorterstatetailstemptemp_keytemp_nodetemp_parent_keys__test__tip_keytip_nodestipstopo_ordertopo_sortupdateuse_setstatevalues\200\001\330\004*\250!\2506\260\021\200\001\340\004\020\220\001\330\004\013\2109\220A\200\001\330\004,\250A\250V\2601\200A\360\036\000\t\026\220^\2401\240D\250\t\260\021\330\010\013\210;\220c\230\021\330\014\023\320\023$\240A\330\014\017\210t\2209\230C\230q\340\020\024\320\024&\240a\240v\250Q\340\020\024\220M\240\026\240q\340\020'\240q\330\020\024\220O\2404\240q\330\024(\250\007\250q\260\013\2701\360\006\000\021\037\230d\240!\2401\330\020\023\320\023(\250\003\2501\340\024\025\340\024\032\230*\240A\240Q\340\033 \320 6\260a\340\014\023\220?\240!\2401\330\014\032\230!\2304\230y\250\005\250Q\330\014\020\320\020\"\240!\2406\250\021\330\010\026\220a\330\010\014\210O\2304\230q\330\014\017\210{\230&\240\004\240A\340\020\033\2308\2401\330\014\017\210|\2302\230[\250\001\330\020\036\230k\250\021\330\010\014\210H\220L\240\002\240!\360\022\000\t\023\220+\230V\2401\240A\240Q\330\010\032\230'\240\021\330\010\031\230\027\240\001\330\010\016\210a\330\014\023\220?\240!\330\014\030\230\004\230F\240\"\240A\330\014\020\220\016\230d\240!\330\020\023\220:\230V\2402\240Q\360\006\000\025\037\230h\240a\330\024\"\240!\2401\200A\360 \000\t\014\2104\320\017 \240\001\240\021\330\014\022\220)\2301\230A\340\010\016\210a\330\010\016\210k\230\021\230,\240a\240u\250A\250Z\260q\270\001\330\014\022\220(\230!\330\014\032\230(\240!\330\014\023\2204\320\027+\2501\250A\330\014\020\320\020\"\240!\2406\250\021\200A\360*\000\t\025\220I\230Q\230a\330\010\026\220n\240A\240T\250\037\270\001\330\010\013\210<\220s\230!\330\014\023\2208\2301\340\010\032\230!\330\010\014\210G\2201\330\014\031\230\036\240q\250\004\250I\260Q\330\014\017\210{\230#\230Q\330\020\026\220h\230a\320\0378\270\001\330\014\032\230!\320\033,\250E\260\030\270\021\330\010\025\220^\2401\320$5\260Q\330\010\013\210;\220c\230\021\340\014\033\2304\230q\240\001\330\014\017\210t\2201\330\020""\027\220y\240\001\240\021\240!\340\014\030\230\t\240\021\240!\330\010\013\210;\220a\320\027(\250\002\250!\330\014\023\2201\340\010\022\220!\330\010\022\220!\340\010\023\220;\230a\230t\2409\250B\250a\360\006\000\t\017\210a\330\010\016\210k\230\021\320\032+\2501\250E\260\026\260q\270\001\330\014\023\320\023$\240A\330\014\017\210t\2209\230G\2401\330\020\027\220w\230a\230t\2401\330\014\017\210t\2206\230\022\2301\330\020\033\2304\230q\360\006\000\t\025\220O\2401\240I\250R\250q\330\010\016\210j\230\003\2301\330\014\023\220>\240\021\240)\2501\330\014\030\230\n\240\"\240A\330\014\017\210t\2201\340\020\021\330\014\031\230\021\230)\2401\330\014\020\220\010\230\001\330\014\017\210t\2206\230\023\230A\330\020\021\330\014\017\210t\2209\230G\2405\250\004\320,<\270A\270T\300\032\3102\310Q\330\020\024\220I\230[\320(8\270\001\270\024\270Q\330\024\"\240/\260\021\260$\260j\300\001\330\024 \240\n\250\"\250A\330\024\027\220z\240\022\240?\260!\2601\330\030!\240\021\240!\330\030&\240a\240y\260\013\2701\340\030%\240Q\240i\250q\330\010\020\220\001\330\010\016\210a\330\010\016\210k\230\021\320\032+\2501\250E\260\026\260q\270\001\330\014\023\320\023$\240A\330\014\017\210t\2204\220q\330\020\035\230Q\230g\240T\250\021\330\010\020\220\t\230\021\230!\330\010\014\210I\220[\240\017\250q\260\001\330\014\023\220>\240\021\240)\2501\330\014\020\220\010\230\001\330\010\013\2104\210q\330\014\032\230!\2304\230\177\250k\270\021\330\010\017\210q\200A\360\016\000\t\022\220\034\230Q\230f\240A\330\010\017\210v\220[\240\001\200A\360\n\000\t\021\220\001\330\010\016\210a\330\010\016\210k\230\021\230$\230i\240q\250\005\250V\2601\260A\330\014\023\320\023$\240A\330\014\017\210t\2209\230C\230u\240C\320'7\260q\270\004\270J\300c\310\021\330\020\024\220H\230A\330\020\035\230Q\230g\240Q\330\010\017\210q\200A\360\"\000\t\023\220$\220l\240!\330\010\013\210?\230!\2309\240C\240r\250\024\250S\260\001\260\024\260Y\270b\300\001\330\014\022\220&\320\030(\250\001\250\024\250Q\340\010\025\220Q\340\010\024\220O\2401\240I\250R\250q""\330\010\016\210j\230\003\2301\360\006\000\r\024\220>\240\021\240)\2501\330\014\030\230\n\240\"\240A\330\014\017\210t\2209\230G\2401\340\020\035\230Q\230l\250$\250a\330\014\020\220\t\230\033\240O\2601\260D\270\001\330\020\030\230\016\240a\240t\250;\260a\330\020\023\2205\230\006\230d\240!\360\006\000\025\033\230&\320 0\260\001\260\024\260Q\330\020\025\220X\230U\240&\250\002\250!\330\020\023\2205\230\006\230c\320!1\260\021\260%\260q\360\006\000\025!\240\n\250\"\250A\330\024\027\220z\240\022\240?\260!\2601\330\030!\240\021\240!\330\030&\240a\240y\260\013\2701\340\030%\240Q\240i\250q\360\006\000\025\032\230\030\240\021\340\010\017\210q\200A\360\022\000\t\023\220$\220l\240!\340\010\024\220O\2401\240I\250R\250q\330\010\016\210j\230\003\2301\360\006\000\r\024\220>\240\021\240)\2501\330\014\030\230\n\240\"\240A\330\014\030\230\004\230F\240\"\240A\330\014\020\220\t\230\033\240O\2601\260D\270\001\330\020\030\230\016\240a\240t\250;\260a\330\020\023\220:\230R\230u\240A\330\024\031\230\030\240\021\330\020\025\220X\230U\240&\250\002\250!\330\020\023\2205\230\006\230c\320!1\260\021\260%\260q\340\024 \240\n\250\"\250A\330\024\027\220z\240\022\240?\260!\2601\330\030!\240\021\240!\330\030&\240a\240y\260\013\2701\340\030%\240Q\240i\250q\360\006\000\025\032\230\030\240\021\200A\360\n\000\t\020\210q\330\010\016\210a\330\010\016\210k\230\021\230$\230i\240q\250\005\250V\2601\260A\330\014\023\320\023$\240A\330\014\017\210\177\230a\230t\240;\250c\260\021\330\020\035\230Q\230f\240A\330\010\017\210q\200A\360\024\000\t\020\210t\2207\230!\2304\230q\200A\360\022\000\t\020\210t\2207\230!\2304\230q\200A\360\036\000\t\020\210t\220;\230a\340\010\026\220a\330\010\014\210I\220[\240\017\250q\260\001\330\014\023\220>\240\021\240&\250\001\330\014\017\320\017!\240\021\240$\240f\250C\250s\260!\2604\260v\270S\300\001\330\020\031\230\021\340\020\031\230\024\230T\240\021\240!\330\014\023\220>\240\021\240-\250q\330\014\017\210u\220C\220q\330\020\033\2301\230J\240a\240q\340\020\034\230H\240A\330\020\035\230Q""\230k\250\021\340\010\021\220\021\330\010\014\210J\220f\230A\230Q\330\014\023\220>\240\021\240-\250q\330\014\023\2205\230\003\2301\330\014\030\230\010\240\001\330\014\026\320\026&\240a\240{\260!\330\014\030\230\017\240q\250\t\260\022\2601\330\014\022\220*\230C\230q\330\020\027\220~\240Q\240i\250q\330\020\034\230J\240b\250\001\330\020\023\2204\220y\240\003\2401\340\024\025\330\020\035\230Q\230h\240d\250!\360\016\000\021\033\320\032*\250!\2504\250z\270\021\330\020\024\220I\230[\250\003\2501\250A\330\024\027\320\027)\250\021\250!\330\030&\240o\260Q\260i\270q\340\030&\240n\260A\260Y\270a\340\024\037\230x\240{\260&\270\002\270!\330\024\030\230\013\2401\330\030\033\230?\250!\250;\260a\360\006\000\031%\240J\250b\260\001\330\030\033\230:\240R\240\177\260a\260q\330\034%\240Q\240a\330\034*\250!\2509\260K\270q\340\034)\250\021\250)\2601\330\030#\2408\2501\330\010\017\210q\200\001\330\004-\250Q\250f\260A\200\001\360\010\000\005\016\210T\220\033\230D\240\010\250\004\250G\2604\260v\270T\300\032\3104\310q\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220t\230:\240W\250E\260\023\260D\270\007\270w\300e\3103\310d\320RW\320W^\320^c\320cf\320fj\320js\320sz\320z{\330\004\007\200q\330\010\017\320\0170\260\004\260A\260W\270K\300w\310a\340\010\017\320\0170\260\004\260A\260W\270K\300q\200\001\360\010\000\005\016\210T\220\037\240\004\240I\250T\260\021\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220t\230>\250\027\260\005\260S\270\004\270H\300G\3101\330\004\007\200q\330\010\017\320\017+\2504\250q\260\007\260{\300'\310\021\340\010\017\320\017+\2504\250q\260\007\260{\300!\200\001\360\010\000\005\016\210T\220\037\240\004\240N\260$\3206F\300d\310,\320VZ\320Zi\320im\320m~\360\000\000\177\001C\002\360\000\000C\002I\002\360\000\000I\002M\002\360\000\000M\002[\002\360\000\000[\002_\002\360\000\000_\002u\002\360\000\000u\002y\002\360\000\000y\002G""\003\360\000\000G\003K\003\360\000\000K\003]\003\360\000\000]\003a\003\360\000\000a\003b\003\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220t\230>\250\027\260\005\260S\270\004\270E\300\027\310\005\310S\320PT\320Ta\320ah\320hm\320mp\320pt\360\000\000u\001J\002\360\000\000J\002Q\002\360\000\000Q\002V\002\360\000\000V\002Y\002\360\000\000Y\002]\002\360\000\000]\002n\002\360\000\000n\002u\002\360\000\000u\002v\002\330\004\007\200q\330\010\017\320\017/\250t\2601\260G\270;\300g\310Q\340\010\017\320\017/\250t\2601\260G\270;\300a\200\001\360\010\000\005\016\210T\320\021&\240d\320*=\270T\320AZ\320Z^\320^q\320qu\320uv\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220t\320\033/\250w\260e\2703\270d\320BZ\320Za\320af\320fi\320im\320m\177\360\000\000@\002G\002\360\000\000G\002L\002\360\000\000L\002O\002\360\000\000O\002S\002\360\000\000S\002Z\002\360\000\000Z\002a\002\360\000\000a\002b\002\330\004\007\200q\330\010\017\320\017-\250T\260\021\260'\270\033\300G\3101\340\010\017\320\017-\250T\260\021\260'\270\033\300A\200\001\330\004(\250\001\250\026\250q\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\220?\240(\250!\2501\330\004\007\200|\2207\230!\330\0101\260\021\3202D\300N\320RS\330\004\013\2101\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\220>\240\030\250\021\250!\330\004\007\200|\2207\230!\330\0100\260\001\3201B\300.\320PQ\330\004\013\2101\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\220:\230X\240Q\240a\330\004\007\200|\2207\230!\330\010,\250A\250]\270.\310\001\330\004\013\2101\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\220<\230x\240q\250\001\330\004\007\200|\2207\230!\330\010.\250a\250\177\270n\310A\330\004\013\2101";
    PyObject *data = NULL;
    CYTHON_UNUSED_VAR(__Pyx_DecompressString);
    #endif
    PyObject **stringtab = __pyx_mstate->__pyx_string_tab;
    Py_ssize_t pos = 0;
    for (int i = 0; i < 161; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyUnicode_DecodeUTF8(bytes + pos, bytes_length, NULL);
      if (likely(string) && i >= 31) PyUnicode_InternInPlace(&string);
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
      stringtab[i] = string;
      pos += bytes_length;
    }
    for (int i = 161; i < 185; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyBytes_FromStringAndSize(bytes + pos, bytes_length);
      stringtab[i] = string;
      pos += bytes_length;
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    Py_XDECREF(data);
    for (Py_ssize_t i = 0; i < 185; i++) {
      if (unlikely(PyObject_Hash(stringtab[i]) == -1)) {
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    #if CYTHON_IMMORTAL_CONSTANTS
    {
      PyObject **table = stringtab + 161;
      for (Py_ssize_t i=0; i<24; ++i) {
        #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
        #if PY_VERSION_HEX < 0x030E0000
        if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
        #else
        if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
        #endif
        {
          Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
        }
        #else
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
        #endif
      }
    }
    #endif
  }
  {
    PyObject **numbertab = __pyx_mstate->__pyx_number_tab + 0;
    int8_t const cint_constants_1[] = {0,1};
    int32_t const cint_constants_4[] = {25525219L,50889907L,238330808L,249680612L};
    for (int i = 0; i < 6; i++) {
      numbertab[i] = PyLong_FromLong((i < 2 ? cint_constants_1[i - 0] : cint_constants_4[i - 2]));
      if (unlikely(!numbertab[i])) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_number_tab;
    for (Py_ssize_t i=0; i<6; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */
typedef struct {
    unsigned int argcount : 2;
    unsigned int num_posonly_args : 1;
    unsigned int num_kwonly_args : 1;
    unsigned int nlocals : 5;
    unsigned int flags : 10;
    unsigned int first_line : 10;
} __Pyx_PyCode_New_function_description;
/* NewCodeObj.proto */
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
);


static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  PyObject* tuple_dedup_map = PyDict_New();
  if (unlikely(!tuple_dedup_map)) return -1;
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[0] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_T_D_G4vT_4q_G1F_a_vWE_Q_q_t_WE, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[0])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[1] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_QfA, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[1])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 120};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_node, __pyx_mstate->__pyx_n_u_real_node};
    __pyx_mstate_global->__pyx_codeobj_tab[2] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy__known_graph_pyx_pyx, __pyx_mstate->__pyx_n_u_get_key, __pyx_mstate->__pyx_kp_b_iso88591_9A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[2])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 10, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 240};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_parent_map, __pyx_mstate->__pyx_n_u_temp_key, __pyx_mstate->__pyx_n_u_temp_parent_keys, __pyx_mstate->__pyx_n_u_temp_node, __pyx_mstate->__pyx_n_u_pos, __pyx_mstate->__pyx_n_u_node, __pyx_mstate->__pyx_n_u_parent_node, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_parent_keys};
    __pyx_mstate_global->__pyx_codeobj_tab[3] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy__known_graph_pyx_pyx, __pyx_mstate->__pyx_n_u_initialize_nodes, __pyx_mstate->__pyx_kp_b_iso88591_A_4_1A_a_k_auAZq_4_1A_6, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[3])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 266};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_temp_node, __pyx_mstate->__pyx_n_u_node, __pyx_mstate->__pyx_n_u_pos, __pyx_mstate->__pyx_n_u_tails};
    __pyx_mstate_global->__pyx_codeobj_tab[4] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy__known_graph_pyx_pyx, __pyx_mstate->__pyx_n_u_find_tails, __pyx_mstate->__pyx_kp_b_iso88591_A_a_k_iq_V1A_A_t9CuC_7q_Jc_HA_Qg, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[4])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 280};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_temp_node, __pyx_mstate->__pyx_n_u_node, __pyx_mstate->__pyx_n_u_pos, __pyx_mstate->__pyx_n_u_tips};
    __pyx_mstate_global->__pyx_codeobj_tab[5] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy__known_graph_pyx_pyx, __pyx_mstate->__pyx_n_u_find_tips, __pyx_mstate->__pyx_kp_b_iso88591_A_q_a_k_iq_V1A_A_at_c_QfA_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[5])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 9, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 293};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_node, __pyx_mstate->__pyx_n_u_child_2, __pyx_mstate->__pyx_n_u_temp, __pyx_mstate->__pyx_n_u_pos, __pyx_mstate->__pyx_n_u_replace, __pyx_mstate->__pyx_n_u_last_item, __pyx_mstate->__pyx_n_u_next_gdfo, __pyx_mstate->__pyx_n_u_pending};
    __pyx_mstate_global->__pyx_codeobj_tab[6] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy__known_graph_pyx_pyx, __pyx_mstate->__pyx_n_u_find_gdfo, __pyx_mstate->__pyx_kp_b_iso88591_A_l_O1IRq_j_1_1_A_F_A_O1D_at_a_R, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[6])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 13, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 328};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_parent_keys, __pyx_mstate->__pyx_n_u_maybe_node, __pyx_mstate->__pyx_n_u_node, __pyx_mstate->__pyx_n_u_parent_node, __pyx_mstate->__pyx_n_u_child_node, __pyx_mstate->__pyx_n_u_parent_gdfo, __pyx_mstate->__pyx_n_u_next_gdfo, __pyx_mstate->__pyx_n_u_existing_parent_keys, __pyx_mstate->__pyx_n_u_pending, __pyx_mstate->__pyx_n_u_pending_popleft, __pyx_mstate->__pyx_n_u_pending_append};
    __pyx_mstate_global->__pyx_codeobj_tab[7] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy__known_graph_pyx_pyx, __pyx_mstate->__pyx_n_u_add_node, __pyx_mstate->__pyx_kp_b_iso88591_A_1D_c_A_t9Cq_avQ_M_q_q_O4q_q_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[7])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 16, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 398};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_keys, __pyx_mstate->__pyx_n_u_maybe_node, __pyx_mstate->__pyx_n_u_maybe_heads, __pyx_mstate->__pyx_n_u_temp_node, __pyx_mstate->__pyx_n_u_node, __pyx_mstate->__pyx_n_u_pos, __pyx_mstate->__pyx_n_u_last_item, __pyx_mstate->__pyx_n_u_min_gdfo, __pyx_mstate->__pyx_n_u_heads_key, __pyx_mstate->__pyx_n_u_candidate_nodes, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_cleanup, __pyx_mstate->__pyx_n_u_pending, __pyx_mstate->__pyx_n_u_parent_node, __pyx_mstate->__pyx_n_u_heads};
    __pyx_mstate_global->__pyx_codeobj_tab[8] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy__known_graph_pyx_pyx, __pyx_mstate->__pyx_n_u_heads, __pyx_mstate->__pyx_kp_b_iso88591_A_IQa_nAT_s_81_G1_q_IQ_Q_ha_8_E, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[8])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 9, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 490};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_node, __pyx_mstate->__pyx_n_u_child_2, __pyx_mstate->__pyx_n_u_temp, __pyx_mstate->__pyx_n_u_pos, __pyx_mstate->__pyx_n_u_replace, __pyx_mstate->__pyx_n_u_last_item, __pyx_mstate->__pyx_n_u_pending, __pyx_mstate->__pyx_n_u_topo_order};
    __pyx_mstate_global->__pyx_codeobj_tab[9] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy__known_graph_pyx_pyx, __pyx_mstate->__pyx_n_u_topo_sort, __pyx_mstate->__pyx_kp_b_iso88591_A_l_9Cr_S_Yb_Q_Q_O1IRq_j_1_1_A_t, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[9])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 14, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 544};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_temp, __pyx_mstate->__pyx_n_u_pos, __pyx_mstate->__pyx_n_u_last_item, __pyx_mstate->__pyx_n_u_node, __pyx_mstate->__pyx_n_u_node2, __pyx_mstate->__pyx_n_u_parent_node, __pyx_mstate->__pyx_n_u_tips, __pyx_mstate->__pyx_n_u_prefix_tips, __pyx_mstate->__pyx_n_u_prefix, __pyx_mstate->__pyx_n_u_tip_nodes, __pyx_mstate->__pyx_n_u_result, __pyx_mstate->__pyx_n_u_pending, __pyx_mstate->__pyx_n_u_parents};
    __pyx_mstate_global->__pyx_codeobj_tab[10] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy__known_graph_pyx_pyx, __pyx_mstate->__pyx_n_u_gc_sort, __pyx_mstate->__pyx_kp_b_iso88591_A_t_a_a_I_q_fCs_4vS_T_q_uCq_1Jaq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[10])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 616};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_tip_key, __pyx_mstate->__pyx_n_u_sorter};
    __pyx_mstate_global->__pyx_codeobj_tab[11] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy__known_graph_pyx_pyx, __pyx_mstate->__pyx_n_u_merge_sort, __pyx_mstate->__pyx_kp_b_iso88591_A_QfA_v, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[11])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 626};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key};
    __pyx_mstate_global->__pyx_codeobj_tab[12] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy__known_graph_pyx_pyx, __pyx_mstate->__pyx_n_u_get_parent_keys, __pyx_mstate->__pyx_kp_b_iso88591_A_t7_4q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[12])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 638};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key};
    __pyx_mstate_global->__pyx_codeobj_tab[13] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy__known_graph_pyx_pyx, __pyx_mstate->__pyx_n_u_get_child_keys, __pyx_mstate->__pyx_kp_b_iso88591_A_t7_4q_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[13])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[14] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_T_IT_G1F_a_vWE_Q_q_t_S_HG1_q_4q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[14])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[15] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[15])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[16] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_T_N_6Fd_VZZiimm_C_C_I_I_M_M, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[16])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[17] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_AV1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[17])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[18] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_T_d_TAZZ_qquuv_G1F_a_vWE_Q_q_t, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[18])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[19] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_6, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[19])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[20] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_pyx_unpickle__KnownGraphNode, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_1_7_1_2DNRS_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[20])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[21] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_pyx_unpickle_KnownGraph, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_XQa_7_A_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[21])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[22] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_pyx_unpickle__MergeSortNode, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_7_0_1B_PQ_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[22])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[23] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_pyx_unpickle__MergeSorter, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_xq_7_a_nA_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[23])) goto bad;
  }
  Py_DECREF(tuple_dedup_map);
  return 0;
  bad:
  Py_DECREF(tuple_dedup_map);
  return -1;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* AssertionsEnabled.init */
  if (likely(__Pyx_init_assertions_enabled() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CommonTypesMetaclass.init */
  if (likely(__pyx_CommonTypesMetaclass_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CachedMethodType.init */
  #if CYTHON_COMPILING_IN_LIMITED_API
  {
      PyObject *typesModule=NULL;
      typesModule = PyImport_ImportModule("types");
      if (typesModule) {
          __pyx_mstate_global->__Pyx_CachedMethodType = PyObject_GetAttrString(typesModule, "MethodType");
          Py_DECREF(typesModule);
      }
  } // error handling follows
  #endif
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CythonFunctionShared.init */
  if (likely(__pyx_CyFunction_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* TupleAndListFromArray (used by fastcall) */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < (0))) {
            Py_DECREF(res);
            return NULL;
        }
        Py_INCREF(src[i]);
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL ||\
        !(CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS)
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length, length2;
        int kind;
        void *data1, *data2;
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        #endif
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length < 0)) return -1;
        #endif
        length2 = __Pyx_PyUnicode_GET_LENGTH(s2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length2 < 0)) return -1;
        #endif
        if (length != length2) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    return (equals == Py_EQ);
return_ne:
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = __Pyx_PyTuple_GET_SIZE(kwnames);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(n == -1)) return NULL;
    #endif
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        if (s == namei) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        int eq = __Pyx_PyUnicode_Equals(s, namei, Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs;
    PyObject *dict;
#if !CYTHON_ASSUME_SAFE_SIZE
    nkwargs = PyTuple_Size(kwnames);
    if (unlikely(nkwargs < 0)) return NULL;
#else
    nkwargs = PyTuple_GET_SIZE(kwnames);
#endif
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
#if !CYTHON_ASSUME_SAFE_MACROS
        PyObject *key = PyTuple_GetItem(kwnames, i);
        if (!key) goto bad;
#else
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
#endif
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* PyObjectCall (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall (used by PyObjectCallOneArg) */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
          #if CYTHON_COMPILING_IN_LIMITED_API
            return PyObject_Vectorcall(func, args, _nargs, NULL);
          #else
            vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
            if (f) {
                return f(func, args, _nargs, NULL);
            }
          #endif
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyObjectCallOneArg (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetAttrStr (used by UnpackUnboundCMethod) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* UnpackUnboundCMethod (used by CallUnboundCMethod0) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
#elif CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject **args, Py_ssize_t nargs, PyObject *kwnames) {
        return _PyObject_Vectorcall
            (method, args ? args+1 : NULL, nargs ? nargs-1 : 0, kwnames);
}
#else
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
    return
#if PY_VERSION_HEX < 0x03090000
    _PyObject_Vectorcall
#else
    PyObject_Vectorcall
#endif
        (method, args ? args+1 : NULL, nargs ? (size_t) nargs-1 : 0, kwnames);
}
#endif
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
     METH_VARARGS | METH_KEYWORDS,
#else
     METH_FASTCALL | METH_KEYWORDS,
#endif
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method, *result=NULL;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    result = method;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            result = unbound_method;
        }
    }
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    if (unlikely(target->method)) {
        Py_DECREF(result);
    } else
#endif
    target->method = result;
    return 0;
}

/* CallUnboundCMethod0 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        if (likely(cfunc->flag == METH_NOARGS))
            return __Pyx_CallCFunction(cfunc, self, NULL);
        if (likely(cfunc->flag == METH_FASTCALL))
            return __Pyx_CallCFunctionFast(cfunc, self, NULL, 0);
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, NULL, 0, NULL);
        if (likely(cfunc->flag == (METH_VARARGS | METH_KEYWORDS)))
            return __Pyx_CallCFunctionWithKeywords(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple, NULL);
        if (cfunc->flag == METH_VARARGS)
            return __Pyx_CallCFunction(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple);
        return __Pyx__CallUnboundCMethod0(cfunc, self);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod0(&tmp_cfunc, self);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod0(cfunc, self);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *result;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
    result = __Pyx_PyObject_CallOneArg(cfunc->method, self);
    return result;
}

/* py_dict_items (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_items, d);
}

/* py_dict_values (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_values, d);
}

/* OwnedDictNext (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue) {
    PyObject *next = NULL;
    if (!*ppos) {
        if (pvalue) {
            PyObject *dictview = pkey ? __Pyx_PyDict_Items(p) : __Pyx_PyDict_Values(p);
            if (unlikely(!dictview)) goto bad;
            *ppos = PyObject_GetIter(dictview);
            Py_DECREF(dictview);
        } else {
            *ppos = PyObject_GetIter(p);
        }
        if (unlikely(!*ppos)) goto bad;
    }
    next = PyIter_Next(*ppos);
    if (!next) {
        if (PyErr_Occurred()) goto bad;
        return 0;
    }
    if (pkey && pvalue) {
        *pkey = __Pyx_PySequence_ITEM(next, 0);
        if (unlikely(*pkey)) goto bad;
        *pvalue = __Pyx_PySequence_ITEM(next, 1);
        if (unlikely(*pvalue)) goto bad;
        Py_DECREF(next);
    } else if (pkey) {
        *pkey = next;
    } else {
        assert(pvalue);
        *pvalue = next;
    }
    return 1;
  bad:
    Py_XDECREF(next);
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
    PyErr_FormatUnraisable("Exception ignored in __Pyx_PyDict_NextRef");
#else
    PyErr_WriteUnraisable(__pyx_mstate_global->__pyx_n_u_Pyx_PyDict_NextRef);
#endif
    if (pkey) *pkey = NULL;
    if (pvalue) *pvalue = NULL;
    return 0;
}
#else // !CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue) {
    int result = PyDict_Next(p, ppos, pkey, pvalue);
    if (likely(result == 1)) {
        if (pkey) Py_INCREF(*pkey);
        if (pvalue) Py_INCREF(*pvalue);
    }
    return result;
}
#endif

/* RaiseDoubleKeywords (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, args, 2);
        }
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, 2, NULL);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod2(&tmp_cfunc, self, arg1, arg2);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *result = NULL;
        PyObject *args = PyTuple_New(2);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
        return result;
    }
#endif
    {
        PyObject *args[4] = {NULL, self, arg1, arg2};
        return __Pyx_PyObject_FastCall(cfunc->method, args+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
    }
}

/* ParseKeywordsImpl (used by ParseKeywords) */
static int __Pyx_ValidateDuplicatePosArgs(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char* function_name)
{
    PyObject ** const *name = argnames;
    while (name != first_kw_arg) {
        PyObject *key = **name;
        int found = PyDict_Contains(kwds, key);
        if (unlikely(found)) {
            if (found == 1) __Pyx_RaiseDoubleKeywordsError(function_name, key);
            goto bad;
        }
        name++;
    }
    return 0;
bad:
    return -1;
}
#if CYTHON_USE_UNICODE_INTERNALS
static CYTHON_INLINE int __Pyx_UnicodeKeywordsEqual(PyObject *s1, PyObject *s2) {
    int kind;
    Py_ssize_t len = PyUnicode_GET_LENGTH(s1);
    if (len != PyUnicode_GET_LENGTH(s2)) return 0;
    kind = PyUnicode_KIND(s1);
    if (kind != PyUnicode_KIND(s2)) return 0;
    const void *data1 = PyUnicode_DATA(s1);
    const void *data2 = PyUnicode_DATA(s2);
    return (memcmp(data1, data2, (size_t) len * (size_t) kind) == 0);
}
#endif
static int __Pyx_MatchKeywordArg_str(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    #if CYTHON_USE_UNICODE_INTERNALS
    Py_hash_t key_hash = ((PyASCIIObject*)key)->hash;
    if (unlikely(key_hash == -1)) {
        key_hash = PyObject_Hash(key);
        if (unlikely(key_hash == -1))
            goto bad;
    }
    #endif
    name = first_kw_arg;
    while (*name) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (key_hash == ((PyASCIIObject*)name_str)->hash && __Pyx_UnicodeKeywordsEqual(name_str, key)) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) {
                *index_found = (size_t) (name - argnames);
                return 1;
            }
        }
        #endif
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (unlikely(key_hash == ((PyASCIIObject*)name_str)->hash)) {
            if (__Pyx_UnicodeKeywordsEqual(name_str, key))
                goto arg_passed_twice;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            if (unlikely(name_str == key)) goto arg_passed_twice;
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) goto arg_passed_twice;
        }
        #endif
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
bad:
    return -1;
}
static int __Pyx_MatchKeywordArg_nostr(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    if (unlikely(!PyUnicode_Check(key))) goto invalid_keyword_type;
    name = first_kw_arg;
    while (*name) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (cmp == 1) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        if (unlikely(cmp == -1)) goto bad;
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (unlikely(cmp != 0)) {
            if (cmp == 1) goto arg_passed_twice;
            else goto bad;
        }
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
bad:
    return -1;
}
static CYTHON_INLINE int __Pyx_MatchKeywordArg(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    return likely(PyUnicode_CheckExact(key)) ?
        __Pyx_MatchKeywordArg_str(key, argnames, first_kw_arg, index_found, function_name) :
        __Pyx_MatchKeywordArg_nostr(key, argnames, first_kw_arg, index_found, function_name);
}
static void __Pyx_RejectUnknownKeyword(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char *function_name)
{
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos = NULL;
    #else
    Py_ssize_t pos = 0;
    #endif
    PyObject *key = NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(kwds);
    while (
        #if CYTHON_AVOID_BORROWED_REFS
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL)
        #else
        PyDict_Next(kwds, &pos, &key, NULL)
        #endif
    ) {
        PyObject** const *name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (!*name) {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp != 1) {
                if (cmp == 0) {
                    PyErr_Format(PyExc_TypeError,
                        "%s() got an unexpected keyword argument '%U'",
                        function_name, key);
                }
                #if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(key);
                #endif
                break;
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        #endif
    }
    __Pyx_END_CRITICAL_SECTION();
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(pos);
    #endif
    assert(PyErr_Occurred());
}
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t extracted = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    name = first_kw_arg;
    while (*name && num_kwargs > extracted) {
        PyObject * key = **name;
        PyObject *value;
        int found = 0;
        #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        found = PyDict_GetItemRef(kwds, key, &value);
        #else
        value = PyDict_GetItemWithError(kwds, key);
        if (value) {
            Py_INCREF(value);
            found = 1;
        } else {
            if (unlikely(PyErr_Occurred())) goto bad;
        }
        #endif
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            extracted++;
        }
        name++;
    }
    if (num_kwargs > extracted) {
        if (ignore_unknown_kwargs) {
            if (unlikely(__Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name) == -1))
                goto bad;
        } else {
            __Pyx_RejectUnknownKeyword(kwds, argnames, first_kw_arg, function_name);
            goto bad;
        }
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t len;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    if (PyDict_Update(kwds2, kwds) < 0) goto bad;
    name = first_kw_arg;
    while (*name) {
        PyObject *key = **name;
        PyObject *value;
#if !CYTHON_COMPILING_IN_LIMITED_API && (PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop))
        int found = PyDict_Pop(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
        }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        int found = PyDict_GetItemRef(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            if (unlikely(PyDict_DelItem(kwds2, key) < 0)) goto bad;
        }
#else
    #if CYTHON_COMPILING_IN_CPYTHON
        value = _PyDict_Pop(kwds2, key, kwds2);
    #else
        value = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, kwds2, key, kwds2);
    #endif
        if (value == kwds2) {
            Py_DECREF(value);
        } else {
            if (unlikely(!value)) goto bad;
            values[name-argnames] = value;
        }
#endif
        name++;
    }
    len = PyDict_Size(kwds2);
    if (len > 0) {
        return __Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name);
    } else if (unlikely(len == -1)) {
        goto bad;
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject *key = NULL;
    PyObject** const * name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    for (Py_ssize_t pos = 0; pos < num_kwargs; pos++) {
#if CYTHON_AVOID_BORROWED_REFS
        key = __Pyx_PySequence_ITEM(kwds, pos);
#else
        key = __Pyx_PyTuple_GET_ITEM(kwds, pos);
#endif
#if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!key)) goto bad;
#endif
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            PyObject *value = kwvalues[pos];
            values[name-argnames] = __Pyx_NewRef(value);
        } else {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp == 1) {
                PyObject *value = kwvalues[pos];
                values[index_found] = __Pyx_NewRef(value);
            } else {
                if (unlikely(cmp == -1)) goto bad;
                if (kwds2) {
                    PyObject *value = kwvalues[pos];
                    if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
                } else if (!ignore_unknown_kwargs) {
                    goto invalid_keyword;
                }
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        key = NULL;
        #endif
    }
    return 0;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    goto bad;
bad:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(key);
    #endif
    return -1;
}

/* ParseKeywords */
static int __Pyx_ParseKeywords(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds)))
        return __Pyx_ParseKeywordsTuple(kwds, kwvalues, argnames, kwds2, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
    else if (kwds2)
        return __Pyx_ParseKeywordDictToDict(kwds, argnames, kwds2, values, num_pos_args, function_name);
    else
        return __Pyx_ParseKeywordDict(kwds, argnames, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* PyObjectFormatAndDecref */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result;
    if (unlikely(!s)) return NULL;
    result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* CIntToDigits (used by CIntToPyUnicode) */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode (used by COrdinalToPyUnicode) */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, const char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp = PySequence_Repeat(padding, uoffset - prepend_sign);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyUnicode_Concat(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyUnicode_Concat(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* COrdinalToPyUnicode (used by CIntToPyUnicode) */
static CYTHON_INLINE int __Pyx_CheckUnicodeValue(int value) {
    return value <= 1114111;
}
static PyObject* __Pyx_PyUnicode_FromOrdinal_Padded(int value, Py_ssize_t ulength, char padding_char) {
    Py_ssize_t padding_length = ulength - 1;
    if (likely((padding_length <= 250) && (value < 0xD800 || value > 0xDFFF))) {
        char chars[256];
        if (value <= 255) {
            memset(chars, padding_char, (size_t) padding_length);
            chars[ulength-1] = (char) value;
            return PyUnicode_DecodeLatin1(chars, ulength, NULL);
        }
        char *cpos = chars + sizeof(chars);
        if (value < 0x800) {
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0xc0 | (value & 0x1f));
        } else if (value < 0x10000) {
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0xe0 | (value & 0x0f));
        } else {
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0xf0 | (value & 0x07));
        }
        cpos -= padding_length;
        memset(cpos, padding_char, (size_t) padding_length);
        return PyUnicode_DecodeUTF8(cpos, chars + sizeof(chars) - cpos, NULL);
    }
    if (value <= 127 && CYTHON_USE_UNICODE_INTERNALS) {
        const char chars[1] = {(char) value};
        return __Pyx_PyUnicode_BuildFromAscii(ulength, chars, 1, 0, padding_char);
    }
    {
        PyObject *uchar, *padding_uchar, *padding, *result;
        padding_uchar = PyUnicode_FromOrdinal(padding_char);
        if (unlikely(!padding_uchar)) return NULL;
        padding = PySequence_Repeat(padding_uchar, padding_length);
        Py_DECREF(padding_uchar);
        if (unlikely(!padding)) return NULL;
        uchar = PyUnicode_FromOrdinal(value);
        if (unlikely(!uchar)) {
            Py_DECREF(padding);
            return NULL;
        }
        result = PyUnicode_Concat(padding, uchar);
        Py_DECREF(padding);
        Py_DECREF(uchar);
        return result;
    }
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_long(long value, Py_ssize_t width, char padding_char) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!(is_unsigned || value == 0 || value > 0) ||
                    !(sizeof(value) <= 2 || value & ~ (long) 0x01fffff || __Pyx_CheckUnicodeValue((int) value)))) {
        PyErr_SetString(PyExc_OverflowError, "%c arg not in range(0x110000)");
        return NULL;
    }
    if (width <= 1) {
        return PyUnicode_FromOrdinal((int) value);
    }
    return __Pyx_PyUnicode_FromOrdinal_Padded((int) value, width, padding_char);
}
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_long(long value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(long)*3+2];
    char *dpos, *end = digits + sizeof(long)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    long remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (long) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (long) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (long) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    if (max_char > 1114111) max_char = 1114111;
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - result_ulength < 0))
        goto overflow;
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = values[i];
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (__Pyx_PyUnicode_READY(uval) == (-1))
            goto bad;
        #endif
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(ulength < 0)) goto bad;
        #endif
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    Py_ssize_t i;
    PyObject *result = NULL;
    PyObject *value_tuple = PyTuple_New(value_count);
    if (unlikely(!value_tuple)) return NULL;
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    for (i=0; i<value_count; i++) {
        if (__Pyx_PyTuple_SET_ITEM(value_tuple, i, values[i]) != (0)) goto bad;
        Py_INCREF(values[i]);
    }
    result = PyUnicode_Join(__pyx_mstate_global->__pyx_empty_unicode, value_tuple);
bad:
    Py_DECREF(value_tuple);
    return result;
#endif
}

/* RejectKeywords */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds) {
    PyObject *key = NULL;
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds))) {
        key = __Pyx_PySequence_ITEM(kwds, 0);
    } else {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *pos = NULL;
#else
        Py_ssize_t pos = 0;
#endif
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
        if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return;
#endif
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL);
#if CYTHON_AVOID_BORROWED_REFS
        Py_XDECREF(pos);
#endif
    }
    if (likely(key)) {
        PyErr_Format(PyExc_TypeError,
            "%s() got an unexpected keyword argument '%U'",
            function_name, key);
        Py_DECREF(key);
    }
}

/* PyErrExceptionMatches (used by GetAttr3) */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore (used by GetAttr3) */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* GetAttr3 */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    int res = PyObject_GetOptionalAttr(o, n, &r);
    return (res != 0) ? r : __Pyx_NewRef(d);
#else
  #if CYTHON_USE_TYPE_SLOTS
    if (likely(PyUnicode_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
  #endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
#endif
}

/* PyObjectGetAttrStrNoError (used by GetBuiltinName) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName (used by GetModuleGlobalName) */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* PyDictVersioning (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* RaiseException */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
#elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}

/* PyObjectVectorCallKwBuilder (used by PyObjectVectorCallMethodKwBuilder) */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    Py_INCREF(key);
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* PyObjectVectorCallMethodKwBuilder */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames) {
    PyObject *result;
    PyObject *obj = PyObject_GetAttr(args[0], name);
    if (unlikely(!obj))
        return NULL;
    result = __Pyx_Object_Vectorcall_CallFromBuilder(obj, args+1, nargsf-1, kwnames);
    Py_DECREF(obj);
    return result;
}
#endif

/* PyObjectFastCallMethod */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf) {
    PyObject *result;
    PyObject *attr = PyObject_GetAttr(args[0], name);
    if (unlikely(!attr))
        return NULL;
    result = __Pyx_PyObject_FastCall(attr, args+1, nargsf - 1);
    Py_DECREF(attr);
    return result;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
    CYTHON_UNUSED_VAR(clineno);
    CYTHON_UNUSED_VAR(lineno);
    CYTHON_UNUSED_VAR(filename);
    CYTHON_MAYBE_UNUSED_VAR(nogil);
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(0);
    }
    ctx = PyUnicode_FromString(name);
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
    if (nogil)
        PyGILState_Release(state);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS)) {
        return __Pyx_PyList_GetItemRefFast(o, wrapped_i, unsafe_shared);
    } else
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyList_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyTuple_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)) {
            return __Pyx_PyList_GetItemRefFast(o, n, unsafe_shared);
        } else if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            return __Pyx_NewRef(PyList_GET_ITEM(o, n));
        }
    } else
    #if !CYTHON_AVOID_BORROWED_REFS
    if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            return __Pyx_NewRef(PyTuple_GET_ITEM(o, n));
        }
    } else
    #endif
#endif
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
    {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (!is_list && mm && mm->mp_subscript) {
            PyObject *r, *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (is_list || likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    (void)wraparound;
    (void)boundscheck;
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
}

/* PyObjectCallNoArg (used by pyfrozenset_new) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
#if CYTHON_COMPILING_IN_LIMITED_API
            || __Pyx_get_runtime_version() >= 0x030A0000
#endif
            )
            return result;
        {
            Py_ssize_t size = __Pyx_PySet_GET_SIZE(result);
            if (likely(size > 0))
                return result;
#if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely(size < 0)) {
                Py_DECREF(result);
                return NULL;
            }
#endif
        }
        Py_DECREF(result);
#endif
    }
    return __Pyx_PyObject_CallNoArg((PyObject*) &PyFrozenSet_Type);
}

/* PyObjectCall2Args (used by CallUnboundCMethod1) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    int was_initialized =  __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return __Pyx_CallCFunction(cfunc, self, arg);
        } else if (flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, &arg, 1);
        } else if (flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, &arg, 1, NULL);
        }
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod1(&tmp_cfunc, self, arg);
    }
#endif
    PyObject* result = __Pyx__CallUnboundCMethod1(cfunc, self, arg);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *args = PyTuple_New(1);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
    } else
#endif
    {
        result = __Pyx_PyObject_Call2Args(cfunc->method, self, arg);
    }
    return result;
}

/* py_dict_pop */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop)
    PyObject *value;
    if (PyDict_Pop(d, key, &value) == 0) {
        if (default_value) {
            Py_INCREF(default_value);
        } else {
            PyErr_SetObject(PyExc_KeyError, key);
        }
        value = default_value;
    }
    return value;
#elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    return _PyDict_Pop(d, key, default_value);
#else
    if (default_value) {
        return __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, d, key, default_value);
    } else {
        return __Pyx_CallUnboundCMethod1(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, d, key);
    }
#endif
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_mstate_global->__pyx_n_u_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!(is_unsigned || value == 0 || value > 0) ||
                    !(sizeof(value) <= 2 || value & ~ (int) 0x01fffff || __Pyx_CheckUnicodeValue((int) value)))) {
        PyErr_SetString(PyExc_OverflowError, "%c arg not in range(0x110000)");
        return NULL;
    }
    if (width <= 1) {
        return PyUnicode_FromOrdinal((int) value);
    }
    return __Pyx_PyUnicode_FromOrdinal_Padded((int) value, width, padding_char);
}
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(int)*3+2];
    char *dpos, *end = digits + sizeof(int)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    int remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (int) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (int) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (int) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* PyLongBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_Fallback___Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, int inplace) {
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject* __Pyx_Unpacked___Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    long a;
    const PY_LONG_LONG llb = intval;
    PY_LONG_LONG lla;
    if (unlikely(__Pyx_PyLong_IsZero(op1))) {
        return __Pyx_NewRef(op2);
    }
    const int is_positive = __Pyx_PyLong_IsPos(op1);
    const digit* digits = __Pyx_PyLong_Digits(op1);
    const Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
    if (likely(size == 1)) {
        a = (long) digits[0];
        if (!is_positive) a *= -1;
    } else {
        switch (size) {
            case 2:
                if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                    a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
            case 3:
                if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                    a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
            case 4:
                if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                    a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
        }
        return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    calculate_long:
        {
            long x;
            x = a + b;
            return PyLong_FromLong(x);
        }
    calculate_long_long:
        {
            PY_LONG_LONG llx;
            llx = lla + llb;
            return PyLong_FromLongLong(llx);
        }
    
}
#endif
static PyObject* __Pyx_Float___Pyx_PyLong_AddObjC(PyObject *float_val, long intval, int zerodivision_check) {
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    double a = __Pyx_PyFloat_AS_DOUBLE(float_val);
        double result;
        
        result = ((double)a) + (double)b;
        return PyFloat_FromDouble(result);
}
static CYTHON_INLINE PyObject* __Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        return __Pyx_Unpacked___Pyx_PyLong_AddObjC(op1, op2, intval, inplace, zerodivision_check);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        return __Pyx_Float___Pyx_PyLong_AddObjC(op1, intval, zerodivision_check);
    }
    return __Pyx_Fallback___Pyx_PyLong_AddObjC(op1, op2, inplace);
}
#endif

/* PyObjectGetMethod (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(tp);
    PyErr_Format(PyExc_AttributeError,
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}
#endif

/* PyObjectCallMethod0 (used by pop) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[1] = {obj};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_CallNoArg;
    return PyObject_VectorcallMethod(method_name, args, 1 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
#endif
}

/* pop */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L) {
    if (__Pyx_IS_TYPE(L, &PySet_Type)) {
        return PySet_Pop(L);
    }
    return __Pyx_PyObject_CallMethod0(L, __pyx_mstate_global->__pyx_n_u_pop);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L) {
    if (likely(PyList_GET_SIZE(L) > (((PyListObject*)L)->allocated >> 1))) {
        __Pyx_SET_SIZE(L, Py_SIZE(L) - 1);
        return PyList_GET_ITEM(L, PyList_GET_SIZE(L));
    }
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyList_Type_pop, L);
}
#endif

/* ArgTypeTestFunc (used by ArgTypeTest) */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    PyObject *extra_info = __pyx_mstate_global->__pyx_empty_unicode;
    int from_annotation_subclass = 0;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (!exact) {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    } else if (exact == 2) {
        if (__Pyx_TypeCheck(obj, type)) {
            from_annotation_subclass = 1;
            extra_info = __pyx_mstate_global->__pyx_kp_u_Note_that_Cython_is_deliberately;
        }
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")"
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        "%s%U"
#endif
        , name, type_name, obj_type_name
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        , (from_annotation_subclass ? ". " : ""), extra_info
#endif
        );
#if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    if (exact == 2 && from_annotation_subclass) {
        PyObject *res;
        PyObject *vargs[2];
        vargs[0] = PyErr_GetRaisedException();
        vargs[1] = extra_info;
        res = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_kp_u_add_note, vargs, 2, NULL);
        Py_XDECREF(res);
        PyErr_SetRaisedException(vargs[0]);
    }
#endif
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* AllocateExtensionType */
static PyObject *__Pyx_AllocateExtensionType(PyTypeObject *t, int is_final) {
    if (is_final || likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
        allocfunc alloc_func = __Pyx_PyType_GetSlot(t, tp_alloc, allocfunc);
        return alloc_func(t, 0);
    } else {
        newfunc tp_new = __Pyx_PyType_TryGetSlot(&PyBaseObject_Type, tp_new, newfunc);
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (!tp_new) {
            PyObject *new_str = PyUnicode_FromString("__new__");
            if (likely(new_str)) {
                PyObject *o = PyObject_CallMethodObjArgs((PyObject *)&PyBaseObject_Type, new_str, t, NULL);
                Py_DECREF(new_str);
                return o;
            } else
                return NULL;
        } else
    #endif
        return tp_new(t, __pyx_mstate_global->__pyx_empty_tuple, 0);
    }
}

/* CallTypeTraverse */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg) {
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
    if (__Pyx_get_runtime_version() < 0x03090000) return 0;
    #endif
    if (!always_call) {
        PyTypeObject *base = __Pyx_PyObject_GetSlot(o, tp_base, PyTypeObject*);
        unsigned long flags = PyType_GetFlags(base);
        if (flags & Py_TPFLAGS_HEAPTYPE) {
            return 0;
        }
    }
    Py_VISIT((PyObject*)Py_TYPE(o));
    return 0;
}
#endif

/* LimitedApiGetTypeDict (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static Py_ssize_t __Pyx_GetTypeDictOffset(void) {
    PyObject *tp_dictoffset_o;
    Py_ssize_t tp_dictoffset;
    tp_dictoffset_o = PyObject_GetAttrString((PyObject*)(&PyType_Type), "__dictoffset__");
    if (unlikely(!tp_dictoffset_o)) return -1;
    tp_dictoffset = PyLong_AsSsize_t(tp_dictoffset_o);
    Py_DECREF(tp_dictoffset_o);
    if (unlikely(tp_dictoffset == 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' doesn't have a dictoffset");
        return -1;
    } else if (unlikely(tp_dictoffset < 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' has an unexpected negative dictoffset. "
            "Please report this as Cython bug");
        return -1;
    }
    return tp_dictoffset;
}
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp) {
    static Py_ssize_t tp_dictoffset = 0;
    if (unlikely(tp_dictoffset == 0)) {
        tp_dictoffset = __Pyx_GetTypeDictOffset();
        if (unlikely(tp_dictoffset == -1 && PyErr_Occurred())) {
            tp_dictoffset = 0; // try again next time?
            return NULL;
        }
    }
    return *(PyObject**)((char*)tp + tp_dictoffset);
}
#endif

/* SetItemOnTypeDict (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_SetItem(tp_dict, k, v);
    if (likely(!result)) {
        PyType_Modified(tp);
        if (unlikely(PyObject_HasAttr(v, __pyx_mstate_global->__pyx_n_u_set_name))) {
            PyObject *setNameResult = PyObject_CallMethodObjArgs(v, __pyx_mstate_global->__pyx_n_u_set_name,  (PyObject *) tp, k, NULL);
            if (!setNameResult) return -1;
            Py_DECREF(setNameResult);
        }
    }
    return result;
}

/* FixUpExtensionType */
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if __PYX_LIMITED_VERSION_HEX > 0x030900B1
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#else
    const PyType_Slot *slot = spec->slots;
    int changed = 0;
#if !CYTHON_COMPILING_IN_LIMITED_API
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
#if !CYTHON_COMPILING_IN_CPYTHON
        const
#endif  // !CYTHON_COMPILING_IN_CPYTHON)
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_vectorcall_offset = memb->offset;
                    changed = 1;
                }
#endif  // CYTHON_METH_FASTCALL
#if !CYTHON_COMPILING_IN_PYPY
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                    Py_DECREF(descr);
                    if (unlikely(set_item_result < 0)) {
                        return -1;
                    }
                    changed = 1;
                }
#endif  // !CYTHON_COMPILING_IN_PYPY
            }
            memb++;
        }
    }
#endif  // !CYTHON_COMPILING_IN_LIMITED_API
#if !CYTHON_COMPILING_IN_PYPY
    slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_getset)
        slot++;
    if (slot && slot->slot == Py_tp_getset) {
        PyGetSetDef *getset = (PyGetSetDef*) slot->pfunc;
        while (getset && getset->name) {
            if (getset->name[0] == '_' && getset->name[1] == '_' && strcmp(getset->name, "__module__") == 0) {
                PyObject *descr = PyDescr_NewGetSet(type, getset);
                if (unlikely(!descr))
                    return -1;
                #if CYTHON_COMPILING_IN_LIMITED_API
                PyObject *pyname = PyUnicode_FromString(getset->name);
                if (unlikely(!pyname)) {
                    Py_DECREF(descr);
                    return -1;
                }
                int set_item_result = __Pyx_SetItemOnTypeDict(type, pyname, descr);
                Py_DECREF(pyname);
                #else
                CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
                int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                #endif
                Py_DECREF(descr);
                if (unlikely(set_item_result < 0)) {
                    return -1;
                }
                changed = 1;
            }
            ++getset;
        }
    }
#else
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#endif  // !CYTHON_COMPILING_IN_PYPY
    if (changed)
        PyType_Modified(type);
#endif  // PY_VERSION_HEX > 0x030900B1
    return 0;
}

/* ValidateBasesTuple (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_SIZE
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (unlikely(n < 0)) return -1;
#endif
    for (i = 1; i < n; i++)
    {
        PyTypeObject *b;
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !CYTHON_USE_TYPE_SLOTS
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
CYTHON_UNUSED static int __Pyx_PyType_HasMultipleInheritance(PyTypeObject *t) {
    while (t) {
        PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
        if (bases) {
            return 1;
        }
        t = __Pyx_PyType_GetSlot(t, tp_base, PyTypeObject*);
    }
    return 0;
}
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !CYTHON_COMPILING_IN_CPYTHON || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    if (!__Pyx_PyType_HasMultipleInheritance(t)) {
        return PyType_Ready(t);
    }
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400) &&\
                !CYTHON_COMPILING_IN_GRAAL
        gc = PyImport_GetModule(__pyx_mstate_global->__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_mstate_global->__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable (used by MergeVTables) */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i=0;
    Py_ssize_t size;
    void** base_vtables;
    __Pyx_TypeName tp_base_name = NULL;
    __Pyx_TypeName base_name = NULL;
    void* unknown = (void*)-1;
    PyObject* bases = __Pyx_PyType_GetSlot(type, tp_bases, PyObject*);
    int base_depth = 0;
    {
        PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        while (base) {
            base_depth += 1;
            base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
        }
    }
    base_vtables = (void**) PyMem_Malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
#if CYTHON_COMPILING_IN_LIMITED_API
    size = PyTuple_Size(bases);
    if (size < 0) goto other_failure;
#else
    size = PyTuple_GET_SIZE(bases);
#endif
    for (i = 1; i < size; i++) {
        PyObject *basei;
        void* base_vtable;
#if CYTHON_AVOID_BORROWED_REFS
        basei = PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#else
        basei = PyTuple_GET_ITEM(bases, i);
#endif
        base_vtable = __Pyx_GetVtable((PyTypeObject*)basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
            }
        }
    }
    PyErr_Clear();
    PyMem_Free(base_vtables);
    return 0;
bad:
    {
        PyTypeObject* basei = NULL;
        PyTypeObject* tp_base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        tp_base_name = __Pyx_PyType_GetFullyQualifiedName(tp_base);
#if CYTHON_AVOID_BORROWED_REFS
        basei = (PyTypeObject*)PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = (PyTypeObject*)PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#else
        basei = (PyTypeObject*)PyTuple_GET_ITEM(bases, i);
#endif
        base_name = __Pyx_PyType_GetFullyQualifiedName(basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
    }
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
really_bad: // bad has failed!
#endif
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
other_failure:
#endif
    PyMem_Free(base_vtables);
    return -1;
}

/* DelItemOnTypeDict (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_DelItem(tp_dict, k);
    if (likely(!result)) PyType_Modified(tp);
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_mstate_global->__pyx_n_u_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_mstate_global->__pyx_n_u_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython);
            if (likely(reduce_cython)) {
                ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_mstate_global->__pyx_n_u_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetFullyQualifiedName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_2_4
#define __PYX_HAVE_RT_ImportType_3_2_4
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size)
{
    PyObject *result = 0;
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API)
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !( defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API) )
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    if (size == 0) {
        return (PyTypeObject *)result;
    }
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_2_4 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_2_4 && (size_t)basicsize > size) {
        if (PyErr_WarnFormat(NULL, 0,
                "%.200s.%.200s size changed, may indicate binary incompatibility. "
                "Expected %zd from C header, got %zd from PyObject",
                module_name, class_name, size, basicsize) < 0) {
            goto bad;
        }
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* HasAttr (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!PyUnicode_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_PyObject_GetAttrStrNoError(o, n);
    if (!r) {
        return (unlikely(PyErr_Occurred())) ? -1 : 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}
#endif

/* ImportImpl (used by Import) */
static int __Pyx__Import_GetModule(PyObject *qualname, PyObject **module) {
    PyObject *imported_module = PyImport_GetModule(qualname);
    if (unlikely(!imported_module)) {
        *module = NULL;
        if (PyErr_Occurred()) {
            return -1;
        }
        return 0;
    }
    *module = imported_module;
    return 1;
}
static int __Pyx__Import_Lookup(PyObject *qualname, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject **module) {
    PyObject *imported_module;
    PyObject *top_level_package_name;
    Py_ssize_t i;
    int status, module_found;
    Py_ssize_t dot_index;
    module_found = __Pyx__Import_GetModule(qualname, &imported_module);
    if (unlikely(!module_found || module_found == -1)) {
        *module = NULL;
        return module_found;
    }
    if (imported_names) {
        for (i = 0; i < len_imported_names; i++) {
            PyObject *imported_name = imported_names[i];
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
            int has_imported_attribute = PyObject_HasAttr(imported_module, imported_name);
#else
            int has_imported_attribute = PyObject_HasAttrWithError(imported_module, imported_name);
            if (unlikely(has_imported_attribute == -1)) goto error;
#endif
            if (!has_imported_attribute) {
                goto not_found;
            }
        }
        *module = imported_module;
        return 1;
    }
    dot_index = PyUnicode_FindChar(qualname, '.', 0, PY_SSIZE_T_MAX, 1);
    if (dot_index == -1) {
        *module = imported_module;
        return 1;
    }
    if (unlikely(dot_index == -2)) goto error;
    top_level_package_name = PyUnicode_Substring(qualname, 0, dot_index);
    if (unlikely(!top_level_package_name)) goto error;
    Py_DECREF(imported_module);
    status = __Pyx__Import_GetModule(top_level_package_name, module);
    Py_DECREF(top_level_package_name);
    return status;
error:
    Py_DECREF(imported_module);
    *module = NULL;
    return -1;
not_found:
    Py_DECREF(imported_module);
    *module = NULL;
    return 0;
}
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *from_list = 0;
    int module_found;
    if (!qualname) {
        qualname = name;
    }
    module_found = __Pyx__Import_Lookup(qualname, imported_names, len_imported_names, &module);
    if (likely(module_found == 1)) {
        return module;
    } else if (unlikely(module_found == -1)) {
        return NULL;
    }
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    if (imported_names) {
#if CYTHON_COMPILING_IN_CPYTHON
        from_list = __Pyx_PyList_FromArray(imported_names, len_imported_names);
        if (unlikely(!from_list))
            goto bad;
#else
        from_list = PyList_New(len_imported_names);
        if (unlikely(!from_list)) goto bad;
        for (Py_ssize_t i=0; i<len_imported_names; ++i) {
            if (PyList_SetItem(from_list, i, __Pyx_NewRef(imported_names[i])) < 0) goto bad;
        }
#endif
    }
    if (level == -1) {
        const char* package_sep = strchr(__Pyx_MODULE_NAME, '.');
        if (package_sep != (0)) {
            module = PyImport_ImportModuleLevelObject(
                name, moddict, empty_dict, from_list, 1);
            if (unlikely(!module)) {
                if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                    goto bad;
                PyErr_Clear();
            }
        }
        level = 0;
    }
    if (!module) {
        module = PyImport_ImportModuleLevelObject(
            name, moddict, empty_dict, from_list, level);
    }
bad:
    Py_XDECREF(from_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level) {
    return __Pyx__Import(name, imported_names, len_imported_names, qualname, __pyx_mstate_global->__pyx_d, level);
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_mstate_global->__pyx_kp_u__5);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400) ||\
                CYTHON_COMPILING_IN_GRAAL
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError, "cannot import name %S", name);
    }
    return value;
}

/* dict_setdefault (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value) {
    PyObject* value;
#if __PYX_LIMITED_VERSION_HEX >= 0x030F0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4)
    PyDict_SetDefaultRef(d, key, default_value, &value);
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    PyObject *args[] = {d, key, default_value};
    value = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_n_u_setdefault, args, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#elif CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethodObjArgs(d, __pyx_mstate_global->__pyx_n_u_setdefault, key, default_value, NULL);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* AddModuleRef (used by FetchSharedCythonModule) */
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  static PyObject *__Pyx_PyImport_AddModuleObjectRef(PyObject *name) {
      PyObject *module_dict = PyImport_GetModuleDict();
      PyObject *m;
      if (PyMapping_GetOptionalItem(module_dict, name, &m) < 0) {
          return NULL;
      }
      if (m != NULL && PyModule_Check(m)) {
          return m;
      }
      Py_XDECREF(m);
      m = PyModule_NewObject(name);
      if (m == NULL)
          return NULL;
      if (PyDict_CheckExact(module_dict)) {
          PyObject *new_m;
          (void)PyDict_SetDefaultRef(module_dict, name, m, &new_m);
          Py_DECREF(m);
          return new_m;
      } else {
           if (PyObject_SetItem(module_dict, name, m) != 0) {
                Py_DECREF(m);
                return NULL;
            }
            return m;
      }
  }
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *py_name = PyUnicode_FromString(name);
      if (!py_name) return NULL;
      PyObject *module = __Pyx_PyImport_AddModuleObjectRef(py_name);
      Py_DECREF(py_name);
      return module;
  }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif

/* FetchSharedCythonModule (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef(__PYX_ABI_MODULE_NAME);
}

/* FetchCommonType (used by CommonTypesMetaclass) */
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject* __Pyx_PyType_FromMetaclass(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *result = __Pyx_PyType_FromModuleAndSpec(module, spec, bases);
    if (result && metaclass) {
        PyObject *old_tp = (PyObject*)Py_TYPE(result);
    Py_INCREF((PyObject*)metaclass);
#if __PYX_LIMITED_VERSION_HEX >= 0x03090000
        Py_SET_TYPE(result, metaclass);
#else
        result->ob_type = metaclass;
#endif
        Py_DECREF(old_tp);
    }
    return result;
}
#else
#define __Pyx_PyType_FromMetaclass(me, mo, s, b) PyType_FromMetaclass(me, mo, s, b)
#endif
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t expected_basicsize) {
    Py_ssize_t basicsize;
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (expected_basicsize == 0) {
        return 0; // size is inherited, nothing useful to check
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
    if (unlikely(!py_basicsize)) return -1;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = NULL;
    if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) return -1;
#else
    basicsize = ((PyTypeObject*) cached_type)->tp_basicsize;
#endif
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module = NULL, *cached_type = NULL, *abi_module_dict, *new_cached_type, *py_object_name;
    int get_item_ref_result;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    py_object_name = PyUnicode_FromString(object_name);
    if (!py_object_name) return NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) goto done;
    abi_module_dict = PyModule_GetDict(abi_module);
    if (!abi_module_dict) goto done;
    get_item_ref_result = __Pyx_PyDict_GetItemRef(abi_module_dict, py_object_name, &cached_type);
    if (get_item_ref_result == 1) {
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else if (unlikely(get_item_ref_result == -1)) {
        goto bad;
    }
    cached_type = __Pyx_PyType_FromMetaclass(
        metaclass,
        CYTHON_USE_MODULE_STATE ? module : abi_module,
        spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    new_cached_type = __Pyx_PyDict_SetDefault(abi_module_dict, py_object_name, cached_type);
    if (unlikely(new_cached_type != cached_type)) {
        if (unlikely(!new_cached_type)) goto bad;
        Py_DECREF(cached_type);
        cached_type = new_cached_type;
        if (__Pyx_VerifyCachedType(
                cached_type,
                object_name,
                spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else {
        Py_DECREF(new_cached_type);
    }
done:
    Py_XDECREF(abi_module);
    Py_DECREF(py_object_name);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CommonTypesMetaclass (used by CythonFunctionShared) */
static PyObject* __pyx_CommonTypesMetaclass_get_module(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED void* context) {
    return PyUnicode_FromString(__PYX_ABI_MODULE_NAME);
}
#if __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject* __pyx_CommonTypesMetaclass_call(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *args, CYTHON_UNUSED PyObject *kwds) {
    PyErr_SetString(PyExc_TypeError, "Cannot instantiate Cython internal types");
    return NULL;
}
static int __pyx_CommonTypesMetaclass_setattr(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *attr, CYTHON_UNUSED PyObject *value) {
    PyErr_SetString(PyExc_TypeError, "Cython internal types are immutable");
    return -1;
}
#endif
static PyGetSetDef __pyx_CommonTypesMetaclass_getset[] = {
    {"__module__", __pyx_CommonTypesMetaclass_get_module, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_CommonTypesMetaclass_slots[] = {
    {Py_tp_getset, (void *)__pyx_CommonTypesMetaclass_getset},
    #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {Py_tp_call, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_new, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_setattro, (void*)__pyx_CommonTypesMetaclass_setattr},
    #endif
    {0, 0}
};
static PyType_Spec __pyx_CommonTypesMetaclass_spec = {
    __PYX_TYPE_MODULE_PREFIX "_common_types_metatype",
    0,
    0,
    Py_TPFLAGS_IMMUTABLETYPE |
    Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT,
    __pyx_CommonTypesMetaclass_slots
};
static int __pyx_CommonTypesMetaclass_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    PyObject *bases = PyTuple_Pack(1, &PyType_Type);
    if (unlikely(!bases)) {
        return -1;
    }
    mstate->__pyx_CommonTypesMetaclassType = __Pyx_FetchCommonTypeFromSpec(NULL, module, &__pyx_CommonTypesMetaclass_spec, bases);
    Py_DECREF(bases);
    if (unlikely(mstate->__pyx_CommonTypesMetaclassType == NULL)) {
        return -1;
    }
    return 0;
}

/* PyMethodNew (used by CythonFunctionShared) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *result;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    #if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    {
        PyObject *args[] = {func, self};
        result = PyObject_Vectorcall(__pyx_mstate_global->__Pyx_CachedMethodType, args, 2, NULL);
    }
    #else
    result = PyObject_CallFunctionObjArgs(__pyx_mstate_global->__Pyx_CachedMethodType, func, self, NULL);
    #endif
    return result;
}
#else
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#endif

/* PyVectorcallFastCallDict (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i;
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos;
    #else
    Py_ssize_t pos;
    #endif
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    #if !CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t nkw = PyDict_Size(kw);
    if (unlikely(nkw == -1)) return NULL;
    #else
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    #endif
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = 0;
    i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (__Pyx_PyDict_NextRef(kw, &pos, &key, &value)) {
        keys_are_strings &=
        #if CYTHON_COMPILING_IN_LIMITED_API
            PyType_GetFlags(Py_TYPE(key));
        #else
            Py_TYPE(key)->tp_flags;
        #endif
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(PyTuple_SetItem(kwnames, i, key) < 0)) goto cleanup;
        #else
        PyTuple_SET_ITEM(kwnames, i, key);
        #endif
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(pos);
    #endif
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    Py_ssize_t kw_size =
        likely(kw == NULL) ?
        0 :
#if !CYTHON_ASSUME_SAFE_SIZE
        PyDict_Size(kw);
#else
        PyDict_GET_SIZE(kw);
#endif
    if (kw_size == 0) {
        return vc(func, args, nargs, NULL);
    }
#if !CYTHON_ASSUME_SAFE_SIZE
    else if (unlikely(kw_size == -1)) {
        return NULL;
    }
#endif
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared (used by CythonFunction) */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunctionNoMethod(PyObject *func, void (*cfunc)(void)) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if ((PyObject*)Py_TYPE(func) == __pyx_mstate_global->__Pyx_CachedMethodType) {
        int result;
        PyObject *newFunc = PyObject_GetAttr(func, __pyx_mstate_global->__pyx_n_u_func);
        if (unlikely(!newFunc)) {
            PyErr_Clear(); // It's only an optimization, so don't throw an error
            return 0;
        }
        result = __Pyx__IsSameCyOrCFunctionNoMethod(newFunc, cfunc);
        Py_DECREF(newFunc);
        return result;
    }
    return __Pyx__IsSameCyOrCFunctionNoMethod(func, cfunc);
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if (PyMethod_Check(func)) {
        func = PyMethod_GET_FUNCTION(func);
    }
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure) {
    PyObject *result;
    CYTHON_UNUSED_VAR(closure);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_doc_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#else
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_name_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    Py_INCREF(op->func_qualname);
    result = op->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
#endif
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_defaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_kwdefaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_annotations_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine_value(__pyx_CyFunctionObject *op) {
    int is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
    if (is_coroutine) {
        PyObject *is_coroutine_value, *module, *fromlist, *marker = __pyx_mstate_global->__pyx_n_u_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_mstate_global->__pyx_n_u_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        is_coroutine_value = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(is_coroutine_value)) {
            return is_coroutine_value;
        }
ignore:
        PyErr_Clear();
    }
    return __Pyx_PyBool_FromLong(is_coroutine);
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    result = __Pyx_CyFunction_get_is_coroutine_value(op);
    if (unlikely(!result))
        return NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    if (op->func_is_coroutine) {
        Py_DECREF(result);
        result = __Pyx_NewRef(op->func_is_coroutine);
    } else {
        op->func_is_coroutine = __Pyx_NewRef(result);
    }
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static void __Pyx_CyFunction_raise_argument_count_error(__pyx_CyFunctionObject *func, const char* message, Py_ssize_t size) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        py_name, message, size);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        name, message, size);
#endif
}
static void __Pyx_CyFunction_raise_type_error(__pyx_CyFunctionObject *func, const char* message) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s",
        py_name, message);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s",
        name, message);
#endif
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {"func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {"func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
#else
    {"func_dict", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
#endif
    {"func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {"__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {"_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    {"__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#endif
#if CYTHON_METH_FASTCALL
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {"__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(args);
    __Pyx_BEGIN_CRITICAL_SECTION(m);
    Py_INCREF(m->func_qualname);
    result = m->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_dict = NULL;
#endif
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func_dict);
#elif PY_VERSION_HEX < 0x030d0000
    _PyObject_ClearManagedDict((PyObject*)m);
#else
    PyObject_ClearManagedDict((PyObject*)m);
#endif
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    Py_CLEAR(m->defaults);
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    {
        int e = __Pyx_call_type_traverse((PyObject*)m, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func_dict);
#else
    {
        int e =
#if PY_VERSION_HEX < 0x030d0000
            _PyObject_VisitManagedDict
#else
            PyObject_VisitManagedDict
#endif
                ((PyObject*)m, visit, arg);
        if (e != 0) return e;
    }
#endif
    __Pyx_VISIT_CONST(m->func_name);
    __Pyx_VISIT_CONST(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    __Pyx_VISIT_CONST(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    Py_VISIT(m->defaults);
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
    PyObject *repr;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    repr = PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
    __Pyx_END_CRITICAL_SECTION();
    return repr;
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void(*)(void))meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes no arguments", size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes exactly one argument", size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    __Pyx_CyFunction_raise_type_error(
        (__pyx_CyFunctionObject*)func, "takes no keyword arguments");
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_SIZE
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(argc < 0)) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            __Pyx_CyFunction_raise_type_error(
                cyfunc, "needs an argument");
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(__Pyx_PyTuple_GET_SIZE(kwnames))) {
        __Pyx_CyFunction_raise_type_error(
            cyfunc, "takes no keyword arguments");
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes no arguments", nargs);
        return NULL;
    }
    return meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes exactly one argument", nargs);
        return NULL;
    }
    return meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    #if PY_VERSION_HEX < 0x030e00A6
    size_t nargs_value = (size_t) nargs;
    #else
    Py_ssize_t nargs_value = nargs;
    #endif
    return ((__Pyx_PyCMethod)(void(*)(void))meth)(self, cls, args, nargs_value, kwnames);
}
#endif
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if CYTHON_METH_FASTCALL
#if defined(Py_TPFLAGS_HAVE_VECTORCALL)
    Py_TPFLAGS_HAVE_VECTORCALL |
#elif defined(_Py_TPFLAGS_HAVE_VECTORCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
#endif // CYTHON_METH_FASTCALL
#if PY_VERSION_HEX >= 0x030C0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_TPFLAGS_MANAGED_DICT |
#endif
    Py_TPFLAGS_IMMUTABLETYPE | Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
static int __pyx_CyFunction_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_CyFunctionType_spec, NULL);
    if (unlikely(mstate->__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func, PyTypeObject *defaults_type) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_CallObject((PyObject*)defaults_type, NULL); // _PyObject_New(defaults_type);
    if (unlikely(!m->defaults))
        return NULL;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_mstate_global->__pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CLineInTraceback (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
#define __Pyx_PyProbablyModule_GetDict(o) __Pyx_XNewRef(PyModule_GetDict(o))
#elif !CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyProbablyModule_GetDict(o) PyObject_GenericGetDict(o, NULL);
#else
PyObject* __Pyx_PyProbablyModule_GetDict(PyObject *o) {
    PyObject **dict_ptr = _PyObject_GetDictPtr(o);
    return dict_ptr ? __Pyx_XNewRef(*dict_ptr) : NULL;
}
#endif
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline = NULL;
    PyObject *ptype, *pvalue, *ptraceback;
    PyObject *cython_runtime_dict;
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
    cython_runtime_dict = __Pyx_PyProbablyModule_GetDict(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, cython_runtime_dict,
            __Pyx_PyDict_SetDefault(cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False))
    }
    if (use_cline == NULL || use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    Py_XDECREF(cython_runtime_dict);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache (used by AddTraceback) */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        if (c_line) {
            py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        } else {
            py_funcname = PyUnicode_FromString(funcname);
        }
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CheckUnpickleChecksum */
static void __Pyx_RaiseUnpickleChecksumError(long checksum, long checksum1, long checksum2, long checksum3, const char *members) {
    PyObject *pickle_module = PyImport_ImportModule("pickle");
    if (unlikely(!pickle_module)) return;
    PyObject *pickle_error = PyObject_GetAttrString(pickle_module, "PickleError");
    Py_DECREF(pickle_module);
    if (unlikely(!pickle_error)) return;
    if (checksum2 == checksum1) {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x) = (%s))",
            checksum, checksum1, members);
    } else if (checksum3 == checksum2) {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x, 0x%x) = (%s))",
            checksum, checksum1, checksum2, members);
    } else {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x, 0x%x, 0x%x) = (%s))",
            checksum, checksum1, checksum2, checksum3, members);
    }
    Py_DECREF(pickle_error);
}
static int __Pyx_CheckUnpickleChecksum(long checksum, long checksum1, long checksum2, long checksum3, const char *members) {
    int found = 0;
    found |= checksum1 == checksum;
    found |= checksum2 == checksum;
    found |= checksum3 == checksum;
    if (likely(found))
        return 0;
    __Pyx_RaiseUnpickleChecksumError(checksum, checksum1, checksum2, checksum3, members);
    return -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* PyObjectCallMethod1 */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* UpdateUnpickledDict */
static int __Pyx__UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index) {
    PyObject *state_dict = __Pyx_PySequence_ITEM(state, index);
    if (unlikely(!state_dict)) {
        return -1;
    }
    int non_empty = PyObject_IsTrue(state_dict);
    if (non_empty == 0) {
        Py_DECREF(state_dict);
        return 0;
    } else if (unlikely(non_empty == -1)) {
        return -1;
    }
    PyObject *dict;
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    dict = PyObject_GetAttrString(obj, "__dict__");
    #else
    dict = PyObject_GenericGetDict(obj, NULL);
    #endif
    if (unlikely(!dict)) {
        Py_DECREF(state_dict);
        return -1;
    }
    int result;
    if (likely(PyDict_CheckExact(dict))) {
        result = PyDict_Update(dict, state_dict);
    } else {
        PyObject *obj_result = __Pyx_PyObject_CallMethod1(dict, __pyx_mstate_global->__pyx_n_u_update, state_dict);
        if (likely(obj_result)) {
            Py_DECREF(obj_result);
            result = 0;
        } else {
            result = -1;
        }
    }
    Py_DECREF(state_dict);
    Py_DECREF(dict);
    return result;
}
static int __Pyx_UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index) {
    Py_ssize_t state_size = __Pyx_PyTuple_GET_SIZE(state);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(state_size == -1)) return -1;
    #endif
    if (state_size <= index) {
        return 0;
    }
    return __Pyx__UpdateUnpickledDict(obj, state, index);
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u__6);
    }
    goto done;
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetRuntimeVersion */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
void __Pyx_init_runtime_version(void) {
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
}
#endif
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    return __Pyx_cached_runtime_version;
#endif
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* NewCodeObj */
#if CYTHON_COMPILING_IN_LIMITED_API
    static PyObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                       PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                       PyObject *fv, PyObject *cell, PyObject* fn,
                                       PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
#elif PY_VERSION_HEX >= 0x030B0000
  static PyCodeObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                         PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                         PyObject *fv, PyObject *cell, PyObject* fn,
                                         PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, __pyx_mstate_global->__pyx_empty_bytes);
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030c00A1
    if (likely(result))
        result->_co_firsttraceable = 0;
    #endif
    return result;
  }
#elif !CYTHON_COMPILING_IN_PYPY
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
) {
    PyObject *code_obj = NULL, *varnames_tuple_dedup = NULL, *code_bytes = NULL;
    Py_ssize_t var_count = (Py_ssize_t) descr.nlocals;
    PyObject *varnames_tuple = PyTuple_New(var_count);
    if (unlikely(!varnames_tuple)) return NULL;
    for (Py_ssize_t i=0; i < var_count; i++) {
        Py_INCREF(varnames[i]);
        if (__Pyx_PyTuple_SET_ITEM(varnames_tuple, i, varnames[i]) != (0)) goto done;
    }
    #if CYTHON_COMPILING_IN_LIMITED_API
    varnames_tuple_dedup = PyDict_GetItem(tuple_dedup_map, varnames_tuple);
    if (!varnames_tuple_dedup) {
        if (unlikely(PyDict_SetItem(tuple_dedup_map, varnames_tuple, varnames_tuple) < 0)) goto done;
        varnames_tuple_dedup = varnames_tuple;
    }
    #else
    varnames_tuple_dedup = PyDict_SetDefault(tuple_dedup_map, varnames_tuple, varnames_tuple);
    if (unlikely(!varnames_tuple_dedup)) goto done;
    #endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(varnames_tuple_dedup);
    #endif
    if (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table != NULL && !CYTHON_COMPILING_IN_GRAAL) {
        Py_ssize_t line_table_length = __Pyx_PyBytes_GET_SIZE(line_table);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(line_table_length == -1)) goto done;
        #endif
        Py_ssize_t code_len = (line_table_length * 2 + 4) & ~3LL;
        code_bytes = PyBytes_FromStringAndSize(NULL, code_len);
        if (unlikely(!code_bytes)) goto done;
        char* c_code_bytes = PyBytes_AsString(code_bytes);
        if (unlikely(!c_code_bytes)) goto done;
        memset(c_code_bytes, 0, (size_t) code_len);
    }
    code_obj = (PyObject*) __Pyx__PyCode_New(
        (int) descr.argcount,
        (int) descr.num_posonly_args,
        (int) descr.num_kwonly_args,
        (int) descr.nlocals,
        0,
        (int) descr.flags,
        code_bytes ? code_bytes : __pyx_mstate_global->__pyx_empty_bytes,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        varnames_tuple_dedup,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        filename,
        funcname,
        (int) descr.first_line,
        (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table) ? line_table : __pyx_mstate_global->__pyx_empty_bytes
    );
done:
    Py_XDECREF(code_bytes);
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(varnames_tuple_dedup);
    #endif
    Py_DECREF(varnames_tuple);
    return code_obj;
}

/* DecompressString */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo) {
    PyObject *module = NULL, *decompress, *compressed_bytes, *decompressed;
    const char* module_name = algo == 3 ? "compression.zstd" : algo == 2 ? "bz2" : "zlib";
    PyObject *methodname = PyUnicode_FromString("decompress");
    if (unlikely(!methodname)) return NULL;
    #if __PYX_LIMITED_VERSION_HEX >= 0x030e0000
    if (algo == 3) {
        PyObject *fromlist = Py_BuildValue("[O]", methodname);
        if (unlikely(!fromlist)) goto bad;
        module = PyImport_ImportModuleLevel("compression.zstd", NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
    } else
    #endif
        module = PyImport_ImportModule(module_name);
    if (unlikely(!module)) goto import_failed;
    decompress = PyObject_GetAttr(module, methodname);
    if (unlikely(!decompress)) goto import_failed;
    {
        #ifdef __cplusplus
            char *memview_bytes = const_cast<char*>(s);
        #else
            #if defined(__clang__)
              #pragma clang diagnostic push
              #pragma clang diagnostic ignored "-Wcast-qual"
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic push
              #pragma GCC diagnostic ignored "-Wcast-qual"
            #endif
            char *memview_bytes = (char*) s;
            #if defined(__clang__)
              #pragma clang diagnostic pop
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic pop
            #endif
        #endif
        #if CYTHON_COMPILING_IN_LIMITED_API && !defined(PyBUF_READ)
        int memview_flags = 0x100;
        #else
        int memview_flags = PyBUF_READ;
        #endif
        compressed_bytes = PyMemoryView_FromMemory(memview_bytes, length, memview_flags);
    }
    if (unlikely(!compressed_bytes)) {
        Py_DECREF(decompress);
        goto bad;
    }
    decompressed = PyObject_CallFunctionObjArgs(decompress, compressed_bytes, NULL);
    Py_DECREF(compressed_bytes);
    Py_DECREF(decompress);
    Py_DECREF(module);
    Py_DECREF(methodname);
    return decompressed;
import_failed:
    PyErr_Format(PyExc_ImportError,
        "Failed to import '%.20s.decompress' - cannot initialise module strings. "
        "String compression was configured with the C macro 'CYTHON_COMPRESS_STRINGS=%d'.",
        module_name, algo);
bad:
    Py_XDECREF(module);
    Py_DECREF(methodname);
    return NULL;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return __Pyx_NewRef(b ? Py_True: Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
