/* Generated by Cython 3.2.4 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "breezy/python-compat.h"
        ],
        "include_dirs": [
            "breezy"
        ],
        "name": "breezy.bzr._btree_serializer_pyx",
        "sources": [
            "breezy/bzr/_btree_serializer_pyx.pyx"
        ]
    },
    "module_name": "breezy.bzr._btree_serializer_pyx"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API)
  #if !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef _MSC_VER
  #pragma message ("Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.")
  #else
  #warning Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_2_4"
#define CYTHON_HEX_VERSION 0x030204F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #ifndef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 1
  #endif
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
  #if defined(CYTHON_IMMORTAL_CONSTANTS) && PY_VERSION_HEX < 0x030C0000
    #undef CYTHON_IMMORTAL_CONSTANTS
    #define CYTHON_IMMORTAL_CONSTANTS 0  // definitely won't work
  #elif !defined(CYTHON_IMMORTAL_CONSTANTS)
    #define CYTHON_IMMORTAL_CONSTANTS (PY_VERSION_HEX >= 0x030C0000 && !CYTHON_USE_MODULE_STATE && CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_COMPRESS_STRINGS
  #define CYTHON_COMPRESS_STRINGS 1
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL)
#endif
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#include <stdint.h>
typedef uintptr_t  __pyx_uintptr_t;
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CInitCode */
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef Py_TPFLAGS_IMMUTABLETYPE
  #define Py_TPFLAGS_IMMUTABLETYPE (1UL << 8)
#endif
#ifndef Py_TPFLAGS_DISALLOW_INSTANTIATION
  #define Py_TPFLAGS_DISALLOW_INSTANTIATION (1UL << 7)
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if PY_VERSION_HEX < 0x03090000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) GraalPyFrame_SetLineNumber((frame), (lineno))
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && PyUnstable_Object_IsUniquelyReferenced(obj)) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#elif CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
enum __Pyx_ReferenceSharing {
  __Pyx_ReferenceSharing_DefinitelyUnique, // We created it so we know it's unshared - no need to check
  __Pyx_ReferenceSharing_OwnStrongReference,
  __Pyx_ReferenceSharing_FunctionArgument,
  __Pyx_ReferenceSharing_SharedReference, // Never trust it to be unshared because it's a global or similar
};
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && PY_VERSION_HEX >= 0x030E0000
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing)\
    (sharing == __Pyx_ReferenceSharing_DefinitelyUnique ? 1 :\
      (sharing == __Pyx_ReferenceSharing_FunctionArgument ? PyUnstable_Object_IsUniqueReferencedTemporary(o) :\
      (sharing == __Pyx_ReferenceSharing_OwnStrongReference ? PyUnstable_Object_IsUniquelyReferenced(o) : 0)))
#elif (CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING) || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)sharing), Py_REFCNT(o) == 1)
#else
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)o), ((void)sharing), 0)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !CYTHON_COMPILING_IN_LIMITED_API && CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) (__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared) ?\
    __Pyx_NewRef(PyList_GET_ITEM(o, i)) : __Pyx_PyList_GetItemRef(o, i))
#else
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) __Pyx_PyList_GetItemRef(o, i)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__breezy__bzr___btree_serializer_pyx
#define __PYX_HAVE_API__breezy__bzr___btree_serializer_pyx
/* Early includes */
#include "python-compat.h"
#include <string.h>
#include <stdio.h>

    #if __PYX_LIMITED_VERSION_HEX < 0x030d0000
    static CYTHON_INLINE PyObject *
    __Pyx_CAPI_PyList_GetItemRef(PyObject *list, Py_ssize_t index)
    {
        PyObject *item = PyList_GetItem(list, index);
        Py_XINCREF(item);
        return item;
    }
    #else
    #define __Pyx_CAPI_PyList_GetItemRef PyList_GetItemRef
    #endif

    #if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030d0000
    static CYTHON_INLINE int
    __Pyx_CAPI_PyList_Extend(PyObject *list, PyObject *iterable)
    {
        return PyList_SetSlice(list, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, iterable);
    }

    static CYTHON_INLINE int
    __Pyx_CAPI_PyList_Clear(PyObject *list)
    {
        return PyList_SetSlice(list, 0, PY_SSIZE_T_MAX, NULL);
    }
    #else
    #define __Pyx_CAPI_PyList_Extend PyList_Extend
    #define __Pyx_CAPI_PyList_Clear PyList_Clear
    #endif
    
#include <stdlib.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifdef CYTHON_FREETHREADING_COMPATIBLE
#if CYTHON_FREETHREADING_COMPATIBLE
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "breezy/bzr/_btree_serializer_pyx.pyx",
  "breezy/bzr/_str_helpers.pxd",
  "(tree fragment)",
  "cpython/type.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* Atomics.proto (used by UnpackUnboundCMethod) */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_nonatomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_nonatomic_ptr_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #define __pyx_nonatomic_ptr_type void*
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_atomic_ptr_type old = _InterlockedCompareExchangePointer(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif

/* CriticalSectionsDefinition.proto (used by CriticalSections) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection void*
#define __Pyx_PyCriticalSection2 void*
#define __Pyx_PyCriticalSection_End(cs)
#define __Pyx_PyCriticalSection2_End(cs)
#else
#define __Pyx_PyCriticalSection PyCriticalSection
#define __Pyx_PyCriticalSection2 PyCriticalSection2
#define __Pyx_PyCriticalSection_End PyCriticalSection_End
#define __Pyx_PyCriticalSection2_End PyCriticalSection2_End
#endif

/* CriticalSections.proto (used by ParseKeywordsImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection_Begin(cs, arg) (void)(cs)
#define __Pyx_PyCriticalSection2_Begin(cs, arg1, arg2) (void)(cs)
#else
#define __Pyx_PyCriticalSection_Begin PyCriticalSection_Begin
#define __Pyx_PyCriticalSection2_Begin PyCriticalSection2_Begin
#endif
#if PY_VERSION_HEX < 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_BEGIN_CRITICAL_SECTION(o) {
#define __Pyx_END_CRITICAL_SECTION() }
#else
#define __Pyx_BEGIN_CRITICAL_SECTION Py_BEGIN_CRITICAL_SECTION
#define __Pyx_END_CRITICAL_SECTION Py_END_CRITICAL_SECTION
#endif

/* IncludeStructmemberH.proto (used by FixUpExtensionType) */
#include <structmember.h>

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser;
struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode;
struct __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record;
typedef struct __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record;

/* "breezy/bzr/_btree_serializer_pyx.pyx":257
 * #       One slightly ugly option would be to cache block offsets in a global.
 * #       However, that leads to thread-safety issues, etc.
 * ctypedef struct gc_chk_sha1_record:             # <<<<<<<<<<<<<<
 *     unsigned long long block_offset
 *     unsigned int block_length
*/
struct __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record {
  unsigned PY_LONG_LONG block_offset;
  unsigned int block_length;
  unsigned int record_start;
  unsigned int record_end;
  char sha1[20];
};

/* "breezy/bzr/_btree_serializer_pyx.pyx":44
 * 
 * 
 * cdef class BTreeLeafParser:             # <<<<<<<<<<<<<<
 *     """Parse the leaf nodes of a BTree index.
 * 
*/
struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser {
  PyObject_HEAD
  struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_vtab;
  PyObject *data;
  int key_length;
  int ref_list_length;
  PyObject *keys;
  char *_cur_str;
  char *_end_str;
  char *_start;
  int _header_found;
};


/* "breezy/bzr/_btree_serializer_pyx.pyx":434
 * 
 * 
 * cdef class GCCHKSHA1LeafNode:             # <<<<<<<<<<<<<<
 *     """Track all the entries for a given leaf node."""
 * 
*/
struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_vtab;
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *records;
  PyObject *last_key;
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *last_record;
  int num_records;
  unsigned char common_shift;
  unsigned char offsets[257];
};



/* "breezy/bzr/_btree_serializer_pyx.pyx":44
 * 
 * 
 * cdef class BTreeLeafParser:             # <<<<<<<<<<<<<<
 *     """Parse the leaf nodes of a BTree index.
 * 
*/

struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser {
  PyObject *(*extract_key)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *, char *);
  int (*process_line)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *);
};
static struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_vtabptr_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser;


/* "breezy/bzr/_btree_serializer_pyx.pyx":434
 * 
 * 
 * cdef class GCCHKSHA1LeafNode:             # <<<<<<<<<<<<<<
 *     """Track all the entries for a given leaf node."""
 * 
*/

struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode {
  PyObject *(*_record_to_value_and_refs)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *);
  PyObject *(*_record_to_item)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *);
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *(*_lookup_record)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *);
  int (*_count_records)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *, char *);
  PyObject *(*_parse_bytes)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, PyObject *);
  char *(*_parse_one_entry)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *, char *, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *);
  int (*_offset_for_sha1)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *);
  PyObject *(*_compute_common)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *);
};
static struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_vtabptr_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyAssertionError_Check.proto */
#define __Pyx_PyExc_AssertionError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_AssertionError)

/* PyObjectCall.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* PyThreadStateGet.proto (used by PyErrFetchRestore) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto (used by RaiseException) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.export */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* TupleAndListFromArray.proto (used by fastcall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto (used by BytesEquals) */
#include <string.h>

/* BytesEquals.proto (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_PySequence_ITEM(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_NewRef(__Pyx_PyTuple_GET_ITEM(args, i))
#else
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_XNewRef(PyTuple_GetItem(args, i))
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_ArgRef_FASTCALL(args, i) __Pyx_NewRef(args[i])
    #define __Pyx_NumKwargs_FASTCALL(kwds) __Pyx_PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
#else
    #define __Pyx_ArgRef_FASTCALL __Pyx_ArgRef_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#if CYTHON_METH_FASTCALL || (CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(args + start, stop - start)
#else
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* py_dict_items.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* CallCFunction.proto (used by CallUnboundCMethod0) */
#define __Pyx_CallCFunction(cfunc, self, args)\
    ((PyCFunction)(void(*)(void))(cfunc)->func)(self, args)
#define __Pyx_CallCFunctionWithKeywords(cfunc, self, args, kwargs)\
    ((PyCFunctionWithKeywords)(void(*)(void))(cfunc)->func)(self, args, kwargs)
#define __Pyx_CallCFunctionFast(cfunc, self, args, nargs)\
    ((__Pyx_PyCFunctionFast)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs)
#define __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, nargs, kwnames)\
    ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs, kwnames)

/* PyObjectCallOneArg.proto (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectGetAttrStr.proto (used by UnpackUnboundCMethod) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* UnpackUnboundCMethod.proto (used by CallUnboundCMethod0) */
typedef struct {
    PyObject *type;
    PyObject **method_name;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && CYTHON_ATOMICS
    __pyx_atomic_int_type initialized;
#endif
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
static CYTHON_INLINE int __Pyx_CachedCFunction_GetAndSetInitializing(__Pyx_CachedCFunction *cfunc) {
#if !CYTHON_ATOMICS
    return 1;
#else
    __pyx_nonatomic_int_type expected = 0;
    if (__pyx_atomic_int_cmp_exchange(&cfunc->initialized, &expected, 1)) {
        return 0;
    }
    return expected;
#endif
}
static CYTHON_INLINE void __Pyx_CachedCFunction_SetFinishedInitializing(__Pyx_CachedCFunction *cfunc) {
#if CYTHON_ATOMICS
    __pyx_atomic_store(&cfunc->initialized, 2);
#endif
}
#else
#define __Pyx_CachedCFunction_GetAndSetInitializing(cfunc) 2
#define __Pyx_CachedCFunction_SetFinishedInitializing(cfunc)
#endif

/* CallUnboundCMethod0.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* py_dict_values.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d);

/* OwnedDictNext.proto (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue);
#else
CYTHON_INLINE
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue);
#endif

/* RaiseDoubleKeywords.proto (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywordsImpl.export */
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name
);
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* CallUnboundCMethod2.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* ParseKeywords.proto */
static CYTHON_INLINE int __Pyx_ParseKeywords(
    PyObject *kwds, PyObject *const *kwvalues, PyObject ** const argnames[],
    PyObject *kwds2, PyObject *values[],
    Py_ssize_t num_pos_args, Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RejectKeywords.export */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds);

/* PyErrExceptionMatches.proto (used by GetAttr3) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyObjectGetAttrStrNoError.proto (used by GetBuiltinName) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto (used by GetModuleGlobalName) */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyDictVersioning.proto (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __Pyx_XNewRef(__pyx_dict_cached_value);\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* PyObjectFastCallMethod.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyObject_FastCallMethod(name, args, nargsf) PyObject_VectorcallMethod(name, args, nargsf, NULL)
#else
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf);
#endif

/* PyValueError_Check.proto */
#define __Pyx_PyExc_ValueError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_ValueError)

/* DivInt[long].proto */
static CYTHON_INLINE long __Pyx_div_long(long, long, int b_is_constant);

/* PyKeyError_Check.proto */
#define __Pyx_PyExc_KeyError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_KeyError)

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* JoinPyUnicode.export */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* PyTypeError_Check.proto */
#define __Pyx_PyExc_TypeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_TypeError)

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* AssertionsEnabled.proto */
#if CYTHON_COMPILING_IN_LIMITED_API  ||  PY_VERSION_HEX >= 0x030C0000
  static int __pyx_assertions_enabled_flag;
  #define __pyx_assertions_enabled() (__pyx_assertions_enabled_flag)
  #if __clang__ || __GNUC__
  __attribute__((no_sanitize("thread")))
  #endif
  static int __Pyx_init_assertions_enabled(void) {
    PyObject *builtins, *debug, *debug_str;
    int flag;
    builtins = PyEval_GetBuiltins();
    if (!builtins) goto bad;
    debug_str = PyUnicode_FromStringAndSize("__debug__", 9);
    if (!debug_str) goto bad;
    debug = PyObject_GetItem(builtins, debug_str);
    Py_DECREF(debug_str);
    if (!debug) goto bad;
    flag = PyObject_IsTrue(debug);
    Py_DECREF(debug);
    if (flag == -1) goto bad;
    __pyx_assertions_enabled_flag = flag;
    return 0;
  bad:
    __pyx_assertions_enabled_flag = 1;
    return -1;
  }
#else
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (!Py_OptimizeFlag)
#endif

/* PyLongCompare.proto */
static CYTHON_INLINE int __Pyx_PyLong_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyObjectCall2Args.proto (used by PyObjectCallMethod1) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectGetMethod.proto (used by PyObjectCallMethod1) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);
#endif

/* PyObjectCallMethod1.proto (used by StringJoin) */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* StringJoin.proto */
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck, unsafe_shared) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck, int unsafe_shared);

/* RaiseUnboundLocalError.proto */
static void __Pyx_RaiseUnboundLocalError(const char *varname);

/* ArgTypeTestFunc.export */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))

/* AllocateExtensionType.proto */
static PyObject *__Pyx_AllocateExtensionType(PyTypeObject *t, int is_final);

/* CallTypeTraverse.proto */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#define __Pyx_call_type_traverse(o, always_call, visit, arg) 0
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg);
#endif

/* LimitedApiGetTypeDict.proto (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp);
#endif

/* SetItemOnTypeDict.proto (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v);
#define __Pyx_SetItemOnTypeDict(tp, k, v) __Pyx__SetItemOnTypeDict((PyTypeObject*)tp, k, v)

/* FixUpExtensionType.proto */
static CYTHON_INLINE int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);

/* PyObjectCallNoArg.proto (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectCallMethod0.proto (used by PyType_Ready) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* ValidateBasesTuple.proto (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto (used by MergeVTables) */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
static int __Pyx_MergeVtables(PyTypeObject *type);

/* DelItemOnTypeDict.proto (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k);
#define __Pyx_DelItemOnTypeDict(tp, k) __Pyx__DelItemOnTypeDict((PyTypeObject*)tp, k)

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_2_4
#define __PYX_HAVE_RT_ImportType_proto_3_2_4
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_2_4 {
   __Pyx_ImportType_CheckSize_Error_3_2_4 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_2_4 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_2_4 = 2
};
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size);
#endif

/* HasAttr.proto (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* ImportImpl.export */
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level);

/* Import.proto */
static CYTHON_INLINE PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level);

/* dict_setdefault.proto (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value);

/* AddModuleRef.proto (used by FetchSharedCythonModule) */
#if ((CYTHON_COMPILING_IN_CPYTHON_FREETHREADING ) ||\
     __PYX_LIMITED_VERSION_HEX < 0x030d0000)
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name);
#else
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#endif

/* FetchSharedCythonModule.proto (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto (used by CommonTypesMetaclass) */
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases);

/* CommonTypesMetaclass.proto (used by CythonFunctionShared) */
static int __pyx_CommonTypesMetaclass_init(PyObject *module);
#define __Pyx_CommonTypesMetaclass_USED

/* PyMethodNew.proto (used by CythonFunctionShared) */
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ);

/* PyVectorcallFastCallDict.proto (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto (used by CythonFunction) */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && CYTHON_METH_FASTCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_dict;
#endif
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    PyObject *defaults;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_mstate_global->__pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_mstate_global->__pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void));
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func,
                                                         PyTypeObject *defaults_type);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CLineInTraceback.proto (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto (used by AddTraceback) */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CheckUnpickleChecksum.proto */
static CYTHON_INLINE int __Pyx_CheckUnpickleChecksum(long checksum, long checksum1, long checksum2, long checksum3, const char *members);

/* PyObjectVectorCallKwBuilder.proto (used by CIntToPy) */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_PY_LONG_LONG(unsigned PY_LONG_LONG value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_int(unsigned int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_char(unsigned char value);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned char __Pyx_PyLong_As_unsigned_char(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* UpdateUnpickledDict.proto */
static int __Pyx_UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
static unsigned long __Pyx_cached_runtime_version = 0;
static void __Pyx_init_runtime_version(void);
#else
#define __Pyx_init_runtime_version()
#endif
static unsigned long __Pyx_get_runtime_version(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* DecompressString.proto */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."

static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_extract_key(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self, char *__pyx_v_last); /* proto*/
static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_process_line(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__record_to_value_and_refs(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__record_to_item(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record); /* proto*/
static __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__lookup_record(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_sha1); /* proto*/
static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__count_records(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_c_content, char *__pyx_v_c_end); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__parse_bytes(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_data); /* proto*/
static char *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__parse_one_entry(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_c_cur, char *__pyx_v_c_end, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_cur_record); /* proto*/
static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__offset_for_sha1(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_sha1); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__compute_common(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto*/

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "cpython.list" */

/* Module declarations from "cpython.mem" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "cpython.tuple" */

/* Module declarations from "libc.stdlib" */

/* Module declarations from "breezy.bzr._str_helpers" */
static CYTHON_INLINE void *__pyx_f_6breezy_3bzr_12_str_helpers__my_memrchr(void *, int, size_t); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(char *, Py_ssize_t); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6breezy_3bzr_12_str_helpers_safe_interned_string_from_size(char *, Py_ssize_t); /*proto*/

/* Module declarations from "breezy.bzr._btree_serializer_pyx" */
static int __pyx_v_6breezy_3bzr_21_btree_serializer_pyx__unhexbuf[256];
static char *__pyx_v_6breezy_3bzr_21_btree_serializer_pyx__hexbuf;
static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__populate_unhexbuf(void); /*proto*/
static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__unhexlify_sha1(char *, char *); /*proto*/
static void __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__hexlify_sha1(char *, char *); /*proto*/
static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__key_to_sha1(PyObject *, char *); /*proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_key(char *); /*proto*/
static unsigned int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_uint(char *); /*proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__format_record(__pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *); /*proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx___pyx_unpickle_BTreeLeafParser__set_state(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *, PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "breezy.bzr._btree_serializer_pyx"
extern int __pyx_module_is_main_breezy__bzr___btree_serializer_pyx;
int __pyx_module_is_main_breezy__bzr___btree_serializer_pyx = 0;

/* Implementation of "breezy.bzr._btree_serializer_pyx" */
/* #### Code section: global_var ### */
/* #### Code section: string_decls ### */
static const char __pyx_k_0_0[] = " 0 0";
static const char __pyx_k_sha1[] = "sha1:";
static const char __pyx_k_type_leaf[] = "type=leaf";
static const char __pyx_k_type_leaf_2[] = "type=leaf\n";
static const char __pyx_k_cur_str__end_str__header_found[] = "_cur_str, _end_str, _header_found, _start, data, key_length, keys, ref_list_length";
static const char __pyx_k_Pyrex_extensions_to_btree_node_p[] = "Pyrex extensions to btree node parsing.";
/* #### Code section: decls ### */
static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser___init__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self, PyObject *__pyx_v_data, PyObject *__pyx_v_key_length, PyObject *__pyx_v_ref_list_length); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_2parse(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_4__reduce_cython__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_6__setstate_cython__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx__parse_leaf_lines(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_data, PyObject *__pyx_v_key_length, PyObject *__pyx_v_ref_list_length); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_2_py_unhexlify(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_as_hex); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_4_py_hexlify(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_as_bin); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_6_py_key_to_sha1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_8_py_sha1_to_key(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_sha1_bin); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode___sizeof__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static void __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_2__dealloc__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_4__init__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_bytes); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7min_key___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7max_key___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_6__contains__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8__getitem__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static Py_ssize_t __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_10__len__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12all_keys(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_14all_items(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_16_get_offset_for_sha1(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_sha1); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_18_get_offsets(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_2__set__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_4__del__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_2__set__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_2__set__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_20__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_22__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_10_parse_into_chk(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bytes, PyObject *__pyx_v_key_length, PyObject *__pyx_v_ref_list_length); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_12_flatten_node(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_node, PyObject *__pyx_v_reference_lists); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_14__pyx_unpickle_BTreeLeafParser(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  PyObject *__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser;
  PyObject *__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode;
  PyTypeObject *__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser;
  PyTypeObject *__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_values;
  PyObject *__pyx_slice[1];
  PyObject *__pyx_codeobj_tab[18];
  PyObject *__pyx_string_tab[156];
  PyObject *__pyx_number_tab[6];
/* #### Code section: module_state_contents ### */
/* CommonTypesMetaclass.module_state_decls */
PyTypeObject *__pyx_CommonTypesMetaclassType;

/* CachedMethodType.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API
PyObject *__Pyx_CachedMethodType;
#endif

/* CythonFunctionShared.module_state_decls */
PyTypeObject *__pyx_CyFunctionType;

/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_Expected_bytes_for_value_not_r __pyx_string_tab[0]
#define __pyx_kp_u_Failed_to_find_the_value_area __pyx_string_tab[1]
#define __pyx_kp_u_Failed_to_parse_block_length __pyx_string_tab[2]
#define __pyx_kp_u_Failed_to_parse_block_offset __pyx_string_tab[3]
#define __pyx_kp_u_Failed_to_parse_record_end __pyx_string_tab[4]
#define __pyx_kp_u_Line_did_not_contain_40_hex_byte __pyx_string_tab[5]
#define __pyx_kp_u_Node_did_not_start_with_type_lea __pyx_string_tab[6]
#define __pyx_kp_u_Note_that_Cython_is_deliberately __pyx_string_tab[7]
#define __pyx_kp_u_Something_went_wrong_while_parsi __pyx_string_tab[8]
#define __pyx_kp_u_We_expect_reference_bits_to_be_b __pyx_string_tab[9]
#define __pyx_kp_u_We_expect_references_to_be_tuple __pyx_string_tab[10]
#define __pyx_kp_u_We_expected_a_tuple_for_node_not __pyx_string_tab[11]
#define __pyx_kp_u_We_failed_to_unhexlify __pyx_string_tab[12]
#define __pyx_kp_u_We_only_support_parsing_byte_str __pyx_string_tab[13]
#define __pyx_kp_u_With_ref_lists_we_expected_4_ent __pyx_string_tab[14]
#define __pyx_kp_u_Without_ref_lists_we_need_at_lea __pyx_string_tab[15]
#define __pyx_kp_u__2 __pyx_string_tab[16]
#define __pyx_kp_u_add_note __pyx_string_tab[17]
#define __pyx_kp_u_breezy_bzr__btree_serializer_pyx_2 __pyx_string_tab[18]
#define __pyx_kp_u_bytes_did_not_start_with_type_le __pyx_string_tab[19]
#define __pyx_kp_u_disable __pyx_string_tab[20]
#define __pyx_kp_u_enable __pyx_string_tab[21]
#define __pyx_kp_u_gc __pyx_string_tab[22]
#define __pyx_kp_u_invalid_key_loop_counter_self_re __pyx_string_tab[23]
#define __pyx_kp_u_invalid_key_wanted_segment_from __pyx_string_tab[24]
#define __pyx_kp_u_is_not_present __pyx_string_tab[25]
#define __pyx_kp_u_isenabled __pyx_string_tab[26]
#define __pyx_kp_u_key_2 __pyx_string_tab[27]
#define __pyx_kp_u_last_self__start __pyx_string_tab[28]
#define __pyx_kp_u_line_did_not_start_with_sha1 __pyx_string_tab[29]
#define __pyx_kp_u_not_a_20_byte_binary_digest __pyx_string_tab[30]
#define __pyx_kp_u_not_a_40_byte_hex_digest __pyx_string_tab[31]
#define __pyx_kp_u_only_1_null_not_2_as_expected __pyx_string_tab[32]
#define __pyx_kp_u_self_data_is_not_a_byte_string __pyx_string_tab[33]
#define __pyx_kp_u_self_last_record_self_records_ca __pyx_string_tab[34]
#define __pyx_kp_u_sha1_bin_must_be_a_str_of_exactl __pyx_string_tab[35]
#define __pyx_kp_u_tree_fragment __pyx_string_tab[36]
#define __pyx_kp_u_tried_to_create_a_string_with_an __pyx_string_tab[37]
#define __pyx_kp_u_unexpected_reference_data_presen __pyx_string_tab[38]
#define __pyx_n_u_BTreeLeafParser __pyx_string_tab[39]
#define __pyx_n_u_BTreeLeafParser___reduce_cython __pyx_string_tab[40]
#define __pyx_n_u_BTreeLeafParser___setstate_cytho __pyx_string_tab[41]
#define __pyx_n_u_BTreeLeafParser_parse __pyx_string_tab[42]
#define __pyx_n_u_GCCHKSHA1LeafNode __pyx_string_tab[43]
#define __pyx_n_u_GCCHKSHA1LeafNode___reduce_cytho __pyx_string_tab[44]
#define __pyx_n_u_GCCHKSHA1LeafNode___setstate_cyt __pyx_string_tab[45]
#define __pyx_n_u_GCCHKSHA1LeafNode___sizeof __pyx_string_tab[46]
#define __pyx_n_u_GCCHKSHA1LeafNode__get_offset_fo __pyx_string_tab[47]
#define __pyx_n_u_GCCHKSHA1LeafNode__get_offsets __pyx_string_tab[48]
#define __pyx_n_u_GCCHKSHA1LeafNode_all_items __pyx_string_tab[49]
#define __pyx_n_u_GCCHKSHA1LeafNode_all_keys __pyx_string_tab[50]
#define __pyx_n_u_Pyx_PyDict_NextRef __pyx_string_tab[51]
#define __pyx_n_u_all_items __pyx_string_tab[52]
#define __pyx_n_u_all_keys __pyx_string_tab[53]
#define __pyx_n_u_as_bin __pyx_string_tab[54]
#define __pyx_n_u_as_bin_sha __pyx_string_tab[55]
#define __pyx_n_u_as_hex __pyx_string_tab[56]
#define __pyx_n_u_asyncio_coroutines __pyx_string_tab[57]
#define __pyx_n_u_breezy_bzr__btree_serializer_pyx __pyx_string_tab[58]
#define __pyx_n_u_byte_count __pyx_string_tab[59]
#define __pyx_n_u_bytes __pyx_string_tab[60]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[61]
#define __pyx_n_u_data __pyx_string_tab[62]
#define __pyx_n_u_dict __pyx_string_tab[63]
#define __pyx_n_u_dict_2 __pyx_string_tab[64]
#define __pyx_n_u_first_ref_list __pyx_string_tab[65]
#define __pyx_n_u_first_reference __pyx_string_tab[66]
#define __pyx_n_u_flat_len __pyx_string_tab[67]
#define __pyx_n_u_flatten_node __pyx_string_tab[68]
#define __pyx_n_u_func __pyx_string_tab[69]
#define __pyx_n_u_get_offset_for_sha1 __pyx_string_tab[70]
#define __pyx_n_u_get_offsets __pyx_string_tab[71]
#define __pyx_n_u_getstate __pyx_string_tab[72]
#define __pyx_n_u_have_reference_lists __pyx_string_tab[73]
#define __pyx_n_u_i __pyx_string_tab[74]
#define __pyx_n_u_is_coroutine __pyx_string_tab[75]
#define __pyx_n_u_item __pyx_string_tab[76]
#define __pyx_n_u_items __pyx_string_tab[77]
#define __pyx_n_u_join __pyx_string_tab[78]
#define __pyx_n_u_key __pyx_string_tab[79]
#define __pyx_n_u_key_len __pyx_string_tab[80]
#define __pyx_n_u_key_length __pyx_string_tab[81]
#define __pyx_n_u_line __pyx_string_tab[82]
#define __pyx_n_u_main __pyx_string_tab[83]
#define __pyx_n_u_module __pyx_string_tab[84]
#define __pyx_n_u_name __pyx_string_tab[85]
#define __pyx_n_u_new __pyx_string_tab[86]
#define __pyx_n_u_next_len __pyx_string_tab[87]
#define __pyx_n_u_node __pyx_string_tab[88]
#define __pyx_n_u_node_len __pyx_string_tab[89]
#define __pyx_n_u_out __pyx_string_tab[90]
#define __pyx_n_u_parse __pyx_string_tab[91]
#define __pyx_n_u_parse_into_chk __pyx_string_tab[92]
#define __pyx_n_u_parse_leaf_lines __pyx_string_tab[93]
#define __pyx_n_u_parser __pyx_string_tab[94]
#define __pyx_n_u_pop __pyx_string_tab[95]
#define __pyx_n_u_py_hexlify __pyx_string_tab[96]
#define __pyx_n_u_py_key_to_sha1 __pyx_string_tab[97]
#define __pyx_n_u_py_sha1_to_key __pyx_string_tab[98]
#define __pyx_n_u_py_unhexlify __pyx_string_tab[99]
#define __pyx_n_u_pyx_checksum __pyx_string_tab[100]
#define __pyx_n_u_pyx_result __pyx_string_tab[101]
#define __pyx_n_u_pyx_state __pyx_string_tab[102]
#define __pyx_n_u_pyx_type __pyx_string_tab[103]
#define __pyx_n_u_pyx_unpickle_BTreeLeafParser __pyx_string_tab[104]
#define __pyx_n_u_pyx_vtable __pyx_string_tab[105]
#define __pyx_n_u_qualname __pyx_string_tab[106]
#define __pyx_n_u_reduce __pyx_string_tab[107]
#define __pyx_n_u_reduce_cython __pyx_string_tab[108]
#define __pyx_n_u_reduce_ex __pyx_string_tab[109]
#define __pyx_n_u_ref_bit __pyx_string_tab[110]
#define __pyx_n_u_ref_bit_len __pyx_string_tab[111]
#define __pyx_n_u_ref_list __pyx_string_tab[112]
#define __pyx_n_u_ref_list_length __pyx_string_tab[113]
#define __pyx_n_u_ref_lists __pyx_string_tab[114]
#define __pyx_n_u_reference __pyx_string_tab[115]
#define __pyx_n_u_reference_lists __pyx_string_tab[116]
#define __pyx_n_u_refs_len __pyx_string_tab[117]
#define __pyx_n_u_result __pyx_string_tab[118]
#define __pyx_n_u_self __pyx_string_tab[119]
#define __pyx_n_u_set_name __pyx_string_tab[120]
#define __pyx_n_u_setdefault __pyx_string_tab[121]
#define __pyx_n_u_setstate __pyx_string_tab[122]
#define __pyx_n_u_setstate_cython __pyx_string_tab[123]
#define __pyx_n_u_sha1_2 __pyx_string_tab[124]
#define __pyx_n_u_sha1_bin __pyx_string_tab[125]
#define __pyx_n_u_sizeof __pyx_string_tab[126]
#define __pyx_n_u_state __pyx_string_tab[127]
#define __pyx_n_u_string_key __pyx_string_tab[128]
#define __pyx_n_u_sys __pyx_string_tab[129]
#define __pyx_n_u_test __pyx_string_tab[130]
#define __pyx_n_u_update __pyx_string_tab[131]
#define __pyx_n_u_use_setstate __pyx_string_tab[132]
#define __pyx_n_u_val __pyx_string_tab[133]
#define __pyx_n_u_value __pyx_string_tab[134]
#define __pyx_n_u_value_len __pyx_string_tab[135]
#define __pyx_n_u_values __pyx_string_tab[136]
#define __pyx_kp_b_ __pyx_string_tab[137]
#define __pyx_kp_b_d __pyx_string_tab[138]
#define __pyx_kp_b_iso88591_0_t_Qa_iq_d_1_s_1_1_q_9Cq_AQ_Qa __pyx_string_tab[139]
#define __pyx_kp_b_iso88591_6_1E_1_q_1 __pyx_string_tab[140]
#define __pyx_kp_b_iso88591_AV_q_6_q __pyx_string_tab[141]
#define __pyx_kp_b_iso88591_A_4_a_Qd_L_a_L_Jb_d_Bd_Q_t1 __pyx_string_tab[142]
#define __pyx_kp_b_iso88591_A_G9A_haq_q __pyx_string_tab[143]
#define __pyx_kp_b_iso88591_A_G9D_4_q_XQa_q __pyx_string_tab[144]
#define __pyx_kp_b_iso88591_A_G9D_at81Ba_q __pyx_string_tab[145]
#define __pyx_kp_b_iso88591_A_a_r_Q __pyx_string_tab[146]
#define __pyx_kp_b_iso88591_A_t_A_6aq __pyx_string_tab[147]
#define __pyx_kp_b_iso88591_Q __pyx_string_tab[148]
#define __pyx_kp_b_iso88591_QfA __pyx_string_tab[149]
#define __pyx_kp_b_iso88591_T_D_4_t9DPWW_hhllsswwx_G1F_a_vW __pyx_string_tab[150]
#define __pyx_kp_b_iso88591_c_3a_AQ __pyx_string_tab[151]
#define __pyx_kp_b_iso88591_q_0_kQR_1_7_1_2DNRS_1 __pyx_string_tab[152]
#define __pyx_kp_b_iso88591_s_83c_D_2_1_j_avQ_1I_aq_1 __pyx_string_tab[153]
#define __pyx_kp_b_iso88591_t_Qhc_9_Q_j_avQ_a_2C1A_q_1 __pyx_string_tab[154]
#define __pyx_kp_b_iso88591_t_Qj_1Jc_j_q __pyx_string_tab[155]
#define __pyx_int_0 __pyx_number_tab[0]
#define __pyx_int_1 __pyx_number_tab[1]
#define __pyx_int_10 __pyx_number_tab[2]
#define __pyx_int_159274753 __pyx_number_tab[3]
#define __pyx_int_2147483648 __pyx_number_tab[4]
#define __pyx_int_4294967295 __pyx_number_tab[5]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser);
  Py_CLEAR(clear_module_state->__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser);
  Py_CLEAR(clear_module_state->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode);
  Py_CLEAR(clear_module_state->__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode);
  for (int i=0; i<1; ++i) { Py_CLEAR(clear_module_state->__pyx_slice[i]); }
  for (int i=0; i<18; ++i) { Py_CLEAR(clear_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<156; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<6; ++i) { Py_CLEAR(clear_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_clear_contents ### */
/* CommonTypesMetaclass.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_clear_end ### */
return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser);
  Py_VISIT(traverse_module_state->__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser);
  Py_VISIT(traverse_module_state->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode);
  Py_VISIT(traverse_module_state->__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode);
  for (int i=0; i<1; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_slice[i]); }
  for (int i=0; i<18; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<156; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<6; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_traverse_contents ### */
/* CommonTypesMetaclass.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_traverse_end ### */
return 0;
}
#endif
/* #### Code section: module_code ### */

/* "_str_helpers.pxd":27
 * 
 * 
 * cdef inline void* _my_memrchr(void *s, int c, size_t n): # cannot_raise             # <<<<<<<<<<<<<<
 *     # memrchr seems to be a GNU extension, so we have to implement it ourselves
 *     cdef char *pos
*/

static CYTHON_INLINE void *__pyx_f_6breezy_3bzr_12_str_helpers__my_memrchr(void *__pyx_v_s, int __pyx_v_c, size_t __pyx_v_n) {
  char *__pyx_v_pos;
  char *__pyx_v_start;
  void *__pyx_r;
  int __pyx_t_1;

  /* "_str_helpers.pxd":32
 *     cdef char *start
 * 
 *     start = <char*>s             # <<<<<<<<<<<<<<
 *     pos = start + n - 1
 *     while pos >= start:
*/
  __pyx_v_start = ((char *)__pyx_v_s);

  /* "_str_helpers.pxd":33
 * 
 *     start = <char*>s
 *     pos = start + n - 1             # <<<<<<<<<<<<<<
 *     while pos >= start:
 *         if pos[0] == c:
*/
  __pyx_v_pos = ((__pyx_v_start + __pyx_v_n) - 1);

  /* "_str_helpers.pxd":34
 *     start = <char*>s
 *     pos = start + n - 1
 *     while pos >= start:             # <<<<<<<<<<<<<<
 *         if pos[0] == c:
 *             return <void*>pos
*/
  while (1) {
    __pyx_t_1 = (__pyx_v_pos >= __pyx_v_start);
    if (!__pyx_t_1) break;

    /* "_str_helpers.pxd":35
 *     pos = start + n - 1
 *     while pos >= start:
 *         if pos[0] == c:             # <<<<<<<<<<<<<<
 *             return <void*>pos
 *         pos = pos - 1
*/
    __pyx_t_1 = ((__pyx_v_pos[0]) == __pyx_v_c);
    if (__pyx_t_1) {

      /* "_str_helpers.pxd":36
 *     while pos >= start:
 *         if pos[0] == c:
 *             return <void*>pos             # <<<<<<<<<<<<<<
 *         pos = pos - 1
 *     return NULL
*/
      __pyx_r = ((void *)__pyx_v_pos);
      goto __pyx_L0;

      /* "_str_helpers.pxd":35
 *     pos = start + n - 1
 *     while pos >= start:
 *         if pos[0] == c:             # <<<<<<<<<<<<<<
 *             return <void*>pos
 *         pos = pos - 1
*/
    }

    /* "_str_helpers.pxd":37
 *         if pos[0] == c:
 *             return <void*>pos
 *         pos = pos - 1             # <<<<<<<<<<<<<<
 *     return NULL
 * 
*/
    __pyx_v_pos = (__pyx_v_pos - 1);
  }

  /* "_str_helpers.pxd":38
 *             return <void*>pos
 *         pos = pos - 1
 *     return NULL             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = NULL;
  goto __pyx_L0;

  /* "_str_helpers.pxd":27
 * 
 * 
 * cdef inline void* _my_memrchr(void *s, int c, size_t n): # cannot_raise             # <<<<<<<<<<<<<<
 *     # memrchr seems to be a GNU extension, so we have to implement it ourselves
 *     cdef char *pos
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "_str_helpers.pxd":41
 * 
 * 
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
*/

static CYTHON_INLINE PyObject *__pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(char *__pyx_v_s, Py_ssize_t __pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("safe_string_from_size", 0);

  /* "_str_helpers.pxd":42
 * 
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
*/
  __pyx_t_1 = (__pyx_v_size < 0);
  if (unlikely(__pyx_t_1)) {

    /* "_str_helpers.pxd":43
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:
 *         raise AssertionError(             # <<<<<<<<<<<<<<
 *             'tried to create a string with an invalid size: %d' % size)
 *     return PyBytes_FromStringAndSize(s, size)
*/
    __pyx_t_3 = NULL;

    /* "_str_helpers.pxd":44
 *     if size < 0:
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)             # <<<<<<<<<<<<<<
 *     return PyBytes_FromStringAndSize(s, size)
 * 
*/
    __pyx_t_4 = PyLong_FromSsize_t(__pyx_v_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyUnicode_Format(__pyx_mstate_global->__pyx_kp_u_tried_to_create_a_string_with_an, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 43, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 43, __pyx_L1_error)

    /* "_str_helpers.pxd":42
 * 
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
*/
  }

  /* "_str_helpers.pxd":45
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
 *     return PyBytes_FromStringAndSize(s, size)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyBytes_FromStringAndSize(__pyx_v_s, __pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "_str_helpers.pxd":41
 * 
 * 
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("breezy.bzr._str_helpers.safe_string_from_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_str_helpers.pxd":48
 * 
 * 
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
*/

static CYTHON_INLINE PyObject *__pyx_f_6breezy_3bzr_12_str_helpers_safe_interned_string_from_size(char *__pyx_v_s, Py_ssize_t __pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("safe_interned_string_from_size", 0);

  /* "_str_helpers.pxd":49
 * 
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
*/
  __pyx_t_1 = (__pyx_v_size < 0);
  if (unlikely(__pyx_t_1)) {

    /* "_str_helpers.pxd":50
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:
 *         raise AssertionError(             # <<<<<<<<<<<<<<
 *             'tried to create a string with an invalid size: %d' % size)
 *     return PyBytes_InternFromStringAndSize(s, size)
*/
    __pyx_t_3 = NULL;

    /* "_str_helpers.pxd":51
 *     if size < 0:
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)             # <<<<<<<<<<<<<<
 *     return PyBytes_InternFromStringAndSize(s, size)
*/
    __pyx_t_4 = PyLong_FromSsize_t(__pyx_v_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyUnicode_Format(__pyx_mstate_global->__pyx_kp_u_tried_to_create_a_string_with_an, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 50, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 50, __pyx_L1_error)

    /* "_str_helpers.pxd":49
 * 
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
*/
  }

  /* "_str_helpers.pxd":52
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
 *     return PyBytes_InternFromStringAndSize(s, size)             # <<<<<<<<<<<<<<
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyBytes_InternFromStringAndSize(__pyx_v_s, __pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "_str_helpers.pxd":48
 * 
 * 
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("breezy.bzr._str_helpers.safe_interned_string_from_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":74
 *     cdef int _header_found
 * 
 *     def __init__(self, data, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *         self.data = data
 *         self.key_length = key_length
*/

/* Python wrapper */
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_key_length = 0;
  PyObject *__pyx_v_ref_list_length = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_data,&__pyx_mstate_global->__pyx_n_u_key_length,&__pyx_mstate_global->__pyx_n_u_ref_list_length,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 74, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 74, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 74, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 74, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 74, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, i); __PYX_ERR(0, 74, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 74, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 74, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 74, __pyx_L3_error)
    }
    __pyx_v_data = values[0];
    __pyx_v_key_length = values[1];
    __pyx_v_ref_list_length = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 74, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.BTreeLeafParser.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser___init__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self), __pyx_v_data, __pyx_v_key_length, __pyx_v_ref_list_length);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser___init__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self, PyObject *__pyx_v_data, PyObject *__pyx_v_key_length, PyObject *__pyx_v_ref_list_length) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":75
 * 
 *     def __init__(self, data, key_length, ref_list_length):
 *         self.data = data             # <<<<<<<<<<<<<<
 *         self.key_length = key_length
 *         self.ref_list_length = ref_list_length
*/
  __Pyx_INCREF(__pyx_v_data);
  __Pyx_GIVEREF(__pyx_v_data);
  __Pyx_GOTREF(__pyx_v_self->data);
  __Pyx_DECREF(__pyx_v_self->data);
  __pyx_v_self->data = __pyx_v_data;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":76
 *     def __init__(self, data, key_length, ref_list_length):
 *         self.data = data
 *         self.key_length = key_length             # <<<<<<<<<<<<<<
 *         self.ref_list_length = ref_list_length
 *         self.keys = []
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_key_length); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 76, __pyx_L1_error)
  __pyx_v_self->key_length = __pyx_t_1;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":77
 *         self.data = data
 *         self.key_length = key_length
 *         self.ref_list_length = ref_list_length             # <<<<<<<<<<<<<<
 *         self.keys = []
 *         self._cur_str = NULL
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_ref_list_length); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 77, __pyx_L1_error)
  __pyx_v_self->ref_list_length = __pyx_t_1;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":78
 *         self.key_length = key_length
 *         self.ref_list_length = ref_list_length
 *         self.keys = []             # <<<<<<<<<<<<<<
 *         self._cur_str = NULL
 *         self._end_str = NULL
*/
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->keys);
  __Pyx_DECREF(__pyx_v_self->keys);
  __pyx_v_self->keys = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":79
 *         self.ref_list_length = ref_list_length
 *         self.keys = []
 *         self._cur_str = NULL             # <<<<<<<<<<<<<<
 *         self._end_str = NULL
 *         self._header_found = 0
*/
  __pyx_v_self->_cur_str = NULL;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":80
 *         self.keys = []
 *         self._cur_str = NULL
 *         self._end_str = NULL             # <<<<<<<<<<<<<<
 *         self._header_found = 0
 *         # keys are tuples
*/
  __pyx_v_self->_end_str = NULL;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":81
 *         self._cur_str = NULL
 *         self._end_str = NULL
 *         self._header_found = 0             # <<<<<<<<<<<<<<
 *         # keys are tuples
 * 
*/
  __pyx_v_self->_header_found = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":74
 *     cdef int _header_found
 * 
 *     def __init__(self, data, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *         self.data = data
 *         self.key_length = key_length
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.BTreeLeafParser.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":84
 *         # keys are tuples
 * 
 *     cdef extract_key(self, char * last):             # <<<<<<<<<<<<<<
 *         """Extract a key.
 * 
*/

static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_extract_key(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self, char *__pyx_v_last) {
  char *__pyx_v_temp_ptr;
  int __pyx_v_loop_counter;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_failure_string = NULL;
  PyObject *__pyx_v_key_element = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("extract_key", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":94
 *         cdef tuple key
 * 
 *         key = PyTuple_New(self.key_length)             # <<<<<<<<<<<<<<
 *         for loop_counter from 0 <= loop_counter < self.key_length:
 *             # grab a key segment
*/
  __pyx_t_1 = PyTuple_New(__pyx_v_self->key_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_key = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":95
 * 
 *         key = PyTuple_New(self.key_length)
 *         for loop_counter from 0 <= loop_counter < self.key_length:             # <<<<<<<<<<<<<<
 *             # grab a key segment
 *             temp_ptr = <char*>memchr(self._start, c'\0', last - self._start)
*/
  __pyx_t_2 = __pyx_v_self->key_length;
  for (__pyx_v_loop_counter = 0; __pyx_v_loop_counter < __pyx_t_2; __pyx_v_loop_counter++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":97
 *         for loop_counter from 0 <= loop_counter < self.key_length:
 *             # grab a key segment
 *             temp_ptr = <char*>memchr(self._start, c'\0', last - self._start)             # <<<<<<<<<<<<<<
 *             if temp_ptr == NULL:
 *                 if loop_counter + 1 == self.key_length:
*/
    __pyx_v_temp_ptr = ((char *)memchr(__pyx_v_self->_start, '\x00', (__pyx_v_last - __pyx_v_self->_start)));

    /* "breezy/bzr/_btree_serializer_pyx.pyx":98
 *             # grab a key segment
 *             temp_ptr = <char*>memchr(self._start, c'\0', last - self._start)
 *             if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *                 if loop_counter + 1 == self.key_length:
 *                     # capture to last
*/
    __pyx_t_3 = (__pyx_v_temp_ptr == NULL);
    if (__pyx_t_3) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":99
 *             temp_ptr = <char*>memchr(self._start, c'\0', last - self._start)
 *             if temp_ptr == NULL:
 *                 if loop_counter + 1 == self.key_length:             # <<<<<<<<<<<<<<
 *                     # capture to last
 *                     temp_ptr = last
*/
      __pyx_t_3 = ((__pyx_v_loop_counter + 1) == __pyx_v_self->key_length);
      if (likely(__pyx_t_3)) {

        /* "breezy/bzr/_btree_serializer_pyx.pyx":101
 *                 if loop_counter + 1 == self.key_length:
 *                     # capture to last
 *                     temp_ptr = last             # <<<<<<<<<<<<<<
 *                 else:
 *                     # Invalid line
*/
        __pyx_v_temp_ptr = __pyx_v_last;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":99
 *             temp_ptr = <char*>memchr(self._start, c'\0', last - self._start)
 *             if temp_ptr == NULL:
 *                 if loop_counter + 1 == self.key_length:             # <<<<<<<<<<<<<<
 *                     # capture to last
 *                     temp_ptr = last
*/
        goto __pyx_L6;
      }

      /* "breezy/bzr/_btree_serializer_pyx.pyx":104
 *                 else:
 *                     # Invalid line
 *                     failure_string = ("invalid key, wanted segment from " +             # <<<<<<<<<<<<<<
 *                         repr(safe_string_from_size(self._start,
 *                                                    last - self._start)))
*/
      /*else*/ {

        /* "breezy/bzr/_btree_serializer_pyx.pyx":105
 *                     # Invalid line
 *                     failure_string = ("invalid key, wanted segment from " +
 *                         repr(safe_string_from_size(self._start,             # <<<<<<<<<<<<<<
 *                                                    last - self._start)))
 *                     raise AssertionError(failure_string)
*/
        __pyx_t_1 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(__pyx_v_self->_start, (__pyx_v_last - __pyx_v_self->_start)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 105, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":104
 *                 else:
 *                     # Invalid line
 *                     failure_string = ("invalid key, wanted segment from " +             # <<<<<<<<<<<<<<
 *                         repr(safe_string_from_size(self._start,
 *                                                    last - self._start)))
*/
        __pyx_t_1 = __Pyx_PyUnicode_ConcatSafe(__pyx_mstate_global->__pyx_kp_u_invalid_key_wanted_segment_from, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 104, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v_failure_string = ((PyObject*)__pyx_t_1);
        __pyx_t_1 = 0;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":107
 *                         repr(safe_string_from_size(self._start,
 *                                                    last - self._start)))
 *                     raise AssertionError(failure_string)             # <<<<<<<<<<<<<<
 *             # capture the key string
 *             if (self.key_length == 1
*/
        __pyx_t_4 = NULL;
        __pyx_t_5 = 1;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_failure_string};
          __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 107, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
        }
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(0, 107, __pyx_L1_error)
      }
      __pyx_L6:;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":98
 *             # grab a key segment
 *             temp_ptr = <char*>memchr(self._start, c'\0', last - self._start)
 *             if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *                 if loop_counter + 1 == self.key_length:
 *                     # capture to last
*/
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":109
 *                     raise AssertionError(failure_string)
 *             # capture the key string
 *             if (self.key_length == 1             # <<<<<<<<<<<<<<
 *                 and (temp_ptr - self._start) == 45
 *                 and strncmp(self._start, b'sha1:', 5) == 0):
*/
    __pyx_t_6 = (__pyx_v_self->key_length == 1);
    if (__pyx_t_6) {
    } else {
      __pyx_t_3 = __pyx_t_6;
      goto __pyx_L8_bool_binop_done;
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":110
 *             # capture the key string
 *             if (self.key_length == 1
 *                 and (temp_ptr - self._start) == 45             # <<<<<<<<<<<<<<
 *                 and strncmp(self._start, b'sha1:', 5) == 0):
 *                 key_element = safe_string_from_size(self._start,
*/
    __pyx_t_6 = ((__pyx_v_temp_ptr - __pyx_v_self->_start) == 45);
    if (__pyx_t_6) {
    } else {
      __pyx_t_3 = __pyx_t_6;
      goto __pyx_L8_bool_binop_done;
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":111
 *             if (self.key_length == 1
 *                 and (temp_ptr - self._start) == 45
 *                 and strncmp(self._start, b'sha1:', 5) == 0):             # <<<<<<<<<<<<<<
 *                 key_element = safe_string_from_size(self._start,
 *                                                     temp_ptr - self._start)
*/
    __pyx_t_6 = (strncmp(__pyx_v_self->_start, __pyx_k_sha1, 5) == 0);
    __pyx_t_3 = __pyx_t_6;
    __pyx_L8_bool_binop_done:;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":109
 *                     raise AssertionError(failure_string)
 *             # capture the key string
 *             if (self.key_length == 1             # <<<<<<<<<<<<<<
 *                 and (temp_ptr - self._start) == 45
 *                 and strncmp(self._start, b'sha1:', 5) == 0):
*/
    if (__pyx_t_3) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":112
 *                 and (temp_ptr - self._start) == 45
 *                 and strncmp(self._start, b'sha1:', 5) == 0):
 *                 key_element = safe_string_from_size(self._start,             # <<<<<<<<<<<<<<
 *                                                     temp_ptr - self._start)
 *             else:
*/
      __pyx_t_1 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(__pyx_v_self->_start, (__pyx_v_temp_ptr - __pyx_v_self->_start)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 112, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_key_element, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":109
 *                     raise AssertionError(failure_string)
 *             # capture the key string
 *             if (self.key_length == 1             # <<<<<<<<<<<<<<
 *                 and (temp_ptr - self._start) == 45
 *                 and strncmp(self._start, b'sha1:', 5) == 0):
*/
      goto __pyx_L7;
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":115
 *                                                     temp_ptr - self._start)
 *             else:
 *                 key_element = safe_interned_string_from_size(self._start,             # <<<<<<<<<<<<<<
 *                                                          temp_ptr - self._start)
 *             # advance our pointer
*/
    /*else*/ {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":116
 *             else:
 *                 key_element = safe_interned_string_from_size(self._start,
 *                                                          temp_ptr - self._start)             # <<<<<<<<<<<<<<
 *             # advance our pointer
 *             self._start = temp_ptr + 1
*/
      __pyx_t_1 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_interned_string_from_size(__pyx_v_self->_start, (__pyx_v_temp_ptr - __pyx_v_self->_start)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 115, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_key_element, __pyx_t_1);
      __pyx_t_1 = 0;
    }
    __pyx_L7:;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":118
 *                                                          temp_ptr - self._start)
 *             # advance our pointer
 *             self._start = temp_ptr + 1             # <<<<<<<<<<<<<<
 *             Py_INCREF(key_element)
 *             PyTuple_SET_ITEM(key, loop_counter, key_element)
*/
    __pyx_v_self->_start = (__pyx_v_temp_ptr + 1);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":119
 *             # advance our pointer
 *             self._start = temp_ptr + 1
 *             Py_INCREF(key_element)             # <<<<<<<<<<<<<<
 *             PyTuple_SET_ITEM(key, loop_counter, key_element)
 *         return key
*/
    Py_INCREF(__pyx_v_key_element);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":120
 *             self._start = temp_ptr + 1
 *             Py_INCREF(key_element)
 *             PyTuple_SET_ITEM(key, loop_counter, key_element)             # <<<<<<<<<<<<<<
 *         return key
 * 
*/
    PyTuple_SET_ITEM(__pyx_v_key, __pyx_v_loop_counter, __pyx_v_key_element);
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":121
 *             Py_INCREF(key_element)
 *             PyTuple_SET_ITEM(key, loop_counter, key_element)
 *         return key             # <<<<<<<<<<<<<<
 * 
 *     cdef int process_line(self) except -1:
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_key);
  __pyx_r = __pyx_v_key;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":84
 *         # keys are tuples
 * 
 *     cdef extract_key(self, char * last):             # <<<<<<<<<<<<<<
 *         """Extract a key.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.BTreeLeafParser.extract_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_failure_string);
  __Pyx_XDECREF(__pyx_v_key_element);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":123
 *         return key
 * 
 *     cdef int process_line(self) except -1:             # <<<<<<<<<<<<<<
 *         """Process a line in the bytes."""
 *         cdef char *last
*/

static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_process_line(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self) {
  char *__pyx_v_last;
  char *__pyx_v_temp_ptr;
  char *__pyx_v_ref_ptr;
  char *__pyx_v_next_start;
  int __pyx_v_loop_counter;
  Py_ssize_t __pyx_v_str_len;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_ref_lists = NULL;
  PyObject *__pyx_v_ref_list = NULL;
  PyObject *__pyx_v_node_value = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  void *__pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("process_line", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":132
 *         cdef Py_ssize_t str_len
 * 
 *         self._start = self._cur_str             # <<<<<<<<<<<<<<
 *         # Find the next newline
 *         last = <char*>memchr(self._start, c'\n', self._end_str - self._start)
*/
  __pyx_t_1 = __pyx_v_self->_cur_str;
  __pyx_v_self->_start = __pyx_t_1;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":134
 *         self._start = self._cur_str
 *         # Find the next newline
 *         last = <char*>memchr(self._start, c'\n', self._end_str - self._start)             # <<<<<<<<<<<<<<
 *         if last == NULL:
 *             # Process until the end of the file
*/
  __pyx_v_last = ((char *)memchr(__pyx_v_self->_start, '\n', (__pyx_v_self->_end_str - __pyx_v_self->_start)));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":135
 *         # Find the next newline
 *         last = <char*>memchr(self._start, c'\n', self._end_str - self._start)
 *         if last == NULL:             # <<<<<<<<<<<<<<
 *             # Process until the end of the file
 *             last = self._end_str
*/
  __pyx_t_2 = (__pyx_v_last == NULL);
  if (__pyx_t_2) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":137
 *         if last == NULL:
 *             # Process until the end of the file
 *             last = self._end_str             # <<<<<<<<<<<<<<
 *             self._cur_str = self._end_str
 *         else:
*/
    __pyx_t_1 = __pyx_v_self->_end_str;
    __pyx_v_last = __pyx_t_1;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":138
 *             # Process until the end of the file
 *             last = self._end_str
 *             self._cur_str = self._end_str             # <<<<<<<<<<<<<<
 *         else:
 *             # And the next string is right after it
*/
    __pyx_t_1 = __pyx_v_self->_end_str;
    __pyx_v_self->_cur_str = __pyx_t_1;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":135
 *         # Find the next newline
 *         last = <char*>memchr(self._start, c'\n', self._end_str - self._start)
 *         if last == NULL:             # <<<<<<<<<<<<<<
 *             # Process until the end of the file
 *             last = self._end_str
*/
    goto __pyx_L3;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":141
 *         else:
 *             # And the next string is right after it
 *             self._cur_str = last + 1             # <<<<<<<<<<<<<<
 *             # The last character is right before the '\n'
 * 
*/
  /*else*/ {
    __pyx_v_self->_cur_str = (__pyx_v_last + 1);
  }
  __pyx_L3:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":144
 *             # The last character is right before the '\n'
 * 
 *         if last == self._start:             # <<<<<<<<<<<<<<
 *             # parsed it all.
 *             return 0
*/
  __pyx_t_2 = (__pyx_v_last == __pyx_v_self->_start);
  if (__pyx_t_2) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":146
 *         if last == self._start:
 *             # parsed it all.
 *             return 0             # <<<<<<<<<<<<<<
 *         if last < self._start:
 *             # Unexpected error condition - fail
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":144
 *             # The last character is right before the '\n'
 * 
 *         if last == self._start:             # <<<<<<<<<<<<<<
 *             # parsed it all.
 *             return 0
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":147
 *             # parsed it all.
 *             return 0
 *         if last < self._start:             # <<<<<<<<<<<<<<
 *             # Unexpected error condition - fail
 *             raise AssertionError("last < self._start")
*/
  __pyx_t_2 = (__pyx_v_last < __pyx_v_self->_start);
  if (unlikely(__pyx_t_2)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":149
 *         if last < self._start:
 *             # Unexpected error condition - fail
 *             raise AssertionError("last < self._start")             # <<<<<<<<<<<<<<
 *         if 0 == self._header_found:
 *             # The first line in a leaf node is the header "type=leaf\n"
*/
    __pyx_t_4 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_last_self__start};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 149, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 149, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":147
 *             # parsed it all.
 *             return 0
 *         if last < self._start:             # <<<<<<<<<<<<<<
 *             # Unexpected error condition - fail
 *             raise AssertionError("last < self._start")
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":150
 *             # Unexpected error condition - fail
 *             raise AssertionError("last < self._start")
 *         if 0 == self._header_found:             # <<<<<<<<<<<<<<
 *             # The first line in a leaf node is the header "type=leaf\n"
 *             if strncmp(b"type=leaf", self._start, last - self._start) == 0:
*/
  __pyx_t_2 = (0 == __pyx_v_self->_header_found);
  if (__pyx_t_2) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":152
 *         if 0 == self._header_found:
 *             # The first line in a leaf node is the header "type=leaf\n"
 *             if strncmp(b"type=leaf", self._start, last - self._start) == 0:             # <<<<<<<<<<<<<<
 *                 self._header_found = 1
 *                 return 0
*/
    __pyx_t_2 = (strncmp(__pyx_k_type_leaf, __pyx_v_self->_start, (__pyx_v_last - __pyx_v_self->_start)) == 0);
    if (likely(__pyx_t_2)) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":153
 *             # The first line in a leaf node is the header "type=leaf\n"
 *             if strncmp(b"type=leaf", self._start, last - self._start) == 0:
 *                 self._header_found = 1             # <<<<<<<<<<<<<<
 *                 return 0
 *             else:
*/
      __pyx_v_self->_header_found = 1;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":154
 *             if strncmp(b"type=leaf", self._start, last - self._start) == 0:
 *                 self._header_found = 1
 *                 return 0             # <<<<<<<<<<<<<<
 *             else:
 *                 raise AssertionError('Node did not start with "type=leaf": %r'
*/
      __pyx_r = 0;
      goto __pyx_L0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":152
 *         if 0 == self._header_found:
 *             # The first line in a leaf node is the header "type=leaf\n"
 *             if strncmp(b"type=leaf", self._start, last - self._start) == 0:             # <<<<<<<<<<<<<<
 *                 self._header_found = 1
 *                 return 0
*/
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":156
 *                 return 0
 *             else:
 *                 raise AssertionError('Node did not start with "type=leaf": %r'             # <<<<<<<<<<<<<<
 *                     % (safe_string_from_size(self._start, last - self._start)))
 * 
*/
    /*else*/ {
      __pyx_t_4 = NULL;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":157
 *             else:
 *                 raise AssertionError('Node did not start with "type=leaf": %r'
 *                     % (safe_string_from_size(self._start, last - self._start)))             # <<<<<<<<<<<<<<
 * 
 *         key = self.extract_key(last)
*/
      __pyx_t_6 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(__pyx_v_self->_start, (__pyx_v_last - __pyx_v_self->_start)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 157, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyUnicode_FormatSafe(__pyx_mstate_global->__pyx_kp_u_Node_did_not_start_with_type_lea, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 157, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_7};
        __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 156, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 156, __pyx_L1_error)
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":150
 *             # Unexpected error condition - fail
 *             raise AssertionError("last < self._start")
 *         if 0 == self._header_found:             # <<<<<<<<<<<<<<
 *             # The first line in a leaf node is the header "type=leaf\n"
 *             if strncmp(b"type=leaf", self._start, last - self._start) == 0:
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":159
 *                     % (safe_string_from_size(self._start, last - self._start)))
 * 
 *         key = self.extract_key(last)             # <<<<<<<<<<<<<<
 *         # find the value area
 *         temp_ptr = <char*>_my_memrchr(self._start, c'\0', last - self._start)
*/
  __pyx_t_3 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self->__pyx_vtab)->extract_key(__pyx_v_self, __pyx_v_last); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_key = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":161
 *         key = self.extract_key(last)
 *         # find the value area
 *         temp_ptr = <char*>_my_memrchr(self._start, c'\0', last - self._start)             # <<<<<<<<<<<<<<
 *         if temp_ptr == NULL:
 *             # Invalid line
*/
  __pyx_t_8 = __pyx_f_6breezy_3bzr_12_str_helpers__my_memrchr(__pyx_v_self->_start, '\x00', (__pyx_v_last - __pyx_v_self->_start)); if (unlikely(__pyx_t_8 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 161, __pyx_L1_error)
  __pyx_v_temp_ptr = ((char *)__pyx_t_8);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":162
 *         # find the value area
 *         temp_ptr = <char*>_my_memrchr(self._start, c'\0', last - self._start)
 *         if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *             # Invalid line
 *             raise AssertionError("Failed to find the value area")
*/
  __pyx_t_2 = (__pyx_v_temp_ptr == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":164
 *         if temp_ptr == NULL:
 *             # Invalid line
 *             raise AssertionError("Failed to find the value area")             # <<<<<<<<<<<<<<
 *         else:
 *             # Because of how conversions were done, we ended up with *lots* of
*/
    __pyx_t_7 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_mstate_global->__pyx_kp_u_Failed_to_find_the_value_area};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 164, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 164, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":162
 *         # find the value area
 *         temp_ptr = <char*>_my_memrchr(self._start, c'\0', last - self._start)
 *         if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *             # Invalid line
 *             raise AssertionError("Failed to find the value area")
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":172
 *             # something like '12607215 328306 0 0', which ends up consuming 1MB
 *             # of memory, just for those strings.
 *             str_len = last - temp_ptr - 1             # <<<<<<<<<<<<<<
 *             if (str_len > 4
 *                 and strncmp(b" 0 0", last - 4, 4) == 0):
*/
  /*else*/ {
    __pyx_v_str_len = ((__pyx_v_last - __pyx_v_temp_ptr) - 1);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":173
 *             # of memory, just for those strings.
 *             str_len = last - temp_ptr - 1
 *             if (str_len > 4             # <<<<<<<<<<<<<<
 *                 and strncmp(b" 0 0", last - 4, 4) == 0):
 *                 # This drops peak mem for bzr.dev from 87.4MB => 86.2MB
*/
    __pyx_t_9 = (__pyx_v_str_len > 4);
    if (__pyx_t_9) {
    } else {
      __pyx_t_2 = __pyx_t_9;
      goto __pyx_L10_bool_binop_done;
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":174
 *             str_len = last - temp_ptr - 1
 *             if (str_len > 4
 *                 and strncmp(b" 0 0", last - 4, 4) == 0):             # <<<<<<<<<<<<<<
 *                 # This drops peak mem for bzr.dev from 87.4MB => 86.2MB
 *                 # For Launchpad 236MB => 232MB
*/
    __pyx_t_9 = (strncmp(__pyx_k_0_0, (__pyx_v_last - 4), 4) == 0);
    __pyx_t_2 = __pyx_t_9;
    __pyx_L10_bool_binop_done:;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":173
 *             # of memory, just for those strings.
 *             str_len = last - temp_ptr - 1
 *             if (str_len > 4             # <<<<<<<<<<<<<<
 *                 and strncmp(b" 0 0", last - 4, 4) == 0):
 *                 # This drops peak mem for bzr.dev from 87.4MB => 86.2MB
*/
    if (__pyx_t_2) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":177
 *                 # This drops peak mem for bzr.dev from 87.4MB => 86.2MB
 *                 # For Launchpad 236MB => 232MB
 *                 value = safe_interned_string_from_size(temp_ptr + 1, str_len)             # <<<<<<<<<<<<<<
 *             else:
 *                 value = safe_string_from_size(temp_ptr + 1, str_len)
*/
      __pyx_t_3 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_interned_string_from_size((__pyx_v_temp_ptr + 1), __pyx_v_str_len); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 177, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_value = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":173
 *             # of memory, just for those strings.
 *             str_len = last - temp_ptr - 1
 *             if (str_len > 4             # <<<<<<<<<<<<<<
 *                 and strncmp(b" 0 0", last - 4, 4) == 0):
 *                 # This drops peak mem for bzr.dev from 87.4MB => 86.2MB
*/
      goto __pyx_L9;
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":179
 *                 value = safe_interned_string_from_size(temp_ptr + 1, str_len)
 *             else:
 *                 value = safe_string_from_size(temp_ptr + 1, str_len)             # <<<<<<<<<<<<<<
 *             # shrink the references end point
 *             last = temp_ptr
*/
    /*else*/ {
      __pyx_t_3 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size((__pyx_v_temp_ptr + 1), __pyx_v_str_len); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 179, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_value = __pyx_t_3;
      __pyx_t_3 = 0;
    }
    __pyx_L9:;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":181
 *                 value = safe_string_from_size(temp_ptr + 1, str_len)
 *             # shrink the references end point
 *             last = temp_ptr             # <<<<<<<<<<<<<<
 * 
 *         if self.ref_list_length:
*/
    __pyx_v_last = __pyx_v_temp_ptr;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":183
 *             last = temp_ptr
 * 
 *         if self.ref_list_length:             # <<<<<<<<<<<<<<
 *             ref_lists = PyTuple_New(self.ref_list_length)
 *             loop_counter = 0
*/
  __pyx_t_2 = (__pyx_v_self->ref_list_length != 0);
  if (__pyx_t_2) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":184
 * 
 *         if self.ref_list_length:
 *             ref_lists = PyTuple_New(self.ref_list_length)             # <<<<<<<<<<<<<<
 *             loop_counter = 0
 *             while loop_counter < self.ref_list_length:
*/
    __pyx_t_3 = PyTuple_New(__pyx_v_self->ref_list_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_ref_lists = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":185
 *         if self.ref_list_length:
 *             ref_lists = PyTuple_New(self.ref_list_length)
 *             loop_counter = 0             # <<<<<<<<<<<<<<
 *             while loop_counter < self.ref_list_length:
 *                 ref_list = []
*/
    __pyx_v_loop_counter = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":186
 *             ref_lists = PyTuple_New(self.ref_list_length)
 *             loop_counter = 0
 *             while loop_counter < self.ref_list_length:             # <<<<<<<<<<<<<<
 *                 ref_list = []
 *                 # extract a reference list
*/
    while (1) {
      __pyx_t_2 = (__pyx_v_loop_counter < __pyx_v_self->ref_list_length);
      if (!__pyx_t_2) break;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":187
 *             loop_counter = 0
 *             while loop_counter < self.ref_list_length:
 *                 ref_list = []             # <<<<<<<<<<<<<<
 *                 # extract a reference list
 *                 loop_counter = loop_counter + 1
*/
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_XDECREF_SET(__pyx_v_ref_list, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":189
 *                 ref_list = []
 *                 # extract a reference list
 *                 loop_counter = loop_counter + 1             # <<<<<<<<<<<<<<
 *                 if last < self._start:
 *                     raise AssertionError("last < self._start")
*/
      __pyx_v_loop_counter = (__pyx_v_loop_counter + 1);

      /* "breezy/bzr/_btree_serializer_pyx.pyx":190
 *                 # extract a reference list
 *                 loop_counter = loop_counter + 1
 *                 if last < self._start:             # <<<<<<<<<<<<<<
 *                     raise AssertionError("last < self._start")
 *                 # find the next reference list end point:
*/
      __pyx_t_2 = (__pyx_v_last < __pyx_v_self->_start);
      if (unlikely(__pyx_t_2)) {

        /* "breezy/bzr/_btree_serializer_pyx.pyx":191
 *                 loop_counter = loop_counter + 1
 *                 if last < self._start:
 *                     raise AssertionError("last < self._start")             # <<<<<<<<<<<<<<
 *                 # find the next reference list end point:
 *                 temp_ptr = <char*>memchr(self._start, c'\t', last - self._start)
*/
        __pyx_t_7 = NULL;
        __pyx_t_5 = 1;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_mstate_global->__pyx_kp_u_last_self__start};
          __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 191, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
        }
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 191, __pyx_L1_error)

        /* "breezy/bzr/_btree_serializer_pyx.pyx":190
 *                 # extract a reference list
 *                 loop_counter = loop_counter + 1
 *                 if last < self._start:             # <<<<<<<<<<<<<<
 *                     raise AssertionError("last < self._start")
 *                 # find the next reference list end point:
*/
      }

      /* "breezy/bzr/_btree_serializer_pyx.pyx":193
 *                     raise AssertionError("last < self._start")
 *                 # find the next reference list end point:
 *                 temp_ptr = <char*>memchr(self._start, c'\t', last - self._start)             # <<<<<<<<<<<<<<
 *                 if temp_ptr == NULL:
 *                     # Only valid for the last list
*/
      __pyx_v_temp_ptr = ((char *)memchr(__pyx_v_self->_start, '\t', (__pyx_v_last - __pyx_v_self->_start)));

      /* "breezy/bzr/_btree_serializer_pyx.pyx":194
 *                 # find the next reference list end point:
 *                 temp_ptr = <char*>memchr(self._start, c'\t', last - self._start)
 *                 if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *                     # Only valid for the last list
 *                     if loop_counter != self.ref_list_length:
*/
      __pyx_t_2 = (__pyx_v_temp_ptr == NULL);
      if (__pyx_t_2) {

        /* "breezy/bzr/_btree_serializer_pyx.pyx":196
 *                 if temp_ptr == NULL:
 *                     # Only valid for the last list
 *                     if loop_counter != self.ref_list_length:             # <<<<<<<<<<<<<<
 *                         # Invalid line
 *                         raise AssertionError(
*/
        __pyx_t_2 = (__pyx_v_loop_counter != __pyx_v_self->ref_list_length);
        if (unlikely(__pyx_t_2)) {

          /* "breezy/bzr/_btree_serializer_pyx.pyx":198
 *                     if loop_counter != self.ref_list_length:
 *                         # Invalid line
 *                         raise AssertionError(             # <<<<<<<<<<<<<<
 *                             "invalid key, loop_counter != self.ref_list_length")
 *                     else:
*/
          __pyx_t_7 = NULL;
          __pyx_t_5 = 1;
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_mstate_global->__pyx_kp_u_invalid_key_loop_counter_self_re};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 198, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 198, __pyx_L1_error)

          /* "breezy/bzr/_btree_serializer_pyx.pyx":196
 *                 if temp_ptr == NULL:
 *                     # Only valid for the last list
 *                     if loop_counter != self.ref_list_length:             # <<<<<<<<<<<<<<
 *                         # Invalid line
 *                         raise AssertionError(
*/
        }

        /* "breezy/bzr/_btree_serializer_pyx.pyx":202
 *                     else:
 *                         # scan to the end of the ref list area
 *                         ref_ptr = last             # <<<<<<<<<<<<<<
 *                         next_start = last
 *                 else:
*/
        /*else*/ {
          __pyx_v_ref_ptr = __pyx_v_last;

          /* "breezy/bzr/_btree_serializer_pyx.pyx":203
 *                         # scan to the end of the ref list area
 *                         ref_ptr = last
 *                         next_start = last             # <<<<<<<<<<<<<<
 *                 else:
 *                     # scan to the end of this ref list
*/
          __pyx_v_next_start = __pyx_v_last;
        }

        /* "breezy/bzr/_btree_serializer_pyx.pyx":194
 *                 # find the next reference list end point:
 *                 temp_ptr = <char*>memchr(self._start, c'\t', last - self._start)
 *                 if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *                     # Only valid for the last list
 *                     if loop_counter != self.ref_list_length:
*/
        goto __pyx_L16;
      }

      /* "breezy/bzr/_btree_serializer_pyx.pyx":206
 *                 else:
 *                     # scan to the end of this ref list
 *                     ref_ptr = temp_ptr             # <<<<<<<<<<<<<<
 *                     next_start = temp_ptr + 1
 *                 # Now, there may be multiple keys in the ref list.
*/
      /*else*/ {
        __pyx_v_ref_ptr = __pyx_v_temp_ptr;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":207
 *                     # scan to the end of this ref list
 *                     ref_ptr = temp_ptr
 *                     next_start = temp_ptr + 1             # <<<<<<<<<<<<<<
 *                 # Now, there may be multiple keys in the ref list.
 *                 while self._start < ref_ptr:
*/
        __pyx_v_next_start = (__pyx_v_temp_ptr + 1);
      }
      __pyx_L16:;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":209
 *                     next_start = temp_ptr + 1
 *                 # Now, there may be multiple keys in the ref list.
 *                 while self._start < ref_ptr:             # <<<<<<<<<<<<<<
 *                     # loop finding keys and extracting them
 *                     temp_ptr = <char*>memchr(self._start, c'\r',
*/
      while (1) {
        __pyx_t_2 = (__pyx_v_self->_start < __pyx_v_ref_ptr);
        if (!__pyx_t_2) break;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":211
 *                 while self._start < ref_ptr:
 *                     # loop finding keys and extracting them
 *                     temp_ptr = <char*>memchr(self._start, c'\r',             # <<<<<<<<<<<<<<
 *                                              ref_ptr - self._start)
 *                     if temp_ptr == NULL:
*/
        __pyx_v_temp_ptr = ((char *)memchr(__pyx_v_self->_start, '\r', (__pyx_v_ref_ptr - __pyx_v_self->_start)));

        /* "breezy/bzr/_btree_serializer_pyx.pyx":213
 *                     temp_ptr = <char*>memchr(self._start, c'\r',
 *                                              ref_ptr - self._start)
 *                     if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *                         # key runs to the end
 *                         temp_ptr = ref_ptr
*/
        __pyx_t_2 = (__pyx_v_temp_ptr == NULL);
        if (__pyx_t_2) {

          /* "breezy/bzr/_btree_serializer_pyx.pyx":215
 *                     if temp_ptr == NULL:
 *                         # key runs to the end
 *                         temp_ptr = ref_ptr             # <<<<<<<<<<<<<<
 * 
 *                     PyList_Append(ref_list, self.extract_key(temp_ptr))
*/
          __pyx_v_temp_ptr = __pyx_v_ref_ptr;

          /* "breezy/bzr/_btree_serializer_pyx.pyx":213
 *                     temp_ptr = <char*>memchr(self._start, c'\r',
 *                                              ref_ptr - self._start)
 *                     if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *                         # key runs to the end
 *                         temp_ptr = ref_ptr
*/
        }

        /* "breezy/bzr/_btree_serializer_pyx.pyx":217
 *                         temp_ptr = ref_ptr
 * 
 *                     PyList_Append(ref_list, self.extract_key(temp_ptr))             # <<<<<<<<<<<<<<
 *                 ref_list = tuple(ref_list)
 *                 Py_INCREF(ref_list)
*/
        __pyx_t_3 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self->__pyx_vtab)->extract_key(__pyx_v_self, __pyx_v_temp_ptr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 217, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_10 = PyList_Append(__pyx_v_ref_list, __pyx_t_3); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 217, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "breezy/bzr/_btree_serializer_pyx.pyx":218
 * 
 *                     PyList_Append(ref_list, self.extract_key(temp_ptr))
 *                 ref_list = tuple(ref_list)             # <<<<<<<<<<<<<<
 *                 Py_INCREF(ref_list)
 *                 PyTuple_SET_ITEM(ref_lists, loop_counter - 1, ref_list)
*/
      __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_v_ref_list); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 218, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF_SET(__pyx_v_ref_list, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":219
 *                     PyList_Append(ref_list, self.extract_key(temp_ptr))
 *                 ref_list = tuple(ref_list)
 *                 Py_INCREF(ref_list)             # <<<<<<<<<<<<<<
 *                 PyTuple_SET_ITEM(ref_lists, loop_counter - 1, ref_list)
 *                 # prepare for the next reference list
*/
      Py_INCREF(__pyx_v_ref_list);

      /* "breezy/bzr/_btree_serializer_pyx.pyx":220
 *                 ref_list = tuple(ref_list)
 *                 Py_INCREF(ref_list)
 *                 PyTuple_SET_ITEM(ref_lists, loop_counter - 1, ref_list)             # <<<<<<<<<<<<<<
 *                 # prepare for the next reference list
 *                 self._start = next_start
*/
      PyTuple_SET_ITEM(__pyx_v_ref_lists, (__pyx_v_loop_counter - 1), __pyx_v_ref_list);

      /* "breezy/bzr/_btree_serializer_pyx.pyx":222
 *                 PyTuple_SET_ITEM(ref_lists, loop_counter - 1, ref_list)
 *                 # prepare for the next reference list
 *                 self._start = next_start             # <<<<<<<<<<<<<<
 *             node_value = (value, ref_lists)
 *         else:
*/
      __pyx_v_self->_start = __pyx_v_next_start;
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":223
 *                 # prepare for the next reference list
 *                 self._start = next_start
 *             node_value = (value, ref_lists)             # <<<<<<<<<<<<<<
 *         else:
 *             if last != self._start:
*/
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_value) != (0)) __PYX_ERR(0, 223, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_ref_lists);
    __Pyx_GIVEREF(__pyx_v_ref_lists);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_ref_lists) != (0)) __PYX_ERR(0, 223, __pyx_L1_error);
    __pyx_v_node_value = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":183
 *             last = temp_ptr
 * 
 *         if self.ref_list_length:             # <<<<<<<<<<<<<<
 *             ref_lists = PyTuple_New(self.ref_list_length)
 *             loop_counter = 0
*/
    goto __pyx_L12;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":225
 *             node_value = (value, ref_lists)
 *         else:
 *             if last != self._start:             # <<<<<<<<<<<<<<
 *                 # unexpected reference data present
 *                 raise AssertionError("unexpected reference data present")
*/
  /*else*/ {
    __pyx_t_2 = (__pyx_v_last != __pyx_v_self->_start);
    if (unlikely(__pyx_t_2)) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":227
 *             if last != self._start:
 *                 # unexpected reference data present
 *                 raise AssertionError("unexpected reference data present")             # <<<<<<<<<<<<<<
 *             node_value = (value, ())
 *         PyList_Append(self.keys, (key, node_value))
*/
      __pyx_t_7 = NULL;
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_mstate_global->__pyx_kp_u_unexpected_reference_data_presen};
        __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 227, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 227, __pyx_L1_error)

      /* "breezy/bzr/_btree_serializer_pyx.pyx":225
 *             node_value = (value, ref_lists)
 *         else:
 *             if last != self._start:             # <<<<<<<<<<<<<<
 *                 # unexpected reference data present
 *                 raise AssertionError("unexpected reference data present")
*/
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":228
 *                 # unexpected reference data present
 *                 raise AssertionError("unexpected reference data present")
 *             node_value = (value, ())             # <<<<<<<<<<<<<<
 *         PyList_Append(self.keys, (key, node_value))
 *         return 0
*/
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 228, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_value) != (0)) __PYX_ERR(0, 228, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_empty_tuple);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_empty_tuple);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_mstate_global->__pyx_empty_tuple) != (0)) __PYX_ERR(0, 228, __pyx_L1_error);
    __pyx_v_node_value = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
  }
  __pyx_L12:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":229
 *                 raise AssertionError("unexpected reference data present")
 *             node_value = (value, ())
 *         PyList_Append(self.keys, (key, node_value))             # <<<<<<<<<<<<<<
 *         return 0
 * 
*/
  __pyx_t_3 = __pyx_v_self->keys;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_key);
  __Pyx_GIVEREF(__pyx_v_key);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_key) != (0)) __PYX_ERR(0, 229, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_node_value);
  __Pyx_GIVEREF(__pyx_v_node_value);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_node_value) != (0)) __PYX_ERR(0, 229, __pyx_L1_error);
  __pyx_t_10 = PyList_Append(__pyx_t_3, __pyx_t_7); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":230
 *             node_value = (value, ())
 *         PyList_Append(self.keys, (key, node_value))
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def parse(self):
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":123
 *         return key
 * 
 *     cdef int process_line(self) except -1:             # <<<<<<<<<<<<<<
 *         """Process a line in the bytes."""
 *         cdef char *last
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.BTreeLeafParser.process_line", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_ref_lists);
  __Pyx_XDECREF(__pyx_v_ref_list);
  __Pyx_XDECREF(__pyx_v_node_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":232
 *         return 0
 * 
 *     def parse(self):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t byte_count
 *         if not PyBytes_CheckExact(self.data):
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_3parse(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_3parse = {"parse", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_3parse, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_3parse(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parse (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("parse", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("parse", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_2parse(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_2parse(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self) {
  Py_ssize_t __pyx_v_byte_count;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parse", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":234
 *     def parse(self):
 *         cdef Py_ssize_t byte_count
 *         if not PyBytes_CheckExact(self.data):             # <<<<<<<<<<<<<<
 *             raise AssertionError('self.data is not a byte string.')
 *         byte_count = PyBytes_GET_SIZE(self.data)
*/
  __pyx_t_1 = __pyx_v_self->data;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = (!PyBytes_CheckExact(__pyx_t_1));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":235
 *         cdef Py_ssize_t byte_count
 *         if not PyBytes_CheckExact(self.data):
 *             raise AssertionError('self.data is not a byte string.')             # <<<<<<<<<<<<<<
 *         byte_count = PyBytes_GET_SIZE(self.data)
 *         self._cur_str = PyBytes_AS_STRING(self.data)
*/
    __pyx_t_3 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_self_data_is_not_a_byte_string};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 235, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":234
 *     def parse(self):
 *         cdef Py_ssize_t byte_count
 *         if not PyBytes_CheckExact(self.data):             # <<<<<<<<<<<<<<
 *             raise AssertionError('self.data is not a byte string.')
 *         byte_count = PyBytes_GET_SIZE(self.data)
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":236
 *         if not PyBytes_CheckExact(self.data):
 *             raise AssertionError('self.data is not a byte string.')
 *         byte_count = PyBytes_GET_SIZE(self.data)             # <<<<<<<<<<<<<<
 *         self._cur_str = PyBytes_AS_STRING(self.data)
 *         # This points to the last character in the string
*/
  __pyx_t_1 = __pyx_v_self->data;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_byte_count = PyBytes_GET_SIZE(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":237
 *             raise AssertionError('self.data is not a byte string.')
 *         byte_count = PyBytes_GET_SIZE(self.data)
 *         self._cur_str = PyBytes_AS_STRING(self.data)             # <<<<<<<<<<<<<<
 *         # This points to the last character in the string
 *         self._end_str = self._cur_str + byte_count
*/
  __pyx_t_1 = __pyx_v_self->data;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_self->_cur_str = PyBytes_AS_STRING(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":239
 *         self._cur_str = PyBytes_AS_STRING(self.data)
 *         # This points to the last character in the string
 *         self._end_str = self._cur_str + byte_count             # <<<<<<<<<<<<<<
 *         while self._cur_str < self._end_str:
 *             self.process_line()
*/
  __pyx_v_self->_end_str = (__pyx_v_self->_cur_str + __pyx_v_byte_count);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":240
 *         # This points to the last character in the string
 *         self._end_str = self._cur_str + byte_count
 *         while self._cur_str < self._end_str:             # <<<<<<<<<<<<<<
 *             self.process_line()
 *         return self.keys
*/
  while (1) {
    __pyx_t_2 = (__pyx_v_self->_cur_str < __pyx_v_self->_end_str);
    if (!__pyx_t_2) break;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":241
 *         self._end_str = self._cur_str + byte_count
 *         while self._cur_str < self._end_str:
 *             self.process_line()             # <<<<<<<<<<<<<<
 *         return self.keys
 * 
*/
    __pyx_t_5 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self->__pyx_vtab)->process_line(__pyx_v_self); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 241, __pyx_L1_error)
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":242
 *         while self._cur_str < self._end_str:
 *             self.process_line()
 *         return self.keys             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->keys);
  __pyx_r = __pyx_v_self->keys;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":232
 *         return 0
 * 
 *     def parse(self):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t byte_count
 *         if not PyBytes_CheckExact(self.data):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.BTreeLeafParser.parse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_4__reduce_cython__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_4__reduce_cython__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._cur_str, self._end_str, self._header_found, self._start, self.data, self.key_length, self.keys, self.ref_list_length)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->_cur_str); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBytes_FromString(__pyx_v_self->_end_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyLong_From_int(__pyx_v_self->_header_found); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_self->_start); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyLong_From_int(__pyx_v_self->key_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyLong_From_int(__pyx_v_self->ref_list_length); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(8); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_3) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_4) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->data);
  __Pyx_GIVEREF(__pyx_v_self->data);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_v_self->data) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 5, __pyx_t_5) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->keys);
  __Pyx_GIVEREF(__pyx_v_self->keys);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 6, __pyx_v_self->keys) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 7, __pyx_t_6) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._cur_str, self._end_str, self._header_found, self._start, self.data, self.key_length, self.keys, self.ref_list_length)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_7 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v__dict = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "(tree fragment)":7
 *     state = (self._cur_str, self._end_str, self._header_found, self._start, self.data, self.key_length, self.keys, self.ref_list_length)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_9 = (__pyx_v__dict != Py_None);
  if (__pyx_t_9) {
  } else {
    __pyx_t_8 = __pyx_t_9;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(2, 7, __pyx_L1_error)
  __pyx_t_8 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_8) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v__dict) != (0)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.data is not None or self.keys is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._cur_str, self._end_str, self._header_found, self._start, self.data, self.key_length, self.keys, self.ref_list_length)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.data is not None or self.keys is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x97e5701, None), state
*/
  /*else*/ {
    __pyx_t_9 = (__pyx_v_self->data != Py_None);
    if (!__pyx_t_9) {
    } else {
      __pyx_t_8 = __pyx_t_9;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_9 = (__pyx_v_self->keys != Py_None);
    __pyx_t_8 = __pyx_t_9;
    __pyx_L6_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_8;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.data is not None or self.keys is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x97e5701, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.data is not None or self.keys is not None
 *     if use_setstate:
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x97e5701, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x97e5701, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_BTreeLeafParser); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_159274753);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_159274753);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_mstate_global->__pyx_int_159274753) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, Py_None) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_7) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_6 = 0;
    __pyx_t_7 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.data is not None or self.keys is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x97e5701, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x97e5701, None), state
 *     else:
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x97e5701, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_BTreeLeafParser__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_BTreeLeafParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_159274753);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_159274753);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_mstate_global->__pyx_int_159274753) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_state) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_7) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_7 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.BTreeLeafParser.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x97e5701, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BTreeLeafParser__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.BTreeLeafParser.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_6__setstate_cython__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_6__setstate_cython__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x97e5701, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_BTreeLeafParser__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(2, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx___pyx_unpickle_BTreeLeafParser__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x97e5701, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BTreeLeafParser__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.BTreeLeafParser.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":245
 * 
 * 
 * def _parse_leaf_lines(data, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     parser = BTreeLeafParser(data, key_length, ref_list_length)
 *     return parser.parse()
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_1_parse_leaf_lines(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_1_parse_leaf_lines = {"_parse_leaf_lines", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_1_parse_leaf_lines, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_1_parse_leaf_lines(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_key_length = 0;
  PyObject *__pyx_v_ref_list_length = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_parse_leaf_lines (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_data,&__pyx_mstate_global->__pyx_n_u_key_length,&__pyx_mstate_global->__pyx_n_u_ref_list_length,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 245, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 245, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 245, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 245, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_parse_leaf_lines", 0) < (0)) __PYX_ERR(0, 245, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_parse_leaf_lines", 1, 3, 3, i); __PYX_ERR(0, 245, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 245, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 245, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 245, __pyx_L3_error)
    }
    __pyx_v_data = values[0];
    __pyx_v_key_length = values[1];
    __pyx_v_ref_list_length = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_parse_leaf_lines", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 245, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._parse_leaf_lines", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx__parse_leaf_lines(__pyx_self, __pyx_v_data, __pyx_v_key_length, __pyx_v_ref_list_length);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx__parse_leaf_lines(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_data, PyObject *__pyx_v_key_length, PyObject *__pyx_v_ref_list_length) {
  struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_parser = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_parse_leaf_lines", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":246
 * 
 * def _parse_leaf_lines(data, key_length, ref_list_length):
 *     parser = BTreeLeafParser(data, key_length, ref_list_length)             # <<<<<<<<<<<<<<
 *     return parser.parse()
 * 
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_2, __pyx_v_data, __pyx_v_key_length, __pyx_v_ref_list_length};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser, __pyx_callargs+__pyx_t_3, (4-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_parser = ((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":247
 * def _parse_leaf_lines(data, key_length, ref_list_length):
 *     parser = BTreeLeafParser(data, key_length, ref_list_length)
 *     return parser.parse()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = ((PyObject *)__pyx_v_parser);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_parse, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":245
 * 
 * 
 * def _parse_leaf_lines(data, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     parser = BTreeLeafParser(data, key_length, ref_list_length)
 *     return parser.parse()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._parse_leaf_lines", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_parser);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":269
 * _hexbuf = b'0123456789abcdef'
 * 
 * cdef _populate_unhexbuf():             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i from 0 <= i < 256:
*/

static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__populate_unhexbuf(void) {
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_populate_unhexbuf", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":271
 * cdef _populate_unhexbuf():
 *     cdef int i
 *     for i from 0 <= i < 256:             # <<<<<<<<<<<<<<
 *         _unhexbuf[i] = -1
 *     for i from 0 <= i < 10: # 0123456789 => map to the raw number
*/
  for (__pyx_v_i = 0; __pyx_v_i < 0x100; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":272
 *     cdef int i
 *     for i from 0 <= i < 256:
 *         _unhexbuf[i] = -1             # <<<<<<<<<<<<<<
 *     for i from 0 <= i < 10: # 0123456789 => map to the raw number
 *         _unhexbuf[(i + c'0')] = i
*/
    (__pyx_v_6breezy_3bzr_21_btree_serializer_pyx__unhexbuf[__pyx_v_i]) = -1;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":273
 *     for i from 0 <= i < 256:
 *         _unhexbuf[i] = -1
 *     for i from 0 <= i < 10: # 0123456789 => map to the raw number             # <<<<<<<<<<<<<<
 *         _unhexbuf[(i + c'0')] = i
 *     for i from 10 <= i < 16: # abcdef => 10, 11, 12, 13, 14, 15, 16
*/
  for (__pyx_v_i = 0; __pyx_v_i < 10; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":274
 *         _unhexbuf[i] = -1
 *     for i from 0 <= i < 10: # 0123456789 => map to the raw number
 *         _unhexbuf[(i + c'0')] = i             # <<<<<<<<<<<<<<
 *     for i from 10 <= i < 16: # abcdef => 10, 11, 12, 13, 14, 15, 16
 *         _unhexbuf[(i - 10 + c'a')] = i
*/
    (__pyx_v_6breezy_3bzr_21_btree_serializer_pyx__unhexbuf[(__pyx_v_i + '0')]) = __pyx_v_i;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":275
 *     for i from 0 <= i < 10: # 0123456789 => map to the raw number
 *         _unhexbuf[(i + c'0')] = i
 *     for i from 10 <= i < 16: # abcdef => 10, 11, 12, 13, 14, 15, 16             # <<<<<<<<<<<<<<
 *         _unhexbuf[(i - 10 + c'a')] = i
 *     for i from 10 <= i < 16: # ABCDEF => 10, 11, 12, 13, 14, 15, 16
*/
  for (__pyx_v_i = 10; __pyx_v_i < 16; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":276
 *         _unhexbuf[(i + c'0')] = i
 *     for i from 10 <= i < 16: # abcdef => 10, 11, 12, 13, 14, 15, 16
 *         _unhexbuf[(i - 10 + c'a')] = i             # <<<<<<<<<<<<<<
 *     for i from 10 <= i < 16: # ABCDEF => 10, 11, 12, 13, 14, 15, 16
 *         _unhexbuf[(i - 10 + c'A')] = i
*/
    (__pyx_v_6breezy_3bzr_21_btree_serializer_pyx__unhexbuf[((__pyx_v_i - 10) + 'a')]) = __pyx_v_i;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":277
 *     for i from 10 <= i < 16: # abcdef => 10, 11, 12, 13, 14, 15, 16
 *         _unhexbuf[(i - 10 + c'a')] = i
 *     for i from 10 <= i < 16: # ABCDEF => 10, 11, 12, 13, 14, 15, 16             # <<<<<<<<<<<<<<
 *         _unhexbuf[(i - 10 + c'A')] = i
 * _populate_unhexbuf()
*/
  for (__pyx_v_i = 10; __pyx_v_i < 16; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":278
 *         _unhexbuf[(i - 10 + c'a')] = i
 *     for i from 10 <= i < 16: # ABCDEF => 10, 11, 12, 13, 14, 15, 16
 *         _unhexbuf[(i - 10 + c'A')] = i             # <<<<<<<<<<<<<<
 * _populate_unhexbuf()
 * 
*/
    (__pyx_v_6breezy_3bzr_21_btree_serializer_pyx__unhexbuf[((__pyx_v_i - 10) + 'A')]) = __pyx_v_i;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":269
 * _hexbuf = b'0123456789abcdef'
 * 
 * cdef _populate_unhexbuf():             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i from 0 <= i < 256:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":282
 * 
 * 
 * cdef int _unhexlify_sha1(char *as_hex, char *as_bin): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Take the hex sha1 in as_hex and make it binary in as_bin
 * 
*/

static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__unhexlify_sha1(char *__pyx_v_as_hex, char *__pyx_v_as_bin) {
  int __pyx_v_top;
  int __pyx_v_bot;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":294
 *     # binascii does this using isupper() and tolower() and ?: syntax. I'm
 *     # guessing a simple lookup array should be faster.
 *     j = 0             # <<<<<<<<<<<<<<
 *     for i from 0 <= i < 20:
 *         top = _unhexbuf[<unsigned char>(as_hex[j])]
*/
  __pyx_v_j = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":295
 *     # guessing a simple lookup array should be faster.
 *     j = 0
 *     for i from 0 <= i < 20:             # <<<<<<<<<<<<<<
 *         top = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1
*/
  for (__pyx_v_i = 0; __pyx_v_i < 20; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":296
 *     j = 0
 *     for i from 0 <= i < 20:
 *         top = _unhexbuf[<unsigned char>(as_hex[j])]             # <<<<<<<<<<<<<<
 *         j = j + 1
 *         bot = _unhexbuf[<unsigned char>(as_hex[j])]
*/
    __pyx_v_top = (__pyx_v_6breezy_3bzr_21_btree_serializer_pyx__unhexbuf[((unsigned char)(__pyx_v_as_hex[__pyx_v_j]))]);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":297
 *     for i from 0 <= i < 20:
 *         top = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1             # <<<<<<<<<<<<<<
 *         bot = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1
*/
    __pyx_v_j = (__pyx_v_j + 1);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":298
 *         top = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1
 *         bot = _unhexbuf[<unsigned char>(as_hex[j])]             # <<<<<<<<<<<<<<
 *         j = j + 1
 *         if top == -1 or bot == -1:
*/
    __pyx_v_bot = (__pyx_v_6breezy_3bzr_21_btree_serializer_pyx__unhexbuf[((unsigned char)(__pyx_v_as_hex[__pyx_v_j]))]);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":299
 *         j = j + 1
 *         bot = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1             # <<<<<<<<<<<<<<
 *         if top == -1 or bot == -1:
 *             return 0
*/
    __pyx_v_j = (__pyx_v_j + 1);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":300
 *         bot = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1
 *         if top == -1 or bot == -1:             # <<<<<<<<<<<<<<
 *             return 0
 *         as_bin[i] = <unsigned char>((top << 4) + bot);
*/
    __pyx_t_2 = (__pyx_v_top == -1L);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_bot == -1L);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_1) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":301
 *         j = j + 1
 *         if top == -1 or bot == -1:
 *             return 0             # <<<<<<<<<<<<<<
 *         as_bin[i] = <unsigned char>((top << 4) + bot);
 *     return 1
*/
      __pyx_r = 0;
      goto __pyx_L0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":300
 *         bot = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1
 *         if top == -1 or bot == -1:             # <<<<<<<<<<<<<<
 *             return 0
 *         as_bin[i] = <unsigned char>((top << 4) + bot);
*/
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":302
 *         if top == -1 or bot == -1:
 *             return 0
 *         as_bin[i] = <unsigned char>((top << 4) + bot);             # <<<<<<<<<<<<<<
 *     return 1
 * 
*/
    (__pyx_v_as_bin[__pyx_v_i]) = ((unsigned char)((__pyx_v_top << 4) + __pyx_v_bot));
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":303
 *             return 0
 *         as_bin[i] = <unsigned char>((top << 4) + bot);
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = 1;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":282
 * 
 * 
 * cdef int _unhexlify_sha1(char *as_hex, char *as_bin): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Take the hex sha1 in as_hex and make it binary in as_bin
 * 
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":306
 * 
 * 
 * def _py_unhexlify(as_hex):             # <<<<<<<<<<<<<<
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if not PyBytes_CheckExact(as_hex) or PyBytes_GET_SIZE(as_hex) != 40:
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_3_py_unhexlify(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_2_py_unhexlify, "For the test infrastructure, just thunks to _unhexlify_sha1");
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_3_py_unhexlify = {"_py_unhexlify", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_3_py_unhexlify, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_2_py_unhexlify};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_3_py_unhexlify(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_as_hex = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_py_unhexlify (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_as_hex,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 306, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 306, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_py_unhexlify", 0) < (0)) __PYX_ERR(0, 306, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_py_unhexlify", 1, 1, 1, i); __PYX_ERR(0, 306, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 306, __pyx_L3_error)
    }
    __pyx_v_as_hex = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_py_unhexlify", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 306, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._py_unhexlify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_2_py_unhexlify(__pyx_self, __pyx_v_as_hex);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_2_py_unhexlify(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_as_hex) {
  PyObject *__pyx_v_as_bin = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_py_unhexlify", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":308
 * def _py_unhexlify(as_hex):
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if not PyBytes_CheckExact(as_hex) or PyBytes_GET_SIZE(as_hex) != 40:             # <<<<<<<<<<<<<<
 *         raise ValueError('not a 40-byte hex digest')
 *     as_bin = PyBytes_FromStringAndSize(NULL, 20)
*/
  __pyx_t_2 = (!PyBytes_CheckExact(__pyx_v_as_hex));
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (PyBytes_GET_SIZE(__pyx_v_as_hex) != 40);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":309
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if not PyBytes_CheckExact(as_hex) or PyBytes_GET_SIZE(as_hex) != 40:
 *         raise ValueError('not a 40-byte hex digest')             # <<<<<<<<<<<<<<
 *     as_bin = PyBytes_FromStringAndSize(NULL, 20)
 *     if _unhexlify_sha1(PyBytes_AS_STRING(as_hex), PyBytes_AS_STRING(as_bin)):
*/
    __pyx_t_4 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_not_a_40_byte_hex_digest};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 309, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 309, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":308
 * def _py_unhexlify(as_hex):
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if not PyBytes_CheckExact(as_hex) or PyBytes_GET_SIZE(as_hex) != 40:             # <<<<<<<<<<<<<<
 *         raise ValueError('not a 40-byte hex digest')
 *     as_bin = PyBytes_FromStringAndSize(NULL, 20)
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":310
 *     if not PyBytes_CheckExact(as_hex) or PyBytes_GET_SIZE(as_hex) != 40:
 *         raise ValueError('not a 40-byte hex digest')
 *     as_bin = PyBytes_FromStringAndSize(NULL, 20)             # <<<<<<<<<<<<<<
 *     if _unhexlify_sha1(PyBytes_AS_STRING(as_hex), PyBytes_AS_STRING(as_bin)):
 *         return as_bin
*/
  __pyx_t_3 = PyBytes_FromStringAndSize(NULL, 20); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_as_bin = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":311
 *         raise ValueError('not a 40-byte hex digest')
 *     as_bin = PyBytes_FromStringAndSize(NULL, 20)
 *     if _unhexlify_sha1(PyBytes_AS_STRING(as_hex), PyBytes_AS_STRING(as_bin)):             # <<<<<<<<<<<<<<
 *         return as_bin
 *     return None
*/
  __pyx_t_6 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__unhexlify_sha1(PyBytes_AS_STRING(__pyx_v_as_hex), PyBytes_AS_STRING(__pyx_v_as_bin)); if (unlikely(__pyx_t_6 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 311, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_6 != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":312
 *     as_bin = PyBytes_FromStringAndSize(NULL, 20)
 *     if _unhexlify_sha1(PyBytes_AS_STRING(as_hex), PyBytes_AS_STRING(as_bin)):
 *         return as_bin             # <<<<<<<<<<<<<<
 *     return None
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_as_bin);
    __pyx_r = __pyx_v_as_bin;
    goto __pyx_L0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":311
 *         raise ValueError('not a 40-byte hex digest')
 *     as_bin = PyBytes_FromStringAndSize(NULL, 20)
 *     if _unhexlify_sha1(PyBytes_AS_STRING(as_hex), PyBytes_AS_STRING(as_bin)):             # <<<<<<<<<<<<<<
 *         return as_bin
 *     return None
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":313
 *     if _unhexlify_sha1(PyBytes_AS_STRING(as_hex), PyBytes_AS_STRING(as_bin)):
 *         return as_bin
 *     return None             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":306
 * 
 * 
 * def _py_unhexlify(as_hex):             # <<<<<<<<<<<<<<
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if not PyBytes_CheckExact(as_hex) or PyBytes_GET_SIZE(as_hex) != 40:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._py_unhexlify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_as_bin);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":316
 * 
 * 
 * cdef void _hexlify_sha1(char *as_bin, char *as_hex): # cannot_raise             # <<<<<<<<<<<<<<
 *     cdef int i, j
 *     cdef char c
*/

static void __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__hexlify_sha1(char *__pyx_v_as_bin, char *__pyx_v_as_hex) {
  int __pyx_v_i;
  int __pyx_v_j;
  char __pyx_v_c;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":320
 *     cdef char c
 * 
 *     j = 0             # <<<<<<<<<<<<<<
 *     for i from 0 <= i < 20:
 *         c = as_bin[i]
*/
  __pyx_v_j = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":321
 * 
 *     j = 0
 *     for i from 0 <= i < 20:             # <<<<<<<<<<<<<<
 *         c = as_bin[i]
 *         as_hex[j] = _hexbuf[(c>>4)&0xf]
*/
  for (__pyx_v_i = 0; __pyx_v_i < 20; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":322
 *     j = 0
 *     for i from 0 <= i < 20:
 *         c = as_bin[i]             # <<<<<<<<<<<<<<
 *         as_hex[j] = _hexbuf[(c>>4)&0xf]
 *         j = j + 1
*/
    __pyx_v_c = (__pyx_v_as_bin[__pyx_v_i]);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":323
 *     for i from 0 <= i < 20:
 *         c = as_bin[i]
 *         as_hex[j] = _hexbuf[(c>>4)&0xf]             # <<<<<<<<<<<<<<
 *         j = j + 1
 *         as_hex[j] = _hexbuf[(c)&0xf]
*/
    (__pyx_v_as_hex[__pyx_v_j]) = (__pyx_v_6breezy_3bzr_21_btree_serializer_pyx__hexbuf[((__pyx_v_c >> 4) & 0xf)]);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":324
 *         c = as_bin[i]
 *         as_hex[j] = _hexbuf[(c>>4)&0xf]
 *         j = j + 1             # <<<<<<<<<<<<<<
 *         as_hex[j] = _hexbuf[(c)&0xf]
 *         j = j + 1
*/
    __pyx_v_j = (__pyx_v_j + 1);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":325
 *         as_hex[j] = _hexbuf[(c>>4)&0xf]
 *         j = j + 1
 *         as_hex[j] = _hexbuf[(c)&0xf]             # <<<<<<<<<<<<<<
 *         j = j + 1
 * 
*/
    (__pyx_v_as_hex[__pyx_v_j]) = (__pyx_v_6breezy_3bzr_21_btree_serializer_pyx__hexbuf[(__pyx_v_c & 0xf)]);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":326
 *         j = j + 1
 *         as_hex[j] = _hexbuf[(c)&0xf]
 *         j = j + 1             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __pyx_v_j = (__pyx_v_j + 1);
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":316
 * 
 * 
 * cdef void _hexlify_sha1(char *as_bin, char *as_hex): # cannot_raise             # <<<<<<<<<<<<<<
 *     cdef int i, j
 *     cdef char c
*/

  /* function exit code */
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":329
 * 
 * 
 * def _py_hexlify(as_bin):             # <<<<<<<<<<<<<<
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyBytes_CheckExact(as_bin):
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_5_py_hexlify(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_4_py_hexlify, "For test infrastructure, thunk to _hexlify_sha1");
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_5_py_hexlify = {"_py_hexlify", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_5_py_hexlify, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_4_py_hexlify};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_5_py_hexlify(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_as_bin = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_py_hexlify (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_as_bin,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 329, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 329, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_py_hexlify", 0) < (0)) __PYX_ERR(0, 329, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_py_hexlify", 1, 1, 1, i); __PYX_ERR(0, 329, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 329, __pyx_L3_error)
    }
    __pyx_v_as_bin = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_py_hexlify", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 329, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._py_hexlify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_4_py_hexlify(__pyx_self, __pyx_v_as_bin);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_4_py_hexlify(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_as_bin) {
  PyObject *__pyx_v_as_hex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_py_hexlify", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":331
 * def _py_hexlify(as_bin):
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyBytes_CheckExact(as_bin):             # <<<<<<<<<<<<<<
 *         raise ValueError('not a 20-byte binary digest')
 *     as_hex = PyBytes_FromStringAndSize(NULL, 40)
*/
  __pyx_t_2 = PyObject_Length(__pyx_v_as_bin); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 331, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 != 20);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (!PyBytes_CheckExact(__pyx_v_as_bin));
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":332
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyBytes_CheckExact(as_bin):
 *         raise ValueError('not a 20-byte binary digest')             # <<<<<<<<<<<<<<
 *     as_hex = PyBytes_FromStringAndSize(NULL, 40)
 *     _hexlify_sha1(PyBytes_AS_STRING(as_bin), PyBytes_AS_STRING(as_hex))
*/
    __pyx_t_5 = NULL;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_mstate_global->__pyx_kp_u_not_a_20_byte_binary_digest};
      __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 332, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":331
 * def _py_hexlify(as_bin):
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyBytes_CheckExact(as_bin):             # <<<<<<<<<<<<<<
 *         raise ValueError('not a 20-byte binary digest')
 *     as_hex = PyBytes_FromStringAndSize(NULL, 40)
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":333
 *     if len(as_bin) != 20 or not PyBytes_CheckExact(as_bin):
 *         raise ValueError('not a 20-byte binary digest')
 *     as_hex = PyBytes_FromStringAndSize(NULL, 40)             # <<<<<<<<<<<<<<
 *     _hexlify_sha1(PyBytes_AS_STRING(as_bin), PyBytes_AS_STRING(as_hex))
 *     return as_hex
*/
  __pyx_t_4 = PyBytes_FromStringAndSize(NULL, 40); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_as_hex = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":334
 *         raise ValueError('not a 20-byte binary digest')
 *     as_hex = PyBytes_FromStringAndSize(NULL, 40)
 *     _hexlify_sha1(PyBytes_AS_STRING(as_bin), PyBytes_AS_STRING(as_hex))             # <<<<<<<<<<<<<<
 *     return as_hex
 * 
*/
  __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__hexlify_sha1(PyBytes_AS_STRING(__pyx_v_as_bin), PyBytes_AS_STRING(__pyx_v_as_hex)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 334, __pyx_L1_error)

  /* "breezy/bzr/_btree_serializer_pyx.pyx":335
 *     as_hex = PyBytes_FromStringAndSize(NULL, 40)
 *     _hexlify_sha1(PyBytes_AS_STRING(as_bin), PyBytes_AS_STRING(as_hex))
 *     return as_hex             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_as_hex);
  __pyx_r = __pyx_v_as_hex;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":329
 * 
 * 
 * def _py_hexlify(as_bin):             # <<<<<<<<<<<<<<
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyBytes_CheckExact(as_bin):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._py_hexlify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_as_hex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":338
 * 
 * 
 * cdef int _key_to_sha1(key, char *sha1): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Map a key into its sha1 content.
 * 
*/

static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__key_to_sha1(PyObject *__pyx_v_key, char *__pyx_v_sha1) {
  char *__pyx_v_c_val;
  PyObject *__pyx_v_p_val;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":348
 *     cdef PyObject *p_val
 * 
 *     if PyTuple_CheckExact(key) and PyTuple_GET_SIZE(key) == 1:             # <<<<<<<<<<<<<<
 *         p_val = <PyObject *>PyTuple_GET_ITEM(key, 0)
 *     else:
*/
  __pyx_t_2 = PyTuple_CheckExact(__pyx_v_key);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (PyTuple_GET_SIZE(__pyx_v_key) == 1);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":349
 * 
 *     if PyTuple_CheckExact(key) and PyTuple_GET_SIZE(key) == 1:
 *         p_val = <PyObject *>PyTuple_GET_ITEM(key, 0)             # <<<<<<<<<<<<<<
 *     else:
 *         # Not a tuple or a PyTuple
*/
    __pyx_v_p_val = ((PyObject *)PyTuple_GET_ITEM(__pyx_v_key, 0));

    /* "breezy/bzr/_btree_serializer_pyx.pyx":348
 *     cdef PyObject *p_val
 * 
 *     if PyTuple_CheckExact(key) and PyTuple_GET_SIZE(key) == 1:             # <<<<<<<<<<<<<<
 *         p_val = <PyObject *>PyTuple_GET_ITEM(key, 0)
 *     else:
*/
    goto __pyx_L3;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":352
 *     else:
 *         # Not a tuple or a PyTuple
 *         return 0             # <<<<<<<<<<<<<<
 *     if (PyBytes_CheckExact(<object>p_val)
 *             and PyBytes_GET_SIZE(<object>p_val) == 45):
*/
  /*else*/ {
    __pyx_r = 0;
    goto __pyx_L0;
  }
  __pyx_L3:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":353
 *         # Not a tuple or a PyTuple
 *         return 0
 *     if (PyBytes_CheckExact(<object>p_val)             # <<<<<<<<<<<<<<
 *             and PyBytes_GET_SIZE(<object>p_val) == 45):
 *         c_val = PyBytes_AS_STRING(<object>p_val)
*/
  __pyx_t_2 = PyBytes_CheckExact(((PyObject *)__pyx_v_p_val));
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":354
 *         return 0
 *     if (PyBytes_CheckExact(<object>p_val)
 *             and PyBytes_GET_SIZE(<object>p_val) == 45):             # <<<<<<<<<<<<<<
 *         c_val = PyBytes_AS_STRING(<object>p_val)
 *     else:
*/
  __pyx_t_2 = (PyBytes_GET_SIZE(((PyObject *)__pyx_v_p_val)) == 45);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":353
 *         # Not a tuple or a PyTuple
 *         return 0
 *     if (PyBytes_CheckExact(<object>p_val)             # <<<<<<<<<<<<<<
 *             and PyBytes_GET_SIZE(<object>p_val) == 45):
 *         c_val = PyBytes_AS_STRING(<object>p_val)
*/
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":355
 *     if (PyBytes_CheckExact(<object>p_val)
 *             and PyBytes_GET_SIZE(<object>p_val) == 45):
 *         c_val = PyBytes_AS_STRING(<object>p_val)             # <<<<<<<<<<<<<<
 *     else:
 *         return 0
*/
    __pyx_v_c_val = PyBytes_AS_STRING(((PyObject *)__pyx_v_p_val));

    /* "breezy/bzr/_btree_serializer_pyx.pyx":353
 *         # Not a tuple or a PyTuple
 *         return 0
 *     if (PyBytes_CheckExact(<object>p_val)             # <<<<<<<<<<<<<<
 *             and PyBytes_GET_SIZE(<object>p_val) == 45):
 *         c_val = PyBytes_AS_STRING(<object>p_val)
*/
    goto __pyx_L6;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":357
 *         c_val = PyBytes_AS_STRING(<object>p_val)
 *     else:
 *         return 0             # <<<<<<<<<<<<<<
 *     if strncmp(c_val, b'sha1:', 5) != 0:
 *         return 0
*/
  /*else*/ {
    __pyx_r = 0;
    goto __pyx_L0;
  }
  __pyx_L6:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":358
 *     else:
 *         return 0
 *     if strncmp(c_val, b'sha1:', 5) != 0:             # <<<<<<<<<<<<<<
 *         return 0
 *     if not _unhexlify_sha1(c_val + 5, sha1):
*/
  __pyx_t_1 = (strncmp(__pyx_v_c_val, __pyx_k_sha1, 5) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":359
 *         return 0
 *     if strncmp(c_val, b'sha1:', 5) != 0:
 *         return 0             # <<<<<<<<<<<<<<
 *     if not _unhexlify_sha1(c_val + 5, sha1):
 *         return 0
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":358
 *     else:
 *         return 0
 *     if strncmp(c_val, b'sha1:', 5) != 0:             # <<<<<<<<<<<<<<
 *         return 0
 *     if not _unhexlify_sha1(c_val + 5, sha1):
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":360
 *     if strncmp(c_val, b'sha1:', 5) != 0:
 *         return 0
 *     if not _unhexlify_sha1(c_val + 5, sha1):             # <<<<<<<<<<<<<<
 *         return 0
 *     return 1
*/
  __pyx_t_3 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__unhexlify_sha1((__pyx_v_c_val + 5), __pyx_v_sha1); if (unlikely(__pyx_t_3 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 360, __pyx_L1_error)
  __pyx_t_1 = (!(__pyx_t_3 != 0));
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":361
 *         return 0
 *     if not _unhexlify_sha1(c_val + 5, sha1):
 *         return 0             # <<<<<<<<<<<<<<
 *     return 1
 * 
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":360
 *     if strncmp(c_val, b'sha1:', 5) != 0:
 *         return 0
 *     if not _unhexlify_sha1(c_val + 5, sha1):             # <<<<<<<<<<<<<<
 *         return 0
 *     return 1
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":362
 *     if not _unhexlify_sha1(c_val + 5, sha1):
 *         return 0
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = 1;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":338
 * 
 * 
 * cdef int _key_to_sha1(key, char *sha1): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Map a key into its sha1 content.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._key_to_sha1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":365
 * 
 * 
 * def _py_key_to_sha1(key):             # <<<<<<<<<<<<<<
 *     """Map a key to a simple sha1 string.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_7_py_key_to_sha1(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_6_py_key_to_sha1, "Map a key to a simple sha1 string.\n\n    This is a testing thunk to the C function.\n    ");
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_7_py_key_to_sha1 = {"_py_key_to_sha1", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_7_py_key_to_sha1, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_6_py_key_to_sha1};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_7_py_key_to_sha1(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_key = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_py_key_to_sha1 (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_key,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 365, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 365, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_py_key_to_sha1", 0) < (0)) __PYX_ERR(0, 365, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_py_key_to_sha1", 1, 1, 1, i); __PYX_ERR(0, 365, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 365, __pyx_L3_error)
    }
    __pyx_v_key = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_py_key_to_sha1", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 365, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._py_key_to_sha1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_6_py_key_to_sha1(__pyx_self, __pyx_v_key);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_6_py_key_to_sha1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_v_as_bin_sha = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_py_key_to_sha1", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":370
 *     This is a testing thunk to the C function.
 *     """
 *     as_bin_sha = PyBytes_FromStringAndSize(NULL, 20)             # <<<<<<<<<<<<<<
 *     if _key_to_sha1(key, PyBytes_AS_STRING(as_bin_sha)):
 *         return as_bin_sha
*/
  __pyx_t_1 = PyBytes_FromStringAndSize(NULL, 20); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_as_bin_sha = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":371
 *     """
 *     as_bin_sha = PyBytes_FromStringAndSize(NULL, 20)
 *     if _key_to_sha1(key, PyBytes_AS_STRING(as_bin_sha)):             # <<<<<<<<<<<<<<
 *         return as_bin_sha
 *     return None
*/
  __pyx_t_2 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__key_to_sha1(__pyx_v_key, PyBytes_AS_STRING(__pyx_v_as_bin_sha)); if (unlikely(__pyx_t_2 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 371, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":372
 *     as_bin_sha = PyBytes_FromStringAndSize(NULL, 20)
 *     if _key_to_sha1(key, PyBytes_AS_STRING(as_bin_sha)):
 *         return as_bin_sha             # <<<<<<<<<<<<<<
 *     return None
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_as_bin_sha);
    __pyx_r = __pyx_v_as_bin_sha;
    goto __pyx_L0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":371
 *     """
 *     as_bin_sha = PyBytes_FromStringAndSize(NULL, 20)
 *     if _key_to_sha1(key, PyBytes_AS_STRING(as_bin_sha)):             # <<<<<<<<<<<<<<
 *         return as_bin_sha
 *     return None
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":373
 *     if _key_to_sha1(key, PyBytes_AS_STRING(as_bin_sha)):
 *         return as_bin_sha
 *     return None             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":365
 * 
 * 
 * def _py_key_to_sha1(key):             # <<<<<<<<<<<<<<
 *     """Map a key to a simple sha1 string.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._py_key_to_sha1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_as_bin_sha);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":376
 * 
 * 
 * cdef tuple _sha1_to_key(char *sha1):             # <<<<<<<<<<<<<<
 *     """Compute a ('sha1:abcd',) key for a given sha1."""
 *     cdef tuple key
*/

static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_key(char *__pyx_v_sha1) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_hexxed = 0;
  char *__pyx_v_c_buf;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_sha1_to_key", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":381
 *     cdef object hexxed
 *     cdef char *c_buf
 *     hexxed = PyBytes_FromStringAndSize(NULL, 45)             # <<<<<<<<<<<<<<
 *     c_buf = PyBytes_AS_STRING(hexxed)
 *     memcpy(c_buf, b'sha1:', 5)
*/
  __pyx_t_1 = PyBytes_FromStringAndSize(NULL, 45); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_hexxed = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":382
 *     cdef char *c_buf
 *     hexxed = PyBytes_FromStringAndSize(NULL, 45)
 *     c_buf = PyBytes_AS_STRING(hexxed)             # <<<<<<<<<<<<<<
 *     memcpy(c_buf, b'sha1:', 5)
 *     _hexlify_sha1(sha1, c_buf+5)
*/
  __pyx_v_c_buf = PyBytes_AS_STRING(__pyx_v_hexxed);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":383
 *     hexxed = PyBytes_FromStringAndSize(NULL, 45)
 *     c_buf = PyBytes_AS_STRING(hexxed)
 *     memcpy(c_buf, b'sha1:', 5)             # <<<<<<<<<<<<<<
 *     _hexlify_sha1(sha1, c_buf+5)
 *     key = PyTuple_New(1)
*/
  (void)(memcpy(__pyx_v_c_buf, __pyx_k_sha1, 5));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":384
 *     c_buf = PyBytes_AS_STRING(hexxed)
 *     memcpy(c_buf, b'sha1:', 5)
 *     _hexlify_sha1(sha1, c_buf+5)             # <<<<<<<<<<<<<<
 *     key = PyTuple_New(1)
 *     Py_INCREF(hexxed)
*/
  __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__hexlify_sha1(__pyx_v_sha1, (__pyx_v_c_buf + 5)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 384, __pyx_L1_error)

  /* "breezy/bzr/_btree_serializer_pyx.pyx":385
 *     memcpy(c_buf, b'sha1:', 5)
 *     _hexlify_sha1(sha1, c_buf+5)
 *     key = PyTuple_New(1)             # <<<<<<<<<<<<<<
 *     Py_INCREF(hexxed)
 *     PyTuple_SET_ITEM(key, 0, hexxed)
*/
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_key = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":386
 *     _hexlify_sha1(sha1, c_buf+5)
 *     key = PyTuple_New(1)
 *     Py_INCREF(hexxed)             # <<<<<<<<<<<<<<
 *     PyTuple_SET_ITEM(key, 0, hexxed)
 *     # This is a bit expensive. To parse 120 keys takes 48us, to return them all
*/
  Py_INCREF(__pyx_v_hexxed);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":387
 *     key = PyTuple_New(1)
 *     Py_INCREF(hexxed)
 *     PyTuple_SET_ITEM(key, 0, hexxed)             # <<<<<<<<<<<<<<
 *     # This is a bit expensive. To parse 120 keys takes 48us, to return them all
 *     # can be done in 66.6us (so 18.6us to build them all).
*/
  PyTuple_SET_ITEM(__pyx_v_key, 0, __pyx_v_hexxed);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":397
 *     # We *could* hang the PyObject form off of the gc_chk_sha1_record for ones
 *     # that we have deserialized. Something to think about, at least.
 *     return key             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_key);
  __pyx_r = __pyx_v_key;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":376
 * 
 * 
 * cdef tuple _sha1_to_key(char *sha1):             # <<<<<<<<<<<<<<
 *     """Compute a ('sha1:abcd',) key for a given sha1."""
 *     cdef tuple key
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._sha1_to_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_hexxed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":400
 * 
 * 
 * def _py_sha1_to_key(sha1_bin):             # <<<<<<<<<<<<<<
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyBytes_CheckExact(sha1_bin) or PyBytes_GET_SIZE(sha1_bin) != 20:
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_9_py_sha1_to_key(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_8_py_sha1_to_key, "Test thunk to check the sha1 mapping.");
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_9_py_sha1_to_key = {"_py_sha1_to_key", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_9_py_sha1_to_key, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_8_py_sha1_to_key};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_9_py_sha1_to_key(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_sha1_bin = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_py_sha1_to_key (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_sha1_bin,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 400, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 400, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_py_sha1_to_key", 0) < (0)) __PYX_ERR(0, 400, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_py_sha1_to_key", 1, 1, 1, i); __PYX_ERR(0, 400, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 400, __pyx_L3_error)
    }
    __pyx_v_sha1_bin = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_py_sha1_to_key", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 400, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._py_sha1_to_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_8_py_sha1_to_key(__pyx_self, __pyx_v_sha1_bin);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_8_py_sha1_to_key(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_sha1_bin) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_py_sha1_to_key", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":402
 * def _py_sha1_to_key(sha1_bin):
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyBytes_CheckExact(sha1_bin) or PyBytes_GET_SIZE(sha1_bin) != 20:             # <<<<<<<<<<<<<<
 *         raise ValueError('sha1_bin must be a str of exactly 20 bytes')
 *     return _sha1_to_key(PyBytes_AS_STRING(sha1_bin))
*/
  __pyx_t_2 = (!PyBytes_CheckExact(__pyx_v_sha1_bin));
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (PyBytes_GET_SIZE(__pyx_v_sha1_bin) != 20);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":403
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyBytes_CheckExact(sha1_bin) or PyBytes_GET_SIZE(sha1_bin) != 20:
 *         raise ValueError('sha1_bin must be a str of exactly 20 bytes')             # <<<<<<<<<<<<<<
 *     return _sha1_to_key(PyBytes_AS_STRING(sha1_bin))
 * 
*/
    __pyx_t_4 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_sha1_bin_must_be_a_str_of_exactl};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 403, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 403, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":402
 * def _py_sha1_to_key(sha1_bin):
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyBytes_CheckExact(sha1_bin) or PyBytes_GET_SIZE(sha1_bin) != 20:             # <<<<<<<<<<<<<<
 *         raise ValueError('sha1_bin must be a str of exactly 20 bytes')
 *     return _sha1_to_key(PyBytes_AS_STRING(sha1_bin))
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":404
 *     if not PyBytes_CheckExact(sha1_bin) or PyBytes_GET_SIZE(sha1_bin) != 20:
 *         raise ValueError('sha1_bin must be a str of exactly 20 bytes')
 *     return _sha1_to_key(PyBytes_AS_STRING(sha1_bin))             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_key(PyBytes_AS_STRING(__pyx_v_sha1_bin)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":400
 * 
 * 
 * def _py_sha1_to_key(sha1_bin):             # <<<<<<<<<<<<<<
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyBytes_CheckExact(sha1_bin) or PyBytes_GET_SIZE(sha1_bin) != 20:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._py_sha1_to_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":407
 * 
 * 
 * cdef unsigned int _sha1_to_uint(char *sha1): # cannot_raise             # <<<<<<<<<<<<<<
 *     cdef unsigned int val
 *     # Must be in MSB, because that is how the content is sorted
*/

static unsigned int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_uint(char *__pyx_v_sha1) {
  unsigned int __pyx_v_val;
  unsigned int __pyx_r;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":413
 *            | ((<unsigned int>(sha1[1]) & 0xff) << 16)
 *            | ((<unsigned int>(sha1[2]) & 0xff) << 8)
 *            | ((<unsigned int>(sha1[3]) & 0xff) << 0))             # <<<<<<<<<<<<<<
 *     return val
 * 
*/
  __pyx_v_val = (((((((unsigned int)(__pyx_v_sha1[0])) & 0xff) << 24) | ((((unsigned int)(__pyx_v_sha1[1])) & 0xff) << 16)) | ((((unsigned int)(__pyx_v_sha1[2])) & 0xff) << 8)) | ((((unsigned int)(__pyx_v_sha1[3])) & 0xff) << 0));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":414
 *            | ((<unsigned int>(sha1[2]) & 0xff) << 8)
 *            | ((<unsigned int>(sha1[3]) & 0xff) << 0))
 *     return val             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = __pyx_v_val;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":407
 * 
 * 
 * cdef unsigned int _sha1_to_uint(char *sha1): # cannot_raise             # <<<<<<<<<<<<<<
 *     cdef unsigned int val
 *     # Must be in MSB, because that is how the content is sorted
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":417
 * 
 * 
 * cdef _format_record(gc_chk_sha1_record *record):             # <<<<<<<<<<<<<<
 *     # This is inefficient to go from a logical state back to a bytes object,
 *     # but it makes things work a bit better internally for now.
*/

static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__format_record(__pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record) {
  PyObject *__pyx_v_block_offset_str = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_format_record", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":420
 *     # This is inefficient to go from a logical state back to a bytes object,
 *     # but it makes things work a bit better internally for now.
 *     if record.block_offset >= 0xFFFFFFFF:             # <<<<<<<<<<<<<<
 *         # Could use %llu which was added to Python 2.7 but it oddly is missing
 *         # from the Python 3 equivalent functions, so hack still needed. :(
*/
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(__pyx_v_record->block_offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_mstate_global->__pyx_int_4294967295, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":423
 *         # Could use %llu which was added to Python 2.7 but it oddly is missing
 *         # from the Python 3 equivalent functions, so hack still needed. :(
 *         block_offset_str = b'%d' % record.block_offset             # <<<<<<<<<<<<<<
 *         value = PyBytes_FromFormat(
 *             '%s %u %u %u', PyBytes_AS_STRING(block_offset_str),
*/
    __pyx_t_2 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(__pyx_v_record->block_offset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyNumber_Remainder(__pyx_mstate_global->__pyx_kp_b_d, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_block_offset_str = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":424
 *         # from the Python 3 equivalent functions, so hack still needed. :(
 *         block_offset_str = b'%d' % record.block_offset
 *         value = PyBytes_FromFormat(             # <<<<<<<<<<<<<<
 *             '%s %u %u %u', PyBytes_AS_STRING(block_offset_str),
 *             record.block_length, record.record_start, record.record_end)
*/
    __pyx_t_1 = PyBytes_FromFormat(((char *)"%s %u %u %u"), PyBytes_AS_STRING(__pyx_v_block_offset_str), __pyx_v_record->block_length, __pyx_v_record->record_start, __pyx_v_record->record_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_value = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":420
 *     # This is inefficient to go from a logical state back to a bytes object,
 *     # but it makes things work a bit better internally for now.
 *     if record.block_offset >= 0xFFFFFFFF:             # <<<<<<<<<<<<<<
 *         # Could use %llu which was added to Python 2.7 but it oddly is missing
 *         # from the Python 3 equivalent functions, so hack still needed. :(
*/
    goto __pyx_L3;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":428
 *             record.block_length, record.record_start, record.record_end)
 *     else:
 *         value = PyBytes_FromFormat(             # <<<<<<<<<<<<<<
 *             '%lu %u %u %u', <unsigned long>record.block_offset,
 *             record.block_length, record.record_start, record.record_end)
*/
  /*else*/ {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":430
 *         value = PyBytes_FromFormat(
 *             '%lu %u %u %u', <unsigned long>record.block_offset,
 *             record.block_length, record.record_start, record.record_end)             # <<<<<<<<<<<<<<
 *     return value
 * 
*/
    __pyx_t_1 = PyBytes_FromFormat(((char *)"%lu %u %u %u"), ((unsigned long)__pyx_v_record->block_offset), __pyx_v_record->block_length, __pyx_v_record->record_start, __pyx_v_record->record_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_value = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":431
 *             '%lu %u %u %u', <unsigned long>record.block_offset,
 *             record.block_length, record.record_start, record.record_end)
 *     return value             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":417
 * 
 * 
 * cdef _format_record(gc_chk_sha1_record *record):             # <<<<<<<<<<<<<<
 *     # This is inefficient to go from a logical state back to a bytes object,
 *     # but it makes things work a bit better internally for now.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._format_record", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_block_offset_str);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":452
 *     cdef unsigned char offsets[257]
 * 
 *     def __sizeof__(self):             # <<<<<<<<<<<<<<
 *         return (
 *             sizeof(GCCHKSHA1LeafNode) +
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_1__sizeof__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_1__sizeof__ = {"__sizeof__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_1__sizeof__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_1__sizeof__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__sizeof__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__sizeof__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__sizeof__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode___sizeof__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode___sizeof__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__sizeof__", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":453
 * 
 *     def __sizeof__(self):
 *         return (             # <<<<<<<<<<<<<<
 *             sizeof(GCCHKSHA1LeafNode) +
 *             sizeof(gc_chk_sha1_record) * self.num_records)
*/
  __Pyx_XDECREF(__pyx_r);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":454
 *     def __sizeof__(self):
 *         return (
 *             sizeof(GCCHKSHA1LeafNode) +             # <<<<<<<<<<<<<<
 *             sizeof(gc_chk_sha1_record) * self.num_records)
 * 
*/
  __pyx_t_1 = __Pyx_PyLong_FromSize_t(((sizeof(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode)) + ((sizeof(__pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record)) * __pyx_v_self->num_records))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":452
 *     cdef unsigned char offsets[257]
 * 
 *     def __sizeof__(self):             # <<<<<<<<<<<<<<
 *         return (
 *             sizeof(GCCHKSHA1LeafNode) +
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.__sizeof__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":457
 *             sizeof(gc_chk_sha1_record) * self.num_records)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.records != NULL:
 *             PyMem_Free(self.records)
*/

/* Python wrapper */
static void __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_2__dealloc__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_2__dealloc__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  int __pyx_t_1;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":458
 * 
 *     def __dealloc__(self):
 *         if self.records != NULL:             # <<<<<<<<<<<<<<
 *             PyMem_Free(self.records)
 *             self.records = NULL
*/
  __pyx_t_1 = (__pyx_v_self->records != NULL);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":459
 *     def __dealloc__(self):
 *         if self.records != NULL:
 *             PyMem_Free(self.records)             # <<<<<<<<<<<<<<
 *             self.records = NULL
 * 
*/
    PyMem_Free(__pyx_v_self->records);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":460
 *         if self.records != NULL:
 *             PyMem_Free(self.records)
 *             self.records = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, bytes):
*/
    __pyx_v_self->records = NULL;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":458
 * 
 *     def __dealloc__(self):
 *         if self.records != NULL:             # <<<<<<<<<<<<<<
 *             PyMem_Free(self.records)
 *             self.records = NULL
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":457
 *             sizeof(gc_chk_sha1_record) * self.num_records)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.records != NULL:
 *             PyMem_Free(self.records)
*/

  /* function exit code */
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":462
 *             self.records = NULL
 * 
 *     def __init__(self, bytes):             # <<<<<<<<<<<<<<
 *         self._parse_bytes(bytes)
 *         self.last_key = None
*/

/* Python wrapper */
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_5__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_5__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_bytes = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_bytes,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 462, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 462, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 462, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, i); __PYX_ERR(0, 462, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 462, __pyx_L3_error)
    }
    __pyx_v_bytes = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 462, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_4__init__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), __pyx_v_bytes);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_4__init__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_bytes) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":463
 * 
 *     def __init__(self, bytes):
 *         self._parse_bytes(bytes)             # <<<<<<<<<<<<<<
 *         self.last_key = None
 *         self.last_record = NULL
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_parse_bytes(__pyx_v_self, __pyx_v_bytes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":464
 *     def __init__(self, bytes):
 *         self._parse_bytes(bytes)
 *         self.last_key = None             # <<<<<<<<<<<<<<
 *         self.last_record = NULL
 * 
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->last_key);
  __Pyx_DECREF(__pyx_v_self->last_key);
  __pyx_v_self->last_key = Py_None;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":465
 *         self._parse_bytes(bytes)
 *         self.last_key = None
 *         self.last_record = NULL             # <<<<<<<<<<<<<<
 * 
 *     property min_key:
*/
  __pyx_v_self->last_record = NULL;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":462
 *             self.records = NULL
 * 
 *     def __init__(self, bytes):             # <<<<<<<<<<<<<<
 *         self._parse_bytes(bytes)
 *         self.last_key = None
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":468
 * 
 *     property min_key:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[0].sha1)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7min_key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7min_key_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7min_key___get__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7min_key___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":469
 *     property min_key:
 *         def __get__(self):
 *             if self.num_records > 0:             # <<<<<<<<<<<<<<
 *                 return _sha1_to_key(self.records[0].sha1)
 *             return None
*/
  __pyx_t_1 = (__pyx_v_self->num_records > 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":470
 *         def __get__(self):
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[0].sha1)             # <<<<<<<<<<<<<<
 *             return None
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_key((__pyx_v_self->records[0]).sha1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":469
 *     property min_key:
 *         def __get__(self):
 *             if self.num_records > 0:             # <<<<<<<<<<<<<<
 *                 return _sha1_to_key(self.records[0].sha1)
 *             return None
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":471
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[0].sha1)
 *             return None             # <<<<<<<<<<<<<<
 * 
 *     property max_key:
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":468
 * 
 *     property min_key:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[0].sha1)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.min_key.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":474
 * 
 *     property max_key:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[self.num_records-1].sha1)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7max_key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7max_key_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7max_key___get__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7max_key___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":475
 *     property max_key:
 *         def __get__(self):
 *             if self.num_records > 0:             # <<<<<<<<<<<<<<
 *                 return _sha1_to_key(self.records[self.num_records-1].sha1)
 *             return None
*/
  __pyx_t_1 = (__pyx_v_self->num_records > 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":476
 *         def __get__(self):
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[self.num_records-1].sha1)             # <<<<<<<<<<<<<<
 *             return None
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_key((__pyx_v_self->records[(__pyx_v_self->num_records - 1)]).sha1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":475
 *     property max_key:
 *         def __get__(self):
 *             if self.num_records > 0:             # <<<<<<<<<<<<<<
 *                 return _sha1_to_key(self.records[self.num_records-1].sha1)
 *             return None
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":477
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[self.num_records-1].sha1)
 *             return None             # <<<<<<<<<<<<<<
 * 
 *     cdef tuple _record_to_value_and_refs(self,
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":474
 * 
 *     property max_key:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[self.num_records-1].sha1)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.max_key.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":479
 *             return None
 * 
 *     cdef tuple _record_to_value_and_refs(self,             # <<<<<<<<<<<<<<
 *                                                gc_chk_sha1_record *record):
 *         """Extract the refs and value part of this record."""
*/

static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__record_to_value_and_refs(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record) {
  PyObject *__pyx_v_value_and_refs = 0;
  PyObject *__pyx_v_empty = 0;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_record_to_value_and_refs", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":484
 *         cdef tuple value_and_refs
 *         cdef tuple empty
 *         value_and_refs = PyTuple_New(2)             # <<<<<<<<<<<<<<
 *         value = _format_record(record)
 *         Py_INCREF(value)
*/
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_value_and_refs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":485
 *         cdef tuple empty
 *         value_and_refs = PyTuple_New(2)
 *         value = _format_record(record)             # <<<<<<<<<<<<<<
 *         Py_INCREF(value)
 *         PyTuple_SET_ITEM(value_and_refs, 0, value)
*/
  __pyx_t_1 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__format_record(__pyx_v_record); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_value = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":486
 *         value_and_refs = PyTuple_New(2)
 *         value = _format_record(record)
 *         Py_INCREF(value)             # <<<<<<<<<<<<<<
 *         PyTuple_SET_ITEM(value_and_refs, 0, value)
 *         # Always empty refs
*/
  Py_INCREF(__pyx_v_value);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":487
 *         value = _format_record(record)
 *         Py_INCREF(value)
 *         PyTuple_SET_ITEM(value_and_refs, 0, value)             # <<<<<<<<<<<<<<
 *         # Always empty refs
 *         empty = PyTuple_New(0)
*/
  PyTuple_SET_ITEM(__pyx_v_value_and_refs, 0, __pyx_v_value);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":489
 *         PyTuple_SET_ITEM(value_and_refs, 0, value)
 *         # Always empty refs
 *         empty = PyTuple_New(0)             # <<<<<<<<<<<<<<
 *         Py_INCREF(empty)
 *         PyTuple_SET_ITEM(value_and_refs, 1, empty)
*/
  __pyx_t_1 = PyTuple_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_empty = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":490
 *         # Always empty refs
 *         empty = PyTuple_New(0)
 *         Py_INCREF(empty)             # <<<<<<<<<<<<<<
 *         PyTuple_SET_ITEM(value_and_refs, 1, empty)
 *         return value_and_refs
*/
  Py_INCREF(__pyx_v_empty);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":491
 *         empty = PyTuple_New(0)
 *         Py_INCREF(empty)
 *         PyTuple_SET_ITEM(value_and_refs, 1, empty)             # <<<<<<<<<<<<<<
 *         return value_and_refs
 * 
*/
  PyTuple_SET_ITEM(__pyx_v_value_and_refs, 1, __pyx_v_empty);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":492
 *         Py_INCREF(empty)
 *         PyTuple_SET_ITEM(value_and_refs, 1, empty)
 *         return value_and_refs             # <<<<<<<<<<<<<<
 * 
 *     cdef tuple _record_to_item(self, gc_chk_sha1_record *record):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value_and_refs);
  __pyx_r = __pyx_v_value_and_refs;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":479
 *             return None
 * 
 *     cdef tuple _record_to_value_and_refs(self,             # <<<<<<<<<<<<<<
 *                                                gc_chk_sha1_record *record):
 *         """Extract the refs and value part of this record."""
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode._record_to_value_and_refs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value_and_refs);
  __Pyx_XDECREF(__pyx_v_empty);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":494
 *         return value_and_refs
 * 
 *     cdef tuple _record_to_item(self, gc_chk_sha1_record *record):             # <<<<<<<<<<<<<<
 *         """Turn a given record back into a fully fledged item.
 *         """
*/

static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__record_to_item(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record) {
  PyObject *__pyx_v_item = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_value_and_refs = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_record_to_item", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":501
 *         cdef tuple value_and_refs
 *         cdef object value
 *         key = _sha1_to_key(record.sha1)             # <<<<<<<<<<<<<<
 *         item = PyTuple_New(2)
 *         Py_INCREF(key)
*/
  __pyx_t_1 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_key(__pyx_v_record->sha1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_key = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":502
 *         cdef object value
 *         key = _sha1_to_key(record.sha1)
 *         item = PyTuple_New(2)             # <<<<<<<<<<<<<<
 *         Py_INCREF(key)
 *         PyTuple_SET_ITEM(item, 0, key)
*/
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_item = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":503
 *         key = _sha1_to_key(record.sha1)
 *         item = PyTuple_New(2)
 *         Py_INCREF(key)             # <<<<<<<<<<<<<<
 *         PyTuple_SET_ITEM(item, 0, key)
 *         value_and_refs = self._record_to_value_and_refs(record)
*/
  Py_INCREF(__pyx_v_key);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":504
 *         item = PyTuple_New(2)
 *         Py_INCREF(key)
 *         PyTuple_SET_ITEM(item, 0, key)             # <<<<<<<<<<<<<<
 *         value_and_refs = self._record_to_value_and_refs(record)
 *         Py_INCREF(value_and_refs)
*/
  PyTuple_SET_ITEM(__pyx_v_item, 0, __pyx_v_key);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":505
 *         Py_INCREF(key)
 *         PyTuple_SET_ITEM(item, 0, key)
 *         value_and_refs = self._record_to_value_and_refs(record)             # <<<<<<<<<<<<<<
 *         Py_INCREF(value_and_refs)
 *         PyTuple_SET_ITEM(item, 1, value_and_refs)
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_record_to_value_and_refs(__pyx_v_self, __pyx_v_record); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_value_and_refs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":506
 *         PyTuple_SET_ITEM(item, 0, key)
 *         value_and_refs = self._record_to_value_and_refs(record)
 *         Py_INCREF(value_and_refs)             # <<<<<<<<<<<<<<
 *         PyTuple_SET_ITEM(item, 1, value_and_refs)
 *         return item
*/
  Py_INCREF(__pyx_v_value_and_refs);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":507
 *         value_and_refs = self._record_to_value_and_refs(record)
 *         Py_INCREF(value_and_refs)
 *         PyTuple_SET_ITEM(item, 1, value_and_refs)             # <<<<<<<<<<<<<<
 *         return item
 * 
*/
  PyTuple_SET_ITEM(__pyx_v_item, 1, __pyx_v_value_and_refs);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":508
 *         Py_INCREF(value_and_refs)
 *         PyTuple_SET_ITEM(item, 1, value_and_refs)
 *         return item             # <<<<<<<<<<<<<<
 * 
 *     cdef gc_chk_sha1_record* _lookup_record(self, char *sha1) except? NULL:
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_item);
  __pyx_r = __pyx_v_item;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":494
 *         return value_and_refs
 * 
 *     cdef tuple _record_to_item(self, gc_chk_sha1_record *record):             # <<<<<<<<<<<<<<
 *         """Turn a given record back into a fully fledged item.
 *         """
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode._record_to_item", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value_and_refs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":510
 *         return item
 * 
 *     cdef gc_chk_sha1_record* _lookup_record(self, char *sha1) except? NULL:             # <<<<<<<<<<<<<<
 *         """Find a gc_chk_sha1_record that matches the sha1 supplied."""
 *         cdef int lo, hi, mid, the_cmp
*/

static __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__lookup_record(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_sha1) {
  int __pyx_v_lo;
  int __pyx_v_hi;
  int __pyx_v_mid;
  int __pyx_v_the_cmp;
  int __pyx_v_offset;
  CYTHON_UNUSED long __pyx_v_local_n_cmp;
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":524
 *         # the offset array dropped us from 23us to 20us and 156 comparisions
 *         # (1.3/key)
 *         offset = self._offset_for_sha1(sha1)             # <<<<<<<<<<<<<<
 *         lo = self.offsets[offset]
 *         hi = self.offsets[offset+1]
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_offset_for_sha1(__pyx_v_self, __pyx_v_sha1); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 524, __pyx_L1_error)
  __pyx_v_offset = __pyx_t_1;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":525
 *         # (1.3/key)
 *         offset = self._offset_for_sha1(sha1)
 *         lo = self.offsets[offset]             # <<<<<<<<<<<<<<
 *         hi = self.offsets[offset+1]
 *         if hi == 255:
*/
  __pyx_v_lo = (__pyx_v_self->offsets[__pyx_v_offset]);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":526
 *         offset = self._offset_for_sha1(sha1)
 *         lo = self.offsets[offset]
 *         hi = self.offsets[offset+1]             # <<<<<<<<<<<<<<
 *         if hi == 255:
 *             # if hi == 255 that means we potentially ran off the end of the
*/
  __pyx_v_hi = (__pyx_v_self->offsets[(__pyx_v_offset + 1)]);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":527
 *         lo = self.offsets[offset]
 *         hi = self.offsets[offset+1]
 *         if hi == 255:             # <<<<<<<<<<<<<<
 *             # if hi == 255 that means we potentially ran off the end of the
 *             # list, so push it up to num_records
*/
  __pyx_t_2 = (__pyx_v_hi == 0xFF);
  if (__pyx_t_2) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":532
 *             # note that if 'lo' == 255, that is ok, because we can start
 *             # searching from that part of the list.
 *             hi = self.num_records             # <<<<<<<<<<<<<<
 *         local_n_cmp = 0
 *         while lo < hi:
*/
    __pyx_t_1 = __pyx_v_self->num_records;
    __pyx_v_hi = __pyx_t_1;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":527
 *         lo = self.offsets[offset]
 *         hi = self.offsets[offset+1]
 *         if hi == 255:             # <<<<<<<<<<<<<<
 *             # if hi == 255 that means we potentially ran off the end of the
 *             # list, so push it up to num_records
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":533
 *             # searching from that part of the list.
 *             hi = self.num_records
 *         local_n_cmp = 0             # <<<<<<<<<<<<<<
 *         while lo < hi:
 *             mid = (lo + hi) // 2
*/
  __pyx_v_local_n_cmp = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":534
 *             hi = self.num_records
 *         local_n_cmp = 0
 *         while lo < hi:             # <<<<<<<<<<<<<<
 *             mid = (lo + hi) // 2
 *             the_cmp = memcmp(self.records[mid].sha1, sha1, 20)
*/
  while (1) {
    __pyx_t_2 = (__pyx_v_lo < __pyx_v_hi);
    if (!__pyx_t_2) break;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":535
 *         local_n_cmp = 0
 *         while lo < hi:
 *             mid = (lo + hi) // 2             # <<<<<<<<<<<<<<
 *             the_cmp = memcmp(self.records[mid].sha1, sha1, 20)
 *             if the_cmp == 0:
*/
    __pyx_v_mid = __Pyx_div_long((__pyx_v_lo + __pyx_v_hi), 2, 1);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":536
 *         while lo < hi:
 *             mid = (lo + hi) // 2
 *             the_cmp = memcmp(self.records[mid].sha1, sha1, 20)             # <<<<<<<<<<<<<<
 *             if the_cmp == 0:
 *                 return &self.records[mid]
*/
    __pyx_v_the_cmp = memcmp((__pyx_v_self->records[__pyx_v_mid]).sha1, __pyx_v_sha1, 20);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":537
 *             mid = (lo + hi) // 2
 *             the_cmp = memcmp(self.records[mid].sha1, sha1, 20)
 *             if the_cmp == 0:             # <<<<<<<<<<<<<<
 *                 return &self.records[mid]
 *             elif the_cmp < 0:
*/
    __pyx_t_2 = (__pyx_v_the_cmp == 0);
    if (__pyx_t_2) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":538
 *             the_cmp = memcmp(self.records[mid].sha1, sha1, 20)
 *             if the_cmp == 0:
 *                 return &self.records[mid]             # <<<<<<<<<<<<<<
 *             elif the_cmp < 0:
 *                 lo = mid + 1
*/
      __pyx_r = (&(__pyx_v_self->records[__pyx_v_mid]));
      goto __pyx_L0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":537
 *             mid = (lo + hi) // 2
 *             the_cmp = memcmp(self.records[mid].sha1, sha1, 20)
 *             if the_cmp == 0:             # <<<<<<<<<<<<<<
 *                 return &self.records[mid]
 *             elif the_cmp < 0:
*/
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":539
 *             if the_cmp == 0:
 *                 return &self.records[mid]
 *             elif the_cmp < 0:             # <<<<<<<<<<<<<<
 *                 lo = mid + 1
 *             else:
*/
    __pyx_t_2 = (__pyx_v_the_cmp < 0);
    if (__pyx_t_2) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":540
 *                 return &self.records[mid]
 *             elif the_cmp < 0:
 *                 lo = mid + 1             # <<<<<<<<<<<<<<
 *             else:
 *                 hi = mid
*/
      __pyx_v_lo = (__pyx_v_mid + 1);

      /* "breezy/bzr/_btree_serializer_pyx.pyx":539
 *             if the_cmp == 0:
 *                 return &self.records[mid]
 *             elif the_cmp < 0:             # <<<<<<<<<<<<<<
 *                 lo = mid + 1
 *             else:
*/
      goto __pyx_L6;
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":542
 *                 lo = mid + 1
 *             else:
 *                 hi = mid             # <<<<<<<<<<<<<<
 *         return NULL
 * 
*/
    /*else*/ {
      __pyx_v_hi = __pyx_v_mid;
    }
    __pyx_L6:;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":543
 *             else:
 *                 hi = mid
 *         return NULL             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, key):
*/
  __pyx_r = NULL;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":510
 *         return item
 * 
 *     cdef gc_chk_sha1_record* _lookup_record(self, char *sha1) except? NULL:             # <<<<<<<<<<<<<<
 *         """Find a gc_chk_sha1_record that matches the sha1 supplied."""
 *         cdef int lo, hi, mid, the_cmp
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode._lookup_record", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":545
 *         return NULL
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
*/

/* Python wrapper */
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_6__contains__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_6__contains__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_key) {
  char __pyx_v_sha1[20];
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":548
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
 *         if _key_to_sha1(key, sha1):             # <<<<<<<<<<<<<<
 *             # If it isn't a sha1 key, then it won't be in this leaf node
 *             record = self._lookup_record(sha1)
*/
  __pyx_t_1 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__key_to_sha1(__pyx_v_key, __pyx_v_sha1); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 548, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":550
 *         if _key_to_sha1(key, sha1):
 *             # If it isn't a sha1 key, then it won't be in this leaf node
 *             record = self._lookup_record(sha1)             # <<<<<<<<<<<<<<
 *             if record != NULL:
 *                 self.last_key = key
*/
    __pyx_t_3 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_lookup_record(__pyx_v_self, __pyx_v_sha1); if (unlikely(__pyx_t_3 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 550, __pyx_L1_error)
    __pyx_v_record = __pyx_t_3;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":551
 *             # If it isn't a sha1 key, then it won't be in this leaf node
 *             record = self._lookup_record(sha1)
 *             if record != NULL:             # <<<<<<<<<<<<<<
 *                 self.last_key = key
 *                 self.last_record = record
*/
    __pyx_t_2 = (__pyx_v_record != NULL);
    if (__pyx_t_2) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":552
 *             record = self._lookup_record(sha1)
 *             if record != NULL:
 *                 self.last_key = key             # <<<<<<<<<<<<<<
 *                 self.last_record = record
 *                 return True
*/
      __Pyx_INCREF(__pyx_v_key);
      __Pyx_GIVEREF(__pyx_v_key);
      __Pyx_GOTREF(__pyx_v_self->last_key);
      __Pyx_DECREF(__pyx_v_self->last_key);
      __pyx_v_self->last_key = __pyx_v_key;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":553
 *             if record != NULL:
 *                 self.last_key = key
 *                 self.last_record = record             # <<<<<<<<<<<<<<
 *                 return True
 *         return False
*/
      __pyx_v_self->last_record = __pyx_v_record;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":554
 *                 self.last_key = key
 *                 self.last_record = record
 *                 return True             # <<<<<<<<<<<<<<
 *         return False
 * 
*/
      __pyx_r = 1;
      goto __pyx_L0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":551
 *             # If it isn't a sha1 key, then it won't be in this leaf node
 *             record = self._lookup_record(sha1)
 *             if record != NULL:             # <<<<<<<<<<<<<<
 *                 self.last_key = key
 *                 self.last_record = record
*/
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":548
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
 *         if _key_to_sha1(key, sha1):             # <<<<<<<<<<<<<<
 *             # If it isn't a sha1 key, then it won't be in this leaf node
 *             record = self._lookup_record(sha1)
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":555
 *                 self.last_record = record
 *                 return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, key):
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":545
 *         return NULL
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":557
 *         return False
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8__getitem__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8__getitem__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_key) {
  char __pyx_v_sha1[20];
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8[3];
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":560
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
 *         record = NULL             # <<<<<<<<<<<<<<
 *         if self.last_record != NULL and key is self.last_key:
 *             record = self.last_record
*/
  __pyx_v_record = NULL;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":561
 *         cdef gc_chk_sha1_record *record
 *         record = NULL
 *         if self.last_record != NULL and key is self.last_key:             # <<<<<<<<<<<<<<
 *             record = self.last_record
 *         elif _key_to_sha1(key, sha1):
*/
  __pyx_t_2 = (__pyx_v_self->last_record != NULL);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_key == __pyx_v_self->last_key);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":562
 *         record = NULL
 *         if self.last_record != NULL and key is self.last_key:
 *             record = self.last_record             # <<<<<<<<<<<<<<
 *         elif _key_to_sha1(key, sha1):
 *             record = self._lookup_record(sha1)
*/
    __pyx_t_3 = __pyx_v_self->last_record;
    __pyx_v_record = __pyx_t_3;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":561
 *         cdef gc_chk_sha1_record *record
 *         record = NULL
 *         if self.last_record != NULL and key is self.last_key:             # <<<<<<<<<<<<<<
 *             record = self.last_record
 *         elif _key_to_sha1(key, sha1):
*/
    goto __pyx_L3;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":563
 *         if self.last_record != NULL and key is self.last_key:
 *             record = self.last_record
 *         elif _key_to_sha1(key, sha1):             # <<<<<<<<<<<<<<
 *             record = self._lookup_record(sha1)
 *         if record == NULL:
*/
  __pyx_t_4 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__key_to_sha1(__pyx_v_key, __pyx_v_sha1); if (unlikely(__pyx_t_4 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 563, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_4 != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":564
 *             record = self.last_record
 *         elif _key_to_sha1(key, sha1):
 *             record = self._lookup_record(sha1)             # <<<<<<<<<<<<<<
 *         if record == NULL:
 *             raise KeyError('key %r is not present' % (key,))
*/
    __pyx_t_3 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_lookup_record(__pyx_v_self, __pyx_v_sha1); if (unlikely(__pyx_t_3 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 564, __pyx_L1_error)
    __pyx_v_record = __pyx_t_3;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":563
 *         if self.last_record != NULL and key is self.last_key:
 *             record = self.last_record
 *         elif _key_to_sha1(key, sha1):             # <<<<<<<<<<<<<<
 *             record = self._lookup_record(sha1)
 *         if record == NULL:
*/
  }
  __pyx_L3:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":565
 *         elif _key_to_sha1(key, sha1):
 *             record = self._lookup_record(sha1)
 *         if record == NULL:             # <<<<<<<<<<<<<<
 *             raise KeyError('key %r is not present' % (key,))
 *         return self._record_to_value_and_refs(record)
*/
  __pyx_t_1 = (__pyx_v_record == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":566
 *             record = self._lookup_record(sha1)
 *         if record == NULL:
 *             raise KeyError('key %r is not present' % (key,))             # <<<<<<<<<<<<<<
 *         return self._record_to_value_and_refs(record)
 * 
*/
    __pyx_t_6 = NULL;
    __pyx_t_7 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_key), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8[0] = __pyx_mstate_global->__pyx_kp_u_key_2;
    __pyx_t_8[1] = __pyx_t_7;
    __pyx_t_8[2] = __pyx_mstate_global->__pyx_kp_u_is_not_present;
    __pyx_t_9 = __Pyx_PyUnicode_Join(__pyx_t_8, 3, 4 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7) + 15, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7));
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_10 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_9};
      __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_KeyError)), __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 566, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 566, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":565
 *         elif _key_to_sha1(key, sha1):
 *             record = self._lookup_record(sha1)
 *         if record == NULL:             # <<<<<<<<<<<<<<
 *             raise KeyError('key %r is not present' % (key,))
 *         return self._record_to_value_and_refs(record)
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":567
 *         if record == NULL:
 *             raise KeyError('key %r is not present' % (key,))
 *         return self._record_to_value_and_refs(record)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_record_to_value_and_refs(__pyx_v_self, __pyx_v_record); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":557
 *         return False
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":569
 *         return self._record_to_value_and_refs(record)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.num_records
 * 
*/

/* Python wrapper */
static Py_ssize_t __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_10__len__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_10__len__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  Py_ssize_t __pyx_r;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":570
 * 
 *     def __len__(self):
 *         return self.num_records             # <<<<<<<<<<<<<<
 * 
 *     def all_keys(self):
*/
  __pyx_r = __pyx_v_self->num_records;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":569
 *         return self._record_to_value_and_refs(record)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.num_records
 * 
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":572
 *         return self.num_records
 * 
 *     def all_keys(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_13all_keys(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_13all_keys = {"all_keys", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_13all_keys, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_13all_keys(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("all_keys (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("all_keys", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("all_keys", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12all_keys(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12all_keys(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  int __pyx_v_i;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("all_keys", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":574
 *     def all_keys(self):
 *         cdef int i
 *         result = []             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < self.num_records:
 *             PyList_Append(result, _sha1_to_key(self.records[i].sha1))
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":575
 *         cdef int i
 *         result = []
 *         for i from 0 <= i < self.num_records:             # <<<<<<<<<<<<<<
 *             PyList_Append(result, _sha1_to_key(self.records[i].sha1))
 *         return result
*/
  __pyx_t_2 = __pyx_v_self->num_records;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":576
 *         result = []
 *         for i from 0 <= i < self.num_records:
 *             PyList_Append(result, _sha1_to_key(self.records[i].sha1))             # <<<<<<<<<<<<<<
 *         return result
 * 
*/
    __pyx_t_1 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_key((__pyx_v_self->records[__pyx_v_i]).sha1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyList_Append(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 576, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":577
 *         for i from 0 <= i < self.num_records:
 *             PyList_Append(result, _sha1_to_key(self.records[i].sha1))
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def all_items(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":572
 *         return self.num_records
 * 
 *     def all_keys(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.all_keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":579
 *         return result
 * 
 *     def all_items(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_15all_items(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_15all_items = {"all_items", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_15all_items, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_15all_items(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("all_items (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("all_items", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("all_items", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_14all_items(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_14all_items(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  int __pyx_v_i;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("all_items", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":581
 *     def all_items(self):
 *         cdef int i
 *         result = []             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < self.num_records:
 *             item = self._record_to_item(&self.records[i])
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":582
 *         cdef int i
 *         result = []
 *         for i from 0 <= i < self.num_records:             # <<<<<<<<<<<<<<
 *             item = self._record_to_item(&self.records[i])
 *             PyList_Append(result, item)
*/
  __pyx_t_2 = __pyx_v_self->num_records;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":583
 *         result = []
 *         for i from 0 <= i < self.num_records:
 *             item = self._record_to_item(&self.records[i])             # <<<<<<<<<<<<<<
 *             PyList_Append(result, item)
 *         return result
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_record_to_item(__pyx_v_self, (&(__pyx_v_self->records[__pyx_v_i]))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 583, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_item, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":584
 *         for i from 0 <= i < self.num_records:
 *             item = self._record_to_item(&self.records[i])
 *             PyList_Append(result, item)             # <<<<<<<<<<<<<<
 *         return result
 * 
*/
    __pyx_t_3 = PyList_Append(__pyx_v_result, __pyx_v_item); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 584, __pyx_L1_error)
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":585
 *             item = self._record_to_item(&self.records[i])
 *             PyList_Append(result, item)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cdef int _count_records(self, char *c_content, char *c_end): # cannot_raise
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":579
 *         return result
 * 
 *     def all_items(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.all_items", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":587
 *         return result
 * 
 *     cdef int _count_records(self, char *c_content, char *c_end): # cannot_raise             # <<<<<<<<<<<<<<
 *         """Count how many records are in this section."""
 *         cdef char *c_cur
*/

static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__count_records(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_c_content, char *__pyx_v_c_end) {
  char *__pyx_v_c_cur;
  int __pyx_v_num_records;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":592
 *         cdef int num_records
 * 
 *         c_cur = c_content             # <<<<<<<<<<<<<<
 *         num_records = 0
 *         while c_cur != NULL and c_cur < c_end:
*/
  __pyx_v_c_cur = __pyx_v_c_content;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":593
 * 
 *         c_cur = c_content
 *         num_records = 0             # <<<<<<<<<<<<<<
 *         while c_cur != NULL and c_cur < c_end:
 *             c_cur = <char *>memchr(c_cur, c'\n', c_end - c_cur);
*/
  __pyx_v_num_records = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":594
 *         c_cur = c_content
 *         num_records = 0
 *         while c_cur != NULL and c_cur < c_end:             # <<<<<<<<<<<<<<
 *             c_cur = <char *>memchr(c_cur, c'\n', c_end - c_cur);
 *             if c_cur == NULL:
*/
  while (1) {
    __pyx_t_2 = (__pyx_v_c_cur != NULL);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_c_cur < __pyx_v_c_end);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L5_bool_binop_done:;
    if (!__pyx_t_1) break;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":595
 *         num_records = 0
 *         while c_cur != NULL and c_cur < c_end:
 *             c_cur = <char *>memchr(c_cur, c'\n', c_end - c_cur);             # <<<<<<<<<<<<<<
 *             if c_cur == NULL:
 *                 break
*/
    __pyx_v_c_cur = ((char *)memchr(__pyx_v_c_cur, '\n', (__pyx_v_c_end - __pyx_v_c_cur)));

    /* "breezy/bzr/_btree_serializer_pyx.pyx":596
 *         while c_cur != NULL and c_cur < c_end:
 *             c_cur = <char *>memchr(c_cur, c'\n', c_end - c_cur);
 *             if c_cur == NULL:             # <<<<<<<<<<<<<<
 *                 break
 *             c_cur = c_cur + 1
*/
    __pyx_t_1 = (__pyx_v_c_cur == NULL);
    if (__pyx_t_1) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":597
 *             c_cur = <char *>memchr(c_cur, c'\n', c_end - c_cur);
 *             if c_cur == NULL:
 *                 break             # <<<<<<<<<<<<<<
 *             c_cur = c_cur + 1
 *             num_records = num_records + 1
*/
      goto __pyx_L4_break;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":596
 *         while c_cur != NULL and c_cur < c_end:
 *             c_cur = <char *>memchr(c_cur, c'\n', c_end - c_cur);
 *             if c_cur == NULL:             # <<<<<<<<<<<<<<
 *                 break
 *             c_cur = c_cur + 1
*/
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":598
 *             if c_cur == NULL:
 *                 break
 *             c_cur = c_cur + 1             # <<<<<<<<<<<<<<
 *             num_records = num_records + 1
 *         return num_records
*/
    __pyx_v_c_cur = (__pyx_v_c_cur + 1);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":599
 *                 break
 *             c_cur = c_cur + 1
 *             num_records = num_records + 1             # <<<<<<<<<<<<<<
 *         return num_records
 * 
*/
    __pyx_v_num_records = (__pyx_v_num_records + 1);
  }
  __pyx_L4_break:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":600
 *             c_cur = c_cur + 1
 *             num_records = num_records + 1
 *         return num_records             # <<<<<<<<<<<<<<
 * 
 *     cdef _parse_bytes(self, data):
*/
  __pyx_r = __pyx_v_num_records;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":587
 *         return result
 * 
 *     cdef int _count_records(self, char *c_content, char *c_end): # cannot_raise             # <<<<<<<<<<<<<<
 *         """Count how many records are in this section."""
 *         cdef char *c_cur
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":602
 *         return num_records
 * 
 *     cdef _parse_bytes(self, data):             # <<<<<<<<<<<<<<
 *         """Parse the bytes 'data' into content."""
 *         cdef char *c_bytes
*/

static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__parse_bytes(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_data) {
  char *__pyx_v_c_bytes;
  char *__pyx_v_c_cur;
  char *__pyx_v_c_end;
  Py_ssize_t __pyx_v_n_bytes;
  int __pyx_v_num_records;
  int __pyx_v_entry;
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_cur_record;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_t_8;
  int __pyx_t_9;
  char *__pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_parse_bytes", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":612
 *         cdef gc_chk_sha1_record *cur_record
 * 
 *         if not PyBytes_CheckExact(data):             # <<<<<<<<<<<<<<
 *             raise TypeError('We only support parsing byte strings.')
 *         # Pass 1, count how many records there will be
*/
  __pyx_t_1 = (!PyBytes_CheckExact(__pyx_v_data));
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":613
 * 
 *         if not PyBytes_CheckExact(data):
 *             raise TypeError('We only support parsing byte strings.')             # <<<<<<<<<<<<<<
 *         # Pass 1, count how many records there will be
 *         n_bytes = PyBytes_GET_SIZE(data)
*/
    __pyx_t_3 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_We_only_support_parsing_byte_str};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 613, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 613, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":612
 *         cdef gc_chk_sha1_record *cur_record
 * 
 *         if not PyBytes_CheckExact(data):             # <<<<<<<<<<<<<<
 *             raise TypeError('We only support parsing byte strings.')
 *         # Pass 1, count how many records there will be
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":615
 *             raise TypeError('We only support parsing byte strings.')
 *         # Pass 1, count how many records there will be
 *         n_bytes = PyBytes_GET_SIZE(data)             # <<<<<<<<<<<<<<
 *         c_bytes = PyBytes_AS_STRING(data)
 *         c_end = c_bytes + n_bytes
*/
  __pyx_v_n_bytes = PyBytes_GET_SIZE(__pyx_v_data);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":616
 *         # Pass 1, count how many records there will be
 *         n_bytes = PyBytes_GET_SIZE(data)
 *         c_bytes = PyBytes_AS_STRING(data)             # <<<<<<<<<<<<<<
 *         c_end = c_bytes + n_bytes
 *         if strncmp(c_bytes, b'type=leaf\n', 10):
*/
  __pyx_v_c_bytes = PyBytes_AS_STRING(__pyx_v_data);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":617
 *         n_bytes = PyBytes_GET_SIZE(data)
 *         c_bytes = PyBytes_AS_STRING(data)
 *         c_end = c_bytes + n_bytes             # <<<<<<<<<<<<<<
 *         if strncmp(c_bytes, b'type=leaf\n', 10):
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"
*/
  __pyx_v_c_end = (__pyx_v_c_bytes + __pyx_v_n_bytes);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":618
 *         c_bytes = PyBytes_AS_STRING(data)
 *         c_end = c_bytes + n_bytes
 *         if strncmp(c_bytes, b'type=leaf\n', 10):             # <<<<<<<<<<<<<<
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"
 *                              % (data[:10],))
*/
  __pyx_t_1 = (strncmp(__pyx_v_c_bytes, __pyx_k_type_leaf_2, 10) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":619
 *         c_end = c_bytes + n_bytes
 *         if strncmp(c_bytes, b'type=leaf\n', 10):
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"             # <<<<<<<<<<<<<<
 *                              % (data[:10],))
 *         c_cur = c_bytes + 10
*/
    __pyx_t_3 = NULL;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":620
 *         if strncmp(c_bytes, b'type=leaf\n', 10):
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"
 *                              % (data[:10],))             # <<<<<<<<<<<<<<
 *         c_cur = c_bytes + 10
 *         num_records = self._count_records(c_cur, c_end)
*/
    __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_data, 0, 10, NULL, NULL, &__pyx_mstate_global->__pyx_slice[0], 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 620, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_t_5), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 620, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":619
 *         c_end = c_bytes + n_bytes
 *         if strncmp(c_bytes, b'type=leaf\n', 10):
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"             # <<<<<<<<<<<<<<
 *                              % (data[:10],))
 *         c_cur = c_bytes + 10
*/
    __pyx_t_5 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_bytes_did_not_start_with_type_le, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 619, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 619, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 619, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":618
 *         c_bytes = PyBytes_AS_STRING(data)
 *         c_end = c_bytes + n_bytes
 *         if strncmp(c_bytes, b'type=leaf\n', 10):             # <<<<<<<<<<<<<<
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"
 *                              % (data[:10],))
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":621
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"
 *                              % (data[:10],))
 *         c_cur = c_bytes + 10             # <<<<<<<<<<<<<<
 *         num_records = self._count_records(c_cur, c_end)
 *         # Now allocate the memory for these items, and go to town
*/
  __pyx_v_c_cur = (__pyx_v_c_bytes + 10);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":622
 *                              % (data[:10],))
 *         c_cur = c_bytes + 10
 *         num_records = self._count_records(c_cur, c_end)             # <<<<<<<<<<<<<<
 *         # Now allocate the memory for these items, and go to town
 *         self.records = <gc_chk_sha1_record*>PyMem_Malloc(num_records *
*/
  __pyx_t_7 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_count_records(__pyx_v_self, __pyx_v_c_cur, __pyx_v_c_end); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 622, __pyx_L1_error)
  __pyx_v_num_records = __pyx_t_7;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":624
 *         num_records = self._count_records(c_cur, c_end)
 *         # Now allocate the memory for these items, and go to town
 *         self.records = <gc_chk_sha1_record*>PyMem_Malloc(num_records *             # <<<<<<<<<<<<<<
 *             (sizeof(unsigned short) + sizeof(gc_chk_sha1_record)))
 *         self.num_records = num_records
*/
  __pyx_v_self->records = ((__pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *)PyMem_Malloc((__pyx_v_num_records * ((sizeof(unsigned short)) + (sizeof(__pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record))))));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":626
 *         self.records = <gc_chk_sha1_record*>PyMem_Malloc(num_records *
 *             (sizeof(unsigned short) + sizeof(gc_chk_sha1_record)))
 *         self.num_records = num_records             # <<<<<<<<<<<<<<
 *         cur_record = self.records
 *         entry = 0
*/
  __pyx_v_self->num_records = __pyx_v_num_records;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":627
 *             (sizeof(unsigned short) + sizeof(gc_chk_sha1_record)))
 *         self.num_records = num_records
 *         cur_record = self.records             # <<<<<<<<<<<<<<
 *         entry = 0
 *         while c_cur != NULL and c_cur < c_end and entry < num_records:
*/
  __pyx_t_8 = __pyx_v_self->records;
  __pyx_v_cur_record = __pyx_t_8;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":628
 *         self.num_records = num_records
 *         cur_record = self.records
 *         entry = 0             # <<<<<<<<<<<<<<
 *         while c_cur != NULL and c_cur < c_end and entry < num_records:
 *             c_cur = self._parse_one_entry(c_cur, c_end, cur_record)
*/
  __pyx_v_entry = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":629
 *         cur_record = self.records
 *         entry = 0
 *         while c_cur != NULL and c_cur < c_end and entry < num_records:             # <<<<<<<<<<<<<<
 *             c_cur = self._parse_one_entry(c_cur, c_end, cur_record)
 *             cur_record = cur_record + 1
*/
  while (1) {
    __pyx_t_9 = (__pyx_v_c_cur != NULL);
    if (__pyx_t_9) {
    } else {
      __pyx_t_1 = __pyx_t_9;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_9 = (__pyx_v_c_cur < __pyx_v_c_end);
    if (__pyx_t_9) {
    } else {
      __pyx_t_1 = __pyx_t_9;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_9 = (__pyx_v_entry < __pyx_v_num_records);
    __pyx_t_1 = __pyx_t_9;
    __pyx_L7_bool_binop_done:;
    if (!__pyx_t_1) break;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":630
 *         entry = 0
 *         while c_cur != NULL and c_cur < c_end and entry < num_records:
 *             c_cur = self._parse_one_entry(c_cur, c_end, cur_record)             # <<<<<<<<<<<<<<
 *             cur_record = cur_record + 1
 *             entry = entry + 1
*/
    __pyx_t_10 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_parse_one_entry(__pyx_v_self, __pyx_v_c_cur, __pyx_v_c_end, __pyx_v_cur_record); if (unlikely(__pyx_t_10 == ((void *)NULL))) __PYX_ERR(0, 630, __pyx_L1_error)
    __pyx_v_c_cur = __pyx_t_10;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":631
 *         while c_cur != NULL and c_cur < c_end and entry < num_records:
 *             c_cur = self._parse_one_entry(c_cur, c_end, cur_record)
 *             cur_record = cur_record + 1             # <<<<<<<<<<<<<<
 *             entry = entry + 1
 *         if (entry != self.num_records
*/
    __pyx_v_cur_record = (__pyx_v_cur_record + 1);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":632
 *             c_cur = self._parse_one_entry(c_cur, c_end, cur_record)
 *             cur_record = cur_record + 1
 *             entry = entry + 1             # <<<<<<<<<<<<<<
 *         if (entry != self.num_records
 *             or c_cur != c_end
*/
    __pyx_v_entry = (__pyx_v_entry + 1);
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":633
 *             cur_record = cur_record + 1
 *             entry = entry + 1
 *         if (entry != self.num_records             # <<<<<<<<<<<<<<
 *             or c_cur != c_end
 *             or cur_record != self.records + self.num_records):
*/
  __pyx_t_9 = (__pyx_v_entry != __pyx_v_self->num_records);
  if (!__pyx_t_9) {
  } else {
    __pyx_t_1 = __pyx_t_9;
    goto __pyx_L11_bool_binop_done;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":634
 *             entry = entry + 1
 *         if (entry != self.num_records
 *             or c_cur != c_end             # <<<<<<<<<<<<<<
 *             or cur_record != self.records + self.num_records):
 *             raise ValueError('Something went wrong while parsing.')
*/
  __pyx_t_9 = (__pyx_v_c_cur != __pyx_v_c_end);
  if (!__pyx_t_9) {
  } else {
    __pyx_t_1 = __pyx_t_9;
    goto __pyx_L11_bool_binop_done;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":635
 *         if (entry != self.num_records
 *             or c_cur != c_end
 *             or cur_record != self.records + self.num_records):             # <<<<<<<<<<<<<<
 *             raise ValueError('Something went wrong while parsing.')
 *         # Pass 3: build the offset map
*/
  __pyx_t_9 = (__pyx_v_cur_record != (__pyx_v_self->records + __pyx_v_self->num_records));
  __pyx_t_1 = __pyx_t_9;
  __pyx_L11_bool_binop_done:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":633
 *             cur_record = cur_record + 1
 *             entry = entry + 1
 *         if (entry != self.num_records             # <<<<<<<<<<<<<<
 *             or c_cur != c_end
 *             or cur_record != self.records + self.num_records):
*/
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":636
 *             or c_cur != c_end
 *             or cur_record != self.records + self.num_records):
 *             raise ValueError('Something went wrong while parsing.')             # <<<<<<<<<<<<<<
 *         # Pass 3: build the offset map
 *         self._compute_common()
*/
    __pyx_t_5 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_mstate_global->__pyx_kp_u_Something_went_wrong_while_parsi};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 636, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 636, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":633
 *             cur_record = cur_record + 1
 *             entry = entry + 1
 *         if (entry != self.num_records             # <<<<<<<<<<<<<<
 *             or c_cur != c_end
 *             or cur_record != self.records + self.num_records):
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":638
 *             raise ValueError('Something went wrong while parsing.')
 *         # Pass 3: build the offset map
 *         self._compute_common()             # <<<<<<<<<<<<<<
 * 
 *     cdef char *_parse_one_entry(self, char *c_cur, char *c_end,
*/
  __pyx_t_2 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_compute_common(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":602
 *         return num_records
 * 
 *     cdef _parse_bytes(self, data):             # <<<<<<<<<<<<<<
 *         """Parse the bytes 'data' into content."""
 *         cdef char *c_bytes
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode._parse_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":640
 *         self._compute_common()
 * 
 *     cdef char *_parse_one_entry(self, char *c_cur, char *c_end,             # <<<<<<<<<<<<<<
 *                                 gc_chk_sha1_record *cur_record) except NULL:
 *         """Read a single sha record from the bytes.
*/

static char *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__parse_one_entry(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_c_cur, char *__pyx_v_c_end, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_cur_record) {
  char *__pyx_v_c_next;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_parse_one_entry", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":648
 *         """
 *         cdef char *c_next
 *         if strncmp(c_cur, 'sha1:', 5):             # <<<<<<<<<<<<<<
 *             raise ValueError('line did not start with sha1: %r'
 *                 % (safe_string_from_size(c_cur, 10),))
*/
  __pyx_t_1 = (strncmp(__pyx_v_c_cur, __pyx_k_sha1, 5) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":649
 *         cdef char *c_next
 *         if strncmp(c_cur, 'sha1:', 5):
 *             raise ValueError('line did not start with sha1: %r'             # <<<<<<<<<<<<<<
 *                 % (safe_string_from_size(c_cur, 10),))
 *         c_cur = c_cur + 5
*/
    __pyx_t_3 = NULL;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":650
 *         if strncmp(c_cur, 'sha1:', 5):
 *             raise ValueError('line did not start with sha1: %r'
 *                 % (safe_string_from_size(c_cur, 10),))             # <<<<<<<<<<<<<<
 *         c_cur = c_cur + 5
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)
*/
    __pyx_t_4 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(__pyx_v_c_cur, 10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 650, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_t_4), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 650, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":649
 *         cdef char *c_next
 *         if strncmp(c_cur, 'sha1:', 5):
 *             raise ValueError('line did not start with sha1: %r'             # <<<<<<<<<<<<<<
 *                 % (safe_string_from_size(c_cur, 10),))
 *         c_cur = c_cur + 5
*/
    __pyx_t_4 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_line_did_not_start_with_sha1, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 649, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 649, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":648
 *         """
 *         cdef char *c_next
 *         if strncmp(c_cur, 'sha1:', 5):             # <<<<<<<<<<<<<<
 *             raise ValueError('line did not start with sha1: %r'
 *                 % (safe_string_from_size(c_cur, 10),))
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":651
 *             raise ValueError('line did not start with sha1: %r'
 *                 % (safe_string_from_size(c_cur, 10),))
 *         c_cur = c_cur + 5             # <<<<<<<<<<<<<<
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)
 *         if c_next == NULL or (c_next - c_cur != 40):
*/
  __pyx_v_c_cur = (__pyx_v_c_cur + 5);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":652
 *                 % (safe_string_from_size(c_cur, 10),))
 *         c_cur = c_cur + 5
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)             # <<<<<<<<<<<<<<
 *         if c_next == NULL or (c_next - c_cur != 40):
 *             raise ValueError('Line did not contain 40 hex bytes')
*/
  __pyx_v_c_next = ((char *)memchr(__pyx_v_c_cur, '\x00', (__pyx_v_c_end - __pyx_v_c_cur)));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":653
 *         c_cur = c_cur + 5
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)
 *         if c_next == NULL or (c_next - c_cur != 40):             # <<<<<<<<<<<<<<
 *             raise ValueError('Line did not contain 40 hex bytes')
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
*/
  __pyx_t_7 = (__pyx_v_c_next == NULL);
  if (!__pyx_t_7) {
  } else {
    __pyx_t_1 = __pyx_t_7;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_7 = ((__pyx_v_c_next - __pyx_v_c_cur) != 40);
  __pyx_t_1 = __pyx_t_7;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":654
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)
 *         if c_next == NULL or (c_next - c_cur != 40):
 *             raise ValueError('Line did not contain 40 hex bytes')             # <<<<<<<<<<<<<<
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
 *             raise ValueError('We failed to unhexlify')
*/
    __pyx_t_4 = NULL;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Line_did_not_contain_40_hex_byte};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 654, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 654, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":653
 *         c_cur = c_cur + 5
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)
 *         if c_next == NULL or (c_next - c_cur != 40):             # <<<<<<<<<<<<<<
 *             raise ValueError('Line did not contain 40 hex bytes')
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":655
 *         if c_next == NULL or (c_next - c_cur != 40):
 *             raise ValueError('Line did not contain 40 hex bytes')
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):             # <<<<<<<<<<<<<<
 *             raise ValueError('We failed to unhexlify')
 *         c_cur = c_next + 1
*/
  __pyx_t_8 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__unhexlify_sha1(__pyx_v_c_cur, __pyx_v_cur_record->sha1); if (unlikely(__pyx_t_8 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 655, __pyx_L1_error)
  __pyx_t_1 = (!(__pyx_t_8 != 0));
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":656
 *             raise ValueError('Line did not contain 40 hex bytes')
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
 *             raise ValueError('We failed to unhexlify')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         if c_cur[0] != c'\0':
*/
    __pyx_t_4 = NULL;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_We_failed_to_unhexlify};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 656, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 656, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":655
 *         if c_next == NULL or (c_next - c_cur != 40):
 *             raise ValueError('Line did not contain 40 hex bytes')
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):             # <<<<<<<<<<<<<<
 *             raise ValueError('We failed to unhexlify')
 *         c_cur = c_next + 1
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":657
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
 *             raise ValueError('We failed to unhexlify')
 *         c_cur = c_next + 1             # <<<<<<<<<<<<<<
 *         if c_cur[0] != c'\0':
 *             raise ValueError('only 1 null, not 2 as expected')
*/
  __pyx_v_c_cur = (__pyx_v_c_next + 1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":658
 *             raise ValueError('We failed to unhexlify')
 *         c_cur = c_next + 1
 *         if c_cur[0] != c'\0':             # <<<<<<<<<<<<<<
 *             raise ValueError('only 1 null, not 2 as expected')
 *         c_cur = c_cur + 1
*/
  __pyx_t_1 = ((__pyx_v_c_cur[0]) != '\x00');
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":659
 *         c_cur = c_next + 1
 *         if c_cur[0] != c'\0':
 *             raise ValueError('only 1 null, not 2 as expected')             # <<<<<<<<<<<<<<
 *         c_cur = c_cur + 1
 *         cur_record.block_offset = strtoull(c_cur, &c_next, 10)
*/
    __pyx_t_4 = NULL;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_only_1_null_not_2_as_expected};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 659, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 659, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":658
 *             raise ValueError('We failed to unhexlify')
 *         c_cur = c_next + 1
 *         if c_cur[0] != c'\0':             # <<<<<<<<<<<<<<
 *             raise ValueError('only 1 null, not 2 as expected')
 *         c_cur = c_cur + 1
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":660
 *         if c_cur[0] != c'\0':
 *             raise ValueError('only 1 null, not 2 as expected')
 *         c_cur = c_cur + 1             # <<<<<<<<<<<<<<
 *         cur_record.block_offset = strtoull(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
*/
  __pyx_v_c_cur = (__pyx_v_c_cur + 1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":661
 *             raise ValueError('only 1 null, not 2 as expected')
 *         c_cur = c_cur + 1
 *         cur_record.block_offset = strtoull(c_cur, &c_next, 10)             # <<<<<<<<<<<<<<
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block offset')
*/
  __pyx_v_cur_record->block_offset = strtoull(__pyx_v_c_cur, (&__pyx_v_c_next), 10);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":662
 *         c_cur = c_cur + 1
 *         cur_record.block_offset = strtoull(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse block offset')
 *         c_cur = c_next + 1
*/
  __pyx_t_7 = (__pyx_v_c_cur == __pyx_v_c_next);
  if (!__pyx_t_7) {
  } else {
    __pyx_t_1 = __pyx_t_7;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_7 = ((__pyx_v_c_next[0]) != ' ');
  __pyx_t_1 = __pyx_t_7;
  __pyx_L10_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":663
 *         cur_record.block_offset = strtoull(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block offset')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
*/
    __pyx_t_4 = NULL;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Failed_to_parse_block_offset};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 663, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 663, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":662
 *         c_cur = c_cur + 1
 *         cur_record.block_offset = strtoull(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse block offset')
 *         c_cur = c_next + 1
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":664
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block offset')
 *         c_cur = c_next + 1             # <<<<<<<<<<<<<<
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
*/
  __pyx_v_c_cur = (__pyx_v_c_next + 1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":665
 *             raise ValueError('Failed to parse block offset')
 *         c_cur = c_next + 1
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)             # <<<<<<<<<<<<<<
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')
*/
  __pyx_v_cur_record->block_length = strtoul(__pyx_v_c_cur, (&__pyx_v_c_next), 10);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":666
 *         c_cur = c_next + 1
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1
*/
  __pyx_t_7 = (__pyx_v_c_cur == __pyx_v_c_next);
  if (!__pyx_t_7) {
  } else {
    __pyx_t_1 = __pyx_t_7;
    goto __pyx_L13_bool_binop_done;
  }
  __pyx_t_7 = ((__pyx_v_c_next[0]) != ' ');
  __pyx_t_1 = __pyx_t_7;
  __pyx_L13_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":667
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)
*/
    __pyx_t_4 = NULL;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Failed_to_parse_block_length};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 667, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 667, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":666
 *         c_cur = c_next + 1
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":668
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1             # <<<<<<<<<<<<<<
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
*/
  __pyx_v_c_cur = (__pyx_v_c_next + 1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":669
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)             # <<<<<<<<<<<<<<
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')
*/
  __pyx_v_cur_record->record_start = strtoul(__pyx_v_c_cur, (&__pyx_v_c_next), 10);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":670
 *         c_cur = c_next + 1
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1
*/
  __pyx_t_7 = (__pyx_v_c_cur == __pyx_v_c_next);
  if (!__pyx_t_7) {
  } else {
    __pyx_t_1 = __pyx_t_7;
    goto __pyx_L16_bool_binop_done;
  }
  __pyx_t_7 = ((__pyx_v_c_next[0]) != ' ');
  __pyx_t_1 = __pyx_t_7;
  __pyx_L16_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":671
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)
*/
    __pyx_t_4 = NULL;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Failed_to_parse_block_length};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 671, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 671, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":670
 *         c_cur = c_next + 1
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":672
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1             # <<<<<<<<<<<<<<
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c'\n':
*/
  __pyx_v_c_cur = (__pyx_v_c_next + 1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":673
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)             # <<<<<<<<<<<<<<
 *         if c_cur == c_next or c_next[0] != c'\n':
 *             raise ValueError('Failed to parse record end')
*/
  __pyx_v_cur_record->record_end = strtoul(__pyx_v_c_cur, (&__pyx_v_c_next), 10);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":674
 *         c_cur = c_next + 1
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c'\n':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse record end')
 *         c_cur = c_next + 1
*/
  __pyx_t_7 = (__pyx_v_c_cur == __pyx_v_c_next);
  if (!__pyx_t_7) {
  } else {
    __pyx_t_1 = __pyx_t_7;
    goto __pyx_L19_bool_binop_done;
  }
  __pyx_t_7 = ((__pyx_v_c_next[0]) != '\n');
  __pyx_t_1 = __pyx_t_7;
  __pyx_L19_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":675
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c'\n':
 *             raise ValueError('Failed to parse record end')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         return c_cur
*/
    __pyx_t_4 = NULL;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Failed_to_parse_record_end};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 675, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 675, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":674
 *         c_cur = c_next + 1
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c'\n':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse record end')
 *         c_cur = c_next + 1
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":676
 *         if c_cur == c_next or c_next[0] != c'\n':
 *             raise ValueError('Failed to parse record end')
 *         c_cur = c_next + 1             # <<<<<<<<<<<<<<
 *         return c_cur
 * 
*/
  __pyx_v_c_cur = (__pyx_v_c_next + 1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":677
 *             raise ValueError('Failed to parse record end')
 *         c_cur = c_next + 1
 *         return c_cur             # <<<<<<<<<<<<<<
 * 
 *     cdef int _offset_for_sha1(self, char *sha1) except -1:
*/
  __pyx_r = __pyx_v_c_cur;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":640
 *         self._compute_common()
 * 
 *     cdef char *_parse_one_entry(self, char *c_cur, char *c_end,             # <<<<<<<<<<<<<<
 *                                 gc_chk_sha1_record *cur_record) except NULL:
 *         """Read a single sha record from the bytes.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode._parse_one_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":679
 *         return c_cur
 * 
 *     cdef int _offset_for_sha1(self, char *sha1) except -1:             # <<<<<<<<<<<<<<
 *         """Find the first interesting 8-bits of this sha1."""
 *         cdef int this_offset
*/

static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__offset_for_sha1(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_sha1) {
  int __pyx_v_this_offset;
  unsigned int __pyx_v_as_uint;
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":683
 *         cdef int this_offset
 *         cdef unsigned int as_uint
 *         as_uint = _sha1_to_uint(sha1)             # <<<<<<<<<<<<<<
 *         this_offset = (as_uint >> self.common_shift) & 0xFF
 *         return this_offset
*/
  __pyx_t_1 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_uint(__pyx_v_sha1); if (unlikely(__pyx_t_1 == ((unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 683, __pyx_L1_error)
  __pyx_v_as_uint = __pyx_t_1;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":684
 *         cdef unsigned int as_uint
 *         as_uint = _sha1_to_uint(sha1)
 *         this_offset = (as_uint >> self.common_shift) & 0xFF             # <<<<<<<<<<<<<<
 *         return this_offset
 * 
*/
  __pyx_v_this_offset = ((__pyx_v_as_uint >> __pyx_v_self->common_shift) & 0xFF);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":685
 *         as_uint = _sha1_to_uint(sha1)
 *         this_offset = (as_uint >> self.common_shift) & 0xFF
 *         return this_offset             # <<<<<<<<<<<<<<
 * 
 *     def _get_offset_for_sha1(self, sha1):
*/
  __pyx_r = __pyx_v_this_offset;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":679
 *         return c_cur
 * 
 *     cdef int _offset_for_sha1(self, char *sha1) except -1:             # <<<<<<<<<<<<<<
 *         """Find the first interesting 8-bits of this sha1."""
 *         cdef int this_offset
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode._offset_for_sha1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":687
 *         return this_offset
 * 
 *     def _get_offset_for_sha1(self, sha1):             # <<<<<<<<<<<<<<
 *         return self._offset_for_sha1(PyBytes_AS_STRING(sha1))
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_17_get_offset_for_sha1(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_17_get_offset_for_sha1 = {"_get_offset_for_sha1", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_17_get_offset_for_sha1, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_17_get_offset_for_sha1(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_sha1 = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_offset_for_sha1 (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_sha1_2,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 687, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 687, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_get_offset_for_sha1", 0) < (0)) __PYX_ERR(0, 687, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_get_offset_for_sha1", 1, 1, 1, i); __PYX_ERR(0, 687, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 687, __pyx_L3_error)
    }
    __pyx_v_sha1 = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_offset_for_sha1", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 687, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode._get_offset_for_sha1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_16_get_offset_for_sha1(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), __pyx_v_sha1);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_16_get_offset_for_sha1(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_sha1) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_offset_for_sha1", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":688
 * 
 *     def _get_offset_for_sha1(self, sha1):
 *         return self._offset_for_sha1(PyBytes_AS_STRING(sha1))             # <<<<<<<<<<<<<<
 * 
 *     cdef _compute_common(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_offset_for_sha1(__pyx_v_self, PyBytes_AS_STRING(__pyx_v_sha1)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 688, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":687
 *         return this_offset
 * 
 *     def _get_offset_for_sha1(self, sha1):             # <<<<<<<<<<<<<<
 *         return self._offset_for_sha1(PyBytes_AS_STRING(sha1))
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode._get_offset_for_sha1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":690
 *         return self._offset_for_sha1(PyBytes_AS_STRING(sha1))
 * 
 *     cdef _compute_common(self):             # <<<<<<<<<<<<<<
 *         cdef unsigned int first
 *         cdef unsigned int this
*/

static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__compute_common(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  unsigned int __pyx_v_first;
  unsigned int __pyx_v_this;
  unsigned int __pyx_v_common_mask;
  unsigned char __pyx_v_common_shift;
  int __pyx_v_i;
  int __pyx_v_offset;
  int __pyx_v_this_offset;
  int __pyx_v_max_offset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  unsigned int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compute_common", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":704
 *         # By XORing the records together, we can determine what bits are set in
 *         # all of them
 *         if self.num_records < 2:             # <<<<<<<<<<<<<<
 *             # Everything is in common if you have 0 or 1 leaves
 *             # So we'll always just shift to the first byte
*/
  __pyx_t_1 = (__pyx_v_self->num_records < 2);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":707
 *             # Everything is in common if you have 0 or 1 leaves
 *             # So we'll always just shift to the first byte
 *             self.common_shift = 24             # <<<<<<<<<<<<<<
 *         else:
 *             common_mask = 0xFFFFFFFF
*/
    __pyx_v_self->common_shift = 24;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":704
 *         # By XORing the records together, we can determine what bits are set in
 *         # all of them
 *         if self.num_records < 2:             # <<<<<<<<<<<<<<
 *             # Everything is in common if you have 0 or 1 leaves
 *             # So we'll always just shift to the first byte
*/
    goto __pyx_L3;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":709
 *             self.common_shift = 24
 *         else:
 *             common_mask = 0xFFFFFFFF             # <<<<<<<<<<<<<<
 *             first = _sha1_to_uint(self.records[0].sha1)
 *             for i from 0 < i < self.num_records:
*/
  /*else*/ {
    __pyx_v_common_mask = 0xFFFFFFFF;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":710
 *         else:
 *             common_mask = 0xFFFFFFFF
 *             first = _sha1_to_uint(self.records[0].sha1)             # <<<<<<<<<<<<<<
 *             for i from 0 < i < self.num_records:
 *                 this = _sha1_to_uint(self.records[i].sha1)
*/
    __pyx_t_2 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_uint((__pyx_v_self->records[0]).sha1); if (unlikely(__pyx_t_2 == ((unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 710, __pyx_L1_error)
    __pyx_v_first = __pyx_t_2;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":711
 *             common_mask = 0xFFFFFFFF
 *             first = _sha1_to_uint(self.records[0].sha1)
 *             for i from 0 < i < self.num_records:             # <<<<<<<<<<<<<<
 *                 this = _sha1_to_uint(self.records[i].sha1)
 *                 common_mask = (~(first ^ this)) & common_mask
*/
    __pyx_t_3 = __pyx_v_self->num_records;
    for (__pyx_v_i = 0+1; __pyx_v_i < __pyx_t_3; __pyx_v_i++) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":712
 *             first = _sha1_to_uint(self.records[0].sha1)
 *             for i from 0 < i < self.num_records:
 *                 this = _sha1_to_uint(self.records[i].sha1)             # <<<<<<<<<<<<<<
 *                 common_mask = (~(first ^ this)) & common_mask
 *             common_shift = 24
*/
      __pyx_t_2 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_uint((__pyx_v_self->records[__pyx_v_i]).sha1); if (unlikely(__pyx_t_2 == ((unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 712, __pyx_L1_error)
      __pyx_v_this = __pyx_t_2;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":713
 *             for i from 0 < i < self.num_records:
 *                 this = _sha1_to_uint(self.records[i].sha1)
 *                 common_mask = (~(first ^ this)) & common_mask             # <<<<<<<<<<<<<<
 *             common_shift = 24
 *             while common_mask & 0x80000000 and common_shift > 0:
*/
      __pyx_v_common_mask = ((~(__pyx_v_first ^ __pyx_v_this)) & __pyx_v_common_mask);
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":714
 *                 this = _sha1_to_uint(self.records[i].sha1)
 *                 common_mask = (~(first ^ this)) & common_mask
 *             common_shift = 24             # <<<<<<<<<<<<<<
 *             while common_mask & 0x80000000 and common_shift > 0:
 *                 common_mask = common_mask << 1
*/
    __pyx_v_common_shift = 24;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":715
 *                 common_mask = (~(first ^ this)) & common_mask
 *             common_shift = 24
 *             while common_mask & 0x80000000 and common_shift > 0:             # <<<<<<<<<<<<<<
 *                 common_mask = common_mask << 1
 *                 common_shift = common_shift - 1
*/
    while (1) {
      __pyx_t_4 = __Pyx_PyLong_From_unsigned_int(__pyx_v_common_mask); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 715, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyNumber_And(__pyx_t_4, __pyx_mstate_global->__pyx_int_2147483648); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 715, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 715, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_6) {
      } else {
        __pyx_t_1 = __pyx_t_6;
        goto __pyx_L8_bool_binop_done;
      }
      __pyx_t_6 = (__pyx_v_common_shift > 0);
      __pyx_t_1 = __pyx_t_6;
      __pyx_L8_bool_binop_done:;
      if (!__pyx_t_1) break;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":716
 *             common_shift = 24
 *             while common_mask & 0x80000000 and common_shift > 0:
 *                 common_mask = common_mask << 1             # <<<<<<<<<<<<<<
 *                 common_shift = common_shift - 1
 *             self.common_shift = common_shift
*/
      __pyx_v_common_mask = (__pyx_v_common_mask << 1);

      /* "breezy/bzr/_btree_serializer_pyx.pyx":717
 *             while common_mask & 0x80000000 and common_shift > 0:
 *                 common_mask = common_mask << 1
 *                 common_shift = common_shift - 1             # <<<<<<<<<<<<<<
 *             self.common_shift = common_shift
 *         offset = 0
*/
      __pyx_v_common_shift = (__pyx_v_common_shift - 1);
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":718
 *                 common_mask = common_mask << 1
 *                 common_shift = common_shift - 1
 *             self.common_shift = common_shift             # <<<<<<<<<<<<<<
 *         offset = 0
 *         max_offset = self.num_records
*/
    __pyx_v_self->common_shift = __pyx_v_common_shift;
  }
  __pyx_L3:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":719
 *                 common_shift = common_shift - 1
 *             self.common_shift = common_shift
 *         offset = 0             # <<<<<<<<<<<<<<
 *         max_offset = self.num_records
 *         # We cap this loop at 254 records. All the other offsets just get
*/
  __pyx_v_offset = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":720
 *             self.common_shift = common_shift
 *         offset = 0
 *         max_offset = self.num_records             # <<<<<<<<<<<<<<
 *         # We cap this loop at 254 records. All the other offsets just get
 *         # filled with 0xff as the singleton saying 'too many'.
*/
  __pyx_t_3 = __pyx_v_self->num_records;
  __pyx_v_max_offset = __pyx_t_3;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":725
 *         # It means that if we have >255 records we have to bisect the second
 *         # half of the list, but this is going to be very rare in practice.
 *         if max_offset > 255:             # <<<<<<<<<<<<<<
 *             max_offset = 255
 *         for i from 0 <= i < max_offset:
*/
  __pyx_t_1 = (__pyx_v_max_offset > 0xFF);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":726
 *         # half of the list, but this is going to be very rare in practice.
 *         if max_offset > 255:
 *             max_offset = 255             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < max_offset:
 *             this_offset = self._offset_for_sha1(self.records[i].sha1)
*/
    __pyx_v_max_offset = 0xFF;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":725
 *         # It means that if we have >255 records we have to bisect the second
 *         # half of the list, but this is going to be very rare in practice.
 *         if max_offset > 255:             # <<<<<<<<<<<<<<
 *             max_offset = 255
 *         for i from 0 <= i < max_offset:
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":727
 *         if max_offset > 255:
 *             max_offset = 255
 *         for i from 0 <= i < max_offset:             # <<<<<<<<<<<<<<
 *             this_offset = self._offset_for_sha1(self.records[i].sha1)
 *             while offset <= this_offset:
*/
  __pyx_t_3 = __pyx_v_max_offset;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_3; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":728
 *             max_offset = 255
 *         for i from 0 <= i < max_offset:
 *             this_offset = self._offset_for_sha1(self.records[i].sha1)             # <<<<<<<<<<<<<<
 *             while offset <= this_offset:
 *                 self.offsets[offset] = i
*/
    __pyx_t_7 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_offset_for_sha1(__pyx_v_self, (__pyx_v_self->records[__pyx_v_i]).sha1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 728, __pyx_L1_error)
    __pyx_v_this_offset = __pyx_t_7;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":729
 *         for i from 0 <= i < max_offset:
 *             this_offset = self._offset_for_sha1(self.records[i].sha1)
 *             while offset <= this_offset:             # <<<<<<<<<<<<<<
 *                 self.offsets[offset] = i
 *                 offset = offset + 1
*/
    while (1) {
      __pyx_t_1 = (__pyx_v_offset <= __pyx_v_this_offset);
      if (!__pyx_t_1) break;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":730
 *             this_offset = self._offset_for_sha1(self.records[i].sha1)
 *             while offset <= this_offset:
 *                 self.offsets[offset] = i             # <<<<<<<<<<<<<<
 *                 offset = offset + 1
 *         while offset < 257:
*/
      (__pyx_v_self->offsets[__pyx_v_offset]) = __pyx_v_i;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":731
 *             while offset <= this_offset:
 *                 self.offsets[offset] = i
 *                 offset = offset + 1             # <<<<<<<<<<<<<<
 *         while offset < 257:
 *             self.offsets[offset] = max_offset
*/
      __pyx_v_offset = (__pyx_v_offset + 1);
    }
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":732
 *                 self.offsets[offset] = i
 *                 offset = offset + 1
 *         while offset < 257:             # <<<<<<<<<<<<<<
 *             self.offsets[offset] = max_offset
 *             offset = offset + 1
*/
  while (1) {
    __pyx_t_1 = (__pyx_v_offset < 0x101);
    if (!__pyx_t_1) break;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":733
 *                 offset = offset + 1
 *         while offset < 257:
 *             self.offsets[offset] = max_offset             # <<<<<<<<<<<<<<
 *             offset = offset + 1
 * 
*/
    (__pyx_v_self->offsets[__pyx_v_offset]) = __pyx_v_max_offset;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":734
 *         while offset < 257:
 *             self.offsets[offset] = max_offset
 *             offset = offset + 1             # <<<<<<<<<<<<<<
 * 
 *     def _get_offsets(self):
*/
    __pyx_v_offset = (__pyx_v_offset + 1);
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":690
 *         return self._offset_for_sha1(PyBytes_AS_STRING(sha1))
 * 
 *     cdef _compute_common(self):             # <<<<<<<<<<<<<<
 *         cdef unsigned int first
 *         cdef unsigned int this
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode._compute_common", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":736
 *             offset = offset + 1
 * 
 *     def _get_offsets(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_19_get_offsets(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_19_get_offsets = {"_get_offsets", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_19_get_offsets, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_19_get_offsets(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_offsets (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_offsets", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_offsets", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_18_get_offsets(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_18_get_offsets(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  int __pyx_v_i;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_offsets", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":738
 *     def _get_offsets(self):
 *         cdef int i
 *         result = []             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < 257:
 *             PyList_Append(result, self.offsets[i])
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":739
 *         cdef int i
 *         result = []
 *         for i from 0 <= i < 257:             # <<<<<<<<<<<<<<
 *             PyList_Append(result, self.offsets[i])
 *         return result
*/
  for (__pyx_v_i = 0; __pyx_v_i < 0x101; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":740
 *         result = []
 *         for i from 0 <= i < 257:
 *             PyList_Append(result, self.offsets[i])             # <<<<<<<<<<<<<<
 *         return result
 * 
*/
    __pyx_t_1 = __Pyx_PyLong_From_unsigned_char((__pyx_v_self->offsets[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 740, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyList_Append(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 740, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":741
 *         for i from 0 <= i < 257:
 *             PyList_Append(result, self.offsets[i])
 *         return result             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":736
 *             offset = offset + 1
 * 
 *     def _get_offsets(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode._get_offsets", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":438
 * 
 *     cdef gc_chk_sha1_record *records
 *     cdef public object last_key             # <<<<<<<<<<<<<<
 *     cdef gc_chk_sha1_record *last_record
 *     cdef public int num_records
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key___get__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->last_key);
  __pyx_r = __pyx_v_self->last_key;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_2__set__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_2__set__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->last_key);
  __Pyx_DECREF(__pyx_v_self->last_key);
  __pyx_v_self->last_key = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_4__del__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_4__del__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->last_key);
  __Pyx_DECREF(__pyx_v_self->last_key);
  __pyx_v_self->last_key = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":440
 *     cdef public object last_key
 *     cdef gc_chk_sha1_record *last_record
 *     cdef public int num_records             # <<<<<<<<<<<<<<
 *     # This is the number of bits to shift to get to the interesting byte. A
 *     # value of 24 means that the very first byte changes across all keys.
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records___get__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->num_records); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.num_records.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_2__set__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_2__set__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 440, __pyx_L1_error)
  __pyx_v_self->num_records = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.num_records.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":446
 *     # ignore. 0 means that at least the first 3 bytes are identical, though
 *     # that is going to be very rare
 *     cdef public unsigned char common_shift             # <<<<<<<<<<<<<<
 *     # This maps an interesting byte to the first record that matches.
 *     # Equivalent to bisect.bisect_left(self.records, sha1), though only taking
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift___get__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->common_shift); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.common_shift.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_2__set__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_2__set__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  unsigned char __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_char(__pyx_v_value); if (unlikely((__pyx_t_1 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 446, __pyx_L1_error)
  __pyx_v_self->common_shift = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.common_shift.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.last_record,self.records cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_21__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_21__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_21__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_21__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_20__reduce_cython__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_20__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.last_record,self.records cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.last_record,self.records cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_last_record_self_records_ca, 0, 0);
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.last_record,self.records cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.last_record,self.records cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.last_record,self.records cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_23__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_23__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_23__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_23__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_22__setstate_cython__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_22__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.last_record,self.records cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.last_record,self.records cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_last_record_self_records_ca, 0, 0);
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.last_record,self.records cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.last_record,self.records cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":744
 * 
 * 
 * def _parse_into_chk(bytes, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     """Parse into a format optimized for chk records."""
 *     assert key_length == 1
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_11_parse_into_chk(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_10_parse_into_chk, "Parse into a format optimized for chk records.");
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_11_parse_into_chk = {"_parse_into_chk", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_11_parse_into_chk, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_10_parse_into_chk};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_11_parse_into_chk(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_bytes = 0;
  PyObject *__pyx_v_key_length = 0;
  PyObject *__pyx_v_ref_list_length = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_parse_into_chk (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_bytes,&__pyx_mstate_global->__pyx_n_u_key_length,&__pyx_mstate_global->__pyx_n_u_ref_list_length,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 744, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 744, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 744, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 744, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_parse_into_chk", 0) < (0)) __PYX_ERR(0, 744, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_parse_into_chk", 1, 3, 3, i); __PYX_ERR(0, 744, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 744, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 744, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 744, __pyx_L3_error)
    }
    __pyx_v_bytes = values[0];
    __pyx_v_key_length = values[1];
    __pyx_v_ref_list_length = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_parse_into_chk", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 744, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._parse_into_chk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_10_parse_into_chk(__pyx_self, __pyx_v_bytes, __pyx_v_key_length, __pyx_v_ref_list_length);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_10_parse_into_chk(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bytes, PyObject *__pyx_v_key_length, PyObject *__pyx_v_ref_list_length) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_parse_into_chk", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":746
 * def _parse_into_chk(bytes, key_length, ref_list_length):
 *     """Parse into a format optimized for chk records."""
 *     assert key_length == 1             # <<<<<<<<<<<<<<
 *     assert ref_list_length == 0
 *     return GCCHKSHA1LeafNode(bytes)
*/
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (__Pyx_PyLong_BoolEqObjC(__pyx_v_key_length, __pyx_mstate_global->__pyx_int_1, 1, 0)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 746, __pyx_L1_error)
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), 0, 0, 0);
      __PYX_ERR(0, 746, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 746, __pyx_L1_error)
  #endif

  /* "breezy/bzr/_btree_serializer_pyx.pyx":747
 *     """Parse into a format optimized for chk records."""
 *     assert key_length == 1
 *     assert ref_list_length == 0             # <<<<<<<<<<<<<<
 *     return GCCHKSHA1LeafNode(bytes)
 * 
*/
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (__Pyx_PyLong_BoolEqObjC(__pyx_v_ref_list_length, __pyx_mstate_global->__pyx_int_0, 0, 0)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 747, __pyx_L1_error)
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), 0, 0, 0);
      __PYX_ERR(0, 747, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 747, __pyx_L1_error)
  #endif

  /* "breezy/bzr/_btree_serializer_pyx.pyx":748
 *     assert key_length == 1
 *     assert ref_list_length == 0
 *     return GCCHKSHA1LeafNode(bytes)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_bytes};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 748, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_r = ((PyObject *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":744
 * 
 * 
 * def _parse_into_chk(bytes, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     """Parse into a format optimized for chk records."""
 *     assert key_length == 1
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._parse_into_chk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":751
 * 
 * 
 * def _flatten_node(node, reference_lists):             # <<<<<<<<<<<<<<
 *     """Convert a node into the serialized form.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_13_flatten_node(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_12_flatten_node, "Convert a node into the serialized form.\n\n    :param node: A tuple representing a node:\n        (index, key_tuple, value, references)\n    :param reference_lists: Does this index have reference lists?\n    :return: (string_key, flattened)\n        string_key  The serialized key for referencing this node\n        flattened   A string with the serialized form for the contents\n    ");
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_13_flatten_node = {"_flatten_node", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_13_flatten_node, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_12_flatten_node};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_13_flatten_node(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_node = 0;
  PyObject *__pyx_v_reference_lists = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_flatten_node (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_node,&__pyx_mstate_global->__pyx_n_u_reference_lists,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 751, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 751, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 751, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_flatten_node", 0) < (0)) __PYX_ERR(0, 751, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_flatten_node", 1, 2, 2, i); __PYX_ERR(0, 751, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 751, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 751, __pyx_L3_error)
    }
    __pyx_v_node = values[0];
    __pyx_v_reference_lists = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_flatten_node", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 751, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._flatten_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_12_flatten_node(__pyx_self, __pyx_v_node, __pyx_v_reference_lists);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_12_flatten_node(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_node, PyObject *__pyx_v_reference_lists) {
  int __pyx_v_have_reference_lists;
  Py_ssize_t __pyx_v_flat_len;
  Py_ssize_t __pyx_v_key_len;
  Py_ssize_t __pyx_v_node_len;
  char *__pyx_v_value;
  Py_ssize_t __pyx_v_value_len;
  char *__pyx_v_out;
  Py_ssize_t __pyx_v_refs_len;
  Py_ssize_t __pyx_v_next_len;
  int __pyx_v_first_ref_list;
  int __pyx_v_first_reference;
  int __pyx_v_i;
  Py_ssize_t __pyx_v_ref_bit_len;
  PyObject *__pyx_v_string_key = NULL;
  PyObject *__pyx_v_ref_lists = NULL;
  PyObject *__pyx_v_ref_list = NULL;
  PyObject *__pyx_v_reference = NULL;
  PyObject *__pyx_v_ref_bit = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_line = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  PyObject *(*__pyx_t_14)(PyObject *);
  PyObject *__pyx_t_15 = NULL;
  char *__pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_flatten_node", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":775
 *     cdef Py_ssize_t ref_bit_len
 * 
 *     if not PyTuple_CheckExact(node):             # <<<<<<<<<<<<<<
 *         raise TypeError('We expected a tuple() for node not: %s'
 *             % type(node))
*/
  __pyx_t_1 = (!PyTuple_CheckExact(__pyx_v_node));
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":776
 * 
 *     if not PyTuple_CheckExact(node):
 *         raise TypeError('We expected a tuple() for node not: %s'             # <<<<<<<<<<<<<<
 *             % type(node))
 *     node_len = len(node)
*/
    __pyx_t_3 = NULL;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":777
 *     if not PyTuple_CheckExact(node):
 *         raise TypeError('We expected a tuple() for node not: %s'
 *             % type(node))             # <<<<<<<<<<<<<<
 *     node_len = len(node)
 *     have_reference_lists = reference_lists
*/
    __pyx_t_4 = __Pyx_PyUnicode_FormatSafe(__pyx_mstate_global->__pyx_kp_u_We_expected_a_tuple_for_node_not, ((PyObject *)Py_TYPE(__pyx_v_node))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 776, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 776, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":775
 *     cdef Py_ssize_t ref_bit_len
 * 
 *     if not PyTuple_CheckExact(node):             # <<<<<<<<<<<<<<
 *         raise TypeError('We expected a tuple() for node not: %s'
 *             % type(node))
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":778
 *         raise TypeError('We expected a tuple() for node not: %s'
 *             % type(node))
 *     node_len = len(node)             # <<<<<<<<<<<<<<
 *     have_reference_lists = reference_lists
 *     if have_reference_lists:
*/
  __pyx_t_6 = PyObject_Length(__pyx_v_node); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 778, __pyx_L1_error)
  __pyx_v_node_len = __pyx_t_6;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":779
 *             % type(node))
 *     node_len = len(node)
 *     have_reference_lists = reference_lists             # <<<<<<<<<<<<<<
 *     if have_reference_lists:
 *         if node_len != 4:
*/
  __pyx_t_7 = __Pyx_PyLong_As_int(__pyx_v_reference_lists); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 779, __pyx_L1_error)
  __pyx_v_have_reference_lists = __pyx_t_7;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":780
 *     node_len = len(node)
 *     have_reference_lists = reference_lists
 *     if have_reference_lists:             # <<<<<<<<<<<<<<
 *         if node_len != 4:
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
*/
  __pyx_t_1 = (__pyx_v_have_reference_lists != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":781
 *     have_reference_lists = reference_lists
 *     if have_reference_lists:
 *         if node_len != 4:             # <<<<<<<<<<<<<<
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 *                 % len(node))
*/
    __pyx_t_1 = (__pyx_v_node_len != 4);
    if (unlikely(__pyx_t_1)) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":782
 *     if have_reference_lists:
 *         if node_len != 4:
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'             # <<<<<<<<<<<<<<
 *                 % len(node))
 *     elif node_len < 3:
*/
      __pyx_t_4 = NULL;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":783
 *         if node_len != 4:
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 *                 % len(node))             # <<<<<<<<<<<<<<
 *     elif node_len < 3:
 *         raise ValueError('Without ref_lists, we need at least 3 entries not: %s'
*/
      __pyx_t_6 = PyObject_Length(__pyx_v_node); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 783, __pyx_L1_error)
      __pyx_t_3 = PyLong_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 783, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PyUnicode_Format(__pyx_mstate_global->__pyx_kp_u_With_ref_lists_we_expected_4_ent, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 783, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_8};
        __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 782, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 782, __pyx_L1_error)

      /* "breezy/bzr/_btree_serializer_pyx.pyx":781
 *     have_reference_lists = reference_lists
 *     if have_reference_lists:
 *         if node_len != 4:             # <<<<<<<<<<<<<<
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 *                 % len(node))
*/
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":780
 *     node_len = len(node)
 *     have_reference_lists = reference_lists
 *     if have_reference_lists:             # <<<<<<<<<<<<<<
 *         if node_len != 4:
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
*/
    goto __pyx_L4;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":784
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 *                 % len(node))
 *     elif node_len < 3:             # <<<<<<<<<<<<<<
 *         raise ValueError('Without ref_lists, we need at least 3 entries not: %s'
 *             % len(node))
*/
  __pyx_t_1 = (__pyx_v_node_len < 3);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":785
 *                 % len(node))
 *     elif node_len < 3:
 *         raise ValueError('Without ref_lists, we need at least 3 entries not: %s'             # <<<<<<<<<<<<<<
 *             % len(node))
 *     # TODO: We can probably do better than string.join(), namely
*/
    __pyx_t_8 = NULL;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":786
 *     elif node_len < 3:
 *         raise ValueError('Without ref_lists, we need at least 3 entries not: %s'
 *             % len(node))             # <<<<<<<<<<<<<<
 *     # TODO: We can probably do better than string.join(), namely
 *     #       when key has only 1 item, we can just grab that string
*/
    __pyx_t_6 = PyObject_Length(__pyx_v_node); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 786, __pyx_L1_error)
    __pyx_t_4 = PyLong_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 786, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyUnicode_Format(__pyx_mstate_global->__pyx_kp_u_Without_ref_lists_we_need_at_lea, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 786, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_3};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 785, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 785, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":784
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 *                 % len(node))
 *     elif node_len < 3:             # <<<<<<<<<<<<<<
 *         raise ValueError('Without ref_lists, we need at least 3 entries not: %s'
 *             % len(node))
*/
  }
  __pyx_L4:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":797
 *     #       We *could* do more work on our own, and grab the actual items
 *     #       lists. For now, just ask people to use a better compiler. :)
 *     string_key = b'\0'.join(node[1])             # <<<<<<<<<<<<<<
 * 
 *     # TODO: instead of using string joins, precompute the final string length,
*/
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_node, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBytes_Join(__pyx_mstate_global->__pyx_kp_b_, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_string_key = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":811
 *     # ref := BYTES (NULL BYTES)*
 *     # value := BYTES
 *     refs_len = 0             # <<<<<<<<<<<<<<
 *     if have_reference_lists:
 *         # Figure out how many bytes it will take to store the references
*/
  __pyx_v_refs_len = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":812
 *     # value := BYTES
 *     refs_len = 0
 *     if have_reference_lists:             # <<<<<<<<<<<<<<
 *         # Figure out how many bytes it will take to store the references
 *         ref_lists = node[3]
*/
  __pyx_t_1 = (__pyx_v_have_reference_lists != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":814
 *     if have_reference_lists:
 *         # Figure out how many bytes it will take to store the references
 *         ref_lists = node[3]             # <<<<<<<<<<<<<<
 *         next_len = len(ref_lists) # TODO: use a Py function
 *         if next_len > 0:
*/
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_node, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 814, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_ref_lists = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":815
 *         # Figure out how many bytes it will take to store the references
 *         ref_lists = node[3]
 *         next_len = len(ref_lists) # TODO: use a Py function             # <<<<<<<<<<<<<<
 *         if next_len > 0:
 *             # If there are no nodes, we don't need to do any work
*/
    __pyx_t_6 = PyObject_Length(__pyx_v_ref_lists); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 815, __pyx_L1_error)
    __pyx_v_next_len = __pyx_t_6;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":816
 *         ref_lists = node[3]
 *         next_len = len(ref_lists) # TODO: use a Py function
 *         if next_len > 0:             # <<<<<<<<<<<<<<
 *             # If there are no nodes, we don't need to do any work
 *             # Otherwise we will need (len - 1) '\t' characters to separate
*/
    __pyx_t_1 = (__pyx_v_next_len > 0);
    if (__pyx_t_1) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":820
 *             # Otherwise we will need (len - 1) '\t' characters to separate
 *             # the reference lists
 *             refs_len = refs_len + (next_len - 1)             # <<<<<<<<<<<<<<
 *             for ref_list in ref_lists:
 *                 next_len = len(ref_list)
*/
      __pyx_v_refs_len = (__pyx_v_refs_len + (__pyx_v_next_len - 1));

      /* "breezy/bzr/_btree_serializer_pyx.pyx":821
 *             # the reference lists
 *             refs_len = refs_len + (next_len - 1)
 *             for ref_list in ref_lists:             # <<<<<<<<<<<<<<
 *                 next_len = len(ref_list)
 *                 if next_len > 0:
*/
      if (likely(PyList_CheckExact(__pyx_v_ref_lists)) || PyTuple_CheckExact(__pyx_v_ref_lists)) {
        __pyx_t_3 = __pyx_v_ref_lists; __Pyx_INCREF(__pyx_t_3);
        __pyx_t_6 = 0;
        __pyx_t_9 = NULL;
      } else {
        __pyx_t_6 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_ref_lists); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 821, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 821, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_9)) {
          if (likely(PyList_CheckExact(__pyx_t_3))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
              #if !CYTHON_ASSUME_SAFE_SIZE
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 821, __pyx_L1_error)
              #endif
              if (__pyx_t_6 >= __pyx_temp) break;
            }
            __pyx_t_2 = __Pyx_PyList_GetItemRefFast(__pyx_t_3, __pyx_t_6, __Pyx_ReferenceSharing_OwnStrongReference);
            ++__pyx_t_6;
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
              #if !CYTHON_ASSUME_SAFE_SIZE
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 821, __pyx_L1_error)
              #endif
              if (__pyx_t_6 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_6));
            #else
            __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_6);
            #endif
            ++__pyx_t_6;
          }
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 821, __pyx_L1_error)
        } else {
          __pyx_t_2 = __pyx_t_9(__pyx_t_3);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 821, __pyx_L1_error)
              PyErr_Clear();
            }
            break;
          }
        }
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_XDECREF_SET(__pyx_v_ref_list, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":822
 *             refs_len = refs_len + (next_len - 1)
 *             for ref_list in ref_lists:
 *                 next_len = len(ref_list)             # <<<<<<<<<<<<<<
 *                 if next_len > 0:
 *                     # We will need (len - 1) '\r' characters to separate the
*/
        __pyx_t_10 = PyObject_Length(__pyx_v_ref_list); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 822, __pyx_L1_error)
        __pyx_v_next_len = __pyx_t_10;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":823
 *             for ref_list in ref_lists:
 *                 next_len = len(ref_list)
 *                 if next_len > 0:             # <<<<<<<<<<<<<<
 *                     # We will need (len - 1) '\r' characters to separate the
 *                     # references
*/
        __pyx_t_1 = (__pyx_v_next_len > 0);
        if (__pyx_t_1) {

          /* "breezy/bzr/_btree_serializer_pyx.pyx":826
 *                     # We will need (len - 1) '\r' characters to separate the
 *                     # references
 *                     refs_len = refs_len + (next_len - 1)             # <<<<<<<<<<<<<<
 *                     for reference in ref_list:
 *                         if not PyTuple_CheckExact(reference):
*/
          __pyx_v_refs_len = (__pyx_v_refs_len + (__pyx_v_next_len - 1));

          /* "breezy/bzr/_btree_serializer_pyx.pyx":827
 *                     # references
 *                     refs_len = refs_len + (next_len - 1)
 *                     for reference in ref_list:             # <<<<<<<<<<<<<<
 *                         if not PyTuple_CheckExact(reference):
 *                             raise TypeError(
*/
          if (likely(PyList_CheckExact(__pyx_v_ref_list)) || PyTuple_CheckExact(__pyx_v_ref_list)) {
            __pyx_t_2 = __pyx_v_ref_list; __Pyx_INCREF(__pyx_t_2);
            __pyx_t_10 = 0;
            __pyx_t_11 = NULL;
          } else {
            __pyx_t_10 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_ref_list); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 827, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_11 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 827, __pyx_L1_error)
          }
          for (;;) {
            if (likely(!__pyx_t_11)) {
              if (likely(PyList_CheckExact(__pyx_t_2))) {
                {
                  Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
                  #if !CYTHON_ASSUME_SAFE_SIZE
                  if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 827, __pyx_L1_error)
                  #endif
                  if (__pyx_t_10 >= __pyx_temp) break;
                }
                __pyx_t_8 = __Pyx_PyList_GetItemRefFast(__pyx_t_2, __pyx_t_10, __Pyx_ReferenceSharing_OwnStrongReference);
                ++__pyx_t_10;
              } else {
                {
                  Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
                  #if !CYTHON_ASSUME_SAFE_SIZE
                  if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 827, __pyx_L1_error)
                  #endif
                  if (__pyx_t_10 >= __pyx_temp) break;
                }
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_8 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_10));
                #else
                __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_10);
                #endif
                ++__pyx_t_10;
              }
              if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 827, __pyx_L1_error)
            } else {
              __pyx_t_8 = __pyx_t_11(__pyx_t_2);
              if (unlikely(!__pyx_t_8)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 827, __pyx_L1_error)
                  PyErr_Clear();
                }
                break;
              }
            }
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_XDECREF_SET(__pyx_v_reference, __pyx_t_8);
            __pyx_t_8 = 0;

            /* "breezy/bzr/_btree_serializer_pyx.pyx":828
 *                     refs_len = refs_len + (next_len - 1)
 *                     for reference in ref_list:
 *                         if not PyTuple_CheckExact(reference):             # <<<<<<<<<<<<<<
 *                             raise TypeError(
 *                                 'We expect references to be tuples not: %r'
*/
            __pyx_t_1 = (!PyTuple_CheckExact(__pyx_v_reference));
            if (unlikely(__pyx_t_1)) {

              /* "breezy/bzr/_btree_serializer_pyx.pyx":829
 *                     for reference in ref_list:
 *                         if not PyTuple_CheckExact(reference):
 *                             raise TypeError(             # <<<<<<<<<<<<<<
 *                                 'We expect references to be tuples not: %r'
 *                                 % type(reference))
*/
              __pyx_t_4 = NULL;

              /* "breezy/bzr/_btree_serializer_pyx.pyx":831
 *                             raise TypeError(
 *                                 'We expect references to be tuples not: %r'
 *                                 % type(reference))             # <<<<<<<<<<<<<<
 *                         next_len = len(reference)
 *                         if next_len > 0:
*/
              __pyx_t_12 = __Pyx_PyUnicode_FormatSafe(__pyx_mstate_global->__pyx_kp_u_We_expect_references_to_be_tuple, ((PyObject *)Py_TYPE(__pyx_v_reference))); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 831, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_12);
              __pyx_t_5 = 1;
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_12};
                __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 829, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_8);
              }
              __Pyx_Raise(__pyx_t_8, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __PYX_ERR(0, 829, __pyx_L1_error)

              /* "breezy/bzr/_btree_serializer_pyx.pyx":828
 *                     refs_len = refs_len + (next_len - 1)
 *                     for reference in ref_list:
 *                         if not PyTuple_CheckExact(reference):             # <<<<<<<<<<<<<<
 *                             raise TypeError(
 *                                 'We expect references to be tuples not: %r'
*/
            }

            /* "breezy/bzr/_btree_serializer_pyx.pyx":832
 *                                 'We expect references to be tuples not: %r'
 *                                 % type(reference))
 *                         next_len = len(reference)             # <<<<<<<<<<<<<<
 *                         if next_len > 0:
 *                             # We will need (len - 1) '\x00' characters to
*/
            __pyx_t_13 = PyObject_Length(__pyx_v_reference); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 832, __pyx_L1_error)
            __pyx_v_next_len = __pyx_t_13;

            /* "breezy/bzr/_btree_serializer_pyx.pyx":833
 *                                 % type(reference))
 *                         next_len = len(reference)
 *                         if next_len > 0:             # <<<<<<<<<<<<<<
 *                             # We will need (len - 1) '\x00' characters to
 *                             # separate the reference key
*/
            __pyx_t_1 = (__pyx_v_next_len > 0);
            if (__pyx_t_1) {

              /* "breezy/bzr/_btree_serializer_pyx.pyx":836
 *                             # We will need (len - 1) '\x00' characters to
 *                             # separate the reference key
 *                             refs_len = refs_len + (next_len - 1)             # <<<<<<<<<<<<<<
 *                             for ref_bit in reference:
 *                                 if not PyBytes_CheckExact(ref_bit):
*/
              __pyx_v_refs_len = (__pyx_v_refs_len + (__pyx_v_next_len - 1));

              /* "breezy/bzr/_btree_serializer_pyx.pyx":837
 *                             # separate the reference key
 *                             refs_len = refs_len + (next_len - 1)
 *                             for ref_bit in reference:             # <<<<<<<<<<<<<<
 *                                 if not PyBytes_CheckExact(ref_bit):
 *                                     raise TypeError(
*/
              if (likely(PyList_CheckExact(__pyx_v_reference)) || PyTuple_CheckExact(__pyx_v_reference)) {
                __pyx_t_8 = __pyx_v_reference; __Pyx_INCREF(__pyx_t_8);
                __pyx_t_13 = 0;
                __pyx_t_14 = NULL;
              } else {
                __pyx_t_13 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_v_reference); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 837, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_8);
                __pyx_t_14 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 837, __pyx_L1_error)
              }
              for (;;) {
                if (likely(!__pyx_t_14)) {
                  if (likely(PyList_CheckExact(__pyx_t_8))) {
                    {
                      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_8);
                      #if !CYTHON_ASSUME_SAFE_SIZE
                      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 837, __pyx_L1_error)
                      #endif
                      if (__pyx_t_13 >= __pyx_temp) break;
                    }
                    __pyx_t_12 = __Pyx_PyList_GetItemRefFast(__pyx_t_8, __pyx_t_13, __Pyx_ReferenceSharing_OwnStrongReference);
                    ++__pyx_t_13;
                  } else {
                    {
                      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_8);
                      #if !CYTHON_ASSUME_SAFE_SIZE
                      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 837, __pyx_L1_error)
                      #endif
                      if (__pyx_t_13 >= __pyx_temp) break;
                    }
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_12 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_13));
                    #else
                    __pyx_t_12 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_13);
                    #endif
                    ++__pyx_t_13;
                  }
                  if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 837, __pyx_L1_error)
                } else {
                  __pyx_t_12 = __pyx_t_14(__pyx_t_8);
                  if (unlikely(!__pyx_t_12)) {
                    PyObject* exc_type = PyErr_Occurred();
                    if (exc_type) {
                      if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 837, __pyx_L1_error)
                      PyErr_Clear();
                    }
                    break;
                  }
                }
                __Pyx_GOTREF(__pyx_t_12);
                __Pyx_XDECREF_SET(__pyx_v_ref_bit, __pyx_t_12);
                __pyx_t_12 = 0;

                /* "breezy/bzr/_btree_serializer_pyx.pyx":838
 *                             refs_len = refs_len + (next_len - 1)
 *                             for ref_bit in reference:
 *                                 if not PyBytes_CheckExact(ref_bit):             # <<<<<<<<<<<<<<
 *                                     raise TypeError(
 *                                         'We expect reference bits to be bytes'
*/
                __pyx_t_1 = (!PyBytes_CheckExact(__pyx_v_ref_bit));
                if (unlikely(__pyx_t_1)) {

                  /* "breezy/bzr/_btree_serializer_pyx.pyx":839
 *                             for ref_bit in reference:
 *                                 if not PyBytes_CheckExact(ref_bit):
 *                                     raise TypeError(             # <<<<<<<<<<<<<<
 *                                         'We expect reference bits to be bytes'
 *                                         ' not: %r' % type(ref_bit))
*/
                  __pyx_t_4 = NULL;

                  /* "breezy/bzr/_btree_serializer_pyx.pyx":841
 *                                     raise TypeError(
 *                                         'We expect reference bits to be bytes'
 *                                         ' not: %r' % type(ref_bit))             # <<<<<<<<<<<<<<
 *                                 refs_len = refs_len + PyBytes_GET_SIZE(ref_bit)
 * 
*/
                  __pyx_t_15 = __Pyx_PyUnicode_FormatSafe(__pyx_mstate_global->__pyx_kp_u_We_expect_reference_bits_to_be_b, ((PyObject *)Py_TYPE(__pyx_v_ref_bit))); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 841, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_15);
                  __pyx_t_5 = 1;
                  {
                    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_15};
                    __pyx_t_12 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                    if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 839, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_12);
                  }
                  __Pyx_Raise(__pyx_t_12, 0, 0, 0);
                  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                  __PYX_ERR(0, 839, __pyx_L1_error)

                  /* "breezy/bzr/_btree_serializer_pyx.pyx":838
 *                             refs_len = refs_len + (next_len - 1)
 *                             for ref_bit in reference:
 *                                 if not PyBytes_CheckExact(ref_bit):             # <<<<<<<<<<<<<<
 *                                     raise TypeError(
 *                                         'We expect reference bits to be bytes'
*/
                }

                /* "breezy/bzr/_btree_serializer_pyx.pyx":842
 *                                         'We expect reference bits to be bytes'
 *                                         ' not: %r' % type(ref_bit))
 *                                 refs_len = refs_len + PyBytes_GET_SIZE(ref_bit)             # <<<<<<<<<<<<<<
 * 
 *     # So we have the (key NULL refs NULL value LF)
*/
                __pyx_v_refs_len = (__pyx_v_refs_len + PyBytes_GET_SIZE(__pyx_v_ref_bit));

                /* "breezy/bzr/_btree_serializer_pyx.pyx":837
 *                             # separate the reference key
 *                             refs_len = refs_len + (next_len - 1)
 *                             for ref_bit in reference:             # <<<<<<<<<<<<<<
 *                                 if not PyBytes_CheckExact(ref_bit):
 *                                     raise TypeError(
*/
              }
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

              /* "breezy/bzr/_btree_serializer_pyx.pyx":833
 *                                 % type(reference))
 *                         next_len = len(reference)
 *                         if next_len > 0:             # <<<<<<<<<<<<<<
 *                             # We will need (len - 1) '\x00' characters to
 *                             # separate the reference key
*/
            }

            /* "breezy/bzr/_btree_serializer_pyx.pyx":827
 *                     # references
 *                     refs_len = refs_len + (next_len - 1)
 *                     for reference in ref_list:             # <<<<<<<<<<<<<<
 *                         if not PyTuple_CheckExact(reference):
 *                             raise TypeError(
*/
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "breezy/bzr/_btree_serializer_pyx.pyx":823
 *             for ref_list in ref_lists:
 *                 next_len = len(ref_list)
 *                 if next_len > 0:             # <<<<<<<<<<<<<<
 *                     # We will need (len - 1) '\r' characters to separate the
 *                     # references
*/
        }

        /* "breezy/bzr/_btree_serializer_pyx.pyx":821
 *             # the reference lists
 *             refs_len = refs_len + (next_len - 1)
 *             for ref_list in ref_lists:             # <<<<<<<<<<<<<<
 *                 next_len = len(ref_list)
 *                 if next_len > 0:
*/
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":816
 *         ref_lists = node[3]
 *         next_len = len(ref_lists) # TODO: use a Py function
 *         if next_len > 0:             # <<<<<<<<<<<<<<
 *             # If there are no nodes, we don't need to do any work
 *             # Otherwise we will need (len - 1) '\t' characters to separate
*/
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":812
 *     # value := BYTES
 *     refs_len = 0
 *     if have_reference_lists:             # <<<<<<<<<<<<<<
 *         # Figure out how many bytes it will take to store the references
 *         ref_lists = node[3]
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":845
 * 
 *     # So we have the (key NULL refs NULL value LF)
 *     key_len = PyBytes_Size(string_key)             # <<<<<<<<<<<<<<
 *     val = node[2]
 *     if not PyBytes_CheckExact(val):
*/
  __pyx_t_6 = PyBytes_Size(__pyx_v_string_key); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 845, __pyx_L1_error)
  __pyx_v_key_len = __pyx_t_6;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":846
 *     # So we have the (key NULL refs NULL value LF)
 *     key_len = PyBytes_Size(string_key)
 *     val = node[2]             # <<<<<<<<<<<<<<
 *     if not PyBytes_CheckExact(val):
 *         raise TypeError('Expected bytes for value not: %r' % type(val))
*/
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_node, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 846, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_val = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":847
 *     key_len = PyBytes_Size(string_key)
 *     val = node[2]
 *     if not PyBytes_CheckExact(val):             # <<<<<<<<<<<<<<
 *         raise TypeError('Expected bytes for value not: %r' % type(val))
 *     value = PyBytes_AS_STRING(val)
*/
  __pyx_t_1 = (!PyBytes_CheckExact(__pyx_v_val));
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":848
 *     val = node[2]
 *     if not PyBytes_CheckExact(val):
 *         raise TypeError('Expected bytes for value not: %r' % type(val))             # <<<<<<<<<<<<<<
 *     value = PyBytes_AS_STRING(val)
 *     value_len = PyBytes_GET_SIZE(val)
*/
    __pyx_t_2 = NULL;
    __pyx_t_8 = __Pyx_PyUnicode_FormatSafe(__pyx_mstate_global->__pyx_kp_u_Expected_bytes_for_value_not_r, ((PyObject *)Py_TYPE(__pyx_v_val))); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 848, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_8};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 848, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 848, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":847
 *     key_len = PyBytes_Size(string_key)
 *     val = node[2]
 *     if not PyBytes_CheckExact(val):             # <<<<<<<<<<<<<<
 *         raise TypeError('Expected bytes for value not: %r' % type(val))
 *     value = PyBytes_AS_STRING(val)
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":849
 *     if not PyBytes_CheckExact(val):
 *         raise TypeError('Expected bytes for value not: %r' % type(val))
 *     value = PyBytes_AS_STRING(val)             # <<<<<<<<<<<<<<
 *     value_len = PyBytes_GET_SIZE(val)
 *     flat_len = (key_len + 1 + refs_len + 1 + value_len + 1)
*/
  __pyx_v_value = PyBytes_AS_STRING(__pyx_v_val);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":850
 *         raise TypeError('Expected bytes for value not: %r' % type(val))
 *     value = PyBytes_AS_STRING(val)
 *     value_len = PyBytes_GET_SIZE(val)             # <<<<<<<<<<<<<<
 *     flat_len = (key_len + 1 + refs_len + 1 + value_len + 1)
 *     line = PyBytes_FromStringAndSize(NULL, flat_len)
*/
  __pyx_v_value_len = PyBytes_GET_SIZE(__pyx_v_val);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":851
 *     value = PyBytes_AS_STRING(val)
 *     value_len = PyBytes_GET_SIZE(val)
 *     flat_len = (key_len + 1 + refs_len + 1 + value_len + 1)             # <<<<<<<<<<<<<<
 *     line = PyBytes_FromStringAndSize(NULL, flat_len)
 *     # Get a pointer to the new buffer
*/
  __pyx_v_flat_len = (((((__pyx_v_key_len + 1) + __pyx_v_refs_len) + 1) + __pyx_v_value_len) + 1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":852
 *     value_len = PyBytes_GET_SIZE(val)
 *     flat_len = (key_len + 1 + refs_len + 1 + value_len + 1)
 *     line = PyBytes_FromStringAndSize(NULL, flat_len)             # <<<<<<<<<<<<<<
 *     # Get a pointer to the new buffer
 *     out = PyBytes_AsString(line)
*/
  __pyx_t_3 = PyBytes_FromStringAndSize(NULL, __pyx_v_flat_len); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_line = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":854
 *     line = PyBytes_FromStringAndSize(NULL, flat_len)
 *     # Get a pointer to the new buffer
 *     out = PyBytes_AsString(line)             # <<<<<<<<<<<<<<
 *     memcpy(out, PyBytes_AsString(string_key), key_len)
 *     out = out + key_len
*/
  __pyx_t_16 = PyBytes_AsString(__pyx_v_line); if (unlikely(__pyx_t_16 == ((void *)NULL))) __PYX_ERR(0, 854, __pyx_L1_error)
  __pyx_v_out = __pyx_t_16;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":855
 *     # Get a pointer to the new buffer
 *     out = PyBytes_AsString(line)
 *     memcpy(out, PyBytes_AsString(string_key), key_len)             # <<<<<<<<<<<<<<
 *     out = out + key_len
 *     out[0] = c'\0'
*/
  __pyx_t_16 = PyBytes_AsString(__pyx_v_string_key); if (unlikely(__pyx_t_16 == ((void *)NULL))) __PYX_ERR(0, 855, __pyx_L1_error)
  (void)(memcpy(__pyx_v_out, __pyx_t_16, __pyx_v_key_len));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":856
 *     out = PyBytes_AsString(line)
 *     memcpy(out, PyBytes_AsString(string_key), key_len)
 *     out = out + key_len             # <<<<<<<<<<<<<<
 *     out[0] = c'\0'
 *     out = out + 1
*/
  __pyx_v_out = (__pyx_v_out + __pyx_v_key_len);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":857
 *     memcpy(out, PyBytes_AsString(string_key), key_len)
 *     out = out + key_len
 *     out[0] = c'\0'             # <<<<<<<<<<<<<<
 *     out = out + 1
 *     if refs_len > 0:
*/
  (__pyx_v_out[0]) = '\x00';

  /* "breezy/bzr/_btree_serializer_pyx.pyx":858
 *     out = out + key_len
 *     out[0] = c'\0'
 *     out = out + 1             # <<<<<<<<<<<<<<
 *     if refs_len > 0:
 *         first_ref_list = 1
*/
  __pyx_v_out = (__pyx_v_out + 1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":859
 *     out[0] = c'\0'
 *     out = out + 1
 *     if refs_len > 0:             # <<<<<<<<<<<<<<
 *         first_ref_list = 1
 *         for ref_list in ref_lists:
*/
  __pyx_t_1 = (__pyx_v_refs_len > 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":860
 *     out = out + 1
 *     if refs_len > 0:
 *         first_ref_list = 1             # <<<<<<<<<<<<<<
 *         for ref_list in ref_lists:
 *             if first_ref_list == 0:
*/
    __pyx_v_first_ref_list = 1;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":861
 *     if refs_len > 0:
 *         first_ref_list = 1
 *         for ref_list in ref_lists:             # <<<<<<<<<<<<<<
 *             if first_ref_list == 0:
 *                 out[0] = c'\t'
*/
    if (unlikely(!__pyx_v_ref_lists)) { __Pyx_RaiseUnboundLocalError("ref_lists"); __PYX_ERR(0, 861, __pyx_L1_error) }
    if (likely(PyList_CheckExact(__pyx_v_ref_lists)) || PyTuple_CheckExact(__pyx_v_ref_lists)) {
      __pyx_t_3 = __pyx_v_ref_lists; __Pyx_INCREF(__pyx_t_3);
      __pyx_t_6 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_ref_lists); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 861, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 861, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 861, __pyx_L1_error)
            #endif
            if (__pyx_t_6 >= __pyx_temp) break;
          }
          __pyx_t_8 = __Pyx_PyList_GetItemRefFast(__pyx_t_3, __pyx_t_6, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_6;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 861, __pyx_L1_error)
            #endif
            if (__pyx_t_6 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_6));
          #else
          __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_6);
          #endif
          ++__pyx_t_6;
        }
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 861, __pyx_L1_error)
      } else {
        __pyx_t_8 = __pyx_t_9(__pyx_t_3);
        if (unlikely(!__pyx_t_8)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 861, __pyx_L1_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_v_ref_list, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":862
 *         first_ref_list = 1
 *         for ref_list in ref_lists:
 *             if first_ref_list == 0:             # <<<<<<<<<<<<<<
 *                 out[0] = c'\t'
 *                 out = out + 1
*/
      __pyx_t_1 = (__pyx_v_first_ref_list == 0);
      if (__pyx_t_1) {

        /* "breezy/bzr/_btree_serializer_pyx.pyx":863
 *         for ref_list in ref_lists:
 *             if first_ref_list == 0:
 *                 out[0] = c'\t'             # <<<<<<<<<<<<<<
 *                 out = out + 1
 *             first_ref_list = 0
*/
        (__pyx_v_out[0]) = '\t';

        /* "breezy/bzr/_btree_serializer_pyx.pyx":864
 *             if first_ref_list == 0:
 *                 out[0] = c'\t'
 *                 out = out + 1             # <<<<<<<<<<<<<<
 *             first_ref_list = 0
 *             first_reference = 1
*/
        __pyx_v_out = (__pyx_v_out + 1);

        /* "breezy/bzr/_btree_serializer_pyx.pyx":862
 *         first_ref_list = 1
 *         for ref_list in ref_lists:
 *             if first_ref_list == 0:             # <<<<<<<<<<<<<<
 *                 out[0] = c'\t'
 *                 out = out + 1
*/
      }

      /* "breezy/bzr/_btree_serializer_pyx.pyx":865
 *                 out[0] = c'\t'
 *                 out = out + 1
 *             first_ref_list = 0             # <<<<<<<<<<<<<<
 *             first_reference = 1
 *             for reference in ref_list:
*/
      __pyx_v_first_ref_list = 0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":866
 *                 out = out + 1
 *             first_ref_list = 0
 *             first_reference = 1             # <<<<<<<<<<<<<<
 *             for reference in ref_list:
 *                 if first_reference == 0:
*/
      __pyx_v_first_reference = 1;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":867
 *             first_ref_list = 0
 *             first_reference = 1
 *             for reference in ref_list:             # <<<<<<<<<<<<<<
 *                 if first_reference == 0:
 *                     out[0] = c'\r'
*/
      if (likely(PyList_CheckExact(__pyx_v_ref_list)) || PyTuple_CheckExact(__pyx_v_ref_list)) {
        __pyx_t_8 = __pyx_v_ref_list; __Pyx_INCREF(__pyx_t_8);
        __pyx_t_10 = 0;
        __pyx_t_11 = NULL;
      } else {
        __pyx_t_10 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_v_ref_list); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 867, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_11 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_8); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 867, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_11)) {
          if (likely(PyList_CheckExact(__pyx_t_8))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_8);
              #if !CYTHON_ASSUME_SAFE_SIZE
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 867, __pyx_L1_error)
              #endif
              if (__pyx_t_10 >= __pyx_temp) break;
            }
            __pyx_t_2 = __Pyx_PyList_GetItemRefFast(__pyx_t_8, __pyx_t_10, __Pyx_ReferenceSharing_OwnStrongReference);
            ++__pyx_t_10;
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_8);
              #if !CYTHON_ASSUME_SAFE_SIZE
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 867, __pyx_L1_error)
              #endif
              if (__pyx_t_10 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_10));
            #else
            __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_10);
            #endif
            ++__pyx_t_10;
          }
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 867, __pyx_L1_error)
        } else {
          __pyx_t_2 = __pyx_t_11(__pyx_t_8);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 867, __pyx_L1_error)
              PyErr_Clear();
            }
            break;
          }
        }
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_XDECREF_SET(__pyx_v_reference, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":868
 *             first_reference = 1
 *             for reference in ref_list:
 *                 if first_reference == 0:             # <<<<<<<<<<<<<<
 *                     out[0] = c'\r'
 *                     out = out + 1
*/
        __pyx_t_1 = (__pyx_v_first_reference == 0);
        if (__pyx_t_1) {

          /* "breezy/bzr/_btree_serializer_pyx.pyx":869
 *             for reference in ref_list:
 *                 if first_reference == 0:
 *                     out[0] = c'\r'             # <<<<<<<<<<<<<<
 *                     out = out + 1
 *                 first_reference = 0
*/
          (__pyx_v_out[0]) = '\r';

          /* "breezy/bzr/_btree_serializer_pyx.pyx":870
 *                 if first_reference == 0:
 *                     out[0] = c'\r'
 *                     out = out + 1             # <<<<<<<<<<<<<<
 *                 first_reference = 0
 *                 next_len = len(reference)
*/
          __pyx_v_out = (__pyx_v_out + 1);

          /* "breezy/bzr/_btree_serializer_pyx.pyx":868
 *             first_reference = 1
 *             for reference in ref_list:
 *                 if first_reference == 0:             # <<<<<<<<<<<<<<
 *                     out[0] = c'\r'
 *                     out = out + 1
*/
        }

        /* "breezy/bzr/_btree_serializer_pyx.pyx":871
 *                     out[0] = c'\r'
 *                     out = out + 1
 *                 first_reference = 0             # <<<<<<<<<<<<<<
 *                 next_len = len(reference)
 *                 for i from 0 <= i < next_len:
*/
        __pyx_v_first_reference = 0;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":872
 *                     out = out + 1
 *                 first_reference = 0
 *                 next_len = len(reference)             # <<<<<<<<<<<<<<
 *                 for i from 0 <= i < next_len:
 *                     if i != 0:
*/
        __pyx_t_13 = PyObject_Length(__pyx_v_reference); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 872, __pyx_L1_error)
        __pyx_v_next_len = __pyx_t_13;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":873
 *                 first_reference = 0
 *                 next_len = len(reference)
 *                 for i from 0 <= i < next_len:             # <<<<<<<<<<<<<<
 *                     if i != 0:
 *                         out[0] = c'\x00'
*/
        __pyx_t_13 = __pyx_v_next_len;
        for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_13; __pyx_v_i++) {

          /* "breezy/bzr/_btree_serializer_pyx.pyx":874
 *                 next_len = len(reference)
 *                 for i from 0 <= i < next_len:
 *                     if i != 0:             # <<<<<<<<<<<<<<
 *                         out[0] = c'\x00'
 *                         out = out + 1
*/
          __pyx_t_1 = (__pyx_v_i != 0);
          if (__pyx_t_1) {

            /* "breezy/bzr/_btree_serializer_pyx.pyx":875
 *                 for i from 0 <= i < next_len:
 *                     if i != 0:
 *                         out[0] = c'\x00'             # <<<<<<<<<<<<<<
 *                         out = out + 1
 *                     ref_bit = reference[i]
*/
            (__pyx_v_out[0]) = '\x00';

            /* "breezy/bzr/_btree_serializer_pyx.pyx":876
 *                     if i != 0:
 *                         out[0] = c'\x00'
 *                         out = out + 1             # <<<<<<<<<<<<<<
 *                     ref_bit = reference[i]
 *                     ref_bit_len = PyBytes_GET_SIZE(ref_bit)
*/
            __pyx_v_out = (__pyx_v_out + 1);

            /* "breezy/bzr/_btree_serializer_pyx.pyx":874
 *                 next_len = len(reference)
 *                 for i from 0 <= i < next_len:
 *                     if i != 0:             # <<<<<<<<<<<<<<
 *                         out[0] = c'\x00'
 *                         out = out + 1
*/
          }

          /* "breezy/bzr/_btree_serializer_pyx.pyx":877
 *                         out[0] = c'\x00'
 *                         out = out + 1
 *                     ref_bit = reference[i]             # <<<<<<<<<<<<<<
 *                     ref_bit_len = PyBytes_GET_SIZE(ref_bit)
 *                     memcpy(out, PyBytes_AS_STRING(ref_bit), ref_bit_len)
*/
          __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_reference, __pyx_v_i, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 877, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_XDECREF_SET(__pyx_v_ref_bit, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "breezy/bzr/_btree_serializer_pyx.pyx":878
 *                         out = out + 1
 *                     ref_bit = reference[i]
 *                     ref_bit_len = PyBytes_GET_SIZE(ref_bit)             # <<<<<<<<<<<<<<
 *                     memcpy(out, PyBytes_AS_STRING(ref_bit), ref_bit_len)
 *                     out = out + ref_bit_len
*/
          __pyx_v_ref_bit_len = PyBytes_GET_SIZE(__pyx_v_ref_bit);

          /* "breezy/bzr/_btree_serializer_pyx.pyx":879
 *                     ref_bit = reference[i]
 *                     ref_bit_len = PyBytes_GET_SIZE(ref_bit)
 *                     memcpy(out, PyBytes_AS_STRING(ref_bit), ref_bit_len)             # <<<<<<<<<<<<<<
 *                     out = out + ref_bit_len
 *     out[0] = c'\0'
*/
          (void)(memcpy(__pyx_v_out, PyBytes_AS_STRING(__pyx_v_ref_bit), __pyx_v_ref_bit_len));

          /* "breezy/bzr/_btree_serializer_pyx.pyx":880
 *                     ref_bit_len = PyBytes_GET_SIZE(ref_bit)
 *                     memcpy(out, PyBytes_AS_STRING(ref_bit), ref_bit_len)
 *                     out = out + ref_bit_len             # <<<<<<<<<<<<<<
 *     out[0] = c'\0'
 *     out = out  + 1
*/
          __pyx_v_out = (__pyx_v_out + __pyx_v_ref_bit_len);
        }

        /* "breezy/bzr/_btree_serializer_pyx.pyx":867
 *             first_ref_list = 0
 *             first_reference = 1
 *             for reference in ref_list:             # <<<<<<<<<<<<<<
 *                 if first_reference == 0:
 *                     out[0] = c'\r'
*/
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":861
 *     if refs_len > 0:
 *         first_ref_list = 1
 *         for ref_list in ref_lists:             # <<<<<<<<<<<<<<
 *             if first_ref_list == 0:
 *                 out[0] = c'\t'
*/
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":859
 *     out[0] = c'\0'
 *     out = out + 1
 *     if refs_len > 0:             # <<<<<<<<<<<<<<
 *         first_ref_list = 1
 *         for ref_list in ref_lists:
*/
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":881
 *                     memcpy(out, PyBytes_AS_STRING(ref_bit), ref_bit_len)
 *                     out = out + ref_bit_len
 *     out[0] = c'\0'             # <<<<<<<<<<<<<<
 *     out = out  + 1
 *     memcpy(out, value, value_len)
*/
  (__pyx_v_out[0]) = '\x00';

  /* "breezy/bzr/_btree_serializer_pyx.pyx":882
 *                     out = out + ref_bit_len
 *     out[0] = c'\0'
 *     out = out  + 1             # <<<<<<<<<<<<<<
 *     memcpy(out, value, value_len)
 *     out = out + value_len
*/
  __pyx_v_out = (__pyx_v_out + 1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":883
 *     out[0] = c'\0'
 *     out = out  + 1
 *     memcpy(out, value, value_len)             # <<<<<<<<<<<<<<
 *     out = out + value_len
 *     out[0] = c'\n'
*/
  (void)(memcpy(__pyx_v_out, __pyx_v_value, __pyx_v_value_len));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":884
 *     out = out  + 1
 *     memcpy(out, value, value_len)
 *     out = out + value_len             # <<<<<<<<<<<<<<
 *     out[0] = c'\n'
 *     return string_key, line
*/
  __pyx_v_out = (__pyx_v_out + __pyx_v_value_len);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":885
 *     memcpy(out, value, value_len)
 *     out = out + value_len
 *     out[0] = c'\n'             # <<<<<<<<<<<<<<
 *     return string_key, line
*/
  (__pyx_v_out[0]) = '\n';

  /* "breezy/bzr/_btree_serializer_pyx.pyx":886
 *     out = out + value_len
 *     out[0] = c'\n'
 *     return string_key, line             # <<<<<<<<<<<<<<
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_string_key);
  __Pyx_GIVEREF(__pyx_v_string_key);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_string_key) != (0)) __PYX_ERR(0, 886, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_line);
  __Pyx_GIVEREF(__pyx_v_line);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_line) != (0)) __PYX_ERR(0, 886, __pyx_L1_error);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":751
 * 
 * 
 * def _flatten_node(node, reference_lists):             # <<<<<<<<<<<<<<
 *     """Convert a node into the serialized form.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._flatten_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_string_key);
  __Pyx_XDECREF(__pyx_v_ref_lists);
  __Pyx_XDECREF(__pyx_v_ref_list);
  __Pyx_XDECREF(__pyx_v_reference);
  __Pyx_XDECREF(__pyx_v_ref_bit);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_line);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_BTreeLeafParser(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x97e5701, 0x92287bf, 0x9b802a2, b'_cur_str, _end_str, _header_found, _start, data, key_length, keys, ref_list_length')
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15__pyx_unpickle_BTreeLeafParser(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_15__pyx_unpickle_BTreeLeafParser = {"__pyx_unpickle_BTreeLeafParser", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15__pyx_unpickle_BTreeLeafParser, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15__pyx_unpickle_BTreeLeafParser(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_BTreeLeafParser (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_BTreeLeafParser", 0) < (0)) __PYX_ERR(2, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BTreeLeafParser", 1, 3, 3, i); __PYX_ERR(2, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(2, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(2, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BTreeLeafParser", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.__pyx_unpickle_BTreeLeafParser", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(2, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_14__pyx_unpickle_BTreeLeafParser(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_14__pyx_unpickle_BTreeLeafParser(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_BTreeLeafParser", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle_BTreeLeafParser(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x97e5701, 0x92287bf, 0x9b802a2, b'_cur_str, _end_str, _header_found, _start, data, key_length, keys, ref_list_length')             # <<<<<<<<<<<<<<
 *     __pyx_result = BTreeLeafParser.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0x97e5701, 0x92287bf, 0x9b802a2, __pyx_k_cur_str__end_str__header_found); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x97e5701, 0x92287bf, 0x9b802a2, b'_cur_str, _end_str, _header_found, _start, data, key_length, keys, ref_list_length')
 *     __pyx_result = BTreeLeafParser.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BTreeLeafParser__set_state(<BTreeLeafParser> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x97e5701, 0x92287bf, 0x9b802a2, b'_cur_str, _end_str, _header_found, _start, data, key_length, keys, ref_list_length')
 *     __pyx_result = BTreeLeafParser.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_BTreeLeafParser__set_state(<BTreeLeafParser> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = BTreeLeafParser.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BTreeLeafParser__set_state(<BTreeLeafParser> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_BTreeLeafParser__set_state(BTreeLeafParser __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(2, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx___pyx_unpickle_BTreeLeafParser__set_state(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x97e5701, 0x92287bf, 0x9b802a2, b'_cur_str, _end_str, _header_found, _start, data, key_length, keys, ref_list_length')
 *     __pyx_result = BTreeLeafParser.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_BTreeLeafParser__set_state(<BTreeLeafParser> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BTreeLeafParser__set_state(<BTreeLeafParser> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_BTreeLeafParser__set_state(BTreeLeafParser __pyx_result, __pyx_state: tuple):
 *     __pyx_result._cur_str = __pyx_state[0]; __pyx_result._end_str = __pyx_state[1]; __pyx_result._header_found = __pyx_state[2]; __pyx_result._start = __pyx_state[3]; __pyx_result.data = __pyx_state[4]; __pyx_result.key_length = __pyx_state[5]; __pyx_result.keys = __pyx_state[6]; __pyx_result.ref_list_length = __pyx_state[7]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_BTreeLeafParser(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x97e5701, 0x92287bf, 0x9b802a2, b'_cur_str, _end_str, _header_found, _start, data, key_length, keys, ref_list_length')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.__pyx_unpickle_BTreeLeafParser", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_BTreeLeafParser__set_state(<BTreeLeafParser> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BTreeLeafParser__set_state(BTreeLeafParser __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._cur_str = __pyx_state[0]; __pyx_result._end_str = __pyx_state[1]; __pyx_result._header_found = __pyx_state[2]; __pyx_result._start = __pyx_state[3]; __pyx_result.data = __pyx_state[4]; __pyx_result.key_length = __pyx_state[5]; __pyx_result.keys = __pyx_state[6]; __pyx_result.ref_list_length = __pyx_state[7]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 8)
*/

static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx___pyx_unpickle_BTreeLeafParser__set_state(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_BTreeLeafParser__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_BTreeLeafParser__set_state(BTreeLeafParser __pyx_result, __pyx_state: tuple):
 *     __pyx_result._cur_str = __pyx_state[0]; __pyx_result._end_str = __pyx_state[1]; __pyx_result._header_found = __pyx_state[2]; __pyx_result._start = __pyx_state[3]; __pyx_result.data = __pyx_state[4]; __pyx_result.key_length = __pyx_state[5]; __pyx_result.keys = __pyx_state[6]; __pyx_result.ref_list_length = __pyx_state[7]             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 8)
*/
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_v___pyx_result->_cur_str = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_v___pyx_result->_end_str = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_header_found = __pyx_t_3;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_v___pyx_result->_start = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->data);
  __Pyx_DECREF(__pyx_v___pyx_result->data);
  __pyx_v___pyx_result->data = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->key_length = __pyx_t_3;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->keys);
  __Pyx_DECREF(__pyx_v___pyx_result->keys);
  __pyx_v___pyx_result->keys = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->ref_list_length = __pyx_t_3;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_BTreeLeafParser__set_state(BTreeLeafParser __pyx_result, __pyx_state: tuple):
 *     __pyx_result._cur_str = __pyx_state[0]; __pyx_result._end_str = __pyx_state[1]; __pyx_result._header_found = __pyx_state[2]; __pyx_result._start = __pyx_state[3]; __pyx_result.data = __pyx_state[4]; __pyx_result.key_length = __pyx_state[5]; __pyx_result.keys = __pyx_state[6]; __pyx_result.ref_list_length = __pyx_state[7]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 8)             # <<<<<<<<<<<<<<
*/
  __pyx_t_3 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 8); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle_BTreeLeafParser__set_state(<BTreeLeafParser> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BTreeLeafParser__set_state(BTreeLeafParser __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._cur_str = __pyx_state[0]; __pyx_result._end_str = __pyx_state[1]; __pyx_result._header_found = __pyx_state[2]; __pyx_result._start = __pyx_state[3]; __pyx_result.data = __pyx_state[4]; __pyx_result.key_length = __pyx_state[5]; __pyx_result.keys = __pyx_state[6]; __pyx_result.ref_list_length = __pyx_state[7]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 8)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.__pyx_unpickle_BTreeLeafParser__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
/* #### Code section: module_exttypes ### */
static struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser;

static PyObject *__pyx_tp_new_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)o);
  p->__pyx_vtab = __pyx_vtabptr_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser;
  p->data = Py_None; Py_INCREF(Py_None);
  p->keys = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser(PyObject *o) {
  struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *p = (struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->data);
  Py_CLEAR(p->keys);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *p = (struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->data) {
    e = (*v)(p->data, a); if (e) return e;
  }
  if (p->keys) {
    e = (*v)(p->keys, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *p = (struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)o;
  tmp = ((PyObject*)p->data);
  p->data = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->keys);
  p->keys = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser[] = {
  {"parse", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_3parse, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser},
  {Py_tp_doc, (void *)PyDoc_STR("Parse the leaf nodes of a BTree index.\n\n    :ivar data: The PyBytes object containing the uncompressed text for the\n        node.\n    :ivar key_length: An integer describing how many pieces the keys have for\n        this index.\n    :ivar ref_list_length: An integer describing how many references this index\n        contains.\n    :ivar keys: A PyList of keys found in this node.\n\n    :ivar _cur_str: A pointer to the start of the next line to parse\n    :ivar _end_str: A pointer to the end of bytes\n    :ivar _start: Pointer to the location within the current line while\n        parsing.\n    :ivar _header_found: True when we have parsed the header for this node\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser},
  {Py_tp_clear, (void *)__pyx_tp_clear_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser},
  {Py_tp_methods, (void *)__pyx_methods_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser},
  {Py_tp_init, (void *)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser},
  {0, 0},
};
static PyType_Spec __pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser_spec = {
  "breezy.bzr._btree_serializer_pyx.BTreeLeafParser",
  sizeof(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser_slots,
};
#else

static PyTypeObject __pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser = {
  PyVarObject_HEAD_INIT(0, 0)
  "breezy.bzr._btree_serializer_pyx.""BTreeLeafParser", /*tp_name*/
  sizeof(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Parse the leaf nodes of a BTree index.\n\n    :ivar data: The PyBytes object containing the uncompressed text for the\n        node.\n    :ivar key_length: An integer describing how many pieces the keys have for\n        this index.\n    :ivar ref_list_length: An integer describing how many references this index\n        contains.\n    :ivar keys: A PyList of keys found in this node.\n\n    :ivar _cur_str: A pointer to the start of the next line to parse\n    :ivar _end_str: A pointer to the end of bytes\n    :ivar _start: Pointer to the location within the current line while\n        parsing.\n    :ivar _header_found: True when we have parsed the header for this node\n    "), /*tp_doc*/
  __pyx_tp_traverse_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser, /*tp_traverse*/
  __pyx_tp_clear_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode;

static PyObject *__pyx_tp_new_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode;
  p->last_key = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode(PyObject *o) {
  struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *p = (struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->last_key);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *p = (struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->last_key) {
    e = (*v)(p->last_key, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *p = (struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)o;
  tmp = ((PyObject*)p->last_key);
  p->last_key = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_sq_item_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyLong_FromSsize_t(i); if(!x) return 0;
  #if CYTHON_USE_TYPE_SLOTS || (!CYTHON_USE_TYPE_SPECS && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  #else
  r = ((binaryfunc)PyType_GetSlot(Py_TYPE(o), Py_mp_subscript))(o, x);
  #endif
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_min_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7min_key_1__get__(o);
}

static PyObject *__pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_max_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7max_key_1__get__(o);
}

static PyObject *__pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_last_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_1__get__(o);
}

static int __pyx_setprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_last_key(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_3__set__(o, v);
  }
  else {
    return __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_num_records(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_1__get__(o);
}

static int __pyx_setprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_num_records(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_common_shift(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_1__get__(o);
}

static int __pyx_setprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_common_shift(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode[] = {
  {"__sizeof__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_1__sizeof__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"all_keys", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_13all_keys, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"all_items", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_15all_items, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_offset_for_sha1", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_17_get_offset_for_sha1, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_offsets", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_19_get_offsets, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_21__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_23__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode[] = {
  {"min_key", __pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_min_key, 0, 0, 0},
  {"max_key", __pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_max_key, 0, 0, 0},
  {"last_key", __pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_last_key, __pyx_setprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_last_key, 0, 0},
  {"num_records", __pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_num_records, __pyx_setprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_num_records, 0, 0},
  {"common_shift", __pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_common_shift, __pyx_setprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_common_shift, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode},
  {Py_sq_length, (void *)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11__len__},
  {Py_sq_item, (void *)__pyx_sq_item_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode},
  {Py_sq_contains, (void *)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7__contains__},
  {Py_mp_length, (void *)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11__len__},
  {Py_mp_subscript, (void *)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_9__getitem__},
  {Py_tp_doc, (void *)PyDoc_STR("Track all the entries for a given leaf node.")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode},
  {Py_tp_clear, (void *)__pyx_tp_clear_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode},
  {Py_tp_methods, (void *)__pyx_methods_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode},
  {Py_tp_getset, (void *)__pyx_getsets_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode},
  {Py_tp_init, (void *)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_5__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode},
  {0, 0},
};
static PyType_Spec __pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode_spec = {
  "breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode",
  sizeof(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_GCCHKSHA1LeafNode = {
  __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_GCCHKSHA1LeafNode = {
  __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11__len__, /*mp_length*/
  __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_9__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "breezy.bzr._btree_serializer_pyx.""GCCHKSHA1LeafNode", /*tp_name*/
  sizeof(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_GCCHKSHA1LeafNode, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_GCCHKSHA1LeafNode, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Track all the entries for a given leaf node."), /*tp_doc*/
  __pyx_tp_traverse_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*tp_traverse*/
  __pyx_tp_clear_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_5__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser = &__pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser.extract_key = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *, char *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_extract_key;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser.process_line = (int (*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_process_line;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser)) __PYX_ERR(0, 44, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser_spec, __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser) < (0)) __PYX_ERR(0, 44, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser = &__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser) < (0)) __PYX_ERR(0, 44, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser->tp_dictoffset && __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser, __pyx_vtabptr_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser) < (0)) __PYX_ERR(0, 44, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser) < (0)) __PYX_ERR(0, 44, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_BTreeLeafParser, (PyObject *) __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser) < (0)) __PYX_ERR(0, 44, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser) < (0)) __PYX_ERR(0, 44, __pyx_L1_error)
  __pyx_vtabptr_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode = &__pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode._record_to_value_and_refs = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__record_to_value_and_refs;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode._record_to_item = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__record_to_item;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode._lookup_record = (__pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *(*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__lookup_record;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode._count_records = (int (*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *, char *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__count_records;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode._parse_bytes = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, PyObject *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__parse_bytes;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode._parse_one_entry = (char *(*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *, char *, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__parse_one_entry;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode._offset_for_sha1 = (int (*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__offset_for_sha1;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode._compute_common = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__compute_common;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode)) __PYX_ERR(0, 434, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode_spec, __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode) < (0)) __PYX_ERR(0, 434, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode = &__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode) < (0)) __PYX_ERR(0, 434, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode->tp_dictoffset && __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, __pyx_vtabptr_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode) < (0)) __PYX_ERR(0, 434, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode) < (0)) __PYX_ERR(0, 434, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_GCCHKSHA1LeafNode, (PyObject *) __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode) < (0)) __PYX_ERR(0, 434, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode) < (0)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_2_4(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  0, 0,
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__btree_serializer_pyx(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__btree_serializer_pyx},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, __Pyx_FREETHREADING_COMPATIBLE},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "_btree_serializer_pyx",
      __pyx_k_Pyrex_extensions_to_btree_node_p, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit__btree_serializer_pyx(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__btree_serializer_pyx(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
    PY_INT64_T current_id = GraalPyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__btree_serializer_pyx(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_btree_serializer_pyx' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "_btree_serializer_pyx" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
    PyErr_Clear();
    __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
    if (!__Pyx_RefNanny)
        Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  
__Pyx_RefNannySetupContext("PyInit__btree_serializer_pyx", 0);
  __Pyx_init_runtime_version();
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Library function declarations ---*/
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__pyx_module_is_main_breezy__bzr___btree_serializer_pyx) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name, __pyx_mstate_global->__pyx_n_u_main) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "breezy.bzr._btree_serializer_pyx")) {
      if (unlikely((PyDict_SetItemString(modules, "breezy.bzr._btree_serializer_pyx", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  (void)__Pyx_modinit_function_export_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_type_init_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  (void)__Pyx_modinit_function_import_code(__pyx_mstate);
  /*--- Execution code ---*/

  /* "breezy/bzr/_btree_serializer_pyx.pyx":41
 *                             safe_string_from_size)
 * 
 * import sys             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_sys, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_sys, __pyx_t_2) < (0)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":232
 *         return 0
 * 
 *     def parse(self):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t byte_count
 *         if not PyBytes_CheckExact(self.data):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_3parse, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_BTreeLeafParser_parse, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__btree_serializer_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser, __pyx_mstate_global->__pyx_n_u_parse, __pyx_t_2) < (0)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_BTreeLeafParser___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__btree_serializer_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x97e5701, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BTreeLeafParser__set_state(self, __pyx_state)
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_BTreeLeafParser___setstate_cytho, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__btree_serializer_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[2])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":245
 * 
 * 
 * def _parse_leaf_lines(data, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     parser = BTreeLeafParser(data, key_length, ref_list_length)
 *     return parser.parse()
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_1_parse_leaf_lines, 0, __pyx_mstate_global->__pyx_n_u_parse_leaf_lines, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__btree_serializer_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[3])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_parse_leaf_lines, __pyx_t_2) < (0)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":267
 * cdef int _unhexbuf[256]
 * cdef char *_hexbuf
 * _hexbuf = b'0123456789abcdef'             # <<<<<<<<<<<<<<
 * 
 * cdef _populate_unhexbuf():
*/
  __pyx_v_6breezy_3bzr_21_btree_serializer_pyx__hexbuf = ((char *)"0123456789abcdef");

  /* "breezy/bzr/_btree_serializer_pyx.pyx":279
 *     for i from 10 <= i < 16: # ABCDEF => 10, 11, 12, 13, 14, 15, 16
 *         _unhexbuf[(i - 10 + c'A')] = i
 * _populate_unhexbuf()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__populate_unhexbuf(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":306
 * 
 * 
 * def _py_unhexlify(as_hex):             # <<<<<<<<<<<<<<
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if not PyBytes_CheckExact(as_hex) or PyBytes_GET_SIZE(as_hex) != 40:
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_3_py_unhexlify, 0, __pyx_mstate_global->__pyx_n_u_py_unhexlify, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__btree_serializer_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[4])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_py_unhexlify, __pyx_t_2) < (0)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":329
 * 
 * 
 * def _py_hexlify(as_bin):             # <<<<<<<<<<<<<<
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyBytes_CheckExact(as_bin):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_5_py_hexlify, 0, __pyx_mstate_global->__pyx_n_u_py_hexlify, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__btree_serializer_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[5])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_py_hexlify, __pyx_t_2) < (0)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":365
 * 
 * 
 * def _py_key_to_sha1(key):             # <<<<<<<<<<<<<<
 *     """Map a key to a simple sha1 string.
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_7_py_key_to_sha1, 0, __pyx_mstate_global->__pyx_n_u_py_key_to_sha1, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__btree_serializer_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[6])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_py_key_to_sha1, __pyx_t_2) < (0)) __PYX_ERR(0, 365, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":400
 * 
 * 
 * def _py_sha1_to_key(sha1_bin):             # <<<<<<<<<<<<<<
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyBytes_CheckExact(sha1_bin) or PyBytes_GET_SIZE(sha1_bin) != 20:
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_9_py_sha1_to_key, 0, __pyx_mstate_global->__pyx_n_u_py_sha1_to_key, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__btree_serializer_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[7])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_py_sha1_to_key, __pyx_t_2) < (0)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":452
 *     cdef unsigned char offsets[257]
 * 
 *     def __sizeof__(self):             # <<<<<<<<<<<<<<
 *         return (
 *             sizeof(GCCHKSHA1LeafNode) +
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_1__sizeof__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GCCHKSHA1LeafNode___sizeof, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__btree_serializer_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[8])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, __pyx_mstate_global->__pyx_n_u_sizeof, __pyx_t_2) < (0)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":572
 *         return self.num_records
 * 
 *     def all_keys(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_13all_keys, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GCCHKSHA1LeafNode_all_keys, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__btree_serializer_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[9])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, __pyx_mstate_global->__pyx_n_u_all_keys, __pyx_t_2) < (0)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":579
 *         return result
 * 
 *     def all_items(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_15all_items, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GCCHKSHA1LeafNode_all_items, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__btree_serializer_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[10])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, __pyx_mstate_global->__pyx_n_u_all_items, __pyx_t_2) < (0)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":687
 *         return this_offset
 * 
 *     def _get_offset_for_sha1(self, sha1):             # <<<<<<<<<<<<<<
 *         return self._offset_for_sha1(PyBytes_AS_STRING(sha1))
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_17_get_offset_for_sha1, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GCCHKSHA1LeafNode__get_offset_fo, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__btree_serializer_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[11])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, __pyx_mstate_global->__pyx_n_u_get_offset_for_sha1, __pyx_t_2) < (0)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":736
 *             offset = offset + 1
 * 
 *     def _get_offsets(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_19_get_offsets, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GCCHKSHA1LeafNode__get_offsets, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__btree_serializer_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[12])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, __pyx_mstate_global->__pyx_n_u_get_offsets, __pyx_t_2) < (0)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.last_record,self.records cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_21__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GCCHKSHA1LeafNode___reduce_cytho, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__btree_serializer_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[13])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.last_record,self.records cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.last_record,self.records cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_23__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GCCHKSHA1LeafNode___setstate_cyt, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__btree_serializer_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[14])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":744
 * 
 * 
 * def _parse_into_chk(bytes, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     """Parse into a format optimized for chk records."""
 *     assert key_length == 1
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_11_parse_into_chk, 0, __pyx_mstate_global->__pyx_n_u_parse_into_chk, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__btree_serializer_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[15])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_parse_into_chk, __pyx_t_2) < (0)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":751
 * 
 * 
 * def _flatten_node(node, reference_lists):             # <<<<<<<<<<<<<<
 *     """Convert a node into the serialized form.
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_13_flatten_node, 0, __pyx_mstate_global->__pyx_n_u_flatten_node, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__btree_serializer_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[16])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 751, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_flatten_node, __pyx_t_2) < (0)) __PYX_ERR(0, 751, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_BTreeLeafParser(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x97e5701, 0x92287bf, 0x9b802a2, b'_cur_str, _end_str, _header_found, _start, data, key_length, keys, ref_list_length')
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_15__pyx_unpickle_BTreeLeafParser, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_BTreeLeafParser, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__btree_serializer_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[17])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_BTreeLeafParser, __pyx_t_2) < (0)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":1
 * # Copyright (C) 2008, 2009, 2010 Canonical Ltd             # <<<<<<<<<<<<<<
 * #
 * # This program is free software; you can redistribute it and/or modify
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_2) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init breezy.bzr._btree_serializer_pyx", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init breezy.bzr._btree_serializer_pyx");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */
/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);

  /* Cached unbound methods */
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.method_name = &__pyx_mstate->__pyx_n_u_items;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.method_name = &__pyx_mstate->__pyx_n_u_values;
  return 0;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":620
 *         if strncmp(c_bytes, b'type=leaf\n', 10):
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"
 *                              % (data[:10],))             # <<<<<<<<<<<<<<
 *         c_cur = c_bytes + 10
 *         num_records = self._count_records(c_cur, c_end)
*/
  __pyx_mstate_global->__pyx_slice[0] = PySlice_New(Py_None, __pyx_mstate_global->__pyx_int_10, Py_None); if (unlikely(!__pyx_mstate_global->__pyx_slice[0])) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_slice[0]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_slice[0]);
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_slice;
    for (Py_ssize_t i=0; i<1; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  {
    const struct { const unsigned int length: 10; } index[] = {{32},{29},{28},{28},{26},{33},{39},{179},{35},{44},{41},{38},{22},{37},{45},{53},{1},{8},{36},{40},{7},{6},{2},{49},{33},{15},{9},{4},{18},{30},{27},{24},{30},{31},{81},{42},{15},{49},{33},{15},{33},{35},{21},{17},{35},{37},{28},{38},{30},{27},{26},{20},{9},{8},{6},{10},{6},{18},{32},{10},{5},{18},{4},{8},{5},{14},{15},{8},{13},{8},{20},{12},{12},{20},{1},{13},{4},{5},{4},{3},{7},{10},{4},{8},{10},{8},{7},{8},{4},{8},{3},{5},{15},{17},{6},{3},{11},{15},{15},{13},{14},{12},{11},{10},{30},{14},{12},{10},{17},{13},{7},{11},{8},{15},{9},{9},{15},{8},{6},{4},{12},{10},{12},{19},{4},{8},{10},{5},{10},{3},{8},{6},{12},{3},{5},{9},{6},{1},{2},{740},{42},{24},{91},{36},{52},{44},{19},{19},{9},{11},{143},{31},{57},{71},{79},{50}};
    #if (CYTHON_COMPRESS_STRINGS) == 3 && __PYX_LIMITED_VERSION_HEX >= 0x030e0000 /* compression: zstd (2176 bytes) */
const char* const cstring = "(\265/\375`\273\016\265C\000\252^\314\0238\000\3336\374\177\367^\303}k\364\253_u|\215:\276\257\373\366\357\311\335\356L\241f6fB\222+\247\202\245\376\334e\371L\323\320\315\222i\346m\3030\274\014SE\031\013\030\001\032\001A\001\270\262$\355\211.C\351\021t\3229\223\037\3362\"\272IDF\245\303]3W\215$\036\315\274/\377n\212Y\365\245\350\252r6\277|\303\020\337\260X\276\371Yn\235-m\312\236l\276G\362&\31167K\034\356\324\245N\034I}\212[\223\214`\022w\362\374m>~6\203:\233Y\314\246\3156'\311n\245\251\363\224\275\361\034U\032\311\362&1O\226\032\271\210\223\036\257.G\2165\026\341y[\275\263\266g\251\263\342\217\225mN\216\347gw\336\033\223\340wn\273\237B\215-\r\206\206B\264Cd\2659\232:2sl\346o$sh\367huy\343\272\034\313q\346\345PL\304\231\254\3118\035\217\331\033\037\247,71\337a9D\225oV\205f\365\210\271.A\033\216\241\315z\024\nu\266\241Y\202\331\344y\213by\343N:\223\026\n,\311\317\346mb\034\316&\3167=\212\266\344\344\210yL\020+\303Km\363Ib)k\226\242\335\231\353>\033f\016f\034\325N}\3345\270V\354m~C\341s,W\001\200\312V\000V\010\002\260\363>3\307!\356\274wc1\207{\363\361\273T\371\222\370\314}\230KeN\236;\233Q\347\206\002\263>/\256\304\231\030\352\006\336\222\353\361RM\345(\2543y\206\345\305QW\217\345\323\335\343\035w\226\302\267{/\250O\361^~1)^\276\315Q\014\357\205\002\361\227Al\362\013\312d;\353\221\237\245\254\034A\255\333xc1\027q\252<t98\352P\351\210\356\335Q\231\217\227$\361\276\367\362\354$\226\263\232\217Q\220ONN\017\216c\276y\336\367\2027\203gxef\336&&\022q\026\312\331<\302\321g\270\255\306\034g\023$\301#:\222\203\230*\343}F\331KB\201#\356+\211,\346f\226\312\237\005\355>\2712\2149\0167\266+'$yj;\243\361\235\235t&\326\343\207\242\262b\261\314\014\337\266\036\311KC\332p\217;\2338;Z\033U\232b\224S\336\032O]\231H\346mk`\372,\301\033\376\303cH\024TD\374\010\006\317\261\260\306|\330)\374N\237\275P\320_\355\034\204v\013\027\256\025\326\005\272\316\266+\330\005\364\017\247\033-;\030F\302O \320\232\245Hh\r\375P\370\n<S]JWR\210\324D\307\230\027\360\256n\037\376\003\301\302_ Xa\275W\357""\210\267\302\233\261\370x\367\261K\200yr\347P\220\353\341\036\217\3137\261Gj\013\313J2\262\316T\222\264\344c\205\361^\274\030\221OC=\024P\253\261b\036\016/F\035Pu\350\310\006\212\203\306\260\240\036\026k\016\305\350\007\032\263+\314\256\263\224\004\215i\2540\365\245\316:>\315gX\2546\224\301[H\254X\305Qc\322\201T\347\355\004\200\301\026\3668>\246G<\2549\217=\213\317xXk~\241?,\2548\305Ic\320\001T\327-\030\000\\[\270\343t\334Yz\204K\247K\270\264r\273\207\002\000\215\255\027\032\200\352\370\231\317}\242C\024\000Pl\301\372\253\203/\037\340\237j=\326!\020V\334m\335\243C<\2546_\363%\035ba\315\365\234\036\263Sp\2559\254\003\237\261w\016\004\353\254\303z\270kXX\003|.\260\364\225\222X\374W\026\330\003\342\315x3&\177 }\352\315z9+\377b\244\342\343~e\353y\364FH>\210\336r<\314\263A\371?\350\014\365\200\370\251\024\312\360\355\327\001;K\207\331FE\177\250\247\263\362\375\313\010z\323#\010\326\"\333\207\352\315\360\024T\025\373s\276C\303\203}F+k\252\027\304KRk*\016\277'\302\343Q\tT\243\264 \255@]<\327\202\220H\332Ia\364\304\363\020\254TS\177j\377M~\362\366oL\237)\351'\021}\235\254\230\036\347\335\244|Z\252\220\321\217v\r\005\000\256\232*\246\266\024\305\326MT\254\272\200\374\210o3Q\221\220\205\025a\306\203Q\371\t\014\2536U)\225HE*>\255G\240:8\275\215\370\230\336\323_\263\316a8\341\207\002y\002-\234~\244\257\224\304\342\303\322\020\265\241 \023\177\302%{\2410\207O\323\252Qt\227\224\006E\272;\006\254Yn\347\340\272\370V/\325%\205\245%\257\237\220\3534;\205\\?\350\257n\356@j\274\022\373[$`\352J\354\307$\326\2004\307\021CC\317a\261\343g\021\336\254\365t\t\030\032{X\003\320/\201\254\250\001\251r\014\321\214\210\210$)H\322\0301\010\202q\244\263\034o\301<\353\010\221\210H\020\005I\222B\226\003\006e\002`W\252v3\370\364G\364\265\226\356\256B\325\013|\321\224\266o\364\236\245\242=\017\356d\347\267+Z\231\361\265\341\003\031\036\361\206\214=\330\024\177\032\330@\301jq \025\275\236*\332[\345\210m\335D\2272\214\206\205!\017\371\262-\002\037);\224\376\210R\336\314*\316\373\206\266\310\361\315.^""\243r]&\2434P\033\342(\214\r\247\312s'\263\005\336\n9\357\266\354\341\020\267tz\201\235J\331b\037\322\317\265\301\240i\200\371\213\323^\361\3063\344\\J\322\355\247Z\306\354;\2757\276\245\315&L\352\352>\"=8Mb\270\010u\232T\343\372?^;<|y\243\3771\306\001\"\360\334\256\332E\357\336cP\010\206\350\273\253W\356\276\210\220\211\271\244\225\220\315\014\362g\357\\\000\276\320\357&\367\002\237\246\342\231\207\377\003\222z\340\275\021\312\271T\252\362{\325\375\2002\034\303@\037\021`\006\265D\224\211\203k\213$\2022z\032\314r\226\366\216\031\210\2015w\346\354\247/\204Jj9\255$w\325\304^\3721\252\227\000^\245\373[\314\003$\241\362\017\2023L5it\265\016\216\2243\311\031N\315\213\364\301\262\346\004\303\363\274\363b\320\366\317w\253i\333\317W\363+\333<b\320\327\267\307\376\307\013D\355\227A*\233x\347\331\005Yt\2576;\032\372\311\036:\362\311\216\020M\036\024\317\322\237\005\376\202,m\000\361\007#\rXx4\307K\216\030\234\211#\310c\035M\322D\222\022\257Y\311Pf|\264\236\026L\315q~\001\370\303\204\024\362B.\365l\307\271c\351h\245\243\221\201\226\303\332\352\033]\364\317\271Z\013m\372\365s3\320<\013R\242\210^6\256\027\320\210I}\211\0371\231x\250@G\037\017\227*\331K\027\301?\020a\334\215\340>^\231\037\315\302\000*\2631j\364\312\371)\002t,\327\367\032\014\217\026\177!\214Ii}\000)B\316t<?Z\255\005\342L\021*\231\322\306\222\312\262\336\354C~& \332\350U\334<\263\276\227\244\210\3306Z\300\323D\002\032\374J\242\tEn\347\212\210\274\214\220v\262\222~\347L\207!\006\301\362\305\350\220\221)\306\315#$\230\3408\251\212>\333\214\311=\001&\367\225p\207\364P;\274\222be\340\014\005\263\003 O@\307\010\343\213\352\215\270$\312\376ME\351\026z\304\216yJ\332k\2773\017C\026\357\234\0351\355\3202|\022\261\274/\033\350\257\3740\234\236m\364\3268(U.\005\221\006\261Sp\021\371\017i2\301\334h\273e\021F<\266\317g\301^6\2342#\355R\2360\207(\244\375\272\212\373\3435tomIOQ0U\004\225b`\323\334\020\204Q\305\207\314 \215PdaFQ&G&9\347\026=\233p\277\031f \312\273\234g\305\326]s~\367x\341&\271\036\342\373l:\006R\232-u\370NV\020""\203\341\240z\037Z4Xh\202\331:\243W! \3128\224\013U\007\217A\346\014+\034\206\334\243h\3678\311\363\204mHg\331I\260\224\360\236\200\021\2033\367\361\007\014\342\261\303)\222\341*\376";
    PyObject *data = __Pyx_DecompressString(cstring, 2176, 3);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) == 2 /* compression: bz2 (2278 bytes) */
const char* const cstring = "BZh91AY&SY\2134\265\320\000\001\354\177\377\377\377\377\377\377\377\377\377\277\375}\334\277\377\377\364\300@@@@@@@@@@@@\000@\000`\010_x\367\252\333\336\366\334s\335\313n\315\021\024Qu\233\300<\rL\21152\237\250\324\364\332\0051\350\324\236\243\323\320jh&\206\321\223S\324\323z\223A\247\244z\231=OS\023\3246\240=M\017I\352fQ\210%\021\244\3204)\34254\032\232z\024\375\024oThh\321\223@d\000\000\000\000\000h\000\000\rOI\204\246CR4i\240h=@\000\000\031\032\000\001\240\000\000\000\r\032\017P\323 $D@\200A\251\246\r&\241\246\032\203M\000\321\240\310\000\032\000\000\000\000\003@\032 \300\t\200\000\000L\000&\000\000\0010\002h`\000\000\000\000\000\224\"LSd\236\246\230\247\351G\251\241\247\251\352yG\250h\007\250z\231\006A\240\000\000\032\007\250\000\003&\004\331\021\305\263\236W\274\370x\000\376\241\236\032\006\203)\236\216x\364g\000\367\275\377\345\325\023YThM\241\260\006\353E\234_\322\233:\313JK\010.\275\211\264\223m1\261\263\215H\317^\265\250\032L%\261\254\206HV\002\253\010\250\255\214\225\"Lpa\272Z\225\212a\024e\031p\025\035kT\350\352\255\265a\220\304f\313\372\363B\216\024\207\234\223\374fI\"\253ub^\272.3\205vK\266\202X\000d|\345\323\312m\250\275\215pU\262\224\0315A\322I\252\334\202v6\326p\274\225\222\032\332\334\013a\212\256\327\225~\254\222:\000\017\214\347\031\207\364\237\t\317'\360\263\027\371\252\252\024\324\036\256\224\372*\316\252\301\370O\370\314\335\247\367\262\206\241^\242\255\320\361\260\243\317\266\311\302\205\247\311\255\231\322u\345\335j\313\242\376DS/D\271\333\225\023'I\023\326\242E\246\300\225h\213uW\r\005D\216\357z\266V\224`;\214d\3415\262\263\0204\254\233\261E\271\227\034\274l\331\351\362\316\206\236.\351q\035\252\360\250\273\223\242\355\223\0238\204\375\022\034\2263\342I\362\243\2357\205\331\365cn\313\302\263\265\232\244\276\352;eD\2376\220\320\220\273K\215u\243\265\013/\035\033+\302^\273\022c\005\231Z@y\355\266\340Gg3\230\241\207\031\203\365\177\236)C^\303n\234\252*\226\324\223v\341\367\230\206\265\030!\277\307\274\253\275*\231\007\240\005\022p\242\006\020""\302\310\214{\303h\372\222(1&H&%\202\021:\224Mt\nJ\330\357\264\273v;\275KP\224\301\232\254/\021h$\022\003\340\375qV\252\370\362\247\262H\017\2521\274\354h\212\030\326:\202/f\037\252\340\024\352\244l/R\205F\201\037IY\301\270\352\225V(\022\204\370Q\213\200\017X\025\"\265m\251\007\257\231\317\312\003-\332\372\327\325\013=\315\314SN\340\375\340\217\\\014@\247n\354\267@R\240)C\241RJ\\\3601\337\371l\344\202\021z\276\373t\271\310m?2\311\026\250\t\010\347\373>\240<\331\372\377:\025W\225u\327F\006:\236\352(8$O\367\347\314!\241S\036\351\310\035P\205o\000M\344\t{\367\231K\240\314\316v(\243\007\004\\\270\254\242\341+R\351\302\t\215\213\323\031\255e\221+\001\256\261:\023Z\010\273\003\367\305\025\252\014v\250)%\014%\222\221&\322\234|\231\nz\231\257\312\365\257\023\266\023\265\350\202P\350f\262\230\231\242\330/\255h\342(\354Z\n\"\245\320$s,X\213.\214\177X\200\037\367\n\221\327\324J\346,\013/\216*\002RKn4\364\264\277B'\n\n\227\004\354\265\360C\230b\246yN\375)\nD\316\257Q\261\216\215\232\315a\330\032\3716M\010L\020g\275\246Q`\335\010Bkw\215\240\323\251\034\250\274y\003\215\270`\035xRqB\266\332\345T\230\232\304*L\202\224\225$.4\303\361\t\255\024\360\207%\371\2312\254\032\030\355t\026]s6\326{\003z\361Fg\021\000Lod\311\313\325\232\360L\360v\320\340*\251Dn\364\274\276\363kNj\362^,\206\035\2332\366G\022\227N\372\035\257T\362\326Wm\232%,\3340\223\267 \303^M\225\271\260\321\221\211$\225\240{\252'&\004i\305\372\2415\026a\226W\371\225(\212\027\260\337\270%\223T,\244\302G;4%\211\323%\223\002\255\250\307dR\373\022\334\250\255kF\313\264\251Mr\362\210\203\031r\2422c\232)=\344\010\266+z\3349\332\205y2\017\005\265Q\312\rJ\315\201\277\242\001\204\276\254r\0224\233\334\007\355\303\260\206%0\352KJ\366\202Q\030\311\307)\214\215(\235\225\313\310\005d\242\"\372\n\341h@\312\263K\300\232\217n\205_l\304p\360\377xm\232\333\207W\202\216\203\007\t\2506\364#q\352\311\020hf\\4$\302\312\212\244\367#h\247\010\347u\026\3262L\366\235\002Y@\346d\234J\260#jNq\233d)\203\003XZ\262\260\334\355\356\034\251\004\345\222""\016&\345\300\323\226\255\334=\342\272\"\203#\336\250\021\305\023y9\311\2266\306J\2011%B\242\247sX\222\274L*\223\262\226\355\270\356\\\222\227{\rd\307aXt\217\362B\030\344dAg\032\372\230U.=\364\226\030\r]e\262^\322\260(8\222--\262r\311\231\303Ht\245\314V\305\256ne\260\242\267$KT\304r\260`5\301B\341\342\016.\036\216\270\030\347\035\273D\254\334\336A\220u\250\005\313\213\025\3244\345)\277q.k&5&\372\"\320\316b\233\035B\360Xq\271\374\034\335\346\316\222\203\2303\260\241\017\267|Sd8\251a3n\376\235.\033;F\3378\032\222\233\263+\207wcccV\314\214\303\371C;\021\322\310\215\333\215L\234\217T\267)n\224\262K@\317\305\240\246\204\226raK6\364\355\033\265\346\335\026\270\032*\271\316QE\344\333~\266y\251\226ru\271\321h\232\223\215C\207\"\315\037\003\345U\004\350m\217AfUA\215o9z\335\010F\027\025\223F\0255\221\236r\023qr\272\321\264\335l)\rC\2042\016(\342\273c\257\333\323\271\260Y\n\031\303\277L\364\321R\026Z+\t\023\236UK\313\350_\005s\321\022\250\271\367H+\342\265\312\024\205\rF\001\\Nj\002\3704D\302d}1C>\010V\342p\031\326\334QU2Z\252D\032\261\361V\357\331$y\233\010\301\033\330\303%da%\220\315\277+\226\265\233\273\315\210>}YK(a\241\213\260k\224\324\326\332R.\3516\034\020\nR#:\267f\255\251\273Y4\342\034\033$\306\241\2611\215.\223\016fS\202`gftt\235\221B\0346\360\311\245\023,\332H\361.\344\n\350`k\352\347l\361\321y\201b\337\337\346t\332\210\352G!\313\243>\344\272\214\006L\005E\0239\016\230\252\250y\221=\311\366\340gyA\310\036 \301u\213\352\023\245\313\357\200o\205\302G\025\212\3311\225\306\370\320\242r\347E\234\200\236\002R\214\201j\363\017\234\275\262\007\301~\337L\335\374*\275w\214\364n\244\350\314\244\262.o\"\321h\332\322#Z\3529&\2202\033\023\224w\366\013\013\223\222'\"\345^Q<\020\005Broa\333\237C[5+)\324\214E\223\372\357\362 U\3174\246\261\320E\024`\316\2544\r*#;\316\217(\266\351\3260U\222\226\246\207\034:\245i\306\210\246'\356\223\324\003\007<q\327|M\354\311i\245\377\243\210\315-s\374\245>\267\254\330#0\316\2316m\352\323\367\nBHi\375\021)',/^N\321\251\267\233\243\371\334\031\311""\375\345e\3507&S\322FW0\272SP\356(\310\021GZ\234}\376\350\220\013\242\303\241\210q!\364\330\200&B\032\345\350\334\022\346G\304/8\221\323\004A\323\203\237\215B\276K\024\365\031\355\263\313\214U\031\367\332\2161@-\232\203\005\241j_r6\324\0138\220\265\267I\345\034\t\017NZ.m\364\356\365{\017\377\027\246\337\247g\206S\310\314b \303Y\256\351\020+(d\3052\233\2069X\2611B\317}[Ch@\024Z\t\322!b\025X\2528i:@I\2263.\205\363U!\320\243D<\315\022\n\224\210/\020I\007\205I!\024`d\362A2^\023\376.\344\212p\241!\026ik\240";
    PyObject *data = __Pyx_DecompressString(cstring, 2278, 2);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) != 0 /* compression: zlib (2153 bytes) */
const char* const cstring = "x\332}V\315S\023\331\026\207\021\236\031\001!\200\212\243\316\334 >@1\022D\237\250\363\254 \352|X\026Qkx\365\346U\335\272\351\276!-Mw\322\367\006\210\365\252\306e\226w\331\313^\3662\313,\263t\331\313^\362'\370'\274sn7\001!\276*\351\276}\317\367\357\374\316\211/\016j\334\220\334$\345\246\344\202T\\\217\3541\273\301\211\343\312\307\344\226\367\222Y6H\245K*\226\003\357*O\345\314\343\354XXc\236\340\244l\273\306\016\261\271\263-\253\375en\245\"\270<-\363\270\341z&\341\216\371\332r81-\023\303\023\303u$\263\034\262\272L\252\374 \311\360\215k\036+\010\311<I\366-Y%\263\262Y\343?\333\234Uf1\3557\256\344\220,\223\344ySV]\207X\202\230\334\266\312\334c\222\333M0\365,\250\333C%\207l\276\330\274\273\372h\2250(\321\343\037\000\021AD\243l\330L\010@\305\255\220r\303\262%\344\202aD\236\374Z!M\267A\034~T\206\370\312\000`r\010\024\252\361\232g\016$\313\244\345:\024\314-g{\036*\200\232\245\265\307\321\372%\263\005\317\277sw\271\254\202\224\354s\007\252\362\\<V\001)\215\022\010\362[\234p\335/H\262\302=\356\030\200\253\005\271\202\2232O[\2306\256\217\356\221\236l\324\354>\212P\nKd\013\213\232\010\016b\235h\t\320\252\364\272\326p\240\037\266Ui\302\255\353 \232\215Z\315\205V\244\211\352L4\304\316\266\310oa\177 \tj[B\212%(\3578\340*4\035\324z\331\010Tv\033\362\224\276\306\031\232\t\375\025\222\334?m\364\214\231&\2053/{\234\177l\336+\177\364\356\321\262\204\017*\270g1\333\372\310=Zk\036\344\341/A\251\017\205\346{\024\372\2173\377\0304\004+\333\234;\370\3346,\007x\017F;\274\271Dl\327\255Q\303m8\310\240\334\317\320j\273\222?J\231&\003\360\225\301>s\260\\\301\267w\261\271\025\317\335%\310IL\241\346q\001\227\226HB\231`@l\254\363i\342\227\352\034\355\223\223q\"kQe\205\307\372\222\221\225\345\273\032\371\262\3450\257\t\332\333\\\310D\264\232\212p\220\222{\335\270\002q\032\266\275\244\355W\010\023\275\316\350\310&\223\354(Kv\262\251y-\306$i2\273K)\002x\026\304\320\214G\252\301\010\357qO&\264ad3\031F\267\214#\2469V\263\214\035\033i\002eP\310\233\3546\204\266d\030\n'\217\0370CB\252+\313\t\301\027\260""\257\200 \323P.\"\021\264w\003\026\222L\355\364\024!:0\333G}\020@\002 \213\331pz\354;\236\"]i\332\210\365\367\020\3405\320`\023\227\223w\3523O\241d\263apj\350b(=\253\000\223\017\035\222\337V\321k\357\325\363\347\277\374\376\356\227b\001\005\270\327\316\\\234\215\325O\345L\264\276JP\275[\351+\333\346\222&{\231BG(v\342\377j\211\263Rf\333\324\222|\367\033\"`\264\240t\263y\000\177\033\260v\351\033~ \337\362J\317\354H\211\t\344@\362\304D\340\004\204e\242\351\030\226\233\007r\301j\2009\020\311\234\347a\316\363}\347\034\211\222\314\247\246\214\201\303C\301\245\364\230\301\313\314\330\301~Sjb.\311\253by\232\314\311\010\367\276\022zTl\246\247\232\342A\302\0337#\245\225\206c\200y\037\374N\242\245\025\222\006\321*\333\343\307n\223\005gQK\320^i\010\207\206\344\203k9\000\t\374\303\310\351\013\326\212.\205\356\302\357\"\204\246\273\256\331\2609\236\034\266\233\274\371>>\016t\302\230\247\316\025\316\340_\323\216&O\313\221.5\252;\351'\256=\212\276\205\376\366j\260\340jM\232\356y<b\n`\242\313\203O\375\206o\274\206\317\336O\002E\374\30117vDc7\371\202\271j\33029\247H\340\021\367mrj8z\017\360\323\203\222H\367$\356E,\256\336`\366Q\241\351d\32033\322\273\340\007Tw\024~\"\323\027\342p\324\343S\353\272\367\203\323k\317\251>\301\247H\354\261\030\334wz\364R\340\341d\002\241u\231\275\201\244}\206\023q;Zv\307c\251u\222\305\245\007\001\347\005/(m\324\200\253\274!x\317\025\2543\375_0\375\300\214\364A\014\3342?\r~Y\036\030\316|\222\321\310\r\277\344\2638s\261e\251\272?\030\217^l\231*\247\n\361\320xK$\207k><\316\177\252\307\231\221\326\232z\256\352\361\350\244\272\355\027\375R<1\251n\252\222b\361\367\223jV\025\321\315\007\177\320\317\242\033CeU\356\313\345\201\341IuIq\277\340o\004\203A\366\313\365\201\341\211V];<\314L\253\r\255\236\231R\367\025Si\210u\305\276d\006\306fT38\027\254\205\353!\213G'\324\250_\214'\256\371\367}\346\327\343\211)\265\346\257\373\354\313\337\006.\021\277\031\236\013\327\332\353m\026O\317\370cA)\236\271\346\257F?\025\332\331v.\276>\033,\206\205\230\344b2\033\314\005,\250""\30737\203\373\311\341\232\277\026\254\007\350\346\306B\320l\237k\257u\326;,\276N\202\021\264\271\031\254F\363k\235l'\027\317\335\016\027\333\205xa1\276\377\240\375^_\221;\0305z\370\262\313\272up1<\336\262\375\034\242u\2415\327bI\225\247 \216f\236tV:\357\273\331n.\036\032\215F\023\331\320D4A\020\0368\251\214\377\235?\013\305Y\201\027N\206+\341\277\333^\007\004#\321\310\\P\014\376\010\013\207C\027\242\013?\034E\312\265\036$\3067\000'\035\271\254\0065\276\255\341V\351\253\233\246\372N\345\342\314UD|\264\365\032\333=\336\372\3137 .\000ZP/@0qY\231\020\276\030cj\361\350\217\276\006\177\014\233=\025Mi\340\342\351+\300\225a\270\232\376\301\237\363\313\301`<1\033\344NtgZ\275\002`\213\250\350\371S\340\014\200.\370/ \316\314\217\276\031\314\242\350G(\260\036\016\306\323\263\321\354J;\007\330NcM\017\"\202}\273\335)\036{?Y\314\255\226\247\262i\335\352<P\246O\311#\255\247\252\016\034\277\000\344\213\262\267\303\\\370\260\235E\371\177\261\310\350*\336\024\342\3148vh\244U\370\004\266c-\n\214\376#x\032\352\273\207\3522x(\036\002\037W\243\361\\\220E\346 \357\363\272K\231K\321\245[A)0\303\234F2\232Y\014\263\341\\\310\0165\256C\376oA\031J\313\3400\335\206V\232\200N\212\"D\225\252\200\256\263J\267\341\0250\031\320\206\246\370\013\020\245\0322H\001s;\245\003\343\023\217N\251\325\350\312<\342\026N\207\377j\227\200\362\251e\356\2336\211|\t\374\312\366\243NA\023<\321\2153\337\307\243\177\307\302\346\201k\323a\t\257 \277h\n\230\026\335z\330f\355:\202s\241uG\225\360p7,\205\225v\021\240\315\014\014_l\275W\327 \304H\270\032\335{\326\221\335\265\317\033\321\346V\264\365g\364g5\252\332\221-\"\261\037\355\037\000\321!\233\202z\351/A,h\303^k\013\2706\215hL\"!\201N\207\360\220\376\303\340J8\034\276k\017\265_v\346\273\331t\355\214G\343\313pUlou~\357\356\177\006\220\277\272\201\024\017\241eO\224\341'c\242i\250O\327qK\241\370'\277\036\221\345\366H\347Iw'*\275\215\207\246\324\263`A\323@\323\342\037>\244\001\274\213V6\272o\242\267\357R^\034\202P\000\323\036\301z2\200\310\033\021I\250\332[tCc\321\030 ""\030\354\205\300;@>\232\274\031\024\202_\243\273\377\354\260N\375\204\233t\017T\003#Z\304\215\262\320\275\331-}\303\317\371O\177)\026]Y\200._\215V\236w\013\335\342I\266\236p\367!<\027\335y\002M\375\255k|\316\236p\247G \241e\356\177Q\206\000\035";
    PyObject *data = __Pyx_DecompressString(cstring, 2153, 1);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #else /* compression: none (4027 bytes) */
const char* const bytes = "Expected bytes for value not: %rFailed to find the value areaFailed to parse block lengthFailed to parse block offsetFailed to parse record endLine did not contain 40 hex bytesNode did not start with \"type=leaf\": %rNote that Cython is deliberately stricter than PEP-484 and rejects subclasses of builtin types. If you need to pass subclasses then set the 'annotation_typing' directive to False.Something went wrong while parsing.We expect reference bits to be bytes not: %rWe expect references to be tuples not: %rWe expected a tuple() for node not: %sWe failed to unhexlifyWe only support parsing byte strings.With ref_lists, we expected 4 entries not: %sWithout ref_lists, we need at least 3 entries not: %s?add_notebreezy/bzr/_btree_serializer_pyx.pyxbytes did not start with 'type=leaf\\n': disableenablegcinvalid key, loop_counter != self.ref_list_lengthinvalid key, wanted segment from  is not presentisenabledkey last < self._startline did not start with sha1: not a 20-byte binary digestnot a 40-byte hex digestonly 1 null, not 2 as expectedself.data is not a byte string.self.last_record,self.records cannot be converted to a Python object for picklingsha1_bin must be a str of exactly 20 bytes(tree fragment)tried to create a string with an invalid size: %dunexpected reference data presentBTreeLeafParserBTreeLeafParser.__reduce_cython__BTreeLeafParser.__setstate_cython__BTreeLeafParser.parseGCCHKSHA1LeafNodeGCCHKSHA1LeafNode.__reduce_cython__GCCHKSHA1LeafNode.__setstate_cython__GCCHKSHA1LeafNode.__sizeof__GCCHKSHA1LeafNode._get_offset_for_sha1GCCHKSHA1LeafNode._get_offsetsGCCHKSHA1LeafNode.all_itemsGCCHKSHA1LeafNode.all_keys__Pyx_PyDict_NextRefall_itemsall_keysas_binas_bin_shaas_hexasyncio.coroutinesbreezy.bzr._btree_serializer_pyxbyte_countbytescline_in_tracebackdata__dict___dictfirst_ref_listfirst_referenceflat_len_flatten_node__func___get_offset_for_sha1_get_offsets__getstate__have_reference_listsi_is_coroutineitemitemsjoinkeykey_lenkey_lengthline__main____module____name""____new__next_lennodenode_lenoutparse_parse_into_chk_parse_leaf_linesparserpop_py_hexlify_py_key_to_sha1_py_sha1_to_key_py_unhexlify__pyx_checksum__pyx_result__pyx_state__pyx_type__pyx_unpickle_BTreeLeafParser__pyx_vtable____qualname____reduce____reduce_cython____reduce_ex__ref_bitref_bit_lenref_listref_list_lengthref_listsreferencereference_listsrefs_lenresultself__set_name__setdefault__setstate____setstate_cython__sha1sha1_bin__sizeof__statestring_keysys__test__updateuse_setstatevalvaluevalue_lenvalues\000%d\200\001\3600\000\005\010\200t\320\013\035\230Q\230a\330\010\016\210i\220q\230\001\330\014\016\210d\220!\2201\330\004\017\210s\220!\2201\330\004\033\2301\330\004\007\200q\330\010\013\2109\220C\220q\330\014\022\220*\230A\230Q\330\020\022\220#\220Q\220a\330\t\022\220\"\220A\330\010\016\210j\230\001\230\021\330\014\016\210c\220\021\220!\360\026\000\005\022\220\025\220e\2301\230D\240\001\240\021\360\034\000\005\020\210q\330\004\007\200q\340\010\024\220D\230\001\230\021\330\010\023\2203\220a\220q\330\010\013\2109\220B\220a\360\010\000\r\030\220y\240\003\2409\250B\250a\330\014\020\220\014\230A\330\020\033\2303\230a\230q\330\020\023\2209\230B\230a\360\006\000\025 \230y\250\003\2509\260B\260a\330\024\030\230\r\240Q\330\030\033\2304\320\0371\260\021\260!\330\034\"\240)\2501\330 !\330 \"\240$\240a\240q\330\030#\2403\240a\240q\330\030\033\2309\240B\240a\360\006\000\035(\240y\260\003\2609\270B\270a\330\034 \240\013\2501\330 #\2404\320'9\270\021\270!\330$*\250)\2601\330()\33035\260T\270\021\270!\330 +\2509\260B\3206F\300a\300q\360\006\000\005\017\210l\230!\2301\330\004\n\210$\210a\210q\330\004\007\200t\320\013\035\230Q\230a\330\010\016\210i\220q\320\030;\2702\270T\300\021\300!\330\004\014\320\014\035\230Q\230a\330\004\020\320\020 \240\001\240\021\330\004\020\220\010\230\002\230\"\230B\230i\240r\250\022\2502\250Z\260r\270\021\330\004\013\320\013$\240A\240V\2501\340\004\n\320\n\032\230!\2301\330\004\n\210!\2105\320\020 \240\001\240\035\250a\330\004\n\210$\210b\220\001\330""\004\007\200q\210\005\210Q\330\004\n\210$\210b\220\001\330\004\007\200y\220\002\220!\330\010\031\230\021\330\010\014\210L\230\001\330\014\017\210\177\230c\240\021\330\020\023\2201\220E\230\021\330\020\026\220d\230\"\230A\330\014\035\230Q\330\014\036\230a\330\014\020\220\r\230Q\330\020\023\320\023#\2403\240a\330\024\027\220q\230\005\230Q\330\024\032\230$\230b\240\001\330\020\"\240!\330\020\033\2303\230a\230q\330\020\024\220G\2309\240A\330\024\027\220r\230\023\230A\330\030\033\2301\230E\240\021\330\030\036\230d\240\"\240A\330\024\036\230i\240q\250\001\330\024\"\320\"2\260!\2601\330\024\032\230!\2305\320 1\260\021\260*\270A\330\024\032\230$\230b\240\001\330\004\007\200q\210\005\210Q\330\004\n\210%\210r\220\021\330\004\n\210!\2105\220\007\220q\330\004\n\210$\210b\220\001\330\004\007\200q\210\005\210Q\330\004\013\210<\220q\200\001\360\n\000\005\022\320\021*\250!\2506\260\021\330\004\007\200|\2201\220E\320\031*\250!\2501\330\010\017\210q\330\004\013\2101\200\001\330\004\r\210_\230A\230V\240<\250q\330\004\013\2106\220\026\220q\200A\340\010\013\2104\320\017!\240\021\240$\240a\330\014\022\220.\240\001\240\021\330\010\025\320\025%\240Q\240d\250!\330\010\014\210L\320\030)\250\021\250$\250a\340\010\014\210L\230\004\230J\240b\250\001\330\010\016\210d\220*\230B\230d\240!\330\014\020\220\r\230Q\330\010\017\210t\2201\200A\340\010\021\220\021\330\010\014\210G\2209\230A\330\014\031\230\021\230(\240$\240h\250a\250q\330\010\017\210q\200A\340\010\021\220\021\330\010\014\210G\2209\230D\240\001\330\014\023\2204\320\027'\240q\250\001\250\024\250X\260Q\260a\330\014\031\230\021\230(\240!\330\010\017\210q\200A\340\010\021\220\021\330\010\014\210G\2209\230D\240\001\330\014\031\230\021\230(\240,\250a\250t\2608\2701\270B\270a\330\010\017\210q\200A\330\010\t\330\014&\240a\330\014'\240r\250\024\250Q\200A\330\010\017\210t\320\023$\240A\320%6\260a\260q\200\001\330\004\n\210+\220Q\200\001\330\004-\250Q\250f\260A\200\001\360\010\000\005\016\210T\220\033\230D\240\013\2504\320/?\270t\3009\310D\320PW""\320W[\320[h\320hl\320ls\320sw\320wx\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220t\2306\240\027\250\005\250S\260\004\260F\270'\300\021\330\004\007\200q\330\010\017\320\0170\260\004\260A\260W\270K\300w\310a\340\010\017\320\0170\260\004\260A\260W\270K\300q\200\001\340\004\013\210;\220c\230\021\330\004\013\320\013\033\2303\230a\330\004\013\320\013\034\230A\230Q\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\220?\240(\250!\2501\330\004\007\200|\2207\230!\330\0101\260\021\3202D\300N\320RS\330\004\013\2101\200\001\340\004\007\200s\210!\2108\2203\220c\230\023\230D\320 2\260!\2601\330\010\016\210j\230\001\230\021\330\004\r\320\r&\240a\240v\250Q\330\004\021\220\021\320\022#\2401\240I\320->\270a\270q\330\004\013\2101\200\001\340\004\007\200t\320\013\035\230Q\230h\240c\320)9\270\021\270(\300#\300Q\330\010\016\210j\230\001\230\021\330\004\r\320\r&\240a\240v\250Q\330\004\007\200\177\220a\320\027(\250\001\250\031\3202C\3001\300A\330\010\017\210q\330\004\013\2101\200\001\340\004\007\200t\320\013\035\230Q\230j\250\003\320+;\2701\270J\300c\310\021\330\010\016\210j\230\001\230\021\330\004\013\210<\220q\320\030)\250\021\250!";
    PyObject *data = NULL;
    CYTHON_UNUSED_VAR(__Pyx_DecompressString);
    #endif
    PyObject **stringtab = __pyx_mstate->__pyx_string_tab;
    Py_ssize_t pos = 0;
    for (int i = 0; i < 137; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyUnicode_DecodeUTF8(bytes + pos, bytes_length, NULL);
      if (likely(string) && i >= 39) PyUnicode_InternInPlace(&string);
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
      stringtab[i] = string;
      pos += bytes_length;
    }
    for (int i = 137; i < 156; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyBytes_FromStringAndSize(bytes + pos, bytes_length);
      stringtab[i] = string;
      pos += bytes_length;
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    Py_XDECREF(data);
    for (Py_ssize_t i = 0; i < 156; i++) {
      if (unlikely(PyObject_Hash(stringtab[i]) == -1)) {
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    #if CYTHON_IMMORTAL_CONSTANTS
    {
      PyObject **table = stringtab + 137;
      for (Py_ssize_t i=0; i<19; ++i) {
        #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
        #if PY_VERSION_HEX < 0x030E0000
        if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
        #else
        if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
        #endif
        {
          Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
        }
        #else
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
        #endif
      }
    }
    #endif
  }
  {
    PyObject **numbertab = __pyx_mstate->__pyx_number_tab + 0;
    int8_t const cint_constants_1[] = {0,1,10};
    int32_t const cint_constants_4[] = {159274753L};
    int64_t const cint_constants_8[] = {2147483648LL,4294967295LL};
    for (int i = 0; i < 6; i++) {
      numbertab[i] = PyLong_FromLongLong((i < 3 ? cint_constants_1[i - 0] : (i < 4 ? cint_constants_4[i - 3] : cint_constants_8[i - 4])));
      if (unlikely(!numbertab[i])) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_number_tab;
    for (Py_ssize_t i=0; i<6; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */
typedef struct {
    unsigned int argcount : 2;
    unsigned int num_posonly_args : 1;
    unsigned int num_kwonly_args : 1;
    unsigned int nlocals : 5;
    unsigned int flags : 10;
    unsigned int first_line : 10;
} __Pyx_PyCode_New_function_description;
/* NewCodeObj.proto */
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
);


static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  PyObject* tuple_dedup_map = PyDict_New();
  if (unlikely(!tuple_dedup_map)) return -1;
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 232};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_byte_count};
    __pyx_mstate_global->__pyx_codeobj_tab[0] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__btree_serializer_pyx_2, __pyx_mstate->__pyx_n_u_parse, __pyx_mstate->__pyx_kp_b_iso88591_A_4_a_Qd_L_a_L_Jb_d_Bd_Q_t1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[0])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[1] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_T_D_4_t9DPWW_hhllsswwx_G1F_a_vW, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[1])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[2] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_QfA, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[2])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 245};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_data, __pyx_mstate->__pyx_n_u_key_length, __pyx_mstate->__pyx_n_u_ref_list_length, __pyx_mstate->__pyx_n_u_parser};
    __pyx_mstate_global->__pyx_codeobj_tab[3] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__btree_serializer_pyx_2, __pyx_mstate->__pyx_n_u_parse_leaf_lines, __pyx_mstate->__pyx_kp_b_iso88591_AV_q_6_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[3])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 306};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_as_hex, __pyx_mstate->__pyx_n_u_as_bin};
    __pyx_mstate_global->__pyx_codeobj_tab[4] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__btree_serializer_pyx_2, __pyx_mstate->__pyx_n_u_py_unhexlify, __pyx_mstate->__pyx_kp_b_iso88591_t_Qhc_9_Q_j_avQ_a_2C1A_q_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[4])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 329};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_as_bin, __pyx_mstate->__pyx_n_u_as_hex};
    __pyx_mstate_global->__pyx_codeobj_tab[5] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__btree_serializer_pyx_2, __pyx_mstate->__pyx_n_u_py_hexlify, __pyx_mstate->__pyx_kp_b_iso88591_s_83c_D_2_1_j_avQ_1I_aq_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[5])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 365};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_as_bin_sha};
    __pyx_mstate_global->__pyx_codeobj_tab[6] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__btree_serializer_pyx_2, __pyx_mstate->__pyx_n_u_py_key_to_sha1, __pyx_mstate->__pyx_kp_b_iso88591_6_1E_1_q_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[6])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 400};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_sha1_bin};
    __pyx_mstate_global->__pyx_codeobj_tab[7] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__btree_serializer_pyx_2, __pyx_mstate->__pyx_n_u_py_sha1_to_key, __pyx_mstate->__pyx_kp_b_iso88591_t_Qj_1Jc_j_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[7])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 452};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[8] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__btree_serializer_pyx_2, __pyx_mstate->__pyx_n_u_sizeof, __pyx_mstate->__pyx_kp_b_iso88591_A_a_r_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[8])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 572};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_i, __pyx_mstate->__pyx_n_u_result};
    __pyx_mstate_global->__pyx_codeobj_tab[9] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__btree_serializer_pyx_2, __pyx_mstate->__pyx_n_u_all_keys, __pyx_mstate->__pyx_kp_b_iso88591_A_G9D_at81Ba_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[9])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 579};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_i, __pyx_mstate->__pyx_n_u_result, __pyx_mstate->__pyx_n_u_item};
    __pyx_mstate_global->__pyx_codeobj_tab[10] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__btree_serializer_pyx_2, __pyx_mstate->__pyx_n_u_all_items, __pyx_mstate->__pyx_kp_b_iso88591_A_G9D_4_q_XQa_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[10])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 687};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_sha1_2};
    __pyx_mstate_global->__pyx_codeobj_tab[11] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__btree_serializer_pyx_2, __pyx_mstate->__pyx_n_u_get_offset_for_sha1, __pyx_mstate->__pyx_kp_b_iso88591_A_t_A_6aq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[11])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 736};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_i, __pyx_mstate->__pyx_n_u_result};
    __pyx_mstate_global->__pyx_codeobj_tab[12] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__btree_serializer_pyx_2, __pyx_mstate->__pyx_n_u_get_offsets, __pyx_mstate->__pyx_kp_b_iso88591_A_G9A_haq_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[12])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[13] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[13])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[14] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[14])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 744};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_bytes, __pyx_mstate->__pyx_n_u_key_length, __pyx_mstate->__pyx_n_u_ref_list_length};
    __pyx_mstate_global->__pyx_codeobj_tab[15] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__btree_serializer_pyx_2, __pyx_mstate->__pyx_n_u_parse_into_chk, __pyx_mstate->__pyx_kp_b_iso88591_c_3a_AQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[15])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 22, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 751};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_node, __pyx_mstate->__pyx_n_u_reference_lists, __pyx_mstate->__pyx_n_u_have_reference_lists, __pyx_mstate->__pyx_n_u_flat_len, __pyx_mstate->__pyx_n_u_key_len, __pyx_mstate->__pyx_n_u_node_len, __pyx_mstate->__pyx_n_u_value, __pyx_mstate->__pyx_n_u_value_len, __pyx_mstate->__pyx_n_u_out, __pyx_mstate->__pyx_n_u_refs_len, __pyx_mstate->__pyx_n_u_next_len, __pyx_mstate->__pyx_n_u_first_ref_list, __pyx_mstate->__pyx_n_u_first_reference, __pyx_mstate->__pyx_n_u_i, __pyx_mstate->__pyx_n_u_ref_bit_len, __pyx_mstate->__pyx_n_u_string_key, __pyx_mstate->__pyx_n_u_ref_lists, __pyx_mstate->__pyx_n_u_ref_list, __pyx_mstate->__pyx_n_u_reference, __pyx_mstate->__pyx_n_u_ref_bit, __pyx_mstate->__pyx_n_u_val, __pyx_mstate->__pyx_n_u_line};
    __pyx_mstate_global->__pyx_codeobj_tab[16] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__btree_serializer_pyx_2, __pyx_mstate->__pyx_n_u_flatten_node, __pyx_mstate->__pyx_kp_b_iso88591_0_t_Qa_iq_d_1_s_1_1_q_9Cq_AQ_Qa, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[16])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[17] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_pyx_unpickle_BTreeLeafParser, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_1_7_1_2DNRS_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[17])) goto bad;
  }
  Py_DECREF(tuple_dedup_map);
  return 0;
  bad:
  Py_DECREF(tuple_dedup_map);
  return -1;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* AssertionsEnabled.init */
  if (likely(__Pyx_init_assertions_enabled() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CommonTypesMetaclass.init */
  if (likely(__pyx_CommonTypesMetaclass_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CachedMethodType.init */
  #if CYTHON_COMPILING_IN_LIMITED_API
  {
      PyObject *typesModule=NULL;
      typesModule = PyImport_ImportModule("types");
      if (typesModule) {
          __pyx_mstate_global->__Pyx_CachedMethodType = PyObject_GetAttrString(typesModule, "MethodType");
          Py_DECREF(typesModule);
      }
  } // error handling follows
  #endif
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CythonFunctionShared.init */
  if (likely(__pyx_CyFunction_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectCall (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
          #if CYTHON_COMPILING_IN_LIMITED_API
            return PyObject_Vectorcall(func, args, _nargs, NULL);
          #else
            vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
            if (f) {
                return f(func, args, _nargs, NULL);
            }
          #endif
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyErrFetchRestore (used by RaiseException) */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* RaiseException */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
#elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}

/* TupleAndListFromArray (used by fastcall) */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < (0))) {
            Py_DECREF(res);
            return NULL;
        }
        Py_INCREF(src[i]);
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL ||\
        !(CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS)
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length, length2;
        int kind;
        void *data1, *data2;
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        #endif
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length < 0)) return -1;
        #endif
        length2 = __Pyx_PyUnicode_GET_LENGTH(s2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length2 < 0)) return -1;
        #endif
        if (length != length2) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    return (equals == Py_EQ);
return_ne:
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = __Pyx_PyTuple_GET_SIZE(kwnames);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(n == -1)) return NULL;
    #endif
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        if (s == namei) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        int eq = __Pyx_PyUnicode_Equals(s, namei, Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs;
    PyObject *dict;
#if !CYTHON_ASSUME_SAFE_SIZE
    nkwargs = PyTuple_Size(kwnames);
    if (unlikely(nkwargs < 0)) return NULL;
#else
    nkwargs = PyTuple_GET_SIZE(kwnames);
#endif
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
#if !CYTHON_ASSUME_SAFE_MACROS
        PyObject *key = PyTuple_GetItem(kwnames, i);
        if (!key) goto bad;
#else
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
#endif
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* PyObjectCallOneArg (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetAttrStr (used by UnpackUnboundCMethod) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* UnpackUnboundCMethod (used by CallUnboundCMethod0) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
#elif CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject **args, Py_ssize_t nargs, PyObject *kwnames) {
        return _PyObject_Vectorcall
            (method, args ? args+1 : NULL, nargs ? nargs-1 : 0, kwnames);
}
#else
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
    return
#if PY_VERSION_HEX < 0x03090000
    _PyObject_Vectorcall
#else
    PyObject_Vectorcall
#endif
        (method, args ? args+1 : NULL, nargs ? (size_t) nargs-1 : 0, kwnames);
}
#endif
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
     METH_VARARGS | METH_KEYWORDS,
#else
     METH_FASTCALL | METH_KEYWORDS,
#endif
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method, *result=NULL;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    result = method;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            result = unbound_method;
        }
    }
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    if (unlikely(target->method)) {
        Py_DECREF(result);
    } else
#endif
    target->method = result;
    return 0;
}

/* CallUnboundCMethod0 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        if (likely(cfunc->flag == METH_NOARGS))
            return __Pyx_CallCFunction(cfunc, self, NULL);
        if (likely(cfunc->flag == METH_FASTCALL))
            return __Pyx_CallCFunctionFast(cfunc, self, NULL, 0);
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, NULL, 0, NULL);
        if (likely(cfunc->flag == (METH_VARARGS | METH_KEYWORDS)))
            return __Pyx_CallCFunctionWithKeywords(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple, NULL);
        if (cfunc->flag == METH_VARARGS)
            return __Pyx_CallCFunction(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple);
        return __Pyx__CallUnboundCMethod0(cfunc, self);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod0(&tmp_cfunc, self);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod0(cfunc, self);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *result;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
    result = __Pyx_PyObject_CallOneArg(cfunc->method, self);
    return result;
}

/* py_dict_items (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_items, d);
}

/* py_dict_values (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_values, d);
}

/* OwnedDictNext (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue) {
    PyObject *next = NULL;
    if (!*ppos) {
        if (pvalue) {
            PyObject *dictview = pkey ? __Pyx_PyDict_Items(p) : __Pyx_PyDict_Values(p);
            if (unlikely(!dictview)) goto bad;
            *ppos = PyObject_GetIter(dictview);
            Py_DECREF(dictview);
        } else {
            *ppos = PyObject_GetIter(p);
        }
        if (unlikely(!*ppos)) goto bad;
    }
    next = PyIter_Next(*ppos);
    if (!next) {
        if (PyErr_Occurred()) goto bad;
        return 0;
    }
    if (pkey && pvalue) {
        *pkey = __Pyx_PySequence_ITEM(next, 0);
        if (unlikely(*pkey)) goto bad;
        *pvalue = __Pyx_PySequence_ITEM(next, 1);
        if (unlikely(*pvalue)) goto bad;
        Py_DECREF(next);
    } else if (pkey) {
        *pkey = next;
    } else {
        assert(pvalue);
        *pvalue = next;
    }
    return 1;
  bad:
    Py_XDECREF(next);
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
    PyErr_FormatUnraisable("Exception ignored in __Pyx_PyDict_NextRef");
#else
    PyErr_WriteUnraisable(__pyx_mstate_global->__pyx_n_u_Pyx_PyDict_NextRef);
#endif
    if (pkey) *pkey = NULL;
    if (pvalue) *pvalue = NULL;
    return 0;
}
#else // !CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue) {
    int result = PyDict_Next(p, ppos, pkey, pvalue);
    if (likely(result == 1)) {
        if (pkey) Py_INCREF(*pkey);
        if (pvalue) Py_INCREF(*pvalue);
    }
    return result;
}
#endif

/* RaiseDoubleKeywords (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, args, 2);
        }
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, 2, NULL);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod2(&tmp_cfunc, self, arg1, arg2);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *result = NULL;
        PyObject *args = PyTuple_New(2);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
        return result;
    }
#endif
    {
        PyObject *args[4] = {NULL, self, arg1, arg2};
        return __Pyx_PyObject_FastCall(cfunc->method, args+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
    }
}

/* ParseKeywordsImpl (used by ParseKeywords) */
static int __Pyx_ValidateDuplicatePosArgs(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char* function_name)
{
    PyObject ** const *name = argnames;
    while (name != first_kw_arg) {
        PyObject *key = **name;
        int found = PyDict_Contains(kwds, key);
        if (unlikely(found)) {
            if (found == 1) __Pyx_RaiseDoubleKeywordsError(function_name, key);
            goto bad;
        }
        name++;
    }
    return 0;
bad:
    return -1;
}
#if CYTHON_USE_UNICODE_INTERNALS
static CYTHON_INLINE int __Pyx_UnicodeKeywordsEqual(PyObject *s1, PyObject *s2) {
    int kind;
    Py_ssize_t len = PyUnicode_GET_LENGTH(s1);
    if (len != PyUnicode_GET_LENGTH(s2)) return 0;
    kind = PyUnicode_KIND(s1);
    if (kind != PyUnicode_KIND(s2)) return 0;
    const void *data1 = PyUnicode_DATA(s1);
    const void *data2 = PyUnicode_DATA(s2);
    return (memcmp(data1, data2, (size_t) len * (size_t) kind) == 0);
}
#endif
static int __Pyx_MatchKeywordArg_str(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    #if CYTHON_USE_UNICODE_INTERNALS
    Py_hash_t key_hash = ((PyASCIIObject*)key)->hash;
    if (unlikely(key_hash == -1)) {
        key_hash = PyObject_Hash(key);
        if (unlikely(key_hash == -1))
            goto bad;
    }
    #endif
    name = first_kw_arg;
    while (*name) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (key_hash == ((PyASCIIObject*)name_str)->hash && __Pyx_UnicodeKeywordsEqual(name_str, key)) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) {
                *index_found = (size_t) (name - argnames);
                return 1;
            }
        }
        #endif
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (unlikely(key_hash == ((PyASCIIObject*)name_str)->hash)) {
            if (__Pyx_UnicodeKeywordsEqual(name_str, key))
                goto arg_passed_twice;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            if (unlikely(name_str == key)) goto arg_passed_twice;
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) goto arg_passed_twice;
        }
        #endif
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
bad:
    return -1;
}
static int __Pyx_MatchKeywordArg_nostr(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    if (unlikely(!PyUnicode_Check(key))) goto invalid_keyword_type;
    name = first_kw_arg;
    while (*name) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (cmp == 1) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        if (unlikely(cmp == -1)) goto bad;
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (unlikely(cmp != 0)) {
            if (cmp == 1) goto arg_passed_twice;
            else goto bad;
        }
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
bad:
    return -1;
}
static CYTHON_INLINE int __Pyx_MatchKeywordArg(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    return likely(PyUnicode_CheckExact(key)) ?
        __Pyx_MatchKeywordArg_str(key, argnames, first_kw_arg, index_found, function_name) :
        __Pyx_MatchKeywordArg_nostr(key, argnames, first_kw_arg, index_found, function_name);
}
static void __Pyx_RejectUnknownKeyword(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char *function_name)
{
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos = NULL;
    #else
    Py_ssize_t pos = 0;
    #endif
    PyObject *key = NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(kwds);
    while (
        #if CYTHON_AVOID_BORROWED_REFS
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL)
        #else
        PyDict_Next(kwds, &pos, &key, NULL)
        #endif
    ) {
        PyObject** const *name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (!*name) {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp != 1) {
                if (cmp == 0) {
                    PyErr_Format(PyExc_TypeError,
                        "%s() got an unexpected keyword argument '%U'",
                        function_name, key);
                }
                #if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(key);
                #endif
                break;
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        #endif
    }
    __Pyx_END_CRITICAL_SECTION();
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(pos);
    #endif
    assert(PyErr_Occurred());
}
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t extracted = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    name = first_kw_arg;
    while (*name && num_kwargs > extracted) {
        PyObject * key = **name;
        PyObject *value;
        int found = 0;
        #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        found = PyDict_GetItemRef(kwds, key, &value);
        #else
        value = PyDict_GetItemWithError(kwds, key);
        if (value) {
            Py_INCREF(value);
            found = 1;
        } else {
            if (unlikely(PyErr_Occurred())) goto bad;
        }
        #endif
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            extracted++;
        }
        name++;
    }
    if (num_kwargs > extracted) {
        if (ignore_unknown_kwargs) {
            if (unlikely(__Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name) == -1))
                goto bad;
        } else {
            __Pyx_RejectUnknownKeyword(kwds, argnames, first_kw_arg, function_name);
            goto bad;
        }
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t len;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    if (PyDict_Update(kwds2, kwds) < 0) goto bad;
    name = first_kw_arg;
    while (*name) {
        PyObject *key = **name;
        PyObject *value;
#if !CYTHON_COMPILING_IN_LIMITED_API && (PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop))
        int found = PyDict_Pop(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
        }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        int found = PyDict_GetItemRef(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            if (unlikely(PyDict_DelItem(kwds2, key) < 0)) goto bad;
        }
#else
    #if CYTHON_COMPILING_IN_CPYTHON
        value = _PyDict_Pop(kwds2, key, kwds2);
    #else
        value = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, kwds2, key, kwds2);
    #endif
        if (value == kwds2) {
            Py_DECREF(value);
        } else {
            if (unlikely(!value)) goto bad;
            values[name-argnames] = value;
        }
#endif
        name++;
    }
    len = PyDict_Size(kwds2);
    if (len > 0) {
        return __Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name);
    } else if (unlikely(len == -1)) {
        goto bad;
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject *key = NULL;
    PyObject** const * name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    for (Py_ssize_t pos = 0; pos < num_kwargs; pos++) {
#if CYTHON_AVOID_BORROWED_REFS
        key = __Pyx_PySequence_ITEM(kwds, pos);
#else
        key = __Pyx_PyTuple_GET_ITEM(kwds, pos);
#endif
#if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!key)) goto bad;
#endif
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            PyObject *value = kwvalues[pos];
            values[name-argnames] = __Pyx_NewRef(value);
        } else {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp == 1) {
                PyObject *value = kwvalues[pos];
                values[index_found] = __Pyx_NewRef(value);
            } else {
                if (unlikely(cmp == -1)) goto bad;
                if (kwds2) {
                    PyObject *value = kwvalues[pos];
                    if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
                } else if (!ignore_unknown_kwargs) {
                    goto invalid_keyword;
                }
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        key = NULL;
        #endif
    }
    return 0;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    goto bad;
bad:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(key);
    #endif
    return -1;
}

/* ParseKeywords */
static int __Pyx_ParseKeywords(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds)))
        return __Pyx_ParseKeywordsTuple(kwds, kwvalues, argnames, kwds2, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
    else if (kwds2)
        return __Pyx_ParseKeywordDictToDict(kwds, argnames, kwds2, values, num_pos_args, function_name);
    else
        return __Pyx_ParseKeywordDict(kwds, argnames, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RejectKeywords */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds) {
    PyObject *key = NULL;
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds))) {
        key = __Pyx_PySequence_ITEM(kwds, 0);
    } else {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *pos = NULL;
#else
        Py_ssize_t pos = 0;
#endif
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
        if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return;
#endif
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL);
#if CYTHON_AVOID_BORROWED_REFS
        Py_XDECREF(pos);
#endif
    }
    if (likely(key)) {
        PyErr_Format(PyExc_TypeError,
            "%s() got an unexpected keyword argument '%U'",
            function_name, key);
        Py_DECREF(key);
    }
}

/* PyErrExceptionMatches (used by GetAttr3) */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* GetAttr3 */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    int res = PyObject_GetOptionalAttr(o, n, &r);
    return (res != 0) ? r : __Pyx_NewRef(d);
#else
  #if CYTHON_USE_TYPE_SLOTS
    if (likely(PyUnicode_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
  #endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
#endif
}

/* PyObjectGetAttrStrNoError (used by GetBuiltinName) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName (used by GetModuleGlobalName) */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* PyDictVersioning (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* PyObjectFastCallMethod */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf) {
    PyObject *result;
    PyObject *attr = PyObject_GetAttr(args[0], name);
    if (unlikely(!attr))
        return NULL;
    result = __Pyx_PyObject_FastCall(attr, args+1, nargsf - 1);
    Py_DECREF(attr);
    return result;
}
#endif

/* DivInt[long] */
static CYTHON_INLINE long __Pyx_div_long(long a, long b, int b_is_constant) {
    long q = a / b;
    long r = a - q*b;
    long adapt_python = (b_is_constant ?
        ((r != 0) & ((r < 0) ^ (b < 0))) :
        ((r != 0) & ((r ^ b) < 0))
    );
    return q - adapt_python;
}

/* PyObjectFormatAndDecref */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result;
    if (unlikely(!s)) return NULL;
    result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    if (max_char > 1114111) max_char = 1114111;
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - result_ulength < 0))
        goto overflow;
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = values[i];
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (__Pyx_PyUnicode_READY(uval) == (-1))
            goto bad;
        #endif
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(ulength < 0)) goto bad;
        #endif
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    Py_ssize_t i;
    PyObject *result = NULL;
    PyObject *value_tuple = PyTuple_New(value_count);
    if (unlikely(!value_tuple)) return NULL;
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    for (i=0; i<value_count; i++) {
        if (__Pyx_PyTuple_SET_ITEM(value_tuple, i, values[i]) != (0)) goto bad;
        Py_INCREF(values[i]);
    }
    result = PyUnicode_Join(__pyx_mstate_global->__pyx_empty_unicode, value_tuple);
bad:
    Py_DECREF(value_tuple);
    return result;
#endif
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
    __Pyx_TypeName obj_type_name;
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyLong_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyLong_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is unsliceable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
bad:
    return NULL;
}

/* PyLongCompare */
static CYTHON_INLINE int __Pyx_PyLong_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 1;
    }
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) == 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 0;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 0;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal == 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = __Pyx_PyFloat_AS_DOUBLE(op1);
        return ((double)a == (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyObjectCall2Args (used by PyObjectCallMethod1) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod (used by PyObjectCallMethod1) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(tp);
    PyErr_Format(PyExc_AttributeError,
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}
#endif

/* PyObjectCallMethod1 (used by StringJoin) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* StringJoin */
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    (void) __Pyx_PyObject_CallMethod1;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030e0000 || defined(PyBytes_Join)
    return PyBytes_Join(sep, values);
#elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000 || defined(_PyBytes_Join)
    return _PyBytes_Join(sep, values);
#else
    return __Pyx_PyObject_CallMethod1(sep, __pyx_mstate_global->__pyx_n_u_join, values);
#endif
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS)) {
        return __Pyx_PyList_GetItemRefFast(o, wrapped_i, unsafe_shared);
    } else
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyList_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyTuple_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)) {
            return __Pyx_PyList_GetItemRefFast(o, n, unsafe_shared);
        } else if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            return __Pyx_NewRef(PyList_GET_ITEM(o, n));
        }
    } else
    #if !CYTHON_AVOID_BORROWED_REFS
    if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            return __Pyx_NewRef(PyTuple_GET_ITEM(o, n));
        }
    } else
    #endif
#endif
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
    {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (!is_list && mm && mm->mp_subscript) {
            PyObject *r, *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (is_list || likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    (void)wraparound;
    (void)boundscheck;
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
}

/* RaiseUnboundLocalError */
static void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* ArgTypeTestFunc (used by ArgTypeTest) */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    PyObject *extra_info = __pyx_mstate_global->__pyx_empty_unicode;
    int from_annotation_subclass = 0;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (!exact) {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    } else if (exact == 2) {
        if (__Pyx_TypeCheck(obj, type)) {
            from_annotation_subclass = 1;
            extra_info = __pyx_mstate_global->__pyx_kp_u_Note_that_Cython_is_deliberately;
        }
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")"
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        "%s%U"
#endif
        , name, type_name, obj_type_name
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        , (from_annotation_subclass ? ". " : ""), extra_info
#endif
        );
#if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    if (exact == 2 && from_annotation_subclass) {
        PyObject *res;
        PyObject *vargs[2];
        vargs[0] = PyErr_GetRaisedException();
        vargs[1] = extra_info;
        res = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_kp_u_add_note, vargs, 2, NULL);
        Py_XDECREF(res);
        PyErr_SetRaisedException(vargs[0]);
    }
#endif
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* AllocateExtensionType */
static PyObject *__Pyx_AllocateExtensionType(PyTypeObject *t, int is_final) {
    if (is_final || likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
        allocfunc alloc_func = __Pyx_PyType_GetSlot(t, tp_alloc, allocfunc);
        return alloc_func(t, 0);
    } else {
        newfunc tp_new = __Pyx_PyType_TryGetSlot(&PyBaseObject_Type, tp_new, newfunc);
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (!tp_new) {
            PyObject *new_str = PyUnicode_FromString("__new__");
            if (likely(new_str)) {
                PyObject *o = PyObject_CallMethodObjArgs((PyObject *)&PyBaseObject_Type, new_str, t, NULL);
                Py_DECREF(new_str);
                return o;
            } else
                return NULL;
        } else
    #endif
        return tp_new(t, __pyx_mstate_global->__pyx_empty_tuple, 0);
    }
}

/* CallTypeTraverse */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg) {
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
    if (__Pyx_get_runtime_version() < 0x03090000) return 0;
    #endif
    if (!always_call) {
        PyTypeObject *base = __Pyx_PyObject_GetSlot(o, tp_base, PyTypeObject*);
        unsigned long flags = PyType_GetFlags(base);
        if (flags & Py_TPFLAGS_HEAPTYPE) {
            return 0;
        }
    }
    Py_VISIT((PyObject*)Py_TYPE(o));
    return 0;
}
#endif

/* LimitedApiGetTypeDict (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static Py_ssize_t __Pyx_GetTypeDictOffset(void) {
    PyObject *tp_dictoffset_o;
    Py_ssize_t tp_dictoffset;
    tp_dictoffset_o = PyObject_GetAttrString((PyObject*)(&PyType_Type), "__dictoffset__");
    if (unlikely(!tp_dictoffset_o)) return -1;
    tp_dictoffset = PyLong_AsSsize_t(tp_dictoffset_o);
    Py_DECREF(tp_dictoffset_o);
    if (unlikely(tp_dictoffset == 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' doesn't have a dictoffset");
        return -1;
    } else if (unlikely(tp_dictoffset < 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' has an unexpected negative dictoffset. "
            "Please report this as Cython bug");
        return -1;
    }
    return tp_dictoffset;
}
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp) {
    static Py_ssize_t tp_dictoffset = 0;
    if (unlikely(tp_dictoffset == 0)) {
        tp_dictoffset = __Pyx_GetTypeDictOffset();
        if (unlikely(tp_dictoffset == -1 && PyErr_Occurred())) {
            tp_dictoffset = 0; // try again next time?
            return NULL;
        }
    }
    return *(PyObject**)((char*)tp + tp_dictoffset);
}
#endif

/* SetItemOnTypeDict (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_SetItem(tp_dict, k, v);
    if (likely(!result)) {
        PyType_Modified(tp);
        if (unlikely(PyObject_HasAttr(v, __pyx_mstate_global->__pyx_n_u_set_name))) {
            PyObject *setNameResult = PyObject_CallMethodObjArgs(v, __pyx_mstate_global->__pyx_n_u_set_name,  (PyObject *) tp, k, NULL);
            if (!setNameResult) return -1;
            Py_DECREF(setNameResult);
        }
    }
    return result;
}

/* FixUpExtensionType */
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if __PYX_LIMITED_VERSION_HEX > 0x030900B1
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#else
    const PyType_Slot *slot = spec->slots;
    int changed = 0;
#if !CYTHON_COMPILING_IN_LIMITED_API
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
#if !CYTHON_COMPILING_IN_CPYTHON
        const
#endif  // !CYTHON_COMPILING_IN_CPYTHON)
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_vectorcall_offset = memb->offset;
                    changed = 1;
                }
#endif  // CYTHON_METH_FASTCALL
#if !CYTHON_COMPILING_IN_PYPY
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                    Py_DECREF(descr);
                    if (unlikely(set_item_result < 0)) {
                        return -1;
                    }
                    changed = 1;
                }
#endif  // !CYTHON_COMPILING_IN_PYPY
            }
            memb++;
        }
    }
#endif  // !CYTHON_COMPILING_IN_LIMITED_API
#if !CYTHON_COMPILING_IN_PYPY
    slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_getset)
        slot++;
    if (slot && slot->slot == Py_tp_getset) {
        PyGetSetDef *getset = (PyGetSetDef*) slot->pfunc;
        while (getset && getset->name) {
            if (getset->name[0] == '_' && getset->name[1] == '_' && strcmp(getset->name, "__module__") == 0) {
                PyObject *descr = PyDescr_NewGetSet(type, getset);
                if (unlikely(!descr))
                    return -1;
                #if CYTHON_COMPILING_IN_LIMITED_API
                PyObject *pyname = PyUnicode_FromString(getset->name);
                if (unlikely(!pyname)) {
                    Py_DECREF(descr);
                    return -1;
                }
                int set_item_result = __Pyx_SetItemOnTypeDict(type, pyname, descr);
                Py_DECREF(pyname);
                #else
                CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
                int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                #endif
                Py_DECREF(descr);
                if (unlikely(set_item_result < 0)) {
                    return -1;
                }
                changed = 1;
            }
            ++getset;
        }
    }
#else
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#endif  // !CYTHON_COMPILING_IN_PYPY
    if (changed)
        PyType_Modified(type);
#endif  // PY_VERSION_HEX > 0x030900B1
    return 0;
}

/* PyObjectCallNoArg (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallMethod0 (used by PyType_Ready) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[1] = {obj};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_CallNoArg;
    return PyObject_VectorcallMethod(method_name, args, 1 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
#endif
}

/* ValidateBasesTuple (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_SIZE
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (unlikely(n < 0)) return -1;
#endif
    for (i = 1; i < n; i++)
    {
        PyTypeObject *b;
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !CYTHON_USE_TYPE_SLOTS
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
CYTHON_UNUSED static int __Pyx_PyType_HasMultipleInheritance(PyTypeObject *t) {
    while (t) {
        PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
        if (bases) {
            return 1;
        }
        t = __Pyx_PyType_GetSlot(t, tp_base, PyTypeObject*);
    }
    return 0;
}
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !CYTHON_COMPILING_IN_CPYTHON || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    if (!__Pyx_PyType_HasMultipleInheritance(t)) {
        return PyType_Ready(t);
    }
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400) &&\
                !CYTHON_COMPILING_IN_GRAAL
        gc = PyImport_GetModule(__pyx_mstate_global->__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_mstate_global->__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable (used by MergeVTables) */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i=0;
    Py_ssize_t size;
    void** base_vtables;
    __Pyx_TypeName tp_base_name = NULL;
    __Pyx_TypeName base_name = NULL;
    void* unknown = (void*)-1;
    PyObject* bases = __Pyx_PyType_GetSlot(type, tp_bases, PyObject*);
    int base_depth = 0;
    {
        PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        while (base) {
            base_depth += 1;
            base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
        }
    }
    base_vtables = (void**) PyMem_Malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
#if CYTHON_COMPILING_IN_LIMITED_API
    size = PyTuple_Size(bases);
    if (size < 0) goto other_failure;
#else
    size = PyTuple_GET_SIZE(bases);
#endif
    for (i = 1; i < size; i++) {
        PyObject *basei;
        void* base_vtable;
#if CYTHON_AVOID_BORROWED_REFS
        basei = PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#else
        basei = PyTuple_GET_ITEM(bases, i);
#endif
        base_vtable = __Pyx_GetVtable((PyTypeObject*)basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
            }
        }
    }
    PyErr_Clear();
    PyMem_Free(base_vtables);
    return 0;
bad:
    {
        PyTypeObject* basei = NULL;
        PyTypeObject* tp_base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        tp_base_name = __Pyx_PyType_GetFullyQualifiedName(tp_base);
#if CYTHON_AVOID_BORROWED_REFS
        basei = (PyTypeObject*)PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = (PyTypeObject*)PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#else
        basei = (PyTypeObject*)PyTuple_GET_ITEM(bases, i);
#endif
        base_name = __Pyx_PyType_GetFullyQualifiedName(basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
    }
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
really_bad: // bad has failed!
#endif
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
other_failure:
#endif
    PyMem_Free(base_vtables);
    return -1;
}

/* DelItemOnTypeDict (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_DelItem(tp_dict, k);
    if (likely(!result)) PyType_Modified(tp);
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_mstate_global->__pyx_n_u_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_mstate_global->__pyx_n_u_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython);
            if (likely(reduce_cython)) {
                ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_mstate_global->__pyx_n_u_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetFullyQualifiedName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_2_4
#define __PYX_HAVE_RT_ImportType_3_2_4
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size)
{
    PyObject *result = 0;
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API)
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !( defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API) )
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    if (size == 0) {
        return (PyTypeObject *)result;
    }
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_2_4 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_2_4 && (size_t)basicsize > size) {
        if (PyErr_WarnFormat(NULL, 0,
                "%.200s.%.200s size changed, may indicate binary incompatibility. "
                "Expected %zd from C header, got %zd from PyObject",
                module_name, class_name, size, basicsize) < 0) {
            goto bad;
        }
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* HasAttr (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!PyUnicode_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_PyObject_GetAttrStrNoError(o, n);
    if (!r) {
        return (unlikely(PyErr_Occurred())) ? -1 : 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}
#endif

/* ImportImpl (used by Import) */
static int __Pyx__Import_GetModule(PyObject *qualname, PyObject **module) {
    PyObject *imported_module = PyImport_GetModule(qualname);
    if (unlikely(!imported_module)) {
        *module = NULL;
        if (PyErr_Occurred()) {
            return -1;
        }
        return 0;
    }
    *module = imported_module;
    return 1;
}
static int __Pyx__Import_Lookup(PyObject *qualname, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject **module) {
    PyObject *imported_module;
    PyObject *top_level_package_name;
    Py_ssize_t i;
    int status, module_found;
    Py_ssize_t dot_index;
    module_found = __Pyx__Import_GetModule(qualname, &imported_module);
    if (unlikely(!module_found || module_found == -1)) {
        *module = NULL;
        return module_found;
    }
    if (imported_names) {
        for (i = 0; i < len_imported_names; i++) {
            PyObject *imported_name = imported_names[i];
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
            int has_imported_attribute = PyObject_HasAttr(imported_module, imported_name);
#else
            int has_imported_attribute = PyObject_HasAttrWithError(imported_module, imported_name);
            if (unlikely(has_imported_attribute == -1)) goto error;
#endif
            if (!has_imported_attribute) {
                goto not_found;
            }
        }
        *module = imported_module;
        return 1;
    }
    dot_index = PyUnicode_FindChar(qualname, '.', 0, PY_SSIZE_T_MAX, 1);
    if (dot_index == -1) {
        *module = imported_module;
        return 1;
    }
    if (unlikely(dot_index == -2)) goto error;
    top_level_package_name = PyUnicode_Substring(qualname, 0, dot_index);
    if (unlikely(!top_level_package_name)) goto error;
    Py_DECREF(imported_module);
    status = __Pyx__Import_GetModule(top_level_package_name, module);
    Py_DECREF(top_level_package_name);
    return status;
error:
    Py_DECREF(imported_module);
    *module = NULL;
    return -1;
not_found:
    Py_DECREF(imported_module);
    *module = NULL;
    return 0;
}
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *from_list = 0;
    int module_found;
    if (!qualname) {
        qualname = name;
    }
    module_found = __Pyx__Import_Lookup(qualname, imported_names, len_imported_names, &module);
    if (likely(module_found == 1)) {
        return module;
    } else if (unlikely(module_found == -1)) {
        return NULL;
    }
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    if (imported_names) {
#if CYTHON_COMPILING_IN_CPYTHON
        from_list = __Pyx_PyList_FromArray(imported_names, len_imported_names);
        if (unlikely(!from_list))
            goto bad;
#else
        from_list = PyList_New(len_imported_names);
        if (unlikely(!from_list)) goto bad;
        for (Py_ssize_t i=0; i<len_imported_names; ++i) {
            if (PyList_SetItem(from_list, i, __Pyx_NewRef(imported_names[i])) < 0) goto bad;
        }
#endif
    }
    if (level == -1) {
        const char* package_sep = strchr(__Pyx_MODULE_NAME, '.');
        if (package_sep != (0)) {
            module = PyImport_ImportModuleLevelObject(
                name, moddict, empty_dict, from_list, 1);
            if (unlikely(!module)) {
                if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                    goto bad;
                PyErr_Clear();
            }
        }
        level = 0;
    }
    if (!module) {
        module = PyImport_ImportModuleLevelObject(
            name, moddict, empty_dict, from_list, level);
    }
bad:
    Py_XDECREF(from_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level) {
    return __Pyx__Import(name, imported_names, len_imported_names, qualname, __pyx_mstate_global->__pyx_d, level);
}

/* dict_setdefault (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value) {
    PyObject* value;
#if __PYX_LIMITED_VERSION_HEX >= 0x030F0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4)
    PyDict_SetDefaultRef(d, key, default_value, &value);
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    PyObject *args[] = {d, key, default_value};
    value = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_n_u_setdefault, args, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#elif CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethodObjArgs(d, __pyx_mstate_global->__pyx_n_u_setdefault, key, default_value, NULL);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* AddModuleRef (used by FetchSharedCythonModule) */
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  static PyObject *__Pyx_PyImport_AddModuleObjectRef(PyObject *name) {
      PyObject *module_dict = PyImport_GetModuleDict();
      PyObject *m;
      if (PyMapping_GetOptionalItem(module_dict, name, &m) < 0) {
          return NULL;
      }
      if (m != NULL && PyModule_Check(m)) {
          return m;
      }
      Py_XDECREF(m);
      m = PyModule_NewObject(name);
      if (m == NULL)
          return NULL;
      if (PyDict_CheckExact(module_dict)) {
          PyObject *new_m;
          (void)PyDict_SetDefaultRef(module_dict, name, m, &new_m);
          Py_DECREF(m);
          return new_m;
      } else {
           if (PyObject_SetItem(module_dict, name, m) != 0) {
                Py_DECREF(m);
                return NULL;
            }
            return m;
      }
  }
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *py_name = PyUnicode_FromString(name);
      if (!py_name) return NULL;
      PyObject *module = __Pyx_PyImport_AddModuleObjectRef(py_name);
      Py_DECREF(py_name);
      return module;
  }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif

/* FetchSharedCythonModule (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef(__PYX_ABI_MODULE_NAME);
}

/* FetchCommonType (used by CommonTypesMetaclass) */
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject* __Pyx_PyType_FromMetaclass(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *result = __Pyx_PyType_FromModuleAndSpec(module, spec, bases);
    if (result && metaclass) {
        PyObject *old_tp = (PyObject*)Py_TYPE(result);
    Py_INCREF((PyObject*)metaclass);
#if __PYX_LIMITED_VERSION_HEX >= 0x03090000
        Py_SET_TYPE(result, metaclass);
#else
        result->ob_type = metaclass;
#endif
        Py_DECREF(old_tp);
    }
    return result;
}
#else
#define __Pyx_PyType_FromMetaclass(me, mo, s, b) PyType_FromMetaclass(me, mo, s, b)
#endif
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t expected_basicsize) {
    Py_ssize_t basicsize;
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (expected_basicsize == 0) {
        return 0; // size is inherited, nothing useful to check
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
    if (unlikely(!py_basicsize)) return -1;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = NULL;
    if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) return -1;
#else
    basicsize = ((PyTypeObject*) cached_type)->tp_basicsize;
#endif
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module = NULL, *cached_type = NULL, *abi_module_dict, *new_cached_type, *py_object_name;
    int get_item_ref_result;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    py_object_name = PyUnicode_FromString(object_name);
    if (!py_object_name) return NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) goto done;
    abi_module_dict = PyModule_GetDict(abi_module);
    if (!abi_module_dict) goto done;
    get_item_ref_result = __Pyx_PyDict_GetItemRef(abi_module_dict, py_object_name, &cached_type);
    if (get_item_ref_result == 1) {
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else if (unlikely(get_item_ref_result == -1)) {
        goto bad;
    }
    cached_type = __Pyx_PyType_FromMetaclass(
        metaclass,
        CYTHON_USE_MODULE_STATE ? module : abi_module,
        spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    new_cached_type = __Pyx_PyDict_SetDefault(abi_module_dict, py_object_name, cached_type);
    if (unlikely(new_cached_type != cached_type)) {
        if (unlikely(!new_cached_type)) goto bad;
        Py_DECREF(cached_type);
        cached_type = new_cached_type;
        if (__Pyx_VerifyCachedType(
                cached_type,
                object_name,
                spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else {
        Py_DECREF(new_cached_type);
    }
done:
    Py_XDECREF(abi_module);
    Py_DECREF(py_object_name);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CommonTypesMetaclass (used by CythonFunctionShared) */
static PyObject* __pyx_CommonTypesMetaclass_get_module(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED void* context) {
    return PyUnicode_FromString(__PYX_ABI_MODULE_NAME);
}
#if __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject* __pyx_CommonTypesMetaclass_call(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *args, CYTHON_UNUSED PyObject *kwds) {
    PyErr_SetString(PyExc_TypeError, "Cannot instantiate Cython internal types");
    return NULL;
}
static int __pyx_CommonTypesMetaclass_setattr(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *attr, CYTHON_UNUSED PyObject *value) {
    PyErr_SetString(PyExc_TypeError, "Cython internal types are immutable");
    return -1;
}
#endif
static PyGetSetDef __pyx_CommonTypesMetaclass_getset[] = {
    {"__module__", __pyx_CommonTypesMetaclass_get_module, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_CommonTypesMetaclass_slots[] = {
    {Py_tp_getset, (void *)__pyx_CommonTypesMetaclass_getset},
    #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {Py_tp_call, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_new, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_setattro, (void*)__pyx_CommonTypesMetaclass_setattr},
    #endif
    {0, 0}
};
static PyType_Spec __pyx_CommonTypesMetaclass_spec = {
    __PYX_TYPE_MODULE_PREFIX "_common_types_metatype",
    0,
    0,
    Py_TPFLAGS_IMMUTABLETYPE |
    Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT,
    __pyx_CommonTypesMetaclass_slots
};
static int __pyx_CommonTypesMetaclass_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    PyObject *bases = PyTuple_Pack(1, &PyType_Type);
    if (unlikely(!bases)) {
        return -1;
    }
    mstate->__pyx_CommonTypesMetaclassType = __Pyx_FetchCommonTypeFromSpec(NULL, module, &__pyx_CommonTypesMetaclass_spec, bases);
    Py_DECREF(bases);
    if (unlikely(mstate->__pyx_CommonTypesMetaclassType == NULL)) {
        return -1;
    }
    return 0;
}

/* PyMethodNew (used by CythonFunctionShared) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *result;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    #if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    {
        PyObject *args[] = {func, self};
        result = PyObject_Vectorcall(__pyx_mstate_global->__Pyx_CachedMethodType, args, 2, NULL);
    }
    #else
    result = PyObject_CallFunctionObjArgs(__pyx_mstate_global->__Pyx_CachedMethodType, func, self, NULL);
    #endif
    return result;
}
#else
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#endif

/* PyVectorcallFastCallDict (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i;
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos;
    #else
    Py_ssize_t pos;
    #endif
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    #if !CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t nkw = PyDict_Size(kw);
    if (unlikely(nkw == -1)) return NULL;
    #else
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    #endif
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = 0;
    i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (__Pyx_PyDict_NextRef(kw, &pos, &key, &value)) {
        keys_are_strings &=
        #if CYTHON_COMPILING_IN_LIMITED_API
            PyType_GetFlags(Py_TYPE(key));
        #else
            Py_TYPE(key)->tp_flags;
        #endif
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(PyTuple_SetItem(kwnames, i, key) < 0)) goto cleanup;
        #else
        PyTuple_SET_ITEM(kwnames, i, key);
        #endif
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(pos);
    #endif
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    Py_ssize_t kw_size =
        likely(kw == NULL) ?
        0 :
#if !CYTHON_ASSUME_SAFE_SIZE
        PyDict_Size(kw);
#else
        PyDict_GET_SIZE(kw);
#endif
    if (kw_size == 0) {
        return vc(func, args, nargs, NULL);
    }
#if !CYTHON_ASSUME_SAFE_SIZE
    else if (unlikely(kw_size == -1)) {
        return NULL;
    }
#endif
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared (used by CythonFunction) */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunctionNoMethod(PyObject *func, void (*cfunc)(void)) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if ((PyObject*)Py_TYPE(func) == __pyx_mstate_global->__Pyx_CachedMethodType) {
        int result;
        PyObject *newFunc = PyObject_GetAttr(func, __pyx_mstate_global->__pyx_n_u_func);
        if (unlikely(!newFunc)) {
            PyErr_Clear(); // It's only an optimization, so don't throw an error
            return 0;
        }
        result = __Pyx__IsSameCyOrCFunctionNoMethod(newFunc, cfunc);
        Py_DECREF(newFunc);
        return result;
    }
    return __Pyx__IsSameCyOrCFunctionNoMethod(func, cfunc);
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if (PyMethod_Check(func)) {
        func = PyMethod_GET_FUNCTION(func);
    }
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure) {
    PyObject *result;
    CYTHON_UNUSED_VAR(closure);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_doc_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#else
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_name_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    Py_INCREF(op->func_qualname);
    result = op->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
#endif
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_defaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_kwdefaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_annotations_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine_value(__pyx_CyFunctionObject *op) {
    int is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
    if (is_coroutine) {
        PyObject *is_coroutine_value, *module, *fromlist, *marker = __pyx_mstate_global->__pyx_n_u_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_mstate_global->__pyx_n_u_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        is_coroutine_value = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(is_coroutine_value)) {
            return is_coroutine_value;
        }
ignore:
        PyErr_Clear();
    }
    return __Pyx_PyBool_FromLong(is_coroutine);
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    result = __Pyx_CyFunction_get_is_coroutine_value(op);
    if (unlikely(!result))
        return NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    if (op->func_is_coroutine) {
        Py_DECREF(result);
        result = __Pyx_NewRef(op->func_is_coroutine);
    } else {
        op->func_is_coroutine = __Pyx_NewRef(result);
    }
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static void __Pyx_CyFunction_raise_argument_count_error(__pyx_CyFunctionObject *func, const char* message, Py_ssize_t size) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        py_name, message, size);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        name, message, size);
#endif
}
static void __Pyx_CyFunction_raise_type_error(__pyx_CyFunctionObject *func, const char* message) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s",
        py_name, message);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s",
        name, message);
#endif
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {"func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {"func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
#else
    {"func_dict", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
#endif
    {"func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {"__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {"_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    {"__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#endif
#if CYTHON_METH_FASTCALL
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {"__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(args);
    __Pyx_BEGIN_CRITICAL_SECTION(m);
    Py_INCREF(m->func_qualname);
    result = m->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_dict = NULL;
#endif
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func_dict);
#elif PY_VERSION_HEX < 0x030d0000
    _PyObject_ClearManagedDict((PyObject*)m);
#else
    PyObject_ClearManagedDict((PyObject*)m);
#endif
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    Py_CLEAR(m->defaults);
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    {
        int e = __Pyx_call_type_traverse((PyObject*)m, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func_dict);
#else
    {
        int e =
#if PY_VERSION_HEX < 0x030d0000
            _PyObject_VisitManagedDict
#else
            PyObject_VisitManagedDict
#endif
                ((PyObject*)m, visit, arg);
        if (e != 0) return e;
    }
#endif
    __Pyx_VISIT_CONST(m->func_name);
    __Pyx_VISIT_CONST(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    __Pyx_VISIT_CONST(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    Py_VISIT(m->defaults);
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
    PyObject *repr;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    repr = PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
    __Pyx_END_CRITICAL_SECTION();
    return repr;
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void(*)(void))meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes no arguments", size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes exactly one argument", size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    __Pyx_CyFunction_raise_type_error(
        (__pyx_CyFunctionObject*)func, "takes no keyword arguments");
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_SIZE
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(argc < 0)) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            __Pyx_CyFunction_raise_type_error(
                cyfunc, "needs an argument");
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(__Pyx_PyTuple_GET_SIZE(kwnames))) {
        __Pyx_CyFunction_raise_type_error(
            cyfunc, "takes no keyword arguments");
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes no arguments", nargs);
        return NULL;
    }
    return meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes exactly one argument", nargs);
        return NULL;
    }
    return meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    #if PY_VERSION_HEX < 0x030e00A6
    size_t nargs_value = (size_t) nargs;
    #else
    Py_ssize_t nargs_value = nargs;
    #endif
    return ((__Pyx_PyCMethod)(void(*)(void))meth)(self, cls, args, nargs_value, kwnames);
}
#endif
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if CYTHON_METH_FASTCALL
#if defined(Py_TPFLAGS_HAVE_VECTORCALL)
    Py_TPFLAGS_HAVE_VECTORCALL |
#elif defined(_Py_TPFLAGS_HAVE_VECTORCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
#endif // CYTHON_METH_FASTCALL
#if PY_VERSION_HEX >= 0x030C0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_TPFLAGS_MANAGED_DICT |
#endif
    Py_TPFLAGS_IMMUTABLETYPE | Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
static int __pyx_CyFunction_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_CyFunctionType_spec, NULL);
    if (unlikely(mstate->__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func, PyTypeObject *defaults_type) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_CallObject((PyObject*)defaults_type, NULL); // _PyObject_New(defaults_type);
    if (unlikely(!m->defaults))
        return NULL;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_mstate_global->__pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CLineInTraceback (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
#define __Pyx_PyProbablyModule_GetDict(o) __Pyx_XNewRef(PyModule_GetDict(o))
#elif !CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyProbablyModule_GetDict(o) PyObject_GenericGetDict(o, NULL);
#else
PyObject* __Pyx_PyProbablyModule_GetDict(PyObject *o) {
    PyObject **dict_ptr = _PyObject_GetDictPtr(o);
    return dict_ptr ? __Pyx_XNewRef(*dict_ptr) : NULL;
}
#endif
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline = NULL;
    PyObject *ptype, *pvalue, *ptraceback;
    PyObject *cython_runtime_dict;
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
    cython_runtime_dict = __Pyx_PyProbablyModule_GetDict(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, cython_runtime_dict,
            __Pyx_PyDict_SetDefault(cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False))
    }
    if (use_cline == NULL || use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    Py_XDECREF(cython_runtime_dict);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache (used by AddTraceback) */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        if (c_line) {
            py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        } else {
            py_funcname = PyUnicode_FromString(funcname);
        }
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CheckUnpickleChecksum */
static void __Pyx_RaiseUnpickleChecksumError(long checksum, long checksum1, long checksum2, long checksum3, const char *members) {
    PyObject *pickle_module = PyImport_ImportModule("pickle");
    if (unlikely(!pickle_module)) return;
    PyObject *pickle_error = PyObject_GetAttrString(pickle_module, "PickleError");
    Py_DECREF(pickle_module);
    if (unlikely(!pickle_error)) return;
    if (checksum2 == checksum1) {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x) = (%s))",
            checksum, checksum1, members);
    } else if (checksum3 == checksum2) {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x, 0x%x) = (%s))",
            checksum, checksum1, checksum2, members);
    } else {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x, 0x%x, 0x%x) = (%s))",
            checksum, checksum1, checksum2, checksum3, members);
    }
    Py_DECREF(pickle_error);
}
static int __Pyx_CheckUnpickleChecksum(long checksum, long checksum1, long checksum2, long checksum3, const char *members) {
    int found = 0;
    found |= checksum1 == checksum;
    found |= checksum2 == checksum;
    found |= checksum3 == checksum;
    if (likely(found))
        return 0;
    __Pyx_RaiseUnpickleChecksumError(checksum, checksum1, checksum2, checksum3, members);
    return -1;
}

/* PyObjectVectorCallKwBuilder (used by CIntToPy) */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    Py_INCREF(key);
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_PY_LONG_LONG(unsigned PY_LONG_LONG value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned PY_LONG_LONG neg_one = (unsigned PY_LONG_LONG) -1, const_zero = (unsigned PY_LONG_LONG) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned PY_LONG_LONG) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned PY_LONG_LONG) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned PY_LONG_LONG),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned PY_LONG_LONG));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_int(unsigned int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_char(unsigned char value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned char neg_one = (unsigned char) -1, const_zero = (unsigned char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned char) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned char) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(unsigned char) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned char) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned char) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned char),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned char));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE unsigned char __Pyx_PyLong_As_unsigned_char(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned char neg_one = (unsigned char) -1, const_zero = (unsigned char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        unsigned char val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (unsigned char) -1;
        val = __Pyx_PyLong_As_unsigned_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned char, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(unsigned char) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned char) >= 2 * PyLong_SHIFT)) {
                            return (unsigned char) (((((unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned char) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned char) >= 3 * PyLong_SHIFT)) {
                            return (unsigned char) (((((((unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned char) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned char) >= 4 * PyLong_SHIFT)) {
                            return (unsigned char) (((((((((unsigned char)digits[3]) << PyLong_SHIFT) | (unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (unsigned char) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(unsigned char) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned char, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(unsigned char) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned char, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(unsigned char) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned char) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned char) (((unsigned char)-1)*(((((unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(unsigned char) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned char) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned char) ((((((unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(unsigned char) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned char) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned char) (((unsigned char)-1)*(((((((unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned char) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned char) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned char) ((((((((unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(unsigned char) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned char) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned char) (((unsigned char)-1)*(((((((((unsigned char)digits[3]) << PyLong_SHIFT) | (unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned char) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned char) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned char) ((((((((((unsigned char)digits[3]) << PyLong_SHIFT) | (unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(unsigned char) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned char, long, PyLong_AsLong(x))
        } else if ((sizeof(unsigned char) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned char, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        unsigned char val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (unsigned char) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (unsigned char) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (unsigned char) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (unsigned char) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(unsigned char) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((unsigned char) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(unsigned char) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((unsigned char) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((unsigned char) 1) << (sizeof(unsigned char) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (unsigned char) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned char");
    return (unsigned char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned char");
    return (unsigned char) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* UpdateUnpickledDict */
static int __Pyx__UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index) {
    PyObject *state_dict = __Pyx_PySequence_ITEM(state, index);
    if (unlikely(!state_dict)) {
        return -1;
    }
    int non_empty = PyObject_IsTrue(state_dict);
    if (non_empty == 0) {
        Py_DECREF(state_dict);
        return 0;
    } else if (unlikely(non_empty == -1)) {
        return -1;
    }
    PyObject *dict;
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    dict = PyObject_GetAttrString(obj, "__dict__");
    #else
    dict = PyObject_GenericGetDict(obj, NULL);
    #endif
    if (unlikely(!dict)) {
        Py_DECREF(state_dict);
        return -1;
    }
    int result;
    if (likely(PyDict_CheckExact(dict))) {
        result = PyDict_Update(dict, state_dict);
    } else {
        PyObject *obj_result = __Pyx_PyObject_CallMethod1(dict, __pyx_mstate_global->__pyx_n_u_update, state_dict);
        if (likely(obj_result)) {
            Py_DECREF(obj_result);
            result = 0;
        } else {
            result = -1;
        }
    }
    Py_DECREF(state_dict);
    Py_DECREF(dict);
    return result;
}
static int __Pyx_UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index) {
    Py_ssize_t state_size = __Pyx_PyTuple_GET_SIZE(state);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(state_size == -1)) return -1;
    #endif
    if (state_size <= index) {
        return 0;
    }
    return __Pyx__UpdateUnpickledDict(obj, state, index);
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u__2);
    }
    goto done;
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetRuntimeVersion */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
void __Pyx_init_runtime_version(void) {
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
}
#endif
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    return __Pyx_cached_runtime_version;
#endif
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* NewCodeObj */
#if CYTHON_COMPILING_IN_LIMITED_API
    static PyObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                       PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                       PyObject *fv, PyObject *cell, PyObject* fn,
                                       PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
#elif PY_VERSION_HEX >= 0x030B0000
  static PyCodeObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                         PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                         PyObject *fv, PyObject *cell, PyObject* fn,
                                         PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, __pyx_mstate_global->__pyx_empty_bytes);
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030c00A1
    if (likely(result))
        result->_co_firsttraceable = 0;
    #endif
    return result;
  }
#elif !CYTHON_COMPILING_IN_PYPY
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
) {
    PyObject *code_obj = NULL, *varnames_tuple_dedup = NULL, *code_bytes = NULL;
    Py_ssize_t var_count = (Py_ssize_t) descr.nlocals;
    PyObject *varnames_tuple = PyTuple_New(var_count);
    if (unlikely(!varnames_tuple)) return NULL;
    for (Py_ssize_t i=0; i < var_count; i++) {
        Py_INCREF(varnames[i]);
        if (__Pyx_PyTuple_SET_ITEM(varnames_tuple, i, varnames[i]) != (0)) goto done;
    }
    #if CYTHON_COMPILING_IN_LIMITED_API
    varnames_tuple_dedup = PyDict_GetItem(tuple_dedup_map, varnames_tuple);
    if (!varnames_tuple_dedup) {
        if (unlikely(PyDict_SetItem(tuple_dedup_map, varnames_tuple, varnames_tuple) < 0)) goto done;
        varnames_tuple_dedup = varnames_tuple;
    }
    #else
    varnames_tuple_dedup = PyDict_SetDefault(tuple_dedup_map, varnames_tuple, varnames_tuple);
    if (unlikely(!varnames_tuple_dedup)) goto done;
    #endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(varnames_tuple_dedup);
    #endif
    if (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table != NULL && !CYTHON_COMPILING_IN_GRAAL) {
        Py_ssize_t line_table_length = __Pyx_PyBytes_GET_SIZE(line_table);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(line_table_length == -1)) goto done;
        #endif
        Py_ssize_t code_len = (line_table_length * 2 + 4) & ~3LL;
        code_bytes = PyBytes_FromStringAndSize(NULL, code_len);
        if (unlikely(!code_bytes)) goto done;
        char* c_code_bytes = PyBytes_AsString(code_bytes);
        if (unlikely(!c_code_bytes)) goto done;
        memset(c_code_bytes, 0, (size_t) code_len);
    }
    code_obj = (PyObject*) __Pyx__PyCode_New(
        (int) descr.argcount,
        (int) descr.num_posonly_args,
        (int) descr.num_kwonly_args,
        (int) descr.nlocals,
        0,
        (int) descr.flags,
        code_bytes ? code_bytes : __pyx_mstate_global->__pyx_empty_bytes,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        varnames_tuple_dedup,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        filename,
        funcname,
        (int) descr.first_line,
        (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table) ? line_table : __pyx_mstate_global->__pyx_empty_bytes
    );
done:
    Py_XDECREF(code_bytes);
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(varnames_tuple_dedup);
    #endif
    Py_DECREF(varnames_tuple);
    return code_obj;
}

/* DecompressString */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo) {
    PyObject *module = NULL, *decompress, *compressed_bytes, *decompressed;
    const char* module_name = algo == 3 ? "compression.zstd" : algo == 2 ? "bz2" : "zlib";
    PyObject *methodname = PyUnicode_FromString("decompress");
    if (unlikely(!methodname)) return NULL;
    #if __PYX_LIMITED_VERSION_HEX >= 0x030e0000
    if (algo == 3) {
        PyObject *fromlist = Py_BuildValue("[O]", methodname);
        if (unlikely(!fromlist)) goto bad;
        module = PyImport_ImportModuleLevel("compression.zstd", NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
    } else
    #endif
        module = PyImport_ImportModule(module_name);
    if (unlikely(!module)) goto import_failed;
    decompress = PyObject_GetAttr(module, methodname);
    if (unlikely(!decompress)) goto import_failed;
    {
        #ifdef __cplusplus
            char *memview_bytes = const_cast<char*>(s);
        #else
            #if defined(__clang__)
              #pragma clang diagnostic push
              #pragma clang diagnostic ignored "-Wcast-qual"
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic push
              #pragma GCC diagnostic ignored "-Wcast-qual"
            #endif
            char *memview_bytes = (char*) s;
            #if defined(__clang__)
              #pragma clang diagnostic pop
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic pop
            #endif
        #endif
        #if CYTHON_COMPILING_IN_LIMITED_API && !defined(PyBUF_READ)
        int memview_flags = 0x100;
        #else
        int memview_flags = PyBUF_READ;
        #endif
        compressed_bytes = PyMemoryView_FromMemory(memview_bytes, length, memview_flags);
    }
    if (unlikely(!compressed_bytes)) {
        Py_DECREF(decompress);
        goto bad;
    }
    decompressed = PyObject_CallFunctionObjArgs(decompress, compressed_bytes, NULL);
    Py_DECREF(compressed_bytes);
    Py_DECREF(decompress);
    Py_DECREF(module);
    Py_DECREF(methodname);
    return decompressed;
import_failed:
    PyErr_Format(PyExc_ImportError,
        "Failed to import '%.20s.decompress' - cannot initialise module strings. "
        "String compression was configured with the C macro 'CYTHON_COMPRESS_STRINGS=%d'.",
        module_name, algo);
bad:
    Py_XDECREF(module);
    Py_DECREF(methodname);
    return NULL;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return __Pyx_NewRef(b ? Py_True: Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
