/* Generated by Cython 3.2.4 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "breezy/bzr/_dirstate_helpers_pyx.h",
            "breezy/python-compat.h"
        ],
        "include_dirs": [
            "breezy/bzr",
            "breezy"
        ],
        "name": "breezy.bzr._dirstate_helpers_pyx",
        "sources": [
            "breezy/bzr/_dirstate_helpers_pyx.pyx"
        ]
    },
    "module_name": "breezy.bzr._dirstate_helpers_pyx"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API)
  #if !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef _MSC_VER
  #pragma message ("Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.")
  #else
  #warning Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_2_4"
#define CYTHON_HEX_VERSION 0x030204F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #ifndef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 1
  #endif
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
  #if defined(CYTHON_IMMORTAL_CONSTANTS) && PY_VERSION_HEX < 0x030C0000
    #undef CYTHON_IMMORTAL_CONSTANTS
    #define CYTHON_IMMORTAL_CONSTANTS 0  // definitely won't work
  #elif !defined(CYTHON_IMMORTAL_CONSTANTS)
    #define CYTHON_IMMORTAL_CONSTANTS (PY_VERSION_HEX >= 0x030C0000 && !CYTHON_USE_MODULE_STATE && CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_COMPRESS_STRINGS
  #define CYTHON_COMPRESS_STRINGS 1
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL)
#endif
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#include <stdint.h>
typedef uintptr_t  __pyx_uintptr_t;
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CInitCode */
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef Py_TPFLAGS_IMMUTABLETYPE
  #define Py_TPFLAGS_IMMUTABLETYPE (1UL << 8)
#endif
#ifndef Py_TPFLAGS_DISALLOW_INSTANTIATION
  #define Py_TPFLAGS_DISALLOW_INSTANTIATION (1UL << 7)
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if PY_VERSION_HEX < 0x03090000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) GraalPyFrame_SetLineNumber((frame), (lineno))
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && PyUnstable_Object_IsUniquelyReferenced(obj)) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#elif CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
enum __Pyx_ReferenceSharing {
  __Pyx_ReferenceSharing_DefinitelyUnique, // We created it so we know it's unshared - no need to check
  __Pyx_ReferenceSharing_OwnStrongReference,
  __Pyx_ReferenceSharing_FunctionArgument,
  __Pyx_ReferenceSharing_SharedReference, // Never trust it to be unshared because it's a global or similar
};
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && PY_VERSION_HEX >= 0x030E0000
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing)\
    (sharing == __Pyx_ReferenceSharing_DefinitelyUnique ? 1 :\
      (sharing == __Pyx_ReferenceSharing_FunctionArgument ? PyUnstable_Object_IsUniqueReferencedTemporary(o) :\
      (sharing == __Pyx_ReferenceSharing_OwnStrongReference ? PyUnstable_Object_IsUniquelyReferenced(o) : 0)))
#elif (CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING) || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)sharing), Py_REFCNT(o) == 1)
#else
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)o), ((void)sharing), 0)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !CYTHON_COMPILING_IN_LIMITED_API && CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) (__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared) ?\
    __Pyx_NewRef(PyList_GET_ITEM(o, i)) : __Pyx_PyList_GetItemRef(o, i))
#else
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) __Pyx_PyList_GetItemRef(o, i)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__breezy__bzr___dirstate_helpers_pyx
#define __PYX_HAVE_API__breezy__bzr___dirstate_helpers_pyx
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "python-compat.h"
#include "_dirstate_helpers_pyx.h"
#include "stdlib.h"
#include "sys/stat.h"
#include "string.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifdef CYTHON_FREETHREADING_COMPATIBLE
#if CYTHON_FREETHREADING_COMPATIBLE
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "breezy/bzr/_dirstate_helpers_pyx.pyx",
  "breezy/bzr/_str_helpers.pxd",
  "(tree fragment)",
  "cpython/type.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* Atomics.proto (used by UnpackUnboundCMethod) */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_nonatomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_nonatomic_ptr_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #define __pyx_nonatomic_ptr_type void*
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_atomic_ptr_type old = _InterlockedCompareExchangePointer(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif

/* CriticalSectionsDefinition.proto (used by CriticalSections) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection void*
#define __Pyx_PyCriticalSection2 void*
#define __Pyx_PyCriticalSection_End(cs)
#define __Pyx_PyCriticalSection2_End(cs)
#else
#define __Pyx_PyCriticalSection PyCriticalSection
#define __Pyx_PyCriticalSection2 PyCriticalSection2
#define __Pyx_PyCriticalSection_End PyCriticalSection_End
#define __Pyx_PyCriticalSection2_End PyCriticalSection2_End
#endif

/* CriticalSections.proto (used by ParseKeywordsImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection_Begin(cs, arg) (void)(cs)
#define __Pyx_PyCriticalSection2_Begin(cs, arg1, arg2) (void)(cs)
#else
#define __Pyx_PyCriticalSection_Begin PyCriticalSection_Begin
#define __Pyx_PyCriticalSection2_Begin PyCriticalSection2_Begin
#endif
#if PY_VERSION_HEX < 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_BEGIN_CRITICAL_SECTION(o) {
#define __Pyx_END_CRITICAL_SECTION() }
#else
#define __Pyx_BEGIN_CRITICAL_SECTION Py_BEGIN_CRITICAL_SECTION
#define __Pyx_END_CRITICAL_SECTION Py_END_CRITICAL_SECTION
#endif

/* IncludeStructmemberH.proto (used by FixUpExtensionType) */
#include <structmember.h>

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader;
struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC;

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":509
 * 
 * 
 * cdef class Reader:             # <<<<<<<<<<<<<<
 *     """Maintain the current location, and return fields as you parse them."""
 * 
*/
struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_vtab;
  PyObject *state;
  PyObject *text;
  char *text_cstr;
  int text_size;
  char *end_cstr;
  char *cur_cstr;
  char *next;
};


/* "breezy/bzr/_dirstate_helpers_pyx.pyx":967
 * 
 * 
 * cdef class ProcessEntryC:             # <<<<<<<<<<<<<<
 * 
 *     cdef int doing_consistency_expansion
*/
struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC {
  PyObject_HEAD
  struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_vtab;
  int doing_consistency_expansion;
  PyObject *old_dirname_to_file_id;
  PyObject *new_dirname_to_file_id;
  PyObject *last_source_parent;
  PyObject *last_target_parent;
  int include_unchanged;
  int partial;
  PyObject *use_filesystem_for_exec;
  PyObject *utf8_decode;
  PyObject *searched_specific_files;
  PyObject *searched_exact_paths;
  PyObject *search_specific_files;
  PyObject *search_specific_file_parents;
  PyObject *state;
  PyObject *current_root;
  PyObject *current_root_unicode;
  PyObject *root_entries;
  int root_entries_pos;
  int root_entries_len;
  PyObject *root_abspath;
  int source_index;
  int target_index;
  int want_unversioned;
  PyObject *tree;
  PyObject *dir_iterator;
  int block_index;
  PyObject *current_block;
  int current_block_pos;
  PyObject *current_block_list;
  PyObject *current_dir_info;
  PyObject *current_dir_list;
  PyObject *_pending_consistent_entries;
  int path_index;
  PyObject *root_dir_info;
  PyObject *bisect_left;
  PyObject *pathjoin;
  PyObject *fstat;
  PyObject *seen_ids;
  PyObject *sha_file;
};



/* "breezy/bzr/_dirstate_helpers_pyx.pyx":509
 * 
 * 
 * cdef class Reader:             # <<<<<<<<<<<<<<
 *     """Maintain the current location, and return fields as you parse them."""
 * 
*/

struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader {
  char *(*get_next)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *, int *);
  PyObject *(*get_next_str)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *);
  int (*_init)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *);
  PyObject *(*_get_entry)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *, int, void **, int *);
};
static struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_vtabptr_6breezy_3bzr_21_dirstate_helpers_pyx_Reader;


/* "breezy/bzr/_dirstate_helpers_pyx.pyx":967
 * 
 * 
 * cdef class ProcessEntryC:             # <<<<<<<<<<<<<<
 * 
 *     cdef int doing_consistency_expansion
*/

struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC {
  PyObject *(*_process_entry)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *, PyObject *, PyObject *);
  int (*_gather_result_for_consistency)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *, PyObject *);
  int (*_update_current_block)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *);
  PyObject *(*_iter_next)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *);
  PyObject *(*_maybe_tree_ref)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *, PyObject *);
  PyObject *(*_loop_one_block)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *);
  PyObject *(*_next_consistent_entries)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *);
  PyObject *(*_path_info)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *, PyObject *, PyObject *);
};
static struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_vtabptr_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyAssertionError_Check.proto */
#define __Pyx_PyExc_AssertionError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_AssertionError)

/* PyObjectCall.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* PyThreadStateGet.proto (used by PyErrFetchRestore) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto (used by RaiseException) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.export */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* TupleAndListFromArray.proto (used by fastcall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto (used by BytesEquals) */
#include <string.h>

/* BytesEquals.proto (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_PySequence_ITEM(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_NewRef(__Pyx_PyTuple_GET_ITEM(args, i))
#else
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_XNewRef(PyTuple_GetItem(args, i))
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_ArgRef_FASTCALL(args, i) __Pyx_NewRef(args[i])
    #define __Pyx_NumKwargs_FASTCALL(kwds) __Pyx_PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
#else
    #define __Pyx_ArgRef_FASTCALL __Pyx_ArgRef_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#if CYTHON_METH_FASTCALL || (CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(args + start, stop - start)
#else
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* py_dict_items.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* CallCFunction.proto (used by CallUnboundCMethod0) */
#define __Pyx_CallCFunction(cfunc, self, args)\
    ((PyCFunction)(void(*)(void))(cfunc)->func)(self, args)
#define __Pyx_CallCFunctionWithKeywords(cfunc, self, args, kwargs)\
    ((PyCFunctionWithKeywords)(void(*)(void))(cfunc)->func)(self, args, kwargs)
#define __Pyx_CallCFunctionFast(cfunc, self, args, nargs)\
    ((__Pyx_PyCFunctionFast)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs)
#define __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, nargs, kwnames)\
    ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs, kwnames)

/* PyObjectCallOneArg.proto (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectGetAttrStr.proto (used by UnpackUnboundCMethod) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* UnpackUnboundCMethod.proto (used by CallUnboundCMethod0) */
typedef struct {
    PyObject *type;
    PyObject **method_name;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && CYTHON_ATOMICS
    __pyx_atomic_int_type initialized;
#endif
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
static CYTHON_INLINE int __Pyx_CachedCFunction_GetAndSetInitializing(__Pyx_CachedCFunction *cfunc) {
#if !CYTHON_ATOMICS
    return 1;
#else
    __pyx_nonatomic_int_type expected = 0;
    if (__pyx_atomic_int_cmp_exchange(&cfunc->initialized, &expected, 1)) {
        return 0;
    }
    return expected;
#endif
}
static CYTHON_INLINE void __Pyx_CachedCFunction_SetFinishedInitializing(__Pyx_CachedCFunction *cfunc) {
#if CYTHON_ATOMICS
    __pyx_atomic_store(&cfunc->initialized, 2);
#endif
}
#else
#define __Pyx_CachedCFunction_GetAndSetInitializing(cfunc) 2
#define __Pyx_CachedCFunction_SetFinishedInitializing(cfunc)
#endif

/* CallUnboundCMethod0.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* py_dict_values.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d);

/* OwnedDictNext.proto (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue);
#else
CYTHON_INLINE
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue);
#endif

/* RaiseDoubleKeywords.proto (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywordsImpl.export */
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name
);
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* CallUnboundCMethod2.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* ParseKeywords.proto */
static CYTHON_INLINE int __Pyx_ParseKeywords(
    PyObject *kwds, PyObject *const *kwvalues, PyObject ** const argnames[],
    PyObject *kwds2, PyObject *values[],
    Py_ssize_t num_pos_args, Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* AssertionsEnabled.proto */
#if CYTHON_COMPILING_IN_LIMITED_API  ||  PY_VERSION_HEX >= 0x030C0000
  static int __pyx_assertions_enabled_flag;
  #define __pyx_assertions_enabled() (__pyx_assertions_enabled_flag)
  #if __clang__ || __GNUC__
  __attribute__((no_sanitize("thread")))
  #endif
  static int __Pyx_init_assertions_enabled(void) {
    PyObject *builtins, *debug, *debug_str;
    int flag;
    builtins = PyEval_GetBuiltins();
    if (!builtins) goto bad;
    debug_str = PyUnicode_FromStringAndSize("__debug__", 9);
    if (!debug_str) goto bad;
    debug = PyObject_GetItem(builtins, debug_str);
    Py_DECREF(debug_str);
    if (!debug) goto bad;
    flag = PyObject_IsTrue(debug);
    Py_DECREF(debug);
    if (flag == -1) goto bad;
    __pyx_assertions_enabled_flag = flag;
    return 0;
  bad:
    __pyx_assertions_enabled_flag = 1;
    return -1;
  }
#else
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (!Py_OptimizeFlag)
#endif

/* PyTypeError_Check.proto */
#define __Pyx_PyExc_TypeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_TypeError)

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* JoinPyUnicode.export */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* DivInt[long].proto */
static CYTHON_INLINE long __Pyx_div_long(long, long, int b_is_constant);

/* PyErrExceptionMatches.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyObjectGetAttrStrNoError.proto (used by GetBuiltinName) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto (used by GetModuleGlobalName) */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyDictVersioning.proto (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __Pyx_XNewRef(__pyx_dict_cached_value);\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* bytes_tailmatch.proto */
static int __Pyx_PyBytes_SingleTailmatch(PyObject* self, PyObject* arg,
                                         Py_ssize_t start, Py_ssize_t end, int direction);
static int __Pyx_PyBytes_Tailmatch(PyObject* self, PyObject* substr,
                                   Py_ssize_t start, Py_ssize_t end, int direction);

/* BuildPyUnicode.proto (used by COrdinalToPyUnicode) */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, const char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* COrdinalToPyUnicode.proto (used by CIntToPyUnicode) */
static CYTHON_INLINE int __Pyx_CheckUnicodeValue(int value);
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromOrdinal_Padded(int value, Py_ssize_t width, char padding_char);

/* GCCDiagnostics.proto (used by CIntToPyUnicode) */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* IncludeStdlibH.proto (used by CIntToPyUnicode) */
#include <stdlib.h>

/* CIntToPyUnicode.proto */
#define __Pyx_PyUnicode_From_int(value, width, padding_char, format_char) (\
    ((format_char) == ('c')) ?\
        __Pyx_uchar___Pyx_PyUnicode_From_int(value, width, padding_char) :\
        __Pyx____Pyx_PyUnicode_From_int(value, width, padding_char, format_char)\
    )
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char);
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char);

/* RejectKeywords.export */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds);

/* PyLongBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static CYTHON_INLINE PyObject* __Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyLong_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyObjectFastCallMethod.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyObject_FastCallMethod(name, args, nargsf) PyObject_VectorcallMethod(name, args, nargsf, NULL)
#else
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf);
#endif

/* PyObjectDelAttr.proto (used by PyObjectSetAttrStr) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
#define __Pyx_PyObject_DelAttr(o, n) PyObject_SetAttr(o, n, NULL)
#else
#define __Pyx_PyObject_DelAttr(o, n) PyObject_DelAttr(o, n)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   __Pyx_PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck, unsafe_shared) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck, int unsafe_shared);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck, unsafe_shared) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck, int unsafe_shared);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* PyKeyError_Check.proto */
#define __Pyx_PyExc_KeyError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_KeyError)

/* PyLongCompare.proto */
static CYTHON_INLINE int __Pyx_PyLong_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyObjectVectorCallKwBuilder.proto (used by PyObjectVectorCallMethodKwBuilder) */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* PyObjectVectorCallMethodKwBuilder.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_VectorcallMethod_CallFromBuilder PyObject_VectorcallMethod
#else
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#endif

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* GetTopmostException.proto (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* CIntToPyUnicode.proto */
#define __Pyx_PyUnicode_From_char(value, width, padding_char, format_char) (\
    ((format_char) == ('c')) ?\
        __Pyx_uchar___Pyx_PyUnicode_From_char(value, width, padding_char) :\
        __Pyx____Pyx_PyUnicode_From_char(value, width, padding_char, format_char)\
    )
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_char(char value, Py_ssize_t width, char padding_char);
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_char(char value, Py_ssize_t width, char padding_char, char format_char);

/* PyOSError_Check.proto */
#define __Pyx_PyExc_OSError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_OSError)

/* PyObjectCallNoArg.proto (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectGetMethod.proto (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);
#endif

/* PyObjectCallMethod0.proto (used by pop) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* pop.proto */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L);
#define __Pyx_PyObject_Pop(L) (likely(PyList_CheckExact(L)) ?\
    __Pyx_PyList_Pop(L) : __Pyx__PyObject_Pop(L))
#else
#define __Pyx_PyList_Pop(L)  __Pyx__PyObject_Pop(L)
#define __Pyx_PyObject_Pop(L)  __Pyx__PyObject_Pop(L)
#endif

/* PyStopIteration_Check.proto */
#define __Pyx_PyExc_StopIteration_Check(obj)  __Pyx_TypeCheck(obj, PyExc_StopIteration)

/* IterNextPlain.proto (used by IterNext) */
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next_Plain(PyObject *iterator);
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *__Pyx_GetBuiltinNext_LimitedAPI(void);
#endif

/* IterNext.proto */
#define __Pyx_PyIter_Next(obj) __Pyx_PyIter_Next2(obj, NULL)
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject *, PyObject *);

/* PyAttributeError_Check.proto */
#define __Pyx_PyExc_AttributeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_AttributeError)

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DelItemInt.proto */
#define __Pyx_DelItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_DelItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_DelItem_Generic(o, to_py_func(i))))
static int __Pyx_DelItem_Generic(PyObject *o, PyObject *j);
static CYTHON_INLINE int __Pyx_DelItemInt_Fast(PyObject *o, Py_ssize_t i,
                                               int is_list, int wraparound);

/* PyUnicodeContains.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_ContainsTF(PyObject* substring, PyObject* text, int eq) {
    int result = PyUnicode_Contains(text, substring);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyLongCompare.proto */
static CYTHON_INLINE int __Pyx_PyLong_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* ArgTypeTestFunc.export */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))

/* AllocateExtensionType.proto */
static PyObject *__Pyx_AllocateExtensionType(PyTypeObject *t, int is_final);

/* CallTypeTraverse.proto */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#define __Pyx_call_type_traverse(o, always_call, visit, arg) 0
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg);
#endif

/* LimitedApiGetTypeDict.proto (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp);
#endif

/* SetItemOnTypeDict.proto (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v);
#define __Pyx_SetItemOnTypeDict(tp, k, v) __Pyx__SetItemOnTypeDict((PyTypeObject*)tp, k, v)

/* FixUpExtensionType.proto */
static CYTHON_INLINE int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);

/* ValidateBasesTuple.proto (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto (used by MergeVTables) */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
static int __Pyx_MergeVtables(PyTypeObject *type);

/* DelItemOnTypeDict.proto (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k);
#define __Pyx_DelItemOnTypeDict(tp, k) __Pyx__DelItemOnTypeDict((PyTypeObject*)tp, k)

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_2_4
#define __PYX_HAVE_RT_ImportType_proto_3_2_4
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_2_4 {
   __Pyx_ImportType_CheckSize_Error_3_2_4 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_2_4 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_2_4 = 2
};
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size);
#endif

/* HasAttr.proto (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* ImportImpl.export */
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level);

/* Import.proto */
static CYTHON_INLINE PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* dict_setdefault.proto (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value);

/* AddModuleRef.proto (used by FetchSharedCythonModule) */
#if ((CYTHON_COMPILING_IN_CPYTHON_FREETHREADING ) ||\
     __PYX_LIMITED_VERSION_HEX < 0x030d0000)
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name);
#else
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#endif

/* FetchSharedCythonModule.proto (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto (used by CommonTypesMetaclass) */
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases);

/* CommonTypesMetaclass.proto (used by CythonFunctionShared) */
static int __pyx_CommonTypesMetaclass_init(PyObject *module);
#define __Pyx_CommonTypesMetaclass_USED

/* PyMethodNew.proto (used by CythonFunctionShared) */
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ);

/* PyVectorcallFastCallDict.proto (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto (used by CythonFunction) */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && CYTHON_METH_FASTCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_dict;
#endif
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    PyObject *defaults;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_mstate_global->__pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_mstate_global->__pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void));
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func,
                                                         PyTypeObject *defaults_type);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CLineInTraceback.proto (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto (used by AddTraceback) */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CheckUnpickleChecksum.proto */
static CYTHON_INLINE int __Pyx_CheckUnpickleChecksum(long checksum, long checksum1, long checksum2, long checksum3, const char *members);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_ptrdiff_t(ptrdiff_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_long(unsigned long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* UpdateUnpickledDict.proto */
static int __Pyx_UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
static unsigned long __Pyx_cached_runtime_version = 0;
static void __Pyx_init_runtime_version(void);
#else
#define __Pyx_init_runtime_version()
#endif
static unsigned long __Pyx_get_runtime_version(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* DecompressString.proto */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."

static char *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_get_next(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self, int *__pyx_v_size); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_get_next_str(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self); /* proto*/
static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader__init(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader__get_entry(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self, int __pyx_v_num_trees, void **__pyx_v_p_current_dirname, int *__pyx_v_new_block); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__process_entry(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_entry, PyObject *__pyx_v_path_info); /* proto*/
static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__gather_result_for_consistency(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_result); /* proto*/
static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__update_current_block(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__iter_next(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__maybe_tree_ref(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_current_path_info); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__loop_one_block(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__next_consistent_entries(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__path_info(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_utf8_path, PyObject *__pyx_v_unicode_path); /* proto*/

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.tuple" */

/* Module declarations from "breezy.bzr._str_helpers" */
static CYTHON_INLINE void *__pyx_f_6breezy_3bzr_12_str_helpers__my_memrchr(void *, int, size_t); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(char *, Py_ssize_t); /*proto*/

/* Module declarations from "breezy.bzr._dirstate_helpers_pyx" */
static int __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx_ERROR_PATH_NOT_FOUND;
static int __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx_ERROR_DIRECTORY;
static PyObject *__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_absent = 0;
static PyObject *__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_file = 0;
static PyObject *__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_directory = 0;
static PyObject *__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_symlink = 0;
static PyObject *__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_relocated = 0;
static PyObject *__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_tree_reference = 0;
static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__is_aligned(void *); /*proto*/
static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_by_dirs(char *, int, char *, int); /*proto*/
static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_path_by_dirblock_intern(char *, int, char *, int); /*proto*/
static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_minikind_from_mode(int); /*proto*/
static unsigned long __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__time_to_unsigned(PyObject *); /*proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__pack_stat(PyObject *); /*proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_update_entry(PyObject *, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static char __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_from_string(PyObject *); /*proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_to_kind(char); /*proto*/
static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__versioned_minikind(char); /*proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx___pyx_unpickle_Reader__set_state(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *, PyObject *); /*proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx___pyx_unpickle_ProcessEntryC__set_state(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *, PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "breezy.bzr._dirstate_helpers_pyx"
extern int __pyx_module_is_main_breezy__bzr___dirstate_helpers_pyx;
int __pyx_module_is_main_breezy__bzr___dirstate_helpers_pyx = 0;

/* Implementation of "breezy.bzr._dirstate_helpers_pyx" */
/* #### Code section: global_var ### */
/* #### Code section: string_decls ### */
static const char __pyx_k_pending_consistent_entries_bise[] = "_pending_consistent_entries, bisect_left, block_index, current_block, current_block_list, current_block_pos, current_dir_info, current_dir_list, current_root, current_root_unicode, dir_iterator, doing_consistency_expansion, fstat, include_unchanged, last_source_parent, last_target_parent, new_dirname_to_file_id, old_dirname_to_file_id, partial, path_index, pathjoin, root_abspath, root_dir_info, root_entries, root_entries_len, root_entries_pos, search_specific_file_parents, search_specific_files, searched_exact_paths, searched_specific_files, seen_ids, sha_file, source_index, state, target_index, tree, use_filesystem_for_exec, utf8_decode, want_unversioned";
static const char __pyx_k_Helper_functions_for_DirState_Th[] = "Helper functions for DirState.\n\nThis is the python implementation for DirState functions.\n";
static const char __pyx_k_cur_cstr_end_cstr_next_state_tex[] = "cur_cstr, end_cstr, next, state, text, text_cstr, text_size";
/* #### Code section: decls ### */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx__py_memrchr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_s, PyObject *__pyx_v_c); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_2lt_by_dirs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path1, PyObject *__pyx_v_path2); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_4_lt_path_by_dirblock(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path1, PyObject *__pyx_v_path2); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6_bisect_path_left(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_paths, PyObject *__pyx_v_path); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_8_bisect_path_right(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_paths, PyObject *__pyx_v_path); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_10bisect_dirblock(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_dirblocks, PyObject *__pyx_v_dirname, PyObject *__pyx_v_lo, PyObject *__pyx_v_hi, CYTHON_UNUSED PyObject *__pyx_v_cache); /* proto */
static int __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader___init__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self, PyObject *__pyx_v_text, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_2_parse_dirblocks(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_4__reduce_cython__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_6__setstate_cython__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_12_read_dirblocks(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_14pack_stat(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_stat_value); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_16update_entry(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_entry, PyObject *__pyx_v_abspath, PyObject *__pyx_v_stat_value); /* proto */
static int __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC___init__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_include_unchanged, PyObject *__pyx_v_use_filesystem_for_exec, PyObject *__pyx_v_search_specific_files, PyObject *__pyx_v_state, PyObject *__pyx_v_source_index, PyObject *__pyx_v_target_index, PyObject *__pyx_v_want_unversioned, PyObject *__pyx_v_tree); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_2__iter__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_4iter_changes(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_6__next__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_23searched_specific_files___get__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_20searched_exact_paths___get__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_8__reduce_cython__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_10__setstate_cython__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_18__pyx_unpickle_Reader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_20__pyx_unpickle_ProcessEntryC(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_6breezy_3bzr_21_dirstate_helpers_pyx_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  PyObject *__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader;
  PyObject *__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC;
  PyTypeObject *__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader;
  PyTypeObject *__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_values;
  __Pyx_CachedCFunction __pyx_umethod_PyList_Type_pop;
  PyObject *__pyx_slice[3];
  PyObject *__pyx_tuple[10];
  PyObject *__pyx_codeobj_tab[17];
  PyObject *__pyx_string_tab[272];
  PyObject *__pyx_number_tab[7];
/* #### Code section: module_state_contents ### */
/* IterNextPlain.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
PyObject *__Pyx_GetBuiltinNext_LimitedAPI_cache;
#endif

/* CommonTypesMetaclass.module_state_decls */
PyTypeObject *__pyx_CommonTypesMetaclassType;

/* CachedMethodType.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API
PyObject *__Pyx_CachedMethodType;
#endif

/* CythonFunctionShared.module_state_decls */
PyTypeObject *__pyx_CyFunctionType;

/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_ __pyx_string_tab[0]
#define __pyx_kp_u_Bad_parse_we_expected_to_end_on __pyx_string_tab[1]
#define __pyx_kp_u_Could_not_find_target_parent_in __pyx_string_tab[2]
#define __pyx_kp_u_Expected_single_character_string __pyx_string_tab[3]
#define __pyx_kp_u_First_character_should_be_null_n __pyx_string_tab[4]
#define __pyx_kp_u_Got_entry_path_mismatch_for_spec __pyx_string_tab[5]
#define __pyx_kp_u_Missing_entry_for_specific_path __pyx_string_tab[6]
#define __pyx_kp_u_Note_that_Cython_is_deliberately __pyx_string_tab[7]
#define __pyx_kp_u_Unsupported_target_index_d __pyx_string_tab[8]
#define __pyx_kp_u_We_could_not_find_the_parent_ent __pyx_string_tab[9]
#define __pyx_kp_u_We_read_the_wrong_number_of_entr __pyx_string_tab[10]
#define __pyx_kp_u__2 __pyx_string_tab[11]
#define __pyx_kp_u__4 __pyx_string_tab[12]
#define __pyx_kp_u__5 __pyx_string_tab[13]
#define __pyx_kp_u__6 __pyx_string_tab[14]
#define __pyx_kp_u__7 __pyx_string_tab[15]
#define __pyx_kp_u__8 __pyx_string_tab[16]
#define __pyx_kp_u_add_note __pyx_string_tab[17]
#define __pyx_kp_u_breezy __pyx_string_tab[18]
#define __pyx_kp_u_breezy_bzr__dirstate_helpers_pyx_2 __pyx_string_tab[19]
#define __pyx_kp_u_breezy_bzr_dirstate __pyx_string_tab[20]
#define __pyx_kp_u_breezy_bzr_inventorytree __pyx_string_tab[21]
#define __pyx_kp_u_breezy_osutils __pyx_string_tab[22]
#define __pyx_kp_u_but_read __pyx_string_tab[23]
#define __pyx_kp_u_but_source_does_not_exist_entry __pyx_string_tab[24]
#define __pyx_kp_u_disable __pyx_string_tab[25]
#define __pyx_kp_u_don_t_know_how_to_compare_source __pyx_string_tab[26]
#define __pyx_kp_u_enable __pyx_string_tab[27]
#define __pyx_kp_u_entry_2 __pyx_string_tab[28]
#define __pyx_kp_u_entry_3 __pyx_string_tab[29]
#define __pyx_kp_u_failed_to_find_trailing_NULL_0_T __pyx_string_tab[30]
#define __pyx_kp_u_for_the_entry __pyx_string_tab[31]
#define __pyx_kp_u_gc __pyx_string_tab[32]
#define __pyx_kp_u_get_next_called_when_cur_str_is __pyx_string_tab[33]
#define __pyx_kp_u_get_next_called_when_there_are_n __pyx_string_tab[34]
#define __pyx_kp_u_is_considered_renamed_from __pyx_string_tab[35]
#define __pyx_kp_u_isenabled __pyx_string_tab[36]
#define __pyx_kp_u_parent_of __pyx_string_tab[37]
#define __pyx_kp_u_path1_must_be_a_bytes_string_no __pyx_string_tab[38]
#define __pyx_kp_u_path1_must_be_a_plain_string_no __pyx_string_tab[39]
#define __pyx_kp_u_path2_must_be_a_bytes_string_no __pyx_string_tab[40]
#define __pyx_kp_u_path2_must_be_a_plain_string_no __pyx_string_tab[41]
#define __pyx_kp_u_path_info __pyx_string_tab[42]
#define __pyx_kp_u_result_is_not_None_r __pyx_string_tab[43]
#define __pyx_kp_u_target_minikind __pyx_string_tab[44]
#define __pyx_kp_u_tree_fragment __pyx_string_tab[45]
#define __pyx_kp_u_tree_reference __pyx_string_tab[46]
#define __pyx_kp_u_tried_to_create_a_string_with_an __pyx_string_tab[47]
#define __pyx_kp_u_unsupported_target_index __pyx_string_tab[48]
#define __pyx_kp_u_you_must_pass_a_python_list_for __pyx_string_tab[49]
#define __pyx_kp_u_you_must_pass_a_python_list_for_2 __pyx_string_tab[50]
#define __pyx_kp_u_you_must_pass_a_string_for_dirna __pyx_string_tab[51]
#define __pyx_kp_u_you_must_pass_a_string_for_path __pyx_string_tab[52]
#define __pyx_n_u_BadFileKindError __pyx_string_tab[53]
#define __pyx_n_u_BzrError __pyx_string_tab[54]
#define __pyx_n_u_DirState __pyx_string_tab[55]
#define __pyx_n_u_DirstateCorrupt __pyx_string_tab[56]
#define __pyx_n_u_EINVAL __pyx_string_tab[57]
#define __pyx_n_u_ENOENT __pyx_string_tab[58]
#define __pyx_n_u_ENOTDIR __pyx_string_tab[59]
#define __pyx_n_u_IN_MEMORY_UNMODIFIED __pyx_string_tab[60]
#define __pyx_n_u_InventoryTreeChange __pyx_string_tab[61]
#define __pyx_n_u_NULLSTAT __pyx_string_tab[62]
#define __pyx_n_u_NULL_PARENT_DETAILS __pyx_string_tab[63]
#define __pyx_n_u_ProcessEntryC __pyx_string_tab[64]
#define __pyx_n_u_ProcessEntryC___reduce_cython __pyx_string_tab[65]
#define __pyx_n_u_ProcessEntryC___setstate_cython __pyx_string_tab[66]
#define __pyx_n_u_ProcessEntryC_iter_changes __pyx_string_tab[67]
#define __pyx_n_u_Pyx_PyDict_NextRef __pyx_string_tab[68]
#define __pyx_n_u_Reader __pyx_string_tab[69]
#define __pyx_n_u_Reader___reduce_cython __pyx_string_tab[70]
#define __pyx_n_u_Reader___setstate_cython __pyx_string_tab[71]
#define __pyx_n_u_Reader__parse_dirblocks __pyx_string_tab[72]
#define __pyx_n_u_S_IEXEC __pyx_string_tab[73]
#define __pyx_n_u_S_ISREG __pyx_string_tab[74]
#define __pyx_n_u__4 __pyx_string_tab[75]
#define __pyx_n_u_absent __pyx_string_tab[76]
#define __pyx_n_u_abspath __pyx_string_tab[77]
#define __pyx_n_u_add __pyx_string_tab[78]
#define __pyx_n_u_asyncio_coroutines __pyx_string_tab[79]
#define __pyx_n_u_b2a_base64 __pyx_string_tab[80]
#define __pyx_n_u_binascii __pyx_string_tab[81]
#define __pyx_n_u_bisect __pyx_string_tab[82]
#define __pyx_n_u_bisect_dirblock __pyx_string_tab[83]
#define __pyx_n_u_bisect_left __pyx_string_tab[84]
#define __pyx_n_u_bisect_path_left __pyx_string_tab[85]
#define __pyx_n_u_bisect_path_right __pyx_string_tab[86]
#define __pyx_n_u_breezy_bzr__dirstate_helpers_pyx __pyx_string_tab[87]
#define __pyx_n_u_c __pyx_string_tab[88]
#define __pyx_n_u_c_2 __pyx_string_tab[89]
#define __pyx_n_u_cache __pyx_string_tab[90]
#define __pyx_n_u_class_getitem __pyx_string_tab[91]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[92]
#define __pyx_n_u_codecs __pyx_string_tab[93]
#define __pyx_n_u_cur __pyx_string_tab[94]
#define __pyx_n_u_cur_cstr __pyx_string_tab[95]
#define __pyx_n_u_cur_size __pyx_string_tab[96]
#define __pyx_n_u_current_block __pyx_string_tab[97]
#define __pyx_n_u_current_dirname __pyx_string_tab[98]
#define __pyx_n_u_cutoff_time __pyx_string_tab[99]
#define __pyx_n_u_d __pyx_string_tab[100]
#define __pyx_n_u_decode __pyx_string_tab[101]
#define __pyx_n_u_dict __pyx_string_tab[102]
#define __pyx_n_u_dict_2 __pyx_string_tab[103]
#define __pyx_n_u_dirblock_state __pyx_string_tab[104]
#define __pyx_n_u_dirblocks __pyx_string_tab[105]
#define __pyx_n_u_dirblocks_2 __pyx_string_tab[106]
#define __pyx_n_u_directory __pyx_string_tab[107]
#define __pyx_n_u_directory_is_tree_reference __pyx_string_tab[108]
#define __pyx_n_u_dirname __pyx_string_tab[109]
#define __pyx_n_u_dirname_cstr __pyx_string_tab[110]
#define __pyx_n_u_dirname_size __pyx_string_tab[111]
#define __pyx_n_u_dirstate __pyx_string_tab[112]
#define __pyx_n_u_encode __pyx_string_tab[113]
#define __pyx_n_u_encode_2 __pyx_string_tab[114]
#define __pyx_n_u_end_of_header __pyx_string_tab[115]
#define __pyx_n_u_ensure_block __pyx_string_tab[116]
#define __pyx_n_u_entries_for_path __pyx_string_tab[117]
#define __pyx_n_u_entry __pyx_string_tab[118]
#define __pyx_n_u_entry_count __pyx_string_tab[119]
#define __pyx_n_u_errno __pyx_string_tab[120]
#define __pyx_n_u_errors __pyx_string_tab[121]
#define __pyx_n_u_expected_entry_count __pyx_string_tab[122]
#define __pyx_n_u_file __pyx_string_tab[123]
#define __pyx_n_u_file_id __pyx_string_tab[124]
#define __pyx_n_u_file_kind_from_stat_mode __pyx_string_tab[125]
#define __pyx_n_u_filename __pyx_string_tab[126]
#define __pyx_n_u_find_block_index_from_key __pyx_string_tab[127]
#define __pyx_n_u_found __pyx_string_tab[128]
#define __pyx_n_u_fstat __pyx_string_tab[129]
#define __pyx_n_u_func __pyx_string_tab[130]
#define __pyx_n_u_get_block_entry_index __pyx_string_tab[131]
#define __pyx_n_u_get_entry __pyx_string_tab[132]
#define __pyx_n_u_getstate __pyx_string_tab[133]
#define __pyx_n_u_hi __pyx_string_tab[134]
#define __pyx_n_u_hi_2 __pyx_string_tab[135]
#define __pyx_n_u_include_unchanged __pyx_string_tab[136]
#define __pyx_n_u_int __pyx_string_tab[137]
#define __pyx_n_u_inventorytree __pyx_string_tab[138]
#define __pyx_n_u_is_coroutine __pyx_string_tab[139]
#define __pyx_n_u_is_executable __pyx_string_tab[140]
#define __pyx_n_u_is_inside __pyx_string_tab[141]
#define __pyx_n_u_is_inside_any __pyx_string_tab[142]
#define __pyx_n_u_items __pyx_string_tab[143]
#define __pyx_n_u_iter_changes __pyx_string_tab[144]
#define __pyx_n_u_kind __pyx_string_tab[145]
#define __pyx_n_u_length __pyx_string_tab[146]
#define __pyx_n_u_lo __pyx_string_tab[147]
#define __pyx_n_u_lo_2 __pyx_string_tab[148]
#define __pyx_n_u_lstat __pyx_string_tab[149]
#define __pyx_n_u_lt_by_dirs __pyx_string_tab[150]
#define __pyx_n_u_lt_path_by_dirblock __pyx_string_tab[151]
#define __pyx_n_u_main __pyx_string_tab[152]
#define __pyx_n_u_mark_modified __pyx_string_tab[153]
#define __pyx_n_u_mid __pyx_string_tab[154]
#define __pyx_n_u_module __pyx_string_tab[155]
#define __pyx_n_u_name __pyx_string_tab[156]
#define __pyx_n_u_ne __pyx_string_tab[157]
#define __pyx_n_u_new __pyx_string_tab[158]
#define __pyx_n_u_new_block __pyx_string_tab[159]
#define __pyx_n_u_next __pyx_string_tab[160]
#define __pyx_n_u_num_entries __pyx_string_tab[161]
#define __pyx_n_u_num_present_parents __pyx_string_tab[162]
#define __pyx_n_u_num_trees __pyx_string_tab[163]
#define __pyx_n_u_obj __pyx_string_tab[164]
#define __pyx_n_u_observed_sha1 __pyx_string_tab[165]
#define __pyx_n_u_os __pyx_string_tab[166]
#define __pyx_n_u_osutils __pyx_string_tab[167]
#define __pyx_n_u_pack_stat __pyx_string_tab[168]
#define __pyx_n_u_parent_directories __pyx_string_tab[169]
#define __pyx_n_u_parse_dirblocks __pyx_string_tab[170]
#define __pyx_n_u_path __pyx_string_tab[171]
#define __pyx_n_u_path1 __pyx_string_tab[172]
#define __pyx_n_u_path2 __pyx_string_tab[173]
#define __pyx_n_u_path_cstr __pyx_string_tab[174]
#define __pyx_n_u_path_size __pyx_string_tab[175]
#define __pyx_n_u_path_utf8 __pyx_string_tab[176]
#define __pyx_n_u_pathjoin __pyx_string_tab[177]
#define __pyx_n_u_paths __pyx_string_tab[178]
#define __pyx_n_u_platform __pyx_string_tab[179]
#define __pyx_n_u_pop __pyx_string_tab[180]
#define __pyx_n_u_prefix __pyx_string_tab[181]
#define __pyx_n_u_py_memrchr __pyx_string_tab[182]
#define __pyx_n_u_pyx_checksum __pyx_string_tab[183]
#define __pyx_n_u_pyx_result __pyx_string_tab[184]
#define __pyx_n_u_pyx_state __pyx_string_tab[185]
#define __pyx_n_u_pyx_type __pyx_string_tab[186]
#define __pyx_n_u_pyx_unpickle_ProcessEntryC __pyx_string_tab[187]
#define __pyx_n_u_pyx_unpickle_Reader __pyx_string_tab[188]
#define __pyx_n_u_pyx_vtable __pyx_string_tab[189]
#define __pyx_n_u_qualname __pyx_string_tab[190]
#define __pyx_n_u_read __pyx_string_tab[191]
#define __pyx_n_u_read_dirblocks __pyx_string_tab[192]
#define __pyx_n_u_read_link __pyx_string_tab[193]
#define __pyx_n_u_reader __pyx_string_tab[194]
#define __pyx_n_u_reduce __pyx_string_tab[195]
#define __pyx_n_u_reduce_cython __pyx_string_tab[196]
#define __pyx_n_u_reduce_ex __pyx_string_tab[197]
#define __pyx_n_u_relocated __pyx_string_tab[198]
#define __pyx_n_u_rsplit __pyx_string_tab[199]
#define __pyx_n_u_s __pyx_string_tab[200]
#define __pyx_n_u_s_2 __pyx_string_tab[201]
#define __pyx_n_u_search_specific_files __pyx_string_tab[202]
#define __pyx_n_u_seek __pyx_string_tab[203]
#define __pyx_n_u_self __pyx_string_tab[204]
#define __pyx_n_u_set_name __pyx_string_tab[205]
#define __pyx_n_u_setdefault __pyx_string_tab[206]
#define __pyx_n_u_setstate __pyx_string_tab[207]
#define __pyx_n_u_setstate_cython __pyx_string_tab[208]
#define __pyx_n_u_sha1_file __pyx_string_tab[209]
#define __pyx_n_u_sha1_provider __pyx_string_tab[210]
#define __pyx_n_u_sha_cutoff_time __pyx_string_tab[211]
#define __pyx_n_u_sha_file __pyx_string_tab[212]
#define __pyx_n_u_source_index __pyx_string_tab[213]
#define __pyx_n_u_split __pyx_string_tab[214]
#define __pyx_n_u_split_root_dirblock_into_conten __pyx_string_tab[215]
#define __pyx_n_u_splitpath __pyx_string_tab[216]
#define __pyx_n_u_st_ctime __pyx_string_tab[217]
#define __pyx_n_u_st_dev __pyx_string_tab[218]
#define __pyx_n_u_st_ino __pyx_string_tab[219]
#define __pyx_n_u_st_mode __pyx_string_tab[220]
#define __pyx_n_u_st_mtime __pyx_string_tab[221]
#define __pyx_n_u_st_size __pyx_string_tab[222]
#define __pyx_n_u_stat __pyx_string_tab[223]
#define __pyx_n_u_stat_and_sha1 __pyx_string_tab[224]
#define __pyx_n_u_stat_value __pyx_string_tab[225]
#define __pyx_n_u_state __pyx_string_tab[226]
#define __pyx_n_u_state_file __pyx_string_tab[227]
#define __pyx_n_u_surrogateescape __pyx_string_tab[228]
#define __pyx_n_u_symlink __pyx_string_tab[229]
#define __pyx_n_u_sys __pyx_string_tab[230]
#define __pyx_n_u_target_index __pyx_string_tab[231]
#define __pyx_n_u_test __pyx_string_tab[232]
#define __pyx_n_u_text __pyx_string_tab[233]
#define __pyx_n_u_tree __pyx_string_tab[234]
#define __pyx_n_u_update __pyx_string_tab[235]
#define __pyx_n_u_update_entry __pyx_string_tab[236]
#define __pyx_n_u_use_filesystem_for_exec __pyx_string_tab[237]
#define __pyx_n_u_use_setstate __pyx_string_tab[238]
#define __pyx_n_u_utf8 __pyx_string_tab[239]
#define __pyx_n_u_utf_8_decode __pyx_string_tab[240]
#define __pyx_n_u_values __pyx_string_tab[241]
#define __pyx_n_u_walkdirs_utf8 __pyx_string_tab[242]
#define __pyx_n_u_want_unversioned __pyx_string_tab[243]
#define __pyx_n_u_win32 __pyx_string_tab[244]
#define __pyx_n_u_winerror __pyx_string_tab[245]
#define __pyx_kp_b__3 __pyx_string_tab[246]
#define __pyx_kp_b__4 __pyx_string_tab[247]
#define __pyx_kp_b__5 __pyx_string_tab[248]
#define __pyx_kp_b_bzr __pyx_string_tab[249]
#define __pyx_kp_b_iso88591_1F __pyx_string_tab[250]
#define __pyx_kp_b_iso88591_6_t_AQ_iq_q_t_Qa_iq_q_q_Qa_AQ_b __pyx_string_tab[251]
#define __pyx_kp_b_iso88591_6_t_AQ_iq_q_t_Qa_iq_q_q_Qa_AQ_b_2 __pyx_string_tab[252]
#define __pyx_kp_b_iso88591_AV1 __pyx_string_tab[253]
#define __pyx_kp_b_iso88591_A_D_33b_t6_F_Be_E_F_a_A_a_d_Bd_D __pyx_string_tab[254]
#define __pyx_kp_b_iso88591_A_q __pyx_string_tab[255]
#define __pyx_kp_b_iso88591_A_r_A_q_AQ_a_PTTU_6OqPYY_y_5_O3 __pyx_string_tab[256]
#define __pyx_kp_b_iso88591_Qa __pyx_string_tab[257]
#define __pyx_kp_b_iso88591_T_D_4wd_gTQ_aab_G1F_a_vWE_Q_q_t __pyx_string_tab[258]
#define __pyx_kp_b_iso88591_T_t_SWWggk_l_A_A_E_E_Y_Y_p_p_t __pyx_string_tab[259]
#define __pyx_kp_b_iso88591_U_5_5_E_V1F_1_HA __pyx_string_tab[260]
#define __pyx_kp_b_iso88591_iq_t_AQ_iq_q_A_t_Qa_iq_q_s_Q_c __pyx_string_tab[261]
#define __pyx_kp_b_iso88591_q_0_kQR_6_7_1 __pyx_string_tab[262]
#define __pyx_kp_b_iso88591_q_0_kQR_7_q0_a_1 __pyx_string_tab[263]
#define __pyx_kp_b_iso88591_q_3c_K2Q_Kq_Bat1_vS_q_7 __pyx_string_tab[264]
#define __pyx_kp_b_iso88591_t_Qa_iq_q_t_Qa_iq_q_9_5Qa_9_5Qh __pyx_string_tab[265]
#define __pyx_kp_b_iso88591_t_Qa_iq_q_t_Qa_iq_q_Bc_Q_aq_1_a __pyx_string_tab[266]
#define __pyx_n_b_a __pyx_string_tab[267]
#define __pyx_n_b_d __pyx_string_tab[268]
#define __pyx_n_b_f __pyx_string_tab[269]
#define __pyx_n_b_l __pyx_string_tab[270]
#define __pyx_n_b_r __pyx_string_tab[271]
#define __pyx_int_0 __pyx_number_tab[0]
#define __pyx_int_neg_1 __pyx_number_tab[1]
#define __pyx_int_1 __pyx_number_tab[2]
#define __pyx_int_2 __pyx_number_tab[3]
#define __pyx_int_3 __pyx_number_tab[4]
#define __pyx_int_37455648 __pyx_number_tab[5]
#define __pyx_int_267571029 __pyx_number_tab[6]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader);
  Py_CLEAR(clear_module_state->__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC);
  Py_CLEAR(clear_module_state->__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC);
  for (int i=0; i<3; ++i) { Py_CLEAR(clear_module_state->__pyx_slice[i]); }
  for (int i=0; i<10; ++i) { Py_CLEAR(clear_module_state->__pyx_tuple[i]); }
  for (int i=0; i<17; ++i) { Py_CLEAR(clear_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<272; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<7; ++i) { Py_CLEAR(clear_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_clear_contents ### */
/* CommonTypesMetaclass.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_clear_end ### */
return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader);
  Py_VISIT(traverse_module_state->__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC);
  Py_VISIT(traverse_module_state->__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC);
  for (int i=0; i<3; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_slice[i]); }
  for (int i=0; i<10; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_tuple[i]); }
  for (int i=0; i<17; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<272; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<7; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_traverse_contents ### */
/* CommonTypesMetaclass.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_traverse_end ### */
return 0;
}
#endif
/* #### Code section: module_code ### */

/* "_str_helpers.pxd":27
 * 
 * 
 * cdef inline void* _my_memrchr(void *s, int c, size_t n): # cannot_raise             # <<<<<<<<<<<<<<
 *     # memrchr seems to be a GNU extension, so we have to implement it ourselves
 *     cdef char *pos
*/

static CYTHON_INLINE void *__pyx_f_6breezy_3bzr_12_str_helpers__my_memrchr(void *__pyx_v_s, int __pyx_v_c, size_t __pyx_v_n) {
  char *__pyx_v_pos;
  char *__pyx_v_start;
  void *__pyx_r;
  int __pyx_t_1;

  /* "_str_helpers.pxd":32
 *     cdef char *start
 * 
 *     start = <char*>s             # <<<<<<<<<<<<<<
 *     pos = start + n - 1
 *     while pos >= start:
*/
  __pyx_v_start = ((char *)__pyx_v_s);

  /* "_str_helpers.pxd":33
 * 
 *     start = <char*>s
 *     pos = start + n - 1             # <<<<<<<<<<<<<<
 *     while pos >= start:
 *         if pos[0] == c:
*/
  __pyx_v_pos = ((__pyx_v_start + __pyx_v_n) - 1);

  /* "_str_helpers.pxd":34
 *     start = <char*>s
 *     pos = start + n - 1
 *     while pos >= start:             # <<<<<<<<<<<<<<
 *         if pos[0] == c:
 *             return <void*>pos
*/
  while (1) {
    __pyx_t_1 = (__pyx_v_pos >= __pyx_v_start);
    if (!__pyx_t_1) break;

    /* "_str_helpers.pxd":35
 *     pos = start + n - 1
 *     while pos >= start:
 *         if pos[0] == c:             # <<<<<<<<<<<<<<
 *             return <void*>pos
 *         pos = pos - 1
*/
    __pyx_t_1 = ((__pyx_v_pos[0]) == __pyx_v_c);
    if (__pyx_t_1) {

      /* "_str_helpers.pxd":36
 *     while pos >= start:
 *         if pos[0] == c:
 *             return <void*>pos             # <<<<<<<<<<<<<<
 *         pos = pos - 1
 *     return NULL
*/
      __pyx_r = ((void *)__pyx_v_pos);
      goto __pyx_L0;

      /* "_str_helpers.pxd":35
 *     pos = start + n - 1
 *     while pos >= start:
 *         if pos[0] == c:             # <<<<<<<<<<<<<<
 *             return <void*>pos
 *         pos = pos - 1
*/
    }

    /* "_str_helpers.pxd":37
 *         if pos[0] == c:
 *             return <void*>pos
 *         pos = pos - 1             # <<<<<<<<<<<<<<
 *     return NULL
 * 
*/
    __pyx_v_pos = (__pyx_v_pos - 1);
  }

  /* "_str_helpers.pxd":38
 *             return <void*>pos
 *         pos = pos - 1
 *     return NULL             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = NULL;
  goto __pyx_L0;

  /* "_str_helpers.pxd":27
 * 
 * 
 * cdef inline void* _my_memrchr(void *s, int c, size_t n): # cannot_raise             # <<<<<<<<<<<<<<
 *     # memrchr seems to be a GNU extension, so we have to implement it ourselves
 *     cdef char *pos
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "_str_helpers.pxd":41
 * 
 * 
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
*/

static CYTHON_INLINE PyObject *__pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(char *__pyx_v_s, Py_ssize_t __pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("safe_string_from_size", 0);

  /* "_str_helpers.pxd":42
 * 
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
*/
  __pyx_t_1 = (__pyx_v_size < 0);
  if (unlikely(__pyx_t_1)) {

    /* "_str_helpers.pxd":43
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:
 *         raise AssertionError(             # <<<<<<<<<<<<<<
 *             'tried to create a string with an invalid size: %d' % size)
 *     return PyBytes_FromStringAndSize(s, size)
*/
    __pyx_t_3 = NULL;

    /* "_str_helpers.pxd":44
 *     if size < 0:
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)             # <<<<<<<<<<<<<<
 *     return PyBytes_FromStringAndSize(s, size)
 * 
*/
    __pyx_t_4 = PyLong_FromSsize_t(__pyx_v_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyUnicode_Format(__pyx_mstate_global->__pyx_kp_u_tried_to_create_a_string_with_an, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 43, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 43, __pyx_L1_error)

    /* "_str_helpers.pxd":42
 * 
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
*/
  }

  /* "_str_helpers.pxd":45
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
 *     return PyBytes_FromStringAndSize(s, size)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyBytes_FromStringAndSize(__pyx_v_s, __pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "_str_helpers.pxd":41
 * 
 * 
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("breezy.bzr._str_helpers.safe_string_from_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_str_helpers.pxd":48
 * 
 * 
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
*/

static CYTHON_INLINE PyObject *__pyx_f_6breezy_3bzr_12_str_helpers_safe_interned_string_from_size(char *__pyx_v_s, Py_ssize_t __pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("safe_interned_string_from_size", 0);

  /* "_str_helpers.pxd":49
 * 
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
*/
  __pyx_t_1 = (__pyx_v_size < 0);
  if (unlikely(__pyx_t_1)) {

    /* "_str_helpers.pxd":50
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:
 *         raise AssertionError(             # <<<<<<<<<<<<<<
 *             'tried to create a string with an invalid size: %d' % size)
 *     return PyBytes_InternFromStringAndSize(s, size)
*/
    __pyx_t_3 = NULL;

    /* "_str_helpers.pxd":51
 *     if size < 0:
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)             # <<<<<<<<<<<<<<
 *     return PyBytes_InternFromStringAndSize(s, size)
*/
    __pyx_t_4 = PyLong_FromSsize_t(__pyx_v_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyUnicode_Format(__pyx_mstate_global->__pyx_kp_u_tried_to_create_a_string_with_an, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 50, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 50, __pyx_L1_error)

    /* "_str_helpers.pxd":49
 * 
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
*/
  }

  /* "_str_helpers.pxd":52
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
 *     return PyBytes_InternFromStringAndSize(s, size)             # <<<<<<<<<<<<<<
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyBytes_InternFromStringAndSize(__pyx_v_s, __pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "_str_helpers.pxd":48
 * 
 * 
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("breezy.bzr._str_helpers.safe_interned_string_from_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":131
 * 
 * 
 * def _py_memrchr(s, c):             # <<<<<<<<<<<<<<
 *     """Just to expose _my_memrchr for testing.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_1_py_memrchr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx__py_memrchr, "Just to expose _my_memrchr for testing.\n\n    :param s: The Python string to search\n    :param c: The character to search for\n    :return: The offset to the last instance of 'c' in s\n    ");
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_1_py_memrchr = {"_py_memrchr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_1_py_memrchr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx__py_memrchr};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_1_py_memrchr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_s = 0;
  PyObject *__pyx_v_c = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_py_memrchr (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_s,&__pyx_mstate_global->__pyx_n_u_c,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 131, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 131, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 131, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_py_memrchr", 0) < (0)) __PYX_ERR(0, 131, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_py_memrchr", 1, 2, 2, i); __PYX_ERR(0, 131, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 131, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 131, __pyx_L3_error)
    }
    __pyx_v_s = values[0];
    __pyx_v_c = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_py_memrchr", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 131, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._py_memrchr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx__py_memrchr(__pyx_self, __pyx_v_s, __pyx_v_c);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx__py_memrchr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_s, PyObject *__pyx_v_c) {
  void *__pyx_v__s;
  void *__pyx_v_found;
  int __pyx_v_length;
  char *__pyx_v__c;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  void *__pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_py_memrchr", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":143
 *     cdef char *_c
 * 
 *     assert PyBytes_Size(c) == 1, 'Expected single character string not %r' % c             # <<<<<<<<<<<<<<
 *     _c = PyBytes_AsString(c)
 *     _s = PyBytes_AsString(s)
*/
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (PyBytes_Size(__pyx_v_c) == 1);
    if (unlikely(!__pyx_t_1)) {
      __pyx_t_2 = __Pyx_PyUnicode_FormatSafe(__pyx_mstate_global->__pyx_kp_u_Expected_single_character_string, __pyx_v_c); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_t_2, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 143, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 143, __pyx_L1_error)
  #endif

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":144
 * 
 *     assert PyBytes_Size(c) == 1, 'Expected single character string not %r' % c
 *     _c = PyBytes_AsString(c)             # <<<<<<<<<<<<<<
 *     _s = PyBytes_AsString(s)
 *     length = PyBytes_Size(s)
*/
  __pyx_v__c = PyBytes_AsString(__pyx_v_c);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":145
 *     assert PyBytes_Size(c) == 1, 'Expected single character string not %r' % c
 *     _c = PyBytes_AsString(c)
 *     _s = PyBytes_AsString(s)             # <<<<<<<<<<<<<<
 *     length = PyBytes_Size(s)
 * 
*/
  __pyx_v__s = PyBytes_AsString(__pyx_v_s);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":146
 *     _c = PyBytes_AsString(c)
 *     _s = PyBytes_AsString(s)
 *     length = PyBytes_Size(s)             # <<<<<<<<<<<<<<
 * 
 *     found = _my_memrchr(_s, _c[0], length)
*/
  __pyx_v_length = PyBytes_Size(__pyx_v_s);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":148
 *     length = PyBytes_Size(s)
 * 
 *     found = _my_memrchr(_s, _c[0], length)             # <<<<<<<<<<<<<<
 *     if found == NULL:
 *         return None
*/
  __pyx_t_3 = __pyx_f_6breezy_3bzr_12_str_helpers__my_memrchr(__pyx_v__s, (__pyx_v__c[0]), __pyx_v_length); if (unlikely(__pyx_t_3 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 148, __pyx_L1_error)
  __pyx_v_found = __pyx_t_3;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":149
 * 
 *     found = _my_memrchr(_s, _c[0], length)
 *     if found == NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <char*>found - <char*>_s
*/
  __pyx_t_1 = (__pyx_v_found == NULL);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":150
 *     found = _my_memrchr(_s, _c[0], length)
 *     if found == NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <char*>found - <char*>_s
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":149
 * 
 *     found = _my_memrchr(_s, _c[0], length)
 *     if found == NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <char*>found - <char*>_s
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":151
 *     if found == NULL:
 *         return None
 *     return <char*>found - <char*>_s             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyLong_From_ptrdiff_t((((char *)__pyx_v_found) - ((char *)__pyx_v__s))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":131
 * 
 * 
 * def _py_memrchr(s, c):             # <<<<<<<<<<<<<<
 *     """Just to expose _my_memrchr for testing.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._py_memrchr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":154
 * 
 * 
 * cdef int _is_aligned(void *ptr): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Is this pointer aligned to an integer size offset?
 * 
*/

static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__is_aligned(void *__pyx_v_ptr) {
  int __pyx_r;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":159
 *     :return: 1 if this pointer is aligned, 0 otherwise.
 *     """
 *     return ((<intptr_t>ptr) & ((sizeof(int))-1)) == 0             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = ((((intptr_t)__pyx_v_ptr) & ((sizeof(int)) - 1)) == 0);
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":154
 * 
 * 
 * cdef int _is_aligned(void *ptr): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Is this pointer aligned to an integer size offset?
 * 
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":162
 * 
 * 
 * cdef int _cmp_by_dirs(char *path1, int size1, char *path2, int size2): # cannot_raise             # <<<<<<<<<<<<<<
 *     cdef unsigned char *cur1
 *     cdef unsigned char *cur2
*/

static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_by_dirs(char *__pyx_v_path1, int __pyx_v_size1, char *__pyx_v_path2, int __pyx_v_size2) {
  unsigned char *__pyx_v_cur1;
  unsigned char *__pyx_v_cur2;
  unsigned char *__pyx_v_end1;
  unsigned char *__pyx_v_end2;
  int *__pyx_v_cur_int1;
  int *__pyx_v_cur_int2;
  int *__pyx_v_end_int1;
  int *__pyx_v_end_int2;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":172
 *     cdef int *end_int2
 * 
 *     if path1 == path2 and size1 == size2:             # <<<<<<<<<<<<<<
 *         return 0
 * 
*/
  __pyx_t_2 = (__pyx_v_path1 == __pyx_v_path2);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_size1 == __pyx_v_size2);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":173
 * 
 *     if path1 == path2 and size1 == size2:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     end1 = <unsigned char*>path1+size1
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":172
 *     cdef int *end_int2
 * 
 *     if path1 == path2 and size1 == size2:             # <<<<<<<<<<<<<<
 *         return 0
 * 
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":175
 *         return 0
 * 
 *     end1 = <unsigned char*>path1+size1             # <<<<<<<<<<<<<<
 *     end2 = <unsigned char*>path2+size2
 * 
*/
  __pyx_v_end1 = (((unsigned char *)__pyx_v_path1) + __pyx_v_size1);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":176
 * 
 *     end1 = <unsigned char*>path1+size1
 *     end2 = <unsigned char*>path2+size2             # <<<<<<<<<<<<<<
 * 
 *     # Use 32-bit comparisons for the matching portion of the string.
*/
  __pyx_v_end2 = (((unsigned char *)__pyx_v_path2) + __pyx_v_size2);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":183
 *     # 99% of the time, these will be aligned, but in case they aren't just skip
 *     # this loop
 *     if _is_aligned(path1) and _is_aligned(path2):             # <<<<<<<<<<<<<<
 *         cur_int1 = <int*>path1
 *         cur_int2 = <int*>path2
*/
  __pyx_t_3 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__is_aligned(__pyx_v_path1); if (unlikely(__pyx_t_3 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 183, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_3 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__is_aligned(__pyx_v_path2); if (unlikely(__pyx_t_3 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 183, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":184
 *     # this loop
 *     if _is_aligned(path1) and _is_aligned(path2):
 *         cur_int1 = <int*>path1             # <<<<<<<<<<<<<<
 *         cur_int2 = <int*>path2
 *         end_int1 = <int*>(path1 + size1 - (size1 % sizeof(int)))
*/
    __pyx_v_cur_int1 = ((int *)__pyx_v_path1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":185
 *     if _is_aligned(path1) and _is_aligned(path2):
 *         cur_int1 = <int*>path1
 *         cur_int2 = <int*>path2             # <<<<<<<<<<<<<<
 *         end_int1 = <int*>(path1 + size1 - (size1 % sizeof(int)))
 *         end_int2 = <int*>(path2 + size2 - (size2 % sizeof(int)))
*/
    __pyx_v_cur_int2 = ((int *)__pyx_v_path2);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":186
 *         cur_int1 = <int*>path1
 *         cur_int2 = <int*>path2
 *         end_int1 = <int*>(path1 + size1 - (size1 % sizeof(int)))             # <<<<<<<<<<<<<<
 *         end_int2 = <int*>(path2 + size2 - (size2 % sizeof(int)))
 * 
*/
    __pyx_t_4 = (sizeof(int));
    if (unlikely(__pyx_t_4 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(0, 186, __pyx_L1_error)
    }
    __pyx_v_end_int1 = ((int *)((__pyx_v_path1 + __pyx_v_size1) - (__pyx_v_size1 % __pyx_t_4)));

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":187
 *         cur_int2 = <int*>path2
 *         end_int1 = <int*>(path1 + size1 - (size1 % sizeof(int)))
 *         end_int2 = <int*>(path2 + size2 - (size2 % sizeof(int)))             # <<<<<<<<<<<<<<
 * 
 *         while cur_int1 < end_int1 and cur_int2 < end_int2:
*/
    __pyx_t_4 = (sizeof(int));
    if (unlikely(__pyx_t_4 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(0, 187, __pyx_L1_error)
    }
    __pyx_v_end_int2 = ((int *)((__pyx_v_path2 + __pyx_v_size2) - (__pyx_v_size2 % __pyx_t_4)));

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":189
 *         end_int2 = <int*>(path2 + size2 - (size2 % sizeof(int)))
 * 
 *         while cur_int1 < end_int1 and cur_int2 < end_int2:             # <<<<<<<<<<<<<<
 *             if cur_int1[0] != cur_int2[0]:
 *                 break
*/
    while (1) {
      __pyx_t_2 = (__pyx_v_cur_int1 < __pyx_v_end_int1);
      if (__pyx_t_2) {
      } else {
        __pyx_t_1 = __pyx_t_2;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_2 = (__pyx_v_cur_int2 < __pyx_v_end_int2);
      __pyx_t_1 = __pyx_t_2;
      __pyx_L11_bool_binop_done:;
      if (!__pyx_t_1) break;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":190
 * 
 *         while cur_int1 < end_int1 and cur_int2 < end_int2:
 *             if cur_int1[0] != cur_int2[0]:             # <<<<<<<<<<<<<<
 *                 break
 *             cur_int1 = cur_int1 + 1
*/
      __pyx_t_1 = ((__pyx_v_cur_int1[0]) != (__pyx_v_cur_int2[0]));
      if (__pyx_t_1) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":191
 *         while cur_int1 < end_int1 and cur_int2 < end_int2:
 *             if cur_int1[0] != cur_int2[0]:
 *                 break             # <<<<<<<<<<<<<<
 *             cur_int1 = cur_int1 + 1
 *             cur_int2 = cur_int2 + 1
*/
        goto __pyx_L10_break;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":190
 * 
 *         while cur_int1 < end_int1 and cur_int2 < end_int2:
 *             if cur_int1[0] != cur_int2[0]:             # <<<<<<<<<<<<<<
 *                 break
 *             cur_int1 = cur_int1 + 1
*/
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":192
 *             if cur_int1[0] != cur_int2[0]:
 *                 break
 *             cur_int1 = cur_int1 + 1             # <<<<<<<<<<<<<<
 *             cur_int2 = cur_int2 + 1
 * 
*/
      __pyx_v_cur_int1 = (__pyx_v_cur_int1 + 1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":193
 *                 break
 *             cur_int1 = cur_int1 + 1
 *             cur_int2 = cur_int2 + 1             # <<<<<<<<<<<<<<
 * 
 *         cur1 = <unsigned char*>cur_int1
*/
      __pyx_v_cur_int2 = (__pyx_v_cur_int2 + 1);
    }
    __pyx_L10_break:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":195
 *             cur_int2 = cur_int2 + 1
 * 
 *         cur1 = <unsigned char*>cur_int1             # <<<<<<<<<<<<<<
 *         cur2 = <unsigned char*>cur_int2
 *     else:
*/
    __pyx_v_cur1 = ((unsigned char *)__pyx_v_cur_int1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":196
 * 
 *         cur1 = <unsigned char*>cur_int1
 *         cur2 = <unsigned char*>cur_int2             # <<<<<<<<<<<<<<
 *     else:
 *         cur1 = <unsigned char*>path1
*/
    __pyx_v_cur2 = ((unsigned char *)__pyx_v_cur_int2);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":183
 *     # 99% of the time, these will be aligned, but in case they aren't just skip
 *     # this loop
 *     if _is_aligned(path1) and _is_aligned(path2):             # <<<<<<<<<<<<<<
 *         cur_int1 = <int*>path1
 *         cur_int2 = <int*>path2
*/
    goto __pyx_L6;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":198
 *         cur2 = <unsigned char*>cur_int2
 *     else:
 *         cur1 = <unsigned char*>path1             # <<<<<<<<<<<<<<
 *         cur2 = <unsigned char*>path2
 * 
*/
  /*else*/ {
    __pyx_v_cur1 = ((unsigned char *)__pyx_v_path1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":199
 *     else:
 *         cur1 = <unsigned char*>path1
 *         cur2 = <unsigned char*>path2             # <<<<<<<<<<<<<<
 * 
 *     while cur1 < end1 and cur2 < end2:
*/
    __pyx_v_cur2 = ((unsigned char *)__pyx_v_path2);
  }
  __pyx_L6:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":201
 *         cur2 = <unsigned char*>path2
 * 
 *     while cur1 < end1 and cur2 < end2:             # <<<<<<<<<<<<<<
 *         if cur1[0] == cur2[0]:
 *             # This character matches, just go to the next one
*/
  while (1) {
    __pyx_t_2 = (__pyx_v_cur1 < __pyx_v_end1);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L16_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_cur2 < __pyx_v_end2);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L16_bool_binop_done:;
    if (!__pyx_t_1) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":202
 * 
 *     while cur1 < end1 and cur2 < end2:
 *         if cur1[0] == cur2[0]:             # <<<<<<<<<<<<<<
 *             # This character matches, just go to the next one
 *             cur1 = cur1 + 1
*/
    __pyx_t_1 = ((__pyx_v_cur1[0]) == (__pyx_v_cur2[0]));
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":204
 *         if cur1[0] == cur2[0]:
 *             # This character matches, just go to the next one
 *             cur1 = cur1 + 1             # <<<<<<<<<<<<<<
 *             cur2 = cur2 + 1
 *             continue
*/
      __pyx_v_cur1 = (__pyx_v_cur1 + 1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":205
 *             # This character matches, just go to the next one
 *             cur1 = cur1 + 1
 *             cur2 = cur2 + 1             # <<<<<<<<<<<<<<
 *             continue
 *         # The current characters do not match
*/
      __pyx_v_cur2 = (__pyx_v_cur2 + 1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":206
 *             cur1 = cur1 + 1
 *             cur2 = cur2 + 1
 *             continue             # <<<<<<<<<<<<<<
 *         # The current characters do not match
 *         if cur1[0] == b'/':
*/
      goto __pyx_L14_continue;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":202
 * 
 *     while cur1 < end1 and cur2 < end2:
 *         if cur1[0] == cur2[0]:             # <<<<<<<<<<<<<<
 *             # This character matches, just go to the next one
 *             cur1 = cur1 + 1
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":208
 *             continue
 *         # The current characters do not match
 *         if cur1[0] == b'/':             # <<<<<<<<<<<<<<
 *             return -1 # Reached the end of path1 segment first
 *         elif cur2[0] == b'/':
*/
    __pyx_t_1 = ((__pyx_v_cur1[0]) == '/');
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":209
 *         # The current characters do not match
 *         if cur1[0] == b'/':
 *             return -1 # Reached the end of path1 segment first             # <<<<<<<<<<<<<<
 *         elif cur2[0] == b'/':
 *             return 1 # Reached the end of path2 segment first
*/
      __pyx_r = -1;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":208
 *             continue
 *         # The current characters do not match
 *         if cur1[0] == b'/':             # <<<<<<<<<<<<<<
 *             return -1 # Reached the end of path1 segment first
 *         elif cur2[0] == b'/':
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":210
 *         if cur1[0] == b'/':
 *             return -1 # Reached the end of path1 segment first
 *         elif cur2[0] == b'/':             # <<<<<<<<<<<<<<
 *             return 1 # Reached the end of path2 segment first
 *         elif cur1[0] < cur2[0]:
*/
    __pyx_t_1 = ((__pyx_v_cur2[0]) == '/');
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":211
 *             return -1 # Reached the end of path1 segment first
 *         elif cur2[0] == b'/':
 *             return 1 # Reached the end of path2 segment first             # <<<<<<<<<<<<<<
 *         elif cur1[0] < cur2[0]:
 *             return -1
*/
      __pyx_r = 1;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":210
 *         if cur1[0] == b'/':
 *             return -1 # Reached the end of path1 segment first
 *         elif cur2[0] == b'/':             # <<<<<<<<<<<<<<
 *             return 1 # Reached the end of path2 segment first
 *         elif cur1[0] < cur2[0]:
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":212
 *         elif cur2[0] == b'/':
 *             return 1 # Reached the end of path2 segment first
 *         elif cur1[0] < cur2[0]:             # <<<<<<<<<<<<<<
 *             return -1
 *         else:
*/
    __pyx_t_1 = ((__pyx_v_cur1[0]) < (__pyx_v_cur2[0]));
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":213
 *             return 1 # Reached the end of path2 segment first
 *         elif cur1[0] < cur2[0]:
 *             return -1             # <<<<<<<<<<<<<<
 *         else:
 *             return 1
*/
      __pyx_r = -1;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":212
 *         elif cur2[0] == b'/':
 *             return 1 # Reached the end of path2 segment first
 *         elif cur1[0] < cur2[0]:             # <<<<<<<<<<<<<<
 *             return -1
 *         else:
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":215
 *             return -1
 *         else:
 *             return 1             # <<<<<<<<<<<<<<
 * 
 *     # We reached the end of at least one of the strings
*/
    /*else*/ {
      __pyx_r = 1;
      goto __pyx_L0;
    }
    __pyx_L14_continue:;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":218
 * 
 *     # We reached the end of at least one of the strings
 *     if cur1 < end1:             # <<<<<<<<<<<<<<
 *         return 1 # Not at the end of cur1, must be at the end of cur2
 *     if cur2 < end2:
*/
  __pyx_t_1 = (__pyx_v_cur1 < __pyx_v_end1);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":219
 *     # We reached the end of at least one of the strings
 *     if cur1 < end1:
 *         return 1 # Not at the end of cur1, must be at the end of cur2             # <<<<<<<<<<<<<<
 *     if cur2 < end2:
 *         return -1 # At the end of cur1, but not at cur2
*/
    __pyx_r = 1;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":218
 * 
 *     # We reached the end of at least one of the strings
 *     if cur1 < end1:             # <<<<<<<<<<<<<<
 *         return 1 # Not at the end of cur1, must be at the end of cur2
 *     if cur2 < end2:
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":220
 *     if cur1 < end1:
 *         return 1 # Not at the end of cur1, must be at the end of cur2
 *     if cur2 < end2:             # <<<<<<<<<<<<<<
 *         return -1 # At the end of cur1, but not at cur2
 *     # We reached the end of both strings
*/
  __pyx_t_1 = (__pyx_v_cur2 < __pyx_v_end2);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":221
 *         return 1 # Not at the end of cur1, must be at the end of cur2
 *     if cur2 < end2:
 *         return -1 # At the end of cur1, but not at cur2             # <<<<<<<<<<<<<<
 *     # We reached the end of both strings
 *     return 0
*/
    __pyx_r = -1;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":220
 *     if cur1 < end1:
 *         return 1 # Not at the end of cur1, must be at the end of cur2
 *     if cur2 < end2:             # <<<<<<<<<<<<<<
 *         return -1 # At the end of cur1, but not at cur2
 *     # We reached the end of both strings
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":223
 *         return -1 # At the end of cur1, but not at cur2
 *     # We reached the end of both strings
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":162
 * 
 * 
 * cdef int _cmp_by_dirs(char *path1, int size1, char *path2, int size2): # cannot_raise             # <<<<<<<<<<<<<<
 *     cdef unsigned char *cur1
 *     cdef unsigned char *cur2
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._cmp_by_dirs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":226
 * 
 * 
 * def lt_by_dirs(path1, path2):             # <<<<<<<<<<<<<<
 *     """Compare two paths directory by directory.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_3lt_by_dirs(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_2lt_by_dirs, "Compare two paths directory by directory.\n\n    This is equivalent to doing::\n\n       operator.lt(path1.split('/'), path2.split('/'))\n\n    The idea is that you should compare path components separately. This\n    differs from plain ``path1 < path2`` for paths like ``'a-b'`` and ``a/b``.\n    \"a-b\" comes after \"a\" but would come before \"a/b\" lexically.\n\n    :param path1: first path\n    :param path2: second path\n    :return: True if path1 comes first, otherwise False\n    ");
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_3lt_by_dirs = {"lt_by_dirs", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_3lt_by_dirs, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_2lt_by_dirs};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_3lt_by_dirs(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_path1 = 0;
  PyObject *__pyx_v_path2 = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lt_by_dirs (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_path1,&__pyx_mstate_global->__pyx_n_u_path2,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 226, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 226, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 226, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "lt_by_dirs", 0) < (0)) __PYX_ERR(0, 226, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("lt_by_dirs", 1, 2, 2, i); __PYX_ERR(0, 226, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 226, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 226, __pyx_L3_error)
    }
    __pyx_v_path1 = values[0];
    __pyx_v_path2 = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lt_by_dirs", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 226, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.lt_by_dirs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_2lt_by_dirs(__pyx_self, __pyx_v_path1, __pyx_v_path2);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_2lt_by_dirs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path1, PyObject *__pyx_v_path2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6[4];
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lt_by_dirs", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":241
 *     :return: True if path1 comes first, otherwise False
 *     """
 *     if not PyBytes_CheckExact(path1):             # <<<<<<<<<<<<<<
 *         raise TypeError("'path1' must be a bytes string, not %s: %r"
 *                         % (type(path1), path1))
*/
  __pyx_t_1 = (!(PyBytes_CheckExact(__pyx_v_path1) != 0));
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":242
 *     """
 *     if not PyBytes_CheckExact(path1):
 *         raise TypeError("'path1' must be a bytes string, not %s: %r"             # <<<<<<<<<<<<<<
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):
*/
    __pyx_t_3 = NULL;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":243
 *     if not PyBytes_CheckExact(path1):
 *         raise TypeError("'path1' must be a bytes string, not %s: %r"
 *                         % (type(path1), path1))             # <<<<<<<<<<<<<<
 *     if not PyBytes_CheckExact(path2):
 *         raise TypeError("'path2' must be a bytes string, not %s: %r"
*/
    __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(((PyObject *)Py_TYPE(__pyx_v_path1))), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_path1), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6[0] = __pyx_mstate_global->__pyx_kp_u_path1_must_be_a_bytes_string_no;
    __pyx_t_6[1] = __pyx_t_4;
    __pyx_t_6[2] = __pyx_mstate_global->__pyx_kp_u_;
    __pyx_t_6[3] = __pyx_t_5;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":242
 *     """
 *     if not PyBytes_CheckExact(path1):
 *         raise TypeError("'path1' must be a bytes string, not %s: %r"             # <<<<<<<<<<<<<<
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):
*/
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_6, 4, 36 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5));
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 242, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_8 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_7};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 242, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 242, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":241
 *     :return: True if path1 comes first, otherwise False
 *     """
 *     if not PyBytes_CheckExact(path1):             # <<<<<<<<<<<<<<
 *         raise TypeError("'path1' must be a bytes string, not %s: %r"
 *                         % (type(path1), path1))
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":244
 *         raise TypeError("'path1' must be a bytes string, not %s: %r"
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):             # <<<<<<<<<<<<<<
 *         raise TypeError("'path2' must be a bytes string, not %s: %r"
 *                         % (type(path2), path2))
*/
  __pyx_t_1 = (!(PyBytes_CheckExact(__pyx_v_path2) != 0));
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":245
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):
 *         raise TypeError("'path2' must be a bytes string, not %s: %r"             # <<<<<<<<<<<<<<
 *                         % (type(path2), path2))
 *     return -1 == _cmp_by_dirs(PyBytes_AsString(path1),
*/
    __pyx_t_7 = NULL;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":246
 *     if not PyBytes_CheckExact(path2):
 *         raise TypeError("'path2' must be a bytes string, not %s: %r"
 *                         % (type(path2), path2))             # <<<<<<<<<<<<<<
 *     return -1 == _cmp_by_dirs(PyBytes_AsString(path1),
 *                               PyBytes_Size(path1),
*/
    __pyx_t_3 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(((PyObject *)Py_TYPE(__pyx_v_path2))), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_path2), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6[0] = __pyx_mstate_global->__pyx_kp_u_path2_must_be_a_bytes_string_no;
    __pyx_t_6[1] = __pyx_t_3;
    __pyx_t_6[2] = __pyx_mstate_global->__pyx_kp_u_;
    __pyx_t_6[3] = __pyx_t_5;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":245
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):
 *         raise TypeError("'path2' must be a bytes string, not %s: %r"             # <<<<<<<<<<<<<<
 *                         % (type(path2), path2))
 *     return -1 == _cmp_by_dirs(PyBytes_AsString(path1),
*/
    __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_6, 4, 36 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3) + 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5));
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_8 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 245, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 245, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":244
 *         raise TypeError("'path1' must be a bytes string, not %s: %r"
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):             # <<<<<<<<<<<<<<
 *         raise TypeError("'path2' must be a bytes string, not %s: %r"
 *                         % (type(path2), path2))
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":247
 *         raise TypeError("'path2' must be a bytes string, not %s: %r"
 *                         % (type(path2), path2))
 *     return -1 == _cmp_by_dirs(PyBytes_AsString(path1),             # <<<<<<<<<<<<<<
 *                               PyBytes_Size(path1),
 *                               PyBytes_AsString(path2),
*/
  __Pyx_XDECREF(__pyx_r);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":250
 *                               PyBytes_Size(path1),
 *                               PyBytes_AsString(path2),
 *                               PyBytes_Size(path2))             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_9 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_by_dirs(PyBytes_AsString(__pyx_v_path1), PyBytes_Size(__pyx_v_path1), PyBytes_AsString(__pyx_v_path2), PyBytes_Size(__pyx_v_path2)); if (unlikely(__pyx_t_9 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 247, __pyx_L1_error)

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":247
 *         raise TypeError("'path2' must be a bytes string, not %s: %r"
 *                         % (type(path2), path2))
 *     return -1 == _cmp_by_dirs(PyBytes_AsString(path1),             # <<<<<<<<<<<<<<
 *                               PyBytes_Size(path1),
 *                               PyBytes_AsString(path2),
*/
  __pyx_t_2 = __Pyx_PyBool_FromLong((-1L == __pyx_t_9)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":226
 * 
 * 
 * def lt_by_dirs(path1, path2):             # <<<<<<<<<<<<<<
 *     """Compare two paths directory by directory.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.lt_by_dirs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":253
 * 
 * 
 * def _lt_path_by_dirblock(path1, path2):             # <<<<<<<<<<<<<<
 *     """Compare two paths based on what directory they are in.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_5_lt_path_by_dirblock(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_4_lt_path_by_dirblock, "Compare two paths based on what directory they are in.\n\n    This generates a sort order, such that all children of a directory are\n    sorted together, and grandchildren are in the same order as the\n    children appear. But all grandchildren come after all children.\n\n    In other words, all entries in a directory are sorted together, and\n    directories are sorted in cmp_by_dirs order.\n\n    :param path1: first path\n    :param path2: the second path\n    :return: True if path1 comes first, otherwise False.\n    ");
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_5_lt_path_by_dirblock = {"_lt_path_by_dirblock", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_5_lt_path_by_dirblock, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_4_lt_path_by_dirblock};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_5_lt_path_by_dirblock(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_path1 = 0;
  PyObject *__pyx_v_path2 = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_lt_path_by_dirblock (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_path1,&__pyx_mstate_global->__pyx_n_u_path2,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 253, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 253, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 253, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_lt_path_by_dirblock", 0) < (0)) __PYX_ERR(0, 253, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_lt_path_by_dirblock", 1, 2, 2, i); __PYX_ERR(0, 253, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 253, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 253, __pyx_L3_error)
    }
    __pyx_v_path1 = values[0];
    __pyx_v_path2 = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_lt_path_by_dirblock", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 253, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._lt_path_by_dirblock", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_4_lt_path_by_dirblock(__pyx_self, __pyx_v_path1, __pyx_v_path2);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_4_lt_path_by_dirblock(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path1, PyObject *__pyx_v_path2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6[4];
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_lt_path_by_dirblock", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":267
 *     :return: True if path1 comes first, otherwise False.
 *     """
 *     if not PyBytes_CheckExact(path1):             # <<<<<<<<<<<<<<
 *         raise TypeError("'path1' must be a plain string, not %s: %r"
 *                         % (type(path1), path1))
*/
  __pyx_t_1 = (!(PyBytes_CheckExact(__pyx_v_path1) != 0));
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":268
 *     """
 *     if not PyBytes_CheckExact(path1):
 *         raise TypeError("'path1' must be a plain string, not %s: %r"             # <<<<<<<<<<<<<<
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):
*/
    __pyx_t_3 = NULL;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":269
 *     if not PyBytes_CheckExact(path1):
 *         raise TypeError("'path1' must be a plain string, not %s: %r"
 *                         % (type(path1), path1))             # <<<<<<<<<<<<<<
 *     if not PyBytes_CheckExact(path2):
 *         raise TypeError("'path2' must be a plain string, not %s: %r"
*/
    __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(((PyObject *)Py_TYPE(__pyx_v_path1))), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_path1), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6[0] = __pyx_mstate_global->__pyx_kp_u_path1_must_be_a_plain_string_no;
    __pyx_t_6[1] = __pyx_t_4;
    __pyx_t_6[2] = __pyx_mstate_global->__pyx_kp_u_;
    __pyx_t_6[3] = __pyx_t_5;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":268
 *     """
 *     if not PyBytes_CheckExact(path1):
 *         raise TypeError("'path1' must be a plain string, not %s: %r"             # <<<<<<<<<<<<<<
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):
*/
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_6, 4, 36 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5));
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_8 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_7};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 268, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":267
 *     :return: True if path1 comes first, otherwise False.
 *     """
 *     if not PyBytes_CheckExact(path1):             # <<<<<<<<<<<<<<
 *         raise TypeError("'path1' must be a plain string, not %s: %r"
 *                         % (type(path1), path1))
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":270
 *         raise TypeError("'path1' must be a plain string, not %s: %r"
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):             # <<<<<<<<<<<<<<
 *         raise TypeError("'path2' must be a plain string, not %s: %r"
 *                         % (type(path2), path2))
*/
  __pyx_t_1 = (!(PyBytes_CheckExact(__pyx_v_path2) != 0));
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":271
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):
 *         raise TypeError("'path2' must be a plain string, not %s: %r"             # <<<<<<<<<<<<<<
 *                         % (type(path2), path2))
 *     # GZ 2017-06-09: This internal function really only needs lt as well.
*/
    __pyx_t_7 = NULL;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":272
 *     if not PyBytes_CheckExact(path2):
 *         raise TypeError("'path2' must be a plain string, not %s: %r"
 *                         % (type(path2), path2))             # <<<<<<<<<<<<<<
 *     # GZ 2017-06-09: This internal function really only needs lt as well.
 *     return (_cmp_path_by_dirblock_intern(PyBytes_AsString(path1),
*/
    __pyx_t_3 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(((PyObject *)Py_TYPE(__pyx_v_path2))), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_path2), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6[0] = __pyx_mstate_global->__pyx_kp_u_path2_must_be_a_plain_string_no;
    __pyx_t_6[1] = __pyx_t_3;
    __pyx_t_6[2] = __pyx_mstate_global->__pyx_kp_u_;
    __pyx_t_6[3] = __pyx_t_5;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":271
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):
 *         raise TypeError("'path2' must be a plain string, not %s: %r"             # <<<<<<<<<<<<<<
 *                         % (type(path2), path2))
 *     # GZ 2017-06-09: This internal function really only needs lt as well.
*/
    __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_6, 4, 36 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3) + 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5));
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_8 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 271, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 271, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":270
 *         raise TypeError("'path1' must be a plain string, not %s: %r"
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):             # <<<<<<<<<<<<<<
 *         raise TypeError("'path2' must be a plain string, not %s: %r"
 *                         % (type(path2), path2))
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":274
 *                         % (type(path2), path2))
 *     # GZ 2017-06-09: This internal function really only needs lt as well.
 *     return (_cmp_path_by_dirblock_intern(PyBytes_AsString(path1),             # <<<<<<<<<<<<<<
 *                                          PyBytes_Size(path1),
 *                                          PyBytes_AsString(path2),
*/
  __Pyx_XDECREF(__pyx_r);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":277
 *                                          PyBytes_Size(path1),
 *                                          PyBytes_AsString(path2),
 *                                          PyBytes_Size(path2)) < 0)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_9 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_path_by_dirblock_intern(PyBytes_AsString(__pyx_v_path1), PyBytes_Size(__pyx_v_path1), PyBytes_AsString(__pyx_v_path2), PyBytes_Size(__pyx_v_path2)); if (unlikely(__pyx_t_9 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 274, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong((__pyx_t_9 < 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":253
 * 
 * 
 * def _lt_path_by_dirblock(path1, path2):             # <<<<<<<<<<<<<<
 *     """Compare two paths based on what directory they are in.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._lt_path_by_dirblock", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":280
 * 
 * 
 * cdef int _cmp_path_by_dirblock_intern(char *path1, int path1_len,             # <<<<<<<<<<<<<<
 *                                       char *path2, int path2_len): # cannot_raise
 *     """Compare two paths by what directory they are in.
*/

static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_path_by_dirblock_intern(char *__pyx_v_path1, int __pyx_v_path1_len, char *__pyx_v_path2, int __pyx_v_path2_len) {
  char *__pyx_v_dirname1;
  int __pyx_v_dirname1_len;
  char *__pyx_v_dirname2;
  int __pyx_v_dirname2_len;
  char *__pyx_v_basename1;
  int __pyx_v_basename1_len;
  char *__pyx_v_basename2;
  int __pyx_v_basename2_len;
  int __pyx_v_cur_len;
  int __pyx_v_cmp_val;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  void *__pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":297
 *     cdef int cmp_val
 * 
 *     if path1_len == 0 and path2_len == 0:             # <<<<<<<<<<<<<<
 *         return 0
 * 
*/
  __pyx_t_2 = (__pyx_v_path1_len == 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_path2_len == 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":298
 * 
 *     if path1_len == 0 and path2_len == 0:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     if path1 == path2 and path1_len == path2_len:
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":297
 *     cdef int cmp_val
 * 
 *     if path1_len == 0 and path2_len == 0:             # <<<<<<<<<<<<<<
 *         return 0
 * 
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":300
 *         return 0
 * 
 *     if path1 == path2 and path1_len == path2_len:             # <<<<<<<<<<<<<<
 *         return 0
 * 
*/
  __pyx_t_2 = (__pyx_v_path1 == __pyx_v_path2);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_path1_len == __pyx_v_path2_len);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":301
 * 
 *     if path1 == path2 and path1_len == path2_len:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     if path1_len == 0:
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":300
 *         return 0
 * 
 *     if path1 == path2 and path1_len == path2_len:             # <<<<<<<<<<<<<<
 *         return 0
 * 
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":303
 *         return 0
 * 
 *     if path1_len == 0:             # <<<<<<<<<<<<<<
 *         return -1
 * 
*/
  __pyx_t_1 = (__pyx_v_path1_len == 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":304
 * 
 *     if path1_len == 0:
 *         return -1             # <<<<<<<<<<<<<<
 * 
 *     if path2_len == 0:
*/
    __pyx_r = -1;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":303
 *         return 0
 * 
 *     if path1_len == 0:             # <<<<<<<<<<<<<<
 *         return -1
 * 
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":306
 *         return -1
 * 
 *     if path2_len == 0:             # <<<<<<<<<<<<<<
 *         return 1
 * 
*/
  __pyx_t_1 = (__pyx_v_path2_len == 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":307
 * 
 *     if path2_len == 0:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     basename1 = <char*>_my_memrchr(path1, b'/', path1_len)
*/
    __pyx_r = 1;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":306
 *         return -1
 * 
 *     if path2_len == 0:             # <<<<<<<<<<<<<<
 *         return 1
 * 
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":309
 *         return 1
 * 
 *     basename1 = <char*>_my_memrchr(path1, b'/', path1_len)             # <<<<<<<<<<<<<<
 * 
 *     if basename1 == NULL:
*/
  __pyx_t_3 = __pyx_f_6breezy_3bzr_12_str_helpers__my_memrchr(__pyx_v_path1, '/', __pyx_v_path1_len); if (unlikely(__pyx_t_3 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 309, __pyx_L1_error)
  __pyx_v_basename1 = ((char *)__pyx_t_3);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":311
 *     basename1 = <char*>_my_memrchr(path1, b'/', path1_len)
 * 
 *     if basename1 == NULL:             # <<<<<<<<<<<<<<
 *         basename1 = path1
 *         basename1_len = path1_len
*/
  __pyx_t_1 = (__pyx_v_basename1 == NULL);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":312
 * 
 *     if basename1 == NULL:
 *         basename1 = path1             # <<<<<<<<<<<<<<
 *         basename1_len = path1_len
 *         dirname1 = b''
*/
    __pyx_v_basename1 = __pyx_v_path1;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":313
 *     if basename1 == NULL:
 *         basename1 = path1
 *         basename1_len = path1_len             # <<<<<<<<<<<<<<
 *         dirname1 = b''
 *         dirname1_len = 0
*/
    __pyx_v_basename1_len = __pyx_v_path1_len;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":314
 *         basename1 = path1
 *         basename1_len = path1_len
 *         dirname1 = b''             # <<<<<<<<<<<<<<
 *         dirname1_len = 0
 *     else:
*/
    __pyx_v_dirname1 = ((char *)"");

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":315
 *         basename1_len = path1_len
 *         dirname1 = b''
 *         dirname1_len = 0             # <<<<<<<<<<<<<<
 *     else:
 *         dirname1 = path1
*/
    __pyx_v_dirname1_len = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":311
 *     basename1 = <char*>_my_memrchr(path1, b'/', path1_len)
 * 
 *     if basename1 == NULL:             # <<<<<<<<<<<<<<
 *         basename1 = path1
 *         basename1_len = path1_len
*/
    goto __pyx_L11;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":317
 *         dirname1_len = 0
 *     else:
 *         dirname1 = path1             # <<<<<<<<<<<<<<
 *         dirname1_len = basename1 - path1
 *         basename1 = basename1 + 1
*/
  /*else*/ {
    __pyx_v_dirname1 = __pyx_v_path1;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":318
 *     else:
 *         dirname1 = path1
 *         dirname1_len = basename1 - path1             # <<<<<<<<<<<<<<
 *         basename1 = basename1 + 1
 *         basename1_len = path1_len - dirname1_len - 1
*/
    __pyx_v_dirname1_len = (__pyx_v_basename1 - __pyx_v_path1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":319
 *         dirname1 = path1
 *         dirname1_len = basename1 - path1
 *         basename1 = basename1 + 1             # <<<<<<<<<<<<<<
 *         basename1_len = path1_len - dirname1_len - 1
 * 
*/
    __pyx_v_basename1 = (__pyx_v_basename1 + 1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":320
 *         dirname1_len = basename1 - path1
 *         basename1 = basename1 + 1
 *         basename1_len = path1_len - dirname1_len - 1             # <<<<<<<<<<<<<<
 * 
 *     basename2 = <char*>_my_memrchr(path2, b'/', path2_len)
*/
    __pyx_v_basename1_len = ((__pyx_v_path1_len - __pyx_v_dirname1_len) - 1);
  }
  __pyx_L11:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":322
 *         basename1_len = path1_len - dirname1_len - 1
 * 
 *     basename2 = <char*>_my_memrchr(path2, b'/', path2_len)             # <<<<<<<<<<<<<<
 * 
 *     if basename2 == NULL:
*/
  __pyx_t_3 = __pyx_f_6breezy_3bzr_12_str_helpers__my_memrchr(__pyx_v_path2, '/', __pyx_v_path2_len); if (unlikely(__pyx_t_3 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 322, __pyx_L1_error)
  __pyx_v_basename2 = ((char *)__pyx_t_3);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":324
 *     basename2 = <char*>_my_memrchr(path2, b'/', path2_len)
 * 
 *     if basename2 == NULL:             # <<<<<<<<<<<<<<
 *         basename2 = path2
 *         basename2_len = path2_len
*/
  __pyx_t_1 = (__pyx_v_basename2 == NULL);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":325
 * 
 *     if basename2 == NULL:
 *         basename2 = path2             # <<<<<<<<<<<<<<
 *         basename2_len = path2_len
 *         dirname2 = b''
*/
    __pyx_v_basename2 = __pyx_v_path2;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":326
 *     if basename2 == NULL:
 *         basename2 = path2
 *         basename2_len = path2_len             # <<<<<<<<<<<<<<
 *         dirname2 = b''
 *         dirname2_len = 0
*/
    __pyx_v_basename2_len = __pyx_v_path2_len;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":327
 *         basename2 = path2
 *         basename2_len = path2_len
 *         dirname2 = b''             # <<<<<<<<<<<<<<
 *         dirname2_len = 0
 *     else:
*/
    __pyx_v_dirname2 = ((char *)"");

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":328
 *         basename2_len = path2_len
 *         dirname2 = b''
 *         dirname2_len = 0             # <<<<<<<<<<<<<<
 *     else:
 *         dirname2 = path2
*/
    __pyx_v_dirname2_len = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":324
 *     basename2 = <char*>_my_memrchr(path2, b'/', path2_len)
 * 
 *     if basename2 == NULL:             # <<<<<<<<<<<<<<
 *         basename2 = path2
 *         basename2_len = path2_len
*/
    goto __pyx_L12;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":330
 *         dirname2_len = 0
 *     else:
 *         dirname2 = path2             # <<<<<<<<<<<<<<
 *         dirname2_len = basename2 - path2
 *         basename2 = basename2 + 1
*/
  /*else*/ {
    __pyx_v_dirname2 = __pyx_v_path2;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":331
 *     else:
 *         dirname2 = path2
 *         dirname2_len = basename2 - path2             # <<<<<<<<<<<<<<
 *         basename2 = basename2 + 1
 *         basename2_len = path2_len - dirname2_len - 1
*/
    __pyx_v_dirname2_len = (__pyx_v_basename2 - __pyx_v_path2);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":332
 *         dirname2 = path2
 *         dirname2_len = basename2 - path2
 *         basename2 = basename2 + 1             # <<<<<<<<<<<<<<
 *         basename2_len = path2_len - dirname2_len - 1
 * 
*/
    __pyx_v_basename2 = (__pyx_v_basename2 + 1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":333
 *         dirname2_len = basename2 - path2
 *         basename2 = basename2 + 1
 *         basename2_len = path2_len - dirname2_len - 1             # <<<<<<<<<<<<<<
 * 
 *     cmp_val = _cmp_by_dirs(dirname1, dirname1_len,
*/
    __pyx_v_basename2_len = ((__pyx_v_path2_len - __pyx_v_dirname2_len) - 1);
  }
  __pyx_L12:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":335
 *         basename2_len = path2_len - dirname2_len - 1
 * 
 *     cmp_val = _cmp_by_dirs(dirname1, dirname1_len,             # <<<<<<<<<<<<<<
 *                            dirname2, dirname2_len)
 *     if cmp_val != 0:
*/
  __pyx_t_4 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_by_dirs(__pyx_v_dirname1, __pyx_v_dirname1_len, __pyx_v_dirname2, __pyx_v_dirname2_len); if (unlikely(__pyx_t_4 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 335, __pyx_L1_error)
  __pyx_v_cmp_val = __pyx_t_4;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":337
 *     cmp_val = _cmp_by_dirs(dirname1, dirname1_len,
 *                            dirname2, dirname2_len)
 *     if cmp_val != 0:             # <<<<<<<<<<<<<<
 *         return cmp_val
 * 
*/
  __pyx_t_1 = (__pyx_v_cmp_val != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":338
 *                            dirname2, dirname2_len)
 *     if cmp_val != 0:
 *         return cmp_val             # <<<<<<<<<<<<<<
 * 
 *     cur_len = basename1_len
*/
    __pyx_r = __pyx_v_cmp_val;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":337
 *     cmp_val = _cmp_by_dirs(dirname1, dirname1_len,
 *                            dirname2, dirname2_len)
 *     if cmp_val != 0:             # <<<<<<<<<<<<<<
 *         return cmp_val
 * 
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":340
 *         return cmp_val
 * 
 *     cur_len = basename1_len             # <<<<<<<<<<<<<<
 *     if basename2_len < basename1_len:
 *         cur_len = basename2_len
*/
  __pyx_v_cur_len = __pyx_v_basename1_len;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":341
 * 
 *     cur_len = basename1_len
 *     if basename2_len < basename1_len:             # <<<<<<<<<<<<<<
 *         cur_len = basename2_len
 * 
*/
  __pyx_t_1 = (__pyx_v_basename2_len < __pyx_v_basename1_len);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":342
 *     cur_len = basename1_len
 *     if basename2_len < basename1_len:
 *         cur_len = basename2_len             # <<<<<<<<<<<<<<
 * 
 *     cmp_val = memcmp(basename1, basename2, cur_len)
*/
    __pyx_v_cur_len = __pyx_v_basename2_len;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":341
 * 
 *     cur_len = basename1_len
 *     if basename2_len < basename1_len:             # <<<<<<<<<<<<<<
 *         cur_len = basename2_len
 * 
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":344
 *         cur_len = basename2_len
 * 
 *     cmp_val = memcmp(basename1, basename2, cur_len)             # <<<<<<<<<<<<<<
 *     if cmp_val != 0:
 *         return cmp_val
*/
  __pyx_v_cmp_val = memcmp(__pyx_v_basename1, __pyx_v_basename2, __pyx_v_cur_len);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":345
 * 
 *     cmp_val = memcmp(basename1, basename2, cur_len)
 *     if cmp_val != 0:             # <<<<<<<<<<<<<<
 *         return cmp_val
 *     if basename1_len == basename2_len:
*/
  __pyx_t_1 = (__pyx_v_cmp_val != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":346
 *     cmp_val = memcmp(basename1, basename2, cur_len)
 *     if cmp_val != 0:
 *         return cmp_val             # <<<<<<<<<<<<<<
 *     if basename1_len == basename2_len:
 *         return 0
*/
    __pyx_r = __pyx_v_cmp_val;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":345
 * 
 *     cmp_val = memcmp(basename1, basename2, cur_len)
 *     if cmp_val != 0:             # <<<<<<<<<<<<<<
 *         return cmp_val
 *     if basename1_len == basename2_len:
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":347
 *     if cmp_val != 0:
 *         return cmp_val
 *     if basename1_len == basename2_len:             # <<<<<<<<<<<<<<
 *         return 0
 *     if basename1_len < basename2_len:
*/
  __pyx_t_1 = (__pyx_v_basename1_len == __pyx_v_basename2_len);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":348
 *         return cmp_val
 *     if basename1_len == basename2_len:
 *         return 0             # <<<<<<<<<<<<<<
 *     if basename1_len < basename2_len:
 *         return -1
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":347
 *     if cmp_val != 0:
 *         return cmp_val
 *     if basename1_len == basename2_len:             # <<<<<<<<<<<<<<
 *         return 0
 *     if basename1_len < basename2_len:
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":349
 *     if basename1_len == basename2_len:
 *         return 0
 *     if basename1_len < basename2_len:             # <<<<<<<<<<<<<<
 *         return -1
 *     return 1
*/
  __pyx_t_1 = (__pyx_v_basename1_len < __pyx_v_basename2_len);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":350
 *         return 0
 *     if basename1_len < basename2_len:
 *         return -1             # <<<<<<<<<<<<<<
 *     return 1
 * 
*/
    __pyx_r = -1;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":349
 *     if basename1_len == basename2_len:
 *         return 0
 *     if basename1_len < basename2_len:             # <<<<<<<<<<<<<<
 *         return -1
 *     return 1
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":351
 *     if basename1_len < basename2_len:
 *         return -1
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = 1;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":280
 * 
 * 
 * cdef int _cmp_path_by_dirblock_intern(char *path1, int path1_len,             # <<<<<<<<<<<<<<
 *                                       char *path2, int path2_len): # cannot_raise
 *     """Compare two paths by what directory they are in.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._cmp_path_by_dirblock_intern", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":354
 * 
 * 
 * def _bisect_path_left(paths, path):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert path into paths.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_7_bisect_path_left(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_6_bisect_path_left, "Return the index where to insert path into paths.\n\n    This uses a path-wise comparison so we get::\n        a\n        a-b\n        a=b\n        a/b\n    Rather than::\n        a\n        a-b\n        a/b\n        a=b\n    :param paths: A list of paths to search through\n    :param path: A single path to insert\n    :return: An offset where 'path' can be inserted.\n    :seealso: bisect.bisect_left\n    ");
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_7_bisect_path_left = {"_bisect_path_left", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_7_bisect_path_left, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_6_bisect_path_left};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_7_bisect_path_left(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_paths = 0;
  PyObject *__pyx_v_path = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bisect_path_left (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_paths,&__pyx_mstate_global->__pyx_n_u_path,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 354, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 354, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 354, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_bisect_path_left", 0) < (0)) __PYX_ERR(0, 354, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_bisect_path_left", 1, 2, 2, i); __PYX_ERR(0, 354, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 354, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 354, __pyx_L3_error)
    }
    __pyx_v_paths = values[0];
    __pyx_v_path = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_bisect_path_left", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 354, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._bisect_path_left", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6_bisect_path_left(__pyx_self, __pyx_v_paths, __pyx_v_path);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6_bisect_path_left(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_paths, PyObject *__pyx_v_path) {
  int __pyx_v__lo;
  int __pyx_v__hi;
  int __pyx_v__mid;
  char *__pyx_v_path_cstr;
  int __pyx_v_path_size;
  char *__pyx_v_cur_cstr;
  int __pyx_v_cur_size;
  void *__pyx_v_cur;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6[4];
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_bisect_path_left", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":381
 *     cdef void *cur
 * 
 *     if not PyList_CheckExact(paths):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))
*/
  __pyx_t_1 = (!(PyList_CheckExact(__pyx_v_paths) != 0));
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":382
 * 
 *     if not PyList_CheckExact(paths):
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"             # <<<<<<<<<<<<<<
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):
*/
    __pyx_t_3 = NULL;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":383
 *     if not PyList_CheckExact(paths):
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))             # <<<<<<<<<<<<<<
 *     if not PyBytes_CheckExact(path):
 *         raise TypeError("you must pass a string for 'path' not: %s %r"
*/
    __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(((PyObject *)Py_TYPE(__pyx_v_paths))), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_paths), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6[0] = __pyx_mstate_global->__pyx_kp_u_you_must_pass_a_python_list_for;
    __pyx_t_6[1] = __pyx_t_4;
    __pyx_t_6[2] = __pyx_mstate_global->__pyx_kp_u__2;
    __pyx_t_6[3] = __pyx_t_5;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":382
 * 
 *     if not PyList_CheckExact(paths):
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"             # <<<<<<<<<<<<<<
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):
*/
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_6, 4, 45 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + 1 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5));
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_8 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_7};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 382, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":381
 *     cdef void *cur
 * 
 *     if not PyList_CheckExact(paths):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":384
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a string for 'path' not: %s %r"
 *                         % (type(path), path))
*/
  __pyx_t_1 = (!(PyBytes_CheckExact(__pyx_v_path) != 0));
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":385
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):
 *         raise TypeError("you must pass a string for 'path' not: %s %r"             # <<<<<<<<<<<<<<
 *                         % (type(path), path))
 * 
*/
    __pyx_t_7 = NULL;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":386
 *     if not PyBytes_CheckExact(path):
 *         raise TypeError("you must pass a string for 'path' not: %s %r"
 *                         % (type(path), path))             # <<<<<<<<<<<<<<
 * 
 *     _hi = len(paths)
*/
    __pyx_t_3 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(((PyObject *)Py_TYPE(__pyx_v_path))), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_path), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6[0] = __pyx_mstate_global->__pyx_kp_u_you_must_pass_a_string_for_path;
    __pyx_t_6[1] = __pyx_t_3;
    __pyx_t_6[2] = __pyx_mstate_global->__pyx_kp_u__2;
    __pyx_t_6[3] = __pyx_t_5;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":385
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):
 *         raise TypeError("you must pass a string for 'path' not: %s %r"             # <<<<<<<<<<<<<<
 *                         % (type(path), path))
 * 
*/
    __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_6, 4, 39 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3) + 1 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5));
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_8 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 385, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 385, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":384
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a string for 'path' not: %s %r"
 *                         % (type(path), path))
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":388
 *                         % (type(path), path))
 * 
 *     _hi = len(paths)             # <<<<<<<<<<<<<<
 *     _lo = 0
 * 
*/
  __pyx_t_9 = PyObject_Length(__pyx_v_paths); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 388, __pyx_L1_error)
  __pyx_v__hi = __pyx_t_9;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":389
 * 
 *     _hi = len(paths)
 *     _lo = 0             # <<<<<<<<<<<<<<
 * 
 *     path_cstr = PyBytes_AsString(path)
*/
  __pyx_v__lo = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":391
 *     _lo = 0
 * 
 *     path_cstr = PyBytes_AsString(path)             # <<<<<<<<<<<<<<
 *     path_size = PyBytes_Size(path)
 * 
*/
  __pyx_v_path_cstr = PyBytes_AsString(__pyx_v_path);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":392
 * 
 *     path_cstr = PyBytes_AsString(path)
 *     path_size = PyBytes_Size(path)             # <<<<<<<<<<<<<<
 * 
 *     while _lo < _hi:
*/
  __pyx_v_path_size = PyBytes_Size(__pyx_v_path);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":394
 *     path_size = PyBytes_Size(path)
 * 
 *     while _lo < _hi:             # <<<<<<<<<<<<<<
 *         _mid = (_lo + _hi) // 2
 *         cur = PyList_GetItem_object_void(paths, _mid)
*/
  while (1) {
    __pyx_t_1 = (__pyx_v__lo < __pyx_v__hi);
    if (!__pyx_t_1) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":395
 * 
 *     while _lo < _hi:
 *         _mid = (_lo + _hi) // 2             # <<<<<<<<<<<<<<
 *         cur = PyList_GetItem_object_void(paths, _mid)
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
*/
    __pyx_v__mid = __Pyx_div_long((__pyx_v__lo + __pyx_v__hi), 2, 1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":396
 *     while _lo < _hi:
 *         _mid = (_lo + _hi) // 2
 *         cur = PyList_GetItem_object_void(paths, _mid)             # <<<<<<<<<<<<<<
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)
*/
    __pyx_v_cur = PyList_GET_ITEM(__pyx_v_paths, __pyx_v__mid);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":397
 *         _mid = (_lo + _hi) // 2
 *         cur = PyList_GetItem_object_void(paths, _mid)
 *         cur_cstr = PyBytes_AS_STRING_void(cur)             # <<<<<<<<<<<<<<
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_path_by_dirblock_intern(cur_cstr, cur_size,
*/
    __pyx_v_cur_cstr = PyBytes_AS_STRING(__pyx_v_cur);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":398
 *         cur = PyList_GetItem_object_void(paths, _mid)
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)             # <<<<<<<<<<<<<<
 *         if _cmp_path_by_dirblock_intern(cur_cstr, cur_size,
 *                                         path_cstr, path_size) < 0:
*/
    __pyx_v_cur_size = PyBytes_GET_SIZE(__pyx_v_cur);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":399
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_path_by_dirblock_intern(cur_cstr, cur_size,             # <<<<<<<<<<<<<<
 *                                         path_cstr, path_size) < 0:
 *             _lo = _mid + 1
*/
    __pyx_t_10 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_path_by_dirblock_intern(__pyx_v_cur_cstr, __pyx_v_cur_size, __pyx_v_path_cstr, __pyx_v_path_size); if (unlikely(__pyx_t_10 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 399, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":400
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_path_by_dirblock_intern(cur_cstr, cur_size,
 *                                         path_cstr, path_size) < 0:             # <<<<<<<<<<<<<<
 *             _lo = _mid + 1
 *         else:
*/
    __pyx_t_1 = (__pyx_t_10 < 0);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":399
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_path_by_dirblock_intern(cur_cstr, cur_size,             # <<<<<<<<<<<<<<
 *                                         path_cstr, path_size) < 0:
 *             _lo = _mid + 1
*/
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":401
 *         if _cmp_path_by_dirblock_intern(cur_cstr, cur_size,
 *                                         path_cstr, path_size) < 0:
 *             _lo = _mid + 1             # <<<<<<<<<<<<<<
 *         else:
 *             _hi = _mid
*/
      __pyx_v__lo = (__pyx_v__mid + 1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":399
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_path_by_dirblock_intern(cur_cstr, cur_size,             # <<<<<<<<<<<<<<
 *                                         path_cstr, path_size) < 0:
 *             _lo = _mid + 1
*/
      goto __pyx_L7;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":403
 *             _lo = _mid + 1
 *         else:
 *             _hi = _mid             # <<<<<<<<<<<<<<
 *     return _lo
 * 
*/
    /*else*/ {
      __pyx_v__hi = __pyx_v__mid;
    }
    __pyx_L7:;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":404
 *         else:
 *             _hi = _mid
 *     return _lo             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_v__lo); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":354
 * 
 * 
 * def _bisect_path_left(paths, path):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert path into paths.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._bisect_path_left", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":407
 * 
 * 
 * def _bisect_path_right(paths, path):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert path into paths.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_9_bisect_path_right(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_8_bisect_path_right, "Return the index where to insert path into paths.\n\n    This uses a path-wise comparison so we get::\n        a\n        a-b\n        a=b\n        a/b\n    Rather than::\n        a\n        a-b\n        a/b\n        a=b\n    :param paths: A list of paths to search through\n    :param path: A single path to insert\n    :return: An offset where 'path' can be inserted.\n    :seealso: bisect.bisect_right\n    ");
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_9_bisect_path_right = {"_bisect_path_right", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_9_bisect_path_right, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_8_bisect_path_right};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_9_bisect_path_right(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_paths = 0;
  PyObject *__pyx_v_path = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bisect_path_right (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_paths,&__pyx_mstate_global->__pyx_n_u_path,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 407, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 407, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 407, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_bisect_path_right", 0) < (0)) __PYX_ERR(0, 407, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_bisect_path_right", 1, 2, 2, i); __PYX_ERR(0, 407, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 407, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 407, __pyx_L3_error)
    }
    __pyx_v_paths = values[0];
    __pyx_v_path = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_bisect_path_right", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 407, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._bisect_path_right", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_8_bisect_path_right(__pyx_self, __pyx_v_paths, __pyx_v_path);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_8_bisect_path_right(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_paths, PyObject *__pyx_v_path) {
  int __pyx_v__lo;
  int __pyx_v__hi;
  int __pyx_v__mid;
  char *__pyx_v_path_cstr;
  int __pyx_v_path_size;
  char *__pyx_v_cur_cstr;
  int __pyx_v_cur_size;
  void *__pyx_v_cur;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6[4];
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_bisect_path_right", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":434
 *     cdef void *cur
 * 
 *     if not PyList_CheckExact(paths):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))
*/
  __pyx_t_1 = (!(PyList_CheckExact(__pyx_v_paths) != 0));
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":435
 * 
 *     if not PyList_CheckExact(paths):
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"             # <<<<<<<<<<<<<<
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):
*/
    __pyx_t_3 = NULL;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":436
 *     if not PyList_CheckExact(paths):
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))             # <<<<<<<<<<<<<<
 *     if not PyBytes_CheckExact(path):
 *         raise TypeError("you must pass a string for 'path' not: %s %r"
*/
    __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(((PyObject *)Py_TYPE(__pyx_v_paths))), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_paths), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6[0] = __pyx_mstate_global->__pyx_kp_u_you_must_pass_a_python_list_for;
    __pyx_t_6[1] = __pyx_t_4;
    __pyx_t_6[2] = __pyx_mstate_global->__pyx_kp_u__2;
    __pyx_t_6[3] = __pyx_t_5;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":435
 * 
 *     if not PyList_CheckExact(paths):
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"             # <<<<<<<<<<<<<<
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):
*/
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_6, 4, 45 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + 1 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5));
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 435, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_8 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_7};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 435, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 435, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":434
 *     cdef void *cur
 * 
 *     if not PyList_CheckExact(paths):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":437
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a string for 'path' not: %s %r"
 *                         % (type(path), path))
*/
  __pyx_t_1 = (!(PyBytes_CheckExact(__pyx_v_path) != 0));
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":438
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):
 *         raise TypeError("you must pass a string for 'path' not: %s %r"             # <<<<<<<<<<<<<<
 *                         % (type(path), path))
 * 
*/
    __pyx_t_7 = NULL;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":439
 *     if not PyBytes_CheckExact(path):
 *         raise TypeError("you must pass a string for 'path' not: %s %r"
 *                         % (type(path), path))             # <<<<<<<<<<<<<<
 * 
 *     _hi = len(paths)
*/
    __pyx_t_3 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(((PyObject *)Py_TYPE(__pyx_v_path))), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 439, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_path), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 439, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6[0] = __pyx_mstate_global->__pyx_kp_u_you_must_pass_a_string_for_path;
    __pyx_t_6[1] = __pyx_t_3;
    __pyx_t_6[2] = __pyx_mstate_global->__pyx_kp_u__2;
    __pyx_t_6[3] = __pyx_t_5;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":438
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):
 *         raise TypeError("you must pass a string for 'path' not: %s %r"             # <<<<<<<<<<<<<<
 *                         % (type(path), path))
 * 
*/
    __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_6, 4, 39 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3) + 1 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5));
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_8 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 438, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":437
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a string for 'path' not: %s %r"
 *                         % (type(path), path))
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":441
 *                         % (type(path), path))
 * 
 *     _hi = len(paths)             # <<<<<<<<<<<<<<
 *     _lo = 0
 * 
*/
  __pyx_t_9 = PyObject_Length(__pyx_v_paths); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 441, __pyx_L1_error)
  __pyx_v__hi = __pyx_t_9;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":442
 * 
 *     _hi = len(paths)
 *     _lo = 0             # <<<<<<<<<<<<<<
 * 
 *     path_cstr = PyBytes_AsString(path)
*/
  __pyx_v__lo = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":444
 *     _lo = 0
 * 
 *     path_cstr = PyBytes_AsString(path)             # <<<<<<<<<<<<<<
 *     path_size = PyBytes_Size(path)
 * 
*/
  __pyx_v_path_cstr = PyBytes_AsString(__pyx_v_path);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":445
 * 
 *     path_cstr = PyBytes_AsString(path)
 *     path_size = PyBytes_Size(path)             # <<<<<<<<<<<<<<
 * 
 *     while _lo < _hi:
*/
  __pyx_v_path_size = PyBytes_Size(__pyx_v_path);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":447
 *     path_size = PyBytes_Size(path)
 * 
 *     while _lo < _hi:             # <<<<<<<<<<<<<<
 *         _mid = (_lo + _hi) // 2
 *         cur = PyList_GetItem_object_void(paths, _mid)
*/
  while (1) {
    __pyx_t_1 = (__pyx_v__lo < __pyx_v__hi);
    if (!__pyx_t_1) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":448
 * 
 *     while _lo < _hi:
 *         _mid = (_lo + _hi) // 2             # <<<<<<<<<<<<<<
 *         cur = PyList_GetItem_object_void(paths, _mid)
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
*/
    __pyx_v__mid = __Pyx_div_long((__pyx_v__lo + __pyx_v__hi), 2, 1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":449
 *     while _lo < _hi:
 *         _mid = (_lo + _hi) // 2
 *         cur = PyList_GetItem_object_void(paths, _mid)             # <<<<<<<<<<<<<<
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)
*/
    __pyx_v_cur = PyList_GET_ITEM(__pyx_v_paths, __pyx_v__mid);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":450
 *         _mid = (_lo + _hi) // 2
 *         cur = PyList_GetItem_object_void(paths, _mid)
 *         cur_cstr = PyBytes_AS_STRING_void(cur)             # <<<<<<<<<<<<<<
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_path_by_dirblock_intern(path_cstr, path_size,
*/
    __pyx_v_cur_cstr = PyBytes_AS_STRING(__pyx_v_cur);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":451
 *         cur = PyList_GetItem_object_void(paths, _mid)
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)             # <<<<<<<<<<<<<<
 *         if _cmp_path_by_dirblock_intern(path_cstr, path_size,
 *                                         cur_cstr, cur_size) < 0:
*/
    __pyx_v_cur_size = PyBytes_GET_SIZE(__pyx_v_cur);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":452
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_path_by_dirblock_intern(path_cstr, path_size,             # <<<<<<<<<<<<<<
 *                                         cur_cstr, cur_size) < 0:
 *             _hi = _mid
*/
    __pyx_t_10 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_path_by_dirblock_intern(__pyx_v_path_cstr, __pyx_v_path_size, __pyx_v_cur_cstr, __pyx_v_cur_size); if (unlikely(__pyx_t_10 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 452, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":453
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_path_by_dirblock_intern(path_cstr, path_size,
 *                                         cur_cstr, cur_size) < 0:             # <<<<<<<<<<<<<<
 *             _hi = _mid
 *         else:
*/
    __pyx_t_1 = (__pyx_t_10 < 0);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":452
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_path_by_dirblock_intern(path_cstr, path_size,             # <<<<<<<<<<<<<<
 *                                         cur_cstr, cur_size) < 0:
 *             _hi = _mid
*/
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":454
 *         if _cmp_path_by_dirblock_intern(path_cstr, path_size,
 *                                         cur_cstr, cur_size) < 0:
 *             _hi = _mid             # <<<<<<<<<<<<<<
 *         else:
 *             _lo = _mid + 1
*/
      __pyx_v__hi = __pyx_v__mid;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":452
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_path_by_dirblock_intern(path_cstr, path_size,             # <<<<<<<<<<<<<<
 *                                         cur_cstr, cur_size) < 0:
 *             _hi = _mid
*/
      goto __pyx_L7;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":456
 *             _hi = _mid
 *         else:
 *             _lo = _mid + 1             # <<<<<<<<<<<<<<
 *     return _lo
 * 
*/
    /*else*/ {
      __pyx_v__lo = (__pyx_v__mid + 1);
    }
    __pyx_L7:;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":457
 *         else:
 *             _lo = _mid + 1
 *     return _lo             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_v__lo); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":407
 * 
 * 
 * def _bisect_path_right(paths, path):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert path into paths.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._bisect_path_right", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":460
 * 
 * 
 * def bisect_dirblock(dirblocks, dirname, lo=0, hi=None, cache=None):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert dirname into the dirblocks.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_11bisect_dirblock(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_10bisect_dirblock, "Return the index where to insert dirname into the dirblocks.\n\n    The return value idx is such that all directories blocks in dirblock[:idx]\n    have names < dirname, and all blocks in dirblock[idx:] have names >=\n    dirname.\n\n    Optional args lo (default 0) and hi (default len(dirblocks)) bound the\n    slice of a to be searched.\n    ");
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_11bisect_dirblock = {"bisect_dirblock", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_11bisect_dirblock, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_10bisect_dirblock};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_11bisect_dirblock(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_dirblocks = 0;
  PyObject *__pyx_v_dirname = 0;
  PyObject *__pyx_v_lo = 0;
  PyObject *__pyx_v_hi = 0;
  CYTHON_UNUSED PyObject *__pyx_v_cache = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bisect_dirblock (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_dirblocks,&__pyx_mstate_global->__pyx_n_u_dirname,&__pyx_mstate_global->__pyx_n_u_lo,&__pyx_mstate_global->__pyx_n_u_hi,&__pyx_mstate_global->__pyx_n_u_cache,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 460, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 460, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 460, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 460, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 460, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 460, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "bisect_dirblock", 0) < (0)) __PYX_ERR(0, 460, __pyx_L3_error)
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)((PyObject*)__pyx_mstate_global->__pyx_int_0)));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("bisect_dirblock", 0, 2, 5, i); __PYX_ERR(0, 460, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 460, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 460, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 460, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 460, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 460, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)((PyObject*)__pyx_mstate_global->__pyx_int_0)));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_dirblocks = values[0];
    __pyx_v_dirname = values[1];
    __pyx_v_lo = values[2];
    __pyx_v_hi = values[3];
    __pyx_v_cache = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bisect_dirblock", 0, 2, 5, __pyx_nargs); __PYX_ERR(0, 460, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.bisect_dirblock", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_10bisect_dirblock(__pyx_self, __pyx_v_dirblocks, __pyx_v_dirname, __pyx_v_lo, __pyx_v_hi, __pyx_v_cache);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_10bisect_dirblock(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_dirblocks, PyObject *__pyx_v_dirname, PyObject *__pyx_v_lo, PyObject *__pyx_v_hi, CYTHON_UNUSED PyObject *__pyx_v_cache) {
  int __pyx_v__lo;
  int __pyx_v__hi;
  int __pyx_v__mid;
  char *__pyx_v_dirname_cstr;
  int __pyx_v_dirname_size;
  char *__pyx_v_cur_cstr;
  int __pyx_v_cur_size;
  void *__pyx_v_cur;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6[4];
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bisect_dirblock", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":479
 *     cdef void *cur
 * 
 *     if not PyList_CheckExact(dirblocks):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a python list for 'dirblocks' not: %s %r"
 *                         % (type(dirblocks), dirblocks))
*/
  __pyx_t_1 = (!(PyList_CheckExact(__pyx_v_dirblocks) != 0));
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":480
 * 
 *     if not PyList_CheckExact(dirblocks):
 *         raise TypeError("you must pass a python list for 'dirblocks' not: %s %r"             # <<<<<<<<<<<<<<
 *                         % (type(dirblocks), dirblocks))
 *     if not PyBytes_CheckExact(dirname):
*/
    __pyx_t_3 = NULL;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":481
 *     if not PyList_CheckExact(dirblocks):
 *         raise TypeError("you must pass a python list for 'dirblocks' not: %s %r"
 *                         % (type(dirblocks), dirblocks))             # <<<<<<<<<<<<<<
 *     if not PyBytes_CheckExact(dirname):
 *         raise TypeError("you must pass a string for dirname not: %s %r"
*/
    __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(((PyObject *)Py_TYPE(__pyx_v_dirblocks))), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_dirblocks), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6[0] = __pyx_mstate_global->__pyx_kp_u_you_must_pass_a_python_list_for_2;
    __pyx_t_6[1] = __pyx_t_4;
    __pyx_t_6[2] = __pyx_mstate_global->__pyx_kp_u__2;
    __pyx_t_6[3] = __pyx_t_5;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":480
 * 
 *     if not PyList_CheckExact(dirblocks):
 *         raise TypeError("you must pass a python list for 'dirblocks' not: %s %r"             # <<<<<<<<<<<<<<
 *                         % (type(dirblocks), dirblocks))
 *     if not PyBytes_CheckExact(dirname):
*/
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_6, 4, 49 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + 1 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5));
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_8 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_7};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 480, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 480, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":479
 *     cdef void *cur
 * 
 *     if not PyList_CheckExact(dirblocks):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a python list for 'dirblocks' not: %s %r"
 *                         % (type(dirblocks), dirblocks))
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":482
 *         raise TypeError("you must pass a python list for 'dirblocks' not: %s %r"
 *                         % (type(dirblocks), dirblocks))
 *     if not PyBytes_CheckExact(dirname):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a string for dirname not: %s %r"
 *                         % (type(dirname), dirname))
*/
  __pyx_t_1 = (!(PyBytes_CheckExact(__pyx_v_dirname) != 0));
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":483
 *                         % (type(dirblocks), dirblocks))
 *     if not PyBytes_CheckExact(dirname):
 *         raise TypeError("you must pass a string for dirname not: %s %r"             # <<<<<<<<<<<<<<
 *                         % (type(dirname), dirname))
 *     if hi is None:
*/
    __pyx_t_7 = NULL;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":484
 *     if not PyBytes_CheckExact(dirname):
 *         raise TypeError("you must pass a string for dirname not: %s %r"
 *                         % (type(dirname), dirname))             # <<<<<<<<<<<<<<
 *     if hi is None:
 *         _hi = len(dirblocks)
*/
    __pyx_t_3 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(((PyObject *)Py_TYPE(__pyx_v_dirname))), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_dirname), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6[0] = __pyx_mstate_global->__pyx_kp_u_you_must_pass_a_string_for_dirna;
    __pyx_t_6[1] = __pyx_t_3;
    __pyx_t_6[2] = __pyx_mstate_global->__pyx_kp_u__2;
    __pyx_t_6[3] = __pyx_t_5;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":483
 *                         % (type(dirblocks), dirblocks))
 *     if not PyBytes_CheckExact(dirname):
 *         raise TypeError("you must pass a string for dirname not: %s %r"             # <<<<<<<<<<<<<<
 *                         % (type(dirname), dirname))
 *     if hi is None:
*/
    __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_6, 4, 40 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3) + 1 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5));
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_8 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 483, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":482
 *         raise TypeError("you must pass a python list for 'dirblocks' not: %s %r"
 *                         % (type(dirblocks), dirblocks))
 *     if not PyBytes_CheckExact(dirname):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a string for dirname not: %s %r"
 *                         % (type(dirname), dirname))
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":485
 *         raise TypeError("you must pass a string for dirname not: %s %r"
 *                         % (type(dirname), dirname))
 *     if hi is None:             # <<<<<<<<<<<<<<
 *         _hi = len(dirblocks)
 *     else:
*/
  __pyx_t_1 = (__pyx_v_hi == Py_None);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":486
 *                         % (type(dirname), dirname))
 *     if hi is None:
 *         _hi = len(dirblocks)             # <<<<<<<<<<<<<<
 *     else:
 *         _hi = hi
*/
    __pyx_t_9 = PyObject_Length(__pyx_v_dirblocks); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 486, __pyx_L1_error)
    __pyx_v__hi = __pyx_t_9;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":485
 *         raise TypeError("you must pass a string for dirname not: %s %r"
 *                         % (type(dirname), dirname))
 *     if hi is None:             # <<<<<<<<<<<<<<
 *         _hi = len(dirblocks)
 *     else:
*/
    goto __pyx_L5;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":488
 *         _hi = len(dirblocks)
 *     else:
 *         _hi = hi             # <<<<<<<<<<<<<<
 * 
 *     _lo = lo
*/
  /*else*/ {
    __pyx_t_10 = __Pyx_PyLong_As_int(__pyx_v_hi); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 488, __pyx_L1_error)
    __pyx_v__hi = __pyx_t_10;
  }
  __pyx_L5:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":490
 *         _hi = hi
 * 
 *     _lo = lo             # <<<<<<<<<<<<<<
 *     dirname_cstr = PyBytes_AsString(dirname)
 *     dirname_size = PyBytes_Size(dirname)
*/
  __pyx_t_10 = __Pyx_PyLong_As_int(__pyx_v_lo); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 490, __pyx_L1_error)
  __pyx_v__lo = __pyx_t_10;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":491
 * 
 *     _lo = lo
 *     dirname_cstr = PyBytes_AsString(dirname)             # <<<<<<<<<<<<<<
 *     dirname_size = PyBytes_Size(dirname)
 * 
*/
  __pyx_v_dirname_cstr = PyBytes_AsString(__pyx_v_dirname);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":492
 *     _lo = lo
 *     dirname_cstr = PyBytes_AsString(dirname)
 *     dirname_size = PyBytes_Size(dirname)             # <<<<<<<<<<<<<<
 * 
 *     while _lo < _hi:
*/
  __pyx_v_dirname_size = PyBytes_Size(__pyx_v_dirname);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":494
 *     dirname_size = PyBytes_Size(dirname)
 * 
 *     while _lo < _hi:             # <<<<<<<<<<<<<<
 *         _mid = (_lo + _hi) // 2
 *         # Grab the dirname for the current dirblock
*/
  while (1) {
    __pyx_t_1 = (__pyx_v__lo < __pyx_v__hi);
    if (!__pyx_t_1) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":495
 * 
 *     while _lo < _hi:
 *         _mid = (_lo + _hi) // 2             # <<<<<<<<<<<<<<
 *         # Grab the dirname for the current dirblock
 *         # cur = dirblocks[_mid][0]
*/
    __pyx_v__mid = __Pyx_div_long((__pyx_v__lo + __pyx_v__hi), 2, 1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":498
 *         # Grab the dirname for the current dirblock
 *         # cur = dirblocks[_mid][0]
 *         cur = PyTuple_GetItem_void_void(             # <<<<<<<<<<<<<<
 *                 PyList_GetItem_object_void(dirblocks, _mid), 0)
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
*/
    __pyx_v_cur = PyTuple_GET_ITEM(PyList_GET_ITEM(__pyx_v_dirblocks, __pyx_v__mid), 0);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":500
 *         cur = PyTuple_GetItem_void_void(
 *                 PyList_GetItem_object_void(dirblocks, _mid), 0)
 *         cur_cstr = PyBytes_AS_STRING_void(cur)             # <<<<<<<<<<<<<<
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_by_dirs(cur_cstr, cur_size, dirname_cstr, dirname_size) < 0:
*/
    __pyx_v_cur_cstr = PyBytes_AS_STRING(__pyx_v_cur);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":501
 *                 PyList_GetItem_object_void(dirblocks, _mid), 0)
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)             # <<<<<<<<<<<<<<
 *         if _cmp_by_dirs(cur_cstr, cur_size, dirname_cstr, dirname_size) < 0:
 *             _lo = _mid + 1
*/
    __pyx_v_cur_size = PyBytes_GET_SIZE(__pyx_v_cur);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":502
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_by_dirs(cur_cstr, cur_size, dirname_cstr, dirname_size) < 0:             # <<<<<<<<<<<<<<
 *             _lo = _mid + 1
 *         else:
*/
    __pyx_t_10 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_by_dirs(__pyx_v_cur_cstr, __pyx_v_cur_size, __pyx_v_dirname_cstr, __pyx_v_dirname_size); if (unlikely(__pyx_t_10 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 502, __pyx_L1_error)
    __pyx_t_1 = (__pyx_t_10 < 0);
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":503
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_by_dirs(cur_cstr, cur_size, dirname_cstr, dirname_size) < 0:
 *             _lo = _mid + 1             # <<<<<<<<<<<<<<
 *         else:
 *             _hi = _mid
*/
      __pyx_v__lo = (__pyx_v__mid + 1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":502
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_by_dirs(cur_cstr, cur_size, dirname_cstr, dirname_size) < 0:             # <<<<<<<<<<<<<<
 *             _lo = _mid + 1
 *         else:
*/
      goto __pyx_L8;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":505
 *             _lo = _mid + 1
 *         else:
 *             _hi = _mid             # <<<<<<<<<<<<<<
 *     return _lo
 * 
*/
    /*else*/ {
      __pyx_v__hi = __pyx_v__mid;
    }
    __pyx_L8:;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":506
 *         else:
 *             _hi = _mid
 *     return _lo             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_v__lo); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":460
 * 
 * 
 * def bisect_dirblock(dirblocks, dirname, lo=0, hi=None, cache=None):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert dirname into the dirblocks.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.bisect_dirblock", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":521
 *     cdef char *next # Pointer to the end of this record
 * 
 *     def __init__(self, text, state):             # <<<<<<<<<<<<<<
 *         self.state = state
 *         self.text = text
*/

/* Python wrapper */
static int __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_text = 0;
  PyObject *__pyx_v_state = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_text,&__pyx_mstate_global->__pyx_n_u_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 521, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 521, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 521, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 521, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, i); __PYX_ERR(0, 521, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 521, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 521, __pyx_L3_error)
    }
    __pyx_v_text = values[0];
    __pyx_v_state = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 521, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader___init__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self), __pyx_v_text, __pyx_v_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader___init__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self, PyObject *__pyx_v_text, PyObject *__pyx_v_state) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":522
 * 
 *     def __init__(self, text, state):
 *         self.state = state             # <<<<<<<<<<<<<<
 *         self.text = text
 *         self.text_cstr = PyBytes_AsString(text)
*/
  __Pyx_INCREF(__pyx_v_state);
  __Pyx_GIVEREF(__pyx_v_state);
  __Pyx_GOTREF(__pyx_v_self->state);
  __Pyx_DECREF(__pyx_v_self->state);
  __pyx_v_self->state = __pyx_v_state;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":523
 *     def __init__(self, text, state):
 *         self.state = state
 *         self.text = text             # <<<<<<<<<<<<<<
 *         self.text_cstr = PyBytes_AsString(text)
 *         self.text_size = PyBytes_Size(text)
*/
  __Pyx_INCREF(__pyx_v_text);
  __Pyx_GIVEREF(__pyx_v_text);
  __Pyx_GOTREF(__pyx_v_self->text);
  __Pyx_DECREF(__pyx_v_self->text);
  __pyx_v_self->text = __pyx_v_text;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":524
 *         self.state = state
 *         self.text = text
 *         self.text_cstr = PyBytes_AsString(text)             # <<<<<<<<<<<<<<
 *         self.text_size = PyBytes_Size(text)
 *         self.end_cstr = self.text_cstr + self.text_size
*/
  __pyx_v_self->text_cstr = PyBytes_AsString(__pyx_v_text);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":525
 *         self.text = text
 *         self.text_cstr = PyBytes_AsString(text)
 *         self.text_size = PyBytes_Size(text)             # <<<<<<<<<<<<<<
 *         self.end_cstr = self.text_cstr + self.text_size
 *         self.cur_cstr = self.text_cstr
*/
  __pyx_v_self->text_size = PyBytes_Size(__pyx_v_text);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":526
 *         self.text_cstr = PyBytes_AsString(text)
 *         self.text_size = PyBytes_Size(text)
 *         self.end_cstr = self.text_cstr + self.text_size             # <<<<<<<<<<<<<<
 *         self.cur_cstr = self.text_cstr
 * 
*/
  __pyx_v_self->end_cstr = (__pyx_v_self->text_cstr + __pyx_v_self->text_size);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":527
 *         self.text_size = PyBytes_Size(text)
 *         self.end_cstr = self.text_cstr + self.text_size
 *         self.cur_cstr = self.text_cstr             # <<<<<<<<<<<<<<
 * 
 *     cdef char *get_next(self, int *size) except NULL:
*/
  __pyx_t_1 = __pyx_v_self->text_cstr;
  __pyx_v_self->cur_cstr = __pyx_t_1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":521
 *     cdef char *next # Pointer to the end of this record
 * 
 *     def __init__(self, text, state):             # <<<<<<<<<<<<<<
 *         self.state = state
 *         self.text = text
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":529
 *         self.cur_cstr = self.text_cstr
 * 
 *     cdef char *get_next(self, int *size) except NULL:             # <<<<<<<<<<<<<<
 *         """Return a pointer to the start of the next field."""
 *         cdef char *next
*/

static char *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_get_next(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self, int *__pyx_v_size) {
  char *__pyx_v_next;
  Py_ssize_t __pyx_v_extra_len;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  char *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_next", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":534
 *         cdef Py_ssize_t extra_len
 * 
 *         if self.cur_cstr == NULL:             # <<<<<<<<<<<<<<
 *             raise AssertionError('get_next() called when cur_str is NULL')
 *         elif self.cur_cstr >= self.end_cstr:
*/
  __pyx_t_1 = (__pyx_v_self->cur_cstr == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":535
 * 
 *         if self.cur_cstr == NULL:
 *             raise AssertionError('get_next() called when cur_str is NULL')             # <<<<<<<<<<<<<<
 *         elif self.cur_cstr >= self.end_cstr:
 *             raise AssertionError('get_next() called when there are no chars'
*/
    __pyx_t_3 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_get_next_called_when_cur_str_is};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 535, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 535, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":534
 *         cdef Py_ssize_t extra_len
 * 
 *         if self.cur_cstr == NULL:             # <<<<<<<<<<<<<<
 *             raise AssertionError('get_next() called when cur_str is NULL')
 *         elif self.cur_cstr >= self.end_cstr:
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":536
 *         if self.cur_cstr == NULL:
 *             raise AssertionError('get_next() called when cur_str is NULL')
 *         elif self.cur_cstr >= self.end_cstr:             # <<<<<<<<<<<<<<
 *             raise AssertionError('get_next() called when there are no chars'
 *                                  ' left')
*/
  __pyx_t_1 = (__pyx_v_self->cur_cstr >= __pyx_v_self->end_cstr);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":537
 *             raise AssertionError('get_next() called when cur_str is NULL')
 *         elif self.cur_cstr >= self.end_cstr:
 *             raise AssertionError('get_next() called when there are no chars'             # <<<<<<<<<<<<<<
 *                                  ' left')
 *         next = self.cur_cstr
*/
    __pyx_t_3 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_get_next_called_when_there_are_n};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 537, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 537, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":536
 *         if self.cur_cstr == NULL:
 *             raise AssertionError('get_next() called when cur_str is NULL')
 *         elif self.cur_cstr >= self.end_cstr:             # <<<<<<<<<<<<<<
 *             raise AssertionError('get_next() called when there are no chars'
 *                                  ' left')
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":539
 *             raise AssertionError('get_next() called when there are no chars'
 *                                  ' left')
 *         next = self.cur_cstr             # <<<<<<<<<<<<<<
 *         self.cur_cstr = <char*>memchr(next, b'\0', self.end_cstr - next)
 *         if self.cur_cstr == NULL:
*/
  __pyx_t_5 = __pyx_v_self->cur_cstr;
  __pyx_v_next = __pyx_t_5;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":540
 *                                  ' left')
 *         next = self.cur_cstr
 *         self.cur_cstr = <char*>memchr(next, b'\0', self.end_cstr - next)             # <<<<<<<<<<<<<<
 *         if self.cur_cstr == NULL:
 *             extra_len = self.end_cstr - next
*/
  __pyx_v_self->cur_cstr = ((char *)memchr(__pyx_v_next, '\x00', (__pyx_v_self->end_cstr - __pyx_v_next)));

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":541
 *         next = self.cur_cstr
 *         self.cur_cstr = <char*>memchr(next, b'\0', self.end_cstr - next)
 *         if self.cur_cstr == NULL:             # <<<<<<<<<<<<<<
 *             extra_len = self.end_cstr - next
 *             raise DirstateCorrupt(self.state,
*/
  __pyx_t_1 = (__pyx_v_self->cur_cstr == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":542
 *         self.cur_cstr = <char*>memchr(next, b'\0', self.end_cstr - next)
 *         if self.cur_cstr == NULL:
 *             extra_len = self.end_cstr - next             # <<<<<<<<<<<<<<
 *             raise DirstateCorrupt(self.state,
 *                 'failed to find trailing NULL (\\0).'
*/
    __pyx_v_extra_len = (__pyx_v_self->end_cstr - __pyx_v_next);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":543
 *         if self.cur_cstr == NULL:
 *             extra_len = self.end_cstr - next
 *             raise DirstateCorrupt(self.state,             # <<<<<<<<<<<<<<
 *                 'failed to find trailing NULL (\\0).'
 *                 ' Trailing garbage: %r'
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_DirstateCorrupt); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 543, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":546
 *                 'failed to find trailing NULL (\\0).'
 *                 ' Trailing garbage: %r'
 *                 % safe_string_from_size(next, extra_len))             # <<<<<<<<<<<<<<
 *         size[0] = self.cur_cstr - next
 *         self.cur_cstr = self.cur_cstr + 1
*/
    __pyx_t_7 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(__pyx_v_next, __pyx_v_extra_len); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 546, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyUnicode_FormatSafe(__pyx_mstate_global->__pyx_kp_u_failed_to_find_trailing_NULL_0_T, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 546, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_self->state, __pyx_t_8};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_4, (3-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 543, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 543, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":541
 *         next = self.cur_cstr
 *         self.cur_cstr = <char*>memchr(next, b'\0', self.end_cstr - next)
 *         if self.cur_cstr == NULL:             # <<<<<<<<<<<<<<
 *             extra_len = self.end_cstr - next
 *             raise DirstateCorrupt(self.state,
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":547
 *                 ' Trailing garbage: %r'
 *                 % safe_string_from_size(next, extra_len))
 *         size[0] = self.cur_cstr - next             # <<<<<<<<<<<<<<
 *         self.cur_cstr = self.cur_cstr + 1
 *         return next
*/
  (__pyx_v_size[0]) = (__pyx_v_self->cur_cstr - __pyx_v_next);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":548
 *                 % safe_string_from_size(next, extra_len))
 *         size[0] = self.cur_cstr - next
 *         self.cur_cstr = self.cur_cstr + 1             # <<<<<<<<<<<<<<
 *         return next
 * 
*/
  __pyx_v_self->cur_cstr = (__pyx_v_self->cur_cstr + 1);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":549
 *         size[0] = self.cur_cstr - next
 *         self.cur_cstr = self.cur_cstr + 1
 *         return next             # <<<<<<<<<<<<<<
 * 
 *     cdef object get_next_str(self):
*/
  __pyx_r = __pyx_v_next;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":529
 *         self.cur_cstr = self.text_cstr
 * 
 *     cdef char *get_next(self, int *size) except NULL:             # <<<<<<<<<<<<<<
 *         """Return a pointer to the start of the next field."""
 *         cdef char *next
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.Reader.get_next", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":551
 *         return next
 * 
 *     cdef object get_next_str(self):             # <<<<<<<<<<<<<<
 *         """Get the next field as a Python string."""
 *         cdef int size
*/

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_get_next_str(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self) {
  int __pyx_v_size;
  char *__pyx_v_next;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_next_str", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":555
 *         cdef int size
 *         cdef char *next
 *         next = self.get_next(&size)             # <<<<<<<<<<<<<<
 *         return safe_string_from_size(next, size)
 * 
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next(__pyx_v_self, (&__pyx_v_size)); if (unlikely(__pyx_t_1 == ((void *)NULL))) __PYX_ERR(0, 555, __pyx_L1_error)
  __pyx_v_next = __pyx_t_1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":556
 *         cdef char *next
 *         next = self.get_next(&size)
 *         return safe_string_from_size(next, size)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _init(self) except -1:
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(__pyx_v_next, __pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":551
 *         return next
 * 
 *     cdef object get_next_str(self):             # <<<<<<<<<<<<<<
 *         """Get the next field as a Python string."""
 *         cdef int size
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.Reader.get_next_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":558
 *         return safe_string_from_size(next, size)
 * 
 *     cdef int _init(self) except -1:             # <<<<<<<<<<<<<<
 *         """Get the pointer ready.
 * 
*/

static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader__init(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self) {
  char *__pyx_v_first;
  int __pyx_v_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":569
 *         cdef int size
 *         # The first field should be an empty string left over from the Header
 *         first = self.get_next(&size)             # <<<<<<<<<<<<<<
 *         if first[0] != b'\0' and size == 0:
 *             raise AssertionError('First character should be null not: %s'
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next(__pyx_v_self, (&__pyx_v_size)); if (unlikely(__pyx_t_1 == ((void *)NULL))) __PYX_ERR(0, 569, __pyx_L1_error)
  __pyx_v_first = __pyx_t_1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":570
 *         # The first field should be an empty string left over from the Header
 *         first = self.get_next(&size)
 *         if first[0] != b'\0' and size == 0:             # <<<<<<<<<<<<<<
 *             raise AssertionError('First character should be null not: %s'
 *                                  % (first,))
*/
  __pyx_t_3 = ((__pyx_v_first[0]) != '\x00');
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_size == 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":571
 *         first = self.get_next(&size)
 *         if first[0] != b'\0' and size == 0:
 *             raise AssertionError('First character should be null not: %s'             # <<<<<<<<<<<<<<
 *                                  % (first,))
 *         return 0
*/
    __pyx_t_5 = NULL;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":572
 *         if first[0] != b'\0' and size == 0:
 *             raise AssertionError('First character should be null not: %s'
 *                                  % (first,))             # <<<<<<<<<<<<<<
 *         return 0
 * 
*/
    __pyx_t_6 = __Pyx_PyBytes_FromString(__pyx_v_first); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_t_6), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":571
 *         first = self.get_next(&size)
 *         if first[0] != b'\0' and size == 0:
 *             raise AssertionError('First character should be null not: %s'             # <<<<<<<<<<<<<<
 *                                  % (first,))
 *         return 0
*/
    __pyx_t_6 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_First_character_should_be_null_n, __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_8 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_6};
      __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 571, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 571, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":570
 *         # The first field should be an empty string left over from the Header
 *         first = self.get_next(&size)
 *         if first[0] != b'\0' and size == 0:             # <<<<<<<<<<<<<<
 *             raise AssertionError('First character should be null not: %s'
 *                                  % (first,))
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":573
 *             raise AssertionError('First character should be null not: %s'
 *                                  % (first,))
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef object _get_entry(self, int num_trees, void **p_current_dirname,
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":558
 *         return safe_string_from_size(next, size)
 * 
 *     cdef int _init(self) except -1:             # <<<<<<<<<<<<<<
 *         """Get the pointer ready.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.Reader._init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":575
 *         return 0
 * 
 *     cdef object _get_entry(self, int num_trees, void **p_current_dirname,             # <<<<<<<<<<<<<<
 *                            int *new_block):
 *         """Extract the next entry.
*/

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader__get_entry(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self, int __pyx_v_num_trees, void **__pyx_v_p_current_dirname, int *__pyx_v_new_block) {
  PyObject *__pyx_v_path_name_file_id_key = 0;
  PyObject *__pyx_v_tmp = 0;
  char *__pyx_v_entry_size_cstr;
  unsigned long __pyx_v_entry_size;
  char *__pyx_v_executable_cstr;
  int __pyx_v_is_executable;
  char *__pyx_v_dirname_cstr;
  char *__pyx_v_trailing;
  int __pyx_v_cur_size;
  CYTHON_UNUSED int __pyx_v_i;
  PyObject *__pyx_v_minikind = 0;
  PyObject *__pyx_v_fingerprint = 0;
  PyObject *__pyx_v_info = 0;
  PyObject *__pyx_v_dirname = NULL;
  PyObject *__pyx_v_cur_dirname = NULL;
  PyObject *__pyx_v_cur_basename = NULL;
  PyObject *__pyx_v_cur_file_id = NULL;
  PyObject *__pyx_v_trees = NULL;
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13[6];
  PyObject *__pyx_t_14 = NULL;
  size_t __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_entry", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":611
 * 
 *         # Read the 'key' information (dirname, name, file_id)
 *         dirname_cstr = self.get_next(&cur_size)             # <<<<<<<<<<<<<<
 *         # Check to see if we have started a new directory block.
 *         # If so, then we need to create a new dirname PyBytes, so that it can
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next(__pyx_v_self, (&__pyx_v_cur_size)); if (unlikely(__pyx_t_1 == ((void *)NULL))) __PYX_ERR(0, 611, __pyx_L1_error)
  __pyx_v_dirname_cstr = __pyx_t_1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":619
 *         # Do the cheap 'length of string' check first. If the string is a
 *         # different length, then we *have* to be a different directory.
 *         if (cur_size != PyBytes_GET_SIZE_void(p_current_dirname[0])             # <<<<<<<<<<<<<<
 *             or strncmp(dirname_cstr,
 *                        # Extract the char* from our current dirname string.  We
*/
  __pyx_t_3 = (__pyx_v_cur_size != PyBytes_GET_SIZE((__pyx_v_p_current_dirname[0])));
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":627
 *                        # <object>
 *                        PyBytes_AS_STRING_void(p_current_dirname[0]),
 *                        cur_size+1) != 0):             # <<<<<<<<<<<<<<
 *             dirname = safe_string_from_size(dirname_cstr, cur_size)
 *             p_current_dirname[0] = <void*>dirname
*/
  __pyx_t_3 = (strncmp(__pyx_v_dirname_cstr, PyBytes_AS_STRING((__pyx_v_p_current_dirname[0])), (__pyx_v_cur_size + 1)) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":619
 *         # Do the cheap 'length of string' check first. If the string is a
 *         # different length, then we *have* to be a different directory.
 *         if (cur_size != PyBytes_GET_SIZE_void(p_current_dirname[0])             # <<<<<<<<<<<<<<
 *             or strncmp(dirname_cstr,
 *                        # Extract the char* from our current dirname string.  We
*/
  if (__pyx_t_2) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":628
 *                        PyBytes_AS_STRING_void(p_current_dirname[0]),
 *                        cur_size+1) != 0):
 *             dirname = safe_string_from_size(dirname_cstr, cur_size)             # <<<<<<<<<<<<<<
 *             p_current_dirname[0] = <void*>dirname
 *             new_block[0] = 1
*/
    __pyx_t_4 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(__pyx_v_dirname_cstr, __pyx_v_cur_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_dirname = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":629
 *                        cur_size+1) != 0):
 *             dirname = safe_string_from_size(dirname_cstr, cur_size)
 *             p_current_dirname[0] = <void*>dirname             # <<<<<<<<<<<<<<
 *             new_block[0] = 1
 *         else:
*/
    (__pyx_v_p_current_dirname[0]) = ((void *)__pyx_v_dirname);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":630
 *             dirname = safe_string_from_size(dirname_cstr, cur_size)
 *             p_current_dirname[0] = <void*>dirname
 *             new_block[0] = 1             # <<<<<<<<<<<<<<
 *         else:
 *             new_block[0] = 0
*/
    (__pyx_v_new_block[0]) = 1;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":619
 *         # Do the cheap 'length of string' check first. If the string is a
 *         # different length, then we *have* to be a different directory.
 *         if (cur_size != PyBytes_GET_SIZE_void(p_current_dirname[0])             # <<<<<<<<<<<<<<
 *             or strncmp(dirname_cstr,
 *                        # Extract the char* from our current dirname string.  We
*/
    goto __pyx_L3;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":632
 *             new_block[0] = 1
 *         else:
 *             new_block[0] = 0             # <<<<<<<<<<<<<<
 * 
 *         # Build up the key that will be used.
*/
  /*else*/ {
    (__pyx_v_new_block[0]) = 0;
  }
  __pyx_L3:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":637
 *         # By using <object>(void *) Pyrex will automatically handle the
 *         # Py_INCREF that we need.
 *         cur_dirname = <object>p_current_dirname[0]             # <<<<<<<<<<<<<<
 *         tmp = PyTuple_New(3)
 *         Py_INCREF(cur_dirname); PyTuple_SET_ITEM(tmp, 0, cur_dirname)
*/
  __pyx_t_4 = ((PyObject *)(__pyx_v_p_current_dirname[0]));
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_cur_dirname = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":638
 *         # Py_INCREF that we need.
 *         cur_dirname = <object>p_current_dirname[0]
 *         tmp = PyTuple_New(3)             # <<<<<<<<<<<<<<
 *         Py_INCREF(cur_dirname); PyTuple_SET_ITEM(tmp, 0, cur_dirname)
 *         cur_basename = self.get_next_str()
*/
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_tmp = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":639
 *         cur_dirname = <object>p_current_dirname[0]
 *         tmp = PyTuple_New(3)
 *         Py_INCREF(cur_dirname); PyTuple_SET_ITEM(tmp, 0, cur_dirname)             # <<<<<<<<<<<<<<
 *         cur_basename = self.get_next_str()
 *         cur_file_id = self.get_next_str()
*/
  Py_INCREF(__pyx_v_cur_dirname);
  PyTuple_SET_ITEM(__pyx_v_tmp, 0, __pyx_v_cur_dirname);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":640
 *         tmp = PyTuple_New(3)
 *         Py_INCREF(cur_dirname); PyTuple_SET_ITEM(tmp, 0, cur_dirname)
 *         cur_basename = self.get_next_str()             # <<<<<<<<<<<<<<
 *         cur_file_id = self.get_next_str()
 *         Py_INCREF(cur_basename); PyTuple_SET_ITEM(tmp, 1, cur_basename)
*/
  __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next_str(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_cur_basename = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":641
 *         Py_INCREF(cur_dirname); PyTuple_SET_ITEM(tmp, 0, cur_dirname)
 *         cur_basename = self.get_next_str()
 *         cur_file_id = self.get_next_str()             # <<<<<<<<<<<<<<
 *         Py_INCREF(cur_basename); PyTuple_SET_ITEM(tmp, 1, cur_basename)
 *         Py_INCREF(cur_file_id); PyTuple_SET_ITEM(tmp, 2, cur_file_id)
*/
  __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next_str(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_cur_file_id = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":642
 *         cur_basename = self.get_next_str()
 *         cur_file_id = self.get_next_str()
 *         Py_INCREF(cur_basename); PyTuple_SET_ITEM(tmp, 1, cur_basename)             # <<<<<<<<<<<<<<
 *         Py_INCREF(cur_file_id); PyTuple_SET_ITEM(tmp, 2, cur_file_id)
 *         path_name_file_id_key = tmp
*/
  Py_INCREF(__pyx_v_cur_basename);
  PyTuple_SET_ITEM(__pyx_v_tmp, 1, __pyx_v_cur_basename);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":643
 *         cur_file_id = self.get_next_str()
 *         Py_INCREF(cur_basename); PyTuple_SET_ITEM(tmp, 1, cur_basename)
 *         Py_INCREF(cur_file_id); PyTuple_SET_ITEM(tmp, 2, cur_file_id)             # <<<<<<<<<<<<<<
 *         path_name_file_id_key = tmp
 * 
*/
  Py_INCREF(__pyx_v_cur_file_id);
  PyTuple_SET_ITEM(__pyx_v_tmp, 2, __pyx_v_cur_file_id);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":644
 *         Py_INCREF(cur_basename); PyTuple_SET_ITEM(tmp, 1, cur_basename)
 *         Py_INCREF(cur_file_id); PyTuple_SET_ITEM(tmp, 2, cur_file_id)
 *         path_name_file_id_key = tmp             # <<<<<<<<<<<<<<
 * 
 *         # Parse all of the per-tree information. current has the information in
*/
  __Pyx_INCREF(__pyx_v_tmp);
  __pyx_v_path_name_file_id_key = __pyx_v_tmp;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":659
 *         #       Especially since this code is pretty much fixed at a max of
 *         #       4GB.
 *         trees = []             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < num_trees:
 *             minikind = self.get_next_str()
*/
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_trees = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":660
 *         #       4GB.
 *         trees = []
 *         for i from 0 <= i < num_trees:             # <<<<<<<<<<<<<<
 *             minikind = self.get_next_str()
 *             fingerprint = self.get_next_str()
*/
  __pyx_t_5 = __pyx_v_num_trees;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_5; __pyx_v_i++) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":661
 *         trees = []
 *         for i from 0 <= i < num_trees:
 *             minikind = self.get_next_str()             # <<<<<<<<<<<<<<
 *             fingerprint = self.get_next_str()
 *             entry_size_cstr = self.get_next(&cur_size)
*/
    __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next_str(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 661, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_minikind, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":662
 *         for i from 0 <= i < num_trees:
 *             minikind = self.get_next_str()
 *             fingerprint = self.get_next_str()             # <<<<<<<<<<<<<<
 *             entry_size_cstr = self.get_next(&cur_size)
 *             entry_size = strtoul(entry_size_cstr, NULL, 10)
*/
    __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next_str(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 662, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_fingerprint, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":663
 *             minikind = self.get_next_str()
 *             fingerprint = self.get_next_str()
 *             entry_size_cstr = self.get_next(&cur_size)             # <<<<<<<<<<<<<<
 *             entry_size = strtoul(entry_size_cstr, NULL, 10)
 *             executable_cstr = self.get_next(&cur_size)
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next(__pyx_v_self, (&__pyx_v_cur_size)); if (unlikely(__pyx_t_1 == ((void *)NULL))) __PYX_ERR(0, 663, __pyx_L1_error)
    __pyx_v_entry_size_cstr = __pyx_t_1;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":664
 *             fingerprint = self.get_next_str()
 *             entry_size_cstr = self.get_next(&cur_size)
 *             entry_size = strtoul(entry_size_cstr, NULL, 10)             # <<<<<<<<<<<<<<
 *             executable_cstr = self.get_next(&cur_size)
 *             is_executable = (executable_cstr[0] == b'y')
*/
    __pyx_v_entry_size = strtoul(__pyx_v_entry_size_cstr, NULL, 10);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":665
 *             entry_size_cstr = self.get_next(&cur_size)
 *             entry_size = strtoul(entry_size_cstr, NULL, 10)
 *             executable_cstr = self.get_next(&cur_size)             # <<<<<<<<<<<<<<
 *             is_executable = (executable_cstr[0] == b'y')
 *             info = self.get_next_str()
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next(__pyx_v_self, (&__pyx_v_cur_size)); if (unlikely(__pyx_t_1 == ((void *)NULL))) __PYX_ERR(0, 665, __pyx_L1_error)
    __pyx_v_executable_cstr = __pyx_t_1;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":666
 *             entry_size = strtoul(entry_size_cstr, NULL, 10)
 *             executable_cstr = self.get_next(&cur_size)
 *             is_executable = (executable_cstr[0] == b'y')             # <<<<<<<<<<<<<<
 *             info = self.get_next_str()
 *             PyList_Append(trees, (
*/
    __pyx_v_is_executable = ((__pyx_v_executable_cstr[0]) == 'y');

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":667
 *             executable_cstr = self.get_next(&cur_size)
 *             is_executable = (executable_cstr[0] == b'y')
 *             info = self.get_next_str()             # <<<<<<<<<<<<<<
 *             PyList_Append(trees, (
 *                 minikind,     # minikind
*/
    __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next_str(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_info, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":671
 *                 minikind,     # minikind
 *                 fingerprint,  # fingerprint
 *                 entry_size,   # size             # <<<<<<<<<<<<<<
 *                 is_executable,# executable
 *                 info,         # packed_stat or revision_id
*/
    __pyx_t_4 = __Pyx_PyLong_From_unsigned_long(__pyx_v_entry_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 671, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":672
 *                 fingerprint,  # fingerprint
 *                 entry_size,   # size
 *                 is_executable,# executable             # <<<<<<<<<<<<<<
 *                 info,         # packed_stat or revision_id
 *             ))
*/
    __pyx_t_6 = __Pyx_PyLong_From_int(__pyx_v_is_executable); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 672, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":669
 *             info = self.get_next_str()
 *             PyList_Append(trees, (
 *                 minikind,     # minikind             # <<<<<<<<<<<<<<
 *                 fingerprint,  # fingerprint
 *                 entry_size,   # size
*/
    __pyx_t_7 = PyTuple_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_minikind);
    __Pyx_GIVEREF(__pyx_v_minikind);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_minikind) != (0)) __PYX_ERR(0, 669, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_fingerprint);
    __Pyx_GIVEREF(__pyx_v_fingerprint);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_fingerprint) != (0)) __PYX_ERR(0, 669, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_4) != (0)) __PYX_ERR(0, 669, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_6) != (0)) __PYX_ERR(0, 669, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_info);
    __Pyx_GIVEREF(__pyx_v_info);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_v_info) != (0)) __PYX_ERR(0, 669, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_6 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":668
 *             is_executable = (executable_cstr[0] == b'y')
 *             info = self.get_next_str()
 *             PyList_Append(trees, (             # <<<<<<<<<<<<<<
 *                 minikind,     # minikind
 *                 fingerprint,  # fingerprint
*/
    __pyx_t_8 = PyList_Append(__pyx_v_trees, __pyx_t_7); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 668, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":677
 * 
 *         # The returned tuple is (key, [trees])
 *         ret = (path_name_file_id_key, trees)             # <<<<<<<<<<<<<<
 *         # Ignore the trailing newline, but assert that it does exist, this
 *         # ensures that we always finish parsing a line on an end-of-entry
*/
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_path_name_file_id_key);
  __Pyx_GIVEREF(__pyx_v_path_name_file_id_key);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_path_name_file_id_key) != (0)) __PYX_ERR(0, 677, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_trees);
  __Pyx_GIVEREF(__pyx_v_trees);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_trees) != (0)) __PYX_ERR(0, 677, __pyx_L1_error);
  __pyx_v_ret = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":681
 *         # ensures that we always finish parsing a line on an end-of-entry
 *         # marker.
 *         trailing = self.get_next(&cur_size)             # <<<<<<<<<<<<<<
 *         if cur_size != 1 or not trailing.startswith(b'\n'):
 *             raise DirstateCorrupt(self.state,
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next(__pyx_v_self, (&__pyx_v_cur_size)); if (unlikely(__pyx_t_1 == ((void *)NULL))) __PYX_ERR(0, 681, __pyx_L1_error)
  __pyx_v_trailing = __pyx_t_1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":682
 *         # marker.
 *         trailing = self.get_next(&cur_size)
 *         if cur_size != 1 or not trailing.startswith(b'\n'):             # <<<<<<<<<<<<<<
 *             raise DirstateCorrupt(self.state,
 *                 'Bad parse, we expected to end on \\n, not: %d %s: %s'
*/
  __pyx_t_3 = (__pyx_v_cur_size != 1);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_PyBytes_FromString(__pyx_v_trailing); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_PyBytes_Tailmatch(__pyx_t_7, __pyx_mstate_global->__pyx_kp_b__3, 0, PY_SSIZE_T_MAX, -1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_9 = (!__pyx_t_3);
  __pyx_t_2 = __pyx_t_9;
  __pyx_L9_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":683
 *         trailing = self.get_next(&cur_size)
 *         if cur_size != 1 or not trailing.startswith(b'\n'):
 *             raise DirstateCorrupt(self.state,             # <<<<<<<<<<<<<<
 *                 'Bad parse, we expected to end on \\n, not: %d %s: %s'
 *                 % (cur_size, safe_string_from_size(trailing, cur_size),
*/
    __pyx_t_6 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DirstateCorrupt); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":685
 *             raise DirstateCorrupt(self.state,
 *                 'Bad parse, we expected to end on \\n, not: %d %s: %s'
 *                 % (cur_size, safe_string_from_size(trailing, cur_size),             # <<<<<<<<<<<<<<
 *                    ret))
 *         return ret
*/
    __pyx_t_10 = __Pyx_PyUnicode_From_int(__pyx_v_cur_size, 0, ' ', 'd'); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(__pyx_v_trailing, __pyx_v_cur_size); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_t_11), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":686
 *                 'Bad parse, we expected to end on \\n, not: %d %s: %s'
 *                 % (cur_size, safe_string_from_size(trailing, cur_size),
 *                    ret))             # <<<<<<<<<<<<<<
 *         return ret
 * 
*/
    __pyx_t_11 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_v_ret), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 686, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13[0] = __pyx_mstate_global->__pyx_kp_u_Bad_parse_we_expected_to_end_on;
    __pyx_t_13[1] = __pyx_t_10;
    __pyx_t_13[2] = __pyx_mstate_global->__pyx_kp_u__2;
    __pyx_t_13[3] = __pyx_t_12;
    __pyx_t_13[4] = __pyx_mstate_global->__pyx_kp_u_;
    __pyx_t_13[5] = __pyx_t_11;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":684
 *         if cur_size != 1 or not trailing.startswith(b'\n'):
 *             raise DirstateCorrupt(self.state,
 *                 'Bad parse, we expected to end on \\n, not: %d %s: %s'             # <<<<<<<<<<<<<<
 *                 % (cur_size, safe_string_from_size(trailing, cur_size),
 *                    ret))
*/
    __pyx_t_14 = __Pyx_PyUnicode_Join(__pyx_t_13, 6, 42 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10) + 1 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12) + 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_11), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11));
    if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 684, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_15 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_6);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_15 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_self->state, __pyx_t_14};
      __pyx_t_7 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_15, (3-__pyx_t_15) | (__pyx_t_15*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 683, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 683, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":682
 *         # marker.
 *         trailing = self.get_next(&cur_size)
 *         if cur_size != 1 or not trailing.startswith(b'\n'):             # <<<<<<<<<<<<<<
 *             raise DirstateCorrupt(self.state,
 *                 'Bad parse, we expected to end on \\n, not: %d %s: %s'
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":687
 *                 % (cur_size, safe_string_from_size(trailing, cur_size),
 *                    ret))
 *         return ret             # <<<<<<<<<<<<<<
 * 
 *     def _parse_dirblocks(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ret);
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":575
 *         return 0
 * 
 *     cdef object _get_entry(self, int num_trees, void **p_current_dirname,             # <<<<<<<<<<<<<<
 *                            int *new_block):
 *         """Extract the next entry.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.Reader._get_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_path_name_file_id_key);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XDECREF(__pyx_v_minikind);
  __Pyx_XDECREF(__pyx_v_fingerprint);
  __Pyx_XDECREF(__pyx_v_info);
  __Pyx_XDECREF(__pyx_v_dirname);
  __Pyx_XDECREF(__pyx_v_cur_dirname);
  __Pyx_XDECREF(__pyx_v_cur_basename);
  __Pyx_XDECREF(__pyx_v_cur_file_id);
  __Pyx_XDECREF(__pyx_v_trees);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":689
 *         return ret
 * 
 *     def _parse_dirblocks(self):             # <<<<<<<<<<<<<<
 *         """Parse all dirblocks in the state file."""
 *         cdef int num_trees
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_3_parse_dirblocks(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_2_parse_dirblocks, "Parse all dirblocks in the state file.");
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_3_parse_dirblocks = {"_parse_dirblocks", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_3_parse_dirblocks, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_2_parse_dirblocks};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_3_parse_dirblocks(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_parse_dirblocks (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_parse_dirblocks", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_parse_dirblocks", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_2_parse_dirblocks(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_2_parse_dirblocks(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self) {
  int __pyx_v_num_trees;
  PyObject *__pyx_v_current_block = 0;
  PyObject *__pyx_v_entry = 0;
  void *__pyx_v_current_dirname;
  int __pyx_v_new_block;
  int __pyx_v_expected_entry_count;
  int __pyx_v_entry_count;
  PyObject *__pyx_v_dirblocks = NULL;
  PyObject *__pyx_v_obj = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9[4];
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_parse_dirblocks", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":699
 *         cdef int entry_count
 * 
 *         num_trees = self.state._num_present_parents() + 1             # <<<<<<<<<<<<<<
 *         expected_entry_count = self.state._num_entries
 * 
*/
  __pyx_t_2 = __pyx_v_self->state;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_num_present_parents, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_2 = __Pyx_PyLong_AddObjC(__pyx_t_1, __pyx_mstate_global->__pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_t_2); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_num_trees = __pyx_t_4;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":700
 * 
 *         num_trees = self.state._num_present_parents() + 1
 *         expected_entry_count = self.state._num_entries             # <<<<<<<<<<<<<<
 * 
 *         # Ignore the first record
*/
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_mstate_global->__pyx_n_u_num_entries); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_t_2); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_expected_entry_count = __pyx_t_4;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":703
 * 
 *         # Ignore the first record
 *         self._init()             # <<<<<<<<<<<<<<
 * 
 *         current_block = []
*/
  __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->_init(__pyx_v_self); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 703, __pyx_L1_error)

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":705
 *         self._init()
 * 
 *         current_block = []             # <<<<<<<<<<<<<<
 *         dirblocks = [(b'', current_block), (b'', [])]
 *         self.state._dirblocks = dirblocks
*/
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_current_block = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":706
 * 
 *         current_block = []
 *         dirblocks = [(b'', current_block), (b'', [])]             # <<<<<<<<<<<<<<
 *         self.state._dirblocks = dirblocks
 *         obj = b''
*/
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b__4);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_b__4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_mstate_global->__pyx_kp_b__4) != (0)) __PYX_ERR(0, 706, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_current_block);
  __Pyx_GIVEREF(__pyx_v_current_block);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_current_block) != (0)) __PYX_ERR(0, 706, __pyx_L1_error);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b__4);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_b__4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_mstate_global->__pyx_kp_b__4) != (0)) __PYX_ERR(0, 706, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1) != (0)) __PYX_ERR(0, 706, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_2) != (0)) __PYX_ERR(0, 706, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, __pyx_t_5) != (0)) __PYX_ERR(0, 706, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_5 = 0;
  __pyx_v_dirblocks = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":707
 *         current_block = []
 *         dirblocks = [(b'', current_block), (b'', [])]
 *         self.state._dirblocks = dirblocks             # <<<<<<<<<<<<<<
 *         obj = b''
 *         current_dirname = <void*>obj
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->state, __pyx_mstate_global->__pyx_n_u_dirblocks_2, __pyx_v_dirblocks) < (0)) __PYX_ERR(0, 707, __pyx_L1_error)

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":708
 *         dirblocks = [(b'', current_block), (b'', [])]
 *         self.state._dirblocks = dirblocks
 *         obj = b''             # <<<<<<<<<<<<<<
 *         current_dirname = <void*>obj
 *         new_block = 0
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b__4);
  __pyx_v_obj = __pyx_mstate_global->__pyx_kp_b__4;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":709
 *         self.state._dirblocks = dirblocks
 *         obj = b''
 *         current_dirname = <void*>obj             # <<<<<<<<<<<<<<
 *         new_block = 0
 *         entry_count = 0
*/
  __pyx_v_current_dirname = ((void *)__pyx_v_obj);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":710
 *         obj = b''
 *         current_dirname = <void*>obj
 *         new_block = 0             # <<<<<<<<<<<<<<
 *         entry_count = 0
 * 
*/
  __pyx_v_new_block = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":711
 *         current_dirname = <void*>obj
 *         new_block = 0
 *         entry_count = 0             # <<<<<<<<<<<<<<
 * 
 *         # TODO: jam 2007-05-07 Consider pre-allocating some space for the
*/
  __pyx_v_entry_count = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":720
 *         #       so), and then truncate. That would give us a malloc + realloc,
 *         #       rather than lots of reallocs.
 *         while self.cur_cstr < self.end_cstr:             # <<<<<<<<<<<<<<
 *             entry = self._get_entry(num_trees, &current_dirname, &new_block)
 *             if new_block:
*/
  while (1) {
    __pyx_t_6 = (__pyx_v_self->cur_cstr < __pyx_v_self->end_cstr);
    if (!__pyx_t_6) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":721
 *         #       rather than lots of reallocs.
 *         while self.cur_cstr < self.end_cstr:
 *             entry = self._get_entry(num_trees, &current_dirname, &new_block)             # <<<<<<<<<<<<<<
 *             if new_block:
 *                 # new block - different dirname
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->_get_entry(__pyx_v_self, __pyx_v_num_trees, (&__pyx_v_current_dirname), (&__pyx_v_new_block)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 721, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_entry, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":722
 *         while self.cur_cstr < self.end_cstr:
 *             entry = self._get_entry(num_trees, &current_dirname, &new_block)
 *             if new_block:             # <<<<<<<<<<<<<<
 *                 # new block - different dirname
 *                 current_block = []
*/
    __pyx_t_6 = (__pyx_v_new_block != 0);
    if (__pyx_t_6) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":724
 *             if new_block:
 *                 # new block - different dirname
 *                 current_block = []             # <<<<<<<<<<<<<<
 *                 PyList_Append(dirblocks,
 *                               (<object>current_dirname, current_block))
*/
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 724, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_current_block, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":726
 *                 current_block = []
 *                 PyList_Append(dirblocks,
 *                               (<object>current_dirname, current_block))             # <<<<<<<<<<<<<<
 *             PyList_Append(current_block, entry)
 *             entry_count = entry_count + 1
*/
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(((PyObject *)__pyx_v_current_dirname));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_current_dirname));
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_current_dirname)) != (0)) __PYX_ERR(0, 726, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_current_block);
      __Pyx_GIVEREF(__pyx_v_current_block);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_current_block) != (0)) __PYX_ERR(0, 726, __pyx_L1_error);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":725
 *                 # new block - different dirname
 *                 current_block = []
 *                 PyList_Append(dirblocks,             # <<<<<<<<<<<<<<
 *                               (<object>current_dirname, current_block))
 *             PyList_Append(current_block, entry)
*/
      __pyx_t_4 = PyList_Append(__pyx_v_dirblocks, __pyx_t_1); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 725, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":722
 *         while self.cur_cstr < self.end_cstr:
 *             entry = self._get_entry(num_trees, &current_dirname, &new_block)
 *             if new_block:             # <<<<<<<<<<<<<<
 *                 # new block - different dirname
 *                 current_block = []
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":727
 *                 PyList_Append(dirblocks,
 *                               (<object>current_dirname, current_block))
 *             PyList_Append(current_block, entry)             # <<<<<<<<<<<<<<
 *             entry_count = entry_count + 1
 *         if entry_count != expected_entry_count:
*/
    __pyx_t_4 = PyList_Append(__pyx_v_current_block, __pyx_v_entry); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 727, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":728
 *                               (<object>current_dirname, current_block))
 *             PyList_Append(current_block, entry)
 *             entry_count = entry_count + 1             # <<<<<<<<<<<<<<
 *         if entry_count != expected_entry_count:
 *             raise DirstateCorrupt(self.state,
*/
    __pyx_v_entry_count = (__pyx_v_entry_count + 1);
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":729
 *             PyList_Append(current_block, entry)
 *             entry_count = entry_count + 1
 *         if entry_count != expected_entry_count:             # <<<<<<<<<<<<<<
 *             raise DirstateCorrupt(self.state,
 *                     'We read the wrong number of entries.'
*/
  __pyx_t_6 = (__pyx_v_entry_count != __pyx_v_expected_entry_count);
  if (unlikely(__pyx_t_6)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":730
 *             entry_count = entry_count + 1
 *         if entry_count != expected_entry_count:
 *             raise DirstateCorrupt(self.state,             # <<<<<<<<<<<<<<
 *                     'We read the wrong number of entries.'
 *                     ' We expected to read %s, but read %s'
*/
    __pyx_t_5 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_DirstateCorrupt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":733
 *                     'We read the wrong number of entries.'
 *                     ' We expected to read %s, but read %s'
 *                     % (expected_entry_count, entry_count))             # <<<<<<<<<<<<<<
 *         self.state._split_root_dirblock_into_contents()
 * 
*/
    __pyx_t_7 = __Pyx_PyUnicode_From_int(__pyx_v_expected_entry_count, 0, ' ', 'd'); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 733, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyUnicode_From_int(__pyx_v_entry_count, 0, ' ', 'd'); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 733, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9[0] = __pyx_mstate_global->__pyx_kp_u_We_read_the_wrong_number_of_entr;
    __pyx_t_9[1] = __pyx_t_7;
    __pyx_t_9[2] = __pyx_mstate_global->__pyx_kp_u_but_read;
    __pyx_t_9[3] = __pyx_t_8;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":731
 *         if entry_count != expected_entry_count:
 *             raise DirstateCorrupt(self.state,
 *                     'We read the wrong number of entries.'             # <<<<<<<<<<<<<<
 *                     ' We expected to read %s, but read %s'
 *                     % (expected_entry_count, entry_count))
*/
    __pyx_t_10 = __Pyx_PyUnicode_Join(__pyx_t_9, 4, 57 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7) + 11 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8), 127);
    if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 731, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      assert(__pyx_t_5);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_self->state, __pyx_t_10};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_3, (3-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 730, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":729
 *             PyList_Append(current_block, entry)
 *             entry_count = entry_count + 1
 *         if entry_count != expected_entry_count:             # <<<<<<<<<<<<<<
 *             raise DirstateCorrupt(self.state,
 *                     'We read the wrong number of entries.'
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":734
 *                     ' We expected to read %s, but read %s'
 *                     % (expected_entry_count, entry_count))
 *         self.state._split_root_dirblock_into_contents()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_v_self->state;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_split_root_dirblock_into_conten, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 734, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":689
 *         return ret
 * 
 *     def _parse_dirblocks(self):             # <<<<<<<<<<<<<<
 *         """Parse all dirblocks in the state file."""
 *         cdef int num_trees
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.Reader._parse_dirblocks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_current_block);
  __Pyx_XDECREF(__pyx_v_entry);
  __Pyx_XDECREF(__pyx_v_dirblocks);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_4__reduce_cython__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_4__reduce_cython__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.cur_cstr, self.end_cstr, self.next, self.state, self.text, self.text_cstr, self.text_size)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->cur_cstr); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBytes_FromString(__pyx_v_self->end_cstr); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_self->next); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_self->text_cstr); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyLong_From_int(__pyx_v_self->text_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(7); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_3) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->state);
  __Pyx_GIVEREF(__pyx_v_self->state);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_v_self->state) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->text);
  __Pyx_GIVEREF(__pyx_v_self->text);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_v_self->text) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 5, __pyx_t_4) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 6, __pyx_t_5) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.cur_cstr, self.end_cstr, self.next, self.state, self.text, self.text_cstr, self.text_size)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_6 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v__dict = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "(tree fragment)":7
 *     state = (self.cur_cstr, self.end_cstr, self.next, self.state, self.text, self.text_cstr, self.text_size)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_8 = (__pyx_v__dict != Py_None);
  if (__pyx_t_8) {
  } else {
    __pyx_t_7 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(2, 7, __pyx_L1_error)
  __pyx_t_7 = __pyx_t_8;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_7) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v__dict) != (0)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.state is not None or self.text is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.cur_cstr, self.end_cstr, self.next, self.state, self.text, self.text_cstr, self.text_size)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.state is not None or self.text is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Reader, (type(self), 0xff2cf55, None), state
*/
  /*else*/ {
    __pyx_t_8 = (__pyx_v_self->state != Py_None);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_8 = (__pyx_v_self->text != Py_None);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L6_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_7;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.state is not None or self.text is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Reader, (type(self), 0xff2cf55, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.state is not None or self.text is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Reader, (type(self), 0xff2cf55, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Reader, (type(self), 0xff2cf55, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_Reader); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_267571029);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_267571029);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_mstate_global->__pyx_int_267571029) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, Py_None) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.state is not None or self.text is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Reader, (type(self), 0xff2cf55, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Reader, (type(self), 0xff2cf55, None), state
 *     else:
 *         return __pyx_unpickle_Reader, (type(self), 0xff2cf55, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Reader__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_Reader); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_267571029);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_267571029);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_mstate_global->__pyx_int_267571029) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_6 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Reader, (type(self), 0xff2cf55, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Reader__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_6__setstate_cython__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_6__setstate_cython__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Reader, (type(self), 0xff2cf55, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Reader__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(2, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx___pyx_unpickle_Reader__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Reader, (type(self), 0xff2cf55, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Reader__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":737
 * 
 * 
 * def _read_dirblocks(state):             # <<<<<<<<<<<<<<
 *     """Read in the dirblocks for the given DirState object.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13_read_dirblocks(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_12_read_dirblocks, "Read in the dirblocks for the given DirState object.\n\n    This is tightly bound to the DirState internal representation. It should be\n    thought of as a member function, which is only separated out so that we can\n    re-write it in pyrex.\n\n    :param state: A DirState object.\n    :return: None\n    :postcondition: The dirblocks will be loaded into the appropriate fields in\n        the DirState object.\n    ");
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_13_read_dirblocks = {"_read_dirblocks", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13_read_dirblocks, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_12_read_dirblocks};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13_read_dirblocks(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_read_dirblocks (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 737, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 737, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_read_dirblocks", 0) < (0)) __PYX_ERR(0, 737, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_read_dirblocks", 1, 1, 1, i); __PYX_ERR(0, 737, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 737, __pyx_L3_error)
    }
    __pyx_v_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_read_dirblocks", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 737, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._read_dirblocks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_12_read_dirblocks(__pyx_self, __pyx_v_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_12_read_dirblocks(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_v_text = NULL;
  struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_reader = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_read_dirblocks", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":749
 *         the DirState object.
 *     """
 *     state._state_file.seek(state._end_of_header)             # <<<<<<<<<<<<<<
 *     text = state._state_file.read()
 *     # TODO: check the crc checksums. crc_measured = zlib.crc32(text)
*/
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_state, __pyx_mstate_global->__pyx_n_u_state_file); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_state, __pyx_mstate_global->__pyx_n_u_end_of_header); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_seek, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 749, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":750
 *     """
 *     state._state_file.seek(state._end_of_header)
 *     text = state._state_file.read()             # <<<<<<<<<<<<<<
 *     # TODO: check the crc checksums. crc_measured = zlib.crc32(text)
 * 
*/
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_state, __pyx_mstate_global->__pyx_n_u_state_file); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __pyx_t_4;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_read, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 750, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_text = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":753
 *     # TODO: check the crc checksums. crc_measured = zlib.crc32(text)
 * 
 *     reader = Reader(text, state)             # <<<<<<<<<<<<<<
 * 
 *     reader._parse_dirblocks()
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_text, __pyx_v_state};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_reader = ((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":755
 *     reader = Reader(text, state)
 * 
 *     reader._parse_dirblocks()             # <<<<<<<<<<<<<<
 *     state._dirblock_state = DirState.IN_MEMORY_UNMODIFIED
 * 
*/
  __pyx_t_4 = ((PyObject *)__pyx_v_reader);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_parse_dirblocks, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 755, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":756
 * 
 *     reader._parse_dirblocks()
 *     state._dirblock_state = DirState.IN_MEMORY_UNMODIFIED             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_DirState); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_IN_MEMORY_UNMODIFIED); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_state, __pyx_mstate_global->__pyx_n_u_dirblock_state, __pyx_t_4) < (0)) __PYX_ERR(0, 756, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":737
 * 
 * 
 * def _read_dirblocks(state):             # <<<<<<<<<<<<<<
 *     """Read in the dirblocks for the given DirState object.
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._read_dirblocks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_text);
  __Pyx_XDECREF((PyObject *)__pyx_v_reader);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":759
 * 
 * 
 * cdef int minikind_from_mode(int mode): # cannot_raise             # <<<<<<<<<<<<<<
 *     # in order of frequency:
 *     if S_ISREG(mode):
*/

static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_minikind_from_mode(int __pyx_v_mode) {
  int __pyx_r;
  int __pyx_t_1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":761
 * cdef int minikind_from_mode(int mode): # cannot_raise
 *     # in order of frequency:
 *     if S_ISREG(mode):             # <<<<<<<<<<<<<<
 *         return c"f"
 *     if S_ISDIR(mode):
*/
  __pyx_t_1 = (S_ISREG(__pyx_v_mode) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":762
 *     # in order of frequency:
 *     if S_ISREG(mode):
 *         return c"f"             # <<<<<<<<<<<<<<
 *     if S_ISDIR(mode):
 *         return c"d"
*/
    __pyx_r = 'f';
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":761
 * cdef int minikind_from_mode(int mode): # cannot_raise
 *     # in order of frequency:
 *     if S_ISREG(mode):             # <<<<<<<<<<<<<<
 *         return c"f"
 *     if S_ISDIR(mode):
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":763
 *     if S_ISREG(mode):
 *         return c"f"
 *     if S_ISDIR(mode):             # <<<<<<<<<<<<<<
 *         return c"d"
 *     if S_ISLNK(mode):
*/
  __pyx_t_1 = (S_ISDIR(__pyx_v_mode) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":764
 *         return c"f"
 *     if S_ISDIR(mode):
 *         return c"d"             # <<<<<<<<<<<<<<
 *     if S_ISLNK(mode):
 *         return c"l"
*/
    __pyx_r = 'd';
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":763
 *     if S_ISREG(mode):
 *         return c"f"
 *     if S_ISDIR(mode):             # <<<<<<<<<<<<<<
 *         return c"d"
 *     if S_ISLNK(mode):
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":765
 *     if S_ISDIR(mode):
 *         return c"d"
 *     if S_ISLNK(mode):             # <<<<<<<<<<<<<<
 *         return c"l"
 *     return 0
*/
  __pyx_t_1 = (S_ISLNK(__pyx_v_mode) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":766
 *         return c"d"
 *     if S_ISLNK(mode):
 *         return c"l"             # <<<<<<<<<<<<<<
 *     return 0
 * 
*/
    __pyx_r = 'l';
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":765
 *     if S_ISDIR(mode):
 *         return c"d"
 *     if S_ISLNK(mode):             # <<<<<<<<<<<<<<
 *         return c"l"
 *     return 0
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":767
 *     if S_ISLNK(mode):
 *         return c"l"
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":759
 * 
 * 
 * cdef int minikind_from_mode(int mode): # cannot_raise             # <<<<<<<<<<<<<<
 *     # in order of frequency:
 *     if S_ISREG(mode):
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":773
 * 
 * 
 * cdef unsigned long _time_to_unsigned(object t):  # cannot_raise             # <<<<<<<<<<<<<<
 *     if PyFloat_Check(t):
 *         t = t.__int__()
*/

static unsigned long __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__time_to_unsigned(PyObject *__pyx_v_t) {
  unsigned long __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  unsigned long __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_time_to_unsigned", 0);
  __Pyx_INCREF(__pyx_v_t);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":774
 * 
 * cdef unsigned long _time_to_unsigned(object t):  # cannot_raise
 *     if PyFloat_Check(t):             # <<<<<<<<<<<<<<
 *         t = t.__int__()
 *     return PyLong_AsUnsignedLongMask(t)
*/
  __pyx_t_1 = (PyFloat_Check(__pyx_v_t) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":775
 * cdef unsigned long _time_to_unsigned(object t):  # cannot_raise
 *     if PyFloat_Check(t):
 *         t = t.__int__()             # <<<<<<<<<<<<<<
 *     return PyLong_AsUnsignedLongMask(t)
 * 
*/
    __pyx_t_3 = __pyx_v_t;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_int, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 775, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF_SET(__pyx_v_t, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":774
 * 
 * cdef unsigned long _time_to_unsigned(object t):  # cannot_raise
 *     if PyFloat_Check(t):             # <<<<<<<<<<<<<<
 *         t = t.__int__()
 *     return PyLong_AsUnsignedLongMask(t)
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":776
 *     if PyFloat_Check(t):
 *         t = t.__int__()
 *     return PyLong_AsUnsignedLongMask(t)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_5 = PyLong_AsUnsignedLongMask(__pyx_v_t); if (unlikely(__pyx_t_5 == ((unsigned long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 776, __pyx_L1_error)
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":773
 * 
 * 
 * cdef unsigned long _time_to_unsigned(object t):  # cannot_raise             # <<<<<<<<<<<<<<
 *     if PyFloat_Check(t):
 *         t = t.__int__()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._time_to_unsigned", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1L;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":779
 * 
 * 
 * cdef _pack_stat(stat_value):             # <<<<<<<<<<<<<<
 *     """return a string representing the stat value's key fields.
 * 
*/

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__pack_stat(PyObject *__pyx_v_stat_value) {
  char __pyx_v_result[(6 * 4)];
  int *__pyx_v_aliased;
  PyObject *__pyx_v_packed = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  unsigned long __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_pack_stat", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":787
 *     cdef char result[6*4] # 6 long ints
 *     cdef int *aliased
 *     aliased = <int *>result             # <<<<<<<<<<<<<<
 *     aliased[0] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_size))
 *     # mtime and ctime will often be floats but get converted to PyInt within
*/
  __pyx_v_aliased = ((int *)__pyx_v_result);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":788
 *     cdef int *aliased
 *     aliased = <int *>result
 *     aliased[0] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_size))             # <<<<<<<<<<<<<<
 *     # mtime and ctime will often be floats but get converted to PyInt within
 *     aliased[1] = htonl(_time_to_unsigned(stat_value.st_mtime))
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_mstate_global->__pyx_n_u_st_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyLong_AsUnsignedLongMask(__pyx_t_1); if (unlikely(__pyx_t_2 == ((unsigned long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  (__pyx_v_aliased[0]) = htonl(__pyx_t_2);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":790
 *     aliased[0] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_size))
 *     # mtime and ctime will often be floats but get converted to PyInt within
 *     aliased[1] = htonl(_time_to_unsigned(stat_value.st_mtime))             # <<<<<<<<<<<<<<
 *     aliased[2] = htonl(_time_to_unsigned(stat_value.st_ctime))
 *     aliased[3] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_dev))
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_mstate_global->__pyx_n_u_st_mtime); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__time_to_unsigned(__pyx_t_1); if (unlikely(__pyx_t_2 == ((unsigned long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  (__pyx_v_aliased[1]) = htonl(__pyx_t_2);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":791
 *     # mtime and ctime will often be floats but get converted to PyInt within
 *     aliased[1] = htonl(_time_to_unsigned(stat_value.st_mtime))
 *     aliased[2] = htonl(_time_to_unsigned(stat_value.st_ctime))             # <<<<<<<<<<<<<<
 *     aliased[3] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_dev))
 *     aliased[4] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_ino))
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_mstate_global->__pyx_n_u_st_ctime); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__time_to_unsigned(__pyx_t_1); if (unlikely(__pyx_t_2 == ((unsigned long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 791, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  (__pyx_v_aliased[2]) = htonl(__pyx_t_2);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":792
 *     aliased[1] = htonl(_time_to_unsigned(stat_value.st_mtime))
 *     aliased[2] = htonl(_time_to_unsigned(stat_value.st_ctime))
 *     aliased[3] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_dev))             # <<<<<<<<<<<<<<
 *     aliased[4] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_ino))
 *     aliased[5] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_mode))
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_mstate_global->__pyx_n_u_st_dev); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyLong_AsUnsignedLongMask(__pyx_t_1); if (unlikely(__pyx_t_2 == ((unsigned long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  (__pyx_v_aliased[3]) = htonl(__pyx_t_2);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":793
 *     aliased[2] = htonl(_time_to_unsigned(stat_value.st_ctime))
 *     aliased[3] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_dev))
 *     aliased[4] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_ino))             # <<<<<<<<<<<<<<
 *     aliased[5] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_mode))
 *     packed = PyBytes_FromStringAndSize(result, 6*4)
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_mstate_global->__pyx_n_u_st_ino); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyLong_AsUnsignedLongMask(__pyx_t_1); if (unlikely(__pyx_t_2 == ((unsigned long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  (__pyx_v_aliased[4]) = htonl(__pyx_t_2);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":794
 *     aliased[3] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_dev))
 *     aliased[4] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_ino))
 *     aliased[5] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_mode))             # <<<<<<<<<<<<<<
 *     packed = PyBytes_FromStringAndSize(result, 6*4)
 *     return _encode(packed)[:-1]
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_mstate_global->__pyx_n_u_st_mode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyLong_AsUnsignedLongMask(__pyx_t_1); if (unlikely(__pyx_t_2 == ((unsigned long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  (__pyx_v_aliased[5]) = htonl(__pyx_t_2);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":795
 *     aliased[4] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_ino))
 *     aliased[5] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_mode))
 *     packed = PyBytes_FromStringAndSize(result, 6*4)             # <<<<<<<<<<<<<<
 *     return _encode(packed)[:-1]
 * 
*/
  __pyx_t_1 = PyBytes_FromStringAndSize(__pyx_v_result, 0x18); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":796
 *     aliased[5] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_mode))
 *     packed = PyBytes_FromStringAndSize(result, 6*4)
 *     return _encode(packed)[:-1]             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_packed};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, -1L, NULL, NULL, &__pyx_mstate_global->__pyx_slice[0], 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":779
 * 
 * 
 * cdef _pack_stat(stat_value):             # <<<<<<<<<<<<<<
 *     """return a string representing the stat value's key fields.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._pack_stat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":799
 * 
 * 
 * def pack_stat(stat_value):             # <<<<<<<<<<<<<<
 *     """Convert stat value into a packed representation quickly with pyrex"""
 *     return _pack_stat(stat_value)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_15pack_stat(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_14pack_stat, "Convert stat value into a packed representation quickly with pyrex");
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_15pack_stat = {"pack_stat", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_15pack_stat, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_14pack_stat};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_15pack_stat(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_stat_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pack_stat (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_stat_value,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 799, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "pack_stat", 0) < (0)) __PYX_ERR(0, 799, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("pack_stat", 1, 1, 1, i); __PYX_ERR(0, 799, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 799, __pyx_L3_error)
    }
    __pyx_v_stat_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pack_stat", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 799, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.pack_stat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_14pack_stat(__pyx_self, __pyx_v_stat_value);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_14pack_stat(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_stat_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pack_stat", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":801
 * def pack_stat(stat_value):
 *     """Convert stat value into a packed representation quickly with pyrex"""
 *     return _pack_stat(stat_value)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__pack_stat(__pyx_v_stat_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":799
 * 
 * 
 * def pack_stat(stat_value):             # <<<<<<<<<<<<<<
 *     """Convert stat value into a packed representation quickly with pyrex"""
 *     return _pack_stat(stat_value)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.pack_stat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":804
 * 
 * 
 * cpdef update_entry(self, entry, abspath, stat_value):             # <<<<<<<<<<<<<<
 *     """Update the entry based on what is actually on disk.
 * 
*/

static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_17update_entry(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_update_entry(PyObject *__pyx_v_self, PyObject *__pyx_v_entry, PyObject *__pyx_v_abspath, PyObject *__pyx_v_stat_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_minikind;
  int __pyx_v_saved_minikind;
  void *__pyx_v_details;
  int __pyx_v_worth_saving;
  PyObject *__pyx_v_packed_stat = NULL;
  PyObject *__pyx_v_saved_link_or_sha1 = NULL;
  PyObject *__pyx_v_saved_file_size = NULL;
  PyObject *__pyx_v_saved_executable = NULL;
  PyObject *__pyx_v_saved_packed_stat = NULL;
  PyObject *__pyx_v_link_or_sha1 = NULL;
  PyObject *__pyx_v_executable = NULL;
  PyObject *__pyx_v_block_index = NULL;
  PyObject *__pyx_v_entry_index = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_dir_present = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_file_present = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update_entry", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":823
 *     cdef void * details
 *     cdef int worth_saving
 *     minikind = minikind_from_mode(stat_value.st_mode)             # <<<<<<<<<<<<<<
 *     if 0 == minikind:
 *         return None
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_mstate_global->__pyx_n_u_st_mode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_minikind_from_mode(__pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 823, __pyx_L1_error)
  __pyx_v_minikind = __pyx_t_3;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":824
 *     cdef int worth_saving
 *     minikind = minikind_from_mode(stat_value.st_mode)
 *     if 0 == minikind:             # <<<<<<<<<<<<<<
 *         return None
 *     packed_stat = _pack_stat(stat_value)
*/
  __pyx_t_4 = (0 == __pyx_v_minikind);
  if (__pyx_t_4) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":825
 *     minikind = minikind_from_mode(stat_value.st_mode)
 *     if 0 == minikind:
 *         return None             # <<<<<<<<<<<<<<
 *     packed_stat = _pack_stat(stat_value)
 *     details = PyList_GetItem_void_void(PyTuple_GetItem_void_void(<void *>entry, 1), 0)
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":824
 *     cdef int worth_saving
 *     minikind = minikind_from_mode(stat_value.st_mode)
 *     if 0 == minikind:             # <<<<<<<<<<<<<<
 *         return None
 *     packed_stat = _pack_stat(stat_value)
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":826
 *     if 0 == minikind:
 *         return None
 *     packed_stat = _pack_stat(stat_value)             # <<<<<<<<<<<<<<
 *     details = PyList_GetItem_void_void(PyTuple_GetItem_void_void(<void *>entry, 1), 0)
 *     saved_minikind = PyBytes_AsString_obj(<PyObject *>PyTuple_GetItem_void_void(details, 0))[0]
*/
  __pyx_t_1 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__pack_stat(__pyx_v_stat_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed_stat = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":827
 *         return None
 *     packed_stat = _pack_stat(stat_value)
 *     details = PyList_GetItem_void_void(PyTuple_GetItem_void_void(<void *>entry, 1), 0)             # <<<<<<<<<<<<<<
 *     saved_minikind = PyBytes_AsString_obj(<PyObject *>PyTuple_GetItem_void_void(details, 0))[0]
 *     if minikind == b'd' and saved_minikind == b't':
*/
  __pyx_v_details = PyList_GET_ITEM(PyTuple_GET_ITEM(((void *)__pyx_v_entry), 1), 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":828
 *     packed_stat = _pack_stat(stat_value)
 *     details = PyList_GetItem_void_void(PyTuple_GetItem_void_void(<void *>entry, 1), 0)
 *     saved_minikind = PyBytes_AsString_obj(<PyObject *>PyTuple_GetItem_void_void(details, 0))[0]             # <<<<<<<<<<<<<<
 *     if minikind == b'd' and saved_minikind == b't':
 *         minikind = b't'
*/
  __pyx_v_saved_minikind = (PyBytes_AsString(((PyObject *)PyTuple_GET_ITEM(__pyx_v_details, 0)))[0]);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":829
 *     details = PyList_GetItem_void_void(PyTuple_GetItem_void_void(<void *>entry, 1), 0)
 *     saved_minikind = PyBytes_AsString_obj(<PyObject *>PyTuple_GetItem_void_void(details, 0))[0]
 *     if minikind == b'd' and saved_minikind == b't':             # <<<<<<<<<<<<<<
 *         minikind = b't'
 *     saved_link_or_sha1 = PyTuple_GetItem_void_object(details, 1)
*/
  __pyx_t_5 = (__pyx_v_minikind == 'd');
  if (__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_5 = (__pyx_v_saved_minikind == 't');
  __pyx_t_4 = __pyx_t_5;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_4) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":830
 *     saved_minikind = PyBytes_AsString_obj(<PyObject *>PyTuple_GetItem_void_void(details, 0))[0]
 *     if minikind == b'd' and saved_minikind == b't':
 *         minikind = b't'             # <<<<<<<<<<<<<<
 *     saved_link_or_sha1 = PyTuple_GetItem_void_object(details, 1)
 *     saved_file_size = PyTuple_GetItem_void_object(details, 2)
*/
    __pyx_v_minikind = 't';

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":829
 *     details = PyList_GetItem_void_void(PyTuple_GetItem_void_void(<void *>entry, 1), 0)
 *     saved_minikind = PyBytes_AsString_obj(<PyObject *>PyTuple_GetItem_void_void(details, 0))[0]
 *     if minikind == b'd' and saved_minikind == b't':             # <<<<<<<<<<<<<<
 *         minikind = b't'
 *     saved_link_or_sha1 = PyTuple_GetItem_void_object(details, 1)
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":831
 *     if minikind == b'd' and saved_minikind == b't':
 *         minikind = b't'
 *     saved_link_or_sha1 = PyTuple_GetItem_void_object(details, 1)             # <<<<<<<<<<<<<<
 *     saved_file_size = PyTuple_GetItem_void_object(details, 2)
 *     saved_executable = PyTuple_GetItem_void_object(details, 3)
*/
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v_details, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_saved_link_or_sha1 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":832
 *         minikind = b't'
 *     saved_link_or_sha1 = PyTuple_GetItem_void_object(details, 1)
 *     saved_file_size = PyTuple_GetItem_void_object(details, 2)             # <<<<<<<<<<<<<<
 *     saved_executable = PyTuple_GetItem_void_object(details, 3)
 *     saved_packed_stat = PyTuple_GetItem_void_object(details, 4)
*/
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v_details, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_saved_file_size = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":833
 *     saved_link_or_sha1 = PyTuple_GetItem_void_object(details, 1)
 *     saved_file_size = PyTuple_GetItem_void_object(details, 2)
 *     saved_executable = PyTuple_GetItem_void_object(details, 3)             # <<<<<<<<<<<<<<
 *     saved_packed_stat = PyTuple_GetItem_void_object(details, 4)
 *     # Deal with pyrex decrefing the objects
*/
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v_details, 3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_saved_executable = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":834
 *     saved_file_size = PyTuple_GetItem_void_object(details, 2)
 *     saved_executable = PyTuple_GetItem_void_object(details, 3)
 *     saved_packed_stat = PyTuple_GetItem_void_object(details, 4)             # <<<<<<<<<<<<<<
 *     # Deal with pyrex decrefing the objects
 *     Py_INCREF(saved_link_or_sha1)
*/
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v_details, 4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_saved_packed_stat = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":836
 *     saved_packed_stat = PyTuple_GetItem_void_object(details, 4)
 *     # Deal with pyrex decrefing the objects
 *     Py_INCREF(saved_link_or_sha1)             # <<<<<<<<<<<<<<
 *     Py_INCREF(saved_file_size)
 *     Py_INCREF(saved_executable)
*/
  Py_INCREF(__pyx_v_saved_link_or_sha1);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":837
 *     # Deal with pyrex decrefing the objects
 *     Py_INCREF(saved_link_or_sha1)
 *     Py_INCREF(saved_file_size)             # <<<<<<<<<<<<<<
 *     Py_INCREF(saved_executable)
 *     Py_INCREF(saved_packed_stat)
*/
  Py_INCREF(__pyx_v_saved_file_size);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":838
 *     Py_INCREF(saved_link_or_sha1)
 *     Py_INCREF(saved_file_size)
 *     Py_INCREF(saved_executable)             # <<<<<<<<<<<<<<
 *     Py_INCREF(saved_packed_stat)
 *     #(saved_minikind, saved_link_or_sha1, saved_file_size,
*/
  Py_INCREF(__pyx_v_saved_executable);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":839
 *     Py_INCREF(saved_file_size)
 *     Py_INCREF(saved_executable)
 *     Py_INCREF(saved_packed_stat)             # <<<<<<<<<<<<<<
 *     #(saved_minikind, saved_link_or_sha1, saved_file_size,
 *     # saved_executable, saved_packed_stat) = entry[1][0]
*/
  Py_INCREF(__pyx_v_saved_packed_stat);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":843
 *     # saved_executable, saved_packed_stat) = entry[1][0]
 * 
 *     if (minikind == saved_minikind             # <<<<<<<<<<<<<<
 *         and packed_stat == saved_packed_stat):
 *         # The stat hasn't changed since we saved, so we can re-use the
*/
  __pyx_t_5 = (__pyx_v_minikind == __pyx_v_saved_minikind);
  if (__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L8_bool_binop_done;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":844
 * 
 *     if (minikind == saved_minikind
 *         and packed_stat == saved_packed_stat):             # <<<<<<<<<<<<<<
 *         # The stat hasn't changed since we saved, so we can re-use the
 *         # saved sha hash.
*/
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_packed_stat, __pyx_v_saved_packed_stat, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 844, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __pyx_t_5;
  __pyx_L8_bool_binop_done:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":843
 *     # saved_executable, saved_packed_stat) = entry[1][0]
 * 
 *     if (minikind == saved_minikind             # <<<<<<<<<<<<<<
 *         and packed_stat == saved_packed_stat):
 *         # The stat hasn't changed since we saved, so we can re-use the
*/
  if (__pyx_t_4) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":847
 *         # The stat hasn't changed since we saved, so we can re-use the
 *         # saved sha hash.
 *         if minikind == b'd':             # <<<<<<<<<<<<<<
 *             return None
 * 
*/
    __pyx_t_4 = (__pyx_v_minikind == 'd');
    if (__pyx_t_4) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":848
 *         # saved sha hash.
 *         if minikind == b'd':
 *             return None             # <<<<<<<<<<<<<<
 * 
 *         # size should also be in packed_stat
*/
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":847
 *         # The stat hasn't changed since we saved, so we can re-use the
 *         # saved sha hash.
 *         if minikind == b'd':             # <<<<<<<<<<<<<<
 *             return None
 * 
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":851
 * 
 *         # size should also be in packed_stat
 *         if saved_file_size == stat_value.st_size:             # <<<<<<<<<<<<<<
 *             return saved_link_or_sha1
 * 
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_mstate_global->__pyx_n_u_st_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyObject_RichCompare(__pyx_v_saved_file_size, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 851, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 851, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_4) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":852
 *         # size should also be in packed_stat
 *         if saved_file_size == stat_value.st_size:
 *             return saved_link_or_sha1             # <<<<<<<<<<<<<<
 * 
 *     # If we have gotten this far, that means that we need to actually
*/
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_saved_link_or_sha1);
      __pyx_r = __pyx_v_saved_link_or_sha1;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":851
 * 
 *         # size should also be in packed_stat
 *         if saved_file_size == stat_value.st_size:             # <<<<<<<<<<<<<<
 *             return saved_link_or_sha1
 * 
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":843
 *     # saved_executable, saved_packed_stat) = entry[1][0]
 * 
 *     if (minikind == saved_minikind             # <<<<<<<<<<<<<<
 *         and packed_stat == saved_packed_stat):
 *         # The stat hasn't changed since we saved, so we can re-use the
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":856
 *     # If we have gotten this far, that means that we need to actually
 *     # process this entry.
 *     link_or_sha1 = None             # <<<<<<<<<<<<<<
 *     worth_saving = 1
 *     if minikind == b'f':
*/
  __Pyx_INCREF(Py_None);
  __pyx_v_link_or_sha1 = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":857
 *     # process this entry.
 *     link_or_sha1 = None
 *     worth_saving = 1             # <<<<<<<<<<<<<<
 *     if minikind == b'f':
 *         executable = self._is_executable(stat_value.st_mode,
*/
  __pyx_v_worth_saving = 1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":858
 *     link_or_sha1 = None
 *     worth_saving = 1
 *     if minikind == b'f':             # <<<<<<<<<<<<<<
 *         executable = self._is_executable(stat_value.st_mode,
 *                                          saved_executable)
*/
  switch (__pyx_v_minikind) {
    case 'f':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":859
 *     worth_saving = 1
 *     if minikind == b'f':
 *         executable = self._is_executable(stat_value.st_mode,             # <<<<<<<<<<<<<<
 *                                          saved_executable)
 *         if self._cutoff_time is None:
*/
    __pyx_t_1 = __pyx_v_self;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_mstate_global->__pyx_n_u_st_mode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":860
 *     if minikind == b'f':
 *         executable = self._is_executable(stat_value.st_mode,
 *                                          saved_executable)             # <<<<<<<<<<<<<<
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()
*/
    __pyx_t_8 = 0;
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_t_7, __pyx_v_saved_executable};
      __pyx_t_6 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_is_executable, __pyx_callargs+__pyx_t_8, (3-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 859, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
    }
    __pyx_v_executable = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":861
 *         executable = self._is_executable(stat_value.st_mode,
 *                                          saved_executable)
 *         if self._cutoff_time is None:             # <<<<<<<<<<<<<<
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time
*/
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_cutoff_time); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 861, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = (__pyx_t_6 == Py_None);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_4) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":862
 *                                          saved_executable)
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()             # <<<<<<<<<<<<<<
 *         if (stat_value.st_mtime < self._cutoff_time
 *             and stat_value.st_ctime < self._cutoff_time
*/
      __pyx_t_7 = __pyx_v_self;
      __Pyx_INCREF(__pyx_t_7);
      __pyx_t_8 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
        __pyx_t_6 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_sha_cutoff_time, __pyx_callargs+__pyx_t_8, (1-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 862, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":861
 *         executable = self._is_executable(stat_value.st_mode,
 *                                          saved_executable)
 *         if self._cutoff_time is None:             # <<<<<<<<<<<<<<
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":863
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time             # <<<<<<<<<<<<<<
 *             and stat_value.st_ctime < self._cutoff_time
 *             and len(entry[1]) > 1
*/
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_mstate_global->__pyx_n_u_st_mtime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_cutoff_time); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_6, __pyx_t_7, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 863, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 863, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L14_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":864
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time
 *             and stat_value.st_ctime < self._cutoff_time             # <<<<<<<<<<<<<<
 *             and len(entry[1]) > 1
 *             and entry[1][1][0] != b'a'):
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_mstate_global->__pyx_n_u_st_ctime); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_cutoff_time); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_LT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 864, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 864, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L14_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":865
 *         if (stat_value.st_mtime < self._cutoff_time
 *             and stat_value.st_ctime < self._cutoff_time
 *             and len(entry[1]) > 1             # <<<<<<<<<<<<<<
 *             and entry[1][1][0] != b'a'):
 *                 # Could check for size changes for further optimised
*/
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 865, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_9 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 865, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__pyx_t_9 > 1);
    if (__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L14_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":866
 *             and stat_value.st_ctime < self._cutoff_time
 *             and len(entry[1]) > 1
 *             and entry[1][1][0] != b'a'):             # <<<<<<<<<<<<<<
 *                 # Could check for size changes for further optimised
 *                 # avoidance of sha1's. However the most prominent case of
*/
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 866, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_6, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 866, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 866, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_5 = (__Pyx_PyBytes_Equals(__pyx_t_6, __pyx_mstate_global->__pyx_n_b_a, Py_NE)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 866, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_4 = __pyx_t_5;
    __pyx_L14_bool_binop_done:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":863
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time             # <<<<<<<<<<<<<<
 *             and stat_value.st_ctime < self._cutoff_time
 *             and len(entry[1]) > 1
*/
    if (__pyx_t_4) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":870
 *                 # avoidance of sha1's. However the most prominent case of
 *                 # over-shaing is during initial add, which this catches.
 *             link_or_sha1 = self._sha1_file(abspath)             # <<<<<<<<<<<<<<
 *             entry[1][0] = (b'f', link_or_sha1, stat_value.st_size,
 *                            executable, packed_stat)
*/
      __pyx_t_7 = __pyx_v_self;
      __Pyx_INCREF(__pyx_t_7);
      __pyx_t_8 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_abspath};
        __pyx_t_6 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_sha1_file, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 870, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_DECREF_SET(__pyx_v_link_or_sha1, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":871
 *                 # over-shaing is during initial add, which this catches.
 *             link_or_sha1 = self._sha1_file(abspath)
 *             entry[1][0] = (b'f', link_or_sha1, stat_value.st_size,             # <<<<<<<<<<<<<<
 *                            executable, packed_stat)
 *         else:
*/
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_mstate_global->__pyx_n_u_st_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 871, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":872
 *             link_or_sha1 = self._sha1_file(abspath)
 *             entry[1][0] = (b'f', link_or_sha1, stat_value.st_size,
 *                            executable, packed_stat)             # <<<<<<<<<<<<<<
 *         else:
 *             # This file is not worth caching the sha1. Either it is too new, or
*/
      __pyx_t_7 = PyTuple_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 871, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_mstate_global->__pyx_n_b_f);
      __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_b_f);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_mstate_global->__pyx_n_b_f) != (0)) __PYX_ERR(0, 871, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_link_or_sha1);
      __Pyx_GIVEREF(__pyx_v_link_or_sha1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_link_or_sha1) != (0)) __PYX_ERR(0, 871, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_6);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_6) != (0)) __PYX_ERR(0, 871, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_executable);
      __Pyx_GIVEREF(__pyx_v_executable);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_v_executable) != (0)) __PYX_ERR(0, 871, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_packed_stat);
      __Pyx_GIVEREF(__pyx_v_packed_stat);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_v_packed_stat) != (0)) __PYX_ERR(0, 871, __pyx_L1_error);
      __pyx_t_6 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":871
 *                 # over-shaing is during initial add, which this catches.
 *             link_or_sha1 = self._sha1_file(abspath)
 *             entry[1][0] = (b'f', link_or_sha1, stat_value.st_size,             # <<<<<<<<<<<<<<
 *                            executable, packed_stat)
 *         else:
*/
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 871, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely((__Pyx_SetItemInt(__pyx_t_6, 0, __pyx_t_7, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference) < 0))) __PYX_ERR(0, 871, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":863
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time             # <<<<<<<<<<<<<<
 *             and stat_value.st_ctime < self._cutoff_time
 *             and len(entry[1]) > 1
*/
      goto __pyx_L13;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":879
 *             # *not* set the IN_MEMORY_MODIFIED flag. (But we'll save the
 *             # updated values if there is *other* data worth saving.)
 *             entry[1][0] = (b'f', b'', stat_value.st_size, executable,             # <<<<<<<<<<<<<<
 *                            DirState.NULLSTAT)
 *             worth_saving = 0
*/
    /*else*/ {
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_mstate_global->__pyx_n_u_st_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 879, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":880
 *             # updated values if there is *other* data worth saving.)
 *             entry[1][0] = (b'f', b'', stat_value.st_size, executable,
 *                            DirState.NULLSTAT)             # <<<<<<<<<<<<<<
 *             worth_saving = 0
 *     elif minikind == b'd':
*/
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_DirState); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 880, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_NULLSTAT); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 880, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":879
 *             # *not* set the IN_MEMORY_MODIFIED flag. (But we'll save the
 *             # updated values if there is *other* data worth saving.)
 *             entry[1][0] = (b'f', b'', stat_value.st_size, executable,             # <<<<<<<<<<<<<<
 *                            DirState.NULLSTAT)
 *             worth_saving = 0
*/
      __pyx_t_6 = PyTuple_New(5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 879, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_mstate_global->__pyx_n_b_f);
      __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_b_f);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_mstate_global->__pyx_n_b_f) != (0)) __PYX_ERR(0, 879, __pyx_L1_error);
      __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b__4);
      __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_b__4);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_mstate_global->__pyx_kp_b__4) != (0)) __PYX_ERR(0, 879, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_7);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_7) != (0)) __PYX_ERR(0, 879, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_executable);
      __Pyx_GIVEREF(__pyx_v_executable);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_v_executable) != (0)) __PYX_ERR(0, 879, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_t_1) != (0)) __PYX_ERR(0, 879, __pyx_L1_error);
      __pyx_t_7 = 0;
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 879, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely((__Pyx_SetItemInt(__pyx_t_1, 0, __pyx_t_6, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference) < 0))) __PYX_ERR(0, 879, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":881
 *             entry[1][0] = (b'f', b'', stat_value.st_size, executable,
 *                            DirState.NULLSTAT)
 *             worth_saving = 0             # <<<<<<<<<<<<<<
 *     elif minikind == b'd':
 *         entry[1][0] = (b'd', b'', 0, False, packed_stat)
*/
      __pyx_v_worth_saving = 0;
    }
    __pyx_L13:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":858
 *     link_or_sha1 = None
 *     worth_saving = 1
 *     if minikind == b'f':             # <<<<<<<<<<<<<<
 *         executable = self._is_executable(stat_value.st_mode,
 *                                          saved_executable)
*/
    break;
    case 'd':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":883
 *             worth_saving = 0
 *     elif minikind == b'd':
 *         entry[1][0] = (b'd', b'', 0, False, packed_stat)             # <<<<<<<<<<<<<<
 *         if saved_minikind != b'd':
 *             # This changed from something into a directory. Make sure we
*/
    __pyx_t_6 = PyTuple_New(5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_n_b_d);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_b_d);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_mstate_global->__pyx_n_b_d) != (0)) __PYX_ERR(0, 883, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b__4);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_b__4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_mstate_global->__pyx_kp_b__4) != (0)) __PYX_ERR(0, 883, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_0);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_0);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_mstate_global->__pyx_int_0) != (0)) __PYX_ERR(0, 883, __pyx_L1_error);
    __Pyx_INCREF(Py_False);
    __Pyx_GIVEREF(Py_False);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 3, Py_False) != (0)) __PYX_ERR(0, 883, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_packed_stat);
    __Pyx_GIVEREF(__pyx_v_packed_stat);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_v_packed_stat) != (0)) __PYX_ERR(0, 883, __pyx_L1_error);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely((__Pyx_SetItemInt(__pyx_t_1, 0, __pyx_t_6, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference) < 0))) __PYX_ERR(0, 883, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":884
 *     elif minikind == b'd':
 *         entry[1][0] = (b'd', b'', 0, False, packed_stat)
 *         if saved_minikind != b'd':             # <<<<<<<<<<<<<<
 *             # This changed from something into a directory. Make sure we
 *             # have a directory block for it. This doesn't happen very
*/
    __pyx_t_4 = (__pyx_v_saved_minikind != 'd');
    if (__pyx_t_4) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":889
 *             # often, so this doesn't have to be super fast.
 *             block_index, entry_index, dir_present, file_present = \
 *                 self._get_block_entry_index(entry[0][0], entry[0][1], 0)             # <<<<<<<<<<<<<<
 *             self._ensure_block(block_index, entry_index,
 *                                pathjoin(entry[0][0], entry[0][1]))
*/
      __pyx_t_1 = __pyx_v_self;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_10 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_11 = __Pyx_GetItemInt(__pyx_t_7, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = 0;
      {
        PyObject *__pyx_callargs[4] = {__pyx_t_1, __pyx_t_10, __pyx_t_11, __pyx_mstate_global->__pyx_int_0};
        __pyx_t_6 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_block_entry_index, __pyx_callargs+__pyx_t_8, (4-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 889, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
        PyObject* sequence = __pyx_t_6;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 4)) {
          if (size > 4) __Pyx_RaiseTooManyValuesError(4);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 888, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_11 = PyTuple_GET_ITEM(sequence, 0);
          __Pyx_INCREF(__pyx_t_11);
          __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1);
          __Pyx_INCREF(__pyx_t_10);
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 2);
          __Pyx_INCREF(__pyx_t_1);
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 3);
          __Pyx_INCREF(__pyx_t_7);
        } else {
          __pyx_t_11 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
          if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 888, __pyx_L1_error)
          __Pyx_XGOTREF(__pyx_t_11);
          __pyx_t_10 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
          if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 888, __pyx_L1_error)
          __Pyx_XGOTREF(__pyx_t_10);
          __pyx_t_1 = __Pyx_PyList_GetItemRefFast(sequence, 2, __Pyx_ReferenceSharing_SharedReference);
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 888, __pyx_L1_error)
          __Pyx_XGOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyList_GetItemRefFast(sequence, 3, __Pyx_ReferenceSharing_SharedReference);
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 888, __pyx_L1_error)
          __Pyx_XGOTREF(__pyx_t_7);
        }
        #else
        {
          Py_ssize_t i;
          PyObject** temps[4] = {&__pyx_t_11,&__pyx_t_10,&__pyx_t_1,&__pyx_t_7};
          for (i=0; i < 4; i++) {
            PyObject* item = __Pyx_PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 888, __pyx_L1_error)
            __Pyx_GOTREF(item);
            *(temps[i]) = item;
          }
        }
        #endif
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else {
        Py_ssize_t index = -1;
        PyObject** temps[4] = {&__pyx_t_11,&__pyx_t_10,&__pyx_t_1,&__pyx_t_7};
        __pyx_t_12 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 888, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_13 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_12);
        for (index=0; index < 4; index++) {
          PyObject* item = __pyx_t_13(__pyx_t_12); if (unlikely(!item)) goto __pyx_L19_unpacking_failed;
          __Pyx_GOTREF(item);
          *(temps[index]) = item;
        }
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_12), 4) < (0)) __PYX_ERR(0, 888, __pyx_L1_error)
        __pyx_t_13 = NULL;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        goto __pyx_L20_unpacking_done;
        __pyx_L19_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_13 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 888, __pyx_L1_error)
        __pyx_L20_unpacking_done:;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":888
 *             # have a directory block for it. This doesn't happen very
 *             # often, so this doesn't have to be super fast.
 *             block_index, entry_index, dir_present, file_present = \             # <<<<<<<<<<<<<<
 *                 self._get_block_entry_index(entry[0][0], entry[0][1], 0)
 *             self._ensure_block(block_index, entry_index,
*/
      __pyx_v_block_index = __pyx_t_11;
      __pyx_t_11 = 0;
      __pyx_v_entry_index = __pyx_t_10;
      __pyx_t_10 = 0;
      __pyx_v_dir_present = __pyx_t_1;
      __pyx_t_1 = 0;
      __pyx_v_file_present = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":890
 *             block_index, entry_index, dir_present, file_present = \
 *                 self._get_block_entry_index(entry[0][0], entry[0][1], 0)
 *             self._ensure_block(block_index, entry_index,             # <<<<<<<<<<<<<<
 *                                pathjoin(entry[0][0], entry[0][1]))
 *         else:
*/
      __pyx_t_7 = __pyx_v_self;
      __Pyx_INCREF(__pyx_t_7);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":891
 *                 self._get_block_entry_index(entry[0][0], entry[0][1], 0)
 *             self._ensure_block(block_index, entry_index,
 *                                pathjoin(entry[0][0], entry[0][1]))             # <<<<<<<<<<<<<<
 *         else:
 *             # Any changes are derived trivially from the stat object, not worth
*/
      __pyx_t_10 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_pathjoin); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 891, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 891, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_14 = __Pyx_GetItemInt(__pyx_t_12, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 891, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_12 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 891, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_15 = __Pyx_GetItemInt(__pyx_t_12, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 891, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_8 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
        assert(__pyx_t_10);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_11, __pyx__function);
        __pyx_t_8 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_10, __pyx_t_14, __pyx_t_15};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_11, __pyx_callargs+__pyx_t_8, (3-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 891, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __pyx_t_8 = 0;
      {
        PyObject *__pyx_callargs[4] = {__pyx_t_7, __pyx_v_block_index, __pyx_v_entry_index, __pyx_t_1};
        __pyx_t_6 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_ensure_block, __pyx_callargs+__pyx_t_8, (4-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 890, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":884
 *     elif minikind == b'd':
 *         entry[1][0] = (b'd', b'', 0, False, packed_stat)
 *         if saved_minikind != b'd':             # <<<<<<<<<<<<<<
 *             # This changed from something into a directory. Make sure we
 *             # have a directory block for it. This doesn't happen very
*/
      goto __pyx_L18;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":895
 *             # Any changes are derived trivially from the stat object, not worth
 *             # re-writing a dirstate for just this
 *             worth_saving = 0             # <<<<<<<<<<<<<<
 *     elif minikind == b'l':
 *         if saved_minikind == b'l':
*/
    /*else*/ {
      __pyx_v_worth_saving = 0;
    }
    __pyx_L18:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":882
 *                            DirState.NULLSTAT)
 *             worth_saving = 0
 *     elif minikind == b'd':             # <<<<<<<<<<<<<<
 *         entry[1][0] = (b'd', b'', 0, False, packed_stat)
 *         if saved_minikind != b'd':
*/
    break;
    case 'l':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":897
 *             worth_saving = 0
 *     elif minikind == b'l':
 *         if saved_minikind == b'l':             # <<<<<<<<<<<<<<
 *             # If the object hasn't changed kind, it isn't worth saving the
 *             # dirstate just for a symlink. The default is 'fast symlinks' which
*/
    __pyx_t_4 = (__pyx_v_saved_minikind == 'l');
    if (__pyx_t_4) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":902
 *             # save the target in the inode entry, rather than separately. So to
 *             # stat, we've already read everything off disk.
 *             worth_saving = 0             # <<<<<<<<<<<<<<
 *         link_or_sha1 = self._read_link(abspath, saved_link_or_sha1)
 *         if self._cutoff_time is None:
*/
      __pyx_v_worth_saving = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":897
 *             worth_saving = 0
 *     elif minikind == b'l':
 *         if saved_minikind == b'l':             # <<<<<<<<<<<<<<
 *             # If the object hasn't changed kind, it isn't worth saving the
 *             # dirstate just for a symlink. The default is 'fast symlinks' which
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":903
 *             # stat, we've already read everything off disk.
 *             worth_saving = 0
 *         link_or_sha1 = self._read_link(abspath, saved_link_or_sha1)             # <<<<<<<<<<<<<<
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()
*/
    __pyx_t_1 = __pyx_v_self;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_8 = 0;
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_v_abspath, __pyx_v_saved_link_or_sha1};
      __pyx_t_6 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_read_link, __pyx_callargs+__pyx_t_8, (3-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 903, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
    }
    __Pyx_DECREF_SET(__pyx_v_link_or_sha1, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":904
 *             worth_saving = 0
 *         link_or_sha1 = self._read_link(abspath, saved_link_or_sha1)
 *         if self._cutoff_time is None:             # <<<<<<<<<<<<<<
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time
*/
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_cutoff_time); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 904, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = (__pyx_t_6 == Py_None);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_4) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":905
 *         link_or_sha1 = self._read_link(abspath, saved_link_or_sha1)
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()             # <<<<<<<<<<<<<<
 *         if (stat_value.st_mtime < self._cutoff_time
 *             and stat_value.st_ctime < self._cutoff_time):
*/
      __pyx_t_1 = __pyx_v_self;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_8 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
        __pyx_t_6 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_sha_cutoff_time, __pyx_callargs+__pyx_t_8, (1-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 905, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":904
 *             worth_saving = 0
 *         link_or_sha1 = self._read_link(abspath, saved_link_or_sha1)
 *         if self._cutoff_time is None:             # <<<<<<<<<<<<<<
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":906
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time             # <<<<<<<<<<<<<<
 *             and stat_value.st_ctime < self._cutoff_time):
 *             entry[1][0] = (b'l', link_or_sha1, stat_value.st_size,
*/
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_mstate_global->__pyx_n_u_st_mtime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 906, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_cutoff_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 906, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = PyObject_RichCompare(__pyx_t_6, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 906, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 906, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L24_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":907
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time
 *             and stat_value.st_ctime < self._cutoff_time):             # <<<<<<<<<<<<<<
 *             entry[1][0] = (b'l', link_or_sha1, stat_value.st_size,
 *                            False, packed_stat)
*/
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_mstate_global->__pyx_n_u_st_ctime); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 907, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_cutoff_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 907, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_7, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 907, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 907, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_4 = __pyx_t_5;
    __pyx_L24_bool_binop_done:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":906
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time             # <<<<<<<<<<<<<<
 *             and stat_value.st_ctime < self._cutoff_time):
 *             entry[1][0] = (b'l', link_or_sha1, stat_value.st_size,
*/
    if (__pyx_t_4) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":908
 *         if (stat_value.st_mtime < self._cutoff_time
 *             and stat_value.st_ctime < self._cutoff_time):
 *             entry[1][0] = (b'l', link_or_sha1, stat_value.st_size,             # <<<<<<<<<<<<<<
 *                            False, packed_stat)
 *         else:
*/
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_mstate_global->__pyx_n_u_st_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 908, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":909
 *             and stat_value.st_ctime < self._cutoff_time):
 *             entry[1][0] = (b'l', link_or_sha1, stat_value.st_size,
 *                            False, packed_stat)             # <<<<<<<<<<<<<<
 *         else:
 *             entry[1][0] = (b'l', b'', stat_value.st_size,
*/
      __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 908, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_mstate_global->__pyx_n_b_l);
      __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_b_l);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_mstate_global->__pyx_n_b_l) != (0)) __PYX_ERR(0, 908, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_link_or_sha1);
      __Pyx_GIVEREF(__pyx_v_link_or_sha1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_link_or_sha1) != (0)) __PYX_ERR(0, 908, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_6);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_6) != (0)) __PYX_ERR(0, 908, __pyx_L1_error);
      __Pyx_INCREF(Py_False);
      __Pyx_GIVEREF(Py_False);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, Py_False) != (0)) __PYX_ERR(0, 908, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_packed_stat);
      __Pyx_GIVEREF(__pyx_v_packed_stat);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_v_packed_stat) != (0)) __PYX_ERR(0, 908, __pyx_L1_error);
      __pyx_t_6 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":908
 *         if (stat_value.st_mtime < self._cutoff_time
 *             and stat_value.st_ctime < self._cutoff_time):
 *             entry[1][0] = (b'l', link_or_sha1, stat_value.st_size,             # <<<<<<<<<<<<<<
 *                            False, packed_stat)
 *         else:
*/
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 908, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely((__Pyx_SetItemInt(__pyx_t_6, 0, __pyx_t_1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference) < 0))) __PYX_ERR(0, 908, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":906
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time             # <<<<<<<<<<<<<<
 *             and stat_value.st_ctime < self._cutoff_time):
 *             entry[1][0] = (b'l', link_or_sha1, stat_value.st_size,
*/
      goto __pyx_L23;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":911
 *                            False, packed_stat)
 *         else:
 *             entry[1][0] = (b'l', b'', stat_value.st_size,             # <<<<<<<<<<<<<<
 *                            False, DirState.NULLSTAT)
 *     if worth_saving:
*/
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_mstate_global->__pyx_n_u_st_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 911, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":912
 *         else:
 *             entry[1][0] = (b'l', b'', stat_value.st_size,
 *                            False, DirState.NULLSTAT)             # <<<<<<<<<<<<<<
 *     if worth_saving:
 *         # Note, even though _mark_modified will only set
*/
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_DirState); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 912, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_NULLSTAT); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 912, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":911
 *                            False, packed_stat)
 *         else:
 *             entry[1][0] = (b'l', b'', stat_value.st_size,             # <<<<<<<<<<<<<<
 *                            False, DirState.NULLSTAT)
 *     if worth_saving:
*/
      __pyx_t_6 = PyTuple_New(5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 911, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_mstate_global->__pyx_n_b_l);
      __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_b_l);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_mstate_global->__pyx_n_b_l) != (0)) __PYX_ERR(0, 911, __pyx_L1_error);
      __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b__4);
      __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_b__4);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_mstate_global->__pyx_kp_b__4) != (0)) __PYX_ERR(0, 911, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1) != (0)) __PYX_ERR(0, 911, __pyx_L1_error);
      __Pyx_INCREF(Py_False);
      __Pyx_GIVEREF(Py_False);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 3, Py_False) != (0)) __PYX_ERR(0, 911, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_7);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_t_7) != (0)) __PYX_ERR(0, 911, __pyx_L1_error);
      __pyx_t_1 = 0;
      __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 911, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely((__Pyx_SetItemInt(__pyx_t_7, 0, __pyx_t_6, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference) < 0))) __PYX_ERR(0, 911, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __pyx_L23:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":896
 *             # re-writing a dirstate for just this
 *             worth_saving = 0
 *     elif minikind == b'l':             # <<<<<<<<<<<<<<
 *         if saved_minikind == b'l':
 *             # If the object hasn't changed kind, it isn't worth saving the
*/
    break;
    default: break;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":913
 *             entry[1][0] = (b'l', b'', stat_value.st_size,
 *                            False, DirState.NULLSTAT)
 *     if worth_saving:             # <<<<<<<<<<<<<<
 *         # Note, even though _mark_modified will only set
 *         # IN_MEMORY_HASH_MODIFIED, it still isn't worth
*/
  __pyx_t_4 = (__pyx_v_worth_saving != 0);
  if (__pyx_t_4) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":916
 *         # Note, even though _mark_modified will only set
 *         # IN_MEMORY_HASH_MODIFIED, it still isn't worth
 *         self._mark_modified([entry])             # <<<<<<<<<<<<<<
 *     return link_or_sha1
 * 
*/
    __pyx_t_7 = __pyx_v_self;
    __Pyx_INCREF(__pyx_t_7);
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 916, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_entry);
    __Pyx_GIVEREF(__pyx_v_entry);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_entry) != (0)) __PYX_ERR(0, 916, __pyx_L1_error);
    __pyx_t_8 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_1};
      __pyx_t_6 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_mark_modified, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 916, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":913
 *             entry[1][0] = (b'l', b'', stat_value.st_size,
 *                            False, DirState.NULLSTAT)
 *     if worth_saving:             # <<<<<<<<<<<<<<
 *         # Note, even though _mark_modified will only set
 *         # IN_MEMORY_HASH_MODIFIED, it still isn't worth
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":917
 *         # IN_MEMORY_HASH_MODIFIED, it still isn't worth
 *         self._mark_modified([entry])
 *     return link_or_sha1             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_link_or_sha1);
  __pyx_r = __pyx_v_link_or_sha1;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":804
 * 
 * 
 * cpdef update_entry(self, entry, abspath, stat_value):             # <<<<<<<<<<<<<<
 *     """Update the entry based on what is actually on disk.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.update_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed_stat);
  __Pyx_XDECREF(__pyx_v_saved_link_or_sha1);
  __Pyx_XDECREF(__pyx_v_saved_file_size);
  __Pyx_XDECREF(__pyx_v_saved_executable);
  __Pyx_XDECREF(__pyx_v_saved_packed_stat);
  __Pyx_XDECREF(__pyx_v_link_or_sha1);
  __Pyx_XDECREF(__pyx_v_executable);
  __Pyx_XDECREF(__pyx_v_block_index);
  __Pyx_XDECREF(__pyx_v_entry_index);
  __Pyx_XDECREF(__pyx_v_dir_present);
  __Pyx_XDECREF(__pyx_v_file_present);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_17update_entry(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_16update_entry, "Update the entry based on what is actually on disk.\n\n    This function only calculates the sha if it needs to - if the entry is\n    uncachable, or clearly different to the first parent's entry, no sha\n    is calculated, and None is returned.\n\n    :param self: The dirstate object this is operating on.\n    :param entry: This is the dirblock entry for the file in question.\n    :param abspath: The path on disk for this file.\n    :param stat_value: The stat value done on the path.\n    :return: None, or The sha1 hexdigest of the file (40 bytes) or link\n        target of a symlink.\n    ");
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_17update_entry = {"update_entry", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_17update_entry, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_16update_entry};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_17update_entry(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_entry = 0;
  PyObject *__pyx_v_abspath = 0;
  PyObject *__pyx_v_stat_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update_entry (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,&__pyx_mstate_global->__pyx_n_u_entry,&__pyx_mstate_global->__pyx_n_u_abspath,&__pyx_mstate_global->__pyx_n_u_stat_value,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 804, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 804, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 804, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 804, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 804, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "update_entry", 0) < (0)) __PYX_ERR(0, 804, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 4; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("update_entry", 1, 4, 4, i); __PYX_ERR(0, 804, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 804, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 804, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 804, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 804, __pyx_L3_error)
    }
    __pyx_v_self = values[0];
    __pyx_v_entry = values[1];
    __pyx_v_abspath = values[2];
    __pyx_v_stat_value = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("update_entry", 1, 4, 4, __pyx_nargs); __PYX_ERR(0, 804, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.update_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_16update_entry(__pyx_self, __pyx_v_self, __pyx_v_entry, __pyx_v_abspath, __pyx_v_stat_value);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_16update_entry(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_entry, PyObject *__pyx_v_abspath, PyObject *__pyx_v_stat_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update_entry", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_update_entry(__pyx_v_self, __pyx_v_entry, __pyx_v_abspath, __pyx_v_stat_value, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.update_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":921
 * 
 * # TODO: Do we want to worry about exceptions here?
 * cdef char _minikind_from_string(object string) except? -1:             # <<<<<<<<<<<<<<
 *     """Convert a python string to a char."""
 *     return PyBytes_AsString(string)[0]
*/

static char __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_from_string(PyObject *__pyx_v_string) {
  char __pyx_r;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":923
 * cdef char _minikind_from_string(object string) except? -1:
 *     """Convert a python string to a char."""
 *     return PyBytes_AsString(string)[0]             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = (PyBytes_AsString(__pyx_v_string)[0]);
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":921
 * 
 * # TODO: Do we want to worry about exceptions here?
 * cdef char _minikind_from_string(object string) except? -1:             # <<<<<<<<<<<<<<
 *     """Convert a python string to a char."""
 *     return PyBytes_AsString(string)[0]
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":940
 * 
 * 
 * cdef object _minikind_to_kind(char minikind):             # <<<<<<<<<<<<<<
 *     """Create a string kind for minikind."""
 *     cdef char _minikind[1]
*/

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_to_kind(char __pyx_v_minikind) {
  char __pyx_v__minikind[1];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_minikind_to_kind", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":943
 *     """Create a string kind for minikind."""
 *     cdef char _minikind[1]
 *     if minikind == b'f':             # <<<<<<<<<<<<<<
 *         return _kind_file
 *     elif minikind == b'd':
*/
  switch (__pyx_v_minikind) {
    case 'f':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":944
 *     cdef char _minikind[1]
 *     if minikind == b'f':
 *         return _kind_file             # <<<<<<<<<<<<<<
 *     elif minikind == b'd':
 *         return _kind_directory
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_file);
    __pyx_r = __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_file;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":943
 *     """Create a string kind for minikind."""
 *     cdef char _minikind[1]
 *     if minikind == b'f':             # <<<<<<<<<<<<<<
 *         return _kind_file
 *     elif minikind == b'd':
*/
    break;
    case 'd':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":946
 *         return _kind_file
 *     elif minikind == b'd':
 *         return _kind_directory             # <<<<<<<<<<<<<<
 *     elif minikind == b'a':
 *         return _kind_absent
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_directory);
    __pyx_r = __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_directory;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":945
 *     if minikind == b'f':
 *         return _kind_file
 *     elif minikind == b'd':             # <<<<<<<<<<<<<<
 *         return _kind_directory
 *     elif minikind == b'a':
*/
    break;
    case 'a':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":948
 *         return _kind_directory
 *     elif minikind == b'a':
 *         return _kind_absent             # <<<<<<<<<<<<<<
 *     elif minikind == b'r':
 *         return _kind_relocated
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_absent);
    __pyx_r = __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_absent;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":947
 *     elif minikind == b'd':
 *         return _kind_directory
 *     elif minikind == b'a':             # <<<<<<<<<<<<<<
 *         return _kind_absent
 *     elif minikind == b'r':
*/
    break;
    case 'r':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":950
 *         return _kind_absent
 *     elif minikind == b'r':
 *         return _kind_relocated             # <<<<<<<<<<<<<<
 *     elif minikind == b'l':
 *         return _kind_symlink
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_relocated);
    __pyx_r = __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_relocated;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":949
 *     elif minikind == b'a':
 *         return _kind_absent
 *     elif minikind == b'r':             # <<<<<<<<<<<<<<
 *         return _kind_relocated
 *     elif minikind == b'l':
*/
    break;
    case 'l':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":952
 *         return _kind_relocated
 *     elif minikind == b'l':
 *         return _kind_symlink             # <<<<<<<<<<<<<<
 *     elif minikind == b't':
 *         return _kind_tree_reference
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_symlink);
    __pyx_r = __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_symlink;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":951
 *     elif minikind == b'r':
 *         return _kind_relocated
 *     elif minikind == b'l':             # <<<<<<<<<<<<<<
 *         return _kind_symlink
 *     elif minikind == b't':
*/
    break;
    case 't':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":954
 *         return _kind_symlink
 *     elif minikind == b't':
 *         return _kind_tree_reference             # <<<<<<<<<<<<<<
 *     _minikind[0] = minikind
 *     raise KeyError(PyBytes_FromStringAndSize(_minikind, 1))
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_tree_reference);
    __pyx_r = __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_tree_reference;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":953
 *     elif minikind == b'l':
 *         return _kind_symlink
 *     elif minikind == b't':             # <<<<<<<<<<<<<<
 *         return _kind_tree_reference
 *     _minikind[0] = minikind
*/
    break;
    default: break;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":955
 *     elif minikind == b't':
 *         return _kind_tree_reference
 *     _minikind[0] = minikind             # <<<<<<<<<<<<<<
 *     raise KeyError(PyBytes_FromStringAndSize(_minikind, 1))
 * 
*/
  (__pyx_v__minikind[0]) = __pyx_v_minikind;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":956
 *         return _kind_tree_reference
 *     _minikind[0] = minikind
 *     raise KeyError(PyBytes_FromStringAndSize(_minikind, 1))             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = PyBytes_FromStringAndSize(__pyx_v__minikind, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 956, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_KeyError)), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 956, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 956, __pyx_L1_error)

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":940
 * 
 * 
 * cdef object _minikind_to_kind(char minikind):             # <<<<<<<<<<<<<<
 *     """Create a string kind for minikind."""
 *     cdef char _minikind[1]
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._minikind_to_kind", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":959
 * 
 * 
 * cdef int _versioned_minikind(char minikind): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Return non-zero if minikind is in fltd"""
 *     return (minikind == b'f' or
*/

static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__versioned_minikind(char __pyx_v_minikind) {
  int __pyx_r;
  int __pyx_t_1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":961
 * cdef int _versioned_minikind(char minikind): # cannot_raise
 *     """Return non-zero if minikind is in fltd"""
 *     return (minikind == b'f' or             # <<<<<<<<<<<<<<
 *             minikind == b'd' or
 *             minikind == b'l' or
*/
  switch (__pyx_v_minikind) {
    case 'f':
    case 'd':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":962
 *     """Return non-zero if minikind is in fltd"""
 *     return (minikind == b'f' or
 *             minikind == b'd' or             # <<<<<<<<<<<<<<
 *             minikind == b'l' or
 *             minikind == b't')
*/
    case 'l':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":963
 *     return (minikind == b'f' or
 *             minikind == b'd' or
 *             minikind == b'l' or             # <<<<<<<<<<<<<<
 *             minikind == b't')
 * 
*/
    case 't':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":961
 * cdef int _versioned_minikind(char minikind): # cannot_raise
 *     """Return non-zero if minikind is in fltd"""
 *     return (minikind == b'f' or             # <<<<<<<<<<<<<<
 *             minikind == b'd' or
 *             minikind == b'l' or
*/
    __pyx_t_1 = 1;
    break;
    default:
    __pyx_t_1 = 0;
    break;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":959
 * 
 * 
 * cdef int _versioned_minikind(char minikind): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Return non-zero if minikind is in fltd"""
 *     return (minikind == b'f' or
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1011
 *     cdef object sha_file
 * 
 *     def __init__(self, include_unchanged, use_filesystem_for_exec,             # <<<<<<<<<<<<<<
 *         search_specific_files, state, source_index, target_index,
 *         want_unversioned, tree):
*/

/* Python wrapper */
static int __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_include_unchanged = 0;
  PyObject *__pyx_v_use_filesystem_for_exec = 0;
  PyObject *__pyx_v_search_specific_files = 0;
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v_source_index = 0;
  PyObject *__pyx_v_target_index = 0;
  PyObject *__pyx_v_want_unversioned = 0;
  PyObject *__pyx_v_tree = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[8] = {0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_include_unchanged,&__pyx_mstate_global->__pyx_n_u_use_filesystem_for_exec,&__pyx_mstate_global->__pyx_n_u_search_specific_files,&__pyx_mstate_global->__pyx_n_u_state,&__pyx_mstate_global->__pyx_n_u_source_index,&__pyx_mstate_global->__pyx_n_u_target_index,&__pyx_mstate_global->__pyx_n_u_want_unversioned,&__pyx_mstate_global->__pyx_n_u_tree,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1011, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  8:
        values[7] = __Pyx_ArgRef_VARARGS(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 1011, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 1011, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 1011, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1011, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1011, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1011, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1011, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1011, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 1011, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 8; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 8, 8, i); __PYX_ERR(0, 1011, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 8)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1011, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1011, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1011, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1011, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1011, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 1011, __pyx_L3_error)
      values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 1011, __pyx_L3_error)
      values[7] = __Pyx_ArgRef_VARARGS(__pyx_args, 7);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 1011, __pyx_L3_error)
    }
    __pyx_v_include_unchanged = values[0];
    __pyx_v_use_filesystem_for_exec = values[1];
    __pyx_v_search_specific_files = values[2];
    __pyx_v_state = values[3];
    __pyx_v_source_index = values[4];
    __pyx_v_target_index = values[5];
    __pyx_v_want_unversioned = values[6];
    __pyx_v_tree = values[7];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 8, 8, __pyx_nargs); __PYX_ERR(0, 1011, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC___init__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self), __pyx_v_include_unchanged, __pyx_v_use_filesystem_for_exec, __pyx_v_search_specific_files, __pyx_v_state, __pyx_v_source_index, __pyx_v_target_index, __pyx_v_want_unversioned, __pyx_v_tree);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC___init__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_include_unchanged, PyObject *__pyx_v_use_filesystem_for_exec, PyObject *__pyx_v_search_specific_files, PyObject *__pyx_v_state, PyObject *__pyx_v_source_index, PyObject *__pyx_v_target_index, PyObject *__pyx_v_want_unversioned, PyObject *__pyx_v_tree) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1014
 *         search_specific_files, state, source_index, target_index,
 *         want_unversioned, tree):
 *         self.doing_consistency_expansion = 0             # <<<<<<<<<<<<<<
 *         self.old_dirname_to_file_id = {}
 *         self.new_dirname_to_file_id = {}
*/
  __pyx_v_self->doing_consistency_expansion = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1015
 *         want_unversioned, tree):
 *         self.doing_consistency_expansion = 0
 *         self.old_dirname_to_file_id = {}             # <<<<<<<<<<<<<<
 *         self.new_dirname_to_file_id = {}
 *         # Are we doing a partial iter_changes?
*/
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->old_dirname_to_file_id);
  __Pyx_DECREF(__pyx_v_self->old_dirname_to_file_id);
  __pyx_v_self->old_dirname_to_file_id = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1016
 *         self.doing_consistency_expansion = 0
 *         self.old_dirname_to_file_id = {}
 *         self.new_dirname_to_file_id = {}             # <<<<<<<<<<<<<<
 *         # Are we doing a partial iter_changes?
 *         self.partial = set(['']).__ne__(search_specific_files)
*/
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->new_dirname_to_file_id);
  __Pyx_DECREF(__pyx_v_self->new_dirname_to_file_id);
  __pyx_v_self->new_dirname_to_file_id = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1018
 *         self.new_dirname_to_file_id = {}
 *         # Are we doing a partial iter_changes?
 *         self.partial = set(['']).__ne__(search_specific_files)             # <<<<<<<<<<<<<<
 *         # Using a list so that we can access the values and change them in
 *         # nested scope. Each one is [path, file_id, entry]
*/
  __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PySet_Add(__pyx_t_3, __pyx_mstate_global->__pyx_kp_u__4) < (0)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_search_specific_files};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_ne, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1018, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_5 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->partial = __pyx_t_5;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1021
 *         # Using a list so that we can access the values and change them in
 *         # nested scope. Each one is [path, file_id, entry]
 *         self.last_source_parent = [None, None]             # <<<<<<<<<<<<<<
 *         self.last_target_parent = [None, None]
 *         if include_unchanged is None:
*/
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, Py_None) != (0)) __PYX_ERR(0, 1021, __pyx_L1_error);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, Py_None) != (0)) __PYX_ERR(0, 1021, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->last_source_parent);
  __Pyx_DECREF(__pyx_v_self->last_source_parent);
  __pyx_v_self->last_source_parent = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1022
 *         # nested scope. Each one is [path, file_id, entry]
 *         self.last_source_parent = [None, None]
 *         self.last_target_parent = [None, None]             # <<<<<<<<<<<<<<
 *         if include_unchanged is None:
 *             self.include_unchanged = False
*/
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1022, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, Py_None) != (0)) __PYX_ERR(0, 1022, __pyx_L1_error);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, Py_None) != (0)) __PYX_ERR(0, 1022, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->last_target_parent);
  __Pyx_DECREF(__pyx_v_self->last_target_parent);
  __pyx_v_self->last_target_parent = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1023
 *         self.last_source_parent = [None, None]
 *         self.last_target_parent = [None, None]
 *         if include_unchanged is None:             # <<<<<<<<<<<<<<
 *             self.include_unchanged = False
 *         else:
*/
  __pyx_t_6 = (__pyx_v_include_unchanged == Py_None);
  if (__pyx_t_6) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1024
 *         self.last_target_parent = [None, None]
 *         if include_unchanged is None:
 *             self.include_unchanged = False             # <<<<<<<<<<<<<<
 *         else:
 *             self.include_unchanged = int(include_unchanged)
*/
    __pyx_v_self->include_unchanged = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1023
 *         self.last_source_parent = [None, None]
 *         self.last_target_parent = [None, None]
 *         if include_unchanged is None:             # <<<<<<<<<<<<<<
 *             self.include_unchanged = False
 *         else:
*/
    goto __pyx_L3;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1026
 *             self.include_unchanged = False
 *         else:
 *             self.include_unchanged = int(include_unchanged)             # <<<<<<<<<<<<<<
 *         self.use_filesystem_for_exec = use_filesystem_for_exec
 *         self.utf8_decode = codecs.utf_8_decode
*/
  /*else*/ {
    __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_v_include_unchanged); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_self->include_unchanged = __pyx_t_5;
  }
  __pyx_L3:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1027
 *         else:
 *             self.include_unchanged = int(include_unchanged)
 *         self.use_filesystem_for_exec = use_filesystem_for_exec             # <<<<<<<<<<<<<<
 *         self.utf8_decode = codecs.utf_8_decode
 *         # for all search_indexs in each path at or under each element of
*/
  __Pyx_INCREF(__pyx_v_use_filesystem_for_exec);
  __Pyx_GIVEREF(__pyx_v_use_filesystem_for_exec);
  __Pyx_GOTREF(__pyx_v_self->use_filesystem_for_exec);
  __Pyx_DECREF(__pyx_v_self->use_filesystem_for_exec);
  __pyx_v_self->use_filesystem_for_exec = __pyx_v_use_filesystem_for_exec;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1028
 *             self.include_unchanged = int(include_unchanged)
 *         self.use_filesystem_for_exec = use_filesystem_for_exec
 *         self.utf8_decode = codecs.utf_8_decode             # <<<<<<<<<<<<<<
 *         # for all search_indexs in each path at or under each element of
 *         # search_specific_files, if the detail is relocated: add the id, and
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_codecs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_utf_8_decode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->utf8_decode);
  __Pyx_DECREF(__pyx_v_self->utf8_decode);
  __pyx_v_self->utf8_decode = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1033
 *         # add the relocated path as one to search if its not searched already.
 *         # If the detail is not relocated, add the id.
 *         self.searched_specific_files = set()             # <<<<<<<<<<<<<<
 *         # When we search exact paths without expanding downwards, we record
 *         # that here.
*/
  __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->searched_specific_files);
  __Pyx_DECREF(__pyx_v_self->searched_specific_files);
  __pyx_v_self->searched_specific_files = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1036
 *         # When we search exact paths without expanding downwards, we record
 *         # that here.
 *         self.searched_exact_paths = set()             # <<<<<<<<<<<<<<
 *         self.search_specific_files = search_specific_files
 *         # The parents up to the root of the paths we are searching.
*/
  __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->searched_exact_paths);
  __Pyx_DECREF(__pyx_v_self->searched_exact_paths);
  __pyx_v_self->searched_exact_paths = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1037
 *         # that here.
 *         self.searched_exact_paths = set()
 *         self.search_specific_files = search_specific_files             # <<<<<<<<<<<<<<
 *         # The parents up to the root of the paths we are searching.
 *         # After all normal paths are returned, these specific items are returned.
*/
  __Pyx_INCREF(__pyx_v_search_specific_files);
  __Pyx_GIVEREF(__pyx_v_search_specific_files);
  __Pyx_GOTREF(__pyx_v_self->search_specific_files);
  __Pyx_DECREF(__pyx_v_self->search_specific_files);
  __pyx_v_self->search_specific_files = __pyx_v_search_specific_files;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1040
 *         # The parents up to the root of the paths we are searching.
 *         # After all normal paths are returned, these specific items are returned.
 *         self.search_specific_file_parents = set()             # <<<<<<<<<<<<<<
 *         # The ids we've sent out in the delta.
 *         self.seen_ids = set()
*/
  __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->search_specific_file_parents);
  __Pyx_DECREF(__pyx_v_self->search_specific_file_parents);
  __pyx_v_self->search_specific_file_parents = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1042
 *         self.search_specific_file_parents = set()
 *         # The ids we've sent out in the delta.
 *         self.seen_ids = set()             # <<<<<<<<<<<<<<
 *         self.state = state
 *         self.current_root = None
*/
  __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1042, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->seen_ids);
  __Pyx_DECREF(__pyx_v_self->seen_ids);
  __pyx_v_self->seen_ids = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1043
 *         # The ids we've sent out in the delta.
 *         self.seen_ids = set()
 *         self.state = state             # <<<<<<<<<<<<<<
 *         self.current_root = None
 *         self.current_root_unicode = None
*/
  __Pyx_INCREF(__pyx_v_state);
  __Pyx_GIVEREF(__pyx_v_state);
  __Pyx_GOTREF(__pyx_v_self->state);
  __Pyx_DECREF(__pyx_v_self->state);
  __pyx_v_self->state = __pyx_v_state;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1044
 *         self.seen_ids = set()
 *         self.state = state
 *         self.current_root = None             # <<<<<<<<<<<<<<
 *         self.current_root_unicode = None
 *         self.root_entries = None
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->current_root);
  __Pyx_DECREF(__pyx_v_self->current_root);
  __pyx_v_self->current_root = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1045
 *         self.state = state
 *         self.current_root = None
 *         self.current_root_unicode = None             # <<<<<<<<<<<<<<
 *         self.root_entries = None
 *         self.root_entries_pos = 0
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->current_root_unicode);
  __Pyx_DECREF(__pyx_v_self->current_root_unicode);
  __pyx_v_self->current_root_unicode = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1046
 *         self.current_root = None
 *         self.current_root_unicode = None
 *         self.root_entries = None             # <<<<<<<<<<<<<<
 *         self.root_entries_pos = 0
 *         self.root_entries_len = 0
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->root_entries);
  __Pyx_DECREF(__pyx_v_self->root_entries);
  __pyx_v_self->root_entries = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1047
 *         self.current_root_unicode = None
 *         self.root_entries = None
 *         self.root_entries_pos = 0             # <<<<<<<<<<<<<<
 *         self.root_entries_len = 0
 *         self.root_abspath = None
*/
  __pyx_v_self->root_entries_pos = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1048
 *         self.root_entries = None
 *         self.root_entries_pos = 0
 *         self.root_entries_len = 0             # <<<<<<<<<<<<<<
 *         self.root_abspath = None
 *         if source_index is None:
*/
  __pyx_v_self->root_entries_len = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1049
 *         self.root_entries_pos = 0
 *         self.root_entries_len = 0
 *         self.root_abspath = None             # <<<<<<<<<<<<<<
 *         if source_index is None:
 *             self.source_index = -1
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->root_abspath);
  __Pyx_DECREF(__pyx_v_self->root_abspath);
  __pyx_v_self->root_abspath = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1050
 *         self.root_entries_len = 0
 *         self.root_abspath = None
 *         if source_index is None:             # <<<<<<<<<<<<<<
 *             self.source_index = -1
 *         else:
*/
  __pyx_t_6 = (__pyx_v_source_index == Py_None);
  if (__pyx_t_6) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1051
 *         self.root_abspath = None
 *         if source_index is None:
 *             self.source_index = -1             # <<<<<<<<<<<<<<
 *         else:
 *             self.source_index = source_index
*/
    __pyx_v_self->source_index = -1;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1050
 *         self.root_entries_len = 0
 *         self.root_abspath = None
 *         if source_index is None:             # <<<<<<<<<<<<<<
 *             self.source_index = -1
 *         else:
*/
    goto __pyx_L4;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1053
 *             self.source_index = -1
 *         else:
 *             self.source_index = source_index             # <<<<<<<<<<<<<<
 *         self.target_index = target_index
 *         self.want_unversioned = want_unversioned
*/
  /*else*/ {
    __pyx_t_5 = __Pyx_PyLong_As_int(__pyx_v_source_index); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1053, __pyx_L1_error)
    __pyx_v_self->source_index = __pyx_t_5;
  }
  __pyx_L4:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1054
 *         else:
 *             self.source_index = source_index
 *         self.target_index = target_index             # <<<<<<<<<<<<<<
 *         self.want_unversioned = want_unversioned
 *         self.tree = tree
*/
  __pyx_t_5 = __Pyx_PyLong_As_int(__pyx_v_target_index); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1054, __pyx_L1_error)
  __pyx_v_self->target_index = __pyx_t_5;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1055
 *             self.source_index = source_index
 *         self.target_index = target_index
 *         self.want_unversioned = want_unversioned             # <<<<<<<<<<<<<<
 *         self.tree = tree
 *         self.dir_iterator = None
*/
  __pyx_t_5 = __Pyx_PyLong_As_int(__pyx_v_want_unversioned); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1055, __pyx_L1_error)
  __pyx_v_self->want_unversioned = __pyx_t_5;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1056
 *         self.target_index = target_index
 *         self.want_unversioned = want_unversioned
 *         self.tree = tree             # <<<<<<<<<<<<<<
 *         self.dir_iterator = None
 *         self.block_index = -1
*/
  __Pyx_INCREF(__pyx_v_tree);
  __Pyx_GIVEREF(__pyx_v_tree);
  __Pyx_GOTREF(__pyx_v_self->tree);
  __Pyx_DECREF(__pyx_v_self->tree);
  __pyx_v_self->tree = __pyx_v_tree;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1057
 *         self.want_unversioned = want_unversioned
 *         self.tree = tree
 *         self.dir_iterator = None             # <<<<<<<<<<<<<<
 *         self.block_index = -1
 *         self.current_block = None
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->dir_iterator);
  __Pyx_DECREF(__pyx_v_self->dir_iterator);
  __pyx_v_self->dir_iterator = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1058
 *         self.tree = tree
 *         self.dir_iterator = None
 *         self.block_index = -1             # <<<<<<<<<<<<<<
 *         self.current_block = None
 *         self.current_block_list = None
*/
  __pyx_v_self->block_index = -1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1059
 *         self.dir_iterator = None
 *         self.block_index = -1
 *         self.current_block = None             # <<<<<<<<<<<<<<
 *         self.current_block_list = None
 *         self.current_block_pos = -1
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->current_block);
  __Pyx_DECREF(__pyx_v_self->current_block);
  __pyx_v_self->current_block = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1060
 *         self.block_index = -1
 *         self.current_block = None
 *         self.current_block_list = None             # <<<<<<<<<<<<<<
 *         self.current_block_pos = -1
 *         self.current_dir_info = None
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->current_block_list);
  __Pyx_DECREF(__pyx_v_self->current_block_list);
  __pyx_v_self->current_block_list = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1061
 *         self.current_block = None
 *         self.current_block_list = None
 *         self.current_block_pos = -1             # <<<<<<<<<<<<<<
 *         self.current_dir_info = None
 *         self.current_dir_list = None
*/
  __pyx_v_self->current_block_pos = -1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1062
 *         self.current_block_list = None
 *         self.current_block_pos = -1
 *         self.current_dir_info = None             # <<<<<<<<<<<<<<
 *         self.current_dir_list = None
 *         self._pending_consistent_entries = []
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->current_dir_info);
  __Pyx_DECREF(__pyx_v_self->current_dir_info);
  __pyx_v_self->current_dir_info = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1063
 *         self.current_block_pos = -1
 *         self.current_dir_info = None
 *         self.current_dir_list = None             # <<<<<<<<<<<<<<
 *         self._pending_consistent_entries = []
 *         self.path_index = 0
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->current_dir_list);
  __Pyx_DECREF(__pyx_v_self->current_dir_list);
  __pyx_v_self->current_dir_list = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1064
 *         self.current_dir_info = None
 *         self.current_dir_list = None
 *         self._pending_consistent_entries = []             # <<<<<<<<<<<<<<
 *         self.path_index = 0
 *         self.root_dir_info = None
*/
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_pending_consistent_entries);
  __Pyx_DECREF(__pyx_v_self->_pending_consistent_entries);
  __pyx_v_self->_pending_consistent_entries = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1065
 *         self.current_dir_list = None
 *         self._pending_consistent_entries = []
 *         self.path_index = 0             # <<<<<<<<<<<<<<
 *         self.root_dir_info = None
 *         self.bisect_left = bisect.bisect_left
*/
  __pyx_v_self->path_index = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1066
 *         self._pending_consistent_entries = []
 *         self.path_index = 0
 *         self.root_dir_info = None             # <<<<<<<<<<<<<<
 *         self.bisect_left = bisect.bisect_left
 *         self.pathjoin = osutils.pathjoin
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->root_dir_info);
  __Pyx_DECREF(__pyx_v_self->root_dir_info);
  __pyx_v_self->root_dir_info = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1067
 *         self.path_index = 0
 *         self.root_dir_info = None
 *         self.bisect_left = bisect.bisect_left             # <<<<<<<<<<<<<<
 *         self.pathjoin = osutils.pathjoin
 *         self.fstat = os.fstat
*/
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_bisect); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1067, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_bisect_left); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1067, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->bisect_left);
  __Pyx_DECREF(__pyx_v_self->bisect_left);
  __pyx_v_self->bisect_left = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1068
 *         self.root_dir_info = None
 *         self.bisect_left = bisect.bisect_left
 *         self.pathjoin = osutils.pathjoin             # <<<<<<<<<<<<<<
 *         self.fstat = os.fstat
 *         self.sha_file = osutils.sha_file
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_osutils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_pathjoin); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->pathjoin);
  __Pyx_DECREF(__pyx_v_self->pathjoin);
  __pyx_v_self->pathjoin = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1069
 *         self.bisect_left = bisect.bisect_left
 *         self.pathjoin = osutils.pathjoin
 *         self.fstat = os.fstat             # <<<<<<<<<<<<<<
 *         self.sha_file = osutils.sha_file
 *         if target_index != 0:
*/
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_fstat); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->fstat);
  __Pyx_DECREF(__pyx_v_self->fstat);
  __pyx_v_self->fstat = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1070
 *         self.pathjoin = osutils.pathjoin
 *         self.fstat = os.fstat
 *         self.sha_file = osutils.sha_file             # <<<<<<<<<<<<<<
 *         if target_index != 0:
 *             # A lot of code in here depends on target_index == 0
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_osutils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_sha_file); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->sha_file);
  __Pyx_DECREF(__pyx_v_self->sha_file);
  __pyx_v_self->sha_file = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1071
 *         self.fstat = os.fstat
 *         self.sha_file = osutils.sha_file
 *         if target_index != 0:             # <<<<<<<<<<<<<<
 *             # A lot of code in here depends on target_index == 0
 *             raise errors.BzrError('unsupported target index')
*/
  __pyx_t_6 = (__Pyx_PyLong_BoolNeObjC(__pyx_v_target_index, __pyx_mstate_global->__pyx_int_0, 0, 0)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 1071, __pyx_L1_error)
  if (unlikely(__pyx_t_6)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1073
 *         if target_index != 0:
 *             # A lot of code in here depends on target_index == 0
 *             raise errors.BzrError('unsupported target index')             # <<<<<<<<<<<<<<
 * 
 *     cdef _process_entry(self, entry, path_info):
*/
    __pyx_t_1 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_errors); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_BzrError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
      assert(__pyx_t_1);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_unsupported_target_index};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1073, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1073, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1071
 *         self.fstat = os.fstat
 *         self.sha_file = osutils.sha_file
 *         if target_index != 0:             # <<<<<<<<<<<<<<
 *             # A lot of code in here depends on target_index == 0
 *             raise errors.BzrError('unsupported target index')
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1011
 *     cdef object sha_file
 * 
 *     def __init__(self, include_unchanged, use_filesystem_for_exec,             # <<<<<<<<<<<<<<
 *         search_specific_files, state, source_index, target_index,
 *         want_unversioned, tree):
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1075
 *             raise errors.BzrError('unsupported target index')
 * 
 *     cdef _process_entry(self, entry, path_info):             # <<<<<<<<<<<<<<
 *         """Compare an entry and real disk to generate delta information.
 * 
*/

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__process_entry(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_entry, PyObject *__pyx_v_path_info) {
  char __pyx_v_target_minikind;
  char __pyx_v_source_minikind;
  PyObject *__pyx_v_file_id = 0;
  int __pyx_v_content_change;
  PyObject *__pyx_v_details_list = 0;
  PyObject *__pyx_v_source_details = NULL;
  PyObject *__pyx_v_target_details = NULL;
  PyObject *__pyx_v_link_or_sha1 = NULL;
  PyObject *__pyx_v_old_dirname = NULL;
  PyObject *__pyx_v_old_basename = NULL;
  PyObject *__pyx_v_old_path = NULL;
  PyObject *__pyx_v_path = NULL;
  PyObject *__pyx_v_old_entry = NULL;
  PyObject *__pyx_v_target_kind = NULL;
  PyObject *__pyx_v_target_exec = NULL;
  PyObject *__pyx_v_statvalue = NULL;
  PyObject *__pyx_v_source_parent_id = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_source_parent_entry = NULL;
  PyObject *__pyx_v_new_dirname = NULL;
  PyObject *__pyx_v_target_parent_id = NULL;
  PyObject *__pyx_v_target_parent_entry = NULL;
  PyObject *__pyx_v_source_exec = NULL;
  PyObject *__pyx_v_changed = NULL;
  PyObject *__pyx_v_old_path_u = NULL;
  PyObject *__pyx_v_path_u = NULL;
  PyObject *__pyx_v_source_kind = NULL;
  PyObject *__pyx_v_parent_entry = NULL;
  PyObject *__pyx_v_parent_id = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  char __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  size_t __pyx_t_9;
  int __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16[8];
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21[4];
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_process_entry", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1097
 *         cdef int content_change
 *         cdef object details_list
 *         file_id = None             # <<<<<<<<<<<<<<
 *         details_list = entry[1]
 *         if -1 == self.source_index:
*/
  __Pyx_INCREF(Py_None);
  __pyx_v_file_id = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1098
 *         cdef object details_list
 *         file_id = None
 *         details_list = entry[1]             # <<<<<<<<<<<<<<
 *         if -1 == self.source_index:
 *             source_details = DirState.NULL_PARENT_DETAILS
*/
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1098, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_details_list = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1099
 *         file_id = None
 *         details_list = entry[1]
 *         if -1 == self.source_index:             # <<<<<<<<<<<<<<
 *             source_details = DirState.NULL_PARENT_DETAILS
 *         else:
*/
  __pyx_t_2 = (-1L == __pyx_v_self->source_index);
  if (__pyx_t_2) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1100
 *         details_list = entry[1]
 *         if -1 == self.source_index:
 *             source_details = DirState.NULL_PARENT_DETAILS             # <<<<<<<<<<<<<<
 *         else:
 *             source_details = details_list[self.source_index]
*/
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_DirState); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_NULL_PARENT_DETAILS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_source_details = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1099
 *         file_id = None
 *         details_list = entry[1]
 *         if -1 == self.source_index:             # <<<<<<<<<<<<<<
 *             source_details = DirState.NULL_PARENT_DETAILS
 *         else:
*/
    goto __pyx_L3;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1102
 *             source_details = DirState.NULL_PARENT_DETAILS
 *         else:
 *             source_details = details_list[self.source_index]             # <<<<<<<<<<<<<<
 *         target_details = details_list[self.target_index]
 *         target_minikind = _minikind_from_string(target_details[0])
*/
  /*else*/ {
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_details_list, __pyx_v_self->source_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_source_details = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1103
 *         else:
 *             source_details = details_list[self.source_index]
 *         target_details = details_list[self.target_index]             # <<<<<<<<<<<<<<
 *         target_minikind = _minikind_from_string(target_details[0])
 *         if path_info is not None and _versioned_minikind(target_minikind):
*/
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_details_list, __pyx_v_self->target_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_target_details = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1104
 *             source_details = details_list[self.source_index]
 *         target_details = details_list[self.target_index]
 *         target_minikind = _minikind_from_string(target_details[0])             # <<<<<<<<<<<<<<
 *         if path_info is not None and _versioned_minikind(target_minikind):
 *             if self.target_index != 0:
*/
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_target_details, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_from_string(__pyx_t_3); if (unlikely(__pyx_t_4 == ((char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_target_minikind = __pyx_t_4;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1105
 *         target_details = details_list[self.target_index]
 *         target_minikind = _minikind_from_string(target_details[0])
 *         if path_info is not None and _versioned_minikind(target_minikind):             # <<<<<<<<<<<<<<
 *             if self.target_index != 0:
 *                 raise AssertionError("Unsupported target index %d" %
*/
  __pyx_t_5 = (__pyx_v_path_info != Py_None);
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_6 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__versioned_minikind(__pyx_v_target_minikind); if (unlikely(__pyx_t_6 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1105, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_6 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_2) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1106
 *         target_minikind = _minikind_from_string(target_details[0])
 *         if path_info is not None and _versioned_minikind(target_minikind):
 *             if self.target_index != 0:             # <<<<<<<<<<<<<<
 *                 raise AssertionError("Unsupported target index %d" %
 *                                      self.target_index)
*/
    __pyx_t_2 = (__pyx_v_self->target_index != 0);
    if (unlikely(__pyx_t_2)) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1107
 *         if path_info is not None and _versioned_minikind(target_minikind):
 *             if self.target_index != 0:
 *                 raise AssertionError("Unsupported target index %d" %             # <<<<<<<<<<<<<<
 *                                      self.target_index)
 *             link_or_sha1 = update_entry(self.state, entry, path_info[4], path_info[3])
*/
      __pyx_t_1 = NULL;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1108
 *             if self.target_index != 0:
 *                 raise AssertionError("Unsupported target index %d" %
 *                                      self.target_index)             # <<<<<<<<<<<<<<
 *             link_or_sha1 = update_entry(self.state, entry, path_info[4], path_info[3])
 *             # The entry may have been modified by update_entry
*/
      __pyx_t_7 = __Pyx_PyLong_From_int(__pyx_v_self->target_index); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1108, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1107
 *         if path_info is not None and _versioned_minikind(target_minikind):
 *             if self.target_index != 0:
 *                 raise AssertionError("Unsupported target index %d" %             # <<<<<<<<<<<<<<
 *                                      self.target_index)
 *             link_or_sha1 = update_entry(self.state, entry, path_info[4], path_info[3])
*/
      __pyx_t_8 = PyUnicode_Format(__pyx_mstate_global->__pyx_kp_u_Unsupported_target_index_d, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_9 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_t_8};
        __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1107, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 1107, __pyx_L1_error)

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1106
 *         target_minikind = _minikind_from_string(target_details[0])
 *         if path_info is not None and _versioned_minikind(target_minikind):
 *             if self.target_index != 0:             # <<<<<<<<<<<<<<
 *                 raise AssertionError("Unsupported target index %d" %
 *                                      self.target_index)
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1109
 *                 raise AssertionError("Unsupported target index %d" %
 *                                      self.target_index)
 *             link_or_sha1 = update_entry(self.state, entry, path_info[4], path_info[3])             # <<<<<<<<<<<<<<
 *             # The entry may have been modified by update_entry
 *             target_details = details_list[self.target_index]
*/
    __pyx_t_3 = __pyx_v_self->state;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_path_info, 4, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_path_info, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_update_entry(__pyx_t_3, __pyx_v_entry, __pyx_t_8, __pyx_t_1, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_link_or_sha1 = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1111
 *             link_or_sha1 = update_entry(self.state, entry, path_info[4], path_info[3])
 *             # The entry may have been modified by update_entry
 *             target_details = details_list[self.target_index]             # <<<<<<<<<<<<<<
 *             target_minikind = _minikind_from_string(target_details[0])
 *         else:
*/
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_details_list, __pyx_v_self->target_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF_SET(__pyx_v_target_details, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1112
 *             # The entry may have been modified by update_entry
 *             target_details = details_list[self.target_index]
 *             target_minikind = _minikind_from_string(target_details[0])             # <<<<<<<<<<<<<<
 *         else:
 *             link_or_sha1 = None
*/
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_target_details, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_from_string(__pyx_t_7); if (unlikely(__pyx_t_4 == ((char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1112, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_target_minikind = __pyx_t_4;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1105
 *         target_details = details_list[self.target_index]
 *         target_minikind = _minikind_from_string(target_details[0])
 *         if path_info is not None and _versioned_minikind(target_minikind):             # <<<<<<<<<<<<<<
 *             if self.target_index != 0:
 *                 raise AssertionError("Unsupported target index %d" %
*/
    goto __pyx_L4;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1114
 *             target_minikind = _minikind_from_string(target_details[0])
 *         else:
 *             link_or_sha1 = None             # <<<<<<<<<<<<<<
 *         # the rest of this function is 0.3 seconds on 50K paths, or
 *         # 0.000006 seconds per call.
*/
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_link_or_sha1 = Py_None;
  }
  __pyx_L4:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1117
 *         # the rest of this function is 0.3 seconds on 50K paths, or
 *         # 0.000006 seconds per call.
 *         source_minikind = _minikind_from_string(source_details[0])             # <<<<<<<<<<<<<<
 *         if ((_versioned_minikind(source_minikind) or source_minikind == b'r')
 *             and _versioned_minikind(target_minikind)):
*/
  __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_source_details, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_from_string(__pyx_t_7); if (unlikely(__pyx_t_4 == ((char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_source_minikind = __pyx_t_4;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1118
 *         # 0.000006 seconds per call.
 *         source_minikind = _minikind_from_string(source_details[0])
 *         if ((_versioned_minikind(source_minikind) or source_minikind == b'r')             # <<<<<<<<<<<<<<
 *             and _versioned_minikind(target_minikind)):
 *             # claimed content in both: diff
*/
  __pyx_t_6 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__versioned_minikind(__pyx_v_source_minikind); if (unlikely(__pyx_t_6 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1118, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_6 != 0);
  if (!__pyx_t_5) {
  } else {
    goto __pyx_L10_next_and;
  }
  __pyx_t_5 = (__pyx_v_source_minikind == 'r');
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_L10_next_and:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1119
 *         source_minikind = _minikind_from_string(source_details[0])
 *         if ((_versioned_minikind(source_minikind) or source_minikind == b'r')
 *             and _versioned_minikind(target_minikind)):             # <<<<<<<<<<<<<<
 *             # claimed content in both: diff
 *             #   r    | fdlt   |      | add source to search, add id path move and perform
*/
  __pyx_t_6 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__versioned_minikind(__pyx_v_target_minikind); if (unlikely(__pyx_t_6 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1119, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_6 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L9_bool_binop_done:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1118
 *         # 0.000006 seconds per call.
 *         source_minikind = _minikind_from_string(source_details[0])
 *         if ((_versioned_minikind(source_minikind) or source_minikind == b'r')             # <<<<<<<<<<<<<<
 *             and _versioned_minikind(target_minikind)):
 *             # claimed content in both: diff
*/
  if (__pyx_t_2) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1125
 *             #   r    | fdlt   |  a   | dangling file that was present in the basis.
 *             #        |        |      | ???
 *             if source_minikind != b'r':             # <<<<<<<<<<<<<<
 *                 old_dirname = entry[0][0]
 *                 old_basename = entry[0][1]
*/
    __pyx_t_2 = (__pyx_v_source_minikind != 'r');
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1126
 *             #        |        |      | ???
 *             if source_minikind != b'r':
 *                 old_dirname = entry[0][0]             # <<<<<<<<<<<<<<
 *                 old_basename = entry[0][1]
 *                 old_path = path = None
*/
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1126, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1126, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_v_old_dirname = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1127
 *             if source_minikind != b'r':
 *                 old_dirname = entry[0][0]
 *                 old_basename = entry[0][1]             # <<<<<<<<<<<<<<
 *                 old_path = path = None
 *             else:
*/
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_old_basename = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1128
 *                 old_dirname = entry[0][0]
 *                 old_basename = entry[0][1]
 *                 old_path = path = None             # <<<<<<<<<<<<<<
 *             else:
 *                 # add the source to the search path to find any children it
*/
      __Pyx_INCREF(Py_None);
      __pyx_v_old_path = Py_None;
      __Pyx_INCREF(Py_None);
      __pyx_v_path = Py_None;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1125
 *             #   r    | fdlt   |  a   | dangling file that was present in the basis.
 *             #        |        |      | ???
 *             if source_minikind != b'r':             # <<<<<<<<<<<<<<
 *                 old_dirname = entry[0][0]
 *                 old_basename = entry[0][1]
*/
      goto __pyx_L12;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1132
 *                 # add the source to the search path to find any children it
 *                 # has.  TODO ? : only add if it is a container ?
 *                 if (not self.doing_consistency_expansion and             # <<<<<<<<<<<<<<
 *                     not is_inside_any(self.searched_specific_files,
 *                                              source_details[1])):
*/
    /*else*/ {
      __pyx_t_5 = (!(__pyx_v_self->doing_consistency_expansion != 0));
      if (__pyx_t_5) {
      } else {
        __pyx_t_2 = __pyx_t_5;
        goto __pyx_L14_bool_binop_done;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1133
 *                 # has.  TODO ? : only add if it is a container ?
 *                 if (not self.doing_consistency_expansion and
 *                     not is_inside_any(self.searched_specific_files,             # <<<<<<<<<<<<<<
 *                                              source_details[1])):
 *                     self.search_specific_files.add(source_details[1])
*/
      __pyx_t_1 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_is_inside_any); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1134
 *                 if (not self.doing_consistency_expansion and
 *                     not is_inside_any(self.searched_specific_files,
 *                                              source_details[1])):             # <<<<<<<<<<<<<<
 *                     self.search_specific_files.add(source_details[1])
 *                     # expanding from a user requested path, parent expansion
*/
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_source_details, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1134, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_8);
        assert(__pyx_t_1);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_8, __pyx__function);
        __pyx_t_9 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_v_self->searched_specific_files, __pyx_t_3};
        __pyx_t_7 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_8, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1133, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1133
 *                 # has.  TODO ? : only add if it is a container ?
 *                 if (not self.doing_consistency_expansion and
 *                     not is_inside_any(self.searched_specific_files,             # <<<<<<<<<<<<<<
 *                                              source_details[1])):
 *                     self.search_specific_files.add(source_details[1])
*/
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1133, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_10 = (!__pyx_t_5);
      __pyx_t_2 = __pyx_t_10;
      __pyx_L14_bool_binop_done:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1132
 *                 # add the source to the search path to find any children it
 *                 # has.  TODO ? : only add if it is a container ?
 *                 if (not self.doing_consistency_expansion and             # <<<<<<<<<<<<<<
 *                     not is_inside_any(self.searched_specific_files,
 *                                              source_details[1])):
*/
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1135
 *                     not is_inside_any(self.searched_specific_files,
 *                                              source_details[1])):
 *                     self.search_specific_files.add(source_details[1])             # <<<<<<<<<<<<<<
 *                     # expanding from a user requested path, parent expansion
 *                     # for delta consistency happens later.
*/
        __pyx_t_8 = __pyx_v_self->search_specific_files;
        __Pyx_INCREF(__pyx_t_8);
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_source_details, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1135, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_3};
          __pyx_t_7 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1135, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1132
 *                 # add the source to the search path to find any children it
 *                 # has.  TODO ? : only add if it is a container ?
 *                 if (not self.doing_consistency_expansion and             # <<<<<<<<<<<<<<
 *                     not is_inside_any(self.searched_specific_files,
 *                                              source_details[1])):
*/
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1140
 *                 # generate the old path; this is needed for stating later
 *                 # as well.
 *                 old_path = source_details[1]             # <<<<<<<<<<<<<<
 *                 old_dirname, old_basename = os.path.split(old_path)
 *                 path = self.pathjoin(entry[0][0], entry[0][1])
*/
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_source_details, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1140, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_v_old_path = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1141
 *                 # as well.
 *                 old_path = source_details[1]
 *                 old_dirname, old_basename = os.path.split(old_path)             # <<<<<<<<<<<<<<
 *                 path = self.pathjoin(entry[0][0], entry[0][1])
 *                 old_entry = self.state._get_entry(self.source_index,
*/
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_os); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1141, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1141, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_3 = __pyx_t_1;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_9 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_old_path};
        __pyx_t_7 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_split, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1141, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
        PyObject* sequence = __pyx_t_7;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1141, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0);
          __Pyx_INCREF(__pyx_t_1);
          __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1);
          __Pyx_INCREF(__pyx_t_3);
        } else {
          __pyx_t_1 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1141, __pyx_L1_error)
          __Pyx_XGOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
          if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1141, __pyx_L1_error)
          __Pyx_XGOTREF(__pyx_t_3);
        }
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1141, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1141, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_8 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1141, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_8);
        index = 0; __pyx_t_1 = __pyx_t_11(__pyx_t_8); if (unlikely(!__pyx_t_1)) goto __pyx_L16_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        index = 1; __pyx_t_3 = __pyx_t_11(__pyx_t_8); if (unlikely(!__pyx_t_3)) goto __pyx_L16_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_3);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_8), 2) < (0)) __PYX_ERR(0, 1141, __pyx_L1_error)
        __pyx_t_11 = NULL;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L17_unpacking_done;
        __pyx_L16_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_11 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1141, __pyx_L1_error)
        __pyx_L17_unpacking_done:;
      }
      __pyx_v_old_dirname = __pyx_t_1;
      __pyx_t_1 = 0;
      __pyx_v_old_basename = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1142
 *                 old_path = source_details[1]
 *                 old_dirname, old_basename = os.path.split(old_path)
 *                 path = self.pathjoin(entry[0][0], entry[0][1])             # <<<<<<<<<<<<<<
 *                 old_entry = self.state._get_entry(self.source_index,
 *                                              path_utf8=old_path)
*/
      __pyx_t_3 = NULL;
      __Pyx_INCREF(__pyx_v_self->pathjoin);
      __pyx_t_1 = __pyx_v_self->pathjoin; 
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1142, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_12 = __Pyx_GetItemInt(__pyx_t_8, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1142, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1142, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_13 = __Pyx_GetItemInt(__pyx_t_8, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1142, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_9 = 1;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        assert(__pyx_t_3);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
        __pyx_t_9 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_t_12, __pyx_t_13};
        __pyx_t_7 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1142, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
      }
      __pyx_v_path = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1143
 *                 old_dirname, old_basename = os.path.split(old_path)
 *                 path = self.pathjoin(entry[0][0], entry[0][1])
 *                 old_entry = self.state._get_entry(self.source_index,             # <<<<<<<<<<<<<<
 *                                              path_utf8=old_path)
 *                 # update the source details variable to be the real
*/
      __pyx_t_1 = __pyx_v_self->state;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_13 = __Pyx_PyLong_From_int(__pyx_v_self->source_index); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1144
 *                 path = self.pathjoin(entry[0][0], entry[0][1])
 *                 old_entry = self.state._get_entry(self.source_index,
 *                                              path_utf8=old_path)             # <<<<<<<<<<<<<<
 *                 # update the source details variable to be the real
 *                 # location.
*/
      __pyx_t_9 = 0;
      {
        PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_1, __pyx_t_13};
        __pyx_t_12 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1143, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_path_utf8, __pyx_v_old_path, __pyx_t_12, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 1143, __pyx_L1_error)
        __pyx_t_7 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_get_entry, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_12);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1143, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
      }
      __pyx_v_old_entry = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1147
 *                 # update the source details variable to be the real
 *                 # location.
 *                 if old_entry == (None, None):             # <<<<<<<<<<<<<<
 *                     raise DirstateCorrupt(self.state._filename,
 *                         "entry '%s/%s' is considered renamed from %r"
*/
      __pyx_t_7 = PyObject_RichCompare(__pyx_v_old_entry, __pyx_mstate_global->__pyx_tuple[0], Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1147, __pyx_L1_error)
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1147, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(__pyx_t_2)) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1148
 *                 # location.
 *                 if old_entry == (None, None):
 *                     raise DirstateCorrupt(self.state._filename,             # <<<<<<<<<<<<<<
 *                         "entry '%s/%s' is considered renamed from %r"
 *                         " but source does not exist\n"
*/
        __pyx_t_12 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_DirstateCorrupt); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1148, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_mstate_global->__pyx_n_u_filename); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1148, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1151
 *                         "entry '%s/%s' is considered renamed from %r"
 *                         " but source does not exist\n"
 *                         "entry: %s" % (entry[0][0], entry[0][1], old_path, entry))             # <<<<<<<<<<<<<<
 *                 source_details = old_entry[1][self.source_index]
 *                 source_minikind = _minikind_from_string(source_details[0])
*/
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_8 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_t_8), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_14 = __Pyx_GetItemInt(__pyx_t_8, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_t_14), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_14 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_old_path), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_15 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_v_entry), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __pyx_t_16[0] = __pyx_mstate_global->__pyx_kp_u_entry_2;
        __pyx_t_16[1] = __pyx_t_3;
        __pyx_t_16[2] = __pyx_mstate_global->__pyx_kp_u__5;
        __pyx_t_16[3] = __pyx_t_8;
        __pyx_t_16[4] = __pyx_mstate_global->__pyx_kp_u_is_considered_renamed_from;
        __pyx_t_16[5] = __pyx_t_14;
        __pyx_t_16[6] = __pyx_mstate_global->__pyx_kp_u_but_source_does_not_exist_entry;
        __pyx_t_16[7] = __pyx_t_15;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1149
 *                 if old_entry == (None, None):
 *                     raise DirstateCorrupt(self.state._filename,
 *                         "entry '%s/%s' is considered renamed from %r"             # <<<<<<<<<<<<<<
 *                         " but source does not exist\n"
 *                         "entry: %s" % (entry[0][0], entry[0][1], old_path, entry))
*/
        __pyx_t_17 = __Pyx_PyUnicode_Join(__pyx_t_16, 8, 7 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3) + 1 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8) + 29 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_14) + 34 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_15), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_14) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15));
        if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1149, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_9 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_13);
          assert(__pyx_t_12);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_13);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_13, __pyx__function);
          __pyx_t_9 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_12, __pyx_t_1, __pyx_t_17};
          __pyx_t_7 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_13, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1148, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
        }
        __Pyx_Raise(__pyx_t_7, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(0, 1148, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1147
 *                 # update the source details variable to be the real
 *                 # location.
 *                 if old_entry == (None, None):             # <<<<<<<<<<<<<<
 *                     raise DirstateCorrupt(self.state._filename,
 *                         "entry '%s/%s' is considered renamed from %r"
*/
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1152
 *                         " but source does not exist\n"
 *                         "entry: %s" % (entry[0][0], entry[0][1], old_path, entry))
 *                 source_details = old_entry[1][self.source_index]             # <<<<<<<<<<<<<<
 *                 source_minikind = _minikind_from_string(source_details[0])
 *             if path_info is None:
*/
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_old_entry, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_13 = __Pyx_GetItemInt(__pyx_t_7, __pyx_v_self->source_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF_SET(__pyx_v_source_details, __pyx_t_13);
      __pyx_t_13 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1153
 *                         "entry: %s" % (entry[0][0], entry[0][1], old_path, entry))
 *                 source_details = old_entry[1][self.source_index]
 *                 source_minikind = _minikind_from_string(source_details[0])             # <<<<<<<<<<<<<<
 *             if path_info is None:
 *                 # the file is missing on disk, show as removed.
*/
      __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_source_details, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_4 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_from_string(__pyx_t_13); if (unlikely(__pyx_t_4 == ((char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1153, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_v_source_minikind = __pyx_t_4;
    }
    __pyx_L12:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1154
 *                 source_details = old_entry[1][self.source_index]
 *                 source_minikind = _minikind_from_string(source_details[0])
 *             if path_info is None:             # <<<<<<<<<<<<<<
 *                 # the file is missing on disk, show as removed.
 *                 content_change = 1
*/
    __pyx_t_2 = (__pyx_v_path_info == Py_None);
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1156
 *             if path_info is None:
 *                 # the file is missing on disk, show as removed.
 *                 content_change = 1             # <<<<<<<<<<<<<<
 *                 target_kind = None
 *                 target_exec = False
*/
      __pyx_v_content_change = 1;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1157
 *                 # the file is missing on disk, show as removed.
 *                 content_change = 1
 *                 target_kind = None             # <<<<<<<<<<<<<<
 *                 target_exec = False
 *             else:
*/
      __Pyx_INCREF(Py_None);
      __pyx_v_target_kind = Py_None;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1158
 *                 content_change = 1
 *                 target_kind = None
 *                 target_exec = False             # <<<<<<<<<<<<<<
 *             else:
 *                 # source and target are both versioned and disk file is present.
*/
      __Pyx_INCREF(Py_False);
      __pyx_v_target_exec = Py_False;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1154
 *                 source_details = old_entry[1][self.source_index]
 *                 source_minikind = _minikind_from_string(source_details[0])
 *             if path_info is None:             # <<<<<<<<<<<<<<
 *                 # the file is missing on disk, show as removed.
 *                 content_change = 1
*/
      goto __pyx_L19;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1161
 *             else:
 *                 # source and target are both versioned and disk file is present.
 *                 target_kind = path_info[2]             # <<<<<<<<<<<<<<
 *                 if target_kind == 'directory':
 *                     if path is None:
*/
    /*else*/ {
      __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_path_info, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1161, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_v_target_kind = __pyx_t_13;
      __pyx_t_13 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1162
 *                 # source and target are both versioned and disk file is present.
 *                 target_kind = path_info[2]
 *                 if target_kind == 'directory':             # <<<<<<<<<<<<<<
 *                     if path is None:
 *                         old_path = path = self.pathjoin(old_dirname, old_basename)
*/
      __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_target_kind, __pyx_mstate_global->__pyx_n_u_directory, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1162, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1163
 *                 target_kind = path_info[2]
 *                 if target_kind == 'directory':
 *                     if path is None:             # <<<<<<<<<<<<<<
 *                         old_path = path = self.pathjoin(old_dirname, old_basename)
 *                     file_id = entry[0][2]
*/
        __pyx_t_2 = (__pyx_v_path == Py_None);
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1164
 *                 if target_kind == 'directory':
 *                     if path is None:
 *                         old_path = path = self.pathjoin(old_dirname, old_basename)             # <<<<<<<<<<<<<<
 *                     file_id = entry[0][2]
 *                     self.new_dirname_to_file_id[path] = file_id
*/
          __pyx_t_7 = NULL;
          __Pyx_INCREF(__pyx_v_self->pathjoin);
          __pyx_t_17 = __pyx_v_self->pathjoin; 
          __pyx_t_9 = 1;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_17))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_17);
            assert(__pyx_t_7);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_17);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_17, __pyx__function);
            __pyx_t_9 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_v_old_dirname, __pyx_v_old_basename};
            __pyx_t_13 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_17, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1164, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
          }
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_DECREF_SET(__pyx_v_old_path, __pyx_t_13);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_DECREF_SET(__pyx_v_path, __pyx_t_13);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1163
 *                 target_kind = path_info[2]
 *                 if target_kind == 'directory':
 *                     if path is None:             # <<<<<<<<<<<<<<
 *                         old_path = path = self.pathjoin(old_dirname, old_basename)
 *                     file_id = entry[0][2]
*/
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1165
 *                     if path is None:
 *                         old_path = path = self.pathjoin(old_dirname, old_basename)
 *                     file_id = entry[0][2]             # <<<<<<<<<<<<<<
 *                     self.new_dirname_to_file_id[path] = file_id
 *                     if source_minikind != b'd':
*/
        __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1165, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_13, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1165, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF_SET(__pyx_v_file_id, __pyx_t_17);
        __pyx_t_17 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1166
 *                         old_path = path = self.pathjoin(old_dirname, old_basename)
 *                     file_id = entry[0][2]
 *                     self.new_dirname_to_file_id[path] = file_id             # <<<<<<<<<<<<<<
 *                     if source_minikind != b'd':
 *                         content_change = 1
*/
        if (unlikely((PyObject_SetItem(__pyx_v_self->new_dirname_to_file_id, __pyx_v_path, __pyx_v_file_id) < 0))) __PYX_ERR(0, 1166, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1167
 *                     file_id = entry[0][2]
 *                     self.new_dirname_to_file_id[path] = file_id
 *                     if source_minikind != b'd':             # <<<<<<<<<<<<<<
 *                         content_change = 1
 *                     else:
*/
        __pyx_t_2 = (__pyx_v_source_minikind != 'd');
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1168
 *                     self.new_dirname_to_file_id[path] = file_id
 *                     if source_minikind != b'd':
 *                         content_change = 1             # <<<<<<<<<<<<<<
 *                     else:
 *                         # directories have no fingerprint
*/
          __pyx_v_content_change = 1;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1167
 *                     file_id = entry[0][2]
 *                     self.new_dirname_to_file_id[path] = file_id
 *                     if source_minikind != b'd':             # <<<<<<<<<<<<<<
 *                         content_change = 1
 *                     else:
*/
          goto __pyx_L22;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1171
 *                     else:
 *                         # directories have no fingerprint
 *                         content_change = 0             # <<<<<<<<<<<<<<
 *                     target_exec = False
 *                 elif target_kind == 'file':
*/
        /*else*/ {
          __pyx_v_content_change = 0;
        }
        __pyx_L22:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1172
 *                         # directories have no fingerprint
 *                         content_change = 0
 *                     target_exec = False             # <<<<<<<<<<<<<<
 *                 elif target_kind == 'file':
 *                     if source_minikind != b'f':
*/
        __Pyx_INCREF(Py_False);
        __pyx_v_target_exec = Py_False;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1162
 *                 # source and target are both versioned and disk file is present.
 *                 target_kind = path_info[2]
 *                 if target_kind == 'directory':             # <<<<<<<<<<<<<<
 *                     if path is None:
 *                         old_path = path = self.pathjoin(old_dirname, old_basename)
*/
        goto __pyx_L20;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1173
 *                         content_change = 0
 *                     target_exec = False
 *                 elif target_kind == 'file':             # <<<<<<<<<<<<<<
 *                     if source_minikind != b'f':
 *                         content_change = 1
*/
      __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_target_kind, __pyx_mstate_global->__pyx_n_u_file, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1173, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1174
 *                     target_exec = False
 *                 elif target_kind == 'file':
 *                     if source_minikind != b'f':             # <<<<<<<<<<<<<<
 *                         content_change = 1
 *                     else:
*/
        __pyx_t_2 = (__pyx_v_source_minikind != 'f');
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1175
 *                 elif target_kind == 'file':
 *                     if source_minikind != b'f':
 *                         content_change = 1             # <<<<<<<<<<<<<<
 *                     else:
 *                         # Check the sha. We can't just rely on the size as
*/
          __pyx_v_content_change = 1;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1174
 *                     target_exec = False
 *                 elif target_kind == 'file':
 *                     if source_minikind != b'f':             # <<<<<<<<<<<<<<
 *                         content_change = 1
 *                     else:
*/
          goto __pyx_L23;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1180
 *                         # content filtering may mean differ sizes actually
 *                         # map to the same content
 *                         if link_or_sha1 is None:             # <<<<<<<<<<<<<<
 *                             # Stat cache miss:
 *                             statvalue, link_or_sha1 = \
*/
        /*else*/ {
          __pyx_t_2 = (__pyx_v_link_or_sha1 == Py_None);
          if (__pyx_t_2) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1183
 *                             # Stat cache miss:
 *                             statvalue, link_or_sha1 = \
 *                                 self.state._sha1_provider.stat_and_sha1(             # <<<<<<<<<<<<<<
 *                                 path_info[4])
 *                             self.state._observed_sha1(entry, link_or_sha1,
*/
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_mstate_global->__pyx_n_u_sha1_provider); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1183, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_13 = __pyx_t_7;
            __Pyx_INCREF(__pyx_t_13);

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1184
 *                             statvalue, link_or_sha1 = \
 *                                 self.state._sha1_provider.stat_and_sha1(
 *                                 path_info[4])             # <<<<<<<<<<<<<<
 *                             self.state._observed_sha1(entry, link_or_sha1,
 *                                 statvalue)
*/
            __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_path_info, 4, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1184, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_9 = 0;
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_t_1};
              __pyx_t_17 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_stat_and_sha1, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1183, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_17);
            }
            if ((likely(PyTuple_CheckExact(__pyx_t_17))) || (PyList_CheckExact(__pyx_t_17))) {
              PyObject* sequence = __pyx_t_17;
              Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
              if (unlikely(size != 2)) {
                if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 1182, __pyx_L1_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              if (likely(PyTuple_CheckExact(sequence))) {
                __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0);
                __Pyx_INCREF(__pyx_t_7);
                __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1);
                __Pyx_INCREF(__pyx_t_1);
              } else {
                __pyx_t_7 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
                if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1182, __pyx_L1_error)
                __Pyx_XGOTREF(__pyx_t_7);
                __pyx_t_1 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
                if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1182, __pyx_L1_error)
                __Pyx_XGOTREF(__pyx_t_1);
              }
              #else
              __pyx_t_7 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1182, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_1 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1182, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              #endif
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            } else {
              Py_ssize_t index = -1;
              __pyx_t_13 = PyObject_GetIter(__pyx_t_17); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1182, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_13);
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __pyx_t_11 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_13);
              index = 0; __pyx_t_7 = __pyx_t_11(__pyx_t_13); if (unlikely(!__pyx_t_7)) goto __pyx_L25_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_7);
              index = 1; __pyx_t_1 = __pyx_t_11(__pyx_t_13); if (unlikely(!__pyx_t_1)) goto __pyx_L25_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_1);
              if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_13), 2) < (0)) __PYX_ERR(0, 1182, __pyx_L1_error)
              __pyx_t_11 = NULL;
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              goto __pyx_L26_unpacking_done;
              __pyx_L25_unpacking_failed:;
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __pyx_t_11 = NULL;
              if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
              __PYX_ERR(0, 1182, __pyx_L1_error)
              __pyx_L26_unpacking_done:;
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1182
 *                         if link_or_sha1 is None:
 *                             # Stat cache miss:
 *                             statvalue, link_or_sha1 = \             # <<<<<<<<<<<<<<
 *                                 self.state._sha1_provider.stat_and_sha1(
 *                                 path_info[4])
*/
            __pyx_v_statvalue = __pyx_t_7;
            __pyx_t_7 = 0;
            __Pyx_DECREF_SET(__pyx_v_link_or_sha1, __pyx_t_1);
            __pyx_t_1 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1185
 *                                 self.state._sha1_provider.stat_and_sha1(
 *                                 path_info[4])
 *                             self.state._observed_sha1(entry, link_or_sha1,             # <<<<<<<<<<<<<<
 *                                 statvalue)
 *                         content_change = (link_or_sha1 != source_details[1])
*/
            __pyx_t_1 = __pyx_v_self->state;
            __Pyx_INCREF(__pyx_t_1);

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1186
 *                                 path_info[4])
 *                             self.state._observed_sha1(entry, link_or_sha1,
 *                                 statvalue)             # <<<<<<<<<<<<<<
 *                         content_change = (link_or_sha1 != source_details[1])
 *                     # Target details is updated at update_entry time
*/
            __pyx_t_9 = 0;
            {
              PyObject *__pyx_callargs[4] = {__pyx_t_1, __pyx_v_entry, __pyx_v_link_or_sha1, __pyx_v_statvalue};
              __pyx_t_17 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_observed_sha1, __pyx_callargs+__pyx_t_9, (4-__pyx_t_9) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1185, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_17);
            }
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1180
 *                         # content filtering may mean differ sizes actually
 *                         # map to the same content
 *                         if link_or_sha1 is None:             # <<<<<<<<<<<<<<
 *                             # Stat cache miss:
 *                             statvalue, link_or_sha1 = \
*/
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1187
 *                             self.state._observed_sha1(entry, link_or_sha1,
 *                                 statvalue)
 *                         content_change = (link_or_sha1 != source_details[1])             # <<<<<<<<<<<<<<
 *                     # Target details is updated at update_entry time
 *                     if self.use_filesystem_for_exec:
*/
          __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_source_details, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1187, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_1 = PyObject_RichCompare(__pyx_v_link_or_sha1, __pyx_t_17, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1187, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          __pyx_t_6 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1187, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_v_content_change = __pyx_t_6;
        }
        __pyx_L23:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1189
 *                         content_change = (link_or_sha1 != source_details[1])
 *                     # Target details is updated at update_entry time
 *                     if self.use_filesystem_for_exec:             # <<<<<<<<<<<<<<
 *                         # We don't need S_ISREG here, because we are sure
 *                         # we are dealing with a file.
*/
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_self->use_filesystem_for_exec); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1189, __pyx_L1_error)
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1192
 *                         # We don't need S_ISREG here, because we are sure
 *                         # we are dealing with a file.
 *                         target_exec = bool(S_IXUSR & path_info[3].st_mode)             # <<<<<<<<<<<<<<
 *                     else:
 *                         target_exec = target_details[3]
*/
          __pyx_t_1 = __Pyx_PyLong_From_int(S_IXUSR); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1192, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_path_info, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1192, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_st_mode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1192, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          __pyx_t_17 = PyNumber_And(__pyx_t_1, __pyx_t_7); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1192, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_17); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1192, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          __pyx_t_17 = __Pyx_PyBool_FromLong((!(!__pyx_t_2))); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1192, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_v_target_exec = __pyx_t_17;
          __pyx_t_17 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1189
 *                         content_change = (link_or_sha1 != source_details[1])
 *                     # Target details is updated at update_entry time
 *                     if self.use_filesystem_for_exec:             # <<<<<<<<<<<<<<
 *                         # We don't need S_ISREG here, because we are sure
 *                         # we are dealing with a file.
*/
          goto __pyx_L27;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1194
 *                         target_exec = bool(S_IXUSR & path_info[3].st_mode)
 *                     else:
 *                         target_exec = target_details[3]             # <<<<<<<<<<<<<<
 *                 elif target_kind == 'symlink':
 *                     if source_minikind != b'l':
*/
        /*else*/ {
          __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_target_details, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1194, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_v_target_exec = __pyx_t_17;
          __pyx_t_17 = 0;
        }
        __pyx_L27:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1173
 *                         content_change = 0
 *                     target_exec = False
 *                 elif target_kind == 'file':             # <<<<<<<<<<<<<<
 *                     if source_minikind != b'f':
 *                         content_change = 1
*/
        goto __pyx_L20;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1195
 *                     else:
 *                         target_exec = target_details[3]
 *                 elif target_kind == 'symlink':             # <<<<<<<<<<<<<<
 *                     if source_minikind != b'l':
 *                         content_change = 1
*/
      __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_target_kind, __pyx_mstate_global->__pyx_n_u_symlink, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1195, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1196
 *                         target_exec = target_details[3]
 *                 elif target_kind == 'symlink':
 *                     if source_minikind != b'l':             # <<<<<<<<<<<<<<
 *                         content_change = 1
 *                     else:
*/
        __pyx_t_2 = (__pyx_v_source_minikind != 'l');
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1197
 *                 elif target_kind == 'symlink':
 *                     if source_minikind != b'l':
 *                         content_change = 1             # <<<<<<<<<<<<<<
 *                     else:
 *                         content_change = (link_or_sha1 != source_details[1])
*/
          __pyx_v_content_change = 1;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1196
 *                         target_exec = target_details[3]
 *                 elif target_kind == 'symlink':
 *                     if source_minikind != b'l':             # <<<<<<<<<<<<<<
 *                         content_change = 1
 *                     else:
*/
          goto __pyx_L28;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1199
 *                         content_change = 1
 *                     else:
 *                         content_change = (link_or_sha1 != source_details[1])             # <<<<<<<<<<<<<<
 *                     target_exec = False
 *                 elif target_kind == 'tree-reference':
*/
        /*else*/ {
          __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_source_details, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1199, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_7 = PyObject_RichCompare(__pyx_v_link_or_sha1, __pyx_t_17, Py_NE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1199, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          __pyx_t_6 = __Pyx_PyLong_As_int(__pyx_t_7); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1199, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_v_content_change = __pyx_t_6;
        }
        __pyx_L28:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1200
 *                     else:
 *                         content_change = (link_or_sha1 != source_details[1])
 *                     target_exec = False             # <<<<<<<<<<<<<<
 *                 elif target_kind == 'tree-reference':
 *                     if source_minikind != b't':
*/
        __Pyx_INCREF(Py_False);
        __pyx_v_target_exec = Py_False;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1195
 *                     else:
 *                         target_exec = target_details[3]
 *                 elif target_kind == 'symlink':             # <<<<<<<<<<<<<<
 *                     if source_minikind != b'l':
 *                         content_change = 1
*/
        goto __pyx_L20;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1201
 *                         content_change = (link_or_sha1 != source_details[1])
 *                     target_exec = False
 *                 elif target_kind == 'tree-reference':             # <<<<<<<<<<<<<<
 *                     if source_minikind != b't':
 *                         content_change = 1
*/
      __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_target_kind, __pyx_mstate_global->__pyx_kp_u_tree_reference, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1201, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1202
 *                     target_exec = False
 *                 elif target_kind == 'tree-reference':
 *                     if source_minikind != b't':             # <<<<<<<<<<<<<<
 *                         content_change = 1
 *                     else:
*/
        __pyx_t_2 = (__pyx_v_source_minikind != 't');
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1203
 *                 elif target_kind == 'tree-reference':
 *                     if source_minikind != b't':
 *                         content_change = 1             # <<<<<<<<<<<<<<
 *                     else:
 *                         content_change = 0
*/
          __pyx_v_content_change = 1;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1202
 *                     target_exec = False
 *                 elif target_kind == 'tree-reference':
 *                     if source_minikind != b't':             # <<<<<<<<<<<<<<
 *                         content_change = 1
 *                     else:
*/
          goto __pyx_L29;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1205
 *                         content_change = 1
 *                     else:
 *                         content_change = 0             # <<<<<<<<<<<<<<
 *                     target_exec = False
 *                 else:
*/
        /*else*/ {
          __pyx_v_content_change = 0;
        }
        __pyx_L29:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1206
 *                     else:
 *                         content_change = 0
 *                     target_exec = False             # <<<<<<<<<<<<<<
 *                 else:
 *                     if path is None:
*/
        __Pyx_INCREF(Py_False);
        __pyx_v_target_exec = Py_False;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1201
 *                         content_change = (link_or_sha1 != source_details[1])
 *                     target_exec = False
 *                 elif target_kind == 'tree-reference':             # <<<<<<<<<<<<<<
 *                     if source_minikind != b't':
 *                         content_change = 1
*/
        goto __pyx_L20;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1208
 *                     target_exec = False
 *                 else:
 *                     if path is None:             # <<<<<<<<<<<<<<
 *                         path = self.pathjoin(old_dirname, old_basename)
 *                     raise errors.BadFileKindError(path, path_info[2])
*/
      /*else*/ {
        __pyx_t_2 = (__pyx_v_path == Py_None);
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1209
 *                 else:
 *                     if path is None:
 *                         path = self.pathjoin(old_dirname, old_basename)             # <<<<<<<<<<<<<<
 *                     raise errors.BadFileKindError(path, path_info[2])
 *             if source_minikind == b'd':
*/
          __pyx_t_17 = NULL;
          __Pyx_INCREF(__pyx_v_self->pathjoin);
          __pyx_t_1 = __pyx_v_self->pathjoin; 
          __pyx_t_9 = 1;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_1);
            assert(__pyx_t_17);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_17);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
            __pyx_t_9 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_17, __pyx_v_old_dirname, __pyx_v_old_basename};
            __pyx_t_7 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1209, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
          }
          __Pyx_DECREF_SET(__pyx_v_path, __pyx_t_7);
          __pyx_t_7 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1208
 *                     target_exec = False
 *                 else:
 *                     if path is None:             # <<<<<<<<<<<<<<
 *                         path = self.pathjoin(old_dirname, old_basename)
 *                     raise errors.BadFileKindError(path, path_info[2])
*/
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1210
 *                     if path is None:
 *                         path = self.pathjoin(old_dirname, old_basename)
 *                     raise errors.BadFileKindError(path, path_info[2])             # <<<<<<<<<<<<<<
 *             if source_minikind == b'd':
 *                 if path is None:
*/
        __pyx_t_1 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_errors); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1210, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_BadFileKindError); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1210, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_path_info, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1210, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        __pyx_t_9 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_13);
          assert(__pyx_t_1);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_13);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_13, __pyx__function);
          __pyx_t_9 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_v_path, __pyx_t_17};
          __pyx_t_7 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_13, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1210, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
        }
        __Pyx_Raise(__pyx_t_7, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(0, 1210, __pyx_L1_error)
      }
      __pyx_L20:;
    }
    __pyx_L19:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1211
 *                         path = self.pathjoin(old_dirname, old_basename)
 *                     raise errors.BadFileKindError(path, path_info[2])
 *             if source_minikind == b'd':             # <<<<<<<<<<<<<<
 *                 if path is None:
 *                     old_path = path = self.pathjoin(old_dirname, old_basename)
*/
    __pyx_t_2 = (__pyx_v_source_minikind == 'd');
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1212
 *                     raise errors.BadFileKindError(path, path_info[2])
 *             if source_minikind == b'd':
 *                 if path is None:             # <<<<<<<<<<<<<<
 *                     old_path = path = self.pathjoin(old_dirname, old_basename)
 *                 if file_id is None:
*/
      __pyx_t_2 = (__pyx_v_path == Py_None);
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1213
 *             if source_minikind == b'd':
 *                 if path is None:
 *                     old_path = path = self.pathjoin(old_dirname, old_basename)             # <<<<<<<<<<<<<<
 *                 if file_id is None:
 *                     file_id = entry[0][2]
*/
        __pyx_t_13 = NULL;
        __Pyx_INCREF(__pyx_v_self->pathjoin);
        __pyx_t_17 = __pyx_v_self->pathjoin; 
        __pyx_t_9 = 1;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_17))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_17);
          assert(__pyx_t_13);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_17);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_17, __pyx__function);
          __pyx_t_9 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_13, __pyx_v_old_dirname, __pyx_v_old_basename};
          __pyx_t_7 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_17, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1213, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
        }
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_DECREF_SET(__pyx_v_old_path, __pyx_t_7);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_DECREF_SET(__pyx_v_path, __pyx_t_7);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1212
 *                     raise errors.BadFileKindError(path, path_info[2])
 *             if source_minikind == b'd':
 *                 if path is None:             # <<<<<<<<<<<<<<
 *                     old_path = path = self.pathjoin(old_dirname, old_basename)
 *                 if file_id is None:
*/
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1214
 *                 if path is None:
 *                     old_path = path = self.pathjoin(old_dirname, old_basename)
 *                 if file_id is None:             # <<<<<<<<<<<<<<
 *                     file_id = entry[0][2]
 *                 self.old_dirname_to_file_id[old_path] = file_id
*/
      __pyx_t_2 = (__pyx_v_file_id == Py_None);
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1215
 *                     old_path = path = self.pathjoin(old_dirname, old_basename)
 *                 if file_id is None:
 *                     file_id = entry[0][2]             # <<<<<<<<<<<<<<
 *                 self.old_dirname_to_file_id[old_path] = file_id
 *             # parent id is the entry for the path in the target tree
*/
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1215, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_7, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1215, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF_SET(__pyx_v_file_id, __pyx_t_17);
        __pyx_t_17 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1214
 *                 if path is None:
 *                     old_path = path = self.pathjoin(old_dirname, old_basename)
 *                 if file_id is None:             # <<<<<<<<<<<<<<
 *                     file_id = entry[0][2]
 *                 self.old_dirname_to_file_id[old_path] = file_id
*/
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1216
 *                 if file_id is None:
 *                     file_id = entry[0][2]
 *                 self.old_dirname_to_file_id[old_path] = file_id             # <<<<<<<<<<<<<<
 *             # parent id is the entry for the path in the target tree
 *             if old_basename and old_dirname == self.last_source_parent[0]:
*/
      if (unlikely((PyObject_SetItem(__pyx_v_self->old_dirname_to_file_id, __pyx_v_old_path, __pyx_v_file_id) < 0))) __PYX_ERR(0, 1216, __pyx_L1_error)

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1211
 *                         path = self.pathjoin(old_dirname, old_basename)
 *                     raise errors.BadFileKindError(path, path_info[2])
 *             if source_minikind == b'd':             # <<<<<<<<<<<<<<
 *                 if path is None:
 *                     old_path = path = self.pathjoin(old_dirname, old_basename)
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1218
 *                 self.old_dirname_to_file_id[old_path] = file_id
 *             # parent id is the entry for the path in the target tree
 *             if old_basename and old_dirname == self.last_source_parent[0]:             # <<<<<<<<<<<<<<
 *                 # use a cached hit for non-root source entries.
 *                 source_parent_id = self.last_source_parent[1]
*/
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_old_basename); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 1218, __pyx_L1_error)
    if (__pyx_t_10) {
    } else {
      __pyx_t_2 = __pyx_t_10;
      goto __pyx_L35_bool_binop_done;
    }
    __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_self->last_source_parent, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_7 = PyObject_RichCompare(__pyx_v_old_dirname, __pyx_t_17, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1218, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 1218, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_2 = __pyx_t_10;
    __pyx_L35_bool_binop_done:;
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1220
 *             if old_basename and old_dirname == self.last_source_parent[0]:
 *                 # use a cached hit for non-root source entries.
 *                 source_parent_id = self.last_source_parent[1]             # <<<<<<<<<<<<<<
 *             else:
 *                 try:
*/
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_self->last_source_parent, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_v_source_parent_id = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1218
 *                 self.old_dirname_to_file_id[old_path] = file_id
 *             # parent id is the entry for the path in the target tree
 *             if old_basename and old_dirname == self.last_source_parent[0]:             # <<<<<<<<<<<<<<
 *                 # use a cached hit for non-root source entries.
 *                 source_parent_id = self.last_source_parent[1]
*/
      goto __pyx_L34;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1222
 *                 source_parent_id = self.last_source_parent[1]
 *             else:
 *                 try:             # <<<<<<<<<<<<<<
 *                     source_parent_id = self.old_dirname_to_file_id[old_dirname]
 *                 except KeyError, _:
*/
    /*else*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
        __Pyx_XGOTREF(__pyx_t_18);
        __Pyx_XGOTREF(__pyx_t_19);
        __Pyx_XGOTREF(__pyx_t_20);
        /*try:*/ {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1223
 *             else:
 *                 try:
 *                     source_parent_id = self.old_dirname_to_file_id[old_dirname]             # <<<<<<<<<<<<<<
 *                 except KeyError, _:
 *                     source_parent_entry = self.state._get_entry(self.source_index,
*/
          __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_self->old_dirname_to_file_id, __pyx_v_old_dirname); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1223, __pyx_L37_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_v_source_parent_id = __pyx_t_7;
          __pyx_t_7 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1222
 *                 source_parent_id = self.last_source_parent[1]
 *             else:
 *                 try:             # <<<<<<<<<<<<<<
 *                     source_parent_id = self.old_dirname_to_file_id[old_dirname]
 *                 except KeyError, _:
*/
        }
        __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
        goto __pyx_L42_try_end;
        __pyx_L37_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1224
 *                 try:
 *                     source_parent_id = self.old_dirname_to_file_id[old_dirname]
 *                 except KeyError, _:             # <<<<<<<<<<<<<<
 *                     source_parent_entry = self.state._get_entry(self.source_index,
 *                                                            path_utf8=old_dirname)
*/
        __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_KeyError))));
        if (__pyx_t_6) {
          __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._process_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_17, &__pyx_t_13) < 0) __PYX_ERR(0, 1224, __pyx_L39_except_error)
          __Pyx_XGOTREF(__pyx_t_7);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_INCREF(__pyx_t_17);
          __pyx_v__ = __pyx_t_17;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1225
 *                     source_parent_id = self.old_dirname_to_file_id[old_dirname]
 *                 except KeyError, _:
 *                     source_parent_entry = self.state._get_entry(self.source_index,             # <<<<<<<<<<<<<<
 *                                                            path_utf8=old_dirname)
 *                     source_parent_id = source_parent_entry[0][2]
*/
          __pyx_t_12 = __pyx_v_self->state;
          __Pyx_INCREF(__pyx_t_12);
          __pyx_t_15 = __Pyx_PyLong_From_int(__pyx_v_self->source_index); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1225, __pyx_L39_except_error)
          __Pyx_GOTREF(__pyx_t_15);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1226
 *                 except KeyError, _:
 *                     source_parent_entry = self.state._get_entry(self.source_index,
 *                                                            path_utf8=old_dirname)             # <<<<<<<<<<<<<<
 *                     source_parent_id = source_parent_entry[0][2]
 *                 if source_parent_id == entry[0][2]:
*/
          __pyx_t_9 = 0;
          {
            PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_12, __pyx_t_15};
            __pyx_t_14 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1225, __pyx_L39_except_error)
            __Pyx_GOTREF(__pyx_t_14);
            if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_path_utf8, __pyx_v_old_dirname, __pyx_t_14, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 1225, __pyx_L39_except_error)
            __pyx_t_1 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_get_entry, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_14);
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1225, __pyx_L39_except_error)
            __Pyx_GOTREF(__pyx_t_1);
          }
          __pyx_v_source_parent_entry = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1227
 *                     source_parent_entry = self.state._get_entry(self.source_index,
 *                                                            path_utf8=old_dirname)
 *                     source_parent_id = source_parent_entry[0][2]             # <<<<<<<<<<<<<<
 *                 if source_parent_id == entry[0][2]:
 *                     # This is the root, so the parent is None
*/
          __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_source_parent_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1227, __pyx_L39_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_14 = __Pyx_GetItemInt(__pyx_t_1, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1227, __pyx_L39_except_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF_SET(__pyx_v_source_parent_id, __pyx_t_14);
          __pyx_t_14 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          goto __pyx_L38_exception_handled;
        }
        goto __pyx_L39_except_error;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1222
 *                 source_parent_id = self.last_source_parent[1]
 *             else:
 *                 try:             # <<<<<<<<<<<<<<
 *                     source_parent_id = self.old_dirname_to_file_id[old_dirname]
 *                 except KeyError, _:
*/
        __pyx_L39_except_error:;
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
        goto __pyx_L1_error;
        __pyx_L38_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
        __pyx_L42_try_end:;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1228
 *                                                            path_utf8=old_dirname)
 *                     source_parent_id = source_parent_entry[0][2]
 *                 if source_parent_id == entry[0][2]:             # <<<<<<<<<<<<<<
 *                     # This is the root, so the parent is None
 *                     source_parent_id = None
*/
      __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1228, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_13, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1228, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_13 = PyObject_RichCompare(__pyx_v_source_parent_id, __pyx_t_17, Py_EQ); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1228, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1228, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1230
 *                 if source_parent_id == entry[0][2]:
 *                     # This is the root, so the parent is None
 *                     source_parent_id = None             # <<<<<<<<<<<<<<
 *                 else:
 *                     self.last_source_parent[0] = old_dirname
*/
        __Pyx_INCREF(Py_None);
        __Pyx_DECREF_SET(__pyx_v_source_parent_id, Py_None);

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1228
 *                                                            path_utf8=old_dirname)
 *                     source_parent_id = source_parent_entry[0][2]
 *                 if source_parent_id == entry[0][2]:             # <<<<<<<<<<<<<<
 *                     # This is the root, so the parent is None
 *                     source_parent_id = None
*/
        goto __pyx_L45;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1232
 *                     source_parent_id = None
 *                 else:
 *                     self.last_source_parent[0] = old_dirname             # <<<<<<<<<<<<<<
 *                     self.last_source_parent[1] = source_parent_id
 *             new_dirname = entry[0][0]
*/
      /*else*/ {
        if (unlikely((__Pyx_SetItemInt(__pyx_v_self->last_source_parent, 0, __pyx_v_old_dirname, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference) < 0))) __PYX_ERR(0, 1232, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1233
 *                 else:
 *                     self.last_source_parent[0] = old_dirname
 *                     self.last_source_parent[1] = source_parent_id             # <<<<<<<<<<<<<<
 *             new_dirname = entry[0][0]
 *             if entry[0][1] and new_dirname == self.last_target_parent[0]:
*/
        if (unlikely((__Pyx_SetItemInt(__pyx_v_self->last_source_parent, 1, __pyx_v_source_parent_id, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference) < 0))) __PYX_ERR(0, 1233, __pyx_L1_error)
      }
      __pyx_L45:;
    }
    __pyx_L34:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1234
 *                     self.last_source_parent[0] = old_dirname
 *                     self.last_source_parent[1] = source_parent_id
 *             new_dirname = entry[0][0]             # <<<<<<<<<<<<<<
 *             if entry[0][1] and new_dirname == self.last_target_parent[0]:
 *                 # use a cached hit for non-root target entries.
*/
    __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_13, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_v_new_dirname = __pyx_t_17;
    __pyx_t_17 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1235
 *                     self.last_source_parent[1] = source_parent_id
 *             new_dirname = entry[0][0]
 *             if entry[0][1] and new_dirname == self.last_target_parent[0]:             # <<<<<<<<<<<<<<
 *                 # use a cached hit for non-root target entries.
 *                 target_parent_id = self.last_target_parent[1]
*/
    __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_13 = __Pyx_GetItemInt(__pyx_t_17, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 1235, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (__pyx_t_10) {
    } else {
      __pyx_t_2 = __pyx_t_10;
      goto __pyx_L47_bool_binop_done;
    }
    __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_self->last_target_parent, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_17 = PyObject_RichCompare(__pyx_v_new_dirname, __pyx_t_13, Py_EQ); __Pyx_XGOTREF(__pyx_t_17); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1235, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_17); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 1235, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_2 = __pyx_t_10;
    __pyx_L47_bool_binop_done:;
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1237
 *             if entry[0][1] and new_dirname == self.last_target_parent[0]:
 *                 # use a cached hit for non-root target entries.
 *                 target_parent_id = self.last_target_parent[1]             # <<<<<<<<<<<<<<
 *             else:
 *                 try:
*/
      __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_self->last_target_parent, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_v_target_parent_id = __pyx_t_17;
      __pyx_t_17 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1235
 *                     self.last_source_parent[1] = source_parent_id
 *             new_dirname = entry[0][0]
 *             if entry[0][1] and new_dirname == self.last_target_parent[0]:             # <<<<<<<<<<<<<<
 *                 # use a cached hit for non-root target entries.
 *                 target_parent_id = self.last_target_parent[1]
*/
      goto __pyx_L46;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1239
 *                 target_parent_id = self.last_target_parent[1]
 *             else:
 *                 try:             # <<<<<<<<<<<<<<
 *                     target_parent_id = self.new_dirname_to_file_id[new_dirname]
 *                 except KeyError, _:
*/
    /*else*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
        __Pyx_XGOTREF(__pyx_t_20);
        __Pyx_XGOTREF(__pyx_t_19);
        __Pyx_XGOTREF(__pyx_t_18);
        /*try:*/ {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1240
 *             else:
 *                 try:
 *                     target_parent_id = self.new_dirname_to_file_id[new_dirname]             # <<<<<<<<<<<<<<
 *                 except KeyError, _:
 *                     # TODO: We don't always need to do the lookup, because the
*/
          __pyx_t_17 = __Pyx_PyObject_GetItem(__pyx_v_self->new_dirname_to_file_id, __pyx_v_new_dirname); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1240, __pyx_L49_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_v_target_parent_id = __pyx_t_17;
          __pyx_t_17 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1239
 *                 target_parent_id = self.last_target_parent[1]
 *             else:
 *                 try:             # <<<<<<<<<<<<<<
 *                     target_parent_id = self.new_dirname_to_file_id[new_dirname]
 *                 except KeyError, _:
*/
        }
        __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
        goto __pyx_L54_try_end;
        __pyx_L49_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1241
 *                 try:
 *                     target_parent_id = self.new_dirname_to_file_id[new_dirname]
 *                 except KeyError, _:             # <<<<<<<<<<<<<<
 *                     # TODO: We don't always need to do the lookup, because the
 *                     #       parent entry will be the same as the source entry.
*/
        __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_KeyError))));
        if (__pyx_t_6) {
          __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._process_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_17, &__pyx_t_13, &__pyx_t_7) < 0) __PYX_ERR(0, 1241, __pyx_L51_except_error)
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_13);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1244
 *                     # TODO: We don't always need to do the lookup, because the
 *                     #       parent entry will be the same as the source entry.
 *                     target_parent_entry = self.state._get_entry(self.target_index,             # <<<<<<<<<<<<<<
 *                                                            path_utf8=new_dirname)
 *                     if target_parent_entry == (None, None):
*/
          __pyx_t_1 = __pyx_v_self->state;
          __Pyx_INCREF(__pyx_t_1);
          __pyx_t_15 = __Pyx_PyLong_From_int(__pyx_v_self->target_index); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1244, __pyx_L51_except_error)
          __Pyx_GOTREF(__pyx_t_15);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1245
 *                     #       parent entry will be the same as the source entry.
 *                     target_parent_entry = self.state._get_entry(self.target_index,
 *                                                            path_utf8=new_dirname)             # <<<<<<<<<<<<<<
 *                     if target_parent_entry == (None, None):
 *                         raise AssertionError(
*/
          __pyx_t_9 = 0;
          {
            PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_1, __pyx_t_15};
            __pyx_t_12 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1244, __pyx_L51_except_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_path_utf8, __pyx_v_new_dirname, __pyx_t_12, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 1244, __pyx_L51_except_error)
            __pyx_t_14 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_get_entry, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_12);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1244, __pyx_L51_except_error)
            __Pyx_GOTREF(__pyx_t_14);
          }
          __pyx_v_target_parent_entry = __pyx_t_14;
          __pyx_t_14 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1246
 *                     target_parent_entry = self.state._get_entry(self.target_index,
 *                                                            path_utf8=new_dirname)
 *                     if target_parent_entry == (None, None):             # <<<<<<<<<<<<<<
 *                         raise AssertionError(
 *                             "Could not find target parent in wt: %s\nparent of: %s"
*/
          __pyx_t_14 = PyObject_RichCompare(__pyx_v_target_parent_entry, __pyx_mstate_global->__pyx_tuple[0], Py_EQ); __Pyx_XGOTREF(__pyx_t_14); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1246, __pyx_L51_except_error)
          __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1246, __pyx_L51_except_error)
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          if (unlikely(__pyx_t_2)) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1247
 *                                                            path_utf8=new_dirname)
 *                     if target_parent_entry == (None, None):
 *                         raise AssertionError(             # <<<<<<<<<<<<<<
 *                             "Could not find target parent in wt: %s\nparent of: %s"
 *                             % (new_dirname, entry))
*/
            __pyx_t_12 = NULL;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1249
 *                         raise AssertionError(
 *                             "Could not find target parent in wt: %s\nparent of: %s"
 *                             % (new_dirname, entry))             # <<<<<<<<<<<<<<
 *                     target_parent_id = target_parent_entry[0][2]
 *                 if target_parent_id == entry[0][2]:
*/
            __pyx_t_15 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_v_new_dirname), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1249, __pyx_L51_except_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_1 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_v_entry), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1249, __pyx_L51_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_21[0] = __pyx_mstate_global->__pyx_kp_u_Could_not_find_target_parent_in;
            __pyx_t_21[1] = __pyx_t_15;
            __pyx_t_21[2] = __pyx_mstate_global->__pyx_kp_u_parent_of;
            __pyx_t_21[3] = __pyx_t_1;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1248
 *                     if target_parent_entry == (None, None):
 *                         raise AssertionError(
 *                             "Could not find target parent in wt: %s\nparent of: %s"             # <<<<<<<<<<<<<<
 *                             % (new_dirname, entry))
 *                     target_parent_id = target_parent_entry[0][2]
*/
            __pyx_t_8 = __Pyx_PyUnicode_Join(__pyx_t_21, 4, 36 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_15) + 12 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1));
            if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1248, __pyx_L51_except_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_9 = 1;
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_t_8};
              __pyx_t_14 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1247, __pyx_L51_except_error)
              __Pyx_GOTREF(__pyx_t_14);
            }
            __Pyx_Raise(__pyx_t_14, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __PYX_ERR(0, 1247, __pyx_L51_except_error)

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1246
 *                     target_parent_entry = self.state._get_entry(self.target_index,
 *                                                            path_utf8=new_dirname)
 *                     if target_parent_entry == (None, None):             # <<<<<<<<<<<<<<
 *                         raise AssertionError(
 *                             "Could not find target parent in wt: %s\nparent of: %s"
*/
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1250
 *                             "Could not find target parent in wt: %s\nparent of: %s"
 *                             % (new_dirname, entry))
 *                     target_parent_id = target_parent_entry[0][2]             # <<<<<<<<<<<<<<
 *                 if target_parent_id == entry[0][2]:
 *                     # This is the root, so the parent is None
*/
          __pyx_t_14 = __Pyx_GetItemInt(__pyx_v_target_parent_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1250, __pyx_L51_except_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_8 = __Pyx_GetItemInt(__pyx_t_14, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1250, __pyx_L51_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_XDECREF_SET(__pyx_v_target_parent_id, __pyx_t_8);
          __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          goto __pyx_L50_exception_handled;
        }
        goto __pyx_L51_except_error;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1239
 *                 target_parent_id = self.last_target_parent[1]
 *             else:
 *                 try:             # <<<<<<<<<<<<<<
 *                     target_parent_id = self.new_dirname_to_file_id[new_dirname]
 *                 except KeyError, _:
*/
        __pyx_L51_except_error:;
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
        goto __pyx_L1_error;
        __pyx_L50_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
        __pyx_L54_try_end:;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1251
 *                             % (new_dirname, entry))
 *                     target_parent_id = target_parent_entry[0][2]
 *                 if target_parent_id == entry[0][2]:             # <<<<<<<<<<<<<<
 *                     # This is the root, so the parent is None
 *                     target_parent_id = None
*/
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1251, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_13 = __Pyx_GetItemInt(__pyx_t_7, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1251, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = PyObject_RichCompare(__pyx_v_target_parent_id, __pyx_t_13, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1251, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1251, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1253
 *                 if target_parent_id == entry[0][2]:
 *                     # This is the root, so the parent is None
 *                     target_parent_id = None             # <<<<<<<<<<<<<<
 *                 else:
 *                     self.last_target_parent[0] = new_dirname
*/
        __Pyx_INCREF(Py_None);
        __Pyx_DECREF_SET(__pyx_v_target_parent_id, Py_None);

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1251
 *                             % (new_dirname, entry))
 *                     target_parent_id = target_parent_entry[0][2]
 *                 if target_parent_id == entry[0][2]:             # <<<<<<<<<<<<<<
 *                     # This is the root, so the parent is None
 *                     target_parent_id = None
*/
        goto __pyx_L58;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1255
 *                     target_parent_id = None
 *                 else:
 *                     self.last_target_parent[0] = new_dirname             # <<<<<<<<<<<<<<
 *                     self.last_target_parent[1] = target_parent_id
 * 
*/
      /*else*/ {
        if (unlikely((__Pyx_SetItemInt(__pyx_v_self->last_target_parent, 0, __pyx_v_new_dirname, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference) < 0))) __PYX_ERR(0, 1255, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1256
 *                 else:
 *                     self.last_target_parent[0] = new_dirname
 *                     self.last_target_parent[1] = target_parent_id             # <<<<<<<<<<<<<<
 * 
 *             source_exec = source_details[3]
*/
        if (unlikely((__Pyx_SetItemInt(__pyx_v_self->last_target_parent, 1, __pyx_v_target_parent_id, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference) < 0))) __PYX_ERR(0, 1256, __pyx_L1_error)
      }
      __pyx_L58:;
    }
    __pyx_L46:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1258
 *                     self.last_target_parent[1] = target_parent_id
 * 
 *             source_exec = source_details[3]             # <<<<<<<<<<<<<<
 *             changed = (content_change
 *                 or source_parent_id != target_parent_id
*/
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_source_details, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_v_source_exec = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1259
 * 
 *             source_exec = source_details[3]
 *             changed = (content_change             # <<<<<<<<<<<<<<
 *                 or source_parent_id != target_parent_id
 *                 or old_basename != entry[0][1]
*/
    if (!__pyx_v_content_change) {
    } else {
      __pyx_t_13 = __Pyx_PyLong_From_int(__pyx_v_content_change); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1259, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_7 = __pyx_t_13;
      __pyx_t_13 = 0;
      goto __pyx_L59_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1260
 *             source_exec = source_details[3]
 *             changed = (content_change
 *                 or source_parent_id != target_parent_id             # <<<<<<<<<<<<<<
 *                 or old_basename != entry[0][1]
 *                 or source_exec != target_exec
*/
    __pyx_t_13 = PyObject_RichCompare(__pyx_v_source_parent_id, __pyx_v_target_parent_id, Py_NE); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1260, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1260, __pyx_L1_error)
    if (!__pyx_t_2) {
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    } else {
      __Pyx_INCREF(__pyx_t_13);
      __pyx_t_7 = __pyx_t_13;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      goto __pyx_L59_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1261
 *             changed = (content_change
 *                 or source_parent_id != target_parent_id
 *                 or old_basename != entry[0][1]             # <<<<<<<<<<<<<<
 *                 or source_exec != target_exec
 *                 )
*/
    __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_13, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = PyObject_RichCompare(__pyx_v_old_basename, __pyx_t_17, Py_NE); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1261, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1261, __pyx_L1_error)
    if (!__pyx_t_2) {
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    } else {
      __Pyx_INCREF(__pyx_t_13);
      __pyx_t_7 = __pyx_t_13;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      goto __pyx_L59_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1262
 *                 or source_parent_id != target_parent_id
 *                 or old_basename != entry[0][1]
 *                 or source_exec != target_exec             # <<<<<<<<<<<<<<
 *                 )
 *             if not changed and not self.include_unchanged:
*/
    __pyx_t_13 = PyObject_RichCompare(__pyx_v_source_exec, __pyx_v_target_exec, Py_NE); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1262, __pyx_L1_error)
    __Pyx_INCREF(__pyx_t_13);
    __pyx_t_7 = __pyx_t_13;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_L59_bool_binop_done:;
    __pyx_v_changed = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1264
 *                 or source_exec != target_exec
 *                 )
 *             if not changed and not self.include_unchanged:             # <<<<<<<<<<<<<<
 *                 return None, False
 *             else:
*/
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 1264, __pyx_L1_error)
    __pyx_t_5 = (!__pyx_t_10);
    if (__pyx_t_5) {
    } else {
      __pyx_t_2 = __pyx_t_5;
      goto __pyx_L64_bool_binop_done;
    }
    __pyx_t_5 = (!(__pyx_v_self->include_unchanged != 0));
    __pyx_t_2 = __pyx_t_5;
    __pyx_L64_bool_binop_done:;
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1265
 *                 )
 *             if not changed and not self.include_unchanged:
 *                 return None, False             # <<<<<<<<<<<<<<
 *             else:
 *                 if old_path is None:
*/
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_mstate_global->__pyx_tuple[1]);
      __pyx_r = __pyx_mstate_global->__pyx_tuple[1];
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1264
 *                 or source_exec != target_exec
 *                 )
 *             if not changed and not self.include_unchanged:             # <<<<<<<<<<<<<<
 *                 return None, False
 *             else:
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1267
 *                 return None, False
 *             else:
 *                 if old_path is None:             # <<<<<<<<<<<<<<
 *                     path = self.pathjoin(old_dirname, old_basename)
 *                     old_path = path
*/
    /*else*/ {
      __pyx_t_2 = (__pyx_v_old_path == Py_None);
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1268
 *             else:
 *                 if old_path is None:
 *                     path = self.pathjoin(old_dirname, old_basename)             # <<<<<<<<<<<<<<
 *                     old_path = path
 *                     old_path_u = self.utf8_decode(old_path, 'surrogateescape')[0]
*/
        __pyx_t_13 = NULL;
        __Pyx_INCREF(__pyx_v_self->pathjoin);
        __pyx_t_17 = __pyx_v_self->pathjoin; 
        __pyx_t_9 = 1;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_17))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_17);
          assert(__pyx_t_13);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_17);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_17, __pyx__function);
          __pyx_t_9 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_13, __pyx_v_old_dirname, __pyx_v_old_basename};
          __pyx_t_7 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_17, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1268, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
        }
        __Pyx_DECREF_SET(__pyx_v_path, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1269
 *                 if old_path is None:
 *                     path = self.pathjoin(old_dirname, old_basename)
 *                     old_path = path             # <<<<<<<<<<<<<<
 *                     old_path_u = self.utf8_decode(old_path, 'surrogateescape')[0]
 *                     path_u = old_path_u
*/
        __Pyx_INCREF(__pyx_v_path);
        __Pyx_DECREF_SET(__pyx_v_old_path, __pyx_v_path);

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1270
 *                     path = self.pathjoin(old_dirname, old_basename)
 *                     old_path = path
 *                     old_path_u = self.utf8_decode(old_path, 'surrogateescape')[0]             # <<<<<<<<<<<<<<
 *                     path_u = old_path_u
 *                 else:
*/
        __pyx_t_17 = NULL;
        __Pyx_INCREF(__pyx_v_self->utf8_decode);
        __pyx_t_13 = __pyx_v_self->utf8_decode; 
        __pyx_t_9 = 1;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_13);
          assert(__pyx_t_17);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_13);
          __Pyx_INCREF(__pyx_t_17);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_13, __pyx__function);
          __pyx_t_9 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_17, __pyx_v_old_path, __pyx_mstate_global->__pyx_n_u_surrogateescape};
          __pyx_t_7 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_13, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1270, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
        }
        __pyx_t_13 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1270, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_old_path_u = __pyx_t_13;
        __pyx_t_13 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1271
 *                     old_path = path
 *                     old_path_u = self.utf8_decode(old_path, 'surrogateescape')[0]
 *                     path_u = old_path_u             # <<<<<<<<<<<<<<
 *                 else:
 *                     old_path_u = self.utf8_decode(old_path, 'surrogateescape')[0]
*/
        __Pyx_INCREF(__pyx_v_old_path_u);
        __pyx_v_path_u = __pyx_v_old_path_u;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1267
 *                 return None, False
 *             else:
 *                 if old_path is None:             # <<<<<<<<<<<<<<
 *                     path = self.pathjoin(old_dirname, old_basename)
 *                     old_path = path
*/
        goto __pyx_L66;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1273
 *                     path_u = old_path_u
 *                 else:
 *                     old_path_u = self.utf8_decode(old_path, 'surrogateescape')[0]             # <<<<<<<<<<<<<<
 *                     if old_path == path:
 *                         path_u = old_path_u
*/
      /*else*/ {
        __pyx_t_7 = NULL;
        __Pyx_INCREF(__pyx_v_self->utf8_decode);
        __pyx_t_17 = __pyx_v_self->utf8_decode; 
        __pyx_t_9 = 1;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_17))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_17);
          assert(__pyx_t_7);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_17);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_17, __pyx__function);
          __pyx_t_9 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_v_old_path, __pyx_mstate_global->__pyx_n_u_surrogateescape};
          __pyx_t_13 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_17, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1273, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
        }
        __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_13, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1273, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_v_old_path_u = __pyx_t_17;
        __pyx_t_17 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1274
 *                 else:
 *                     old_path_u = self.utf8_decode(old_path, 'surrogateescape')[0]
 *                     if old_path == path:             # <<<<<<<<<<<<<<
 *                         path_u = old_path_u
 *                     else:
*/
        __pyx_t_17 = PyObject_RichCompare(__pyx_v_old_path, __pyx_v_path, Py_EQ); __Pyx_XGOTREF(__pyx_t_17); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1274, __pyx_L1_error)
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_17); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1274, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1275
 *                     old_path_u = self.utf8_decode(old_path, 'surrogateescape')[0]
 *                     if old_path == path:
 *                         path_u = old_path_u             # <<<<<<<<<<<<<<
 *                     else:
 *                         path_u = self.utf8_decode(path, 'surrogateescape')[0]
*/
          __Pyx_INCREF(__pyx_v_old_path_u);
          __pyx_v_path_u = __pyx_v_old_path_u;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1274
 *                 else:
 *                     old_path_u = self.utf8_decode(old_path, 'surrogateescape')[0]
 *                     if old_path == path:             # <<<<<<<<<<<<<<
 *                         path_u = old_path_u
 *                     else:
*/
          goto __pyx_L67;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1277
 *                         path_u = old_path_u
 *                     else:
 *                         path_u = self.utf8_decode(path, 'surrogateescape')[0]             # <<<<<<<<<<<<<<
 *                 source_kind = _minikind_to_kind(source_minikind)
 *                 return InventoryTreeChange(entry[0][2],
*/
        /*else*/ {
          __pyx_t_13 = NULL;
          __Pyx_INCREF(__pyx_v_self->utf8_decode);
          __pyx_t_7 = __pyx_v_self->utf8_decode; 
          __pyx_t_9 = 1;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_7);
            assert(__pyx_t_13);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
            __pyx_t_9 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_13, __pyx_v_path, __pyx_mstate_global->__pyx_n_u_surrogateescape};
            __pyx_t_17 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1277, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
          }
          __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_17, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1277, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          __pyx_v_path_u = __pyx_t_7;
          __pyx_t_7 = 0;
        }
        __pyx_L67:;
      }
      __pyx_L66:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1278
 *                     else:
 *                         path_u = self.utf8_decode(path, 'surrogateescape')[0]
 *                 source_kind = _minikind_to_kind(source_minikind)             # <<<<<<<<<<<<<<
 *                 return InventoryTreeChange(entry[0][2],
 *                        (old_path_u, path_u),
*/
      __pyx_t_7 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_to_kind(__pyx_v_source_minikind); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_v_source_kind = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1279
 *                         path_u = self.utf8_decode(path, 'surrogateescape')[0]
 *                 source_kind = _minikind_to_kind(source_minikind)
 *                 return InventoryTreeChange(entry[0][2],             # <<<<<<<<<<<<<<
 *                        (old_path_u, path_u),
 *                        content_change,
*/
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_17 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_InventoryTreeChange); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_14 = __Pyx_GetItemInt(__pyx_t_8, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1280
 *                 source_kind = _minikind_to_kind(source_minikind)
 *                 return InventoryTreeChange(entry[0][2],
 *                        (old_path_u, path_u),             # <<<<<<<<<<<<<<
 *                        content_change,
 *                        (True, True),
*/
      __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1280, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_v_old_path_u);
      __Pyx_GIVEREF(__pyx_v_old_path_u);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_old_path_u) != (0)) __PYX_ERR(0, 1280, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_path_u);
      __Pyx_GIVEREF(__pyx_v_path_u);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_v_path_u) != (0)) __PYX_ERR(0, 1280, __pyx_L1_error);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1281
 *                 return InventoryTreeChange(entry[0][2],
 *                        (old_path_u, path_u),
 *                        content_change,             # <<<<<<<<<<<<<<
 *                        (True, True),
 *                        (source_parent_id, target_parent_id),
*/
      __pyx_t_12 = __Pyx_PyLong_From_int(__pyx_v_content_change); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1283
 *                        content_change,
 *                        (True, True),
 *                        (source_parent_id, target_parent_id),             # <<<<<<<<<<<<<<
 *                        (self.utf8_decode(old_basename, 'surrogateescape')[0], self.utf8_decode(entry[0][1], 'surrogateescape')[0]),
 *                        (source_kind, target_kind),
*/
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1283, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_source_parent_id);
      __Pyx_GIVEREF(__pyx_v_source_parent_id);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_source_parent_id) != (0)) __PYX_ERR(0, 1283, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_target_parent_id);
      __Pyx_GIVEREF(__pyx_v_target_parent_id);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_target_parent_id) != (0)) __PYX_ERR(0, 1283, __pyx_L1_error);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1284
 *                        (True, True),
 *                        (source_parent_id, target_parent_id),
 *                        (self.utf8_decode(old_basename, 'surrogateescape')[0], self.utf8_decode(entry[0][1], 'surrogateescape')[0]),             # <<<<<<<<<<<<<<
 *                        (source_kind, target_kind),
 *                        (source_exec, target_exec)), changed
*/
      __pyx_t_3 = NULL;
      __Pyx_INCREF(__pyx_v_self->utf8_decode);
      __pyx_t_22 = __pyx_v_self->utf8_decode; 
      __pyx_t_9 = 1;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_22))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_22);
        assert(__pyx_t_3);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_22);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_22, __pyx__function);
        __pyx_t_9 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_old_basename, __pyx_mstate_global->__pyx_n_u_surrogateescape};
        __pyx_t_15 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_22, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
        if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1284, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
      }
      __pyx_t_22 = __Pyx_GetItemInt(__pyx_t_15, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_22);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_t_3 = NULL;
      __Pyx_INCREF(__pyx_v_self->utf8_decode);
      __pyx_t_23 = __pyx_v_self->utf8_decode; 
      __pyx_t_24 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_24);
      __pyx_t_25 = __Pyx_GetItemInt(__pyx_t_24, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_25);
      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
      __pyx_t_9 = 1;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_23))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_23);
        assert(__pyx_t_3);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_23);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_23, __pyx__function);
        __pyx_t_9 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_t_25, __pyx_mstate_global->__pyx_n_u_surrogateescape};
        __pyx_t_15 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_23, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
        __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
        if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1284, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
      }
      __pyx_t_23 = __Pyx_GetItemInt(__pyx_t_15, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_23);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_GIVEREF(__pyx_t_22);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_22) != (0)) __PYX_ERR(0, 1284, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_23);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_23) != (0)) __PYX_ERR(0, 1284, __pyx_L1_error);
      __pyx_t_22 = 0;
      __pyx_t_23 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1285
 *                        (source_parent_id, target_parent_id),
 *                        (self.utf8_decode(old_basename, 'surrogateescape')[0], self.utf8_decode(entry[0][1], 'surrogateescape')[0]),
 *                        (source_kind, target_kind),             # <<<<<<<<<<<<<<
 *                        (source_exec, target_exec)), changed
 *         elif source_minikind == b'a' and _versioned_minikind(target_minikind):
*/
      __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_23);
      __Pyx_INCREF(__pyx_v_source_kind);
      __Pyx_GIVEREF(__pyx_v_source_kind);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_v_source_kind) != (0)) __PYX_ERR(0, 1285, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_target_kind);
      __Pyx_GIVEREF(__pyx_v_target_kind);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_v_target_kind) != (0)) __PYX_ERR(0, 1285, __pyx_L1_error);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1286
 *                        (self.utf8_decode(old_basename, 'surrogateescape')[0], self.utf8_decode(entry[0][1], 'surrogateescape')[0]),
 *                        (source_kind, target_kind),
 *                        (source_exec, target_exec)), changed             # <<<<<<<<<<<<<<
 *         elif source_minikind == b'a' and _versioned_minikind(target_minikind):
 *             # looks like a new file
*/
      __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_22);
      __Pyx_INCREF(__pyx_v_source_exec);
      __Pyx_GIVEREF(__pyx_v_source_exec);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_v_source_exec) != (0)) __PYX_ERR(0, 1286, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_target_exec);
      __Pyx_GIVEREF(__pyx_v_target_exec);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_v_target_exec) != (0)) __PYX_ERR(0, 1286, __pyx_L1_error);
      __pyx_t_9 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_13))) {
        __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_13);
        assert(__pyx_t_17);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_17);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_13, __pyx__function);
        __pyx_t_9 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[9] = {__pyx_t_17, __pyx_t_14, __pyx_t_8, __pyx_t_12, __pyx_mstate_global->__pyx_tuple[2], __pyx_t_1, __pyx_t_15, __pyx_t_23, __pyx_t_22};
        __pyx_t_7 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_13, __pyx_callargs+__pyx_t_9, (9-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
        __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1279, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1279
 *                         path_u = self.utf8_decode(path, 'surrogateescape')[0]
 *                 source_kind = _minikind_to_kind(source_minikind)
 *                 return InventoryTreeChange(entry[0][2],             # <<<<<<<<<<<<<<
 *                        (old_path_u, path_u),
 *                        content_change,
*/
      __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_GIVEREF(__pyx_t_7);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_7) != (0)) __PYX_ERR(0, 1279, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_changed);
      __Pyx_GIVEREF(__pyx_v_changed);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_v_changed) != (0)) __PYX_ERR(0, 1279, __pyx_L1_error);
      __pyx_t_7 = 0;
      __pyx_r = __pyx_t_13;
      __pyx_t_13 = 0;
      goto __pyx_L0;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1118
 *         # 0.000006 seconds per call.
 *         source_minikind = _minikind_from_string(source_details[0])
 *         if ((_versioned_minikind(source_minikind) or source_minikind == b'r')             # <<<<<<<<<<<<<<
 *             and _versioned_minikind(target_minikind)):
 *             # claimed content in both: diff
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1287
 *                        (source_kind, target_kind),
 *                        (source_exec, target_exec)), changed
 *         elif source_minikind == b'a' and _versioned_minikind(target_minikind):             # <<<<<<<<<<<<<<
 *             # looks like a new file
 *             path = self.pathjoin(entry[0][0], entry[0][1])
*/
  __pyx_t_5 = (__pyx_v_source_minikind == 'a');
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L68_bool_binop_done;
  }
  __pyx_t_6 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__versioned_minikind(__pyx_v_target_minikind); if (unlikely(__pyx_t_6 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1287, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_6 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L68_bool_binop_done:;
  if (__pyx_t_2) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1289
 *         elif source_minikind == b'a' and _versioned_minikind(target_minikind):
 *             # looks like a new file
 *             path = self.pathjoin(entry[0][0], entry[0][1])             # <<<<<<<<<<<<<<
 *             # parent id is the entry for the path in the target tree
 *             # TODO: these are the same for an entire directory: cache em.
*/
    __pyx_t_7 = NULL;
    __Pyx_INCREF(__pyx_v_self->pathjoin);
    __pyx_t_22 = __pyx_v_self->pathjoin; 
    __pyx_t_23 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_15 = __Pyx_GetItemInt(__pyx_t_23, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __pyx_t_23 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_23, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __pyx_t_9 = 1;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_22))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_22);
      assert(__pyx_t_7);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_22);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_22, __pyx__function);
      __pyx_t_9 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_t_15, __pyx_t_1};
      __pyx_t_13 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_22, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
      if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
    }
    __pyx_v_path = __pyx_t_13;
    __pyx_t_13 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1292
 *             # parent id is the entry for the path in the target tree
 *             # TODO: these are the same for an entire directory: cache em.
 *             parent_entry = self.state._get_entry(self.target_index,             # <<<<<<<<<<<<<<
 *                                                  path_utf8=entry[0][0])
 *             if parent_entry is None:
*/
    __pyx_t_22 = __pyx_v_self->state;
    __Pyx_INCREF(__pyx_t_22);
    __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->target_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1293
 *             # TODO: these are the same for an entire directory: cache em.
 *             parent_entry = self.state._get_entry(self.target_index,
 *                                                  path_utf8=entry[0][0])             # <<<<<<<<<<<<<<
 *             if parent_entry is None:
 *                 raise DirstateCorrupt(self.state,
*/
    __pyx_t_15 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_15, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_9 = 0;
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_22, __pyx_t_1};
      __pyx_t_15 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1292, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_path_utf8, __pyx_t_7, __pyx_t_15, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 1292, __pyx_L1_error)
      __pyx_t_13 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_get_entry, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_15);
      __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1292, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
    }
    __pyx_v_parent_entry = __pyx_t_13;
    __pyx_t_13 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1294
 *             parent_entry = self.state._get_entry(self.target_index,
 *                                                  path_utf8=entry[0][0])
 *             if parent_entry is None:             # <<<<<<<<<<<<<<
 *                 raise DirstateCorrupt(self.state,
 *                     "We could not find the parent entry in index %d"
*/
    __pyx_t_2 = (__pyx_v_parent_entry == Py_None);
    if (unlikely(__pyx_t_2)) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1295
 *                                                  path_utf8=entry[0][0])
 *             if parent_entry is None:
 *                 raise DirstateCorrupt(self.state,             # <<<<<<<<<<<<<<
 *                     "We could not find the parent entry in index %d"
 *                     " for the entry: %s"
*/
      __pyx_t_15 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_DirstateCorrupt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1295, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1298
 *                     "We could not find the parent entry in index %d"
 *                     " for the entry: %s"
 *                     % (self.target_index, entry[0]))             # <<<<<<<<<<<<<<
 *             parent_id = parent_entry[0][2]
 *             if parent_id == entry[0][2]:
*/
      __pyx_t_1 = __Pyx_PyUnicode_From_int(__pyx_v_self->target_index, 0, ' ', 'd'); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_22 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_22);
      __pyx_t_23 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Str(__pyx_t_22), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_23);
      __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
      __pyx_t_21[0] = __pyx_mstate_global->__pyx_kp_u_We_could_not_find_the_parent_ent;
      __pyx_t_21[1] = __pyx_t_1;
      __pyx_t_21[2] = __pyx_mstate_global->__pyx_kp_u_for_the_entry;
      __pyx_t_21[3] = __pyx_t_23;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1296
 *             if parent_entry is None:
 *                 raise DirstateCorrupt(self.state,
 *                     "We could not find the parent entry in index %d"             # <<<<<<<<<<<<<<
 *                     " for the entry: %s"
 *                     % (self.target_index, entry[0]))
*/
      __pyx_t_22 = __Pyx_PyUnicode_Join(__pyx_t_21, 4, 44 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1) + 16 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_23), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_23));
      if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1296, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_22);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
      __pyx_t_9 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_7);
        assert(__pyx_t_15);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_15);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
        __pyx_t_9 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_15, __pyx_v_self->state, __pyx_t_22};
        __pyx_t_13 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1295, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
      }
      __Pyx_Raise(__pyx_t_13, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __PYX_ERR(0, 1295, __pyx_L1_error)

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1294
 *             parent_entry = self.state._get_entry(self.target_index,
 *                                                  path_utf8=entry[0][0])
 *             if parent_entry is None:             # <<<<<<<<<<<<<<
 *                 raise DirstateCorrupt(self.state,
 *                     "We could not find the parent entry in index %d"
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1299
 *                     " for the entry: %s"
 *                     % (self.target_index, entry[0]))
 *             parent_id = parent_entry[0][2]             # <<<<<<<<<<<<<<
 *             if parent_id == entry[0][2]:
 *                 parent_id = None
*/
    __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_parent_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_13, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_v_parent_id = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1300
 *                     % (self.target_index, entry[0]))
 *             parent_id = parent_entry[0][2]
 *             if parent_id == entry[0][2]:             # <<<<<<<<<<<<<<
 *                 parent_id = None
 *             if path_info is not None:
*/
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_13 = __Pyx_GetItemInt(__pyx_t_7, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyObject_RichCompare(__pyx_v_parent_id, __pyx_t_13, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1300, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1300, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1301
 *             parent_id = parent_entry[0][2]
 *             if parent_id == entry[0][2]:
 *                 parent_id = None             # <<<<<<<<<<<<<<
 *             if path_info is not None:
 *                 # Present on disk:
*/
      __Pyx_INCREF(Py_None);
      __Pyx_DECREF_SET(__pyx_v_parent_id, Py_None);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1300
 *                     % (self.target_index, entry[0]))
 *             parent_id = parent_entry[0][2]
 *             if parent_id == entry[0][2]:             # <<<<<<<<<<<<<<
 *                 parent_id = None
 *             if path_info is not None:
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1302
 *             if parent_id == entry[0][2]:
 *                 parent_id = None
 *             if path_info is not None:             # <<<<<<<<<<<<<<
 *                 # Present on disk:
 *                 if self.use_filesystem_for_exec:
*/
    __pyx_t_2 = (__pyx_v_path_info != Py_None);
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1304
 *             if path_info is not None:
 *                 # Present on disk:
 *                 if self.use_filesystem_for_exec:             # <<<<<<<<<<<<<<
 *                     # We need S_ISREG here, because we aren't sure if this
 *                     # is a file or not.
*/
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_self->use_filesystem_for_exec); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1304, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1308
 *                     # is a file or not.
 *                     target_exec = bool(
 *                         S_ISREG(path_info[3].st_mode)             # <<<<<<<<<<<<<<
 *                         and S_IXUSR & path_info[3].st_mode)
 *                 else:
*/
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_path_info, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1308, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_st_mode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1308, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_6 = __Pyx_PyLong_As_int(__pyx_t_13); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1308, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_5 = (S_ISREG(__pyx_t_6) != 0);
        if (__pyx_t_5) {
        } else {
          __pyx_t_2 = __pyx_t_5;
          goto __pyx_L74_bool_binop_done;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1309
 *                     target_exec = bool(
 *                         S_ISREG(path_info[3].st_mode)
 *                         and S_IXUSR & path_info[3].st_mode)             # <<<<<<<<<<<<<<
 *                 else:
 *                     target_exec = target_details[3]
*/
        __pyx_t_13 = __Pyx_PyLong_From_int(S_IXUSR); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1309, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_path_info, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1309, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_st_mode); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1309, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_22);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_And(__pyx_t_13, __pyx_t_22); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1309, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1309, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_2 = __pyx_t_5;
        __pyx_L74_bool_binop_done:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1307
 *                     # We need S_ISREG here, because we aren't sure if this
 *                     # is a file or not.
 *                     target_exec = bool(             # <<<<<<<<<<<<<<
 *                         S_ISREG(path_info[3].st_mode)
 *                         and S_IXUSR & path_info[3].st_mode)
*/
        __pyx_t_7 = __Pyx_PyBool_FromLong((!(!__pyx_t_2))); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1307, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_v_target_exec = __pyx_t_7;
        __pyx_t_7 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1304
 *             if path_info is not None:
 *                 # Present on disk:
 *                 if self.use_filesystem_for_exec:             # <<<<<<<<<<<<<<
 *                     # We need S_ISREG here, because we aren't sure if this
 *                     # is a file or not.
*/
        goto __pyx_L73;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1311
 *                         and S_IXUSR & path_info[3].st_mode)
 *                 else:
 *                     target_exec = target_details[3]             # <<<<<<<<<<<<<<
 *                 return InventoryTreeChange(entry[0][2],
 *                        (None, self.utf8_decode(path, 'surrogateescape')[0]),
*/
      /*else*/ {
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_target_details, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1311, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_v_target_exec = __pyx_t_7;
        __pyx_t_7 = 0;
      }
      __pyx_L73:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1312
 *                 else:
 *                     target_exec = target_details[3]
 *                 return InventoryTreeChange(entry[0][2],             # <<<<<<<<<<<<<<
 *                        (None, self.utf8_decode(path, 'surrogateescape')[0]),
 *                        True,
*/
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_22 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_InventoryTreeChange); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_15 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_23 = __Pyx_GetItemInt(__pyx_t_15, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_23);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1313
 *                     target_exec = target_details[3]
 *                 return InventoryTreeChange(entry[0][2],
 *                        (None, self.utf8_decode(path, 'surrogateescape')[0]),             # <<<<<<<<<<<<<<
 *                        True,
 *                        (False, True),
*/
      __pyx_t_1 = NULL;
      __Pyx_INCREF(__pyx_v_self->utf8_decode);
      __pyx_t_12 = __pyx_v_self->utf8_decode; 
      __pyx_t_9 = 1;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_12);
        assert(__pyx_t_1);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_12, __pyx__function);
        __pyx_t_9 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_v_path, __pyx_mstate_global->__pyx_n_u_surrogateescape};
        __pyx_t_15 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_12, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1313, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
      }
      __pyx_t_12 = __Pyx_GetItemInt(__pyx_t_15, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1313, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1313, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 0, Py_None) != (0)) __PYX_ERR(0, 1313, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_12);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_12) != (0)) __PYX_ERR(0, 1313, __pyx_L1_error);
      __pyx_t_12 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1316
 *                        True,
 *                        (False, True),
 *                        (None, parent_id),             # <<<<<<<<<<<<<<
 *                        (None, self.utf8_decode(entry[0][1], 'surrogateescape')[0]),
 *                        (None, path_info[2]),
*/
      __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1316, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 0, Py_None) != (0)) __PYX_ERR(0, 1316, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_parent_id);
      __Pyx_GIVEREF(__pyx_v_parent_id);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_v_parent_id) != (0)) __PYX_ERR(0, 1316, __pyx_L1_error);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1317
 *                        (False, True),
 *                        (None, parent_id),
 *                        (None, self.utf8_decode(entry[0][1], 'surrogateescape')[0]),             # <<<<<<<<<<<<<<
 *                        (None, path_info[2]),
 *                        (None, target_exec)), True
*/
      __pyx_t_8 = NULL;
      __Pyx_INCREF(__pyx_v_self->utf8_decode);
      __pyx_t_14 = __pyx_v_self->utf8_decode; 
      __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1317, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_25 = __Pyx_GetItemInt(__pyx_t_17, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1317, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_25);
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_t_9 = 1;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_14))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_14);
        assert(__pyx_t_8);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_14);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_14, __pyx__function);
        __pyx_t_9 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_8, __pyx_t_25, __pyx_mstate_global->__pyx_n_u_surrogateescape};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_14, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1317, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __pyx_t_14 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1317, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1317, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, Py_None) != (0)) __PYX_ERR(0, 1317, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_14);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_14) != (0)) __PYX_ERR(0, 1317, __pyx_L1_error);
      __pyx_t_14 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1318
 *                        (None, parent_id),
 *                        (None, self.utf8_decode(entry[0][1], 'surrogateescape')[0]),
 *                        (None, path_info[2]),             # <<<<<<<<<<<<<<
 *                        (None, target_exec)), True
 *             else:
*/
      __pyx_t_14 = __Pyx_GetItemInt(__pyx_v_path_info, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1318, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1318, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_25);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 0, Py_None) != (0)) __PYX_ERR(0, 1318, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_14);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_14) != (0)) __PYX_ERR(0, 1318, __pyx_L1_error);
      __pyx_t_14 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1319
 *                        (None, self.utf8_decode(entry[0][1], 'surrogateescape')[0]),
 *                        (None, path_info[2]),
 *                        (None, target_exec)), True             # <<<<<<<<<<<<<<
 *             else:
 *                 # Its a missing file, report it as such.
*/
      __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1319, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_14, 0, Py_None) != (0)) __PYX_ERR(0, 1319, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_target_exec);
      __Pyx_GIVEREF(__pyx_v_target_exec);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_v_target_exec) != (0)) __PYX_ERR(0, 1319, __pyx_L1_error);
      __pyx_t_9 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_13))) {
        __pyx_t_22 = PyMethod_GET_SELF(__pyx_t_13);
        assert(__pyx_t_22);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_22);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_13, __pyx__function);
        __pyx_t_9 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[9] = {__pyx_t_22, __pyx_t_23, __pyx_t_15, Py_True, __pyx_mstate_global->__pyx_tuple[3], __pyx_t_12, __pyx_t_1, __pyx_t_25, __pyx_t_14};
        __pyx_t_7 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_13, __pyx_callargs+__pyx_t_9, (9-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
        __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1312, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1312
 *                 else:
 *                     target_exec = target_details[3]
 *                 return InventoryTreeChange(entry[0][2],             # <<<<<<<<<<<<<<
 *                        (None, self.utf8_decode(path, 'surrogateescape')[0]),
 *                        True,
*/
      __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_GIVEREF(__pyx_t_7);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_7) != (0)) __PYX_ERR(0, 1312, __pyx_L1_error);
      __Pyx_INCREF(Py_True);
      __Pyx_GIVEREF(Py_True);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 1, Py_True) != (0)) __PYX_ERR(0, 1312, __pyx_L1_error);
      __pyx_t_7 = 0;
      __pyx_r = __pyx_t_13;
      __pyx_t_13 = 0;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1302
 *             if parent_id == entry[0][2]:
 *                 parent_id = None
 *             if path_info is not None:             # <<<<<<<<<<<<<<
 *                 # Present on disk:
 *                 if self.use_filesystem_for_exec:
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1322
 *             else:
 *                 # Its a missing file, report it as such.
 *                 return InventoryTreeChange(entry[0][2],             # <<<<<<<<<<<<<<
 *                        (None, self.utf8_decode(path, 'surrogateescape')[0]),
 *                        False,
*/
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_7 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_InventoryTreeChange); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1322, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_25 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1322, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_25);
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_25, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1322, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1323
 *                 # Its a missing file, report it as such.
 *                 return InventoryTreeChange(entry[0][2],
 *                        (None, self.utf8_decode(path, 'surrogateescape')[0]),             # <<<<<<<<<<<<<<
 *                        False,
 *                        (False, True),
*/
      __pyx_t_12 = NULL;
      __Pyx_INCREF(__pyx_v_self->utf8_decode);
      __pyx_t_15 = __pyx_v_self->utf8_decode; 
      __pyx_t_9 = 1;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_15))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_15);
        assert(__pyx_t_12);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_15);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_15, __pyx__function);
        __pyx_t_9 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_12, __pyx_v_path, __pyx_mstate_global->__pyx_n_u_surrogateescape};
        __pyx_t_25 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_15, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1323, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_25);
      }
      __pyx_t_15 = __Pyx_GetItemInt(__pyx_t_25, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1323, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
      __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1323, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_25);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 0, Py_None) != (0)) __PYX_ERR(0, 1323, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_15);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_15) != (0)) __PYX_ERR(0, 1323, __pyx_L1_error);
      __pyx_t_15 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1326
 *                        False,
 *                        (False, True),
 *                        (None, parent_id),             # <<<<<<<<<<<<<<
 *                        (None, self.utf8_decode(entry[0][1], 'surrogateescape')[0]),
 *                        (None, None),
*/
      __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1326, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 0, Py_None) != (0)) __PYX_ERR(0, 1326, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_parent_id);
      __Pyx_GIVEREF(__pyx_v_parent_id);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_v_parent_id) != (0)) __PYX_ERR(0, 1326, __pyx_L1_error);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1327
 *                        (False, True),
 *                        (None, parent_id),
 *                        (None, self.utf8_decode(entry[0][1], 'surrogateescape')[0]),             # <<<<<<<<<<<<<<
 *                        (None, None),
 *                        (None, False)), True
*/
      __pyx_t_23 = NULL;
      __Pyx_INCREF(__pyx_v_self->utf8_decode);
      __pyx_t_22 = __pyx_v_self->utf8_decode; 
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1327, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_8, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1327, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_9 = 1;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_22))) {
        __pyx_t_23 = PyMethod_GET_SELF(__pyx_t_22);
        assert(__pyx_t_23);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_22);
        __Pyx_INCREF(__pyx_t_23);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_22, __pyx__function);
        __pyx_t_9 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_23, __pyx_t_17, __pyx_mstate_global->__pyx_n_u_surrogateescape};
        __pyx_t_12 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_22, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
        if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1327, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
      }
      __pyx_t_22 = __Pyx_GetItemInt(__pyx_t_12, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1327, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_22);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1327, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 0, Py_None) != (0)) __PYX_ERR(0, 1327, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_22);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_22) != (0)) __PYX_ERR(0, 1327, __pyx_L1_error);
      __pyx_t_22 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1329
 *                        (None, self.utf8_decode(entry[0][1], 'surrogateescape')[0]),
 *                        (None, None),
 *                        (None, False)), True             # <<<<<<<<<<<<<<
 *         elif _versioned_minikind(source_minikind) and target_minikind == b'a':
 *             # unversioned, possibly, or possibly not deleted: we dont care.
*/
      __pyx_t_9 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_14))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_14);
        assert(__pyx_t_7);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_14);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_14, __pyx__function);
        __pyx_t_9 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[9] = {__pyx_t_7, __pyx_t_1, __pyx_t_25, Py_False, __pyx_mstate_global->__pyx_tuple[3], __pyx_t_15, __pyx_t_12, __pyx_mstate_global->__pyx_tuple[0], __pyx_mstate_global->__pyx_tuple[1]};
        __pyx_t_13 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_14, __pyx_callargs+__pyx_t_9, (9-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1322, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1322
 *             else:
 *                 # Its a missing file, report it as such.
 *                 return InventoryTreeChange(entry[0][2],             # <<<<<<<<<<<<<<
 *                        (None, self.utf8_decode(path, 'surrogateescape')[0]),
 *                        False,
*/
      __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1322, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_GIVEREF(__pyx_t_13);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13) != (0)) __PYX_ERR(0, 1322, __pyx_L1_error);
      __Pyx_INCREF(Py_True);
      __Pyx_GIVEREF(Py_True);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_14, 1, Py_True) != (0)) __PYX_ERR(0, 1322, __pyx_L1_error);
      __pyx_t_13 = 0;
      __pyx_r = __pyx_t_14;
      __pyx_t_14 = 0;
      goto __pyx_L0;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1287
 *                        (source_kind, target_kind),
 *                        (source_exec, target_exec)), changed
 *         elif source_minikind == b'a' and _versioned_minikind(target_minikind):             # <<<<<<<<<<<<<<
 *             # looks like a new file
 *             path = self.pathjoin(entry[0][0], entry[0][1])
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1330
 *                        (None, None),
 *                        (None, False)), True
 *         elif _versioned_minikind(source_minikind) and target_minikind == b'a':             # <<<<<<<<<<<<<<
 *             # unversioned, possibly, or possibly not deleted: we dont care.
 *             # if its still on disk, *and* theres no other entry at this
*/
  __pyx_t_6 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__versioned_minikind(__pyx_v_source_minikind); if (unlikely(__pyx_t_6 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1330, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_6 != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L76_bool_binop_done;
  }
  __pyx_t_5 = (__pyx_v_target_minikind == 'a');
  __pyx_t_2 = __pyx_t_5;
  __pyx_L76_bool_binop_done:;
  if (__pyx_t_2) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1335
 *             # path [we dont know this in this routine at the moment -
 *             # perhaps we should change this - then it would be an unknown.
 *             old_path = self.pathjoin(entry[0][0], entry[0][1])             # <<<<<<<<<<<<<<
 *             # parent id is the entry for the path in the target tree
 *             parent_id = self.state._get_entry(self.source_index, path_utf8=entry[0][0])[0][2]
*/
    __pyx_t_13 = NULL;
    __Pyx_INCREF(__pyx_v_self->pathjoin);
    __pyx_t_12 = __pyx_v_self->pathjoin; 
    __pyx_t_15 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_25 = __Pyx_GetItemInt(__pyx_t_15, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_15 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_15, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_9 = 1;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
      assert(__pyx_t_13);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_12);
      __Pyx_INCREF(__pyx_t_13);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_12, __pyx__function);
      __pyx_t_9 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_13, __pyx_t_25, __pyx_t_1};
      __pyx_t_14 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_12, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1335, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
    }
    __pyx_v_old_path = __pyx_t_14;
    __pyx_t_14 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1337
 *             old_path = self.pathjoin(entry[0][0], entry[0][1])
 *             # parent id is the entry for the path in the target tree
 *             parent_id = self.state._get_entry(self.source_index, path_utf8=entry[0][0])[0][2]             # <<<<<<<<<<<<<<
 *             if parent_id == entry[0][2]:
 *                 parent_id = None
*/
    __pyx_t_12 = __pyx_v_self->state;
    __Pyx_INCREF(__pyx_t_12);
    __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->source_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_25 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_13 = __Pyx_GetItemInt(__pyx_t_25, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __pyx_t_9 = 0;
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_12, __pyx_t_1};
      __pyx_t_25 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1337, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_25);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_path_utf8, __pyx_t_13, __pyx_t_25, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 1337, __pyx_L1_error)
      __pyx_t_14 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_get_entry, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_25);
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
      if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1337, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
    }
    __pyx_t_25 = __Pyx_GetItemInt(__pyx_t_14, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_GetItemInt(__pyx_t_25, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __pyx_v_parent_id = __pyx_t_14;
    __pyx_t_14 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1338
 *             # parent id is the entry for the path in the target tree
 *             parent_id = self.state._get_entry(self.source_index, path_utf8=entry[0][0])[0][2]
 *             if parent_id == entry[0][2]:             # <<<<<<<<<<<<<<
 *                 parent_id = None
 *             return InventoryTreeChange(
*/
    __pyx_t_14 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_25 = __Pyx_GetItemInt(__pyx_t_14, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = PyObject_RichCompare(__pyx_v_parent_id, __pyx_t_25, Py_EQ); __Pyx_XGOTREF(__pyx_t_14); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1339
 *             parent_id = self.state._get_entry(self.source_index, path_utf8=entry[0][0])[0][2]
 *             if parent_id == entry[0][2]:
 *                 parent_id = None             # <<<<<<<<<<<<<<
 *             return InventoryTreeChange(
 *                    entry[0][2],
*/
      __Pyx_INCREF(Py_None);
      __Pyx_DECREF_SET(__pyx_v_parent_id, Py_None);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1338
 *             # parent id is the entry for the path in the target tree
 *             parent_id = self.state._get_entry(self.source_index, path_utf8=entry[0][0])[0][2]
 *             if parent_id == entry[0][2]:             # <<<<<<<<<<<<<<
 *                 parent_id = None
 *             return InventoryTreeChange(
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1340
 *             if parent_id == entry[0][2]:
 *                 parent_id = None
 *             return InventoryTreeChange(             # <<<<<<<<<<<<<<
 *                    entry[0][2],
 *                    (self.utf8_decode(old_path, 'surrogateescape')[0], None),
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_25 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_InventoryTreeChange); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1341
 *                 parent_id = None
 *             return InventoryTreeChange(
 *                    entry[0][2],             # <<<<<<<<<<<<<<
 *                    (self.utf8_decode(old_path, 'surrogateescape')[0], None),
 *                    True,
*/
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_12 = __Pyx_GetItemInt(__pyx_t_1, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1342
 *             return InventoryTreeChange(
 *                    entry[0][2],
 *                    (self.utf8_decode(old_path, 'surrogateescape')[0], None),             # <<<<<<<<<<<<<<
 *                    True,
 *                    (True, False),
*/
    __pyx_t_15 = NULL;
    __Pyx_INCREF(__pyx_v_self->utf8_decode);
    __pyx_t_7 = __pyx_v_self->utf8_decode; 
    __pyx_t_9 = 1;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_7);
      assert(__pyx_t_15);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
      __pyx_t_9 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_15, __pyx_v_old_path, __pyx_mstate_global->__pyx_n_u_surrogateescape};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1342, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_7) != (0)) __PYX_ERR(0, 1342, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, Py_None) != (0)) __PYX_ERR(0, 1342, __pyx_L1_error);
    __pyx_t_7 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1345
 *                    True,
 *                    (True, False),
 *                    (parent_id, None),             # <<<<<<<<<<<<<<
 *                    (self.utf8_decode(entry[0][1], 'surrogateescape')[0], None),
 *                    (_minikind_to_kind(source_minikind), None),
*/
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_parent_id);
    __Pyx_GIVEREF(__pyx_v_parent_id);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_parent_id) != (0)) __PYX_ERR(0, 1345, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, Py_None) != (0)) __PYX_ERR(0, 1345, __pyx_L1_error);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1346
 *                    (True, False),
 *                    (parent_id, None),
 *                    (self.utf8_decode(entry[0][1], 'surrogateescape')[0], None),             # <<<<<<<<<<<<<<
 *                    (_minikind_to_kind(source_minikind), None),
 *                    (source_details[3], None)), True
*/
    __pyx_t_22 = NULL;
    __Pyx_INCREF(__pyx_v_self->utf8_decode);
    __pyx_t_17 = __pyx_v_self->utf8_decode; 
    __pyx_t_23 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_t_23, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __pyx_t_9 = 1;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_17))) {
      __pyx_t_22 = PyMethod_GET_SELF(__pyx_t_17);
      assert(__pyx_t_22);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_17);
      __Pyx_INCREF(__pyx_t_22);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_17, __pyx__function);
      __pyx_t_9 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_22, __pyx_t_8, __pyx_mstate_global->__pyx_n_u_surrogateescape};
      __pyx_t_15 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_17, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1346, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
    }
    __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_15, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_17);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_17) != (0)) __PYX_ERR(0, 1346, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 1, Py_None) != (0)) __PYX_ERR(0, 1346, __pyx_L1_error);
    __pyx_t_17 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1347
 *                    (parent_id, None),
 *                    (self.utf8_decode(entry[0][1], 'surrogateescape')[0], None),
 *                    (_minikind_to_kind(source_minikind), None),             # <<<<<<<<<<<<<<
 *                    (source_details[3], None)), True
 *         elif _versioned_minikind(source_minikind) and target_minikind == b'r':
*/
    __pyx_t_17 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_to_kind(__pyx_v_source_minikind); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_17);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_17) != (0)) __PYX_ERR(0, 1347, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, Py_None) != (0)) __PYX_ERR(0, 1347, __pyx_L1_error);
    __pyx_t_17 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1348
 *                    (self.utf8_decode(entry[0][1], 'surrogateescape')[0], None),
 *                    (_minikind_to_kind(source_minikind), None),
 *                    (source_details[3], None)), True             # <<<<<<<<<<<<<<
 *         elif _versioned_minikind(source_minikind) and target_minikind == b'r':
 *             # a rename; could be a true rename, or a rename inherited from
*/
    __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_source_details, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1348, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1348, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_GIVEREF(__pyx_t_17);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_17) != (0)) __PYX_ERR(0, 1348, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_22, 1, Py_None) != (0)) __PYX_ERR(0, 1348, __pyx_L1_error);
    __pyx_t_17 = 0;
    __pyx_t_9 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_13);
      assert(__pyx_t_25);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_13);
      __Pyx_INCREF(__pyx_t_25);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_13, __pyx__function);
      __pyx_t_9 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[9] = {__pyx_t_25, __pyx_t_12, __pyx_t_1, Py_True, __pyx_mstate_global->__pyx_tuple[4], __pyx_t_7, __pyx_t_15, __pyx_t_8, __pyx_t_22};
      __pyx_t_14 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_13, __pyx_callargs+__pyx_t_9, (9-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1340, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1340
 *             if parent_id == entry[0][2]:
 *                 parent_id = None
 *             return InventoryTreeChange(             # <<<<<<<<<<<<<<
 *                    entry[0][2],
 *                    (self.utf8_decode(old_path, 'surrogateescape')[0], None),
*/
    __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14) != (0)) __PYX_ERR(0, 1340, __pyx_L1_error);
    __Pyx_INCREF(Py_True);
    __Pyx_GIVEREF(Py_True);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 1, Py_True) != (0)) __PYX_ERR(0, 1340, __pyx_L1_error);
    __pyx_t_14 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1330
 *                        (None, None),
 *                        (None, False)), True
 *         elif _versioned_minikind(source_minikind) and target_minikind == b'a':             # <<<<<<<<<<<<<<
 *             # unversioned, possibly, or possibly not deleted: we dont care.
 *             # if its still on disk, *and* theres no other entry at this
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1349
 *                    (_minikind_to_kind(source_minikind), None),
 *                    (source_details[3], None)), True
 *         elif _versioned_minikind(source_minikind) and target_minikind == b'r':             # <<<<<<<<<<<<<<
 *             # a rename; could be a true rename, or a rename inherited from
 *             # a renamed parent. TODO: handle this efficiently. Its not
*/
  __pyx_t_6 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__versioned_minikind(__pyx_v_source_minikind); if (unlikely(__pyx_t_6 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1349, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_6 != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L79_bool_binop_done;
  }
  __pyx_t_5 = (__pyx_v_target_minikind == 'r');
  __pyx_t_2 = __pyx_t_5;
  __pyx_L79_bool_binop_done:;
  if (__pyx_t_2) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1354
 *             # common case to rename dirs though, so a correct but slow
 *             # implementation will do.
 *             if (not self.doing_consistency_expansion and             # <<<<<<<<<<<<<<
 *                 not is_inside_any(self.searched_specific_files,
 *                     target_details[1])):
*/
    __pyx_t_5 = (!(__pyx_v_self->doing_consistency_expansion != 0));
    if (__pyx_t_5) {
    } else {
      __pyx_t_2 = __pyx_t_5;
      goto __pyx_L82_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1355
 *             # implementation will do.
 *             if (not self.doing_consistency_expansion and
 *                 not is_inside_any(self.searched_specific_files,             # <<<<<<<<<<<<<<
 *                     target_details[1])):
 *                 self.search_specific_files.add(target_details[1])
*/
    __pyx_t_14 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_is_inside_any); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1355, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1356
 *             if (not self.doing_consistency_expansion and
 *                 not is_inside_any(self.searched_specific_files,
 *                     target_details[1])):             # <<<<<<<<<<<<<<
 *                 self.search_specific_files.add(target_details[1])
 *                 # We don't expand the specific files parents list here as
*/
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_target_details, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1356, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_22))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_22);
      assert(__pyx_t_14);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_22);
      __Pyx_INCREF(__pyx_t_14);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_22, __pyx__function);
      __pyx_t_9 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_14, __pyx_v_self->searched_specific_files, __pyx_t_8};
      __pyx_t_13 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_22, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
      if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1355, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1355
 *             # implementation will do.
 *             if (not self.doing_consistency_expansion and
 *                 not is_inside_any(self.searched_specific_files,             # <<<<<<<<<<<<<<
 *                     target_details[1])):
 *                 self.search_specific_files.add(target_details[1])
*/
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1355, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_10 = (!__pyx_t_5);
    __pyx_t_2 = __pyx_t_10;
    __pyx_L82_bool_binop_done:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1354
 *             # common case to rename dirs though, so a correct but slow
 *             # implementation will do.
 *             if (not self.doing_consistency_expansion and             # <<<<<<<<<<<<<<
 *                 not is_inside_any(self.searched_specific_files,
 *                     target_details[1])):
*/
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1357
 *                 not is_inside_any(self.searched_specific_files,
 *                     target_details[1])):
 *                 self.search_specific_files.add(target_details[1])             # <<<<<<<<<<<<<<
 *                 # We don't expand the specific files parents list here as
 *                 # the path is absent in target and won't create a delta with
*/
      __pyx_t_22 = __pyx_v_self->search_specific_files;
      __Pyx_INCREF(__pyx_t_22);
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_target_details, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1357, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_22, __pyx_t_8};
        __pyx_t_13 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1357, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
      }
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1354
 *             # common case to rename dirs though, so a correct but slow
 *             # implementation will do.
 *             if (not self.doing_consistency_expansion and             # <<<<<<<<<<<<<<
 *                 not is_inside_any(self.searched_specific_files,
 *                     target_details[1])):
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1349
 *                    (_minikind_to_kind(source_minikind), None),
 *                    (source_details[3], None)), True
 *         elif _versioned_minikind(source_minikind) and target_minikind == b'r':             # <<<<<<<<<<<<<<
 *             # a rename; could be a true rename, or a rename inherited from
 *             # a renamed parent. TODO: handle this efficiently. Its not
*/
    goto __pyx_L8;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1361
 *                 # the path is absent in target and won't create a delta with
 *                 # missing parent.
 *         elif ((source_minikind == b'r' or source_minikind == b'a') and             # <<<<<<<<<<<<<<
 *               (target_minikind == b'r' or target_minikind == b'a')):
 *             # neither of the selected trees contain this path,
*/
  switch (__pyx_v_source_minikind) {
    case 'r':
    case 'a':
    __pyx_t_10 = 1;
    break;
    default:
    __pyx_t_10 = 0;
    break;
  }
  if (__pyx_t_10) {
  } else {
    __pyx_t_2 = __pyx_t_10;
    goto __pyx_L84_bool_binop_done;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1362
 *                 # missing parent.
 *         elif ((source_minikind == b'r' or source_minikind == b'a') and
 *               (target_minikind == b'r' or target_minikind == b'a')):             # <<<<<<<<<<<<<<
 *             # neither of the selected trees contain this path,
 *             # so skip over it. This is not currently directly tested, but
*/
  switch (__pyx_v_target_minikind) {
    case 'r':
    case 'a':
    __pyx_t_10 = 1;
    break;
    default:
    __pyx_t_10 = 0;
    break;
  }
  __pyx_t_2 = __pyx_t_10;
  __pyx_L84_bool_binop_done:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1361
 *                 # the path is absent in target and won't create a delta with
 *                 # missing parent.
 *         elif ((source_minikind == b'r' or source_minikind == b'a') and             # <<<<<<<<<<<<<<
 *               (target_minikind == b'r' or target_minikind == b'a')):
 *             # neither of the selected trees contain this path,
*/
  if (likely(__pyx_t_2)) {
    goto __pyx_L8;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1368
 *             pass
 *         else:
 *             raise AssertionError("don't know how to compare "             # <<<<<<<<<<<<<<
 *                 "source_minikind=%r, target_minikind=%r"
 *                 % (source_minikind, target_minikind))
*/
  /*else*/ {
    __pyx_t_8 = NULL;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1370
 *             raise AssertionError("don't know how to compare "
 *                 "source_minikind=%r, target_minikind=%r"
 *                 % (source_minikind, target_minikind))             # <<<<<<<<<<<<<<
 *             ## import pdb;pdb.set_trace()
 *         return None, None
*/
    __pyx_t_22 = __Pyx_PyUnicode_From_char(__pyx_v_source_minikind, 0, ' ', 'd'); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_14 = __Pyx_PyUnicode_From_char(__pyx_v_target_minikind, 0, ' ', 'd'); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_21[0] = __pyx_mstate_global->__pyx_kp_u_don_t_know_how_to_compare_source;
    __pyx_t_21[1] = __pyx_t_22;
    __pyx_t_21[2] = __pyx_mstate_global->__pyx_kp_u_target_minikind;
    __pyx_t_21[3] = __pyx_t_14;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1368
 *             pass
 *         else:
 *             raise AssertionError("don't know how to compare "             # <<<<<<<<<<<<<<
 *                 "source_minikind=%r, target_minikind=%r"
 *                 % (source_minikind, target_minikind))
*/
    __pyx_t_15 = __Pyx_PyUnicode_Join(__pyx_t_21, 4, 42 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_22) + 18 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_14), 127);
    if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_9 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_15};
      __pyx_t_13 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1368, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
    }
    __Pyx_Raise(__pyx_t_13, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __PYX_ERR(0, 1368, __pyx_L1_error)
  }
  __pyx_L8:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1372
 *                 % (source_minikind, target_minikind))
 *             ## import pdb;pdb.set_trace()
 *         return None, None             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_tuple[0]);
  __pyx_r = __pyx_mstate_global->__pyx_tuple[0];
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1075
 *             raise errors.BzrError('unsupported target index')
 * 
 *     cdef _process_entry(self, entry, path_info):             # <<<<<<<<<<<<<<
 *         """Compare an entry and real disk to generate delta information.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._process_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_file_id);
  __Pyx_XDECREF(__pyx_v_details_list);
  __Pyx_XDECREF(__pyx_v_source_details);
  __Pyx_XDECREF(__pyx_v_target_details);
  __Pyx_XDECREF(__pyx_v_link_or_sha1);
  __Pyx_XDECREF(__pyx_v_old_dirname);
  __Pyx_XDECREF(__pyx_v_old_basename);
  __Pyx_XDECREF(__pyx_v_old_path);
  __Pyx_XDECREF(__pyx_v_path);
  __Pyx_XDECREF(__pyx_v_old_entry);
  __Pyx_XDECREF(__pyx_v_target_kind);
  __Pyx_XDECREF(__pyx_v_target_exec);
  __Pyx_XDECREF(__pyx_v_statvalue);
  __Pyx_XDECREF(__pyx_v_source_parent_id);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_source_parent_entry);
  __Pyx_XDECREF(__pyx_v_new_dirname);
  __Pyx_XDECREF(__pyx_v_target_parent_id);
  __Pyx_XDECREF(__pyx_v_target_parent_entry);
  __Pyx_XDECREF(__pyx_v_source_exec);
  __Pyx_XDECREF(__pyx_v_changed);
  __Pyx_XDECREF(__pyx_v_old_path_u);
  __Pyx_XDECREF(__pyx_v_path_u);
  __Pyx_XDECREF(__pyx_v_source_kind);
  __Pyx_XDECREF(__pyx_v_parent_entry);
  __Pyx_XDECREF(__pyx_v_parent_id);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1374
 *         return None, None
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_3__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_3__iter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_2__iter__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_2__iter__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1375
 * 
 *     def __iter__(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def iter_changes(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1374
 *         return None, None
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1377
 *         return self
 * 
 *     def iter_changes(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_5iter_changes(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_5iter_changes = {"iter_changes", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_5iter_changes, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_5iter_changes(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iter_changes (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("iter_changes", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("iter_changes", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_4iter_changes(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_4iter_changes(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iter_changes", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1378
 * 
 *     def iter_changes(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cdef int _gather_result_for_consistency(self, result) except -1:
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1377
 *         return self
 * 
 *     def iter_changes(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1380
 *         return self
 * 
 *     cdef int _gather_result_for_consistency(self, result) except -1:             # <<<<<<<<<<<<<<
 *         """Check a result we will yield to make sure we are consistent later.
 * 
*/

static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__gather_result_for_consistency(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_result) {
  PyObject *__pyx_v_new_path = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_gather_result_for_consistency", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1387
 *         :param result: A result tuple.
 *         """
 *         if not self.partial or not result.file_id:             # <<<<<<<<<<<<<<
 *             return 0
 *         self.seen_ids.add(result.file_id)
*/
  __pyx_t_2 = (!(__pyx_v_self->partial != 0));
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_mstate_global->__pyx_n_u_file_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = (!__pyx_t_2);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1388
 *         """
 *         if not self.partial or not result.file_id:
 *             return 0             # <<<<<<<<<<<<<<
 *         self.seen_ids.add(result.file_id)
 *         new_path = result.path[1]
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1387
 *         :param result: A result tuple.
 *         """
 *         if not self.partial or not result.file_id:             # <<<<<<<<<<<<<<
 *             return 0
 *         self.seen_ids.add(result.file_id)
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1389
 *         if not self.partial or not result.file_id:
 *             return 0
 *         self.seen_ids.add(result.file_id)             # <<<<<<<<<<<<<<
 *         new_path = result.path[1]
 *         if new_path:
*/
  __pyx_t_5 = __pyx_v_self->seen_ids;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_mstate_global->__pyx_n_u_file_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_6};
    __pyx_t_3 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1390
 *             return 0
 *         self.seen_ids.add(result.file_id)
 *         new_path = result.path[1]             # <<<<<<<<<<<<<<
 *         if new_path:
 *             # Not the root and not a delete: queue up the parents of the path.
*/
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_mstate_global->__pyx_n_u_path); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_3, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_new_path = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1391
 *         self.seen_ids.add(result.file_id)
 *         new_path = result.path[1]
 *         if new_path:             # <<<<<<<<<<<<<<
 *             # Not the root and not a delete: queue up the parents of the path.
 *             self.search_specific_file_parents.update(
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_new_path); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1391, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1393
 *         if new_path:
 *             # Not the root and not a delete: queue up the parents of the path.
 *             self.search_specific_file_parents.update(             # <<<<<<<<<<<<<<
 *                 osutils.parent_directories(new_path.encode('utf8')))
 *             # Add the root directory which parent_directories does not
*/
    __pyx_t_3 = __pyx_v_self->search_specific_file_parents;
    __Pyx_INCREF(__pyx_t_3);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1394
 *             # Not the root and not a delete: queue up the parents of the path.
 *             self.search_specific_file_parents.update(
 *                 osutils.parent_directories(new_path.encode('utf8')))             # <<<<<<<<<<<<<<
 *             # Add the root directory which parent_directories does not
 *             # provide.
*/
    __pyx_t_8 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_osutils); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_parent_directories); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_11 = __pyx_v_new_path;
    __Pyx_INCREF(__pyx_t_11);
    __pyx_t_7 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_mstate_global->__pyx_n_u_utf8};
      __pyx_t_9 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_encode_2, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
    }
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_10);
      assert(__pyx_t_8);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_10, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_9};
      __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_10, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    }
    __pyx_t_7 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
      __pyx_t_6 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_update, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1393, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1397
 *             # Add the root directory which parent_directories does not
 *             # provide.
 *             self.search_specific_file_parents.add(b'')             # <<<<<<<<<<<<<<
 *         return 0
 * 
*/
    __pyx_t_5 = __pyx_v_self->search_specific_file_parents;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_7 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_mstate_global->__pyx_kp_b__4};
      __pyx_t_6 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1397, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1391
 *         self.seen_ids.add(result.file_id)
 *         new_path = result.path[1]
 *         if new_path:             # <<<<<<<<<<<<<<
 *             # Not the root and not a delete: queue up the parents of the path.
 *             self.search_specific_file_parents.update(
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1398
 *             # provide.
 *             self.search_specific_file_parents.add(b'')
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _update_current_block(self) except -1:
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1380
 *         return self
 * 
 *     cdef int _gather_result_for_consistency(self, result) except -1:             # <<<<<<<<<<<<<<
 *         """Check a result we will yield to make sure we are consistent later.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._gather_result_for_consistency", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_new_path);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1400
 *         return 0
 * 
 *     cdef int _update_current_block(self) except -1:             # <<<<<<<<<<<<<<
 *         if (self.block_index < len(self.state._dirblocks) and
 *             is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):
*/

static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__update_current_block(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  size_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_update_current_block", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1401
 * 
 *     cdef int _update_current_block(self) except -1:
 *         if (self.block_index < len(self.state._dirblocks) and             # <<<<<<<<<<<<<<
 *             is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):
 *             self.current_block = self.state._dirblocks[self.block_index]
*/
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_mstate_global->__pyx_n_u_dirblocks_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1401, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_v_self->block_index < __pyx_t_3);
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1402
 *     cdef int _update_current_block(self) except -1:
 *         if (self.block_index < len(self.state._dirblocks) and
 *             is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):             # <<<<<<<<<<<<<<
 *             self.current_block = self.state._dirblocks[self.block_index]
 *             self.current_block_list = self.current_block[1]
*/
  __pyx_t_5 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_is_inside); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_mstate_global->__pyx_n_u_dirblocks_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetItemInt(__pyx_t_7, __pyx_v_self->block_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_8, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_9 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
    assert(__pyx_t_5);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
    __pyx_t_9 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_self->current_root, __pyx_t_7};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 1402, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1401
 * 
 *     cdef int _update_current_block(self) except -1:
 *         if (self.block_index < len(self.state._dirblocks) and             # <<<<<<<<<<<<<<
 *             is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):
 *             self.current_block = self.state._dirblocks[self.block_index]
*/
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1403
 *         if (self.block_index < len(self.state._dirblocks) and
 *             is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):
 *             self.current_block = self.state._dirblocks[self.block_index]             # <<<<<<<<<<<<<<
 *             self.current_block_list = self.current_block[1]
 *             self.current_block_pos = 0
*/
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_mstate_global->__pyx_n_u_dirblocks_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_self->block_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GIVEREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_v_self->current_block);
    __Pyx_DECREF(__pyx_v_self->current_block);
    __pyx_v_self->current_block = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1404
 *             is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):
 *             self.current_block = self.state._dirblocks[self.block_index]
 *             self.current_block_list = self.current_block[1]             # <<<<<<<<<<<<<<
 *             self.current_block_pos = 0
 *         else:
*/
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_self->current_block, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_v_self->current_block_list);
    __Pyx_DECREF(__pyx_v_self->current_block_list);
    __pyx_v_self->current_block_list = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1405
 *             self.current_block = self.state._dirblocks[self.block_index]
 *             self.current_block_list = self.current_block[1]
 *             self.current_block_pos = 0             # <<<<<<<<<<<<<<
 *         else:
 *             self.current_block = None
*/
    __pyx_v_self->current_block_pos = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1401
 * 
 *     cdef int _update_current_block(self) except -1:
 *         if (self.block_index < len(self.state._dirblocks) and             # <<<<<<<<<<<<<<
 *             is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):
 *             self.current_block = self.state._dirblocks[self.block_index]
*/
    goto __pyx_L3;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1407
 *             self.current_block_pos = 0
 *         else:
 *             self.current_block = None             # <<<<<<<<<<<<<<
 *             self.current_block_list = None
 *         return 0
*/
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->current_block);
    __Pyx_DECREF(__pyx_v_self->current_block);
    __pyx_v_self->current_block = Py_None;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1408
 *         else:
 *             self.current_block = None
 *             self.current_block_list = None             # <<<<<<<<<<<<<<
 *         return 0
 * 
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->current_block_list);
    __Pyx_DECREF(__pyx_v_self->current_block_list);
    __pyx_v_self->current_block_list = Py_None;
  }
  __pyx_L3:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1409
 *             self.current_block = None
 *             self.current_block_list = None
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __next__(self):
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1400
 *         return 0
 * 
 *     cdef int _update_current_block(self) except -1:             # <<<<<<<<<<<<<<
 *         if (self.block_index < len(self.state._dirblocks) and
 *             is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._update_current_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1411
 *         return 0
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         # Simple thunk to allow tail recursion without pyrex confusion
 *         return self._iter_next()
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_7__next__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_7__next__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__next__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_6__next__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_6__next__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__next__", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1413
 *     def __next__(self):
 *         # Simple thunk to allow tail recursion without pyrex confusion
 *         return self._iter_next()             # <<<<<<<<<<<<<<
 * 
 *     cdef _iter_next(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_iter_next(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1411
 *         return 0
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         # Simple thunk to allow tail recursion without pyrex confusion
 *         return self._iter_next()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC.__next__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1415
 *         return self._iter_next()
 * 
 *     cdef _iter_next(self):             # <<<<<<<<<<<<<<
 *         """Iterate over the changes."""
 *         # This function single steps through an iterator. As such while loops
*/

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__iter_next(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  PyObject *__pyx_v_current_dirname = 0;
  PyObject *__pyx_v_current_blockname = 0;
  char *__pyx_v_current_dirname_c;
  char *__pyx_v_current_blockname_c;
  int __pyx_v_path_handled;
  CYTHON_UNUSED PyObject *__pyx_v_searched_specific_files = NULL;
  PyObject *__pyx_v_entry = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_changed = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_root_stat = NULL;
  PyObject *__pyx_v_e = NULL;
  int __pyx_v_new_executable;
  PyObject *__pyx_v_e_winerror = NULL;
  PyObject *__pyx_v_win_errors = NULL;
  PyObject *__pyx_v_bzr_index = NULL;
  PyObject *__pyx_v_initial_key = NULL;
  PyObject *__pyx_v_current_path_info = NULL;
  PyObject *__pyx_v_current_entry = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  size_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_iter_next", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1468
 *         cdef int advance_entry, advance_path
 *         cdef int path_handled
 *         searched_specific_files = self.searched_specific_files             # <<<<<<<<<<<<<<
 *         # Are we walking a root?
 *         while self.root_entries_pos < self.root_entries_len:
*/
  __pyx_t_1 = __pyx_v_self->searched_specific_files;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_searched_specific_files = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1470
 *         searched_specific_files = self.searched_specific_files
 *         # Are we walking a root?
 *         while self.root_entries_pos < self.root_entries_len:             # <<<<<<<<<<<<<<
 *             entry = self.root_entries[self.root_entries_pos]
 *             self.root_entries_pos = self.root_entries_pos + 1
*/
  while (1) {
    __pyx_t_2 = (__pyx_v_self->root_entries_pos < __pyx_v_self->root_entries_len);
    if (!__pyx_t_2) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1471
 *         # Are we walking a root?
 *         while self.root_entries_pos < self.root_entries_len:
 *             entry = self.root_entries[self.root_entries_pos]             # <<<<<<<<<<<<<<
 *             self.root_entries_pos = self.root_entries_pos + 1
 *             result, changed = self._process_entry(entry, self.root_dir_info)
*/
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self->root_entries, __pyx_v_self->root_entries_pos, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1471, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_entry, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1472
 *         while self.root_entries_pos < self.root_entries_len:
 *             entry = self.root_entries[self.root_entries_pos]
 *             self.root_entries_pos = self.root_entries_pos + 1             # <<<<<<<<<<<<<<
 *             result, changed = self._process_entry(entry, self.root_dir_info)
 *             if changed is not None:
*/
    __pyx_v_self->root_entries_pos = (__pyx_v_self->root_entries_pos + 1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1473
 *             entry = self.root_entries[self.root_entries_pos]
 *             self.root_entries_pos = self.root_entries_pos + 1
 *             result, changed = self._process_entry(entry, self.root_dir_info)             # <<<<<<<<<<<<<<
 *             if changed is not None:
 *                 if changed:
*/
    __pyx_t_1 = __pyx_v_self->root_dir_info;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_process_entry(__pyx_v_self, __pyx_v_entry, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1473, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1473, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1);
        __Pyx_INCREF(__pyx_t_4);
      } else {
        __pyx_t_1 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1473, __pyx_L1_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1473, __pyx_L1_error)
        __Pyx_XGOTREF(__pyx_t_4);
      }
      #else
      __pyx_t_1 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1473, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1473, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1473, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_5);
      index = 0; __pyx_t_1 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_1)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 2) < (0)) __PYX_ERR(0, 1473, __pyx_L1_error)
      __pyx_t_6 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1473, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_result, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_changed, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1474
 *             self.root_entries_pos = self.root_entries_pos + 1
 *             result, changed = self._process_entry(entry, self.root_dir_info)
 *             if changed is not None:             # <<<<<<<<<<<<<<
 *                 if changed:
 *                     self._gather_result_for_consistency(result)
*/
    __pyx_t_2 = (__pyx_v_changed != Py_None);
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1475
 *             result, changed = self._process_entry(entry, self.root_dir_info)
 *             if changed is not None:
 *                 if changed:             # <<<<<<<<<<<<<<
 *                     self._gather_result_for_consistency(result)
 *                 if changed or self.include_unchanged:
*/
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1475, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1476
 *             if changed is not None:
 *                 if changed:
 *                     self._gather_result_for_consistency(result)             # <<<<<<<<<<<<<<
 *                 if changed or self.include_unchanged:
 *                     return result
*/
        __pyx_t_7 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_gather_result_for_consistency(__pyx_v_self, __pyx_v_result); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1476, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1475
 *             result, changed = self._process_entry(entry, self.root_dir_info)
 *             if changed is not None:
 *                 if changed:             # <<<<<<<<<<<<<<
 *                     self._gather_result_for_consistency(result)
 *                 if changed or self.include_unchanged:
*/
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1477
 *                 if changed:
 *                     self._gather_result_for_consistency(result)
 *                 if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                     return result
 *         # Have we finished the prior root, or never started one ?
*/
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1477, __pyx_L1_error)
      if (!__pyx_t_8) {
      } else {
        __pyx_t_2 = __pyx_t_8;
        goto __pyx_L10_bool_binop_done;
      }
      __pyx_t_8 = (__pyx_v_self->include_unchanged != 0);
      __pyx_t_2 = __pyx_t_8;
      __pyx_L10_bool_binop_done:;
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1478
 *                     self._gather_result_for_consistency(result)
 *                 if changed or self.include_unchanged:
 *                     return result             # <<<<<<<<<<<<<<
 *         # Have we finished the prior root, or never started one ?
 *         if self.current_root is None:
*/
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_result);
        __pyx_r = __pyx_v_result;
        goto __pyx_L0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1477
 *                 if changed:
 *                     self._gather_result_for_consistency(result)
 *                 if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                     return result
 *         # Have we finished the prior root, or never started one ?
*/
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1474
 *             self.root_entries_pos = self.root_entries_pos + 1
 *             result, changed = self._process_entry(entry, self.root_dir_info)
 *             if changed is not None:             # <<<<<<<<<<<<<<
 *                 if changed:
 *                     self._gather_result_for_consistency(result)
*/
    }
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1480
 *                     return result
 *         # Have we finished the prior root, or never started one ?
 *         if self.current_root is None:             # <<<<<<<<<<<<<<
 *             # TODO: the pending list should be lexically sorted?  the
 *             # interface doesn't require it.
*/
  __pyx_t_2 = (__pyx_v_self->current_root == Py_None);
  if (__pyx_t_2) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1483
 *             # TODO: the pending list should be lexically sorted?  the
 *             # interface doesn't require it.
 *             try:             # <<<<<<<<<<<<<<
 *                 self.current_root = self.search_specific_files.pop()
 *             except KeyError, _:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      /*try:*/ {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1484
 *             # interface doesn't require it.
 *             try:
 *                 self.current_root = self.search_specific_files.pop()             # <<<<<<<<<<<<<<
 *             except KeyError, _:
 *                 raise StopIteration()
*/
        __pyx_t_3 = __Pyx_PyObject_Pop(__pyx_v_self->search_specific_files); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1484, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_v_self->current_root);
        __Pyx_DECREF(__pyx_v_self->current_root);
        __pyx_v_self->current_root = __pyx_t_3;
        __pyx_t_3 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1483
 *             # TODO: the pending list should be lexically sorted?  the
 *             # interface doesn't require it.
 *             try:             # <<<<<<<<<<<<<<
 *                 self.current_root = self.search_specific_files.pop()
 *             except KeyError, _:
*/
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      goto __pyx_L18_try_end;
      __pyx_L13_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1485
 *             try:
 *                 self.current_root = self.search_specific_files.pop()
 *             except KeyError, _:             # <<<<<<<<<<<<<<
 *                 raise StopIteration()
 *             self.searched_specific_files.add(self.current_root)
*/
      __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_KeyError))));
      if (__pyx_t_7) {
        __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._iter_next", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_4, &__pyx_t_1) < 0) __PYX_ERR(0, 1485, __pyx_L15_except_error)
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v__ = __pyx_t_4;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1486
 *                 self.current_root = self.search_specific_files.pop()
 *             except KeyError, _:
 *                 raise StopIteration()             # <<<<<<<<<<<<<<
 *             self.searched_specific_files.add(self.current_root)
 *             # process the entries for this containing directory: the rest will be
*/
        __pyx_t_12 = NULL;
        __pyx_t_13 = 1;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, NULL};
          __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_StopIteration)), __pyx_callargs+__pyx_t_13, (1-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1486, __pyx_L15_except_error)
          __Pyx_GOTREF(__pyx_t_5);
        }
        __Pyx_Raise(__pyx_t_5, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __PYX_ERR(0, 1486, __pyx_L15_except_error)
      }
      goto __pyx_L15_except_error;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1483
 *             # TODO: the pending list should be lexically sorted?  the
 *             # interface doesn't require it.
 *             try:             # <<<<<<<<<<<<<<
 *                 self.current_root = self.search_specific_files.pop()
 *             except KeyError, _:
*/
      __pyx_L15_except_error:;
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      goto __pyx_L1_error;
      __pyx_L18_try_end:;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1487
 *             except KeyError, _:
 *                 raise StopIteration()
 *             self.searched_specific_files.add(self.current_root)             # <<<<<<<<<<<<<<
 *             # process the entries for this containing directory: the rest will be
 *             # found by their parents recursively.
*/
    __pyx_t_4 = __pyx_v_self->searched_specific_files;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_13 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_self->current_root};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1487, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1490
 *             # process the entries for this containing directory: the rest will be
 *             # found by their parents recursively.
 *             self.root_entries = self.state._entries_for_path(self.current_root)             # <<<<<<<<<<<<<<
 *             self.root_entries_len = len(self.root_entries)
 *             self.current_root_unicode = self.current_root.decode('utf8')
*/
    __pyx_t_4 = __pyx_v_self->state;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_13 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_self->current_root};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_entries_for_path, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->root_entries);
    __Pyx_DECREF(__pyx_v_self->root_entries);
    __pyx_v_self->root_entries = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1491
 *             # found by their parents recursively.
 *             self.root_entries = self.state._entries_for_path(self.current_root)
 *             self.root_entries_len = len(self.root_entries)             # <<<<<<<<<<<<<<
 *             self.current_root_unicode = self.current_root.decode('utf8')
 *             self.root_abspath = self.tree.abspath(self.current_root_unicode)
*/
    __pyx_t_1 = __pyx_v_self->root_entries;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_14 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1491, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_self->root_entries_len = __pyx_t_14;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1492
 *             self.root_entries = self.state._entries_for_path(self.current_root)
 *             self.root_entries_len = len(self.root_entries)
 *             self.current_root_unicode = self.current_root.decode('utf8')             # <<<<<<<<<<<<<<
 *             self.root_abspath = self.tree.abspath(self.current_root_unicode)
 *             try:
*/
    __pyx_t_4 = __pyx_v_self->current_root;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_13 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_n_u_utf8};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_decode, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1492, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->current_root_unicode);
    __Pyx_DECREF(__pyx_v_self->current_root_unicode);
    __pyx_v_self->current_root_unicode = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1493
 *             self.root_entries_len = len(self.root_entries)
 *             self.current_root_unicode = self.current_root.decode('utf8')
 *             self.root_abspath = self.tree.abspath(self.current_root_unicode)             # <<<<<<<<<<<<<<
 *             try:
 *                 root_stat = os.lstat(self.root_abspath)
*/
    __pyx_t_4 = __pyx_v_self->tree;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_13 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_self->current_root_unicode};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_abspath, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1493, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->root_abspath);
    __Pyx_DECREF(__pyx_v_self->root_abspath);
    __pyx_v_self->root_abspath = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1494
 *             self.current_root_unicode = self.current_root.decode('utf8')
 *             self.root_abspath = self.tree.abspath(self.current_root_unicode)
 *             try:             # <<<<<<<<<<<<<<
 *                 root_stat = os.lstat(self.root_abspath)
 *             except OSError, e:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1495
 *             self.root_abspath = self.tree.abspath(self.current_root_unicode)
 *             try:
 *                 root_stat = os.lstat(self.root_abspath)             # <<<<<<<<<<<<<<
 *             except OSError, e:
 *                 if e.errno == errno.ENOENT:
*/
        __pyx_t_4 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1495, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_lstat); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1495, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_13 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
          assert(__pyx_t_4);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
          __pyx_t_13 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_self->root_abspath};
          __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1495, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_1);
        }
        __pyx_v_root_stat = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1494
 *             self.current_root_unicode = self.current_root.decode('utf8')
 *             self.root_abspath = self.tree.abspath(self.current_root_unicode)
 *             try:             # <<<<<<<<<<<<<<
 *                 root_stat = os.lstat(self.root_abspath)
 *             except OSError, e:
*/
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1504
 *                     raise
 *             else:
 *                 self.root_dir_info = (b'', self.current_root,             # <<<<<<<<<<<<<<
 *                     osutils.file_kind_from_stat_mode(root_stat.st_mode), root_stat,
 *                     self.root_abspath)
*/
      /*else:*/ {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1505
 *             else:
 *                 self.root_dir_info = (b'', self.current_root,
 *                     osutils.file_kind_from_stat_mode(root_stat.st_mode), root_stat,             # <<<<<<<<<<<<<<
 *                     self.root_abspath)
 *                 if self.root_dir_info[2] == 'directory':
*/
        __pyx_t_5 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_osutils); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1505, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_file_kind_from_stat_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1505, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_root_stat, __pyx_mstate_global->__pyx_n_u_st_mode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1505, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_13 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
          assert(__pyx_t_5);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
          __pyx_t_13 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
          __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1505, __pyx_L23_except_error)
          __Pyx_GOTREF(__pyx_t_1);
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1504
 *                     raise
 *             else:
 *                 self.root_dir_info = (b'', self.current_root,             # <<<<<<<<<<<<<<
 *                     osutils.file_kind_from_stat_mode(root_stat.st_mode), root_stat,
 *                     self.root_abspath)
*/
        __pyx_t_3 = PyTuple_New(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1504, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b__4);
        __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_b__4);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_mstate_global->__pyx_kp_b__4) != (0)) __PYX_ERR(0, 1504, __pyx_L23_except_error);
        __Pyx_INCREF(__pyx_v_self->current_root);
        __Pyx_GIVEREF(__pyx_v_self->current_root);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self->current_root) != (0)) __PYX_ERR(0, 1504, __pyx_L23_except_error);
        __Pyx_GIVEREF(__pyx_t_1);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_1) != (0)) __PYX_ERR(0, 1504, __pyx_L23_except_error);
        __Pyx_INCREF(__pyx_v_root_stat);
        __Pyx_GIVEREF(__pyx_v_root_stat);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_root_stat) != (0)) __PYX_ERR(0, 1504, __pyx_L23_except_error);
        __Pyx_INCREF(__pyx_v_self->root_abspath);
        __Pyx_GIVEREF(__pyx_v_self->root_abspath);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_v_self->root_abspath) != (0)) __PYX_ERR(0, 1504, __pyx_L23_except_error);
        __pyx_t_1 = 0;
        __Pyx_GIVEREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_v_self->root_dir_info);
        __Pyx_DECREF(__pyx_v_self->root_dir_info);
        __pyx_v_self->root_dir_info = __pyx_t_3;
        __pyx_t_3 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1507
 *                     osutils.file_kind_from_stat_mode(root_stat.st_mode), root_stat,
 *                     self.root_abspath)
 *                 if self.root_dir_info[2] == 'directory':             # <<<<<<<<<<<<<<
 *                     if self.tree._directory_is_tree_reference(
 *                         self.current_root_unicode):
*/
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_self->root_dir_info, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1507, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_directory, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1507, __pyx_L23_except_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1508
 *                     self.root_abspath)
 *                 if self.root_dir_info[2] == 'directory':
 *                     if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *                         self.current_root_unicode):
 *                         self.root_dir_info = self.root_dir_info[:2] + \
*/
          __pyx_t_1 = __pyx_v_self->tree;
          __Pyx_INCREF(__pyx_t_1);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1509
 *                 if self.root_dir_info[2] == 'directory':
 *                     if self.tree._directory_is_tree_reference(
 *                         self.current_root_unicode):             # <<<<<<<<<<<<<<
 *                         self.root_dir_info = self.root_dir_info[:2] + \
 *                             ('tree-reference',) + self.root_dir_info[3:]
*/
          __pyx_t_13 = 0;
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_self->current_root_unicode};
            __pyx_t_3 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_directory_is_tree_reference, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1508, __pyx_L23_except_error)
            __Pyx_GOTREF(__pyx_t_3);
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1508
 *                     self.root_abspath)
 *                 if self.root_dir_info[2] == 'directory':
 *                     if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *                         self.current_root_unicode):
 *                         self.root_dir_info = self.root_dir_info[:2] + \
*/
          __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1508, __pyx_L23_except_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (__pyx_t_2) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1510
 *                     if self.tree._directory_is_tree_reference(
 *                         self.current_root_unicode):
 *                         self.root_dir_info = self.root_dir_info[:2] + \             # <<<<<<<<<<<<<<
 *                             ('tree-reference',) + self.root_dir_info[3:]
 *             if not self.root_entries and not self.root_dir_info:
*/
            __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_self->root_dir_info, 0, 2, NULL, NULL, &__pyx_mstate_global->__pyx_slice[1], 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1510, __pyx_L23_except_error)
            __Pyx_GOTREF(__pyx_t_3);

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1511
 *                         self.current_root_unicode):
 *                         self.root_dir_info = self.root_dir_info[:2] + \
 *                             ('tree-reference',) + self.root_dir_info[3:]             # <<<<<<<<<<<<<<
 *             if not self.root_entries and not self.root_dir_info:
 *                 # this specified path is not present at all, skip it.
*/
            __pyx_t_1 = PyNumber_Add(__pyx_t_3, __pyx_mstate_global->__pyx_tuple[5]); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1510, __pyx_L23_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_self->root_dir_info, 3, 0, NULL, NULL, &__pyx_mstate_global->__pyx_slice[2], 1, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1511, __pyx_L23_except_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_4 = PyNumber_Add(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1511, __pyx_L23_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1510
 *                     if self.tree._directory_is_tree_reference(
 *                         self.current_root_unicode):
 *                         self.root_dir_info = self.root_dir_info[:2] + \             # <<<<<<<<<<<<<<
 *                             ('tree-reference',) + self.root_dir_info[3:]
 *             if not self.root_entries and not self.root_dir_info:
*/
            __Pyx_GIVEREF(__pyx_t_4);
            __Pyx_GOTREF(__pyx_v_self->root_dir_info);
            __Pyx_DECREF(__pyx_v_self->root_dir_info);
            __pyx_v_self->root_dir_info = __pyx_t_4;
            __pyx_t_4 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1508
 *                     self.root_abspath)
 *                 if self.root_dir_info[2] == 'directory':
 *                     if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *                         self.current_root_unicode):
 *                         self.root_dir_info = self.root_dir_info[:2] + \
*/
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1507
 *                     osutils.file_kind_from_stat_mode(root_stat.st_mode), root_stat,
 *                     self.root_abspath)
 *                 if self.root_dir_info[2] == 'directory':             # <<<<<<<<<<<<<<
 *                     if self.tree._directory_is_tree_reference(
 *                         self.current_root_unicode):
*/
        }
      }
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L26_try_end;
      __pyx_L21_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1496
 *             try:
 *                 root_stat = os.lstat(self.root_abspath)
 *             except OSError, e:             # <<<<<<<<<<<<<<
 *                 if e.errno == errno.ENOENT:
 *                     # the path does not exist: let _process_entry know that.
*/
      __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_OSError))));
      if (__pyx_t_7) {
        __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._iter_next", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_3, &__pyx_t_1) < 0) __PYX_ERR(0, 1496, __pyx_L23_except_error)
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_e = __pyx_t_3;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1497
 *                 root_stat = os.lstat(self.root_abspath)
 *             except OSError, e:
 *                 if e.errno == errno.ENOENT:             # <<<<<<<<<<<<<<
 *                     # the path does not exist: let _process_entry know that.
 *                     self.root_dir_info = None
*/
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_mstate_global->__pyx_n_u_errno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1497, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_errno); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1497, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_ENOENT); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1497, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_12 = PyObject_RichCompare(__pyx_t_5, __pyx_t_15, Py_EQ); __Pyx_XGOTREF(__pyx_t_12); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1497, __pyx_L23_except_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1497, __pyx_L23_except_error)
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (likely(__pyx_t_2)) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1499
 *                 if e.errno == errno.ENOENT:
 *                     # the path does not exist: let _process_entry know that.
 *                     self.root_dir_info = None             # <<<<<<<<<<<<<<
 *                 else:
 *                     # some other random error: hand it up.
*/
          __Pyx_INCREF(Py_None);
          __Pyx_GIVEREF(Py_None);
          __Pyx_GOTREF(__pyx_v_self->root_dir_info);
          __Pyx_DECREF(__pyx_v_self->root_dir_info);
          __pyx_v_self->root_dir_info = Py_None;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1497
 *                 root_stat = os.lstat(self.root_abspath)
 *             except OSError, e:
 *                 if e.errno == errno.ENOENT:             # <<<<<<<<<<<<<<
 *                     # the path does not exist: let _process_entry know that.
 *                     self.root_dir_info = None
*/
          goto __pyx_L31;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1502
 *                 else:
 *                     # some other random error: hand it up.
 *                     raise             # <<<<<<<<<<<<<<
 *             else:
 *                 self.root_dir_info = (b'', self.current_root,
*/
        /*else*/ {
          __Pyx_GIVEREF(__pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_3);
          __Pyx_XGIVEREF(__pyx_t_1);
          __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_3, __pyx_t_1);
          __pyx_t_4 = 0;  __pyx_t_3 = 0;  __pyx_t_1 = 0; 
          __PYX_ERR(0, 1502, __pyx_L23_except_error)
        }
        __pyx_L31:;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L22_exception_handled;
      }
      goto __pyx_L23_except_error;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1494
 *             self.current_root_unicode = self.current_root.decode('utf8')
 *             self.root_abspath = self.tree.abspath(self.current_root_unicode)
 *             try:             # <<<<<<<<<<<<<<
 *                 root_stat = os.lstat(self.root_abspath)
 *             except OSError, e:
*/
      __pyx_L23_except_error:;
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      goto __pyx_L1_error;
      __pyx_L22_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      __pyx_L26_try_end:;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1512
 *                         self.root_dir_info = self.root_dir_info[:2] + \
 *                             ('tree-reference',) + self.root_dir_info[3:]
 *             if not self.root_entries and not self.root_dir_info:             # <<<<<<<<<<<<<<
 *                 # this specified path is not present at all, skip it.
 *                 # (tail recursion, can do a loop once the full structure is
*/
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_self->root_entries); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1512, __pyx_L1_error)
    __pyx_t_16 = (!__pyx_t_8);
    if (__pyx_t_16) {
    } else {
      __pyx_t_2 = __pyx_t_16;
      goto __pyx_L33_bool_binop_done;
    }
    __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_v_self->root_dir_info); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 1512, __pyx_L1_error)
    __pyx_t_8 = (!__pyx_t_16);
    __pyx_t_2 = __pyx_t_8;
    __pyx_L33_bool_binop_done:;
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1516
 *                 # (tail recursion, can do a loop once the full structure is
 *                 # known).
 *                 return self._iter_next()             # <<<<<<<<<<<<<<
 *             path_handled = 0
 *             self.root_entries_pos = 0
*/
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_iter_next(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1516, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1512
 *                         self.root_dir_info = self.root_dir_info[:2] + \
 *                             ('tree-reference',) + self.root_dir_info[3:]
 *             if not self.root_entries and not self.root_dir_info:             # <<<<<<<<<<<<<<
 *                 # this specified path is not present at all, skip it.
 *                 # (tail recursion, can do a loop once the full structure is
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1517
 *                 # known).
 *                 return self._iter_next()
 *             path_handled = 0             # <<<<<<<<<<<<<<
 *             self.root_entries_pos = 0
 *             # XXX Clarity: This loop is duplicated a out the self.current_root
*/
    __pyx_v_path_handled = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1518
 *                 return self._iter_next()
 *             path_handled = 0
 *             self.root_entries_pos = 0             # <<<<<<<<<<<<<<
 *             # XXX Clarity: This loop is duplicated a out the self.current_root
 *             # is None guard above: if we return from it, it completes there
*/
    __pyx_v_self->root_entries_pos = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1523
 *             # (and the following if block cannot trigger because
 *             # path_handled must be true, so the if block is not # duplicated.
 *             while self.root_entries_pos < self.root_entries_len:             # <<<<<<<<<<<<<<
 *                 entry = self.root_entries[self.root_entries_pos]
 *                 self.root_entries_pos = self.root_entries_pos + 1
*/
    while (1) {
      __pyx_t_2 = (__pyx_v_self->root_entries_pos < __pyx_v_self->root_entries_len);
      if (!__pyx_t_2) break;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1524
 *             # path_handled must be true, so the if block is not # duplicated.
 *             while self.root_entries_pos < self.root_entries_len:
 *                 entry = self.root_entries[self.root_entries_pos]             # <<<<<<<<<<<<<<
 *                 self.root_entries_pos = self.root_entries_pos + 1
 *                 result, changed = self._process_entry(entry, self.root_dir_info)
*/
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self->root_entries, __pyx_v_self->root_entries_pos, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_entry, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1525
 *             while self.root_entries_pos < self.root_entries_len:
 *                 entry = self.root_entries[self.root_entries_pos]
 *                 self.root_entries_pos = self.root_entries_pos + 1             # <<<<<<<<<<<<<<
 *                 result, changed = self._process_entry(entry, self.root_dir_info)
 *                 if changed is not None:
*/
      __pyx_v_self->root_entries_pos = (__pyx_v_self->root_entries_pos + 1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1526
 *                 entry = self.root_entries[self.root_entries_pos]
 *                 self.root_entries_pos = self.root_entries_pos + 1
 *                 result, changed = self._process_entry(entry, self.root_dir_info)             # <<<<<<<<<<<<<<
 *                 if changed is not None:
 *                     path_handled = -1
*/
      __pyx_t_1 = __pyx_v_self->root_dir_info;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_process_entry(__pyx_v_self, __pyx_v_entry, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1526, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
        PyObject* sequence = __pyx_t_3;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1526, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0);
          __Pyx_INCREF(__pyx_t_1);
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1);
          __Pyx_INCREF(__pyx_t_4);
        } else {
          __pyx_t_1 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1526, __pyx_L1_error)
          __Pyx_XGOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1526, __pyx_L1_error)
          __Pyx_XGOTREF(__pyx_t_4);
        }
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1526, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1526, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_12 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1526, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_6 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_12);
        index = 0; __pyx_t_1 = __pyx_t_6(__pyx_t_12); if (unlikely(!__pyx_t_1)) goto __pyx_L37_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        index = 1; __pyx_t_4 = __pyx_t_6(__pyx_t_12); if (unlikely(!__pyx_t_4)) goto __pyx_L37_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_12), 2) < (0)) __PYX_ERR(0, 1526, __pyx_L1_error)
        __pyx_t_6 = NULL;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        goto __pyx_L38_unpacking_done;
        __pyx_L37_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_6 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1526, __pyx_L1_error)
        __pyx_L38_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_result, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_changed, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1527
 *                 self.root_entries_pos = self.root_entries_pos + 1
 *                 result, changed = self._process_entry(entry, self.root_dir_info)
 *                 if changed is not None:             # <<<<<<<<<<<<<<
 *                     path_handled = -1
 *                     if changed:
*/
      __pyx_t_2 = (__pyx_v_changed != Py_None);
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1528
 *                 result, changed = self._process_entry(entry, self.root_dir_info)
 *                 if changed is not None:
 *                     path_handled = -1             # <<<<<<<<<<<<<<
 *                     if changed:
 *                         self._gather_result_for_consistency(result)
*/
        __pyx_v_path_handled = -1;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1529
 *                 if changed is not None:
 *                     path_handled = -1
 *                     if changed:             # <<<<<<<<<<<<<<
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:
*/
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1529, __pyx_L1_error)
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1530
 *                     path_handled = -1
 *                     if changed:
 *                         self._gather_result_for_consistency(result)             # <<<<<<<<<<<<<<
 *                     if changed or self.include_unchanged:
 *                         return result
*/
          __pyx_t_7 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_gather_result_for_consistency(__pyx_v_self, __pyx_v_result); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1530, __pyx_L1_error)

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1529
 *                 if changed is not None:
 *                     path_handled = -1
 *                     if changed:             # <<<<<<<<<<<<<<
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:
*/
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1531
 *                     if changed:
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                         return result
 *             # handle unversioned specified paths:
*/
        __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1531, __pyx_L1_error)
        if (!__pyx_t_8) {
        } else {
          __pyx_t_2 = __pyx_t_8;
          goto __pyx_L42_bool_binop_done;
        }
        __pyx_t_8 = (__pyx_v_self->include_unchanged != 0);
        __pyx_t_2 = __pyx_t_8;
        __pyx_L42_bool_binop_done:;
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1532
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:
 *                         return result             # <<<<<<<<<<<<<<
 *             # handle unversioned specified paths:
 *             if self.want_unversioned and not path_handled and self.root_dir_info:
*/
          __Pyx_XDECREF(__pyx_r);
          __Pyx_INCREF(__pyx_v_result);
          __pyx_r = __pyx_v_result;
          goto __pyx_L0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1531
 *                     if changed:
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                         return result
 *             # handle unversioned specified paths:
*/
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1527
 *                 self.root_entries_pos = self.root_entries_pos + 1
 *                 result, changed = self._process_entry(entry, self.root_dir_info)
 *                 if changed is not None:             # <<<<<<<<<<<<<<
 *                     path_handled = -1
 *                     if changed:
*/
      }
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1534
 *                         return result
 *             # handle unversioned specified paths:
 *             if self.want_unversioned and not path_handled and self.root_dir_info:             # <<<<<<<<<<<<<<
 *                 new_executable = bool(
 *                     stat.S_ISREG(self.root_dir_info[3].st_mode)
*/
    __pyx_t_8 = (__pyx_v_self->want_unversioned != 0);
    if (__pyx_t_8) {
    } else {
      __pyx_t_2 = __pyx_t_8;
      goto __pyx_L45_bool_binop_done;
    }
    __pyx_t_8 = (!(__pyx_v_path_handled != 0));
    if (__pyx_t_8) {
    } else {
      __pyx_t_2 = __pyx_t_8;
      goto __pyx_L45_bool_binop_done;
    }
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_self->root_dir_info); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1534, __pyx_L1_error)
    __pyx_t_2 = __pyx_t_8;
    __pyx_L45_bool_binop_done:;
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1536
 *             if self.want_unversioned and not path_handled and self.root_dir_info:
 *                 new_executable = bool(
 *                     stat.S_ISREG(self.root_dir_info[3].st_mode)             # <<<<<<<<<<<<<<
 *                     and stat.S_IEXEC & self.root_dir_info[3].st_mode)
 *                 return InventoryTreeChange(
*/
      __pyx_t_4 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_stat); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1536, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_S_ISREG); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1536, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self->root_dir_info, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1536, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_st_mode); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1536, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_13 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_12);
        assert(__pyx_t_4);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_12, __pyx__function);
        __pyx_t_13 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_15};
        __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_12, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1536, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
      }
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1536, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_8) {
      } else {
        __pyx_t_2 = __pyx_t_8;
        goto __pyx_L48_bool_binop_done;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1537
 *                 new_executable = bool(
 *                     stat.S_ISREG(self.root_dir_info[3].st_mode)
 *                     and stat.S_IEXEC & self.root_dir_info[3].st_mode)             # <<<<<<<<<<<<<<
 *                 return InventoryTreeChange(
 *                        None,
*/
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_stat); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1537, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_S_IEXEC); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1537, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_self->root_dir_info, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1537, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_st_mode); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1537, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyNumber_And(__pyx_t_12, __pyx_t_15); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1537, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1537, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_2 = __pyx_t_8;
      __pyx_L48_bool_binop_done:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1535
 *             # handle unversioned specified paths:
 *             if self.want_unversioned and not path_handled and self.root_dir_info:
 *                 new_executable = bool(             # <<<<<<<<<<<<<<
 *                     stat.S_ISREG(self.root_dir_info[3].st_mode)
 *                     and stat.S_IEXEC & self.root_dir_info[3].st_mode)
*/
      __pyx_v_new_executable = (!(!__pyx_t_2));

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1538
 *                     stat.S_ISREG(self.root_dir_info[3].st_mode)
 *                     and stat.S_IEXEC & self.root_dir_info[3].st_mode)
 *                 return InventoryTreeChange(             # <<<<<<<<<<<<<<
 *                        None,
 *                        (None, self.current_root_unicode),
*/
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_15 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_InventoryTreeChange); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1538, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1540
 *                 return InventoryTreeChange(
 *                        None,
 *                        (None, self.current_root_unicode),             # <<<<<<<<<<<<<<
 *                        True,
 *                        (False, False),
*/
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1540, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, Py_None) != (0)) __PYX_ERR(0, 1540, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_self->current_root_unicode);
      __Pyx_GIVEREF(__pyx_v_self->current_root_unicode);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_self->current_root_unicode) != (0)) __PYX_ERR(0, 1540, __pyx_L1_error);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1544
 *                        (False, False),
 *                        (None, None),
 *                        (None, splitpath(self.current_root_unicode)[-1]),             # <<<<<<<<<<<<<<
 *                        (None, self.root_dir_info[2]),
 *                        (None, new_executable)
*/
      __pyx_t_5 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_splitpath); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_13 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_17))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_17);
        assert(__pyx_t_5);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_17);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_17, __pyx__function);
        __pyx_t_13 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_self->current_root_unicode};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_17, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1544, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_1, -1L, long, 1, __Pyx_PyLong_From_long, 0, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, Py_None) != (0)) __PYX_ERR(0, 1544, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_17);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_17) != (0)) __PYX_ERR(0, 1544, __pyx_L1_error);
      __pyx_t_17 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1545
 *                        (None, None),
 *                        (None, splitpath(self.current_root_unicode)[-1]),
 *                        (None, self.root_dir_info[2]),             # <<<<<<<<<<<<<<
 *                        (None, new_executable)
 *                       )
*/
      __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_self->root_dir_info, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1545, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1545, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, Py_None) != (0)) __PYX_ERR(0, 1545, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_17);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_17) != (0)) __PYX_ERR(0, 1545, __pyx_L1_error);
      __pyx_t_17 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1546
 *                        (None, splitpath(self.current_root_unicode)[-1]),
 *                        (None, self.root_dir_info[2]),
 *                        (None, new_executable)             # <<<<<<<<<<<<<<
 *                       )
 *             # If we reach here, the outer flow continues, which enters into the
*/
      __pyx_t_17 = __Pyx_PyBool_FromLong(__pyx_v_new_executable); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1546, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_18 = PyTuple_New(2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1546, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_18, 0, Py_None) != (0)) __PYX_ERR(0, 1546, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_17);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_t_17) != (0)) __PYX_ERR(0, 1546, __pyx_L1_error);
      __pyx_t_17 = 0;
      __pyx_t_13 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_12);
        assert(__pyx_t_15);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_15);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_12, __pyx__function);
        __pyx_t_13 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[9] = {__pyx_t_15, Py_None, __pyx_t_4, Py_True, __pyx_mstate_global->__pyx_tuple[6], __pyx_mstate_global->__pyx_tuple[0], __pyx_t_1, __pyx_t_5, __pyx_t_18};
        __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_12, __pyx_callargs+__pyx_t_13, (9-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1538, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
      }
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1534
 *                         return result
 *             # handle unversioned specified paths:
 *             if self.want_unversioned and not path_handled and self.root_dir_info:             # <<<<<<<<<<<<<<
 *                 new_executable = bool(
 *                     stat.S_ISREG(self.root_dir_info[3].st_mode)
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1480
 *                     return result
 *         # Have we finished the prior root, or never started one ?
 *         if self.current_root is None:             # <<<<<<<<<<<<<<
 *             # TODO: the pending list should be lexically sorted?  the
 *             # interface doesn't require it.
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1550
 *             # If we reach here, the outer flow continues, which enters into the
 *             # per-root setup logic.
 *         if (self.current_dir_info is None and self.current_block is None and not             # <<<<<<<<<<<<<<
 *             self.doing_consistency_expansion):
 *             # setup iteration of this root:
*/
  __pyx_t_8 = (__pyx_v_self->current_dir_info == Py_None);
  if (__pyx_t_8) {
  } else {
    __pyx_t_2 = __pyx_t_8;
    goto __pyx_L51_bool_binop_done;
  }
  __pyx_t_8 = (__pyx_v_self->current_block == Py_None);
  if (__pyx_t_8) {
  } else {
    __pyx_t_2 = __pyx_t_8;
    goto __pyx_L51_bool_binop_done;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1551
 *             # per-root setup logic.
 *         if (self.current_dir_info is None and self.current_block is None and not
 *             self.doing_consistency_expansion):             # <<<<<<<<<<<<<<
 *             # setup iteration of this root:
 *             self.current_dir_list = None
*/
  __pyx_t_8 = (!(__pyx_v_self->doing_consistency_expansion != 0));
  __pyx_t_2 = __pyx_t_8;
  __pyx_L51_bool_binop_done:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1550
 *             # If we reach here, the outer flow continues, which enters into the
 *             # per-root setup logic.
 *         if (self.current_dir_info is None and self.current_block is None and not             # <<<<<<<<<<<<<<
 *             self.doing_consistency_expansion):
 *             # setup iteration of this root:
*/
  if (__pyx_t_2) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1553
 *             self.doing_consistency_expansion):
 *             # setup iteration of this root:
 *             self.current_dir_list = None             # <<<<<<<<<<<<<<
 *             if self.root_dir_info and self.root_dir_info[2] == 'tree-reference':
 *                 self.current_dir_info = None
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->current_dir_list);
    __Pyx_DECREF(__pyx_v_self->current_dir_list);
    __pyx_v_self->current_dir_list = Py_None;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1554
 *             # setup iteration of this root:
 *             self.current_dir_list = None
 *             if self.root_dir_info and self.root_dir_info[2] == 'tree-reference':             # <<<<<<<<<<<<<<
 *                 self.current_dir_info = None
 *             else:
*/
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_self->root_dir_info); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1554, __pyx_L1_error)
    if (__pyx_t_8) {
    } else {
      __pyx_t_2 = __pyx_t_8;
      goto __pyx_L55_bool_binop_done;
    }
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_self->root_dir_info, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_tree_reference, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1554, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __pyx_t_8;
    __pyx_L55_bool_binop_done:;
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1555
 *             self.current_dir_list = None
 *             if self.root_dir_info and self.root_dir_info[2] == 'tree-reference':
 *                 self.current_dir_info = None             # <<<<<<<<<<<<<<
 *             else:
 *                 self.dir_iterator = osutils._walkdirs_utf8(self.root_abspath,
*/
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_GOTREF(__pyx_v_self->current_dir_info);
      __Pyx_DECREF(__pyx_v_self->current_dir_info);
      __pyx_v_self->current_dir_info = Py_None;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1554
 *             # setup iteration of this root:
 *             self.current_dir_list = None
 *             if self.root_dir_info and self.root_dir_info[2] == 'tree-reference':             # <<<<<<<<<<<<<<
 *                 self.current_dir_info = None
 *             else:
*/
      goto __pyx_L54;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1557
 *                 self.current_dir_info = None
 *             else:
 *                 self.dir_iterator = osutils._walkdirs_utf8(self.root_abspath,             # <<<<<<<<<<<<<<
 *                     prefix=self.current_root)
 *                 self.path_index = 0
*/
    /*else*/ {
      __pyx_t_12 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_osutils); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1557, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_walkdirs_utf8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1557, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1558
 *             else:
 *                 self.dir_iterator = osutils._walkdirs_utf8(self.root_abspath,
 *                     prefix=self.current_root)             # <<<<<<<<<<<<<<
 *                 self.path_index = 0
 *                 try:
*/
      __pyx_t_13 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_5);
        assert(__pyx_t_12);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
        __pyx_t_13 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_12, __pyx_v_self->root_abspath};
        __pyx_t_18 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1557, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_18);
        if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_prefix, __pyx_v_self->current_root, __pyx_t_18, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 1557, __pyx_L1_error)
        __pyx_t_3 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_18);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1557, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1557
 *                 self.current_dir_info = None
 *             else:
 *                 self.dir_iterator = osutils._walkdirs_utf8(self.root_abspath,             # <<<<<<<<<<<<<<
 *                     prefix=self.current_root)
 *                 self.path_index = 0
*/
      __Pyx_GIVEREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_v_self->dir_iterator);
      __Pyx_DECREF(__pyx_v_self->dir_iterator);
      __pyx_v_self->dir_iterator = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1559
 *                 self.dir_iterator = osutils._walkdirs_utf8(self.root_abspath,
 *                     prefix=self.current_root)
 *                 self.path_index = 0             # <<<<<<<<<<<<<<
 *                 try:
 *                     self.current_dir_info = next(self.dir_iterator)
*/
      __pyx_v_self->path_index = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1560
 *                     prefix=self.current_root)
 *                 self.path_index = 0
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]
*/
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        /*try:*/ {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1561
 *                 self.path_index = 0
 *                 try:
 *                     self.current_dir_info = next(self.dir_iterator)             # <<<<<<<<<<<<<<
 *                     self.current_dir_list = self.current_dir_info[1]
 *                 except OSError, e:
*/
          __pyx_t_3 = __pyx_v_self->dir_iterator;
          __Pyx_INCREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyIter_Next(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1561, __pyx_L57_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GIVEREF(__pyx_t_5);
          __Pyx_GOTREF(__pyx_v_self->current_dir_info);
          __Pyx_DECREF(__pyx_v_self->current_dir_info);
          __pyx_v_self->current_dir_info = __pyx_t_5;
          __pyx_t_5 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1562
 *                 try:
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]             # <<<<<<<<<<<<<<
 *                 except OSError, e:
 *                     # there may be directories in the inventory even though
*/
          __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_self->current_dir_info, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1562, __pyx_L57_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_5);
          __Pyx_GOTREF(__pyx_v_self->current_dir_list);
          __Pyx_DECREF(__pyx_v_self->current_dir_list);
          __pyx_v_self->current_dir_list = __pyx_t_5;
          __pyx_t_5 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1560
 *                     prefix=self.current_root)
 *                 self.path_index = 0
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]
*/
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1586
 *                         raise
 *                 else:
 *                     if self.current_dir_info[0][0] == b'':             # <<<<<<<<<<<<<<
 *                         # remove .bzr from iteration
 *                         bzr_index = self.bisect_left(self.current_dir_list, (b'.bzr',))
*/
        /*else:*/ {
          __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_self->current_dir_info, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1586, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1586, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_2 = (__Pyx_PyBytes_Equals(__pyx_t_3, __pyx_mstate_global->__pyx_kp_b__4, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1586, __pyx_L59_except_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (__pyx_t_2) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1588
 *                     if self.current_dir_info[0][0] == b'':
 *                         # remove .bzr from iteration
 *                         bzr_index = self.bisect_left(self.current_dir_list, (b'.bzr',))             # <<<<<<<<<<<<<<
 *                         if self.current_dir_list[bzr_index][0] != b'.bzr':
 *                             raise AssertionError()
*/
            __pyx_t_5 = NULL;
            __Pyx_INCREF(__pyx_v_self->bisect_left);
            __pyx_t_18 = __pyx_v_self->bisect_left; 
            __pyx_t_13 = 1;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_18))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_18);
              assert(__pyx_t_5);
              PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_18);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(__pyx__function);
              __Pyx_DECREF_SET(__pyx_t_18, __pyx__function);
              __pyx_t_13 = 0;
            }
            #endif
            {
              PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_self->current_dir_list, __pyx_mstate_global->__pyx_tuple[7]};
              __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_18, __pyx_callargs+__pyx_t_13, (3-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1588, __pyx_L59_except_error)
              __Pyx_GOTREF(__pyx_t_3);
            }
            __pyx_v_bzr_index = __pyx_t_3;
            __pyx_t_3 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1589
 *                         # remove .bzr from iteration
 *                         bzr_index = self.bisect_left(self.current_dir_list, (b'.bzr',))
 *                         if self.current_dir_list[bzr_index][0] != b'.bzr':             # <<<<<<<<<<<<<<
 *                             raise AssertionError()
 *                         del self.current_dir_list[bzr_index]
*/
            __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_self->current_dir_list, __pyx_v_bzr_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1589, __pyx_L59_except_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_18 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1589, __pyx_L59_except_error)
            __Pyx_GOTREF(__pyx_t_18);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_2 = (__Pyx_PyBytes_Equals(__pyx_t_18, __pyx_mstate_global->__pyx_kp_b_bzr, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1589, __pyx_L59_except_error)
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            if (unlikely(__pyx_t_2)) {

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1590
 *                         bzr_index = self.bisect_left(self.current_dir_list, (b'.bzr',))
 *                         if self.current_dir_list[bzr_index][0] != b'.bzr':
 *                             raise AssertionError()             # <<<<<<<<<<<<<<
 *                         del self.current_dir_list[bzr_index]
 *             initial_key = (self.current_root, b'', b'')
*/
              __pyx_t_3 = NULL;
              __pyx_t_13 = 1;
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
                __pyx_t_18 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_13, (1-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1590, __pyx_L59_except_error)
                __Pyx_GOTREF(__pyx_t_18);
              }
              __Pyx_Raise(__pyx_t_18, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              __PYX_ERR(0, 1590, __pyx_L59_except_error)

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1589
 *                         # remove .bzr from iteration
 *                         bzr_index = self.bisect_left(self.current_dir_list, (b'.bzr',))
 *                         if self.current_dir_list[bzr_index][0] != b'.bzr':             # <<<<<<<<<<<<<<
 *                             raise AssertionError()
 *                         del self.current_dir_list[bzr_index]
*/
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1591
 *                         if self.current_dir_list[bzr_index][0] != b'.bzr':
 *                             raise AssertionError()
 *                         del self.current_dir_list[bzr_index]             # <<<<<<<<<<<<<<
 *             initial_key = (self.current_root, b'', b'')
 *             self.block_index, _ = self.state._find_block_index_from_key(initial_key)
*/
            if (unlikely((PyObject_DelItem(__pyx_v_self->current_dir_list, __pyx_v_bzr_index) < 0))) __PYX_ERR(0, 1591, __pyx_L59_except_error)

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1586
 *                         raise
 *                 else:
 *                     if self.current_dir_info[0][0] == b'':             # <<<<<<<<<<<<<<
 *                         # remove .bzr from iteration
 *                         bzr_index = self.bisect_left(self.current_dir_list, (b'.bzr',))
*/
          }
        }
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L62_try_end;
        __pyx_L57_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1563
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]
 *                 except OSError, e:             # <<<<<<<<<<<<<<
 *                     # there may be directories in the inventory even though
 *                     # this path is not a file on disk: so mark it as end of
*/
        __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_OSError))));
        if (__pyx_t_7) {
          __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._iter_next", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_18, &__pyx_t_3, &__pyx_t_5) < 0) __PYX_ERR(0, 1563, __pyx_L59_except_error)
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_3);
          __Pyx_XGOTREF(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_3);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1567
 *                     # this path is not a file on disk: so mark it as end of
 *                     # iterator
 *                     if e.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):             # <<<<<<<<<<<<<<
 *                         self.current_dir_info = None
 *                     elif sys.platform == 'win32':
*/
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_mstate_global->__pyx_n_u_errno); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1567, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_errno); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1567, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_ENOENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1567, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_t_12, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1567, __pyx_L59_except_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1567, __pyx_L59_except_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_8) {
          } else {
            __pyx_t_2 = __pyx_t_8;
            goto __pyx_L68_bool_binop_done;
          }
          __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_errno); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1567, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_ENOTDIR); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1567, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_t_12, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1567, __pyx_L59_except_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1567, __pyx_L59_except_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_8) {
          } else {
            __pyx_t_2 = __pyx_t_8;
            goto __pyx_L68_bool_binop_done;
          }
          __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_errno); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1567, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_EINVAL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1567, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_t_12, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1567, __pyx_L59_except_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1567, __pyx_L59_except_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_2 = __pyx_t_8;
          __pyx_L68_bool_binop_done:;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_8 = __pyx_t_2;
          if (__pyx_t_8) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1568
 *                     # iterator
 *                     if e.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):
 *                         self.current_dir_info = None             # <<<<<<<<<<<<<<
 *                     elif sys.platform == 'win32':
 *                         # on win32, python2.4 has e.errno == ERROR_DIRECTORY, but
*/
            __Pyx_INCREF(Py_None);
            __Pyx_GIVEREF(Py_None);
            __Pyx_GOTREF(__pyx_v_self->current_dir_info);
            __Pyx_DECREF(__pyx_v_self->current_dir_info);
            __pyx_v_self->current_dir_info = Py_None;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1567
 *                     # this path is not a file on disk: so mark it as end of
 *                     # iterator
 *                     if e.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):             # <<<<<<<<<<<<<<
 *                         self.current_dir_info = None
 *                     elif sys.platform == 'win32':
*/
            goto __pyx_L67;
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1569
 *                     if e.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):
 *                         self.current_dir_info = None
 *                     elif sys.platform == 'win32':             # <<<<<<<<<<<<<<
 *                         # on win32, python2.4 has e.errno == ERROR_DIRECTORY, but
 *                         # python 2.5 has e.errno == EINVAL,
*/
          __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_sys); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1569, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_platform); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1569, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_win32, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1569, __pyx_L59_except_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (likely(__pyx_t_8)) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1573
 *                         # python 2.5 has e.errno == EINVAL,
 *                         #            and e.winerror == ERROR_DIRECTORY
 *                         try:             # <<<<<<<<<<<<<<
 *                             e_winerror = e.winerror
 *                         except AttributeError, _:
*/
            {
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __Pyx_ExceptionSave(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
              __Pyx_XGOTREF(__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_20);
              __Pyx_XGOTREF(__pyx_t_21);
              /*try:*/ {

                /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1574
 *                         #            and e.winerror == ERROR_DIRECTORY
 *                         try:
 *                             e_winerror = e.winerror             # <<<<<<<<<<<<<<
 *                         except AttributeError, _:
 *                             e_winerror = None
*/
                __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_mstate_global->__pyx_n_u_winerror); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1574, __pyx_L71_error)
                __Pyx_GOTREF(__pyx_t_1);
                __pyx_v_e_winerror = __pyx_t_1;
                __pyx_t_1 = 0;

                /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1573
 *                         # python 2.5 has e.errno == EINVAL,
 *                         #            and e.winerror == ERROR_DIRECTORY
 *                         try:             # <<<<<<<<<<<<<<
 *                             e_winerror = e.winerror
 *                         except AttributeError, _:
*/
              }
              __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
              __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
              __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
              goto __pyx_L78_try_end;
              __pyx_L71_error:;
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1575
 *                         try:
 *                             e_winerror = e.winerror
 *                         except AttributeError, _:             # <<<<<<<<<<<<<<
 *                             e_winerror = None
 *                         win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)
*/
              __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_AttributeError))));
              if (__pyx_t_7) {
                __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._iter_next", __pyx_clineno, __pyx_lineno, __pyx_filename);
                if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_12, &__pyx_t_4) < 0) __PYX_ERR(0, 1575, __pyx_L73_except_error)
                __Pyx_XGOTREF(__pyx_t_1);
                __Pyx_XGOTREF(__pyx_t_12);
                __Pyx_XGOTREF(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_12);
                __pyx_v__ = __pyx_t_12;

                /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1576
 *                             e_winerror = e.winerror
 *                         except AttributeError, _:
 *                             e_winerror = None             # <<<<<<<<<<<<<<
 *                         win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)
 *                         if (e.errno in win_errors or e_winerror in win_errors):
*/
                __Pyx_INCREF(Py_None);
                __Pyx_XDECREF_SET(__pyx_v_e_winerror, Py_None);
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                goto __pyx_L72_exception_handled;
              }
              goto __pyx_L73_except_error;

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1573
 *                         # python 2.5 has e.errno == EINVAL,
 *                         #            and e.winerror == ERROR_DIRECTORY
 *                         try:             # <<<<<<<<<<<<<<
 *                             e_winerror = e.winerror
 *                         except AttributeError, _:
*/
              __pyx_L73_except_error:;
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
              goto __pyx_L59_except_error;
              __pyx_L72_exception_handled:;
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
              __pyx_L78_try_end:;
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1577
 *                         except AttributeError, _:
 *                             e_winerror = None
 *                         win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)             # <<<<<<<<<<<<<<
 *                         if (e.errno in win_errors or e_winerror in win_errors):
 *                             self.current_dir_info = None
*/
            __pyx_t_4 = __Pyx_PyLong_From_int(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx_ERROR_DIRECTORY); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1577, __pyx_L59_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_12 = __Pyx_PyLong_From_int(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx_ERROR_PATH_NOT_FOUND); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1577, __pyx_L59_except_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1577, __pyx_L59_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_4);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4) != (0)) __PYX_ERR(0, 1577, __pyx_L59_except_error);
            __Pyx_GIVEREF(__pyx_t_12);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_12) != (0)) __PYX_ERR(0, 1577, __pyx_L59_except_error);
            __pyx_t_4 = 0;
            __pyx_t_12 = 0;
            __pyx_v_win_errors = __pyx_t_1;
            __pyx_t_1 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1578
 *                             e_winerror = None
 *                         win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)
 *                         if (e.errno in win_errors or e_winerror in win_errors):             # <<<<<<<<<<<<<<
 *                             self.current_dir_info = None
 *                         else:
*/
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_mstate_global->__pyx_n_u_errno); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1578, __pyx_L59_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_v_win_errors, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1578, __pyx_L59_except_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (!__pyx_t_2) {
            } else {
              __pyx_t_8 = __pyx_t_2;
              goto __pyx_L82_bool_binop_done;
            }
            __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_e_winerror, __pyx_v_win_errors, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1578, __pyx_L59_except_error)
            __pyx_t_8 = __pyx_t_2;
            __pyx_L82_bool_binop_done:;
            if (likely(__pyx_t_8)) {

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1579
 *                         win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)
 *                         if (e.errno in win_errors or e_winerror in win_errors):
 *                             self.current_dir_info = None             # <<<<<<<<<<<<<<
 *                         else:
 *                             # Will this really raise the right exception ?
*/
              __Pyx_INCREF(Py_None);
              __Pyx_GIVEREF(Py_None);
              __Pyx_GOTREF(__pyx_v_self->current_dir_info);
              __Pyx_DECREF(__pyx_v_self->current_dir_info);
              __pyx_v_self->current_dir_info = Py_None;

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1578
 *                             e_winerror = None
 *                         win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)
 *                         if (e.errno in win_errors or e_winerror in win_errors):             # <<<<<<<<<<<<<<
 *                             self.current_dir_info = None
 *                         else:
*/
              goto __pyx_L81;
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1582
 *                         else:
 *                             # Will this really raise the right exception ?
 *                             raise             # <<<<<<<<<<<<<<
 *                     else:
 *                         raise
*/
            /*else*/ {
              __Pyx_GIVEREF(__pyx_t_18);
              __Pyx_GIVEREF(__pyx_t_3);
              __Pyx_XGIVEREF(__pyx_t_5);
              __Pyx_ErrRestoreWithState(__pyx_t_18, __pyx_t_3, __pyx_t_5);
              __pyx_t_18 = 0;  __pyx_t_3 = 0;  __pyx_t_5 = 0; 
              __PYX_ERR(0, 1582, __pyx_L59_except_error)
            }
            __pyx_L81:;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1569
 *                     if e.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):
 *                         self.current_dir_info = None
 *                     elif sys.platform == 'win32':             # <<<<<<<<<<<<<<
 *                         # on win32, python2.4 has e.errno == ERROR_DIRECTORY, but
 *                         # python 2.5 has e.errno == EINVAL,
*/
            goto __pyx_L67;
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1584
 *                             raise
 *                     else:
 *                         raise             # <<<<<<<<<<<<<<
 *                 else:
 *                     if self.current_dir_info[0][0] == b'':
*/
          /*else*/ {
            __Pyx_GIVEREF(__pyx_t_18);
            __Pyx_GIVEREF(__pyx_t_3);
            __Pyx_XGIVEREF(__pyx_t_5);
            __Pyx_ErrRestoreWithState(__pyx_t_18, __pyx_t_3, __pyx_t_5);
            __pyx_t_18 = 0;  __pyx_t_3 = 0;  __pyx_t_5 = 0; 
            __PYX_ERR(0, 1584, __pyx_L59_except_error)
          }
          __pyx_L67:;
          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          goto __pyx_L58_exception_handled;
        }
        goto __pyx_L59_except_error;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1560
 *                     prefix=self.current_root)
 *                 self.path_index = 0
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]
*/
        __pyx_L59_except_error:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        goto __pyx_L1_error;
        __pyx_L58_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        __pyx_L62_try_end:;
      }
    }
    __pyx_L54:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1592
 *                             raise AssertionError()
 *                         del self.current_dir_list[bzr_index]
 *             initial_key = (self.current_root, b'', b'')             # <<<<<<<<<<<<<<
 *             self.block_index, _ = self.state._find_block_index_from_key(initial_key)
 *             if self.block_index == 0:
*/
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1592, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_self->current_root);
    __Pyx_GIVEREF(__pyx_v_self->current_root);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_self->current_root) != (0)) __PYX_ERR(0, 1592, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b__4);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_b__4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_mstate_global->__pyx_kp_b__4) != (0)) __PYX_ERR(0, 1592, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b__4);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_b__4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_mstate_global->__pyx_kp_b__4) != (0)) __PYX_ERR(0, 1592, __pyx_L1_error);
    __pyx_v_initial_key = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1593
 *                         del self.current_dir_list[bzr_index]
 *             initial_key = (self.current_root, b'', b'')
 *             self.block_index, _ = self.state._find_block_index_from_key(initial_key)             # <<<<<<<<<<<<<<
 *             if self.block_index == 0:
 *                 # we have processed the total root already, but because the
*/
    __pyx_t_3 = __pyx_v_self->state;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_13 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_initial_key};
      __pyx_t_5 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_find_block_index_from_key, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1593, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
      PyObject* sequence = __pyx_t_5;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1593, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_t_18 = PyTuple_GET_ITEM(sequence, 1);
        __Pyx_INCREF(__pyx_t_18);
      } else {
        __pyx_t_3 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1593, __pyx_L1_error)
        __Pyx_XGOTREF(__pyx_t_3);
        __pyx_t_18 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
        if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1593, __pyx_L1_error)
        __Pyx_XGOTREF(__pyx_t_18);
      }
      #else
      __pyx_t_3 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1593, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_18 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1593, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      #endif
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_1 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1593, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_1);
      index = 0; __pyx_t_3 = __pyx_t_6(__pyx_t_1); if (unlikely(!__pyx_t_3)) goto __pyx_L84_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_18 = __pyx_t_6(__pyx_t_1); if (unlikely(!__pyx_t_18)) goto __pyx_L84_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_18);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_1), 2) < (0)) __PYX_ERR(0, 1593, __pyx_L1_error)
      __pyx_t_6 = NULL;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L85_unpacking_done;
      __pyx_L84_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_6 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1593, __pyx_L1_error)
      __pyx_L85_unpacking_done:;
    }
    __pyx_t_7 = __Pyx_PyLong_As_int(__pyx_t_3); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1593, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->block_index = __pyx_t_7;
    __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_18);
    __pyx_t_18 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1594
 *             initial_key = (self.current_root, b'', b'')
 *             self.block_index, _ = self.state._find_block_index_from_key(initial_key)
 *             if self.block_index == 0:             # <<<<<<<<<<<<<<
 *                 # we have processed the total root already, but because the
 *                 # initial key matched it we should skip it here.
*/
    __pyx_t_8 = (__pyx_v_self->block_index == 0);
    if (__pyx_t_8) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1597
 *                 # we have processed the total root already, but because the
 *                 # initial key matched it we should skip it here.
 *                 self.block_index = self.block_index + 1             # <<<<<<<<<<<<<<
 *             self._update_current_block()
 *         # walk until both the directory listing and the versioned metadata
*/
      __pyx_v_self->block_index = (__pyx_v_self->block_index + 1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1594
 *             initial_key = (self.current_root, b'', b'')
 *             self.block_index, _ = self.state._find_block_index_from_key(initial_key)
 *             if self.block_index == 0:             # <<<<<<<<<<<<<<
 *                 # we have processed the total root already, but because the
 *                 # initial key matched it we should skip it here.
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1598
 *                 # initial key matched it we should skip it here.
 *                 self.block_index = self.block_index + 1
 *             self._update_current_block()             # <<<<<<<<<<<<<<
 *         # walk until both the directory listing and the versioned metadata
 *         # are exhausted.
*/
    __pyx_t_7 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_update_current_block(__pyx_v_self); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1598, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1550
 *             # If we reach here, the outer flow continues, which enters into the
 *             # per-root setup logic.
 *         if (self.current_dir_info is None and self.current_block is None and not             # <<<<<<<<<<<<<<
 *             self.doing_consistency_expansion):
 *             # setup iteration of this root:
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1601
 *         # walk until both the directory listing and the versioned metadata
 *         # are exhausted.
 *         while (self.current_dir_info is not None             # <<<<<<<<<<<<<<
 *             or self.current_block is not None):
 *             # Uncommon case - a missing directory or an unversioned directory:
*/
  while (1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1602
 *         # are exhausted.
 *         while (self.current_dir_info is not None
 *             or self.current_block is not None):             # <<<<<<<<<<<<<<
 *             # Uncommon case - a missing directory or an unversioned directory:
 *             if (self.current_dir_info and self.current_block
*/
    __pyx_t_2 = (__pyx_v_self->current_dir_info != Py_None);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_8 = __pyx_t_2;
      goto __pyx_L89_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_self->current_block != Py_None);
    __pyx_t_8 = __pyx_t_2;
    __pyx_L89_bool_binop_done:;
    if (!__pyx_t_8) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1604
 *             or self.current_block is not None):
 *             # Uncommon case - a missing directory or an unversioned directory:
 *             if (self.current_dir_info and self.current_block             # <<<<<<<<<<<<<<
 *                 and self.current_dir_info[0][0] != self.current_block[0]):
 *                 # Work around pyrex broken heuristic - current_dirname has
*/
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_self->current_dir_info); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1604, __pyx_L1_error)
    if (__pyx_t_2) {
    } else {
      __pyx_t_8 = __pyx_t_2;
      goto __pyx_L92_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1605
 *             # Uncommon case - a missing directory or an unversioned directory:
 *             if (self.current_dir_info and self.current_block
 *                 and self.current_dir_info[0][0] != self.current_block[0]):             # <<<<<<<<<<<<<<
 *                 # Work around pyrex broken heuristic - current_dirname has
 *                 # the same scope as current_dirname_c
*/
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_self->current_block); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1604, __pyx_L1_error)
    if (__pyx_t_2) {
    } else {
      __pyx_t_8 = __pyx_t_2;
      goto __pyx_L92_bool_binop_done;
    }
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_self->current_dir_info, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_18 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_self->current_block, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_18, __pyx_t_5, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1605, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1605, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_8 = __pyx_t_2;
    __pyx_L92_bool_binop_done:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1604
 *             or self.current_block is not None):
 *             # Uncommon case - a missing directory or an unversioned directory:
 *             if (self.current_dir_info and self.current_block             # <<<<<<<<<<<<<<
 *                 and self.current_dir_info[0][0] != self.current_block[0]):
 *                 # Work around pyrex broken heuristic - current_dirname has
*/
    if (__pyx_t_8) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1608
 *                 # Work around pyrex broken heuristic - current_dirname has
 *                 # the same scope as current_dirname_c
 *                 current_dirname = self.current_dir_info[0][0]             # <<<<<<<<<<<<<<
 *                 current_dirname_c = PyBytes_AS_STRING_void(
 *                     <void *>current_dirname)
*/
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_self->current_dir_info, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1608, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1608, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_current_dirname, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1609
 *                 # the same scope as current_dirname_c
 *                 current_dirname = self.current_dir_info[0][0]
 *                 current_dirname_c = PyBytes_AS_STRING_void(             # <<<<<<<<<<<<<<
 *                     <void *>current_dirname)
 *                 current_blockname = self.current_block[0]
*/
      __pyx_v_current_dirname_c = PyBytes_AS_STRING(((void *)__pyx_v_current_dirname));

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1611
 *                 current_dirname_c = PyBytes_AS_STRING_void(
 *                     <void *>current_dirname)
 *                 current_blockname = self.current_block[0]             # <<<<<<<<<<<<<<
 *                 current_blockname_c = PyBytes_AS_STRING_void(
 *                     <void *>current_blockname)
*/
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_self->current_block, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1611, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_current_blockname, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1612
 *                     <void *>current_dirname)
 *                 current_blockname = self.current_block[0]
 *                 current_blockname_c = PyBytes_AS_STRING_void(             # <<<<<<<<<<<<<<
 *                     <void *>current_blockname)
 *                 # In the python generator we evaluate this if block once per
*/
      __pyx_v_current_blockname_c = PyBytes_AS_STRING(((void *)__pyx_v_current_blockname));

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1618
 *                 # evaluated once per path: we could cache the result before
 *                 # doing the while loop and probably save time.
 *                 if _cmp_by_dirs(current_dirname_c,             # <<<<<<<<<<<<<<
 *                     PyBytes_Size(current_dirname),
 *                     current_blockname_c,
*/
      __pyx_t_7 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_by_dirs(__pyx_v_current_dirname_c, PyBytes_Size(__pyx_v_current_dirname), __pyx_v_current_blockname_c, PyBytes_Size(__pyx_v_current_blockname)); if (unlikely(__pyx_t_7 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1618, __pyx_L1_error)

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1621
 *                     PyBytes_Size(current_dirname),
 *                     current_blockname_c,
 *                     PyBytes_Size(current_blockname)) < 0:             # <<<<<<<<<<<<<<
 *                     # filesystem data refers to paths not covered by the
 *                     # dirblock.  this has two possibilities:
*/
      __pyx_t_8 = (__pyx_t_7 < 0);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1618
 *                 # evaluated once per path: we could cache the result before
 *                 # doing the while loop and probably save time.
 *                 if _cmp_by_dirs(current_dirname_c,             # <<<<<<<<<<<<<<
 *                     PyBytes_Size(current_dirname),
 *                     current_blockname_c,
*/
      if (__pyx_t_8) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1632
 *                     # recurse into unknown directories.
 *                     # We are doing a loop
 *                     while self.path_index < len(self.current_dir_list):             # <<<<<<<<<<<<<<
 *                         current_path_info = self.current_dir_list[self.path_index]
 *                         # dont descend into this unversioned path if it is
*/
        while (1) {
          __pyx_t_5 = __pyx_v_self->current_dir_list;
          __Pyx_INCREF(__pyx_t_5);
          __pyx_t_14 = PyObject_Length(__pyx_t_5); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1632, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_8 = (__pyx_v_self->path_index < __pyx_t_14);
          if (!__pyx_t_8) break;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1633
 *                     # We are doing a loop
 *                     while self.path_index < len(self.current_dir_list):
 *                         current_path_info = self.current_dir_list[self.path_index]             # <<<<<<<<<<<<<<
 *                         # dont descend into this unversioned path if it is
 *                         # a dir
*/
          __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_self->current_dir_list, __pyx_v_self->path_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1633, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_XDECREF_SET(__pyx_v_current_path_info, __pyx_t_5);
          __pyx_t_5 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1636
 *                         # dont descend into this unversioned path if it is
 *                         # a dir
 *                         if current_path_info[2] in ('directory',             # <<<<<<<<<<<<<<
 *                                                     'tree-reference'):
 *                             del self.current_dir_list[self.path_index]
*/
          __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_current_path_info, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1636, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_directory, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1636, __pyx_L1_error)
          if (!__pyx_t_2) {
          } else {
            __pyx_t_8 = __pyx_t_2;
            goto __pyx_L99_bool_binop_done;
          }
          __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_5, __pyx_mstate_global->__pyx_kp_u_tree_reference, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1636, __pyx_L1_error)
          __pyx_t_8 = __pyx_t_2;
          __pyx_L99_bool_binop_done:;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_2 = __pyx_t_8;
          if (__pyx_t_2) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1638
 *                         if current_path_info[2] in ('directory',
 *                                                     'tree-reference'):
 *                             del self.current_dir_list[self.path_index]             # <<<<<<<<<<<<<<
 *                             self.path_index = self.path_index - 1
 *                         self.path_index = self.path_index + 1
*/
            if (unlikely((__Pyx_DelItemInt(__pyx_v_self->current_dir_list, __pyx_v_self->path_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_SharedReference) < 0))) __PYX_ERR(0, 1638, __pyx_L1_error)

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1639
 *                                                     'tree-reference'):
 *                             del self.current_dir_list[self.path_index]
 *                             self.path_index = self.path_index - 1             # <<<<<<<<<<<<<<
 *                         self.path_index = self.path_index + 1
 *                         if self.want_unversioned:
*/
            __pyx_v_self->path_index = (__pyx_v_self->path_index - 1);

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1636
 *                         # dont descend into this unversioned path if it is
 *                         # a dir
 *                         if current_path_info[2] in ('directory',             # <<<<<<<<<<<<<<
 *                                                     'tree-reference'):
 *                             del self.current_dir_list[self.path_index]
*/
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1640
 *                             del self.current_dir_list[self.path_index]
 *                             self.path_index = self.path_index - 1
 *                         self.path_index = self.path_index + 1             # <<<<<<<<<<<<<<
 *                         if self.want_unversioned:
 *                             if current_path_info[2] == 'directory':
*/
          __pyx_v_self->path_index = (__pyx_v_self->path_index + 1);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1641
 *                             self.path_index = self.path_index - 1
 *                         self.path_index = self.path_index + 1
 *                         if self.want_unversioned:             # <<<<<<<<<<<<<<
 *                             if current_path_info[2] == 'directory':
 *                                 if self.tree._directory_is_tree_reference(
*/
          __pyx_t_2 = (__pyx_v_self->want_unversioned != 0);
          if (__pyx_t_2) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1642
 *                         self.path_index = self.path_index + 1
 *                         if self.want_unversioned:
 *                             if current_path_info[2] == 'directory':             # <<<<<<<<<<<<<<
 *                                 if self.tree._directory_is_tree_reference(
 *                                     self.utf8_decode(current_path_info[0], 'surrogateescape')[0]):
*/
            __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_current_path_info, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1642, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_directory, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1642, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (__pyx_t_2) {

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1643
 *                         if self.want_unversioned:
 *                             if current_path_info[2] == 'directory':
 *                                 if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *                                     self.utf8_decode(current_path_info[0], 'surrogateescape')[0]):
 *                                     current_path_info = current_path_info[:2] + \
*/
              __pyx_t_3 = __pyx_v_self->tree;
              __Pyx_INCREF(__pyx_t_3);

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1644
 *                             if current_path_info[2] == 'directory':
 *                                 if self.tree._directory_is_tree_reference(
 *                                     self.utf8_decode(current_path_info[0], 'surrogateescape')[0]):             # <<<<<<<<<<<<<<
 *                                     current_path_info = current_path_info[:2] + \
 *                                         ('tree-reference',) + current_path_info[3:]
*/
              __pyx_t_1 = NULL;
              __Pyx_INCREF(__pyx_v_self->utf8_decode);
              __pyx_t_12 = __pyx_v_self->utf8_decode; 
              __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_current_path_info, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1644, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_13 = 1;
              #if CYTHON_UNPACK_METHODS
              if (likely(PyMethod_Check(__pyx_t_12))) {
                __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_12);
                assert(__pyx_t_1);
                PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_12);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(__pyx__function);
                __Pyx_DECREF_SET(__pyx_t_12, __pyx__function);
                __pyx_t_13 = 0;
              }
              #endif
              {
                PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_t_4, __pyx_mstate_global->__pyx_n_u_surrogateescape};
                __pyx_t_18 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_12, __pyx_callargs+__pyx_t_13, (3-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1644, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_18);
              }
              __pyx_t_12 = __Pyx_GetItemInt(__pyx_t_18, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1644, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              __pyx_t_13 = 0;
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_12};
                __pyx_t_5 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_directory_is_tree_reference, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1643, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_5);
              }

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1643
 *                         if self.want_unversioned:
 *                             if current_path_info[2] == 'directory':
 *                                 if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *                                     self.utf8_decode(current_path_info[0], 'surrogateescape')[0]):
 *                                     current_path_info = current_path_info[:2] + \
*/
              __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1643, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              if (__pyx_t_2) {

                /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1645
 *                                 if self.tree._directory_is_tree_reference(
 *                                     self.utf8_decode(current_path_info[0], 'surrogateescape')[0]):
 *                                     current_path_info = current_path_info[:2] + \             # <<<<<<<<<<<<<<
 *                                         ('tree-reference',) + current_path_info[3:]
 *                             new_executable = bool(
*/
                __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_current_path_info, 0, 2, NULL, NULL, &__pyx_mstate_global->__pyx_slice[1], 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1645, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_5);

                /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1646
 *                                     self.utf8_decode(current_path_info[0], 'surrogateescape')[0]):
 *                                     current_path_info = current_path_info[:2] + \
 *                                         ('tree-reference',) + current_path_info[3:]             # <<<<<<<<<<<<<<
 *                             new_executable = bool(
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
*/
                __pyx_t_12 = PyNumber_Add(__pyx_t_5, __pyx_mstate_global->__pyx_tuple[5]); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1645, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_12);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_current_path_info, 3, 0, NULL, NULL, &__pyx_mstate_global->__pyx_slice[2], 1, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1646, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_5);
                __pyx_t_3 = PyNumber_Add(__pyx_t_12, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1646, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF_SET(__pyx_v_current_path_info, __pyx_t_3);
                __pyx_t_3 = 0;

                /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1643
 *                         if self.want_unversioned:
 *                             if current_path_info[2] == 'directory':
 *                                 if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *                                     self.utf8_decode(current_path_info[0], 'surrogateescape')[0]):
 *                                     current_path_info = current_path_info[:2] + \
*/
              }

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1642
 *                         self.path_index = self.path_index + 1
 *                         if self.want_unversioned:
 *                             if current_path_info[2] == 'directory':             # <<<<<<<<<<<<<<
 *                                 if self.tree._directory_is_tree_reference(
 *                                     self.utf8_decode(current_path_info[0], 'surrogateescape')[0]):
*/
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1648
 *                                         ('tree-reference',) + current_path_info[3:]
 *                             new_executable = bool(
 *                                 stat.S_ISREG(current_path_info[3].st_mode)             # <<<<<<<<<<<<<<
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)
 *                             return InventoryTreeChange(
*/
            __pyx_t_5 = NULL;
            __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_stat); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1648, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_S_ISREG); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1648, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_18);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __pyx_t_12 = __Pyx_GetItemInt(__pyx_v_current_path_info, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1648, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_st_mode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1648, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __pyx_t_13 = 1;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_18))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_18);
              assert(__pyx_t_5);
              PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_18);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(__pyx__function);
              __Pyx_DECREF_SET(__pyx_t_18, __pyx__function);
              __pyx_t_13 = 0;
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
              __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_18, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1648, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
            }
            __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1648, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (__pyx_t_8) {
            } else {
              __pyx_t_2 = __pyx_t_8;
              goto __pyx_L104_bool_binop_done;
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1649
 *                             new_executable = bool(
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)             # <<<<<<<<<<<<<<
 *                             return InventoryTreeChange(
 *                                 None,
*/
            __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_stat); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1649, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_S_IEXEC); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1649, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_18);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_current_path_info, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1649, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_st_mode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1649, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = PyNumber_And(__pyx_t_18, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1649, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1649, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_2 = __pyx_t_8;
            __pyx_L104_bool_binop_done:;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1647
 *                                     current_path_info = current_path_info[:2] + \
 *                                         ('tree-reference',) + current_path_info[3:]
 *                             new_executable = bool(             # <<<<<<<<<<<<<<
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)
*/
            __pyx_v_new_executable = (!(!__pyx_t_2));

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1650
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)
 *                             return InventoryTreeChange(             # <<<<<<<<<<<<<<
 *                                 None,
 *                                 (None, self.utf8_decode(current_path_info[0], 'surrogateescape')[0]),
*/
            __Pyx_XDECREF(__pyx_r);
            __pyx_t_4 = NULL;
            __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_InventoryTreeChange); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1650, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_18);

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1652
 *                             return InventoryTreeChange(
 *                                 None,
 *                                 (None, self.utf8_decode(current_path_info[0], 'surrogateescape')[0]),             # <<<<<<<<<<<<<<
 *                                 True,
 *                                 (False, False),
*/
            __pyx_t_12 = NULL;
            __Pyx_INCREF(__pyx_v_self->utf8_decode);
            __pyx_t_1 = __pyx_v_self->utf8_decode; 
            __pyx_t_15 = __Pyx_GetItemInt(__pyx_v_current_path_info, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1652, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_13 = 1;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_1);
              assert(__pyx_t_12);
              PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(__pyx__function);
              __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
              __pyx_t_13 = 0;
            }
            #endif
            {
              PyObject *__pyx_callargs[3] = {__pyx_t_12, __pyx_t_15, __pyx_mstate_global->__pyx_n_u_surrogateescape};
              __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_13, (3-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1652, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
            }
            __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1652, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1652, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_INCREF(Py_None);
            __Pyx_GIVEREF(Py_None);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, Py_None) != (0)) __PYX_ERR(0, 1652, __pyx_L1_error);
            __Pyx_GIVEREF(__pyx_t_1);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1) != (0)) __PYX_ERR(0, 1652, __pyx_L1_error);
            __pyx_t_1 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1656
 *                                 (False, False),
 *                                 (None, None),
 *                                 (None, self.utf8_decode(current_path_info[1], 'surrogateescape')[0]),             # <<<<<<<<<<<<<<
 *                                 (None, current_path_info[2]),
 *                                 (None, new_executable))
*/
            __pyx_t_15 = NULL;
            __Pyx_INCREF(__pyx_v_self->utf8_decode);
            __pyx_t_12 = __pyx_v_self->utf8_decode; 
            __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_current_path_info, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1656, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            __pyx_t_13 = 1;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_12))) {
              __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_12);
              assert(__pyx_t_15);
              PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_15);
              __Pyx_INCREF(__pyx__function);
              __Pyx_DECREF_SET(__pyx_t_12, __pyx__function);
              __pyx_t_13 = 0;
            }
            #endif
            {
              PyObject *__pyx_callargs[3] = {__pyx_t_15, __pyx_t_17, __pyx_mstate_global->__pyx_n_u_surrogateescape};
              __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_12, __pyx_callargs+__pyx_t_13, (3-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1656, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
            }
            __pyx_t_12 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1656, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1656, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_INCREF(Py_None);
            __Pyx_GIVEREF(Py_None);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, Py_None) != (0)) __PYX_ERR(0, 1656, __pyx_L1_error);
            __Pyx_GIVEREF(__pyx_t_12);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_12) != (0)) __PYX_ERR(0, 1656, __pyx_L1_error);
            __pyx_t_12 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1657
 *                                 (None, None),
 *                                 (None, self.utf8_decode(current_path_info[1], 'surrogateescape')[0]),
 *                                 (None, current_path_info[2]),             # <<<<<<<<<<<<<<
 *                                 (None, new_executable))
 *                     # This dir info has been handled, go to the next
*/
            __pyx_t_12 = __Pyx_GetItemInt(__pyx_v_current_path_info, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1657, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1657, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_INCREF(Py_None);
            __Pyx_GIVEREF(Py_None);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_17, 0, Py_None) != (0)) __PYX_ERR(0, 1657, __pyx_L1_error);
            __Pyx_GIVEREF(__pyx_t_12);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_12) != (0)) __PYX_ERR(0, 1657, __pyx_L1_error);
            __pyx_t_12 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1658
 *                                 (None, self.utf8_decode(current_path_info[1], 'surrogateescape')[0]),
 *                                 (None, current_path_info[2]),
 *                                 (None, new_executable))             # <<<<<<<<<<<<<<
 *                     # This dir info has been handled, go to the next
 *                     self.path_index = 0
*/
            __pyx_t_12 = __Pyx_PyBool_FromLong(__pyx_v_new_executable); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1658, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1658, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_INCREF(Py_None);
            __Pyx_GIVEREF(Py_None);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 0, Py_None) != (0)) __PYX_ERR(0, 1658, __pyx_L1_error);
            __Pyx_GIVEREF(__pyx_t_12);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_12) != (0)) __PYX_ERR(0, 1658, __pyx_L1_error);
            __pyx_t_12 = 0;
            __pyx_t_13 = 1;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_18))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_18);
              assert(__pyx_t_4);
              PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_18);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(__pyx__function);
              __Pyx_DECREF_SET(__pyx_t_18, __pyx__function);
              __pyx_t_13 = 0;
            }
            #endif
            {
              PyObject *__pyx_callargs[9] = {__pyx_t_4, Py_None, __pyx_t_5, Py_True, __pyx_mstate_global->__pyx_tuple[6], __pyx_mstate_global->__pyx_tuple[0], __pyx_t_1, __pyx_t_17, __pyx_t_15};
              __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_18, __pyx_callargs+__pyx_t_13, (9-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1650, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
            }
            __pyx_r = __pyx_t_3;
            __pyx_t_3 = 0;
            goto __pyx_L0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1641
 *                             self.path_index = self.path_index - 1
 *                         self.path_index = self.path_index + 1
 *                         if self.want_unversioned:             # <<<<<<<<<<<<<<
 *                             if current_path_info[2] == 'directory':
 *                                 if self.tree._directory_is_tree_reference(
*/
          }
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1660
 *                                 (None, new_executable))
 *                     # This dir info has been handled, go to the next
 *                     self.path_index = 0             # <<<<<<<<<<<<<<
 *                     self.current_dir_list = None
 *                     try:
*/
        __pyx_v_self->path_index = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1661
 *                     # This dir info has been handled, go to the next
 *                     self.path_index = 0
 *                     self.current_dir_list = None             # <<<<<<<<<<<<<<
 *                     try:
 *                         self.current_dir_info = next(self.dir_iterator)
*/
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        __Pyx_GOTREF(__pyx_v_self->current_dir_list);
        __Pyx_DECREF(__pyx_v_self->current_dir_list);
        __pyx_v_self->current_dir_list = Py_None;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1662
 *                     self.path_index = 0
 *                     self.current_dir_list = None
 *                     try:             # <<<<<<<<<<<<<<
 *                         self.current_dir_info = next(self.dir_iterator)
 *                         self.current_dir_list = self.current_dir_info[1]
*/
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_9);
          /*try:*/ {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1663
 *                     self.current_dir_list = None
 *                     try:
 *                         self.current_dir_info = next(self.dir_iterator)             # <<<<<<<<<<<<<<
 *                         self.current_dir_list = self.current_dir_info[1]
 *                     except StopIteration, _:
*/
            __pyx_t_3 = __pyx_v_self->dir_iterator;
            __Pyx_INCREF(__pyx_t_3);
            __pyx_t_18 = __Pyx_PyIter_Next(__pyx_t_3); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1663, __pyx_L106_error)
            __Pyx_GOTREF(__pyx_t_18);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GIVEREF(__pyx_t_18);
            __Pyx_GOTREF(__pyx_v_self->current_dir_info);
            __Pyx_DECREF(__pyx_v_self->current_dir_info);
            __pyx_v_self->current_dir_info = __pyx_t_18;
            __pyx_t_18 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1664
 *                     try:
 *                         self.current_dir_info = next(self.dir_iterator)
 *                         self.current_dir_list = self.current_dir_info[1]             # <<<<<<<<<<<<<<
 *                     except StopIteration, _:
 *                         self.current_dir_info = None
*/
            __pyx_t_18 = __Pyx_GetItemInt(__pyx_v_self->current_dir_info, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1664, __pyx_L106_error)
            __Pyx_GOTREF(__pyx_t_18);
            __Pyx_GIVEREF(__pyx_t_18);
            __Pyx_GOTREF(__pyx_v_self->current_dir_list);
            __Pyx_DECREF(__pyx_v_self->current_dir_list);
            __pyx_v_self->current_dir_list = __pyx_t_18;
            __pyx_t_18 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1662
 *                     self.path_index = 0
 *                     self.current_dir_list = None
 *                     try:             # <<<<<<<<<<<<<<
 *                         self.current_dir_info = next(self.dir_iterator)
 *                         self.current_dir_list = self.current_dir_info[1]
*/
          }
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          goto __pyx_L113_try_end;
          __pyx_L106_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1665
 *                         self.current_dir_info = next(self.dir_iterator)
 *                         self.current_dir_list = self.current_dir_info[1]
 *                     except StopIteration, _:             # <<<<<<<<<<<<<<
 *                         self.current_dir_info = None
 *                 else: #(dircmp > 0)
*/
          __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_StopIteration))));
          if (__pyx_t_7) {
            __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._iter_next", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_18, &__pyx_t_3, &__pyx_t_15) < 0) __PYX_ERR(0, 1665, __pyx_L108_except_error)
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_3);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_3);

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1666
 *                         self.current_dir_list = self.current_dir_info[1]
 *                     except StopIteration, _:
 *                         self.current_dir_info = None             # <<<<<<<<<<<<<<
 *                 else: #(dircmp > 0)
 *                     # We have a dirblock entry for this location, but there
*/
            __Pyx_INCREF(Py_None);
            __Pyx_GIVEREF(Py_None);
            __Pyx_GOTREF(__pyx_v_self->current_dir_info);
            __Pyx_DECREF(__pyx_v_self->current_dir_info);
            __pyx_v_self->current_dir_info = Py_None;
            __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            goto __pyx_L107_exception_handled;
          }
          goto __pyx_L108_except_error;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1662
 *                     self.path_index = 0
 *                     self.current_dir_list = None
 *                     try:             # <<<<<<<<<<<<<<
 *                         self.current_dir_info = next(self.dir_iterator)
 *                         self.current_dir_list = self.current_dir_info[1]
*/
          __pyx_L108_except_error:;
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_9);
          goto __pyx_L1_error;
          __pyx_L107_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_9);
          __pyx_L113_try_end:;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1618
 *                 # evaluated once per path: we could cache the result before
 *                 # doing the while loop and probably save time.
 *                 if _cmp_by_dirs(current_dirname_c,             # <<<<<<<<<<<<<<
 *                     PyBytes_Size(current_dirname),
 *                     current_blockname_c,
*/
        goto __pyx_L95;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1675
 *                     # need to handle all of the files that are contained
 *                     # within.
 *                     while self.current_block_pos < len(self.current_block_list):             # <<<<<<<<<<<<<<
 *                         current_entry = self.current_block_list[self.current_block_pos]
 *                         self.current_block_pos = self.current_block_pos + 1
*/
      /*else*/ {
        while (1) {
          __pyx_t_15 = __pyx_v_self->current_block_list;
          __Pyx_INCREF(__pyx_t_15);
          __pyx_t_14 = PyObject_Length(__pyx_t_15); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1675, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_t_2 = (__pyx_v_self->current_block_pos < __pyx_t_14);
          if (!__pyx_t_2) break;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1676
 *                     # within.
 *                     while self.current_block_pos < len(self.current_block_list):
 *                         current_entry = self.current_block_list[self.current_block_pos]             # <<<<<<<<<<<<<<
 *                         self.current_block_pos = self.current_block_pos + 1
 *                         # entry referring to file not present on disk.
*/
          __pyx_t_15 = __Pyx_GetItemInt(__pyx_v_self->current_block_list, __pyx_v_self->current_block_pos, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1676, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_XDECREF_SET(__pyx_v_current_entry, __pyx_t_15);
          __pyx_t_15 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1677
 *                     while self.current_block_pos < len(self.current_block_list):
 *                         current_entry = self.current_block_list[self.current_block_pos]
 *                         self.current_block_pos = self.current_block_pos + 1             # <<<<<<<<<<<<<<
 *                         # entry referring to file not present on disk.
 *                         # advance the entry only, after processing.
*/
          __pyx_v_self->current_block_pos = (__pyx_v_self->current_block_pos + 1);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1680
 *                         # entry referring to file not present on disk.
 *                         # advance the entry only, after processing.
 *                         result, changed = self._process_entry(current_entry, None)             # <<<<<<<<<<<<<<
 *                         if changed is not None:
 *                             if changed:
*/
          __pyx_t_15 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_process_entry(__pyx_v_self, __pyx_v_current_entry, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1680, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          if ((likely(PyTuple_CheckExact(__pyx_t_15))) || (PyList_CheckExact(__pyx_t_15))) {
            PyObject* sequence = __pyx_t_15;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 1680, __pyx_L1_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0);
              __Pyx_INCREF(__pyx_t_3);
              __pyx_t_18 = PyTuple_GET_ITEM(sequence, 1);
              __Pyx_INCREF(__pyx_t_18);
            } else {
              __pyx_t_3 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
              if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1680, __pyx_L1_error)
              __Pyx_XGOTREF(__pyx_t_3);
              __pyx_t_18 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
              if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1680, __pyx_L1_error)
              __Pyx_XGOTREF(__pyx_t_18);
            }
            #else
            __pyx_t_3 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1680, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_18 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1680, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_18);
            #endif
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_17 = PyObject_GetIter(__pyx_t_15); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1680, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __pyx_t_6 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_17);
            index = 0; __pyx_t_3 = __pyx_t_6(__pyx_t_17); if (unlikely(!__pyx_t_3)) goto __pyx_L118_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_3);
            index = 1; __pyx_t_18 = __pyx_t_6(__pyx_t_17); if (unlikely(!__pyx_t_18)) goto __pyx_L118_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_18);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_17), 2) < (0)) __PYX_ERR(0, 1680, __pyx_L1_error)
            __pyx_t_6 = NULL;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            goto __pyx_L119_unpacking_done;
            __pyx_L118_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __pyx_t_6 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 1680, __pyx_L1_error)
            __pyx_L119_unpacking_done:;
          }
          __Pyx_XDECREF_SET(__pyx_v_result, __pyx_t_3);
          __pyx_t_3 = 0;
          __Pyx_XDECREF_SET(__pyx_v_changed, __pyx_t_18);
          __pyx_t_18 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1681
 *                         # advance the entry only, after processing.
 *                         result, changed = self._process_entry(current_entry, None)
 *                         if changed is not None:             # <<<<<<<<<<<<<<
 *                             if changed:
 *                                 self._gather_result_for_consistency(result)
*/
          __pyx_t_2 = (__pyx_v_changed != Py_None);
          if (__pyx_t_2) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1682
 *                         result, changed = self._process_entry(current_entry, None)
 *                         if changed is not None:
 *                             if changed:             # <<<<<<<<<<<<<<
 *                                 self._gather_result_for_consistency(result)
 *                             if changed or self.include_unchanged:
*/
            __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1682, __pyx_L1_error)
            if (__pyx_t_2) {

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1683
 *                         if changed is not None:
 *                             if changed:
 *                                 self._gather_result_for_consistency(result)             # <<<<<<<<<<<<<<
 *                             if changed or self.include_unchanged:
 *                                 return result
*/
              __pyx_t_7 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_gather_result_for_consistency(__pyx_v_self, __pyx_v_result); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1683, __pyx_L1_error)

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1682
 *                         result, changed = self._process_entry(current_entry, None)
 *                         if changed is not None:
 *                             if changed:             # <<<<<<<<<<<<<<
 *                                 self._gather_result_for_consistency(result)
 *                             if changed or self.include_unchanged:
*/
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1684
 *                             if changed:
 *                                 self._gather_result_for_consistency(result)
 *                             if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                                 return result
 *                     self.block_index = self.block_index + 1
*/
            __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1684, __pyx_L1_error)
            if (!__pyx_t_8) {
            } else {
              __pyx_t_2 = __pyx_t_8;
              goto __pyx_L123_bool_binop_done;
            }
            __pyx_t_8 = (__pyx_v_self->include_unchanged != 0);
            __pyx_t_2 = __pyx_t_8;
            __pyx_L123_bool_binop_done:;
            if (__pyx_t_2) {

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1685
 *                                 self._gather_result_for_consistency(result)
 *                             if changed or self.include_unchanged:
 *                                 return result             # <<<<<<<<<<<<<<
 *                     self.block_index = self.block_index + 1
 *                     self._update_current_block()
*/
              __Pyx_XDECREF(__pyx_r);
              __Pyx_INCREF(__pyx_v_result);
              __pyx_r = __pyx_v_result;
              goto __pyx_L0;

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1684
 *                             if changed:
 *                                 self._gather_result_for_consistency(result)
 *                             if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                                 return result
 *                     self.block_index = self.block_index + 1
*/
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1681
 *                         # advance the entry only, after processing.
 *                         result, changed = self._process_entry(current_entry, None)
 *                         if changed is not None:             # <<<<<<<<<<<<<<
 *                             if changed:
 *                                 self._gather_result_for_consistency(result)
*/
          }
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1686
 *                             if changed or self.include_unchanged:
 *                                 return result
 *                     self.block_index = self.block_index + 1             # <<<<<<<<<<<<<<
 *                     self._update_current_block()
 *                 continue # next loop-on-block/dir
*/
        __pyx_v_self->block_index = (__pyx_v_self->block_index + 1);

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1687
 *                                 return result
 *                     self.block_index = self.block_index + 1
 *                     self._update_current_block()             # <<<<<<<<<<<<<<
 *                 continue # next loop-on-block/dir
 *             result = self._loop_one_block()
*/
        __pyx_t_7 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_update_current_block(__pyx_v_self); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1687, __pyx_L1_error)
      }
      __pyx_L95:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1688
 *                     self.block_index = self.block_index + 1
 *                     self._update_current_block()
 *                 continue # next loop-on-block/dir             # <<<<<<<<<<<<<<
 *             result = self._loop_one_block()
 *             if result is not None:
*/
      goto __pyx_L87_continue;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1604
 *             or self.current_block is not None):
 *             # Uncommon case - a missing directory or an unversioned directory:
 *             if (self.current_dir_info and self.current_block             # <<<<<<<<<<<<<<
 *                 and self.current_dir_info[0][0] != self.current_block[0]):
 *                 # Work around pyrex broken heuristic - current_dirname has
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1689
 *                     self._update_current_block()
 *                 continue # next loop-on-block/dir
 *             result = self._loop_one_block()             # <<<<<<<<<<<<<<
 *             if result is not None:
 *                 return result
*/
    __pyx_t_15 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_loop_one_block(__pyx_v_self); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1689, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_XDECREF_SET(__pyx_v_result, __pyx_t_15);
    __pyx_t_15 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1690
 *                 continue # next loop-on-block/dir
 *             result = self._loop_one_block()
 *             if result is not None:             # <<<<<<<<<<<<<<
 *                 return result
 *         if len(self.search_specific_files):
*/
    __pyx_t_2 = (__pyx_v_result != Py_None);
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1691
 *             result = self._loop_one_block()
 *             if result is not None:
 *                 return result             # <<<<<<<<<<<<<<
 *         if len(self.search_specific_files):
 *             # More supplied paths to process
*/
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_result);
      __pyx_r = __pyx_v_result;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1690
 *                 continue # next loop-on-block/dir
 *             result = self._loop_one_block()
 *             if result is not None:             # <<<<<<<<<<<<<<
 *                 return result
 *         if len(self.search_specific_files):
*/
    }
    __pyx_L87_continue:;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1692
 *             if result is not None:
 *                 return result
 *         if len(self.search_specific_files):             # <<<<<<<<<<<<<<
 *             # More supplied paths to process
 *             self.current_root = None
*/
  __pyx_t_15 = __pyx_v_self->search_specific_files;
  __Pyx_INCREF(__pyx_t_15);
  __pyx_t_14 = PyObject_Length(__pyx_t_15); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1692, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_2 = (__pyx_t_14 != 0);
  if (__pyx_t_2) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1694
 *         if len(self.search_specific_files):
 *             # More supplied paths to process
 *             self.current_root = None             # <<<<<<<<<<<<<<
 *             return self._iter_next()
 *         # Start expanding more conservatively, adding paths the user may not
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->current_root);
    __Pyx_DECREF(__pyx_v_self->current_root);
    __pyx_v_self->current_root = Py_None;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1695
 *             # More supplied paths to process
 *             self.current_root = None
 *             return self._iter_next()             # <<<<<<<<<<<<<<
 *         # Start expanding more conservatively, adding paths the user may not
 *         # have intended but required for consistent deltas.
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_15 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_iter_next(__pyx_v_self); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1695, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_r = __pyx_t_15;
    __pyx_t_15 = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1692
 *             if result is not None:
 *                 return result
 *         if len(self.search_specific_files):             # <<<<<<<<<<<<<<
 *             # More supplied paths to process
 *             self.current_root = None
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1698
 *         # Start expanding more conservatively, adding paths the user may not
 *         # have intended but required for consistent deltas.
 *         self.doing_consistency_expansion = 1             # <<<<<<<<<<<<<<
 *         if not self._pending_consistent_entries:
 *             self._pending_consistent_entries = self._next_consistent_entries()
*/
  __pyx_v_self->doing_consistency_expansion = 1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1699
 *         # have intended but required for consistent deltas.
 *         self.doing_consistency_expansion = 1
 *         if not self._pending_consistent_entries:             # <<<<<<<<<<<<<<
 *             self._pending_consistent_entries = self._next_consistent_entries()
 *         while self._pending_consistent_entries:
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_self->_pending_consistent_entries); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1699, __pyx_L1_error)
  __pyx_t_8 = (!__pyx_t_2);
  if (__pyx_t_8) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1700
 *         self.doing_consistency_expansion = 1
 *         if not self._pending_consistent_entries:
 *             self._pending_consistent_entries = self._next_consistent_entries()             # <<<<<<<<<<<<<<
 *         while self._pending_consistent_entries:
 *             result, changed = self._pending_consistent_entries.pop()
*/
    __pyx_t_15 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_next_consistent_entries(__pyx_v_self); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_15);
    __Pyx_GOTREF(__pyx_v_self->_pending_consistent_entries);
    __Pyx_DECREF(__pyx_v_self->_pending_consistent_entries);
    __pyx_v_self->_pending_consistent_entries = __pyx_t_15;
    __pyx_t_15 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1699
 *         # have intended but required for consistent deltas.
 *         self.doing_consistency_expansion = 1
 *         if not self._pending_consistent_entries:             # <<<<<<<<<<<<<<
 *             self._pending_consistent_entries = self._next_consistent_entries()
 *         while self._pending_consistent_entries:
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1701
 *         if not self._pending_consistent_entries:
 *             self._pending_consistent_entries = self._next_consistent_entries()
 *         while self._pending_consistent_entries:             # <<<<<<<<<<<<<<
 *             result, changed = self._pending_consistent_entries.pop()
 *             if changed is not None:
*/
  while (1) {
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_self->_pending_consistent_entries); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1701, __pyx_L1_error)
    if (!__pyx_t_8) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1702
 *             self._pending_consistent_entries = self._next_consistent_entries()
 *         while self._pending_consistent_entries:
 *             result, changed = self._pending_consistent_entries.pop()             # <<<<<<<<<<<<<<
 *             if changed is not None:
 *                 return result
*/
    __pyx_t_15 = __Pyx_PyObject_Pop(__pyx_v_self->_pending_consistent_entries); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1702, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    if ((likely(PyTuple_CheckExact(__pyx_t_15))) || (PyList_CheckExact(__pyx_t_15))) {
      PyObject* sequence = __pyx_t_15;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1702, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_18 = PyTuple_GET_ITEM(sequence, 0);
        __Pyx_INCREF(__pyx_t_18);
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1);
        __Pyx_INCREF(__pyx_t_3);
      } else {
        __pyx_t_18 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
        if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1702, __pyx_L1_error)
        __Pyx_XGOTREF(__pyx_t_18);
        __pyx_t_3 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1702, __pyx_L1_error)
        __Pyx_XGOTREF(__pyx_t_3);
      }
      #else
      __pyx_t_18 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1702, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __pyx_t_3 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1702, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      #endif
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_17 = PyObject_GetIter(__pyx_t_15); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1702, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_t_6 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_17);
      index = 0; __pyx_t_18 = __pyx_t_6(__pyx_t_17); if (unlikely(!__pyx_t_18)) goto __pyx_L130_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_18);
      index = 1; __pyx_t_3 = __pyx_t_6(__pyx_t_17); if (unlikely(!__pyx_t_3)) goto __pyx_L130_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_17), 2) < (0)) __PYX_ERR(0, 1702, __pyx_L1_error)
      __pyx_t_6 = NULL;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      goto __pyx_L131_unpacking_done;
      __pyx_L130_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_t_6 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1702, __pyx_L1_error)
      __pyx_L131_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_result, __pyx_t_18);
    __pyx_t_18 = 0;
    __Pyx_XDECREF_SET(__pyx_v_changed, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1703
 *         while self._pending_consistent_entries:
 *             result, changed = self._pending_consistent_entries.pop()
 *             if changed is not None:             # <<<<<<<<<<<<<<
 *                 return result
 *         raise StopIteration()
*/
    __pyx_t_8 = (__pyx_v_changed != Py_None);
    if (__pyx_t_8) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1704
 *             result, changed = self._pending_consistent_entries.pop()
 *             if changed is not None:
 *                 return result             # <<<<<<<<<<<<<<
 *         raise StopIteration()
 * 
*/
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_result);
      __pyx_r = __pyx_v_result;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1703
 *         while self._pending_consistent_entries:
 *             result, changed = self._pending_consistent_entries.pop()
 *             if changed is not None:             # <<<<<<<<<<<<<<
 *                 return result
 *         raise StopIteration()
*/
    }
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1705
 *             if changed is not None:
 *                 return result
 *         raise StopIteration()             # <<<<<<<<<<<<<<
 * 
 *     cdef object _maybe_tree_ref(self, current_path_info):
*/
  __pyx_t_3 = NULL;
  __pyx_t_13 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_15 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_StopIteration)), __pyx_callargs+__pyx_t_13, (1-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1705, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
  }
  __Pyx_Raise(__pyx_t_15, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __PYX_ERR(0, 1705, __pyx_L1_error)

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1415
 *         return self._iter_next()
 * 
 *     cdef _iter_next(self):             # <<<<<<<<<<<<<<
 *         """Iterate over the changes."""
 *         # This function single steps through an iterator. As such while loops
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._iter_next", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_current_dirname);
  __Pyx_XDECREF(__pyx_v_current_blockname);
  __Pyx_XDECREF(__pyx_v_searched_specific_files);
  __Pyx_XDECREF(__pyx_v_entry);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_changed);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_root_stat);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_e_winerror);
  __Pyx_XDECREF(__pyx_v_win_errors);
  __Pyx_XDECREF(__pyx_v_bzr_index);
  __Pyx_XDECREF(__pyx_v_initial_key);
  __Pyx_XDECREF(__pyx_v_current_path_info);
  __Pyx_XDECREF(__pyx_v_current_entry);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1707
 *         raise StopIteration()
 * 
 *     cdef object _maybe_tree_ref(self, current_path_info):             # <<<<<<<<<<<<<<
 *         if self.tree._directory_is_tree_reference(
 *             self.utf8_decode(current_path_info[0], 'surrogateescape')[0]):
*/

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__maybe_tree_ref(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_current_path_info) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_maybe_tree_ref", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1708
 * 
 *     cdef object _maybe_tree_ref(self, current_path_info):
 *         if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *             self.utf8_decode(current_path_info[0], 'surrogateescape')[0]):
 *             return current_path_info[:2] + \
*/
  __pyx_t_2 = __pyx_v_self->tree;
  __Pyx_INCREF(__pyx_t_2);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1709
 *     cdef object _maybe_tree_ref(self, current_path_info):
 *         if self.tree._directory_is_tree_reference(
 *             self.utf8_decode(current_path_info[0], 'surrogateescape')[0]):             # <<<<<<<<<<<<<<
 *             return current_path_info[:2] + \
 *                 ('tree-reference',) + current_path_info[3:]
*/
  __pyx_t_4 = NULL;
  __Pyx_INCREF(__pyx_v_self->utf8_decode);
  __pyx_t_5 = __pyx_v_self->utf8_decode; 
  __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_current_path_info, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = 1;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    assert(__pyx_t_4);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
    __pyx_t_7 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_6, __pyx_mstate_global->__pyx_n_u_surrogateescape};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_7, (3-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1709, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_directory_is_tree_reference, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1708, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1708
 * 
 *     cdef object _maybe_tree_ref(self, current_path_info):
 *         if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *             self.utf8_decode(current_path_info[0], 'surrogateescape')[0]):
 *             return current_path_info[:2] + \
*/
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1708, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_8) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1710
 *         if self.tree._directory_is_tree_reference(
 *             self.utf8_decode(current_path_info[0], 'surrogateescape')[0]):
 *             return current_path_info[:2] + \             # <<<<<<<<<<<<<<
 *                 ('tree-reference',) + current_path_info[3:]
 *         else:
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_current_path_info, 0, 2, NULL, NULL, &__pyx_mstate_global->__pyx_slice[1], 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1710, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1711
 *             self.utf8_decode(current_path_info[0], 'surrogateescape')[0]):
 *             return current_path_info[:2] + \
 *                 ('tree-reference',) + current_path_info[3:]             # <<<<<<<<<<<<<<
 *         else:
 *             return current_path_info
*/
    __pyx_t_5 = PyNumber_Add(__pyx_t_1, __pyx_mstate_global->__pyx_tuple[5]); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1710, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_current_path_info, 3, 0, NULL, NULL, &__pyx_mstate_global->__pyx_slice[2], 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1711, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyNumber_Add(__pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1711, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1708
 * 
 *     cdef object _maybe_tree_ref(self, current_path_info):
 *         if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *             self.utf8_decode(current_path_info[0], 'surrogateescape')[0]):
 *             return current_path_info[:2] + \
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1713
 *                 ('tree-reference',) + current_path_info[3:]
 *         else:
 *             return current_path_info             # <<<<<<<<<<<<<<
 * 
 *     cdef object _loop_one_block(self):
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_current_path_info);
    __pyx_r = __pyx_v_current_path_info;
    goto __pyx_L0;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1707
 *         raise StopIteration()
 * 
 *     cdef object _maybe_tree_ref(self, current_path_info):             # <<<<<<<<<<<<<<
 *         if self.tree._directory_is_tree_reference(
 *             self.utf8_decode(current_path_info[0], 'surrogateescape')[0]):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._maybe_tree_ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1715
 *             return current_path_info
 * 
 *     cdef object _loop_one_block(self):             # <<<<<<<<<<<<<<
 *             # current_dir_info and current_block refer to the same directory -
 *             # this is the common case code.
*/

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__loop_one_block(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  PyObject *__pyx_v_current_entry = 0;
  PyObject *__pyx_v_current_path_info = 0;
  int __pyx_v_path_handled;
  char __pyx_v_minikind;
  int __pyx_v_cmp_result;
  PyObject *__pyx_v_disk_kind = NULL;
  long __pyx_v_advance_entry;
  long __pyx_v_advance_path;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_changed = NULL;
  int __pyx_v_new_executable;
  PyObject *__pyx_v_relpath_unicode = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  char __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  size_t __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  Py_ssize_t __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_loop_one_block", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1728
 *             # PyBytes_AsStringAndSize(disk_kind, &temp_str, &temp_str_length)
 *             # if not strncmp(temp_str, "directory", temp_str_length):
 *             if (self.current_block is not None and             # <<<<<<<<<<<<<<
 *                 self.current_block_pos < PyList_GET_SIZE(self.current_block_list)):
 *                 current_entry = PyList_GET_ITEM(self.current_block_list,
*/
  __pyx_t_2 = (__pyx_v_self->current_block != Py_None);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1729
 *             # if not strncmp(temp_str, "directory", temp_str_length):
 *             if (self.current_block is not None and
 *                 self.current_block_pos < PyList_GET_SIZE(self.current_block_list)):             # <<<<<<<<<<<<<<
 *                 current_entry = PyList_GET_ITEM(self.current_block_list,
 *                     self.current_block_pos)
*/
  __pyx_t_3 = __pyx_v_self->current_block_list;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_2 = (__pyx_v_self->current_block_pos < PyList_GET_SIZE(__pyx_t_3));
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1728
 *             # PyBytes_AsStringAndSize(disk_kind, &temp_str, &temp_str_length)
 *             # if not strncmp(temp_str, "directory", temp_str_length):
 *             if (self.current_block is not None and             # <<<<<<<<<<<<<<
 *                 self.current_block_pos < PyList_GET_SIZE(self.current_block_list)):
 *                 current_entry = PyList_GET_ITEM(self.current_block_list,
*/
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1730
 *             if (self.current_block is not None and
 *                 self.current_block_pos < PyList_GET_SIZE(self.current_block_list)):
 *                 current_entry = PyList_GET_ITEM(self.current_block_list,             # <<<<<<<<<<<<<<
 *                     self.current_block_pos)
 *                 # accomodate pyrex
*/
    __pyx_t_3 = __pyx_v_self->current_block_list;
    __Pyx_INCREF(__pyx_t_3);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1731
 *                 self.current_block_pos < PyList_GET_SIZE(self.current_block_list)):
 *                 current_entry = PyList_GET_ITEM(self.current_block_list,
 *                     self.current_block_pos)             # <<<<<<<<<<<<<<
 *                 # accomodate pyrex
 *                 Py_INCREF(current_entry)
*/
    __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_v_self->current_block_pos); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_current_entry = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1733
 *                     self.current_block_pos)
 *                 # accomodate pyrex
 *                 Py_INCREF(current_entry)             # <<<<<<<<<<<<<<
 *             else:
 *                 current_entry = None
*/
    Py_INCREF(__pyx_v_current_entry);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1728
 *             # PyBytes_AsStringAndSize(disk_kind, &temp_str, &temp_str_length)
 *             # if not strncmp(temp_str, "directory", temp_str_length):
 *             if (self.current_block is not None and             # <<<<<<<<<<<<<<
 *                 self.current_block_pos < PyList_GET_SIZE(self.current_block_list)):
 *                 current_entry = PyList_GET_ITEM(self.current_block_list,
*/
    goto __pyx_L3;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1735
 *                 Py_INCREF(current_entry)
 *             else:
 *                 current_entry = None             # <<<<<<<<<<<<<<
 *             if (self.current_dir_info is not None and
 *                 self.path_index < PyList_GET_SIZE(self.current_dir_list)):
*/
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_current_entry = Py_None;
  }
  __pyx_L3:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1736
 *             else:
 *                 current_entry = None
 *             if (self.current_dir_info is not None and             # <<<<<<<<<<<<<<
 *                 self.path_index < PyList_GET_SIZE(self.current_dir_list)):
 *                 current_path_info = PyList_GET_ITEM(self.current_dir_list,
*/
  __pyx_t_2 = (__pyx_v_self->current_dir_info != Py_None);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1737
 *                 current_entry = None
 *             if (self.current_dir_info is not None and
 *                 self.path_index < PyList_GET_SIZE(self.current_dir_list)):             # <<<<<<<<<<<<<<
 *                 current_path_info = PyList_GET_ITEM(self.current_dir_list,
 *                     self.path_index)
*/
  __pyx_t_4 = __pyx_v_self->current_dir_list;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_2 = (__pyx_v_self->path_index < PyList_GET_SIZE(__pyx_t_4));
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1736
 *             else:
 *                 current_entry = None
 *             if (self.current_dir_info is not None and             # <<<<<<<<<<<<<<
 *                 self.path_index < PyList_GET_SIZE(self.current_dir_list)):
 *                 current_path_info = PyList_GET_ITEM(self.current_dir_list,
*/
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1738
 *             if (self.current_dir_info is not None and
 *                 self.path_index < PyList_GET_SIZE(self.current_dir_list)):
 *                 current_path_info = PyList_GET_ITEM(self.current_dir_list,             # <<<<<<<<<<<<<<
 *                     self.path_index)
 *                 # accomodate pyrex
*/
    __pyx_t_4 = __pyx_v_self->current_dir_list;
    __Pyx_INCREF(__pyx_t_4);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1739
 *                 self.path_index < PyList_GET_SIZE(self.current_dir_list)):
 *                 current_path_info = PyList_GET_ITEM(self.current_dir_list,
 *                     self.path_index)             # <<<<<<<<<<<<<<
 *                 # accomodate pyrex
 *                 Py_INCREF(current_path_info)
*/
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_v_self->path_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_current_path_info = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1741
 *                     self.path_index)
 *                 # accomodate pyrex
 *                 Py_INCREF(current_path_info)             # <<<<<<<<<<<<<<
 *                 disk_kind = PyTuple_GET_ITEM(current_path_info, 2)
 *                 # accomodate pyrex
*/
    Py_INCREF(__pyx_v_current_path_info);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1742
 *                 # accomodate pyrex
 *                 Py_INCREF(current_path_info)
 *                 disk_kind = PyTuple_GET_ITEM(current_path_info, 2)             # <<<<<<<<<<<<<<
 *                 # accomodate pyrex
 *                 Py_INCREF(disk_kind)
*/
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_v_current_path_info, 2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1742, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_disk_kind = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1744
 *                 disk_kind = PyTuple_GET_ITEM(current_path_info, 2)
 *                 # accomodate pyrex
 *                 Py_INCREF(disk_kind)             # <<<<<<<<<<<<<<
 *                 if disk_kind == "directory":
 *                     current_path_info = self._maybe_tree_ref(current_path_info)
*/
    Py_INCREF(__pyx_v_disk_kind);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1745
 *                 # accomodate pyrex
 *                 Py_INCREF(disk_kind)
 *                 if disk_kind == "directory":             # <<<<<<<<<<<<<<
 *                     current_path_info = self._maybe_tree_ref(current_path_info)
 *             else:
*/
    __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_disk_kind, __pyx_mstate_global->__pyx_n_u_directory, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1745, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1746
 *                 Py_INCREF(disk_kind)
 *                 if disk_kind == "directory":
 *                     current_path_info = self._maybe_tree_ref(current_path_info)             # <<<<<<<<<<<<<<
 *             else:
 *                 current_path_info = None
*/
      __pyx_t_3 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_maybe_tree_ref(__pyx_v_self, __pyx_v_current_path_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1746, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF_SET(__pyx_v_current_path_info, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1745
 *                 # accomodate pyrex
 *                 Py_INCREF(disk_kind)
 *                 if disk_kind == "directory":             # <<<<<<<<<<<<<<
 *                     current_path_info = self._maybe_tree_ref(current_path_info)
 *             else:
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1736
 *             else:
 *                 current_entry = None
 *             if (self.current_dir_info is not None and             # <<<<<<<<<<<<<<
 *                 self.path_index < PyList_GET_SIZE(self.current_dir_list)):
 *                 current_path_info = PyList_GET_ITEM(self.current_dir_list,
*/
    goto __pyx_L6;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1748
 *                     current_path_info = self._maybe_tree_ref(current_path_info)
 *             else:
 *                 current_path_info = None             # <<<<<<<<<<<<<<
 *             while (current_entry is not None or current_path_info is not None):
 *                 advance_entry = -1
*/
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_current_path_info = Py_None;
  }
  __pyx_L6:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1749
 *             else:
 *                 current_path_info = None
 *             while (current_entry is not None or current_path_info is not None):             # <<<<<<<<<<<<<<
 *                 advance_entry = -1
 *                 advance_path = -1
*/
  while (1) {
    __pyx_t_2 = (__pyx_v_current_entry != Py_None);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L12_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_current_path_info != Py_None);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L12_bool_binop_done:;
    if (!__pyx_t_1) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1750
 *                 current_path_info = None
 *             while (current_entry is not None or current_path_info is not None):
 *                 advance_entry = -1             # <<<<<<<<<<<<<<
 *                 advance_path = -1
 *                 result = None
*/
    __pyx_v_advance_entry = -1L;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1751
 *             while (current_entry is not None or current_path_info is not None):
 *                 advance_entry = -1
 *                 advance_path = -1             # <<<<<<<<<<<<<<
 *                 result = None
 *                 changed = None
*/
    __pyx_v_advance_path = -1L;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1752
 *                 advance_entry = -1
 *                 advance_path = -1
 *                 result = None             # <<<<<<<<<<<<<<
 *                 changed = None
 *                 path_handled = 0
*/
    __Pyx_INCREF(Py_None);
    __Pyx_XDECREF_SET(__pyx_v_result, Py_None);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1753
 *                 advance_path = -1
 *                 result = None
 *                 changed = None             # <<<<<<<<<<<<<<
 *                 path_handled = 0
 *                 if current_entry is None:
*/
    __Pyx_INCREF(Py_None);
    __Pyx_XDECREF_SET(__pyx_v_changed, Py_None);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1754
 *                 result = None
 *                 changed = None
 *                 path_handled = 0             # <<<<<<<<<<<<<<
 *                 if current_entry is None:
 *                     # unversioned -  the check for path_handled when the path
*/
    __pyx_v_path_handled = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1755
 *                 changed = None
 *                 path_handled = 0
 *                 if current_entry is None:             # <<<<<<<<<<<<<<
 *                     # unversioned -  the check for path_handled when the path
 *                     # is advanced will yield this path if needed.
*/
    __pyx_t_1 = (__pyx_v_current_entry == Py_None);
    if (__pyx_t_1) {
      goto __pyx_L14;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1759
 *                     # is advanced will yield this path if needed.
 *                     pass
 *                 elif current_path_info is None:             # <<<<<<<<<<<<<<
 *                     # no path is fine: the per entry code will handle it.
 *                     result, changed = self._process_entry(current_entry,
*/
    __pyx_t_1 = (__pyx_v_current_path_info == Py_None);
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1761
 *                 elif current_path_info is None:
 *                     # no path is fine: the per entry code will handle it.
 *                     result, changed = self._process_entry(current_entry,             # <<<<<<<<<<<<<<
 *                         current_path_info)
 *                 else:
*/
      __pyx_t_3 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_process_entry(__pyx_v_self, __pyx_v_current_entry, __pyx_v_current_path_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1761, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
        PyObject* sequence = __pyx_t_3;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1761, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0);
          __Pyx_INCREF(__pyx_t_4);
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1);
          __Pyx_INCREF(__pyx_t_5);
        } else {
          __pyx_t_4 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1761, __pyx_L1_error)
          __Pyx_XGOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1761, __pyx_L1_error)
          __Pyx_XGOTREF(__pyx_t_5);
        }
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1761, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1761, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_6 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1761, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_7 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_6);
        index = 0; __pyx_t_4 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_4)) goto __pyx_L15_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        index = 1; __pyx_t_5 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_5)) goto __pyx_L15_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 2) < (0)) __PYX_ERR(0, 1761, __pyx_L1_error)
        __pyx_t_7 = NULL;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L16_unpacking_done;
        __pyx_L15_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_7 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1761, __pyx_L1_error)
        __pyx_L16_unpacking_done:;
      }
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_changed, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1759
 *                     # is advanced will yield this path if needed.
 *                     pass
 *                 elif current_path_info is None:             # <<<<<<<<<<<<<<
 *                     # no path is fine: the per entry code will handle it.
 *                     result, changed = self._process_entry(current_entry,
*/
      goto __pyx_L14;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1764
 *                         current_path_info)
 *                 else:
 *                     minikind = _minikind_from_string(             # <<<<<<<<<<<<<<
 *                         current_entry[1][self.target_index][0])
 *                     cmp_result = ((current_path_info[1] > current_entry[0][1]) -
*/
    /*else*/ {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1765
 *                 else:
 *                     minikind = _minikind_from_string(
 *                         current_entry[1][self.target_index][0])             # <<<<<<<<<<<<<<
 *                     cmp_result = ((current_path_info[1] > current_entry[0][1]) -
 *                                   (current_path_info[1] < current_entry[0][1]))
*/
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_current_entry, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1765, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_self->target_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1765, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1765, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1764
 *                         current_path_info)
 *                 else:
 *                     minikind = _minikind_from_string(             # <<<<<<<<<<<<<<
 *                         current_entry[1][self.target_index][0])
 *                     cmp_result = ((current_path_info[1] > current_entry[0][1]) -
*/
      __pyx_t_8 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_from_string(__pyx_t_3); if (unlikely(__pyx_t_8 == ((char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1764, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_minikind = __pyx_t_8;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1766
 *                     minikind = _minikind_from_string(
 *                         current_entry[1][self.target_index][0])
 *                     cmp_result = ((current_path_info[1] > current_entry[0][1]) -             # <<<<<<<<<<<<<<
 *                                   (current_path_info[1] < current_entry[0][1]))
 *                     if (cmp_result or minikind == b'a' or minikind == b'r'):
*/
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_current_path_info, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_current_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_5, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_GT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1766, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1767
 *                         current_entry[1][self.target_index][0])
 *                     cmp_result = ((current_path_info[1] > current_entry[0][1]) -
 *                                   (current_path_info[1] < current_entry[0][1]))             # <<<<<<<<<<<<<<
 *                     if (cmp_result or minikind == b'a' or minikind == b'r'):
 *                         # The current path on disk doesn't match the dirblock
*/
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_current_path_info, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1767, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_current_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1767, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_3, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1767, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyObject_RichCompare(__pyx_t_4, __pyx_t_6, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1767, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1766
 *                     minikind = _minikind_from_string(
 *                         current_entry[1][self.target_index][0])
 *                     cmp_result = ((current_path_info[1] > current_entry[0][1]) -             # <<<<<<<<<<<<<<
 *                                   (current_path_info[1] < current_entry[0][1]))
 *                     if (cmp_result or minikind == b'a' or minikind == b'r'):
*/
      __pyx_t_6 = PyNumber_Subtract(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_9 = __Pyx_PyLong_As_int(__pyx_t_6); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1766, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_v_cmp_result = __pyx_t_9;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1768
 *                     cmp_result = ((current_path_info[1] > current_entry[0][1]) -
 *                                   (current_path_info[1] < current_entry[0][1]))
 *                     if (cmp_result or minikind == b'a' or minikind == b'r'):             # <<<<<<<<<<<<<<
 *                         # The current path on disk doesn't match the dirblock
 *                         # record. Either the dirblock record is marked as
*/
      __pyx_t_2 = (__pyx_v_cmp_result != 0);
      if (!__pyx_t_2) {
      } else {
        __pyx_t_1 = __pyx_t_2;
        goto __pyx_L18_bool_binop_done;
      }
      switch (__pyx_v_minikind) {
        case 'a':
        case 'r':
        __pyx_t_2 = 1;
        break;
        default:
        __pyx_t_2 = 0;
        break;
      }
      __pyx_t_1 = __pyx_t_2;
      __pyx_L18_bool_binop_done:;
      if (__pyx_t_1) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1777
 *                         # Compare the basename for these files to determine
 *                         # which comes first
 *                         if cmp_result < 0:             # <<<<<<<<<<<<<<
 *                             # extra file on disk: pass for now, but only
 *                             # increment the path, not the entry
*/
        __pyx_t_1 = (__pyx_v_cmp_result < 0);
        if (__pyx_t_1) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1780
 *                             # extra file on disk: pass for now, but only
 *                             # increment the path, not the entry
 *                             advance_entry = 0             # <<<<<<<<<<<<<<
 *                         else:
 *                             # entry referring to file not present on disk.
*/
          __pyx_v_advance_entry = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1777
 *                         # Compare the basename for these files to determine
 *                         # which comes first
 *                         if cmp_result < 0:             # <<<<<<<<<<<<<<
 *                             # extra file on disk: pass for now, but only
 *                             # increment the path, not the entry
*/
          goto __pyx_L20;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1784
 *                             # entry referring to file not present on disk.
 *                             # advance the entry only, after processing.
 *                             result, changed = self._process_entry(current_entry,             # <<<<<<<<<<<<<<
 *                                 None)
 *                             advance_path = 0
*/
        /*else*/ {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1785
 *                             # advance the entry only, after processing.
 *                             result, changed = self._process_entry(current_entry,
 *                                 None)             # <<<<<<<<<<<<<<
 *                             advance_path = 0
 *                     else:
*/
          __pyx_t_6 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_process_entry(__pyx_v_self, __pyx_v_current_entry, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1784, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
            PyObject* sequence = __pyx_t_6;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 1784, __pyx_L1_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0);
              __Pyx_INCREF(__pyx_t_3);
              __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1);
              __Pyx_INCREF(__pyx_t_5);
            } else {
              __pyx_t_3 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
              if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1784, __pyx_L1_error)
              __Pyx_XGOTREF(__pyx_t_3);
              __pyx_t_5 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
              if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1784, __pyx_L1_error)
              __Pyx_XGOTREF(__pyx_t_5);
            }
            #else
            __pyx_t_3 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1784, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_5 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1784, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            #endif
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_4 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1784, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_7 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_4);
            index = 0; __pyx_t_3 = __pyx_t_7(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L21_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_3);
            index = 1; __pyx_t_5 = __pyx_t_7(__pyx_t_4); if (unlikely(!__pyx_t_5)) goto __pyx_L21_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_5);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_4), 2) < (0)) __PYX_ERR(0, 1784, __pyx_L1_error)
            __pyx_t_7 = NULL;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            goto __pyx_L22_unpacking_done;
            __pyx_L21_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_7 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 1784, __pyx_L1_error)
            __pyx_L22_unpacking_done:;
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1784
 *                             # entry referring to file not present on disk.
 *                             # advance the entry only, after processing.
 *                             result, changed = self._process_entry(current_entry,             # <<<<<<<<<<<<<<
 *                                 None)
 *                             advance_path = 0
*/
          __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_3);
          __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v_changed, __pyx_t_5);
          __pyx_t_5 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1786
 *                             result, changed = self._process_entry(current_entry,
 *                                 None)
 *                             advance_path = 0             # <<<<<<<<<<<<<<
 *                     else:
 *                         # paths are the same,and the dirstate entry is not
*/
          __pyx_v_advance_path = 0;
        }
        __pyx_L20:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1768
 *                     cmp_result = ((current_path_info[1] > current_entry[0][1]) -
 *                                   (current_path_info[1] < current_entry[0][1]))
 *                     if (cmp_result or minikind == b'a' or minikind == b'r'):             # <<<<<<<<<<<<<<
 *                         # The current path on disk doesn't match the dirblock
 *                         # record. Either the dirblock record is marked as
*/
        goto __pyx_L17;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1790
 *                         # paths are the same,and the dirstate entry is not
 *                         # absent or renamed.
 *                         result, changed = self._process_entry(current_entry,             # <<<<<<<<<<<<<<
 *                             current_path_info)
 *                         if changed is not None:
*/
      /*else*/ {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1791
 *                         # absent or renamed.
 *                         result, changed = self._process_entry(current_entry,
 *                             current_path_info)             # <<<<<<<<<<<<<<
 *                         if changed is not None:
 *                             path_handled = -1
*/
        __pyx_t_6 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_process_entry(__pyx_v_self, __pyx_v_current_entry, __pyx_v_current_path_info); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1790, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
          PyObject* sequence = __pyx_t_6;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 1790, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0);
            __Pyx_INCREF(__pyx_t_5);
            __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1);
            __Pyx_INCREF(__pyx_t_3);
          } else {
            __pyx_t_5 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1790, __pyx_L1_error)
            __Pyx_XGOTREF(__pyx_t_5);
            __pyx_t_3 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1790, __pyx_L1_error)
            __Pyx_XGOTREF(__pyx_t_3);
          }
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1790, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1790, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_4 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1790, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_7 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_4);
          index = 0; __pyx_t_5 = __pyx_t_7(__pyx_t_4); if (unlikely(!__pyx_t_5)) goto __pyx_L23_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_5);
          index = 1; __pyx_t_3 = __pyx_t_7(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L23_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_3);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_4), 2) < (0)) __PYX_ERR(0, 1790, __pyx_L1_error)
          __pyx_t_7 = NULL;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          goto __pyx_L24_unpacking_done;
          __pyx_L23_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_7 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 1790, __pyx_L1_error)
          __pyx_L24_unpacking_done:;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1790
 *                         # paths are the same,and the dirstate entry is not
 *                         # absent or renamed.
 *                         result, changed = self._process_entry(current_entry,             # <<<<<<<<<<<<<<
 *                             current_path_info)
 *                         if changed is not None:
*/
        __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_5);
        __pyx_t_5 = 0;
        __Pyx_DECREF_SET(__pyx_v_changed, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1792
 *                         result, changed = self._process_entry(current_entry,
 *                             current_path_info)
 *                         if changed is not None:             # <<<<<<<<<<<<<<
 *                             path_handled = -1
 *                             if not changed and not self.include_unchanged:
*/
        __pyx_t_1 = (__pyx_v_changed != Py_None);
        if (__pyx_t_1) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1793
 *                             current_path_info)
 *                         if changed is not None:
 *                             path_handled = -1             # <<<<<<<<<<<<<<
 *                             if not changed and not self.include_unchanged:
 *                                 changed = None
*/
          __pyx_v_path_handled = -1;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1794
 *                         if changed is not None:
 *                             path_handled = -1
 *                             if not changed and not self.include_unchanged:             # <<<<<<<<<<<<<<
 *                                 changed = None
 *                 # >- loop control starts here:
*/
          __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1794, __pyx_L1_error)
          __pyx_t_10 = (!__pyx_t_2);
          if (__pyx_t_10) {
          } else {
            __pyx_t_1 = __pyx_t_10;
            goto __pyx_L27_bool_binop_done;
          }
          __pyx_t_10 = (!(__pyx_v_self->include_unchanged != 0));
          __pyx_t_1 = __pyx_t_10;
          __pyx_L27_bool_binop_done:;
          if (__pyx_t_1) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1795
 *                             path_handled = -1
 *                             if not changed and not self.include_unchanged:
 *                                 changed = None             # <<<<<<<<<<<<<<
 *                 # >- loop control starts here:
 *                 # >- entry
*/
            __Pyx_INCREF(Py_None);
            __Pyx_DECREF_SET(__pyx_v_changed, Py_None);

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1794
 *                         if changed is not None:
 *                             path_handled = -1
 *                             if not changed and not self.include_unchanged:             # <<<<<<<<<<<<<<
 *                                 changed = None
 *                 # >- loop control starts here:
*/
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1792
 *                         result, changed = self._process_entry(current_entry,
 *                             current_path_info)
 *                         if changed is not None:             # <<<<<<<<<<<<<<
 *                             path_handled = -1
 *                             if not changed and not self.include_unchanged:
*/
        }
      }
      __pyx_L17:;
    }
    __pyx_L14:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1798
 *                 # >- loop control starts here:
 *                 # >- entry
 *                 if advance_entry and current_entry is not None:             # <<<<<<<<<<<<<<
 *                     self.current_block_pos = self.current_block_pos + 1
 *                     if self.current_block_pos < PyList_GET_SIZE(self.current_block_list):
*/
    __pyx_t_10 = (__pyx_v_advance_entry != 0);
    if (__pyx_t_10) {
    } else {
      __pyx_t_1 = __pyx_t_10;
      goto __pyx_L30_bool_binop_done;
    }
    __pyx_t_10 = (__pyx_v_current_entry != Py_None);
    __pyx_t_1 = __pyx_t_10;
    __pyx_L30_bool_binop_done:;
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1799
 *                 # >- entry
 *                 if advance_entry and current_entry is not None:
 *                     self.current_block_pos = self.current_block_pos + 1             # <<<<<<<<<<<<<<
 *                     if self.current_block_pos < PyList_GET_SIZE(self.current_block_list):
 *                         current_entry = self.current_block_list[self.current_block_pos]
*/
      __pyx_v_self->current_block_pos = (__pyx_v_self->current_block_pos + 1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1800
 *                 if advance_entry and current_entry is not None:
 *                     self.current_block_pos = self.current_block_pos + 1
 *                     if self.current_block_pos < PyList_GET_SIZE(self.current_block_list):             # <<<<<<<<<<<<<<
 *                         current_entry = self.current_block_list[self.current_block_pos]
 *                     else:
*/
      __pyx_t_6 = __pyx_v_self->current_block_list;
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_1 = (__pyx_v_self->current_block_pos < PyList_GET_SIZE(__pyx_t_6));
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (__pyx_t_1) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1801
 *                     self.current_block_pos = self.current_block_pos + 1
 *                     if self.current_block_pos < PyList_GET_SIZE(self.current_block_list):
 *                         current_entry = self.current_block_list[self.current_block_pos]             # <<<<<<<<<<<<<<
 *                     else:
 *                         current_entry = None
*/
        __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_self->current_block_list, __pyx_v_self->current_block_pos, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1801, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF_SET(__pyx_v_current_entry, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1800
 *                 if advance_entry and current_entry is not None:
 *                     self.current_block_pos = self.current_block_pos + 1
 *                     if self.current_block_pos < PyList_GET_SIZE(self.current_block_list):             # <<<<<<<<<<<<<<
 *                         current_entry = self.current_block_list[self.current_block_pos]
 *                     else:
*/
        goto __pyx_L32;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1803
 *                         current_entry = self.current_block_list[self.current_block_pos]
 *                     else:
 *                         current_entry = None             # <<<<<<<<<<<<<<
 *                 # >- path
 *                 if advance_path and current_path_info is not None:
*/
      /*else*/ {
        __Pyx_INCREF(Py_None);
        __Pyx_DECREF_SET(__pyx_v_current_entry, Py_None);
      }
      __pyx_L32:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1798
 *                 # >- loop control starts here:
 *                 # >- entry
 *                 if advance_entry and current_entry is not None:             # <<<<<<<<<<<<<<
 *                     self.current_block_pos = self.current_block_pos + 1
 *                     if self.current_block_pos < PyList_GET_SIZE(self.current_block_list):
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1805
 *                         current_entry = None
 *                 # >- path
 *                 if advance_path and current_path_info is not None:             # <<<<<<<<<<<<<<
 *                     if not path_handled:
 *                         # unversioned in all regards
*/
    __pyx_t_10 = (__pyx_v_advance_path != 0);
    if (__pyx_t_10) {
    } else {
      __pyx_t_1 = __pyx_t_10;
      goto __pyx_L34_bool_binop_done;
    }
    __pyx_t_10 = (__pyx_v_current_path_info != Py_None);
    __pyx_t_1 = __pyx_t_10;
    __pyx_L34_bool_binop_done:;
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1806
 *                 # >- path
 *                 if advance_path and current_path_info is not None:
 *                     if not path_handled:             # <<<<<<<<<<<<<<
 *                         # unversioned in all regards
 *                         if self.want_unversioned:
*/
      __pyx_t_1 = (!(__pyx_v_path_handled != 0));
      if (__pyx_t_1) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1808
 *                     if not path_handled:
 *                         # unversioned in all regards
 *                         if self.want_unversioned:             # <<<<<<<<<<<<<<
 *                             new_executable = bool(
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
*/
        __pyx_t_1 = (__pyx_v_self->want_unversioned != 0);
        if (__pyx_t_1) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1810
 *                         if self.want_unversioned:
 *                             new_executable = bool(
 *                                 stat.S_ISREG(current_path_info[3].st_mode)             # <<<<<<<<<<<<<<
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)
 *                             relpath_unicode = self.utf8_decode(current_path_info[0], 'surrogateescape')[0]
*/
          __pyx_t_3 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_stat); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1810, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_S_ISREG); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1810, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_current_path_info, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1810, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_st_mode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1810, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_12 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            assert(__pyx_t_3);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
            __pyx_t_12 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_11};
            __pyx_t_6 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1810, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
          }
          __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 1810, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (__pyx_t_10) {
          } else {
            __pyx_t_1 = __pyx_t_10;
            goto __pyx_L38_bool_binop_done;
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1811
 *                             new_executable = bool(
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)             # <<<<<<<<<<<<<<
 *                             relpath_unicode = self.utf8_decode(current_path_info[0], 'surrogateescape')[0]
 *                             if changed is not None:
*/
          __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_stat); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1811, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_S_IEXEC); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1811, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_current_path_info, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1811, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_st_mode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1811, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_6 = PyNumber_And(__pyx_t_4, __pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1811, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 1811, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_1 = __pyx_t_10;
          __pyx_L38_bool_binop_done:;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1809
 *                         # unversioned in all regards
 *                         if self.want_unversioned:
 *                             new_executable = bool(             # <<<<<<<<<<<<<<
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)
*/
          __pyx_v_new_executable = (!(!__pyx_t_1));

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1812
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)
 *                             relpath_unicode = self.utf8_decode(current_path_info[0], 'surrogateescape')[0]             # <<<<<<<<<<<<<<
 *                             if changed is not None:
 *                                 raise AssertionError(
*/
          __pyx_t_11 = NULL;
          __Pyx_INCREF(__pyx_v_self->utf8_decode);
          __pyx_t_4 = __pyx_v_self->utf8_decode; 
          __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_current_path_info, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1812, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_12 = 1;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_4);
            assert(__pyx_t_11);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
            __pyx_t_12 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_11, __pyx_t_3, __pyx_mstate_global->__pyx_n_u_surrogateescape};
            __pyx_t_6 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1812, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
          }
          __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_6, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1812, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF_SET(__pyx_v_relpath_unicode, __pyx_t_4);
          __pyx_t_4 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1813
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)
 *                             relpath_unicode = self.utf8_decode(current_path_info[0], 'surrogateescape')[0]
 *                             if changed is not None:             # <<<<<<<<<<<<<<
 *                                 raise AssertionError(
 *                                     "result is not None: %r" % result)
*/
          __pyx_t_1 = (__pyx_v_changed != Py_None);
          if (unlikely(__pyx_t_1)) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1814
 *                             relpath_unicode = self.utf8_decode(current_path_info[0], 'surrogateescape')[0]
 *                             if changed is not None:
 *                                 raise AssertionError(             # <<<<<<<<<<<<<<
 *                                     "result is not None: %r" % result)
 *                             result = InventoryTreeChange(
*/
            __pyx_t_6 = NULL;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1815
 *                             if changed is not None:
 *                                 raise AssertionError(
 *                                     "result is not None: %r" % result)             # <<<<<<<<<<<<<<
 *                             result = InventoryTreeChange(
 *                                 None,
*/
            __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_mstate_global->__pyx_kp_u_result_is_not_None_r, __pyx_v_result); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1815, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_12 = 1;
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_3};
              __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1814, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
            }
            __Pyx_Raise(__pyx_t_4, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __PYX_ERR(0, 1814, __pyx_L1_error)

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1813
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)
 *                             relpath_unicode = self.utf8_decode(current_path_info[0], 'surrogateescape')[0]
 *                             if changed is not None:             # <<<<<<<<<<<<<<
 *                                 raise AssertionError(
 *                                     "result is not None: %r" % result)
*/
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1816
 *                                 raise AssertionError(
 *                                     "result is not None: %r" % result)
 *                             result = InventoryTreeChange(             # <<<<<<<<<<<<<<
 *                                 None,
 *                                 (None, relpath_unicode),
*/
          __pyx_t_3 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_InventoryTreeChange); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1816, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1818
 *                             result = InventoryTreeChange(
 *                                 None,
 *                                 (None, relpath_unicode),             # <<<<<<<<<<<<<<
 *                                 True,
 *                                 (False, False),
*/
          __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1818, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_INCREF(Py_None);
          __Pyx_GIVEREF(Py_None);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 0, Py_None) != (0)) __PYX_ERR(0, 1818, __pyx_L1_error);
          __Pyx_INCREF(__pyx_v_relpath_unicode);
          __Pyx_GIVEREF(__pyx_v_relpath_unicode);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_v_relpath_unicode) != (0)) __PYX_ERR(0, 1818, __pyx_L1_error);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1822
 *                                 (False, False),
 *                                 (None, None),
 *                                 (None, self.utf8_decode(current_path_info[1], 'surrogateescape')[0]),             # <<<<<<<<<<<<<<
 *                                 (None, current_path_info[2]),
 *                                 (None, new_executable))
*/
          __pyx_t_13 = NULL;
          __Pyx_INCREF(__pyx_v_self->utf8_decode);
          __pyx_t_14 = __pyx_v_self->utf8_decode; 
          __pyx_t_15 = __Pyx_GetItemInt(__pyx_v_current_path_info, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1822, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_12 = 1;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_14))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_14);
            assert(__pyx_t_13);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_14);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_14, __pyx__function);
            __pyx_t_12 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_13, __pyx_t_15, __pyx_mstate_global->__pyx_n_u_surrogateescape};
            __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_14, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1822, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
          }
          __pyx_t_14 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1822, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1822, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_INCREF(Py_None);
          __Pyx_GIVEREF(Py_None);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, Py_None) != (0)) __PYX_ERR(0, 1822, __pyx_L1_error);
          __Pyx_GIVEREF(__pyx_t_14);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_14) != (0)) __PYX_ERR(0, 1822, __pyx_L1_error);
          __pyx_t_14 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1823
 *                                 (None, None),
 *                                 (None, self.utf8_decode(current_path_info[1], 'surrogateescape')[0]),
 *                                 (None, current_path_info[2]),             # <<<<<<<<<<<<<<
 *                                 (None, new_executable))
 *                             changed = True
*/
          __pyx_t_14 = __Pyx_GetItemInt(__pyx_v_current_path_info, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1823, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1823, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_INCREF(Py_None);
          __Pyx_GIVEREF(Py_None);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 0, Py_None) != (0)) __PYX_ERR(0, 1823, __pyx_L1_error);
          __Pyx_GIVEREF(__pyx_t_14);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_14) != (0)) __PYX_ERR(0, 1823, __pyx_L1_error);
          __pyx_t_14 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1824
 *                                 (None, self.utf8_decode(current_path_info[1], 'surrogateescape')[0]),
 *                                 (None, current_path_info[2]),
 *                                 (None, new_executable))             # <<<<<<<<<<<<<<
 *                             changed = True
 *                         # dont descend into this unversioned path if it is
*/
          __pyx_t_14 = __Pyx_PyBool_FromLong(__pyx_v_new_executable); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1824, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1824, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_INCREF(Py_None);
          __Pyx_GIVEREF(Py_None);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 0, Py_None) != (0)) __PYX_ERR(0, 1824, __pyx_L1_error);
          __Pyx_GIVEREF(__pyx_t_14);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_14) != (0)) __PYX_ERR(0, 1824, __pyx_L1_error);
          __pyx_t_14 = 0;
          __pyx_t_12 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
            assert(__pyx_t_3);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
            __pyx_t_12 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[9] = {__pyx_t_3, Py_None, __pyx_t_11, Py_True, __pyx_mstate_global->__pyx_tuple[6], __pyx_mstate_global->__pyx_tuple[0], __pyx_t_5, __pyx_t_15, __pyx_t_13};
            __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_12, (9-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1816, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
          }
          __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_4);
          __pyx_t_4 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1825
 *                                 (None, current_path_info[2]),
 *                                 (None, new_executable))
 *                             changed = True             # <<<<<<<<<<<<<<
 *                         # dont descend into this unversioned path if it is
 *                         # a dir
*/
          __Pyx_INCREF(Py_True);
          __Pyx_DECREF_SET(__pyx_v_changed, Py_True);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1808
 *                     if not path_handled:
 *                         # unversioned in all regards
 *                         if self.want_unversioned:             # <<<<<<<<<<<<<<
 *                             new_executable = bool(
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
*/
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1828
 *                         # dont descend into this unversioned path if it is
 *                         # a dir
 *                         if current_path_info[2] in ('directory'):             # <<<<<<<<<<<<<<
 *                             del self.current_dir_list[self.path_index]
 *                             self.path_index = self.path_index - 1
*/
        __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_current_path_info, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1828, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = (__Pyx_PyUnicode_ContainsTF(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_directory, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1828, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (__pyx_t_1) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1829
 *                         # a dir
 *                         if current_path_info[2] in ('directory'):
 *                             del self.current_dir_list[self.path_index]             # <<<<<<<<<<<<<<
 *                             self.path_index = self.path_index - 1
 *                     # dont descend the disk iterator into any tree
*/
          if (unlikely((__Pyx_DelItemInt(__pyx_v_self->current_dir_list, __pyx_v_self->path_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_SharedReference) < 0))) __PYX_ERR(0, 1829, __pyx_L1_error)

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1830
 *                         if current_path_info[2] in ('directory'):
 *                             del self.current_dir_list[self.path_index]
 *                             self.path_index = self.path_index - 1             # <<<<<<<<<<<<<<
 *                     # dont descend the disk iterator into any tree
 *                     # paths.
*/
          __pyx_v_self->path_index = (__pyx_v_self->path_index - 1);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1828
 *                         # dont descend into this unversioned path if it is
 *                         # a dir
 *                         if current_path_info[2] in ('directory'):             # <<<<<<<<<<<<<<
 *                             del self.current_dir_list[self.path_index]
 *                             self.path_index = self.path_index - 1
*/
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1806
 *                 # >- path
 *                 if advance_path and current_path_info is not None:
 *                     if not path_handled:             # <<<<<<<<<<<<<<
 *                         # unversioned in all regards
 *                         if self.want_unversioned:
*/
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1833
 *                     # dont descend the disk iterator into any tree
 *                     # paths.
 *                     if current_path_info[2] == 'tree-reference':             # <<<<<<<<<<<<<<
 *                         del self.current_dir_list[self.path_index]
 *                         self.path_index = self.path_index - 1
*/
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_current_path_info, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1833, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_tree_reference, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1833, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_1) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1834
 *                     # paths.
 *                     if current_path_info[2] == 'tree-reference':
 *                         del self.current_dir_list[self.path_index]             # <<<<<<<<<<<<<<
 *                         self.path_index = self.path_index - 1
 *                     self.path_index = self.path_index + 1
*/
        if (unlikely((__Pyx_DelItemInt(__pyx_v_self->current_dir_list, __pyx_v_self->path_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_SharedReference) < 0))) __PYX_ERR(0, 1834, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1835
 *                     if current_path_info[2] == 'tree-reference':
 *                         del self.current_dir_list[self.path_index]
 *                         self.path_index = self.path_index - 1             # <<<<<<<<<<<<<<
 *                     self.path_index = self.path_index + 1
 *                     if self.path_index < len(self.current_dir_list):
*/
        __pyx_v_self->path_index = (__pyx_v_self->path_index - 1);

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1833
 *                     # dont descend the disk iterator into any tree
 *                     # paths.
 *                     if current_path_info[2] == 'tree-reference':             # <<<<<<<<<<<<<<
 *                         del self.current_dir_list[self.path_index]
 *                         self.path_index = self.path_index - 1
*/
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1836
 *                         del self.current_dir_list[self.path_index]
 *                         self.path_index = self.path_index - 1
 *                     self.path_index = self.path_index + 1             # <<<<<<<<<<<<<<
 *                     if self.path_index < len(self.current_dir_list):
 *                         current_path_info = self.current_dir_list[self.path_index]
*/
      __pyx_v_self->path_index = (__pyx_v_self->path_index + 1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1837
 *                         self.path_index = self.path_index - 1
 *                     self.path_index = self.path_index + 1
 *                     if self.path_index < len(self.current_dir_list):             # <<<<<<<<<<<<<<
 *                         current_path_info = self.current_dir_list[self.path_index]
 *                         if current_path_info[2] == 'directory':
*/
      __pyx_t_4 = __pyx_v_self->current_dir_list;
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_16 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_16 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1837, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_1 = (__pyx_v_self->path_index < __pyx_t_16);
      if (__pyx_t_1) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1838
 *                     self.path_index = self.path_index + 1
 *                     if self.path_index < len(self.current_dir_list):
 *                         current_path_info = self.current_dir_list[self.path_index]             # <<<<<<<<<<<<<<
 *                         if current_path_info[2] == 'directory':
 *                             current_path_info = self._maybe_tree_ref(
*/
        __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_self->current_dir_list, __pyx_v_self->path_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1838, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF_SET(__pyx_v_current_path_info, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1839
 *                     if self.path_index < len(self.current_dir_list):
 *                         current_path_info = self.current_dir_list[self.path_index]
 *                         if current_path_info[2] == 'directory':             # <<<<<<<<<<<<<<
 *                             current_path_info = self._maybe_tree_ref(
 *                                 current_path_info)
*/
        __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_current_path_info, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1839, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_directory, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1839, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (__pyx_t_1) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1840
 *                         current_path_info = self.current_dir_list[self.path_index]
 *                         if current_path_info[2] == 'directory':
 *                             current_path_info = self._maybe_tree_ref(             # <<<<<<<<<<<<<<
 *                                 current_path_info)
 *                     else:
*/
          __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_maybe_tree_ref(__pyx_v_self, __pyx_v_current_path_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1840, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF_SET(__pyx_v_current_path_info, __pyx_t_4);
          __pyx_t_4 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1839
 *                     if self.path_index < len(self.current_dir_list):
 *                         current_path_info = self.current_dir_list[self.path_index]
 *                         if current_path_info[2] == 'directory':             # <<<<<<<<<<<<<<
 *                             current_path_info = self._maybe_tree_ref(
 *                                 current_path_info)
*/
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1837
 *                         self.path_index = self.path_index - 1
 *                     self.path_index = self.path_index + 1
 *                     if self.path_index < len(self.current_dir_list):             # <<<<<<<<<<<<<<
 *                         current_path_info = self.current_dir_list[self.path_index]
 *                         if current_path_info[2] == 'directory':
*/
        goto __pyx_L43;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1843
 *                                 current_path_info)
 *                     else:
 *                         current_path_info = None             # <<<<<<<<<<<<<<
 *                 if changed is not None:
 *                     # Found a result on this pass, yield it
*/
      /*else*/ {
        __Pyx_INCREF(Py_None);
        __Pyx_DECREF_SET(__pyx_v_current_path_info, Py_None);
      }
      __pyx_L43:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1805
 *                         current_entry = None
 *                 # >- path
 *                 if advance_path and current_path_info is not None:             # <<<<<<<<<<<<<<
 *                     if not path_handled:
 *                         # unversioned in all regards
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1844
 *                     else:
 *                         current_path_info = None
 *                 if changed is not None:             # <<<<<<<<<<<<<<
 *                     # Found a result on this pass, yield it
 *                     if changed:
*/
    __pyx_t_1 = (__pyx_v_changed != Py_None);
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1846
 *                 if changed is not None:
 *                     # Found a result on this pass, yield it
 *                     if changed:             # <<<<<<<<<<<<<<
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:
*/
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1846, __pyx_L1_error)
      if (__pyx_t_1) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1847
 *                     # Found a result on this pass, yield it
 *                     if changed:
 *                         self._gather_result_for_consistency(result)             # <<<<<<<<<<<<<<
 *                     if changed or self.include_unchanged:
 *                         return result
*/
        __pyx_t_9 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_gather_result_for_consistency(__pyx_v_self, __pyx_v_result); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 1847, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1846
 *                 if changed is not None:
 *                     # Found a result on this pass, yield it
 *                     if changed:             # <<<<<<<<<<<<<<
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:
*/
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1848
 *                     if changed:
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                         return result
 *             if self.current_block is not None:
*/
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 1848, __pyx_L1_error)
      if (!__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L48_bool_binop_done;
      }
      __pyx_t_10 = (__pyx_v_self->include_unchanged != 0);
      __pyx_t_1 = __pyx_t_10;
      __pyx_L48_bool_binop_done:;
      if (__pyx_t_1) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1849
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:
 *                         return result             # <<<<<<<<<<<<<<
 *             if self.current_block is not None:
 *                 self.block_index = self.block_index + 1
*/
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_result);
        __pyx_r = __pyx_v_result;
        goto __pyx_L0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1848
 *                     if changed:
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                         return result
 *             if self.current_block is not None:
*/
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1844
 *                     else:
 *                         current_path_info = None
 *                 if changed is not None:             # <<<<<<<<<<<<<<
 *                     # Found a result on this pass, yield it
 *                     if changed:
*/
    }
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1850
 *                     if changed or self.include_unchanged:
 *                         return result
 *             if self.current_block is not None:             # <<<<<<<<<<<<<<
 *                 self.block_index = self.block_index + 1
 *                 self._update_current_block()
*/
  __pyx_t_1 = (__pyx_v_self->current_block != Py_None);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1851
 *                         return result
 *             if self.current_block is not None:
 *                 self.block_index = self.block_index + 1             # <<<<<<<<<<<<<<
 *                 self._update_current_block()
 *             if self.current_dir_info is not None:
*/
    __pyx_v_self->block_index = (__pyx_v_self->block_index + 1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1852
 *             if self.current_block is not None:
 *                 self.block_index = self.block_index + 1
 *                 self._update_current_block()             # <<<<<<<<<<<<<<
 *             if self.current_dir_info is not None:
 *                 self.path_index = 0
*/
    __pyx_t_9 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_update_current_block(__pyx_v_self); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 1852, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1850
 *                     if changed or self.include_unchanged:
 *                         return result
 *             if self.current_block is not None:             # <<<<<<<<<<<<<<
 *                 self.block_index = self.block_index + 1
 *                 self._update_current_block()
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1853
 *                 self.block_index = self.block_index + 1
 *                 self._update_current_block()
 *             if self.current_dir_info is not None:             # <<<<<<<<<<<<<<
 *                 self.path_index = 0
 *                 self.current_dir_list = None
*/
  __pyx_t_1 = (__pyx_v_self->current_dir_info != Py_None);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1854
 *                 self._update_current_block()
 *             if self.current_dir_info is not None:
 *                 self.path_index = 0             # <<<<<<<<<<<<<<
 *                 self.current_dir_list = None
 *                 try:
*/
    __pyx_v_self->path_index = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1855
 *             if self.current_dir_info is not None:
 *                 self.path_index = 0
 *                 self.current_dir_list = None             # <<<<<<<<<<<<<<
 *                 try:
 *                     self.current_dir_info = next(self.dir_iterator)
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->current_dir_list);
    __Pyx_DECREF(__pyx_v_self->current_dir_list);
    __pyx_v_self->current_dir_list = Py_None;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1856
 *                 self.path_index = 0
 *                 self.current_dir_list = None
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      /*try:*/ {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1857
 *                 self.current_dir_list = None
 *                 try:
 *                     self.current_dir_info = next(self.dir_iterator)             # <<<<<<<<<<<<<<
 *                     self.current_dir_list = self.current_dir_info[1]
 *                 except StopIteration, _:
*/
        __pyx_t_4 = __pyx_v_self->dir_iterator;
        __Pyx_INCREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyIter_Next(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1857, __pyx_L52_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GIVEREF(__pyx_t_6);
        __Pyx_GOTREF(__pyx_v_self->current_dir_info);
        __Pyx_DECREF(__pyx_v_self->current_dir_info);
        __pyx_v_self->current_dir_info = __pyx_t_6;
        __pyx_t_6 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1858
 *                 try:
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]             # <<<<<<<<<<<<<<
 *                 except StopIteration, _:
 *                     self.current_dir_info = None
*/
        __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_self->current_dir_info, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1858, __pyx_L52_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_6);
        __Pyx_GOTREF(__pyx_v_self->current_dir_list);
        __Pyx_DECREF(__pyx_v_self->current_dir_list);
        __pyx_v_self->current_dir_list = __pyx_t_6;
        __pyx_t_6 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1856
 *                 self.path_index = 0
 *                 self.current_dir_list = None
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]
*/
      }
      __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
      __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
      goto __pyx_L57_try_end;
      __pyx_L52_error:;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1859
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]
 *                 except StopIteration, _:             # <<<<<<<<<<<<<<
 *                     self.current_dir_info = None
 * 
*/
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_StopIteration))));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._loop_one_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_4, &__pyx_t_13) < 0) __PYX_ERR(0, 1859, __pyx_L54_except_error)
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v__ = __pyx_t_4;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1860
 *                     self.current_dir_list = self.current_dir_info[1]
 *                 except StopIteration, _:
 *                     self.current_dir_info = None             # <<<<<<<<<<<<<<
 * 
 *     cdef object _next_consistent_entries(self):
*/
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        __Pyx_GOTREF(__pyx_v_self->current_dir_info);
        __Pyx_DECREF(__pyx_v_self->current_dir_info);
        __pyx_v_self->current_dir_info = Py_None;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        goto __pyx_L53_exception_handled;
      }
      goto __pyx_L54_except_error;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1856
 *                 self.path_index = 0
 *                 self.current_dir_list = None
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]
*/
      __pyx_L54_except_error:;
      __Pyx_XGIVEREF(__pyx_t_17);
      __Pyx_XGIVEREF(__pyx_t_18);
      __Pyx_XGIVEREF(__pyx_t_19);
      __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
      goto __pyx_L1_error;
      __pyx_L53_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_17);
      __Pyx_XGIVEREF(__pyx_t_18);
      __Pyx_XGIVEREF(__pyx_t_19);
      __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
      __pyx_L57_try_end:;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1853
 *                 self.block_index = self.block_index + 1
 *                 self._update_current_block()
 *             if self.current_dir_info is not None:             # <<<<<<<<<<<<<<
 *                 self.path_index = 0
 *                 self.current_dir_list = None
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1715
 *             return current_path_info
 * 
 *     cdef object _loop_one_block(self):             # <<<<<<<<<<<<<<
 *             # current_dir_info and current_block refer to the same directory -
 *             # this is the common case code.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._loop_one_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_current_entry);
  __Pyx_XDECREF(__pyx_v_current_path_info);
  __Pyx_XDECREF(__pyx_v_disk_kind);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_changed);
  __Pyx_XDECREF(__pyx_v_relpath_unicode);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1862
 *                     self.current_dir_info = None
 * 
 *     cdef object _next_consistent_entries(self):             # <<<<<<<<<<<<<<
 *         """Grabs the next specific file parent case to consider.
 * 
*/

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__next_consistent_entries(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  PyObject *__pyx_v_results = NULL;
  PyObject *__pyx_v_path_utf8 = NULL;
  PyObject *__pyx_v_path_entries = NULL;
  PyObject *__pyx_v_selected_entries = NULL;
  int __pyx_v_found_item;
  PyObject *__pyx_v_candidate_entry = NULL;
  PyObject *__pyx_v_path_info = NULL;
  PyObject *__pyx_v_entry = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_changed = NULL;
  PyObject *__pyx_v_entry_path_utf8 = NULL;
  PyObject *__pyx_v_initial_key = NULL;
  PyObject *__pyx_v_block_index = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_current_block = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13[4];
  PyObject *(*__pyx_t_14)(PyObject *);
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16[7];
  PyObject *__pyx_t_17 = NULL;
  int __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_next_consistent_entries", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1867
 *         :return: A list of the results, each of which is as for _process_entry.
 *         """
 *         results = []             # <<<<<<<<<<<<<<
 *         while self.search_specific_file_parents:
 *             # Process the parent directories for the paths we were iterating.
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_results = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1868
 *         """
 *         results = []
 *         while self.search_specific_file_parents:             # <<<<<<<<<<<<<<
 *             # Process the parent directories for the paths we were iterating.
 *             # Even in extremely large trees this should be modest, so currently
*/
  while (1) {
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_self->search_specific_file_parents); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1868, __pyx_L1_error)
    if (!__pyx_t_2) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1872
 *             # Even in extremely large trees this should be modest, so currently
 *             # no attempt is made to optimise.
 *             path_utf8 = self.search_specific_file_parents.pop()             # <<<<<<<<<<<<<<
 *             if path_utf8 in self.searched_exact_paths:
 *                 # We've examined this path.
*/
    __pyx_t_1 = __Pyx_PyObject_Pop(__pyx_v_self->search_specific_file_parents); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1872, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_path_utf8, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1873
 *             # no attempt is made to optimise.
 *             path_utf8 = self.search_specific_file_parents.pop()
 *             if path_utf8 in self.searched_exact_paths:             # <<<<<<<<<<<<<<
 *                 # We've examined this path.
 *                 continue
*/
    __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_path_utf8, __pyx_v_self->searched_exact_paths, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1873, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1875
 *             if path_utf8 in self.searched_exact_paths:
 *                 # We've examined this path.
 *                 continue             # <<<<<<<<<<<<<<
 *             if is_inside_any(self.searched_specific_files, path_utf8):
 *                 # We've examined this path.
*/
      goto __pyx_L3_continue;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1873
 *             # no attempt is made to optimise.
 *             path_utf8 = self.search_specific_file_parents.pop()
 *             if path_utf8 in self.searched_exact_paths:             # <<<<<<<<<<<<<<
 *                 # We've examined this path.
 *                 continue
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1876
 *                 # We've examined this path.
 *                 continue
 *             if is_inside_any(self.searched_specific_files, path_utf8):             # <<<<<<<<<<<<<<
 *                 # We've examined this path.
 *                 continue
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_is_inside_any); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_self->searched_specific_files, __pyx_v_path_utf8};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1876, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1876, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1878
 *             if is_inside_any(self.searched_specific_files, path_utf8):
 *                 # We've examined this path.
 *                 continue             # <<<<<<<<<<<<<<
 *             path_entries = self.state._entries_for_path(path_utf8)
 *             # We need either one or two entries. If the path in
*/
      goto __pyx_L3_continue;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1876
 *                 # We've examined this path.
 *                 continue
 *             if is_inside_any(self.searched_specific_files, path_utf8):             # <<<<<<<<<<<<<<
 *                 # We've examined this path.
 *                 continue
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1879
 *                 # We've examined this path.
 *                 continue
 *             path_entries = self.state._entries_for_path(path_utf8)             # <<<<<<<<<<<<<<
 *             # We need either one or two entries. If the path in
 *             # self.target_index has moved (so the entry in source_index is in
*/
    __pyx_t_4 = __pyx_v_self->state;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_path_utf8};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_entries_for_path, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1879, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_path_entries, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1884
 *             # 'ar') then we need to also look for the entry for this path in
 *             # self.source_index, to output the appropriate delete-or-rename.
 *             selected_entries = []             # <<<<<<<<<<<<<<
 *             found_item = False
 *             for candidate_entry in path_entries:
*/
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_selected_entries, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1885
 *             # self.source_index, to output the appropriate delete-or-rename.
 *             selected_entries = []
 *             found_item = False             # <<<<<<<<<<<<<<
 *             for candidate_entry in path_entries:
 *                 # Find entries present in target at this path:
*/
    __pyx_v_found_item = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1886
 *             selected_entries = []
 *             found_item = False
 *             for candidate_entry in path_entries:             # <<<<<<<<<<<<<<
 *                 # Find entries present in target at this path:
 *                 if candidate_entry[1][self.target_index][0] not in (b'a', b'r'):
*/
    if (likely(PyList_CheckExact(__pyx_v_path_entries)) || PyTuple_CheckExact(__pyx_v_path_entries)) {
      __pyx_t_1 = __pyx_v_path_entries; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_path_entries); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1886, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1886, __pyx_L1_error)
            #endif
            if (__pyx_t_6 >= __pyx_temp) break;
          }
          __pyx_t_4 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_6, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_6;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1886, __pyx_L1_error)
            #endif
            if (__pyx_t_6 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6));
          #else
          __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_6);
          #endif
          ++__pyx_t_6;
        }
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1886, __pyx_L1_error)
      } else {
        __pyx_t_4 = __pyx_t_7(__pyx_t_1);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1886, __pyx_L1_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_candidate_entry, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1888
 *             for candidate_entry in path_entries:
 *                 # Find entries present in target at this path:
 *                 if candidate_entry[1][self.target_index][0] not in (b'a', b'r'):             # <<<<<<<<<<<<<<
 *                     found_item = True
 *                     selected_entries.append(candidate_entry)
*/
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_candidate_entry, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_4, __pyx_v_self->target_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_8 = (__Pyx_PyBytes_Equals(__pyx_t_4, __pyx_mstate_global->__pyx_n_b_a, Py_NE)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1888, __pyx_L1_error)
      if (__pyx_t_8) {
      } else {
        __pyx_t_2 = __pyx_t_8;
        goto __pyx_L10_bool_binop_done;
      }
      __pyx_t_8 = (__Pyx_PyBytes_Equals(__pyx_t_4, __pyx_mstate_global->__pyx_n_b_r, Py_NE)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1888, __pyx_L1_error)
      __pyx_t_2 = __pyx_t_8;
      __pyx_L10_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = __pyx_t_2;
      if (__pyx_t_8) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1889
 *                 # Find entries present in target at this path:
 *                 if candidate_entry[1][self.target_index][0] not in (b'a', b'r'):
 *                     found_item = True             # <<<<<<<<<<<<<<
 *                     selected_entries.append(candidate_entry)
 *                 # Find entries present in source at this path:
*/
        __pyx_v_found_item = 1;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1890
 *                 if candidate_entry[1][self.target_index][0] not in (b'a', b'r'):
 *                     found_item = True
 *                     selected_entries.append(candidate_entry)             # <<<<<<<<<<<<<<
 *                 # Find entries present in source at this path:
 *                 elif (self.source_index is not None and
*/
        __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_selected_entries, __pyx_v_candidate_entry); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 1890, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1888
 *             for candidate_entry in path_entries:
 *                 # Find entries present in target at this path:
 *                 if candidate_entry[1][self.target_index][0] not in (b'a', b'r'):             # <<<<<<<<<<<<<<
 *                     found_item = True
 *                     selected_entries.append(candidate_entry)
*/
        goto __pyx_L9;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1892
 *                     selected_entries.append(candidate_entry)
 *                 # Find entries present in source at this path:
 *                 elif (self.source_index is not None and             # <<<<<<<<<<<<<<
 *                     candidate_entry[1][self.source_index][0] not in (b'a', b'r')):
 *                     found_item = True
*/
      __pyx_t_4 = __Pyx_PyLong_From_int(__pyx_v_self->source_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1892, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = (__pyx_t_4 != Py_None);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_2) {
      } else {
        __pyx_t_8 = __pyx_t_2;
        goto __pyx_L12_bool_binop_done;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1893
 *                 # Find entries present in source at this path:
 *                 elif (self.source_index is not None and
 *                     candidate_entry[1][self.source_index][0] not in (b'a', b'r')):             # <<<<<<<<<<<<<<
 *                     found_item = True
 *                     if candidate_entry[1][self.target_index][0] == b'a':
*/
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_candidate_entry, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_4, __pyx_v_self->source_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_10 = (__Pyx_PyBytes_Equals(__pyx_t_4, __pyx_mstate_global->__pyx_n_b_a, Py_NE)); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 1893, __pyx_L1_error)
      if (__pyx_t_10) {
      } else {
        __pyx_t_2 = __pyx_t_10;
        goto __pyx_L14_bool_binop_done;
      }
      __pyx_t_10 = (__Pyx_PyBytes_Equals(__pyx_t_4, __pyx_mstate_global->__pyx_n_b_r, Py_NE)); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 1893, __pyx_L1_error)
      __pyx_t_2 = __pyx_t_10;
      __pyx_L14_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_10 = __pyx_t_2;
      __pyx_t_8 = __pyx_t_10;
      __pyx_L12_bool_binop_done:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1892
 *                     selected_entries.append(candidate_entry)
 *                 # Find entries present in source at this path:
 *                 elif (self.source_index is not None and             # <<<<<<<<<<<<<<
 *                     candidate_entry[1][self.source_index][0] not in (b'a', b'r')):
 *                     found_item = True
*/
      if (__pyx_t_8) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1894
 *                 elif (self.source_index is not None and
 *                     candidate_entry[1][self.source_index][0] not in (b'a', b'r')):
 *                     found_item = True             # <<<<<<<<<<<<<<
 *                     if candidate_entry[1][self.target_index][0] == b'a':
 *                         # Deleted, emit it here.
*/
        __pyx_v_found_item = 1;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1895
 *                     candidate_entry[1][self.source_index][0] not in (b'a', b'r')):
 *                     found_item = True
 *                     if candidate_entry[1][self.target_index][0] == b'a':             # <<<<<<<<<<<<<<
 *                         # Deleted, emit it here.
 *                         selected_entries.append(candidate_entry)
*/
        __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_candidate_entry, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1895, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_4, __pyx_v_self->target_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1895, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1895, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_8 = (__Pyx_PyBytes_Equals(__pyx_t_4, __pyx_mstate_global->__pyx_n_b_a, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1895, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (__pyx_t_8) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1897
 *                     if candidate_entry[1][self.target_index][0] == b'a':
 *                         # Deleted, emit it here.
 *                         selected_entries.append(candidate_entry)             # <<<<<<<<<<<<<<
 *                     else:
 *                         # renamed, emit it when we process the directory it
*/
          __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_selected_entries, __pyx_v_candidate_entry); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 1897, __pyx_L1_error)

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1895
 *                     candidate_entry[1][self.source_index][0] not in (b'a', b'r')):
 *                     found_item = True
 *                     if candidate_entry[1][self.target_index][0] == b'a':             # <<<<<<<<<<<<<<
 *                         # Deleted, emit it here.
 *                         selected_entries.append(candidate_entry)
*/
          goto __pyx_L16;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1901
 *                         # renamed, emit it when we process the directory it
 *                         # ended up at.
 *                         self.search_specific_file_parents.add(             # <<<<<<<<<<<<<<
 *                             candidate_entry[1][self.target_index][1])
 *             if not found_item:
*/
        /*else*/ {
          __pyx_t_3 = __pyx_v_self->search_specific_file_parents;
          __Pyx_INCREF(__pyx_t_3);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1902
 *                         # ended up at.
 *                         self.search_specific_file_parents.add(
 *                             candidate_entry[1][self.target_index][1])             # <<<<<<<<<<<<<<
 *             if not found_item:
 *                 raise AssertionError(
*/
          __pyx_t_11 = __Pyx_GetItemInt(__pyx_v_candidate_entry, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1902, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_12 = __Pyx_GetItemInt(__pyx_t_11, __pyx_v_self->target_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1902, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = __Pyx_GetItemInt(__pyx_t_12, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1902, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_5 = 0;
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_11};
            __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1901, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __pyx_L16:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1892
 *                     selected_entries.append(candidate_entry)
 *                 # Find entries present in source at this path:
 *                 elif (self.source_index is not None and             # <<<<<<<<<<<<<<
 *                     candidate_entry[1][self.source_index][0] not in (b'a', b'r')):
 *                     found_item = True
*/
      }
      __pyx_L9:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1886
 *             selected_entries = []
 *             found_item = False
 *             for candidate_entry in path_entries:             # <<<<<<<<<<<<<<
 *                 # Find entries present in target at this path:
 *                 if candidate_entry[1][self.target_index][0] not in (b'a', b'r'):
*/
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1903
 *                         self.search_specific_file_parents.add(
 *                             candidate_entry[1][self.target_index][1])
 *             if not found_item:             # <<<<<<<<<<<<<<
 *                 raise AssertionError(
 *                     "Missing entry for specific path parent %r, %r" % (
*/
    __pyx_t_8 = (!__pyx_v_found_item);
    if (unlikely(__pyx_t_8)) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1904
 *                             candidate_entry[1][self.target_index][1])
 *             if not found_item:
 *                 raise AssertionError(             # <<<<<<<<<<<<<<
 *                     "Missing entry for specific path parent %r, %r" % (
 *                     path_utf8, path_entries))
*/
      __pyx_t_4 = NULL;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1906
 *                 raise AssertionError(
 *                     "Missing entry for specific path parent %r, %r" % (
 *                     path_utf8, path_entries))             # <<<<<<<<<<<<<<
 *             path_info = self._path_info(path_utf8, path_utf8.decode('utf8'))
 *             for entry in selected_entries:
*/
      __pyx_t_11 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_path_utf8), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1906, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_3 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_path_entries), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1906, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_13[0] = __pyx_mstate_global->__pyx_kp_u_Missing_entry_for_specific_path;
      __pyx_t_13[1] = __pyx_t_11;
      __pyx_t_13[2] = __pyx_mstate_global->__pyx_kp_u__6;
      __pyx_t_13[3] = __pyx_t_3;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1905
 *             if not found_item:
 *                 raise AssertionError(
 *                     "Missing entry for specific path parent %r, %r" % (             # <<<<<<<<<<<<<<
 *                     path_utf8, path_entries))
 *             path_info = self._path_info(path_utf8, path_utf8.decode('utf8'))
*/
      __pyx_t_12 = __Pyx_PyUnicode_Join(__pyx_t_13, 4, 39 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_11) + 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3));
      if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1905, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_12};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1904, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 1904, __pyx_L1_error)

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1903
 *                         self.search_specific_file_parents.add(
 *                             candidate_entry[1][self.target_index][1])
 *             if not found_item:             # <<<<<<<<<<<<<<
 *                 raise AssertionError(
 *                     "Missing entry for specific path parent %r, %r" % (
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1907
 *                     "Missing entry for specific path parent %r, %r" % (
 *                     path_utf8, path_entries))
 *             path_info = self._path_info(path_utf8, path_utf8.decode('utf8'))             # <<<<<<<<<<<<<<
 *             for entry in selected_entries:
 *                 if entry[0][2] in self.seen_ids:
*/
    __pyx_t_12 = __pyx_v_path_utf8;
    __Pyx_INCREF(__pyx_t_12);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_mstate_global->__pyx_n_u_utf8};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_decode, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1907, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_t_12 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_path_info(__pyx_v_self, __pyx_v_path_utf8, __pyx_t_1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1907, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_path_info, __pyx_t_12);
    __pyx_t_12 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1908
 *                     path_utf8, path_entries))
 *             path_info = self._path_info(path_utf8, path_utf8.decode('utf8'))
 *             for entry in selected_entries:             # <<<<<<<<<<<<<<
 *                 if entry[0][2] in self.seen_ids:
 *                     continue
*/
    __pyx_t_12 = __pyx_v_selected_entries; __Pyx_INCREF(__pyx_t_12);
    __pyx_t_6 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_12);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1908, __pyx_L1_error)
        #endif
        if (__pyx_t_6 >= __pyx_temp) break;
      }
      __pyx_t_1 = __Pyx_PyList_GetItemRefFast(__pyx_t_12, __pyx_t_6, __Pyx_ReferenceSharing_OwnStrongReference);
      ++__pyx_t_6;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1908, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_entry, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1909
 *             path_info = self._path_info(path_utf8, path_utf8.decode('utf8'))
 *             for entry in selected_entries:
 *                 if entry[0][2] in self.seen_ids:             # <<<<<<<<<<<<<<
 *                     continue
 *                 result, changed = self._process_entry(entry, path_info)
*/
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1909, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_1, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1909, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = (__Pyx_PySequence_ContainsTF(__pyx_t_4, __pyx_v_self->seen_ids, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1909, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_8) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1910
 *             for entry in selected_entries:
 *                 if entry[0][2] in self.seen_ids:
 *                     continue             # <<<<<<<<<<<<<<
 *                 result, changed = self._process_entry(entry, path_info)
 *                 if changed is None:
*/
        goto __pyx_L19_continue;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1909
 *             path_info = self._path_info(path_utf8, path_utf8.decode('utf8'))
 *             for entry in selected_entries:
 *                 if entry[0][2] in self.seen_ids:             # <<<<<<<<<<<<<<
 *                     continue
 *                 result, changed = self._process_entry(entry, path_info)
*/
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1911
 *                 if entry[0][2] in self.seen_ids:
 *                     continue
 *                 result, changed = self._process_entry(entry, path_info)             # <<<<<<<<<<<<<<
 *                 if changed is None:
 *                     raise AssertionError(
*/
      __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_process_entry(__pyx_v_self, __pyx_v_entry, __pyx_v_path_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1911, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
        PyObject* sequence = __pyx_t_4;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1911, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0);
          __Pyx_INCREF(__pyx_t_1);
          __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1);
          __Pyx_INCREF(__pyx_t_3);
        } else {
          __pyx_t_1 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1911, __pyx_L1_error)
          __Pyx_XGOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
          if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1911, __pyx_L1_error)
          __Pyx_XGOTREF(__pyx_t_3);
        }
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1911, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1911, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_11 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1911, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_14 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_11);
        index = 0; __pyx_t_1 = __pyx_t_14(__pyx_t_11); if (unlikely(!__pyx_t_1)) goto __pyx_L22_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        index = 1; __pyx_t_3 = __pyx_t_14(__pyx_t_11); if (unlikely(!__pyx_t_3)) goto __pyx_L22_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_3);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_11), 2) < (0)) __PYX_ERR(0, 1911, __pyx_L1_error)
        __pyx_t_14 = NULL;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L23_unpacking_done;
        __pyx_L22_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_14 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1911, __pyx_L1_error)
        __pyx_L23_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_result, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_changed, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1912
 *                     continue
 *                 result, changed = self._process_entry(entry, path_info)
 *                 if changed is None:             # <<<<<<<<<<<<<<
 *                     raise AssertionError(
 *                         "Got entry<->path mismatch for specific path "
*/
      __pyx_t_8 = (__pyx_v_changed == Py_None);
      if (unlikely(__pyx_t_8)) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1913
 *                 result, changed = self._process_entry(entry, path_info)
 *                 if changed is None:
 *                     raise AssertionError(             # <<<<<<<<<<<<<<
 *                         "Got entry<->path mismatch for specific path "
 *                         "%r entry %r path_info %r " % (
*/
        __pyx_t_3 = NULL;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1916
 *                         "Got entry<->path mismatch for specific path "
 *                         "%r entry %r path_info %r " % (
 *                         path_utf8, entry, path_info))             # <<<<<<<<<<<<<<
 *                 # Only include changes - we're outside the users requested
 *                 # expansion.
*/
        __pyx_t_1 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_path_utf8), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1916, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_entry), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1916, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_15 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_path_info), __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1916, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __pyx_t_16[0] = __pyx_mstate_global->__pyx_kp_u_Got_entry_path_mismatch_for_spec;
        __pyx_t_16[1] = __pyx_t_1;
        __pyx_t_16[2] = __pyx_mstate_global->__pyx_kp_u_entry_3;
        __pyx_t_16[3] = __pyx_t_11;
        __pyx_t_16[4] = __pyx_mstate_global->__pyx_kp_u_path_info;
        __pyx_t_16[5] = __pyx_t_15;
        __pyx_t_16[6] = __pyx_mstate_global->__pyx_kp_u__2;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1914
 *                 if changed is None:
 *                     raise AssertionError(
 *                         "Got entry<->path mismatch for specific path "             # <<<<<<<<<<<<<<
 *                         "%r entry %r path_info %r " % (
 *                         path_utf8, entry, path_info))
*/
        __pyx_t_17 = __Pyx_PyUnicode_Join(__pyx_t_16, 7, 44 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1) + 7 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_11) + 11 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_15) + 1, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15));
        if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1914, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_5 = 1;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_17};
          __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AssertionError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1913, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 1913, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1912
 *                     continue
 *                 result, changed = self._process_entry(entry, path_info)
 *                 if changed is None:             # <<<<<<<<<<<<<<
 *                     raise AssertionError(
 *                         "Got entry<->path mismatch for specific path "
*/
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1919
 *                 # Only include changes - we're outside the users requested
 *                 # expansion.
 *                 if changed:             # <<<<<<<<<<<<<<
 *                     self._gather_result_for_consistency(result)
 *                     if (result.kind[0] == 'directory' and
*/
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1919, __pyx_L1_error)
      if (__pyx_t_8) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1920
 *                 # expansion.
 *                 if changed:
 *                     self._gather_result_for_consistency(result)             # <<<<<<<<<<<<<<
 *                     if (result.kind[0] == 'directory' and
 *                         result.kind[1] != 'directory'):
*/
        __pyx_t_18 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_gather_result_for_consistency(__pyx_v_self, __pyx_v_result); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 1920, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1921
 *                 if changed:
 *                     self._gather_result_for_consistency(result)
 *                     if (result.kind[0] == 'directory' and             # <<<<<<<<<<<<<<
 *                         result.kind[1] != 'directory'):
 *                         # This stopped being a directory, the old children have
*/
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_mstate_global->__pyx_n_u_kind); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1921, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_4, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1921, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_directory, Py_EQ)); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 1921, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        if (__pyx_t_10) {
        } else {
          __pyx_t_8 = __pyx_t_10;
          goto __pyx_L27_bool_binop_done;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1922
 *                     self._gather_result_for_consistency(result)
 *                     if (result.kind[0] == 'directory' and
 *                         result.kind[1] != 'directory'):             # <<<<<<<<<<<<<<
 *                         # This stopped being a directory, the old children have
 *                         # to be included.
*/
        __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_mstate_global->__pyx_n_u_kind); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1922, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_17, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1922, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_directory, Py_NE)); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 1922, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_8 = __pyx_t_10;
        __pyx_L27_bool_binop_done:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1921
 *                 if changed:
 *                     self._gather_result_for_consistency(result)
 *                     if (result.kind[0] == 'directory' and             # <<<<<<<<<<<<<<
 *                         result.kind[1] != 'directory'):
 *                         # This stopped being a directory, the old children have
*/
        if (__pyx_t_8) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1925
 *                         # This stopped being a directory, the old children have
 *                         # to be included.
 *                         if entry[1][self.source_index][0] == b'r':             # <<<<<<<<<<<<<<
 *                             # renamed, take the source path
 *                             entry_path_utf8 = entry[1][self.source_index][1]
*/
          __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1925, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_4, __pyx_v_self->source_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1925, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_17, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1925, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          __pyx_t_8 = (__Pyx_PyBytes_Equals(__pyx_t_4, __pyx_mstate_global->__pyx_n_b_r, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1925, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (__pyx_t_8) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1927
 *                         if entry[1][self.source_index][0] == b'r':
 *                             # renamed, take the source path
 *                             entry_path_utf8 = entry[1][self.source_index][1]             # <<<<<<<<<<<<<<
 *                         else:
 *                             entry_path_utf8 = path_utf8
*/
            __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1927, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_4, __pyx_v_self->source_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1927, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_17, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1927, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_XDECREF_SET(__pyx_v_entry_path_utf8, __pyx_t_4);
            __pyx_t_4 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1925
 *                         # This stopped being a directory, the old children have
 *                         # to be included.
 *                         if entry[1][self.source_index][0] == b'r':             # <<<<<<<<<<<<<<
 *                             # renamed, take the source path
 *                             entry_path_utf8 = entry[1][self.source_index][1]
*/
            goto __pyx_L29;
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1929
 *                             entry_path_utf8 = entry[1][self.source_index][1]
 *                         else:
 *                             entry_path_utf8 = path_utf8             # <<<<<<<<<<<<<<
 *                         initial_key = (entry_path_utf8, b'', b'')
 *                         block_index, _ = self.state._find_block_index_from_key(
*/
          /*else*/ {
            __Pyx_INCREF(__pyx_v_path_utf8);
            __Pyx_XDECREF_SET(__pyx_v_entry_path_utf8, __pyx_v_path_utf8);
          }
          __pyx_L29:;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1930
 *                         else:
 *                             entry_path_utf8 = path_utf8
 *                         initial_key = (entry_path_utf8, b'', b'')             # <<<<<<<<<<<<<<
 *                         block_index, _ = self.state._find_block_index_from_key(
 *                             initial_key)
*/
          __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1930, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_INCREF(__pyx_v_entry_path_utf8);
          __Pyx_GIVEREF(__pyx_v_entry_path_utf8);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_entry_path_utf8) != (0)) __PYX_ERR(0, 1930, __pyx_L1_error);
          __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b__4);
          __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_b__4);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_mstate_global->__pyx_kp_b__4) != (0)) __PYX_ERR(0, 1930, __pyx_L1_error);
          __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_b__4);
          __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_b__4);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_mstate_global->__pyx_kp_b__4) != (0)) __PYX_ERR(0, 1930, __pyx_L1_error);
          __Pyx_XDECREF_SET(__pyx_v_initial_key, ((PyObject*)__pyx_t_4));
          __pyx_t_4 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1931
 *                             entry_path_utf8 = path_utf8
 *                         initial_key = (entry_path_utf8, b'', b'')
 *                         block_index, _ = self.state._find_block_index_from_key(             # <<<<<<<<<<<<<<
 *                             initial_key)
 *                         if block_index == 0:
*/
          __pyx_t_17 = __pyx_v_self->state;
          __Pyx_INCREF(__pyx_t_17);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1932
 *                         initial_key = (entry_path_utf8, b'', b'')
 *                         block_index, _ = self.state._find_block_index_from_key(
 *                             initial_key)             # <<<<<<<<<<<<<<
 *                         if block_index == 0:
 *                             # The children of the root are in block index 1.
*/
          __pyx_t_5 = 0;
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_17, __pyx_v_initial_key};
            __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_find_block_index_from_key, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1931, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
          }
          if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
            PyObject* sequence = __pyx_t_4;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 1931, __pyx_L1_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_17 = PyTuple_GET_ITEM(sequence, 0);
              __Pyx_INCREF(__pyx_t_17);
              __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1);
              __Pyx_INCREF(__pyx_t_3);
            } else {
              __pyx_t_17 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
              if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1931, __pyx_L1_error)
              __Pyx_XGOTREF(__pyx_t_17);
              __pyx_t_3 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
              if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1931, __pyx_L1_error)
              __Pyx_XGOTREF(__pyx_t_3);
            }
            #else
            __pyx_t_17 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1931, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            __pyx_t_3 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1931, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_15 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1931, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_14 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_15);
            index = 0; __pyx_t_17 = __pyx_t_14(__pyx_t_15); if (unlikely(!__pyx_t_17)) goto __pyx_L30_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_17);
            index = 1; __pyx_t_3 = __pyx_t_14(__pyx_t_15); if (unlikely(!__pyx_t_3)) goto __pyx_L30_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_3);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_15), 2) < (0)) __PYX_ERR(0, 1931, __pyx_L1_error)
            __pyx_t_14 = NULL;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            goto __pyx_L31_unpacking_done;
            __pyx_L30_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __pyx_t_14 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 1931, __pyx_L1_error)
            __pyx_L31_unpacking_done:;
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1931
 *                             entry_path_utf8 = path_utf8
 *                         initial_key = (entry_path_utf8, b'', b'')
 *                         block_index, _ = self.state._find_block_index_from_key(             # <<<<<<<<<<<<<<
 *                             initial_key)
 *                         if block_index == 0:
*/
          __Pyx_XDECREF_SET(__pyx_v_block_index, __pyx_t_17);
          __pyx_t_17 = 0;
          __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_3);
          __pyx_t_3 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1933
 *                         block_index, _ = self.state._find_block_index_from_key(
 *                             initial_key)
 *                         if block_index == 0:             # <<<<<<<<<<<<<<
 *                             # The children of the root are in block index 1.
 *                             block_index = block_index + 1
*/
          __pyx_t_8 = (__Pyx_PyLong_BoolEqObjC(__pyx_v_block_index, __pyx_mstate_global->__pyx_int_0, 0, 0)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1933, __pyx_L1_error)
          if (__pyx_t_8) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1935
 *                         if block_index == 0:
 *                             # The children of the root are in block index 1.
 *                             block_index = block_index + 1             # <<<<<<<<<<<<<<
 *                         current_block = None
 *                         if block_index < len(self.state._dirblocks):
*/
            __pyx_t_4 = __Pyx_PyLong_AddObjC(__pyx_v_block_index, __pyx_mstate_global->__pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1935, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF_SET(__pyx_v_block_index, __pyx_t_4);
            __pyx_t_4 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1933
 *                         block_index, _ = self.state._find_block_index_from_key(
 *                             initial_key)
 *                         if block_index == 0:             # <<<<<<<<<<<<<<
 *                             # The children of the root are in block index 1.
 *                             block_index = block_index + 1
*/
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1936
 *                             # The children of the root are in block index 1.
 *                             block_index = block_index + 1
 *                         current_block = None             # <<<<<<<<<<<<<<
 *                         if block_index < len(self.state._dirblocks):
 *                             current_block = self.state._dirblocks[block_index]
*/
          __Pyx_INCREF(Py_None);
          __Pyx_XDECREF_SET(__pyx_v_current_block, Py_None);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1937
 *                             block_index = block_index + 1
 *                         current_block = None
 *                         if block_index < len(self.state._dirblocks):             # <<<<<<<<<<<<<<
 *                             current_block = self.state._dirblocks[block_index]
 *                             if not is_inside(
*/
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_mstate_global->__pyx_n_u_dirblocks_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1937, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_19 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_19 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1937, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyLong_FromSsize_t(__pyx_t_19); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1937, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = PyObject_RichCompare(__pyx_v_block_index, __pyx_t_4, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1937, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1937, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (__pyx_t_8) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1938
 *                         current_block = None
 *                         if block_index < len(self.state._dirblocks):
 *                             current_block = self.state._dirblocks[block_index]             # <<<<<<<<<<<<<<
 *                             if not is_inside(
 *                                 entry_path_utf8, current_block[0]):
*/
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_mstate_global->__pyx_n_u_dirblocks_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1938, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_v_block_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1938, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF_SET(__pyx_v_current_block, __pyx_t_4);
            __pyx_t_4 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1939
 *                         if block_index < len(self.state._dirblocks):
 *                             current_block = self.state._dirblocks[block_index]
 *                             if not is_inside(             # <<<<<<<<<<<<<<
 *                                 entry_path_utf8, current_block[0]):
 *                                 # No entries for this directory at all.
*/
            __pyx_t_3 = NULL;
            __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_is_inside); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1939, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1940
 *                             current_block = self.state._dirblocks[block_index]
 *                             if not is_inside(
 *                                 entry_path_utf8, current_block[0]):             # <<<<<<<<<<<<<<
 *                                 # No entries for this directory at all.
 *                                 current_block = None
*/
            __pyx_t_15 = __Pyx_GetItemInt(__pyx_v_current_block, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1940, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_5 = 1;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_17))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_17);
              assert(__pyx_t_3);
              PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_17);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(__pyx__function);
              __Pyx_DECREF_SET(__pyx_t_17, __pyx__function);
              __pyx_t_5 = 0;
            }
            #endif
            {
              PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_entry_path_utf8, __pyx_t_15};
              __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_17, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1939, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1939
 *                         if block_index < len(self.state._dirblocks):
 *                             current_block = self.state._dirblocks[block_index]
 *                             if not is_inside(             # <<<<<<<<<<<<<<
 *                                 entry_path_utf8, current_block[0]):
 *                                 # No entries for this directory at all.
*/
            __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1939, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_10 = (!__pyx_t_8);
            if (__pyx_t_10) {

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1942
 *                                 entry_path_utf8, current_block[0]):
 *                                 # No entries for this directory at all.
 *                                 current_block = None             # <<<<<<<<<<<<<<
 *                         if current_block is not None:
 *                             for entry in current_block[1]:
*/
              __Pyx_INCREF(Py_None);
              __Pyx_DECREF_SET(__pyx_v_current_block, Py_None);

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1939
 *                         if block_index < len(self.state._dirblocks):
 *                             current_block = self.state._dirblocks[block_index]
 *                             if not is_inside(             # <<<<<<<<<<<<<<
 *                                 entry_path_utf8, current_block[0]):
 *                                 # No entries for this directory at all.
*/
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1937
 *                             block_index = block_index + 1
 *                         current_block = None
 *                         if block_index < len(self.state._dirblocks):             # <<<<<<<<<<<<<<
 *                             current_block = self.state._dirblocks[block_index]
 *                             if not is_inside(
*/
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1943
 *                                 # No entries for this directory at all.
 *                                 current_block = None
 *                         if current_block is not None:             # <<<<<<<<<<<<<<
 *                             for entry in current_block[1]:
 *                                 if entry[1][self.source_index][0] in (b'a', b'r'):
*/
          __pyx_t_10 = (__pyx_v_current_block != Py_None);
          if (__pyx_t_10) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1944
 *                                 current_block = None
 *                         if current_block is not None:
 *                             for entry in current_block[1]:             # <<<<<<<<<<<<<<
 *                                 if entry[1][self.source_index][0] in (b'a', b'r'):
 *                                     # Not in the source tree, so doesn't have to be
*/
            __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_current_block, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1944, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
              __pyx_t_17 = __pyx_t_4; __Pyx_INCREF(__pyx_t_17);
              __pyx_t_19 = 0;
              __pyx_t_7 = NULL;
            } else {
              __pyx_t_19 = -1; __pyx_t_17 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1944, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_17);
              __pyx_t_7 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_17); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1944, __pyx_L1_error)
            }
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            for (;;) {
              if (likely(!__pyx_t_7)) {
                if (likely(PyList_CheckExact(__pyx_t_17))) {
                  {
                    Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_17);
                    #if !CYTHON_ASSUME_SAFE_SIZE
                    if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1944, __pyx_L1_error)
                    #endif
                    if (__pyx_t_19 >= __pyx_temp) break;
                  }
                  __pyx_t_4 = __Pyx_PyList_GetItemRefFast(__pyx_t_17, __pyx_t_19, __Pyx_ReferenceSharing_OwnStrongReference);
                  ++__pyx_t_19;
                } else {
                  {
                    Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_17);
                    #if !CYTHON_ASSUME_SAFE_SIZE
                    if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1944, __pyx_L1_error)
                    #endif
                    if (__pyx_t_19 >= __pyx_temp) break;
                  }
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_4 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_17, __pyx_t_19));
                  #else
                  __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_17, __pyx_t_19);
                  #endif
                  ++__pyx_t_19;
                }
                if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1944, __pyx_L1_error)
              } else {
                __pyx_t_4 = __pyx_t_7(__pyx_t_17);
                if (unlikely(!__pyx_t_4)) {
                  PyObject* exc_type = PyErr_Occurred();
                  if (exc_type) {
                    if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1944, __pyx_L1_error)
                    PyErr_Clear();
                  }
                  break;
                }
              }
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF_SET(__pyx_v_entry, __pyx_t_4);
              __pyx_t_4 = 0;

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1945
 *                         if current_block is not None:
 *                             for entry in current_block[1]:
 *                                 if entry[1][self.source_index][0] in (b'a', b'r'):             # <<<<<<<<<<<<<<
 *                                     # Not in the source tree, so doesn't have to be
 *                                     # included.
*/
              __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1945, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_15 = __Pyx_GetItemInt(__pyx_t_4, __pyx_v_self->source_index, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1945, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_15, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1945, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __pyx_t_8 = (__Pyx_PyBytes_Equals(__pyx_t_4, __pyx_mstate_global->__pyx_n_b_a, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1945, __pyx_L1_error)
              if (!__pyx_t_8) {
              } else {
                __pyx_t_10 = __pyx_t_8;
                goto __pyx_L39_bool_binop_done;
              }
              __pyx_t_8 = (__Pyx_PyBytes_Equals(__pyx_t_4, __pyx_mstate_global->__pyx_n_b_r, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1945, __pyx_L1_error)
              __pyx_t_10 = __pyx_t_8;
              __pyx_L39_bool_binop_done:;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_8 = __pyx_t_10;
              if (__pyx_t_8) {

                /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1948
 *                                     # Not in the source tree, so doesn't have to be
 *                                     # included.
 *                                     continue             # <<<<<<<<<<<<<<
 *                                 # Path of the entry itself.
 *                                 self.search_specific_file_parents.add(
*/
                goto __pyx_L36_continue;

                /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1945
 *                         if current_block is not None:
 *                             for entry in current_block[1]:
 *                                 if entry[1][self.source_index][0] in (b'a', b'r'):             # <<<<<<<<<<<<<<
 *                                     # Not in the source tree, so doesn't have to be
 *                                     # included.
*/
              }

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1950
 *                                     continue
 *                                 # Path of the entry itself.
 *                                 self.search_specific_file_parents.add(             # <<<<<<<<<<<<<<
 *                                     self.pathjoin(*entry[0][:2]))
 *                 if changed or self.include_unchanged:
*/
              __pyx_t_15 = __pyx_v_self->search_specific_file_parents;
              __Pyx_INCREF(__pyx_t_15);

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1951
 *                                 # Path of the entry itself.
 *                                 self.search_specific_file_parents.add(
 *                                     self.pathjoin(*entry[0][:2]))             # <<<<<<<<<<<<<<
 *                 if changed or self.include_unchanged:
 *                     results.append((result, changed))
*/
              __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1951, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_11 = __Pyx_PyObject_GetSlice(__pyx_t_3, 0, 2, NULL, NULL, &__pyx_mstate_global->__pyx_slice[1], 0, 1, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1951, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1951, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __pyx_t_11 = __Pyx_PyObject_Call(__pyx_v_self->pathjoin, __pyx_t_3, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1951, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_5 = 0;
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_15, __pyx_t_11};
                __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
                __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1950, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_4);
              }
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1944
 *                                 current_block = None
 *                         if current_block is not None:
 *                             for entry in current_block[1]:             # <<<<<<<<<<<<<<
 *                                 if entry[1][self.source_index][0] in (b'a', b'r'):
 *                                     # Not in the source tree, so doesn't have to be
*/
              __pyx_L36_continue:;
            }
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1943
 *                                 # No entries for this directory at all.
 *                                 current_block = None
 *                         if current_block is not None:             # <<<<<<<<<<<<<<
 *                             for entry in current_block[1]:
 *                                 if entry[1][self.source_index][0] in (b'a', b'r'):
*/
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1921
 *                 if changed:
 *                     self._gather_result_for_consistency(result)
 *                     if (result.kind[0] == 'directory' and             # <<<<<<<<<<<<<<
 *                         result.kind[1] != 'directory'):
 *                         # This stopped being a directory, the old children have
*/
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1919
 *                 # Only include changes - we're outside the users requested
 *                 # expansion.
 *                 if changed:             # <<<<<<<<<<<<<<
 *                     self._gather_result_for_consistency(result)
 *                     if (result.kind[0] == 'directory' and
*/
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1952
 *                                 self.search_specific_file_parents.add(
 *                                     self.pathjoin(*entry[0][:2]))
 *                 if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                     results.append((result, changed))
 *             self.searched_exact_paths.add(path_utf8)
*/
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 1952, __pyx_L1_error)
      if (!__pyx_t_10) {
      } else {
        __pyx_t_8 = __pyx_t_10;
        goto __pyx_L43_bool_binop_done;
      }
      __pyx_t_10 = (__pyx_v_self->include_unchanged != 0);
      __pyx_t_8 = __pyx_t_10;
      __pyx_L43_bool_binop_done:;
      if (__pyx_t_8) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1953
 *                                     self.pathjoin(*entry[0][:2]))
 *                 if changed or self.include_unchanged:
 *                     results.append((result, changed))             # <<<<<<<<<<<<<<
 *             self.searched_exact_paths.add(path_utf8)
 *         return results
*/
        __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1953, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        __Pyx_INCREF(__pyx_v_result);
        __Pyx_GIVEREF(__pyx_v_result);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_v_result) != (0)) __PYX_ERR(0, 1953, __pyx_L1_error);
        __Pyx_INCREF(__pyx_v_changed);
        __Pyx_GIVEREF(__pyx_v_changed);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_v_changed) != (0)) __PYX_ERR(0, 1953, __pyx_L1_error);
        __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_results, __pyx_t_17); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 1953, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1952
 *                                 self.search_specific_file_parents.add(
 *                                     self.pathjoin(*entry[0][:2]))
 *                 if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                     results.append((result, changed))
 *             self.searched_exact_paths.add(path_utf8)
*/
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1908
 *                     path_utf8, path_entries))
 *             path_info = self._path_info(path_utf8, path_utf8.decode('utf8'))
 *             for entry in selected_entries:             # <<<<<<<<<<<<<<
 *                 if entry[0][2] in self.seen_ids:
 *                     continue
*/
      __pyx_L19_continue:;
    }
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1954
 *                 if changed or self.include_unchanged:
 *                     results.append((result, changed))
 *             self.searched_exact_paths.add(path_utf8)             # <<<<<<<<<<<<<<
 *         return results
 * 
*/
    __pyx_t_17 = __pyx_v_self->searched_exact_paths;
    __Pyx_INCREF(__pyx_t_17);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_17, __pyx_v_path_utf8};
      __pyx_t_12 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
      if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1954, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
    }
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_L3_continue:;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1955
 *                     results.append((result, changed))
 *             self.searched_exact_paths.add(path_utf8)
 *         return results             # <<<<<<<<<<<<<<
 * 
 *     cdef object _path_info(self, utf8_path, unicode_path):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_results);
  __pyx_r = __pyx_v_results;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1862
 *                     self.current_dir_info = None
 * 
 *     cdef object _next_consistent_entries(self):             # <<<<<<<<<<<<<<
 *         """Grabs the next specific file parent case to consider.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._next_consistent_entries", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_results);
  __Pyx_XDECREF(__pyx_v_path_utf8);
  __Pyx_XDECREF(__pyx_v_path_entries);
  __Pyx_XDECREF(__pyx_v_selected_entries);
  __Pyx_XDECREF(__pyx_v_candidate_entry);
  __Pyx_XDECREF(__pyx_v_path_info);
  __Pyx_XDECREF(__pyx_v_entry);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_changed);
  __Pyx_XDECREF(__pyx_v_entry_path_utf8);
  __Pyx_XDECREF(__pyx_v_initial_key);
  __Pyx_XDECREF(__pyx_v_block_index);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_current_block);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1957
 *         return results
 * 
 *     cdef object _path_info(self, utf8_path, unicode_path):             # <<<<<<<<<<<<<<
 *         """Generate path_info for unicode_path.
 * 
*/

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__path_info(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_utf8_path, PyObject *__pyx_v_unicode_path) {
  PyObject *__pyx_v_abspath = NULL;
  PyObject *__pyx_v_stat = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_utf8_basename = NULL;
  PyObject *__pyx_v_dir_info = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_path_info", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1962
 *         :return: None if unicode_path does not exist, or a path_info tuple.
 *         """
 *         abspath = self.tree.abspath(unicode_path)             # <<<<<<<<<<<<<<
 *         try:
 *             stat = os.lstat(abspath)
*/
  __pyx_t_2 = __pyx_v_self->tree;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_unicode_path};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_abspath, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1962, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_abspath = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1963
 *         """
 *         abspath = self.tree.abspath(unicode_path)
 *         try:             # <<<<<<<<<<<<<<
 *             stat = os.lstat(abspath)
 *         except OSError, e:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1964
 *         abspath = self.tree.abspath(unicode_path)
 *         try:
 *             stat = os.lstat(abspath)             # <<<<<<<<<<<<<<
 *         except OSError, e:
 *             if e.errno == errno.ENOENT:
*/
      __pyx_t_2 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_os); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1964, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_lstat); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1964, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_3 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_8);
        assert(__pyx_t_2);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_8, __pyx__function);
        __pyx_t_3 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_abspath};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_8, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1964, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __pyx_v_stat = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1963
 *         """
 *         abspath = self.tree.abspath(unicode_path)
 *         try:             # <<<<<<<<<<<<<<
 *             stat = os.lstat(abspath)
 *         except OSError, e:
*/
    }
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1965
 *         try:
 *             stat = os.lstat(abspath)
 *         except OSError, e:             # <<<<<<<<<<<<<<
 *             if e.errno == errno.ENOENT:
 *                 # the path does not exist.
*/
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_OSError))));
    if (__pyx_t_9) {
      __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._path_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_8, &__pyx_t_2) < 0) __PYX_ERR(0, 1965, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __pyx_v_e = __pyx_t_8;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1966
 *             stat = os.lstat(abspath)
 *         except OSError, e:
 *             if e.errno == errno.ENOENT:             # <<<<<<<<<<<<<<
 *                 # the path does not exist.
 *                 return None
*/
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_mstate_global->__pyx_n_u_errno); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1966, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_errno); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1966, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_ENOENT); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1966, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = PyObject_RichCompare(__pyx_t_7, __pyx_t_11, Py_EQ); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1966, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 1966, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (likely(__pyx_t_12)) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1968
 *             if e.errno == errno.ENOENT:
 *                 # the path does not exist.
 *                 return None             # <<<<<<<<<<<<<<
 *             else:
 *                 raise
*/
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L6_except_return;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1966
 *             stat = os.lstat(abspath)
 *         except OSError, e:
 *             if e.errno == errno.ENOENT:             # <<<<<<<<<<<<<<
 *                 # the path does not exist.
 *                 return None
*/
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1970
 *                 return None
 *             else:
 *                 raise             # <<<<<<<<<<<<<<
 *         utf8_basename = utf8_path.rsplit(b'/', 1)[-1]
 *         dir_info = (utf8_path, utf8_basename,
*/
      /*else*/ {
        __Pyx_GIVEREF(__pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_2);
        __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_8, __pyx_t_2);
        __pyx_t_1 = 0;  __pyx_t_8 = 0;  __pyx_t_2 = 0; 
        __PYX_ERR(0, 1970, __pyx_L5_except_error)
      }
    }
    goto __pyx_L5_except_error;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1963
 *         """
 *         abspath = self.tree.abspath(unicode_path)
 *         try:             # <<<<<<<<<<<<<<
 *             stat = os.lstat(abspath)
 *         except OSError, e:
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1971
 *             else:
 *                 raise
 *         utf8_basename = utf8_path.rsplit(b'/', 1)[-1]             # <<<<<<<<<<<<<<
 *         dir_info = (utf8_path, utf8_basename,
 *             osutils.file_kind_from_stat_mode(stat.st_mode), stat,
*/
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_utf8_path, __pyx_mstate_global->__pyx_n_u_rsplit); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[8], NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_8, -1L, long, 1, __Pyx_PyLong_From_long, 0, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_utf8_basename = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1973
 *         utf8_basename = utf8_path.rsplit(b'/', 1)[-1]
 *         dir_info = (utf8_path, utf8_basename,
 *             osutils.file_kind_from_stat_mode(stat.st_mode), stat,             # <<<<<<<<<<<<<<
 *             abspath)
 *         if dir_info[2] == 'directory':
*/
  __pyx_t_8 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_osutils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_file_kind_from_stat_mode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat, __pyx_mstate_global->__pyx_n_u_st_mode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_10);
    assert(__pyx_t_8);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_10);
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_10, __pyx__function);
    __pyx_t_3 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_1};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_10, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1973, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1972
 *                 raise
 *         utf8_basename = utf8_path.rsplit(b'/', 1)[-1]
 *         dir_info = (utf8_path, utf8_basename,             # <<<<<<<<<<<<<<
 *             osutils.file_kind_from_stat_mode(stat.st_mode), stat,
 *             abspath)
*/
  __pyx_t_10 = PyTuple_New(5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_v_utf8_path);
  __Pyx_GIVEREF(__pyx_v_utf8_path);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_utf8_path) != (0)) __PYX_ERR(0, 1972, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_utf8_basename);
  __Pyx_GIVEREF(__pyx_v_utf8_basename);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_v_utf8_basename) != (0)) __PYX_ERR(0, 1972, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_2) != (0)) __PYX_ERR(0, 1972, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_stat);
  __Pyx_GIVEREF(__pyx_v_stat);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_v_stat) != (0)) __PYX_ERR(0, 1972, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_abspath);
  __Pyx_GIVEREF(__pyx_v_abspath);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 4, __pyx_v_abspath) != (0)) __PYX_ERR(0, 1972, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_v_dir_info = ((PyObject*)__pyx_t_10);
  __pyx_t_10 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1975
 *             osutils.file_kind_from_stat_mode(stat.st_mode), stat,
 *             abspath)
 *         if dir_info[2] == 'directory':             # <<<<<<<<<<<<<<
 *             if self.tree._directory_is_tree_reference(
 *                 unicode_path):
*/
  __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v_dir_info, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_12 = (__Pyx_PyUnicode_Equals(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_directory, Py_EQ)); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 1975, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (__pyx_t_12) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1976
 *             abspath)
 *         if dir_info[2] == 'directory':
 *             if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *                 unicode_path):
 *                 self.root_dir_info = self.root_dir_info[:2] + \
*/
    __pyx_t_2 = __pyx_v_self->tree;
    __Pyx_INCREF(__pyx_t_2);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1977
 *         if dir_info[2] == 'directory':
 *             if self.tree._directory_is_tree_reference(
 *                 unicode_path):             # <<<<<<<<<<<<<<
 *                 self.root_dir_info = self.root_dir_info[:2] + \
 *                     ('tree-reference',) + self.root_dir_info[3:]
*/
    __pyx_t_3 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_unicode_path};
      __pyx_t_10 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_directory_is_tree_reference, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1976, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1976
 *             abspath)
 *         if dir_info[2] == 'directory':
 *             if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *                 unicode_path):
 *                 self.root_dir_info = self.root_dir_info[:2] + \
*/
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 1976, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (__pyx_t_12) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1978
 *             if self.tree._directory_is_tree_reference(
 *                 unicode_path):
 *                 self.root_dir_info = self.root_dir_info[:2] + \             # <<<<<<<<<<<<<<
 *                     ('tree-reference',) + self.root_dir_info[3:]
 *         return dir_info
*/
      __pyx_t_10 = __Pyx_PyObject_GetSlice(__pyx_v_self->root_dir_info, 0, 2, NULL, NULL, &__pyx_mstate_global->__pyx_slice[1], 0, 1, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1978, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1979
 *                 unicode_path):
 *                 self.root_dir_info = self.root_dir_info[:2] + \
 *                     ('tree-reference',) + self.root_dir_info[3:]             # <<<<<<<<<<<<<<
 *         return dir_info
*/
      __pyx_t_2 = PyNumber_Add(__pyx_t_10, __pyx_mstate_global->__pyx_tuple[5]); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1978, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_GetSlice(__pyx_v_self->root_dir_info, 3, 0, NULL, NULL, &__pyx_mstate_global->__pyx_slice[2], 1, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1979, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1979, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1978
 *             if self.tree._directory_is_tree_reference(
 *                 unicode_path):
 *                 self.root_dir_info = self.root_dir_info[:2] + \             # <<<<<<<<<<<<<<
 *                     ('tree-reference',) + self.root_dir_info[3:]
 *         return dir_info
*/
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_v_self->root_dir_info);
      __Pyx_DECREF(__pyx_v_self->root_dir_info);
      __pyx_v_self->root_dir_info = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1976
 *             abspath)
 *         if dir_info[2] == 'directory':
 *             if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *                 unicode_path):
 *                 self.root_dir_info = self.root_dir_info[:2] + \
*/
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1975
 *             osutils.file_kind_from_stat_mode(stat.st_mode), stat,
 *             abspath)
 *         if dir_info[2] == 'directory':             # <<<<<<<<<<<<<<
 *             if self.tree._directory_is_tree_reference(
 *                 unicode_path):
*/
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1980
 *                 self.root_dir_info = self.root_dir_info[:2] + \
 *                     ('tree-reference',) + self.root_dir_info[3:]
 *         return dir_info             # <<<<<<<<<<<<<<
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_dir_info);
  __pyx_r = __pyx_v_dir_info;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1957
 *         return results
 * 
 *     cdef object _path_info(self, utf8_path, unicode_path):             # <<<<<<<<<<<<<<
 *         """Generate path_info for unicode_path.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._path_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_abspath);
  __Pyx_XDECREF(__pyx_v_stat);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_utf8_basename);
  __Pyx_XDECREF(__pyx_v_dir_info);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":978
 *     cdef object use_filesystem_for_exec
 *     cdef object utf8_decode
 *     cdef readonly object searched_specific_files             # <<<<<<<<<<<<<<
 *     cdef readonly object searched_exact_paths
 *     cdef object search_specific_files
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_23searched_specific_files_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_23searched_specific_files_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_23searched_specific_files___get__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_23searched_specific_files___get__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->searched_specific_files);
  __pyx_r = __pyx_v_self->searched_specific_files;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":979
 *     cdef object utf8_decode
 *     cdef readonly object searched_specific_files
 *     cdef readonly object searched_exact_paths             # <<<<<<<<<<<<<<
 *     cdef object search_specific_files
 *     # The parents up to the root of the paths we are searching.
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_20searched_exact_paths_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_20searched_exact_paths_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_20searched_exact_paths___get__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_20searched_exact_paths___get__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->searched_exact_paths);
  __pyx_r = __pyx_v_self->searched_exact_paths;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_9__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_8__reduce_cython__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_8__reduce_cython__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._pending_consistent_entries, self.bisect_left, self.block_index, self.current_block, self.current_block_list, self.current_block_pos, self.current_dir_info, self.current_dir_list, self.current_root, self.current_root_unicode, self.dir_iterator, self.doing_consistency_expansion, self.fstat, self.include_unchanged, self.last_source_parent, self.last_target_parent, self.new_dirname_to_file_id, self.old_dirname_to_file_id, self.partial, self.path_index, self.pathjoin, self.root_abspath, self.root_dir_info, self.root_entries, self.root_entries_len, self.root_entries_pos, self.search_specific_file_parents, self.search_specific_files, self.searched_exact_paths, self.searched_specific_files, self.seen_ids, self.sha_file, self.source_index, self.state, self.target_index, self.tree, self.use_filesystem_for_exec, self.utf8_decode, self.want_unversioned)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->block_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_v_self->current_block_pos); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyLong_From_int(__pyx_v_self->doing_consistency_expansion); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyLong_From_int(__pyx_v_self->include_unchanged); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyLong_From_int(__pyx_v_self->partial); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyLong_From_int(__pyx_v_self->path_index); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyLong_From_int(__pyx_v_self->root_entries_len); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyLong_From_int(__pyx_v_self->root_entries_pos); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyLong_From_int(__pyx_v_self->source_index); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyLong_From_int(__pyx_v_self->target_index); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyLong_From_int(__pyx_v_self->want_unversioned); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = PyTuple_New(39); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_INCREF(__pyx_v_self->_pending_consistent_entries);
  __Pyx_GIVEREF(__pyx_v_self->_pending_consistent_entries);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v_self->_pending_consistent_entries) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->bisect_left);
  __Pyx_GIVEREF(__pyx_v_self->bisect_left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_v_self->bisect_left) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_t_1) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->current_block);
  __Pyx_GIVEREF(__pyx_v_self->current_block);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 3, __pyx_v_self->current_block) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->current_block_list);
  __Pyx_GIVEREF(__pyx_v_self->current_block_list);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 4, __pyx_v_self->current_block_list) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 5, __pyx_t_2) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->current_dir_info);
  __Pyx_GIVEREF(__pyx_v_self->current_dir_info);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 6, __pyx_v_self->current_dir_info) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->current_dir_list);
  __Pyx_GIVEREF(__pyx_v_self->current_dir_list);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 7, __pyx_v_self->current_dir_list) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->current_root);
  __Pyx_GIVEREF(__pyx_v_self->current_root);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 8, __pyx_v_self->current_root) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->current_root_unicode);
  __Pyx_GIVEREF(__pyx_v_self->current_root_unicode);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 9, __pyx_v_self->current_root_unicode) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->dir_iterator);
  __Pyx_GIVEREF(__pyx_v_self->dir_iterator);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 10, __pyx_v_self->dir_iterator) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 11, __pyx_t_3) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->fstat);
  __Pyx_GIVEREF(__pyx_v_self->fstat);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 12, __pyx_v_self->fstat) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 13, __pyx_t_4) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->last_source_parent);
  __Pyx_GIVEREF(__pyx_v_self->last_source_parent);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 14, __pyx_v_self->last_source_parent) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->last_target_parent);
  __Pyx_GIVEREF(__pyx_v_self->last_target_parent);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 15, __pyx_v_self->last_target_parent) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->new_dirname_to_file_id);
  __Pyx_GIVEREF(__pyx_v_self->new_dirname_to_file_id);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 16, __pyx_v_self->new_dirname_to_file_id) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->old_dirname_to_file_id);
  __Pyx_GIVEREF(__pyx_v_self->old_dirname_to_file_id);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 17, __pyx_v_self->old_dirname_to_file_id) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 18, __pyx_t_5) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 19, __pyx_t_6) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->pathjoin);
  __Pyx_GIVEREF(__pyx_v_self->pathjoin);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 20, __pyx_v_self->pathjoin) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->root_abspath);
  __Pyx_GIVEREF(__pyx_v_self->root_abspath);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 21, __pyx_v_self->root_abspath) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->root_dir_info);
  __Pyx_GIVEREF(__pyx_v_self->root_dir_info);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 22, __pyx_v_self->root_dir_info) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->root_entries);
  __Pyx_GIVEREF(__pyx_v_self->root_entries);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 23, __pyx_v_self->root_entries) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 24, __pyx_t_7) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 25, __pyx_t_8) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->search_specific_file_parents);
  __Pyx_GIVEREF(__pyx_v_self->search_specific_file_parents);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 26, __pyx_v_self->search_specific_file_parents) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->search_specific_files);
  __Pyx_GIVEREF(__pyx_v_self->search_specific_files);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 27, __pyx_v_self->search_specific_files) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->searched_exact_paths);
  __Pyx_GIVEREF(__pyx_v_self->searched_exact_paths);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 28, __pyx_v_self->searched_exact_paths) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->searched_specific_files);
  __Pyx_GIVEREF(__pyx_v_self->searched_specific_files);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 29, __pyx_v_self->searched_specific_files) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->seen_ids);
  __Pyx_GIVEREF(__pyx_v_self->seen_ids);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 30, __pyx_v_self->seen_ids) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->sha_file);
  __Pyx_GIVEREF(__pyx_v_self->sha_file);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 31, __pyx_v_self->sha_file) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_9);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 32, __pyx_t_9) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->state);
  __Pyx_GIVEREF(__pyx_v_self->state);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 33, __pyx_v_self->state) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_10);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 34, __pyx_t_10) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->tree);
  __Pyx_GIVEREF(__pyx_v_self->tree);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 35, __pyx_v_self->tree) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->use_filesystem_for_exec);
  __Pyx_GIVEREF(__pyx_v_self->use_filesystem_for_exec);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 36, __pyx_v_self->use_filesystem_for_exec) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->utf8_decode);
  __Pyx_GIVEREF(__pyx_v_self->utf8_decode);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 37, __pyx_v_self->utf8_decode) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_11);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 38, __pyx_t_11) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_12);
  __pyx_t_12 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._pending_consistent_entries, self.bisect_left, self.block_index, self.current_block, self.current_block_list, self.current_block_pos, self.current_dir_info, self.current_dir_list, self.current_root, self.current_root_unicode, self.dir_iterator, self.doing_consistency_expansion, self.fstat, self.include_unchanged, self.last_source_parent, self.last_target_parent, self.new_dirname_to_file_id, self.old_dirname_to_file_id, self.partial, self.path_index, self.pathjoin, self.root_abspath, self.root_dir_info, self.root_entries, self.root_entries_len, self.root_entries_pos, self.search_specific_file_parents, self.search_specific_files, self.searched_exact_paths, self.searched_specific_files, self.seen_ids, self.sha_file, self.source_index, self.state, self.target_index, self.tree, self.use_filesystem_for_exec, self.utf8_decode, self.want_unversioned)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_12 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_v__dict = __pyx_t_12;
  __pyx_t_12 = 0;

  /* "(tree fragment)":7
 *     state = (self._pending_consistent_entries, self.bisect_left, self.block_index, self.current_block, self.current_block_list, self.current_block_pos, self.current_dir_info, self.current_dir_list, self.current_root, self.current_root_unicode, self.dir_iterator, self.doing_consistency_expansion, self.fstat, self.include_unchanged, self.last_source_parent, self.last_target_parent, self.new_dirname_to_file_id, self.old_dirname_to_file_id, self.partial, self.path_index, self.pathjoin, self.root_abspath, self.root_dir_info, self.root_entries, self.root_entries_len, self.root_entries_pos, self.search_specific_file_parents, self.search_specific_files, self.searched_exact_paths, self.searched_specific_files, self.seen_ids, self.sha_file, self.source_index, self.state, self.target_index, self.tree, self.use_filesystem_for_exec, self.utf8_decode, self.want_unversioned)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_14 = (__pyx_v__dict != Py_None);
  if (__pyx_t_14) {
  } else {
    __pyx_t_13 = __pyx_t_14;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_14 < 0))) __PYX_ERR(2, 7, __pyx_L1_error)
  __pyx_t_13 = __pyx_t_14;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_13) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_12 = PyTuple_New(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v__dict) != (0)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_11 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_11));
    __pyx_t_11 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._pending_consistent_entries is not None or self.bisect_left is not None or self.current_block is not None or self.current_block_list is not None or self.current_dir_info is not None or self.current_dir_list is not None or self.current_root is not None or self.current_root_unicode is not None or self.dir_iterator is not None or self.fstat is not None or self.last_source_parent is not None or self.last_target_parent is not None or self.new_dirname_to_file_id is not None or self.old_dirname_to_file_id is not None or self.pathjoin is not None or self.root_abspath is not None or self.root_dir_info is not None or self.root_entries is not None or self.search_specific_file_parents is not None or self.search_specific_files is not None or self.searched_exact_paths is not None or self.searched_specific_files is not None or self.seen_ids is not None or self.sha_file is not None or self.state is not None or self.tree is not None or self.use_filesystem_for_exec is not None or self.utf8_decode is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._pending_consistent_entries, self.bisect_left, self.block_index, self.current_block, self.current_block_list, self.current_block_pos, self.current_dir_info, self.current_dir_list, self.current_root, self.current_root_unicode, self.dir_iterator, self.doing_consistency_expansion, self.fstat, self.include_unchanged, self.last_source_parent, self.last_target_parent, self.new_dirname_to_file_id, self.old_dirname_to_file_id, self.partial, self.path_index, self.pathjoin, self.root_abspath, self.root_dir_info, self.root_entries, self.root_entries_len, self.root_entries_pos, self.search_specific_file_parents, self.search_specific_files, self.searched_exact_paths, self.searched_specific_files, self.seen_ids, self.sha_file, self.source_index, self.state, self.target_index, self.tree, self.use_filesystem_for_exec, self.utf8_decode, self.want_unversioned)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._pending_consistent_entries is not None or self.bisect_left is not None or self.current_block is not None or self.current_block_list is not None or self.current_dir_info is not None or self.current_dir_list is not None or self.current_root is not None or self.current_root_unicode is not None or self.dir_iterator is not None or self.fstat is not None or self.last_source_parent is not None or self.last_target_parent is not None or self.new_dirname_to_file_id is not None or self.old_dirname_to_file_id is not None or self.pathjoin is not None or self.root_abspath is not None or self.root_dir_info is not None or self.root_entries is not None or self.search_specific_file_parents is not None or self.search_specific_files is not None or self.searched_exact_paths is not None or self.searched_specific_files is not None or self.seen_ids is not None or self.sha_file is not None or self.state is not None or self.tree is not None or self.use_filesystem_for_exec is not None or self.utf8_decode is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x23b8720, None), state
*/
  /*else*/ {
    __pyx_t_14 = (__pyx_v_self->_pending_consistent_entries != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->bisect_left != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->current_block != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->current_block_list != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->current_dir_info != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->current_dir_list != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->current_root != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->current_root_unicode != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->dir_iterator != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->fstat != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->last_source_parent != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->last_target_parent != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->new_dirname_to_file_id != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->old_dirname_to_file_id != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->pathjoin != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->root_abspath != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->root_dir_info != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->root_entries != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->search_specific_file_parents != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->search_specific_files != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->searched_exact_paths != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->searched_specific_files != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->seen_ids != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->sha_file != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->state != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->tree != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->use_filesystem_for_exec != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->utf8_decode != Py_None);
    __pyx_t_13 = __pyx_t_14;
    __pyx_L6_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_13;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._pending_consistent_entries is not None or self.bisect_left is not None or self.current_block is not None or self.current_block_list is not None or self.current_dir_info is not None or self.current_dir_list is not None or self.current_root is not None or self.current_root_unicode is not None or self.dir_iterator is not None or self.fstat is not None or self.last_source_parent is not None or self.last_target_parent is not None or self.new_dirname_to_file_id is not None or self.old_dirname_to_file_id is not None or self.pathjoin is not None or self.root_abspath is not None or self.root_dir_info is not None or self.root_entries is not None or self.search_specific_file_parents is not None or self.search_specific_files is not None or self.searched_exact_paths is not None or self.searched_specific_files is not None or self.seen_ids is not None or self.sha_file is not None or self.state is not None or self.tree is not None or self.use_filesystem_for_exec is not None or self.utf8_decode is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x23b8720, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._pending_consistent_entries is not None or self.bisect_left is not None or self.current_block is not None or self.current_block_list is not None or self.current_dir_info is not None or self.current_dir_list is not None or self.current_root is not None or self.current_root_unicode is not None or self.dir_iterator is not None or self.fstat is not None or self.last_source_parent is not None or self.last_target_parent is not None or self.new_dirname_to_file_id is not None or self.old_dirname_to_file_id is not None or self.pathjoin is not None or self.root_abspath is not None or self.root_dir_info is not None or self.root_entries is not None or self.search_specific_file_parents is not None or self.search_specific_files is not None or self.searched_exact_paths is not None or self.searched_specific_files is not None or self.seen_ids is not None or self.sha_file is not None or self.state is not None or self.tree is not None or self.use_filesystem_for_exec is not None or self.utf8_decode is not None
 *     if use_setstate:
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x23b8720, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x23b8720, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_ProcessEntryC); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = PyTuple_New(3); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_37455648);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_37455648);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_mstate_global->__pyx_int_37455648) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 2, Py_None) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_11);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_11) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_12);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_12) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_v_state) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_11 = 0;
    __pyx_t_12 = 0;
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._pending_consistent_entries is not None or self.bisect_left is not None or self.current_block is not None or self.current_block_list is not None or self.current_dir_info is not None or self.current_dir_list is not None or self.current_root is not None or self.current_root_unicode is not None or self.dir_iterator is not None or self.fstat is not None or self.last_source_parent is not None or self.last_target_parent is not None or self.new_dirname_to_file_id is not None or self.old_dirname_to_file_id is not None or self.pathjoin is not None or self.root_abspath is not None or self.root_dir_info is not None or self.root_entries is not None or self.search_specific_file_parents is not None or self.search_specific_files is not None or self.searched_exact_paths is not None or self.searched_specific_files is not None or self.seen_ids is not None or self.sha_file is not None or self.state is not None or self.tree is not None or self.use_filesystem_for_exec is not None or self.utf8_decode is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x23b8720, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x23b8720, None), state
 *     else:
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x23b8720, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ProcessEntryC__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_ProcessEntryC); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = PyTuple_New(3); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_37455648);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_37455648);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_mstate_global->__pyx_int_37455648) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_v_state) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_10);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_10) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_12);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_12) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_10 = 0;
    __pyx_t_12 = 0;
    __pyx_r = __pyx_t_11;
    __pyx_t_11 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x23b8720, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ProcessEntryC__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_11__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_10__setstate_cython__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_10__setstate_cython__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x23b8720, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ProcessEntryC__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(2, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx___pyx_unpickle_ProcessEntryC__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x23b8720, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ProcessEntryC__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_Reader(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xff2cf55, 0x6e2f153, 0x65fc0dd, b'cur_cstr, end_cstr, next, state, text, text_cstr, text_size')
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_19__pyx_unpickle_Reader(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_19__pyx_unpickle_Reader = {"__pyx_unpickle_Reader", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_19__pyx_unpickle_Reader, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_19__pyx_unpickle_Reader(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Reader (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_Reader", 0) < (0)) __PYX_ERR(2, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Reader", 1, 3, 3, i); __PYX_ERR(2, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(2, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(2, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Reader", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.__pyx_unpickle_Reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(2, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_18__pyx_unpickle_Reader(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_18__pyx_unpickle_Reader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Reader", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle_Reader(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xff2cf55, 0x6e2f153, 0x65fc0dd, b'cur_cstr, end_cstr, next, state, text, text_cstr, text_size')             # <<<<<<<<<<<<<<
 *     __pyx_result = Reader.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0xff2cf55, 0x6e2f153, 0x65fc0dd, __pyx_k_cur_cstr_end_cstr_next_state_tex); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xff2cf55, 0x6e2f153, 0x65fc0dd, b'cur_cstr, end_cstr, next, state, text, text_cstr, text_size')
 *     __pyx_result = Reader.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Reader__set_state(<Reader> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xff2cf55, 0x6e2f153, 0x65fc0dd, b'cur_cstr, end_cstr, next, state, text, text_cstr, text_size')
 *     __pyx_result = Reader.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Reader__set_state(<Reader> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = Reader.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Reader__set_state(<Reader> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Reader__set_state(Reader __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(2, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx___pyx_unpickle_Reader__set_state(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xff2cf55, 0x6e2f153, 0x65fc0dd, b'cur_cstr, end_cstr, next, state, text, text_cstr, text_size')
 *     __pyx_result = Reader.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Reader__set_state(<Reader> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Reader__set_state(<Reader> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Reader__set_state(Reader __pyx_result, __pyx_state: tuple):
 *     __pyx_result.cur_cstr = __pyx_state[0]; __pyx_result.end_cstr = __pyx_state[1]; __pyx_result.next = __pyx_state[2]; __pyx_result.state = __pyx_state[3]; __pyx_result.text = __pyx_state[4]; __pyx_result.text_cstr = __pyx_state[5]; __pyx_result.text_size = __pyx_state[6]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_Reader(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xff2cf55, 0x6e2f153, 0x65fc0dd, b'cur_cstr, end_cstr, next, state, text, text_cstr, text_size')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.__pyx_unpickle_Reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Reader__set_state(<Reader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Reader__set_state(Reader __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result.cur_cstr = __pyx_state[0]; __pyx_result.end_cstr = __pyx_state[1]; __pyx_result.next = __pyx_state[2]; __pyx_result.state = __pyx_state[3]; __pyx_result.text = __pyx_state[4]; __pyx_result.text_cstr = __pyx_state[5]; __pyx_result.text_size = __pyx_state[6]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 7)
*/

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx___pyx_unpickle_Reader__set_state(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Reader__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Reader__set_state(Reader __pyx_result, __pyx_state: tuple):
 *     __pyx_result.cur_cstr = __pyx_state[0]; __pyx_result.end_cstr = __pyx_state[1]; __pyx_result.next = __pyx_state[2]; __pyx_result.state = __pyx_state[3]; __pyx_result.text = __pyx_state[4]; __pyx_result.text_cstr = __pyx_state[5]; __pyx_result.text_size = __pyx_state[6]             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 7)
*/
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_v___pyx_result->cur_cstr = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_v___pyx_result->end_cstr = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_v___pyx_result->next = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->state);
  __Pyx_DECREF(__pyx_v___pyx_result->state);
  __pyx_v___pyx_result->state = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->text);
  __Pyx_DECREF(__pyx_v___pyx_result->text);
  __pyx_v___pyx_result->text = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_v___pyx_result->text_cstr = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->text_size = __pyx_t_3;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Reader__set_state(Reader __pyx_result, __pyx_state: tuple):
 *     __pyx_result.cur_cstr = __pyx_state[0]; __pyx_result.end_cstr = __pyx_state[1]; __pyx_result.next = __pyx_state[2]; __pyx_result.state = __pyx_state[3]; __pyx_result.text = __pyx_state[4]; __pyx_result.text_cstr = __pyx_state[5]; __pyx_result.text_size = __pyx_state[6]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 7)             # <<<<<<<<<<<<<<
*/
  __pyx_t_3 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 7); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle_Reader__set_state(<Reader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Reader__set_state(Reader __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result.cur_cstr = __pyx_state[0]; __pyx_result.end_cstr = __pyx_state[1]; __pyx_result.next = __pyx_state[2]; __pyx_result.state = __pyx_state[3]; __pyx_result.text = __pyx_state[4]; __pyx_result.text_cstr = __pyx_state[5]; __pyx_result.text_size = __pyx_state[6]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 7)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.__pyx_unpickle_Reader__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_ProcessEntryC(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x23b8720, 0xfa409dc, 0x21d99cd, b'_pending_consistent_entries, bisect_left, block_index, current_block, current_block_list, current_block_pos, current_dir_info, current_dir_list, current_root, current_root_unicode, dir_iterator, doing_consistency_expansion, fstat, include_unchanged, last_source_parent, last_target_parent, new_dirname_to_file_id, old_dirname_to_file_id, partial, path_index, pathjoin, root_abspath, root_dir_info, root_entries, root_entries_len, root_entries_pos, search_specific_file_parents, search_specific_files, searched_exact_paths, searched_specific_files, seen_ids, sha_file, source_index, state, target_index, tree, use_filesystem_for_exec, utf8_decode, want_unversioned')
*/

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_21__pyx_unpickle_ProcessEntryC(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_21__pyx_unpickle_ProcessEntryC = {"__pyx_unpickle_ProcessEntryC", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_21__pyx_unpickle_ProcessEntryC, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_21__pyx_unpickle_ProcessEntryC(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_ProcessEntryC (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_ProcessEntryC", 0) < (0)) __PYX_ERR(2, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ProcessEntryC", 1, 3, 3, i); __PYX_ERR(2, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(2, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(2, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ProcessEntryC", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.__pyx_unpickle_ProcessEntryC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(2, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_20__pyx_unpickle_ProcessEntryC(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_20__pyx_unpickle_ProcessEntryC(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ProcessEntryC", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle_ProcessEntryC(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x23b8720, 0xfa409dc, 0x21d99cd, b'_pending_consistent_entries, bisect_left, block_index, current_block, current_block_list, current_block_pos, current_dir_info, current_dir_list, current_root, current_root_unicode, dir_iterator, doing_consistency_expansion, fstat, include_unchanged, last_source_parent, last_target_parent, new_dirname_to_file_id, old_dirname_to_file_id, partial, path_index, pathjoin, root_abspath, root_dir_info, root_entries, root_entries_len, root_entries_pos, search_specific_file_parents, search_specific_files, searched_exact_paths, searched_specific_files, seen_ids, sha_file, source_index, state, target_index, tree, use_filesystem_for_exec, utf8_decode, want_unversioned')             # <<<<<<<<<<<<<<
 *     __pyx_result = ProcessEntryC.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0x23b8720, 0xfa409dc, 0x21d99cd, __pyx_k_pending_consistent_entries_bise); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x23b8720, 0xfa409dc, 0x21d99cd, b'_pending_consistent_entries, bisect_left, block_index, current_block, current_block_list, current_block_pos, current_dir_info, current_dir_list, current_root, current_root_unicode, dir_iterator, doing_consistency_expansion, fstat, include_unchanged, last_source_parent, last_target_parent, new_dirname_to_file_id, old_dirname_to_file_id, partial, path_index, pathjoin, root_abspath, root_dir_info, root_entries, root_entries_len, root_entries_pos, search_specific_file_parents, search_specific_files, searched_exact_paths, searched_specific_files, seen_ids, sha_file, source_index, state, target_index, tree, use_filesystem_for_exec, utf8_decode, want_unversioned')
 *     __pyx_result = ProcessEntryC.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ProcessEntryC__set_state(<ProcessEntryC> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x23b8720, 0xfa409dc, 0x21d99cd, b'_pending_consistent_entries, bisect_left, block_index, current_block, current_block_list, current_block_pos, current_dir_info, current_dir_list, current_root, current_root_unicode, dir_iterator, doing_consistency_expansion, fstat, include_unchanged, last_source_parent, last_target_parent, new_dirname_to_file_id, old_dirname_to_file_id, partial, path_index, pathjoin, root_abspath, root_dir_info, root_entries, root_entries_len, root_entries_pos, search_specific_file_parents, search_specific_files, searched_exact_paths, searched_specific_files, seen_ids, sha_file, source_index, state, target_index, tree, use_filesystem_for_exec, utf8_decode, want_unversioned')
 *     __pyx_result = ProcessEntryC.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ProcessEntryC__set_state(<ProcessEntryC> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = ProcessEntryC.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ProcessEntryC__set_state(<ProcessEntryC> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_ProcessEntryC__set_state(ProcessEntryC __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(2, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx___pyx_unpickle_ProcessEntryC__set_state(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x23b8720, 0xfa409dc, 0x21d99cd, b'_pending_consistent_entries, bisect_left, block_index, current_block, current_block_list, current_block_pos, current_dir_info, current_dir_list, current_root, current_root_unicode, dir_iterator, doing_consistency_expansion, fstat, include_unchanged, last_source_parent, last_target_parent, new_dirname_to_file_id, old_dirname_to_file_id, partial, path_index, pathjoin, root_abspath, root_dir_info, root_entries, root_entries_len, root_entries_pos, search_specific_file_parents, search_specific_files, searched_exact_paths, searched_specific_files, seen_ids, sha_file, source_index, state, target_index, tree, use_filesystem_for_exec, utf8_decode, want_unversioned')
 *     __pyx_result = ProcessEntryC.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ProcessEntryC__set_state(<ProcessEntryC> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ProcessEntryC__set_state(<ProcessEntryC> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_ProcessEntryC__set_state(ProcessEntryC __pyx_result, __pyx_state: tuple):
 *     __pyx_result._pending_consistent_entries = __pyx_state[0]; __pyx_result.bisect_left = __pyx_state[1]; __pyx_result.block_index = __pyx_state[2]; __pyx_result.current_block = __pyx_state[3]; __pyx_result.current_block_list = __pyx_state[4]; __pyx_result.current_block_pos = __pyx_state[5]; __pyx_result.current_dir_info = __pyx_state[6]; __pyx_result.current_dir_list = __pyx_state[7]; __pyx_result.current_root = __pyx_state[8]; __pyx_result.current_root_unicode = __pyx_state[9]; __pyx_result.dir_iterator = __pyx_state[10]; __pyx_result.doing_consistency_expansion = __pyx_state[11]; __pyx_result.fstat = __pyx_state[12]; __pyx_result.include_unchanged = __pyx_state[13]; __pyx_result.last_source_parent = __pyx_state[14]; __pyx_result.last_target_parent = __pyx_state[15]; __pyx_result.new_dirname_to_file_id = __pyx_state[16]; __pyx_result.old_dirname_to_file_id = __pyx_state[17]; __pyx_result.partial = __pyx_state[18]; __pyx_result.path_index = __pyx_state[19]; __pyx_result.pathjoin = __pyx_state[20]; __pyx_result.root_abspath = __pyx_state[21]; __pyx_result.root_dir_info = __pyx_state[22]; __pyx_result.root_entries = __pyx_state[23]; __pyx_result.root_entries_len = __pyx_state[24]; __pyx_result.root_entries_pos = __pyx_state[25]; __pyx_result.search_specific_file_parents = __pyx_state[26]; __pyx_result.search_specific_files = __pyx_state[27]; __pyx_result.searched_exact_paths = __pyx_state[28]; __pyx_result.searched_specific_files = __pyx_state[29]; __pyx_result.seen_ids = __pyx_state[30]; __pyx_result.sha_file = __pyx_state[31]; __pyx_result.source_index = __pyx_state[32]; __pyx_result.state = __pyx_state[33]; __pyx_result.target_index = __pyx_state[34]; __pyx_result.tree = __pyx_state[35]; __pyx_result.use_filesystem_for_exec = __pyx_state[36]; __pyx_result.utf8_decode = __pyx_state[37]; __pyx_result.want_unversioned = __pyx_state[38]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_ProcessEntryC(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x23b8720, 0xfa409dc, 0x21d99cd, b'_pending_consistent_entries, bisect_left, block_index, current_block, current_block_list, current_block_pos, current_dir_info, current_dir_list, current_root, current_root_unicode, dir_iterator, doing_consistency_expansion, fstat, include_unchanged, last_source_parent, last_target_parent, new_dirname_to_file_id, old_dirname_to_file_id, partial, path_index, pathjoin, root_abspath, root_dir_info, root_entries, root_entries_len, root_entries_pos, search_specific_file_parents, search_specific_files, searched_exact_paths, searched_specific_files, seen_ids, sha_file, source_index, state, target_index, tree, use_filesystem_for_exec, utf8_decode, want_unversioned')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.__pyx_unpickle_ProcessEntryC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_ProcessEntryC__set_state(<ProcessEntryC> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ProcessEntryC__set_state(ProcessEntryC __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._pending_consistent_entries = __pyx_state[0]; __pyx_result.bisect_left = __pyx_state[1]; __pyx_result.block_index = __pyx_state[2]; __pyx_result.current_block = __pyx_state[3]; __pyx_result.current_block_list = __pyx_state[4]; __pyx_result.current_block_pos = __pyx_state[5]; __pyx_result.current_dir_info = __pyx_state[6]; __pyx_result.current_dir_list = __pyx_state[7]; __pyx_result.current_root = __pyx_state[8]; __pyx_result.current_root_unicode = __pyx_state[9]; __pyx_result.dir_iterator = __pyx_state[10]; __pyx_result.doing_consistency_expansion = __pyx_state[11]; __pyx_result.fstat = __pyx_state[12]; __pyx_result.include_unchanged = __pyx_state[13]; __pyx_result.last_source_parent = __pyx_state[14]; __pyx_result.last_target_parent = __pyx_state[15]; __pyx_result.new_dirname_to_file_id = __pyx_state[16]; __pyx_result.old_dirname_to_file_id = __pyx_state[17]; __pyx_result.partial = __pyx_state[18]; __pyx_result.path_index = __pyx_state[19]; __pyx_result.pathjoin = __pyx_state[20]; __pyx_result.root_abspath = __pyx_state[21]; __pyx_result.root_dir_info = __pyx_state[22]; __pyx_result.root_entries = __pyx_state[23]; __pyx_result.root_entries_len = __pyx_state[24]; __pyx_result.root_entries_pos = __pyx_state[25]; __pyx_result.search_specific_file_parents = __pyx_state[26]; __pyx_result.search_specific_files = __pyx_state[27]; __pyx_result.searched_exact_paths = __pyx_state[28]; __pyx_result.searched_specific_files = __pyx_state[29]; __pyx_result.seen_ids = __pyx_state[30]; __pyx_result.sha_file = __pyx_state[31]; __pyx_result.source_index = __pyx_state[32]; __pyx_result.state = __pyx_state[33]; __pyx_result.target_index = __pyx_state[34]; __pyx_result.tree = __pyx_state[35]; __pyx_result.use_filesystem_for_exec = __pyx_state[36]; __pyx_result.utf8_decode = __pyx_state[37]; __pyx_result.want_unversioned = __pyx_state[38]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 39)
*/

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx___pyx_unpickle_ProcessEntryC__set_state(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ProcessEntryC__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_ProcessEntryC__set_state(ProcessEntryC __pyx_result, __pyx_state: tuple):
 *     __pyx_result._pending_consistent_entries = __pyx_state[0]; __pyx_result.bisect_left = __pyx_state[1]; __pyx_result.block_index = __pyx_state[2]; __pyx_result.current_block = __pyx_state[3]; __pyx_result.current_block_list = __pyx_state[4]; __pyx_result.current_block_pos = __pyx_state[5]; __pyx_result.current_dir_info = __pyx_state[6]; __pyx_result.current_dir_list = __pyx_state[7]; __pyx_result.current_root = __pyx_state[8]; __pyx_result.current_root_unicode = __pyx_state[9]; __pyx_result.dir_iterator = __pyx_state[10]; __pyx_result.doing_consistency_expansion = __pyx_state[11]; __pyx_result.fstat = __pyx_state[12]; __pyx_result.include_unchanged = __pyx_state[13]; __pyx_result.last_source_parent = __pyx_state[14]; __pyx_result.last_target_parent = __pyx_state[15]; __pyx_result.new_dirname_to_file_id = __pyx_state[16]; __pyx_result.old_dirname_to_file_id = __pyx_state[17]; __pyx_result.partial = __pyx_state[18]; __pyx_result.path_index = __pyx_state[19]; __pyx_result.pathjoin = __pyx_state[20]; __pyx_result.root_abspath = __pyx_state[21]; __pyx_result.root_dir_info = __pyx_state[22]; __pyx_result.root_entries = __pyx_state[23]; __pyx_result.root_entries_len = __pyx_state[24]; __pyx_result.root_entries_pos = __pyx_state[25]; __pyx_result.search_specific_file_parents = __pyx_state[26]; __pyx_result.search_specific_files = __pyx_state[27]; __pyx_result.searched_exact_paths = __pyx_state[28]; __pyx_result.searched_specific_files = __pyx_state[29]; __pyx_result.seen_ids = __pyx_state[30]; __pyx_result.sha_file = __pyx_state[31]; __pyx_result.source_index = __pyx_state[32]; __pyx_result.state = __pyx_state[33]; __pyx_result.target_index = __pyx_state[34]; __pyx_result.tree = __pyx_state[35]; __pyx_result.use_filesystem_for_exec = __pyx_state[36]; __pyx_result.utf8_decode = __pyx_state[37]; __pyx_result.want_unversioned = __pyx_state[38]             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 39)
*/
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_pending_consistent_entries);
  __Pyx_DECREF(__pyx_v___pyx_result->_pending_consistent_entries);
  __pyx_v___pyx_result->_pending_consistent_entries = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->bisect_left);
  __Pyx_DECREF(__pyx_v___pyx_result->bisect_left);
  __pyx_v___pyx_result->bisect_left = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->block_index = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->current_block);
  __Pyx_DECREF(__pyx_v___pyx_result->current_block);
  __pyx_v___pyx_result->current_block = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->current_block_list);
  __Pyx_DECREF(__pyx_v___pyx_result->current_block_list);
  __pyx_v___pyx_result->current_block_list = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->current_block_pos = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->current_dir_info);
  __Pyx_DECREF(__pyx_v___pyx_result->current_dir_info);
  __pyx_v___pyx_result->current_dir_info = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->current_dir_list);
  __Pyx_DECREF(__pyx_v___pyx_result->current_dir_list);
  __pyx_v___pyx_result->current_dir_list = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->current_root);
  __Pyx_DECREF(__pyx_v___pyx_result->current_root);
  __pyx_v___pyx_result->current_root = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->current_root_unicode);
  __Pyx_DECREF(__pyx_v___pyx_result->current_root_unicode);
  __pyx_v___pyx_result->current_root_unicode = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->dir_iterator);
  __Pyx_DECREF(__pyx_v___pyx_result->dir_iterator);
  __pyx_v___pyx_result->dir_iterator = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->doing_consistency_expansion = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->fstat);
  __Pyx_DECREF(__pyx_v___pyx_result->fstat);
  __pyx_v___pyx_result->fstat = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->include_unchanged = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->last_source_parent);
  __Pyx_DECREF(__pyx_v___pyx_result->last_source_parent);
  __pyx_v___pyx_result->last_source_parent = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->last_target_parent);
  __Pyx_DECREF(__pyx_v___pyx_result->last_target_parent);
  __pyx_v___pyx_result->last_target_parent = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->new_dirname_to_file_id);
  __Pyx_DECREF(__pyx_v___pyx_result->new_dirname_to_file_id);
  __pyx_v___pyx_result->new_dirname_to_file_id = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 17, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->old_dirname_to_file_id);
  __Pyx_DECREF(__pyx_v___pyx_result->old_dirname_to_file_id);
  __pyx_v___pyx_result->old_dirname_to_file_id = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 18, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->partial = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 19, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->path_index = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 20, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->pathjoin);
  __Pyx_DECREF(__pyx_v___pyx_result->pathjoin);
  __pyx_v___pyx_result->pathjoin = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 21, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->root_abspath);
  __Pyx_DECREF(__pyx_v___pyx_result->root_abspath);
  __pyx_v___pyx_result->root_abspath = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 22, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->root_dir_info);
  __Pyx_DECREF(__pyx_v___pyx_result->root_dir_info);
  __pyx_v___pyx_result->root_dir_info = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 23, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->root_entries);
  __Pyx_DECREF(__pyx_v___pyx_result->root_entries);
  __pyx_v___pyx_result->root_entries = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 24, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->root_entries_len = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 25, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->root_entries_pos = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 26, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->search_specific_file_parents);
  __Pyx_DECREF(__pyx_v___pyx_result->search_specific_file_parents);
  __pyx_v___pyx_result->search_specific_file_parents = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 27, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->search_specific_files);
  __Pyx_DECREF(__pyx_v___pyx_result->search_specific_files);
  __pyx_v___pyx_result->search_specific_files = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 28, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->searched_exact_paths);
  __Pyx_DECREF(__pyx_v___pyx_result->searched_exact_paths);
  __pyx_v___pyx_result->searched_exact_paths = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 29, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->searched_specific_files);
  __Pyx_DECREF(__pyx_v___pyx_result->searched_specific_files);
  __pyx_v___pyx_result->searched_specific_files = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 30, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->seen_ids);
  __Pyx_DECREF(__pyx_v___pyx_result->seen_ids);
  __pyx_v___pyx_result->seen_ids = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 31, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->sha_file);
  __Pyx_DECREF(__pyx_v___pyx_result->sha_file);
  __pyx_v___pyx_result->sha_file = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 32, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->source_index = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 33, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->state);
  __Pyx_DECREF(__pyx_v___pyx_result->state);
  __pyx_v___pyx_result->state = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 34, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->target_index = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 35, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->tree);
  __Pyx_DECREF(__pyx_v___pyx_result->tree);
  __pyx_v___pyx_result->tree = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 36, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->use_filesystem_for_exec);
  __Pyx_DECREF(__pyx_v___pyx_result->use_filesystem_for_exec);
  __pyx_v___pyx_result->use_filesystem_for_exec = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 37, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->utf8_decode);
  __Pyx_DECREF(__pyx_v___pyx_result->utf8_decode);
  __pyx_v___pyx_result->utf8_decode = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 38, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->want_unversioned = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_ProcessEntryC__set_state(ProcessEntryC __pyx_result, __pyx_state: tuple):
 *     __pyx_result._pending_consistent_entries = __pyx_state[0]; __pyx_result.bisect_left = __pyx_state[1]; __pyx_result.block_index = __pyx_state[2]; __pyx_result.current_block = __pyx_state[3]; __pyx_result.current_block_list = __pyx_state[4]; __pyx_result.current_block_pos = __pyx_state[5]; __pyx_result.current_dir_info = __pyx_state[6]; __pyx_result.current_dir_list = __pyx_state[7]; __pyx_result.current_root = __pyx_state[8]; __pyx_result.current_root_unicode = __pyx_state[9]; __pyx_result.dir_iterator = __pyx_state[10]; __pyx_result.doing_consistency_expansion = __pyx_state[11]; __pyx_result.fstat = __pyx_state[12]; __pyx_result.include_unchanged = __pyx_state[13]; __pyx_result.last_source_parent = __pyx_state[14]; __pyx_result.last_target_parent = __pyx_state[15]; __pyx_result.new_dirname_to_file_id = __pyx_state[16]; __pyx_result.old_dirname_to_file_id = __pyx_state[17]; __pyx_result.partial = __pyx_state[18]; __pyx_result.path_index = __pyx_state[19]; __pyx_result.pathjoin = __pyx_state[20]; __pyx_result.root_abspath = __pyx_state[21]; __pyx_result.root_dir_info = __pyx_state[22]; __pyx_result.root_entries = __pyx_state[23]; __pyx_result.root_entries_len = __pyx_state[24]; __pyx_result.root_entries_pos = __pyx_state[25]; __pyx_result.search_specific_file_parents = __pyx_state[26]; __pyx_result.search_specific_files = __pyx_state[27]; __pyx_result.searched_exact_paths = __pyx_state[28]; __pyx_result.searched_specific_files = __pyx_state[29]; __pyx_result.seen_ids = __pyx_state[30]; __pyx_result.sha_file = __pyx_state[31]; __pyx_result.source_index = __pyx_state[32]; __pyx_result.state = __pyx_state[33]; __pyx_result.target_index = __pyx_state[34]; __pyx_result.tree = __pyx_state[35]; __pyx_result.use_filesystem_for_exec = __pyx_state[36]; __pyx_result.utf8_decode = __pyx_state[37]; __pyx_result.want_unversioned = __pyx_state[38]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 39)             # <<<<<<<<<<<<<<
*/
  __pyx_t_2 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 39); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle_ProcessEntryC__set_state(<ProcessEntryC> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ProcessEntryC__set_state(ProcessEntryC __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._pending_consistent_entries = __pyx_state[0]; __pyx_result.bisect_left = __pyx_state[1]; __pyx_result.block_index = __pyx_state[2]; __pyx_result.current_block = __pyx_state[3]; __pyx_result.current_block_list = __pyx_state[4]; __pyx_result.current_block_pos = __pyx_state[5]; __pyx_result.current_dir_info = __pyx_state[6]; __pyx_result.current_dir_list = __pyx_state[7]; __pyx_result.current_root = __pyx_state[8]; __pyx_result.current_root_unicode = __pyx_state[9]; __pyx_result.dir_iterator = __pyx_state[10]; __pyx_result.doing_consistency_expansion = __pyx_state[11]; __pyx_result.fstat = __pyx_state[12]; __pyx_result.include_unchanged = __pyx_state[13]; __pyx_result.last_source_parent = __pyx_state[14]; __pyx_result.last_target_parent = __pyx_state[15]; __pyx_result.new_dirname_to_file_id = __pyx_state[16]; __pyx_result.old_dirname_to_file_id = __pyx_state[17]; __pyx_result.partial = __pyx_state[18]; __pyx_result.path_index = __pyx_state[19]; __pyx_result.pathjoin = __pyx_state[20]; __pyx_result.root_abspath = __pyx_state[21]; __pyx_result.root_dir_info = __pyx_state[22]; __pyx_result.root_entries = __pyx_state[23]; __pyx_result.root_entries_len = __pyx_state[24]; __pyx_result.root_entries_pos = __pyx_state[25]; __pyx_result.search_specific_file_parents = __pyx_state[26]; __pyx_result.search_specific_files = __pyx_state[27]; __pyx_result.searched_exact_paths = __pyx_state[28]; __pyx_result.searched_specific_files = __pyx_state[29]; __pyx_result.seen_ids = __pyx_state[30]; __pyx_result.sha_file = __pyx_state[31]; __pyx_result.source_index = __pyx_state[32]; __pyx_result.state = __pyx_state[33]; __pyx_result.target_index = __pyx_state[34]; __pyx_result.tree = __pyx_state[35]; __pyx_result.use_filesystem_for_exec = __pyx_state[36]; __pyx_result.utf8_decode = __pyx_state[37]; __pyx_result.want_unversioned = __pyx_state[38]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 39)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.__pyx_unpickle_ProcessEntryC__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
/* #### Code section: module_exttypes ### */
static struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_Reader;

static PyObject *__pyx_tp_new_6breezy_3bzr_21_dirstate_helpers_pyx_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6breezy_3bzr_21_dirstate_helpers_pyx_Reader;
  p->state = Py_None; Py_INCREF(Py_None);
  p->text = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6breezy_3bzr_21_dirstate_helpers_pyx_Reader(PyObject *o) {
  struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *p = (struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6breezy_3bzr_21_dirstate_helpers_pyx_Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->state);
  Py_CLEAR(p->text);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_6breezy_3bzr_21_dirstate_helpers_pyx_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *p = (struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->state) {
    e = (*v)(p->state, a); if (e) return e;
  }
  if (p->text) {
    e = (*v)(p->text, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6breezy_3bzr_21_dirstate_helpers_pyx_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *p = (struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)o;
  tmp = ((PyObject*)p->state);
  p->state = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->text);
  p->text = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6breezy_3bzr_21_dirstate_helpers_pyx_Reader[] = {
  {"_parse_dirblocks", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_3_parse_dirblocks, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_2_parse_dirblocks},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6breezy_3bzr_21_dirstate_helpers_pyx_Reader},
  {Py_tp_doc, (void *)PyDoc_STR("Maintain the current location, and return fields as you parse them.")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6breezy_3bzr_21_dirstate_helpers_pyx_Reader},
  {Py_tp_clear, (void *)__pyx_tp_clear_6breezy_3bzr_21_dirstate_helpers_pyx_Reader},
  {Py_tp_methods, (void *)__pyx_methods_6breezy_3bzr_21_dirstate_helpers_pyx_Reader},
  {Py_tp_init, (void *)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6breezy_3bzr_21_dirstate_helpers_pyx_Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader_spec = {
  "breezy.bzr._dirstate_helpers_pyx.Reader",
  sizeof(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader_slots,
};
#else

static PyTypeObject __pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "breezy.bzr._dirstate_helpers_pyx.""Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6breezy_3bzr_21_dirstate_helpers_pyx_Reader, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Maintain the current location, and return fields as you parse them."), /*tp_doc*/
  __pyx_tp_traverse_6breezy_3bzr_21_dirstate_helpers_pyx_Reader, /*tp_traverse*/
  __pyx_tp_clear_6breezy_3bzr_21_dirstate_helpers_pyx_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6breezy_3bzr_21_dirstate_helpers_pyx_Reader, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6breezy_3bzr_21_dirstate_helpers_pyx_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC;

static PyObject *__pyx_tp_new_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)o);
  p->__pyx_vtab = __pyx_vtabptr_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC;
  p->old_dirname_to_file_id = Py_None; Py_INCREF(Py_None);
  p->new_dirname_to_file_id = Py_None; Py_INCREF(Py_None);
  p->last_source_parent = Py_None; Py_INCREF(Py_None);
  p->last_target_parent = Py_None; Py_INCREF(Py_None);
  p->use_filesystem_for_exec = Py_None; Py_INCREF(Py_None);
  p->utf8_decode = Py_None; Py_INCREF(Py_None);
  p->searched_specific_files = Py_None; Py_INCREF(Py_None);
  p->searched_exact_paths = Py_None; Py_INCREF(Py_None);
  p->search_specific_files = Py_None; Py_INCREF(Py_None);
  p->search_specific_file_parents = Py_None; Py_INCREF(Py_None);
  p->state = Py_None; Py_INCREF(Py_None);
  p->current_root = Py_None; Py_INCREF(Py_None);
  p->current_root_unicode = Py_None; Py_INCREF(Py_None);
  p->root_entries = Py_None; Py_INCREF(Py_None);
  p->root_abspath = Py_None; Py_INCREF(Py_None);
  p->tree = Py_None; Py_INCREF(Py_None);
  p->dir_iterator = Py_None; Py_INCREF(Py_None);
  p->current_block = Py_None; Py_INCREF(Py_None);
  p->current_block_list = Py_None; Py_INCREF(Py_None);
  p->current_dir_info = Py_None; Py_INCREF(Py_None);
  p->current_dir_list = Py_None; Py_INCREF(Py_None);
  p->_pending_consistent_entries = Py_None; Py_INCREF(Py_None);
  p->root_dir_info = Py_None; Py_INCREF(Py_None);
  p->bisect_left = Py_None; Py_INCREF(Py_None);
  p->pathjoin = Py_None; Py_INCREF(Py_None);
  p->fstat = Py_None; Py_INCREF(Py_None);
  p->seen_ids = Py_None; Py_INCREF(Py_None);
  p->sha_file = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC(PyObject *o) {
  struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *p = (struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->old_dirname_to_file_id);
  Py_CLEAR(p->new_dirname_to_file_id);
  Py_CLEAR(p->last_source_parent);
  Py_CLEAR(p->last_target_parent);
  Py_CLEAR(p->use_filesystem_for_exec);
  Py_CLEAR(p->utf8_decode);
  Py_CLEAR(p->searched_specific_files);
  Py_CLEAR(p->searched_exact_paths);
  Py_CLEAR(p->search_specific_files);
  Py_CLEAR(p->search_specific_file_parents);
  Py_CLEAR(p->state);
  Py_CLEAR(p->current_root);
  Py_CLEAR(p->current_root_unicode);
  Py_CLEAR(p->root_entries);
  Py_CLEAR(p->root_abspath);
  Py_CLEAR(p->tree);
  Py_CLEAR(p->dir_iterator);
  Py_CLEAR(p->current_block);
  Py_CLEAR(p->current_block_list);
  Py_CLEAR(p->current_dir_info);
  Py_CLEAR(p->current_dir_list);
  Py_CLEAR(p->_pending_consistent_entries);
  Py_CLEAR(p->root_dir_info);
  Py_CLEAR(p->bisect_left);
  Py_CLEAR(p->pathjoin);
  Py_CLEAR(p->fstat);
  Py_CLEAR(p->seen_ids);
  Py_CLEAR(p->sha_file);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *p = (struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->old_dirname_to_file_id) {
    e = (*v)(p->old_dirname_to_file_id, a); if (e) return e;
  }
  if (p->new_dirname_to_file_id) {
    e = (*v)(p->new_dirname_to_file_id, a); if (e) return e;
  }
  if (p->last_source_parent) {
    e = (*v)(p->last_source_parent, a); if (e) return e;
  }
  if (p->last_target_parent) {
    e = (*v)(p->last_target_parent, a); if (e) return e;
  }
  if (p->use_filesystem_for_exec) {
    e = (*v)(p->use_filesystem_for_exec, a); if (e) return e;
  }
  if (p->utf8_decode) {
    e = (*v)(p->utf8_decode, a); if (e) return e;
  }
  if (p->searched_specific_files) {
    e = (*v)(p->searched_specific_files, a); if (e) return e;
  }
  if (p->searched_exact_paths) {
    e = (*v)(p->searched_exact_paths, a); if (e) return e;
  }
  if (p->search_specific_files) {
    e = (*v)(p->search_specific_files, a); if (e) return e;
  }
  if (p->search_specific_file_parents) {
    e = (*v)(p->search_specific_file_parents, a); if (e) return e;
  }
  if (p->state) {
    e = (*v)(p->state, a); if (e) return e;
  }
  if (p->current_root) {
    e = (*v)(p->current_root, a); if (e) return e;
  }
  if (p->current_root_unicode) {
    e = (*v)(p->current_root_unicode, a); if (e) return e;
  }
  if (p->root_entries) {
    e = (*v)(p->root_entries, a); if (e) return e;
  }
  if (p->root_abspath) {
    e = (*v)(p->root_abspath, a); if (e) return e;
  }
  if (p->tree) {
    e = (*v)(p->tree, a); if (e) return e;
  }
  if (p->dir_iterator) {
    e = (*v)(p->dir_iterator, a); if (e) return e;
  }
  if (p->current_block) {
    e = (*v)(p->current_block, a); if (e) return e;
  }
  if (p->current_block_list) {
    e = (*v)(p->current_block_list, a); if (e) return e;
  }
  if (p->current_dir_info) {
    e = (*v)(p->current_dir_info, a); if (e) return e;
  }
  if (p->current_dir_list) {
    e = (*v)(p->current_dir_list, a); if (e) return e;
  }
  if (p->_pending_consistent_entries) {
    e = (*v)(p->_pending_consistent_entries, a); if (e) return e;
  }
  if (p->root_dir_info) {
    e = (*v)(p->root_dir_info, a); if (e) return e;
  }
  if (p->bisect_left) {
    e = (*v)(p->bisect_left, a); if (e) return e;
  }
  if (p->pathjoin) {
    e = (*v)(p->pathjoin, a); if (e) return e;
  }
  if (p->fstat) {
    e = (*v)(p->fstat, a); if (e) return e;
  }
  if (p->seen_ids) {
    e = (*v)(p->seen_ids, a); if (e) return e;
  }
  if (p->sha_file) {
    e = (*v)(p->sha_file, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *p = (struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)o;
  tmp = ((PyObject*)p->old_dirname_to_file_id);
  p->old_dirname_to_file_id = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->new_dirname_to_file_id);
  p->new_dirname_to_file_id = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->last_source_parent);
  p->last_source_parent = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->last_target_parent);
  p->last_target_parent = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->use_filesystem_for_exec);
  p->use_filesystem_for_exec = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->utf8_decode);
  p->utf8_decode = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->searched_specific_files);
  p->searched_specific_files = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->searched_exact_paths);
  p->searched_exact_paths = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->search_specific_files);
  p->search_specific_files = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->search_specific_file_parents);
  p->search_specific_file_parents = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->state);
  p->state = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->current_root);
  p->current_root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->current_root_unicode);
  p->current_root_unicode = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->root_entries);
  p->root_entries = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->root_abspath);
  p->root_abspath = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->tree);
  p->tree = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->dir_iterator);
  p->dir_iterator = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->current_block);
  p->current_block = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->current_block_list);
  p->current_block_list = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->current_dir_info);
  p->current_dir_info = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->current_dir_list);
  p->current_dir_list = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_pending_consistent_entries);
  p->_pending_consistent_entries = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->root_dir_info);
  p->root_dir_info = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->bisect_left);
  p->bisect_left = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->pathjoin);
  p->pathjoin = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->fstat);
  p->fstat = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->seen_ids);
  p->seen_ids = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->sha_file);
  p->sha_file = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_searched_specific_files(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_23searched_specific_files_1__get__(o);
}

static PyObject *__pyx_getprop_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_searched_exact_paths(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_20searched_exact_paths_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_7__next__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  PyObject *res = __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_7__next__(self);
  if (!res && !PyErr_Occurred()) { PyErr_SetNone(PyExc_StopIteration); }
  return res;
}

static PyMethodDef __pyx_methods_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC[] = {
  {"iter_changes", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_5iter_changes, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__next__", (PyCFunction)__pyx_specialmethod___pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_7__next__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC[] = {
  {"searched_specific_files", __pyx_getprop_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_searched_specific_files, 0, 0, 0},
  {"searched_exact_paths", __pyx_getprop_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_searched_exact_paths, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC},
  {Py_tp_clear, (void *)__pyx_tp_clear_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC},
  {Py_tp_iter, (void *)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_3__iter__},
  {Py_tp_iternext, (void *)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_7__next__},
  {Py_tp_methods, (void *)__pyx_methods_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC},
  {Py_tp_getset, (void *)__pyx_getsets_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC},
  {Py_tp_init, (void *)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC},
  {0, 0},
};
static PyType_Spec __pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC_spec = {
  "breezy.bzr._dirstate_helpers_pyx.ProcessEntryC",
  sizeof(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC_slots,
};
#else

static PyTypeObject __pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC = {
  PyVarObject_HEAD_INIT(0, 0)
  "breezy.bzr._dirstate_helpers_pyx.""ProcessEntryC", /*tp_name*/
  sizeof(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC, /*tp_traverse*/
  __pyx_tp_clear_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_3__iter__, /*tp_iter*/
  __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_7__next__, /*tp_iternext*/
  __pyx_methods_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_absent = Py_None; Py_INCREF(Py_None);
  __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_file = Py_None; Py_INCREF(Py_None);
  __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_directory = Py_None; Py_INCREF(Py_None);
  __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_symlink = Py_None; Py_INCREF(Py_None);
  __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_relocated = Py_None; Py_INCREF(Py_None);
  __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_tree_reference = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_6breezy_3bzr_21_dirstate_helpers_pyx_Reader = &__pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_Reader;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_Reader.get_next = (char *(*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *, int *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_get_next;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_Reader.get_next_str = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_get_next_str;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_Reader._init = (int (*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader__init;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_Reader._get_entry = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *, int, void **, int *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader__get_entry;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader)) __PYX_ERR(0, 509, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader_spec, __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader) < (0)) __PYX_ERR(0, 509, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader = &__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader) < (0)) __PYX_ERR(0, 509, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader->tp_dictoffset && __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader, __pyx_vtabptr_6breezy_3bzr_21_dirstate_helpers_pyx_Reader) < (0)) __PYX_ERR(0, 509, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader) < (0)) __PYX_ERR(0, 509, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_Reader, (PyObject *) __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader) < (0)) __PYX_ERR(0, 509, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader) < (0)) __PYX_ERR(0, 509, __pyx_L1_error)
  __pyx_vtabptr_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC = &__pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC._process_entry = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *, PyObject *, PyObject *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__process_entry;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC._gather_result_for_consistency = (int (*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *, PyObject *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__gather_result_for_consistency;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC._update_current_block = (int (*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__update_current_block;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC._iter_next = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__iter_next;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC._maybe_tree_ref = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *, PyObject *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__maybe_tree_ref;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC._loop_one_block = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__loop_one_block;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC._next_consistent_entries = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__next_consistent_entries;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC._path_info = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *, PyObject *, PyObject *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__path_info;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC)) __PYX_ERR(0, 967, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC_spec, __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC) < (0)) __PYX_ERR(0, 967, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC = &__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC) < (0)) __PYX_ERR(0, 967, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC->tp_dictoffset && __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC, __pyx_vtabptr_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC) < (0)) __PYX_ERR(0, 967, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC) < (0)) __PYX_ERR(0, 967, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_ProcessEntryC, (PyObject *) __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC) < (0)) __PYX_ERR(0, 967, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC) < (0)) __PYX_ERR(0, 967, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_2_4(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  0, 0,
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__dirstate_helpers_pyx(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__dirstate_helpers_pyx},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, __Pyx_FREETHREADING_COMPATIBLE},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "_dirstate_helpers_pyx",
      __pyx_k_Helper_functions_for_DirState_Th, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit__dirstate_helpers_pyx(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__dirstate_helpers_pyx(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
    PY_INT64_T current_id = GraalPyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__dirstate_helpers_pyx(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_dirstate_helpers_pyx' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "_dirstate_helpers_pyx" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
    PyErr_Clear();
    __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
    if (!__Pyx_RefNanny)
        Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  
__Pyx_RefNannySetupContext("PyInit__dirstate_helpers_pyx", 0);
  __Pyx_init_runtime_version();
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Library function declarations ---*/
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__pyx_module_is_main_breezy__bzr___dirstate_helpers_pyx) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name, __pyx_mstate_global->__pyx_n_u_main) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "breezy.bzr._dirstate_helpers_pyx")) {
      if (unlikely((PyDict_SetItemString(modules, "breezy.bzr._dirstate_helpers_pyx", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  (void)__Pyx_modinit_function_export_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_type_init_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  (void)__Pyx_modinit_function_import_code(__pyx_mstate);
  /*--- Execution code ---*/

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":24
 * """
 * 
 * import binascii             # <<<<<<<<<<<<<<
 * import bisect
 * import codecs
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_binascii, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_binascii, __pyx_t_2) < (0)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":25
 * 
 * import binascii
 * import bisect             # <<<<<<<<<<<<<<
 * import codecs
 * import errno
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_bisect, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_bisect, __pyx_t_2) < (0)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":26
 * import binascii
 * import bisect
 * import codecs             # <<<<<<<<<<<<<<
 * import errno
 * import os
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_codecs, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_codecs, __pyx_t_2) < (0)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":27
 * import bisect
 * import codecs
 * import errno             # <<<<<<<<<<<<<<
 * import os
 * import stat
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_errno, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_errno, __pyx_t_2) < (0)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":28
 * import codecs
 * import errno
 * import os             # <<<<<<<<<<<<<<
 * import stat
 * import sys
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_os, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_os, __pyx_t_2) < (0)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":29
 * import errno
 * import os
 * import stat             # <<<<<<<<<<<<<<
 * import sys
 * 
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_stat, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_stat, __pyx_t_2) < (0)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":30
 * import os
 * import stat
 * import sys             # <<<<<<<<<<<<<<
 * 
 * from .. import errors, osutils
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_sys, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_sys, __pyx_t_2) < (0)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":32
 * import sys
 * 
 * from .. import errors, osutils             # <<<<<<<<<<<<<<
 * from ..osutils import (is_inside, is_inside_any, parent_directories, pathjoin,
 *                        splitpath)
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_errors,__pyx_mstate_global->__pyx_n_u_osutils};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u__4, __pyx_imported_names, 2, __pyx_mstate_global->__pyx_kp_u_breezy, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_errors,__pyx_mstate_global->__pyx_n_u_osutils};
    for (__pyx_t_3=0; __pyx_t_3 < 2; __pyx_t_3++) {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 32, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 32, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":33
 * 
 * from .. import errors, osutils
 * from ..osutils import (is_inside, is_inside_any, parent_directories, pathjoin,             # <<<<<<<<<<<<<<
 *                        splitpath)
 * from .dirstate import DirState, DirstateCorrupt
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_is_inside,__pyx_mstate_global->__pyx_n_u_is_inside_any,__pyx_mstate_global->__pyx_n_u_parent_directories,__pyx_mstate_global->__pyx_n_u_pathjoin,__pyx_mstate_global->__pyx_n_u_splitpath};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_osutils, __pyx_imported_names, 5, __pyx_mstate_global->__pyx_kp_u_breezy_osutils, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_is_inside,__pyx_mstate_global->__pyx_n_u_is_inside_any,__pyx_mstate_global->__pyx_n_u_parent_directories,__pyx_mstate_global->__pyx_n_u_pathjoin,__pyx_mstate_global->__pyx_n_u_splitpath};
    for (__pyx_t_3=0; __pyx_t_3 < 5; __pyx_t_3++) {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 33, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 33, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":35
 * from ..osutils import (is_inside, is_inside_any, parent_directories, pathjoin,
 *                        splitpath)
 * from .dirstate import DirState, DirstateCorrupt             # <<<<<<<<<<<<<<
 * from .inventorytree import InventoryTreeChange
 * 
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_DirState,__pyx_mstate_global->__pyx_n_u_DirstateCorrupt};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_dirstate, __pyx_imported_names, 2, __pyx_mstate_global->__pyx_kp_u_breezy_bzr_dirstate, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_DirState,__pyx_mstate_global->__pyx_n_u_DirstateCorrupt};
    for (__pyx_t_3=0; __pyx_t_3 < 2; __pyx_t_3++) {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 35, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 35, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":36
 *                        splitpath)
 * from .dirstate import DirState, DirstateCorrupt
 * from .inventorytree import InventoryTreeChange             # <<<<<<<<<<<<<<
 * 
 * from cpython.tuple cimport PyTuple_New, PyTuple_SET_ITEM
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_InventoryTreeChange};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_inventorytree, __pyx_imported_names, 1, __pyx_mstate_global->__pyx_kp_u_breezy_bzr_inventorytree, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_InventoryTreeChange};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 36, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 36, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":46
 * # XXX: Perhaps we could get it from a windows header ?
 * cdef int ERROR_PATH_NOT_FOUND
 * ERROR_PATH_NOT_FOUND = 3             # <<<<<<<<<<<<<<
 * cdef int ERROR_DIRECTORY
 * ERROR_DIRECTORY = 267
*/
  __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx_ERROR_PATH_NOT_FOUND = 3;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":48
 * ERROR_PATH_NOT_FOUND = 3
 * cdef int ERROR_DIRECTORY
 * ERROR_DIRECTORY = 267             # <<<<<<<<<<<<<<
 * 
 * cdef extern from "python-compat.h":
*/
  __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx_ERROR_DIRECTORY = 0x10B;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":131
 * 
 * 
 * def _py_memrchr(s, c):             # <<<<<<<<<<<<<<
 *     """Just to expose _my_memrchr for testing.
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_1_py_memrchr, 0, __pyx_mstate_global->__pyx_n_u_py_memrchr, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__dirstate_helpers_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_py_memrchr, __pyx_t_2) < (0)) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":226
 * 
 * 
 * def lt_by_dirs(path1, path2):             # <<<<<<<<<<<<<<
 *     """Compare two paths directory by directory.
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_3lt_by_dirs, 0, __pyx_mstate_global->__pyx_n_u_lt_by_dirs, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__dirstate_helpers_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_lt_by_dirs, __pyx_t_2) < (0)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":253
 * 
 * 
 * def _lt_path_by_dirblock(path1, path2):             # <<<<<<<<<<<<<<
 *     """Compare two paths based on what directory they are in.
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_5_lt_path_by_dirblock, 0, __pyx_mstate_global->__pyx_n_u_lt_path_by_dirblock, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__dirstate_helpers_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[2])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_lt_path_by_dirblock, __pyx_t_2) < (0)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":354
 * 
 * 
 * def _bisect_path_left(paths, path):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert path into paths.
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_7_bisect_path_left, 0, __pyx_mstate_global->__pyx_n_u_bisect_path_left, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__dirstate_helpers_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[3])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_bisect_path_left, __pyx_t_2) < (0)) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":407
 * 
 * 
 * def _bisect_path_right(paths, path):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert path into paths.
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_9_bisect_path_right, 0, __pyx_mstate_global->__pyx_n_u_bisect_path_right, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__dirstate_helpers_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[4])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_bisect_path_right, __pyx_t_2) < (0)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":460
 * 
 * 
 * def bisect_dirblock(dirblocks, dirname, lo=0, hi=None, cache=None):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert dirname into the dirblocks.
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_11bisect_dirblock, 0, __pyx_mstate_global->__pyx_n_u_bisect_dirblock, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__dirstate_helpers_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[5])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[9]);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_bisect_dirblock, __pyx_t_2) < (0)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":689
 *         return ret
 * 
 *     def _parse_dirblocks(self):             # <<<<<<<<<<<<<<
 *         """Parse all dirblocks in the state file."""
 *         cdef int num_trees
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_3_parse_dirblocks, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Reader__parse_dirblocks, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__dirstate_helpers_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[6])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader, __pyx_mstate_global->__pyx_n_u_parse_dirblocks, __pyx_t_2) < (0)) __PYX_ERR(0, 689, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Reader___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__dirstate_helpers_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[7])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Reader, (type(self), 0xff2cf55, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Reader__set_state(self, __pyx_state)
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Reader___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__dirstate_helpers_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[8])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":737
 * 
 * 
 * def _read_dirblocks(state):             # <<<<<<<<<<<<<<
 *     """Read in the dirblocks for the given DirState object.
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_13_read_dirblocks, 0, __pyx_mstate_global->__pyx_n_u_read_dirblocks, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__dirstate_helpers_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[9])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 737, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_read_dirblocks, __pyx_t_2) < (0)) __PYX_ERR(0, 737, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":770
 * 
 * 
 * _encode = binascii.b2a_base64             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_binascii); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_b2a_base64); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_encode, __pyx_t_4) < (0)) __PYX_ERR(0, 770, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":799
 * 
 * 
 * def pack_stat(stat_value):             # <<<<<<<<<<<<<<
 *     """Convert stat value into a packed representation quickly with pyrex"""
 *     return _pack_stat(stat_value)
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_15pack_stat, 0, __pyx_mstate_global->__pyx_n_u_pack_stat, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__dirstate_helpers_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[10])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pack_stat, __pyx_t_4) < (0)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":804
 * 
 * 
 * cpdef update_entry(self, entry, abspath, stat_value):             # <<<<<<<<<<<<<<
 *     """Update the entry based on what is actually on disk.
 * 
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_17update_entry, 0, __pyx_mstate_global->__pyx_n_u_update_entry, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__dirstate_helpers_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[11])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_update_entry, __pyx_t_4) < (0)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":932
 * cdef object _kind_relocated
 * cdef object _kind_tree_reference
 * _kind_absent = "absent"             # <<<<<<<<<<<<<<
 * _kind_file = "file"
 * _kind_directory = "directory"
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_absent);
  __Pyx_XGOTREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_absent);
  __Pyx_DECREF_SET(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_absent, __pyx_mstate_global->__pyx_n_u_absent);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_absent);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":933
 * cdef object _kind_tree_reference
 * _kind_absent = "absent"
 * _kind_file = "file"             # <<<<<<<<<<<<<<
 * _kind_directory = "directory"
 * _kind_symlink = "symlink"
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_file);
  __Pyx_XGOTREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_file);
  __Pyx_DECREF_SET(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_file, __pyx_mstate_global->__pyx_n_u_file);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_file);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":934
 * _kind_absent = "absent"
 * _kind_file = "file"
 * _kind_directory = "directory"             # <<<<<<<<<<<<<<
 * _kind_symlink = "symlink"
 * _kind_relocated = "relocated"
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_directory);
  __Pyx_XGOTREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_directory);
  __Pyx_DECREF_SET(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_directory, __pyx_mstate_global->__pyx_n_u_directory);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_directory);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":935
 * _kind_file = "file"
 * _kind_directory = "directory"
 * _kind_symlink = "symlink"             # <<<<<<<<<<<<<<
 * _kind_relocated = "relocated"
 * _kind_tree_reference = "tree-reference"
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_symlink);
  __Pyx_XGOTREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_symlink);
  __Pyx_DECREF_SET(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_symlink, __pyx_mstate_global->__pyx_n_u_symlink);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_symlink);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":936
 * _kind_directory = "directory"
 * _kind_symlink = "symlink"
 * _kind_relocated = "relocated"             # <<<<<<<<<<<<<<
 * _kind_tree_reference = "tree-reference"
 * 
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_relocated);
  __Pyx_XGOTREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_relocated);
  __Pyx_DECREF_SET(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_relocated, __pyx_mstate_global->__pyx_n_u_relocated);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_relocated);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":937
 * _kind_symlink = "symlink"
 * _kind_relocated = "relocated"
 * _kind_tree_reference = "tree-reference"             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_tree_reference);
  __Pyx_XGOTREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_tree_reference);
  __Pyx_DECREF_SET(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_tree_reference, __pyx_mstate_global->__pyx_kp_u_tree_reference);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u_tree_reference);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1377
 *         return self
 * 
 *     def iter_changes(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_5iter_changes, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ProcessEntryC_iter_changes, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__dirstate_helpers_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[12])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC, __pyx_mstate_global->__pyx_n_u_iter_changes, __pyx_t_4) < (0)) __PYX_ERR(0, 1377, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_9__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ProcessEntryC___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__dirstate_helpers_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[13])); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_4) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x23b8720, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ProcessEntryC__set_state(self, __pyx_state)
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_11__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ProcessEntryC___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__dirstate_helpers_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[14])); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_4) < (0)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_Reader(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xff2cf55, 0x6e2f153, 0x65fc0dd, b'cur_cstr, end_cstr, next, state, text, text_cstr, text_size')
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_19__pyx_unpickle_Reader, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_Reader, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__dirstate_helpers_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[15])); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_Reader, __pyx_t_4) < (0)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * cdef extern from *:             # <<<<<<<<<<<<<<
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_21__pyx_unpickle_ProcessEntryC, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_ProcessEntryC, NULL, __pyx_mstate_global->__pyx_n_u_breezy_bzr__dirstate_helpers_pyx, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[16])); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_ProcessEntryC, __pyx_t_4) < (0)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1
 * # Copyright (C) 2007-2010 Canonical Ltd             # <<<<<<<<<<<<<<
 * #
 * # This program is free software; you can redistribute it and/or modify
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_4) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init breezy.bzr._dirstate_helpers_pyx", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init breezy.bzr._dirstate_helpers_pyx");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */
/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);

  /* Cached unbound methods */
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.method_name = &__pyx_mstate->__pyx_n_u_items;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.method_name = &__pyx_mstate->__pyx_n_u_values;
  __pyx_mstate->__pyx_umethod_PyList_Type_pop.type = (PyObject*)&PyList_Type;
  __pyx_mstate->__pyx_umethod_PyList_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  return 0;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":796
 *     aliased[5] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_mode))
 *     packed = PyBytes_FromStringAndSize(result, 6*4)
 *     return _encode(packed)[:-1]             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_mstate_global->__pyx_slice[0] = PySlice_New(Py_None, __pyx_mstate_global->__pyx_int_neg_1, Py_None); if (unlikely(!__pyx_mstate_global->__pyx_slice[0])) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_slice[0]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_slice[0]);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1147
 *                 # update the source details variable to be the real
 *                 # location.
 *                 if old_entry == (None, None):             # <<<<<<<<<<<<<<
 *                     raise DirstateCorrupt(self.state._filename,
 *                         "entry '%s/%s' is considered renamed from %r"
*/
  __pyx_mstate_global->__pyx_tuple[0] = PyTuple_Pack(2, Py_None, Py_None); if (unlikely(!__pyx_mstate_global->__pyx_tuple[0])) __PYX_ERR(0, 1147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[0]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[0]);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1265
 *                 )
 *             if not changed and not self.include_unchanged:
 *                 return None, False             # <<<<<<<<<<<<<<
 *             else:
 *                 if old_path is None:
*/
  __pyx_mstate_global->__pyx_tuple[1] = PyTuple_Pack(2, Py_None, Py_False); if (unlikely(!__pyx_mstate_global->__pyx_tuple[1])) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[1]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[1]);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1282
 *                        (old_path_u, path_u),
 *                        content_change,
 *                        (True, True),             # <<<<<<<<<<<<<<
 *                        (source_parent_id, target_parent_id),
 *                        (self.utf8_decode(old_basename, 'surrogateescape')[0], self.utf8_decode(entry[0][1], 'surrogateescape')[0]),
*/
  __pyx_mstate_global->__pyx_tuple[2] = PyTuple_Pack(2, Py_True, Py_True); if (unlikely(!__pyx_mstate_global->__pyx_tuple[2])) __PYX_ERR(0, 1282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[2]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[2]);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1315
 *                        (None, self.utf8_decode(path, 'surrogateescape')[0]),
 *                        True,
 *                        (False, True),             # <<<<<<<<<<<<<<
 *                        (None, parent_id),
 *                        (None, self.utf8_decode(entry[0][1], 'surrogateescape')[0]),
*/
  __pyx_mstate_global->__pyx_tuple[3] = PyTuple_Pack(2, Py_False, Py_True); if (unlikely(!__pyx_mstate_global->__pyx_tuple[3])) __PYX_ERR(0, 1315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[3]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[3]);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1344
 *                    (self.utf8_decode(old_path, 'surrogateescape')[0], None),
 *                    True,
 *                    (True, False),             # <<<<<<<<<<<<<<
 *                    (parent_id, None),
 *                    (self.utf8_decode(entry[0][1], 'surrogateescape')[0], None),
*/
  __pyx_mstate_global->__pyx_tuple[4] = PyTuple_Pack(2, Py_True, Py_False); if (unlikely(!__pyx_mstate_global->__pyx_tuple[4])) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[4]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[4]);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1510
 *                     if self.tree._directory_is_tree_reference(
 *                         self.current_root_unicode):
 *                         self.root_dir_info = self.root_dir_info[:2] + \             # <<<<<<<<<<<<<<
 *                             ('tree-reference',) + self.root_dir_info[3:]
 *             if not self.root_entries and not self.root_dir_info:
*/
  __pyx_mstate_global->__pyx_slice[1] = PySlice_New(Py_None, __pyx_mstate_global->__pyx_int_2, Py_None); if (unlikely(!__pyx_mstate_global->__pyx_slice[1])) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_slice[1]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_slice[1]);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1511
 *                         self.current_root_unicode):
 *                         self.root_dir_info = self.root_dir_info[:2] + \
 *                             ('tree-reference',) + self.root_dir_info[3:]             # <<<<<<<<<<<<<<
 *             if not self.root_entries and not self.root_dir_info:
 *                 # this specified path is not present at all, skip it.
*/
  __pyx_mstate_global->__pyx_tuple[5] = PyTuple_Pack(1, __pyx_mstate_global->__pyx_kp_u_tree_reference); if (unlikely(!__pyx_mstate_global->__pyx_tuple[5])) __PYX_ERR(0, 1511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[5]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[5]);
  __pyx_mstate_global->__pyx_slice[2] = PySlice_New(__pyx_mstate_global->__pyx_int_3, Py_None, Py_None); if (unlikely(!__pyx_mstate_global->__pyx_slice[2])) __PYX_ERR(0, 1511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_slice[2]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_slice[2]);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1542
 *                        (None, self.current_root_unicode),
 *                        True,
 *                        (False, False),             # <<<<<<<<<<<<<<
 *                        (None, None),
 *                        (None, splitpath(self.current_root_unicode)[-1]),
*/
  __pyx_mstate_global->__pyx_tuple[6] = PyTuple_Pack(2, Py_False, Py_False); if (unlikely(!__pyx_mstate_global->__pyx_tuple[6])) __PYX_ERR(0, 1542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[6]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[6]);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1588
 *                     if self.current_dir_info[0][0] == b'':
 *                         # remove .bzr from iteration
 *                         bzr_index = self.bisect_left(self.current_dir_list, (b'.bzr',))             # <<<<<<<<<<<<<<
 *                         if self.current_dir_list[bzr_index][0] != b'.bzr':
 *                             raise AssertionError()
*/
  __pyx_mstate_global->__pyx_tuple[7] = PyTuple_Pack(1, __pyx_mstate_global->__pyx_kp_b_bzr); if (unlikely(!__pyx_mstate_global->__pyx_tuple[7])) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[7]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[7]);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1971
 *             else:
 *                 raise
 *         utf8_basename = utf8_path.rsplit(b'/', 1)[-1]             # <<<<<<<<<<<<<<
 *         dir_info = (utf8_path, utf8_basename,
 *             osutils.file_kind_from_stat_mode(stat.st_mode), stat,
*/
  __pyx_mstate_global->__pyx_tuple[8] = PyTuple_Pack(2, __pyx_mstate_global->__pyx_kp_b__5, __pyx_mstate_global->__pyx_int_1); if (unlikely(!__pyx_mstate_global->__pyx_tuple[8])) __PYX_ERR(0, 1971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[8]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[8]);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":460
 * 
 * 
 * def bisect_dirblock(dirblocks, dirname, lo=0, hi=None, cache=None):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert dirname into the dirblocks.
 * 
*/
  __pyx_mstate_global->__pyx_tuple[9] = PyTuple_Pack(3, ((PyObject*)__pyx_mstate_global->__pyx_int_0), Py_None, Py_None); if (unlikely(!__pyx_mstate_global->__pyx_tuple[9])) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[9]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[9]);
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_tuple;
    for (Py_ssize_t i=0; i<10; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_slice;
    for (Py_ssize_t i=0; i<3; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  {
    const struct { const unsigned int length: 11; } index[] = {{2},{42},{36},{39},{36},{44},{39},{179},{27},{44},{57},{1},{0},{1},{2},{1},{1},{8},{7},{36},{19},{24},{14},{11},{34},{7},{42},{6},{7},{7},{55},{16},{2},{38},{46},{29},{9},{12},{36},{36},{36},{36},{11},{22},{18},{15},{14},{49},{24},{45},{49},{40},{39},{16},{8},{8},{15},{6},{6},{7},{20},{19},{8},{19},{13},{31},{33},{26},{20},{6},{24},{26},{23},{7},{7},{0},{6},{7},{3},{18},{10},{8},{6},{15},{11},{17},{18},{32},{1},{2},{5},{17},{18},{6},{3},{8},{8},{13},{15},{12},{1},{6},{8},{5},{15},{9},{10},{9},{28},{7},{12},{12},{8},{7},{6},{14},{13},{17},{5},{11},{5},{6},{20},{4},{7},{24},{9},{26},{5},{5},{8},{22},{10},{12},{2},{3},{17},{7},{13},{13},{14},{9},{13},{5},{12},{4},{6},{2},{3},{5},{10},{20},{8},{14},{4},{10},{8},{6},{7},{9},{4},{12},{20},{9},{3},{14},{2},{7},{9},{18},{16},{4},{5},{5},{9},{9},{9},{8},{5},{8},{3},{6},{11},{14},{12},{11},{10},{28},{21},{14},{12},{4},{15},{10},{6},{10},{17},{13},{9},{6},{1},{2},{21},{4},{4},{12},{10},{12},{19},{10},{14},{16},{8},{12},{5},{34},{9},{8},{6},{6},{7},{8},{7},{4},{13},{10},{5},{11},{15},{7},{3},{12},{8},{4},{4},{6},{12},{23},{12},{4},{12},{6},{14},{16},{5},{8},{1},{0},{1},{4},{11},{191},{191},{11},{204},{7},{707},{11},{134},{1509},{59},{219},{55},{56},{97},{104},{103},{1},{1},{1},{1},{1}};
    #if (CYTHON_COMPRESS_STRINGS) == 3 && __PYX_LIMITED_VERSION_HEX >= 0x030e0000 /* compression: zstd (3763 bytes) */
const char* const cstring = "(\265/\375`\020\033Mu\000\372\242\374\"H\340\316l\0030\3143\004\343c6C0\036a\030\037\263cv\\\013\263\021\206\023\266\232A2%D\022\2144D$W.\035\314\375:\r\177Lv\247\210\310~@?y\334T\346c}\311\260\250\276\234I\205&\t\200\277%#\365\001\373\001C\002-0\354\313Mb\303\353%\346\332o\257;\306%\326\345\2121\310\310\253w]\252\234E\226r\342\232\275Hp\213\213\005\317\\3\231\246\245\27162;!y\235\224\234\220\254\250\300\\pL\313\212\312K\005\303NZNHf3\330k\005\366Jq\275\300N^>\267r\256\216t\031\337$\374F\204\253\234\277P4\272Kn\245\212\222|\233\254\356\373FW\363f\243y\227\307\232\3375>k^\352R\237\232K\360\300\321r\307G\252\302<\272\3663\335x\244\220\356\214\346\272\366<r\314\276\257\270\221\325b\215\034B\023\247\362s45.\247\346\356\251Z\262\374\251\321\261\253)\367\035\363\265\254As9ztx\033\235\342\222\306A\3232\3113m6\202xL\230\351\030SR\\`\034\335\231\225\352\017\267\235n\034\306\251e\215Z\033k\324\3646zV\252\232\032%s\347\020\372\276\273V\315Q\227|U\273\377\350\350\366kW5\367<\256\356\322\r\002b\303\313\341\316u\016\3255\254}6q\010\277!C@x7ZsRc\323\314!@\306\217\220\021f\343V\225\327d\315\313Q\303\230\016F\033\255=Ws\271\265\247. \272\243\271D\3576\337\346\310\302\251\272}N3k\364\272\217\251\356:\035W\2257\236'#H\3279\236\367\350^\311\361T\307\256{\317hI\256}\3546\315\363hA\203\005\t,/\2261\345\241\332g\256\311\215js\354\273\346y#\243\3175?j\257\302\372V\031\313}n\365\034k\245\317\270\276qc\315\347\034\010\t\036x\334\004E\327t\324d\315w\276K\270\003\356\373\3103\317\265r\324(-\335\375U3\246{\364\251\366\232\306\263\337\347h\242\253:2r\360j:~\2443c\252\216f\252G\325\217\246\366\245\316\331\350Q\327\214\037\272\363\234\033\315,2r\010\027\000ZdV@)\350b}\003\236\026\344\344{\271\037\034\257\306'P*\351\241L\027\243\026*\027C\353\277\202\037\351\253\340\253\220\340\035}\022v\303t\241\321\302\305\242\345\276\264\256\266\301\316*\377D~\020|\024:\202-\237\330\352\211\241\340jG\3659\353\200\005p\nz\200\344\257\203\257BGp\345\013]]\313\022_\261\024T\255\\\225\353\021\340\r\340\033\340z\334""\216\324[\341kV\204\r\261\031\245\240\207%{\rNlF(\250j=\355\230\361f\204vV\301\334\213\365\030\364\302\320\216\256%\345\037\373\010Hn\251\325}\005\034A\317\327\261\273\227*\005c\337\365\017\210\257gQ\022]`h\341a\261\362\225\270-7B(\350a\021\022R\250\220\356\032S\247\352\356|\305>\315574\327`\254\001\343t_\315\334\352m<\356\274\216\343u\235\313e\3274\225\363U\257h\336u\325\367\235.\261\367\251\352\323\255\252\034\357T\236\313\2476OT\367\233\306Y\367\216\273\337\334Fc~K7\312\321\274\352v\032\275\225;\227\256\032\363Z\311q\032\207m\366\231Ss8\337\323\255\3324\234~\315\271\256\266r\255\357\270\356\244\212m\244\332;\326<\207[S\\\2035\272bs\343\366\271o<\3430\212;\212\252(V\321\232\247u^\232\361\3266zE\313t\357\032\346a\\\246\353\211O\363\344\351\033\016csc\215\356\023w\255\246\234k\237\256j\211y\030k\276\366\032>Ul\246\032\343\276161\326\246\251q\333\343\363T\375Zb\354{\311\327\262\227\332T\313[\231I]f\356q\2721N\373\2151\326\2652\227%.}\215\2669\366\245\356\233\334\273\3549\335\244\005\014\272g\022g\nd\327\351\366\253\312\325\022D\335e\312\353\3314\263Ou'\263\311\253E6\362*\361:\3012\314\265\231u\313\251\003&\026\023K\t\007\247\304\343\225\316\263D\222%\231\254\264RZ!\345\344\220R\244 \241(\251V#\221hH\244\030\022)\205D\0321Z\261bts3\312yDC3\"!\031\211\030\211\020\335n\"\032M\324\322\"\202\211`\241\233\233\320n\027J\323\220\315\026JI\t\215\010\215\000\361x\240u\005\2358\001\222\221\001\241\240\200L@&\037\357?\271\334\207\206\346\023\023\363y\275>\256\217\313\203^\\\264\270\321^\032\314\013#{O\0005\200 |\336\rQ\275;1 \315\304\267,\214\027\230\341\367R\272\033.\253\315\260\243\323\200\351\355r\211I\357\024\n\222\371\3154/M\354\235\351\320:~\246\257\306\265\340\212-\311\347`g\275\224\207E\346\260\323\372+\272\225\333\341Q\265\013<>^\233\333\016\337\263\003Ja\376\305~\024~\253\237\213O\303\347\361\221n\314\025)\300\371\371\247\200\r\370/\3776~\263\005/\215W#\003\327\"\370\006b\237\301\017\341raZ\006\255\313\337\261\002X\226\001\013\300\024l}\216\226\001Gp\365\013}\266v\003""\000\352\"\325\202#\250\372\335\270\tZ\014\377\205u\261\013X\025\253:\005[\356\312M\361-\374\003\250U\333\377\310o}\000\337\363I6\306\212X\3627\370\252\377bW\256\331\212\264\020~\310~Z\033K\242\225\375\025=\013\020<\014(\3703v\243\244\341qm\034\342{\n\352\330\021vC\357\324\372\240\237\342b\003\260\246\236ItQj\201\372\235\370Z\276\n\004^\002\277\r\276\225\217\302\200w\300\317\303\247\362\265\004\274\002~1>\227\317B\201\267@(\330b\263[\301]\371\241\177\201\005\261\274\0234/\315\314;\023\363\306\260\274:\235\025\014\357\307\306\273\021\n\272\314\336\031\313;@\207\000\320\003\024\301zs\373 \004 \024\354\005\350q\227\013a\205\361f\210.\036\260\210\371.\000xc\256\204\001;\033\032\301\3346hUm\310\352w\326C\317\343\222\355\3077p\n\371\252_\340\247\201\365\262\024d\004wt=\014\240\n\350\232\360\352\\\224\266\325\212\032\247\340\2525\277\310\207\300\232=\014S\034=\315\014!AP\347f\027$\301\033\373\007D\022\324\202\331\347\241\372m\360xO\357\201\317\177\300f\360\260\262V\353\177\376\"?\204\2623\323C\255b:\024tPVo\t:e\235\024PDH\251\375\340\001\202\370\n\374\014\254\n\002\037CK\354\275\001;\374D\337\204\361]\351|t@~\260\320C/\366\243\371:@\310\010\246^\014\217\007\326\325\266Z\006\001\336\000Y\305\226=t\271\247\007-t\276\355\216\334\216\213\327\325\216\370\007L\257)\024\276\341K\373\372'\366\202\247\275\360W\337\205=\320\013E\027\004b.\035\030`\204_\006\323\005\3110\311\310(\301\310\311\301\340\034#M1n7\214\363\304\300\030\243\244\004\003C$\225\223\223\272\271I\365\236\302\341R{\247\3163\225sJF&%\223\245VVR))\251\224I*%\002\265\333\2418G\251*\312fC\225%*&\006\345\342\202jiA\251\250\240JJN8\334i\357\323y\236r>\311\310\234HHN\"D\330\330\210\242N\024eD\361E\024e\242\010\023E\223p\357P\353\220$C\032\232\360\345%\0341\"\014I\302\320eR\241\302\264\256\246\3634\221\244\tc\223\214\214I&3\275^&\227\253\264b\005\347\245Z\255d\232\245\234K33%\027\027\022\016\016\211\307#\355v$\235\216d\263\221ZZH0\030\t\005\205TR2\312\311\031\335n#\032md\232#\214G%J\340\340\210llD:\235\310f\023\235\247\250,""E'N\210``D''\242\327K$\"\022\302\301A\201\"43\023\212\211\t\315f!\025\025PN\016(E\n\220\215\r\250w\020\212\276\274\200d2\020\014\266\256\037\235\356\223\246\237\275?$\371\201\301>$H|LL>\"\037\021\317\315\3555_\023\277\270\304\373A\307\003G\200\353\217\364\323\300\361\242J\240``P2\031\n\365B\241\\'\032\355T\226\247\231\231S\211\022\247\223\223\323\351%\342\340\210<\236h\263\211\347)\2320!\302\300\210b\212(\216\010\3034\014ma(\013\303\025SM\215\251w\323l\026\203.\250r\202\016\021\031\232\031\021\021I\222$\035\0022\204DQ\230\310\271=\362q-\253\024\221\021\021\221@DD\244$\005\005\251Q\030\346\221\010`\323#\002%;))\362j~\235*\252\260\265o-XNg\357\212\267\016t\245_*Mb\346\024D\232\377t\032n\n\367:\006L3M\227\210\366'\202\337\377\306\324\377\267x)l\234\023\342\031\266\265P\3541\251\017\236\322\240\253\tmw\255\206\230jHh\211\230\251e\320\274\357\230):0@\n'\263s\2358\376\277'\265\3139e\007S\022\355\343\363\0321*\027\367\211\342\227w\354\202\022\033\220V\3046v\357\335%\255\356h\200\257\200\256\341\240\203{\324\273<)\353\304X%\3164z&\350\206\277\365oS\245\016\210k\025\177\272\311\250\022\366j\356G\307\331\372\265~\304SC\264\251\270\3118\336J\354)\327t\307l\316\243=\373\232\315\341h\213\243\n8[\2775\035y\253i\351[\214\350%\022+\344\226a\331L\315\331\002\025|\250d\030JI\003\256\305\337\254\236<\317\343\2111\260J<\274\300\365\237\355#\310IC,\343\010\304C6S\3636\250J\266\t^\024q\2561\005\240\024\352\207\025\272\033#\362\227\304|F\365\030\322\334\316\302\305S\271\211\3463\222\263\212l\352V44e\317\034;)f\307\366\344\271\362\325\024\337Yt?[\363{\3673\265\2579\217\241\260,/\3218\257\234\367\370q\365\316\021x\302bH\344b\213\267~\013\340\340\211\003\370\023\2636sp\317\276J7/\334|n\276^Hz\326\004t\265\274X\320\005\027\004\326\003\337M\203\214X\344\300\243\240\234\370\313o\233;\207\207\010\006z\257D\373\225\030>\202l\221,\341\364\363\276\213$8\315^\223An\226?\320\033\203m\374\033\022B\263\367\250\245?\344\031\234c\260\251\3534\013Q\322A\r\016x\3447\202\2256\314\220}\224\346*\215\245>\310>""\024\356\033\003\233\013\"\302\214\226f\212\363\376\211\363\216\275\241\3517\367\245/\247\317p0\024}\353\033\312\275]c\2164\2248&\215Q@3\326\347\375\215\227a\004\250]\270NM\275\244\263\340Jo~\010:\213\203\3313\304\341[\347\010\237k\316\277\033\014{\315@\001|\353\261D\237D\177\224s\322\244#\354\026`9\301d\253\213AW\014\236\\>\212\251\320'\013\006%\236Nh\031\335T\343u\341&\007\224C\276 K\227\324\227k\267\371\204\016y\223\353$\330\013\027\305\340\333-\t\223\367[\257\256\246k$ts\204\231\337\016\006\314\260\3210lQ\370d\316\302\220\300#\213\337E\202g\234\n=\361\334Z\003\342n]\225n\007\036\306\265\307\022\265\027\371\373\265\233\035\357\013\316\214\033\221CX\246\221t%e\233\341B\311a\241C\003\026%\2674\314\254\014\345\251\003\231\373\364\251;\227$\252\210\341\335E\322\2215)\\\377\346\177W\024\326\233l\351\027|\210U\0055\017\240\"\256!\346\nk\007\374\256\n\2134;'\212\031.1\325\253*\013\007\357\305=\034\305\3418\347}\355\303\362\211\210\353;\204#\r\337\360\214^|\241u\224\020\201\325\337\304\214m\322\00008\233\021\010\247T\230k\365Q\337\250wW\232\240\361\264\327\365\217\235\310\001:/:\226\315[\212\256\032\325F?\254\277j\210\373u\261&H\030\231\013;\222\371!B\021\202\225\"\251\0060\000\303\245\306}%\277\267\311-\264\030n\321\347\370\211EM*\341\376\031P\303\027\246\370\315u\271PO\212\252\245\244\304\307\205\204a\277)y_+\220r\270\300R\265k\274\332\274\022\346T.%4E\217\250\212w\036b\007l\220T\347\243\320<\351`\021\001\310\364\315\250~t\340.\27467\026\272$m\215?\242\004\013s\013\325V\262I\263~\024\002\025\\\201p\204\321\203K\301\t}W\311\213\177\016\371\245\347b]W\314\256\022\213,d\032\255\214V:N \026<\030\320\236\352\215\027&\307\007\007\321\010{\2132\0168\361\201\263\214\220\274!\\t)\007\274\356lEFg\2450\261\211\320\0335Y5usw\016\344s[\253l\215s\264[\342Q\257\340i\026skX\007\236\302g\237\t7\324\301\373\226#v\356\217\357\004#I\244#$Dl\357\2158\250\264\026\216h\336Y\304`\353\331\024\317\346\245\305tx\004\202\017X`}\341\366\216V{p\022\201]\257\000\331l\355\355\216r\tF\353-\274hI\344""\2035\276\325\304\340\364\001`=\363h\257\001eGy\307<\016\031\300^=\277@\365=\212\355\263\307\321\371)\244Y\304|\277\206A\246\032\277\233YL)\216U\275i\245 \313\006\013Xp\000A9\032\240\313\246\205\017a\225cO\225nf}\305\324\351:?#f\006\301u9y\244\337\320\331\325@\366\004X}\024gh\177 \227eC\225x\330R\231kN\004\031\363\365\206\242\354\214^I\302vo\317\300\033\302N\343;\350\033\250\033H\034\270\242&\215\307\000\3231\264\237\265;\240qY\337\257\375\353\3104\3630qG\031N\020\313\254rn\023\022\251\322\331TT\017\020\010K]F\322\256'\030\255\026\234\247{\0242\261\033\\\271\325\033\201\254\345\007\320g\230\272\251@z\264\024D\201\200W\255\035\212bc\022U\347\211\024\"R\351\007s\237,~\213j\312v\226\"\374\036\216z\206\261\320\254\321\323\350\271\025\016\256\350\256]s.e\026?\350@\000A\350\200\265?:\314\200q\347\\\037\335\241B\370Uj\005";
    PyObject *data = __Pyx_DecompressString(cstring, 3763, 3);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) == 2 /* compression: bz2 (3977 bytes) */
const char* const cstring = "BZh91AY&SY\372\374%S\000\004\000\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\300@@@@@@@@@@@@\000@\000`\016\177-M\032w\247Z\363\316\275Q\243x5\336\354\\0i\353\245\017A\240\266\365\275 \003\335\0044\324\247\244\321=\r\0051\243F\247\251\354\221\251\351O=L\231\006T\375\002\237\244\364\232\215\232D\332OQ\342\2006\t<\223\324\330M2j\036Q\340\220jiLS\306\2314h\214F S\304\211\371D\366E\007\265\036\224\361L\217P4\000\000\000i\210\323FCj4\321\350\232l\240\320\230ML\210&\242yF\324\311\264\203C\324\331@\0324\320\006\200\r1\000\006C@h\000\000\000\001)\242\020L\232\001\032\006\204S\301\251\2656\2315M\246\223\rM6\324jzA\241\240\000\000\000\000\000\000\0104\304\300\203\000\023M4\323\000\021\200\214L\230\004\300\000F\021\200\000F\000L\2310\230\201\"\246\204F\202i\351\2413j\231\242\031\251\221\211\246i\000\3204h\000\001\240h\000\000\000\000a4\037\340@*\306\005K\313\031\206X\311\024Ap\356m\243\242\317\345\212@\203916\230\323hA\344\371\351\347\350h\232@~\344\211\016\021i\021\305\tB\004\272q\251\322\302\021\2100\246X(\203GK\343\321\251\253\251\263\261\277\305\373\362\363s\310\216;\243F\206\333q\033\206\333m\266\334h\335P\273L-\337\333\257\330\025\376\177\332;\367\357\232\220\220\204\245)\020%)L*\207\214\003\016B\004\014@\002#\021\211\020\202\210\010#\273\331\333\355\341\227\313\301\323/\233\260^\245\350\204\313M4\324\026\232hB\206\362\004?\354\371\365\366\270j\376\270\177\316%\353\335\267\235JJ\220\245)JT0\302\275\360o\324^\343\356\016\366}<J\342\271\r\030\222\020\020\204<\204>\210\023\335\307\261b\3332b\206\237\005\024a\347f\315\3230\003\2367\030\306.8b\212(\241\212(\25554o\022\304Ib\260\006\333H\033c\013y?\031\370\2717\233\212\224j\351\332\034\304^#@\374\020A\004\010\202\010\002U\234\031`\272&\343\336\343\331\3603VX\364\307\253U\276L\306RR\227\322\224\245)T5\305j\0377F\253\274\335\336b\367\374Vr\340\246\340\003tm\353\351\340nFE\032c\2168\244\216K&\006\304\021\321\331\203i;\276+\370vU\346f\230\033\322\004b\002\232i\2444#dLc\031\023\031\033\030\311\006`\0052""\010$\240\0279y\275\370~G,F\307\307\316\220\346\276\003\263\001\004\023\200\216\260\270\321\"\305\213\026,U\2568\243\362L:\031w\270!\336\303\203V\3648y6\255\213`TS\255KZ\327\n\326\265\255qy\027\200\335\2576]\250n\327]Ui\351c5f\006\307\323\275`-kZ\326\265\255k\\Q\230z\007\257q\356\017*\214R\233\344D\003\t\231\231\234:\341q\3040\223\215\310\313`\036\371\002\r\206\216\"\226\311\016)\n\027\024\036\211\3458\210\234\307i\273<\311\021\034v}+L\005\254M\202\262\214\336IRc\2665@S\2464\250\211\031\206\177b1\214\343>\214\271p\371\327x\252B\002\362t\206l\331:\3423\312\352\252\256\260\"\201\312__6;\334\tJ\354\204Z\263E\212\303PA\356\202\031\306x\360G#\272\375\321Y\226|AkE=\357\337\212\037\243\363L\374=\031\257?\024\030\r\252 \014\000Y\376\324\310\021\260[o\020}a\3438\262%\351\365j\322\357v\323,`\306\361\003\232\231\350Y|\355\207\226\365\030\231\326\317\036\005\216\013\246\036\343\027\347\2755F2\221\362^\211R\227b\256\332D \311\352\032t\204+\270\246i\240f\240\321\001\r\354fPY\232*\212\031\333J\202\244 \311\227X\265pU\316$\336\030\314 \215b\354`k\247U\036\024j\004\t a\302Y\323\331\002y<\221\266\030\004\261\250\022\0369%\221GV\n\2012'\254\021\347[GO\355\367\035\366\254\267%\002\346\314\366b\010\031@\215\265\016p\177\303h\317\017\037G\250\375\316\020\377x\203\t(7f\273bQ\326\030\022\033\317\246\372\307z\233\300U\227\226\243W\223\346\360\267\333~\022\274\327\032\010\215\234\010tX\r5\267<\343\030\232\314\033\217\274\354\274\257\323\362\372\305\207\234z\006\342u\337\032-\013N\"\003\365e\211\315eC\265%7DH\260\372\013\212\236c\373X\0360Gf\215d\377\275\376\333\003\317p\330\215\256\326!\2661[q\300&v\300\260\241'K\326\324l\354(1\004\010h\376u@/\201Fa\370 $@\375\204o4\033\331,\371_\212\243D\034\263\346\267\256\373\336s\027\246[\235~\246\331\343\002\321\215v1\001\346\0051\334\355H{\026\232TE\320-\375\002\247\337\002\020\220\233\320?S\342\020C,\350B\034`\3257W\373\027P\341\306\211\014C\266\3453L\2634\315\332\333\331\361\355j\307\220\304\356\243\276q\206Yi\224\322\342\021\322W\013|\211\016\276""\375\360\203>wh @\253,\036\215\324k\007C\235\302Y<E\021\353(\376\034\026~\365Oa\250O\225\346D\016\262\341\3455\257\335o\230N\0262\237\004}`+\260d8\222\310\316\305\231[ak\022\311d\265:\003\254\304u\204\234T\363\345\313)2DD\312\2659\241i\277\322=\270>\332hnm\2273v\213\300\317#\277\036\004Nj\220\304\320\252s\016\312\324\266*\364\374\357\0321\266[\034\201\374%AVYK\tBW\261\302W6*\212\211\203\330\311W\246\265\202t\236\376\314\027p\037\006\205])\237\002\275\267\344FGk*\212\247\341x\014\2579\204\265\211Z$\022\276\036\305;\271\313;\367\036k\232U\224\344\343{\021\333lZ\005t\244$\226\"\260P\0262o~\0136z\356FR\023\n\343\243K\001q\315\345\337\271\304E\264n\016C\316xC\217a[\n~u\"\256T\211\364\212$)\021h\024Y\374\250)\3166c\342S\254^{S\233\330\325\340vJ\207\327\214\332\225\324`\324\013\312@\207c\267%3\330VG3N\200\326\231K\230/&\222!\tq\224\242\231\371\271\270L/\253Up\243\006\303\326\246ZP\245\264\316`\224MN\023\336A\216:)\254\265\020\346\301\354\217\026\3255\025W\203\022\310\014\302\233\364)MeB\323}9\332H\216\246\200X\327)\241\0053q!\031\214\342j\327\021#8\327\255kJ\240\234\035(\234\216\006\351\355\036\351\356\237a\272@X\016i\213\345sO\270\252\222\236\347\301`H:\244DnH\016c\\q\311\310u\005\035\214i\035^\233A\200\3016Eg\003\332\232J\216v\210\324E\275!\223\320\255\245\010g-\246\200\"P\200S\005kiU\254\200\031\303?\261\360z\216\2571\017\341G\214\366\223J\214\024\211\236\224\336~I\272\225\235\217]v\327\246\275\377j\342\333W|]\242\236>n\253e8\217\"\262+\343M1\212\204\355\234F<\263\235\202\23625\222\355\300\2635\027-\333\010\233\314-\005\301(\345\310|\221\213\222\3135\303\002\n}\263\002\311Cv5\203\306\326;[\215m,\032\211+\204\315:\273\202\244-\344\366\2227]\305\016\325\263\260X<C\030\211\342\213\02107M\275\315=+\216\253\316\333\243$(0C\220\354:\023?\367\211JP\356\374\221\020\361\310H\226\232#n\231 \332\236\212\331\234\221;Y\242O\321\"h\016i \214\351Y#\330=\243\236\321\316&m\362h\306z#\315j\370\234_D\370\363R\005\250\355\366n\276\3453\277W.\313;\253\261S\037\037\013a`d\246Wd\311\247\200""\203\203\002\230\240v33\2706:\"QPV^\311JQ\350\357\025\005\343\020;BK\211\261=J\010P\010\003\215\227R1\305\254\\#[\242\014\350\272\"\207\233\010Y!\240\301\036Y\231\230\001\241P)\231!\246.\264\320 \254hM\357\325\312\272\374\003SP$\312\006bv\020\205\256b\234tbU\335\223m\333\001\220\207\243\364\222\315\307\254\303$)\212Z\266\223v\253\317\206\212F\235\266\026)\340\254\315aK\225\255z\004\204\255wAn%\210\270Ft\310\322\324\211\036\272\244v+\033\274\034>z\363\227:\363.e\352.\255\323\342\343g\271\365G\001\242m\006\370>mU\246(\312R\321&i\023\342\202\t\321\236\374\353\235e\265\206\256\345\317\001\276\326\010\336\362A\223\240\245K\273\323\350\346\345\311\346\\\033F\t\231\222\343\031\036H\031\330Y\0229\241!\304\314\033\220\355\221\271P\277\244\263\024\345\177-\351\357\033\\{v\205Ne`\314R\266\361\001!\2359\340\244n8\333\020\310\332<\364\032\340w\334\213\273\332+nS\366\273\326P\023\350\266\3258\315\224tjA\032I\023\354 \265x\264\333^\336\323\014iw\n\302\263\270(\320\304\257\371\0358\005\330-\304\220\300b\335h+\004\031\0233\316\031\240g\034C\225\340\271w\024W\256\031h\3248\323\020M(.6v\005\252M\263\266\354dC\370:\016b\363\333Z\357\355\2262\356\265Xr\021\256\031*\207\037\r\002\224\352\214\r\003n\364\017\303\250\304\007\261\226\372lDUA\265\203\237[R\351\341qq]m\305\336\266V\211\335+\r\307\255\202\325-0\316\331{\216\025\357ZD\300\341\277\231vWeqc\305m\361\r\020$?\303wa$\240\276\255\354pI%L\260,\221\rk\200\350\204Js(\023:\nIl\320\211\214\024\000u0\0220\242\021\027\006\265\\\001\211\367b\321\244\202\3259\022\206\033\3430\200\266\303S\317\304\212\254\307\334\336\301\r\277S\347\305s\346a36\371\304\204\221g\036F\252\302\277Z\307k)\306\325\241\256\364\325(.\272\301\231\010\304\244xe2lua\317\231\306\362\246\r\325z\340c\013\257\306\213\033\033UM\256\264\366\2647\036m\270B\033\377/a\255F\231\371\246\346>\033\243\341=y.\3414\315\274!\300G\025C\336Q\024f+\213\250\254\2109F\004\316\342\341\343<7\220\362k!VH\303L\3118r\0352\224\230\207%(\365\253\361\347~\352\347\200\366\2708\210\020\241\316\244\233d""\207\224\271{.\033\252\202r\276#\221\375\311\312\261\2123\360\275\t\310\215\215\300\336\016\355\201T\2331a\2111\342\032_C\257\212\025e\341\310\222\234\247\014V\306Xp\".b\257u\205\231\274u\300P\266\314\305\311z\002\235\327@\235\242:d4\026J\260hL\222H[{8\233\331\325\352\257*\365n\345\004^\257\013\244\374\033\3740I\210z:xr\250\021*\245\022tf\340v\252\240\336\t\251\003t\256\310Q\276\330!\230\344\310:\356\023\243p\021\300y\314\232\354\207\205\327\301^U\315}\371X\265\2057r\373\326P\243\216\023\261{89\337\341\263\027N(!3\272(\2363<\252}FV\004\211\031\264\263\334,\262}\355\356]}\375\253\030\261[\014'\007\036\303\016\210\211\006\320\372u\223\254\353\240\353\3139u.1E\205\263\234\313*$L\224=P;W:\250n\321J\255\313\343KQ&>:\031\256Mf\365\227\251\241\231)n\211\234.\251[&q\367\r\312\n\360\224\023L\005\354\311/\035L\016\007\020+\277\320\241\262d\222+0T0am|!\0073\325\305\214\367\355|\t\010\003\010(W\304P\220T\247\3309\211\214S\033.\240\211*Tw\356\376\232=\326\246\n.\226\004\212\024\246iN\035l\364E\034}E\016\035m6P\205\272\373\000\270z\353^*\212\311\234\221\216\210/,\266bB.\255\240\020\333t\032\207\246p8\034]f\227\236sT\365\302\322Z_\024\002\004\214\316\003\267:\030\235KJ<\"\257Hl\350\031\300\360 \340I\035\345\213saG\001\252\252\004\363*\270\275\025z\212\203AvU\264\275y\216\265\312,\003\234\350\212b\264\342\350\312\302T\346\302\242\031\205\371\226D\261\001Q\212\362B\232+,\230\346\275Y%0\206\252\267?\n_z\330t\021f\230\306\"\225~X\257Z\265\021\305m\207\226,f\003\320\032\036U$\310Y\"\0034TW\266\321\010a\232_o\032\354\343\014\r7\205O\276\313y\263$\220$\350)@\024\035\306\252\325;g\320%\177\005d\023\361\004v\000W\204\207\277\205S\377\020S\222\223\241\377^\350\016\361\027@\377\205`\203\223\326s8\306\262>RA\277\343\307\010:\367\024(n(\361\215\273J\254\230/\207\273\261\355\327\203\303%\347\334\000\353\300\313\014\005/\207\023\216p4\264P\251\367\223\024\0178{\271\204\316\025\037.\002\330c\270\000\247\220T\004R\035b\315,\352h\234\340\233%f\217@\tO\314\254\365}\203\017~\246\200^'\000#\210\242T;""\331\034\216\324\004.\245@>m\014\372x\264\303\030\254*\243\272o(\315\353{\014\032\017\376Y_\340\200\207\306,\2122@\364\202\227\372\233\3010\n\220O\276S\253#x\007\000vkO\200\271\355\"\005\345\3528\014\352\277!0>\031SUmZ\327k\276\022\214Z\342\304\030\371\016>\033\022\2558\025\221\324\035\326\323\027\303\244cH.}\337T\321\354u\320\2559<\023t\2378\232\3527\325d\307\260\225t\001\233\007\222\263\355Y{\343\\o\017Vu\335\330\225`sV\272\306\274\020\250\314\344\347\267\243\323bO\301\201F\305\337n*\2253c\241\353\251C3\2376\334\337\350\010\303\323\027N\326M\033\022\341z\254\334m%\014\370!\005\360S;\204n(\002\003\325\366\321.g\372\354\\e\376m\320\362\351\213\026\260g\254h\302W\264H\35735\251\354\365\016XyS\203U\345\231\3068\n\263\000\030\254R\222\245\260UXTMc\355:\014L\360\"\277\265\215\373\332\263\035\207\034\344\327\n\021\230D\241\27010\221\020\255\035\244J\306x(\253\252c\322P@\207\206C\304<*\235\r\330\"\262\002\nl\212\031\335\030\314 Q\002\024\314\321\002\003GQA\000\323k\245\034>\036\346\336\336\232r\366\352\032\254\321H\343\035u\343\334\033\227\347\326\353{V\257\006\365\245\235\353\021\350\3418\273\315a\242\255\\\031\326\"$\327+\337\013\264\213V.\221u\2340\276\027[e\2050\302\366\244a/\215]\231\037\030\233=W\032\333\005\255\353\r\205\327\013l\345[Z\367\275l\242\325\276,\366\312\322)K\t\212.\030#`)\031a\310\210\227\232L\211HE\233\350\326-\032\315d\265\306\333\3165\320!7\377\027rE8P\220\372\374%S";
    PyObject *data = __Pyx_DecompressString(cstring, 3977, 2);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) != 0 /* compression: zlib (3776 bytes) */
const char* const cstring = "x\332\315\227M[\023Y\026\307yI\310{\000\305\326n\035\255\264v\243\216\215\215\332j\333\3353\006\010H\213A\010/\302\2005\225J%)\222T%U\025 \330\210\313,kY\313Zf\231e\226,Y\3262K>\202\037a\316\377V\202\250\3303\363\314f|\244~\367\375\236{\3569\347\236<\341&\204\014W\0264]\272\303\355H\234\264[\226DC\312p\206\312IJ\206S\025nC\271\303)\252\361\204\233T\253\305\014\212\\V\246.C\320r\222\201\271\222bp\262\302\355\320\230Dw\276.+\271\242\304\211yA\023\250A\343tC\243&6\375;mZ\326t\343tg\236\255\235\2268\245Z,\272\333\315\320HZY\253\375\372\303\337\312\202\221\347J\262^\022\0141\317eU\232A\373\310YY\344X\327\013Y\307\206\356\3703\372;B&UC\342\214\274`p\2235#Og\223u.#\025\345\264\244\t\206T\2541!\231@4H\341^&^\376\360\340\361\003N\240\323j\322\026\035L\347\364jZ,\n\272.\351\234\232\345\322U\271h\320\321\215ZY\322\307\270\331,WS\253\234\"\271\n,\323\270\323\023\214\274\244p:\351\214\n\334\250\240\3201\005CV\025\236\246\223\364\243\\F\326h\023y[\302\354i\241\250Kc\313\212^-\227U\215\335\211\253qR\276\264\313}\227Y%\365~r%\264n\347\250\256&H4w4\215\325$\301\035\261\243\251\270\210j\211\216\215C`\250\014\361W?\276\1776\201\273{\207\033\373\273\220\311\360\264\213\224\326$i\2576\346\342nzO\273\313gp\225\244=>/\025\313\222\246\363\345\332\356\030\375u\207\356ic\335!\247\232de\233\266U\265\232Am\235vU\257\032rQ\277CZ5:\233\243\244\253UM\224\270\214J\032\304I\245]Y7\202\354|OHc\272\220.J\031U\0315\270\202\242\356py\372#\341E\265\004Mtf\363%Y\221\013\244\212\337$\005\343]\355\214v\355E\220\213\356\221]-jT\2075%\227\347\346\270\233\033?\336\032\343\226\272m9AK\0139\351\t\0311\2633(\264#KN\244\333\341\025i\327\270y\213\023\205\"\326\334\301\225\213U\215'\313\202\265a\311/\214\242\225H^\310\254\250\3145t\256(e\215QL\023UE\2273\324\017CT\204\0221\253\251%\352r\317\223\tv\256]\315>\341Fa\362\343\243\\\251J>F.%p\351\232A\312s=\2209\363\031c\312E\201\254\345\2631\367\376\203u\356\375\371:\314\005yY\311\302\244\364j\321\300\211\320\221T\025\246\310;\035\313\376pK7a\026tD!W\242S\335B\355\007M""\312\222\002\024Q\202\265\262\333\022\311H\014\354\331\211-;2\271\272\000\223\337\026\2122b\320\036\326\317T\277\340C\360U&8sT\022\335\r\nE\262/v\267\354p\372\250\033\216\376\355`\262\362tQ\025\013_\232\320\021\022ci(.\361\337\216c\002t\226\2438=Mf\372\234$Oh\232\252M\354i\214S\262\226\202sMu\234lR\325\264j\331H\314&W\342s\211\344|\"\271D\337\245\251\331\305\331$\377\"\361b~q\215_N\276\230\237\232\235\236ML\315v\375p\211T<IA/'\301DSK\361%\220\177\031_\244\005\370\251\304R|v.\365RSEI\327\023\260\367\311\217*c<O\306Y%O\023\231^x\376\323n\212{n\2348{\200LQ\227\027\231\000:\365\325v\351o\212b1\237$OY\224\262\213\024\017$\315\375~\276\331I\373g\273t{\330\023\307\237\334Q\212\237M\274JL\022R\213\211\031!M~d\320\027\372\246X'\3505E\224\3251Q\325T\212I\212\244\247\357\t|Z\320\245\207\017\322\262\"\350\242,\247\311\367D\303\375\236\254\333\251\302o\371N\231\031\377g\r\232\234\313\033\247\342\341\231aT\344EQ\020\363\022\317\263\027\204'\303%5\225P%\231\310\243x\nU\242\224\026\304\202\250f$Q\247@\203X#\222\t\261\230C\366OD`\340\231x\335J\307\376x\261j\250\331,o\310\024Ph>\255\301S\037\311\330A\367X<\023\355Dy\037\324\350>Xd>\374\207\222\254\363\360W\376\304_;\273\235lJ\322u\313\220\360\344\3444\226$p\277T\311\360j\226\264\201\353\243\232^\325$\367\020|\347\275\342\311C\2302Y\370e\037\236\336C\205\014IST\t\276\241w\2373\376Tw\226\274\010\177\274\234a@\300\341\021M\331)\371\022vG\007\023\020\017Bg[\0262\334\201\005\251\226\245\2452Y6\203\317V\025\221T\206\010\366A\302Zg\002Z\335:\212\356I\371\274L\377eE,Vi3\232\315\354>\303\323\024\003\237So#\324yb\206\250H\273t\247\006\202>Ud\366,\234\024xA\251\301B\364\323\336\204\363\321irF\276\250\362E\265\010\021\212$)\2733\235/v,\322\255\273\362\363%\212\341<\240\025\240\020J\244H\272\222L\022R\255JJ\243\177L?`\347\273\343~\330\nx\336x\3120N\356\n\3452E\177X\237\373T\351h\302\tu5\275\305\253\344\201\3326m\242\347\205qU\357\244\002e\241c{\356\224\023#\303\212\23784\316\300\3364\366 \261\023\301\320X\001V\306\nU#\373\030\205-UVXl\247\307""\312 3*\225\3252I\227\225w\311\355\370\222T\322\304\274\306\303\005I\211\022\255N\242\262\232\373~\271\345\316U\242\210\364\317-U\225\262,\026HA\037E\267O\372\334\230\3446n\263\253\204\006+U\241\350\352\024\271\217\373\371\340r\254Jn_\320:\223;\021\220\377,\026\2364\220\361Q\221\246\303w5:\253l\350<%\242\002\235\216\357\346\307\314\324\251Q*\350R1\313\002h\347f\251\224\221\262\002;\357IX\345\317\010\261\354\316\330:n\251\254\251\333HVP;\035bPe\273\271)\031s\017&\025\357~5U=\025r\310\025T\262|rg\262\0256\200\335\230\301\213l-\032)m\323WVT\335uZ\240\323\205\013\207\214\314=){gr\261\n%\006U\311\025\337\3752\201(,\2529\252I\272(P\"_+A\321zM\357\244%\256#\363\006\226\246\357\256\001\263\255\22634\303\375\272\016^\325;\313\321D\212\322\010O\314Y\251\271\2533\030 \375\361\217y7\334\272\342\360;B\261\300\234\021\375;\002Yz\225b\200\246\323o\003)\263#+\367\357\321\207E\264\340]<\027\357z\333\236\2776\306\033\323\315\330\273\336\367\017{\274\376w\206\023\272b\305\255\205\266?Z\227\315\212\325\333\276|\315\252\330\376\006\r\365\241\367/\326\202%|\322\353\263+\307\236`\375z}\241.\264\251\020;\366\0149C\234\335k\017\267=Cf\030\013b\300\215z\332\354m\373\207L\2179aJ\326}\266\220\023\275\331\350m\\lR\307y\347\374\255\306p#\306Jh\035n\373CNh\324\2564\202\315X\373\346\375\346/\255\364ao;|\316\374\316\324\254\341c*\304\332\236P}\374\377]\374Ps\274}\363^\363vk\242%@\374\330q\367\014\035\361\333\236\033v\334^i\214\277\213\277\037\351\t\\0\247\254\001\347\352\375\346\375f\272EK_\263\014\373ac\370\375@O R\237\246\351\376K$\275\177\004\2028\227\3574\022M\332+L=cv/NG\222}c\215\222\3744$\336\366\177e\n\357\317\365\004\006\353\031\363\2665ae\354X;<B[\204\354q\373y\243\342\374\030?\034>\244\266\301z\345\030\007o\017\271z\2726\352\214>\246\375\303_[\303\326]:O\370\033\353\216\235&m\372C\365_M\335\212\3410w\355a\373\206-\264G.\034_\374\241\261\300\004q\316\375\324\\xG[\323\222t9\337\367x\207\234\301\030\r\274\335\210\343*\264\372\371\272\333\333\366\234#\241pg\036R\346\367\266\340\214>=\354=""\274tt\321y\271\344,-\267=\027\234\013\320\353\035\347\341\374Q\305y\271\346\254m8\033\233\316\346k,T3\373\315\237,\2175\337\270\337\240S\2377\307\333\236\257\235\257\0374\343\3155\322\265\347+\347\253\361\346p\363V\213\332/:\027\3575c\315\237[$\302%\347\322\375\346xs\266EW\033qo\376\014\274\367\367x\003\365\200y\036Z\014\327\347\254~k\034\227\020\256\377lN\232\225v\230\366cu't\231\014b\2134\303\232h\332\010$a\342\270B\216\267\375\027\314%\353\032\331\3057d\r\267nC\207\017\352o\255\024\351;L\006\027\263\331\035\376nn\331}L\237\341!3h\335\266'\354\014\331\024Uh\r3A\027q\303J\263)\246\327\\0\323V\257u\316\032\267&-\335\216\321\276\221+\326\003\353M#\326\030o\207\207\315a\363[3n\256X\017\355\253\215\275\326p\3732\031\343\373hO\344\334\251\236\013\366V3\330\212\265/_\267i\316ek\274\0358g^7\351\036\241\0022y\363+s\333Z\266\257\333\253tw$\364\337\255\353\326\002\266\372\306\372\213]\242\263\014\2158#\210#d\205\315oI\361K\255\013-\241\245\221M\3358d\307\210Z\202\265O\223\257\301\027\274\215\205F\272\331\333\034i.\267b\255{\255\205\226\340\312\375a\343\356\036A\267\331\177\3214\254_l\301\2565{\377K\255\235\241\202om\362\276\341O\024\320\213\366\2334\203n\253\342\272\331<i\225\231%s\320c\302\023\322\267@\326L\267\033\255/\231\227\255);\324x\320\330ifZ7\351\250\271\243%g\201\354Rp\204t\333\023\256\317\320}M\223\277\010Xt\273\276J\2277\202\320\204\240EG\252\034\263s=\262G\033\027\032\"\251c\245\365\350\320\025\200\034\303\031\034\265\r\322\351\014\305\274\334\341\302\361\307-\037\244p\206\3576\214\346\337\016G\016_;\251Ug5\347\344\n\357{z\212\275\361>B\274/\001$\372\326\200\265\276M`\263\257\014\224\373\014\300\350\233\351'\314\364?\007\236\367\257\003\353\375\257\201\327\375U\240\332_\003j\375\317<\204g\2369`\316\263\005ly\024@\361l\003\333\236=`\317\223\364\022\222\336E`\321\233\003r\336\002P\360>\035 <\035\230\002\246\0066\201\315\001\001\020\006\366\200\275\201\267\300\333\201g>\354\347\233\003\346|k\300\232o\023\330\364\345\201\274\257\010\024}o\2007\276\003\340\3007\357'""\314\373S@\312\237\006\322\376,\220\365\327\200\232\177\037\330\367\277\014\020^\006\226\200\245\200\016\350\201\035`'0\037\304*\301\024\220\nn\001[A\005P\202\317B\220,4\007\314\205V\201\325\320?\200\177\204\262@6\264\005l\205j@-\264\017\354\207\022a\\Gx\026\230\r\277\002^\2057\200\215\260\010\210\341\034\220\013\307#\270\277H\002HDR@*\262\n\254F^E\376c\373r.?n\371Z\325\303\311C\343\3507\347\025\357\360\031'\003#q\n{PY\357\004\214`\242o\006\230\351\373\035\370\275/\t$\373\322@\272O\006\344>\005P\372*@\245\257\nT\273\306\223\004\222\375) \325\277\002\254tM\251\010\024\373u@\357\337\005v\373\337\000o\372\017\200\203\376\027\260\232\027\236%`\311\263\006\254y6\200\r\317?\201\177v\355k\037\330\367L\300\224&\274\t \341\235\005f\275\253\300\252\3675\360\332+\002\2427\013d\275[\300\226\267\002T\274\273\300\256w\037\330\377\304 _\001\257\006x\200\037\310\000\231\201\034\220\033(\000\205\201\003\340``\032\3266\355{\016<\367%\201\244o\021X\364m\001[\276\nP\361m\003\333\276\032P\363\355\003\373\276e\230\340\262\177\003\330\360\013\200\340\317\000\031\177\036\310\3735@\353\332\354[\340\255?\016\363\214\007\022@\"\220\002R\201u`=\300\003| \r\244\003Y \033\250\002\325\300\037\300\037\2018,8\036\234\002\246\202\317\200g\301\025`%\270\tl\006\323@:(\001RP\006\344\340\014,x&\224\004\222\241\024\220\n\255\000+\241u`=T\001*\241]`\267k\362Oa\317O\303S\300Tx\035X\017\013\200\020\316\002\331\260\014\310\341\022P\nO\303\310\247#/\200\027\221E`1\262\014,G\326\200\265\210\010\210\221-`+\242\002jD\003\264\3106\260\035y\032\305\266\321\031`&:\007\314E\347\201\371h\nHE\327\201\365\250\000\010\321,\220\215\312\200\034-\001\245\250\016\350\321=`/z\000\034D'\006a|\203\323\300\364 \017\360\203Y ;X\000\n\203\n\240\014j\2006x\000\034\014N\017a\302\320\314\320\207 ?\326\310PF\362\250\365\327\303\325\243\370\361\307-\013\024\344/\365x\203\365+\346\262\025\263~\242\027\217\036\241\237(!M\330H\005)\374\257\260\020@\tk\320\t\342\345\366\004\234\300\025\353\231\035wzn6\006\032r\263\202\324\353\213yq\330M""\306\276\220\027\007\335\347PGf\214.\221\336L\312<)\273D6L}\224\346R\316`\307\221\347\374J\023z\317\316\222\361\234\016:Qz\373\333C\267)3\371\245\231k-|!]\246\254\262b\005)KL\266\306\216\372\216bg\346\372\307\036\022\317\341~l\206Z\277\034\026\234\205E\010\360\320\272D\351%\223\370\017\363\021e\246~\254\371us\354\260\367\317\246\375\206\037\007\264\365\207iw\221\020?=\214\036\t\335\237\026t\ta&X?\035Gtb\317\217\356\035-0]#/\216\235.E\352\033(\034Sx~N\023<\224m\013\224\273\217\273!:\005\315\260\2347T\177d~o\365!G\247\r\256t\356\350\254[\370\363\337.\324K{9a:\253s\353\347\3260\345q\267(\363n\n\355\323\265<~6\321>W\377\227}\360\344L\220\021\\\261\026\234\253c\r\241Qi_\275\315\022\315\217jB&[\324\376\005V\356\017W";
    PyObject *data = __Pyx_DecompressString(cstring, 3776, 1);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #else /* compression: none (7184 bytes) */
const char* const bytes = ": Bad parse, we expected to end on \\n, not: Could not find target parent in wt: Expected single character string not %rFirst character should be null not: Got entry<->path mismatch for specific path Missing entry for specific path parent Note that Cython is deliberately stricter than PEP-484 and rejects subclasses of builtin types. If you need to pass subclasses then set the 'annotation_typing' directive to False.Unsupported target index %dWe could not find the parent entry in index We read the wrong number of entries. We expected to read  /, .?add_notebreezy.breezy/bzr/_dirstate_helpers_pyx.pyxbreezy.bzr.dirstatebreezy.bzr.inventorytreebreezy.osutils, but read  but source does not exist\nentry: disabledon't know how to compare source_minikind=enableentry ' entry failed to find trailing NULL (\\0). Trailing garbage: %r for the entry: gcget_next() called when cur_str is NULLget_next() called when there are no chars left' is considered renamed from isenabled\nparent of: 'path1' must be a bytes string, not 'path1' must be a plain string, not 'path2' must be a bytes string, not 'path2' must be a plain string, not  path_info result is not None: %r, target_minikind=(tree fragment)tree-referencetried to create a string with an invalid size: %dunsupported target indexyou must pass a python list for 'paths' not: you must pass a python list for 'dirblocks' not: you must pass a string for dirname not: you must pass a string for 'path' not: BadFileKindErrorBzrErrorDirStateDirstateCorruptEINVALENOENTENOTDIRIN_MEMORY_UNMODIFIEDInventoryTreeChangeNULLSTATNULL_PARENT_DETAILSProcessEntryCProcessEntryC.__reduce_cython__ProcessEntryC.__setstate_cython__ProcessEntryC.iter_changes__Pyx_PyDict_NextRefReaderReader.__reduce_cython__Reader.__setstate_cython__Reader._parse_dirblocksS_IEXECS_ISREGabsentabspathaddasyncio.coroutinesb2a_base64binasciibisectbisect_dirblockbisect_left_bisect_path_left_bisect_path_rightbreezy.bzr._dirstate_helpers_pyxc_ccache__class_getitem__cline_in_tracebackcode""cscurcur_cstrcur_sizecurrent_blockcurrent_dirname_cutoff_timeddecode__dict___dict_dirblock_statedirblocks_dirblocksdirectory_directory_is_tree_referencedirnamedirname_cstrdirname_sizedirstate_encodeencode_end_of_header_ensure_block_entries_for_pathentryentry_counterrnoerrorsexpected_entry_countfilefile_idfile_kind_from_stat_mode_filename_find_block_index_from_keyfoundfstat__func___get_block_entry_index_get_entry__getstate__hi_hiinclude_unchanged__int__inventorytree_is_coroutine_is_executableis_insideis_inside_anyitemsiter_changeskindlengthlo_lolstatlt_by_dirs_lt_path_by_dirblock__main___mark_modified_mid__module____name____ne____new__new_blocknext_num_entries_num_present_parentsnum_treesobj_observed_sha1ososutilspack_statparent_directories_parse_dirblockspathpath1path2path_cstrpath_sizepath_utf8pathjoinpathsplatformpopprefix_py_memrchr__pyx_checksum__pyx_result__pyx_state__pyx_type__pyx_unpickle_ProcessEntryC__pyx_unpickle_Reader__pyx_vtable____qualname__read_read_dirblocks_read_linkreader__reduce____reduce_cython____reduce_ex__relocatedrsplits_ssearch_specific_filesseekself__set_name__setdefault__setstate____setstate_cython___sha1_file_sha1_provider_sha_cutoff_timesha_filesource_indexsplit_split_root_dirblock_into_contentssplitpathst_ctimest_devst_inost_modest_mtimest_sizestatstat_and_sha1stat_valuestate_state_filesurrogateescapesymlinksystarget_index__test__texttreeupdateupdate_entryuse_filesystem_for_execuse_setstateutf8utf_8_decodevalues_walkdirs_utf8want_unversionedwin32winerror\n/.bzr\200\001\330\004+\2501\250F\260!\200\001\3606\000\005\010\200t\320\013\034\230A\230Q\330\010\016\210i\220q\230\001\330\033\037\230q\240\010\250\001\330\004\007\200t\320\013\035\230Q\230a\330\010\016\210i\220q\230\001\330\033\037\230q\240\007\240q\340\004\n\210#\210Q\210a\330\004\n\210!\340\004\020\320\020 \240\001\240\021\330\004\020\220\014\230A\230Q\340\004\n\210$\210b\220\001\330\010\020\220\004\220B\220e\2303\230a\330\010\016\320\016(\250\001\250\027\260\001\330\010\023\320""\023)\250\021\250!\330\010\023\320\023(\250\001\250\021\330\010\013\320\013'\240q\250\n\260!\330(3\260;\270b\300\001\330\014\022\220%\220r\230\021\340\014\022\220!\330\004\013\2101\200\001\3606\000\005\010\200t\320\013\034\230A\230Q\330\010\016\210i\220q\230\001\330\033\037\230q\240\010\250\001\330\004\007\200t\320\013\035\230Q\230a\330\010\016\210i\220q\230\001\330\033\037\230q\240\007\240q\340\004\n\210#\210Q\210a\330\004\n\210!\340\004\020\320\020 \240\001\240\021\330\004\020\220\014\230A\230Q\340\004\n\210$\210b\220\001\330\010\020\220\004\220B\220e\2303\230a\330\010\016\320\016(\250\001\250\027\260\001\330\010\023\320\023)\250\021\250!\330\010\023\320\023(\250\001\250\021\330\010\013\320\013'\240q\250\013\2601\330(2\260*\270B\270a\330\014\022\220!\340\014\022\220%\220r\230\021\330\004\013\2101\200\001\330\004$\240A\240V\2501\200A\360\024\000\t\025\220D\230\006\320\0363\2603\260b\270\001\330\010\037\230t\2406\250\021\360\006\000\t\r\210F\220!\340\010\030\230\001\330\010\024\220B\220e\320\033,\250E\260\021\330\010\014\210F\220.\240\001\330\010\016\210a\330\010\032\230'\240\021\330\010\024\220A\330\010\026\220a\360\022\000\t\017\210d\220*\230B\230d\240!\330\014\024\220D\230\013\2401\240K\250q\3200A\300\021\300!\330\014\017\210q\340\020 \240\001\330\020\035\230Q\230a\330\037'\320'8\270\001\330\014\031\230\021\230/\250\021\330\014\032\230,\240b\250\001\330\010\013\210<\220s\230!\330\014\022\220/\240\021\240$\240a\330\024\025\340\027-\250Q\330\010\014\210F\320\0225\260Q\200A\330\010\017\210q\200\001\360&\000\005\020\320\017!\240\021\240*\250A\330\004\007\200r\210\023\210A\330\010\017\210q\330\004\022\220*\230A\230Q\330\004\016\320\016&\240a\320'@\300\001\300\030\310\027\320PT\320TU\330\004\025\320\025)\250\021\250,\3206O\310q\320PY\320Y\\\320\\]\320]^\330\004\007\200y\220\003\2205\230\004\230O\2503\250a\330\010\023\2201\330\004\031\320\0314\260A\260Y\270a\330\004\026\320\0261\260\021\260)\2701\330\004\027\320\0272\260!\2609\270A\330\004\030\320\0303\2601\260I\270Q""\340\004\r\210Q\210a\330\004\r\210Q\210a\330\004\r\210Q\210a\330\004\r\210Q\210a\360\010\000\005\t\210\t\220\023\220A\330\010\014\210L\230\003\2301\360\006\000\t\014\2109\220C\220q\330\014\023\2201\360\006\000\t\014\320\013\033\2303\230j\250\001\330\014\023\2201\360\010\000\005\024\2201\330\004\023\2201\330\004\007\200y\220\003\2201\330\010\025\220T\230\037\250\001\250\032\2601\330)*\330\010\013\2104\210~\230S\240\001\330\014\020\320\020!\240\021\330\010\014\210J\220j\240\002\240$\240a\330\014\020\220\n\230*\240B\240d\250!\330\014\020\220\003\2201\220E\230\021\230$\230b\240\001\330\014\020\220\005\220Q\220b\230\001\230\022\2301\230C\230s\240!\360\010\000\r\034\2304\230{\250!\2501\330\014\021\220\021\220\"\220A\220V\2306\240\036\250z\270\021\330\033'\240q\360\016\000\r\022\220\021\220\"\220A\220V\2306\240\025\240j\260\n\270!\330\033#\2401\330\014\033\2301\330\t\022\220#\220Q\330\010\r\210Q\210b\220\001\220\026\220v\230U\240#\240W\250A\330\010\013\210?\230#\230Q\360\010\000\r\032\230\035\240m\2601\330\020\024\320\024+\2501\250E\260\021\260\"\260A\260T\270\025\270a\270r\300\021\300$\300a\330\014\020\220\016\230a\230}\250A\330\037'\240q\250\005\250Q\250b\260\001\260\024\260U\270!\2702\270Q\270a\360\010\000\r\034\2301\330\t\022\220#\220Q\330\010\013\210?\230#\230Q\360\n\000\r\034\2301\330\010\027\220t\230;\240a\240y\260\001\330\010\013\2104\210~\230S\240\001\330\014\020\320\020!\240\021\330\010\014\210J\220j\240\002\240$\240a\330\014\020\220\n\230*\240B\240d\250!\330\014\021\220\021\220\"\220A\220V\2306\240\036\250z\270\021\330\033\"\240!\340\014\021\220\021\220\"\220A\220V\2306\240\025\240j\260\001\330\033\"\240(\250!\330\004\007\200q\360\006\000\t\r\210O\2301\230A\230Q\330\004\013\2101\200\001\340\004\013\210:\220Q\220a\200\001\360\010\000\005\016\210T\220\033\230D\240\013\2504\250w\260d\270(\300$\300g\310T\320Q]\320]a\320ab\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220t\2307\240'\250\025""\250c\260\024\260V\2707\300!\330\004\007\200q\330\010\017\320\017'\240t\2501\250G\260;\270g\300Q\340\010\017\320\017'\240t\2501\250G\260;\270a\200\001\360\010\000\005\016\210T\320\021/\250t\260>\300\024\300^\320SW\320Wg\320gk\360\000\000l\001A\002\360\000\000A\002E\002\360\000\000E\002Y\002\360\000\000Y\002]\002\360\000\000]\002p\002\360\000\000p\002t\002\360\000\000t\002G\003\360\000\000G\003K\003\360\000\000K\003Z\003\360\000\000Z\003^\003\360\000\000^\003u\003\360\000\000u\003y\003\360\000\000y\003H\004\360\000\000H\004L\004\360\000\000L\004j\004\360\000\000j\004n\004\360\000\000n\004v\004\360\000\000v\004z\004\360\000\000z\004N\005\360\000\000N\005R\005\360\000\000R\005g\005\360\000\000g\005k\005\360\000\000k\005@\006\360\000\000@\006D\006\360\000\000D\006]\006\360\000\000]\006a\006\360\000\000a\006z\006\360\000\000z\006~\006\360\000\000~\006H\007\360\000\000H\007L\007\360\000\000L\007Y\007\360\000\000Y\007]\007\360\000\000]\007h\007\360\000\000h\007l\007\360\000\000l\007{\007\360\000\000{\007\177\007\360\000\000\177\007O\010\360\000\000O\010S\010\360\000\000S\010b\010\360\000\000b\010f\010\360\000\000f\010y\010\360\000\000y\010}\010\360\000\000}\010P\t\360\000\000P\tT\t\360\000\000T\ts\t\360\000\000s\tw\t\360\000\000w\tO\n\360\000\000O\nS\n\360\000\000S\nj\n\360\000\000j\nn\n\360\000\000n\nH\013\360\000\000H\013L\013\360\000\000L\013W\013\360\000\000W\013[\013\360\000\000[\013f\013\360\000\000f\013j\013\360\000\000j\013y\013\360\000\000y\013}\013\360\000\000}\013E\014\360\000\000E\014I\014\360\000\000I\014X\014\360\000\000X\014\\\014\360\000\000\\\014c\014\360\000\000c\014g\014\360\000\000g\014A\r\360\000\000A\rE\r\360\000\000E\rS\r\360\000\000S\rW\r\360\000\000W\rX\r\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220t\320\0338\270\007\270u\300C\300t\310=\320X_\320_d\320dg\320gk\320kz\360\000\000{\001B\002\360\000\000B\002G\002\360\000\000G\002J\002\360\000\000J\002N\002\360\000\000N""\002b\002\360\000\000b\002i\002\360\000\000i\002n\002\360\000\000n\002q\002\360\000\000q\002u\002\360\000\000u\002G\003\360\000\000G\003N\003\360\000\000N\003S\003\360\000\000S\003V\003\360\000\000V\003Z\003\360\000\000Z\003l\003\360\000\000l\003s\003\360\000\000s\003x\003\360\000\000x\003{\003\360\000\000{\003\177\003\360\000\000\177\003M\004\360\000\000M\004T\004\360\000\000T\004Y\004\360\000\000Y\004\\\004\360\000\000\\\004`\004\360\000\000`\004v\004\360\000\000v\004}\004\360\000\000}\004B\005\360\000\000B\005E\005\360\000\000E\005I\005\360\000\000I\005W\005\360\000\000W\005^\005\360\000\000^\005c\005\360\000\000c\005f\005\360\000\000f\005j\005\360\000\000j\005q\005\360\000\000q\005x\005\360\000\000x\005}\005\360\000\000}\005@\006\360\000\000@\006D\006\360\000\000D\006X\006\360\000\000X\006_\006\360\000\000_\006d\006\360\000\000d\006g\006\360\000\000g\006k\006\360\000\000k\006\177\006\360\000\000\177\006F\007\360\000\000F\007K\007\360\000\000K\007N\007\360\000\000N\007R\007\360\000\000R\007j\007\360\000\000j\007q\007\360\000\000q\007v\007\360\000\000v\007y\007\360\000\000y\007}\007\360\000\000}\007U\010\360\000\000U\010\\\010\360\000\000\\\010a\010\360\000\000a\010d\010\360\000\000d\010h\010\360\000\000h\010r\010\360\000\000r\010y\010\360\000\000y\010~\010\360\000\000~\010A\t\360\000\000A\tE\t\360\000\000E\tS\t\360\000\000S\tZ\t\360\000\000Z\t_\t\360\000\000_\tb\t\360\000\000b\tf\t\360\000\000f\tu\t\360\000\000u\t|\t\360\000\000|\tA\n\360\000\000A\nD\n\360\000\000D\nH\n\360\000\000H\nV\n\360\000\000V\n]\n\360\000\000]\nb\n\360\000\000b\ne\n\360\000\000e\ni\n\360\000\000i\nG\013\360\000\000G\013N\013\360\000\000N\013S\013\360\000\000S\013V\013\360\000\000V\013Z\013\360\000\000Z\013q\013\360\000\000q\013x\013\360\000\000x\013}\013\360\000\000}\013@\014\360\000\000@\014D\014\360\000\000D\014Z\014\360\000\000Z\014a\014\360\000\000a\014f\014\360\000\000f\014i\014\360\000\000i\014m\014\360\000\000m\014F\r\360\000\000F\rM\r\360\000\000M\rR\r\360\000\000R\rU\r\360\000""\000U\rY\r\360\000\000Y\rc\r\360\000\000c\rj\r\360\000\000j\ro\r\360\000\000o\rr\r\360\000\000r\rv\r\360\000\000v\r@\016\360\000\000@\016G\016\360\000\000G\016L\016\360\000\000L\016O\016\360\000\000O\016S\016\360\000\000S\016Z\016\360\000\000Z\016a\016\360\000\000a\016f\016\360\000\000f\016i\016\360\000\000i\016m\016\360\000\000m\016s\016\360\000\000s\016z\016\360\000\000z\016\177\016\360\000\000\177\016B\017\360\000\000B\017F\017\360\000\000F\017_\017\360\000\000_\017f\017\360\000\000f\017k\017\360\000\000k\017n\017\360\000\000n\017r\017\360\000\000r\017\177\017\360\000\000\177\017F\020\360\000\000F\020G\020\330\004\007\200q\330\010\017\320\017.\250d\260!\2607\270+\300W\310A\340\010\017\320\017.\250d\260!\2607\270+\300Q\200\001\360\030\000\005\n\210\034\220U\230!\2305\240\001\330\004\013\2105\220\014\230E\240\021\360\006\000\005\016\210V\2201\220F\230!\340\004\n\320\n\033\2301\330\004\t\320\t\034\230H\240A\320\000(\250\006\250i\260q\360&\000\005\010\200t\320\013\034\230A\230Q\330\010\016\210i\220q\230\001\330\033\037\230q\240\014\250A\330\004\007\200t\320\013\035\230Q\230a\330\010\016\210i\220q\230\001\330\033\037\230q\240\n\250!\330\004\007\200s\210#\210Q\330\010\016\210c\220\021\220!\340\010\016\210a\340\004\n\210!\330\004\023\320\023#\2401\240A\330\004\023\220<\230q\240\001\340\004\n\210$\210b\220\001\330\010\020\220\004\220B\220e\2303\230a\360\006\000\t\017\320\016'\240q\330\020*\250!\250;\260g\270Q\330\010\023\320\023)\250\021\250!\330\010\023\320\023(\250\001\250\021\330\010\013\210<\220q\230\n\240*\250N\270.\310\002\310!\330\014\022\220%\220r\230\021\340\014\022\220!\330\004\013\2101\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\2206\230\030\240\021\240!\330\004\007\200|\2207\230!\330\010(\250\001\250\031\260.\300\001\330\004\013\2101\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\220=\240\010\250\001\250\021\330\004\007\200|\2207\230!\330\010/\250q\3200@\300\016\310a\330\004\013\2101\200\001\360\030\000\005\014\210<""\220q\230\003\2303\230c\320!K\3102\310Q\330\004\t\320\t\031\230\021\230!\330\004\t\320\t\031\230\021\230!\330\004\r\210\\\230\021\230!\340\004\014\210K\220q\230\004\230B\230a\230t\2401\330\004\007\200v\210S\220\001\330\010\017\210q\330\004\013\2107\220&\230\002\230'\240\021\200\001\360\034\000\005\010\200t\320\013\035\230Q\230a\330\010\016\210i\220q\230\001\330\033\037\230q\240\010\250\001\330\004\007\200t\320\013\035\230Q\230a\330\010\016\210i\220q\230\001\330\033\037\230q\240\010\250\001\340\004\014\320\014(\250\001\320)9\270\021\270!\330)5\260Q\260a\330)9\270\021\270!\330)5\260Q\260h\270b\300\001\200\001\360\036\000\005\010\200t\320\013\035\230Q\230a\330\010\016\210i\220q\230\001\330\033\037\230q\240\010\250\001\330\004\007\200t\320\013\035\230Q\230a\330\010\016\210i\220q\230\001\330\033\037\230q\240\010\250\001\330\004\014\210B\210c\220\034\230Q\320\036.\250a\250q\330\036*\250!\2501\330\036.\250a\250q\330\036*\250!\2501adflr";
    PyObject *data = NULL;
    CYTHON_UNUSED_VAR(__Pyx_DecompressString);
    #endif
    PyObject **stringtab = __pyx_mstate->__pyx_string_tab;
    Py_ssize_t pos = 0;
    for (int i = 0; i < 246; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyUnicode_DecodeUTF8(bytes + pos, bytes_length, NULL);
      if (likely(string) && i >= 53) PyUnicode_InternInPlace(&string);
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
      stringtab[i] = string;
      pos += bytes_length;
    }
    for (int i = 246; i < 272; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyBytes_FromStringAndSize(bytes + pos, bytes_length);
      stringtab[i] = string;
      pos += bytes_length;
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    Py_XDECREF(data);
    for (Py_ssize_t i = 0; i < 272; i++) {
      if (unlikely(PyObject_Hash(stringtab[i]) == -1)) {
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    #if CYTHON_IMMORTAL_CONSTANTS
    {
      PyObject **table = stringtab + 246;
      for (Py_ssize_t i=0; i<26; ++i) {
        #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
        #if PY_VERSION_HEX < 0x030E0000
        if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
        #else
        if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
        #endif
        {
          Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
        }
        #else
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
        #endif
      }
    }
    #endif
  }
  {
    PyObject **numbertab = __pyx_mstate->__pyx_number_tab + 0;
    int8_t const cint_constants_1[] = {0,-1,1,2,3};
    int32_t const cint_constants_4[] = {37455648L,267571029L};
    for (int i = 0; i < 7; i++) {
      numbertab[i] = PyLong_FromLong((i < 5 ? cint_constants_1[i - 0] : cint_constants_4[i - 5]));
      if (unlikely(!numbertab[i])) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_number_tab;
    for (Py_ssize_t i=0; i<7; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */
typedef struct {
    unsigned int argcount : 3;
    unsigned int num_posonly_args : 1;
    unsigned int num_kwonly_args : 1;
    unsigned int nlocals : 4;
    unsigned int flags : 10;
    unsigned int first_line : 11;
} __Pyx_PyCode_New_function_description;
/* NewCodeObj.proto */
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
);


static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  PyObject* tuple_dedup_map = PyDict_New();
  if (unlikely(!tuple_dedup_map)) return -1;
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 131};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_s, __pyx_mstate->__pyx_n_u_c, __pyx_mstate->__pyx_n_u_s_2, __pyx_mstate->__pyx_n_u_found, __pyx_mstate->__pyx_n_u_length, __pyx_mstate->__pyx_n_u_c_2};
    __pyx_mstate_global->__pyx_codeobj_tab[0] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__dirstate_helpers_pyx_2, __pyx_mstate->__pyx_n_u_py_memrchr, __pyx_mstate->__pyx_kp_b_iso88591_q_3c_K2Q_Kq_Bat1_vS_q_7, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[0])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 226};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_path1, __pyx_mstate->__pyx_n_u_path2};
    __pyx_mstate_global->__pyx_codeobj_tab[1] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__dirstate_helpers_pyx_2, __pyx_mstate->__pyx_n_u_lt_by_dirs, __pyx_mstate->__pyx_kp_b_iso88591_t_Qa_iq_q_t_Qa_iq_q_Bc_Q_aq_1_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[1])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 253};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_path1, __pyx_mstate->__pyx_n_u_path2};
    __pyx_mstate_global->__pyx_codeobj_tab[2] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__dirstate_helpers_pyx_2, __pyx_mstate->__pyx_n_u_lt_path_by_dirblock, __pyx_mstate->__pyx_kp_b_iso88591_t_Qa_iq_q_t_Qa_iq_q_9_5Qa_9_5Qh, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[2])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 10, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 354};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_paths, __pyx_mstate->__pyx_n_u_path, __pyx_mstate->__pyx_n_u_lo_2, __pyx_mstate->__pyx_n_u_hi_2, __pyx_mstate->__pyx_n_u_mid, __pyx_mstate->__pyx_n_u_path_cstr, __pyx_mstate->__pyx_n_u_path_size, __pyx_mstate->__pyx_n_u_cur_cstr, __pyx_mstate->__pyx_n_u_cur_size, __pyx_mstate->__pyx_n_u_cur};
    __pyx_mstate_global->__pyx_codeobj_tab[3] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__dirstate_helpers_pyx_2, __pyx_mstate->__pyx_n_u_bisect_path_left, __pyx_mstate->__pyx_kp_b_iso88591_6_t_AQ_iq_q_t_Qa_iq_q_q_Qa_AQ_b, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[3])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 10, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 407};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_paths, __pyx_mstate->__pyx_n_u_path, __pyx_mstate->__pyx_n_u_lo_2, __pyx_mstate->__pyx_n_u_hi_2, __pyx_mstate->__pyx_n_u_mid, __pyx_mstate->__pyx_n_u_path_cstr, __pyx_mstate->__pyx_n_u_path_size, __pyx_mstate->__pyx_n_u_cur_cstr, __pyx_mstate->__pyx_n_u_cur_size, __pyx_mstate->__pyx_n_u_cur};
    __pyx_mstate_global->__pyx_codeobj_tab[4] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__dirstate_helpers_pyx_2, __pyx_mstate->__pyx_n_u_bisect_path_right, __pyx_mstate->__pyx_kp_b_iso88591_6_t_AQ_iq_q_t_Qa_iq_q_q_Qa_AQ_b_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[4])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 13, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 460};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_dirblocks, __pyx_mstate->__pyx_n_u_dirname, __pyx_mstate->__pyx_n_u_lo, __pyx_mstate->__pyx_n_u_hi, __pyx_mstate->__pyx_n_u_cache, __pyx_mstate->__pyx_n_u_lo_2, __pyx_mstate->__pyx_n_u_hi_2, __pyx_mstate->__pyx_n_u_mid, __pyx_mstate->__pyx_n_u_dirname_cstr, __pyx_mstate->__pyx_n_u_dirname_size, __pyx_mstate->__pyx_n_u_cur_cstr, __pyx_mstate->__pyx_n_u_cur_size, __pyx_mstate->__pyx_n_u_cur};
    __pyx_mstate_global->__pyx_codeobj_tab[5] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__dirstate_helpers_pyx_2, __pyx_mstate->__pyx_n_u_bisect_dirblock, __pyx_mstate->__pyx_kp_b_iso88591_iq_t_AQ_iq_q_A_t_Qa_iq_q_s_Q_c, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[5])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 10, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 689};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_num_trees, __pyx_mstate->__pyx_n_u_current_block, __pyx_mstate->__pyx_n_u_entry, __pyx_mstate->__pyx_n_u_current_dirname, __pyx_mstate->__pyx_n_u_new_block, __pyx_mstate->__pyx_n_u_expected_entry_count, __pyx_mstate->__pyx_n_u_entry_count, __pyx_mstate->__pyx_n_u_dirblocks, __pyx_mstate->__pyx_n_u_obj};
    __pyx_mstate_global->__pyx_codeobj_tab[6] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__dirstate_helpers_pyx_2, __pyx_mstate->__pyx_n_u_parse_dirblocks, __pyx_mstate->__pyx_kp_b_iso88591_A_D_33b_t6_F_Be_E_F_a_A_a_d_Bd_D, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[6])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[7] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_T_D_4wd_gTQ_aab_G1F_a_vWE_Q_q_t, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[7])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[8] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_AV1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[8])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 737};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_text, __pyx_mstate->__pyx_n_u_reader};
    __pyx_mstate_global->__pyx_codeobj_tab[9] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__dirstate_helpers_pyx_2, __pyx_mstate->__pyx_n_u_read_dirblocks, __pyx_mstate->__pyx_kp_b_iso88591_U_5_5_E_V1F_1_HA, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[9])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 799};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_stat_value};
    __pyx_mstate_global->__pyx_codeobj_tab[10] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__dirstate_helpers_pyx_2, __pyx_mstate->__pyx_n_u_pack_stat, __pyx_mstate->__pyx_kp_b_iso88591_Qa, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[10])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 804};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_entry, __pyx_mstate->__pyx_n_u_abspath, __pyx_mstate->__pyx_n_u_stat_value};
    __pyx_mstate_global->__pyx_codeobj_tab[11] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__dirstate_helpers_pyx_2, __pyx_mstate->__pyx_n_u_update_entry, __pyx_mstate->__pyx_kp_b_iso88591_A_r_A_q_AQ_a_PTTU_6OqPYY_y_5_O3, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[11])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1377};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[12] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_breezy_bzr__dirstate_helpers_pyx_2, __pyx_mstate->__pyx_n_u_iter_changes, __pyx_mstate->__pyx_kp_b_iso88591_A_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[12])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[13] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_T_t_SWWggk_l_A_A_E_E_Y_Y_p_p_t, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[13])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[14] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_1F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[14])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[15] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_pyx_unpickle_Reader, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_6_7_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[15])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[16] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_pyx_unpickle_ProcessEntryC, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_7_q0_a_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[16])) goto bad;
  }
  Py_DECREF(tuple_dedup_map);
  return 0;
  bad:
  Py_DECREF(tuple_dedup_map);
  return -1;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* AssertionsEnabled.init */
  if (likely(__Pyx_init_assertions_enabled() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CommonTypesMetaclass.init */
  if (likely(__pyx_CommonTypesMetaclass_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CachedMethodType.init */
  #if CYTHON_COMPILING_IN_LIMITED_API
  {
      PyObject *typesModule=NULL;
      typesModule = PyImport_ImportModule("types");
      if (typesModule) {
          __pyx_mstate_global->__Pyx_CachedMethodType = PyObject_GetAttrString(typesModule, "MethodType");
          Py_DECREF(typesModule);
      }
  } // error handling follows
  #endif
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CythonFunctionShared.init */
  if (likely(__pyx_CyFunction_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectCall (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
          #if CYTHON_COMPILING_IN_LIMITED_API
            return PyObject_Vectorcall(func, args, _nargs, NULL);
          #else
            vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
            if (f) {
                return f(func, args, _nargs, NULL);
            }
          #endif
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyErrFetchRestore (used by RaiseException) */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* RaiseException */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
#elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}

/* TupleAndListFromArray (used by fastcall) */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < (0))) {
            Py_DECREF(res);
            return NULL;
        }
        Py_INCREF(src[i]);
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL ||\
        !(CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS)
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length, length2;
        int kind;
        void *data1, *data2;
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        #endif
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length < 0)) return -1;
        #endif
        length2 = __Pyx_PyUnicode_GET_LENGTH(s2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length2 < 0)) return -1;
        #endif
        if (length != length2) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    return (equals == Py_EQ);
return_ne:
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = __Pyx_PyTuple_GET_SIZE(kwnames);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(n == -1)) return NULL;
    #endif
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        if (s == namei) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        int eq = __Pyx_PyUnicode_Equals(s, namei, Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs;
    PyObject *dict;
#if !CYTHON_ASSUME_SAFE_SIZE
    nkwargs = PyTuple_Size(kwnames);
    if (unlikely(nkwargs < 0)) return NULL;
#else
    nkwargs = PyTuple_GET_SIZE(kwnames);
#endif
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
#if !CYTHON_ASSUME_SAFE_MACROS
        PyObject *key = PyTuple_GetItem(kwnames, i);
        if (!key) goto bad;
#else
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
#endif
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* PyObjectCallOneArg (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetAttrStr (used by UnpackUnboundCMethod) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* UnpackUnboundCMethod (used by CallUnboundCMethod0) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
#elif CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject **args, Py_ssize_t nargs, PyObject *kwnames) {
        return _PyObject_Vectorcall
            (method, args ? args+1 : NULL, nargs ? nargs-1 : 0, kwnames);
}
#else
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
    return
#if PY_VERSION_HEX < 0x03090000
    _PyObject_Vectorcall
#else
    PyObject_Vectorcall
#endif
        (method, args ? args+1 : NULL, nargs ? (size_t) nargs-1 : 0, kwnames);
}
#endif
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
     METH_VARARGS | METH_KEYWORDS,
#else
     METH_FASTCALL | METH_KEYWORDS,
#endif
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method, *result=NULL;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    result = method;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            result = unbound_method;
        }
    }
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    if (unlikely(target->method)) {
        Py_DECREF(result);
    } else
#endif
    target->method = result;
    return 0;
}

/* CallUnboundCMethod0 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        if (likely(cfunc->flag == METH_NOARGS))
            return __Pyx_CallCFunction(cfunc, self, NULL);
        if (likely(cfunc->flag == METH_FASTCALL))
            return __Pyx_CallCFunctionFast(cfunc, self, NULL, 0);
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, NULL, 0, NULL);
        if (likely(cfunc->flag == (METH_VARARGS | METH_KEYWORDS)))
            return __Pyx_CallCFunctionWithKeywords(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple, NULL);
        if (cfunc->flag == METH_VARARGS)
            return __Pyx_CallCFunction(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple);
        return __Pyx__CallUnboundCMethod0(cfunc, self);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod0(&tmp_cfunc, self);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod0(cfunc, self);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *result;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
    result = __Pyx_PyObject_CallOneArg(cfunc->method, self);
    return result;
}

/* py_dict_items (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_items, d);
}

/* py_dict_values (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_values, d);
}

/* OwnedDictNext (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue) {
    PyObject *next = NULL;
    if (!*ppos) {
        if (pvalue) {
            PyObject *dictview = pkey ? __Pyx_PyDict_Items(p) : __Pyx_PyDict_Values(p);
            if (unlikely(!dictview)) goto bad;
            *ppos = PyObject_GetIter(dictview);
            Py_DECREF(dictview);
        } else {
            *ppos = PyObject_GetIter(p);
        }
        if (unlikely(!*ppos)) goto bad;
    }
    next = PyIter_Next(*ppos);
    if (!next) {
        if (PyErr_Occurred()) goto bad;
        return 0;
    }
    if (pkey && pvalue) {
        *pkey = __Pyx_PySequence_ITEM(next, 0);
        if (unlikely(*pkey)) goto bad;
        *pvalue = __Pyx_PySequence_ITEM(next, 1);
        if (unlikely(*pvalue)) goto bad;
        Py_DECREF(next);
    } else if (pkey) {
        *pkey = next;
    } else {
        assert(pvalue);
        *pvalue = next;
    }
    return 1;
  bad:
    Py_XDECREF(next);
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
    PyErr_FormatUnraisable("Exception ignored in __Pyx_PyDict_NextRef");
#else
    PyErr_WriteUnraisable(__pyx_mstate_global->__pyx_n_u_Pyx_PyDict_NextRef);
#endif
    if (pkey) *pkey = NULL;
    if (pvalue) *pvalue = NULL;
    return 0;
}
#else // !CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue) {
    int result = PyDict_Next(p, ppos, pkey, pvalue);
    if (likely(result == 1)) {
        if (pkey) Py_INCREF(*pkey);
        if (pvalue) Py_INCREF(*pvalue);
    }
    return result;
}
#endif

/* RaiseDoubleKeywords (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, args, 2);
        }
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, 2, NULL);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod2(&tmp_cfunc, self, arg1, arg2);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *result = NULL;
        PyObject *args = PyTuple_New(2);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
        return result;
    }
#endif
    {
        PyObject *args[4] = {NULL, self, arg1, arg2};
        return __Pyx_PyObject_FastCall(cfunc->method, args+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
    }
}

/* ParseKeywordsImpl (used by ParseKeywords) */
static int __Pyx_ValidateDuplicatePosArgs(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char* function_name)
{
    PyObject ** const *name = argnames;
    while (name != first_kw_arg) {
        PyObject *key = **name;
        int found = PyDict_Contains(kwds, key);
        if (unlikely(found)) {
            if (found == 1) __Pyx_RaiseDoubleKeywordsError(function_name, key);
            goto bad;
        }
        name++;
    }
    return 0;
bad:
    return -1;
}
#if CYTHON_USE_UNICODE_INTERNALS
static CYTHON_INLINE int __Pyx_UnicodeKeywordsEqual(PyObject *s1, PyObject *s2) {
    int kind;
    Py_ssize_t len = PyUnicode_GET_LENGTH(s1);
    if (len != PyUnicode_GET_LENGTH(s2)) return 0;
    kind = PyUnicode_KIND(s1);
    if (kind != PyUnicode_KIND(s2)) return 0;
    const void *data1 = PyUnicode_DATA(s1);
    const void *data2 = PyUnicode_DATA(s2);
    return (memcmp(data1, data2, (size_t) len * (size_t) kind) == 0);
}
#endif
static int __Pyx_MatchKeywordArg_str(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    #if CYTHON_USE_UNICODE_INTERNALS
    Py_hash_t key_hash = ((PyASCIIObject*)key)->hash;
    if (unlikely(key_hash == -1)) {
        key_hash = PyObject_Hash(key);
        if (unlikely(key_hash == -1))
            goto bad;
    }
    #endif
    name = first_kw_arg;
    while (*name) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (key_hash == ((PyASCIIObject*)name_str)->hash && __Pyx_UnicodeKeywordsEqual(name_str, key)) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) {
                *index_found = (size_t) (name - argnames);
                return 1;
            }
        }
        #endif
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (unlikely(key_hash == ((PyASCIIObject*)name_str)->hash)) {
            if (__Pyx_UnicodeKeywordsEqual(name_str, key))
                goto arg_passed_twice;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            if (unlikely(name_str == key)) goto arg_passed_twice;
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) goto arg_passed_twice;
        }
        #endif
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
bad:
    return -1;
}
static int __Pyx_MatchKeywordArg_nostr(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    if (unlikely(!PyUnicode_Check(key))) goto invalid_keyword_type;
    name = first_kw_arg;
    while (*name) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (cmp == 1) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        if (unlikely(cmp == -1)) goto bad;
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (unlikely(cmp != 0)) {
            if (cmp == 1) goto arg_passed_twice;
            else goto bad;
        }
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
bad:
    return -1;
}
static CYTHON_INLINE int __Pyx_MatchKeywordArg(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    return likely(PyUnicode_CheckExact(key)) ?
        __Pyx_MatchKeywordArg_str(key, argnames, first_kw_arg, index_found, function_name) :
        __Pyx_MatchKeywordArg_nostr(key, argnames, first_kw_arg, index_found, function_name);
}
static void __Pyx_RejectUnknownKeyword(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char *function_name)
{
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos = NULL;
    #else
    Py_ssize_t pos = 0;
    #endif
    PyObject *key = NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(kwds);
    while (
        #if CYTHON_AVOID_BORROWED_REFS
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL)
        #else
        PyDict_Next(kwds, &pos, &key, NULL)
        #endif
    ) {
        PyObject** const *name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (!*name) {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp != 1) {
                if (cmp == 0) {
                    PyErr_Format(PyExc_TypeError,
                        "%s() got an unexpected keyword argument '%U'",
                        function_name, key);
                }
                #if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(key);
                #endif
                break;
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        #endif
    }
    __Pyx_END_CRITICAL_SECTION();
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(pos);
    #endif
    assert(PyErr_Occurred());
}
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t extracted = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    name = first_kw_arg;
    while (*name && num_kwargs > extracted) {
        PyObject * key = **name;
        PyObject *value;
        int found = 0;
        #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        found = PyDict_GetItemRef(kwds, key, &value);
        #else
        value = PyDict_GetItemWithError(kwds, key);
        if (value) {
            Py_INCREF(value);
            found = 1;
        } else {
            if (unlikely(PyErr_Occurred())) goto bad;
        }
        #endif
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            extracted++;
        }
        name++;
    }
    if (num_kwargs > extracted) {
        if (ignore_unknown_kwargs) {
            if (unlikely(__Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name) == -1))
                goto bad;
        } else {
            __Pyx_RejectUnknownKeyword(kwds, argnames, first_kw_arg, function_name);
            goto bad;
        }
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t len;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    if (PyDict_Update(kwds2, kwds) < 0) goto bad;
    name = first_kw_arg;
    while (*name) {
        PyObject *key = **name;
        PyObject *value;
#if !CYTHON_COMPILING_IN_LIMITED_API && (PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop))
        int found = PyDict_Pop(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
        }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        int found = PyDict_GetItemRef(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            if (unlikely(PyDict_DelItem(kwds2, key) < 0)) goto bad;
        }
#else
    #if CYTHON_COMPILING_IN_CPYTHON
        value = _PyDict_Pop(kwds2, key, kwds2);
    #else
        value = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, kwds2, key, kwds2);
    #endif
        if (value == kwds2) {
            Py_DECREF(value);
        } else {
            if (unlikely(!value)) goto bad;
            values[name-argnames] = value;
        }
#endif
        name++;
    }
    len = PyDict_Size(kwds2);
    if (len > 0) {
        return __Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name);
    } else if (unlikely(len == -1)) {
        goto bad;
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject *key = NULL;
    PyObject** const * name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    for (Py_ssize_t pos = 0; pos < num_kwargs; pos++) {
#if CYTHON_AVOID_BORROWED_REFS
        key = __Pyx_PySequence_ITEM(kwds, pos);
#else
        key = __Pyx_PyTuple_GET_ITEM(kwds, pos);
#endif
#if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!key)) goto bad;
#endif
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            PyObject *value = kwvalues[pos];
            values[name-argnames] = __Pyx_NewRef(value);
        } else {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp == 1) {
                PyObject *value = kwvalues[pos];
                values[index_found] = __Pyx_NewRef(value);
            } else {
                if (unlikely(cmp == -1)) goto bad;
                if (kwds2) {
                    PyObject *value = kwvalues[pos];
                    if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
                } else if (!ignore_unknown_kwargs) {
                    goto invalid_keyword;
                }
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        key = NULL;
        #endif
    }
    return 0;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    goto bad;
bad:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(key);
    #endif
    return -1;
}

/* ParseKeywords */
static int __Pyx_ParseKeywords(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds)))
        return __Pyx_ParseKeywordsTuple(kwds, kwvalues, argnames, kwds2, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
    else if (kwds2)
        return __Pyx_ParseKeywordDictToDict(kwds, argnames, kwds2, values, num_pos_args, function_name);
    else
        return __Pyx_ParseKeywordDict(kwds, argnames, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyObjectFormatAndDecref */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result;
    if (unlikely(!s)) return NULL;
    result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    if (max_char > 1114111) max_char = 1114111;
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - result_ulength < 0))
        goto overflow;
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = values[i];
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (__Pyx_PyUnicode_READY(uval) == (-1))
            goto bad;
        #endif
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(ulength < 0)) goto bad;
        #endif
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    Py_ssize_t i;
    PyObject *result = NULL;
    PyObject *value_tuple = PyTuple_New(value_count);
    if (unlikely(!value_tuple)) return NULL;
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    for (i=0; i<value_count; i++) {
        if (__Pyx_PyTuple_SET_ITEM(value_tuple, i, values[i]) != (0)) goto bad;
        Py_INCREF(values[i]);
    }
    result = PyUnicode_Join(__pyx_mstate_global->__pyx_empty_unicode, value_tuple);
bad:
    Py_DECREF(value_tuple);
    return result;
#endif
}

/* DivInt[long] */
static CYTHON_INLINE long __Pyx_div_long(long a, long b, int b_is_constant) {
    long q = a / b;
    long r = a - q*b;
    long adapt_python = (b_is_constant ?
        ((r != 0) & ((r < 0) ^ (b < 0))) :
        ((r != 0) & ((r ^ b) < 0))
    );
    return q - adapt_python;
}

/* PyErrExceptionMatches (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyObjectGetAttrStrNoError (used by GetBuiltinName) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName (used by GetModuleGlobalName) */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* PyDictVersioning (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* bytes_tailmatch */
static int __Pyx_PyBytes_SingleTailmatch(PyObject* self, PyObject* arg,
                                         Py_ssize_t start, Py_ssize_t end, int direction) {
    char* self_ptr;
    Py_ssize_t self_len;
    char* sub_ptr;
    Py_ssize_t sub_len;
    int retval;
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030b0000
    PyObject *converted_arg = NULL;
    #else
    Py_buffer view;
    view.obj = NULL;
    #endif
    #if !(CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE)
    if (PyBytes_AsStringAndSize(self, &self_ptr, &self_len) == -1) return -1;
    #else
    self_ptr = PyBytes_AS_STRING(self);
    self_len = PyBytes_GET_SIZE(self);
    #endif
    if (PyBytes_Check(arg)) {
        #if !(CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE)
        if (PyBytes_AsStringAndSize(arg, &sub_ptr, &sub_len) == -1) return -1;
        #else
        sub_ptr = PyBytes_AS_STRING(arg);
        sub_len = PyBytes_GET_SIZE(arg);
        #endif
    }
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030b0000
    else if (PyByteArray_Check(arg)) {
        sub_ptr = PyByteArray_AsString(arg);
        if (unlikely(!sub_ptr)) return -1;
        sub_len = PyByteArray_Size(arg);
        if (unlikely(sub_len < 0)) return -1;
    } else {
        PyObject *as_memoryview = PyMemoryView_FromObject(arg);
        if (!as_memoryview) return -1;
        Py_DECREF(as_memoryview);
        converted_arg = PyBytes_FromObject(arg);
        if (!converted_arg) return -1;
        if (PyBytes_AsStringAndSize(converted_arg, &sub_ptr, &sub_len) == -1) {
            Py_DECREF(converted_arg);
            return -1;
        }
    }
    #else // LIMITED_API >= 030B0000 or !LIMITED_API
    else {
        if (unlikely(PyObject_GetBuffer(arg, &view, PyBUF_SIMPLE) == -1))
            return -1;
        sub_ptr = (char*) view.buf;
        sub_len = view.len;
    }
    #endif
    if (end > self_len)
        end = self_len;
    else if (end < 0)
        end += self_len;
    if (end < 0)
        end = 0;
    if (start < 0)
        start += self_len;
    if (start < 0)
        start = 0;
    if (direction > 0) {
        if (end-sub_len > start)
            start = end - sub_len;
    }
    if (start + sub_len <= end)
        retval = !memcmp(self_ptr+start, sub_ptr, (size_t)sub_len);
    else
        retval = 0;
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030b0000
    Py_XDECREF(converted_arg);
    #else
    if (view.obj)
        PyBuffer_Release(&view);
    #endif
    return retval;
}
static int __Pyx_PyBytes_TailmatchTuple(PyObject* self, PyObject* substrings,
                                        Py_ssize_t start, Py_ssize_t end, int direction) {
    Py_ssize_t i, count = __Pyx_PyTuple_GET_SIZE(substrings);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(count < 0)) return -1;
    #endif
    for (i = 0; i < count; i++) {
        int result;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        result = __Pyx_PyBytes_SingleTailmatch(self, PyTuple_GET_ITEM(substrings, i),
                                               start, end, direction);
#else
        PyObject* sub = __Pyx_PySequence_ITEM(substrings, i);
        if (unlikely(!sub)) return -1;
        result = __Pyx_PyBytes_SingleTailmatch(self, sub, start, end, direction);
        Py_DECREF(sub);
#endif
        if (result) {
            return result;
        }
    }
    return 0;
}
static int __Pyx_PyBytes_Tailmatch(PyObject* self, PyObject* substr,
                                   Py_ssize_t start, Py_ssize_t end, int direction) {
    if (unlikely(PyTuple_Check(substr))) {
        return __Pyx_PyBytes_TailmatchTuple(self, substr, start, end, direction);
    }
    return __Pyx_PyBytes_SingleTailmatch(self, substr, start, end, direction);
}

/* CIntToDigits (used by CIntToPyUnicode) */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode (used by COrdinalToPyUnicode) */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, const char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp = PySequence_Repeat(padding, uoffset - prepend_sign);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyUnicode_Concat(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyUnicode_Concat(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* COrdinalToPyUnicode (used by CIntToPyUnicode) */
static CYTHON_INLINE int __Pyx_CheckUnicodeValue(int value) {
    return value <= 1114111;
}
static PyObject* __Pyx_PyUnicode_FromOrdinal_Padded(int value, Py_ssize_t ulength, char padding_char) {
    Py_ssize_t padding_length = ulength - 1;
    if (likely((padding_length <= 250) && (value < 0xD800 || value > 0xDFFF))) {
        char chars[256];
        if (value <= 255) {
            memset(chars, padding_char, (size_t) padding_length);
            chars[ulength-1] = (char) value;
            return PyUnicode_DecodeLatin1(chars, ulength, NULL);
        }
        char *cpos = chars + sizeof(chars);
        if (value < 0x800) {
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0xc0 | (value & 0x1f));
        } else if (value < 0x10000) {
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0xe0 | (value & 0x0f));
        } else {
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0xf0 | (value & 0x07));
        }
        cpos -= padding_length;
        memset(cpos, padding_char, (size_t) padding_length);
        return PyUnicode_DecodeUTF8(cpos, chars + sizeof(chars) - cpos, NULL);
    }
    if (value <= 127 && CYTHON_USE_UNICODE_INTERNALS) {
        const char chars[1] = {(char) value};
        return __Pyx_PyUnicode_BuildFromAscii(ulength, chars, 1, 0, padding_char);
    }
    {
        PyObject *uchar, *padding_uchar, *padding, *result;
        padding_uchar = PyUnicode_FromOrdinal(padding_char);
        if (unlikely(!padding_uchar)) return NULL;
        padding = PySequence_Repeat(padding_uchar, padding_length);
        Py_DECREF(padding_uchar);
        if (unlikely(!padding)) return NULL;
        uchar = PyUnicode_FromOrdinal(value);
        if (unlikely(!uchar)) {
            Py_DECREF(padding);
            return NULL;
        }
        result = PyUnicode_Concat(padding, uchar);
        Py_DECREF(padding);
        Py_DECREF(uchar);
        return result;
    }
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!(is_unsigned || value == 0 || value > 0) ||
                    !(sizeof(value) <= 2 || value & ~ (int) 0x01fffff || __Pyx_CheckUnicodeValue((int) value)))) {
        PyErr_SetString(PyExc_OverflowError, "%c arg not in range(0x110000)");
        return NULL;
    }
    if (width <= 1) {
        return PyUnicode_FromOrdinal((int) value);
    }
    return __Pyx_PyUnicode_FromOrdinal_Padded((int) value, width, padding_char);
}
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(int)*3+2];
    char *dpos, *end = digits + sizeof(int)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    int remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (int) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (int) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (int) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* RejectKeywords */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds) {
    PyObject *key = NULL;
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds))) {
        key = __Pyx_PySequence_ITEM(kwds, 0);
    } else {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *pos = NULL;
#else
        Py_ssize_t pos = 0;
#endif
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
        if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return;
#endif
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL);
#if CYTHON_AVOID_BORROWED_REFS
        Py_XDECREF(pos);
#endif
    }
    if (likely(key)) {
        PyErr_Format(PyExc_TypeError,
            "%s() got an unexpected keyword argument '%U'",
            function_name, key);
        Py_DECREF(key);
    }
}

/* PyLongBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_Fallback___Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, int inplace) {
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject* __Pyx_Unpacked___Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    long a;
    const PY_LONG_LONG llb = intval;
    PY_LONG_LONG lla;
    if (unlikely(__Pyx_PyLong_IsZero(op1))) {
        return __Pyx_NewRef(op2);
    }
    const int is_positive = __Pyx_PyLong_IsPos(op1);
    const digit* digits = __Pyx_PyLong_Digits(op1);
    const Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
    if (likely(size == 1)) {
        a = (long) digits[0];
        if (!is_positive) a *= -1;
    } else {
        switch (size) {
            case 2:
                if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                    a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
            case 3:
                if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                    a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
            case 4:
                if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                    a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
        }
        return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    calculate_long:
        {
            long x;
            x = a + b;
            return PyLong_FromLong(x);
        }
    calculate_long_long:
        {
            PY_LONG_LONG llx;
            llx = lla + llb;
            return PyLong_FromLongLong(llx);
        }
    
}
#endif
static PyObject* __Pyx_Float___Pyx_PyLong_AddObjC(PyObject *float_val, long intval, int zerodivision_check) {
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    double a = __Pyx_PyFloat_AS_DOUBLE(float_val);
        double result;
        
        result = ((double)a) + (double)b;
        return PyFloat_FromDouble(result);
}
static CYTHON_INLINE PyObject* __Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        return __Pyx_Unpacked___Pyx_PyLong_AddObjC(op1, op2, intval, inplace, zerodivision_check);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        return __Pyx_Float___Pyx_PyLong_AddObjC(op1, intval, zerodivision_check);
    }
    return __Pyx_Fallback___Pyx_PyLong_AddObjC(op1, op2, inplace);
}
#endif

/* PyObjectFastCallMethod */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf) {
    PyObject *result;
    PyObject *attr = PyObject_GetAttr(args[0], name);
    if (unlikely(!attr))
        return NULL;
    result = __Pyx_PyObject_FastCall(attr, args+1, nargsf - 1);
    Py_DECREF(attr);
    return result;
}
#endif

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* GetAttr3 */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    int res = PyObject_GetOptionalAttr(o, n, &r);
    return (res != 0) ? r : __Pyx_NewRef(d);
#else
  #if CYTHON_USE_TYPE_SLOTS
    if (likely(PyUnicode_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
  #endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
#endif
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
    __Pyx_TypeName obj_type_name;
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyLong_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyLong_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is unsliceable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
bad:
    return NULL;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS)) {
        return __Pyx_PyList_GetItemRefFast(o, wrapped_i, unsafe_shared);
    } else
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyList_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyTuple_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)) {
            return __Pyx_PyList_GetItemRefFast(o, n, unsafe_shared);
        } else if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            return __Pyx_NewRef(PyList_GET_ITEM(o, n));
        }
    } else
    #if !CYTHON_AVOID_BORROWED_REFS
    if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            return __Pyx_NewRef(PyTuple_GET_ITEM(o, n));
        }
    } else
    #endif
#endif
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
    {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (!is_list && mm && mm->mp_subscript) {
            PyObject *r, *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (is_list || likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    (void)wraparound;
    (void)boundscheck;
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (unlikely(!j)) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE && !CYTHON_AVOID_BORROWED_REFS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared))) {
            Py_INCREF(v);
            return PyList_SetItem(o, n, v);
        } else if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old;
            Py_INCREF(v);
            old = PyList_GET_ITEM(o, n);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 0;
        }
    } else
#endif
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
    {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (!is_list && mm && mm->mp_ass_subscript) {
            int r;
            PyObject *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return -1;
            r = mm->mp_ass_subscript(o, key, v);
            Py_DECREF(key);
            return r;
        }
        if (is_list || likely(sm && sm->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return sm->sq_ass_item(o, i, v);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_SetItem(o, i, v);
    }
#endif
    (void)wraparound;
    (void)boundscheck;
    return __Pyx_SetItemInt_Generic(o, PyLong_FromSsize_t(i), v);
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* PyLongCompare */
static CYTHON_INLINE int __Pyx_PyLong_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 0;
    }
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) != 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 1;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 1;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal != 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = __Pyx_PyFloat_AS_DOUBLE(op1);
        return ((double)a != (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* PyObjectVectorCallKwBuilder (used by PyObjectVectorCallMethodKwBuilder) */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    Py_INCREF(key);
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* PyObjectVectorCallMethodKwBuilder */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames) {
    PyObject *result;
    PyObject *obj = PyObject_GetAttr(args[0], name);
    if (unlikely(!obj))
        return NULL;
    result = __Pyx_Object_Vectorcall_CallFromBuilder(obj, args+1, nargsf-1, kwnames);
    Py_DECREF(obj);
    return result;
}
#endif

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_mstate_global->__pyx_n_u_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* GetTopmostException (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C0000
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#elif __PYX_LIMITED_VERSION_HEX > 0x030C0000
    local_value = PyErr_GetRaisedException();
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
#if __PYX_LIMITED_VERSION_HEX > 0x030C0000
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
#else
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
#endif // __PYX_LIMITED_VERSION_HEX > 0x030C0000
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#elif __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    PyErr_SetHandledException(local_value);
    Py_XDECREF(local_value);
    Py_XDECREF(local_type);
    Py_XDECREF(local_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
#if __PYX_LIMITED_VERSION_HEX <= 0x030C0000
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
#endif
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_char(char value, Py_ssize_t width, char padding_char) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const char neg_one = (char) -1, const_zero = (char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!(is_unsigned || value == 0 || value > 0) ||
                    !(sizeof(value) <= 2 || value & ~ (char) 0x01fffff || __Pyx_CheckUnicodeValue((int) value)))) {
        PyErr_SetString(PyExc_OverflowError, "%c arg not in range(0x110000)");
        return NULL;
    }
    if (width <= 1) {
        return PyUnicode_FromOrdinal((int) value);
    }
    return __Pyx_PyUnicode_FromOrdinal_Padded((int) value, width, padding_char);
}
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_char(char value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(char)*3+2];
    char *dpos, *end = digits + sizeof(char)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    char remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const char neg_one = (char) -1, const_zero = (char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (char) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (char) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (char) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* PyObjectCallNoArg (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(tp);
    PyErr_Format(PyExc_AttributeError,
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}
#endif

/* PyObjectCallMethod0 (used by pop) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[1] = {obj};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_CallNoArg;
    return PyObject_VectorcallMethod(method_name, args, 1 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
#endif
}

/* pop */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L) {
    if (__Pyx_IS_TYPE(L, &PySet_Type)) {
        return PySet_Pop(L);
    }
    return __Pyx_PyObject_CallMethod0(L, __pyx_mstate_global->__pyx_n_u_pop);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L) {
    if (likely(PyList_GET_SIZE(L) > (((PyListObject*)L)->allocated >> 1))) {
        __Pyx_SET_SIZE(L, Py_SIZE(L) - 1);
        return PyList_GET_ITEM(L, PyList_GET_SIZE(L));
    }
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyList_Type_pop, L);
}
#endif

/* IterNextPlain (used by IterNext) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *__Pyx_GetBuiltinNext_LimitedAPI(void) {
    if (unlikely(!__pyx_mstate_global->__Pyx_GetBuiltinNext_LimitedAPI_cache))
        __pyx_mstate_global->__Pyx_GetBuiltinNext_LimitedAPI_cache = __Pyx_GetBuiltinName(__pyx_mstate_global->__pyx_n_u_next);
    return __pyx_mstate_global->__Pyx_GetBuiltinNext_LimitedAPI_cache;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next_Plain(PyObject *iterator) {
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    PyObject *result;
    PyObject *next = __Pyx_GetBuiltinNext_LimitedAPI();
    if (unlikely(!next)) return NULL;
    result = PyObject_CallFunctionObjArgs(next, iterator, NULL);
    return result;
#else
    (void)__Pyx_GetBuiltinName; // only for early limited API
    iternextfunc iternext = __Pyx_PyObject_GetIterNextFunc(iterator);
    assert(iternext);
    return iternext(iterator);
#endif
}

/* IterNext */
static PyObject *__Pyx_PyIter_Next2Default(PyObject* defval) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (!defval || unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return NULL;
        __Pyx_PyErr_Clear();
        Py_INCREF(defval);
        return defval;
    }
    if (defval) {
        Py_INCREF(defval);
        return defval;
    }
    __Pyx_PyErr_SetNone(PyExc_StopIteration);
    return NULL;
}
static void __Pyx_PyIter_Next_ErrorNoIterator(PyObject *iterator) {
    __Pyx_TypeName iterator_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(iterator));
    PyErr_Format(PyExc_TypeError,
        __Pyx_FMT_TYPENAME " object is not an iterator", iterator_type_name);
    __Pyx_DECREF_TypeName(iterator_type_name);
}
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject* iterator, PyObject* defval) {
    PyObject* next;
#if !CYTHON_COMPILING_IN_LIMITED_API
    iternextfunc iternext = __Pyx_PyObject_TryGetSlot(iterator, tp_iternext, iternextfunc);
    if (likely(iternext)) {
        next = iternext(iterator);
        if (likely(next))
            return next;
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
        if (unlikely(iternext == &_PyObject_NextNotImplemented))
            return NULL;
    #endif
    } else if (CYTHON_USE_TYPE_SLOTS) {
        __Pyx_PyIter_Next_ErrorNoIterator(iterator);
        return NULL;
    } else
#endif
    if (unlikely(!PyIter_Check(iterator))) {
        __Pyx_PyIter_Next_ErrorNoIterator(iterator);
        return NULL;
    } else {
        next = defval ? PyIter_Next(iterator) : __Pyx_PyIter_Next_Plain(iterator);
        if (likely(next))
            return next;
    }
    return __Pyx_PyIter_Next2Default(defval);
}

/* DelItemInt */
static int __Pyx_DelItem_Generic(PyObject *o, PyObject *j) {
    int r;
    if (unlikely(!j)) return -1;
    r = PyObject_DelItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_DelItemInt_Fast(PyObject *o, Py_ssize_t i,
                                               int is_list, CYTHON_NCP_UNUSED int wraparound) {
#if !CYTHON_USE_TYPE_SLOTS
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_DelItem(o, i);
    }
#else
    PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
    PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
    if ((!is_list) && mm && mm->mp_ass_subscript) {
        PyObject *key = PyLong_FromSsize_t(i);
        return likely(key) ? mm->mp_ass_subscript(o, key, (PyObject *)NULL) : -1;
    }
    if (likely(sm && sm->sq_ass_item)) {
        if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
            Py_ssize_t l = sm->sq_length(o);
            if (likely(l >= 0)) {
                i += l;
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    return -1;
                PyErr_Clear();
            }
        }
        return sm->sq_ass_item(o, i, (PyObject *)NULL);
    }
#endif
    return __Pyx_DelItem_Generic(o, PyLong_FromSsize_t(i));
}

/* PyLongCompare */
static CYTHON_INLINE int __Pyx_PyLong_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 1;
    }
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) == 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 0;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 0;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal == 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = __Pyx_PyFloat_AS_DOUBLE(op1);
        return ((double)a == (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* ArgTypeTestFunc (used by ArgTypeTest) */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    PyObject *extra_info = __pyx_mstate_global->__pyx_empty_unicode;
    int from_annotation_subclass = 0;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (!exact) {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    } else if (exact == 2) {
        if (__Pyx_TypeCheck(obj, type)) {
            from_annotation_subclass = 1;
            extra_info = __pyx_mstate_global->__pyx_kp_u_Note_that_Cython_is_deliberately;
        }
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")"
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        "%s%U"
#endif
        , name, type_name, obj_type_name
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        , (from_annotation_subclass ? ". " : ""), extra_info
#endif
        );
#if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    if (exact == 2 && from_annotation_subclass) {
        PyObject *res;
        PyObject *vargs[2];
        vargs[0] = PyErr_GetRaisedException();
        vargs[1] = extra_info;
        res = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_kp_u_add_note, vargs, 2, NULL);
        Py_XDECREF(res);
        PyErr_SetRaisedException(vargs[0]);
    }
#endif
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* AllocateExtensionType */
static PyObject *__Pyx_AllocateExtensionType(PyTypeObject *t, int is_final) {
    if (is_final || likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
        allocfunc alloc_func = __Pyx_PyType_GetSlot(t, tp_alloc, allocfunc);
        return alloc_func(t, 0);
    } else {
        newfunc tp_new = __Pyx_PyType_TryGetSlot(&PyBaseObject_Type, tp_new, newfunc);
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (!tp_new) {
            PyObject *new_str = PyUnicode_FromString("__new__");
            if (likely(new_str)) {
                PyObject *o = PyObject_CallMethodObjArgs((PyObject *)&PyBaseObject_Type, new_str, t, NULL);
                Py_DECREF(new_str);
                return o;
            } else
                return NULL;
        } else
    #endif
        return tp_new(t, __pyx_mstate_global->__pyx_empty_tuple, 0);
    }
}

/* CallTypeTraverse */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg) {
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
    if (__Pyx_get_runtime_version() < 0x03090000) return 0;
    #endif
    if (!always_call) {
        PyTypeObject *base = __Pyx_PyObject_GetSlot(o, tp_base, PyTypeObject*);
        unsigned long flags = PyType_GetFlags(base);
        if (flags & Py_TPFLAGS_HEAPTYPE) {
            return 0;
        }
    }
    Py_VISIT((PyObject*)Py_TYPE(o));
    return 0;
}
#endif

/* LimitedApiGetTypeDict (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static Py_ssize_t __Pyx_GetTypeDictOffset(void) {
    PyObject *tp_dictoffset_o;
    Py_ssize_t tp_dictoffset;
    tp_dictoffset_o = PyObject_GetAttrString((PyObject*)(&PyType_Type), "__dictoffset__");
    if (unlikely(!tp_dictoffset_o)) return -1;
    tp_dictoffset = PyLong_AsSsize_t(tp_dictoffset_o);
    Py_DECREF(tp_dictoffset_o);
    if (unlikely(tp_dictoffset == 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' doesn't have a dictoffset");
        return -1;
    } else if (unlikely(tp_dictoffset < 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' has an unexpected negative dictoffset. "
            "Please report this as Cython bug");
        return -1;
    }
    return tp_dictoffset;
}
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp) {
    static Py_ssize_t tp_dictoffset = 0;
    if (unlikely(tp_dictoffset == 0)) {
        tp_dictoffset = __Pyx_GetTypeDictOffset();
        if (unlikely(tp_dictoffset == -1 && PyErr_Occurred())) {
            tp_dictoffset = 0; // try again next time?
            return NULL;
        }
    }
    return *(PyObject**)((char*)tp + tp_dictoffset);
}
#endif

/* SetItemOnTypeDict (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_SetItem(tp_dict, k, v);
    if (likely(!result)) {
        PyType_Modified(tp);
        if (unlikely(PyObject_HasAttr(v, __pyx_mstate_global->__pyx_n_u_set_name))) {
            PyObject *setNameResult = PyObject_CallMethodObjArgs(v, __pyx_mstate_global->__pyx_n_u_set_name,  (PyObject *) tp, k, NULL);
            if (!setNameResult) return -1;
            Py_DECREF(setNameResult);
        }
    }
    return result;
}

/* FixUpExtensionType */
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if __PYX_LIMITED_VERSION_HEX > 0x030900B1
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#else
    const PyType_Slot *slot = spec->slots;
    int changed = 0;
#if !CYTHON_COMPILING_IN_LIMITED_API
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
#if !CYTHON_COMPILING_IN_CPYTHON
        const
#endif  // !CYTHON_COMPILING_IN_CPYTHON)
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_vectorcall_offset = memb->offset;
                    changed = 1;
                }
#endif  // CYTHON_METH_FASTCALL
#if !CYTHON_COMPILING_IN_PYPY
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                    Py_DECREF(descr);
                    if (unlikely(set_item_result < 0)) {
                        return -1;
                    }
                    changed = 1;
                }
#endif  // !CYTHON_COMPILING_IN_PYPY
            }
            memb++;
        }
    }
#endif  // !CYTHON_COMPILING_IN_LIMITED_API
#if !CYTHON_COMPILING_IN_PYPY
    slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_getset)
        slot++;
    if (slot && slot->slot == Py_tp_getset) {
        PyGetSetDef *getset = (PyGetSetDef*) slot->pfunc;
        while (getset && getset->name) {
            if (getset->name[0] == '_' && getset->name[1] == '_' && strcmp(getset->name, "__module__") == 0) {
                PyObject *descr = PyDescr_NewGetSet(type, getset);
                if (unlikely(!descr))
                    return -1;
                #if CYTHON_COMPILING_IN_LIMITED_API
                PyObject *pyname = PyUnicode_FromString(getset->name);
                if (unlikely(!pyname)) {
                    Py_DECREF(descr);
                    return -1;
                }
                int set_item_result = __Pyx_SetItemOnTypeDict(type, pyname, descr);
                Py_DECREF(pyname);
                #else
                CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
                int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                #endif
                Py_DECREF(descr);
                if (unlikely(set_item_result < 0)) {
                    return -1;
                }
                changed = 1;
            }
            ++getset;
        }
    }
#else
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#endif  // !CYTHON_COMPILING_IN_PYPY
    if (changed)
        PyType_Modified(type);
#endif  // PY_VERSION_HEX > 0x030900B1
    return 0;
}

/* ValidateBasesTuple (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_SIZE
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (unlikely(n < 0)) return -1;
#endif
    for (i = 1; i < n; i++)
    {
        PyTypeObject *b;
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !CYTHON_USE_TYPE_SLOTS
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
CYTHON_UNUSED static int __Pyx_PyType_HasMultipleInheritance(PyTypeObject *t) {
    while (t) {
        PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
        if (bases) {
            return 1;
        }
        t = __Pyx_PyType_GetSlot(t, tp_base, PyTypeObject*);
    }
    return 0;
}
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !CYTHON_COMPILING_IN_CPYTHON || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    if (!__Pyx_PyType_HasMultipleInheritance(t)) {
        return PyType_Ready(t);
    }
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400) &&\
                !CYTHON_COMPILING_IN_GRAAL
        gc = PyImport_GetModule(__pyx_mstate_global->__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_mstate_global->__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable (used by MergeVTables) */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i=0;
    Py_ssize_t size;
    void** base_vtables;
    __Pyx_TypeName tp_base_name = NULL;
    __Pyx_TypeName base_name = NULL;
    void* unknown = (void*)-1;
    PyObject* bases = __Pyx_PyType_GetSlot(type, tp_bases, PyObject*);
    int base_depth = 0;
    {
        PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        while (base) {
            base_depth += 1;
            base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
        }
    }
    base_vtables = (void**) PyMem_Malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
#if CYTHON_COMPILING_IN_LIMITED_API
    size = PyTuple_Size(bases);
    if (size < 0) goto other_failure;
#else
    size = PyTuple_GET_SIZE(bases);
#endif
    for (i = 1; i < size; i++) {
        PyObject *basei;
        void* base_vtable;
#if CYTHON_AVOID_BORROWED_REFS
        basei = PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#else
        basei = PyTuple_GET_ITEM(bases, i);
#endif
        base_vtable = __Pyx_GetVtable((PyTypeObject*)basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
            }
        }
    }
    PyErr_Clear();
    PyMem_Free(base_vtables);
    return 0;
bad:
    {
        PyTypeObject* basei = NULL;
        PyTypeObject* tp_base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        tp_base_name = __Pyx_PyType_GetFullyQualifiedName(tp_base);
#if CYTHON_AVOID_BORROWED_REFS
        basei = (PyTypeObject*)PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = (PyTypeObject*)PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#else
        basei = (PyTypeObject*)PyTuple_GET_ITEM(bases, i);
#endif
        base_name = __Pyx_PyType_GetFullyQualifiedName(basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
    }
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
really_bad: // bad has failed!
#endif
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
other_failure:
#endif
    PyMem_Free(base_vtables);
    return -1;
}

/* DelItemOnTypeDict (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_DelItem(tp_dict, k);
    if (likely(!result)) PyType_Modified(tp);
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_mstate_global->__pyx_n_u_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_mstate_global->__pyx_n_u_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython);
            if (likely(reduce_cython)) {
                ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_mstate_global->__pyx_n_u_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetFullyQualifiedName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_2_4
#define __PYX_HAVE_RT_ImportType_3_2_4
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size)
{
    PyObject *result = 0;
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API)
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !( defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API) )
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    if (size == 0) {
        return (PyTypeObject *)result;
    }
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_2_4 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_2_4 && (size_t)basicsize > size) {
        if (PyErr_WarnFormat(NULL, 0,
                "%.200s.%.200s size changed, may indicate binary incompatibility. "
                "Expected %zd from C header, got %zd from PyObject",
                module_name, class_name, size, basicsize) < 0) {
            goto bad;
        }
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* HasAttr (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!PyUnicode_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_PyObject_GetAttrStrNoError(o, n);
    if (!r) {
        return (unlikely(PyErr_Occurred())) ? -1 : 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}
#endif

/* ImportImpl (used by Import) */
static int __Pyx__Import_GetModule(PyObject *qualname, PyObject **module) {
    PyObject *imported_module = PyImport_GetModule(qualname);
    if (unlikely(!imported_module)) {
        *module = NULL;
        if (PyErr_Occurred()) {
            return -1;
        }
        return 0;
    }
    *module = imported_module;
    return 1;
}
static int __Pyx__Import_Lookup(PyObject *qualname, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject **module) {
    PyObject *imported_module;
    PyObject *top_level_package_name;
    Py_ssize_t i;
    int status, module_found;
    Py_ssize_t dot_index;
    module_found = __Pyx__Import_GetModule(qualname, &imported_module);
    if (unlikely(!module_found || module_found == -1)) {
        *module = NULL;
        return module_found;
    }
    if (imported_names) {
        for (i = 0; i < len_imported_names; i++) {
            PyObject *imported_name = imported_names[i];
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
            int has_imported_attribute = PyObject_HasAttr(imported_module, imported_name);
#else
            int has_imported_attribute = PyObject_HasAttrWithError(imported_module, imported_name);
            if (unlikely(has_imported_attribute == -1)) goto error;
#endif
            if (!has_imported_attribute) {
                goto not_found;
            }
        }
        *module = imported_module;
        return 1;
    }
    dot_index = PyUnicode_FindChar(qualname, '.', 0, PY_SSIZE_T_MAX, 1);
    if (dot_index == -1) {
        *module = imported_module;
        return 1;
    }
    if (unlikely(dot_index == -2)) goto error;
    top_level_package_name = PyUnicode_Substring(qualname, 0, dot_index);
    if (unlikely(!top_level_package_name)) goto error;
    Py_DECREF(imported_module);
    status = __Pyx__Import_GetModule(top_level_package_name, module);
    Py_DECREF(top_level_package_name);
    return status;
error:
    Py_DECREF(imported_module);
    *module = NULL;
    return -1;
not_found:
    Py_DECREF(imported_module);
    *module = NULL;
    return 0;
}
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *from_list = 0;
    int module_found;
    if (!qualname) {
        qualname = name;
    }
    module_found = __Pyx__Import_Lookup(qualname, imported_names, len_imported_names, &module);
    if (likely(module_found == 1)) {
        return module;
    } else if (unlikely(module_found == -1)) {
        return NULL;
    }
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    if (imported_names) {
#if CYTHON_COMPILING_IN_CPYTHON
        from_list = __Pyx_PyList_FromArray(imported_names, len_imported_names);
        if (unlikely(!from_list))
            goto bad;
#else
        from_list = PyList_New(len_imported_names);
        if (unlikely(!from_list)) goto bad;
        for (Py_ssize_t i=0; i<len_imported_names; ++i) {
            if (PyList_SetItem(from_list, i, __Pyx_NewRef(imported_names[i])) < 0) goto bad;
        }
#endif
    }
    if (level == -1) {
        const char* package_sep = strchr(__Pyx_MODULE_NAME, '.');
        if (package_sep != (0)) {
            module = PyImport_ImportModuleLevelObject(
                name, moddict, empty_dict, from_list, 1);
            if (unlikely(!module)) {
                if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                    goto bad;
                PyErr_Clear();
            }
        }
        level = 0;
    }
    if (!module) {
        module = PyImport_ImportModuleLevelObject(
            name, moddict, empty_dict, from_list, level);
    }
bad:
    Py_XDECREF(from_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level) {
    return __Pyx__Import(name, imported_names, len_imported_names, qualname, __pyx_mstate_global->__pyx_d, level);
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_mstate_global->__pyx_kp_u__7);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400) ||\
                CYTHON_COMPILING_IN_GRAAL
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError, "cannot import name %S", name);
    }
    return value;
}

/* dict_setdefault (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value) {
    PyObject* value;
#if __PYX_LIMITED_VERSION_HEX >= 0x030F0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4)
    PyDict_SetDefaultRef(d, key, default_value, &value);
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    PyObject *args[] = {d, key, default_value};
    value = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_n_u_setdefault, args, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#elif CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethodObjArgs(d, __pyx_mstate_global->__pyx_n_u_setdefault, key, default_value, NULL);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* AddModuleRef (used by FetchSharedCythonModule) */
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  static PyObject *__Pyx_PyImport_AddModuleObjectRef(PyObject *name) {
      PyObject *module_dict = PyImport_GetModuleDict();
      PyObject *m;
      if (PyMapping_GetOptionalItem(module_dict, name, &m) < 0) {
          return NULL;
      }
      if (m != NULL && PyModule_Check(m)) {
          return m;
      }
      Py_XDECREF(m);
      m = PyModule_NewObject(name);
      if (m == NULL)
          return NULL;
      if (PyDict_CheckExact(module_dict)) {
          PyObject *new_m;
          (void)PyDict_SetDefaultRef(module_dict, name, m, &new_m);
          Py_DECREF(m);
          return new_m;
      } else {
           if (PyObject_SetItem(module_dict, name, m) != 0) {
                Py_DECREF(m);
                return NULL;
            }
            return m;
      }
  }
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *py_name = PyUnicode_FromString(name);
      if (!py_name) return NULL;
      PyObject *module = __Pyx_PyImport_AddModuleObjectRef(py_name);
      Py_DECREF(py_name);
      return module;
  }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif

/* FetchSharedCythonModule (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef(__PYX_ABI_MODULE_NAME);
}

/* FetchCommonType (used by CommonTypesMetaclass) */
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject* __Pyx_PyType_FromMetaclass(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *result = __Pyx_PyType_FromModuleAndSpec(module, spec, bases);
    if (result && metaclass) {
        PyObject *old_tp = (PyObject*)Py_TYPE(result);
    Py_INCREF((PyObject*)metaclass);
#if __PYX_LIMITED_VERSION_HEX >= 0x03090000
        Py_SET_TYPE(result, metaclass);
#else
        result->ob_type = metaclass;
#endif
        Py_DECREF(old_tp);
    }
    return result;
}
#else
#define __Pyx_PyType_FromMetaclass(me, mo, s, b) PyType_FromMetaclass(me, mo, s, b)
#endif
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t expected_basicsize) {
    Py_ssize_t basicsize;
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (expected_basicsize == 0) {
        return 0; // size is inherited, nothing useful to check
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
    if (unlikely(!py_basicsize)) return -1;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = NULL;
    if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) return -1;
#else
    basicsize = ((PyTypeObject*) cached_type)->tp_basicsize;
#endif
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module = NULL, *cached_type = NULL, *abi_module_dict, *new_cached_type, *py_object_name;
    int get_item_ref_result;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    py_object_name = PyUnicode_FromString(object_name);
    if (!py_object_name) return NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) goto done;
    abi_module_dict = PyModule_GetDict(abi_module);
    if (!abi_module_dict) goto done;
    get_item_ref_result = __Pyx_PyDict_GetItemRef(abi_module_dict, py_object_name, &cached_type);
    if (get_item_ref_result == 1) {
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else if (unlikely(get_item_ref_result == -1)) {
        goto bad;
    }
    cached_type = __Pyx_PyType_FromMetaclass(
        metaclass,
        CYTHON_USE_MODULE_STATE ? module : abi_module,
        spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    new_cached_type = __Pyx_PyDict_SetDefault(abi_module_dict, py_object_name, cached_type);
    if (unlikely(new_cached_type != cached_type)) {
        if (unlikely(!new_cached_type)) goto bad;
        Py_DECREF(cached_type);
        cached_type = new_cached_type;
        if (__Pyx_VerifyCachedType(
                cached_type,
                object_name,
                spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else {
        Py_DECREF(new_cached_type);
    }
done:
    Py_XDECREF(abi_module);
    Py_DECREF(py_object_name);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CommonTypesMetaclass (used by CythonFunctionShared) */
static PyObject* __pyx_CommonTypesMetaclass_get_module(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED void* context) {
    return PyUnicode_FromString(__PYX_ABI_MODULE_NAME);
}
#if __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject* __pyx_CommonTypesMetaclass_call(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *args, CYTHON_UNUSED PyObject *kwds) {
    PyErr_SetString(PyExc_TypeError, "Cannot instantiate Cython internal types");
    return NULL;
}
static int __pyx_CommonTypesMetaclass_setattr(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *attr, CYTHON_UNUSED PyObject *value) {
    PyErr_SetString(PyExc_TypeError, "Cython internal types are immutable");
    return -1;
}
#endif
static PyGetSetDef __pyx_CommonTypesMetaclass_getset[] = {
    {"__module__", __pyx_CommonTypesMetaclass_get_module, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_CommonTypesMetaclass_slots[] = {
    {Py_tp_getset, (void *)__pyx_CommonTypesMetaclass_getset},
    #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {Py_tp_call, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_new, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_setattro, (void*)__pyx_CommonTypesMetaclass_setattr},
    #endif
    {0, 0}
};
static PyType_Spec __pyx_CommonTypesMetaclass_spec = {
    __PYX_TYPE_MODULE_PREFIX "_common_types_metatype",
    0,
    0,
    Py_TPFLAGS_IMMUTABLETYPE |
    Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT,
    __pyx_CommonTypesMetaclass_slots
};
static int __pyx_CommonTypesMetaclass_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    PyObject *bases = PyTuple_Pack(1, &PyType_Type);
    if (unlikely(!bases)) {
        return -1;
    }
    mstate->__pyx_CommonTypesMetaclassType = __Pyx_FetchCommonTypeFromSpec(NULL, module, &__pyx_CommonTypesMetaclass_spec, bases);
    Py_DECREF(bases);
    if (unlikely(mstate->__pyx_CommonTypesMetaclassType == NULL)) {
        return -1;
    }
    return 0;
}

/* PyMethodNew (used by CythonFunctionShared) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *result;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    #if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    {
        PyObject *args[] = {func, self};
        result = PyObject_Vectorcall(__pyx_mstate_global->__Pyx_CachedMethodType, args, 2, NULL);
    }
    #else
    result = PyObject_CallFunctionObjArgs(__pyx_mstate_global->__Pyx_CachedMethodType, func, self, NULL);
    #endif
    return result;
}
#else
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#endif

/* PyVectorcallFastCallDict (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i;
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos;
    #else
    Py_ssize_t pos;
    #endif
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    #if !CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t nkw = PyDict_Size(kw);
    if (unlikely(nkw == -1)) return NULL;
    #else
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    #endif
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = 0;
    i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (__Pyx_PyDict_NextRef(kw, &pos, &key, &value)) {
        keys_are_strings &=
        #if CYTHON_COMPILING_IN_LIMITED_API
            PyType_GetFlags(Py_TYPE(key));
        #else
            Py_TYPE(key)->tp_flags;
        #endif
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(PyTuple_SetItem(kwnames, i, key) < 0)) goto cleanup;
        #else
        PyTuple_SET_ITEM(kwnames, i, key);
        #endif
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(pos);
    #endif
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    Py_ssize_t kw_size =
        likely(kw == NULL) ?
        0 :
#if !CYTHON_ASSUME_SAFE_SIZE
        PyDict_Size(kw);
#else
        PyDict_GET_SIZE(kw);
#endif
    if (kw_size == 0) {
        return vc(func, args, nargs, NULL);
    }
#if !CYTHON_ASSUME_SAFE_SIZE
    else if (unlikely(kw_size == -1)) {
        return NULL;
    }
#endif
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared (used by CythonFunction) */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunctionNoMethod(PyObject *func, void (*cfunc)(void)) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if ((PyObject*)Py_TYPE(func) == __pyx_mstate_global->__Pyx_CachedMethodType) {
        int result;
        PyObject *newFunc = PyObject_GetAttr(func, __pyx_mstate_global->__pyx_n_u_func);
        if (unlikely(!newFunc)) {
            PyErr_Clear(); // It's only an optimization, so don't throw an error
            return 0;
        }
        result = __Pyx__IsSameCyOrCFunctionNoMethod(newFunc, cfunc);
        Py_DECREF(newFunc);
        return result;
    }
    return __Pyx__IsSameCyOrCFunctionNoMethod(func, cfunc);
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if (PyMethod_Check(func)) {
        func = PyMethod_GET_FUNCTION(func);
    }
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure) {
    PyObject *result;
    CYTHON_UNUSED_VAR(closure);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_doc_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#else
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_name_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    Py_INCREF(op->func_qualname);
    result = op->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
#endif
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_defaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_kwdefaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_annotations_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine_value(__pyx_CyFunctionObject *op) {
    int is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
    if (is_coroutine) {
        PyObject *is_coroutine_value, *module, *fromlist, *marker = __pyx_mstate_global->__pyx_n_u_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_mstate_global->__pyx_n_u_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        is_coroutine_value = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(is_coroutine_value)) {
            return is_coroutine_value;
        }
ignore:
        PyErr_Clear();
    }
    return __Pyx_PyBool_FromLong(is_coroutine);
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    result = __Pyx_CyFunction_get_is_coroutine_value(op);
    if (unlikely(!result))
        return NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    if (op->func_is_coroutine) {
        Py_DECREF(result);
        result = __Pyx_NewRef(op->func_is_coroutine);
    } else {
        op->func_is_coroutine = __Pyx_NewRef(result);
    }
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static void __Pyx_CyFunction_raise_argument_count_error(__pyx_CyFunctionObject *func, const char* message, Py_ssize_t size) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        py_name, message, size);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        name, message, size);
#endif
}
static void __Pyx_CyFunction_raise_type_error(__pyx_CyFunctionObject *func, const char* message) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s",
        py_name, message);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s",
        name, message);
#endif
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {"func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {"func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
#else
    {"func_dict", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
#endif
    {"func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {"__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {"_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    {"__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#endif
#if CYTHON_METH_FASTCALL
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {"__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(args);
    __Pyx_BEGIN_CRITICAL_SECTION(m);
    Py_INCREF(m->func_qualname);
    result = m->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_dict = NULL;
#endif
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func_dict);
#elif PY_VERSION_HEX < 0x030d0000
    _PyObject_ClearManagedDict((PyObject*)m);
#else
    PyObject_ClearManagedDict((PyObject*)m);
#endif
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    Py_CLEAR(m->defaults);
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    {
        int e = __Pyx_call_type_traverse((PyObject*)m, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func_dict);
#else
    {
        int e =
#if PY_VERSION_HEX < 0x030d0000
            _PyObject_VisitManagedDict
#else
            PyObject_VisitManagedDict
#endif
                ((PyObject*)m, visit, arg);
        if (e != 0) return e;
    }
#endif
    __Pyx_VISIT_CONST(m->func_name);
    __Pyx_VISIT_CONST(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    __Pyx_VISIT_CONST(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    Py_VISIT(m->defaults);
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
    PyObject *repr;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    repr = PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
    __Pyx_END_CRITICAL_SECTION();
    return repr;
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void(*)(void))meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes no arguments", size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes exactly one argument", size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    __Pyx_CyFunction_raise_type_error(
        (__pyx_CyFunctionObject*)func, "takes no keyword arguments");
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_SIZE
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(argc < 0)) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            __Pyx_CyFunction_raise_type_error(
                cyfunc, "needs an argument");
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(__Pyx_PyTuple_GET_SIZE(kwnames))) {
        __Pyx_CyFunction_raise_type_error(
            cyfunc, "takes no keyword arguments");
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes no arguments", nargs);
        return NULL;
    }
    return meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes exactly one argument", nargs);
        return NULL;
    }
    return meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    #if PY_VERSION_HEX < 0x030e00A6
    size_t nargs_value = (size_t) nargs;
    #else
    Py_ssize_t nargs_value = nargs;
    #endif
    return ((__Pyx_PyCMethod)(void(*)(void))meth)(self, cls, args, nargs_value, kwnames);
}
#endif
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if CYTHON_METH_FASTCALL
#if defined(Py_TPFLAGS_HAVE_VECTORCALL)
    Py_TPFLAGS_HAVE_VECTORCALL |
#elif defined(_Py_TPFLAGS_HAVE_VECTORCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
#endif // CYTHON_METH_FASTCALL
#if PY_VERSION_HEX >= 0x030C0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_TPFLAGS_MANAGED_DICT |
#endif
    Py_TPFLAGS_IMMUTABLETYPE | Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
static int __pyx_CyFunction_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_CyFunctionType_spec, NULL);
    if (unlikely(mstate->__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func, PyTypeObject *defaults_type) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_CallObject((PyObject*)defaults_type, NULL); // _PyObject_New(defaults_type);
    if (unlikely(!m->defaults))
        return NULL;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_mstate_global->__pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CLineInTraceback (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
#define __Pyx_PyProbablyModule_GetDict(o) __Pyx_XNewRef(PyModule_GetDict(o))
#elif !CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyProbablyModule_GetDict(o) PyObject_GenericGetDict(o, NULL);
#else
PyObject* __Pyx_PyProbablyModule_GetDict(PyObject *o) {
    PyObject **dict_ptr = _PyObject_GetDictPtr(o);
    return dict_ptr ? __Pyx_XNewRef(*dict_ptr) : NULL;
}
#endif
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline = NULL;
    PyObject *ptype, *pvalue, *ptraceback;
    PyObject *cython_runtime_dict;
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
    cython_runtime_dict = __Pyx_PyProbablyModule_GetDict(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, cython_runtime_dict,
            __Pyx_PyDict_SetDefault(cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False))
    }
    if (use_cline == NULL || use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    Py_XDECREF(cython_runtime_dict);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache (used by AddTraceback) */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        if (c_line) {
            py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        } else {
            py_funcname = PyUnicode_FromString(funcname);
        }
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CheckUnpickleChecksum */
static void __Pyx_RaiseUnpickleChecksumError(long checksum, long checksum1, long checksum2, long checksum3, const char *members) {
    PyObject *pickle_module = PyImport_ImportModule("pickle");
    if (unlikely(!pickle_module)) return;
    PyObject *pickle_error = PyObject_GetAttrString(pickle_module, "PickleError");
    Py_DECREF(pickle_module);
    if (unlikely(!pickle_error)) return;
    if (checksum2 == checksum1) {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x) = (%s))",
            checksum, checksum1, members);
    } else if (checksum3 == checksum2) {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x, 0x%x) = (%s))",
            checksum, checksum1, checksum2, members);
    } else {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x, 0x%x, 0x%x) = (%s))",
            checksum, checksum1, checksum2, checksum3, members);
    }
    Py_DECREF(pickle_error);
}
static int __Pyx_CheckUnpickleChecksum(long checksum, long checksum1, long checksum2, long checksum3, const char *members) {
    int found = 0;
    found |= checksum1 == checksum;
    found |= checksum2 == checksum;
    found |= checksum3 == checksum;
    if (likely(found))
        return 0;
    __Pyx_RaiseUnpickleChecksumError(checksum, checksum1, checksum2, checksum3, members);
    return -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_ptrdiff_t(ptrdiff_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const ptrdiff_t neg_one = (ptrdiff_t) -1, const_zero = (ptrdiff_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(ptrdiff_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(ptrdiff_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(ptrdiff_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(ptrdiff_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(ptrdiff_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(ptrdiff_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(ptrdiff_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_long(unsigned long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned long neg_one = (unsigned long) -1, const_zero = (unsigned long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallMethod1 */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* UpdateUnpickledDict */
static int __Pyx__UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index) {
    PyObject *state_dict = __Pyx_PySequence_ITEM(state, index);
    if (unlikely(!state_dict)) {
        return -1;
    }
    int non_empty = PyObject_IsTrue(state_dict);
    if (non_empty == 0) {
        Py_DECREF(state_dict);
        return 0;
    } else if (unlikely(non_empty == -1)) {
        return -1;
    }
    PyObject *dict;
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    dict = PyObject_GetAttrString(obj, "__dict__");
    #else
    dict = PyObject_GenericGetDict(obj, NULL);
    #endif
    if (unlikely(!dict)) {
        Py_DECREF(state_dict);
        return -1;
    }
    int result;
    if (likely(PyDict_CheckExact(dict))) {
        result = PyDict_Update(dict, state_dict);
    } else {
        PyObject *obj_result = __Pyx_PyObject_CallMethod1(dict, __pyx_mstate_global->__pyx_n_u_update, state_dict);
        if (likely(obj_result)) {
            Py_DECREF(obj_result);
            result = 0;
        } else {
            result = -1;
        }
    }
    Py_DECREF(state_dict);
    Py_DECREF(dict);
    return result;
}
static int __Pyx_UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index) {
    Py_ssize_t state_size = __Pyx_PyTuple_GET_SIZE(state);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(state_size == -1)) return -1;
    #endif
    if (state_size <= index) {
        return 0;
    }
    return __Pyx__UpdateUnpickledDict(obj, state, index);
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u__8);
    }
    goto done;
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetRuntimeVersion */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
void __Pyx_init_runtime_version(void) {
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
}
#endif
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    return __Pyx_cached_runtime_version;
#endif
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* NewCodeObj */
#if CYTHON_COMPILING_IN_LIMITED_API
    static PyObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                       PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                       PyObject *fv, PyObject *cell, PyObject* fn,
                                       PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
#elif PY_VERSION_HEX >= 0x030B0000
  static PyCodeObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                         PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                         PyObject *fv, PyObject *cell, PyObject* fn,
                                         PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, __pyx_mstate_global->__pyx_empty_bytes);
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030c00A1
    if (likely(result))
        result->_co_firsttraceable = 0;
    #endif
    return result;
  }
#elif !CYTHON_COMPILING_IN_PYPY
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
) {
    PyObject *code_obj = NULL, *varnames_tuple_dedup = NULL, *code_bytes = NULL;
    Py_ssize_t var_count = (Py_ssize_t) descr.nlocals;
    PyObject *varnames_tuple = PyTuple_New(var_count);
    if (unlikely(!varnames_tuple)) return NULL;
    for (Py_ssize_t i=0; i < var_count; i++) {
        Py_INCREF(varnames[i]);
        if (__Pyx_PyTuple_SET_ITEM(varnames_tuple, i, varnames[i]) != (0)) goto done;
    }
    #if CYTHON_COMPILING_IN_LIMITED_API
    varnames_tuple_dedup = PyDict_GetItem(tuple_dedup_map, varnames_tuple);
    if (!varnames_tuple_dedup) {
        if (unlikely(PyDict_SetItem(tuple_dedup_map, varnames_tuple, varnames_tuple) < 0)) goto done;
        varnames_tuple_dedup = varnames_tuple;
    }
    #else
    varnames_tuple_dedup = PyDict_SetDefault(tuple_dedup_map, varnames_tuple, varnames_tuple);
    if (unlikely(!varnames_tuple_dedup)) goto done;
    #endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(varnames_tuple_dedup);
    #endif
    if (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table != NULL && !CYTHON_COMPILING_IN_GRAAL) {
        Py_ssize_t line_table_length = __Pyx_PyBytes_GET_SIZE(line_table);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(line_table_length == -1)) goto done;
        #endif
        Py_ssize_t code_len = (line_table_length * 2 + 4) & ~3LL;
        code_bytes = PyBytes_FromStringAndSize(NULL, code_len);
        if (unlikely(!code_bytes)) goto done;
        char* c_code_bytes = PyBytes_AsString(code_bytes);
        if (unlikely(!c_code_bytes)) goto done;
        memset(c_code_bytes, 0, (size_t) code_len);
    }
    code_obj = (PyObject*) __Pyx__PyCode_New(
        (int) descr.argcount,
        (int) descr.num_posonly_args,
        (int) descr.num_kwonly_args,
        (int) descr.nlocals,
        0,
        (int) descr.flags,
        code_bytes ? code_bytes : __pyx_mstate_global->__pyx_empty_bytes,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        varnames_tuple_dedup,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        filename,
        funcname,
        (int) descr.first_line,
        (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table) ? line_table : __pyx_mstate_global->__pyx_empty_bytes
    );
done:
    Py_XDECREF(code_bytes);
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(varnames_tuple_dedup);
    #endif
    Py_DECREF(varnames_tuple);
    return code_obj;
}

/* DecompressString */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo) {
    PyObject *module = NULL, *decompress, *compressed_bytes, *decompressed;
    const char* module_name = algo == 3 ? "compression.zstd" : algo == 2 ? "bz2" : "zlib";
    PyObject *methodname = PyUnicode_FromString("decompress");
    if (unlikely(!methodname)) return NULL;
    #if __PYX_LIMITED_VERSION_HEX >= 0x030e0000
    if (algo == 3) {
        PyObject *fromlist = Py_BuildValue("[O]", methodname);
        if (unlikely(!fromlist)) goto bad;
        module = PyImport_ImportModuleLevel("compression.zstd", NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
    } else
    #endif
        module = PyImport_ImportModule(module_name);
    if (unlikely(!module)) goto import_failed;
    decompress = PyObject_GetAttr(module, methodname);
    if (unlikely(!decompress)) goto import_failed;
    {
        #ifdef __cplusplus
            char *memview_bytes = const_cast<char*>(s);
        #else
            #if defined(__clang__)
              #pragma clang diagnostic push
              #pragma clang diagnostic ignored "-Wcast-qual"
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic push
              #pragma GCC diagnostic ignored "-Wcast-qual"
            #endif
            char *memview_bytes = (char*) s;
            #if defined(__clang__)
              #pragma clang diagnostic pop
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic pop
            #endif
        #endif
        #if CYTHON_COMPILING_IN_LIMITED_API && !defined(PyBUF_READ)
        int memview_flags = 0x100;
        #else
        int memview_flags = PyBUF_READ;
        #endif
        compressed_bytes = PyMemoryView_FromMemory(memview_bytes, length, memview_flags);
    }
    if (unlikely(!compressed_bytes)) {
        Py_DECREF(decompress);
        goto bad;
    }
    decompressed = PyObject_CallFunctionObjArgs(decompress, compressed_bytes, NULL);
    Py_DECREF(compressed_bytes);
    Py_DECREF(decompress);
    Py_DECREF(module);
    Py_DECREF(methodname);
    return decompressed;
import_failed:
    PyErr_Format(PyExc_ImportError,
        "Failed to import '%.20s.decompress' - cannot initialise module strings. "
        "String compression was configured with the C macro 'CYTHON_COMPRESS_STRINGS=%d'.",
        module_name, algo);
bad:
    Py_XDECREF(module);
    Py_DECREF(methodname);
    return NULL;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return __Pyx_NewRef(b ? Py_True: Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
